#include <assert.h>

#include "libcad3d.h"


CAMERA::CAMERA (RPY Frame /* = RPY_0 */) :
  CAD_FILE (),
  Location (Frame)
{
  PRO_METHOD ("CAMERA", "")

  Create ();

  PRO_END
}


CAMERA::~CAMERA ()
{
  PRO_METHOD ("CAMERA", "")

  PRO_END
}


bool CAMERA::Read (DATAFILE &File, double Version)
{
  PRO_METHOD ("CAMERA", "")

  if (File.Parse("<\\[CAMERA\\]>\n") < 0)
    PRO_RETURN ("Key [CAMERA] not found", FALSE)

  if (! CAD_FILE::Read(File, Version))
    PRO_RETURN ("Reading FILE failed", FALSE)

  if (File.Parse("  <Position:>  (%f) / (%f) / (%f)\n"
		 "  <Rotation:>  (%f) / (%f) / (%f)\n",
		 &Location.px, &Location.py, &Location.pz,
		 &Location.y,  &Location.p,  &Location.r) < 6)
    PRO_RETURN ("Reading DATA failed", FALSE)

  Create ();

  PRO_END
  return TRUE;
}
 

bool CAMERA::Write (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  PRO_METHOD ("CAMERA", "")

  if (fp == NULL)
    fp = stdout;

  Space   (fp, Indent);
  fprintf (fp, "[CAMERA]\n");

  if (! CAD_FILE::Write(fp))
    PRO_RETURN ("Writing FILE failed", FALSE)

  Space   (fp, Indent);
  fprintf (fp, "Position:  %.2f / %.2f / %.2f\n", 
	       Location.px, Location.py, Location.pz);

  Space   (fp, Indent);
  fprintf (fp, "Rotation:  %8.2f / %8.2f / %8.2f\n", 
	       Location.y, Location.p, Location.r);

  PRO_END
  return TRUE;
}


bool CAMERA::Create ()
{
  PRO_METHOD ("CAMERA", "")

  SetID ("CAMERA");
  glNewList (Id, GL_COMPILE);
  glIndexi  (Id);

  // Transformation: Camera -> World
  glTranslated (Location.px, Location.py, Location.pz);
  glRotated (Location.y, 1.0, 0.0, 0.0);
  glRotated (Location.p, 0.0, 1.0, 0.0);
  glRotated (Location.r, 0.0, 0.0, 1.0);

  glEndList ();

  PRO_END
  return TRUE;
}


void CAMERA::Place (RPY Frame /* = RPY_0 */)
{
  PRO_METHOD ("CAMERA", "")

  Location = Frame;
  Create ();

  PRO_END
}
 

void CAMERA::Rotate (double rx, double ry, double rz)
{
  PRO_METHOD ("CAMERA", "")
  
  Location.y = MOD (Location.y+rx, -180.0, 180.0, 360.0);
  Location.p = MOD (Location.p+ry, -180.0, 180.0, 360.0);
  Location.r = MOD (Location.r+rz, -180.0, 180.0, 360.0);

  Create ();

  PRO_END
}


RPY CAMERA::GetLocation () const
{
  return Location;
}
