#include <assert.h>

#include "libcad3d.h"


#define TRANSPARENCY 1.0              // default: no transparency


char *ColorNames[] =         // => Dialog oder fr Speicherung beibehalten ?
{
  "OBJECT",
  "VALUES",
  "BLACK",
  "GREY_50",
  "CORNSILK",
  "GREY",
  "SILVER",
  "WHITE",
  "DARK_RED",
  "RED",
  "LITE_RED",
  "DARK_ORANGE",
  "ORANGE",
  "GOLD",
  "YELLOW",
  "LEMONCHIFFON",
  "FOREST_GREEN",
  "LIME",  
  "GREEN",
  "SEA_GREEN",
  "TURQUOISE",
  "STEEL_BLUE",
  "DARK_BLUE",
  "BLUE",
  "DODGER_BLUE",
  "LITE_BLUE",
  "CYAN",
  "PURPLE",
  "ORCHIDEE",
  "DEEPPINK",
  "MAGENTA",
  "CHERRY",
  "BROWN",
  "SIENNA",
  "PERU",
  "TAN",
  "PEACHPUFF",
  NULL,
  "TEXTURE"
};


int COLOR::COLORS = (COLOR::TEXTURE+2);


GLfloat ColorTable[][4] = 
{
  {1.00, 1.00, 1.00, TRANSPARENCY},      // Object                (-1)
  {0.75, 0.75, 0.75, TRANSPARENCY},      // Values                 (0)
  {0.00, 0.00, 0.00, TRANSPARENCY},      // Black
  {0.50, 0.50, 0.50, TRANSPARENCY},      // Grey50
  {0.54, 0.53, 0.47, TRANSPARENCY},      // CornSilk4
  {0.70, 0.70, 0.70, TRANSPARENCY},      // Grey
  {0.90, 0.90, 0.90, TRANSPARENCY},      // Silver                 (5)
  {1.00, 1.00, 1.00, TRANSPARENCY},      // White
  {0.80, 0.15, 0.15, TRANSPARENCY},      // FireBrick3
  {1.00, 0.00, 0.00, TRANSPARENCY},      // Red
  {1.00, 0.68, 0.72, TRANSPARENCY},      // LightPink1
  {0.93, 0.46, 0.00, TRANSPARENCY},      // DarkOrange2           (10)
  {1.00, 0.60, 0.30, TRANSPARENCY},      // Orange
  {1.00, 0.84, 0.00, TRANSPARENCY},      // Gold      
  {1.00, 1.00, 0.00, TRANSPARENCY},      // Yellow       
  {1.00, 0.98, 0.80, TRANSPARENCY},      // LemonChiffon
  {0.13, 0.54, 0.13, TRANSPARENCY},      // ForrestGreen          (15)
  {0.20, 0.80, 0.20, TRANSPARENCY},      // Lime (Green)
  {0.00, 0.80, 0.00, TRANSPARENCY},      // Green
  {0.26, 0.80, 0.50, TRANSPARENCY},      // SeaGreen3
  {0.00, 0.52, 0.54, TRANSPARENCY},      // Turquoise4
  {0.27, 0.51, 0.70, TRANSPARENCY},      // Steel Blue            (20)
  {0.00, 0.00, 0.54, TRANSPARENCY},      // Blue4         
  {0.00, 0.00, 1.00, TRANSPARENCY},      // Blue
  {0.11, 0.52, 0.93, TRANSPARENCY},      // DodgerBlue2
  {0.00, 0.75, 1.00, TRANSPARENCY},      // DeepSkyBlue (Lt Blue)
  {0.00, 1.00, 1.00, TRANSPARENCY},      // Cyan                  (25)
  {1.00, 0.00, 1.00, TRANSPARENCY},      // Purple
  {0.73, 0.33, 0.82, TRANSPARENCY},      // MediumOrchid
  {1.00, 0.08, 0.57, TRANSPARENCY},      // DeepPink     
  {1.00, 0.00, 1.00, TRANSPARENCY},      // Magenta
  {1.00, 0.70, 1.00, TRANSPARENCY},      // Cherry / Plum1        (30)
  {0.55, 0.27, 0.16, TRANSPARENCY},      // (Saddle)Brown
  {0.63, 0.32, 0.18, TRANSPARENCY},      // Sienna (Brown)
  {0.80, 0.52, 0.25, TRANSPARENCY},      // Peru (25)
  {0.80, 0.70, 0.60, TRANSPARENCY},      // Tan
  {1.00, 0.85, 0.72, TRANSPARENCY},      // PeachPuff
};

#define MAX_COLORS (sizeof(ColorTable) / (sizeof(GLfloat)*4))


COLOR::COLOR (ID ColorID /* = OBJECT */) :
  CAD_FILE (),
  UNIT     (),
  ColorID  (ColorID)
{
  PRO_METHOD ("COLOR", "")

  // Initialize color components
  Color[RED]   = 1.0;
  Color[GREEN] = 1.0;
  Color[BLUE]  = 1.0;
  Color[ALPHA] = 1.0;

  SetID (ColorID);

  PRO_END
}


COLOR::COLOR (double Red, double Green, double Blue,
	      double Alpha /* = 1.0 */) :
  CAD_FILE (),
  UNIT     (),
  ColorID  (RGB)
{
  PRO_METHOD ("COLOR", "\n")

  Color[RED]   = Red;
  Color[GREEN] = Green;
  Color[BLUE]  = Blue;
  Color[ALPHA] = Alpha;

  PRO_END
}


COLOR &COLOR::operator = (const COLOR &Source)
{
  int count;

  if (this != &Source)                    // check for self-assignment
  {
    ColorID = Source.ColorID;
    for (count = RED; count <= ALPHA; ++count)
      Color[count] = Source.Color[count];
  }

  return *this;
} 


double &COLOR::operator [] (const int &i)
{
  assert (i >= RED && i <= ALPHA);

  if (i < ALPHA)
    ColorID = RGB;

  return Color[i];
}


double COLOR::operator [] (const int &i) const
{
  assert (i >= RED && i <= ALPHA);
 
  return Color[i];
}


bool COLOR::Read (DATAFILE &File, double Version)
{
  int Id;

  PRO_METHOD ("COLOR", "\n")

  if (File.Parse("<\\[COLOR\\]>\n") < 0)
    PRO_RETURN ("Key [COLOR] not found", FALSE)

  if (! CAD_FILE::Read(File, Version))
    PRO_RETURN ("Reading FILE failed", FALSE)

  if (File.Parse("  <ID:>      (OBJECT|VALUES|BLACK|GREY_50|CORNSILK|GREY|"
                               "SILVER|WHITE|DARK_RED|RED|LITE_RED|"
                               "DARK_ORANGE|ORANGE|GOLD|YELLOW|LEMONCHIFFON|"
                               "FOREST_FREEN|LIME|GREEN|SEA_GREEN|TURQUISE|"
                               "STEEL_BLUE|DARK_BLUE|BLUE|DODGER_BLUE|"
		               "LITE_BLUE|CYAN|PURPLE|ORCHIDEE|DEEPPINK|"
		               "MAGENTA|CHERRY|BROWN|SIENNA|PERU|TAN|"
                               "PEACHPUFF|NONE|TEXTURE)\n"
		 "  <Values:>  (%f) (%f) (%f) (%f)\n",
		 &Id,
		 &Color[RED], &Color[GREEN], &Color[BLUE], &Color[ALPHA]) < 5)
    PRO_RETURN ("Reading DATA failed", FALSE)

  ColorID = (ID)Id;

  Create ();

  PRO_END
  return TRUE;
}


bool COLOR::Write (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  PRO_METHOD ("COLOR", "\n")

  if (fp == NULL)
    fp = stdout;

  Space   (fp, Indent);
  fprintf (fp, "[COLOR]\n");
  Indent += INDENT;

  if (! CAD_FILE::Write(fp, Indent))
    PRO_RETURN ("Writing FILE failed", FALSE)

  Space   (fp, Indent);
  fprintf (fp, "ID:      %s\n", ColorNames[ColorID]);

  Space   (fp, Indent);
  fprintf (fp, "Values:  %.3f %.3f %.3f %.3f\n",
	       Color[RED], Color[GREEN], Color[BLUE], Color[ALPHA]);

  PRO_END
  return TRUE;
}


bool COLOR::Create ()
{
  if (ColorID >= RGB)
  {
    Id = OpenGL->GetID (Id, GetName());
    glNewList (Id, GL_COMPILE);
    glIndexi  (Id);

      glColor4f (Color[RED], Color[GREEN], Color[BLUE], Color[ALPHA]);

    glEndList ();
  }

  return TRUE;
}


void COLOR::Set () const
{
  if (ColorID >= RGB)
    glColor4f (Color[RED], Color[GREEN], Color[BLUE], Color[ALPHA]);
}


COLOR COLOR::Complement () const
{
  return COLOR (1.0-Color[RED],  1.0-Color[GREEN], 
		1.0-Color[BLUE],     Color[ALPHA]);
}


void COLOR::SetID (ID ColorID /* = COLOR_DEFAULT */)
{
  if (ColorID >= BLACK && ColorID <= PEACHPUFF)
  {
    Color[RED]   = ColorTable[abs(ColorID+1)][RED];
    Color[GREEN] = ColorTable[abs(ColorID+1)][GREEN];
    Color[BLUE]  = ColorTable[abs(ColorID+1)][BLUE];
    /* 
       The alpha value does not belong to the color-id. If so, we had to
       restore the alpha value e.g. after setting the object color in part
       creation.
       Color[ALPHA] = ColorTable[abs(ColorID+1)][ALPHA];
    */
  }

  // Do not change this order, the id has been changed in upper commands !!! 
  this->ColorID = ColorID;               
}


void COLOR::SetValues (double Red, double Green, double Blue, 
		       double Alpha /* = 1.0 */)
{
  PRO_METHOD ("COLOR", "")

  Color[RED]   = LIMIT (0.0, Red,   1.0);
  Color[GREEN] = LIMIT (0.0, Green, 1.0);
  Color[BLUE]  = LIMIT (0.0, Blue,  1.0);
  Color[ALPHA] = LIMIT (0.0, Alpha, 1.0);

  // ColorID = RBGB ???

  PRO_END
}


void COLOR::SetRed (double Red)
{
  PRO_METHOD ("COLOR", "")

  Color[RED] = LIMIT (0.0, Red, 1.0);

  PRO_END
}


void COLOR::SetGreen  (double Green)
{
  PRO_METHOD ("COLOR", "")

  Color[GREEN] = LIMIT (0.0, Green, 1.0);

  PRO_END
}


void COLOR::SetBlue (double Blue)
{
  PRO_METHOD ("COLOR", "")

  Color[BLUE] = LIMIT (0.0, Blue, 1.0);

  PRO_END
}


void COLOR::SetAlpha (double Alpha)
{
  PRO_METHOD ("COLOR", "")

  Color[ALPHA] = LIMIT (0.0, Alpha, 1.0);

  PRO_END
}


double COLOR::GetRed () const
{
  return Color[RED];
}


double COLOR::GetGreen () const
{
  return Color[GREEN];
}


double COLOR::GetBlue () const
{
  return Color[BLUE];
}


double COLOR::GetAlpha () const
{
  return Color[ALPHA];
}


COLOR::ID COLOR::GetID () const
{
  return ColorID;
}


char *COLOR::GetName ()
{
  if (ColorID >= OBJECT && ColorID <= TEXTURE)
    return ColorNames[ColorID];
  else
    return NULL;
}

