#include "libcad3d.h"


ARROW::ARROW (WORLD  *World,   const char *Name,     H_MATRIX Frame,
	      double  Radius1, double      Len1,     
	      double  Radius2, double      Len2, 
	      double  Angle,   int         Facettes, double   Segment, 
	      const   MATERIAL &Material) :
  PART3D (NULL, World, Material)
{
  Part = PARTCreateArrow (Name, Frame, 
			  Radius1, Len1, Radius2, Len2, Angle, 
			  Facettes, Segment, Material.GetColorID());
  Initialize ();
}


BOLID::BOLID (WORLD  *World,   const char *Name,      H_MATRIX Frame,
	      double  Radius1, double      Radius2,   double   Radius3,
	      double  Height,  int         Facettes1, int      Facettes2, 
	      double  Segment, const       MATERIAL  &Material) :
  PART3D (NULL, World, Material)
{
  Part = PARTCreateBolid (Name, Frame, 
			  Radius1, Radius2, Radius3, Height, 
			  Facettes1, Facettes2, Segment, 
			  Material.GetColorID());
  Initialize ();
}


CONE::CONE (WORLD  *World,    const char *Name,    H_MATRIX Frame,
	    double  Radius1,  double      Radius2, double   Radius3,
	    double  SizeZ,    double      Segment, int      Facettes, 
	    const   MATERIAL &Material) :
  PART3D (NULL, World, Material)
{
  Part = PARTCreateCone (Name, Frame,
			 Radius1, Radius2, Radius3, SizeZ, Segment, Facettes, 
			 Material.GetColorID());
  Initialize ();
}


GEAR::GEAR (WORLD  *World,   const char *Name,    H_MATRIX Frame,
	    double  Radius1, double      Radius2, double   Radius3,
	    double  Depth1,  double      Depth2,  int      Teeth, 
	    double  Skew,    const       MATERIAL &Material) :
  PART3D (NULL, World, Material)
{
  Part = PARTCreateGear (Name, Frame, 
			 Radius1, Radius2, Radius3, Depth1, Depth2, 
			 Teeth, Skew, Material.GetColorID());
  Initialize ();
}


ICOSAHEDRON::ICOSAHEDRON (WORLD *World,    const char *Name,    H_MATRIX Frame,
			  double  Radius1, double      Radius2, 
			  int     Density, int         Iterations, 
			  const   MATERIAL &Material) :
  PART3D (NULL, World, Material)
{
  Part = PARTCreateIcosahedron (Name, Frame, 
				Radius1, Radius2, Density, Iterations, 
				Material.GetColorID());
  Initialize ();
}


CYLINDER::CYLINDER (WORLD *World,    const char *Name,     H_MATRIX Frame,
		    double Radius,   double      Depth,    double   Segment, 
		    int    Facettes, const       MATERIAL &Material) :
  PART3D (NULL, World, Material)
{
  Part = PARTCreateCylinder (Name, Frame, Radius, Depth, Segment, Facettes, 
			     Material.GetColorID());
  Initialize ();
}


PIPE::PIPE (WORLD *World,       const char *Name,        H_MATRIX Frame,
	    double Radius1,     double      Radius2, 
	    double Thickness1,  double      Thickness2,
	    double SizeZ,       double      Segment,    int Facettes, 
	    const  MATERIAL    &Material) :
  PART3D (NULL, World, Material)
{
  Part = PARTCreatePipe (Name, Frame, 
			 Radius1, Radius2, Thickness1, Thickness2,
			 SizeZ, Segment, Facettes, Material.GetColorID());
  Initialize ();
}


PRISM::PRISM (WORLD *World,     const char *Name,    H_MATRIX Frame,
	      double  Width1,   double      Width2,  double   OffsX,
	      double  Depth1,   double      Depth2,  double   OffsY,
	      double  Height1,  double      Height2, double   OffsZ,
	      const   MATERIAL &Material) :
  PART3D (NULL, World, Material)
{
  Part = PARTCreatePrism (Name, Frame,
			  Width1, Width2, OffsX,
			  Depth1, Depth2, OffsY,
			  Height1, Height2, OffsZ,
			  Material.GetColorID());
  Initialize ();
}


QUADER::QUADER (WORLD *World,    const char *Name,  H_MATRIX Frame,
		double SizeX,    double      SizeY, double   SizeZ, 
		const  MATERIAL &Material) :
  PART3D (NULL, World, Material)
{
  Part = PARTCreateQuader (Name, Frame,
			   SizeX, SizeY, SizeZ, Material.GetColorID());
  Initialize ();
}


ROTATIONAL::ROTATIONAL (WORLD  *World,    const char *Name,     H_MATRIX Frame,
			int     Number,   POINT2D    *Points, 
		        double  Segment,  int         Facettes, 
			double   Radius,  const       MATERIAL &Material) :
  PART3D (NULL, World, Material)
{
  Part = PARTCreateRot (Name, Frame, 
			Number, Points, Segment, Facettes, Radius, 
			Material.GetColorID());
  Initialize ();
}


SCREW::SCREW (WORLD *World,      const char *Name,      H_MATRIX Frame,
	      double  Radius1,   double      Radius2,
	      double  Height,    double      Increase,  double   Offset,
	      int     Facettes,  THREAD      Direction,
	      const   MATERIAL  &Material) :
  PART3D (NULL, World, Material)
{ 
  Part = PARTCreateScrew (Name, Frame,
			  Radius1, Radius2, Height, Increase, Offset,
			  Facettes, (Direction==RIGHT_THREAD),
			  Material.GetColorID());
  Initialize ();
}


SLICE::SLICE (WORLD *World,    const char *Name,      H_MATRIX Frame,
	      double  Radius1, double      Radius2,   double   Radius3,
	      double  SizeZ,   int         Facettes1, int      Facettes2, 
	      double  Segment, const       MATERIAL   &Material) :
  PART3D (NULL, World, Material)
{
  Part = PARTCreateSlice (Name, Frame,
			  Radius1, Radius2, Radius3, SizeZ, 
			  Facettes1, Facettes2, Segment, 
			  Material.GetColorID());
  Initialize ();
}


SPHERE::SPHERE (WORLD *World,     const char *Name,     H_MATRIX Frame,    
		double Radius,    double      SegmentX, double   SegmentZ, 
		int    Facettes,  const       MATERIAL  &Material) :
  PART3D (NULL, World, Material)
{
  Part = PARTCreateSphere (Name, Frame, 
			   Radius, SegmentX, SegmentZ, Facettes, 
			   Material.GetColorID());
  Initialize ();
}


SPRING::SPRING (WORLD *World,       const char *Name,      H_MATRIX Frame,
		double  Radius1,    double      Radius2,   double   Radius3,
		double  Height,     double      Increase,
		int     Facettes1,  int         Facettes2, THREAD   Direction,
		const   MATERIAL   &Material) :
  PART3D (NULL, World, Material)
{ 
  Part = PARTCreateSpring (Name, Frame,
			   Radius1, Radius2, Radius3, Height, Increase,
			   Facettes1, Facettes2, (Direction==RIGHT_THREAD),
			   Material.GetColorID());
  Initialize ();
}


SWEEP::SWEEP  (WORLD *World,    const char *Name,   H_MATRIX Frame,
	       int    Number,   POINT2D    *Points, double   Height, 
	       const  MATERIAL &Material) :
  PART3D (NULL, World, Material)
{
  Part = PARTCreateSweep (Name, Frame,
			  Number, Points, Height, Material.GetColorID());
  Initialize ();
}


TORUS::TORUS (WORLD *World,       const char *Name,      H_MATRIX Frame,
	      double  Radius1,    double      Radius2,
	      int     Facettes1,  int         Facettes2, double Segment, 
	      const   MATERIAL   &Material) :
  PART3D (NULL, World, Material)
{
  Part = PARTCreateTorus (Name, Frame,
			  Radius1, Radius2, Facettes1, Facettes2, Segment, 
			  Material.GetColorID());
  Initialize ();
}


WEDGE::WEDGE (WORLD *World,     const char *Name,     H_MATRIX Frame,
	      double  Width1,   double      Width2, 
	      double  Height1,  double      Height2, 
	      double  Depth1,   double      Depth2, 
	      const   MATERIAL &Material) :
  PART3D (NULL, World, Material)
{
  Part = PARTCreateWedge (Name, Frame,
			  Width1, Width2, Height1, Height2, Depth1, Depth2, 
			  Material.GetColorID());
  Initialize ();
}
