#include "libcad3d.h"


static void WORLD_Action (DEVICE3D *DevList, ACTION_FUNC Func, void *Ptr);


bool WORLD::Remove (DEVICE3D *Dev)
{
  bool      Result;
  DEVICE3D *Ptr = DevList;

  PRO_METHOD ("WORLD", "")

  if (Dev == NULL && DevList != NULL)
    PRO_RETURN ("Dev is NULL", TRUE)

  if ((Result = (DevList == Dev)))
    DevList = Dev->GetNext ();
  else
    while (! Result && Ptr->GetNext() != NULL)
    {
      if ((Result = (Ptr->GetNext() == Dev)))
	Ptr->SetNext (Dev->GetNext());
      else
	Result = Ptr->Remove (Dev);

      Ptr = Ptr->GetNext ();
    }

  if (Result)
  {
    DevNum--;
    PartNum -= Dev->GetPartNo ();
    Dev->SetFix  (NULL);
    Dev->SetNext (NULL);
  }

  PRO_END
  return Result;
}


bool WORLD::Remove (PART3D *Part)
{
  DEVICE3D *Ptr;

  PRO_METHOD ("WORLD", "")

  Ptr = DevList;
  while (Ptr != NULL)
  {
    if (Ptr->Remove(Part))
    {
      PartNum -= Part->GetPartNumber (CHAIN_MASK | ATTACH_MASK);
      /* 
	 Maybe, we should remove the device if it contains no other parts,
	 but removing is not deleting, and it might be difficult to free the 
         removed device pointer, later. So let's keep the empty device.
	 if (Ptr->GetPartNumber(CHAIN_MASK | ATTACH_MASK) == 0)
	   Remove (Ptr);
      */
      PRO_RETURN ("Part removed", TRUE)
    }
    Ptr = Ptr->GetNext ();
  }

  // TODO: Falls letztes Part im Device: auch Device entfernen ? (-> ja)

  PRO_END
  return TRUE;
}


bool WORLD::Delete (DEVICE3D *Dev)
{
  bool Result = TRUE;

  PRO_METHOD ("WORLD", "\n")

  if (Dev != NULL && (Result = Remove(Dev)))
    delete (Dev);   // TODO: werden alle Parts gelscht ?

  PRO_END
  return Result;
}


bool WORLD::Delete (PART3D *Part)
{
  bool      Result = TRUE;
  DEVICE3D *Dev;

  PRO_METHOD ("WORLD", "")

  if (Part != NULL && (Result = Remove(Part)))
  {
    Dev = Part->GetParent ();           // Get parent device

    Part->SetParent (NULL);             // disconnect and Delete the part
    Part->SetNext   (NULL);
    delete (Part);

    // if we have deleted the last part in a device, we also have to delete the
    // device.
    if (Dev != NULL && Dev->GetPartNumber() == 0)  // Delete parent, if empty
      Delete (Dev);
  }

  PRO_END
  return Result;
}

/* 
   Methoden:
   - Part in Next-Liste eines Parts anhngen
   - Part als Chain an Part anhaengen
   - Dev in Dev-Liste hinzufgen

   WICHTIG:
   Bei allen nderungen: 
   1. Zeiger korrekt setzen
   2. Previous-Zeiger aktualisieren
   3. Counter der Devices bzw. Parts im Device aktualisieren
*/



static void WORLD_Action (PART3D *PartList, ACTION_FUNC Func, void *DataPtr)
{
  PART3D   *Ptr, *Chain;
  DEVICE3D *Attach;

  Ptr = PartList;
  while (Ptr != NULL)
  {
    Func (Ptr->GetParent(), Ptr, DataPtr);

    if ((Chain = Ptr->GetChain()) != NULL)
      WORLD_Action (Chain, Func, DataPtr);

    if ((Attach = Ptr->GetAttach()) != NULL)
      WORLD_Action (Attach, Func, DataPtr);

    Ptr = Ptr->GetNext ();
  }
}

static void WORLD_Action (DEVICE3D *DevList, ACTION_FUNC Func, void *DataPtr)
{
  DEVICE3D *Ptr;
  
  Ptr = DevList;
  while (Ptr != NULL)
  {
    Func (Ptr, NULL, DataPtr);

    WORLD_Action (Ptr->GetPartList(), Func, DataPtr);

    Ptr = Ptr->GetNext ();
  }
}

void WORLD::UserAction (ACTION_FUNC Func, void *DataPtr)
{
  PRO_METHOD ("WORLD", "")

  if (Func == NULL)
    return;

  WORLD_Action (DevList, Func, DataPtr);

  PRO_END
}



static void WORLD_CalcDHP (DEVICE3D *Dev, PART3D *Part, void *Ptr)
{
  if (Part != NULL && Part->GetPrev() != NULL)
  {
    // Try to compute Denavit Hartenberg-Parameter (Test)
    double   Theta, d, a, Alpha;
    H_MATRIX M = Part->GetMatrix ();

    FRAME An (M.m[N][X], M.m[N][Y], M.m[N][Z], 
	      M.m[O][X], M.m[O][Y], M.m[O][Z], 
	      M.m[A][X], M.m[A][Y], M.m[A][Z], 
	      M.m[P][X], M.m[P][Y], M.m[P][Z]);
    An.GetDHP (&Theta, &d, &a, &Alpha);
    printf ("DHP %s: THETA=%7.2f D=%6.2f A=%6.2f ALPHA=%7.2f\n", 
	    Part->GetName(), Theta, d, a, Alpha);
  }
}


void WORLD::CalcDHP ()
{
  UserAction (WORLD_CalcDHP, NULL);
}
