#include <sys/stat.h>
#include <assert.h>
#include <stdio.h>
#include <stdarg.h>

#include "libcad3d.h"


CAD_FILE::CAD_FILE ()
{
}


bool CAD_FILE::Read (DATAFILE &File, double Version)
{
  return TRUE;
}


bool CAD_FILE::Write (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  return TRUE;
}


void CAD_FILE::Space (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  int count;

  if (fp == NULL)
    fp = stdout;

  for (count = 0; count < Indent; ++count)
    fprintf (fp, " ");
}



static void  WORLD_ReadDev  (FILE *fp, WORLD *World, H_MATRIX WTP, int Number);
static void  WORLD_WriteDev (FILE *fp, DEVICE3D *DevList, H_MATRIX WTP);


int Indent (FILE *fp /* = NULL */, int Indent /* = 0 */)
{
  int count, Result = 0;

  if (fp == NULL)
    fp = stdout;

  for (count = 0; count < Indent; ++count)
    Result += fprintf (fp, "  ");

  return Result;
}


bool WORLD::ReadLine (FILE *fp, bool show /* = FALSE */, 
		      char *Result /* = NULL */, int Len /* = 0 */)	      
{
  int  count = 0;
  char letter = ' ';

  while (! feof(fp) && letter != '\n')
  {
    letter = fgetc (fp);

    if (Result != NULL && count < Len)
      Result[count++] = letter;
      
    if (show)
      PRO1 ("\\%c", letter)
  }

  if (Result != NULL)
    Result[Len-1] = '\0';

  return (! feof(fp));
}



DEVICE3D *WORLD::LoadPart (const char *FileName)
{
  char      PartName[PATH_MAX+1];
  char      DevName[PATH_MAX+1];
  char     *DotPtr;
  PART     *Part;
  DEVICE3D *Dev;

  PRO_METHOD ("WORLD", "\n")

  TOOLGetFileName (FileName, PartName, PATH_MAX);

  if ((Part = PARTLoad(FileName, PartName, FALSE)) == NULL)
    PRO_RETURN ("Error loading part", NULL)

  PRO0 ("loaded\n")

  // create device name
  strcpy (DevName, PartName);
  if ((DotPtr = strrchr(DevName, '.')) != NULL)
    *DotPtr = '\0';
  strcat (DevName, ".dev");

  Win->Select ();
  Dev = new DEVICE3D (DevName, new PART3D(Part, this), this);
  AddDevice (Dev);

  if (AlphaNum > 0)
    Win->SetBlending (WINDOW3D::TRANSPARENT);
  Fill ();

  PRO_END
  return Dev;
}


bool WORLD::LoadDXF (const char *FileName)
{
  FILE     *fp;
  PART     *Part;
  DEVICE3D *Dev;

  if ((fp = fopen(FileName, "r")) == NULL)
    return FALSE;

  if ((Part = PARTLoadDXF(fp, "test", FALSE)) == NULL)
  {
    PRO0 ("Error loadind DXF")
    return FALSE;
  }
  else
    PRO0 ("done !\n")

  Dev = new DEVICE3D ("test", new PART3D(Part, this), this);
  AddDevice (Dev);

  if (AlphaNum > 0)
    Win->SetBlending (WINDOW3D::TRANSPARENT);
  Fill ();

  return TRUE;
}


bool WORLD::LoadWorld (const char *FileName,
		       const char *DataPath /* = NULL */)
{
  int  State = -1;
  char Name[PATH_MAX+1];

  PRO_METHOD ("WORLD", "\n")

  assert (FileName != NULL);

  if (strcasecmp(FileName+strlen(FileName)-4, ".dxf") == 0)
    State = (LoadDXF(FileName) == TRUE);
  elsif (strcasecmp(FileName+strlen(FileName)-4, ".tcf") == 0)
    State = (Load(FileName) == TRUE);

  if (State < 0)
    PRO_RETURN1 ("Bad Status: %d", State, FALSE)

  TOOLGetFileName (FileName, Name, PATH_MAX);
  SetName (Name);

  if (AlphaNum > 0)
    Win->SetBlending (WINDOW3D::TRANSPARENT);
  Fill ();

  PRO_END
  return (State >= 0);
}


bool WORKCELL::LoadWorld (const char *FileSpec, 
			  const char *DataPath /* = "~/tuxcad/data" */)
{
  int  State = -1;
  char Name[PATH_MAX+1];

  PRO_METHOD ("WORKCELL", "\n")

  assert (FileSpec != NULL);

  Win->Select (WINDOW3D::LOCK);

  if (strcasecmp(FileSpec+strlen(FileSpec)-4, ".dxf") == 0)
    PRO_RETURN ("", WORLD::LoadWorld(FileSpec))
  elsif (strcasecmp(FileSpec+strlen(FileSpec)-4, ".tcf") == 0)
    PRO_RETURN ("", WORLD::LoadWorld(FileSpec))
  else
    PRO_RETURN ("Unknown file format", FALSE)

  if (State < 0)
    PRO_RETURN1 ("Bad Status: %d", State, FALSE);

  TOOLGetFileName (FileSpec, Name, PATH_MAX);
  SetName (Name);

  if (AlphaNum > 0)
    Win->SetBlending (WINDOW3D::TRANSPARENT);
  Fill ();

  Win->Select (WINDOW3D::UNLOCK);

  PRO_END
  return (State >= 0);
}


static void WORLD_ReadPart (FILE *fp, WORLD *World, DEVICE3D *Dev, 
			    H_MATRIX WTP, int Number)
{
  int         count;
  char        PartName[PATH_MAX+1];
  char        PartFile[PATH_MAX+1];
  // PART       *Part;
  H_MATRIX    Frame /* , Save[2] */;
  /*
  PART3D   *Ptr, *Next;
  DEVICE3D *Attach;
  */

  for (count = 0; count < Number; ++count)
  {
    World->ReadLine (fp, TRUE);
    fscanf (fp, "  PartName: %s\n", PartName);
    fscanf (fp, "  PartFile: %s\n", PartFile);
    fscanf (fp, "  Frame: %lf %lf %lf %lf\n", 
	    &Frame.m[N][X], &Frame.m[O][X], &Frame.m[A][X], &Frame.m[P][X]);
    fscanf (fp, "         %lf %lf %lf %lf\n", 
	    &Frame.m[N][Y], &Frame.m[O][Y], &Frame.m[A][Y], &Frame.m[P][Y]);
    fscanf (fp, "         %lf %lf %lf %lf\n", 
	    &Frame.m[N][Z], &Frame.m[O][Z], &Frame.m[A][Z], &Frame.m[P][Z]);
    fscanf (fp, "         %lf %lf %lf %lf\n", 
	    &Frame.m[N][W], &Frame.m[O][W], &Frame.m[A][W], &Frame.m[P][W]);

    World->ReadLine (fp);
    World->ReadLine (fp, TRUE);

    World->ReadLine (fp);
    World->ReadLine (fp, TRUE);
  }

#if 0
  Next = PartList;
  while (Next != NULL)
  {
    Save[0] = WTP;     //  glPushMatrix ();
    Part = Next->GetPart ();

    Part->WTP = GEOMatrixProd (WTP, Next->GetRel());    // update Part-Position
    //    glMultMatrixd ((GLdouble*)Next->GetRel().m);

    printf ("FileName: %s\n", Part->FileName);
    printf ("  %8.5f %8.5f %8.5f %8.5f\n", 
	    WTP.m[N][X], WTP.m[O][X], WTP.m[A][X], WTP.m[P][X]);
    printf ("  %8.5f %8.5f %8.5f %8.5f\n", 
	    WTP.m[N][Y], WTP.m[O][Y], WTP.m[A][Y], WTP.m[P][Y]);
    printf ("  %8.5f %8.5f %8.5f %8.5f\n", 
	    WTP.m[N][Z], WTP.m[O][Z], WTP.m[A][Z], WTP.m[P][Z]);
    printf ("  %8.5f %8.5f %8.5f %8.5f\n", 
	    WTP.m[N][W], WTP.m[O][W], WTP.m[A][W], WTP.m[P][W]);

    if ((Ptr = Next->GetChain()) != NULL)
    {
      Save[1] = WTP;  //      glPushMatrix ();
      WORLD_LoadPart (fp, World, Dev, Part->WTP, 0);
      WTP = Save[1]; // glPopMatrix ();
    }

    if ((Attach = Next->GetAttach()) != NULL)
    {
      Save[1] = WTP;  //      glPushMatrix ();
      WORLD_ReadDev (fp, World, Part->WTP, 0);
      WTP = Save[1]; // glPopMatrix ();
    }
    WTP = Save[0];   // glPopMatrix ();

    Next = Next->GetNext ();
  }
#endif
}


static void WORLD_ReadDev (FILE *fp, WORLD *World, H_MATRIX WTP, int Number)
{
  int       count, PartNum;
  char      Line[80];
  DEVICE3D *Dev;

  for (count = 0; count < Number; ++count)
  {
    fscanf (fp, "  DevName: %s\n", Line);
    Dev = new DEVICE3D (Line, NULL, World);
    fscanf (fp, "  Parts: %d\n", &PartNum);
    printf ("# Parts: %d\n", PartNum);
    World->ReadLine (fp);

    WORLD_ReadPart (fp, World, Dev, WTP, PartNum);

    World->AddDevice (Dev);
  }
}


bool WORLD::Load (const char *FileSpec)
{
  int   DevNum;
  FILE *fp;
  
  if ((fp = fopen(FileSpec, "r")) == NULL)
  {
    PRO1 ("Error opening file: %s\n", FileSpec)
    return FALSE;
  }

  ReadLine (fp);                        // TuxCAD: WORKCELL
  ReadLine (fp);                        // Version
  ReadLine (fp);

  ReadLine (fp);                        // [Comment]
  ReadLine (fp);

  ReadLine (fp);                        // [Grid]
  ReadLine (fp);

  ReadLine (fp);                        // World
  ReadLine (fp);

  ReadLine (fp, TRUE);                  // [Devices]
  fscanf (fp, "  Number: %d\n", &DevNum);
  ReadLine (fp);

  WORLD_ReadDev (fp, this, Matrix_1, DevNum);

  fclose (fp);

  return TRUE;
}


static void WORLD_WritePart (FILE *fp, PART3D *PartList, H_MATRIX WTP)
{
  int       Index = 1;
  PART     *Part;
  H_MATRIX  Save[2];
  PART3D   *Ptr, *Next;
  DEVICE3D *Attach;

  Next = PartList;
  while (Next != NULL)
  {
    Save[0] = WTP;     //  glPushMatrix ();
    Part = Next->GetPart ();

    WTP = GEOMatrixProd (WTP, Part->WTP);
    //    WTP = GEOMatrixProd (WTP, Next->GetRel());    // update Part-Position
    //    glMultMatrixd ((GLdouble*)Next->GetRel().m);

    fprintf (fp, "    Part: %d\n", Index++);
    fprintf (fp, "    PartName: %s\n", Part->PartName);
    fprintf (fp, "    PartFile: %s\n", Part->FileName);
    fprintf (fp, "    Frame:    %8.5f %8.5f %8.5f %8.5f\n", 
	         WTP.m[N][X], WTP.m[O][X], WTP.m[A][X], WTP.m[P][X]);
    fprintf (fp, "              %8.5f %8.5f %8.5f %8.5f\n", 
	         WTP.m[N][Y], WTP.m[O][Y], WTP.m[A][Y], WTP.m[P][Y]);
    fprintf (fp, "              %8.5f %8.5f %8.5f %8.5f\n", 
	         WTP.m[N][Z], WTP.m[O][Z], WTP.m[A][Z], WTP.m[P][Z]);
    fprintf (fp, "              %8.5f %8.5f %8.5f %8.5f\n", 
	         WTP.m[N][W], WTP.m[O][W], WTP.m[A][W], WTP.m[P][W]);
    fprintf (fp, "\n");

    Ptr = Next->GetChain ();
    fprintf (fp, "    Chain: %d\n", (Ptr==NULL)? 0:Ptr->GetPartNumber());
    if (Ptr != NULL)
    {
      Save[1] = WTP;     // glPushMatrix ();
      WORLD_WritePart (fp, Ptr, Part->WTP);
      WTP = Save[1];     // glPopMatrix ();
    }
    fprintf (fp, "\n");

    Attach = Next->GetAttach ();
    fprintf (fp, "    Attach: %d\n", 
	         (Attach==NULL)? 0:Attach->GetPartNumber());
    if ((Attach) != NULL)
    {
      Save[1] = WTP;  //      glPushMatrix ();
      WORLD_WriteDev (fp, Attach, Part->WTP);
      WTP = Save[1]; // glPopMatrix ();
    }
    fprintf (fp, "\n");

    WTP = Save[0];   // glPopMatrix ();

    Next = Next->GetNext ();
  }
}


static void WORLD_WriteDev (FILE *fp, DEVICE3D *DevList, H_MATRIX WTP)
{
  int       Index = 1;
  PART3D   *PartList;
  DEVICE3D *Dev;

  Dev = DevList;
  while (Dev != NULL)
  {
    PartList = Dev->GetPartList();
    fprintf (fp, "  Device %d\n", Index++);
    fprintf (fp, "  DevName: %s\n", Dev->GetName());
    fprintf (fp, "  Parts: %d\n", 
	         (PartList==NULL)? 0:PartList->GetPartNumber());
    fprintf (fp, "\n");
    Dev->Write (fp, 2);

    WORLD_WritePart (fp, Dev->GetPartList(), WTP);
    Dev = Dev->GetNext ();
  }
}


bool WORLD::Write (FILE *fp /* = NULL */) const
{
  int count;

  if (fp == NULL)
    fp = stdout;

  fprintf (fp, "TUXCAD\n");
  fprintf (fp, "Version: 0.0106\n");
  fprintf (fp, "\n");

  fprintf (fp, "[Comment]\n");
  fprintf (fp, "\n");

  fprintf (fp, "[Grid]\n");
  fprintf (fp, "\n");

  fprintf (fp, "[World]\n");
  fprintf (fp, "  Name:         %s\n", Name);
  fprintf (fp, "  PolygonMode:  ");
  switch (Win->GetPolygonMode())
  {
    case WINDOW3D::POINTS :
      fprintf (fp, "POINTS\n");
      break;

    case WINDOW3D::WIRE_FRAME :
      fprintf (fp, "WIRE-FRAME\n");
      break;

    case WINDOW3D::HIDDEN_LINE :
      fprintf (fp, "HIDDEN-LINE\n");
      break;

    case WINDOW3D::FACES :
      fprintf (fp, "FACES\n");
      break;

    default :
      fprintf (fp, "INVALID Mode\n");
  }
  fprintf (fp, "  Blending:     %s\n",
	       (Win->GetBlending()==WINDOW3D::SOLID)? "SOLID":"TRANSPARENT");
  fprintf (fp, "  Range:        %.3g %.3g %.3g ... %.3g %.3g %.3g\n",
	        Range[MIN_X], Range[MIN_Y], Range[MIN_Z], 
	        Range[MAX_X], Range[MAX_Y], Range[MAX_Z]);
  fprintf (fp, "  Part-Highlight:\n");
  PartHighLight.Write (fp, 4);
  fprintf (fp, "  Dev->Highlight:\n");
  DevHighLight.Write (fp, 4);
  fprintf (fp, "\n");

  fprintf (fp, "[Lightning]\n");
  fprintf (fp, "  Lights:       %d\n", OpenGL->GetMaxLights());
  fprintf (fp, "\n");
  for (count = 0; count < OpenGL->GetMaxLights(); ++count)
  {
    Lights[count]->Write (fp, 2);
    fprintf (fp, "\n");
  }

  fprintf (fp, "[Data]\n");
  fprintf (fp, "  Devices:    %d\n", GetDevNum());
  //	       (DevList==NULL)? 0:DevList->GetDevNumber());
  fprintf (fp, "  Parts:      %d\n", GetPartNum());
  fprintf (fp, "\n");
  WORLD_WriteDev (fp, DevList, Matrix_1);
  fprintf (fp, "\n");

  fclose (fp);
  return TRUE;
}


bool WORLD::Save (const char *FileName)
{
  FILE *fp = NULL;
  
  if ((fp = fopen(FileName, "w+")) == NULL)
  {
    PRO1 ("Error opening file: %s\n", FileName)
    return FALSE;
  }

  return Write (fp);
}



