#include <stdarg.h>

#include <libtools.h>

#include "libcad3d.h"



#define FONT "-adobe-helvetica-medium-r-normal--17-120-100-100-p-88-iso8859-1"


static int VisualAttr[] = 
{
  GLX_RGBA,
  GLX_RED_SIZE,     1,
  GLX_GREEN_SIZE,   1,
  GLX_BLUE_SIZE,    1,
  // GLX_ALPHA_SIZE,    1,
  GLX_DEPTH_SIZE,   1,
  //  GLX_STENCIL_SIZE, 1,
  GLX_DOUBLEBUFFER,
  None
};

static OPENGL GLParam;                       // Global attribute

OPENGL *OpenGL = &GLParam;


OPENGL::OPENGL () :
  VisualInfo  (NULL),
  StencilSize (0),
  MaxLights   (8),
  FontBase    (0)
{
  PRO_METHOD ("OPENGL", "")

  PRO_END
}


OPENGL::~OPENGL ()
{
  PRO_METHOD ("~OPENGL", "")

  if (VisualInfo != NULL)
    XFree (VisualInfo);

  PRO_END
}


void OPENGL::Initialize ()
{
  int Config;
  /*
    XVisualInfo    VIProto;
    XVisualInfo   *pVI;
    unsigned long  Mask;
  */

  PRO_METHOD ("OPENGL", "\n")

  glXQueryVersion (XDisplay, &Version[0], &Version[1]);
  PRO2 ("OpenGL-Version: %d.%d\n", Version[0], Version[1])

  glXQueryExtension (XDisplay, &Extension[0], &Extension[1]);
  PRO2 ("Extensions:     %d.%d\n", Extension[0], Extension[1])

  /*
  VIProto.screen = XScreen;
  pVI = XGetVisualInfo (XDisplay, VisualScreenMask, &VIProto, &VINum);
  */

  if ((VisualInfo = glXChooseVisual(XDisplay, XScreen, VisualAttr)) == NULL)
    PRO_ABORT ("Error: Selecting RGB + double-buffered visual failed\n")

  // print some infos about the selected visual (run 'glxinfo' for details)
  glXGetConfig (XDisplay, VisualInfo, GLX_STENCIL_SIZE, &Config);
  PRO1 ("Selected Visual ID: 0x%lx\n", VisualInfo->visualid)

  Context = glXCreateContext (XDisplay, VisualInfo, NULL, GL_TRUE);
  if (Context == NULL)
    PRO_ABORT ("Error creating context")

  PRO1 ("Direct rendering: %s\n", glXIsDirect(XDisplay, Context)? "YES":"NO")

#if 0
  glutInit (NULL, 0);
  glutInitDisplayMode (GLUT_RGB);
#endif

  // Get vendor string
  if (glGetString(GL_VENDOR) == NULL)
    Vendor[0] = '\0';
  else
  {
    strncpy (Vendor, (char *)glGetString(GL_VENDOR), 39);
    Vendor[39] = '\0';
  }

  glXQueryVersion   (XDisplay, &Version[0],   &Version[1]);
  glXQueryExtension (XDisplay, &Extension[0], &Extension[1]);

  //  glXGetConfig (XDisplay, VisualInfo, GLX_STENCIL_SIZE, &StencilSize);

  glGetIntegerv (GL_MAX_LIGHTS,       (GLint *)&MaxLights);
  glGetIntegerv (GL_MAX_TEXTURE_SIZE, (GLint *)&TextureSize);

  FontBase = MakeFont (FONT);
 
  PRO_END
}


bool OPENGL::Write (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  if (fp == NULL)
    fp = stdout;

  fprintf (fp, "[OPENGL]\n");
  fprintf (fp, "  Vendor:       %s\n",      Vendor);
  fprintf (fp, "  Version:      %d.%d\n",   Version[0], Version[1]);
  fprintf (fp, "  Extension:    %d / %d\n", Extension[0], Extension[1]);
  fprintf (fp, "  StencilSize:  %d\n",      StencilSize);
  fprintf (fp, "  TextureSize:  %d\n",      TextureSize);
  fprintf (fp, "  Max Lights:   %d\n",      MaxLights);

  PRO_END
  return TRUE;
}


int OPENGL::GetID (int ID /* = -1 */, const char *Name /* = NULL */)
{
  PRO_METHOD ("OPENGL", "")

  if (ID < 0)
    ID = glGenLists (1);
  elsif (glIsList(ID))
    FreeID (ID);

  PRO2 ("\\ID %s: %d...", (Name == NULL)?"":Name, ID)

  PRO_END
  return ID;
}


void OPENGL::FreeID (int ID)
{
  if (ID >= 0)
  {
    PRO1 ("Free OpenGL-ID %d\n", ID)
    glDeleteLists (ID, 1);
  }
}


int OPENGL::GetMaxLights () const
{ 
  return MaxLights;
}


GLuint OPENGL::GetFontBase () 
{
  if (FontBase == 0)
    FontBase = MakeFont (FONT);

  return FontBase;
}


GLuint OPENGL::MakeFont (const char *FontSpec /* = "fixed" */, ...)
{
  unsigned int  First, Last;
  char          FontName[256];
  XFontStruct  *FontInfo;
  Font          FontID;
  va_list       Arguments;
  GLuint        ID;

  PRO_METHOD ("OPENGL", "\n")

  va_start (Arguments, FontSpec);
  vsprintf (FontName, FontSpec, Arguments);
  va_end   (Arguments);

  if ((FontInfo = XLoadQueryFont(XDisplay, FontName)) == NULL)
    PRO_RETURN ("No font found", 0)

  FontID = FontInfo->fid;
  First  = FontInfo->min_char_or_byte2;
  Last   = FontInfo->max_char_or_byte2;

  if ((ID = glGenLists((GLuint)(Last-First+1))) == 0)
    PRO_RETURN ("Out of display lists", 0)

  glXUseXFont (FontID, First, Last-First+1, ID+First);

  PRO_END
  return ID;
}


GLXContext OPENGL::GetContext () const
{
  return Context;
}


XVisualInfo *OPENGL::GetVisualInfo ()
{
  return VisualInfo;
}


/* 
   TODO:
   Features der benutzten OpenGL-Implementierung ermitteln/ausgeben/abfragen
   Liste mit Bauteilen, die global, aber nur einmal verfgbar sein mssen:
   z.B. Koordinatensysteme, Lichtquellen ?, ...
*/
