#include "libcad3d.h"


LINE_STYLE::LINE_STYLE (COLOR Color /* = COLOR(COLOR::BLACK) */,
			int   Width /* = 1 */, int Pattern /* = 0xFFFF */) :
  CAD_FILE     (),
  UNIT         (),
  Color        (Color),
  Width        (Width),
  Pattern      (Pattern),
  Antialiasing (FALSE)
{
}


bool LINE_STYLE::Read (DATAFILE &File, double Version)
{
  PRO_METHOD ("LINE_STYLE", "\n")

  if (File.Parse("<\\[LINE\\]>\n") < 0)
    return FALSE;

  if (! CAD_FILE::Read(File, Version))
    return FALSE;

  if (! Color.Read(File, Version))
    return FALSE;

  if (File.Parse("  <Width:>     (%d)\n"
		 "  <Pattern:>   (%x)\n"
		 "  <AntiAlias:> (NO|YES)\n",
		 &Width, &Pattern, &Antialiasing) < 3)
    return FALSE;

  Create ();

  PRO_END
  return TRUE;
}


bool LINE_STYLE::Write (FILE *fp /* = NULL */, int Indent /* = 0 */) const
{
  PRO_METHOD ("LINE_STYLE", "\n")

  if (fp == NULL)
    fp = stdout;

  Space  (fp, Indent);
  fprintf (fp, "[LINE]\n");
  Indent += INDENT;

  if (! CAD_FILE::Write(fp, Indent))
    return FALSE;

  Color.Write (fp, Indent);

  Space   (fp, Indent);
  fprintf (fp, "Width:     %d\n", Width);

  Space   (fp, Indent);
  fprintf (fp, "Pattern:   0x%x\n", Pattern);

  Space   (fp, Indent);
  fprintf (fp, "AntiAlias: %s\n", Antialiasing?"YES":"NO");

  PRO_END
  return TRUE;
}


bool LINE_STYLE::Create ()
{
  Id = OpenGL->GetID (Id, "Line");
  glNewList (Id, GL_COMPILE);
  glIndexi  (Id);

    glLineWidth (Width);
    if (Pattern != 0xFFFF)
    {
      glEnable (GL_LINE_STIPPLE);
      glLineStipple (1 /* factor */, Pattern);
      glDisable (GL_LINE_STIPPLE);
    }

  glEndList ();

  return TRUE;
}


void LINE_STYLE::Set () const
{
  if (Id > 0)
    glCallList (Id);
}


void LINE_STYLE::SetColor (COLOR Color /* = COLOR(COLOR::BLACK) */)
{
  this->Color = Color;
}


void LINE_STYLE::SetWidth (int Width /* = 1 */)
{
  this->Width = Width;
}


void LINE_STYLE::SetPattern (int Pattern /* = 0xFFFF */)
{
  this->Pattern = Pattern;
}


void LINE_STYLE::SetAntialiasing (bool Mode)
{
  Antialiasing = Mode;
}


COLOR &LINE_STYLE::GetColor ()
{
  return Color;
}


int LINE_STYLE::GetWidth () const
{
  return Width;
}


int LINE_STYLE::GetPattern () const
{
  return Pattern;
}


bool LINE_STYLE::GetAntiAliasing () const
{
  return Antialiasing;
}
