#include <sys/stat.h>
#include <stdarg.h>
#include <ctype.h>

#include "libcad3d.h"

// This code is very ugly and unsave. So, feel free to change it. Maybe you
// might use lexx for this job.
// Here's a description of the accepted syntax:
// <foo>           search for the occurrance of 'foo'
// (%T)            expects exact one value of type T (see below)
// (%20l)          reads until end of line, but returns max. 20 characters
// (KEY1|KEY2|...) expects one of the Keywords 'KEY1' *or* 'KEY2' or ...
//                 with=KEY1=1, KEY2=2, ...
// [...]           expects one or none value (e.g. [%d] max one decimal value
// {KEY1|KEY2|...} accepts a combination of the keywords separated by spaces
//                 e.g. 'KEY1 KEY2'. The returned value comply to the binary 
//                 OR-bombination of this bits.
// #               starts a comment. Rest of line will be ignored
// Construtcs may also be nested: e.g. [<foo> (%d)]\n looks for a OPTIONAL 
// line beginning with foo and a following decimal value. This feature has not
// been tested for all combinations and may fail in some cases.
// The follwoing data type are realized:
// %d   a decimal value (int)
// %f   a floating point value (double)
// %l   reads end of line 
//      (the length of the returning string may be restrincted (e.g. %128l)
// %c   reads one character
// %s   reads a string, length might be also restrincted


#define ID_WORDS  64
#define ID_LENGTH 25

#define MAX_STRING_LEN 128


/* Parse-Strukturen */
enum TC_PARSE_CODE
{
  NOCODE = 0, 
  KEYCODE, TEXTCODE, REMCCODE, REMSCODE,
  INTCODE, HEXCODE, FLOATCODE, CHARCODE, STRINGCODE, LINECODE, DUALCODE, IDCODE
};

enum TC_BACKET_CODE
{
  NOBRACE = 0, 
  LESSBRACE, RNDBRACE, NDXBRACE, SETBRACE, SETSETBRACE
};


typedef struct 
{
  int   IBracket;                       /* inner bracket              */
  int   OBracket;                       /* outer bracket              */
  int   Code;                           /* value code                 */
  int   Length;                         /* max. String length         */
  int   Number;                         /* amount of IDs              */
  int   Amount;                         /* {{...}}-frequency          */
  char  IDWord[ID_WORDS][ID_LENGTH];    /* max 30 ID-words a 25 chars */
  char  Stop;                           /* Stop at this char          */
} PARSE_INFO;


static dual Error = FALSE;



DATAFILE::DATAFILE (const char *FileName, char Comment /* = '#' */) :
  Buffer (NULL),
  LineNo (0),
  Ptr    (NULL)
{
         int    Index = 0, count = 0;
         char   Letter;
         FILE *fp = NULL;
  struct stat  Info;

  if ((fp = fopen(FileName, "r")) == NULL)
    printf ("ERROR Opening file '%s'\n", FileName);  // TODO: EXCEPTION

  if (stat(FileName, &Info) != 0)
    printf ("ERROR Reading File\n");  // TODO: EXCEPTION

  Buffer = new char[Info.st_size+2];

  do
  {
    Buffer[Index] = (char)getc (fp);
    if (Buffer[Index] == Comment && Comment != '\0')
    {
      Buffer[Index] = '\n';
      do 
	count++;
      while ((Letter = getc(fp)) != '\0' && Letter != '\n' && Letter != EOF);
    }
  } while (((count++) <= Info.st_size) && (Buffer[Index++] != '\0'));

  fclose (fp);  
  
  Buffer[Index-1] = '\n';
  Buffer[Index]   = '\0';

  Ptr = Buffer;
}


DATAFILE::~DATAFILE ()
{
  if (Buffer != NULL)
    delete Buffer;
}


int DATAFILE::ReadLine ()
{
  return 0;
}


int DATAFILE::Parse (const char *Mask, ...)
{
  bool         Escape;
  int          Result = 0;
  int          IDVal;
  int          IntVal;
  double       DoubleVal;
  char         CharVal;
  char        *TextVal = NULL;

  bool        *RetDual = NULL;
  int         *RetInt = NULL;
  int         *RetID = NULL;
  double      *RetDouble = NULL;
  char        *RetChar = NULL;
  char        *RetText = NULL;
  char        *RetLine = NULL;
  void       **RetMulti = NULL;
  char        *RetData = NULL;

  va_list      Arguments;
  dual         ready = FALSE;
  dual         EOL = TRUE;
  dual         found = TRUE, ok = FALSE;
  int          count, mc = 0;
  int          repeat = 0, Size;
  char         KeyWord[80] = "";
  char         TextWord[80] = "";
  char         String[1024];
  PARSE_INFO   Parse;
  char        *Start;
  char        *Save = NULL;

  PRO_START ("\n")

  if (Ptr == NULL)
    PRO_RETURN ("No text found", 0)

  Start = Ptr;

  va_start (Arguments, Mask);

  Parse.OBracket = NOBRACE;
  Parse.Amount   = 0;
  while (Mask[mc] != '\0')
  {
    /* ***** nchsten Maskeneintrag suchen & auswerten: ***** */
    ready = FALSE;
    Parse.IBracket = NOBRACE;
    Parse.Code     = NOCODE;
    Parse.Number   = 0;
    while (! ready && Mask[mc] != '\0')
    {
      switch (Mask[mc])
      {
        case '<' : 
          if (Parse.IBracket != NOBRACE)
            Parse.OBracket = Parse.IBracket;
          Parse.IBracket = LESSBRACE;
          break;

        case '(' : 
          if (Parse.IBracket != NOBRACE)
            Parse.OBracket = Parse.IBracket;
          Parse.IBracket = RNDBRACE;
          break;

        case '[' : 
          Parse.IBracket = NDXBRACE;
          break;

        case '{' : 
          if (Mask[mc+1] == '{')
          {
            Parse.IBracket = SETSETBRACE;
            mc++;
            repeat = mc;
            Size = 0;
            while (Mask[mc] != '}' || Mask[mc+1] != '}')
              if (Mask[mc++] == '%')
              {
                if (isdigit(Mask[mc]))
                {
                  count = 0;
                  do
                    String[count++] = Mask[mc++];
                  while (isdigit(Mask[mc]));
                  String[count] = '\0';
                  Parse.Length = atoi (String);
                }
                else
                  Parse.Length = MAX_STRING_LEN;

                switch (Mask[mc])
                {
                  case 'd' : Size += sizeof(int);
                             break;
                  case 'x' : Size += sizeof(long);
                             break;
                  case 'f' : Size += sizeof(double);
                             break;
                  case 'c' : Size += sizeof(char);
                             break;
                  case 's' : Size += sizeof(char) * Parse.Length;
                             break;
                  case 'l' : Size += sizeof(char) * Parse.Length;
                             break;
                   default : PRO_RETURN ("Unknown Data-type", -1)
                }
                mc++;
              }
            mc = repeat;
            RetData   = (char *)malloc (Size * Parse.Amount);
            RetMulti  = (void **)va_arg (Arguments, void **);
            *RetMulti = (void *)RetData;
          }
          else
            Parse.IBracket = SETBRACE;
          break; 

        case '%' : 
          mc++;
          if (Parse.IBracket != NOBRACE)
          {
            if (isdigit(Mask[mc]))
            {
              count = 0;
              do
                String[count++] = Mask[mc++];
              while (isdigit(Mask[mc]));
              String[count] = '\0';
              Parse.Length = atoi (String);
            }
            else
              Parse.Length = MAX_STRING_LEN;

            switch (Mask[mc])
            {
              case 'd' : Parse.Code = INTCODE;
                         break;
  	      case 'x' : Parse.Code = HEXCODE;
		         break;
              case 'f' : Parse.Code = FLOATCODE;
                         break;
              case 'c' : Parse.Code = CHARCODE;
                         break;
              case 's' : Parse.Code = STRINGCODE;
                         break;
              case 'l' : Parse.Code = LINECODE;
                         break;
               default : PRO_RETURN ("Unknown data-type", -1)
            }
            mc++;
            while (Mask[mc] != ')' && Mask[mc] != ']' && Mask[mc] != '}')
              mc++;
	    if (Parse.Code == STRINGCODE)
	      Parse.Stop = Mask[mc+1];  
          }
          else
          {
            Parse.Code = TEXTCODE;
            TextWord[0] = '%';
            count = 1;
            while (Mask[mc] != ' ' && Mask[mc] != '\t' && Mask[mc] != '\n')
              TextWord[count++] = Mask[mc++];
            TextWord[count] = '\0';
          }
          ready = TRUE;
          break;

        case ' ' : 
        case '\n' : 
        case '\t' : 
          break;

        case '#' :
          if (Mask[mc+1] == '#')
          { 
            Parse.Code = REMSCODE; mc++;
          }
         else
            Parse.Code = REMCCODE;
          ready = TRUE;
          break;

        case ')' : 
          if (Parse.OBracket != RNDBRACE)
            PRO_RETURN ("  Too much ')'-Brackets", -2)
          Parse.OBracket = NOBRACE;
          break;

        case ']' : 
          if (Parse.OBracket != NDXBRACE)
	    PRO_RETURN ("Too much ']'-Brackets", -2)
          Parse.OBracket = NOBRACE;
          break;

        case '}' :
          if (Parse.OBracket != SETBRACE && Parse.OBracket != SETSETBRACE)
	    PRO_RETURN ("Too much '}'-Brackets", -2)

          if (Parse.OBracket == SETSETBRACE)
            if (Mask[mc+1] != '}')
	      PRO_RETURN ("Bad close-brackets '}}'", -2)
            else if (Parse.Amount > 1)
            {
              Parse.Amount--;
	      mc = repeat;
            }
            else
            {
              Parse.OBracket = NOBRACE; 
	      mc++;
            }
          else
            Parse.OBracket = NOBRACE;                        
          break;

        default :
          if (Parse.IBracket == NOBRACE)
          {
	    Escape = FALSE;
            count = 0;
            while (Mask[mc] != ' ' && Mask[mc] != '\t' && Mask[mc] != '\n' &&
		   (Escape || 
		    Mask[mc] != '(' && Mask[mc] != '['  && Mask[mc] != '{'))
	    {
	      if (! (Escape = (Mask[mc] == '\\')))
		TextWord[count++] = Mask[mc];
	      mc++;
	    }
            TextWord[count] = '\0';
            Parse.Code = TEXTCODE;
	    Parse.Length = count;
            ready = TRUE;
	    if (Mask[mc] == '(' || Mask[mc] == '[' || Mask[mc] == '{')
	      mc--;
          }
          else if (Parse.IBracket == LESSBRACE)
          {
            count = 0;
            while (Mask[mc] != '>')
	    {
              if (Mask[mc] != '\\')
                KeyWord[count++] = Mask[mc];
              mc++;
	    }
            KeyWord[count] = '\0';
            Parse.Code = KEYCODE;
            ready = TRUE;
          }
          else if (Parse.IBracket > LESSBRACE)
          {
            Parse.Number = 1; count = 0;
            Parse.Length = MAX_STRING_LEN;
            Parse.Code = IDCODE;
            IDVal = 0;
            while (Mask[mc] != ')' && Mask[mc] != ']' && Mask[mc] != '}')
            {
              if (Mask[mc] == '|')
              {
                Parse.IDWord[Parse.Number-1][count] = '\0';
                Parse.Number++; count = 0;
              }
              else
              {
                Parse.IDWord[Parse.Number-1][count] = Mask[mc];
                count++;
              }
              mc++;
            }
            Parse.IDWord[Parse.Number-1][count] = '\0';
            if (Mask[mc] == '}' && Mask[mc+1] == '}')
              mc++;
            ready = TRUE;
          }
      }
      mc++;
    }

    if (!ready && Mask[mc] == '\0' && 
        (Parse.IBracket != NOBRACE || Parse.Code != NOCODE))
      PRO_RETURN ("End of Mask reached !", -3)

    /* geparsten Code in der Datei ermitteln und in Ergebnis schreiben */
    if (Parse.Code == KEYCODE && 
        (Parse.OBracket != SETSETBRACE || Parse.Amount > 0))
    {
      PRO1 ("Parse.KeyWord '%s'\n", KeyWord)
      Save = Ptr;
      GetString (String, 1024, '\0', TRUE);
      while (!(found = (strcmp(String, KeyWord) == 0)) && CheckData())
        GetString (String, 1024, '\0', TRUE);
      if (! CheckData())
      {
        if (Parse.OBracket == NDXBRACE)
        {
          PRO0 ("unnessesary")
	  Ptr = Save;
        }
        else
	  PRO_RETURN ("NOT FOUND", -4)
      }
      PRO1 ("found '%s'\n", String)
    }
    else if (Parse.Code == TEXTCODE &&
             (Parse.OBracket != SETSETBRACE || Parse.Amount > 0))
    {
      PRO1 ("Parse.TextWord '%s'\n", TextWord)
      GetString (String, Parse.Length, '\0', FALSE);
      if (strcmp (String, TextWord) != 0)
	PRO_RETURN ("NOT FOUND", -4)
      PRO1 ("found '%s'", String)
    }
    else if (Parse.Code != NOCODE && 
             (Parse.OBracket != SETSETBRACE || Parse.Amount > 0))
    {
      PRO0 ("Parse...")
      if (Parse.IBracket == NDXBRACE)
        Save = Ptr;
      switch (Parse.Code)
      {
        case REMCCODE :
          CharVal = GetChar (TRUE);
          PRO1 ("Char-Remark %d", CharVal)
          break;

        case REMSCODE : 
          GetString (String, 1024, '\0', TRUE);
          PRO1 ("String-Remark %s", String) 
          break;

        case INTCODE :
          PRO0 ("\\INT\n")
          if (Parse.IBracket == RNDBRACE || Parse.IBracket == NDXBRACE)
          {
            IntVal = GetInt (FALSE);
            if (Parse.OBracket != SETSETBRACE)
            {
              RetInt = (int*)va_arg(Arguments, int*);
              if (Parse.OBracket != NDXBRACE || found)
                if (! Error)
                  Parse.Amount = *RetInt = IntVal;
                else if (Parse.IBracket == NDXBRACE)
                {
		  Ptr = Save;
                  Parse.Amount = 0;
                }
                else
		  PRO_RETURN ("Invalid brace", -2)
              else
		Ptr = Save;
            }
            else if (Parse.Amount > 0)
            {
              memcpy ((void*)RetData, (void*)&IntVal, sizeof(int));
              RetData += sizeof (int);
            }
            PRO1 ("INT: %d\n", IntVal)
            Result++;
          }
          break;

        case HEXCODE :
          PRO0 ("\\HEX\n")
          if (Parse.IBracket == RNDBRACE || Parse.IBracket == NDXBRACE)
          {
            IntVal = GetHex (FALSE);
            if (Parse.OBracket != SETSETBRACE)
            {
              RetInt = (int*)va_arg(Arguments, int*);
              if (Parse.OBracket != NDXBRACE || found)
                if (! Error)
                  Parse.Amount = *RetInt = IntVal;
                else if (Parse.IBracket == NDXBRACE)
                {
		  Ptr = Save;
                  Parse.Amount = 0;
                }
                else
		  PRO_RETURN ("Invalid brace", -2)
              else
		Ptr = Save;
            }
            else if (Parse.Amount > 0)
            {
              memcpy ((void*)RetData, (void*)&IntVal, sizeof(int));
              RetData += sizeof (int);
            }
            PRO1 ("INT: %d\n", IntVal)
            Result++;
          }
          break;

        case FLOATCODE :
          PRO0 ("\\DOUBLE\n")
          if (Parse.IBracket == RNDBRACE || Parse.IBracket == NDXBRACE)
          {
            DoubleVal = GetDouble (FALSE);
            if (Parse.OBracket != SETSETBRACE)
            {
              RetDouble = (double*)va_arg(Arguments, double*);
              if (Parse.OBracket != NDXBRACE || found)
                if (! Error)
                  *RetDouble = DoubleVal;
                else if (Parse.IBracket == NDXBRACE)
		  Ptr = Save;
                else
		  PRO_RETURN ("Invalid brace", -2)
              else
		Ptr = Save;
            }
            else if (Parse.Amount > 0)
            {
              memcpy ((void*)RetData, (void*)&DoubleVal, sizeof(double));
              RetData += sizeof (double);
            }
            PRO1 ("DOUBLE: %f\n", DoubleVal)
	    Result++;
          }
          break;

        case CHARCODE :
          PRO0 ("\\CHAR\n")
          if (Parse.IBracket == RNDBRACE || Parse.IBracket==NDXBRACE)
          {
            CharVal = GetChar (FALSE);
            if (Parse.OBracket != SETSETBRACE)
            {
              RetChar = (char*)va_arg(Arguments, char*);
              if (Parse.OBracket != NDXBRACE || found)
                if (ok)
                  *RetChar = CharVal;
                else if (Parse.IBracket == NDXBRACE)
		  Ptr = Save;
                else
		  PRO_RETURN ("Invalid brace", -2)
              else
		Ptr = Save;
            }
            else if (Parse.Amount > 0)
            {
              memcpy ((void*)RetData, (void*)&CharVal, sizeof(char));
              RetData += sizeof (char);
            }
	    PRO1 ("CHAR: <%c>\n", CharVal)
	    Result++;
          }
          break;

        case STRINGCODE :
          PRO0 ("\\STRING\n")
          if (Parse.IBracket == RNDBRACE || Parse.IBracket == NDXBRACE)
          {
            TextVal = (char*)malloc (sizeof(char) * Parse.Length);
	    ok = (GetString(TextVal, Parse.Length, Parse.Stop, TRUE) != NULL);
            if (Parse.OBracket != SETSETBRACE)
            {
              RetText = (char*)va_arg(Arguments, char*);
              if (Parse.OBracket != NDXBRACE || found)
	      {
                if (ok || Parse.IBracket == NDXBRACE)
                  strcpy (RetText, TextVal);
                else
                {
                  free (TextVal);
		  PRO_RETURN ("Invalid index-brace", -2)
                }
	      }
            }
            else if (Parse.Amount > 0)
            {
              memcpy ((void*)RetData, (void*)TextVal, 
                      sizeof(char)*Parse.Length);
              RetData += sizeof (char) * Parse.Length;
            }
            PRO1 ("TEXT: <%s>\n", TextVal)
            free (TextVal);
            Result++;
          }
          break;

        case LINECODE :
          PRO0 ("\\LINE\n")
          if (Parse.IBracket == RNDBRACE || Parse.IBracket == NDXBRACE)
          {
            TextVal = (char*)malloc (sizeof(char) * Parse.Length);
            GetLine (TextVal, Parse.Length);
            if (Parse.OBracket != SETSETBRACE)
            {
              RetLine = (char*)va_arg(Arguments, char*);
              if (Parse.OBracket != NDXBRACE || found)
	      {
                if (ok || Parse.IBracket == NDXBRACE)
                  strcpy (RetLine, TextVal);
                else
                {
                  free (TextVal);
		  PRO_RETURN ("Invalid index-braces", -2)
		}
	      }
            }
            else if (Parse.Amount > 0)
            {
              memcpy ((void*)RetData, (void*)TextVal,
		      sizeof(char) * Parse.Length);
              RetData += sizeof(char) * Parse.Length;
            }
          }
          free (TextVal);
          PRO1 ("LINE: <%s>\n", TextVal)
          Result++;
          break;

        case IDCODE :
          PRO0 ("\\ID\n")
          IDVal = 0;
          count = 0;
          do
          {
            if (!(EOL= ! GetString(String, Parse.Length, '\0', TRUE)))
            {
              ok = FALSE;
              count = 0;
              while (count < Parse.Number && ! ok)
                if ((ok = (strcmp(String, Parse.IDWord[count]) == 0)))
                  IDVal += 1 << count;
                else
                  count++;
              if (! ok)
		PRO_RETURN ("Invalid Keyword", -5)
            }
          }
          while (! EOL && Parse.IBracket == SETBRACE && Parse.Number > 1);

          if (Parse.Number == 2 && Parse.IBracket != SETBRACE)
          {
            RetDual = (bool *)va_arg (Arguments, bool *);
            if (found)
              *RetDual = (bool)(IDVal-1);
            PRO1 ("ID: %d\n", *RetDual)
            Result++;
          }
          else
          {
            RetID = (int *)va_arg (Arguments, int *);
            if (found)
	    {
              if (Parse.IBracket == RNDBRACE)
                *RetID = count;
              else
                *RetID = IDVal;
	      PRO1 ("ID: %d\n", IDVal)
	      Result++;
	    }
          }
          break;

        default : 
	  break;
      }
    }
  }

  PRO_END
  return Result;
}


bool DATAFILE::IsSpace (char Letter)
{
  return (isspace(Letter) && Letter != '' && Letter != '');
}


char DATAFILE::SkipSpace (bool EOL)
{
  char Letter;

  do
    if ((Letter = GetChar()) == '\\')
      if (*Ptr == '\\')
      {
        do
          Letter = GetChar ();
        while (Letter != '\n');
        Letter = GetChar ();
      }
  while (IsSpace(Letter) && (Letter != '\n' || !EOL) && Letter != '\0');

  return Letter;
}


bool DATAFILE::CheckData ()
{
  int   Result;
  char *Save;

  if ((Save = Ptr) == NULL)
    return FALSE;

  while(IsSpace(*Ptr) && *Ptr != '\0')
    Ptr++;
  
  Result = (*Ptr != '\0');

  Ptr = Save;
  return Result;
}


char DATAFILE::GetChar (bool Skip /* = FALSE */)
{
  char Letter;

  //  PRO_METHOD ("DATAFILE", "")

  if (Ptr == NULL)
    PRO_RETURN ("No data found", '\0')

  if (Skip)
    Letter = SkipSpace (FALSE);
  elsif ((Letter = *Ptr) != '\0')
    Ptr++;

  //  PRO_END
  return Letter;
}


int DATAFILE::GetInt (bool Skip)
{
  int  count = 0, Result = 0;
  char Letter;
  char Value[20];

  PRO_METHOD ("DATAFILE", "")

  if (Ptr == NULL)
    PRO_RETURN ("No data found", 0)

  Error = FALSE;

  if (Skip)
    do
      Letter = GetChar ();
    while (!isdigit(Letter) && Letter != '-' && Letter != '+');
  else   
    Letter = SkipSpace (FALSE);

  if (Letter == '\0' ||
      (!isdigit(Letter) && Letter != '-' && Letter != '+'))
  {
    Error = TRUE;
    PRO_RETURN ("Invalid charater found", 0)
  }

  while ((isdigit(Letter) || Letter == '-' || Letter == '+'
          || Letter == 'E' || Letter == 'e') && count < 19)
  {
    Value[count++] = Letter;
    Letter = GetChar ();
  }

  Ptr--;
  Value[count] = '\0';
  Result = atoi (Value);

  PRO_END
  return Result;
}


unsigned long DATAFILE::GetHex (bool Skip)
{
  int           count = 0;
  unsigned long Result = 0;
  char          Letter;
  char          Value[20];

  PRO_METHOD ("DATAFILE", "")

  if (Ptr == NULL)
    PRO_RETURN ("No data found", 0)

  Error = FALSE;

  if (Skip)
    do
      Letter = GetChar ();
    while (Letter != '0');
  else   
    Letter = SkipSpace (FALSE);

  if (Letter == '\0' || Letter != '0')
  {
    Error = TRUE;
    PRO_RETURN ("Invalid charater found", 0)
  }

  while ((isxdigit(Letter) || Letter == 'x') && count < 19)
  {
    Value[count++] = Letter;
    Letter = GetChar ();
  }

  Ptr--;
  Value[count] = '\0';
  Result = strtol (Value, NULL, 16);

  PRO_END
  return Result;
}


double DATAFILE::GetDouble (bool Skip)
{
  int    Dot = 0, Exp = 0, Sgn = 0;
  int    count = 0;
  char   Letter;
  char   Value[20];
  double Result = 0.0;

  PRO_METHOD ("DATAFILE", "\n")

  Error = FALSE;

  if (Ptr == NULL)
    PRO_RETURN ("No data found", 0.0)

  if (Skip)
    do
      Letter = GetChar ();
    while (!isdigit(Letter) && Letter != '\0' && 
           Letter != '-' && Letter != '+' && Letter != '.');
  else   
    Letter = SkipSpace (FALSE);

  if (Letter == '\0' ||
      (!isdigit(Letter) && Letter != '-' && Letter != '+' && Letter != '.'))
  {
    Error = TRUE;
    PRO_RETURN ("Invalid character found", 0.0)
  }

  // this is not completly correct e.g. while there can be more than one 
  // exponent sign, but this is not to dangerous
  while ((isdigit(Letter) || 
	  (Dot < 1 && Letter == '.') || 
	  (Sgn < 2 && (Letter == '-' || Letter == '+')) ||
	  (Exp < 1 && (Letter == 'E' || Letter == 'e'))) &&
	 Letter != '\0' && count < 19)
  {
    if (Letter == '.')
      Dot++;
    if (Letter == '-' || Letter == Exp)
      Sgn++;
    if (Letter == 'E' || Letter == 'e')
      Exp++;

    Value[count++] = Letter;
    Letter = GetChar ();
  }
  Ptr--;
  Value[count] = '\0';
  Result = atof (Value);

  PRO_END
  return Result;
}


char *DATAFILE::GetString (char *Result, int MaxLen, char StopChar, bool EOL)
{
  int  count = 0;
  char Letter;  

  PRO_METHOD ("DATAFILE", "\n")

  if (Ptr == NULL)
    PRO_RETURN ("No data found", NULL)
  if (Result == NULL || MaxLen == 0)
    PRO_RETURN ("Invalid buffer", NULL)

  Error = FALSE;

  Result[0] = '\0';
  
  if (((Letter=SkipSpace(EOL)) == '\n') && EOL)
  {
    Error = TRUE;
    PRO_RETURN ("Nothing found", NULL)
  }

  while (count < MaxLen && Letter != '\0' && Letter != StopChar && 
         Letter != ' ' && Letter != '\t' && (Letter != '\n' || ! EOL))
  {
    Result[count++] = Letter;
    Letter = GetChar ();
  }
  Ptr--;
  Result[count] = '\0';

  PRO_END
  return Result;
}


char *DATAFILE::GetLine (char *Result, int MaxLen)
{
  int  count = 0;
  char Letter;

  PRO_METHOD ("DATAFILE", "\n")

  if (Ptr == NULL)
    PRO_RETURN ("No data found", NULL)
  if (Result == NULL || MaxLen == 0)
    PRO_RETURN ("invalid buffer", NULL)

  Error = FALSE;

  Result[0] = '\0';

  if ((Letter = SkipSpace(TRUE)) == '\n')
  {
    Error = TRUE;
    PRO_RETURN ("Empty Line found", NULL)
  }

  Result[count++] = Letter;
  Letter = GetChar ();
  while (count < MaxLen-1 && Letter != '\n' && Letter != '\0')
  {
    if (Letter == '\\')
    {
      if (*Ptr == '\\')
        do
          Letter = GetChar ();
        while (Letter != '\n' && Letter != '\0');
      Result[count++] = '\\';
    }
    Result[count++] = Letter;
    
    Letter = GetChar ();
  }
  Ptr--;
  Result[count] = '\0';
  
  PRO_END
  return Result;
}



