#include <libtools.h>
#include <libprg.h>

#include "libcad3d.h"


PLOT3D::PLOT3D (const char *Name, 
		int SizeX /* = 250 */, int SizeY /* = 250 */) :
  WORLD  (Name, SizeX, SizeY),
  Rows   (0),
  Cols   (0),
  Values (NULL)
{
  PRO_METHOD ("PLOT3D", "\n")

  PRO_END
}


PLOT3D::~PLOT3D ()
{
  printf ("Delete PLOT3D\n");
  DeleteValues ();
}


void PLOT3D::Draw (VIEW *View, GLenum Mode /* = GL_RENDER */)
{
#if 0
  int      countX, countY;
  double   MinZ, MaxZ;
  MATERIAL Material (COLOR::GOLD);

  Win->GetView(0)->Set ();

  Material.Set ();

  MinZ = MaxZ = Values[0][0].z;
  for (countY = 0; countY < Rows; ++countY)
    for (countX = 0; countX < Cols; ++countX)
    {
      if (Values[countX][countY].z < MinZ)
	MinZ = Values[countX][countY].z;
      if (Values[countX][countY].z > MaxZ)
	MaxZ = Values[countX][countY].z;
    }

  for (countY = 0; countY < Rows-1; ++countY)
    for (countX = 0; countX < Cols-1; ++countX)
    {
      glBegin (GL_TRIANGLE_FAN);
        glNormal3d (0.0, 0.0, 1.0);
        glVertex3d (Values[countX  ][countY  ].x, 
		    Values[countX  ][countY  ].y, 
		    Values[countX  ][countY  ].z); 
	glVertex3d (Values[countX+1][countY  ].x-0.1, 
		    Values[countX+1][countY  ].y, 
		    Values[countX+1][countY  ].z); 
	glVertex3d (Values[countX+1][countY+1].x-0.1, 
		    Values[countX+1][countY+1].y-0.1, 
		    Values[countX+1][countY+1].z); 
	glVertex3d (Values[countX  ][countY+1].x, 
		    Values[countX  ][countY+1].y-0.1, 
		    Values[countX  ][countY+1].z); 
      glEnd ();
    }

  Win->Flush ();

#else
  WORLD::Draw (View, Mode);
#endif
}


void PLOT3D::Redraw (int ViewNo /* = -1 */)
{
  VIEW *View;

  if ((View = Win->GetView(0)) != NULL)
  {
    View->Set ();
    Draw (View);
    Win->Flush ();
  }
}


void PLOT3D::GetSize (double *MinX, double *MaxX,
		      double *MinY, double *MaxY,
		      double *MinZ, double *MaxZ)
{
  int countX, countY;

  *MinX = *MaxX = 0.0;
  *MinY = *MaxY = 0.0;
  *MinZ = *MaxZ = 0.0;

  if (Rows > 0 && Cols > 0) 
  {
    *MinX = *MaxX = Values[0][0].x;
    *MinY = *MaxY = Values[0][0].y;
    *MinZ = *MaxZ = Values[0][0].z;

    for (countY = 0; countY < Rows; ++countY)
      for (countX = 0; countX < Cols; ++countX)
      {
	if (Values[countX][countY].x < *MinX)
	  *MinX = Values[countX][countY].x;
	if (Values[countX][countY].x > *MaxX)
	  *MaxX = Values[countX][countY].x;
	if (Values[countX][countY].y < *MinY)
	  *MinY = Values[countX][countY].y;
	if (Values[countX][countY].y > *MaxY)
	  *MaxY = Values[countX][countY].y;
	if (Values[countX][countY].z < *MinZ)
	  *MinZ = Values[countX][countY].z;
	if (Values[countX][countY].z > *MaxZ)
	  *MaxZ = Values[countX][countY].z;
      }
  }
}

#if 0
static PART *PLOT3D_CreateCell (const char *Name;
			        POINT3D p1, POINT3D p2, POINT3D p3, POINT3D p4)
{
  int       count;
  VECTOR    Points[8];
  POLYGON  *Polygon[6];
  POLYEDER *Polyeder;
  OBJECT   *Obj;
  PART     *Part;

  Part = PARTNewPart ();

  Part->ObjectNo = 1;
  if (Name != NULL)
  {
    Part->PartName = (char *)malloc (sizeof(char) * (strlen(Name)+1));
    strcpy (Part->PartName, Name);
  }

  Points[0] = GEOMakeVector (p1.x, p1.y,  0.0);
  Points[1] = GEOMakeVector (p2.x, p2.y,  0.0);
  Points[2] = GEOMakeVector (p3.x, p3.y,  0.0);
  Points[3] = GEOMakeVector (p4.x, p4.y,  0.0);
  Points[4] = GEOMakeVector (p1.x, p1.y, p1.z);
  Points[5] = GEOMakeVector (p2.x, p2.y, p2.z);
  Points[6] = GEOMakeVector (p3.x, p3.y, p3.z);
  Points[7] = GEOMakeVector (p4.x, p4.y, p4.z);

  Obj = PARTNewObject ();
  Obj->ColorIndex = - COLOR::GOLD;
  Obj->VertexNo   = 8;
  Obj->VertexList = (VERTEX *)malloc (sizeof(VERTEX) * Obj->VertexNo);
  for (count = 0; count < 8; ++count)
  {
    Obj->VertexList[count].Pos = Points[count];
    Obj->VertexList[count].Normal = Vector_0;
  }

  Obj->Type = PART_PLANE;
  Polyeder = GEONewPolyeder ();
  Polyeder->Number = 6;

  Polygon[0] = 

  for (count = 0; count < 6; ++count)
  {
    Polygon = PART_VaMakePolygon (count, Obj->VertexList, 4,
				  PrismData[count][0], PrismData[count][1], 
				  PrismData[count][2], PrismData[count][3]);
    if (Polyeder->List == NULL)
      Polyeder->List = Polygon;
    else
      Polyeder->Last->Next = Polygon;
    Polyeder->Last = Polygon;
  }
  Obj->Data.Polyeder = Polyeder;
  Obj->Info = PARTCalcObjectInfo (Obj);

  Part->ObjectList = Obj;
  PARTCalcPartInfo (Part);

  return Part;
}
#endif


void PLOT3D::SetValues (int Rows, int Cols, POINT3D **Points)
{
  int       countX, countY;
  double    Avg;
  double    MinZ, MaxZ;
  double    Component;
  MATERIAL  Material (COLOR::GOLD);
  POINT3D  *VertexList;
  OBJECT   *Obj;
  PART3D   *Part;
  DEVICE3D *Dev;

  PRO_METHOD ("PLOT3D", "")

  DeleteValues ();

  if (Rows == 0 || Cols == 0 || Points == NULL)
    PRO_ABORT ("No data")

  this->Rows = Rows;
  this->Cols = Cols;

  MinZ = MaxZ = Points[0][0].z;

  Floor->SetMode (FLOOR::OFF);

  // Initialize Lights:
  Values = new POINT3D *[Cols];
  for (countX = 0; countX < Cols; ++countX)
  {
    Values[countX] = new POINT3D [Rows];
    for (countY = 0; countY < Rows; ++countY)
    {
      Values[countX][countY] = Points[countX][countY];
      if (Points[countX][countY].z < MinZ)
	MinZ = Points[countX][countY].z;
      if (Points[countX][countY].z > MaxZ)
	MaxZ = Points[countX][countY].z;
    }
  }

  Dev = new DEVICE3D ("plot", NULL, this);
  AddDevice (Dev);

  MinZ -= 1.0;

  for (countY = 0; countY < Rows-1; ++countY)
    for (countX = 0; countX < Cols-1; ++countX)
    {
      Avg = 0.25 * (  Values[countX  ][countY  ].z
		    + Values[countX+1][countY  ].z
		    + Values[countX+1][countY+1].z
		    + Values[countX  ][countY+1].z);
      Component = (Avg-MinZ) / (MaxZ-MinZ);
      Material.SetColor (Component, Component, Component);
      Obj = OBJCreateQuader (GEOMakeTrans(Values[countX][countY].x, 
					  Values[countX][countY].y, 0.0),
			     Values[countX+1][countY].x -
			       Values[countX][countY].x, 
			     Values[countX][countY+1].y - 
	                       Values[countX][countY].y,
			     Avg, Material.GetColorID());
      VertexList = Obj->VertexList;

      VertexList[4].z = Values[countX  ][countY  ].z - MinZ;
      VertexList[5].z = Values[countX+1][countY  ].z - MinZ;
      VertexList[6].z = Values[countX+1][countY+1].z - MinZ;
      VertexList[7].z = Values[countX  ][countY+1].z - MinZ;

#if 1
      // Activate for a small thickness 
      VertexList[0].z = VertexList[4].z - 0.5;
      VertexList[1].z = VertexList[5].z - 0.5;
      VertexList[2].z = VertexList[6].z - 0.5;
      VertexList[3].z = VertexList[7].z - 0.5;
#endif

      Obj->Info = OBJCalcInfo (Obj);
      Part = new PART3D (PARTCreate("test", Obj), this, Material);
      Dev->AddPart (Part);
    }

  Fill ();

  PRO_END
}
  

void PLOT3D::DeleteValues ()
{
  int count;

  if (Values != NULL)
  {
    for (count = 0; count < Cols; ++count)
      delete Values[count];
    delete Values;
  }

  Values = NULL;
  Rows   = 0;
  Cols   = 0;
}
