#include <assert.h>
#include <math.h>

#include <X11/cursorfont.h>
#include <X11/keysym.h>

#include <libtools.h>

#include "libcad3d.h"


static void  WORKCELL_PartListDev  (DEVICE3D  *DevList, 
				    PART3D   **PartList, int &No);


WORKCELL::WORKCELL (char *Name /* = NULL */,
		    int SizeX  /* = 250  */, int SizeY /* = 250 */) :
  WORLD (Name, SizeX, SizeY),
  Axes  (50.0, OpenGL->GetFontBase())
{
  PRO_METHOD ("WORKCELL", "\n")

  PRO_END
}


WORKCELL::WORKCELL (Widget Parent, const char *Name /* = NULL */) :
  WORLD (Parent, Name),
  Axes  (50.0, OpenGL->GetFontBase())
{
  PRO_METHOD ("WORKCELL", "\n")

  PRO_END
}


WORKCELL::~WORKCELL ()
{
}


void WORKCELL::Draw (VIEW *View, GLenum Mode /* = GL_RENDER */)
{
  // * TEST * TEST * TEST *
#if 0
  COORDSYSTEM *CS = new COORDSYSTEM (500.0, XDisplay, "fixed");
  CS->Draw ();
  delete (CS);
#endif

  WORLD::Draw (View, Mode);

#if 0
   delete CS;
#endif
}


void WORKCELL::DrawFrame (H_MATRIX Matrix)
{
  int   count;
  VIEW *View;

  for (count = 0; count < 4; ++count)
  {
    View = Win->GetView (count);
    if (View != NULL)
      Axes.Draw ();
  }

  Win->Flush ();
}


// =======================================================================

static void GetWorldSize (DEVICE3D *DevList, double *MinX, double *MaxX, 
			                     double *MinY, double *MaxY,
	                                     double *MinZ, double *MaxZ)
{
  PART     *Part;
  PART3D   *Ptr, *Next;
  DEVICE3D *Dev;

  Dev = DevList;

  while (Dev != NULL)
  {
    Ptr = Dev->GetPartList ();
    while (Ptr != NULL)
    {
      Next = Ptr;
      while (Next != NULL)
      {
	Part = Next->GetPart ();
	*MinX = MIN (*MinX, Part->WTP.m[P][X] - Part->Info.Bubble);
	*MaxX = MAX (*MaxX, Part->WTP.m[P][X] + Part->Info.Bubble);
	*MinY = MIN (*MinY, Part->WTP.m[P][Y] - Part->Info.Bubble);
	*MaxY = MAX (*MaxY, Part->WTP.m[P][Y] + Part->Info.Bubble);
	*MinZ = MIN (*MinZ, Part->WTP.m[P][Z] - Part->Info.Bubble);
	*MaxZ = MAX (*MaxZ, Part->WTP.m[P][Z] + Part->Info.Bubble);
	Next = Next->GetNext ();
      }

      if (Ptr->GetAttach() != NULL)
	GetWorldSize (Ptr->GetAttach(), MinX, MaxX, MinY, MaxY, MinZ, MaxZ);

      Ptr = Ptr->GetChain ();
    }
    Dev = Dev->GetNext ();
  }
}


PART3D *WORKCELL::GetMount (char *DevName)
{
  DEVICE3D *Dev;

  if ((Dev = GetDev(DevName)) == NULL)
    return NULL;

  return Dev->GetMount ();
}


static void WORKCELL_GetDevList (DEVICE3D *Data, DEVICE3D **DevList, int &No)
{
  PART3D   *Part;
  DEVICE3D *Dev;

  if ((Dev = Data) == NULL)
    return;

  while (Dev != NULL)
  {
    DevList[No++] = Dev;

    Part = Dev->GetPartList ();
    while (Part != NULL)
    {
      if (Part->GetAttach() != NULL)
	WORKCELL_GetDevList (Part->GetAttach(), DevList, No);

      Part = Part->GetChain ();
    }  
    Dev = Dev->GetNext ();
  }
}


int WORKCELL::GetDevList (DEVICE3D ***List /* = NULL */)
{
  int No = 0;

  if (DevList != NULL)
  {
    *List = new DEVICE3D *[GetDevNum()];
    WORKCELL_GetDevList (DevList, *List, No);
  }

  return GetDevNum ();
}


static void WORKCELL_PartListPart (PART3D *List, PART3D **PartList, int &No)
{
  PART3D   *Ptr, *Next;
  DEVICE3D *Attach;

  Next = List;
  while (Next != NULL)
  {
    PartList[No++] = Next;
    if ((Ptr = Next->GetChain()) != NULL)
      WORKCELL_PartListPart (Ptr, PartList, No);
    if ((Attach = Next->GetAttach()) != NULL)
      WORKCELL_PartListDev (Attach, PartList, No);
    Next = Next->GetNext ();
  }
}

static void WORKCELL_PartListDev (DEVICE3D *DevList, PART3D **PartList, 
				  int &No)
{
  DEVICE3D *Dev;

  Dev = DevList;
  while (Dev != NULL)
  {
    WORKCELL_PartListPart (Dev->GetPartList(), PartList, No);
    Dev = Dev->GetNext ();
  }
}


int WORKCELL::GetPartList (PART3D ***List /* = NULL */)
{
  int No = 0;

  if (List != NULL)
  {
    *List = new PART3D *[GetPartNum()];

    WORKCELL_PartListDev (DevList, *List, No);
  }
  return No;
}


bool WORKCELL::KeyHandle (KeySym Key, int Mask)
{
  bool processed;

  if (Mask < 0)                                   // Key release: not used
    return WORLD::KeyHandle (Key, Mask);

  switch (Key)
  {
#if 0
    case XK_Escape :
      PRO0 ("Escape\n")
      SetAction (NORMAL_MODE);
      Redraw ();
      Win->SetMouseCursor (ToolNORMAL);      
      processed = TRUE;
      break;
#endif

    case XK_Delete :
      PRO0 ("Delete\n")
      processed = TRUE;
      break;

    case XK_q :
      if (Mask & ControlMask)
	PRO0 ("Ctrl-Q pressed\n")
      processed = TRUE;
      break;

    case XK_1 :
      PRO0 ("Split Mode 1\n")
      SetViews (1);
      processed = TRUE;
      break;

    case XK_2 :
      PRO0 ("Split Mode 2\n")
      SetViews (2 | 4);
      processed = TRUE;
      break;

    case XK_4 :
      PRO0 ("Split Mode 4\n")
      SetViews (8 | 16 | 32 | 64);
      processed = TRUE;
      break;

    case XK_5 :
      PRO0 ("Split Mode 5\n");
      SetViews (128 | 256 | 512 | 1024);
      processed = TRUE;
      break;

    default :
      PRO2 ("Key: %ld [%d]\n", Key, Mask)
      processed = WORLD::KeyHandle (Key, Mask);
      break;
  }

  return processed;
}
