#ifndef _DEBUG_H
#define _DEBUG_H


#include <stdio.h>
#include <stdlib.h>

#ifdef __cplusplus
extern "C" {
#endif

/*
   Decide where to write protocol-informations, if no protocol-file was 
   opened. Can be be changed to suits your taste (maybe stdout)
*/
#define STREAM stderr

#define DEB_OVERWRITE     0
/*
#define DEB_OTO_OVERWRITE  1
#define DEB_ASK_OVERWRITE 2
*/
#define DEB_APPEND        3


#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef elsif
#define elsif else if
#endif

/* Type definitions: */
#ifndef byte
typedef unsigned char  byte;
#endif

#ifndef word
typedef unsigned short word;
#endif

#ifndef bitmask8
typedef unsigned char bitmask8;
#endif

#ifndef bitmask16
typedef unsigned short bitmask16;
#endif

#ifndef bitmask32
typedef unsigned int bitmask32;
#endif


/*
  First, I called this datatype 'bool', but in version 2.6.0 of g++ this
  type was defined for for use in c++. So I renamed it into 'dual' !
*/
#ifndef dual
typedef unsigned int   dual;
#endif

#ifndef card
typedef unsigned int   card;
#endif

#ifndef boolean
typedef int boolean;
#endif

#ifndef PI
#define PI 3.14159265358979323846
#endif

#ifndef MIN
#define MIN(x, y) (((x) < (y))? (x) : (y))
#endif

#ifndef MAX
#define MAX(x, y) (((x) > (y))? (x) : (y))
#endif

#ifndef SIGN
#define SIGN(x) (((x)<0)? (-1):((x)>0)? 1:0)
#endif

#ifndef ABS_MOD
#define ABS_MOD(x,m) (((x)>=0 && (x)<m)? (x):((x)%m))
#endif

#ifndef SQR
#define SQR(x) ((x) * (x))
#endif


#define LIMIT(Min,Val,Max) (Val<Min)? (Min):((Val>Max)? (Max):(Val))

void   DEBOpenProtocol    (const char *FileName, int Type);
void   DEBWriteProtocol   (const char *Line, ...);
void   DEBCloseProtocol   (void);


#ifdef _PROTOCOL

#include <mem.h>            /* Do we use the momory-library by default ? */

#define PRO_NO_LF '\\'

#define DEB_FUNCTION    "["__FUNCTION__"]..."
#define DEB_FUNCTION_CR "["__FUNCTION__"]\n"

#define PRO_POS \
        { \
          DEBWriteProtocol ("%s,%4d: ", __FILE__, __LINE__); \
          for (deb_count = 0; \
	       deb_count < 15-(short)strlen(__FILE__); \
	       deb_count++) \
            DEBWriteProtocol (" "); \
          for (deb_count = 0; deb_count < deb_col; deb_count++) \
            DEBWriteProtocol ("  "); \
        }

#define PRO_MSG(fmt,Msg) \
        { \
          deb_col = MAX (deb_col-1, 0); \
          if (deb_cr) \
            PRO_POS; \
          DEBWriteProtocol (fmt, Msg); \
          deb_cr = TRUE; \
        }
#define PRO_MSG1(fmt,Val) \
        { \
          deb_col = MAX (deb_col-1, 0); \
          if (deb_cr) \
            PRO_POS; \
          DEBWriteProtocol (fmt, Val); \
	  DEBWriteProtocol ("\n"); \
          deb_cr = TRUE; \
        }

#define PRO0(s) \
        { \
          if ((s)[0] != PRO_NO_LF) \
            PRO_POS; \
          DEBWriteProtocol ((s)+((s)[0]==PRO_NO_LF)); \
        }
#define PRO1(s,p1) \
        { \
          if (s[0] != PRO_NO_LF) \
            PRO_POS; \
          DEBWriteProtocol (s+(s[0]==PRO_NO_LF), p1); \
        }
#define PRO2(s,p1,p2) \
        { \
	  if (s[0] != PRO_NO_LF) \
	    PRO_POS; \
          DEBWriteProtocol (s+(s[0]==PRO_NO_LF), p1, p2); \
        }
#define PRO3(s,p1,p2,p3) \
        { \
          if (s[0] != PRO_NO_LF) \
            PRO_POS; \
          DEBWriteProtocol (s+(s[0]==PRO_NO_LF), p1, p2, p3); \
        }
#define PRO4(s,p1,p2,p3,p4) \
        { \
          if (s[0] != PRO_NO_LF) \
            PRO_POS; \
          DEBWriteProtocol (s+(s[0]==PRO_NO_LF), p1, p2, p3, p4); \
        }
#define PRO5(s,p1,p2,p3,p4,p5) \
        { \
          if (s[0] != PRO_NO_LF) \
            PRO_POS; \
          DEBWriteProtocol (s+(s[0]==PRO_NO_LF), p1, p2, p3, p4, p5); \
        }
#define PRO6(s,p1,p2,p3,p4,p5,p6) \
        { \
          if (s[0] != PRO_NO_LF) \
            PRO_POS; \
          DEBWriteProtocol (s+(s[0]==PRO_NO_LF), p1, p2, p3, p4, p5, p6); \
        }

#define PRO_START(Msg) \
        { \
          PRO_POS \
	  if (Msg == NULL || Msg[0] == '\0' || Msg[0] == ' ') \
	    DEBWriteProtocol ("[%s]...", __FUNCTION__); \
	  else if (Msg[0] == '\n') \
	    DEBWriteProtocol ("[%s] %s", __FUNCTION__, Msg); \
	  else \
            DEBWriteProtocol (Msg); \
          deb_col++; \
          deb_cr = (Msg[MAX(0,(short)strlen(Msg)-1)] == '\n')? 1:0; \
        }
#define PRO_METHOD(Class,Msg) \
        { \
          PRO_POS \
	  if ((Msg) == NULL || (Msg)[0] == '\0' || (Msg)[0] == ' ') \
	    DEBWriteProtocol ("[%s.%s]...", Class, __FUNCTION__); \
          else if ((Msg)[0] == '\n') \
            DEBWriteProtocol ("[%s.%s] %s", Class, __FUNCTION__, Msg); \
          else \
            DEBWriteProtocol (Msg); \
          deb_col++; \
          deb_cr = (Msg[MAX(0,(short)strlen(Msg)-1)] == '\n')? 1:0; \
        }
#define PRO_ABORT(Msg) \
        { \
          PRO_MSG ("ABORT: %s\n", Msg); \
          return; \
        }
#define PRO_ABORT1(Msg,Val) \
        { \
          PRO_MSG1 ("ABORT: " Msg, Val); \
          return; \
        }
#define PRO_RETURN(Msg,Result) \
        { \
          PRO_MSG ("RETURN: %s\n", Msg); \
          return Result; \
        }
#define PRO_RETURN1(Msg,Val,Result) \
        { \
          PRO_MSG1 ("RETURN: " Msg, Val); \
          return Result; \
        }
#define PRO_END \
        { \
          deb_col = MAX (deb_col-1, 0); \
          if (deb_cr) \
            PRO_POS; \
          PRO0 ("\\ready\n"); \
          deb_cr = TRUE; \
        }


extern          short deb_count;
extern          short deb_col;
extern unsigned char  deb_cr;

#else

#define PRO_POS ;
#define PRO_MSG(fmt,Msg) \
        { \
	  fprintf (STREAM, "\n"); \
          fprintf (STREAM, fmt, Msg); \
          fflush  (STREAM); \
        }
#define PRO_MSG1(fmt,Val) \
        { \
	  fprintf (STREAM, "\n"); \
          fprintf (STREAM, fmt, Val); \
          fflush  (STREAM); \
        }

#define PRO0(s) { /* */ }
#define PRO1(s,p1) { /* */ }
#define PRO2(s,p1,p2) { /* */ }
#define PRO3(s,p1,p2,p3) { /* */ }
#define PRO4(s,p1,p2,p3,p4) { /* */ }
#define PRO5(s,p1,p2,p3,p4,p5) { /* */ }
#define PRO6(s,p1,p2,p3,p4,p5,p6) { /* */ }

#define PRO_START(Msg) /* */
#define PRO_METHOD(Class,Msg) /* */
#define PRO_ABORT(Msg) return;
#define PRO_ABORT1(Msg,Val) return;
#define PRO_RETURN(Msg,Result) return Result;
#define PRO_RETURN1(Msg,Val,Result) return Result;
#define PRO_END /* */
#endif


#ifdef _ABORT
#define PRO_FINISH(x) exit(1)
#else
#define PRO_FINISH(x) return x;
#endif

#ifdef _ABORT
#define PRO_EXIT exit(1)
#else
#define PRO_EXIT return;
#endif

#define PRO_ERROR(Msg) \
        { \
          PRO_MSG ("ERROR: %s\n", Msg); \
          PRO_EXIT; \
        }
#define PRO_ERROR0(Msg,Result) \
        { \
          PRO_MSG ("ERROR: %s\n", Msg); \
          PRO_FINISH (Result); \
        }
#define PRO_ERROR_VAL(Msg,Val) \
        { \
          PRO_MSG1 ("ERROR: " Msg "\n", Val); \
          PRO_EXIT; \
        }
#define PRO_ERROR_VAL1(Msg,Val,Result) \
        { \
          PRO_MSG1 ("ERROR: " Msg "\n", Val); \
          PRO_FINISH (Result); \
        }

#define PRO_WARNING(Msg,Result) \
        { \
          PRO_MSG ("WARNING: %s\n", Msg); \
	  PRO_FINISH (Result); \
	}
#define PRO_WARNING1(Msg,Val,Result) \
        { \
          PRO_MSG1 ("WARNING: " Msg "\n", Val); \
	  PRO_FINISH (Result); \
	}

#ifdef __cplusplus
} /* for C++ V2.0 */
#endif

#endif
