#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <debug.h>

#include "ArrayP.h"

#ifndef MAX
#define MAX(x,y) (((x) > (y)) ? (x) : (y))
#endif


/* Prototypes: */
static void             ClassInitialize      ();
static void             ClassPartInitialize  ();
static void             Initialize           ();
static void             Resize               ();
static void             ConstraintInitialize ();
static XtGeometryResult GeometryManager      ();
static void             ChangeManaged        ();

static XtGeometryResult QueryGeometry (ArrayWidget w, 
				       XtWidgetGeometry *request,
				       XtWidgetGeometry *geom_return);
static Boolean Layout (ArrayWidget mw, Dimension width, Dimension height, 
		       Boolean force_relayout);

static MAlignmentType defXAlignment = M_ALIGN_LEFT;
static MAlignmentType defYAlignment = M_ALIGN_CENTER;



/* Resources: */
static XtResource resources[] = 
{
  { 
    MNcellDistance,   XtCThickness,    XtRInt,      sizeof(int),
    XtOffsetOf(ArrayRec, array.cell_distance), XtRImmediate,  (XtPointer)0 
  },
  { 
    MNautoFill,       XtCBoolean,      XtRBoolean,  sizeof(Boolean),
    XtOffsetOf(ArrayRec, array.auto_fill), XtRImmediate, (XtPointer)False 
  },
  { 
    MNfixedWidth,     XtCBoolean,      XtRBoolean,  sizeof(Boolean),
    XtOffsetOf(ArrayRec, array.fixed_width), XtRImmediate, (XtPointer)False 
  },
  { 
    MNfixedHeight,    XtCBoolean,      XtRBoolean,  sizeof(Boolean),
    XtOffsetOf(ArrayRec, array.fixed_height), XtRImmediate, (XtPointer)False 
  },
  { 
    XtNborderWidth,   XtCBorderWidth,  XtRInt,      sizeof(Dimension),
    XtOffsetOf(ArrayRec, core.border_width), XtRImmediate, (XtPointer)0 
  }
};

static XtResource arrayConstraintResources[] = 
{
  { 
    MNcolumn,          XtCPosition,     XtRInt,     sizeof(Position),
    XtOffsetOf(ArrayConstraintsRec, array.column), XtRImmediate,
    (XtPointer)0
  },
  { 
    MNrow,             XtCPosition,     XtRInt,     sizeof (Position),
    XtOffsetOf(ArrayConstraintsRec, array.row),    XtRImmediate,
    (XtPointer)0 
  },
  { 
    MNcolumns,         XtCWidth,        XtRInt,     sizeof(Dimension),
    XtOffsetOf(ArrayConstraintsRec, array.columns), XtRImmediate, 
    (XtPointer)1
  },
  { 
    MNrows,            XtCHeight,       XtRInt,     sizeof(Dimension),
    XtOffsetOf (ArrayConstraintsRec, array.rows),   XtRImmediate, 
    (XtPointer)1
  },
  { 
    MNallowResize,     XtCBoolean,      XtRBoolean, sizeof(Boolean),
    XtOffsetOf (ArrayConstraintsRec, array.allow_resize), XtRImmediate,
    (XtPointer)False 
  },
  { 
    MNxAlignment,      MCxAlignment,   MRAlignmentType, sizeof(MAlignmentType),
    XtOffsetOf(ArrayConstraintsRec, array.x_alignment), MRAlignmentType, 
    (XtPointer)&defXAlignment 
  },
  { 
    MNyAlignment,      MCyAlignment,   MRAlignmentType, sizeof(MAlignmentType),
    XtOffsetOf(ArrayConstraintsRec, array.y_alignment), MRAlignmentType, 
    (XtPointer)&defYAlignment 
  },
  { 
    MNpadX,            XtCWidth,       XtRInt,          sizeof(Dimension),
    XtOffsetOf(ArrayConstraintsRec, array.padx),        XtRImmediate,
    (XtPointer) 0 
  },
  { 
    MNpadY,            XtCHeight,      XtRInt,          sizeof(Dimension),
    XtOffsetOf(ArrayConstraintsRec, array.pady),        XtRImmediate,
    (XtPointer) 0 
  }
};


/*--------------------------------------------------------------------------
 * Structure of Widet class
 *--------------------------------------------------------------------------*/

ArrayClassRec arrayClassRec = 
{
  /* core_class fields */
  { 
    #ifdef USE_MOTIF
      /* superclass         */    (WidgetClass)&xmManagerClassRec,
    #else
      /* superclass         */    (WidgetClass)&constraintClassRec,
    #endif
    /* class_name         */    "Array",
    /* widget_size        */    sizeof(ArrayRec),
    /* class_initialize   */    ClassInitialize,
    /* class_part_init    */    ClassPartInitialize,
    /* class_inited       */    FALSE,
    /* initialize         */    Initialize,
    /* initialize_hook    */    NULL,
    /* realize            */    XtInheritRealize,
    /* actions            */    NULL,
    /* num_actions        */    0,
    /* resources          */    resources,
    /* num_resources      */    XtNumber(resources),
    /* xrm_class          */    NULLQUARK,
    /* compress_motion    */    TRUE,
    /* compress_exposure  */    XtExposeCompressMaximal,
    /* compress_enterleave*/    TRUE,
    /* visible_interest   */    FALSE,
    /* destroy            */    NULL,
    /* resize             */    Resize,
    /* expose             */    XtInheritExpose,
    /* set_values         */    NULL,
    /* set_values_hook    */    NULL,
    /* set_values_almost  */    XtInheritSetValuesAlmost,
    /* get_values_hook    */    NULL,
    /* accept_focus       */    NULL,
    /* version            */    XtVersion,
    /* callback_private   */    NULL,
    /* tm_table           */    XtInheritTranslations,
    /* query_geometry     */	(XtGeometryHandler)QueryGeometry,
    /* display_accelerator*/	NULL,
    /* extension          */	NULL,
  },
  /* composite_class fields */
  { 
    /* geometry_manager   */   GeometryManager,
    /* change_managed     */   ChangeManaged,
    /* insert_child       */   XtInheritInsertChild,
    /* delete_child       */   XtInheritDeleteChild,
    /* extension          */   NULL
  },
  /* constraint_class fields */
  {
    /* subresourses       */   arrayConstraintResources,
    /* subresource_count  */   XtNumber(arrayConstraintResources),
    /* constraint_size    */   sizeof(ArrayConstraintsRec),
    /* initialize         */   ConstraintInitialize,
    /* destroy            */   NULL,
    /* set_values         */   NULL,
    /* extension          */   NULL
  },
  #ifdef USE_MOTIF
    /* Here is the XmManager class record. */ 
    {
      /* translations */                 XtInheritTranslations,
      /* syn_resources */                NULL,
      /* num_syn_resources */            0,
      /* syn_constraint_resources */     NULL,
      /* num_syn_constraint_resources */ 0,
      /* parent_process */               XmInheritParentProcess,
      /* extension */                    NULL
    },
  #endif
  /* array_class fields */
  { 
    /* layout */                         (LayoutProc)Layout,
    /* extension */                      NULL
  }
};

WidgetClass arrayWidgetClass = (WidgetClass)&arrayClassRec;

/*--------------------------------------------------------------------------
 * Ressource-Converter for Alignment
 *--------------------------------------------------------------------------*/

static XrmQuark AlignLeft, AlignRight, AlignTop, AlignBottom;
static XrmQuark AlignCenter, AlignFill;

#define Compare(quark, val) \
          if (q == quark) \
          { \
	    al = val; \
	    toVal->size = sizeof(MAlignmentType); \
	    toVal->addr = (XPointer) al; \
	    return; \
	  }


static void StringToAlignmentType (XrmValuePtr args, Cardinal *nArgs,
				   XrmValuePtr fromVal, XrmValuePtr toVal) 
{
  static MAlignmentType al;
         XrmQuark        q;

  q = XrmStringToQuark (fromVal->addr);
  
       Compare (AlignLeft,   M_ALIGN_LEFT)
  else Compare (AlignRight,  M_ALIGN_RIGHT)
  else Compare (AlignTop,    M_ALIGN_TOP)
  else Compare (AlignBottom, M_ALIGN_BOTTOM)
  else Compare (AlignCenter, M_ALIGN_CENTER)
  else Compare (AlignFill,   M_ALIGN_FILL)

  XtStringConversionWarning (fromVal->addr, "alignmentType");
  toVal->addr = NULL;
  toVal->size = 0;
}

#undef Compare


/* Widget-class init (part 1) */
static void ClassInitialize (void) 
{
  PRO_START (" ")

  AlignLeft   = XrmPermStringToQuark ("left");
  AlignRight  = XrmPermStringToQuark ("right");
  AlignTop    = XrmPermStringToQuark ("top");
  AlignBottom = XrmPermStringToQuark ("bottom");
  AlignCenter = XrmPermStringToQuark ("center");
  AlignFill   = XrmPermStringToQuark ("fill");

  XtAddConverter (XtRString, MRAlignmentType, StringToAlignmentType,
		  (XtConvertArgList)NULL, 0);

  PRO_END
}


/* Widget-class init (part 2) */
static void ClassPartInitialize (WidgetClass class) 
{
  ArrayWidgetClass m = (ArrayWidgetClass)class;

  PRO_START (" ")

  if (m->array_class.layout == XtInheritLayout)
    m->array_class.layout = (LayoutProc)Layout;

  PRO_END
}


/* Widget-instance init */
static void Initialize (Widget request, Widget new, ArgList args, 
			Cardinal num_args) 
{
  ArrayWidget mw = (ArrayWidget)new;
  
  PRO_START (" ")

  mw->array.no_refigure = False;

  PRO_END
}

/* Init constraint-resources of children */
static void ConstraintInitialize (Widget request, Widget   new, 
				  ArgList args,   Cardinal num_args) 
{
  ArrayConstraints c = (ArrayConstraints)new->core.constraints;
  
  PRO_START (" ")

  /* Check range of values */
  if (c->array.column < 0) 
    c->array.column = 0;
  else if (c->array.column >= M_ARRAY_WIDTH_MAX) 
    c->array.column = M_ARRAY_WIDTH_MAX - 1;

  if (c->array.row < 0) 
    c->array.row = 0;
  else if (c->array.row >= M_ARRAY_HEIGHT_MAX) 
    c->array.row = M_ARRAY_HEIGHT_MAX - 1;

  if (c->array.columns < 1) 
    c->array.columns = 1;
  else if (c->array.columns + c->array.column > M_ARRAY_WIDTH_MAX)
    c->array.columns = M_ARRAY_WIDTH_MAX - c->array.column;

  if (c->array.rows < 1) 
    c->array.rows = 1;
  else if (c->array.rows + c->array.row > M_ARRAY_HEIGHT_MAX)
    c->array.rows = M_ARRAY_HEIGHT_MAX - c->array.row;

  PRO_END
}

/* Calc position of widgets */
static void LayoutChild (Widget W, Position X, Position Y, 
			 Dimension Width, Dimension Height) 
{
  ArrayConstraints c       = (ArrayConstraints)W->core.constraints;
  Dimension        wWidth  = W->core.width;
  Dimension        wHeight = W->core.height;
  Dimension        bw      = W->core.border_width;
  Position         newX, newY;
  Dimension        newWidth, newHeight;

  switch (c->array.x_alignment) 
  {
    case M_ALIGN_RIGHT:
      newX = X + Width - wWidth - 2 * bw;
      newWidth = wWidth;
      break;
    
    case M_ALIGN_CENTER:
      newX = X + (Width - wWidth - 2 * bw) / 2;
      newWidth = wWidth;
      break;

    case M_ALIGN_FILL:
      newX = X;
      newWidth = Width - 2 * bw;
      break;

    case M_ALIGN_LEFT:
    default:
      newX = X;
      newWidth = wWidth;
  }
  
  switch (c->array.y_alignment) 
  {
    case M_ALIGN_BOTTOM:
      newY = Y + Height - wHeight - 2 * bw;
      newHeight = wHeight;
      break;
    
    case M_ALIGN_CENTER:
      newY = Y + (Height - wHeight - 2 * bw) / 2;
      newHeight = wHeight;
      break;
    
    case M_ALIGN_FILL:
      newY = Y;
      newHeight = Height - 2 * bw;
      break;

    case M_ALIGN_TOP:
    default:
      newY = Y;
      newHeight = wHeight;
  }
  
  XtMoveWidget (W, newX, newY);
  if ((newWidth != wWidth) || (newHeight != wHeight))
    XtResizeWidget (W, newWidth, newHeight, W->core.border_width);
}


/* Compute layout and move/resize children */
static Boolean Layout (ArrayWidget mw, Dimension width, Dimension height, 
		       Boolean force_relayout) 
{
  Boolean           ret_val;
  int               count;
  int               rows, columns;
  int               row, column;
  int               w, h, b;
  int               px, py;
  int               maxx, maxy;
  int               nChildren = mw->composite.num_children;
  int               cellDistance = mw->array.cell_distance;
  double            FactorX, FactorY;
  Dimension         ReturnX, ReturnY;
  Dimension         cellWidth[M_ARRAY_WIDTH_MAX];
  Dimension         cellHeight[M_ARRAY_HEIGHT_MAX];
  Position          cellX[M_ARRAY_WIDTH_MAX];
  Position          cellY[M_ARRAY_HEIGHT_MAX];
  WidgetList        children = mw->composite.children;
  Widget           *child;
  ArrayConstraints  c;
  XtWidgetGeometry  childGeometry;

  PRO_START (" ")
  
  /* Mark all childeren's resources as LayoutPending */
  for (child = children; child - children < nChildren; ++child) 
  {
    ArrayConstraints m = (ArrayConstraints) (*child)->core.constraints;
    m->array.layout_state = LayoutPending;
  }
  
  /* Initialize size of child-widgets */
  rows = columns = 0;
  for (count = 0; count < M_ARRAY_WIDTH_MAX; ++count)
    cellWidth[count] = cellX[count] = 0;
  for (count = 0; count < M_ARRAY_HEIGHT_MAX; ++count)
    cellHeight[count] = cellY[count] = 0;

  /* Compute size of children and set it */
  for (child = children; child - children < nChildren; ++child) 
    if (XtIsManaged (*child)) 
    {
      c = (ArrayConstraints)(*child)->core.constraints;

      row = c->array.row;
      column = c->array.column;

      if ((c->array.columns == 1) || (c->array.rows == 1)) 
      {
	childGeometry.request_mode = CWWidth | CWHeight | CWBorderWidth;
	XtQueryGeometry (*child, NULL, &childGeometry);

	if (c->array.columns == 1) 
        {
	  columns = MAX (columns, column);
	  if (childGeometry.request_mode & CWWidth)
	    w = childGeometry.width;
	  else
	    w = (*child)->core.width;

	  if (childGeometry.request_mode & CWBorderWidth)
	    b = childGeometry.border_width;
	  else
	    b = (*child)->core.border_width;

	  cellWidth[column] = MAX (cellWidth[column], 
				   w + 2 * b + 2 * c->array.padx);
	}

	if (c->array.rows == 1) 
        {
	  rows = MAX (rows, row);
	  if (childGeometry.request_mode & CWHeight)
	    h = childGeometry.height;
	  else
	    h = (*child)->core.height;

	  if (childGeometry.request_mode & CWBorderWidth)
	    b = childGeometry.border_width;
	  else
	    b = (*child)->core.border_width;

	  cellHeight[row] = MAX (cellHeight[row], 
				 h + 2 * b + 2 * c->array.pady);
	}
      }
    }

  /* shall the width/height be constant */
  if (mw->array.fixed_width) 
  {
    w = cellWidth[0];

    for (count = 1; count <= columns; ++count) 
      w = MAX (w, cellWidth[count]);
    for (count = 0; count <= columns; ++count)
      cellWidth[count] = w;
  }
  if (mw->array.fixed_height) 
  {
    h = cellHeight[0];

    for (count = 1; count <= rows; ++count) 
      h = MAX (h, cellHeight[count]);
    for (count = 0; count <= rows; ++count) 
      cellHeight[count] = h;
  }

  /* Compute start-position of cells */
  cellX[0] = 0;
  for (count = 0; count < columns; ++count)
    cellX[count+1] = cellX[count] + cellWidth[count] + cellDistance;

  cellY[0] = 0;
  for (count = 0; count < rows; ++count)
    cellY[count+1] = cellY[count] + cellHeight[count] + cellDistance;

  maxx = MAX (1, cellX[columns] + cellWidth[columns]);
  maxy = MAX (1, cellY[rows]    + cellHeight[rows]);

  /* Adjust width and height, if the array-widget shall use the complete
     available space */
  if (mw->array.auto_fill && mw->core.width && mw->core.height) 
  {
    FactorX = mw->core.width  / maxx;
    FactorY = mw->core.height / maxy;

    for (count = 0; count <= columns; ++count) 
    {
      cellWidth[count] = cellWidth[count] * FactorX;
      cellX[count]     = cellX[count]     * FactorX;
    }

    for (count = 0; count <= rows; ++count) 
    {
      cellHeight[count] = cellHeight[count] * FactorY;
      cellY[count]      = cellY[count]      * FactorY;
    }

    maxx = mw->core.width;
    maxy = mw->core.height;
  }
  
  /* place child-widgets */
  for (child = children; child - children < nChildren; ++child) 
    if (XtIsManaged (*child)) 
    {
      ArrayConstraints c = (ArrayConstraints) (*child)->core.constraints;
      row    = c->array.row;
      column = c->array.column;
      
      switch (c->array.layout_state) 
      {
        case LayoutPending:
	  {
	    w = 0;
	    h = 0;
	    px = c->array.padx;
	    py = c->array.pady;

	    for (count=0; count < c->array.columns; ++count)
	      w += cellWidth[count + column];
	    w += (c->array.columns - 1) * cellDistance;
	    if (2 * px + 1 > w) 
	      px = (w - 1) / 2;

	    for (count = 0; count < c->array.rows; ++count)
	      h += cellHeight[count + row];
	    h += (c->array.rows - 1) * cellDistance;
	    if (2 * py + 1 > h) 
	      py = (h - 1) / 2;

	    LayoutChild (*child, cellX[column] + px, cellY[row] + py, 
			 w - 2 * px, h - 2 * py);
	    c->array.layout_state = LayoutDone;
	  }
	  break;

        case LayoutDone:
	  break;

        default :
        case LayoutInProgress:
	  {
	    String subs[2];
	    Cardinal nSubs = 2;
	  
	    subs[0] = (*child)->core.name;
	    subs[1] = (*child)->core.parent->core.name;
	    XtAppWarningMsg (XtWidgetToApplicationContext (*child),
			     "constraintLoop", "arrayLayout", "ArrayError", 
			     "constraint loop detected while laying out "
			     "child '%s' in ArrayWidget '%s'",
			     subs, &nSubs);
	  }
      }
    }

  /* ask parent widget for new size */
  if ((maxx != mw->core.width) && (maxy != mw->core.height)) 
  {
    XtGeometryResult result;
    
    result = XtMakeResizeRequest ((Widget) mw, 
				  (Dimension) maxx, (Dimension) maxy,
				  &ReturnX, &ReturnY);
    maxx = ReturnX;
    maxy = ReturnY;
    if (result == XtGeometryAlmost)
      result = XtMakeResizeRequest ((Widget) mw, 
				    (Dimension) maxx, (Dimension) maxy, 
				    NULL, NULL);
    ret_val = (result == XtGeometryYes);
  } 
  else
    ret_val = False;

  PRO_END
  return ret_val;
}

/*--------------------------------------------------------------------------
 * Call procedure for layout computing
 *--------------------------------------------------------------------------*/
static void RefigureLocations (ArrayWidget W) 
{
  if (W->array.no_refigure)
    W->array.needs_relayout = True;
  else 
  {
    (*((ArrayWidgetClass) W->core.widget_class)->array_class.layout)
      (W, W->core.width, W->core.height);
    W->array.needs_relayout = False;
  }
}

/*--------------------------------------------------------------------------
 * Compute resize requests of childs
 *--------------------------------------------------------------------------*/
static XtGeometryResult GeometryManager (Widget w, XtWidgetGeometry *request, 
					 XtWidgetGeometry *reply) 
{
#if 1
  return XtGeometryNo;
#else
  ArrayConstraints array = (ArrayConstraints)w->core.constraints;
  XtWidgetGeometry allowed;

  PRO_START (" ")

  if ((request->request_mode & ~(XtCWQueryOnly | CWWidth | CWHeight)) ||
      !array->array.allow_resize)
    PRO_RETURN ("ready", XtGeometryNo)

  if (request->request_mode & CWWidth)
    allowed.width = request->width;
  else
    allowed.width = w->core.width;

  if (request->request_mode & CWHeight)
    allowed.height = request->height;
  else
    allowed.height = w->core.height;

  if ((allowed.width == w->core.width) && (allowed.height == w->core.height))
    return XtGeometryNo;

  if (!(request->request_mode & XtCWQueryOnly))
    RefigureLocations ((ArrayWidget)w->core.parent);
  
  PRO_END
  return XtGeometryYes;
#endif
}

/*--------------------------------------------------------------------------
 * Handle size request of parent-widget 
 *--------------------------------------------------------------------------*/
static XtGeometryResult QueryGeometry (ArrayWidget mw, 
				       XtWidgetGeometry *request,
				       XtWidgetGeometry *geom_return) {
  int               count;
  int               rows, columns;
  int               maxx, maxy;
  int               nChildren = mw->composite.num_children;
  int               cellDistance = mw->array.cell_distance;
  Dimension         cellWidth[M_ARRAY_WIDTH_MAX];
  Dimension         cellHeight[M_ARRAY_HEIGHT_MAX];
  Position          cellX[M_ARRAY_WIDTH_MAX];
  Position          cellY[M_ARRAY_HEIGHT_MAX];
  WidgetList        children = mw->composite.children;
  Widget           *child;
  XtWidgetGeometry  childGeometry;

  PRO_START (" ")
  
  /* Compute and set size of childs */
  rows = columns = 0;

  for (count = 0; count < M_ARRAY_WIDTH_MAX; ++count)
    cellWidth[count] = cellX[count] = 0;
  for (count = 0; count < M_ARRAY_HEIGHT_MAX; ++count)
    cellHeight[count] = cellY[count] = 0;

  for (child = children; child - children < nChildren; ++child) 
  {
    if (XtIsManaged (*child)) 
    {
      ArrayConstraints c = (ArrayConstraints) (*child)->core.constraints;
      int row = c->array.row;
      int column = c->array.column;

      if ((c->array.columns == 1) || (c->array.rows == 1)) 
      {
	childGeometry.request_mode = CWWidth | CWHeight | CWBorderWidth;
	XtQueryGeometry (*child, NULL, &childGeometry);

	if (c->array.columns == 1) 
	{
	  columns = MAX (columns, column);
	  cellWidth[column] = MAX (cellWidth[column], 
				   childGeometry.width + 
				   2 * childGeometry.border_width);
	}
	if (c->array.rows == 1) 
	{
	  rows = MAX (rows, row);
	  cellHeight[row] = MAX (cellHeight[row], 
				 childGeometry.height +
				 2 * childGeometry.border_width);
	}
      }
    }
  }  

  /* Compute start-position of cells */
  for (count = 0; count < columns; ++count)
    cellX[count+1] = cellX[count] + cellWidth[count] + cellDistance;

  for (count = 0; count < rows; ++count)
    cellY[count+1] = cellY[count] + cellHeight[count] + cellDistance;
  maxx = cellX[columns] + cellWidth[columns];
  maxy = cellY[rows] + cellHeight[rows];

  /* Insert requested fields */
  geom_return->request_mode = CWWidth | CWHeight;
  geom_return->width = maxx;
  geom_return->height = maxy;
  
  PRO_END
  return XtGeometryYes;
}

/*--------------------------------------------------------------------------
 * Handle resize of fields 
 *--------------------------------------------------------------------------*/
static void Resize (Widget w) 
{
  PRO_START (" ")

  RefigureLocations ((ArrayWidget) w);

  PRO_END
}

/*--------------------------------------------------------------------------
 * Resize managing of childs
 *--------------------------------------------------------------------------*/
static void ChangeManaged (Widget w) 
{
  PRO_START (" ")
  
  RefigureLocations ((ArrayWidget) w);

  PRO_END
}
