#ifndef ARRAYP_h
#define ARRAYP_h

#include "Array.h"

#ifdef USE_MOTIF
#include <Xm/ManagerP.h>
#else
#include <X11/ConstrainP.h>
#endif

typedef enum 
{
  LayoutPending,
  LayoutInProgress, 
  LayoutDone
} LayoutState;


#define XtInheritLayout ((Boolean(*)(ArrayWidget,Dimension,Dimension))_XtInherit)

#define MRAlignmentType     "AlignmentType"

typedef Boolean (*LayoutProc)(ArrayWidget aw, Dimension x, Dimension y); 


typedef struct 
{
  LayoutProc layout;
  caddr_t    extension;
} ArrayClassPart;

/*
 * Layout(
 *	ArrayWidget w	- the widget whose children are to be configured
 *	Dimension w, h	- bounding box of layout to be calculated
 *
 *  Stores preferred geometry in w->array.preferred_{width,height}.
 *  If w->array.resize_in_layout is True, then a geometry request
 *  may be made for the preferred bounding box if necessary.
 *
 *  Returns True if a geometry request was granted, False otherwise.
 */

typedef struct _ArrayClassRec 
{
  CoreClassPart	      core_class;
  CompositeClassPart  composite_class;
  ConstraintClassPart constraint_class;
  #ifdef USE_MOTIF
    XmManagerClassPart  manager_class;
  #endif
  ArrayClassPart      array_class;
} ArrayClassRec;

extern ArrayClassRec arrayClassRec;


typedef struct _ArrayPart 
{
  /* resources */
  int		cell_distance;      /* default distance between cells */
  Boolean       auto_fill;          /* Fill free space automatically */
  Boolean       fixed_width;        /* Cells have fixed width */
  Boolean       fixed_height;       /* Cells have fixed height */

  /* private state */
  int		no_refigure;	    /* no re-layout while > 0		 */
  Boolean	needs_relayout;	    /* next time no_refigure == 0	 */
} ArrayPart;


typedef struct _ArrayRec 
{
  CorePart	 core;
  CompositePart	 composite;
  ConstraintPart constraint;
  #ifdef USE_MOTIF
    XmManagerPart  manager;
  #endif
  ArrayPart	 array;
} ArrayRec;

typedef struct _ArrayConstraintsPart 
{
  /*
   * Constraint Resources.
   */
  Position       row,column;           /* Position in der Marrray */
  Dimension      rows, columns;        /* Groesse in Array-Zellen */
  Boolean        allow_resize;         /* Widget darf sich vergroessern */
  MAlignmentType x_alignment;          /* Positionierung in Array-Zelle */
  MAlignmentType y_alignment;
  Dimension      padx, pady;           /* Platz um das Widget */

  /*
   * Private contstraint resources.
   */
  LayoutState layout_state;            /* temporary layout state */
} ArrayConstraintsPart;


typedef struct _ArrayConstraintsRec 
{
  #ifdef USE_MOTIF
    XmManagerConstraintPart manager;
  #endif
  ArrayConstraintsPart array;
} ArrayConstraintsRec, *ArrayConstraints;


#endif
