/* 
   Icon-Wiget found in Thomas Berlage: OSF/Motif und das X-Window System
   Erweitern und angepat von S. Vietinghoff
*/

#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <Xm/Xm.h>
#include <Xm/XmP.h>
#include <Xm/Label.h>
#include <Xm/LabelP.h>
#include <Xm/DrawP.h>
#include "Icon.h"
#include "IconP.h"


/* Static routine definitions */

static void     Initialize    ();
static Boolean  SetValues     ();
static void     Redisplay     ();
static void     Destroy       ();

static void     Arm           ();
static void     Activate      ();
static void     Disarm        ();
static void     Enter         ();
static void     Leave         ();
static void     Help          ();

static void     GetPixmapSize ();
static void     DrawShadow    (IconWidget widget, Boolean really);


/* Default translation tabel and action list */

static char defaultTranslations[] = 
  "<EnterWindow>:            Enter()  \n\
   <LeaveWindow>:            Leave()";

static XtActionsRec actionsList[] =
{
  {"Arm",       (XtActionProc)Arm},
  {"Activate",  (XtActionProc)Activate},
  {"Disarm",    (XtActionProc)Disarm},
  {"Enter",     (XtActionProc)Enter},
  {"Leave",     (XtActionProc)Leave},
  {"Help",      (XtActionProc)Help}
};

static XtResource resources[] = 
{
  {
    XmNshadowThickness, XmCShadowThickness, XmRShort, sizeof(short),
    XtOffset(XmPrimitiveWidget, primitive.shadow_thickness), 
    XmRImmediate, (caddr_t)2
  },
  {
    XmNiconOffset, XmCIconOffset, XmRShort, sizeof(short),
    XtOffset(IconWidget, icon.Offset), 
    XmRImmediate, (caddr_t)0
  },
  {
    XmNshadowType, XmCShadowType, XmRShadowType, sizeof(unsigned char),
    XtOffset(IconWidget, icon.shadow_type), 
    XmRImmediate, (caddr_t)XmSHADOW_OUT
  },
  {
    XmNactivateCallback, XmCCallback, XmRCallback, sizeof(XtCallbackList),
    XtOffset(IconWidget, icon.activate_callback), 
    XmRPointer, (caddr_t)NULL
  },
  {
    XmNarmCallback, XmCCallback, XmRCallback, sizeof(XtCallbackList),
    XtOffset(IconWidget, icon.arm_callback), 
    XmRPointer, (caddr_t)NULL
  },
  {
    XmNdisarmCallback, XmCCallback, XmRCallback, sizeof(XtCallbackList),
    XtOffset(IconWidget, icon.disarm_callback), 
    XmRPointer, (caddr_t)NULL
  }
};


/* The IconWidget class record definition */

IconWidgetClassRec iconWidgetClassRec = 
{
  /* Core */
  {
    (WidgetClass)&xmLabelClassRec,   /* superclass                   */
    "IconWidget",                    /* class_name                   */
    sizeof(IconWidgetRec),           /* widget_size                  */
    NULL,                            /* class_initialize             */
    NULL,                            /* class_part_initialize        */
    FALSE,                           /* class_inited                 */
    (XtInitProc)Initialize,          /* initialize                   */
    NULL,                            /* initialize_hook              */
    XtInheritRealize,                /* realize                      */
    actionsList,                     /* actions                      */
    XtNumber(actionsList),           /* num_actions                  */
    resources,                       /* resources                    */
    XtNumber(resources),             /* num_resources                */
    NULLQUARK,                       /* xrm_class                    */
    TRUE,                            /* compress_motion              */
    TRUE,                            /* compress_exposure            */
    TRUE,                            /* compress_enterleave          */
    FALSE,                           /* visible_interest             */
    Destroy,                         /* destroy                      */
    XtInheritResize,                 /* resize                       */
    (XtExposeProc)Redisplay,         /* expose                       */
    (XtSetValuesFunc)SetValues,      /* set_values                   */
    NULL,                            /* set_values_hook              */
    XtInheritSetValuesAlmost,        /* set_values_almost            */
    NULL,                            /* get_values_hook              */
    NULL,                            /* accept_focus                 */
    XtVersion,                       /* version                      */
    NULL,                            /* callback private             */
    defaultTranslations,             /* tm_table                     */
    XtInheritQueryGeometry,          /* query_geometry               */
    NULL,                            /* display_acclerator           */
    NULL                             /* extension                    */
  },
  /* Primitive */
  {
    XmInheritBorderHighlight,        /* Primitive border_highlight   */
    XmInheritBorderUnhighlight,      /* Primitive border_unhighlight */
    XtInheritTranslations,           /* translations                 */
    NULL,                            /* arm and activate             */
    NULL,                            /* get_resources                */
    0,                               /* num_get_resources            */
    NULL                             /* extension                    */
  },
  /* Label */
  {
    XmInheritSetOverrideCallback,    /* SetOverideCallback           */
    XmInheritMenuProc,               /* SetWhichButton               */
    XtInheritTranslations,           /* menu traversal xlation       */
    NULL                             /* extension                    */
  },
  /* IconWidget */
  {
    NULL                             /* extension                    */
  }
};

WidgetClass iconWidgetClass = (WidgetClass)&iconWidgetClassRec;


/* *************************************************** */
/* Initialize                                          */
/* Get geometry if pixmap already there                */
/* *************************************************** */
static void Initialize (IconWidget Request, IconWidget New)
{
  if (New->icon.shadow_type != XmSHADOW_IN && 
      New->icon.shadow_type != XmSHADOW_OUT)
    New->icon.shadow_type = XmSHADOW_OUT;

  GetPixmapSize (New);
  New->icon.Armed = FALSE;
}

/* *************************************************** */
/* GetPixmapSize                                       */
/* *************************************************** */
static void GetPixmapSize (IconWidget iw)
{
  unsigned int w, h, Dummy;
  int          Junk;
  Window       Root;

  if (iw->label.pixmap != XmUNSPECIFIED_PIXMAP)
  {
    XGetGeometry (XtDisplay(iw), iw->label.pixmap,
		  &Root, &Junk, &Junk, &w, &h, &Dummy, &Dummy);
    iw->icon.Width  = (short)w;
    iw->icon.Height = (short)h;
  }
}

/* *************************************************** */
/* Redisplay                                           */
/* Redraw on expose events                             */
/* *************************************************** */
static void Redisplay (Widget widget, XEvent *Event, Region region)
{
  short      avail;
  IconWidget iw = (IconWidget)widget;
  short      shadow = iw->primitive.shadow_thickness +
                      iw->primitive.highlight_thickness;

  /* use the label's expose first */
  (* xmLabelClassRec.core_class.expose) (widget, Event, region);

  /* copy the pixmap into the space left by margin_top */
  if (iw->label.pixmap != XmUNSPECIFIED_PIXMAP)
  {
    avail = iw->core.width - 2 * shadow - 2 * iw->label.margin_width - 
            iw->label.margin_left - iw->label.margin_right - iw->icon.Width;
    XCopyArea (XtDisplay(widget), iw->label.pixmap, XtWindow(widget),
	       iw->label.normal_GC, 0, 0,
	       iw->icon.Width,
	       (iw->icon.Height > iw->label.margin_top) ?
	         iw->label.margin_top : iw->icon.Height,
	       avail/2 + shadow + 
	         iw->label.margin_width - iw->label.margin_left,
	       iw->label.TextRect.y - iw->icon.Height - iw->icon.Offset);
  }

  /* Draw the shadow */
  DrawShadow (iw, TRUE);

  /* for keyboard traversal */
#if 0
  if (iw->primitive.highlighted)
    _XmHighlightBorder (widget);
  else if (_XmDifferentBackground(widget, XtParent(widget)))
    _XmUnhighlightBorder (widget);
#endif
}

/* *************************************************** */
/* DrawShadow                                          */
/* *************************************************** */
static void DrawShadow (IconWidget widget, Boolean really)
{
  Boolean in = (widget->icon.Armed && really) ||
               (widget->icon.shadow_type == XmSHADOW_IN);

  if ((widget->primitive.shadow_thickness) > 0 && XtIsRealized((Widget)widget))
    XmeDrawShadows (XtDisplay(widget), XtWindow(widget),
		   in ? widget->primitive.bottom_shadow_GC : 
		        widget->primitive.top_shadow_GC,
		   in ? widget->primitive.top_shadow_GC :
		        widget->primitive.bottom_shadow_GC,
		   widget->primitive.shadow_thickness,
		   widget->primitive.highlight_thickness,
		   widget->primitive.highlight_thickness,
		   (int)widget->core.width -
		        2 * widget->primitive.highlight_thickness,
		   (int)widget->core.height -
		        2 * widget->primitive.highlight_thickness,
		    widget->icon.shadow_type);
}

/* *************************************************** */
/* SetValues                                           */
/* Update geometry, pixmap might have changed          */
/* *************************************************** */
static Boolean SetValues (Widget Current, Widget Request, Widget NewW)
{
  IconWidget cur = (IconWidget)Current;
  IconWidget New = (IconWidget)NewW;

  if (New->icon.shadow_type != XmSHADOW_IN &&
      New->icon.shadow_type != XmSHADOW_OUT)
    New->icon.shadow_type = XmSHADOW_OUT;

  GetPixmapSize (New);

  return (New->icon.shadow_type != cur->icon.shadow_type ||
  	  New->primitive.highlight_thickness != 
	    cur->primitive.highlight_thickness ||
	  New->primitive.shadow_thickness != 
	    cur->primitive.shadow_thickness);
}

/* *************************************************** */
/* Destroy                                             */
/* Clean up allocated resources when the widget is     */
/* destroyed                                           */
/* *************************************************** */
static void Destroy (IconWidget widget)
{
  XtRemoveAllCallbacks ((Widget)widget, XmNactivateCallback);
  XtRemoveAllCallbacks ((Widget)widget, XmNarmCallback);
  XtRemoveAllCallbacks ((Widget)widget, XmNdisarmCallback);
}

/* *************************************************** */
/* Enter                                               */
/* *************************************************** */
static void Enter (IconWidget widget, XEvent *Event)
{
#if 0
  _XmPrimitiveEnter (widget, Event);
#endif

  if (widget->icon.Armed)
    DrawShadow (widget, True);
}

/* *************************************************** */
/* Leave                                               */
/* *************************************************** */
static void Leave (IconWidget iw, XEvent *Event)
{
#if 0
  _XmPrimitiveLeave (iw, Event);
#endif
  
  if (iw->icon.Armed)
    DrawShadow (iw, False);
}

/* *************************************************** */
/* Arm                                                 */
/* *************************************************** */
static void Arm (IconWidget iw, XEvent *Event)
{
  XmAnyCallbackStruct CallValue;

  iw->icon.Armed = True;
  DrawShadow (iw, True);

  if (iw->icon.arm_callback)
  {
    XFlush (XtDisplay(iw));
    CallValue.reason = XmCR_ARM;
    CallValue.event  = Event;
    XtCallCallbackList ((Widget)iw, iw->icon.arm_callback, &CallValue);
  }
}

/* *************************************************** */
/* Activate                                            */
/* *************************************************** */
static void Activate (IconWidget iw, XEvent *Event)
{
  XmAnyCallbackStruct CallValue;

  if ((Event->xany.type == ButtonPress || Event->xany.type == ButtonRelease) &&
      (Event->xbutton.x > iw->core.width || 
       Event->xbutton.y > iw->core.height))
    return;

  CallValue.reason = XmCR_ACTIVATE;
  CallValue.event  = Event;
  XtCallCallbackList ((Widget)iw, iw->icon.activate_callback, &CallValue);
}


/* *************************************************** */
/* Disarm                                              */
/* *************************************************** */
static void Disarm (IconWidget iw, XEvent *Event)
{
  XmAnyCallbackStruct CallValue;

  iw->icon.Armed = False;
  DrawShadow (iw, True);

  if (iw->icon.arm_callback)
  {
    XFlush (XtDisplay(iw));
    CallValue.reason = XmCR_ARM;
    CallValue.event  = Event;
    XtCallCallbackList ((Widget)iw, iw->icon.disarm_callback, &CallValue);
  }
}

/* *************************************************** */
/* Help                                                */
/* *************************************************** */
static void Help (IconWidget iw, XEvent *Event)
{
  XmAnyCallbackStruct CallValue;

  CallValue.reason = XmCR_HELP;
  CallValue.event  = Event;
  XtCallCallbackList ((Widget)iw, iw->primitive.help_callback, &CallValue);
}

/* *************************************************** */
/* CreateIconLabel                                     */
/* Create an instance of an icon as label              */
/* *************************************************** */
Widget CreateIconLabel (Widget Parent, char *Name, 
			ArgList arglist, Cardinal ArgCount, char *PixMap)
{
  Widget widget;
  Pixmap Pix;
  Arg    args[1];

  widget = XtCreateManagedWidget (Name, iconWidgetClass, 
				  Parent, arglist, ArgCount);
  if (PixMap != NULL && PixMap != "")
  {
    Pix = XmGetPixmap (((IconWidget)widget)->core.screen, PixMap,
		       ((IconWidget)widget)->primitive.foreground,
		       ((IconWidget)widget)->core.background_pixel);
    if (Pix != XmUNSPECIFIED_PIXMAP)
    {
      XtSetArg (args[0], XmNlabelPixmap, Pix);
      XtSetValues (widget, args, 1);
    }
  }
  return widget;
}


/* *************************************************** */
/* CreateIconButton                                    */
/* Create an instance of an icon as button             */
/* *************************************************** */
#define BUTTON_TRANS \
        "#override <Btn1Down>:  Arm()\n \
                   <Btn1Up>:    Activate() Disarm"

Widget CreateIconButton (Widget  Parent, char     *Name,
			 ArgList Arguments, Cardinal ArgCount,
			 char *PixMap)
{
  Widget         w;
  Pixmap         Pix;
  Arg            args[1];
  XtTranslations parsed = XtParseTranslationTable (BUTTON_TRANS);

  w = XtCreateManagedWidget (Name, iconWidgetClass, 
			     Parent, Arguments, ArgCount);
  if (PixMap != NULL && PixMap != "")
  {
    Pix = XmGetPixmap (((IconWidget)w)->core.screen, PixMap,
		       ((IconWidget)w)->primitive.foreground,
		       ((IconWidget)w)->core.background_pixel);
    if (Pix != XmUNSPECIFIED_PIXMAP)
    {
      XtSetArg (args[0], XmNlabelPixmap, Pix);
      XtSetValues (w, args, 1);
    }
  }

  XtAugmentTranslations (w, parsed);

  return w;
}


