/* ************************************ */
/* *** (c) Stefan Vietinghoff       *** */
/* *** Functions for setting dialog *** */
/* *** attributes                   *** */
/* ************************************ */

#include <stdlib.h>
#include <math.h>

#include "dlgData.h"
#include "dlgSegment.h"

/* ***** eigene globale Variablen: ***** */
static int n = 0;
static Arg args[NUM_ARGS];

/*----------------------------------------------------------------------
  Funktion: Setzt eine berschrift ber ein Box. Dies kann sowohl eine
            Box sein, welche ein Item enthalt, als auch eine Box, die
	    weitere Unterboxen enthlt.
	    Ist Title ein NULL-Zeiger, werden evtl. zuvor gesetzte 
	    berschriften gelscht, andernfalls gendertert
	    Der Titel wird in die interne Struktur kopiert, kann/mu
	    also vom Benutzter freigegeben werden.

  INPUT:  Dlg    Dialog der die Box enthlt
          BoxID  Nummer der Box, dessen berschrift gesetzt werden soll
	  Title  berschrift (auch als Resourceeintrag)
  OUTPUT: ---

  Bsp.: DIALOG *Dlg = NULL;
        ...
	DLGSetTitle (Dlg, 12, "Eintrag");
  ----------------------------------------------------------------------*/
void DLGSetTitle (DIALOG *Dlg, int BoxID, const char *Title, char Position)
{
  DLGBOX *Ptr;

  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR ("Try to set title of an empty dialog")

  if ((Ptr = DLG_GetBox(Dlg->Box, DLGBox(BoxID))) == NULL)
    PRO_ERROR_VAL ("The dialog contains no BoxID %d", BoxID)

  DLG_strncp (Ptr->Title, Title, DLG_TLEN);
  Ptr->TitlePos = Position;

  PRO_END
}


/*----------------------------------------------------------------------
  Funktion: Setzt/ndert eine Umrandung einer Dialogbox (NICHT
            rekursiv fr alle Unterboxen mit)

  INPUT:  Dlg    Dialog mit der BoxNummer
          BoxID  Nummer der Box deren Umrandung gendert werden soll
	  State  TRUE, falls ein Rand um die Box dargestellt werden soll
	         FALSE, sonst
  OUTPUT: ---

  Bsp.: DIALOG *Dlg = NULL;
        ...
	DLGSetBorder (Dlg, 12, ETCHEDIN);
  ----------------------------------------------------------------------*/
void DLGSetBorder (DIALOG *Dlg, int BoxID, char *Style)
{
  DLGBOX *Ptr;

  PRO_START ("")

  if (Dlg == NULL)
    PRO_ABORT ("empty/invalid dialog")

  if ((Ptr = DLG_GetBox(Dlg->Box, DLGBox(BoxID))) == NULL)
    PRO_ABORT1 ("Dialog contains no such BoxID %d", BoxID)

  /* change border to new value */
  if (Style == NULL)
    Ptr->Border = DlgNO_LINE[0];
  else
    Ptr->Border = Style[0];

  PRO_END
}


/*----------------------------------------------------------------------
  Funktion: legt die Position der Boxbeschriftung fr eine Box mit allen
            ihren Unterboxen fest.
	    Trennung von DLGSetLabelPos um rekursiv die Unterboxen zu ndern.

  INPUT:  Box       Zeiger auf die Box ab der die Label-Positionen gesetzt
                    werden sollen.
	  Position  neue Postion des/der Label(s) (siehe Dokumentation)
  OUTPUT: ---
  ----------------------------------------------------------------------*/
static void DLG_SetBoxLabelPos (DLGBOX *Box, 
				unsigned long LabelPos, unsigned long ValuePos)
{
  int     countX, countY;
  DLGBOX *Ptr;

  if (Box == NULL)
    return;

  /* check, if box is non-empty item-box: */
  if (Box->Dir == BOXELEM && Box->Item != NULL)
  {
    Box->Item->LabelPos = LabelPos;
    Box->Item->ValuePos = ValuePos;
  }
  else
    for (countY = 0; countY < Box->Rows; ++countY)
      for (countX = 0; countX < Box->Cols; ++countX)
	if ((Ptr = Box->List[countX][countY]) != NULL)
	  DLG_SetBoxLabelPos (Ptr, LabelPos, ValuePos);
}


/*----------------------------------------------------------------------
  Funktion: Exportierte Version, um die Beschriftungposition  der
            Dialogeintrge zu setzten. (greift auf DLG_SetBoxLabelPos zu)

  INPUT:  Dlg    Zeiger auf den Dialog in dem sich die Box(en)
                 befindet/n
		 BoxID  Nummer der Box auf die und deren Unterboxen sich die
		 Operation erstreckt.
  OUTPUT: ---

  Bsp.: DIALOG *Dlg = NULL;
        ...
	DLGSetLabelPos (Dlg, 12, FRONT);
  ----------------------------------------------------------------------*/
void DLGSetLabelPos (DIALOG *Dlg, int BoxID, 
		     unsigned long LabelPos, unsigned long ValuePos)
{
  DLGBOX *Ptr;

  PRO_START ("")
  PRO1 ("\\%d...", BoxID)

  if (Dlg == NULL)
    PRO_ERROR ("Invalid/empty dialog")

  /* Box suchen: */
  if ((Ptr = DLG_GetBox(Dlg->Box, DLGBox(BoxID))) == NULL)
    PRO_ERROR_VAL ("Dialog contains no BoxID %d", BoxID)

  DLG_SetBoxLabelPos (Ptr, LabelPos, ValuePos);

  PRO_END
}  


void DLG_SetBoxAlignment (DLGBOX *Box, int AdjustX, int AdjustY)
{
  int     countX, countY;
  DLGBOX *Ptr;

  if (Box == NULL)
    return;

  if (AdjustX != DlgNO_CHANGE)
    Box->AlignX = AdjustX;
  if (AdjustY != DlgNO_CHANGE)
    Box->AlignY = AdjustY;
  
  if (Box->Dir == BOXGRID || Box->Dir == BOXARRAY)
  {
    for (countY = 0; countY < Box->Rows; ++countY)
      for (countX = 0; countX < Box->Cols; ++countX)
	if ((Ptr = Box->List[countX][countY]) != NULL)
	  DLG_SetBoxAlignment (Ptr, AdjustX, AdjustY);
  }
}


void DLGSetBoxAlignment (DIALOG *Dlg, int BoxID, int AdjustX, int AdjustY)
{
  DLGBOX *Ptr;

  PRO_START ("")
  PRO1 ("\\%d...", BoxID)

  if (Dlg == NULL)
    PRO_ERROR ("Invalid/empty dialog")

  if ((Ptr = DLG_GetBox(Dlg->Box, DLGBox(BoxID))) == NULL)
    PRO_ERROR_VAL ("Dialog contains no BoxID %d", BoxID)

  DLG_SetBoxAlignment (Ptr, AdjustX, AdjustY);

  PRO_END
}


/* ********************************************************************** */
/* ** Sets the CB-function a dialog box and all it's subboxes.         ** */
/* ** To overwrite existing data, first set the function to NULL !!!   ** */
/* ********************************************************************** */
static void DLG_SetFunction (DLGBOX *Box, DLGFUNCTION Function)
{
  int      countX, countY;
  DLGBOX  *Ptr;
  DLGELEM *Elem;
  
  if (Box == NULL)
    return;

  if (Box->Dir == BOXELEM && Box->Item != NULL)
  {
    Elem = Box->Item;

    switch (Elem->Type)    /* insert CB-function: */
    {
      case  DlgLABEL : break;
      case   DlgBOOL : if (Elem->Value.drec.FPtr == NULL || Function == NULL)
	                 Elem->Value.drec.FPtr = Function;
	               break;
      case    DlgINT : if (Elem->Value.irec.FPtr == NULL || Function == NULL)
	                 Elem->Value.irec.FPtr = Function;
                       break;
      case DlgDOUBLE : if (Elem->Value.rrec.FPtr == NULL || Function == NULL)
	                 Elem->Value.rrec.FPtr = Function;
                       break;   
      case   DlgTEXT : if (Elem->Value.srec.FPtr == NULL || Function == NULL)
	                 Elem->Value.srec.FPtr = Function;
	               break;
      case  DlgSCALE : if (Elem->Value.zrec.FPtr == NULL || Function == NULL)
	                 Elem->Value.zrec.FPtr = Function;
                       break;
      case    DlgBAR : if (Elem->Value.brec.FPtr == NULL || Function == NULL)
	                 Elem->Value.brec.FPtr = Function;
                       break;
      case   DlgUPDN : if (Elem->Value.urec.FPtr == NULL || Function == NULL)
	                 Elem->Value.urec.FPtr = Function;
                       break; 
      case   DlgLIST : if (Elem->Value.lrec.FPtr == NULL || Function == NULL)
	                 Elem->Value.lrec.FPtr = Function;
                       break;
      case  DlgOMENU : if (Elem->Value.orec.FPtr == NULL || Function == NULL)
	                 Elem->Value.orec.FPtr = Function;
                       break;
      case   DlgTBOX : if (Elem->Value.trec.FPtr == NULL || Function == NULL)
	                 Elem->Value.trec.FPtr = Function;
                       break;
      case   DlgRBOX : if (Elem->Value.trec.FPtr == NULL || Function == NULL)
	                 Elem->Value.trec.FPtr = Function;
                       break;
      case   DlgPROC : if (Elem->Value.crec.FPtr == NULL || Function == NULL)
	                 Elem->Value.crec.FPtr = Function;
                       break;
      case  DlgARROW : if (Elem->Value.arec.FPtr == NULL || Function == NULL)
	                 Elem->Value.arec.FPtr = Function;
                       break;
      case    DlgPIC : if (Elem->Value.prec.FPtr != NULL || Function == NULL)
                         Elem->Value.prec.FPtr = Function;
                       break;
      case  DlgIMAGE : if (Elem->Value.xrec.FPtr != NULL || Function == NULL)
                         Elem->Value.xrec.FPtr = Function;
                       break;
      case   DlgFSEL : 
      case DlgDIRSEL : if (Elem->Value.grec.FPtr == NULL || Function == NULL)
	                 Elem->Value.grec.FPtr = Function;
                       break;
      case   DlgDIAL : if (Elem->Value.erec.FPtr == NULL || Function == NULL)
	                 Elem->Value.erec.FPtr = Function;
                       break;
      case   DlgSEGM : break;
      case   DlgNOTE : {
	                 NOTEPAGE *Page;

			 if (Elem->Value.nrec.FPtr == NULL || Function == NULL)
			   Elem->Value.nrec.FPtr = Function;

			 Page = Elem->Value.nrec.Page;
			 while (Page != NULL)
			 {
			   DLG_SetFunction (Page->Box, Function);
			   Page = Page->Next;
		         }
	               }
	               break;
      case   DlgTREE : if (Elem->Value.krec.FPtr == NULL || Function == NULL)
	                 Elem->Value.krec.FPtr = Function;
	               break;
      case   DlgSEP  : break;
      case   DlgBOX  : break;
      case   DlgUSER : break;
      default        : PRO_ERROR_VAL ("Invalid dialog element %d\n", Elem->Type)
    }
  }
  else
    for (countY = 0; countY < Box->Rows; ++countY)
      for (countX = 0; countX < Box->Cols; ++countX)
	if ((Ptr = Box->List[countX][countY]) != NULL)
	  DLG_SetFunction (Ptr, Function);
}

/*----------------------------------------------------------------------
  Funktion: mit dieser Funktion wird ein Funktionsaufruf an ein 
            Dialogeintrag gekoppelt, d.h, jedesmal, wenn dieser
	    Eintrag sein Wert ndert, wird die angegeben Funktion 
	    aufgerufen, welche als Parameter einen Zeiger auf die
	    aktuellen Dialog-Struktur (bitte nicht lschen), 
	    sowie eine den genderten Wert erhlt (Dieser Wert
	    hngt vom Typ der Items ab -> Dokumentation)
	    Existiert die Box mit der Nummer BoxID nicht springt
	    die Routine zuck.

  INPUT:  Dlg       Dialog in dem sich der Dialogereintrag befindet
          BoxID     Nummer der Box an die die Funktion gehngt werden soll
	  Function  Funktion, die bei Wertnderung aufgerufen werden soll
  OUTPUT: ---

  Bsp.: DIALOG *Dlg = NULL;
        ...
	DLGSetFunction (Dlg, 12, StartMich);
  ----------------------------------------------------------------------*/
void DLGSetFunction (DIALOG *Dlg, int BoxID, DLGFUNCTION Function)
{
  DLGBOX *Ptr;

  PRO_START ("")
  PRO1 ("\\%d...", BoxID)

  if (Dlg == NULL)
    PRO_ERROR ("Try to set box-function of an empty dialog")

  if ((Ptr = DLG_GetBox(Dlg->Box, DLGBox(BoxID))) == NULL)
    PRO_ERROR_VAL ("The dialog contains no BoxID %d", BoxID)

  DLG_SetFunction (Ptr, Function);

  PRO_END
}


void DLGSetBoxPosition (DIALOG *Dlg, int BoxID, int PosX, int PosY)
{
  DLGBOX  *Ptr;
  DLGELEM *Elem;

  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR ("dialog is empty/uninitialized")

  if ((Ptr = DLG_GetBox(Dlg->Box, DLGBox(BoxID))) == NULL)
    PRO_ERROR_VAL ("The dialog contains no BoxID %d", BoxID)

  Elem = Ptr->Item;
  Elem->PosX = PosX;
  Elem->PosX = PosY;

  PRO_END
}


/*----------------------------------------------------------------------
  Funktion: legt fest, ob eine Box aktiv (sensitiv) oder inaktiv (in-
            sensitiv) sein soll, d.h. ob Wertnderungen/Aktivierungen erlaubt 
	    sind oder eben nicht.

  INPUT:  Dlg     Dialog der die Box enthlt
          BoxID   Bezugspunkt/-box der Operation
	  Active  TRUE, Wertnderung sind fr dieses Item erlaubt
	          FALSE, Wertnderungen werden unterbunden
  OUTPUT: ---

  Bsp.: DIALOG *Dlg = NULL;
        ...
	DLGSetBoxActive (Dlg, 12, FALSE);
	(Wert in der Box 12 nicht mehr gendert werden)  
  ----------------------------------------------------------------------*/
void DLGSetBoxActive (DIALOG *Dlg, int BoxID, dual active)
{
  DLGBOX *Ptr;

  PRO_START  ("\n")

  if (Dlg == NULL)
    PRO_ERROR ("dialog is empty/uninitialized")

  if (TOOLScriptRunning())
    PRO_ABORT ("Script running")

  if ((Ptr = DLG_GetBox(Dlg->Box, DLGBox(BoxID))) == NULL)
  {
    PRO1 ("BoxID: %d\n", BoxID)
    PRO_ERROR_VAL ("The dialog contains no BoxID %d", BoxID)
  }

  if (Ptr->Dir != BOXELEM || Ptr->Item == NULL)     /* nichtleere Item-Box */
  {
    if (Ptr->Box != NULL)                           /* Item eingetragen ?  */
      XtSetSensitive (Ptr->Box, (Boolean)active);
    Ptr->active = active;
  }
  else                                               /* Box mit Unterboxen */
  {
    if (Ptr->Item->Master != NULL)
      XtSetSensitive (Ptr->Item->Master, active);
    Ptr->Item->active = active;
  }
  XFlush (XDisplay);

  PRO_END
}


void DLGShowBox (DIALOG *Dlg, int BoxID, dual visible)
{
  DLGBOX  *Ptr;
  DLGELEM *Elem;

  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR ("dialog is empty/uninitialized")

  if (TOOLScriptRunning())
    PRO_ABORT ("Script running")

  if ((Ptr = DLG_GetBox(Dlg->Box, DLGBox(BoxID))) == NULL)
  {
    PRO1 ("BoxID: %d\n", BoxID)
    PRO_ERROR_VAL ("The dialog contains no BoxID %d", BoxID)
  }

  Elem = Ptr->Item;
  if (! visible)
    Elem->LabelPos = DlgNOWHERE;
  Elem->hidden = ! visible;

  if (Elem->WidgetID != NULL)
  {
    if (! visible)
    {
      XtUnmapWidget (XtParent(Elem->WidgetID));
      XtUnrealizeWidget (XtParent(Elem->WidgetID));
    }
    else 
    {
      XtRealizeWidget (XtParent(Elem->WidgetID));
      XtManageChild (XtParent(Elem->WidgetID));
      XtMapWidget (XtParent(Elem->WidgetID));
    }
    XFlush (XDisplay);
  }

  PRO_END 
}


void DLGSetButtonActive (DIALOG *Dlg, int Button, dual active)
{
  int count;

  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR ("dialog is empty/uninitialized")

  if (TOOLScriptRunning())
    PRO_ABORT ("Script running")

  if (Button < 0 || Button > Dlg->Buttons.Number)
    PRO_ABORT ("Number out of range")

  if (Button == 0)
  {
    for (count = 0; count < Dlg->Buttons.Number; ++count)
      if (Dlg->Buttons.Data[count].Button != NULL)
	XtSetSensitive (Dlg->Buttons.Data[count].Button, active);
  }
  elsif (Dlg->Buttons.Data[Button-1].Button != NULL)
    XtSetSensitive (Dlg->Buttons.Data[Button-1].Button, active);

  XFlush (XDisplay);

  PRO_END
}


/*----------------------------------------------------------------------
  Funktion: legt fest, ob eine Box aktiv (sensitiv) oder inaktiv (in-
            sensitiv) sein soll, d.h. ob Wertnderungen/Aktivierungen 
	    erlaubt sind oder eben nicht.
  
  INPUT:  Dlg     Dialog der die Box enthlt
          BoxID   Bezugspunkt/-box der Operation
	  active  TRUE, Wertnderung sind fr dieses Item erlaubt
	          FALSE, Wertnderungen werden unterbunden
  OUTPUT: ---

  Bsp.: DIALOG *Dlg = NULL;
        ...
	DLGSetBoxActive (Dlg, 12, FALSE);
	(Wert in der Box 12 nicht mehr gendert werden)  
  ----------------------------------------------------------------------*/
void DLGSetDlgActive (DIALOG *Dlg, dual active)
{
  PRO_START ("")

  if (! TOOLIsInitialized())
    PRO_ERROR ("Dlg.-Lib has NOT been initialized\n")

  if (Dlg == NULL)
    PRO_ERROR ("dialog is empty/uninitialized")

  if (TOOLScriptRunning())
    PRO_ABORT ("Script running")

  if (Dlg->Form != NULL)
    XtSetSensitive (Dlg->Form, (Boolean)active);
  Dlg->active = active;
  XFlush (XDisplay);

  PRO_END
}


/*----------------------------------------------------------------------
  Funktion: Setzt die Farbe mit der ein Dialogeintrag dargestellt
            wird.

  INPUT:  Dlg      Dialog in dem sich die Box befindet
          BoxID    Box auf die sich die nderung bezieht
	  r, g, b  Werte der Grundfarben (rot, grn, blau) 
	           jeweils im Wertebereich [0..255]
	  front    TRUE  Einstellung der Vordergrundfarbe
	           FALSE      "       "  Hintergrundfarbe
  OUTPUT: ---

  Bsp.: DIALOG *Dlg = NULL;
        ...
	DLGSetColor (Dlg, 10, 255, 255, 0);  
	Stellt die Box mit der Nummer 10 in gelber Farbe dar
  ----------------------------------------------------------------------*/
void DLGSetColor (DIALOG *Dlg, int BoxID, int r, int g, int b, dual front)
{
  XColor   Color;
  Pixmap   Pix1, Pix2;
  DLGBOX   *Ptr = NULL;
  DLGELEM *Elem = NULL;

  PRO_START ("\n")

  if (Dlg == NULL)
    PRO_ERROR ("Try to set box-color of an empty dialog")

  if ((Ptr = DLG_GetBox(Dlg->Box, DLGBox(BoxID))) == NULL)
    PRO_ERROR_VAL ("The dialog contains no BoxID %d", BoxID)

  if (Ptr->Dir != BOXELEM || (Elem = Ptr->Item) == NULL)
    PRO_ERROR_VAL ("Box %d contains no element", BoxID)

  Color.pad   = 0;
  Color.red   = 256 * r;     /* # Farbwerte pro Grundfarbe = 16 */
  Color.green = 256 * g;     /* => 65536 / 16 = 4096            */
  Color.blue  = 256 * b;
  if (! XAllocColor(XDisplay, CMap, &Color))
    PRO_ABORT ("too much Colors allocated\n")

  PRO4 ("(%d|%d|%d)->%ld...\n", r, g, b, Color.pixel)

  /*
    TODO:
    alte Farbe freigeben (XFreeColors)
  */
  if (Elem->WidgetID == NULL)
  {
    if (front)
      Elem->fg = (Pixel)Color.pixel;
    else
      Elem->bg = (Pixel)Color.pixel;
    return;
  }
  else
    switch (Elem->Type)
    {
      case DlgEMPTY :
	break;

      /* 
	 TODO:
	 restliche Eintrge (soweit sinnvoll)
      */
      case  DlgLABEL :
      case   DlgBOOL : 
      case    DlgINT :
      case DlgDOUBLE :
      case   DlgTEXT :
      case    DlgBAR :
      case   DlgUPDN :
      case   DlgLIST :
      case   DlgTBOX :
      case   DlgRBOX :
      case   DlgPROC :
      case   DlgFSEL :
      case DlgDIRSEL :
      case    DlgSEP :
      case    DlgBOX :
        break;

      case DlgOMENU :
	if (front)
	  XtSetArg (args[0], XmNforeground, Color.pixel);
	else
	  XtSetArg (args[0], XmNbackground, Color.pixel);

	XtSetValues (XmOptionButtonGadget(Elem->WidgetID), args, 1);
	break;

      case DlgSCALE :
	if (front)
	  XtSetArg (args[0], XmNforeground, Color.pixel);
	else
	  XtSetArg (args[0], XmNbackground, Color.pixel);
	XtSetValues (Elem->WidgetID, args, 1);
	break;

      case DlgARROW :
	/* !!! Prfen, ob Center Arrow-> neu erzeugen */
	if (front)
	  XtSetArg (args[0], XmNforeground, Color.pixel);
	else
	  XtSetArg (args[0], XmNbackground, Color.pixel);
	XtSetValues (Elem->WidgetID, args, 1);
	break;

      case DlgPIC :
	if (front)
	  Elem->Value.prec.fg = Color.pixel;
	else
	  Elem->Value.prec.bg = Color.pixel;
	if (Elem->WidgetID != NULL)
	{
	  Pix1 = XCreatePixmapFromBitmapData (XDisplay,
					      DefaultRootWindow (XDisplay),
					      Elem->Value.prec.Data, 
					      Elem->Value.prec.SizeX,
					      Elem->Value.prec.SizeY, 
					      Elem->Value.prec.fg,
					      Elem->Value.prec.bg,
					      TOOLGetDepth());
	  n = 0;
	  if (Elem->Value.prec.FPtr != NULL)
	  {
	    Pix2 = XCreatePixmapFromBitmapData (XDisplay,
						DefaultRootWindow(XDisplay), 
						Elem->Value.prec.iData, 
						Elem->Value.prec.SizeX,
						Elem->Value.prec.SizeY, 
						Elem->Value.prec.fg,
						Elem->Value.prec.bg,
						TOOLGetDepth());
	    XtSetArg (args[n], XmNlabelInsensitivePixmap, Pix2); n++;
	  }
	  XtSetArg (args[n], XmNlabelPixmap, Pix1); n++;
	  XtSetArg (args[n], XmNforeground, Elem->Value.prec.fg); n++;
	  XtSetArg (args[n], XmNbackground, Elem->Value.prec.bg); n++;
	  XtSetValues (Elem->WidgetID, args, n);
	}
	break;

      default:
	PRO_ERROR_VAL ("Box %d has an unknown type", Elem->Type)
    }

  PRO_END
}


void DLGSetArguments (DIALOG *Dlg, int BoxID, ArgList Arguments, int Number)
{
  int      OldNum;
  ArgList  OldArgs = NULL;
  DLGBOX  *Ptr;
  DLGELEM *Elem;

  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR ("dialog is empty/uninitialized")

  if (TOOLScriptRunning())
    PRO_ABORT ("Script running")

  if ((Ptr = DLG_GetBox(Dlg->Box, DLGBox(BoxID))) == NULL)
    PRO_ERROR_VAL ("The dialog contains no BoxID %d", BoxID)

  Elem    = Ptr->Item;
  OldNum  = Elem->NumArgs;
  OldArgs = Elem->UserArgs;
  Elem->UserArgs = XtMergeArgLists (OldArgs, OldNum, Arguments, Number);
  Elem->NumArgs += Number;

  if (OldArgs != NULL)
    XtFree ((char*)OldArgs);

  if (Elem->WidgetID != NULL)
  {
    XtUnmanageChild (Elem->Dlg->Shell);
    XtUnmanageChild (XtParent(Elem->WidgetID));
    /*  XtUnmapWidget (XtParent(Elem->WidgetID)); */
    /* XtUnrealizeWidget (XtParent(Elem->WidgetID)); */

    XtSetValues (Elem->WidgetID, Arguments, Number);

    /* XtRealizeWidget (XtParent(Elem->WidgetID)); */
    XtManageChild (XtParent(Elem->WidgetID));
    XtManageChild (Elem->Dlg->Shell);
    /*   XtMapWidget (XtParent(Elem->WidgetID)); */
  }
 
  PRO_END
}


/*----------------------------------------------------------------------
  Funktion: Setzt den Mauszeiger an eine gegebene Position
            (diese Funktion sollte i.d.R. nicht benutzt werden)
  ----------------------------------------------------------------------*/
void DLGSetMousePos (DIALOG *Dlg, int BoxID)
{
  Dimension  width, height;
  Position   XPos, YPos;
  int        SumX, SumY;
  char      *Geometry = NULL;
  Widget     Box = NULL;
  DLGBOX    *Ptr = NULL;
  DLGELEM   *Elem = NULL;

  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR ("Bad Dialog")

  if ((Ptr = DLG_GetBox(Dlg->Box, DLGBox(BoxID))) == NULL)
    PRO_ERROR_VAL ("Dialog contains no such BoxID %d", BoxID)

  if (Ptr->Dir != BOXELEM || Ptr->Item  == NULL)
    PRO_ERROR_VAL ("Box %d contains no element", BoxID)

  Elem = Ptr->Item;

  if (Elem->WidgetID != NULL)
  {
    SumX = 0;
    SumY = 0;

    Box = Elem->WidgetID;
    do
    {
      XtSetArg (args[0], XmNx, &XPos);
      XtSetArg (args[1], XmNy, &YPos);
      XtGetValues (Box, args, 2);

      SumX += XPos;
      SumY += YPos;

      Box = XtParent (Box);
    } while (Box != Dlg->Form);

    XtSetArg (args[0], XmNheight,   &height);
    XtSetArg (args[1], XmNwidth,    &width);
    XtSetArg (args[2], XmNgeometry, Geometry);
    XtGetValues (Elem->WidgetID, args, 3);

    PRO3 ("setting MousePointer to Box %d: (%d/%d)\n", BoxID, SumX, SumY)
    PRO3 ("Size: %dx%d = %s...\n", height, width, Geometry)

    XWarpPointer (XDisplay, None, XtWindow(Dlg->Shell),
		  None, None, None, None, SumX+width/2, SumY+height/2);
  }
  else
    PRO1 ("No valid widget in box %d\n", BoxID)

  PRO_END
}


void DLGSetMouseCursor (DIALOG *Dlg, int CursorID)
{
  PRO_START ("")

  if (Dlg == NULL)
    PRO_ABORT ("dialog is empty/uninitialized")

  if (Dlg->Shell == NULL || TOOLScriptRunning())
    PRO_ABORT ("Try to set Cursor of an closed Dialog")

  TOOLSetMouseCursor (Dlg->Shell, CursorID, NULL, NULL);

  PRO_END
}


/*----------------------------------------------------------------------
  Funktion: Mit dieser Funktion kann der Benutzer eigene Informationen
            in einem Dialog speichern, die dann z.B. in einer Callback-
	    Routine wieder ausgelesen werden knnen. 
	    Der Benutzer ist fr die Allokierung und Freigabe des bentigten
	    Speichers verantwortlich.
	    Eine Anwendungsmglichkeit ist z.B. das Speichern einer
	    Fensternummer, um fr mehrere Fenster jeweils einen gleichen 
	    Dialog zu ffnen und bei nderungen innerhalb einem dieser
	    Dialog das zugehrige Fenster zu ermitteln.

  INPUT:  Dlg   Dialog, in dem die Information gespeichert werden soll.
          Info  Zeiger auf den Speicherbereich mit der Daten
  OUTPUT: none

  Bsp.: int        WinNr = 3;
        DIALOG *Dlg = NULL;
	...
	DLGSetUserData (Dlg, 0, &WinNr);
  ----------------------------------------------------------------------*/
static void DLG_SetUserData (DLGBOX *Box, void *Data)
{
  int     countX, countY;
  DLGBOX *Ptr = NULL;

  if (Box->UserData == NULL || Data == NULL)
    Box->UserData = Data;

  if (Box->Dir != BOXELEM)
    for (countY = 0; countY < Box->Rows; ++countY)
      for (countX = 0; countX < Box->Cols; ++countX)
	if ((Ptr = Box->List[countX][countY]) != NULL)
	  DLG_SetUserData (Ptr, Data);
}


void DLGSetUserData (DIALOG *Dlg, int BoxID, void *Data)
{
  DLGBOX *Ptr = NULL;

  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR ("dialog is empty/uninitialized")

  if ((Ptr = DLG_GetBox(Dlg->Box, DLGBox(BoxID))) == NULL)
    PRO_ERROR_VAL ("Dialog contains no such BoxID %d", BoxID)

  DLG_SetUserData (Ptr, Data);

  PRO_END
}


/*----------------------------------------------------------------------
  Funktion: Setzt die Startposition des Dialog, falls dieser nicht 
            im USERPOS-Stil erzeugt wird und nicht durch einen Eintrag
	    in der Resourcedatei berschrieben wird

  INPUT:  Dlg         zu positionierender Dialog
          xPos, yPos  Koordinaten der linken oberen Ecke des Dialog
	              fr 0 < Pos < 1 in Prozent
		      sonst           in absoluten Pixeln
  OUTPUT: ...

  Bsp.: DIALOG *Dlg;
        ...
	DLGSetPos (Dlg, 150.0, 200.0);
  ----------------------------------------------------------------------*/
void DLGSetPos (DIALOG *Dlg, double PosX, double PosY)
{
  int ScrX, ScrY;

  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR ("dialog is empty/uninitialized")

  ScrX = DisplayWidth  (XDisplay, XScreen);
  ScrY = DisplayHeight (XDisplay, XScreen);

  if (PosX > 0.0 && PosX < 1.0)
    Dlg->PosX = (int)(PosX * ScrX);
  elsif (PosX >= 1.0)
    Dlg->PosX = MIN (abs(PosX), ScrX);
  else
    Dlg->PosX = -1;

  if (PosY > 0.0 && PosY < 1.0)
    Dlg->PosY = (int)(PosY * ScrY);
  elsif (PosY >= 1.0)
    Dlg->PosY = MIN (abs(PosY), ScrY);
  else
    Dlg->PosY = -1;

  Dlg->PosSet = (PosY >= 0.0 && PosY >= 0.0);

  PRO_END
}


void DLGGetPos (DIALOG *Dlg, int *PosX, int *PosY)
{
  PRO_START ("")

  if (Dlg == NULL)
  {
    if (PosX != NULL)
      *PosX = 0;
    if (PosY != NULL)
      *PosY = 0;
  }
  else
  {
    DLG_UpdateSizeAndPos (Dlg);

    if (PosX != NULL)
      *PosX = Dlg->PosX;
    if (PosY != NULL)
      *PosY = Dlg->PosY;
  }

  PRO_END
}


void DLGGetSize (DIALOG *Dlg, int *SizeX, int *SizeY)
{
  PRO_START ("")

  if (Dlg == NULL)
  {
    if (SizeX != NULL)
      *SizeX = 0;
    if (SizeY != NULL)
      *SizeY = 0;
  }
  else
  {
    DLG_UpdateSizeAndPos (Dlg);

    if (SizeX != NULL)
      *SizeX = Dlg->SizeX;
    if (SizeY != NULL)
      *SizeY = Dlg->SizeY;
  }

  PRO_END
}


/*----------------------------------------------------------------------
  Funktion: Setzt die (Anfangs-)gre eines Dialogfensters

  INPUT:  Dlg           Dialog der aus-/ein-geblendet werden soll
          SizeX, Sizey  Gre des Dialog 
	                0    : Gre wird automatisch berechnet (default)
	                0 < 1: Gre in Prozent vom Bildschirm
			> 1  : Gre absolut in Pixeln
	  Resize        Die Gre des Fensters lt sich verndern
  OUTPUT: ---

  Bsp.: DIALOG *Dlg = NULL
        ...
	DLGSetSize (Dlg, 300, 200, TRUE);
  ----------------------------------------------------------------------*/
void DLGSetSize (DIALOG *Dlg, double SizeX, double SizeY, dual Resize)
{
  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR ("dialog is empty/uninitialized")

  Dlg->Resize = Resize;

  TOOLGetScrValue (XDisplay, XScreen, SizeX, SizeY, &Dlg->SizeX, &Dlg->SizeY);

  PRO_END
}


/*----------------------------------------------------------------------
  Funktion: ermglicht es, den Dialog zwischenzeitlich auszublenden
            bzw. wieder einzublenden.

  INPUT:  Dlg    Dialog der aus-/ein-geblendet werden soll
          State  Zustand:  TRUE = Dialog wird eingeblendet
	                  FALSE =  "" wird ausgeblendet
  OUTPUT: ---

  Bsp.: DIALOG *Dlg = NULL
        ...
	DLGShowDialog (Dlg, FALSE);
  ----------------------------------------------------------------------*/
void DLGShowDialog (DIALOG *Dlg, dual State)
{
  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR ("dialog is empty/uninitialized")

  if (Dlg->Shell == NULL)
    PRO_ABORT ("dialog is not open");

  if (TOOLScriptRunning())
    PRO_ABORT ("script running")

  if (! State)
  {
    XtUnmapWidget (Dlg->Shell);
    XtUnmanageChild (Dlg->Shell);                 /* Dialog verstecken */
  }
  else 
  {
    XtManageChild (Dlg->Shell);                   /* Dlg. sichtbar machen */
    XtMapWidget (Dlg->Shell);
  }

  XFlush (XDisplay);

  PRO_END
}


void DLGSetCheckFunction (DIALOG *Dlg, CHECKFUNCTION CheckFct)
{
  PRO_START ("")

  if (Dlg == NULL)
    PRO_ABORT ("Empty dialog")

  Dlg->CheckFct = CheckFct;

  PRO_END
}


void DLGSetHelpFunction (DIALOG *Dlg, CALLFUNCTION HelpFct)
{
  int count;

  PRO_START ("")

  if (Dlg == NULL)
    PRO_ABORT ("dialog is empty/uninitialized")

  Dlg->HelpFct = HelpFct;

  if (! Dlg->Buttons.changed)
  {
    count = 0;
    while (count < Dlg->Buttons.Number && 
	   Dlg->Buttons.Data[count].Type != HELP_ID)
      count++;

    if (count == Dlg->Buttons.Number)
    {
      Dlg->Buttons.Number++;
      Dlg->Buttons.Data = (DLGBUTTON*)realloc (Dlg->Buttons.Data, 
					       sizeof(DLGBUTTON)*
					       Dlg->Buttons.Number);
      Dlg->Buttons.Data[count].ID     = DlgHELP;
      Dlg->Buttons.Data[count].Type   = HELP_ID;
      Dlg->Buttons.Data[count].Number = 2;
      strcpy (Dlg->Buttons.Data[count].Name, "help");
      Dlg->Buttons.Data[count].FPtr   = NULL;
      Dlg->Buttons.Data[count].Dlg    = Dlg;
      Dlg->Buttons.Data[count].Button = NULL;
    }
  }

  PRO_END
}


static void DLG_SetBoxDefault (DIALOG *Dlg, DLGBOX *Box)
{
  dual     changed = FALSE;
  int      countX, countY;
  DLGELEM *Elem;
  DLGBOX  *Ptr;

  if (Box == NULL)
    return;

  /* check, if box is non-empty item-box: */
  if (Box->Dir == BOXELEM && (Elem = Box->Item) != NULL)
  {
    switch (Elem->Type)
    {
      case DlgBOOL :
	if ((changed = (Elem->Value.drec.Value != Elem->Value.drec.Default)))
	  DLGSetBool (Dlg, Box->BoxID, Elem->Value.drec.Default);
	break;

      case DlgINT :
	if ((changed = (Elem->Value.irec.Value != Elem->Value.irec.Default)))
	  DLGSetInt (Dlg, Box->BoxID, Elem->Value.irec.Default);
	break;

      case DlgDOUBLE :
	if ((changed = (Elem->Value.rrec.Value != Elem->Value.rrec.Default)))
	  DLGSetDouble (Dlg, Box->BoxID, Elem->Value.rrec.Default);
	break;

      case DlgTEXT :
	if ((changed = (strcmp(Elem->Value.srec.Text,
			       Elem->Value.srec.Default) != 0)))
	  DLGSetText (Dlg, Box->BoxID, Elem->Value.srec.Default);
	break;

      case DlgSCALE :
	if ((changed = (Elem->Value.zrec.Val != Elem->Value.zrec.Default)))
	  DLGSetScale (Dlg, Box->BoxID, Elem->Value.zrec.Default);
	break;

      case DlgBAR :
	if ((changed = (Elem->Value.brec.Val != Elem->Value.brec.Default)))
	  DLGSetBar (Dlg, Box->BoxID, Elem->Value.brec.Default);
	break;

      case DlgUPDN :
	if ((changed = (Elem->Value.urec.Val != Elem->Value.urec.Default)))
	  DLGSetUpDn (Dlg, Box->BoxID, Elem->Value.urec.Default);
	break;

      case DlgLIST :
	if ((changed = (Elem->Value.lrec.Result != Elem->Value.lrec.Default)))
	  DLGSetList (Dlg, Box->BoxID, Elem->Value.lrec.Default);
	break;

      case DlgOMENU :
	if ((changed = (Elem->Value.orec.Result != Elem->Value.orec.Default)))
	  DLGSetOMenu (Dlg, Box->BoxID, Elem->Value.orec.Default);
	break;

      case DlgTBOX :
	{
	  int count = 0;
	  
	  while (count < Elem->Value.trec.Number &&
		 Elem->Value.trec.Toggle[count].Value ==
		   Elem->Value.trec.Default[count])
	    count++;
	  if ((changed = (count < Elem->Value.trec.Number)))
	    DLGSetTBox (Dlg, Box->BoxID, Elem->Value.trec.Default);
	}
	break;

      case DlgRBOX :
	{
	  int count = 0;

	  while (count < Elem->Value.trec.Number && 
		 Elem->Value.trec.Default[count] == FALSE)
	    count++;
	  if ((changed = (count < Elem->Value.trec.Number &&
			  Elem->Value.trec.Toggle[count].Value != TRUE)))
	    DLGSetRBox (Dlg, Box->BoxID, count);
	}
	break;

      case DlgDIAL :
	if ((changed = (Elem->Value.erec.Val != Elem->Value.erec.Default)))
	  DLGSetDial (Dlg, Box->BoxID, Elem->Value.erec.Default);
	break;

      case DlgSEGM :
	if ((changed = (Elem->Value.hrec.Val != Elem->Value.hrec.Default)))
	  DLGSetSegment (Dlg, Box->BoxID, Elem->Value.hrec.Default);
	break;

      case DlgNOTE :
	{
	  NOTEPAGE *Page;

	  Page = Elem->Value.nrec.Page;
	  while (Page != NULL)
	  {
	    DLG_SetBoxDefault (Dlg, Page->Box);
	    Page = Page->Next;
	  }
	}
	break;

#if 0
	/*  TODO: */
      case DlgTREE :
	DLGSetTree (Dlg, Box->BoxID, Elem->Value.krec.Default);
	break;
#endif

      case DlgLABEL :
      case DlgPROC :
      case DlgARROW :
      case DlgPIC :
      case DlgIMAGE :
      case DlgUSER :
      default :
	break;
    }
    /* The callback routine might be moved in DLGSet...-functions. But     */
    /* this might produce endless recursions if two boxes changes the      */
    /* value of each other.                                                */
    if (changed)
    {
      PRO1 ("Box %d set to default value\n", Elem->BoxID)
      DLG_RunBoxCB (Dlg, Elem);
    }
  }
  else
    for (countY = 0; countY < Box->Rows; ++countY)
      for (countX = 0; countX < Box->Cols; ++countX)
	if ((Ptr = Box->List[countX][countY]) != NULL)
	  DLG_SetBoxDefault (Dlg, Ptr);
}


void DLGSetDefault (DIALOG *Dlg, int BoxID)
{
  DLGBOX *Ptr;

  PRO_START ("")
  PRO1 ("\\%d...", BoxID)

  if (Dlg == NULL)
    PRO_ERROR ("Invalid/empty dialog")

  /* Box suchen: */
  if ((Ptr = DLG_GetBox(Dlg->Box, DLGBox(BoxID))) == NULL)
    PRO_ERROR_VAL ("Dialog contains no BoxID %d", BoxID)

  DLG_SetBoxDefault (Dlg, Ptr);

  PRO_END
}  



