#include <unistd.h>
#include <math.h>

#include <X11/keysym.h>

#include "dlgCB.h"
#include "dlgFS.h"

#include "dlgDial.h"

#ifdef USER_INPUT
#include <ctype.h>
#endif

#define BS  8             /* BackSpace */
#define DEL 127           /* Delete    */


/* global timer-data, because only one repeating at a time possible */
TIMER_DATA Timer = {TRUE, NULL, NULL, 0, 0, NULL, 0};


/* ************************************************** */
/* *** Callback-Routinen der einzelnen Item-Typen *** */
/* ************************************************** */

#ifdef USER_INPUT
  static int  Move = 1;
#endif
static Arg args[NUM_ARGS];


static int DLG_RunCB (DLGFUNCTION FPtr, DIALOG *Dlg, int BoxID)
{
  int CBResult = DlgCONFIRM;

  DLG_UpdateSizeAndPos (Dlg);

  if (FPtr != NULL)
  {
    (Dlg->locked)++;
    CBResult = FPtr (Dlg, BoxID);
    (Dlg->locked)--;
  }

  return CBResult;
}


int DLG_RunBoxCB (DIALOG *Dlg, DLGELEM *Elem)
{
  int CBResult = DlgCONFIRM;

  switch (Elem->Type)
  {
    case DlgBOOL :
      CBResult = DLG_RunCB (Elem->Value.drec.FPtr, Dlg, Elem->BoxID);
      break;

    case DlgINT :
      CBResult = DLG_RunCB (Elem->Value.irec.FPtr, Dlg, Elem->BoxID);
      break;
      
    case DlgDOUBLE :
      CBResult = DLG_RunCB (Elem->Value.rrec.FPtr, Dlg, Elem->BoxID);
      break;

    case DlgTEXT :
      CBResult = DLG_RunCB (Elem->Value.srec.FPtr, Dlg, Elem->BoxID);
      break;

    case DlgSCALE :
      CBResult = DLG_RunCB (Elem->Value.zrec.FPtr, Dlg, Elem->BoxID);
      break;

    case DlgBAR :
      CBResult = DLG_RunCB (Elem->Value.brec.FPtr, Dlg, Elem->BoxID);
      break;
    
    case DlgUPDN :
      CBResult = DLG_RunCB (Elem->Value.urec.FPtr, Dlg, Elem->BoxID);
      break;

    case DlgLIST :
      CBResult = DLG_RunCB (Elem->Value.lrec.FPtr, Dlg, Elem->BoxID);
      break;

    case DlgOMENU :
      CBResult = DLG_RunCB (Elem->Value.orec.FPtr, Dlg, Elem->BoxID);
      break;

    case DlgTBOX :
    case DlgRBOX :
      CBResult = DLG_RunCB (Elem->Value.trec.FPtr, Dlg, Elem->BoxID);
      break;
    
    case DlgPROC :
      CBResult = DLG_RunCB (Elem->Value.crec.FPtr, Dlg, Elem->BoxID);
      break;

    case DlgARROW :
      CBResult = DLG_RunCB (Elem->Value.arec.FPtr, Dlg, Elem->BoxID);
      break;

    case DlgPIC :
      CBResult = DLG_RunCB (Elem->Value.prec.FPtr, Dlg, Elem->BoxID);
      break;
     
    case DlgIMAGE :
      CBResult = DLG_RunCB (Elem->Value.xrec.FPtr, Dlg, Elem->BoxID);
      break;
    
    case DlgDIAL :
      CBResult = DLG_RunCB (Elem->Value.erec.FPtr, Dlg, Elem->BoxID);
      break;
    
    case DlgSEGM :
      CBResult = DLG_RunCB (Elem->Value.hrec.FPtr, Dlg, Elem->BoxID);
      break;

    case DlgFSEL :
    case DlgDIRSEL :
      CBResult = DLG_RunCB (Elem->Value.grec.FPtr, Dlg, Elem->BoxID);
      break;

    case DlgNOTE :
      CBResult = DLG_RunCB (Elem->Value.nrec.FPtr, Dlg, Elem->BoxID);
      break;
    
    case DlgTREE :
      CBResult = DLG_RunCB (Elem->Value.krec.FPtr, Dlg, Elem->BoxID);
      break;

    case DlgLABEL :
    case DlgUSER :    
    default :
      break;
  }

  return CBResult;
}


void DLG_TextCB (Widget widget, DLGELEM *Elem, XmAnyCallbackStruct *Data)
{
  char *Text = NULL, *Backup;

  if (Elem == NULL)
    return;

  Backup = Elem->Value.srec.Text;

  XtVaGetValues (widget, XmNvalue, &Text, NULL);
  Elem->Value.srec.Text = DLG_strcp (Text);

  PRO1 ("TEXT: set to [%s]\n", Elem->Value.srec.Text)

  if (DLG_RunBoxCB(Elem->Dlg, Elem) == DlgDENY)
    DLGSetText (Elem->Dlg, Elem->BoxID, Backup);
  elsif (Backup != NULL)
    free (Backup);
}


void DLG_ProcCB (Widget widget, DLGELEM *Elem, XtPointer Data)
{
  if (Elem == NULL)
    return;

  PRO0 ("PROC: called\n")

  DLG_RunBoxCB (Elem->Dlg, Elem);
}


void DLG_ArrowCB (Widget widget, DLGELEM *Elem, XtPointer Data)
{
  if (Elem == NULL)
    return;

  PRO0 ("ARROW: called\n")

  DLG_RunBoxCB (Elem->Dlg, Elem);
}


void DLG_DualCB (Widget widget, DLGELEM *Elem, XtPointer Data)
{
  dual    Backup;
  Boolean Result;

  if (Elem == NULL)
    return;

  Backup = Elem->Value.drec.Value;

  XtVaGetValues (widget, XmNset, &Result, NULL);
  Elem->Value.drec.Value = (dual)Result;
 
  PRO2 ("%s set to %c\n", Elem->Name, (dual)Result ? 'T':'F')

  if (DLG_RunBoxCB(Elem->Dlg, Elem) == DlgDENY)
    DLGSetBool (Elem->Dlg, Elem->BoxID, Backup);
}

#ifndef VAL_INPUT

void DLG_IntCB (Widget widget, DLGELEM *Elem, 
	                       XmTextVerifyCallbackStruct *Data)
{
  static int   Backup = 0;
         char  New[DLG_VLEN];
         char *ValStr;
         char *ErrPos = NULL;

  if (Data == NULL || Elem == NULL)
    return;

#if 0
/* XmCR_MOVING_INSERTING_CURSOR, XmCR_VALUE_CHANGED */
if (Data->reason == XmCR_MODIFYING_TEXT_VALUE)
  printf ("LENGTH %d: '%s'\n", Data->text->length, Data->text->ptr);

  printf ("REASON: %d\n", Data->reason);
#endif

  if (Data->reason == XmCR_FOCUS)
  {
    Backup = Elem->Value.irec.Value;
    XtSetArg (args[0], XmNcursorPosition, Elem->Value.irec.Digits);
    XtSetValues (Elem->WidgetID, args, 1);
    return;
  }

  ValStr = XmTextFieldGetString (widget);
  DLG_strncp (New, ValStr, DLG_VLEN);
  Elem->Value.irec.Value = (int)strtol (New, &ErrPos, 10);
  if (ErrPos == New)
    Elem->Value.irec.Value = 0;
  elsif (ErrPos != NULL && ErrPos[0] != '\0')
  {
    PRO1 ("INT-Error: '%s'\n", New);
    XBell (XDisplay, 50);
    Elem->Value.irec.Value = 0;
    XtVaSetValues (Elem->WidgetID, XmNvalue, "0", NULL);
  }
  XtFree (ValStr);

  PRO1 ("INT: set to %d\n", Elem->Value.irec.Value)

  if (Data->reason == XmCR_LOSING_FOCUS)
  {
    if (ErrPos == New)
    {
      DLG_Val2Str (New, Elem->Value.irec.Value, Elem->Value.irec.Digits, 0); 
      XtVaSetValues (Elem->WidgetID, XmNvalue, New, NULL);
    }
    if (DLG_RunBoxCB(Elem->Dlg, Elem) == DlgDENY)
      DLGSetInt (Elem->Dlg, Elem->BoxID, Backup);
  }
}  

#else

static dual DLG_InsertChar (char *Target, char *Source, int Pos, char Letter)
{
  int count = 0, len;

  len = strlen (Source);
  if (Source[0] != ' ' || Pos > len)
  {
    strcpy (Target, Source);
    return FALSE;
  }
  else
  {
    while (count < Pos-1 && count < len-1)
    {
      Target[count] = Source[count+1];
      count++;
    }
    Target[count++] = Letter;
    while (count < len)
    {
      Target[count] = Source[count];
      count++;
    }
    Target[count] = '\0';
  }
  return TRUE;
}

static dual DLG_DelChar (char *Target, char *Source, int Pos)
{
  int count, len;

  len = strlen (Source);
  if (Pos >= len)
  {
    strcpy (Target, Source);
    return FALSE;
  }
  else
  {
    count = len;
    while (count > Pos && count > 0)
    {
      Target[count] = Source[count];
      count--;
    }
    while (count > 0)
    {
      Target[count] = Source[count-1];
      count--;
    }
    Target[0] = ' ';
  }
  return TRUE;
}


#if USER_INPUT
void DLG_MoveCB (Widget widget, DLGELEM *Elem,
		                XmTextVerifyCallbackStruct *Data)
{
  char *Text;

  if (Elem == NULL || Data == NULL)
    return;

  if (Move < 1)
  {
    XtSetArg (args[0], XmNverifyBell, False);
    XtSetValues (widget, args, 1);
    if (Move < 0)
      Data->doit = FALSE;
    Move = 1;
    return;
  }

  PRO0 ("MOVING: ")

  Text = XmTextGetString (widget);
  while (Text[Data->newInsert] == ' ')
    (Data->newInsert)++;

  PRO1 ("\\Pos: %d\n", Data->newInsert)

  Move = 0;
  XmTextSetInsertionPosition (widget, Data->newInsert);
/*
    TODO: 
    prfen, ob neue Pos sichtbar ist...
    if (XmTextGetTopCharacter(widget) > 0)
      XmTextSetTopCharacter (widget, 0);
*/

  Data->doit = False;
}
#endif


void DLG_IntCB (Widget widget, DLGELEM *Elem, XmTextVerifyCallbackStruct *Data)
{
  int   len, Result;
  int   Pos, ePos;
  char  letter = ' ';
  char  NewText[32];
  char *Text;
 
  if (Elem == NULL || Data == NULL)
    return;

  Text = XmTextGetString (widget);
  PRO0 ("\\\n")
  PRO2 ("startPos: %d endPos: %d\n", Data->startPos, Data->endPos)
  PRO3 ("Text: <%s> currIns: %d newIns: %d\n",
	Text, Data->currInsert, Data->newInsert)

  PRO0 ("MODIFYING: ")

  if (Data->text->length > 1)
  {
    /* TODO:
       paste-operation bearbeiten
    */
    Data->doit = False;
    return;
  }
  else
  {
    len = strlen(Text);

    if (Data->text->length > 0 && Data->text->ptr != NULL)
      letter = Data->text->ptr[0];
    elsif (Data->currInsert == Data->newInsert)
      letter = (Data->currInsert == Data->endPos)? BS : DEL;

    Pos = 0;
    while (Pos < len && Text[Pos] == ' ')
      Pos++;
    ePos = Pos;
    while (ePos < len && toupper(Text[ePos]) != 'E')
      ePos++;

    switch (letter)
    {
      case '0'...'9' :
	PRO1 ("\\DIGIT: %c\n", letter)
	if (Text[len-1] == '0' && (Text[len-2] == '-' || Text[len-2] == ' '))
	{
	  strcpy (NewText, Text);
	  NewText[len-1] = letter;
	}
	else
	  Data->doit = DLG_InsertChar (NewText, Text, 
				       Data->currInsert, letter);
	break;

      case BS :
	PRO0 ("\\BS\n")
	Data->doit = DLG_DelChar (NewText, Text, Data->currInsert-1);
	break;

      case DEL :
	PRO0 ("\\DELETE\n")
	Data->doit = DLG_DelChar (NewText, Text, Data->currInsert);
	XmTextSetInsertionPosition (widget, Data->newInsert+1);
	break;

      case '-' :
	PRO0 ("\\MINUS\n")
	if (Data->currInsert > ePos)
	  if (Text[ePos+1] != '-')
	    Data->doit = DLG_InsertChar (NewText, Text, ePos+1, '-');
	  else
	    DLG_DelChar (NewText, Text, ePos+1);
	else
	{
	  strcpy (NewText, Text);

	  if (Text[Pos] == '-')
	    NewText[Pos] = ' ';
	  elsif (Pos > 0)
	    NewText[Pos-1] = '-';
	  else
            Data->doit = FALSE;
	}
	Move = -1;
	break;

      case '+' :
	PRO0 ("\\PLUS\n")
	strcpy (NewText, Text);

	if (Data->currInsert > ePos)
	{
	  if (Text[ePos+1] == '-')
	    DLG_DelChar (NewText, Text, ePos+1);
	}
	elsif (Text[Pos] == '-')
	    NewText[Pos] = ' ';

	Move = -1;
	break;        

      case 'e' :
      case 'E' :
	PRO0 ("\\EXPONENT\n")
	if ((Data->doit = (ePos == len)))
	  Data->doit = DLG_InsertChar (NewText, Text, Data->currInsert, 'E');
	break;

      default :
	PRO0 ("\\INVALID CHAR\n")
	Data->doit = False;
    }
    if (NewText[Elem->Value.irec.Digits-1] == ' ')
      NewText[Elem->Value.irec.Digits-1] = '0';
    NewText[Elem->Value.irec.Digits] = '\0';

    Elem->Value.irec.Value = 0;
    Result = strtol (NewText, (char**)NULL, 10);
    if (errno == ERANGE)
      Data->doit = False;

    if (! Data->doit)
    {
      XtSetArg (args[0], XmNverifyBell, True);
      XtSetValues (widget, args, 1);
      Move = 1;
      return;
    }
 
    if (Data->startPos != Data->endPos)
      Move = -1;
    
    Elem->Value.irec.Value = Result;

    if (Data->text->length > 0 && Data->text->ptr != NULL)
      XtFree (Data->text->ptr);
    Data->text->ptr = XtMalloc (sizeof(char) * (Elem->Value.irec.Digits+1));
    strcpy (Data->text->ptr, NewText);
    Data->text->ptr[Elem->Value.irec.Digits] = '\0';

    Data->text->length = Elem->Value.irec.Digits;
    Data->startPos     = 0;
    Data->endPos       = Data->text->length;
/*
    XmTextSetTopCharacter (widget, 0);
*/
  }
  PRO2 ("NEW Text: <%s> Wert: %d\n", Data->text->ptr, Result)
}

void DLG_IntCallCB (Widget widget, DLGELEM *Elem, XtPointer Data)
{
  if (Elem == NULL)
    return;

  DLG_RunBoxCB (Elem->Dlg, Elem->BoxID);
}
#endif


void DLG_DoubleCB (Widget widget, DLGELEM *Elem, 
		                  XmTextVerifyCallbackStruct *Data)
{
  static double  Backup = 0.0; 
         char    New[DLG_VLEN];
         char   *ValStr;
         char   *ErrPos = NULL;

  if (Data == NULL || Elem == NULL)
    return;

  if (Data->reason == XmCR_FOCUS)
  {
    Backup = Elem->Value.rrec.Value;
    XtSetArg (args[0], XmNcursorPosition, 
	               Elem->Value.rrec.Digits-Elem->Value.rrec.Decimals-1);
    XtSetValues (Elem->WidgetID, args, 1);
    return;
  }

  ValStr = XmTextGetString (widget);
  DLG_strncp (New, ValStr, DLG_VLEN);
  Elem->Value.rrec.Value = strtod (New, &ErrPos);
  if (ErrPos == New)
    Elem->Value.rrec.Value = 0.0;
  elsif (ErrPos != NULL && ErrPos[0] != '\0')
  {
    PRO1 ("DOUBLE-Error: '%s'\n", New);
    XBell (XDisplay, 50);
    Elem->Value.rrec.Value = 0.0;
    XtVaSetValues (Elem->WidgetID, XmNvalue, "0.0", NULL);
  }
  XtFree (ValStr);

  PRO1 ("DOUBLE: set to %f\n", Elem->Value.rrec.Value)

  if (Data->reason == XmCR_LOSING_FOCUS)
  {
    if (ErrPos == New)
    {
      DLG_Val2Str (New, Elem->Value.rrec.Value, Elem->Value.rrec.Digits,
		        Elem->Value.rrec.Decimals); 
      XtVaSetValues (Elem->WidgetID, XmNvalue, New, NULL);
    }
    if (DLG_RunBoxCB(Elem->Dlg, Elem) == DlgDENY)
      DLGSetDouble (Elem->Dlg, Elem->BoxID, Backup);
  }
}


#if USER_INPUT
void DLG_DoubleValCB (Widget widget, DLGELEM *Elem, 
		                   XmTextVerifyCallbackStruct *Data)
{
  int     count;
  int     pPos, ePos;
  double  Val; 
  char    NewChar;
  char   *Text;

  if (Elem == NULL || Data == NULL)
    return;

  Text = XmTextGetString (widget);
  PRO2 ("startPos: %d endPos: %d\n", Data->startPos, Data->endPos)
  PRO3 ("Text: <%s> currIns: %d newIns: %d\n",
	Text, Data->currInsert, Data->newInsert)

  count = 0;
  while (count < Elem->Value.rrec.Digits && Text[count] != '.')
    count++;
  pPos = count;

  count = 0;
  while (count < Elem->Value.rrec.Digits && Text[count] != 'e')
    count++;
  ePos = count;

  if (Data->reason == XmCR_MODIFYING_TEXT_VALUE)
  {
    PRO0 ("MODIFYING: ")

    XtSetArg (args[0], XmNverifyBell, True);
    XtSetValues (widget, args, 1);

    if (Data->text->length > 1)      /* keine "paste"-Operationen erlauben */
    {
      Data->doit = False;
      return;
    }

    if (Data->text->length == 0 || Data->text->ptr == NULL)
    {
      Data->text->ptr = XtMalloc (sizeof(char) * (Elem->Value.rrec.Digits+1));

      PRO0 ("\\Backspace\n")
      if (Text[Data->currInsert-1] != ' ')
      {
	if (Data->currInsert-1 < pPos || Data->currInsert-1 >= ePos)
	{
	  for (count = Elem->Value.rrec.Digits-1; count > Data->currInsert-1; 
	       count--)
	    Data->text->ptr[count] = Text[count];
	  for (count = Data->currInsert-1; count > 0; count--)
	    Data->text->ptr[count] = Text[count-1];
	  Data->text->ptr[0] = ' ';
	  if (Data->text->ptr[pPos-1] == ' ')
	    Data->text->ptr[pPos-1] = '0';
	}
	elsif (Data->currInsert-1 > pPos)
	{
	  for (count = 0; count < Data->currInsert-1; ++count)
	    Data->text->ptr[count] = Text[count];
	  for (count = Data->currInsert-1; count < ePos-1; ++count)
	    Data->text->ptr[count] = Text[count+1];
	  Data->text->ptr[ePos-1] = '0';
	  for (count = ePos; count < Elem->Value.rrec.Digits; ++count)
	    Data->text->ptr[count] = Text[count];
	}
	else
	  Data->doit = False;
      }
      else
	Data->doit = False;

      /* NewInsert=...kann Cursorverschiebung nach links nicht aufhalten ?!? */
      /* Versuch ber Key-Accelerator */
    }
    else
    {
      NewChar = *(Data->text->ptr);
      PRO1 ("Zeichen: %c\n", NewChar)

      XtFree (Data->text->ptr);
      Data->text->ptr = XtMalloc (sizeof(char) * (Elem->Value.rrec.Digits+1));

      if (isdigit (NewChar))
	if ((Data->currInsert <= pPos || Data->currInsert > ePos) &&
	    Text[0] == ' ')
	{
	  count = 0;
	  while (Text[count+1] == ' ')
	  {
	    Data->text->ptr[count] = ' ';
	    count++;
	  }
	  if (Text[count+1] == '0' && Data->currInsert <= pPos)
	  {
	    Data->text->ptr[count] = ' ';
	    count++;
	  }
	  elsif (Text[count+1] == '-' && Text[count+2] == '0' &&
		 Data->currInsert <= pPos)
	  {
	    Data->text->ptr[count] = ' ';
	    Data->text->ptr[count+1] = '-';
	    count += 2;
	  }
	  while (count < Data->currInsert-1)
	  {
	    Data->text->ptr[count] = Text[count+1];
	    count++;
	  }
	  Data->text->ptr[count] = NewChar;
	  for (count = Data->currInsert; 
	       count < Elem->Value.rrec.Digits; ++count)
	    Data->text->ptr[count] = Text[count];
	}
        elsif (Data->currInsert > pPos)
	{
	  for (count = Elem->Value.rrec.Digits-1; count >= ePos; count--)
	    Data->text->ptr[count] = Text[count];
	  while (count > Data->currInsert)
	  {
	    Data->text->ptr[count] = Text[count-1];
	    count--;
	  }
	  Data->text->ptr[count] = NewChar;
	  for (count = Data->currInsert-1; count >= 0; count--)
	    Data->text->ptr[count] = Text[count];
	  XmTextSetInsertionPosition (widget, Data->currInsert+1);
	}
	else
	{
	  PRO0 ("To many characters\n")
	  Data->doit = False;
	}
      else
      {
	PRO1 ("ePos=%d\n", ePos)

	switch (NewChar)
	{
	  case '-' : if (Data->currInsert <= ePos)
	             {
		       count = 0;
		       while (Text[count] == ' ')
		       {
			 Data->text->ptr[count] = ' ';
			 count++;
		       }
 
		       if (Text[count] == '-')
			 Data->text->ptr[count] = ' ';
		       elsif (count == 0)
		       {
			 PRO0 ("Was nun ?\n")
			 Data->doit = False;
		       }
		       else
		       {
			 Data->text->ptr[count-1] = '-';
			 Data->text->ptr[count] = Text[count];
		       }
		       count++;
		       while (count < Elem->Value.rrec.Digits)
		       {
			 Data->text->ptr[count] = Text[count];
			 count++;
		       }
		     }
	             elsif (Text[ePos+1] == '-')
		     {
		       for (count = Elem->Value.rrec.Digits-1; count > ePos+1;
			    count--)
			 Data->text->ptr[count] = Text[count];
		       for (count = ePos+1; count > 0; count--)
			 Data->text->ptr[count] = Text[count-1];
		       Data->text->ptr[0] = ' ';
		     }
	             elsif (Text[0] == ' ')
		     {
		       for (count = 0; count < ePos; ++count)
			 Data->text->ptr[count] = Text[count+1];
		       Data->text->ptr[count] = '-';
		       for (count = ePos+1; count < Elem->Value.rrec.Digits;
			    ++count)
			 Data->text->ptr[count] = Text[count];
		     }
	             else
		     {
		       PRO0 ("To many characters for operation\n")
		       Data->doit = False;
		     }
	             break;
	  case '+' : if (Data->currInsert <= ePos)
	             {
		       count = 0;
		       while (Text[count] == ' ')
		       {
			 Data->text->ptr[count] = ' ';
			 count++;
		       }
		       if (Text[count] == '-')
			 Data->text->ptr[count] = ' ';
		       else
			 Data->text->ptr[count] = Text[count];
		       count++;
		       while (count < Elem->Value.rrec.Digits)
		       {
			 Data->text->ptr[count] = Text[count];
			 count++;
		       }
		     }
	             elsif (Text[ePos+1] == '-')
		     {
		       for (count = Elem->Value.rrec.Digits; count > ePos+1;
			    count--)
			 Data->text->ptr[count] = Text[count];
		       for (count = ePos+1; count > 0; count--)
			 Data->text->ptr[count] = Text[count-1];
		       Data->text->ptr[0] = ' ';
		     }
	             break;
	  case 'e' : if (ePos == Elem->Value.rrec.Digits && Text[0] == ' '
		     /* && (Text[pPos-1] != '0' || Text[pPos-2] != ' ') */ )
	             {
		       for (count = 0; count < Elem->Value.rrec.Digits-1;
			    ++count)
			 Data->text->ptr[count] = Text[count+1];
		       Data->text->ptr[count] = 'e';
		       XmTextSetInsertionPosition (widget, 
						   Elem->Value.irec.Digits);
                     }
	             else
		       Data->doit = False;
	             break;
	  case '.' : Data->text->length = 0;
	             Data->newInsert = pPos + 1;
	             XtSetArg (args[0], XmNverifyBell, False);
	             XtSetValues (widget, args, 1);
	             XmTextSetInsertionPosition (widget, pPos+1);
	             return;
	             break;
	   default : 
	             PRO0 ("invalid Character\n")
	             Data->doit = False;
	}
      }
    }
    if (Data->doit == False)
      return;

    Data->text->ptr[Elem->Value.rrec.Digits] = '\0';

    Data->text->length = Elem->Value.rrec.Digits;
    Data->startPos = 0;
    Data->endPos   = Data->text->length;

    PRO2 ("Inserting <%s>(%d)\n", Data->text->ptr, Data->text->length)

    XmTextSetTopCharacter (widget, 0);

    Val = atof (Data->text->ptr);
    Elem->Value.rrec.Value = Val;
    PRO2 ("neuer Text: <%s> Wert: %f\n", Data->text->ptr, Val)
  }
}


void DLG_DoubleCallCB (Widget widget, DLGELEM *Elem, XtPointer Data)
{
  if (Elem == NULL)
    return;

  DLG_RunBoxCB (Elem->Dlg, Elem->Value.BoxID);
}
#endif


void DLG_ScaleCB (Widget widget, DLGELEM *Elem, XtPointer Data)
{
  int    Result;
  double Backup;

  if (Data == NULL || Elem == NULL)
    return;

  Backup = Elem->Value.zrec.Val;

  XtSetArg (args[0], XmNvalue, &Result);
  XtGetValues (widget, args, 1);
  Elem->Value.zrec.Val = Result / pow(10.0, Elem->Value.zrec.Decimals);

  PRO1 ("[SCALE]: set to %f\n", Elem->Value.zrec.Val)

  if (DLG_RunBoxCB(Elem->Dlg, Elem) == DlgDENY)
    DLGSetScale (Elem->Dlg, Elem->BoxID, Backup);
}


void DLG_BarCB (Widget widget, DLGELEM *Elem, XtPointer Data)
{
  int    Result;
  double Backup;

  if (Data == NULL || Elem == NULL)
    return;

  Backup = Elem->Value.brec.Val;

  XtSetArg (args[0], XmNvalue, &Result);
  XtGetValues (widget, args, 1);
  Elem->Value.brec.Val = (double)Result / 100.0;

  PRO1 ("[BAR] set to %7.2f\n", Elem->Value.brec.Val)

  if (DLG_RunBoxCB(Elem->Dlg, Elem) == DlgDENY)
    DLGSetBar (Elem->Dlg, Elem->BoxID, Backup);
}


static void DLG_UpTimer (DLGELEM *Elem, int Param)
{
  double Backup;
  char   New[DLG_VLEN];

  if (Elem->Value.urec.Val >= Elem->Value.urec.Max &&
      Elem->Value.urec.Min < Elem->Value.urec.Max)
    return;

  Backup = Elem->Value.urec.Val;

  if (Elem->Value.urec.Val + Elem->Value.urec.Step < Elem->Value.urec.Max ||
      Elem->Value.urec.Min >= Elem->Value.urec.Max)
    Elem->Value.urec.Val += Elem->Value.urec.Step;
  else
    Elem->Value.urec.Val = Elem->Value.urec.Max;

  PRO1 ("[UP]: set to %f\n", Elem->Value.urec.Val)

  DLG_Val2Str (New, Elem->Value.urec.Val, Elem->Value.urec.Length,
	            Elem->Value.urec.Decimals); 
  XtVaSetValues (Elem->WidgetID, XmNvalue, New, NULL);

  if (DLG_RunBoxCB(Elem->Dlg, Elem) == DlgDENY)
    DLGSetUpDn (Elem->Dlg, Elem->BoxID, Backup);
}


void DLG_UpCB (Widget widget, DLGELEM *Elem, XmAnyCallbackStruct *Data)
{
  if (Data == NULL || Elem == NULL)
    return;

  Timer.widget   = widget;
  Timer.Data     = (void*)Elem;
  Timer.start    = TRUE;
  Timer.Function = (CBFUNCTION)DLG_UpTimer;
  Timer.Param    = 0;

  XtSetArg (args[0], XmNwidth,  &Timer.SizeX);
  XtSetArg (args[1], XmNheight, &Timer.SizeY);
  XtGetValues (widget, args, 2);

  DLG_Timer (&Timer, (XtIntervalId)0);
}


static void DLG_DnTimer (DLGELEM *Elem, int Param)
{
  double Backup;
  char   New[DLG_VLEN];

  if (Elem->Value.urec.Val <= Elem->Value.urec.Min &&
      Elem->Value.urec.Min <= Elem->Value.urec.Max)
    return;

  Backup = Elem->Value.urec.Val;

  if (Elem->Value.urec.Val - Elem->Value.urec.Step > Elem->Value.urec.Min ||
      Elem->Value.urec.Min >= Elem->Value.urec.Max)
    Elem->Value.urec.Val -= Elem->Value.urec.Step;
  else
    Elem->Value.urec.Val = Elem->Value.urec.Min;

  PRO1 ("[DOWN]: set to %f\n", Elem->Value.urec.Val)

  DLG_Val2Str (New, Elem->Value.urec.Val, Elem->Value.urec.Length, 
		    Elem->Value.urec.Decimals);
  XtVaSetValues (Elem->WidgetID, XmNvalue, New, NULL);

  if (DLG_RunBoxCB(Elem->Dlg, Elem) == DlgDENY)
    DLGSetUpDn (Elem->Dlg, Elem->BoxID, Backup);
}

void DLG_DnCB (Widget widget, DLGELEM *Elem, XtPointer Data)
{
  if (Data == NULL || Elem == NULL)
    return;

  Timer.widget   = widget;
  Timer.Data     = (void*)Elem;
  Timer.start    = TRUE;
  Timer.Function = (CBFUNCTION)DLG_DnTimer;
  Timer.Param    = 0;

  XtSetArg (args[0], XmNwidth,  &Timer.SizeX);
  XtSetArg (args[1], XmNheight, &Timer.SizeY);
  XtGetValues (widget, args, 2);

  DLG_Timer (&Timer, (XtIntervalId)0);
}


void DLG_UpDnValCB (Widget widget, DLGELEM *Elem, 
		                   XmTextVerifyCallbackStruct *Data)
{
          double  Val;
  static  double  Backup = 0.0;
          char    New[DLG_VLEN];
          char   *ErrPos = NULL;

  if (Data == NULL || Elem == NULL)
    return;

  if (Data->reason == XmCR_FOCUS)
  {
    Backup = Elem->Value.urec.Val;
    return;
  }

  if (Data->reason == XmCR_LOSING_FOCUS)
  {
    DLG_strncp (New, XmTextGetString(widget), DLG_VLEN);
    Val = strtod (New, &ErrPos);

    if (Elem->Value.urec.Min < Elem->Value.urec.Max)
      Elem->Value.urec.Val = 
	DLG_UPDN_VAL(Elem->Value.urec.Min,  Elem->Value.urec.Max, 
		     Elem->Value.urec.Step, Val);
    else
      Elem->Value.urec.Val = Val;

    PRO1 ("UPDN: set to %f\n", Elem->Value.urec.Val)

    DLG_Val2Str (New, Elem->Value.urec.Val, Elem->Value.urec.Length,
		      Elem->Value.urec.Decimals); 
    XtSetArg (args[0], XmNvalue, New);
    XtSetValues (Elem->WidgetID, args, 1);

    if (DLG_RunBoxCB(Elem->Dlg, Elem) == DlgDENY)
      DLGSetUpDn (Elem->Dlg, Elem->BoxID, Backup);
  }
}  


void DLG_ListCB (Widget widget, DLGELEM *Elem, XmListCallbackStruct *Data)
{
  int Backup;

  if (Data == NULL || Elem == NULL)
    return;

  Backup = Elem->Value.lrec.Result;

  if (Elem->Value.lrec.Result == Data->item_position-1)
    Elem->Value.lrec.Result = -1;
  else
    Elem->Value.lrec.Result = Data->item_position-1;

  if (Elem->Value.lrec.Result > 0)
    if (Elem->Value.lrec.Result < Elem->Value.lrec.Number)
      PRO1 ("[LIST]: %s\n", Elem->Value.lrec.Item[Elem->Value.lrec.Result])

  if (DLG_RunBoxCB(Elem->Dlg, Elem) == DlgDENY)
    DLGSetList (Elem->Dlg, Elem->BoxID, Backup);
}


void DLG_ToggleCB (Widget widget, SWITCHREC *UserData, XtPointer Data)
{
  dual    Backup;
  Boolean Result;
  
  if (Data == NULL || UserData == NULL)
    return;

  Backup = UserData->Value;

  XtSetArg (args[0], XmNset, &Result);
  XtGetValues (widget, args, 1);
  UserData->Value = (dual)Result;
 
  PRO2 ("[TOGGLE]: %s set to %c\n", UserData->Item, UserData->Value ? 'T':'F')

  if (DLG_RunBoxCB(UserData->Parent->Dlg, UserData->Parent) == DlgDENY)
  {
    UserData->Value = Backup;
    XtVaSetValues (widget, XmNset, (Boolean)Backup, NULL);
  }
}

void DLG_RadioCB (Widget widget, SWITCHREC *UserData, XtPointer Data)
{
  Boolean Result;
  int     count, Backup;

  if (Data == NULL || UserData == NULL)
    return;

  count = 0;
  while (count < UserData->Parent->Value.trec.Number && 
	 UserData->Parent->Value.trec.Toggle[count].Value == FALSE)
    count++;
  Backup = count;

  XtSetArg (args[0], XmNset, &Result);
  XtGetValues (widget, args, 1);

  if ((dual)Result)
  {
    for (count = 0; count < UserData->Parent->Value.trec.Number; ++count)
      UserData->Parent->Value.trec.Toggle[count].Value = FALSE;

    UserData->Value = (dual)Result;

    if (DLG_RunBoxCB(UserData->Parent->Dlg, UserData->Parent) == DlgDENY)
      DLGSetRBox (UserData->Parent->Dlg, UserData->Parent->BoxID, Backup);
  }
}


void DLG_MenuCB (Widget widget, ITEMREC *UserData, XtPointer Data)
{
  int Backup;

  if (Data == NULL || UserData == NULL)
    return;

  Backup = UserData->Parent->Value.orec.Result;

  UserData->Parent->Value.orec.Result = UserData->Index;

  PRO1 ("[MENU]: set to %s\n", UserData->Label)

  if (DLG_RunBoxCB(UserData->Parent->Dlg, UserData->Parent) == DlgDENY)
    DLGSetOMenu (UserData->Parent->Dlg, UserData->Parent->BoxID, Backup);
}


void DLG_PicCB (Widget widget, DLGELEM *Elem, XtPointer Data)
{
  if (Elem == NULL)
    return;

  PRO0 ("[BMAP/PICB]: called\n")

  DLG_RunBoxCB (Elem->Dlg, Elem);
}


void DLG_ImageCB (Widget widget, DLGELEM *Elem, XtPointer Data)
{
  if (Elem == NULL)
    return;

  PRO0 ("[IMAGE]: called\n")

  DLG_RunBoxCB (Elem->Dlg, Elem);
}


#if (XmVERSION>=2)
  void DLG_NoteCB (Widget widget, DLGELEM *Elem, 
		                  XmNotebookCallbackStruct *Data)
  {
    if (Elem == NULL)
      return;

    PRO1 ("[NOTEBOOK]: called: %d\n", Data->page_number)

    Elem->Value.nrec.ActPage = Data->page_number;

    DLG_RunBoxCB (Elem->Dlg, Elem);
  }


  static TREEITEM *DLG_FindTreeItem (TREEITEM *List, Widget widget)
  {
    TREEITEM *Ptr, *Item;

    Ptr = List;
    while (Ptr != NULL && Ptr->gadget != widget)
    {
      if (Ptr->Sub != NULL && 
	  (Item = DLG_FindTreeItem(Ptr->Sub, widget)) != NULL)
	return Item;

      Ptr = Ptr->Next;
    }
    return Ptr;
  }

  void DLG_TreeCB (Widget widget, DLGELEM *Elem, 
		                  XmContainerSelectCallbackStruct *Data)
  {
    TREEITEM *Item;

    if (Data->selected_item_count != 1)
    {
      PRO1 ("Multiple Tree-Selection not supported (%d)\n",
	    Data->selected_item_count)
      return;
    }

    if ((Item = DLG_FindTreeItem(Elem->Value.krec.Data, 
				 Data->selected_items[0])) == NULL)
      return;

    Elem->Value.krec.Result = Item;
    Elem->Value.krec.Reason = DlgOTHER;

    if (Data->event != NULL)
    {
      PRO1 ("TreeCB: Type = %d\n", Data->event->type)

      /* Don't call the User Callback routine on cursor-key events */
      if (Data->event->type == KeyPress)
      {
	XKeyPressedEvent *Key = (XKeyPressedEvent *)Data->event;
	KeySym            Symbol = XLookupKeysym (Key, 0);

	if (Symbol == XK_Down || Symbol == XK_Up || 
	    Symbol == XK_Left || Symbol == XK_Right)
	  Elem->Value.krec.Reason = DlgCURSOR;
	elsif (Symbol == XK_Return)
	  Elem->Value.krec.Reason = DlgRETURN;
      }
      elsif (Data->event->type == ButtonRelease)
      {
	XButtonEvent *Button = (XButtonEvent *)Data->event;

	Elem->Value.krec.Reason = DlgMOUSE1 + (int)Button->button;
      }
    }
    DLG_RunBoxCB (Elem->Dlg, Elem);
  }
#endif


void DLG_DialCB (Widget widget, DLGELEM *Elem, DialCallbackStruct *Data)
{
  double Backup;

  if (Elem == NULL)
    return;

  Backup = Elem->Value.erec.Val;

  PRO1 ("[DIAL]: set to %d\n", Data->Value)

  Elem->Value.erec.Val = Data->Value / pow (10, Elem->Value.erec.Decimals);

  if (DLG_RunBoxCB(Elem->Dlg, Elem) == DlgDENY)
  {
    /* DLGSetDialValue (Elem->Dlg, Elem->BoxID, Backup); */
  }
}


void DLG_SegmentCB (Widget widget, DLGELEM *Elem, DialCallbackStruct *Data)
{
  double Backup;

  if (Elem == NULL)
    return;

  Backup = Elem->Value.hrec.Val;

  PRO1 ("[SEGMENT]: set to %d\n", Data->Value)

  Elem->Value.hrec.Val = Data->Value;

  if (DLG_RunBoxCB(Elem->Dlg, Elem) == DlgDENY)
    DLGSetSegment (Elem->Dlg, Elem->BoxID, Backup);
}


void DLG_FileSelCB (Widget widget, DLGELEM *Elem, XtPointer Data)
{ 
  char      DirMask[PATH_MAX+1] = "";
  char     *Text = NULL;
  XmString  Label = NULL;
  Widget    FileSel, Shell;

  if (Elem->Value.grec.Open)            /* prohibit multiple calls */
    return;

  DLGSetButtonActive (Elem->Dlg, 0, FALSE);
  Elem->Value.grec.Open = TRUE;

  Text = TOOLGetResString (XtParent(widget), "topLevelShellWidgetClass", 
			                     Elem->Name, "title", Elem->Name);
  Shell = XtVaCreatePopupShell (Elem->Name, topLevelShellWidgetClass, AppShell,
				            XmNallowShellResize, True,
				            XmNx,                FSXPos+30,
				            XmNy,                FSYPos+20,
				            XmNtopOffset,        10,
				            XmNbottomOffset,     10,
				            XmNrightOffset,      10,
				            XmNtitle,            Text,
				            NULL);

  if (Elem->Value.grec.Button == 0)
    Text = TOOLGetResString (XtParent(widget), "xmFileSelectionBoxWidgetClass",
			     Elem->Name, "dirMask", Elem->Value.grec.Mask);
  else
    Text = Elem->Value.grec.Mask;
  if (strcmp(Text, "") != 0)
    TOOLParseFileName (Text, DirMask, PATH_MAX);

  FileSel = DLG_CreateFileSel (Shell, DirMask, Elem->Value.grec.Default, 15,
			              Elem->Value.grec.Files);
  XtAddCallback (FileSel, XmNokCallback, 
		          (XtCallbackProc)DLG_FSelOkCB,     Elem);
  XtAddCallback (FileSel, XmNcancelCallback, 
		          (XtCallbackProc)DLG_FSelCancelCB, Elem);

  PRO0 ("\\Opening Sub-FileSelector...")

  XtManageChild (Shell);
  XtRealizeWidget (Shell);
  XtPopup (Shell, XtGrabNone);
  XFlush (XDisplay);

  TOOLMainLoop (AppCon, Shell);

  PRO0 ("\\finished...")
  /* Evaluate result */
  if (Elem->Value.grec.Button == 1)
  {
    Label = XmStringLtoRCreate (Elem->Value.grec.Result, NORMAL_FONT);
    XtVaSetValues (widget, XmNrecomputeSize, FALSE, XmNalignment, 
     ((int)strlen(Elem->Value.grec.Result)<Elem->Value.grec.Length)?
       XmALIGNMENT_BEGINNING:XmALIGNMENT_END,
     XmNlabelString,  Label, NULL);
    XmStringFree (Label);
    strcpy (Elem->Value.grec.Default, Elem->Value.grec.Result);
  }

  Elem->Value.grec.Open = FALSE;
  DLGSetButtonActive (Elem->Dlg, 0, TRUE);
}


void DLG_FSelChangeCB (Widget widget, DLGELEM *Elem, XmAnyCallbackStruct *Ptr)
{
  if (Elem != NULL)
    DLG_strncp (Elem->Value.grec.Result, XmTextGetString(widget), PATH_MAX);
}


void DLG_FSelOkCB (Widget widget, DLGELEM *Elem, 
		                  XmFileSelectionBoxCallbackStruct *Data)
{
  char *Result, *Backup;
  char  Directory[PATH_MAX+1] = "";

  if (Data == NULL || Elem == NULL)
    return;

  Backup = Elem->Value.grec.Result;

  Elem->Value.grec.Button    = 1;
  Elem->Value.grec.Result[0] = '\0';

  if (Elem->Value.grec.Files)
  {
    XmStringGetLtoR (Data->value, XmSTRING_DEFAULT_CHARSET, &Result);
    DLG_strncp (Elem->Value.grec.Result, Result, PATH_MAX);
    XtFree (Result);
  }

  XmStringGetLtoR (Data->mask, XmSTRING_DEFAULT_CHARSET, &Result);
  DLG_strncp (Elem->Value.grec.Mask, Result, PATH_MAX);
  if (! Elem->Value.grec.Files)
  {
    TOOLGetFileDir (Result, Directory, PATH_MAX);
    DLG_strncp (Elem->Value.grec.Result, Directory, PATH_MAX);
  }
  XtFree (Result);

  PRO1 ("[FSEL]: selected %s\n", Elem->Value.grec.Result)

  if (Elem->Value.grec.Length != 0)
  {
    XtUnmanageChild (XtParent(widget));  
    XtDestroyWidget (XtParent(widget));
  }

  if (DLG_RunBoxCB(Elem->Dlg, Elem) == DlgDENY)
  {
    if (Elem->Value.grec.Files)
      DLGSetFileSel (Elem->Dlg, Elem->BoxID, Backup);
    else
      DLGSetDirSel  (Elem->Dlg, Elem->BoxID, Backup);
  }
}


void DLG_FSelCancelCB (Widget widget, DLGELEM *Elem, XtPointer Data)
{
  if (Data == NULL || Elem == NULL)
    return;
  
  Elem->Value.grec.Button = 2;
  if (Elem->Value.grec.Length != 0)
  {
    XtUnmanageChild (XtParent(widget));  
    XtDestroyWidget (XtParent(widget));
  }

  PRO0 ("[FSEL]: canceled\n")

  DLG_RunBoxCB (Elem->Dlg, Elem);
}


/* ********************************************** */
/* *** Callback-Routinen des gesamten Dialogs *** */
/* ********************************************** */
void DLG_ButtonCB (Widget widget, DLGBUTTON *Button, XtPointer Data)
{
  int          Number;
  CALLFUNCTION CBFct;

  PRO0 ("Dialog: ")
  switch (Button->Type)
  {
    case      OK_ID : PRO0 ("\\OK...\n")
	              break;
    case  CANCEL_ID : PRO0 ("\\CANCEL...\n")
		      break;
    case   CLOSE_ID : PRO0 ("\\CLOSE...\n")
		      break;
    case    SAVE_ID : PRO0 ("\\SAVE...\n")
		      break;
    case    HELP_ID : PRO0 ("\\HELP...\n")
		      break;
    case    STOP_ID : PRO0 ("\\STOP...\n")
		      break;
    case   APPLY_ID : PRO0 ("\\APPLY...\n")
		      break;
    case  ACCEPT_ID : PRO0 ("\\ACCEPT...\n")
		      break;
    case    DONE_ID : PRO0 ("\\DONE...\n")
		      break;
    case DEFAULT_ID : PRO0 ("\\DEFAULT...\n")
		      break;
            default : PRO1 ("\\USER...: %d\n", Button->ID)
		      break;
  }

  if (Button->Dlg->locked > 0)          /* another function is running:  */
    return;                             /*   forbid closing dialog.      */

  /* Check, if the dialog values are valid */
  if (Button->Dlg->CheckFct != NULL && 
      ! Button->Dlg->CheckFct(Button->Dlg, Button->ID))
    return;

  Number = Button->Number;

  CBFct = Button->Dlg->CBFct;
  
  DLG_UpdateSizeAndPos (Button->Dlg);

  (Button->Dlg->locked)++;
  if (Button->Dlg->Buttons.Data[Number].FPtr != NULL)
    Button->Dlg->Buttons.Data[Number].FPtr (Button->Dlg, Button->ID);
  elsif (Button->Dlg->Buttons.Data[Number].Type == HELP_ID &&
	 Button->Dlg->HelpFct != NULL)
    Button->Dlg->HelpFct (Button->Dlg, Button->ID);
  elsif (Button->Dlg->Buttons.Data[Number].Type == DEFAULT_ID)
    DLGSetDefault (Button->Dlg, Button->Dlg->Box->BoxID);
  else
  {
    Button->Dlg->Buttons.Result = Button->ID;
    XtPopdown (Button->Dlg->Shell);
    XtDestroyWidget (Button->Dlg->Shell);
  }
  (Button->Dlg->locked)--;

  /* TODO:
     falls mglich: Prfen, ob der Dialog in Button-CB-Fkt gelscht wurde 
  */
  /* call CB-function for non-modal dialogs */
  if (CBFct != NULL)
    CBFct (Button->Dlg, Button->ID);
}


/* **************************************************** */
/* ** This function will be called on program-exit,  ** */
/* ** if there's still a dialog open                 ** */
/* ** INPUT:  widget  app-shell (uninteresting)      ** */
/* **         Dlg     Dialog, which must be freed    ** */
/* **         Ptr     Callback-structure             ** */
/* ** OUTPUT: ---                                    ** */
/* **************************************************** */
void DLG_ExitCB (Widget widget, DIALOG *Dlg, XmAnyCallbackStruct *Ptr)
{
  PRO_START ("\n")

  if (Dlg == NULL)
    PRO_ABORT ("Invalid dialog")

  DLG_UpdateSizeAndPos (Dlg);

  DLGFreeDialog (Dlg);

  PRO_END
}


void DLG_HelpCB (Widget widget, DIALOG *Dlg, XmAnyCallbackStruct *Data)
{
  PRO_START ("")

  if (Dlg == NULL)
    PRO_ABORT ("Invalid dialog")

  DLG_RunCB ((DLGFUNCTION)Dlg->HelpFct, Dlg, -1);

  PRO_END
}


void DLG_Help2CB (Widget widget, DLGELEM *Elem, XmAnyCallbackStruct *Data)
{
  PRO_START ("")

  if (Elem == NULL)
    PRO_ABORT ("Invalid Elem")

  PRO1 ("\\HELP: BoxID = %d...", Elem->BoxID)

  DLG_RunCB ((DLGFUNCTION)Elem->Dlg->HelpFct, Elem->Dlg, Elem->BoxID);

  PRO_END
}






