#include <math.h>

#include "dlgData.h"
#include "dlgCB.h"
#include "dlgSegment.h"

static Arg args[NUM_ARGS];

unsigned char nopic_bits[] = 
{
  0x00, 0x00, 0x00, 0x00, 0xe0, 0x77, 0x5c, 0x00, 0x40, 0x2c, 0x66, 0x00,
  0x40, 0x28, 0x42, 0x00, 0x40, 0x2c, 0x01, 0x00, 0xc0, 0x27, 0x01, 0x00,
  0x40, 0x20, 0x01, 0x00, 0x40, 0x20, 0x42, 0x00, 0x40, 0x20, 0x66, 0x00,
  0xe0, 0x70, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x00,
  0x40, 0xc3, 0x3c, 0x00, 0x80, 0x24, 0x09, 0x00, 0x80, 0x24, 0x09, 0x00,
  0x80, 0x24, 0x09, 0x00, 0x80, 0x24, 0x09, 0x00, 0xc0, 0xcd, 0x30, 0x00,
  0x00, 0x00, 0x00, 0x00, 0xc8, 0xff, 0x3f, 0x06, 0xc4, 0xff, 0x3f, 0x04,
  0x04, 0x00, 0x00, 0x04, 0xce, 0x6c, 0x0d, 0x07, 0x24, 0x49, 0x92, 0x04,
  0x24, 0x49, 0x92, 0x04, 0x24, 0x49, 0x92, 0x04, 0x24, 0x49, 0x92, 0x04,
  0xc4, 0xb0, 0x37, 0x0b, 0x00, 0x00, 0x00, 0x00, 
};


void DLGChangeInt (DIALOG *Dlg, int BoxID, int Digits)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgINT)) == NULL)
    PRO_ABORT ("Invalid Box")

  Elem->Value.irec.Digits = Digits;

  if (Elem->WidgetID != NULL)
  {
    XtSetArg (args[0], XmNcolumns,   Digits);
    XtSetArg (args[1], XmNmaxLength, Digits+1);
    XtSetValues (Elem->WidgetID, args, 2);
  }
  PRO_END
}


void DLGChangeDouble (DIALOG *Dlg, int BoxID, int Digits, int Decimals)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgDOUBLE)) == NULL)
    PRO_ABORT ("Invalid Box")

  Elem->Value.rrec.Digits   = Digits;
  Elem->Value.rrec.Decimals = Decimals;

  if (Elem->WidgetID != NULL)
  {
    XtSetArg (args[0], XmNcolumns,   Digits);
    XtSetArg (args[1], XmNmaxLength, Digits+1);
    XtSetValues (Elem->WidgetID, args, 2);
  }
  PRO_END
}


void DLGChangeText (DIALOG *Dlg, int BoxID, int Length, int Rows)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgTEXT)) == NULL)
    PRO_ABORT ("Invalid Box")

  if (Elem->Value.srec.Length != Length)
  {
    Elem->Value.srec.Length = Length;
    Elem->Value.srec.Text = (char*) realloc(Elem->Value.srec.Text,
					    sizeof(char) * Length+1);
    Elem->Value.srec.Text[Length] = '\0';
  }
  Elem->Value.srec.Rows = Rows;

  if (Elem->WidgetID != NULL)
  {
    XtSetArg (args[0], XmNmaxLength, Elem->Value.srec.Length);
    XtSetArg (args[1], XmNvalue,     Elem->Value.srec.Text);
    XtSetArg (args[2], XmNrows,      Elem->Value.srec.Rows);
    if (abs(Rows) > 1)
      XtSetArg (args[3], XmNeditMode, XmMULTI_LINE_EDIT);
    else
      XtSetArg (args[3], XmNeditMode, XmSINGLE_LINE_EDIT);

    XtSetValues (Elem->WidgetID, args, 4);
    /* TODO:
       Unmanagen + neu-managen fr Grennderung
       Problem: Gesamtdialog darf nicht vergrert werden
       -> eigene Funktion, ResizeWidget (Dlg, BoxID) ?
    */
  }
  PRO_END
}


void DLGChangeScale (DIALOG *Dlg, int BoxID, double Min, double Max, 
		                                int Decimals)
{
  double   Factor;
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgSCALE)) == NULL)
    PRO_ABORT ("Invalid Box")

  Elem->Value.zrec.Min      = Min;
  Elem->Value.zrec.Max      = Max;
  Elem->Value.zrec.Val      = LIMIT (Min, Max, Elem->Value.zrec.Val);
  Elem->Value.zrec.Decimals = Decimals;
  Elem->Value.zrec.reverse  = (Min > Max);

  if (Elem->WidgetID != NULL)
  {
    Factor = pow (10.0, Elem->Value.zrec.Decimals);
    XtSetArg (args[0], XmNminimum, (int)(Min * Factor + GREATER(Min)));
    XtSetArg (args[1], XmNmaximum, (int)(Max * Factor + GREATER(Max)));
    XtSetArg (args[2], XmNvalue,   (int)(Elem->Value.zrec.Val * Factor +
					 GREATER(Elem->Value.zrec.Val)));
    XtSetArg (args[3], XmNdecimalPoints, Elem->Value.zrec.Decimals);

    if (Elem->Value.zrec.Side == DlgVERTICAL)
      XtSetArg (args[4], XmNprocessingDirection, 
		Elem->Value.zrec.reverse? XmMAX_ON_BOTTOM : XmMAX_ON_TOP);
    else
      XtSetArg (args[4], XmNprocessingDirection, 
		Elem->Value.zrec.reverse? XmMAX_ON_LEFT : XmMAX_ON_RIGHT);

    XtSetValues (Elem->WidgetID, args, 5);
  }
  PRO_END
}


void DLGChangeBar (DIALOG *Dlg, int BoxID, double Min, double Max,
		                              double Step)
{
  double   swap;
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgBAR)) == NULL)
    PRO_ABORT ("Invalid Box")

  if ((Elem->Value.brec.reverse = (Min > Max)))
  {
    swap = Min;
    Min  = Max;
    Max  = swap;
  }

  Elem->Value.brec.Min  = Min;
  Elem->Value.brec.Max  = Max;
  Elem->Value.brec.Val  = LIMIT (Min, Max, Elem->Value.zrec.Val);
  Elem->Value.brec.Step = Step;

  if (Elem->WidgetID != NULL)
  {
    XtSetArg (args[0], XmNminimum,   (int)(Min *100)); 
    XtSetArg (args[1], XmNmaximum,   (int)(Max *100)); 
    XtSetArg (args[2], XmNvalue,     (int)(Elem->Value.brec.Val * 100)); 
    XtSetArg (args[3], XmNincrement, (int)(Step*100));

    if (Elem->Value.brec.Side == DlgVERTICAL)
      XtSetArg (args[4], XmNprocessingDirection, 
		Elem->Value.brec.reverse ? XmMAX_ON_BOTTOM : XmMAX_ON_TOP);
    else
      XtSetArg (args[4], XmNprocessingDirection, 
		Elem->Value.brec.reverse? XmMAX_ON_LEFT : XmMAX_ON_RIGHT);

    XtSetValues (Elem->WidgetID, args, 5);
  }
  PRO_END
}


void DLGChangeUpDn (DIALOG *Dlg, int BoxID, double Min, double Max,
		                               double Step)
{
  double   swap;
  char     New[PATH_MAX+1] = "";
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgUPDN)) == NULL)
    PRO_ABORT ("Invalid Box")

  if (Min > Max)
  {
    swap = Min;
    Min  = Max;
    Max  = swap;
  }

  Elem->Value.urec.Min  = Min;
  Elem->Value.urec.Max  = Max;
  Elem->Value.urec.Step = Step;
  Elem->Value.urec.Val  = DLG_UPDN_VAL (Min, Max, Step, Elem->Value.urec.Val);

  Elem->Value.urec.Decimals = 
    MAX (DLG_GetDecimals(Min), (MAX(DLG_GetDecimals(Max),
	 MAX(DLG_GetDecimals(Elem->Value.urec.Val), DLG_GetDecimals(Step)))));
  Elem->Value.urec.Length =
    MAX (Elem->Value.urec.Length, (MAX(DLG_GetDigits(Min),
	 DLG_GetDigits(Max)) + Elem->Value.urec.Decimals + 1));

  if (Elem->WidgetID != NULL)
  {
    DLG_Val2Str (New, Elem->Value.urec.Val, Elem->Value.urec.Length,
		      Elem->Value.urec.Decimals);
    XtSetArg (args[0], XmNvalue, New);
    XtSetValues (Elem->WidgetID, args, 1);
  }

  PRO_END  
}


void DLGChangeList (DIALOG *Dlg, int BoxID, int Lines, int Number, ...)
{
  int             count;
  char            New[PATH_MAX+1] = "";
  char           *Text = NULL;
  char          **ItemList = NULL;
  DLGELEM        *Elem;
  XmStringTable   List;
  va_list         Arguments;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgLIST)) == NULL)
    PRO_ABORT ("Invalid Box")

  Elem->Value.lrec.Number = abs (Number);
  Elem->Value.lrec.Lines  = abs (Lines);
  Elem->Value.lrec.Item = 
    (ITEMTEXT *)realloc (Elem->Value.lrec.Item,
			 sizeof(ITEMTEXT) * Elem->Value.lrec.Number);

  va_start (Arguments, Number);

  if (Number > 0)
    for (count = 0; count < Number; ++count)
      DLG_strncp (Elem->Value.lrec.Item[count], 
		  va_arg(Arguments, char*), DLG_LLEN);
  else
  {
    Number *= -1;
    ItemList = va_arg (Arguments, char**);
    for (count = 0; count < Number; ++count)
      DLG_strncp (Elem->Value.lrec.Item[count], ItemList[count], DLG_LLEN);
  }

  Elem->Value.lrec.Result = LIMIT (0, Number-1, Elem->Value.lrec.Result);

  if (Elem->WidgetID != NULL)
  {
    List = (XmStringTable)malloc (sizeof(XmString) * Number);
    for (count = 0; count < Number; ++count)
    {
      sprintf (New, "item[%d]", count);
      if (count < Number)
	Text = TOOLGetResString (Elem->WidgetID, "xmListWidgetClass",
				 "list", New, Elem->Value.lrec.Item[count]);
      else
	Text = TOOLGetResString (Elem->WidgetID, "xmListWidgetClass",
				                 "list", New, New);
      if (Elem->Value.lrec.Item[count][0] == '\001')
	List[count] = XmStringCreate (Elem->Value.lrec.Item[count]+1, "font1");
      else
	List[count] = XmStringCreate (Elem->Value.lrec.Item[count], 
				      NORMAL_FONT);
    }
    XtSetArg (args[0], XmNitemCount,        Number);
    XtSetArg (args[1], XmNitems,            List);
    XtSetArg (args[2], XmNvisibleItemCount, Lines);
    XtSetValues (Elem->WidgetID, args, 3);
    for (count = 0; count < Number; ++count)
      XmStringFree (List[count]);
    free (List);

    if (Elem->Value.lrec.Result >= 0)
      XmListSelectPos (Elem->WidgetID, Elem->Value.lrec.Result+1, False);
    else
    {
      XtSetArg (args[0], XmNselectedItemCount, 0);
      XtSetValues (Elem->WidgetID, args, 1);
    }
  }
  va_end (Arguments);

  PRO_END  
}


void DLGChangeOMenu (DIALOG *Dlg, int BoxID, int Number, ...)
{
  int        count;
  char     **ItemList = NULL;
  XmString   Label;
  Widget     Menu;
  va_list    Arguments;
  DLGELEM   *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgOMENU)) == NULL)
    PRO_ABORT ("Invalid Box")

  for (count = 0; count < Elem->Value.orec.Number; ++count)
    if (Elem->Value.orec.Item[count].WidgetID != NULL) 
      XtDestroyWidget (Elem->Value.orec.Item[count].WidgetID);

  Elem->Value.orec.Number = abs (Number);
  Elem->Value.orec.Item = 
    (ITEMREC*)realloc (Elem->Value.orec.Item,
		       sizeof(ITEMREC) * Elem->Value.orec.Number);

  va_start (Arguments, Number);
  if (Number > 0)
    for (count = 0; count < Number; ++count)
      DLG_strncp (Elem->Value.orec.Item[count].Label, 
		  va_arg(Arguments, char*), DLG_LLEN);
  else
  {
    Number *= -1;
    ItemList = va_arg (Arguments, char**);
    for (count = 0; count < Number; ++count)
      DLG_strncp (Elem->Value.orec.Item[count].Label,
		  ItemList[count], DLG_LLEN);
  }
  va_end (Arguments);

  Elem->Value.orec.Result = LIMIT (0, Number-1, Elem->Value.orec.Result);

  if (Elem->WidgetID != NULL)
  {
    XtSetArg (args[0], XmNsubMenuId, &Menu);
    XtGetValues (Elem->WidgetID, args, 1);

    for (count = 0; count < Number; ++count)
    {
      Elem->Value.orec.Item[count].Index  = count;
      Elem->Value.orec.Item[count].Parent = Elem;

      if (Elem->Value.orec.Item[count].Label[0] != '\0')
      {
	Label = XmStringLtoRCreate (Elem->Value.orec.Item[count].Label,
				    NORMAL_FONT);
	XtSetArg (args[0], XmNlabelString, Label);
	Elem->Value.orec.Item[count].WidgetID =
	  XmCreatePushButtonGadget (Menu, "entry", args, 1);
	XtAddCallback (Elem->Value.orec.Item[count].WidgetID, 
		       XmNactivateCallback, (XtCallbackProc)DLG_MenuCB, 
		       &Elem->Value.orec.Item[count]);
	XmStringFree (Label);
      }
      else
        Elem->Value.orec.Item[count].WidgetID =
	  XmCreateSeparatorGadget (Menu, "sep", NULL, 0);

      XtManageChild (Elem->Value.orec.Item[count].WidgetID);
    }

    XtSetArg (args[0], XmNsubMenuId, Menu);
    XtSetArg (args[1], XmNmenuHistory,
	      Elem->Value.orec.Item[Elem->Value.orec.Result].WidgetID);
    XtGetValues (Elem->WidgetID, args, 2);
  }

  PRO_END
}


void DLGChangeProc (DIALOG *Dlg, int BoxID, char *Name, 
		    DLGFUNCTION Function)
{
  XmString  Label = NULL;
  DLGELEM  *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgPROC)) == NULL)
    PRO_ABORT ("Invalid Box")

  DLG_strncp (Elem->Name, Name, DLG_TLEN);
  Elem->Value.crec.FPtr = Function;
  if (Elem->WidgetID != NULL)
  {
    Label = XmStringLtoRCreate (Elem->Name, NORMAL_FONT);
    XtSetArg (args[0], XmNlabelString, Label);

    XtSetValues (Elem->WidgetID, args, 1);

    XmStringFree (Label);
  }

  PRO_END
}


void DLGChangeArrow (DIALOG *Dlg, int BoxID, int Direction)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgARROW)) == NULL)
    PRO_ABORT ("Invalid Box")

  if ((Elem->Value.arec.Dir == DlgMIDDLE) ^ (Direction == DlgMIDDLE))
    PRO_ABORT ("changing from/to middle-button not allowed")

  Elem->Value.arec.Dir = Direction;

  if (Elem->WidgetID != NULL)
  {
    switch (Direction)
    {
      case DlgLEFT : 
	XtSetArg (args[0], XmNarrowDirection, XmARROW_LEFT);
	break;

      case DlgRIGHT : 
	XtSetArg (args[0], XmNarrowDirection, XmARROW_RIGHT);
	break;

      case DlgDOWN : 
	XtSetArg (args[0], XmNarrowDirection, XmARROW_DOWN);
	break;

      default : 
	XtSetArg (args[0], XmNarrowDirection, XmARROW_UP);
    }
    XtSetValues (Elem->WidgetID, args, 1);
  }

  PRO_END
}


void DLGChangePic (DIALOG *Dlg, int BoxID, int SizeX, int SizeY, void *Data)
{
  int      n = 0;
  char     FileName[PATH_MAX+1];
  Pixmap   pix1, pix2;
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgPIC)) == NULL)
    PRO_ABORT ("Invalid Box")

  Elem->Value.prec.SizeX = SizeX;
  Elem->Value.prec.SizeY = SizeY;
  Elem->Value.prec.PicName[0] = '\0';

  if (Elem->Value.prec.Data != NULL)
    free (Elem->Value.prec.Data);
  Elem->Value.prec.Data = NULL;

  if (SizeX == 0 || SizeY == 0)
  {
    DLG_strncp (Elem->Value.prec.PicName, (char*)Data, PATH_MAX);
    TOOLParseFileName (Elem->Value.prec.PicName, FileName, PATH_MAX);
    Elem->Value.prec.Data = TOOLLoadXBMData (FileName,
					     &Elem->Value.prec.SizeX, 
					     &Elem->Value.prec.SizeY,
					     NULL, NULL);
    if (Elem->Value.prec.Data == NULL)
    {
      Elem->Value.prec.SizeX = nopic_width;
      Elem->Value.prec.SizeY = nopic_height;
      Elem->Value.prec.Data = TOOLCopyXBM (nopic_bits, Elem->Value.prec.SizeX,
		                                       Elem->Value.prec.SizeY);
    }
  }
  else
    Elem->Value.prec.Data = TOOLCopyXBM ((byte*)Data, Elem->Value.prec.SizeX,
					              Elem->Value.prec.SizeY);

  if (Elem->Value.prec.Data == NULL)
    Elem->Value.prec.Data = TOOLMakeEmptyXBM (Elem->Value.prec.SizeX,
					      Elem->Value.prec.SizeY);

  if (Elem->WidgetID != NULL)
  {
    pix1 = XCreatePixmapFromBitmapData (XDisplay,
					DefaultRootWindow(XDisplay),
					Elem->Value.prec.Data, 
					Elem->Value.prec.SizeX,
					Elem->Value.prec.SizeY, 
					Elem->Value.prec.fg,
					Elem->Value.prec.bg,
					TOOLGetDepth());
    XtSetArg (args[n], XmNlabelPixmap, pix1); n++;
    XtSetArg (args[n], XmNforeground,  Elem->Value.prec.fg); n++;
    XtSetArg (args[n], XmNbackground,  Elem->Value.prec.bg); n++;

    if (Elem->Value.prec.FPtr != NULL)
    {
      if (Elem->Value.prec.iData != NULL)
	free (Elem->Value.prec.iData);
      Elem->Value.prec.iData = TOOLLightenXBM (Elem->Value.prec.Data, 
					       Elem->Value.prec.SizeX,
					       Elem->Value.prec.SizeY);
      pix2 = XCreatePixmapFromBitmapData (XDisplay,
					  DefaultRootWindow(XDisplay), 
					  Elem->Value.prec.iData, 
					  Elem->Value.prec.SizeX,
					  Elem->Value.prec.SizeY, 
					  Elem->Value.prec.fg,
					  Elem->Value.prec.bg,
					  TOOLGetDepth());
      XtSetArg (args[n], XmNlabelInsensitivePixmap, pix2); n++;
      XtSetArg (args[n], XmNrecomputeSize, True); n++;
    }
    XtSetValues (Elem->WidgetID, args, n);
  }

  PRO_END
}


void DLGChangeImage (DIALOG *Dlg, int BoxID, XImage *Image)
{
  int      n = 0;
  Pixmap   pix1;
  DLGELEM *Elem;

  PRO_START ("\n")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgIMAGE)) == NULL)
    PRO_ABORT ("Invalid Box")

  if (Elem->Value.xrec.Image != NULL)
    TOOLDestroyImage (Elem->Value.xrec.Image);
  Elem->Value.xrec.Image = TOOLCopyImage (Image);

  if (Elem->WidgetID != NULL)
  {
    pix1 = TOOLImage2Pixmap (Elem->Value.xrec.Image);
    XtSetArg (args[n], XmNlabelPixmap, pix1); n++;

    if (Elem->Value.xrec.FPtr != NULL)
    {
      XtSetArg (args[n], XmNlabelInsensitivePixmap, pix1); n++;
      XtSetArg (args[n], XmNrecomputeSize,          True); n++;
    }
    XtSetValues (Elem->WidgetID, args, n);
  }

  PRO_END
}


void DLGChangeSegment (DIALOG *Dlg, int BoxID, int Digits, int Decimals)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgSEGM)) == NULL)
    PRO_ABORT ("Invalid Box")

  Elem->Value.hrec.Digits   = Digits;
  Elem->Value.hrec.Decimals = Decimals;

  if (Elem->WidgetID != NULL)
  {
    XtSetArg (args[0], DlgNvalue, 
	      (int)(Elem->Value.hrec.Val * 
		    pow(10.0, Decimals) + GREATER(Elem->Value.hrec.Val)));
    XtSetArg (args[1], DlgNdigits,   Digits);
    XtSetArg (args[2], DlgNdecimals, Decimals);
    XtSetValues (Elem->WidgetID, args, 3);
  }

  PRO_END
}


static void DLG_SetTreeState (TREEITEM *List, dual State)
{
  TREEITEM *Ptr = List;

  while (Ptr != NULL)
  {
    if (Ptr->gadget != NULL)
      XtVaSetValues (Ptr->gadget, 
		     XmNoutlineState, State? XmEXPANDED:XmCOLLAPSED,
		     NULL);

    if (Ptr->Sub != NULL)
      DLG_SetTreeState (Ptr->Sub, State);

    Ptr = Ptr->Next;
  }
}

void DLGSetTreeState (DIALOG *Dlg, int BoxID, dual State)
{
  DLGELEM  *Elem;


  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgTREE)) == NULL)
    PRO_ABORT ("Invalid Box")
   
  DLG_SetTreeState (Elem->Value.krec.Data, State);

  XmContainerRelayout (Elem->WidgetID);
}

