/* ************************************** */
/* *** (c) Stefan Vietinghoff         *** */
/* *** dialog structure administation *** */
/* ************************************** */

#include "dlgData.h"


#define DlgID   1
#define DlgPOS  2
#define DlgTAB  4
#define DlgNAME 8


static dual DLG_CheckBox (DLGBOX *Box, BOX_ID BoxID)
{
  if (BoxID.Mask == DlgID)
    return (Box->BoxID == BoxID.BoxID);

  if (BoxID.Mask == DlgNAME)
    if (strcmp(Box->Title, BoxID.Name) == 0 ||
	(Box->Dir == BOXELEM && strcmp(Box->Item->Name, BoxID.Name) == 0))
      return TRUE;

  return FALSE;
}   


/*----------------------------------------------------------------------
  Funktion: durchsucht einen Dialogeintrag und die von ihm ausgehenden
            Untereintrge nach eine Box-Identifikationsnummer und liefert
	    bei Erfolg den Zeiger auf die Box, andernfalls einen NULL-
	    Zeiger zurck. 

  INPUT:  Box        Zeiger auf die Box, von der ausgehend gesucht
                      werden soll
	  BoxID       Identifikations-Nummer der gesuchten Box
  OUTPUT: (PBOXHEAD)  Zeiger auf den gewnschten Eintrag bzw. NULL-Zeiger
  ----------------------------------------------------------------------*/
DLGBOX *DLG_GetBox (DLGBOX *Box, BOX_ID BoxID)
{
  int       countX, countY;
  DLGBOX   *Ptr;
#if (XmVERSION>=2)
  DLGBOX   *Result = NULL;
  NOTEPAGE *Page;
#endif

  if (Box == NULL)                     /* No (more) entries    */
    return NULL;

  if (DLG_CheckBox(Box, BoxID))        /* found ?              */
    return Box;
  elsif (Box->List != NULL)
  {
    if (BoxID.Mask == DlgTAB && Box->BoxID == BoxID.BoxID)
      return Box->List[BoxID.x][BoxID.y];
    else
      for (countY = 0; countY < Box->Rows; ++countY)
	for (countX = 0; countX < Box->Cols; ++countX)
	  if ((Ptr = Box->List[countX][countY]) != NULL &&
	      (DLG_CheckBox(Ptr,BoxID) || 
	       (Ptr = DLG_GetBox(Ptr, BoxID)) != NULL))
	    return Ptr;
  }
#if (XmVERSION>=2)
  elsif (Box->Dir == BOXELEM && Box->Item->Type == DlgNOTE)
  {
    PRO1 ("Searching box %d in notebook-pages...\n", Box->BoxID)
    Page = Box->Item->Value.nrec.Page;
    while (Page != NULL)
    {
      if ((Result = DLG_GetBox(Page->Box, BoxID)) != NULL)
	return Result;
      Page = Page->Next;
    }
  }
#endif

  return NULL;
}


/*----------------------------------------------------------------------
  Function: Searches for a box with given ID and elem-type

  INPUT:  Dlg         Dialog, which may contain the box
	  BoxID       Number of searched box
          Type        assumed type of box-element
  OUTPUT: (DLGELEM*)  Pointer to element. NULL, if element wasn't found
                      or has an other type
  NOTE: Function is prepared for identifiing a box by it's name (BOX_ID)
  ----------------------------------------------------------------------*/
DLGELEM *DLG_CheckBoxType (DIALOG *Dlg, BOX_ID BoxID, int Type)
{
  DLGBOX  *Ptr = NULL;
  DLGELEM *Result = NULL;

  if (Dlg == NULL)
    PRO_ERROR0 ("invalid/empty dialog", NULL)

  if ((Ptr = DLG_GetBox(Dlg->Box, BoxID)) == NULL)
    PRO_ERROR_VAL1 ("BoxID %d does not exist", BoxID.BoxID, NULL)

  if (Ptr->Dir != BOXELEM || Ptr->Item == NULL)
    PRO_ERROR_VAL1 ("Box %d contains no value", BoxID.BoxID, NULL)

  Result = Ptr->Item;
  if (Type >= 0 && Result->Type != Type)
    PRO_ERROR_VAL1 ("Box %d has invalid Type", BoxID.BoxID, NULL)

  return Result;
}


/* *********************************************** */
/* ** Deallocates a dialog entry and all it     ** */
/* ** sub-entries.                              ** */
/* ** INPUT:  Box  Pointer to dialog head       ** */
/* ** OUTPUT: ---                               ** */
/* *********************************************** */
void DLG_FreeBox (DLGBOX *Box)
{
  int      countX, countY;
  DLGBOX  *Ptr;
  DLGELEM *Elem, *Next;

  if (Box == NULL)
    return;

  if ((Box->Dir == BOXELEM || Box->Dir == BOXLIST) && Box->Item != NULL)
  {
    Elem = Box->Item;
    while (Elem != NULL)
    {
      Next = Elem->Next;
      DLGFreeElem (Elem);
      Elem = Next;
    }
  }
  else
  {
    for (countY = 0; countY < Box->Rows; ++countY)
      for (countX = 0; countX < Box->Cols; ++countX)
	if ((Ptr = Box->List[countX][countY]) != NULL)
	  DLG_FreeBox (Ptr);
    for (countX = 0; countX < Box->Cols; ++countX)
      if (Box->List[countX] != NULL)
	free (Box->List[countX]);
    if (Box->List != NULL)
      free (Box->List);
  }
  free (Box);
}


static void DLG_FreeTreeItem (TREEITEM *List)
{
  TREEITEM *Ptr, *Next;

  Ptr = List;
  while (Ptr != NULL)
  {
    Next = Ptr->Next;
    if (Ptr->Sub != NULL)
      DLG_FreeTreeItem (Ptr->Sub);
    /* the gadget will be freed when destroying the shell widget
      if (Ptr->gadget != NULL)
        XtDestroyWidget (Ptr->gadget);
    */
    free (Ptr);
    Ptr = Next;
  }
}


dual DLGFreeTree (DIALOG *Dlg, int BoxID)
{
  DLGELEM *Elem;

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgTREE)) == NULL)
    PRO_RETURN ("DLGFreeTree: Invalid Box", FALSE)

  DLG_FreeTreeItem (Elem->Value.krec.Data);
  Elem->Value.krec.Data = NULL;

  XmContainerRelayout (Elem->WidgetID);
 
  return TRUE;
}


dual DLGBuildTree (DIALOG *Dlg, int BoxID)
{
  DLGELEM *Elem;

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgTREE)) == NULL)
    PRO_RETURN ("DLGFreeTree: Invalid Box", FALSE)

  DLG_CreateWidget (Elem->Master, Elem, TRUE);

  return TRUE;
}


/* ************************************* */
/* ** Free memory used by elements    ** */
/* ** INPUT:  Elem  dialog element    ** */
/* ** OUTPUT: ---                     ** */
/* ** EXAMPLE:                        ** */
/* **   DLGELEM *Elem;                ** */
/* **   ...                           ** */
/* **   DLGFreeElem (Elem);           ** */
/* ************************************* */       
void DLGFreeElem (DLGELEM *Elem)
{
  if (Elem == NULL)
    return;

  /* free type-specific memory */
  switch (Elem->Type)
  {
    case DlgEMPTY :
    case DlgLABEL :
    case DlgBOOL :
    case DlgINT :
    case DlgDOUBLE :
    case DlgSCALE :
    case DlgBAR :
    case DlgUPDN : 
    case DlgPROC :
    case DlgARROW :
    case DlgFSEL :
    case DlgDIRSEL :
    case DlgDIAL :
    case DlgSEGM :
      break;

    case DlgTEXT : 
      free (Elem->Value.srec.Text);
      free (Elem->Value.srec.Default);
      break;

    case DlgLIST : 
      free (Elem->Value.lrec.Item);
      break;

    case DlgOMENU : 
      if (Elem->Value.orec.Item != NULL)
	free (Elem->Value.orec.Item);
      break;

    case DlgTBOX : 
    case DlgRBOX :
      if (Elem->Value.trec.Toggle != NULL)
	free (Elem->Value.trec.Toggle);
      if (Elem->Value.trec.Result != NULL)
	free (Elem->Value.trec.Result);
      if (Elem->Value.trec.Default != NULL)
	free (Elem->Value.trec.Default);
      break; 

    case DlgPIC : 
      if (Elem->Value.prec.Data != NULL)
	free (Elem->Value.prec.Data);
      if (Elem->Value.prec.iData != NULL)
	free (Elem->Value.prec.iData);
      break;

    case DlgIMAGE :
      if (Elem->Value.xrec.Image != NULL)
	TOOLDestroyImage (Elem->Value.xrec.Image);
      break;

#if (XmVERSION>=2)
    case DlgNOTE :
      {
	NOTEPAGE *Page, *Next;

	Page = Elem->Value.nrec.Page;
	while (Page != NULL)
	{
	  Next = Page->Next;
	  if (Page->Box != NULL)
	    DLG_FreeBox (Page->Box);
	  free (Page);
	  Page = Next;
	}
      }
      break;

    case DlgTREE :
      DLG_FreeTreeItem (Elem->Value.krec.Data);
      break;
#endif

    default : 
      PRO1 ("Type %d: NO memory allocated\n", Elem->Type)
  }
  if (Elem->UserArgs != NULL)
    XtFree ((char*)Elem->UserArgs);
  free (Elem);
}


/*----------------------------------------------------------------------
  Funktion: Aufteilung einer Dialogbox in eine Anzahl von Unterboxen
            Existiert die se zu teilende Box nicht bricht die Routine
	    ab. Gleiches passiert, wenn bereits ein Item eingetragen ist

  INPUT:  Dlg        Dialogstruktur, die die zu teilende Box enthlt
          BoxID      Box-Nummer die in Unterboxen aufgeteilt werden soll
	  Separator  Angabe, ob Trennlinien zw. den Boxen gezeichnet 
	             werden sollen.
	  Number     Anzahl der zu erzeugenen Unterboxen
	  IDList     Liste der Box-Nummern die den Unterboxen zugewiesen
	             werden sollen.
          Dir        Richtung der Aufteilung 
                     (BOXCOLS = Spalten, BOXROWS = Zeilen)
  OUTPUT: (dual)  Success of this operation
  ----------------------------------------------------------------------*/
dual DLG_SplitBox (DIALOG *Dlg, int BoxID,  char *Separator,
		                int Cols, int Rows, int *IDList,
		                int Dir)
{
  int     count, Number;
  int     countX, countY;
  char    Style;
  DLGBOX *Box, *Ptr;

  PRO_START ("\n")

  if (Dlg == NULL)
    PRO_RETURN ("invalid/empty dialog", FALSE)

  /* search box... */
  if ((Box = DLG_GetBox(Dlg->Box, DLGBox(BoxID))) == NULL)
    PRO_ERROR_VAL1 ("Dialog contains no BoxID %d", BoxID, FALSE)
  elsif (Box->Rows > 0 || Box->Cols > 0 || Box->List != NULL)
    PRO_ERROR_VAL1 ("Box %d is full or allready splitted", BoxID, FALSE)
  else                                            /* otherwise split it now  */
  {
    Number = Cols * Rows;
    PRO2 ("Split box %d into %d sub-boxes: ", BoxID, Number)
    for (count = 0; count < Number; ++count)
      PRO1 ("\\[%d] ", IDList[count])
    PRO0 ("\\\n")

    Box->Dir = Dir;
    PRO1 ("Separators: '%s'\n", Separator)
    /* search until Number+1 to check for separator behind the last row/col */
    for (count = 1; count < Number+1; ++count)
      if ((Style = DLG_GetSeparator(Separator, Number)) == 'X')
      {
	PRO1 ("Invalid separator style '%c' !", Style)
	Separator[0] = '\0';
      }

    strncpy (Box->Sep, Separator, MAX_SEP);
    Box->Sep[MAX_SEP] = '\0';

    Box->List = (DLGBOX***)malloc (sizeof(DLGBOX**) * Cols);
    for (count = 0; count < Cols; ++count)
      Box->List[count] = (DLGBOX**)malloc (sizeof(DLGBOX*) * Rows);

    count = 0;
    for (countY = 0; countY < Rows; ++countY)
      for (countX = 0; countX < Cols; ++countX)
      {
	if (DLG_GetBox(Dlg->Box, DLGBox(IDList[count])) == NULL)
	{
	  Box->List[countX][countY] = (DLGBOX*)malloc (sizeof(DLGBOX));
	  Ptr = Box->List[countX][countY];
	  Ptr->BoxID    = IDList[count];
	  Ptr->Title[0] = '\0';
	  Ptr->TitlePos = DlgMIDDLE;
	  Ptr->AlignX   = (Cols==1)? DlgFILL : DlgLEFT;
	  Ptr->AlignY   = (Rows==1)? DlgFILL : DlgUP;
	  Ptr->Parent   = NULL;
	  Ptr->Main     = NULL;
	  Ptr->Box      = NULL;
	  Ptr->Border   = DlgNO_LINE[0];
	  Ptr->active   = TRUE;
	  Ptr->Sep[0]   = '\0';
	  Ptr->Dir      = BOXNONE;
	  Ptr->Number   = 0;
	  Ptr->Item     = NULL;
	  Ptr->Rows     = 0;
	  Ptr->Cols     = 0;
	  Ptr->List     = NULL;
	  Ptr->UserData = NULL;
	}
	else
	{
	  Box->List[countX][countY] = NULL;
	  PRO_ERROR_VAL1 ("Box %d already exists !", IDList[count], FALSE)
	}
	count++;
      }

    Box->Cols = Cols;
    Box->Rows = Rows;
  }
  PRO_END
  return TRUE;
}


/*----------------------------------------------------------------------
  Funktion: Aufteilung einer Dialogbox in eine Menge von Spalten.
            Existiert die zu teilende Dialogbox nicht, werden keine
	    nderungen vorgenommen.

  INPUT:  Dlg        Dialog mit der zu teilenden Box
          BoxID      Nummer der zu teilenden Box
	  Number     Anzahl der gewnschtne Spalten
	  Separator  Art der Trennung(en) (Linienwahl)
	  ...        Auflistung der ID-Nummern der zu erzeugenden Spalten
  OUTPUT: ---

  Bsp.: DIALOG *Dlg = NULL;
        ...
	DLGMakeDlgCols (Dlg, 20, 3, DlgNO_LINE DlgSINGLE_LINE, 21, 22, 23);
	(slit box 20 into three columns with the id's 21, 22 and 23. Box 22
	 22 and 23 are separated with a single line.)
  ----------------------------------------------------------------------*/
void DLGMakeCols (DIALOG *Dlg, int BoxID, int Number, char *Separators, ...)
{
  int      count, SubBoxes;
  int     *IDList;
  va_list  Arguments;

  PRO_START ("\n")

  if (Dlg == NULL)
    PRO_ERROR ("invalid/empty dialog")

  if (Number == 0)
    PRO_ABORT ("Splitting a box into 0 cols makes no sense")

  va_start (Arguments, Separators);

  /* create list of col-id's: */
  if (Number > 0)
  {
    SubBoxes = Number;
    IDList = (int *)malloc (sizeof(int) * Number);
    for (count = 0; count < Number; ++count)
      IDList[count] = va_arg(Arguments, int);
  }
  else
  {
    SubBoxes = -Number;
    IDList   = va_arg (Arguments, int *);
  }
  va_end (Arguments);
  
  DLG_SplitBox (Dlg, BoxID, Separators, SubBoxes, 1, IDList, BOXCOLS);

  if (Number > 0)
    free (IDList);

  PRO_END
}


/*----------------------------------------------------------------------
  Funktion: Aufteilung einer Dialogbox in Zeilen (analog zu 
            DLGMakeDlgCols)

  INPUT:  Dlg        Zeiger auf den Dialog mit der zu teilenden Box
          BoxID      Nummer der Box, aus der Zeilen gemacht werden sollen
	  Number     Anzahl der zu erzeugenden Zeilen
	  Separator  Art der Trennung
	  ...        Liste mit den BoxID-Nummern der Zeilen
  OUTPUT: ---

  Bsp.: DIALOG *Dlg = NULL;
        ...
	DLGMakeDlgRows (Dlg, 15, 4, SINGLELINE, 16, 17, 18 ,19);
  ----------------------------------------------------------------------*/
void DLGMakeRows (DIALOG *Dlg, int BoxID, int Number, char *Separators, ...)
{
  int      count, SubBoxes;
  int     *IDList;
  va_list  Arguments;

  PRO_START ("\n")

  if (Dlg == NULL)
    PRO_ERROR ("invalid/empty dialog")

  if (Number == 0)
    PRO_ABORT ("Splitting a box into 0 rows makes no sense")

  va_start (Arguments, Separators);
  if (Number > 0)
  {
    SubBoxes = Number;
    IDList = (int *)malloc (sizeof(int) * Number);
    for (count = 0; count < Number; ++count)
      IDList[count] = va_arg (Arguments, int);
  }
  else
  {
    SubBoxes = -Number;
    IDList   = va_arg (Arguments, int *);
  }
  va_end (Arguments);

  DLG_SplitBox (Dlg, BoxID, Separators, 1, SubBoxes, IDList, BOXROWS);

  if (Number > 0)
    free (IDList);

  PRO_END
}


/* ******************************************** */
/* ** Split a box into an array of subboxes. ** */
/* ** The new boxes are numbered as follows: ** */
/* **   BoxIDxxyy                            ** */
/* ** INPUT:  Dlg   dialog containing the    ** */
/* **               box                      ** */
/* **         BoxID Number of box to split   ** */
/* **         Cols  Number of array-columns  ** */
/* **         Rows  Number of array-rows     ** */
/* ** OUTPUT: ---                            ** */
/* ** EXAMPLE:                               ** */
/* **   DIALOG *Dlg;                         ** */
/* **   ...                                  ** */
/* **   DLGMakeGrid (Dlg, 10, 3, 4);         ** */
/* ******************************************** */
void DLGMakeGrid (DIALOG *Dlg, int BoxID, int Cols, int Rows)
{
  int  count;
  int  countX, countY;
  int *IDList;

  /* TODO:
     Rand um Zellen definieren
  */

  PRO_START ("\n")

  if (Rows <= 0 || Cols <= 0)
    PRO_ABORT ("rows and cols must be greater than 0")

  IDList = (int *)malloc (sizeof(int) * (Rows * Cols));

  count = 0;
  for (countY = 0; countY < Rows; ++countY)
    for (countX = 0; countX < Cols; ++countX)
      IDList[count++] = DlgGRID (BoxID, countX, countY);

  DLG_SplitBox (Dlg, BoxID, DlgNO_LINE, Cols, Rows, IDList, BOXGRID);
 
  free (IDList);

  PRO_END
}


void DLGMakeArray (DIALOG *Dlg, int BoxID, int Cols, int Rows)
{
  int  count;
  int  countX, countY;
  int *IDList;

  PRO_START ("\n")

  if (Rows <= 0 || Cols <= 0)
    PRO_ABORT ("rows and cols must be greater than 0")

  IDList = (int *)malloc (sizeof(int) * (Rows * Cols));

  count = 0;
  for (countY = 0; countY < Rows; ++countY)
    for (countX = 0; countX < Cols; ++countX)
      IDList[count++] = DlgGRID (BoxID, countX, countY);

  DLG_SplitBox (Dlg, BoxID, DlgNO_LINE, Cols, Rows, IDList, BOXARRAY);
 
  free (IDList);

  PRO_END
}


/*----------------------------------------------------------------------
  Function: by default the dialog will be created with an 'ok' and
            a 'cancel'-button. This function overwrites this defaults
	    and allows it, to create the button manually. When this
	    function is called the default-buttons will be erased.
	    The buttons will be displayed in the order in which they
	    are inserted.

  INPUT:  Dlg       dialog
          Type      tpye of button:
	              NO_ID needed, if NO button is be generated
		      OK_ID defines an OK-button
		      CANCEL_ID defines an CANCEL-button
		      SAVE_ID Button with save-title
		      HELP_ID Button with title 'help'
          Label     title of the button also used for resource-file
	  Placement where to place the button
	  Function  function, which will be called if button is activated.
	            If NULL is given, the dialog window will be closed,
	  ButtonID  Number which will used as parameter 
  OUTPUT: ---

  Bsp.: DIALOG *Dlg = NULL;
        ...
	DLGMakeButton (Dlg, USER_ID, "apply", NULL, 8);
  ----------------------------------------------------------------------*/
void DLGMakeButton (DIALOG *Dlg, int Type, char *Label, int Placement,
		    CALLFUNCTION Function, int ButtonID)
{
  int Number;

  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR ("invalid/empty dialog")

  if (! Dlg->Buttons.changed)
  {
    Dlg->Buttons.Number = 0;
    Dlg->Buttons.changed = TRUE;
    if (Dlg->Buttons.Data != NULL)
      free (Dlg->Buttons.Data);
    Dlg->Buttons.Data = NULL;
  }

  Number = Dlg->Buttons.Number;

  if (Type != NO_ID)
  {
    Dlg->Buttons.Data =
      (DLGBUTTON*)realloc (Dlg->Buttons.Data, sizeof(DLGBUTTON)*(Number+1));

    Dlg->Buttons.Data[Number].ID             = ButtonID;
    Dlg->Buttons.Data[Number].Type           = Type;
    Dlg->Buttons.Data[Number].Number         = Number;
    strncpy (Dlg->Buttons.Data[Number].Name, Label, DLG_LLEN);
    Dlg->Buttons.Data[Number].Name[DLG_LLEN] = '\0';
    Dlg->Buttons.Data[Number].Placement      = Placement;
    Dlg->Buttons.Data[Number].FPtr           = Function;
    Dlg->Buttons.Data[Number].Dlg            = Dlg;
    Dlg->Buttons.Data[Number].Button         = NULL;
    Dlg->Buttons.Data[Number].Width          = 0;

    if (Placement & DlgBUTTON_FILL)
      (Dlg->Buttons.Types[0])++;
    if (Placement & DlgBUTTON_MIDDLE)
      (Dlg->Buttons.Types[1])++;
    if (Placement & DlgBUTTON_EVEN)
      (Dlg->Buttons.Types[2])++;

    (Dlg->Buttons.Number)++;
  }

  PRO_END
}


/*----------------------------------------------------------------------
  Funktion: gibt rekursiv ab einer gegebenen Box Application-Defaultwerte
            in ein Datei aus

  INPUT:  fp      Zeiger auf die Dateiposition, ab der geschrieben werden soll
          Box     Zeiger auf die Dialogpostion ab der die (Unter-)boxen 
	          bearbeitet werden sollen
	  DlgName Name des Dialog (wird fr die Eintrge bentigt)
  OUTPUT: ---
  ----------------------------------------------------------------------*/
void DLG_WriteBoxApp (FILE *fp, DLGBOX *Box, char *DlgName)
{
  int      count;
  int      countX, countY;
  DLGBOX  *Ptr;
  DLGELEM *Elem;

  if (Box == NULL)
    return;

  if (Box->Title[0] != '\0')
  {
    fprintf (fp, "! Headline Box: %d\n", Box->BoxID);
    fprintf (fp, "*%s*%s.labelString:\t\t\t%s\n",
	         DlgName, Box->Title, Box->Title);
    fprintf (fp, "\n");
  }

  if (Box->Dir != BOXELEM && Box->Dir != BOXLIST)
  {
    for (countY = 0; countY < Box->Rows; ++countY)
      for (countX = 0; countX < Box->Cols; ++countX)
	if ((Ptr = Box->List[countX][countY]) != NULL)
	  DLG_WriteBoxApp (fp, Ptr, DlgName);  
  }
  elsif (Box->Dir == BOXELEM && (Elem = Box->Item) != NULL)
  {
    while (Elem != NULL)
    {
      fprintf (fp, "! Box: %d\n", Elem->BoxID);
      if (Elem->LabelPos != DlgNOWHERE)
	fprintf (fp, "*%s*%s*label.labelString:\t\t%s\n",
		     DlgName, Elem->Name, Elem->Name);
      switch (Elem->Type)
      {
        case DlgEMPTY : 
	  break;

        case DlgLABEL :
	  fprintf (fp, "*%s*%s*labelString:\t\t%s\n",
		       DlgName, Elem->Name, Elem->Name);
	  break;

        case DlgBOOL :
	  fprintf (fp, "*%s*%s*button.labelString:\t\t%s\n",
		       DlgName, Elem->Name, Elem->Name);
	  fprintf (fp, "*%s*%s*button.foreground:\t\tBlack\n",
		       DlgName, Elem->Name);
	  break;

        case DlgINT :
        case DlgDOUBLE :
	  fprintf (fp, "*%s*%s*value.foreground:\t\t\tBlack\n",
		       DlgName, Elem->Name);
	  break;

        case DlgTEXT :
	  fprintf (fp, "*%s*%s*text.foreground:\t\t\tBlack\n", 
		       DlgName, Elem->Name);
	  break;

        case DlgSCALE :
	  fprintf (fp, "*%s*%s*scale.foreground:\t\tBlack\n", 
		       DlgName, Elem->Name);
	  break;

        case DlgBAR :
	  fprintf (fp, "*%s*%s*bar.foreground:\t\t\tBlack\n", 
		       DlgName, Elem->Name);
	  break;

        case DlgUPDN :
	  fprintf (fp, "*%s*%s*upDown.labelString:\t\t%s\n",
		       DlgName, Elem->Name, Elem->Name);
	  fprintf (fp, "*%s*%s*up.foreground:\t\t\tBlack\n",
		       DlgName, Elem->Name);
	  fprintf (fp, "*%s*%s*down.foreground:\t\tBlack\n",
		       DlgName, Elem->Name);
	  fprintf (fp, "*%s*%s*value.foreground:\t\tBlack\n",
		       DlgName, Elem->Name);
	  break;

        case DlgLIST :
	  fprintf (fp, "*%s*%s*list.foreground:\t\t\tBlack\n", 
		       DlgName, Elem->Name);
	  for (count = 0; count < Elem->Value.lrec.Number; ++count)
	    fprintf (fp, "*%s*%s*list.item[%d].labelString:\t%s\n",
		         DlgName, Elem->Name, count,
		         Elem->Value.lrec.Item[count]);
	  break;

        case DlgOMENU :
	  fprintf (fp, "*%s*%s*menu*foreground:\t\t\tBlack\n", 
		       DlgName, Elem->Name);
	  fprintf (fp, "*%s*%s*oMenu.labelString:\t\t%s\n",
		       DlgName, Elem->Name, Elem->Name);
	  fprintf (fp, "*%s*%s*oMenu*foreground:\t\tBlack\n",
		       DlgName, Elem->Name);
	  for (count = 0; count < Elem->Value.orec.Number; ++count)
	    fprintf (fp, "*%s*%s*option[%d].labelString:\t\t%s\n", 
		         DlgName, Elem->Name, 
		         count, Elem->Value.orec.Item[count].Label);
	  break;

        case DlgTBOX :
	  fprintf (fp, "*%s*%s*toggle*highlightColor:\t\t\tWhite\n",
		       DlgName, Elem->Name);
	  for (count = 0; count < Elem->Value.trec.Number; ++count)
	    fprintf (fp, "*%s*%s*toggle.button[%d].labelString:\t%s\n",
		         DlgName, Elem->Name, count,
		         Elem->Value.trec.Toggle[count].Item);
	  break;

        case DlgRBOX :
	  fprintf (fp, "*%s*%s*radio*highlightColor:\t\t\tWhite\n",
		       DlgName, Elem->Name);
	  for (count = 0; count < Elem->Value.trec.Number; ++count)
	    fprintf (fp, "*%s*%s*radio.button[%d].labelString:\t%s\n",
		         DlgName, Elem->Name, count,
		         Elem->Value.trec.Toggle[count].Item);
	  break;

        case DlgPROC :
	  fprintf (fp, "*%s*%s*button.labelString:\t\t%s\n",
		       DlgName, Elem->Name, Elem->Name);
	  fprintf (fp, "*%s*%s*button.foreground:\t\tBlack\n",
		       DlgName, Elem->Name);
	  break;

        case DlgARROW :
	  fprintf (fp, "*%s*%s*button.foreground:\t\tBlack\n",
		       DlgName, Elem->Name);
	  break;

        case DlgFSEL :
        case DlgDIRSEL :
	  {
	    char WidgetName[10];

	    strcpy (WidgetName, (Elem->Type==DlgFSEL)? "fileSel":"dirSel");

	    if (Elem->Value.grec.Length > 0)
	    {
	      fprintf (fp, "*%s*%s*button.foreground:\t\tBlack\n",
		           DlgName, Elem->Name);
	      fprintf (fp, "*%s*%s*%s.default:\t\t%s\n",
		           DlgName, Elem->Name, WidgetName, 
		           Elem->Value.grec.Default);
	    }
	    fprintf (fp, "*%s*%s*%s.title:\t\t%s\n",
		         DlgName, Elem->Name, WidgetName, Elem->Name);
	    fprintf (fp, "*%s*%s*%s.dirMask:\t\t%s\n",
		         DlgName, Elem->Name, WidgetName, 
		         Elem->Value.grec.Mask);
	    fprintf (fp, "*%s*%s*foreground:\t\tBlack\n",
		         WidgetName, Elem->Name);
	  }
	  /* Buttons, Labels */
	  break;

        case DlgPIC :
	  if (Elem->Value.prec.FPtr == NULL)
	  {
	    if (Elem->Value.prec.PicName != NULL)
	      fprintf (fp, "*%s*%s*pic.fileName:\t\t%s\n", 
		           DlgName, Elem->Name, Elem->Value.prec.PicName);
	    if (Elem->Value.prec.Format != DlgXPM)
	      fprintf (fp, "*%s*%s*pic.foreground:\t\tBlack\n",
		           DlgName, Elem->Name);
	  }
	  else
	  {
	    if (Elem->Value.prec.PicName != NULL)
	      fprintf (fp, "*%s*%s*button.fileName:\t\t%s\n", 
		           DlgName, Elem->Name, Elem->Value.prec.PicName);
	    if (Elem->Value.prec.Format != DlgXPM)
	      fprintf (fp, "*%s*%s*button.foreground:\t\tBlack\n",
		           DlgName, Elem->Name);
	  }
	  break;

#if(XmVERSION>=2)
        case DlgNOTE :
	  {
	    NOTEPAGE *Ptr;

	    fprintf (fp, "*%s*%s*foreground:\t\t\tBlack\n", 
		         DlgName, Elem->Name);
	    fprintf (fp, "*%s*%s*backPageForeground:\t\t\tBlack\n", 
		         DlgName, Elem->Name);
	    fprintf (fp, "*%s*%s*frameForeground:\t\t\tBlack\n",
		         DlgName, Elem->Name);

	    if ((Ptr = Elem->Value.nrec.Page) != NULL)
	      while (Ptr != NULL)
	      {
		DLG_WriteBoxApp (fp, Ptr->Box, DlgName);
		Ptr = Ptr->Next;
	      }
	  }
	  break;

        case DlgTREE :
	  fprintf (fp, "*%s*%s*foreground:\t\t\tBlack\n",
  		       DlgName, Elem->Name);
	  break;
#endif

        case DlgDIAL :
	  fprintf (fp, "*%s*%s*foreground:\t\t\tBlack\n", DlgName, Elem->Name);
	  break;

        case DlgSEGM :
	  fprintf (fp, "*%s*%s*foreground:\t\t\tGreen\n", DlgName, Elem->Name);
	  fprintf (fp, "*%s*%s*background:\t\t\tBlack\n", DlgName, Elem->Name);
	  break;

        default : 
	  fprintf (fp, "! unkown element-type\n");
      }
      Elem = Elem->Next;
    }
    fprintf (fp, "\n");
  }  
}


/*----------------------------------------------------------------------
  Funktion: Erzeugt aus einer Dialogstruktur eine grundlegende 
            Application-Default-Datei (Resourcedatei).

  INPUT:  fp      Zeiger auf die Dateiposition, ab der geschrieben werden soll
          Dlg     Dialog, zu dem die App-Defs-Eintrge geschrieben werden
	          sollen
	  DlgName Name des Dialogs fr den Resourcennamen.
  OUTPUT: ---

  Bsp.: FILE   *FilePtr;
        DLGBOX *Box;
	...
	DLGWriteAppDefs (FilePtr, Box, "TestDlg");
  ----------------------------------------------------------------------*/
void DLGWriteAppDefs (FILE *fp, DIALOG *Dlg)
{
  int  count;
  char Upper[DLG_LLEN+1];

  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR ("Try to get create AppDefs of an empty dialog")

  if (fp == NULL)
    PRO_ERROR ("Invalid file-pointer (NULL) specified")

  if (Dlg->Box != NULL)
  {
    TOOLstrtoupper (Upper, Dlg->Title);
    fprintf (fp, "!\n");
    fprintf (fp, "! *** %s ***\n", Dlg->Title);
    fprintf (fp, "!\n");
    fprintf (fp, "*%s.title:\t\t\t%s\n",          Dlg->Title, Upper);
    fprintf (fp, "*%s.style:\t\t\tUSERPOS\n",     Dlg->Title);
    fprintf (fp, "*%s*background:\t\t\tGrey50\n", Dlg->Title);
    fprintf (fp, "*%s*foreground:\t\t\tBlack\n",  Dlg->Title);
    fprintf (fp, "*%s*xPosition:\t\t\t%d\n",      Dlg->Title, Dlg->PosX);
    fprintf (fp, "*%s*yPosition:\t\t\t%d\n",      Dlg->Title, Dlg->PosY);
    fprintf (fp, "\n");
    if (Dlg->Box->Title[0] != '\0')
    {
      fprintf (fp, "*%s*%s.foreground:\t\tRed\n", Dlg->Title, Dlg->Box->Title);
      fprintf (fp, "*%s*%s.fontList:"
	           "\t\t-adobe-*-*-*-*-*-24-*-*-*-*-*-*-*\n", 
	           Dlg->Title, Dlg->Box->Title);
    }
    fprintf (fp, "\n");

    DLG_WriteBoxApp (fp, Dlg->Box, Dlg->Title);

    fprintf (fp, "! Buttons:\n");
    for (count = 0; count < Dlg->Buttons.Number; ++count)
      if (Dlg->Buttons.Data[count].Type != NO_ID)
      {
	TOOLstrtoupper (Upper, Dlg->Buttons.Data[count].Name);
	fprintf (fp, "*%s*%s.labelString:\t\t%s\n", 
		     Dlg->Title, Dlg->Buttons.Data[count].Name, Upper);

	switch (Dlg->Buttons.Data[count].Type)
	{
  	  case OK_ID :
	    fprintf (fp, "*%s*%s*accelerators:\t\t"
		         "#override <Key>Return: ArmAndActivate()",
		         Dlg->Title, Dlg->Buttons.Data[count].Name);
	    break;

  	  case CANCEL_ID :
	    fprintf (fp, "*%s*%s*accelerators:\t\t"
                         "#override <Key>osfCancel: ArmAndActivate()",
		         Dlg->Title, Dlg->Buttons.Data[count].Name);
	    break;

    	  default :
	    break;
	}
      }
  }

  PRO_END
}


/*----------------------------------------------------------------------
  Funktion: schreibt rekursiv alle ab einer geg. Dlg-Box enthaltenen
            Unterboxen in eine Datei.

  INPUT:  fp      Zeiger auf die Dateiposition, ab der geschrieben werden soll
          Box     Zeiger auf die Dialogpostion ab der die (Unter-)boxen 
	          ausgegeben werden sollen
	  Indent  Verschachtelungstiefe (fr rekursiven Aufruf)
  OUTPUT: ---
  ----------------------------------------------------------------------*/
void DLG_WriteBox (FILE *fp, DLGBOX *Box, int Indent)
{
  int      count, Default;
  int      countX, countY;
  DLGBOX  *Ptr;
  DLGELEM *Elem;
  
  if (Box == NULL)
    return;

  /* TODO:
     Auch fr BOXLIST und Notebook-Eintrge
  */
  for (count = 0; count < Indent; ++count)
    fprintf (fp, "  ");
  fprintf (fp, "Box: <ID = %d | Typ: ", Box->BoxID);
  switch (Box->Dir)
  {
    case BOXELEM : fprintf (fp, "ELEM");
	           break;
    case BOXROWS : fprintf (fp, "ROWS");
                   break;
    case BOXCOLS : fprintf (fp, "COLS");
                   break;
    case BOXLIST : fprintf (fp, "LIST");
	           break;
    case BOXGRID : fprintf (fp, "ARRAY");
                   break;
    default : fprintf (fp, "NONE");
  }
  fprintf (fp, " | #Elem: %dx%d>: ", Box->Cols, Box->Rows);

  if (Box->Dir == BOXELEM && Box->Item != NULL)
  {
    Elem = Box->Item;
    fprintf (fp, "Label: %s\n", Elem->Name);
    for (count = 0; count <= Indent+1; ++count)
      fprintf (fp, "  ");
    fprintf (fp, "       ");
    if (Elem->PosX > 0 || Elem->PosY > 0)
      fprintf (fp, "(%3d/%3d)  ", Elem->PosX, Elem->PosY);
    switch (Elem->Type)
    {
      case DlgEMPTY : 
	fprintf (fp, "NONE []");
	break;

      case DlgLABEL : 
	fprintf (fp, "LABEL [%s]", Elem->Value.frec.Text);
	break;

      case DlgBOOL : 
	fprintf (fp, "DUAL [%s]", Elem->Value.drec.Value ? "TRUE":"FALSE");
	break;

      case DlgINT : 
	fprintf (fp, "GANZ [%d | %d]", 
		     Elem->Value.irec.Value, Elem->Value.irec.Digits);
	break;

      case DlgDOUBLE : 
	fprintf (fp, "REAL [%f | %d | %d]",   Elem->Value.rrec.Value,
		     Elem->Value.rrec.Digits, Elem->Value.rrec.Decimals);
	  break;

      case DlgTEXT : 
	fprintf (fp, "TEXT [%d | %s | %d | %d]", Elem->Value.srec.Length,
		     Elem->Value.srec.Text,      Elem->Value.srec.Cols,
	             Elem->Value.srec.Rows);
	break;

      case DlgSCALE : 
	fprintf (fp, "SCALE [%f | %f | %f | %s | %s]", 
		     Elem->Value.zrec.Min, Elem->Value.zrec.Max, 
		     Elem->Value.zrec.Val,
		     (Elem->Value.zrec.Side==DlgVERTICAL)? "ver":"hor",
		     Elem->Value.zrec.reverse ? "<-":"->");
	break;

      case DlgBAR : 
	fprintf (fp, "BAR [%f | %f | %f | %f | %s | %s]",
		     Elem->Value.brec.Min, Elem->Value.brec.Max,
	             Elem->Value.brec.Val, Elem->Value.brec.Step,
		     (Elem->Value.brec.Side==DlgVERTICAL)? "ver":"hor",
		     Elem->Value.brec.reverse ? "<-":"->");
	break;

      case DlgUPDN : 
	fprintf (fp, "UPDN [%f | %f | %f | %f]",
		     Elem->Value.urec.Min, Elem->Value.urec.Max,
		     Elem->Value.urec.Val, Elem->Value.urec.Step);
	break;

      case DlgLIST : 
	fprintf (fp, "LIST [%d ", Elem->Value.lrec.Number);
	for (count = 0; count < Elem->Value.lrec.Number; ++count)
	  fprintf (fp, "| %s ", Elem->Value.lrec.Item[count]);
	fprintf (fp, "%d | %d]", 
		     Elem->Value.lrec.Result, Elem->Value.lrec.Lines);
	break;

      case DlgOMENU : 
	fprintf (fp, "OMENU [%d ", Elem->Value.orec.Number);
	for (count = 0; count < Elem->Value.orec.Number; ++count)
	  if (Elem->Value.orec.Item[count].Label[0] != '\0')
	    fprintf (fp, "| %s ", Elem->Value.orec.Item[count].Label);
	  else
	    fprintf (fp, "| NULL ");
	fprintf (fp, "| %d]", Elem->Value.orec.Result);
	break;

      case DlgTBOX : 
	fprintf (fp, "TBOX [%d | %d ", 
		     Elem->Value.trec.RowCol, Elem->Value.trec.Number);
	for (count = 0; count < Elem->Value.trec.Number; ++count)
	  fprintf (fp, "| (%s=%c) ", Elem->Value.trec.Toggle[count].Item,
		       Elem->Value.trec.Toggle[count].Value ? 'T':'F');
	fprintf (fp, "]");
	break;

      case DlgRBOX : 
	fprintf (fp, "RBOX [%d | %d ", 
		     Elem->Value.trec.RowCol, Elem->Value.trec.Number);
	Default = 0;
	for (count = 0; count < Elem->Value.trec.Number; ++count)
	{
	  fprintf (fp, "| (%s) ", Elem->Value.trec.Toggle[count].Item);
	  if (Elem->Value.trec.Toggle[count].Value)
	    Default = count;
	}
	fprintf (fp, "| %d]", Default);
	break;

      case DlgPROC : 
	fprintf (fp, "PROC []");
	break;

      case DlgARROW : 
	fprintf (fp, "ARROW [");
	switch (Elem->Value.arec.Dir)
	{
	  case   DlgLEFT : fprintf (fp, "left");
	                   break;
	  case     DlgUP : fprintf (fp, "up");
	                   break;
	  case  DlgRIGHT : fprintf (fp, "right");
	                   break;
	  case   DlgDOWN : fprintf (fp, "down");
	                   break;
	  case DlgMIDDLE : fprintf (fp, "middle");
	                   break;
	         default : fprintf (fp, "unknown");
	}
	fprintf (fp, "]");
	break;

      case DlgFSEL : 
	fprintf (fp, "FSEL [%s | %d | %s | %s]", Elem->Value.grec.Mask, 
		     Elem->Value.grec.Length,    Elem->Value.grec.Result,
		     (Elem->Value.grec.Length==0)? "permanent":"call");
	break;

      case DlgDIRSEL : 
	fprintf (fp, "DIRSEL [%s | %d | %s | %s]", Elem->Value.grec.Mask, 
		     Elem->Value.grec.Length,    Elem->Value.grec.Result,
		     (Elem->Value.grec.Length==0)? "permanent":"call");
	break;

      case DlgPIC : 
	fprintf (fp, "BMAP (%d) [%d x %d:%s]", Elem->Value.prec.Format,
		     Elem->Value.prec.SizeX,   Elem->Value.prec.SizeY,
		     Elem->Value.prec.PicName);
	break;

#if(XmVERSION>=2)
      case DlgNOTE :
	break;

      case DlgTREE :
	break;
#endif

      case DlgDIAL :
	break;

      case DlgSEGM :
	break;

      default : 
	fprintf (fp, "unknown Item-Type !");
    }
    fprintf (fp, "\n");
  }
  else
  {
    fprintf (fp, "(Sep: %s) Title: %s\n", Box->Sep, Box->Title);
    Indent++;
    for (countY = 0; countY < Box->Rows; ++countY)
      for (countX = 0; countX < Box->Cols; ++countX)
	if ((Ptr = Box->List[countX][countY]) != NULL)
	{
	  fprintf (fp, "[%3d]=", Ptr->BoxID);
	  DLG_WriteBox (fp, Ptr, Indent);
	}
    Indent--;
  }
}


/*----------------------------------------------------------------------
  Funktion: schreibt die wichtigen Daten des Inhalt eines Dialog in
            eine Datei mit dem angegebenen Namen

  INPUT:  Dlg       Dialog dessen Inhalt in die Datei soll
          FileName  Name der zu erzeugenden Datei
  OUTPUT: ---

  Bsp.: DIALOG *Dlg;
        ...
	DLGWritecDialog (Dlg, "/usr/stefan/daten/dialog.dat");
  ----------------------------------------------------------------------*/
void DLGWriteDialog (DIALOG *Dlg, char *FileName)
{
  FILE *fp = NULL;

  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR ("Try to write data of an empty dialog")

  if ((fp = fopen(FileName, "w+")) == NULL)
    PRO_ABORT ("Couldn't open the file")

  if (Dlg->Box != NULL)
  {
    fprintf (fp, "[%3d]=", Dlg->Box->BoxID);
    DLG_WriteBox (fp, Dlg->Box, 0);
  }

  fclose (fp);

  PRO_END
}


BOX_ID DLGBox (int BoxID)
{
  BOX_ID Result;

  Result.Mask    = DlgID;
  Result.BoxID   = BoxID;
  Result.x       = -1;
  Result.y       = -1;
  Result.Name[0] = '\0';

  return Result;
}


#if 0
/* Not yet needed */
BOX_ID DLGPos (int BoxID, int x, int y)
{
  BOX_ID Result;

  Result.Mask    = DlgPOS;
  Result.BoxID   = BoxID;
  Result.x       = x;
  Result.y       = y;
  Result.Name[0] = '\0';

  return Result;
}


BOX_ID DLGTab (int BoxID, int c, int r)
{
  BOX_ID Result;

  Result.Mask    = DlgTAB;
  Result.BoxID   = BoxID;
  Result.x       = c;
  Result.y       = r;
  Result.Name[0] = '\0';

  return Result;
}
#endif


BOX_ID DLGName (char *Name)
{
  BOX_ID Result;

  Result.Mask    = DlgNAME;
  Result.BoxID   = 0;
  Result.x       = -1;
  Result.y       = -1;
  strncpy (Result.Name, Name, DLG_NLEN);
  Result.Name[DLG_NLEN] = '\0';

  return Result;
}

