#include <unistd.h>
#include <math.h>

#include <Xm/DrawingA.h>

/* TODO: 
   besserer 3D-Style (X-Resourcen fr Schattendicke benutzen)
   bessere Zeigerberechnung, falls Scale oval
*/

#include "dlgUtils.h"
#include "dlgDnD.h"
#include "dlgDial.h"

#ifndef PI
#define PI 3.14159265
#endif

#define SIZE_X 70               /* Default Size */
#define SIZE_Y 120

#ifndef GRAD
#define GRAD(x) ((x) * 180.0 / PI)
#endif
#ifndef RAD
#define RAD(x) ((x) * PI / 180.0)
#endif

enum TURN_DIRECTION
{
  TURN_NONE = -1,
  TURN_LEFT,
  TURN_RIGHT
};

typedef struct
{
  double d;
  int    a;
} PTR_POINT;

typedef struct
{
  XmString       Title;

  int            Min, Max, Step;
  int            Old, Val;
  int            Digits, Decimals;
  int            Ticks, SubTicks;
  int            TickShadow;

  Dimension      SizeX, SizeY;
  Dimension      Thickness;
  Pixel          shadow;
  Widget         Value, Area;
  Widget         Main;

  XtCallbackProc CBProc;
  caddr_t        UserData;
} DIALDATA;


typedef struct
{
  XmFontList Fonts;
} FONTRES, *PFONTRES;


typedef struct
{
  Widget     widget;
  DIALDATA  *Dial;
  dual       start;
  int        x, y;
  int        SizeX, SizeY;
  dual       dir;
  XEvent    *Event;
} DIAL_TIMER;

static DIAL_TIMER Timer = {NULL, NULL, TRUE, 0, 0, 0, 0, TURN_NONE, NULL};
static GC         DialGC;
#if (XmVERSION>=2)
  static Atom       StringAtom;
  static Atom       CompTxtAtom;
#endif

/* Default-Values: */
static XtResource Resources[] =
{
  /* Name, Class, Type, Size, Location, ..., Default */

  /* Form-Widget */
  {XmNwidth,           XmCWidth,    
                       XtRDimension, sizeof(Dimension),
                       XtOffsetOf(DIALDATA, SizeX), 
                       XtRImmediate, (XtPointer) 70},
  {XmNheight,          XmCHeight,   XtRDimension, sizeof(Dimension),
                       XtOffsetOf(DIALDATA, SizeY),  
                       XtRImmediate, (XtPointer)120},
  {XmNshadowThickness, XmCShadowThickness, XtRDimension, sizeof(Dimension),
                       XtOffsetOf(DIALDATA, Thickness), 
                       XtRImmediate, (XtPointer)  3},
  /* TODO:
     Shadow-Color, Skalierungsfarbe
  */

  /* Titel */
  {XmNfontList,        XmCFontList, XmRFontList, sizeof(XmFontList),
                       XtOffset(PFONTRES, Fonts),
                       XmRString, "fixed"},

  /* Value range */
  {DlgNminimum,        DlgCMinimum, XtRInt,            sizeof(int),
                       XtOffsetOf(DIALDATA, Min),  
                       XtRImmediate, (XtPointer)   0},
  {DlgNmaximum,        DlgCMaximum, XtRInt,            sizeof(int),
                       XtOffsetOf(DIALDATA, Max), 
                       XtRImmediate, (XtPointer)1000},
  {DlgNvalue,          DlgCValue  , XtRInt,            sizeof(int),
                       XtOffsetOf(DIALDATA, Val),
                       XtRImmediate, (XtPointer)   0},
  {DlgNstep,           DlgCStep,     XtRInt,           sizeof(int),
                       XtOffsetOf(DIALDATA, Step), 
                       XtRImmediate, (XtPointer)   1},
  {DlgNdigits,         DlgCDigits,   XtRInt,           sizeof(int),
                       XtOffsetOf(DIALDATA, Digits),
                       XtRImmediate, (XtPointer)   5},
  {DlgNdecimals,       DlgCDecimals, XtRInt,           sizeof(int),
                       XtOffsetOf(DIALDATA, Decimals),
                       XtRImmediate, (XtPointer)   1},

  /* geometry-info */
  {DlgNticks,          DlgCTicks,    XtRInt,           sizeof(int),
                       XtOffsetOf(DIALDATA, Ticks),
                       XtRImmediate, (XtPointer) 6},
  {DlgNsubTicks,       DlgCSubTicks, XtRInt,           sizeof(int),
                       XtOffsetOf(DIALDATA, SubTicks),  
                       XtRImmediate, (XtPointer) 4},
  {DlgNtickShadow,     DlgCTickShadow, XtRInt,        sizeof(int),
		       XtOffsetOf(DIALDATA, TickShadow),
                       XtRImmediate, (XtPointer) 1}
};

static Arg       args[NUM_ARGS];
static Pixel     bg, fg;
static PTR_POINT Pointer[][2] = {{{0.60,    0}, {4.01,    0}},
				 {{0.33,  160}, {1.37,  132}}, 
				 {{0.27,  180}, {0.70,   180}},
				 {{0.33, -160}, {1.37, -132}}};

/* function prototypes: */
static void DLG_CallCallback (DIALDATA *Dial, XEvent *event);
static void DLG_TurnLeft     (Widget widget, DIALDATA *Dial, 
			        XmArrowButtonCallbackStruct *Data);
static void DLG_TurnRight    (Widget widget, DIALDATA *Dial,
			        XmArrowButtonCallbackStruct *Data);
static void DLG_ChangeValue  (Widget widget, DIALDATA *Dial, 
			        XmTextVerifyCallbackStruct *Data);
static void DLG_ResizeDial   (Widget widget, DIALDATA *Dial,
			        XmAnyCallbackStruct *Data);
static void DLG_UpdateValue  (DIALDATA *Dial, double Value);


#if (XmVERSION>=2)
static void DLG_DialTransfer (Widget widget, XtPointer Data,
			      Atom *Selection, Atom *Type, XtPointer Value,
			      unsigned long *Length, int *Format)
{
  double    Val = 0.0;
  DIALDATA *Dial = (DIALDATA *)Data;

#if 0
  if (*Type == StringAtom)
  {
#endif
PRO1 ("TRANFER: %ld\n", Length)
    ((char *)Value)[MIN(10,(int)(*Length)/sizeof(char))] = '\0';
PRO1 ("TRANFER: '%s'\n", (char *)Value)
      Val = atof ((char *)Value) * pow (10.0, Dial->Decimals);
#if 0
  }
  else
  {
    PRO2 ("Type %s  Format %d\n", 
	  XmGetAtomName(XtDisplay(widget), *Type), *Format)
    XtVaSetValues (widget, XmNtransferStatus,   XmTRANSFER_FAILURE,
		           XmNnumDropTransfers, 0,
		           NULL);
  }
#endif

  DLG_UpdateValue (Dial, Val);
}

static void DLG_DialDrop (Widget widget, XtPointer UserData,
			                 XmDropProcCallbackStruct *Data)
{
  unsigned char            State = XmTRANSFER_FAILURE;
  int                      n = 0;
  Cardinal                 count, TargetNo;
  Cardinal                 TransferNo = 0;
  Atom                    *Exports;
  XmDropTransferEntryRec   Targets[1];
  Arg                      args[4];
  DIALDATA                *Dial = NULL;

  XtVaGetValues (Data->dragContext, XmNexportTargets,    &Exports,
		                    XmNnumExportTargets, &TargetNo,
		                    NULL);
  XtVaGetValues (widget, XmNuserData, &Dial, NULL);

  if (Data->dropSiteStatus == XmVALID_DROP_SITE)
  {
    Data->operation = XmDROP_COPY;
    switch (Data->reason)
    {
      case XmCR_DROP_MESSAGE :
	{
	  PRO1 ("DIAL Drop: DROP-MESSAGE: %d\n", TargetNo)

	  for (count = 0; count < TargetNo; ++count)
	    if (Exports[count] == StringAtom ||
		Exports[count] == CompTxtAtom)
	    {
	      Targets[0].target      = Exports[count];
	      Targets[0].client_data = (XtPointer)Dial;
	      XtSetArg (args[n], XmNtransferProc,  DLG_DialTransfer); n++;
	      XtSetArg (args[n], XmNdropTransfers, Targets);          n++;
	      TransferNo = 1;
	      State = XmTRANSFER_SUCCESS;
	      break;
	    }
        }
        break;

      case XmCR_DRAG_MOTION:
	printf ("DIAL Drop: DRAG_MOTION\n");
	break;

      case XmCR_OPERATION_CHANGED:
      default :
	printf ("DIAL Drop: OPERATION CHANGED\n");
    }
  }

  XtSetArg (args[n], XmNnumDropTransfers, TransferNo); n++;
  XtSetArg (args[n], XmNtransferStatus,   State);      n++;
  XmDropTransferStart (Data->dragContext, args, n);
}
#endif


void CalcShadowParam (void)
{
  int    count, Index, Number;
  double px[3], py[3];
  double w10, w12, w;

  Number = sizeof(Pointer) / (sizeof(PTR_POINT)*2);
  for (count = 0; count < Number; ++count)
  {
    Index = (count+Number-1) % Number;
    px[0] =  Pointer[Index][0].d * cos(Pointer[Index][0].a);
    py[0] = -Pointer[Index][0].d * sin(Pointer[Index][0].a);
    PRO4 ("P1[%f, %d] = (%f / %f)\n", 
	  Pointer[Index][0].d, Pointer[Index][0].a, px[0], py[0])

    Index = count;
    px[1] =  Pointer[Index][0].d * cos(Pointer[Index][0].a);
    py[1] = -Pointer[Index][0].d * sin(Pointer[Index][0].a);
    PRO4 ("P2[%f, %d] = (%f / %f)\n", 
	  Pointer[Index][0].d, Pointer[Index][0].a, px[1], py[1])

    Index = (count+1) % Number;
    px[2] =  Pointer[Index][0].d * cos(Pointer[Index][0].a);
    py[2] = -Pointer[Index][0].d * sin(Pointer[Index][0].a);
    PRO4 ("P3[%f, %d] = (%f / %f)\n", 
	  Pointer[Index][0].d, Pointer[Index][0].a, px[2], py[2])

    w10 = atan2 (py[0]-py[1], px[0]-px[1]);
    w12 = atan2 (py[2]-py[1], px[2]-px[1]);
    PRO2 ("P12: %f = %f\n", GRAD(w10), fmod(GRAD(w10) + 360.0, 360.0))
    PRO2 ("P23: %f = %f\n", GRAD(w12), fmod(GRAD(w12) + 360.0, 360.0))
  
    PRO1 ("Test: %f\n", fmod(w12-w10 + 2*PI, 2*PI))

    w = fmod (w10 + w12, PI);
    PRO1 ("w: %f\n", GRAD(w))
  }
}


static void DLG_RedrawPointer (DIALDATA *Dial, int CenterX, int CenterY)
{
  int    i, count, num;
  double Angle, Radius;
  XPoint Points[5];
  XPoint Points2[5];
  XPoint Poly[5];

  Radius = MIN (CenterX, CenterY) * 0.8;

  num = sizeof(Pointer) / (sizeof(PTR_POINT)*2);
  /* clear old pointer */
  i = 240 + (-300 * (Dial->Old - Dial->Min)) / (Dial->Max - Dial->Min);
  XSetForeground (XDisplay, DialGC, bg);
  for (count = 0; count < num; ++count)
  {
    Points[count].x = CenterX + (int)(Radius * Pointer[count][0].d *
				      cos(RAD(i+Pointer[count][0].a)));
    Points[count].y = CenterY - (int)(Radius * Pointer[count][0].d *
				      sin(RAD(i+Pointer[count][0].a)));
    Points2[count].x = 
      Points[count].x + (int)(Dial->Thickness * Pointer[count][1].d *
			      cos(RAD(i+Pointer[count][1].a)));
    Points2[count].y = 
      Points[count].y - (int)(Dial->Thickness * Pointer[count][1].d *
			      sin(RAD(i+Pointer[count][1].a)));
  }
  Points[num].x = Points[0].x;
  Points[num].y = Points[0].y;
  Points2[num].x = Points2[0].x;
  Points2[num].y = Points2[0].y;
  for (i = 0; i < num; ++i)
  {
    Poly[0] = Points[i];
    Poly[1] = Points2[i];
    Poly[2] = Points2[i+1];
    Poly[3] = Points[i+1];
    Poly[4] = Poly[0];

    XFillPolygon (XDisplay, XtWindow(Dial->Area), DialGC, Poly, 5, 
		  Convex, CoordModeOrigin);
  }

  /* draw new pointer */
  i = 240 + (-300 * (Dial->Val - Dial->Min)) / (Dial->Max - Dial->Min);
  for (count = 0; count < num; ++count)
  {
    Points[count].x = CenterX + (int)(Radius * Pointer[count][0].d *
				      cos(RAD(i+Pointer[count][0].a)));
    Points[count].y = CenterY - (int)(Radius * Pointer[count][0].d *
				      sin(RAD(i+Pointer[count][0].a)));
    Points2[count].x = 
      Points[count].x + (int)(Dial->Thickness * Pointer[count][1].d *
			      cos(RAD(i+Pointer[count][1].a)));
    Points2[count].y = 
      Points[count].y - (int)(Dial->Thickness * Pointer[count][1].d *
			      sin(RAD(i+Pointer[count][1].a)));
  }
  Points[num].x = Points[0].x;
  Points[num].y = Points[0].y;
  Points2[num].x = Points2[0].x;
  Points2[num].y = Points2[0].y;

  for (i = 0; i < num; ++i)
  {
    Angle = atan2 (Points[i+1].y - Points[i].y, Points[i+1].x - Points[i].x);
    if (Angle > -3.0*PI/4.0 && Angle < 1.0*PI/4.0)
      XSetForeground (XDisplay, DialGC, fg);
    else
      XSetForeground (XDisplay, DialGC, Dial->shadow);

    Poly[0] = Points[i];
    Poly[1] = Points2[i];
    Poly[2] = Points2[i+1];
    Poly[3] = Points[i+1];
    Poly[4] = Poly[0];

    XFillPolygon (XDisplay, XtWindow(Dial->Area), DialGC, Poly, 5, 
		  Convex, CoordModeOrigin);
  }
}


static void DLG_UpdateValue (DIALDATA *Dial, double Value)
{
  int    count, len, ticks;
  int    CenterX, CenterY;
  int    x1, y1, x2, y2;
  double OffsX, OffsY;
  double Step, Angle;
  char   ValStr[16];

  Dial->Old = Dial->Val;
  Dial->Val = LIMIT (Dial->Min, Dial->Max, Value);

  CenterX = Dial->SizeX / 2;
  CenterY = Dial->SizeY / 2;

  ticks = (Dial->Ticks -1) * (Dial->SubTicks+1);
  Step = -(5.0/3.0 * PI) / (double)ticks;

  for (count = 0; count <= ticks; ++count)
  {
    len = ((count % (Dial->SubTicks+1)) == 0)? 5 : 2;
    Angle = PI/3.0 + count*Step;
    OffsX = cos (Angle);
    OffsY = sin (Angle);

    /* TODO: 
       Position ohne absolute Distanz -10 bestimmen
    */
    x1 = CenterX + (Dial->SizeX/2-10) * OffsX;
    y1 = CenterY + (Dial->SizeY/2-10) * OffsY;
    x2 = x1 + len * OffsX;
    y2 = y1 + len * OffsY;
    XSetForeground (XDisplay, DialGC, Dial->shadow);
    XDrawLine (XDisplay, XtWindow(Dial->Area), DialGC, x1, y1, x2, y2);

    if (Dial->TickShadow != 0)
    {
      x1+=Dial->TickShadow; y1+=Dial->TickShadow;
      x2+=Dial->TickShadow; y2+=Dial->TickShadow;
      XSetForeground (XDisplay, DialGC, fg);
      XDrawLine (XDisplay, XtWindow(Dial->Area), DialGC, x1, y1, x2, y2);
    }
  } 

  /* Redraw the pointer */
  DLG_RedrawPointer (Dial, CenterX, CenterY);

  /* Update value-string */
  DLG_Val2Str (ValStr, Dial->Val / pow(10, Dial->Decimals), 
	               Dial->Digits, Dial->Decimals);
  XmTextSetString (Dial->Value, ValStr);
  XFlush (XDisplay); 
}


static double DLG_Pos2Val (DIALDATA *Dial, int x, int y)
{
  int    dx, dy;
  double Angle, Result;

  dx =  x - Dial->SizeX / 2;
  dy = Dial->SizeY / 2 - y;

  if (abs(dx) < 3 && abs(dy) < 3) 
    return Dial->Val;

  Angle = GRAD (atan2(dy, dx));

  if (Angle > 270.0) 
    Angle -= 360.0;
  if (Angle < -90.0) 
    Angle += 360.0;

  Result = (Dial->Max - Dial->Min) * (240.0-Angle) / 300.0 + Dial->Min;

  return Result;
}


static void DLG_DialEventHandle (Widget widget, DIALDATA *Dial, XEvent *Event)
{
  switch (Event->type)
  {
    case MotionNotify :
      {
	double        New;
	XMotionEvent *Motion = (XMotionEvent *)Event;

	if (Motion->state & Button1Mask)
	{
	  New = DLG_Pos2Val (Dial, Motion->x, Motion->y);
	  DLG_UpdateValue (Dial, New);
	}
      }
      break;

    case ButtonPress :
      {
	double New;

	if (Event->xbutton.button == Button1)
	{
	  New = DLG_Pos2Val (Dial, Event->xbutton.x, Event->xbutton.y);
	  DLG_UpdateValue (Dial, New);
	}
      }
      break;

    case Expose :
    case ConfigureNotify :
    {
      XExposeEvent *ExpEvent = (XExposeEvent*)Event;

      if (ExpEvent->window == XtWindow(Dial->Area))
      {
	PRO0 ("REDRAW...\n")
	XClearWindow (XDisplay, XtWindow(Dial->Area));
	DLG_UpdateValue (Dial, Dial->Val);
       	DLG_CallCallback (Dial, Event);
      }
    }
    break;

    default : 
      PRO1 ("Event: %d\n", Event->type)
  }
}


/* ********************************************* */
/* *** Userfunctions for the Dial-Class *** */
/* ********************************************* */
Widget DLGCreateDial (Widget Parent, String Name, 
		      ArgList user_args, Cardinal user_num, 
		      XtCallbackProc CBProc, XtPointer UserData)
{
  char      ValStr[20];
  int       n = 0; 
  DIALDATA *Dial = NULL;
  Widget    Up, Dn;
  ArgList   MergeArgs;
#if (XmVERSION>=2)
    Atom Targets[2];
#endif

  PRO_START ("\n")

#if 0
CalcShadowParam ();
#endif

  DialGC = DefaultGC (XDisplay, XScreen);

  Dial = (DIALDATA*)XtMalloc (sizeof(DIALDATA));
  Dial->CBProc   = CBProc;
  Dial->UserData = UserData;
  Dial->shadow   = TOOLGetColor ("white");
  /* TODO: 
     Zeigerfarbe mit DLGGetDlgString setzen... 
  */

  n = 0;
  XtSetArg (args[n], XmNuserData, Dial); n++;
  XtSetArg (args[n], XmNwidth,    SIZE_X); n++;
  XtSetArg (args[n], XmNheight,   SIZE_Y); n++;
  MergeArgs = XtMergeArgLists (args, n, user_args, user_num);
  Dial->Main = XtCreateManagedWidget (Name, xmFormWidgetClass, 
				      Parent, MergeArgs, user_num+n);
  XtFree ((char*)MergeArgs);

  /* Init "Dial"-record using user_args, resource-entries or def-values */
  XtGetSubresources (Dial->Main, Dial, Name, DIAL_CLASS, 
		     Resources, XtNumber(Resources), user_args, user_num);

  PRO2 ("PARAMETER: %d x %d\n",     Dial->SizeX, Dial->SizeY)
  PRO3 ("           %d - %d [%d]\n",Dial->Min, Dial->Max, Dial->Step)
  PRO3 ("           %d (%d: %d)\n", Dial->Val, Dial->Digits, Dial->Decimals)

  if (Dial->Min == Dial->Max)
    PRO_RETURN ("Min = Max not allowed", Dial->Main)

  /* TODO: 
     Setzen der pointer-Struktur mit shadow-width 
  */
  XtSetArg (args[0], XmNbackground, &bg);
  XtSetArg (args[1], XmNforeground, &fg);
  XtGetValues (Dial->Main, args, 2);

  n = 0;
  XtSetArg (args[n], XmNbottomAttachment,  XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNarrowDirection,    XmARROW_LEFT); n++;
  XtSetArg (args[n], XmNrubberPositioning, True); n++;
  XtSetArg (args[n], XmNfractionBase,      100); n++;
  XtSetArg (args[n], XmNleftAttachment,    XmATTACH_POSITION); n++;
  XtSetArg (args[n], XmNleftPosition,      0); n++;
  XtSetArg (args[n], XmNrightAttachment,   XmATTACH_POSITION); n++;
  XtSetArg (args[n], XmNrightPosition,     50); n++;
  Dn = XtCreateManagedWidget ("down", xmArrowButtonWidgetClass, 
			              Dial->Main, args, n);
  XtAddCallback (Dn, XmNarmCallback, (XtCallbackProc)DLG_TurnLeft, Dial);

  n = 0;
  XtSetArg (args[n], XmNbottomAttachment,   XmATTACH_WIDGET); n++;
  XtSetArg (args[n], XmNarrowDirection,     XmARROW_RIGHT); n++;
  XtSetArg (args[n], XmNrubberPositioning,  True); n++;
  XtSetArg (args[n], XmNfractionBase,       100); n++;
  XtSetArg (args[n], XmNleftAttachment,     XmATTACH_POSITION); n++;
  XtSetArg (args[n], XmNleftPosition,       50); n++;
  XtSetArg (args[n], XmNrightAttachment,    XmATTACH_POSITION); n++;
  XtSetArg (args[n], XmNrightPosition,      100); n++;
  Up = XtCreateManagedWidget ("up", xmArrowButtonWidgetClass, 
			            Dial->Main, args, n);
  XtAddCallback (Up, XmNarmCallback, (XtCallbackProc)DLG_TurnRight, Dial);

  DLG_Val2Str (ValStr, Dial->Val / pow(10, Dial->Decimals), 
	               Dial->Digits, Dial->Decimals);
  XtSetArg (args[0], XmNbottomAttachment, XmATTACH_WIDGET);
  XtSetArg (args[1], XmNbottomWidget,     Dn);
  XtSetArg (args[2], XmNleftAttachment,   XmATTACH_FORM);
  XtSetArg (args[3], XmNrightAttachment,  XmATTACH_FORM);
  XtSetArg (args[4], XmNvalue,            ValStr);
  XtSetArg (args[5], XmNcolumns,          Dial->Digits);
  Dial->Value = XtCreateManagedWidget ("value", xmTextWidgetClass, 
				                Dial->Main, args, 6);
  XtAddCallback (Dial->Value, XmNvalueChangedCallback, 
		              (XtCallbackProc)DLG_ChangeValue, Dial);
  XtAddCallback (Dial->Value, XmNlosingFocusCallback,
		              (XtCallbackProc)DLG_ChangeValue, Dial);

  n = 0;
  XtSetArg (args[n], XmNtopAttachment,     XmATTACH_FORM);   n++;
  XtSetArg (args[n], XmNleftAttachment,    XmATTACH_FORM);   n++;
  XtSetArg (args[n], XmNrightAttachment,   XmATTACH_FORM);   n++;
  XtSetArg (args[n], XmNbottomAttachment,  XmATTACH_WIDGET); n++;
  XtSetArg (args[n], XmNbottomWidget,      Dial->Value);     n++;
#if (XmVERSION>=2)
  XtSetArg (args[n], XmNuserData,          Dial);            n++;
#endif
  Dial->Area = XtCreateManagedWidget ("icon", xmDrawingAreaWidgetClass,
				              Dial->Main, args, n);
  XtAddCallback (Dial->Area, XmNresizeCallback, 
		             (XtCallbackProc)DLG_ResizeDial, Dial);
  XtAddEventHandler (Dial->Area, StructureNotifyMask | ExposureMask | 
		                 ButtonPressMask | PointerMotionMask,
		                 False,
		                 (XtEventHandler)DLG_DialEventHandle, Dial);
  /* dial will be drawn after creation by event-generation */
#if (XmVERSION>=2)
    StringAtom  = XmInternAtom (XDisplay, "STRING",    False);
    CompTxtAtom = XmInternAtom (XDisplay, "COMP_TEXT", False);

    Targets[0] = StringAtom;
    Targets[1] = CompTxtAtom;  /* unused jet */
    XtSetArg (args[0], XmNimportTargets,    Targets);
    XtSetArg (args[1], XmNnumImportTargets, 2);
    XtSetArg (args[2], XmNdragProc,         NULL);
    XtSetArg (args[3], XmNdropProc,         DLG_DialDrop);
    XtSetArg (args[4], XmNdropSiteActivity, XmDROP_SITE_ACTIVE);
    XmDropSiteRegister (Dial->Area, args, 5);
#endif

  PRO_END
  return Dial->Main;
}


/* ****************** */
/* *** CALLBACKS: *** */
/* ****************** */
static void DLG_CallCallback (DIALDATA *Dial, XEvent *event)
{
  DialCallbackStruct CB;

  if (Dial->CBProc != NULL)
  {
    CB.reason = XmCR_VALUE_CHANGED;
    CB.event  = event;
    CB.Value  = Dial->Val;
    CB.Old    = Dial->Old;

    (*Dial->CBProc) (Dial->Main, Dial->UserData, &CB);
  }
}


static void DLG_DialTimer (DIAL_TIMER *Data, XtIntervalId Timer)
{
  int           PosX, PosY, x, y;
  unsigned int  Mask;
  double        New;
  Window        RootWin, ChildWin;
  DIALDATA     *Dial;

  XQueryPointer(XDisplay, XtWindow(Data->widget),
		&RootWin, &ChildWin, &x, &y, &PosX, &PosY, &Mask);
  if (Mask & Button1Mask)
  {
    Dial = Data->Dial;
    if (Data->dir == TURN_RIGHT)
      New = Dial->Val + Dial->Step;
    else
      New = Dial->Val - Dial->Step;

    DLG_UpdateValue (Dial, New);

    XtAppAddTimeOut (AppCon, Data->start? DLG_START_DELAY : DLG_NEXT_DELAY, 
			     (XtTimerCallbackProc)DLG_DialTimer, Data);
  }
  Data->start = FALSE;
}

static void DLG_TurnLeft (Widget widget, DIALDATA *Dial, 
		                         XmArrowButtonCallbackStruct *Data)
{
  Timer.widget   = widget;
  Timer.Dial     = Dial;
  Timer.start    = TRUE;
  Timer.dir      = TURN_LEFT;
  Timer.Event    = Data->event;

  XtSetArg (args[0], XmNwidth,  &Timer.SizeX);
  XtSetArg (args[1], XmNheight, &Timer.SizeY);
  XtGetValues (widget, args, 2);

  DLG_DialTimer (&Timer, (XtIntervalId)0);
}


static void DLG_TurnRight (Widget widget, DIALDATA *Dial,
		                          XmArrowButtonCallbackStruct *Data)
{
  Timer.widget   = widget;
  Timer.Dial     = Dial;
  Timer.start    = TRUE;
  Timer.dir      = TURN_RIGHT;
  Timer.Event    = Data->event;

  XtSetArg (args[0], XmNwidth,  &Timer.SizeX);
  XtSetArg (args[1], XmNheight, &Timer.SizeY);
  XtGetValues (widget, args, 2);

  DLG_DialTimer (&Timer, (XtIntervalId)0);
}


static void DLG_ResizeDial (Widget widget, DIALDATA *Dial,
		                           XmAnyCallbackStruct *Data)
{
  Dimension width, height;

  PRO0 ("RESIZE\n")

  XtSetArg (args[0], XmNwidth,  &width);
  XtSetArg (args[1], XmNheight, &height);
  XtGetValues (widget, args, 2);

  Dial->SizeX = width;
  Dial->SizeY = height;

/*
  XClearWindow (XDisplay, XtWindow(Dial->Area));
  DLG_UpdateValue (Dial, Dial->Val);
*/
}


static void DLG_ChangeValue (Widget widget, DIALDATA *Dial, 
		                            XmTextVerifyCallbackStruct *Data)
{
  char    New[DLG_VLEN];
  char   *ErrPos = NULL; 
  double  Value;

  DLG_CallCallback (Dial, Data->event);

  DLG_strncp (New, XmTextGetString(widget), DLG_VLEN);
  Value = strtod (New, &ErrPos) * pow (10.0, Dial->Decimals);

  if (Data->reason == XmCR_LOSING_FOCUS)
    DLG_UpdateValue (Dial, Value);
}


void DLGSetDialValue (Widget widget, int Value)
{
  DIALDATA *Dial = NULL;

  PRO_START ("")

  XtSetArg (args[0], XmNuserData, &Dial);
  XtGetValues (widget, args, 1);

  DLG_UpdateValue (Dial, Value);

  PRO_END
}


double DLGGetDialValue (Widget widget)
{
  double    Result = 0.0;
  DIALDATA *Dial = NULL;

  PRO_START ("")

  XtSetArg (args[0], XmNuserData, &Dial);
  XtGetValues (widget, args, 1);

  Result = Dial->Val / pow(10.0, Dial->Decimals);

  PRO_END
  return Result;
}
