/* ******************************* */
/* *** (c) Stefan Vietinghoff  *** */
/* *** dialog-box functions    *** */
/* ******************************* */

#include "dlgUtils.h"
#include "dlgScript.h"


static void Msg1CB ();          /* Dialogbox button 1..3 CB-function */
static void Msg2CB ();
static void Msg3CB ();

static int  n;
static int  DLGXPos = -1;
static int  DLGYPos = -1;


/* ******************* */
/* *** Dialogboxen *** */
/* ******************* */

static unsigned char DLG_GetDlgType (int Type)
{
  PRO_START (" ")

  switch (Type)
  {
    case DlgERROR :
      PRO_RETURN ("Error", XmDIALOG_ERROR)
    case DlgINFO :
      PRO_RETURN ("Info", XmDIALOG_INFORMATION)
    case DlgMESSAGE :
      PRO_RETURN ("Message", XmDIALOG_MESSAGE)
    case DlgQUESTION :
      PRO_RETURN ("Question", XmDIALOG_QUESTION)
    case DlgWARNING :
      PRO_RETURN ("Warning", XmDIALOG_WARNING)
    case DlgWORKING :
      PRO_RETURN ("Working", XmDIALOG_WORKING)
    default :
      PRO_RETURN ("Invalid Dialog type", XmDIALOG_MESSAGE)
  }

  PRO_END
  return XmDIALOG_MESSAGE;
}


/*----------------------------------------------------------------------
  Funktion: setzt die Bildschirmposition der folgenden Dialogbox(en)
            Bei einem Wert von 0 < Pos < 1 wird der Wert als Prozent-
	    angabe gewertet, andererseits als Absolute Pixelangabe.
	    Vorteil gegenber einem zustzlichem Unterscheidungsflag:
	      + keine zustzlichen Parameter
	      + X- und Y Position knnen unabhngig im Typ sein
  
  INPUT:  xPos, yPos  gewnschte  Position der linken oberen Dialogbox-
                      ecke
  OUTPUT: ---
  
  Bsp.: DLGSetDialogPos (450, 300);
  ----------------------------------------------------------------------*/
void DLGSetDialogPos (double xPos, double yPos)
{
  int ScrX, ScrY;

  ScrX = DisplayWidth  (XDisplay, XScreen);
  ScrY = DisplayHeight (XDisplay, XScreen);

  if (xPos > 0.0 && xPos < 1.0)
    DLGXPos = (int)(xPos * ScrX);
  elsif (xPos >= 0.0)
    DLGXPos = MIN ((int)xPos, ScrX);
  else
    DLGXPos = -1;

  if (yPos > 0.0 && yPos < 1.0)
    DLGYPos = (int)(yPos * ScrY);
  elsif (yPos >= 0.0)
    DLGYPos = MIN ((int)yPos, ScrY);
  else
    DLGYPos = -1;
}
  

/*----------------------------------------------------------------------
  Funktion: Eventhandler (fr Dialogboxen ohne Knpfe), der beim 
            anklicken mit der Maus das Fenster schliet
  
  INPUT:  widget    Shell-Widget der Dialogbox
          UserData  Rckgabewert (ID des gedrckten Knopfes)
	  Event     Event (Mausclick)
  OUTPUT: ---
  ----------------------------------------------------------------------*/
static void DLG_DlgEventHandle (Widget widget, int *UserData, XEvent *Event)
{
PRO0 ("DlgEvent\n")
  if (Event->type && (Event->xbutton.button == Button1)
	 /*	      || Event->xbutton.button == Button2 ||
		      Event->xbutton. */)
  {
    XtPopdown (XtParent(widget));
    XtDestroyWidget (XtParent(widget));

    *UserData = 0;
PRO0 ("called\n")
  }
}


/*----------------------------------------------------------------------
  Funktion: Erstellung einer Dialogbox (Dialog, welcher eine Meldung
            darstellt und 1 bis 3 verschieden Beendigungs-Knpfe enthlt)
  
  INPUT:  Type      Dialogbox-Typ (siehe dazu Dokumentation)
          DialogID  Identifier mit dem Eintrge in der Resource-Datei
	            gekennzeichnet werden. (keine Leerzeichen)
	  Msg       Nachricht, die ausgegeben werden soll
	            bzw. falls Type = USERMSG den Dateinamen des Bitmaps
	  Number    Anzahl der gewnschten Beendigungsknpfe
	  ...       Auflistung der Knopf-Identifier (zur Indentifikation)
	            in der Resourcedatei (drfen keine Leerzeichen enthalten)
  OUTPUT: (int)     Nummer des gedrckten Knopfes (von 1 ab gezhlt)
                    0, falls Number nicht in 1..3 (knopfloser Dialog)

  EXAMPLE: 
    int Button;
    ...
    Button = DLGOpenTextDialog (WARNMSG, "WARN_BOX", 
                                "Allgemeine Schutzverletzung !",
				2, "PRIMA", "TOLL");
    if (Button == 1)     oder      switch (Button)
      ...                          { ...
  ----------------------------------------------------------------------*/
int DLGOpenTextDialog (char *DialogID, int Type, char *Msg, int Number, ...)
{
  Arg           args[NUM_ARGS];
  va_list       Arguments;
  unsigned int  w, h, border, bits;
  int           x, y, count, Result = 0;
  int           xMouse = 0, yMouse = 0;
  int           DlgType;
  char          geo_string[12];
  char         *Text = NULL;
  char         *ScriptMsg = NULL;
  char         *Button[3];
  Pixel         back, front;
  Widget        Shell, Dialog;
  Window        Dummy;
  XmString      Message;
  XmString      OKText = NULL;
  XmString      CANCELText = NULL;
  XmString      HELPText = NULL;

  PRO_START (" ")

  if (Number < 0 || Number > 3)
    PRO_RETURN ("Invalid Number of Buttons", 0)

  if (! TOOLIsInitialized())
    PRO_ERROR0 ("Dlg.-Lib has NOT been initialized\n", 0)

  va_start (Arguments, Number);
  for (count = 0; count < Number; ++count)
    Button[count] = va_arg (Arguments, char *);
  va_end (Arguments);

  if (TOOLScriptRunning())
  {
    Result = Number;
    if ((ScriptMsg = DLG_RunDlgScript(DialogID, Button, &Result)) == NULL)
	 PRO_RETURN ("Script-Button", Result)
    else
      PRO1 ("Error while execution command-script: %s\n", ScriptMsg)
  }

  /* Set background color: */
  if ((Text = TOOLGetResString(AppShell, "xmDialogShellWidgetClass", 
			       DialogID, "background", NULL)) == NULL)
    switch (Type)
    {
      case DlgERROR :
      case DlgWARNING : 
	back = TOOLGetColor ("Red");
	break;
      case DlgQUESTION : 
	back = TOOLGetColor ("Green");
	break;
      default : 
	XtSetArg (args[0], XmNbackground, &back);
	XtGetValues (AppShell, args, 1);
    }
  else
    back = TOOLGetColor (Text);

  /* set foreground-color... */
  Text = TOOLGetResString (AppShell, "xmDialogShellWidgetClass", 
			             DialogID, "foreground", "Black");
  front = TOOLGetColor (Text);

  /* global recource entries: */
  Shell = XtVaCreatePopupShell (DialogID, xmDialogShellWidgetClass,
				          AppShell, 
				          XmNallowShellResize, TRUE,
				          XmNtopOffset,          10,
				          XmNbottomOffset,       10,
				          XmNrightOffset,        10,
				          XmNbackground,       back,
				          XmNforeground,      front,
				          NULL);

  DlgType = DLG_GetDlgType (Type);

  n = 0;
  Text = TOOLGetResString (Shell, "xmDialogShellWidgetClass",
			          "message", "labelString", Msg);
  Message = XmStringLtoRCreate (Text, NORMAL_FONT);
  XtSetArg (args[n], XmNmessageString,    Message);               n++;
  XtSetArg (args[n], XmNmessageAlignment, XmALIGNMENT_BEGINNING); n++;

  /* Resource fr die Buttons setzen: */
  switch (Number)
  {
    case 1 :
      Text = TOOLGetResString (Shell, "xmDialogShellWidgetClass", Button[0],
			              "labelString", Button[0]);
      CANCELText = XmStringLtoRCreate (Text, NORMAL_FONT);
      XtSetArg (args[n], XmNcancelLabelString, CANCELText); n++;
      break;

    case 2 :
      Text = TOOLGetResString (Shell, "xmDialogShellWidgetClass", Button[0],
			              "labelString", Button[0]);
      OKText = XmStringLtoRCreate (Text, NORMAL_FONT);
      XtSetArg (args[n], XmNokLabelString, OKText); n++;

      Text = TOOLGetResString (Shell, "xmDialogShellWidgetClass", Button[1],
			              "labelString", Button[1]);
      HELPText = XmStringLtoRCreate (Text, NORMAL_FONT);
      XtSetArg (args[n], XmNhelpLabelString, HELPText); n++;
      break;

    case 3 :
      Text = TOOLGetResString (Shell, "xmDialogShellWidgetClass", Button[0],
			              "labelString", Button[0]);
      OKText = XmStringLtoRCreate (Text, NORMAL_FONT);
      XtSetArg (args[n], XmNokLabelString, OKText); n++;

      Text = TOOLGetResString (Shell, "xmDialogShellWidgetClass", Button[1],
			              "labelString", Button[1]);
      CANCELText = XmStringLtoRCreate (Text, NORMAL_FONT);
      XtSetArg (args[n], XmNcancelLabelString, CANCELText); n++;

      Text = TOOLGetResString (Shell, "xmDialogShellWidgetClass", Button[2],
			              "labelString", Button[2]);
      HELPText = XmStringLtoRCreate (Text, NORMAL_FONT);
      XtSetArg (args[n], XmNhelpLabelString, HELPText); n++;

    case 0 :
    default :
      break;
  }

  /* Jetzt die MessageBox erzeugen: */
  if (Number > 0)
  {
    XtSetArg (args[n], XmNdefaultButtonType, XmDIALOG_OK_BUTTON); n++;
  }
  XtSetArg (args[n], XmNbackground, back);     n++;
  XtSetArg (args[n], XmNforeground, front);    n++;
  XtSetArg (args[n], XmNdialogType, DlgType);  n++;

  /* TODO:
     Mit dialogType strtzt xmetex bei ExitPrg ab !!! 
  */

  Dialog = XmCreateMessageBox (Shell, "messageBox", args, n);

  XmStringFree (Message);

  if (OKText != NULL)
    XmStringFree (OKText);
  if (CANCELText != NULL)
    XmStringFree (CANCELText);
  if (HELPText != NULL)
    XmStringFree (HELPText);

  if (Number == 0)
    XtAddEventHandler (Dialog, ButtonPressMask, False,
	                       (XtEventHandler)DLG_DlgEventHandle, &Result);

  /* unmanage unused buttons: */
  switch (Number)
  {
    case 0 : 
      XtUnmanageChild (XmMessageBoxGetChild(Dialog, XmDIALOG_OK_BUTTON));
      XtUnmanageChild (XmMessageBoxGetChild(Dialog, XmDIALOG_CANCEL_BUTTON));
      XtUnmanageChild (XmMessageBoxGetChild(Dialog, XmDIALOG_HELP_BUTTON));
      XtUnmanageChild (XmMessageBoxGetChild(Dialog, XmDIALOG_SEPARATOR));
      break;

    case 1 : 
      XtAddCallback (Dialog, XmNcancelCallback, 
		             (XtCallbackProc)Msg1CB, &Result);
      XtUnmanageChild (XmMessageBoxGetChild(Dialog, XmDIALOG_OK_BUTTON));
      XtUnmanageChild (XmMessageBoxGetChild(Dialog, XmDIALOG_HELP_BUTTON));
      break;
    case 2 :
      XtAddCallback (Dialog, XmNokCallback,   (XtCallbackProc)Msg1CB, &Result);
      XtAddCallback (Dialog, XmNhelpCallback, (XtCallbackProc)Msg2CB, &Result);
      XtUnmanageChild (XmMessageBoxGetChild(Dialog, XmDIALOG_CANCEL_BUTTON));
      break;
    default :
      XtAddCallback (Dialog, XmNokCallback,
		             (XtCallbackProc)Msg1CB, &Result);
      XtAddCallback (Dialog, XmNcancelCallback,
		             (XtCallbackProc)Msg2CB, &Result);
      XtAddCallback (Dialog, XmNhelpCallback,
		             (XtCallbackProc)Msg3CB, &Result);
  }

  XtManageChild (Dialog);
  XtSetMappedWhenManaged (Shell, False);
  XtManageChild (Shell);

  if (Number == 0)
    Number = 1;

  if (DLGXPos == -1 && DLGYPos == -1)
  {
    XGetGeometry (XDisplay, XtWindow(Shell), &Dummy, 
		  &x, &y, &w, &h, &border, &bits);
    border = 10;                         /* 2 x border width [pixel] */
    TOOLGetMousePos (&xMouse, &yMouse);
    x = MIN (MAX(0, xMouse-(int)w/(2*Number)),   /* whole dialog visible ? */
	     DisplayWidth(XDisplay, XScreen)-(int)(w+border));
    y = MIN (MAX(0, yMouse-(int)h+30),
	     DisplayHeight(XDisplay, XScreen)-(int)(h+border));
  }
  else
  {
    x = DLGXPos;
    y = DLGYPos;
  }
  sprintf (geo_string, "+%d+%d", x, y);
  XtSetArg (args[0], XmNgeometry, geo_string);
  XtSetArg (args[1], XmNx, x);
  XtSetArg (args[2], XmNy, y);
  XtSetValues (Shell, args, 3);

  XtMapWidget (Shell);
  XtRealizeWidget (Shell); 
  XtPopup (Shell, XtGrabExclusive);   /* Fokus auf das Fenster beschrnken */

  PRO0 ("\\MainLoop...")
  TOOLMainLoop (AppCon, Shell);

  XUngrabPointer (XDisplay, CurrentTime);
  XFlush (XDisplay);

  PRO_END
  return Result;
}


/*----------------------------------------------------------------------
  Funktion: Mit der Funktion DLGOpenPicDialog wird ein Dialogfenster
            erstellt in dem anstelle der vordefinierten Symbole ein 
	    selbtdefiniertes Bitmap angezeigt wird.
  
  INPUT:  DialogID  Identifier mit dem Attribute in der Resourcedatei
                    angesprochen werden. (keine Leerzeichen)
	  XSize,
	  YSize     Gre des Bildes bzw. 0,0 falls Bild aus Datei geladen
	            werden soll.
	  Data      Dateiname bzw. Daten des Bildes; NULL erzeugt leeres
	            Bild.
	  Number    Anzahl der gewnschten Beendigungsknpfe
	  ...       Auflistung der Knopf-Identifier (zur Indentifikation
	            in der Resourcedatei (drfen keine Leerzeichen enthalten)
  OUTPUT: (int)     Nummer des gedrckten Knopfes (von 1 ab gezhlt)
                    0, falls Number nicht in 1..3
  
  Bsp.: int Button;
        ...
        Button = DLGOpenPicDialog ("WARN_BOX", DLG_XBM, 0, 0, 
	                           "pic/stefan/bild.xbm", 2, "OK", "CANCEL");
	if (Button == 1)     bzw.      switch (Button)
	  ...                          { ...
  ----------------------------------------------------------------------*/
int DLGOpenPicDialog (char *DialogID, int Format, int XSize, int YSize, 
		      void *Data,     char *Msg,  int Number, ...)
{
  unsigned int  w, h, border, bits;
  int           count;
  int           x, y, Result = 0;
  int           xMouse = 0, yMouse = 0;
  char          geo_string[12];
  char          FileName[PATH_MAX+1] = "";
  char         *Text, *PicData = NULL;
  char         *ScriptMsg = NULL;
  char         *Button[3];
  va_list       Arguments;
  Arg           args[NUM_ARGS];
  Pixel         bg, fg;
  Pixmap        PixMap = XmUNSPECIFIED_PIXMAP;
  Widget        Shell, Dialog;
  Window        Dummy;
  XmString      Message = NULL; 
  XmString      OKText = NULL, CANCELText = NULL, HELPText = NULL;

  PRO_START ("\n")

  if (! TOOLIsInitialized())
    PRO_ERROR0 ("Dlg-Lib has NOT been initialized\n", 0)

  if (Number < 0 || Number > 3)
    PRO_RETURN ("Invalid amount of buttons", 0)

  va_start (Arguments, Number);
  for (count = 0; count < Number; ++count)
    Button[count] = va_arg(Arguments, char*);
  va_end (Arguments);

  if (TOOLScriptRunning())
  {
    Result = Number;
    if ((ScriptMsg = DLG_RunDlgScript(DialogID, Button, &Result)) == NULL)
	 PRO_RETURN ("Script-Button", Result)
    else
      PRO1 ("Error while execution command-script: %s\n", ScriptMsg)
  }

  /* permanent attributes: */
  XtSetArg (args[0], XmNallowShellResize, True);
  XtSetArg (args[1], XmNtopOffset,        10);
  XtSetArg (args[2], XmNbottomOffset,     10);
  XtSetArg (args[3], XmNrightOffset,      10);
  Shell = XtCreatePopupShell (DialogID, xmDialogShellWidgetClass,
			                AppShell, args, 4);

  /* get colors: */
  Text = TOOLGetResString (Shell, "xmDialogShellWidgetClass", 
			          "image", "background", "Grey");
  bg = TOOLGetColor (Text);

  Text = TOOLGetResString (Shell, "xmDialogShellWidgetClass", 
			          "image", "foreground", "Black");
  fg = TOOLGetColor (Text);

  n = 0;
  if (XSize == 0 || YSize == 0)
  {
    Text = TOOLGetResString (Shell, "xmDialogShellWidgetClass",
			            "image", "fileName", (char*)Data);
    TOOLParseFileName (Text, FileName, PATH_MAX);

    if (Format == DlgXPM)
      PixMap = TOOLLoadXPM (FileName, NULL, NULL, bg);
    else
      PixMap = XmGetPixmap (XtScreen(Shell), FileName, fg, bg);
  }
  elsif (Data == NULL)
  {
    PicData = (char*)TOOLMakeEmptyXBM (XSize, YSize);
    PixMap = TOOLMakeXBM (XDisplay, PicData, XSize, YSize, fg, bg);
  }
  elsif (Format == DlgXPM)
    PixMap = TOOLMakeXPM (Data, NULL, NULL, bg);
  else
    PixMap = TOOLMakeXBM (XDisplay, Data, XSize, YSize, fg, bg);

#if 0
  if (PixMap == XmUNSPECIFIED_PIXMAP)
  {
    PRO0 ("\\Bitmap-ERROR...")
    PixMap = TOOLMakeXPM (error_xpm, NULL, NULL, bg);
  }
#endif

  XtSetArg (args[n], XmNsymbolPixmap, PixMap); n++;
  if (Msg != NULL)
  {
    Text = TOOLGetResString (Shell, "xmDialogShellWidgetClass",
					           "message", "labelString", Msg);
    Message = XmStringLtoRCreate (Text, NORMAL_FONT);
    XtSetArg (args[n], XmNmessageString,    Message); n++;
    XtSetArg (args[n], XmNmessageAlignment, XmALIGNMENT_BEGINNING); n++;
  }

  if (PicData != NULL)
    free (PicData);

  /* set button-attributes: */
  switch (Number)
  {
    case 1 :
      Text = TOOLGetResString (Shell, "xmDialogShellWidgetClass", Button[0],
						        "labelString", Button[0]);
      CANCELText = XmStringLtoRCreate (Text, NORMAL_FONT);
      XtSetArg (args[n], XmNcancelLabelString, CANCELText); n++;
      break;

    case 2 :
      Text = TOOLGetResString (Shell, "xmDialogShellWidgetClass", Button[0],
			                       "labelString", Button[0]);
      OKText = XmStringLtoRCreate (Text, NORMAL_FONT);
      XtSetArg (args[n], XmNokLabelString, OKText); n++;

      Text = TOOLGetResString (Shell, "xmDialogShellWidgetClass", Button[1],
			              "labelString", Button[1]);
      HELPText = XmStringLtoRCreate (Text, NORMAL_FONT);
      XtSetArg (args[n], XmNhelpLabelString, HELPText); n++;
      break;

    case 3 :
      Text = TOOLGetResString (Shell, "xmDialogShellWidgetClass", Button[0],
			              "labelString", Button[0]);
      OKText = XmStringLtoRCreate (Text, NORMAL_FONT);
      XtSetArg (args[n], XmNokLabelString, OKText); n++;

      Text = TOOLGetResString (Shell, "xmDialogShellWidgetClass", Button[1],
			              "labelString", Button[1]);
      CANCELText = XmStringLtoRCreate (Text, NORMAL_FONT);
      XtSetArg (args[n], XmNcancelLabelString, CANCELText); n++;

      Text = TOOLGetResString (Shell, "xmDialogShellWidgetClass", Button[2],
			              "labelString", Button[2]);
      HELPText = XmStringLtoRCreate (Text, NORMAL_FONT);
      XtSetArg (args[n], XmNhelpLabelString, HELPText); n++;

    case 0 :
    default :
      break;
  }

  /* let's create the message-box... */
  XtSetArg (args[n], XmNdefaultButtonType, XmDIALOG_OK_BUTTON); n++;
  XtSetArg (args[n], XmNdialogType,        XmDIALOG_INFORMATION); n++;
  Dialog = XmCreateMessageBox (Shell, "symbolBox", args, n);

  if (Msg != NULL)
    XmStringFree (Message);
  if (OKText != NULL)
    XmStringFree (OKText);
  if (CANCELText != NULL)
  XmStringFree (CANCELText);
  if (HELPText != NULL)
  XmStringFree (HELPText);

  if (Number == 0)
    XtAddEventHandler (Dialog, ButtonPressMask, False,
		                     (XtEventHandler)DLG_DlgEventHandle, &Result);

  /* unmanage unused buttons : */
  switch (Number)
  {
    case 0 :
      XtUnmanageChild (XmMessageBoxGetChild(Dialog, XmDIALOG_OK_BUTTON));
      XtUnmanageChild (XmMessageBoxGetChild(Dialog, XmDIALOG_CANCEL_BUTTON));
      XtUnmanageChild (XmMessageBoxGetChild(Dialog, XmDIALOG_HELP_BUTTON));
      XtUnmanageChild (XmMessageBoxGetChild(Dialog, XmDIALOG_SEPARATOR));
      break;
    case 1 : 
      XtAddCallback (Dialog, XmNcancelCallback,
		             (XtCallbackProc)Msg1CB, &Result);
      XtUnmanageChild (XmMessageBoxGetChild(Dialog, XmDIALOG_OK_BUTTON));
      XtUnmanageChild (XmMessageBoxGetChild(Dialog, XmDIALOG_HELP_BUTTON));
      break;
    case 2 :
      XtAddCallback (Dialog, XmNokCallback,   (XtCallbackProc)Msg1CB, &Result);
      XtAddCallback (Dialog, XmNhelpCallback, (XtCallbackProc)Msg2CB, &Result);
      XtUnmanageChild (XmMessageBoxGetChild(Dialog, XmDIALOG_CANCEL_BUTTON));
      break;
    default :
      XtAddCallback (Dialog, XmNokCallback,
		             (XtCallbackProc)Msg1CB, &Result);
      XtAddCallback (Dialog, XmNcancelCallback,
		             (XtCallbackProc)Msg2CB, &Result);
      XtAddCallback (Dialog, XmNhelpCallback,
		             (XtCallbackProc)Msg3CB, &Result);
  }

  if (Number == 0)
    Number = 1;

  XtSetMappedWhenManaged (Shell, False);
  XtManageChild (Dialog);

  if (DLGXPos == -1 && DLGYPos == -1)
  {
    XGetGeometry (XDisplay, XtWindow(Shell), &Dummy, 
                  &x, &y, &w, &h, &border, &bits);
    border = 10;
    TOOLGetMousePos (&xMouse, &yMouse);
    x = MIN (MAX(0, xMouse-(int)w/(2*Number)),
	     DisplayWidth(XDisplay, XScreen)-(int)(w+border));
    y = MIN (MAX(0, yMouse-(int)h+30),
	     DisplayHeight(XDisplay, XScreen)-(int)(h+border));
  }
  else
  {
    x = DLGXPos;
    y = DLGYPos;
  }
  sprintf (geo_string, "+%d+%d", x, y);
  XtSetArg (args[0], XmNgeometry, geo_string);
  XtSetArg (args[1], XmNx,        x);
  XtSetArg (args[2], XmNy,        y);
  XtSetValues (Shell, args, 3);

  XtManageChild (Shell);
  XtRealizeWidget (Shell);
  XtPopup (Shell, XtGrabExclusive);
  XFlush (XDisplay);

  PRO0 ("\\MainLoop...")
  TOOLMainLoop (AppCon, Shell);

  XmDestroyPixmap (XtScreen(Shell), PixMap);

  PRO_END
  return Result;
}


/* Callbackroutinen fr die Massage-Box: */
void Msg1CB (Widget widget, int *UserData, XtPointer Data)
{
  XtPopdown (XtParent(widget)); 
  XtDestroyWidget (XtParent(widget)); 

  *UserData = 1;
}

void Msg2CB (Widget widget, int *UserData, XtPointer Data)
{
  XtPopdown (XtParent(widget)); 
  XtDestroyWidget (XtParent(widget)); 

  *UserData = 2;
}

void Msg3CB (Widget widget, int *UserData, XtPointer Data)
{
  XtPopdown (XtParent(widget)); 
  XtDestroyWidget (XtParent(widget)); 

  *UserData = 3;
}
