#include <math.h>

#include "dlgData.h"
#include "dlgCB.h"
#include "dlgDnD.h"
#include "dlgFS.h"
#include "dlgDial.h"
#include "dlgSegment.h"


/* ***** Fest definierte Bitmaps: ***** */
/* Centerfeld des ARROW-Buttons im sensitiver Zustand */
static byte  CenterBMP[] = {0xf0, 0x01, 0xfc, 0x07, 0x1e, 0x0f, 0x06, 0x0c,
			    0x07, 0x1c, 0x03, 0x18, 0x03, 0x18, 0x03, 0x18,
			    0x07, 0x1c, 0x06, 0x0c, 0x1e, 0x0f, 0xfc, 0x07,
			    0xf0, 0x01};
/* CENTER-Feld des ARROW-Buttons im insensitiven Zustand */
static byte iCenterBMP[] = {0x50, 0x01, 0xa8, 0x02, 0x14, 0x05, 0x02, 0x08,
			    0x05, 0x14, 0x02, 0x08, 0x01, 0x10, 0x02, 0x08,
			    0x05, 0x14, 0x02, 0x08, 0x14, 0x05, 0xa8, 0x02,
			    0x50, 0x01};

static int  n = 0;
static Arg  args[NUM_ARGS];

static Widget DLG_CreateLabel (Widget Parent, DLGELEM *Elem)
{
  char     *Text = NULL;
  XmString  Label = NULL;
  Widget    Result;

#if 0
  /* TODO:
     Einbauen...
  */
  XmFontListEntry Entry;

  if ((Entry = XmFontListEntryLoad(XDisplay, "8x13bold", 
	          	   XmFONT_IS_FONT, XmFONTLIST_DEFAULT_TAG)) == NULL)
    PRO0 ("Font can not be created !\n")
  XmFontListAppendEntry (NULL, Entry);
#endif

  Text = TOOLGetResString (Parent, "xmLabelWidgetClass",  "label", 
			           "labelString", Elem->Value.frec.Text);

  Label = XmStringLtoRCreate (Text, NORMAL_FONT);
  Result = XtVaCreateManagedWidget ("label", xmLabelWidgetClass, Parent, 
				    XmNlabelString, Label,
	                            XmNalignment,   XmALIGNMENT_END,
				    NULL);
  if (Text[0] != '\0')
    XmStringFree (Label);

  DLG_BuildLabel (Parent, Result, Elem);

  Elem->WidgetID = Result;

  return Result;
}


static Widget DLG_CreateBool (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  Widget Result;

  Result = XtVaCreateManagedWidget (Elem->Name, xmToggleButtonWidgetClass, 
                                    Parent,
                                    XmNset, Elem->Value.drec.Value,
                                    NULL);
  XtAddCallback (Result, XmNvalueChangedCallback, 
		         (XtCallbackProc)DLG_DualCB, Elem);
  
  DLG_BuildLabel (Parent, Result, Elem);

  if (accelerate)
  {
    if (Elem->Dlg->Ok != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Ok);
    if (Elem->Dlg->Cancel != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Cancel);
  }

  Elem->WidgetID = Result;

  return Result;
}


static Widget DLG_CreateInt (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  char   Text[DLG_LLEN];
  Widget Result;

  DLG_Val2Str (Text, (double)Elem->Value.irec.Value, 
	             Elem->Value.irec.Digits, 0);

  XtSetArg (args[0], XmNbottomAttachment, XmATTACH_FORM);
  XtSetArg (args[1], XmNcolumns,          Elem->Value.irec.Digits);
  XtSetArg (args[2], XmNmaxLength,        Elem->Value.irec.Digits+1);
  XtSetArg (args[3], XmNvalue,            Text);
  XtSetArg (args[4], XmNalignment,        XmALIGNMENT_END);
  Result = XtCreateManagedWidget ("value", xmTextFieldWidgetClass, Parent, args, 5);

  XtAddCallback (Result, XmNfocusCallback, DLG_IntCB, Elem);
#ifdef USER_INPUT
  XtAddCallback (Result, XmNmodifyVerifyCallback,
		         (XtCallbackProc)DLG_IntCB,     Elem);
  XtAddCallback (Result, XmNmotionVerifyCallback,
		         (XtCallbackProc)DLG_MoveCB,    Elem);
  XtAddCallback (Result, XmNlosingFocusCallback,
		         (XtCallbackProc)DLG_IntCallCB, Elem);
#else
#if 1
  XtAddCallback (Result, XmNmodifyVerifyCallback,
		         (XtCallbackProc)DLG_IntCB,     Elem);
#endif
  XtAddCallback (Result, XmNactivateCallback,
		         (XtCallbackProc)DLG_IntCB, Elem);
  XtAddCallback (Result, XmNvalueChangedCallback, 
      		         (XtCallbackProc)DLG_IntCB, Elem);
  XtAddCallback (Result, XmNlosingFocusCallback,
		         (XtCallbackProc)DLG_IntCB, Elem);
#endif

  DLG_BuildLabel (Parent, Result, Elem);

  if (accelerate)
  {
    if (Elem->Dlg->Ok != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Ok);
    if (Elem->Dlg->Cancel != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Cancel);
  }

  Elem->WidgetID = Result;

  return Result;
}


static Widget DLG_CreateDouble (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  char   Text[DLG_LLEN];
  Widget Result;

  XtSetArg (args[0], XmNbottomAttachment, XmATTACH_FORM);
  XtSetArg (args[1], XmNcolumns,          Elem->Value.rrec.Digits);
  XtSetArg (args[2], XmNmaxLength,        Elem->Value.rrec.Digits+1);
  DLG_Val2Str (Text, Elem->Value.rrec.Value, Elem->Value.rrec.Digits,
		     Elem->Value.rrec.Decimals);
  XtSetArg (args[3], XmNvalue, Text);
  Result = XtCreateManagedWidget ("value", xmTextWidgetClass, Parent, args, 4);

  XtAddCallback (Result, XmNfocusCallback, 
		         (XtCallbackProc)DLG_DoubleCB, Elem);
#ifdef USER_INPUT
  XtAddCallback (Result, XmNmodifyVerifyCallback, 
		         (XtCallbackProc)DLG_DoubleCB, Elem);
  XtAddCallback (Result, XmNmotionVerifyCallback, 
		         (XtCallbackProc)DLG_MoveCB, Elem);
  XtAddCallback (Result, XmNlosingFocusCallback,  
		         (XtCallbackProc)DLG_DoubleCallCB, Elem);
#else
  XtAddCallback (Result, XmNactivateCallback,
		         (XtCallbackProc)DLG_DoubleCB, Elem);
  XtAddCallback (Result, XmNvalueChangedCallback, 
		         (XtCallbackProc)DLG_DoubleCB, Elem);
  XtAddCallback (Result, XmNlosingFocusCallback,  
		         (XtCallbackProc)DLG_DoubleCB, Elem);
#endif

  DLG_BuildLabel (Parent, Result, Elem);

  if (accelerate)
  {
    if (Elem->Dlg->Ok != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Ok);
    if (Elem->Dlg->Cancel != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Cancel);  
  }

  Elem->WidgetID = Result;

  return Result;
}


static Widget DLG_CreateArrow (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  Pixel  fg, bg;
  Pixmap pix1, pix2;
  Widget Result;

  if (Elem->Value.arec.Dir == DlgMIDDLE)
  {
    XtSetArg (args[0], XmNshadowType,        XmSHADOW_ETCHED_IN);
    XtSetArg (args[1], XmNpushButtonEnabled, TRUE);
    XtSetArg (args[2], XmNalignment,         XmALIGNMENT_CENTER);
    Result = XtCreateManagedWidget("button", xmPushButtonWidgetClass, 
				             Parent, args, 3);

    XtSetArg (args[0], XmNforeground, &fg);
    XtSetArg (args[1], XmNbackground, &bg);
    XtGetValues (Result, args, 2);

    pix1 = XCreatePixmapFromBitmapData (XDisplay, DefaultRootWindow(XDisplay),
					(char*)CenterBMP, 13, 13, fg, bg, 
					TOOLGetDepth());
    pix2 = XCreatePixmapFromBitmapData (XDisplay, DefaultRootWindow(XDisplay),
					(char*)iCenterBMP, 13, 13, fg, bg, 
					TOOLGetDepth());
    XtSetArg (args[0], XmNlabelType,              XmPIXMAP);
    XtSetArg (args[1], XmNlabelPixmap,            pix1);
    XtSetArg (args[2], XmNlabelInsensitivePixmap, pix2);
    XtSetArg (args[3], XmNmarginWidth,            1);
    XtSetArg (args[4], XmNmarginHeight,           1);
    XtSetValues (Result, args, 5);

    XtAddCallback (Result, XmNactivateCallback, 
		           (XtCallbackProc)DLG_ArrowCB, Elem);
  }
  else
  {              
    switch (Elem->Value.arec.Dir)
    {
      case DlgRIGHT : XtSetArg (args[0], XmNarrowDirection, XmARROW_RIGHT);
                      break;
      case  DlgLEFT : XtSetArg (args[0], XmNarrowDirection, XmARROW_LEFT);
                      break;
      case  DlgDOWN : XtSetArg (args[0], XmNarrowDirection, XmARROW_DOWN);
	              break;
            default : XtSetArg (args[0], XmNarrowDirection, XmARROW_UP);
    }
    Result = XtCreateManagedWidget ("button", xmArrowButtonWidgetClass,
				              Parent, args, 1);
    XtAddCallback (Result, XmNarmCallback, 
		           (XtCallbackProc)DLG_ArrowCB, Elem);
  }
  DLG_BuildLabel (Parent, Result, Elem);

  if (accelerate)
  {
    if (Elem->Dlg->Ok != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Ok);
    if (Elem->Dlg->Cancel != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Cancel);
  }

  Elem->WidgetID = Result;

  return Result;
}

static Widget DLG_CreateBar (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  double Distance;
  Widget Result;

  Distance = Elem->Value.brec.Max - Elem->Value.brec.Min;
  /*
  XtSetArg (args[n], XmNtopAttachment, XmATTACH_FORM); n++;
  */
  XtSetArg (args[n], XmNminimum,    (int)(Elem->Value.brec.Min * 100.0)); n++;
  XtSetArg (args[n], XmNmaximum,    (int)(Elem->Value.brec.Max*100.0 +
				          Distance * 10.0)); n++;
  XtSetArg (args[n], XmNvalue,      (int)(Elem->Value.brec.Val * 100.0)); n++;
  XtSetArg (args[n], XmNincrement,  (int)(Elem->Value.brec.Step * 100.0)); n++;
  XtSetArg (args[n], XmNsliderSize, (int)(Distance * 10.0)); n++;

  if (Elem->Value.brec.Side == DlgVERTICAL)
  {
    XtSetArg (args[n], XmNorientation, XmVERTICAL); n++;
    if (Elem->Value.brec.reverse)
    {
      XtSetArg (args[n], XmNprocessingDirection, XmMAX_ON_BOTTOM); n++;
    }
  }
  else
  {
    XtSetArg (args[n], XmNorientation, XmHORIZONTAL); n++;
    if (Elem->Value.brec.reverse)
    { 
      XtSetArg (args[n], XmNprocessingDirection, XmMAX_ON_LEFT); n++;
    }
  }

  Result = XtCreateManagedWidget (Elem->Name, xmScrollBarWidgetClass, 
				              Parent, args, n);
  DLG_BuildLabel (Parent, Result, Elem);

  XtAddCallback (Result, XmNvalueChangedCallback, 
		         (XtCallbackProc)DLG_BarCB, Elem);

  if (accelerate)
  {
    if (Elem->Dlg->Ok != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Ok);
    if (Elem->Dlg->Cancel != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Cancel);  
  }

  Elem->WidgetID = Result;

  return Result;
}


static Widget DLG_CreateCall (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  char     *Text = NULL;
  XmString  Label = NULL;
  Widget    Result;

  Text = TOOLGetResString (Parent, "xmPushButtonWidgetClass", "button",
			           "labelString", Elem->Name);
  Label = XmStringLtoRCreate (Text, NORMAL_FONT);
  XtSetArg (args[n], XmNlabelString, Label); n++;
  Result = XtCreateManagedWidget ("button", xmPushButtonWidgetClass, 
				            Parent, args, n);
  XmStringFree (Label);

  DLG_BuildLabel (Parent, Result, Elem);

  XtAddCallback (Result, XmNactivateCallback, 
		         (XtCallbackProc)DLG_ProcCB, Elem);

  if (accelerate)
  {
    if (Elem->Dlg->Ok != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Ok);
    if (Elem->Dlg->Cancel != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Cancel);   
  }

  Elem->WidgetID = Result;

 return Result;
}


static Widget DLG_CreateFSel (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  char      Default[PATH_MAX+1] = "";
  char      Mask[PATH_MAX+1] = "";
  char      FileName[PATH_MAX+1] = "";
  char      WidgetName[10];
  char     *Def = NULL;
  XmString  Label = NULL;
  Widget    Frame;
  Widget    Result;

  if (Elem->Value.grec.Length == 0)
  {
    /* Insert the file dialog directly */
    strcpy (WidgetName, (Elem->Type==DlgFSEL)? "fileSel":"dirSel");
    XtSetArg (args[n], XmNshadowType, XmSHADOW_ETCHED_IN); n++;
    Frame = XtCreateManagedWidget ("border", xmFrameWidgetClass,
				             Parent, args, 3);

    Def = TOOLGetResString (Frame, "xmFileSelectionBoxWidgetClass", 
			    WidgetName, "mask", Elem->Value.grec.Mask);
    if (strcmp(Def, "") != 0)
      TOOLParseFileName (Def, Mask, PATH_MAX);

    TOOLParseFileName (Elem->Value.grec.Default, Default, PATH_MAX);

    Result = DLG_CreateFileSel (Frame, Mask, Default,
				             2, Elem->Value.grec.Files);
    XtAddCallback (XmFileSelectionBoxGetChild(Result, XmDIALOG_TEXT),
		   XmNvalueChangedCallback, 
		   (XtCallbackProc)DLG_FSelChangeCB, Elem);
  }
  else
  {
    strcpy (Default, Elem->Value.grec.Result);
    Def = TOOLGetResString (Parent, "xmPushButtonWidgetClass",
			            Elem->Name, "default", Default); 
    TOOLParseFileName (Def, FileName, PATH_MAX);
    Label = XmStringLtoRCreate (FileName, NORMAL_FONT);
    XtSetArg (args[n], XmNalignment, 
	                 ((int)strlen(FileName) < Elem->Value.grec.Length) ?
  		           XmALIGNMENT_BEGINNING : XmALIGNMENT_END);  n++;
    XtSetArg (args[n], XmNlabelString,   Label); n++;
    /* enlarging will crash X - so prohibit resizing of the pushbutton */
    XtSetArg (args[n], XmNrecomputeSize, FALSE); n++;
    XtSetArg (args[n], XmNallowResize,   FALSE); n++;
    Result = XtCreateManagedWidget ("button", xmPushButtonWidgetClass, 
				              Parent, args, n);
    XmStringFree (Label);
    XtAddCallback (Result, XmNactivateCallback, 
		           (XtCallbackProc)DLG_FileSelCB, Elem);

    if (accelerate)
    {
      if (Elem->Dlg->Ok != NULL)
	XtInstallAccelerators (Result, Elem->Dlg->Ok);
      if (Elem->Dlg->Cancel != NULL)
	XtInstallAccelerators (Result, Elem->Dlg->Cancel);
    }
  }
  DLG_BuildLabel (Parent, Result, Elem);

  Elem->WidgetID = Result;

  return Result;
}


static Widget DLG_CreateList (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  int            count, Number;
  char          *Text = NULL;
  char           Name[12];
  XmStringTable  List;
  Widget         Result;

  Number = TOOLGetResInt (Parent, "xmListWidgetClass", "list", 
			          "number", Elem->Value.lrec.Number);

  XtSetArg (args[n], XmNlistSizePolicy,         XmVARIABLE); n++;
  XtSetArg (args[n], XmNscrollBarDisplayPolicy, XmAS_NEEDED); n++;
  Result = XmCreateScrolledList (Parent, "list", args, n);

  n = 0;
  List = (XmStringTable)malloc (sizeof(XmString) * Number);
  for (count = 0; count < Number; ++count)
  {
    sprintf (Name, "item[%d]", count);
    if (count < Elem->Value.lrec.Number)
      Text = TOOLGetResString (Result, "xmListWidgetClass", Name,
			       "labelString", Elem->Value.lrec.Item[count]);
    else
      Text = TOOLGetResString (Result, "xmListWidgetClass", Name,
			       "labelString", Name);
    /*
      TODO:
      Funktion, welche den String anhand von Steuercodes im Text aus 
      verschiedenen Fonts zusammensetzt (XmStringSegmentCreate)
    */
    if (Text[0] == '\001')
      List[count] = XmStringCreate (Text+1, "font1");
    else
      List[count] = XmStringCreate (Text, NORMAL_FONT);
  }
  XtSetArg (args[n], XmNitemCount,        Number); n++;
  XtSetArg (args[n], XmNitems,            List); n++;
  XtSetArg (args[n], XmNselectionPolicy,  XmSINGLE_SELECT); n++;
  XtSetArg (args[n], XmNvisibleItemCount, Elem->Value.lrec.Lines); n++;

  XtSetValues (Result, args, n);
  XtManageChild (Result);

  for (count = 0; count < Number; ++count)
    XmStringFree (List[count]);
  free (List);

  if (Elem->Value.lrec.Result >= 0 &&
      Elem->Value.lrec.Result < Elem->Value.lrec.Number)
    XmListSelectPos (Result, Elem->Value.lrec.Result+1, FALSE);

  if (Number > Elem->Value.lrec.Lines)
    DLG_BuildLabel (Parent, XtParent(Result), Elem);
  else
    DLG_BuildLabel (Parent, Result, Elem);

  XtAddCallback (Result, XmNsingleSelectionCallback, 
		         (XtCallbackProc)DLG_ListCB, Elem);

  if (accelerate)
  {
    if (Elem->Dlg->Ok != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Ok);
    if (Elem->Dlg->Cancel != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Cancel);
  }

  Elem->WidgetID = Result;

  return Result;
}


static Widget DLG_CreateOMenu (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  int       count;
  char      Name[12];
  char     *Text = NULL;
  XmString  Label = NULL;
  Widget    Menu, Result;

  Menu = XmCreatePulldownMenu (Parent, "menu", args, n);

  for (count = 0; count < Elem->Value.orec.Number; ++count)
  {
    if (Elem->Value.orec.Item[count].Label[0] != '\0')
    {
      sprintf (Name, "option[%d]", count);
      Text = TOOLGetResString (Menu, "xmPushButtonWidgetClass", 
			       Name, "labelString",
			       Elem->Value.orec.Item[count].Label);
      Label = XmStringLtoRCreate (Text, NORMAL_FONT);
      XtSetArg (args[0], XmNlabelString, Label);
      Elem->Value.orec.Item[count].WidgetID = 
	XmCreatePushButtonGadget (Menu, Name, args, 1);
      XtAddCallback (Elem->Value.orec.Item[count].WidgetID, 
		     XmNactivateCallback, (XtCallbackProc)DLG_MenuCB, 
		     &Elem->Value.orec.Item[count]);
      XmStringFree (Label);
    }
    else
    {
      sprintf (Name, "option[%d]", count);
      Elem->Value.orec.Item[count].WidgetID = 
	XmCreateSeparatorGadget (Menu, Name, NULL, 0);
    }
    XtManageChild (Elem->Value.orec.Item[count].WidgetID);
  }

  n = 0;
#if 0
  if (Elem->LabelPos == DlgWEST || Elem->LabelPos == DlgCENTER)
  {
    Text = TOOLGetResString (Parent, "xmMenuShellWidgetClass", 
			             "oMenu", "labelString", Elem->Name);
    Label = XmStringLtoRCreate ("", NORMAL_FONT);
    XtSetArg (args[n], XmNlabelString, Label); n++;
  }
#endif
  XtSetArg (args[n], XmNsubMenuId, Menu); n++;
  XtSetArg (args[n], XmNmenuHistory, 
	    Elem->Value.orec.Item[Elem->Value.orec.Result].WidgetID); n++;
  Result = XmCreateOptionMenu (Parent, "oMenu", args, n);
  XtManageChild (Result);

  /* Change color, if set by user */
  n = 0;
  if ((int)(Elem->fg) != -1)
  {
    XtSetArg (args[n], XmNforeground, Elem->fg); n++;
  }
  if ((int)(Elem->bg) != -1)
  {
    XtSetArg (args[n], XmNbackground, Elem->bg); n++;
  }
  if (n > 0)
    XtSetValues (XmOptionButtonGadget(Result), args, n);

  /* create label */
#if 0
  if (Elem->LabelPos == DlgWEST)
    XmStringFree (Label);
#endif
  DLG_BuildLabel (Parent, Result, Elem);

  if (accelerate)
  {
    if (Elem->Dlg->Ok != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Ok);
    if (Elem->Dlg->Cancel != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Cancel);
  }

  Elem->WidgetID = Result;

  return Result;
}


static Widget DLG_CreateRBox (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  int       count;
  char      Name[12];
  char     *Text = NULL;
  XmString  Label = NULL;
  Widget    Radio;
  Widget    Result;

  XtSetArg (args[n], XmNshadowType, XmSHADOW_ETCHED_IN); n++;
  Result = XtCreateManagedWidget ("border", xmFrameWidgetClass, 
				            Parent, args, n);
  DLG_BuildLabel (Parent, Result, Elem);

  XtSetArg (args[0], XmNradioBehavior,  TRUE);
  XtSetArg (args[1], XmNradioAlwaysOne, TRUE);
  XtSetArg (args[2], XmNpacking, XmPACK_COLUMN);
  if (Elem->Value.trec.RowCol > 0)
  {
    XtSetArg (args[3], XmNorientation, XmVERTICAL);
    XtSetArg (args[4], XmNnumColumns,  Elem->Value.trec.RowCol);
  }
  else
  {
    XtSetArg (args[3], XmNorientation, XmHORIZONTAL);
    XtSetArg (args[4], XmNnumColumns,  -Elem->Value.trec.RowCol);
  }
  Radio = XtCreateManagedWidget ("radio", xmRowColumnWidgetClass,
				          Result, args, 5);

  for (count = 0; count < Elem->Value.trec.Number; ++count)
  {
    sprintf (Name, "button[%d]", count);
    Text = TOOLGetResString (Radio, "xmToggleButtonWidgetClass", Name, 
			            "labelString", 
			            Elem->Value.trec.Toggle[count].Item);
    Label = XmStringLtoRCreate (Text, NORMAL_FONT);
    XtSetArg (args[0], XmNlabelString, Label);
    XtSetArg (args[1], XmNset, (Boolean)Elem->Value.trec.Toggle[count].Value);
    Elem->Value.trec.Toggle[count].WidgetID = 
      XtCreateManagedWidget(Name, xmToggleButtonWidgetClass, Radio, args, 2);
    XmStringFree (Label);

    XtAddCallback (Elem->Value.trec.Toggle[count].WidgetID, 
		   XmNvalueChangedCallback, (XtCallbackProc)DLG_RadioCB, 
		   &Elem->Value.trec.Toggle[count]);

    if (accelerate)
    {
      if (Elem->Dlg->Ok != NULL)
	XtInstallAccelerators (Elem->Value.trec.Toggle[count].WidgetID, 
			       Elem->Dlg->Ok);
      if (Elem->Dlg->Cancel != NULL)
	XtInstallAccelerators (Elem->Value.trec.Toggle[count].WidgetID, 
			       Elem->Dlg->Cancel);  
    }
  }
  Elem->WidgetID = Result;

  return Result;
}


static Widget DLG_CreateScale (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  double Factor;
  Widget Result;

  Factor = pow (10.0, Elem->Value.zrec.Decimals);
  XtSetArg (args[n], XmNminimum, (int)(Elem->Value.zrec.Min * Factor +
				       GREATER(Elem->Value.zrec.Min))); n++;
  XtSetArg (args[n], XmNmaximum, (int)(Elem->Value.zrec.Max * Factor + 
				       GREATER(Elem->Value.zrec.Max))); n++;
  XtSetArg (args[n], XmNvalue,   (int)(Elem->Value.zrec.Val * Factor + 
			               GREATER(Elem->Value.zrec.Val))); n++;
  if (Elem->Value.zrec.Max - Elem->Value.zrec.Min > 10.0)
  {
    XtSetArg (args[n], XmNscaleMultiple, 10); n++;
  }
  XtSetArg (args[n], XmNshowValue, TRUE); n++;

  if (Elem->Value.zrec.Decimals > 0)
  {
    XtSetArg (args[n], XmNdecimalPoints, Elem->Value.zrec.Decimals); n++;
  }

  XtSetArg (args[n], XmNorientation, (Elem->Value.zrec.Side==DlgVERTICAL)? 
	                                XmVERTICAL:XmHORIZONTAL); n++;
  if (Elem->Value.zrec.reverse)
  {
    XtSetArg (args[n], XmNprocessingDirection, 
	               (Elem->Value.zrec.Side==DlgVERTICAL)?
	                 XmMAX_ON_BOTTOM:XmMAX_ON_LEFT); n++;
  }

  Result = XtCreateManagedWidget ("scale", xmScaleWidgetClass, 
				           Parent, args, n);
  XtAddCallback (Result, XmNvalueChangedCallback, 
		         (XtCallbackProc)DLG_ScaleCB, Elem);

  if (accelerate)
  {
    if (Elem->Dlg->Ok != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Ok);
    if (Elem->Dlg->Cancel != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Cancel);  
  }
  Elem->WidgetID = Result;

  DLG_BuildLabel (Parent, Result, Elem);

  return Result;
}


static Widget DLG_CreateText (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  Widget Result;
  
  XtSetArg (args[n], XmNcolumns,      abs(Elem->Value.srec.Cols)); n++;
  XtSetArg (args[n], XmNrows,         abs(Elem->Value.srec.Rows)); n++;
  XtSetArg (args[n], XmNmaxLength,    Elem->Value.srec.Length); n++;
  XtSetArg (args[n], XmNvalue,        Elem->Value.srec.Text); n++;
  XtSetArg (args[n], XmNresizeWidth,  FALSE); n++;
  XtSetArg (args[n], XmNresizeHeight, FALSE); n++;
  if (abs(Elem->Value.srec.Rows) > 1)
  {
    XtSetArg (args[n], XmNeditMode, XmMULTI_LINE_EDIT); n++;
  }
  if (Elem->Value.srec.Cols < 0 || Elem->Value.srec.Rows < 0)
  {
    XtSetArg (args[n], XmNscrollHorizontal, (Elem->Value.srec.Cols < 0)); n++;
    XtSetArg (args[n], XmNscrollVertical,   (Elem->Value.srec.Rows < 0)); n++;
    Result = XmCreateScrolledText (Parent, "text", args, n);
  }
  else
    Result = XtCreateWidget ("text", xmTextWidgetClass, Parent, args, n);
  XtManageChild (Result);
  DLG_BuildLabel (Parent, Result, Elem);

  XtAddCallback (Result, XmNactivateCallback,
		         (XtCallbackProc)DLG_TextCB, Elem);
  XtAddCallback (Result, XmNlosingFocusCallback,
		         (XtCallbackProc)DLG_TextCB, Elem);

  Elem->WidgetID = Result;

  return Result;
}


static Widget DLG_CreateTBox (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  int       count;
  char      Name[12];
  char     *Text = NULL;
  XmString  Label = NULL;
  Widget    Toggle;
  Widget    Result;

  XtSetArg (args[n], XmNshadowType, XmSHADOW_ETCHED_IN); n++;
  Result = XtCreateManagedWidget ("border", xmFrameWidgetClass,
				            Parent, args, n);
  DLG_BuildLabel (Parent, Result, Elem);

  XtSetArg (args[0], XmNradioBehavior,  FALSE);
  XtSetArg (args[1], XmNradioAlwaysOne, FALSE);
  if (Elem->Value.trec.RowCol > 0)
  {
    XtSetArg (args[2], XmNorientation, XmVERTICAL);
    XtSetArg (args[3], XmNpacking,     XmPACK_COLUMN);
    XtSetArg (args[4], XmNnumColumns,  Elem->Value.trec.RowCol);
  }
  else
  {
    XtSetArg (args[2], XmNorientation, XmHORIZONTAL);
    XtSetArg (args[3], XmNpacking,     XmPACK_COLUMN);
    XtSetArg (args[4], XmNnumColumns,  -1*Elem->Value.trec.RowCol);
  }
  Toggle = XtCreateManagedWidget ("toggle", xmRowColumnWidgetClass,
					    Result, args, 5);

  for (count = 0; count < Elem->Value.trec.Number; ++count)
  {
    sprintf (Name, "button[%d]", count);
    Text = TOOLGetResString (Toggle, "xmToggleButtonWidgetClass", Name,
			     "labelString", 
			     Elem->Value.trec.Toggle[count].Item);
    Label = XmStringLtoRCreate (Text, NORMAL_FONT);
    XtSetArg (args[0], XmNlabelString, Label);
    XtSetArg (args[1], XmNset, (Boolean)Elem->Value.trec.Toggle[count].Value);
    Elem->Value.trec.Toggle[count].WidgetID = 
      XtCreateManagedWidget(Name, xmToggleButtonWidgetClass, Toggle, args, 2);
    XmStringFree (Label);

    XtAddCallback (Elem->Value.trec.Toggle[count].WidgetID, 
		   XmNvalueChangedCallback, (XtCallbackProc)DLG_ToggleCB,  
		   &Elem->Value.trec.Toggle[count]);
    if (accelerate)
    {
      if (Elem->Dlg->Ok != NULL)
	XtInstallAccelerators (Elem->Value.trec.Toggle[count].WidgetID, 
			       Elem->Dlg->Ok);
      if (Elem->Dlg->Cancel != NULL)
	XtInstallAccelerators (Elem->Value.trec.Toggle[count].WidgetID, 
			       Elem->Dlg->Cancel);  
    }
  }

  Elem->WidgetID = Toggle;

  return Result;
}


static Widget DLG_CreateUpDn (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  char   Text[DLG_LLEN];
  Widget Result;
  Widget Up, Down, Value;

  Result = XtCreateManagedWidget (Elem->Name, xmFormWidgetClass, 
				              Parent, args, n);
  DLG_BuildLabel (Parent, Result, Elem);

  DLG_Val2Str (Text, Elem->Value.urec.Val, Elem->Value.urec.Length, 
		                           Elem->Value.urec.Decimals);
  XtSetArg (args[0], XmNrightAttachment,  XmATTACH_FORM);
  XtSetArg (args[1], XmNbottomAttachment, XmATTACH_FORM);
  XtSetArg (args[2], XmNcolumns,          Elem->Value.urec.Length);
  XtSetArg (args[3], XmNmaxLength,        Elem->Value.urec.Length);
  XtSetArg (args[4], XmNvalue,            Text);
  XtSetArg (args[5], XmNcursorPosition,   Elem->Value.urec.Length);
  Value = XtCreateManagedWidget ("value", xmTextWidgetClass, Result, args, 6);

  XtAddCallback (Value, XmNfocusCallback, 
		        (XtCallbackProc)DLG_UpDnValCB, Elem);
  XtAddCallback (Value, XmNactivateCallback,
		        (XtCallbackProc)DLG_UpDnValCB, Elem);
  XtAddCallback (Value, XmNvalueChangedCallback,
		        (XtCallbackProc)DLG_UpDnValCB, Elem);
  XtAddCallback (Value, XmNlosingFocusCallback,
		        (XtCallbackProc)DLG_UpDnValCB, Elem);
  if (accelerate)
  {
    if (Elem->Dlg->Ok != NULL)
      XtInstallAccelerators (Value, Elem->Dlg->Ok);
    if (Elem->Dlg->Cancel != NULL)
      XtInstallAccelerators (Value, Elem->Dlg->Cancel);
  }

  XtSetArg (args[0], XmNtopAttachment,    XmATTACH_FORM);
  XtSetArg (args[1], XmNbottomAttachment, XmATTACH_FORM);
  XtSetArg (args[2], XmNrightAttachment,  XmATTACH_WIDGET);
  XtSetArg (args[3], XmNrightWidget,      Value);
  XtSetArg (args[4], XmNarrowDirection,   XmARROW_UP);
  Up = XtCreateManagedWidget ("up", xmArrowButtonWidgetClass, Result, args, 5);
  XtAddCallback (Up, XmNarmCallback, (XtCallbackProc)DLG_UpCB, Elem);
  if (accelerate)
  {
    if (Elem->Dlg->Ok != NULL)
      XtInstallAccelerators (Up, Elem->Dlg->Ok);
    if (Elem->Dlg->Cancel != NULL)
      XtInstallAccelerators (Up, Elem->Dlg->Cancel);  
  }

  XtSetArg (args[0], XmNtopAttachment,    XmATTACH_FORM);
  XtSetArg (args[1], XmNbottomAttachment, XmATTACH_FORM);
  XtSetArg (args[2], XmNrightAttachment,  XmATTACH_WIDGET);
  XtSetArg (args[3], XmNrightWidget,      Up);
  XtSetArg (args[4], XmNarrowDirection,   XmARROW_DOWN);
  Down = XtCreateManagedWidget ("down", xmArrowButtonWidgetClass,
				        Result, args, 5);
  XtAddCallback (Down, XmNarmCallback, (XtCallbackProc)DLG_DnCB, Elem);
  if (accelerate)
  {
    if (Elem->Dlg->Ok != NULL)
      XtInstallAccelerators (Down, Elem->Dlg->Ok);
    if (Elem->Dlg->Cancel != NULL)
      XtInstallAccelerators (Down, Elem->Dlg->Cancel);
  }

  Elem->WidgetID = Value;

  return Result;
}


static Widget DLG_CreateBMap (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  dual    Button; 
  char    Class[7];
  char    FileName[PATH_MAX+1];
  char   *Text = NULL;
  Pixmap  pix1, pix2;
  Widget  Result;

  Button = (Elem->Value.prec.FPtr != NULL);
  strcpy (Class, Button? "button":"icon");

  XtSetArg (args[n], XmNpushButtonEnabled, Button); n++;
  if (Button)
  {
    XtSetArg (args[n], XmNshadowType, XmSHADOW_ETCHED_IN); n++;
    Result = XtCreateManagedWidget (Class, xmPushButtonWidgetClass, 
				           Parent, args, n);
  }
  else
    Result = XtCreateManagedWidget (Class, xmLabelWidgetClass, 
				           Parent, args, n);

  if (Elem->Value.prec.bg == 0)
  {
    XtSetArg (args[0], XmNbackground, &Elem->Value.prec.bg);
    XtGetValues (Result, args, 1);
  }

  if (Elem->Value.prec.fg == 0)
  {
    XtSetArg (args[0], XmNforeground, &Elem->Value.prec.fg);
    XtGetValues (Result, args, 1);
  }

  if (Elem->Value.prec.PicName[0] != '\0')
  {
    Text = TOOLGetResString (Result, "xmPushButtonWidgetClass", Class, 
			             "fileName", Elem->Value.prec.PicName);
    TOOLParseFileName (Text, FileName, PATH_MAX);
    Elem->Value.prec.Data = (char*)TOOLLoadXBMData (FileName, 
						    &Elem->Value.prec.SizeX,
						    &Elem->Value.prec.SizeY,
						    NULL, NULL);
    if (Elem->Value.prec.Data == NULL)
    {
      Elem->Value.prec.SizeX = nopic_width;
      Elem->Value.prec.SizeY = nopic_height;
      Elem->Value.prec.Data = TOOLCopyXBM (nopic_bits, Elem->Value.prec.SizeX,
					               Elem->Value.prec.SizeY);
    }
  }

  if (Elem->Value.prec.Data == NULL)
    Elem->Value.prec.Data = TOOLMakeEmptyXBM (Elem->Value.prec.SizeX,
					      Elem->Value.prec.SizeY);

  pix1 = XCreatePixmapFromBitmapData (XDisplay, DefaultRootWindow(XDisplay),
				      Elem->Value.prec.Data,
				      Elem->Value.prec.SizeX,
				      Elem->Value.prec.SizeY,
				      Elem->Value.prec.fg,
				      Elem->Value.prec.bg,
				      TOOLGetDepth());

  n = 0;
  XtSetArg (args[n], XmNlabelType,   XmPIXMAP); n++;
  XtSetArg (args[n], XmNlabelPixmap, pix1); n++;
  XtSetArg (args[n], XmNforeground,  Elem->Value.prec.fg); n++;
  XtSetArg (args[n], XmNbackground,  Elem->Value.prec.bg); n++;

  if (Button)
  {
    Elem->Value.prec.iData = TOOLLightenXBM (Elem->Value.prec.Data,
					     Elem->Value.prec.SizeX,
					     Elem->Value.prec.SizeY);
    pix2 = XCreatePixmapFromBitmapData (XDisplay,
					DefaultRootWindow(XDisplay),
					Elem->Value.prec.iData,
					Elem->Value.prec.SizeX,
					Elem->Value.prec.SizeY,
					Elem->Value.prec.fg,
					Elem->Value.prec.bg,
					TOOLGetDepth());
    XtSetArg (args[n], XmNlabelInsensitivePixmap, pix2); n++;
  }

  XtSetValues (Result, args, n);

  if (Button)
  {
    DLG_BuildLabel (Parent, Result, Elem);

    XtAddCallback (Result, XmNactivateCallback, 
		           (XtCallbackProc)DLG_PicCB, Elem);

    if (accelerate)
    {
      if (Elem->Dlg->Ok != NULL)
	XtInstallAccelerators (Result, Elem->Dlg->Ok);
      if (Elem->Dlg->Cancel != NULL)
	XtInstallAccelerators (Result, Elem->Dlg->Cancel);
    }
  }
  Elem->WidgetID = Result;

  return Result;
}


static Widget DLG_CreatePic (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  dual    Button;
  char    FileName[PATH_MAX+1] = "";
  char    Class[7];
  char   *Text = NULL;
  Pixel   Color;
  Pixmap  PixMap;
  Widget  Result;
 
  if (Elem->Value.prec.Format != DlgXPM)
  /* TODO:
     sonst laden mit XmGetPixmap (d.h. 1 Fkt: CreatePic+CreateBMap) ?
  */
    return NULL;
 
  Button = (Elem->Value.prec.FPtr != NULL);
  strcpy (Class, Button? "button":"icon");
 
  if (! Button)
    Result = XtCreateManagedWidget (Class, xmLabelWidgetClass,
                                           Parent, args, n);
  else
  {
    XtSetArg (args[n], XmNshadowType, XmSHADOW_ETCHED_IN); n++;
    XtSetArg (args[n], XmNpushButtonEnabled,        TRUE); n++;
    Result = XtCreateWidget (Class, xmPushButtonWidgetClass, Parent, args, n);
    XtAddCallback (Result, XmNactivateCallback,
		           (XtCallbackProc)DLG_PicCB, Elem);
  }
 
  XtVaGetValues (Result, XmNbackground, &Color, NULL);
 
  if (Elem->Value.prec.PicName[0] != '\0')
  {
    Text = TOOLGetResString (Parent, Button?
                             "xmPushButtonWidgetClass":"xmLabelWidgetClass",
                             Class, "fileName", Elem->Value.prec.PicName);
    TOOLParseFileName (Text, FileName, PATH_MAX);
  }
 
  if (FileName[0] != '\0')
    PixMap = TOOLLoadXPM (FileName, NULL, NULL, Color);
  else
    PixMap = TOOLMakeXPM (Elem->Value.prec.Data, NULL, NULL, Color);
 
  XtSetArg (args[0], XmNlabelType,              XmPIXMAP);
  XtSetArg (args[1], XmNlabelPixmap,            PixMap);
  XtSetArg (args[2], XmNlabelInsensitivePixmap, PixMap);
  XtSetValues (Result, args, 3);
 
  XtManageChild (Result);           
  DLG_BuildLabel (Parent, Result, Elem);
 
  Elem->WidgetID = Result;
 
  return Result;
}
   

static Widget DLG_CreateImage (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  dual    Button; 
  char    Class[7];
  Pixmap  PixMap;
  Widget  Result;

  Button = (Elem->Value.xrec.FPtr != NULL);
  strcpy (Class, Button? "button":"icon");

  if (! Button)
    Result = XtCreateManagedWidget (Class, xmLabelWidgetClass, 
				           Parent, args, n);
  else
  {
    XtSetArg (args[n], XmNshadowType,        XmSHADOW_ETCHED_IN); n++;
    XtSetArg (args[n], XmNpushButtonEnabled, TRUE);               n++;
    Result = XtCreateWidget (Class, xmPushButtonWidgetClass, Parent, args, n);
    XtAddCallback (Result, XmNactivateCallback, 
		           (XtCallbackProc)DLG_ImageCB, Elem);
  }

  if (Elem->Value.xrec.Image != NULL)
  {
    PixMap = TOOLImage2Pixmap (Elem->Value.xrec.Image);
    XtVaSetValues (Result, XmNlabelType,              XmPIXMAP,
		           XmNlabelPixmap,            PixMap,
    		           XmNlabelInsensitivePixmap, PixMap,
       		           NULL);
  }

  XtManageChild (Result);

  DLG_BuildLabel (Parent, Result, Elem);
  
  Elem->WidgetID = Result;

  return Result;
}


#if (XmVERSION>=2)
static Widget DLG_CreateNote (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  int       count = 1;
  char      Name[10];
  NOTEPAGE *Ptr;
  Widget    Result, Page, Child;

  XtSetArg (args[n], XmNcurrentPageNumber, Elem->Value.nrec.ActPage); n++;
  XtSetArg (args[n], XmNlastPageNumber,    Elem->Value.nrec.Number);  n++;
  XtSetArg (args[n], XmNbackPagePlacement, XmTOP_RIGHT);              n++;

  switch (Elem->Value.nrec.Style)
  {
    case DlgSOLID :
      XtSetArg (args[n], XmNbindingType, XmSOLID);      n++;
      XtSetArg (args[n], XmNorientation, XmHORIZONTAL); n++;
      break;

    case DlgSPIRAL :
      XtSetArg (args[n], XmNbindingType, XmSPIRAL);     n++;
      XtSetArg (args[n], XmNorientation, XmHORIZONTAL); n++;
      break;

    default :
    case DlgNOBINDING :
      XtSetArg (args[n], XmNbindingType, XmNONE);     n++;
      XtSetArg (args[n], XmNorientation, XmVERTICAL); n++;
  }
  Result = XmCreateNotebook (Parent, Elem->Name, args, n);
  XtAddCallback (Result, XmNpageChangedCallback, 
		         (XtCallbackProc)DLG_NoteCB, Elem); 

  Ptr = Elem->Value.nrec.Page;
  while (Ptr != NULL)
  {
    PRO2 ("NOTE-PAGE %d: %s\n", count+1, Ptr->Label)

    sprintf (Name, "page[%d]", count);
    Page = XtCreateManagedWidget (Name, xmFormWidgetClass, Result, args, 0);

#if 1
    DLG_CreateBox (Page, Page, Page, Ptr->Box, DlgVERTICAL);
#else
    DLG_CreateBox (Page, Ptr->Box);
#endif
    switch (Ptr->Type)
    {
      case DlgMAJOR :
	XtSetArg (args[0], XmNnotebookChildType, XmMAJOR_TAB);
	break;

      case DlgMINOR :
	XtSetArg (args[0], XmNnotebookChildType, XmMINOR_TAB);
	break;

      default :
      case DlgPAGE :
	XtSetArg (args[0], XmNnotebookChildType, XmPAGE);
    }
    Child = XtCreateManagedWidget (Ptr->Label, xmPushButtonWidgetClass, 
				               Result, args, 1);

    Ptr = Ptr->Next;
    count++;
  }

  XtManageChild (Result);

  DLG_BuildLabel (Parent, Result, Elem);

  Elem->WidgetID = Result;

  return Result;   
}


static void DLG_CreateTreeItem (Widget Container, Widget Parent, DLGELEM *Elem,
				TREEITEM *List, Pixel bg, dual accelerate)
{
  TREEITEM *Ptr = List;
  XmString  Details[1];
  XmString  Name;
  Pixmap    Icon;

  while (Ptr != NULL)
  {
    if (Ptr->Icon != NULL)
      Icon = TOOLMakeXPM (Ptr->Icon, NULL, NULL, bg);
    else
      Icon = XmUNSPECIFIED_PIXMAP;

    if (Icon != XmUNSPECIFIED_PIXMAP)
    {
      Name = XmStringCreateLocalized (Ptr->Label);
      Details[0] = XmStringCreateLocalized (" ");
      Ptr->gadget =
	XtVaCreateManagedWidget (Ptr->Label, xmIconGadgetClass, Container,
				 XmNlabelString,     Name, 
				 XmNentryParent,     Parent,
				 XmNsmallIconPixmap, Icon,
				 XmNoutlineState,
				   (Ptr->State)? XmEXPANDED:XmCOLLAPSED,
				 XmNshadowThickness, 0,
				 XmNdetailCount,     1,
				 XmNdetail,          Details,
				 XmNmarginHeight,    0,
				 NULL);
      XmStringFree (Details[0]);
      XmStringFree (Name);
    }
    else
      Ptr->gadget = 
	XtVaCreateManagedWidget (Ptr->Label, xmPushButtonGadgetClass, 
				 Container,
				 XmNentryParent,     Parent,
				 XmNmarginHeight,    0,
				 XmNoutlineState,
				   (Ptr->State)? XmEXPANDED:XmCOLLAPSED,
				 NULL);

    XtAddCallback (Ptr->gadget, 
		   XmNhelpCallback, (XtCallbackProc)DLG_Help2CB, Elem);

    if (accelerate)
    {
      if (Elem->Dlg->Ok != NULL)
	XtInstallAccelerators (Ptr->gadget, Elem->Dlg->Ok);
      if (Elem->Dlg->Cancel != NULL)
	XtInstallAccelerators (Ptr->gadget, Elem->Dlg->Cancel);
    }

    if (Ptr->Sub != NULL)
      DLG_CreateTreeItem (Container, Ptr->gadget, Elem, Ptr->Sub, bg,
			  accelerate);

    Ptr = Ptr->Next;
  }
}


static Widget DLG_CreateTree (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  Pixel  fg, bg;
  Widget Result;

  if (Elem->Value.krec.Lines > 0)
  {
    Parent = XtVaCreateManagedWidget ("treeScrWin",
				      xmScrolledWindowWidgetClass, Parent,
				      XmNscrollingPolicy,          XmAUTOMATIC,
				      XmNscrollBarDisplayPolicy,   XmAS_NEEDED,
				      XmNvisualPolicy,             XmCONSTANT,
				      XmNheight, Elem->Value.krec.Lines * 20,
				      XmNwidth,                    250,
				      NULL);
    XtAddCallback (Parent, XmNhelpCallback, (XtCallbackProc)DLG_Help2CB, Elem);
  }

  n = 0;
  XtSetArg (args[n], XmNlayoutType,               XmDETAIL);          n++;
  XtSetArg (args[n], XmNentryViewType,            XmSMALL_ICON);      n++;
  XtSetArg (args[n], XmNdetailColumnHeadingCount, 0);                 n++;
  XtSetArg (args[n], XmNselectColor,     TOOLGetColor("lightblue"));  n++;
  XtSetArg (args[n], XmNoutlineIndentation,       15);                n++;
  XtSetArg (args[n], XmNoutlineLineStyle,         XmSINGLE);          n++;
  XtSetArg (args[n], XmNsmallCellHeight,          8);                 n++;
  XtSetArg (args[n], XmNautomaticSelection,       XmNO_AUTO_SELECT);  n++;
  /* Set the selection-mode to no-auto-select, because if the applications 
     opens an popupmenu on selection, the mouse focus will change to this 
     window. If the mouse button is release then (like for short clicks) the
     menu would disappear. */

  if (Elem->Value.krec.ClosePic != XmUNSPECIFIED_PIXMAP)
  {
    XtSetArg (args[n], XmNcollapsedStatePixmap, Elem->Value.krec.ClosePic);
    n++;
  }
  if (Elem->Value.krec.OpenPic != XmUNSPECIFIED_PIXMAP)
  {
    XtSetArg (args[n], XmNexpandedStatePixmap, Elem->Value.krec.OpenPic);  
    n++;
  }
#if 0
  XtSetArg (args[n], XmNspatialIncludeModel,   XmAPPEND);           n++;
  XtSetArg (args[n], XmNspatialStyle,          XmGRID);             n++;
  XtSetArg (args[n], XmNspatialResizeModel,    XmGROW_MINOR);       n++;
  XtSetArg (args[n], XmNspatialSnapModel,      XmCENTER);           n++;
  XtSetArg (args[n], XmNselectionPolicy,       XmEXTENDED_SELECT);  n++;
  XtSetArg (args[n], XmNdetailColumnHeading,   Headings);           n++;
#endif

  Result = XtCreateManagedWidget (Elem->Name, xmContainerWidgetClass, 
				              Parent, args, n);

  XtAddCallback (Result, XmNdefaultActionCallback, 
		         (XtCallbackProc)DLG_TreeCB, Elem);
  XtAddCallback (Result, XmNselectionCallback, 
		         (XtCallbackProc)DLG_TreeCB, Elem);

  XtVaGetValues (Result, XmNforeground, &fg,
		         XmNbackground, &bg,
		         NULL);
  DLG_CreateTreeItem (Result, NULL, Elem, Elem->Value.krec.Data, bg, 
		                    accelerate);

  XmContainerRelayout (Result);

  Elem->WidgetID = Result;
  return Result;
}

#endif


static Widget DLG_CreateDial (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  double Factor;
  Widget Result;

  Factor = pow (10.0, Elem->Value.erec.Decimals);
  XtSetArg (args[n], DlgNminimum, (int)(Elem->Value.erec.Min * Factor + 
					GREATER(Elem->Value.erec.Min))); n++;
  XtSetArg (args[n], DlgNmaximum, (int)(Elem->Value.erec.Max * Factor + 
					GREATER(Elem->Value.erec.Max))); n++;
  XtSetArg (args[n], DlgNvalue,   (int)(Elem->Value.erec.Val * Factor +
					GREATER(Elem->Value.erec.Val))); n++;
  XtSetArg (args[n], DlgNdigits,   Elem->Value.erec.Digits);   n++;
  XtSetArg (args[n], DlgNdecimals, Elem->Value.erec.Decimals); n++;

  Result = DLGCreateDial (Parent, "dial", args, n, DLG_DialCB, 
			          (XtPointer)Elem);

  if (accelerate)
  {
    if (Elem->Dlg->Ok != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Ok);
    if (Elem->Dlg->Cancel != NULL)
      XtInstallAccelerators (Result, Elem->Dlg->Cancel);
  }

  DLG_BuildLabel (Parent, Result, Elem);

  Elem->WidgetID = Result;

  return Result;
}


static Widget DLG_CreateSegment (Widget Parent, DLGELEM *Elem)
{
  double Factor;
  Widget Result;
#if (XmVERSION>=2)
  static char    SegmTrans[] = "<Btn2Down>:DragAction()";
  XtTranslations NewTable;
#endif

  Factor = pow (10.0, Elem->Value.hrec.Decimals);
  XtSetArg (args[n], DlgNvalue,   (int)(Elem->Value.hrec.Val * Factor +
					GREATER(Elem->Value.hrec.Val))); n++;
  XtSetArg (args[n], DlgNdigits,   Elem->Value.hrec.Digits); n++;
  XtSetArg (args[n], DlgNdecimals, Elem->Value.hrec.Decimals); n++;
  Result = XtCreateManagedWidget ("segment", segmentWidgetClass, 
				             Parent,args, n);
  Elem->WidgetID = Result;

#if (XmVERSION>=2)
  NewTable = XtParseTranslationTable (SegmTrans);
  XtOverrideTranslations (Result, NewTable);
#endif

  DLG_BuildLabel (Parent, Result, Elem);

  return Result;
}

static Widget DLG_CreateUser (Widget Parent, DLGELEM *Elem)
{
  Widget Result;

  Result = Elem->Value.wrec.FPtr (Elem->Dlg, Parent, Elem->Value.wrec.Param);

  Elem->WidgetID = Result;

  return Result;
}


Widget DLGCreateSep (Widget Parent, DLGELEM *Elem)
{
  Widget Result;

  if (abs(Elem->Value.mrec.x2 - Elem->Value.mrec.x1) <
      abs(Elem->Value.mrec.y2 - Elem->Value.mrec.y1))
    XtSetArg (args[n], XmNorientation, XmHORIZONTAL);
  else
    XtSetArg (args[n], XmNorientation, XmVERTICAL);
  n++;

  Result = XtCreateManagedWidget ("separator", xmSeparatorWidgetClass, 
				               Parent, args, n);

  return Result;
}


Widget DLG_CreateSep (Widget Parent, Widget Last, DLGBOX *Box, int Pos)
{
  char   Style;
  Widget Result = NULL;

  PRO_START ("")

  if ((Style = DLG_GetSeparator(Box->Sep, Pos)) == DlgNO_LINE[0])
    PRO_RETURN ("Seperator Type is NONE", Last)

  PRO2 ("\\<%d %c>...", Box->BoxID, Style)

  if (Style == DlgSINGLE_LINE[0])
    XtSetArg (args[0], XmNseparatorType, XmSINGLE_LINE);
  elsif (Style == DlgETCHED_IN[0])
    XtSetArg (args[0], XmNseparatorType, XmSHADOW_ETCHED_IN);
  elsif (Style == DlgETCHED_OUT[0])
    XtSetArg (args[0], XmNseparatorType, XmSHADOW_ETCHED_OUT);
  elsif (Style == DlgDASH_LINE[0])
    XtSetArg (args[0], XmNseparatorType, XmSINGLE_DASHED_LINE);
  elsif (Style == DlgDOUBLE_LINE[0])
    XtSetArg (args[0], XmNseparatorType, XmDOUBLE_LINE);
  else
  {
    PRO1 ("Style: %d\n", Style)
    PRO_ERROR0 ("Invalid separator", Last)
  }

  if (Box->Dir == BOXROWS)
  {
    XtSetArg (args[1], XmNtopAttachment,   XmATTACH_WIDGET);
    XtSetArg (args[2], XmNtopWidget,       Last);
    XtSetArg (args[3], XmNtopOffset,       3);
    XtSetArg (args[4], XmNleftAttachment,  XmATTACH_FORM);
    XtSetArg (args[5], XmNrightAttachment, XmATTACH_FORM);
    XtSetArg (args[6], XmNorientation,     XmHORIZONTAL);
  }
  else
  {
    XtSetArg (args[1], XmNleftAttachment,   XmATTACH_WIDGET);
    XtSetArg (args[2], XmNleftWidget,       Last);
    XtSetArg (args[3], XmNleftOffset,       3);
    XtSetArg (args[4], XmNtopAttachment,    XmATTACH_FORM);
    XtSetArg (args[5], XmNbottomAttachment, XmATTACH_FORM);
    XtSetArg (args[6], XmNorientation,      XmVERTICAL);
  }

  XtSetArg (args[7], XmNmargin, 0); n++;
  Result = XtCreateManagedWidget ("separator", xmSeparatorWidgetClass, 
				               Parent, args, 8);

  PRO_END
  return Result;
}


static Widget DLG_CreateFrame (Widget Parent, DLGELEM *Elem)
{
  Widget Result;

  XtSetArg (args[n], XmNwidth, 
	             abs(Elem->Value.mrec.x2 - Elem->Value.mrec.x1)); n++;
  XtSetArg (args[n], XmNheight, 
	             abs(Elem->Value.mrec.y2 - Elem->Value.mrec.y1)); n++;
  XtSetArg (args[n], XmNseparatorType, 
	             (Elem->Value.mrec.Style==DlgETCHED_OUT[0])? 
	             XmSHADOW_ETCHED_OUT : XmSHADOW_ETCHED_IN); n++;
  Result = XtCreateManagedWidget ("frame", xmFrameWidgetClass, 
				           Parent, args, n);

  Elem->WidgetID = Result;

  return Result;
}


Widget DLG_CreateWidget (Widget Parent, DLGELEM *Elem, dual accelerate)
{
  Widget Result = NULL;

  PRO_START ("\n")

  Elem->Master = Parent;

  n = 0;

#if 1
  XtSetArg (args[0], MNxAlignment, M_ALIGN_FILL);
  XtSetArg (args[1], MNyAlignment, M_ALIGN_CENTER);
  n = 2;
#endif


  switch (Elem->Type)
  {
    case DlgLABEL : 
      PRO0 ("<LABEL>\n")
      Result = DLG_CreateLabel (Parent, Elem);
      break;
      
    case DlgBOOL : 
      PRO0 ("<BOOL>\n")
      Result = DLG_CreateBool (Parent, Elem, accelerate);
      break;

    case DlgINT :
      PRO0 ("<INT>\n")
      Result = DLG_CreateInt (Parent, Elem, accelerate);
      break;

    case DlgDOUBLE : 
      PRO0 ("<DOUBLE>\n")
      Result = DLG_CreateDouble (Parent, Elem, accelerate);
      break;

    case DlgTEXT : 
      PRO0 ("<TEXT>\n")
      Result = DLG_CreateText (Parent, Elem, accelerate);
      break;

    case DlgSCALE : 
      PRO0 ("<SCALE>\n")
      Result = DLG_CreateScale (Parent, Elem, accelerate);
      break;

    case DlgBAR : 
      PRO0 ("<BAR>\n")
      Result = DLG_CreateBar (Parent, Elem, accelerate);
      break;                      

    case DlgUPDN : 
      PRO0 ("<UPDN>\n")
      Result = DLG_CreateUpDn (Parent, Elem, accelerate);
      break;

    case DlgLIST : 
      PRO0 ("<LIST>\n")
      Result = DLG_CreateList (Parent, Elem, accelerate);
      break;

    case DlgOMENU : 
      PRO0 ("<OMENU>\n")
      Result = DLG_CreateOMenu (Parent, Elem, accelerate);
      break;

    case DlgTBOX : 
      PRO0 ("<TOGGLE>\n")
      Result = DLG_CreateTBox (Parent, Elem, accelerate);
      break;

    case DlgRBOX :  
      PRO0 ("<RADIO>\n")
      Result = DLG_CreateRBox (Parent, Elem, accelerate);
      break;

    case DlgPROC : 
      PRO0 ("<PROC>\n")
      Result = DLG_CreateCall (Parent, Elem, accelerate);
      break;

    case DlgARROW :
      PRO0 ("<ARROW>\n")
      Result = DLG_CreateArrow (Parent, Elem, accelerate);
      break;

    case DlgPIC :
      {
	dual accel;

	accel = (Elem->Value.prec.FPtr == NULL)? FALSE : accelerate;
	PRO0 ("<PIC>\n")
	if (Elem->Value.prec.Format == DlgXBM)
	  Result = DLG_CreateBMap (Parent, Elem, accel);
	else
	  Result = DLG_CreatePic (Parent, Elem, accel);
      }
      break;

    case DlgIMAGE :
      {
	dual accel;

	accel = (Elem->Value.prec.FPtr == NULL)? FALSE : accelerate;
	PRO0 ("<IMAGE>\n")
	Result = DLG_CreateImage (Parent, Elem, accel);
      }
      break;

#if (XmVERSION>=2)
    case DlgNOTE :
      PRO0 ("<NOTEBOOK>\n")
      Result = DLG_CreateNote (Parent, Elem, accelerate);
      break;

    case DlgTREE :
      PRO0 ("<TREE>\n")
      Result = DLG_CreateTree (Parent, Elem, accelerate);
      break;
#endif

    case DlgFSEL :
      PRO0 ("<FILESELECTOR>\n")
      Result = DLG_CreateFSel (Parent, Elem, accelerate);
      break;

    case DlgDIRSEL :
      PRO0 ("<DIRSELECTOR>\n")
      Result = DLG_CreateFSel (Parent, Elem, accelerate);
      break;

    case DlgDIAL :
      PRO0 ("<DIAL>\n")
      Result = DLG_CreateDial (Parent, Elem, accelerate);
      break;

    case DlgSEGM :
      PRO0 ("<SEGMENT>\n")
      Result = DLG_CreateSegment (Parent, Elem);
      break;

    case DlgSEP :
      PRO0 ("<SEPERATOR>\n")
      Result = DLGCreateSep (Parent, Elem);
      break;

    case DlgBOX :
      PRO0 ("<FRAME>\n")
      Result = DLG_CreateFrame (Parent, Elem);
      break;

    case DlgUSER :
      PRO0 ("<WIDGET>\n")
      Result = DLG_CreateUser (Parent, Elem);
      break;

    default : 
      break;
  }
  if (Elem->UserArgs != NULL)
    XtSetValues (Result, Elem->UserArgs, Elem->NumArgs);

  XtAddCallback (Result, XmNhelpCallback, (XtCallbackProc)DLG_Help2CB, Elem);

  PRO_END
  return Result;
}


Widget DLGCreateWidget (Widget Parent, DLGELEM *Elem, ArgList args, int Num)
{
  Widget Form;

  PRO_START ("")

  if (Elem == NULL)
    PRO_RETURN ("Invalid element-pointer", NULL)

  Form = XtCreateManagedWidget ("form", xmFormWidgetClass, Parent, args, Num);

  DLG_CreateWidget (Form, Elem, FALSE);

  PRO_END
  return Form;
}


Widget DLGVaCreateWidget (Widget Parent, DLGELEM *Elem, ...)
{
  Widget  Form;
  va_list Arguments;
  
  PRO_START ("")

  va_start (Arguments, Elem);
  Form = XtVaCreateManagedWidget ("form", xmFormWidgetClass,
				          Parent, Arguments);
  va_end (Arguments);
  
  DLG_CreateWidget (Form, Elem, FALSE);

  PRO_END
  return Form;
}


