/* ******************************** */
/* *** (c) Stefan Vietinghoff   *** */
/* *** fileselector management  *** */
/* ******************************** */

#ifdef sunos
#include <sys/types.h>
#include <sys/dir.h>
#define dirent direct
#else
#include <dirent.h>
#endif

#include <sys/stat.h>

#include <Xm/PanedW.h>

#include "dlgUtils.h"
#include "dlgScript.h"

#include "folder.xpm"
#include "plus.xpm"
#include "minus.xpm"

/* TODO:
   - Directoryliste auch bei Tab aktualisieren, d.h wie bei RETURN 
      (XmDirSearchProc auslsen ?) 
     Funktioniert falls verzeichnis komplettiert wird (mit endendem /)
   - bei Tab auch die Fileliste aktualisieren ?
   - Leerzeichen nach Komma erlauben
   - HOME/END und Ctrl-Y Untersttzung
*/

typedef struct
{
  char  *FileName;
  char  *DirMask;
  Widget Shell;
} DLG_FSDATA;

typedef struct
{
  char          *Directory;
  unsigned char  MaskType;
  int            Number;
  char           Pattern[10][40];
  char           Mask[PATH_MAX+1];
  char           LastMask[PATH_MAX+1];
} DLG_FILESEARCH;

typedef struct
{
  char           Extension[20];
  Pixmap         Pic;
} DLG_FSELICON;

typedef struct DIR_ENTRY
{
         char       DirName[PATH_MAX+1];
         dual       build;
         Widget     Entry;
  struct DIR_ENTRY *Next;
} DIR_ENTRY;


typedef struct DLG_FSELDATA
{
  char          FileName[PATH_MAX+1];
  char          DirMask[PATH_MAX+1];
  Widget        Shell;
  Widget        ScrWin;
  Widget        List;
  Widget        Filter, Selection;
  DIR_ENTRY    *Entry;
  int           IconNo;
  DLG_FSELICON *Icons;
  Pixel         bg;
} DLG_FSELDATA;


static void DLG_FSEditCB      ();
static void DLG_FSOkCB        ();      /* FileSelektor OK-Button      */  
static void DLG_FSCancelCB    ();      /*          CANCEL-Button      */
static void DLG_FSSearchProc  ();

static dual           Tab     = FALSE;
static dual           changed = FALSE;
static DLG_FILESEARCH Search;

/* ********************** */
/* *** File-Selelctor *** */
/* ********************** */

/*----------------------------------------------------------------------
  Funktion: legt die Position fest an der der Fileselektor erscheinen
            soll.
  
  INPUT:  xPos, yPos   Koordinaten der linken, oberen Fileselektor-Ecke
  OUTPUT: ---
  
  Bsp.: DLGSetFSPos (100.0, 300.0); absolut: X=100  Y=300
        bzw.
	DLGSetFSPos (0.1, 0,3);     enspr.:  X=10%  Y=30%
  ----------------------------------------------------------------------*/
void DLGSetFSPos (double xPos, double yPos)
{
  int ScrX, ScrY;

  ScrX = DisplayWidth  (XDisplay, XScreen);
  ScrY = DisplayHeight (XDisplay, XScreen);

  if (xPos > 0.0 && xPos < 1.0)
    FSXPos = (int)(xPos * ScrX);
  elsif (xPos >= 0.0)
    FSXPos = MIN ((int)xPos, ScrX);
  else
    FSXPos = -1;

  if (yPos > 0.0 && yPos < 1.0)
    FSYPos = (int)(yPos * ScrY);
  elsif (yPos >= 0.0)
    FSYPos = MIN ((int)yPos, ScrY);
  else
    FSYPos = -1;
}


/* ******************************************************* */
/* ** Creates an file selection widget with some        ** */
/* ** convenience functions like completing filenames   ** */
/* ** using TAB like in bash or zshell.                 ** */
/* ** INPUT:  Parent   parent widget (e.g.:shell/form)  ** */
/* **         DirMask  default-mask of displayed files  ** */
/* **         FileName Default selection                ** */
/* **         Buttons  bit-mask of displayed buttons    ** */
/* **         Files    TRUE for file selection          ** */
/* **                  FALSE for directory selection    ** */
/* ** OUTPUT: (Widget) created widget                   ** */
/* ******************************************************* */
Widget DLG_CreateFileSel (Widget Parent, char *DirMask, char *FileName,
			                 int   Buttons, dual  Files)
{
  int             n = 0;
  char           *Text = NULL;
  char            Name[PATH_MAX+1] = "";
  char            Default[PATH_MAX+1] = "";
  char            WidgetName[16];
  XmString        Label;
  Arg             args[NUM_ARGS];
  Widget          Filter, Select;
  Widget          Button, Result;
  XtTranslations  Translation;

  if (! TOOLIsInitialized())
  {
    PRO0 ("Dlg.-Lib has NOT been initialized\n")
    return NULL;
  }

  if ((Text = TOOLGetResString(Parent, "xmFileSelectionBoxWidgetClass",
			       "fileSelector", "dirMask", DirMask)) != NULL)
  {
    TOOLParseFileName (Text, Name, PATH_MAX);
    Label = XmStringLtoRCreate (Name, NORMAL_FONT);
  }
  else
    Label = XmStringLtoRCreate ("/*", NORMAL_FONT); /* */

  XtSetArg (args[0], XmNdirMask,         Label);
  XtSetArg (args[1], XmNfileSearchProc,  DLG_FSSearchProc);
  XtSetArg (args[n], XmNuserData,       &Files);
  n = 3;
  if (Buttons < 7)
  {
    XtSetArg (args[n], XmNmarginHeight, 3); n++;
    XtSetArg (args[n], XmNmarginWidth,  3); n++;
  }

  strcpy (WidgetName, Files? "fileSelector":"dirSelector");
  Result = XtCreateManagedWidget (WidgetName, xmFileSelectionBoxWidgetClass,
				              Parent, args, n);
  XmStringFree (Label);

  Filter = XmFileSelectionBoxGetChild (Result, XmDIALOG_FILTER_TEXT);
  Select = XmFileSelectionBoxGetChild (Result, XmDIALOG_TEXT);
  XtAddCallback (Filter, XmNmodifyVerifyCallback, 
		         (XtCallbackProc)DLG_FSEditCB, Result);
  XtAddCallback (Select, XmNmodifyVerifyCallback, 
		         (XtCallbackProc)DLG_FSEditCB, Result);

  XtVaSetValues (Filter, XmNverifyBell, False, NULL);

  n = 1;
  if (FileName != NULL && FileName[0] != '\0')
  {
    TOOLParseFileName (FileName, Default, PATH_MAX);
    XtSetArg (args[n], XmNvalue, Default); n++;
  }
  XtSetValues (Select, args, n);

  if (! (Buttons & 1))
  {
    Button = XtNameToWidget (Result, "OK");
    XtUnmanageChild (Button);
  }

  if (! (Buttons & 2))
  {
    Button = XtNameToWidget (Result, "Apply");
    XtUnmanageChild (Button);
  }

  if (! (Buttons & 4))
  {
    Button = XtNameToWidget (Result, "Cancel");
    XtUnmanageChild (Button);
  }

  XtUnmanageChild (XmFileSelectionBoxGetChild(Result, XmDIALOG_HELP_BUTTON));
  if (! Files)
  {
    XtUnmanageChild (XtParent(XmFileSelectionBoxGetChild(Result, 
							 XmDIALOG_LIST)));
    XtUnmanageChild (XmFileSelectionBoxGetChild(Result, XmDIALOG_LIST_LABEL));
    XtUnmanageChild (XmFileSelectionBoxGetChild(Result, XmDIALOG_TEXT));
    XtUnmanageChild (XmFileSelectionBoxGetChild(Result, 
						XmDIALOG_SELECTION_LABEL));
  }

  /* forbid traversing to the next tab group: */
  Translation = XtParseTranslationTable ("#override <Key>Tab: self-insert()");
  XtOverrideTranslations (Select, Translation);
  XtOverrideTranslations (Filter, Translation);

  return Result;
}


/* ******************************************************* */
/* ** opens a modal file- or directory-selector.        ** */
/* ** The Result will be returned twice:                ** */
/* ** 1st  with parameter 'Result'                      ** */
/* ** 2nd  as return-value (points onto 'Result')       ** */
/* ** Memory for the result must be allocated by the    ** */
/* ** user.                                             ** */
/* ** INPUT:  Title    Name in title-bar of selector    ** */
/* **         Result   Default-entry and return-mem     ** */
/* **         Length   length of allocated memory       ** */
/* **         DirMask  restriction for shown entries    ** */
/* **         ...      (var. parameter)                 ** */
/* ** OUTPUT: (char *) Pointer onto 'Result'            ** */
/* ** EXAMPLE:                                          ** */
/* **   char Home[40] = "/users/stefan"                 ** */
/* **   char FileName[100];                             ** */
/* **   ...                                             ** */
/* **   printf ("Selected filename is: %s",             ** */
/* **           DLGOpenFileSel("select", FileName, 100, ** */
/* **                          "%s/data/foo.*", Home)); ** */
/* ******************************************************* */
char *DLGOpenFileSel (const char *Title, 
		      char *Result, int Length, char *Mask, ...)
{
  char       *Msg = NULL;
  char        FileDir[PATH_MAX+1];
  char        FileSpec[PATH_MAX+1];
  char        FileMask[PATH_MAX+1];
  Arg         args[NUM_ARGS];
  Widget      FileSel;
  DLG_FSDATA  FSData = {NULL, NULL, NULL};
  va_list     Arguments;

  PRO_START ("\n")

  if (! TOOLIsInitialized())
    PRO_ERROR0 ("Dlg.-Lib has NOT been initialized\n", NULL)

#if 0
  /* unlock cursor if it is locked from other windows */
  XUngrabPointer (XDisplay, CurrentTime);
  XFlush (XDisplay);    
#endif

  if (TOOLScriptRunning())
  {
    if ((Msg = DLG_RunFSScript(Title, Result, Length)) != NULL)
    {
      PRO1 ("Error while execution command-script: %s\n", Msg)
      PRO_ERROR0 ("Error in command-script", NULL)
    }
    PRO_RETURN ("Script-File", Result)
  }

  /* create filesel-shell */
  XtSetArg (args[0], XmNallowShellResize, TRUE);
  XtSetArg (args[1], XmNx,                FSXPos);
  XtSetArg (args[2], XmNy,                FSYPos);
  FSData.Shell = XtCreatePopupShell (Title, topLevelShellWidgetClass, 
		   	                    AppShell, args, 3);

  va_start (Arguments, Mask);
  vsprintf (FileMask, Mask, Arguments);
  va_end   (Arguments);

  TOOLGetFileDir (FileMask, FileDir, PATH_MAX);
  strncpy (FileSpec, Result, PATH_MAX);
  sprintf (Result, "%s/%s", FileDir, FileSpec);
  FileSel = DLG_CreateFileSel (FSData.Shell, FileMask, Result, 15, TRUE);

  XtAddCallback (FileSel, XmNokCallback,
		          (XtCallbackProc)DLG_FSOkCB,     &FSData);
  XtAddCallback (FileSel, XmNcancelCallback, 
		          (XtCallbackProc)DLG_FSCancelCB, &FSData);

  XtManageChild   (FSData.Shell);
  XtRealizeWidget (FSData.Shell);
  XtPopup         (FSData.Shell, XtGrabNone);  

  /*  XFlush (XDisplay); */

  PRO0 ("\\MainLoop...\n")
  TOOLMainLoop (AppCon, FSData.Shell);

  if (FSData.FileName == NULL && FSData.DirMask == NULL)
  {
    Result[0] = '\0';
    PRO_RETURN ("Aborted", NULL)
  }

  strncpy (Result, FSData.FileName, Length);
  Result[Length-1] = '\0';
  strncpy (Mask, FSData.DirMask, Length);
  Mask[Length-1] = '\0';

  if (FSData.FileName != NULL)
    free (FSData.FileName);
  if (FSData.DirMask != NULL)
    free (FSData.DirMask);

  PRO_END
  return Result;
}  


char *DLGOpenDirSel (const char *Title, 
		     char *Result, int Length, char *Mask, ...)
{
  char       *Msg = NULL;
  char        DirMask[PATH_MAX+1];
  Arg         args[NUM_ARGS];
  Widget      FileSel;
  DLG_FSDATA  FSData = {NULL, NULL, NULL};
  va_list     Arguments;

  PRO_START ("\n")

  if (! TOOLIsInitialized())
    PRO_ERROR0 ("Dlg.-Lib has NOT been initialized\n", NULL)

#if 0
  /* unlock cursor if it is locked by another windows */
  XUngrabPointer (XDisplay, CurrentTime);
  XFlush (XDisplay);    
#endif

  if (TOOLScriptRunning())
  {
    if ((Msg = DLG_RunFSScript(Title, Result, Length)) != NULL)
    {
      PRO1 ("Error while execution command-script: %s\n", Msg)
      PRO_ERROR0 ("Error in command-script", NULL)
    }
    else
    {
      TOOLGetFileDir (Result, Mask, Length);
      PRO_RETURN ("Script-Dir", Mask)
    }
  }

  /* create filesel-shell */
  XtSetArg (args[0], XmNallowShellResize, TRUE);
  XtSetArg (args[1], XmNx,                FSXPos);
  XtSetArg (args[2], XmNy,                FSYPos);
  FSData.Shell = XtCreatePopupShell (Title, topLevelShellWidgetClass, 
		  	                    AppShell, args, 3);

  va_start (Arguments, Mask);
  vsprintf (DirMask, Mask, Arguments);
  va_end   (Arguments);

  FileSel = DLG_CreateFileSel (FSData.Shell, DirMask, Result, 15, FALSE);

  XtAddCallback (FileSel, XmNokCallback,
		          (XtCallbackProc)DLG_FSOkCB,     &FSData);
  XtAddCallback (FileSel, XmNcancelCallback, 
		          (XtCallbackProc)DLG_FSCancelCB, &FSData);

  XtManageChild   (FSData.Shell);
  XtRealizeWidget (FSData.Shell);
  XtPopup         (FSData.Shell, XtGrabNone);  

  /*  XFlush (XDisplay); */

  PRO0 ("\\MainLoop...\n")
  TOOLMainLoop (AppCon, FSData.Shell);

  if (FSData.FileName == NULL && FSData.DirMask == NULL)
  {
    Mask[0] = '\0';
    PRO_RETURN ("Aborted", NULL)
  }

  TOOLGetFileDir (FSData.DirMask, Result, Length);
  strncpy (Mask, FSData.DirMask, Length);
  Mask[Length-1] = '\0';

  if (FSData.FileName != NULL)
    free (FSData.FileName);
  if (FSData.DirMask != NULL)
    free (FSData.DirMask);

  PRO_END
  return Result;
}


static int DLG_MatchFile (const char *FileName)
{
  int count;

  for (count = 0; count < Search.Number; ++count)
    if (TOOLTestPattern(Search.Pattern[count], FileName))
      return 1;

  return 0;
}


static int DLG_GetPatterns (char *Mask)
{
  int count = 0;
  int Number = 0;
  int Index = 0;

  PRO_START ("")

  while (Mask[Index] != '\0' && Number < 10 && count < 39)
  {
    if (Mask[Index] != ',')
      Search.Pattern[Number][count++] = Mask[Index++];
    else
    {
      Search.Pattern[Number][count] = '\0';
      Number++;
      count = 0;
      do
	Index++;
      while (Mask[Index] == ' ');
    }
  }

  if (count > 0)
  {
    Search.Pattern[Number][count] = '\0';
    Number++;
  }
  Search.Number = Number;

  PRO_END
  return Number;
}


static int DLG_SelectFiles (const struct dirent *Entry)
{
         char FileName[PATH_MAX+1];
  struct stat Buffer;

  if (strcmp(Entry->d_name, ".") != 0 && strcmp(Entry->d_name, "..") != 0)
  {
    PRO1 ("FILE '%s'\n", Entry->d_name)
    sprintf (FileName, "%s/%s", Search.Directory, Entry->d_name);
    if (lstat(FileName, &Buffer) == 0 && 
	(Search.MaskType == XmFILE_ANY_TYPE ||
	 (Buffer.st_mode & S_IFMT) != S_IFDIR))
      return DLG_MatchFile (Entry->d_name);
  }

  return 0;
}


static int DLG_SelectDirs (const struct dirent *Entry)
{
         char FileName[PATH_MAX+1];
  struct stat Buffer;

  PRO1 ("FILE '%s'\n", Entry->d_name)

  if (strcmp(Entry->d_name, ".") == 0 || strcmp(Entry->d_name, "..") == 0)
    return 1;

  sprintf (FileName, "%s/%s", Search.Directory, Entry->d_name);
  if (lstat(FileName, &Buffer) == 0 && 
      (Buffer.st_mode & S_IFMT) == S_IFDIR)
    return DLG_MatchFile (Entry->d_name);

  return 0;
}


static void DLG_FSSearchProc (Widget widget, 
			      XmFileSelectionBoxCallbackStruct *Data)
{
  dual           *Files;
  int             count, Number = 0;
  int             Patterns = 0;
  char            FileName[PATH_MAX+1];
  char           *Mask = NULL;
  char           *Pattern = NULL;
  XmString        DirSpec;
  XmStringTable   MatchList = NULL;
  Arg             args[NUM_ARGS];
  struct dirent **Entry = NULL;

  PRO_START ("\n")

  XmStringGetLtoR (Data->dir,     XmFONTLIST_DEFAULT_TAG, &Search.Directory);
  PRO1 ("Dir: %s\n", Search.Directory)
  XmStringGetLtoR (Data->pattern, XmFONTLIST_DEFAULT_TAG, &Pattern);
  PRO1 ("Pattern: %s\n", Pattern)
  XmStringGetLtoR (Data->mask,    XmFONTLIST_DEFAULT_TAG, &Mask);
  PRO1 ("Mask: %s\n", Mask)

  strncpy (Search.Mask, Pattern, PATH_MAX);
  Search.Mask[PATH_MAX] = '\0';

  XtVaGetValues (widget, XmNuserData, &Files, NULL);

#if 0
  /* Beim 2. ffnen d. FileSelektors wird die Liste nicht richtig initialis. */ 
  if (strcmp(Mask, Search.LastMask) == 0)   
  {
    XtSetArg (args[0], XmNlistUpdated, False);
    XtSetValues (widget, args, 1);
    PRO_ABORT ("MASK NOT CHANGED")
  }
#endif
  strncpy (Search.LastMask, Mask, PATH_MAX);
  Search.LastMask[PATH_MAX] = '\0';

  XtSetArg (args[0], XmNfileTypeMask, &Search.MaskType);
  XtGetValues (widget, args, 1);

  Patterns = DLG_GetPatterns (Pattern);

  if (*Files)
    Number = scandir (Search.Directory, &Entry, DLG_SelectFiles, alphasort);
  else
    Number = scandir (Search.Directory, &Entry, DLG_SelectDirs, alphasort);

  if (Number > 0)
  {
    MatchList = (XmStringTable)malloc (sizeof(XmString) * Number);
    for (count = 0; count < Number; ++count)
    {
      sprintf (FileName, "%s%s", Search.Directory, Entry[count]->d_name);
      MatchList[count] = XmStringCreate (FileName, NORMAL_FONT);
      free (Entry[count]);
    }
    free (Entry);
  }

  DirSpec = XmStringLtoRCreate (Search.Directory, NORMAL_FONT);
  if (*Files)
  {
    XtSetArg (args[0], XmNfileListItemCount, Number);
    XtSetArg (args[1], XmNfileListItems,     MatchList);
  }
  else
  {
    XtSetArg (args[0], XmNdirListItemCount, Number);
    XtSetArg (args[1], XmNdirListItems,     MatchList);
  }
  XtSetArg (args[2], XmNlistUpdated,       True);
  XtSetArg (args[3], XmNdirSpec,           DirSpec);
  XtSetValues (widget, args, 4);
  if (MatchList != NULL)
    free (MatchList);

  PRO_END
}


static void DLG_FSEditCB (Widget widget, Widget FileSel,
			                 XmTextVerifyCallbackStruct *Data)
{
  int   InsPos, RetCode;
  int   Pos;
  char *OldMask;
  char  Complete[PATH_MAX+1];
  char  NewMask[PATH_MAX+1];

  PRO_START ("\n")

  if (changed || Data == NULL)
    PRO_ABORT ("CHANGED")

  OldMask = XmTextGetString(widget);
  InsPos  = XmTextGetInsertionPosition(widget);
#if 1
  PRO1 ("DIRMASK: '%s'\n", OldMask)
  PRO1 ("Ins:     %d\n",   InsPos)
  PRO1 ("Length:  %d\n",   Data->text->length)
  if (Data->reason == XmCR_VALUE_CHANGED)
    PRO1 ("String:  '%s'\n", Data->text->ptr)
  PRO3 ("Start: %d  End: %d  Current: %d\n",
	Data->startPos, Data->endPos, Data->currInsert)
#endif

  if (Data->reason == XmCR_MODIFYING_TEXT_VALUE && Data->text->length == 1)
  {
    switch (Data->text->ptr[0])
    {
      case ' '  : 
      case '\t' :
	TOOLParseFileName (OldMask, NewMask, PATH_MAX);
	Pos = strlen(NewMask)-strlen(OldMask)+InsPos;
        strncpy (Complete, NewMask, Pos);
	Complete[Pos] = '\0';
	RetCode = DLG_CompleteFileName (Complete, PATH_MAX);
	InsPos = strlen(Complete);
	if (Complete[InsPos-1] == '/' && NewMask[Pos] == '/')
	  Pos++;
        strcat (Complete, NewMask+Pos);
        switch (RetCode)
        {
	  case  1 : /* exactly one match */
            PRO1 ("Completed exactly: %s\n", Complete)
	    changed = TRUE;
	    break;

          case  2 : /* mulitple matches */
	    PRO1 ("Completed partialy: %s\n", Complete)
	    changed = TRUE;
            /* TODO: Directory-List update... */
	    break;

	  default : /* no matching entry */
            PRO1 ("Error while completion (Bad FileSpec): %s\n", Complete)
            XBell (XDisplay, 100);
        }

	XtFree (Data->text->ptr);
	Data->text->ptr = NULL;
	Data->text->length = 0;
	Data->doit = FALSE;

	if (Complete[InsPos-1] == '/')
	{
	  TOOLGetFileDir (Complete, NewMask, PATH_MAX);
	  strcat (NewMask, "/");
	  strcat (NewMask, Search.Mask);
	  XmFileSelectionDoSearch (FileSel,
				   XmStringLtoRCreate(NewMask, NORMAL_FONT));
	}
/*	  else
  	    sprintf (NewMask, "%s*", Complete);
*/

	/* DIR-List aktualisieren... */
	XmTextSetString (widget, Complete);
	XmTextSetInsertionPosition (widget, InsPos);
        Tab = ! Tab;
	break;

      case    1 : /* Ctrl-A (HOME) */
	Data->doit = FALSE;
	XmTextSetInsertionPosition (widget, 0);
	Tab = FALSE;
	break;

      case    5 : /* Ctrl-E (END) */
	Data->doit = FALSE;
	XmTextSetInsertionPosition (widget, strlen(OldMask));
	Tab = FALSE;
	break;

      case   11 : /* Ctrl-K */
	Data->doit = FALSE;
	strncpy (NewMask, OldMask, InsPos);
	NewMask[InsPos] = '\0';
	changed = TRUE;
	XmTextSetString (widget, NewMask);
	XmTextSetInsertionPosition (widget, InsPos);
	Tab = FALSE;
	break;

      default : 
	Tab = FALSE;
    }
    changed = FALSE;
  }
  else
    PRO0 ("PASTE\n")

  PRO_END
}


/* Callback-Routinen fr den Fileselektor: */
/* *************************************** */
static void DLG_FSOkCB (Widget widget, DLG_FSDATA *UserData,
			               XmFileSelectionBoxCallbackStruct *Data)
{
  char *Result;

  printf ("OK-CB\n");

  if (Data == NULL)
    return;

  printf ("Close\n");
  
  /* Dateiname auslesen: */
  XmStringGetLtoR (Data->value, NORMAL_FONT, &Result);
  if (UserData->FileName != NULL)
    free (UserData->FileName);
  UserData->FileName = DLG_strcp(Result);
  XtFree (Result);

  XmStringGetLtoR (Data->mask, NORMAL_FONT, &Result);
  if (UserData->DirMask != NULL)
    free (UserData->DirMask);
  UserData->DirMask = DLG_strcp(Result);
  XtFree (Result);

  /* Close file-selector 
     Works fine on slow Pentium system, but wont finish TOOLMainLoop in 
     some cases on an fast Athlon system:
       XtUnmanageChild (UserData->Shell);
  */
  XtDestroyWidget (UserData->Shell);
}


static void DLG_FSCancelCB (Widget widget, DLG_FSDATA *UserData, 
			                   XtPointer Data)
{
  if (UserData->FileName != NULL)
    free (UserData->FileName);
  UserData->FileName = NULL;

  if (UserData->DirMask != NULL)
    free (UserData->DirMask);
  UserData->DirMask = NULL;

  /* Close file-selector (see comment in FSOkCB) */
  XtDestroyWidget (UserData->Shell);
}


/* *********************************************************************** */

void DLG_TreeCB2 ();

static int DLG_DirSelect (const struct dirent *Entry)
{
  return 1;
}

static int DLG_FileSelect (const struct dirent *Entry)
{
  return 1;
}

static dual DLG_HasSubDirs (const char *Directory)
{
         dual     Result = FALSE;
         int      Number, count;
         char     DirName[PATH_MAX+1];
	 char     Separator[2] = "/";
  struct stat     Buffer;
  struct dirent **Entry = NULL;
  
  if ((Number = scandir(Directory, &Entry, DLG_DirSelect, alphasort)) > 0)
  {
    if (Directory[strlen(Directory)-1] == '/')
      Separator[0] = '\0';

    for (count = 0; count < Number; ++count)
    {
      if (! Result && 
	  strcmp(Entry[count]->d_name, ".") != 0 &&
	  strcmp(Entry[count]->d_name, "..") != 0)
      {
	sprintf (DirName, "%s%s%s", 
		 Directory, Separator, Entry[count]->d_name);
	if (stat(DirName, &Buffer) == 0)
	  if ((Buffer.st_mode & S_IFMT) & S_IFDIR)
	    Result = TRUE;
      }
      free (Entry[count]);
    }
    free (Entry);
  }
  return Result;
}  


void DLG_MakeDirectoryTree (Widget Tree, Widget Parent, DLG_FSELDATA *FSelData,
			    char *Selected, char *Current, char *Mask)
{
         dual            Fit;
         int             count, index, Number;
         unsigned char   Expand;
	 char            Separator[2] = "/";
	 char            DirName[PATH_MAX+1];
	 char            NextDir[PATH_MAX+1] = "";
	 XmString        Name;
	 Widget          Gadget;
	 DIR_ENTRY      *Line;
  struct stat            Buffer;
  struct dirent        **Entry = NULL;

  if ((Number = scandir(Current, &Entry, DLG_DirSelect, alphasort)) > 0)
  {
    /* Extract next directory */
    count = strlen (Current);
    Fit = (strncmp(Current, Selected, count) == 0);
    if (Fit)
    {
      if (Selected[count] == '/')
	count++;

      index = 0; 
      while (Selected[count] != '\0' && Selected[count] != '/' && 
	     index < PATH_MAX)
	NextDir[index++] = Selected[count++];
      NextDir[index] = '\0';
    }

    if (Current[strlen(Current)-1] == '/')
      Separator[0] = '\0';

    /* Process each subdirectory */
    for (count = 0; count < Number; ++count)
    {
      if (strcmp(Entry[count]->d_name, ".") != 0 &&
	  strcmp(Entry[count]->d_name, "..") != 0)
      {
	sprintf (DirName, "%s%s%s", Current, Separator, Entry[count]->d_name);
	if (stat(DirName, &Buffer) == 0)
	  if ((Buffer.st_mode & S_IFMT) & S_IFDIR)
	  {
	    if (Fit)
	      Expand =
		(strcmp(NextDir, 
			Entry[count]->d_name)==0)? XmEXPANDED : XmCOLLAPSED;
	    else
	      Expand = XmCOLLAPSED;

	    Name = XmStringCreateLocalized (Entry[count]->d_name);	
	    Gadget = XtVaCreateManagedWidget (Entry[count]->d_name, 
					      xmIconGadgetClass, Tree,
					      XmNlabelString,     Name,
					      XmNentryParent,     Parent,
					      XmNsmallIconPixmap, 
					        FSelData->Icons[0].Pic,
					      XmNmarginHeight,    0,
					      XmNshadowThickness, 0,
					      XmNbackground,      FSelData->bg,
					      XmNdetailCount,     1,
					      XmNoutlineState,    Expand,
					      NULL);
	    XmStringFree (Name);

	    Line = (DIR_ENTRY *)malloc (sizeof(DIR_ENTRY));
	    strcpy (Line->DirName, DirName);
	    Line->Entry = Gadget;
	    Line->build = (Expand == XmEXPANDED);
	    Line->Next  = FSelData->Entry;
	    FSelData->Entry = Line;

	    if (Expand == XmEXPANDED)
	      DLG_MakeDirectoryTree (Tree, Gadget, FSelData, Selected, 
				     DirName, Mask);
	    elsif (DLG_HasSubDirs(DirName))
	      /* Create unmanaged (unvisible) widget in order to create an
		 collapsed icon for this directory */
	      Gadget = XtVaCreateWidget ("", 
					 xmPushButtonGadgetClass, Tree,
					 XmNentryParent,     Gadget,
					 XmNmarginHeight,    0,
					 XmNshadowThickness, 0,
					 XmNdetailCount,     1,
					 XmNbackground,      FSelData->bg,
					 NULL);
	  }
      }
      free (Entry[count]);
    }
    free (Entry);
  }
} 


void DLG_MakeFileList (DLG_FSELDATA *FSelData, 
		       const char *DirName, const char *DirMask)
{
         int        count, Index = 0;
         int        Number;
	 char       Separator[2] = "/";
	 char       FileName[PATH_MAX+1];
         XmString   FileString;
  struct stat       Buffer;
  struct dirent   **Entry = NULL;

  XmListDeleteAllItems (FSelData->List);
  if ((Number = scandir(DirName, &Entry, DLG_FileSelect, alphasort)) > 0)
  {
    if (DirName[strlen(DirName)-1] == '/')
      Separator[0] = '\0';
    
    for (count = 0; count < Number; ++count)
    {
      if (strcmp(Entry[count]->d_name, ".") != 0 &&
	  strcmp(Entry[count]->d_name, "..") != 0)
      {
	sprintf (FileName, "%s%s%s", DirName, Separator, Entry[count]->d_name);
	if (stat(FileName, &Buffer) == 0)
	  if (! ((Buffer.st_mode & S_IFMT) & S_IFDIR))
	  {
	    FileString = XmStringCreateLocalized (Entry[count]->d_name);
	    XmListAddItem (FSelData->List, FileString, Index++);
	    XmStringFree (FileString);
	  }
      }
      free (Entry[count]);
    }
    free (Entry);
  }
}

Widget DLG_CreateFileSel2 (Widget Parent,
			   DLG_FSELDATA *FSelData, dual Files)
{
  int      n;
  Pixel    fg;
  XmString Label;
  Widget   Form, Form1, Form2;
  Widget   PanedWin;
  Widget   Mask, Result;
  Widget   Tree;
  /*   Widget Mask, Selection; */
  Pixmap ClosePic, OpenPic;
  Arg    args[NUM_ARGS];

printf ("Patterns: %d\n", DLG_GetPatterns(FSelData->DirMask));

  Form = XtVaCreateManagedWidget ("fsForm", xmFormWidgetClass, Parent,
				  XmNwidth,  400,
				  XmNheight, 350,
				  NULL);

  Form1 = XtVaCreateManagedWidget ("header", xmFormWidgetClass, Form,
				   XmNtopAttachment,   XmATTACH_FORM,
				   XmNleftAttachment,  XmATTACH_FORM,
				   XmNrightAttachment, XmATTACH_FORM,
				   NULL);

  Label = XmStringCreateLocalized ("mask");
  Mask = XtVaCreateManagedWidget ("mask", xmLabelWidgetClass, Form1,
				  XmNtopAttachment,    XmATTACH_FORM,
				  XmNleftAttachment,   XmATTACH_FORM,
				  XmNbottomAttachment, XmATTACH_FORM,
				  XmNlabelString,      Label,
				  NULL);
  XmStringFree (Label);
				   
  FSelData->Filter = 
    XtVaCreateManagedWidget ("filter", xmTextWidgetClass, Form1, 
			     XmNtopAttachment,   XmATTACH_FORM,
			     XmNleftAttachment,  XmATTACH_WIDGET,
			     XmNleftWidget,      Mask,
			     XmNrightAttachment, XmATTACH_FORM,
			     XmNvalue,           FSelData->DirMask,
			     NULL);
			     

  Form2 = XtVaCreateManagedWidget ("footer", xmFormWidgetClass, Form,
				   XmNleftAttachment,   XmATTACH_FORM,
				   XmNrightAttachment,  XmATTACH_FORM,
				   XmNbottomAttachment, XmATTACH_FORM,
				   NULL);

  Label = XmStringCreateLocalized ("result");
  Result = XtVaCreateManagedWidget ("result", xmLabelWidgetClass, Form2,
				  XmNtopAttachment,    XmATTACH_FORM,
				  XmNleftAttachment,   XmATTACH_FORM,
				  XmNbottomAttachment, XmATTACH_FORM,
				  XmNlabelString,      Label,
				  NULL);
  XmStringFree (Label);
				   
  FSelData->Selection = 
    XtVaCreateManagedWidget ("selection", xmTextWidgetClass, Form2,
			     XmNtopAttachment,   XmATTACH_FORM,
			     XmNleftAttachment,  XmATTACH_WIDGET,
			     XmNleftWidget,      Result,
			     XmNrightAttachment, XmATTACH_FORM,
			     XmNvalue,           FSelData->FileName,
			     NULL);

  PanedWin = XtVaCreateManagedWidget ("panedWin", xmPanedWindowWidgetClass, 
				      Form,
				      XmNorientation,      XmHORIZONTAL, 
				      XmNtopAttachment,    XmATTACH_WIDGET,
				      XmNtopWidget,        Form1,
				      XmNleftAttachment,   XmATTACH_FORM,
				      XmNrightAttachment,  XmATTACH_FORM,
				      XmNbottomAttachment, XmATTACH_WIDGET,
				      XmNbottomWidget,     Form2,
				      NULL);

  FSelData->ScrWin = 
    XtVaCreateManagedWidget ("dirScrWin", 
			     xmScrolledWindowWidgetClass, PanedWin,
			     XmNresizeWidth,              True,
			     XmNresizeHeight,             True,
			     XmNpaneMinimum,              200,
			     XmNscrollingPolicy,          XmAUTOMATIC,
			     XmNscrollBarDisplayPolicy,   XmAS_NEEDED,
			     XmNscrollHorizontal,         False,
			     XmNscrollVertical,           True,
			     NULL);

  n = 0;
  fg = TOOLGetColor ("LightBlue");
  XtSetArg (args[n], XmNlayoutType,         XmDETAIL);      n++;
  XtSetArg (args[n], XmNentryViewType,      XmSMALL_ICON);  n++;
  XtSetArg (args[n], XmNoutlineIndentation, 12);            n++;
  XtSetArg (args[n], XmNoutlineLineStyle,   XmSINGLE);      n++;
  XtSetArg (args[n], XmNsmallCellHeight,    8);             n++;
  XtSetArg (args[n], XmNbackground,         FSelData->bg);  n++;
  XtSetArg (args[n], XmNselectColor,        fg);            n++;
  ClosePic = TOOLMakeXPM (plus_xpm, NULL, NULL, FSelData->bg);
  if (ClosePic != XmUNSPECIFIED_PIXMAP)
  {
    XtSetArg (args[n], XmNcollapsedStatePixmap, ClosePic); n++;
  }

  OpenPic = TOOLMakeXPM (minus_xpm, NULL, NULL, FSelData->bg);
  if (OpenPic != XmUNSPECIFIED_PIXMAP)
  {
    XtSetArg (args[n], XmNexpandedStatePixmap, OpenPic);  n++;
  }
  Tree = XtCreateManagedWidget ("dirTree", xmContainerWidgetClass, 
				           FSelData->ScrWin, args, n);

  XtAddCallback (Tree, XmNdefaultActionCallback, 
		       (XtCallbackProc)DLG_TreeCB2, FSelData);
  XtAddCallback (Tree, XmNoutlineChangedCallback, 
		       (XtCallbackProc)DLG_TreeCB2, FSelData);

  DLG_MakeDirectoryTree (Tree, NULL, FSelData, "/users/stefan/TUXCAD", "/",
			 FSelData->DirMask);

  XmContainerRelayout (Tree);

  n = 0;
  XtSetArg (args[n], XmNlistSizePolicy,         XmVARIABLE);      n++;
  XtSetArg (args[n], XmNscrollBarDisplayPolicy, XmAS_NEEDED);     n++;
  XtSetArg (args[n], XmNitemCount,              0);               n++;
  XtSetArg (args[n], XmNselectionPolicy,        XmSINGLE_SELECT); n++;
  FSelData->List = XmCreateScrolledList (PanedWin, "fileList", args, n);

  DLG_MakeFileList (FSelData, "/users/stefan/TUXCAD", FSelData->DirMask);

  XtManageChild (FSelData->List);

  return Form;
}


char *DLGOpenFileSel2 (const char *Title, 
                       char *Result, int Length, char *Mask, ...)
{
  char          FileMask[PATH_MAX+1];
  Arg           args[NUM_ARGS];
  Widget        FileSel;
  DIR_ENTRY    *Ptr, *Next;
  DLG_FSELDATA  FSelData = {"", "", NULL, NULL, NULL, NULL, NULL, NULL,
			    0, NULL, 0};
  va_list       Arguments;

  PRO_START ("\n")

  if (! TOOLIsInitialized())
    PRO_ERROR0 ("Dlg.-Lib has NOT been initialized\n", NULL)

  va_start (Arguments, Mask);
  vsprintf (FileMask, Mask, Arguments);
  va_end   (Arguments);

  strcpy (FSelData.DirMask, FileMask);
  strcpy (FSelData.FileName, Result);

  /* create filesel-shell */
  XtSetArg (args[0], XmNallowShellResize, TRUE);
  XtSetArg (args[1], XmNx,                FSXPos);
  XtSetArg (args[2], XmNy,                FSYPos);
  FSelData.Shell = XtCreatePopupShell (Title, topLevelShellWidgetClass, 
				              AppShell, args, 3);

  XtVaGetValues (FSelData.Shell, XmNbackground, &FSelData.bg, NULL);

  /* Add some demo-MIME-Types */
  FSelData.IconNo = 2;
  FSelData.Icons = 
    (DLG_FSELICON *)malloc (sizeof(DLG_FSELICON) * FSelData.IconNo);
  FSelData.Icons[0].Pic = TOOLMakeXPM (folder_xpm, NULL, NULL, FSelData.bg);
  strcpy (FSelData.Icons[1].Extension, "world");
  FSelData.Icons[1].Pic = TOOLMakeXPM (plus_xpm, NULL, NULL, FSelData.bg);

  FileSel = DLG_CreateFileSel2 (FSelData.Shell, &FSelData, TRUE);

  XtManageChild   (FSelData.Shell);
  XtRealizeWidget (FSelData.Shell);
  XtPopup         (FSelData.Shell, XtGrabNone);  

  PRO0 ("\\MainLoop...\n")
  TOOLMainLoop (AppCon, FSelData.Shell);

  if (FSelData.FileName == NULL && FSelData.DirMask == NULL)
  {
    Result[0] = '\0';
    PRO_RETURN ("Aborted", NULL)
  }

  /* Evaluate result */
  strncpy (Result, FSelData.FileName, Length);
  Result[Length-1] = '\0';
  strncpy (Mask, FSelData.DirMask, Length);
  Mask[Length-1] = '\0';

  /* Clean up memory */
  Ptr = FSelData.Entry;
  while (Ptr != NULL)
  {
    Next = Ptr->Next;
    free (Ptr);
    Ptr = Next;
  }

  if (FSelData.FileName != NULL)
    free (FSelData.FileName);
  if (FSelData.DirMask != NULL)
    free (FSelData.DirMask);

  PRO_END
  return Result;
}  


void DLG_TreeCB2 (Widget widget, DLG_FSELDATA *FSelData,
		  XmAnyCallbackStruct *Data)
{
  DIR_ENTRY *Ptr = NULL;

  if (Data->reason == XmCR_DEFAULT_ACTION)
  {
    XmContainerSelectCallbackStruct *SelectData =
      (XmContainerSelectCallbackStruct *)Data;
    unsigned char Expand;

    XtVaGetValues (SelectData->selected_items[0], XmNoutlineState, &Expand,
		                                  NULL);
    Expand = (Expand == XmEXPANDED)? XmCOLLAPSED : XmEXPANDED;
    XtVaSetValues (SelectData->selected_items[0], XmNoutlineState, Expand, 
		                                  NULL);
  }
  else
  {
    int       Dummy;
    int       Size, Inc, Page;
    Dimension PosX, PosY;
    Widget    ScrBar;
    XmContainerOutlineCallbackStruct *OutlineData = 
      (XmContainerOutlineCallbackStruct *)Data;

    Ptr = FSelData->Entry;
    while (Ptr != NULL && Ptr->Entry != OutlineData->item)
    Ptr = Ptr->Next;

    /* On expansion add data, if we haven't build it before */
    if (OutlineData->reason == XmCR_EXPANDED)
    {
      if (! Ptr->build)
      {
	Ptr->build = TRUE;
	TOOLSetMouseCursor (FSelData->Shell, ToolWATCH, NULL, NULL);
	DLG_MakeDirectoryTree (widget, OutlineData->item, FSelData, 
			       Ptr->DirName, Ptr->DirName, FSelData->DirMask);
      }

      /* Set selected directory on top of scrolled window */
      XtVaGetValues (FSelData->ScrWin, XmNverticalScrollBar, &ScrBar, NULL);
      if (ScrBar != NULL)
      {
	XtVaGetValues (OutlineData->item, XmNx, &PosX, XmNy, &PosY, NULL);
	XmScrollBarGetValues (ScrBar, &Dummy, &Size, &Inc, &Page);
	XmScrollBarSetValues (ScrBar, PosY, Size, Inc, Page, True);    
      }
      TOOLSetMouseCursor (FSelData->Shell, ToolNORMAL, NULL, NULL);
    }
  }

  if (Ptr != NULL)
    DLG_MakeFileList (FSelData, Ptr->DirName, FSelData->DirMask);
}



