#include <X11/IntrinsicP.h>

#include "dlgGlobal.h"


#define H_POS(x) ((x & (DlgNORTH_WEST | DlgWEST | DlgSOUTH_WEST))? (-1) : \
		  (x & (DlgNORTH_EAST | DlgEAST | DlgSOUTH_EAST))? 1 : 0)
#define V_POS(y) ((y & (DlgNORTH_WEST | DlgNORTH | DlgNORTH_EAST))? (-1) : \
		  (y & (DlgSOUTH_WEST | DlgSOUTH | DlgSOUTH_EAST))? 1 : 0)


static int n = 0;
static Arg args[NUM_ARGS];


/* ******************************************************* */
/* ** Positions the lable- and value-widget in a common ** */
/* ** form. The position is defined in each element     ** */
/* ** INPUT:  Parent  patent form-widget                ** */
/* **         Value   created value-widget              ** */
/* **         Elem    structure containing label and    ** */
/* **                 label-position                    ** */
/* ** OUTPUT: ---                                       ** */
/* ******************************************************* */
void DLG_BuildLabel (Widget Parent, Widget Value, DLGELEM *Elem)
{
  int       LabelHPos, LabelVPos;
  int       ValueHPos, ValueVPos;
  int       MinX, MinY;
  char     *Text = NULL;
  XmString  LabelStr;
  Widget    Fill;

  if (Elem == NULL)
    return;

  PRO_START ("")
  PRO2 ("\\Box [%d] Title: <%s>\n", Elem->BoxID, Elem->Name)

  if (Elem->LabelPos == Elem->ValuePos)
    PRO_ABORT ("Label has same position as value")

  LabelHPos = H_POS (Elem->LabelPos);
  LabelVPos = V_POS (Elem->LabelPos);
  ValueHPos = H_POS (Elem->ValuePos);
  ValueVPos = V_POS (Elem->ValuePos);
  MinX = MIN (LabelHPos, ValueHPos);
  MinY = MIN (LabelVPos, ValueVPos);

  /* create label and position it */
  if (Elem->LabelPos != DlgNOWHERE)
  {
    n = 0;
    Text = TOOLGetResString (Parent, "xmLabelWidgetClass", "label",
                                     "labelString", Elem->Name);
    LabelStr = XmStringLtoRCreate (Text, NORMAL_FONT);
    XtSetArg (args[n], XmNlabelString, LabelStr); n++;

    if (Elem->LabelPos & (DlgCENTER | 
         DlgWEST | DlgNORTH_WEST | DlgNORTH | DlgSOUTH | DlgSOUTH_WEST))
    {
      XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    }
    if (Elem->LabelPos & (DlgCENTER |
         DlgWEST | DlgNORTH_WEST | DlgNORTH | DlgNORTH_EAST | DlgEAST))
    {
      XtSetArg (args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    }
    if (Elem->LabelPos & (DlgCENTER | 
         DlgNORTH | DlgNORTH_EAST | DlgEAST | DlgSOUTH_EAST | DlgSOUTH))
    {
      XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    }
    if (Elem->LabelPos & (DlgCENTER |
         DlgWEST | DlgEAST | DlgSOUTH_EAST | DlgSOUTH | DlgSOUTH_WEST))
    {
      XtSetArg (args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    }
    Elem->Label = XtCreateManagedWidget ("label", xmLabelWidgetClass, 
					          Parent, args, n);
    XmStringFree (LabelStr);
  }

  /* Place value-widget */
  if (Elem->ValuePos != DlgCENTER)
  {
    n = 0;
    if (Elem->ValuePos & (DlgWEST | DlgNORTH_WEST | DlgSOUTH_WEST))
    {
      XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    }
    if (Elem->ValuePos & (DlgNORTH_WEST | DlgNORTH | DlgNORTH_EAST))
    {
      XtSetArg (args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    }
    if (Elem->ValuePos & (DlgNORTH_EAST | DlgEAST | DlgSOUTH_EAST))
    {
      XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    }
    if (Elem->ValuePos & (DlgSOUTH_EAST | DlgSOUTH | DlgSOUTH_WEST))
    {
      XtSetArg (args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    }
    XtSetValues (Value, args, n);
  }

  /* fill-widget necessary ? */
  n = 0;
  if (abs(ValueHPos-LabelHPos) == 2 || abs(ValueVPos-LabelVPos) == 2)
  {
    LabelStr = XmStringLtoRCreate (" ", NORMAL_FONT);
    XtSetArg (args[n], XmNlabelString, LabelStr); n++;

    if (abs(ValueHPos - LabelHPos) > 1)
    {
      XtSetArg (args[n], XmNleftAttachment,  XmATTACH_WIDGET); n++;
      XtSetArg (args[n], XmNrightAttachment, XmATTACH_WIDGET); n++;
      if (LabelHPos < ValueHPos)
      {
	XtSetArg (args[n], XmNleftWidget,  Elem->Label); n++;
	XtSetArg (args[n], XmNrightWidget, Value);       n++;
      }
      else
      {
	XtSetArg (args[n], XmNleftWidget,  Value);       n++;
	XtSetArg (args[n], XmNrightWidget, Elem->Label); n++;
      }
    }
    if (abs(ValueVPos - LabelVPos) > 1)
    {
      XtSetArg (args[n], XmNtopAttachment,    XmATTACH_WIDGET); n++;
      XtSetArg (args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
      if (LabelVPos < ValueVPos)
      {
	XtSetArg (args[n], XmNtopWidget,    Elem->Label); n++;
	XtSetArg (args[n], XmNbottomWidget, Value);       n++;
      }
      else
      {
	XtSetArg (args[n], XmNtopWidget,    Value);       n++;
	XtSetArg (args[n], XmNbottomWidget, Elem->Label); n++;
      }
    }
    Fill = XtCreateManagedWidget ("fill", xmLabelWidgetClass, Parent, args, n);
    XmStringFree (LabelStr);
    XtSetMappedWhenManaged (Fill, FALSE);
  }
  else
  {
    if ((MinX >= 0 && LabelHPos != ValueHPos) ||
	(MinY >= 0 && LabelVPos != ValueVPos))
    {
      LabelStr = XmStringLtoRCreate (" ", NORMAL_FONT);
      XtSetArg (args[n], XmNlabelString, LabelStr); n++;
      if (MinX >= 0)
      {
	XtSetArg (args[n], XmNleftAttachment,  XmATTACH_FORM); n++;
	XtSetArg (args[n], XmNrightAttachment, XmATTACH_WIDGET); n++;
	XtSetArg (args[n], XmNrightWidget, 
		  (ValueHPos <= LabelHPos || 
		   Elem->LabelPos == DlgNOWHERE)? Value : Elem->Label); n++;
      }
      if (MinY >= 0)
      {
	XtSetArg (args[n], XmNtopAttachment,    XmATTACH_FORM); n++;
	XtSetArg (args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
	XtSetArg (args[n], XmNbottomWidget, 
		  (ValueHPos <= LabelHPos ||
		   Elem->LabelPos == DlgNOWHERE)? Value : Elem->Label); n++;
      }
      Fill = XtCreateManagedWidget ("fill", xmLabelWidgetClass, 
				            Parent, args, n);
      XtSetMappedWhenManaged (Fill, FALSE);
      XmStringFree (LabelStr);
    }

    if (Elem->LabelPos == DlgNOWHERE)
      PRO_ABORT ("done")

    if ((ValueHPos == 0 && LabelHPos != ValueHPos) || 
	(ValueVPos == 0 && LabelVPos != ValueVPos))
    {
      n = 0;
      if (ValueHPos == 0)
      {
	if (LabelHPos < ValueHPos)
	{
	  XtSetArg (args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
	  XtSetArg (args[n], XmNleftWidget,     Elem->Label);     n++;
	}
	elsif (LabelHPos > ValueHPos)
	{
	  XtSetArg (args[n], XmNrightAttachment, XmATTACH_WIDGET); n++;
	  XtSetArg (args[n], XmNrightWidget,     Elem->Label);     n++;
	}
      }
      if (ValueVPos == 0)
      {
	if (LabelVPos < ValueVPos)
	{
	  XtSetArg (args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
	  XtSetArg (args[n], XmNtopWidget,     Elem->Label);     n++;
	}
	elsif (LabelVPos > ValueVPos)
	{
	  XtSetArg (args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
	  XtSetArg (args[n], XmNbottomWidget,     Elem->Label);     n++;
	}
      }
      if (n > 0)
	XtSetValues (Value, args, n);
    }
    elsif (LabelHPos == 0 || LabelVPos == 0)
    {
      n = 0;
      if (LabelHPos == 0)
      {
	if (ValueHPos < LabelHPos)
	{
	  XtSetArg (args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
	  XtSetArg (args[n], XmNleftWidget,     Value);           n++;
	}
	elsif (ValueHPos > LabelHPos)
	{
	  XtSetArg (args[n], XmNrightAttachment, XmATTACH_WIDGET); n++;
	  XtSetArg (args[n], XmNrightWidget,     Value);           n++;
	}
      }
      if (LabelVPos == 0)
      {
	if (ValueVPos < LabelVPos)
	{
	  XtSetArg (args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
	  XtSetArg (args[n], XmNtopWidget,     Value);           n++;
	}
	elsif (ValueVPos > LabelVPos)
	{
	  XtSetArg (args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
	  XtSetArg (args[n], XmNbottomWidget,     Value);           n++;
	}
      }
      if (n > 0)
	XtSetValues (Elem->Label, args, n);
    }
  }

  PRO_END
}


static void DLG_AdjustBox (DLGBOX *Box)
{
  Dimension        Value;
#if 1
#if 0
  Position         PosX, PosY;
  Dimension        SizeX, SizeY;
#endif
  XtWidgetGeometry Geo1, Geo2;

  if (Box->Parent == NULL || Box->Box == NULL)
    return;

  PRO1 ("ADJUST BOX %d:  ", Box->BoxID)

  if (Box->AlignX == DlgRIGHT || Box->AlignX == DlgFILL || 
      Box->AlignX == DlgMIDDLE ||
      Box->AlignY == DlgDOWN || Box->AlignY == DlgFILL ||
      Box->AlignY == DlgMIDDLE)
  {
    Geo1.request_mode = CWWidth | CWHeight;
    if (Box->Dir == BOXELEM)
      XtQueryGeometry (XtParent(Box->Parent), NULL, &Geo1);
    elsif (Box->Dir == BOXGRID)
      XtQueryGeometry (Box->Parent, NULL, &Geo1);
    else
      XtQueryGeometry (XtParent(Box->Parent), NULL, &Geo1);
    if (! (Geo1.request_mode & (CWWidth | CWHeight)))
    {
      PRO0 ("\\GEO1 failed\n");
      return;
    }
    PRO2 ("\\[%dx%d]: ", Geo1.width, Geo1.height)

    Geo2.request_mode = CWX | CWY | CWWidth | CWHeight;
    XtQueryGeometry (Box->Main, NULL, &Geo2);
    if (! (Geo2.request_mode & (CWX | CWY | CWWidth | CWHeight)))
    {
      PRO0 ("\\GEO2 failed\n")
      return;
    }
    PRO4 ("\\(%d/%d) %dx%d", Geo2.x, Geo2.y, Geo2.width, Geo2.height)

#if 0
    PosX  = Geo2.x;
    PosY  = Geo2.y;
    SizeX = Geo2.width;
    SizeY = Geo2.height;

#if 0
    if (Box->AlignX == DlgRIGHT)
      PosX = FormW-ValW;
    elsif (Box->AlignX == DlgMIDDLE)
      PosX = (FormW-ValW)/2;
    elsif (Box->AlignX == DlgFILL)
      SizeX = FormW;

    if (Box->AlignY == DlgDOWN)
      PosY = FormH-ValH;
    elsif (Box->AlignY == DlgMIDDLE)
      PosY = (FormH-ValH)/2;
    elsif (Box->AlignY == DlgFILL)
      SizeY = FormH;
#else
    if (Box->AlignX == DlgRIGHT)
      PosX += Geo1.width-(Geo2.x+Geo2.width)-3;
    elsif (Box->AlignX == DlgMIDDLE)
      PosX += (Geo1.width-Geo2.width)/2;

    if (Box->AlignY == DlgDOWN)
      PosY += Geo1.height-(Geo2.y+Geo2.height)-3;
    elsif (Box->AlignY == DlgMIDDLE)
      PosY += (Geo1.height-Geo2.height)/2;
#endif
    PRO4 ("SET TO %d/%d: %dx%d\n", PosX, PosY, SizeX, SizeY)
    XtConfigureWidget (Box->Main, PosX, PosY, SizeX, SizeY, 0);

#else

    n = 0;
    if (Box->AlignX == DlgMIDDLE)
    {  
      Value = (Geo1.width-Geo2.width)/2; 
      PRO1 ("\\  X-MIDDLE: %d", Value)
      XtSetArg (args[n], XmNleftOffset, Value); n++;
    }
    elsif (Box->AlignX == DlgRIGHT)
    {
      Value = MAX (0, Geo1.width-(Geo2.x+Geo2.width)-3);
      PRO1 ("\\  X-RIGHT: %d", Value)
      XtSetArg (args[n], XmNleftOffset, Value); n++;
    }
#if 0
    elsif (Box->AlignX == DlgFILL)
    {
      Value = Geo1.width - Geo2.x-3; 
      PRO1 ("\\  X-FILL: %d", Value)
      XtSetArg (args[n], XmNwidth, Value); n++;
    }
#endif

    if (Box->AlignY == DlgMIDDLE)
    {
      Value = (Geo1.height-Geo2.height) / 2;
      PRO1 ("\\  Y-MIDDLE: %d", Value)
      XtSetArg (args[n], XmNtopOffset, Value); n++;
    }
    elsif (Box->AlignY == DlgDOWN)
    {
      Value = MAX (0, Geo1.height-(Geo2.y+Geo2.height)-3);
      PRO1 ("\\  Y-DOWN: %d", Value)
      XtSetArg (args[n], XmNtopOffset, Value); n++;
    }
    PRO0 ("\\\n")

    if (n > 0)
      XtSetValues (Box->Main, args, n);
#endif
  }
  else
    PRO0 ("\\clean\n")
#endif
}


static void DLG_AdjustElem (DLGELEM *Elem)
{
  dual         SetX, SetY;
  int          WinX, WinY;
  int          PosX, PosY;
  unsigned int FormW, FormH;
  unsigned int ValW, ValH;
  unsigned int border, bits;
  Window       Dummy;

  SetX = (H_POS(Elem->LabelPos) == 0 &&
	  (Elem->ValuePos & (DlgNORTH | DlgCENTER | DlgSOUTH)));
  SetY = (V_POS(Elem->LabelPos) == 0 &&
	  (Elem->ValuePos & (DlgWEST | DlgCENTER | DlgEAST)));

  if (SetX || SetY)
  {
    XGetGeometry (XDisplay, XtWindow(Elem->Master), &Dummy,
		  &WinX, &WinY, &FormW, &FormH, &border, &bits);
    XGetGeometry (XDisplay, XtWindow(Elem->WidgetID), &Dummy,
		  &WinX, &WinY, &ValW, &ValH, &border, &bits);

    PRO3 ("ELEM: %d  POS: %d/%d...", Elem->BoxID, WinX, WinY)
    PRO4 ("\\GEO: [%dx%d] %dx%d\n", FormW, FormH, ValW, ValH)

    n = 0;
    if (SetX && (PosX = ((FormW-ValW)/2)) != WinX)
    {
      XtSetArg (args[n], XmNleftOffset, PosX); n++;
    }
    if (SetY && (PosY = ((FormH-ValH)/2)) != WinY)
    {
      XtSetArg (args[n], XmNtopOffset, PosY); n++;
    }
    XtSetValues (Elem->WidgetID, args, n);
  }
}


void DLG_AdjustDlg (DLGBOX *Box)
{
  int      countX, countY;
  DLGBOX  *Ptr;
  DLGELEM *Elem;

  PRO_START ("\n")

  if (Box == NULL)
    PRO_ABORT ("Empty box")

  if (Box->Dir == BOXROWS || Box->Dir == BOXCOLS)
  {
    /* TODO:
       Anzahl FILL/MIDDLE berechnen -> Divisor fuer Platzausnutzung 
    */
    for (countY = 0; countY < Box->Rows; ++countY)
      for (countX = 0; countX < Box->Cols; ++countX)
	if ((Ptr = Box->List[countX][countY]) != NULL)
	{
	  if (Ptr->Dir == BOXROWS || Ptr->Dir == BOXCOLS)
	    DLG_AdjustBox (Ptr);
	  DLG_AdjustDlg (Ptr);
	}
  }
  elsif (Box->Dir == BOXGRID)
  {
    DLG_AdjustBox (Box);
    /* TODO:
       AdjustDlg fr alle Subboxen 
    */
  }
  elsif ((Elem = Box->Item) != NULL)
    while (Elem != NULL)
    {
      DLG_AdjustBox  (Box);
      DLG_AdjustElem (Elem);
      Elem = Elem->Next;
    }
  else
    PRO_ABORT ("No Item")

  PRO_END
}


#if 0
#define SHOW_FRAME(x) XtParent(x)
#else
#define SHOW_FRAME(x) x
#endif

void DLG_AdjustArrays (DLGBOX *Box)
{
  int               countX, countY;
  int               SizeX[100], SizeY[100];
  int               PosX, PosY;
  XtWidgetGeometry  Geo;
  DLGBOX           *Ptr;
  DLGELEM          *Elem;

  PRO_START ("\n")

  if (Box == NULL)
    PRO_ABORT ("Empty box")

  if (Box->Dir == BOXROWS || Box->Dir == BOXCOLS)
  {
    for (countY = 0; countY < Box->Rows; ++countY)
      for (countX = 0; countX < Box->Cols; ++countX)
	if ((Ptr = Box->List[countX][countY]) != NULL)
	  DLG_AdjustArrays (Ptr);
  }
  elsif (Box->Dir == BOXGRID)
  {
    /* find out the max. size of each row/cols */
    for (countX = 0; countX < Box->Cols; ++countX)
      SizeX[countX] = 0;
    for (countY = 0; countY < Box->Rows; ++countY)
      SizeY[countY] = 0;

    for (countY = 0; countY < Box->Rows; ++countY)
      for (countX = 0; countX < Box->Cols; ++countX)
	if ((Ptr = Box->List[countX][countY]) != NULL && Ptr->Box != NULL)
	{
	  Geo.request_mode = CWWidth | CWHeight;
	  XtQueryGeometry (SHOW_FRAME(Ptr->Box), NULL, &Geo);
	  if (Geo.request_mode & (CWWidth | CWHeight))
	  {
	    PRO2 ("PARENT-GEOMETRY: %d x %d\n", Geo.width, Geo.height)
	    SizeX[countX] = MAX (SizeX[countX], Geo.width);
	    SizeY[countY] = MAX (SizeY[countY], Geo.height);
	  }
	}

    /* Align all boxes */
    PosY = 0;
    for (countY = 0; countY < Box->Rows; ++countY)
    {
      PosX = 0;
      for (countX = 0; countX < Box->Cols; ++countX)
      {
        if ((Ptr = Box->List[countX][countY]) != NULL && Ptr->Box != NULL)
	{
	  PRO5 ("PLACE %d: %3d/%3d: %dx%d\n", 
		Ptr->BoxID, PosX, PosY, SizeX[countX], SizeY[countY])
          /* The y-position must be set to 0, while the x position must be
	     set to the calculated postion, don't know why :-| */
	  XtConfigureWidget (SHOW_FRAME(Ptr->Box),
			     PosX, 0, SizeX[countX], SizeY[countY], 0);
	}
	PosX += SizeX[countX];
      }
      PosY += SizeY[countY];
    }
  }
  elsif (Box->Dir == BOXARRAY)
  {
    for (countY = 0; countY < Box->Rows; ++countY)
      for (countX = 0; countX < Box->Cols; ++countX)
        if ((Ptr = Box->List[countX][countY]) != NULL && Ptr->Box != NULL)
	{
	  Elem = Ptr->Item;
	  XtVaSetValues (Ptr->Main, 
			 MNcolumn,      countX,
			 MNrow,         countY,
			 MNxAlignment, M_ALIGN_FILL,
			 MNyAlignment, M_ALIGN_CENTER,
			 MNpadX,       0,
			 MNpadY,       0,
			 NULL);
	  if (Elem != NULL)
	    XtVaSetValues (Elem->WidgetID, 
			   MNxAlignment, M_ALIGN_FILL,
			   MNyAlignment, M_ALIGN_CENTER,
			   MNpadX,       0,
			   MNpadY,       0,
			   NULL);
	}
  }

  PRO_END
}
