#include <math.h>

#include "dlgData.h"

/*----------------------------------------------------------------------
  Funktion: prft, ob eine gegebene Dialogstruktur leer ist
            (Wird nur fr DLGMakeDialog bentigt)

  INPUT:  Dlg     Dialog, der auf Inhalt geprft werden soll
  OUTPUT: (bool)  TRUE, falls der Dialog leer ist
                  FALSE, sonst

  Bsp.: DIALOG *Dlg;
        ...
	if (DLGIsEmpty (Dlg))
	  ...
  ----------------------------------------------------------------------*/
dual DLGIsEmpty (DIALOG *Dlg)
{
  return (Dlg == NULL || Dlg->Box == NULL);
}


/*----------------------------------------------------------------------
  Funktion: DLGGetExitButton liefert die Nummer der Taste, mit der
            der Dialog beendet wurde. Wird bei der Benutzung von
	    DLGOpenDialog bentigt, um festzustellen, mit welcher
	    Taste der Dialog beendet wurde, da dieser in bei 
	    DLGOpenDialog bei bettigen des CANCEL-Buttons nicht
	    automatisch gelscht wird.

  INPUT:  Dlg    Dialog, zu dem die Beendigungstaste gesucht ist
  OUTPUT: (INT)  -1, falls Dialog noch nicht beendet wurde,
                     bzw. falls der Dialog leer ist. 
		  1, falls der Dialog mit OK beendet wurde.
		  0, falls der Dialog mit CANCEL beendet wurde.

  Bsp.: INT     Button;
        DIALOG *Dlg;
	...
	Button = DLGGetExitButton (Dlg))
	  ...
  ----------------------------------------------------------------------*/
int DLGGetExitButton (DIALOG *Dlg)
{
  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR0 ("Try to get exit-button of an empty dialog", -1)

  PRO1 ("\\%d...", Dlg->Buttons.Result)

  PRO_END
  return Dlg->Buttons.Result;
}


/*----------------------------------------------------------------------
  Funktion: Liefert die in einem Dialog eingetragenen Benutzer-
            informationen bzw. NULL, falls keine Infos eingetragen sind.

  INPUT:  Dlg     Dialog, dessen Info gesucht ist.
  OUTPUT: (void*) Zeiger auf die eingetragene Struktur. Der Anwender mu
                  wissen, was er ursprnglich mal reingeschrieben hat.
	          Die Freigabe des belegten Speicher mu vom Anwender erfolgen.
  
  Bsp.: DLGUSTER *Dlg;
        ...
        Result = (...*)DLGGetUserData (Dlg, 0);
  ----------------------------------------------------------------------*/
void *DLGGetUserData (DIALOG *Dlg, int BoxID)
{
  DLGBOX *Ptr;

  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR0 ("Try to get user-data of an empty dialog", NULL)

  if ((Ptr = DLG_GetBox(Dlg->Box, DLGBox(BoxID))) == NULL)
    PRO_ERROR_VAL1 ("Dialog contains no such BoxID %d", BoxID, NULL)

  PRO_END
  return Ptr->UserData;
}


/*----------------------------------------------------------------------
  Funktion: liest den Wert eines Dialog-Items, welches durch die 
            Box-Numer gegeben ist, falls der Dialog/die Box nicht
	    existiert oder die Box weitere Unterboxen enthlt, wird
	    ein NULL-Zeiger zurckgeliefert.
	    Ansonsten mu der zurckgelieferte Zeiger in den dazu
	    passenden Typ (siehe Dokumentation) gecastet werden.
	    In der Regel kann - sorgflltige Programmierung vorausgesetzt -
	    jedoch auf berprfung der Gltigkeit des Ergebnisses entfallen. 

  INPUT:  Dlg      Dialog aus dem der Item-Inhalt gelesen werden soll
          BoxID    Nummber der Box, die den auszulesenden Eintrag enthlt
  OUTPUT: (void*)  Zeiger auf das Ergebnis bzw. NULL, falls einer der oben
                   genannten Flle eintritt.
  
  Bsp.: void       *Ergebnis;        
        int         Zahl;         
	DIALOG  *Dlg;
	...                                       (Box 10 ist GANZ-Item)
	if ((Ergebnis = DLGGetValue (Dlg, 10)) != NULL)   
	  Zahl = *(PINT)Ergebnis;

	bzw. ohne Prfung:
	Zahl = *(PINT)DLGGetValue (Dlg, 10);
  ----------------------------------------------------------------------*/
void *DLGGetValue (DIALOG *Dlg, int BoxID)
{
  int       count, Number;
  DLGELEM  *Elem = NULL;
  DLGBOX   *Ptr = NULL;
  void     *Result = NULL;

  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR0 ("\\Try to get value of an empty dialog", NULL)

  if ((Ptr = DLG_GetBox(Dlg->Box, DLGBox(BoxID))) == NULL)
    PRO_ERROR_VAL1 ("The dialog contains no BoxID %d", BoxID, NULL)
 
  if (Ptr->Dir != BOXELEM || Ptr->Item == NULL)
    PRO_ERROR_VAL1 ("Box %d contains no element", BoxID, NULL)

  PRO1 ("\\[%d]...", BoxID)

  /* Je nach Item-Type die genderten Werte zurckliefern: */
  Elem = Ptr->Item;
  switch (Elem->Type)
  {
    case  DlgEMPTY :
    case  DlgLABEL : break;
    case   DlgBOOL : Result = &Elem->Value.drec.Value;
                     break;
    case    DlgINT : Result = &Elem->Value.irec.Value;
                     break;
    case DlgDOUBLE : Result = &Elem->Value.rrec.Value;
                     break;
    case   DlgTEXT : Result = Elem->Value.srec.Text;
                     break;
    case  DlgSCALE : Result = &Elem->Value.zrec.Val;
                     break;
    case    DlgBAR : Result = &Elem->Value.brec.Val;
                     break;
    case   DlgUPDN : Result = &Elem->Value.urec.Val;
                     break;
    case   DlgLIST : Result = &Elem->Value.lrec.Result;
                     break;
    case  DlgOMENU : Result = &Elem->Value.orec.Result;
                     break;
    case   DlgTBOX : if (Elem->Value.trec.Result != NULL)
                       free (Elem->Value.trec.Result);
                     Elem->Value.trec.Result = 
		       (dual*) malloc (sizeof(dual) * Elem->Value.trec.Number);
		     for (count = 0; count < Elem->Value.trec.Number; ++count)
		       Elem->Value.trec.Result[count] = 
			 Elem->Value.trec.Toggle[count].Value;
		     Result = &Elem->Value.trec.Result;
		     break;
    case   DlgRBOX : Number = -1;
                     for (count = 0; count < Elem->Value.trec.Number; ++count)
		       if (Elem->Value.trec.Toggle[count].Value)
			 Number = count;
		     Result = &Number;
		     break;
    case   DlgPROC : PRO0 ("Proc has NO result...")
                     break;
    case  DlgARROW : PRO0 ("Arrow has NO result...")
                     break;
    case   DlgPIC  : PRO0 ("Picture/-button has NO result...")
                     break;
    case   DlgFSEL : Result = Elem->Value.grec.Result;
                     break;
    case   DlgDIAL : Result = &Elem->Value.erec.Val;
                     break;
           default : break;
  }

  PRO_END
  return Result;
}

/* ------------- Convenience Functions ------------------ */

dual DLGGetBool (DIALOG *Dlg, int BoxID)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgBOOL)) == NULL)
    PRO_RETURN1 ("Invalid BoxID %d", BoxID, FALSE)

  PRO_END
  return Elem->Value.drec.Value;
}


int DLGGetInt (DIALOG *Dlg, int BoxID)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgINT)) == NULL)
    PRO_RETURN1 ("Invalid BoxID %d", BoxID, 0)

  PRO_END
  return Elem->Value.irec.Value;
}


double DLGGetDouble (DIALOG *Dlg, int BoxID)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgDOUBLE)) == NULL)
    PRO_RETURN1 ("Invalid BoxID %d", BoxID, 0.0)

  PRO_END
  return Elem->Value.rrec.Value;
}


char *DLGGetText (DIALOG *Dlg, int BoxID)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgTEXT)) == NULL)
    PRO_RETURN1 ("Invalid BoxID %d", BoxID, NULL)

  PRO_END
  return Elem->Value.srec.Text;
}


double DLGGetScale (DIALOG *Dlg, int BoxID)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgSCALE)) == NULL)
    PRO_RETURN1 ("Invalid BoxID %d", BoxID, 0.0)

  PRO_END
  return Elem->Value.zrec.Val;
}


double DLGGetBar (DIALOG *Dlg, int BoxID)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgBAR)) == NULL)
    PRO_RETURN1 ("Invalid BoxID %d", BoxID, 0.0)

  PRO_END
  return Elem->Value.brec.Val;
}


double DLGGetUpDn (DIALOG *Dlg, int BoxID)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgUPDN)) == NULL)
    PRO_RETURN1 ("Invalid BoxID %d", BoxID, 0.0)

  PRO_END
  return Elem->Value.urec.Val;
}


int DLGGetList (DIALOG *Dlg, int BoxID)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgLIST)) == NULL)
    PRO_RETURN1 ("Invalid BoxID %d", BoxID, -1)

  PRO_END
  return Elem->Value.lrec.Result;
}


int DLGGetOMenu (DIALOG *Dlg, int BoxID)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgOMENU)) == NULL)
    PRO_RETURN1 ("Invalid BoxID %d", BoxID, -1)

  PRO_END
  return Elem->Value.orec.Result;
}


dual DLGGetTBox (DIALOG *Dlg, int BoxID, int Number)
{
  dual     Result = FALSE;
  int      count;
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgTBOX)) == NULL)
    PRO_RETURN1 ("Invalid BoxID %d", BoxID, FALSE)

  if (Elem->Value.trec.Result != NULL)
    free (Elem->Value.trec.Result);
  Elem->Value.trec.Result = 
    (dual*)malloc (sizeof(dual) * Elem->Value.trec.Number);
  for (count = 0; count < Elem->Value.trec.Number; ++count)
    Elem->Value.trec.Result[count] = Elem->Value.trec.Toggle[count].Value;

  if (Number < Elem->Value.trec.Number)
    Result = Elem->Value.trec.Result[Number];
  else
    PRO2 ("Button %d is not in ToggleBox (# %d)", 
	  Number, Elem->Value.trec.Number)

  PRO_END
  return Result;
}


int DLGGetRBox (DIALOG *Dlg, int BoxID)
{
  int      count;
  int      Result = -1;
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgRBOX)) == NULL)
    PRO_RETURN1 ("Invalid BoxID %d", BoxID, -1)

  for (count = 0; count < Elem->Value.trec.Number; ++count)
    if (Elem->Value.trec.Toggle[count].Value)
      Result = count;

  PRO_END
  return Result;
}


char *DLGGetFileSel (DIALOG *Dlg, int BoxID)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgFSEL)) == NULL)
    PRO_RETURN1 ("Invalid BoxID %d", BoxID, NULL)

  PRO_END
  return Elem->Value.grec.Result;
}


char *DLGGetDirSel (DIALOG *Dlg, int BoxID)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgDIRSEL)) == NULL)
    PRO_RETURN1 ("Invalid BoxID %d", BoxID, NULL)

  PRO_END
  return Elem->Value.grec.Result;
}


#if (XmVERSION>=2)
  int DLGGetNote (DIALOG *Dlg, int BoxID)
  {
    DLGELEM *Elem;

    PRO_START ("")

    if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgNOTE)) == NULL)
      PRO_RETURN1 ("Invalid BoxID %d", BoxID, -1)

    PRO_END
    return Elem->Value.nrec.ActPage;
  }

  TREEITEM *DLGGetTree (DIALOG *Dlg, int BoxID)
  {
    DLGELEM *Elem;

    PRO_START ("")

    if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgTREE)) == NULL)
      PRO_RETURN1 ("Invalid BoxID %d", BoxID, NULL)

    PRO_END
    return Elem->Value.krec.Result;
  }

  int DLGGetReason (DIALOG *Dlg, int BoxID)
  {
    DLGELEM *Elem;

    PRO_START ("")

    if ((Elem =DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgTREE)) == NULL)
      PRO_RETURN ("Only Trees supported, yet", -1)

    PRO_END
    return Elem->Value.krec.Reason;
  }
#endif


double DLGGetDial (DIALOG *Dlg, int BoxID)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgDIAL)) == NULL)
    PRO_RETURN1 ("Invalid BoxID %d", BoxID, -1)

  PRO_END
  return Elem->Value.erec.Val;
}



Widget DLGGetBoxWidget (DIALOG *Dlg, int BoxID)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), -1)) == NULL)
    PRO_RETURN1 ("Invalid BoxID %d", BoxID, NULL)

  PRO_END
  return Elem->WidgetID;
}

