/* **************************************** */
/* *** (c) Stefan Vietinghoff           *** */
/* *** Functions for inserting elements *** */
/* **************************************** */
#include "dlgData.h"


/* *************************************** */
/* ** Creates a new element with        ** */
/* ** default-values                    ** */
/* ** INPUT:  Title       Name of elem  ** */
/* ** OUTPUT: (DLGELEM*)  Created elem  ** */
/* *************************************** */
static DLGELEM *DLG_NewElem (char *Name)
{
  DLGELEM *Elem;

  if ((Elem = (DLGELEM*)calloc(1, sizeof(DLGELEM))) == NULL)
    return NULL;

  DLG_strncp (Elem->Name, Name, DLG_TLEN);
  Elem->BoxID    = 0;
  Elem->PosX     = -1;
  Elem->PosY     = -1;
  Elem->LabelPos = DlgWEST;
  Elem->ValuePos = DlgEAST;
  Elem->hidden   = FALSE;
  Elem->active   = TRUE;
  Elem->Dlg      = NULL;
  Elem->NumArgs  = 0;
  Elem->UserArgs = NULL;
  Elem->fg       = -1;
  Elem->bg       = -1;
  Elem->Master   = NULL;
  Elem->Label    = NULL;
  Elem->WidgetID = NULL;
  Elem->Type     = DlgEMPTY;
  Elem->Next     = NULL;

  return Elem;
}


/* ********************************************* */
/* ** Inserts an element into a box of a      ** */
/* ** given dialog                            ** */
/* ** INPUT:  Dlg    Dialog to put box in     ** */
/* **         Box    target-position          ** */
/* **         Elem   Element to insert        ** */
/* ** OUTPUT: (bool) TRUE on success          ** */
/* ** NOTE:   This version is prepaired for   ** */
/* **         using box-names to identificate ** */
/* **         the box (BOX_ID)                ** */
/* ******************************************** */
static dual DLG_InsertElem (DIALOG *Dlg, BOX_ID BoxID, DLGELEM *Elem)
{
  DLGBOX  *Box;
  DLGELEM *Ptr;

  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR0 ("Empty or invalid Dialog", FALSE)

  if (Elem == NULL)
    PRO_RETURN ("Nothing to insert", FALSE)

  if ((Box = DLG_GetBox(Dlg->Box, BoxID)) == NULL)
    PRO_ERROR_VAL1 ("The dialog contains no BoxID %d", BoxID.BoxID, FALSE)

  if ((Elem->PosX != -1 && Elem->PosY != -1) &&
      (Box->Dir == BOXLIST || (Box->Dir == BOXNONE && Box->Item == NULL)))
  {
    Elem->BoxID = BoxID.BoxID;
    Elem->Dlg   = Dlg;

    Box->Dir = BOXLIST;
    if (Box->Item == NULL)
      Box->Item = Elem;
    else
    {
      Ptr = Box->Item;
      while (Ptr->Next != NULL)
	Ptr = Ptr->Next;
      Ptr->Next = Elem;
    }
    (Box->Number)++;
  }
  elsif (Box->Dir == BOXNONE && Box->Item == NULL)
  {
    Elem->BoxID = BoxID.BoxID;
    Elem->Dlg   = Dlg;

    Box->Dir   = BOXELEM;
    Box->Item  = Elem;
  } 
  else
    PRO_ERROR0 ("Box is not atomar", FALSE)             

  PRO_END
  return TRUE;
}


#if 0
/* **************************************************** */
/* ** Generates a dialog element with given          ** */
/* ** type and parameter.                            ** */
/* ** INPUT:  Type      type of element              ** */
/* **         x, y      Offset stored in elem-struct ** */
/* **         Title     Name of produced element     ** */
/* **         Arguments type-specific parameter      ** */
/* **************************************************** */
static DLGELEM *DLG_MakeElem (int Type, int x, int y, 
			                char *Title, va_list Arguments)
{
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type = Type;
  Elem->PosX = x;
  Elem->PosY = y;

  switch (Type)
  {
    case DlgLABEL :
      Elem->LabelPos = DlgNOWHERE;
      DLG_strncp (Elem->Value.frec.Text, va_arg(Arguments, char*), DLG_LLEN);
      break;

    case DlgBOOL :
      Elem->LabelPos           = DlgNOWHERE;
      Elem->ValuePos           = DlgWEST;
      Elem->Value.drec.Value   = va_arg (Arguments, dual);
      Elem->Value.drec.Default = Elem->Value.drec.Value;
      Elem->Value.drec.FPtr    = NULL;
      break;

    case DlgINT :
      Elem->LabelPos           = DlgWEST;
      Elem->Value.irec.Value   = va_arg (Arguments, int);
      Elem->Value.irec.Digits  = va_arg (Arguments, int);
      Elem->Value.irec.Default = Elem->Value.irec.Value;
      Elem->Value.irec.FPtr    = NULL;
      break;

    case DlgDOUBLE :
      Elem->LabelPos            = DlgWEST;
      Elem->Value.rrec.Value    = va_arg (Arguments, double);
      Elem->Value.rrec.Digits   = va_arg (Arguments, int);
      Elem->Value.rrec.Decimals = va_arg (Arguments, int);
      Elem->Value.rrec.Default  = Elem->Value.rrec.Value;
      Elem->Value.rrec.FPtr     = NULL;
      break;

    case DlgTEXT :
      Elem->LabelPos           = DlgWEST;
      Elem->Value.srec.Length  = va_arg (Arguments, int);
      Elem->Value.srec.Text    = DLG_strcp (va_arg(Arguments, char*));
      Elem->Value.srec.Cols    = va_arg (Arguments, int);
      Elem->Value.srec.Rows    = va_arg (Arguments, int);
      Elem->Value.srec.Default = DLG_strcp (Elem->Value.srec.Text);
      Elem->Value.srec.FPtr    = NULL;
      break;

    case DlgSCALE :
      {
	double swap;

	Elem->LabelPos            = DlgWEST;
	Elem->Value.zrec.Min      = va_arg (Arguments, double);
	Elem->Value.zrec.Max      = va_arg (Arguments, double);
	Elem->Value.zrec.Val      = va_arg (Arguments, double);
	Elem->Value.zrec.Decimals = va_arg (Arguments, int);
	Elem->Value.zrec.Side     = va_arg (Arguments, int);

	/* prepare correct order (min,max) for motif (but save original) */
	if ((Elem->Value.zrec.reverse = 
	     (Elem->Value.zrec.Min > Elem->Value.zrec.Max)))
	{
	  swap                 = Elem->Value.zrec.Min;
	  Elem->Value.zrec.Min = Elem->Value.zrec.Max;
	  Elem->Value.zrec.Max = swap;
	}

	/* range-test*/
	if (Elem->Value.zrec.Val < Elem->Value.zrec.Min)
	  Elem->Value.zrec.Val = Elem->Value.zrec.Min;
	elsif (Elem->Value.zrec.Val > Elem->Value.zrec.Max)
	  Elem->Value.zrec.Val = Elem->Value.zrec.Max;

	Elem->Value.zrec.Default = Elem->Value.zrec.Val;

	Elem->Value.zrec.FPtr = NULL;
      }
      break;

    case DlgBAR :
      {
	double swap;

	Elem->LabelPos        = DlgWEST;
	Elem->Value.brec.Min  = va_arg (Arguments, double);
	Elem->Value.brec.Max  = va_arg (Arguments, double);
	Elem->Value.brec.Val  = va_arg (Arguments, double);
	Elem->Value.brec.Step = va_arg (Arguments, double);
	Elem->Value.brec.Side = va_arg (Arguments, int);

	if ((Elem->Value.brec.reverse = 
	     (Elem->Value.brec.Min > Elem->Value.brec.Max)))
	{
	  swap                 = Elem->Value.brec.Min;
	  Elem->Value.brec.Min = Elem->Value.brec.Max;
	  Elem->Value.brec.Max = swap;
	}

	if (Elem->Value.brec.Val < Elem->Value.brec.Min)
	  Elem->Value.brec.Val = Elem->Value.brec.Min;
	elsif (Elem->Value.brec.Val > Elem->Value.brec.Max)
	  Elem->Value.brec.Val = Elem->Value.brec.Max;

	Elem->Value.brec.Default = Elem->Value.brec.Val;

	Elem->Value.brec.FPtr = NULL;
      }
      break;

    case DlgUPDN :
      Elem->LabelPos        = DlgWEST;
      Elem->Value.urec.Min  = va_arg (Arguments, double);
      Elem->Value.urec.Max  = va_arg (Arguments, double);
      Elem->Value.urec.Val  = va_arg (Arguments, double);
      Elem->Value.urec.Step = va_arg (Arguments, double);

      if (Elem->Value.urec.Min < Elem->Value.urec.Max)
	   Elem->Value.urec.Val = 
		DLG_UPDN_VAL (Elem->Value.urec.Min,  Elem->Value.urec.Max,
			      Elem->Value.urec.Step, Elem->Value.urec.Val);
      Elem->Value.urec.Decimals = 
	   MAX (DLG_GetDecimals(Elem->Value.urec.Min),
		   (MAX(DLG_GetDecimals(Elem->Value.urec.Max),
		        MAX(DLG_GetDecimals(Elem->Value.urec.Val),
			    DLG_GetDecimals(Elem->Value.urec.Step)))));

      Elem->Value.urec.Length = MAX (DLG_GetDigits(Elem->Value.urec.Min),
				     DLG_GetDigits(Elem->Value.urec.Max)) + 
	                          Elem->Value.urec.Decimals + 2;
      if (Elem->Value.urec.Min >= Elem->Value.urec.Max)
	Elem->Value.urec.Length += 3;
	Elem->Value.urec.Default = Elem->Value.urec.Val;

      Elem->Value.urec.FPtr  = NULL;
      break;

    case DlgLIST :
      {
	int    count;
	char **ItemList = NULL;

	Elem->LabelPos          = DlgNORTH_WEST;
	Elem->ValuePos          = DlgWEST;
	Elem->Value.lrec.Lines  = va_arg (Arguments, int);
	Elem->Value.lrec.Number = va_arg (Arguments, int);

	if (Elem->Value.lrec.Number > 0)          /* single parameters */
	{
	  Elem->Value.lrec.Item = 
	    (ITEMTEXT *)malloc (sizeof(ITEMTEXT) * Elem->Value.lrec.Number);
	  for (count = 0; count < Elem->Value.lrec.Number; ++count)
	    DLG_strncp (Elem->Value.lrec.Item[count], 
			va_arg(Arguments, char*), DLG_LLEN);
	}
	elsif (Elem->Value.lrec.Number == 0)
	  Elem->Value.lrec.Item = NULL;
	else                                      /* list of entries */
	{
	  Elem->Value.lrec.Number *= -1;
	  ItemList = va_arg (Arguments, char**);
	  Elem->Value.lrec.Item =
	    (ITEMTEXT *)malloc (sizeof(ITEMTEXT) * Elem->Value.lrec.Number);
	  for (count = 0; count < Elem->Value.lrec.Number; ++count)
	    DLG_strncp (Elem->Value.lrec.Item[count], ItemList[count], 
			DLG_LLEN);
	}

	Elem->Value.lrec.Result = va_arg (Arguments, int);
	if (Elem->Value.lrec.Result >= Elem->Value.lrec.Number)
	{
	  Elem->Value.lrec.Result = 0;
	  PRO1 ("WARNING: Default is out of range: %d\n", 
		Elem->Value.lrec.Result)
        }
	
	Elem->Value.lrec.Default = Elem->Value.lrec.Result;
	Elem->Value.lrec.FPtr = NULL;
      }
      break;

    case DlgOMENU :
      {
	int    count;
	char **ItemList = NULL;

	Elem->LabelPos          = DlgWEST;
	Elem->Value.orec.Number = va_arg (Arguments, int);

	if (Elem->Value.orec.Number > 0)
	{
	  Elem->Value.orec.Item = 
	    (ITEMREC*)malloc (sizeof(ITEMREC) * Elem->Value.orec.Number);
	  for (count = 0; count < Elem->Value.orec.Number; ++count)
	  {
	    DLG_strncp (Elem->Value.orec.Item[count].Label, 
			va_arg(Arguments, char*), DLG_LLEN);
	    Elem->Value.orec.Item[count].Index    = count;
	    Elem->Value.orec.Item[count].Parent   = Elem;
	    Elem->Value.orec.Item[count].WidgetID = NULL;
	  }
	}
	elsif (Elem->Value.orec.Number == 0)
	  Elem->Value.orec.Item = NULL;
	else
	{
	  Elem->Value.orec.Number *= -1;
	  ItemList = va_arg (Arguments, char**);
	  Elem->Value.orec.Item =
	    (ITEMREC*)malloc (sizeof(ITEMREC) * Elem->Value.orec.Number);
	  for (count = 0; count < Elem->Value.orec.Number; ++count)
	  {
	    DLG_strncp (Elem->Value.orec.Item[count].Label,
			ItemList[count], DLG_LLEN);
	    Elem->Value.orec.Item[count].Index    = count;
	    Elem->Value.orec.Item[count].Parent   = Elem;
	    Elem->Value.orec.Item[count].WidgetID = NULL;
	  }
	}
	if ((Elem->Value.orec.Result = 
	     abs(va_arg(Arguments, int))) > Elem->Value.orec.Number)
	{
	  Elem->Value.orec.Result = 0;
	  PRO1 ("WARNING: Default is out of range: %d\n", 
		Elem->Value.orec.Result)
	}

	Elem->Value.orec.Default = Elem->Value.orec.Result;
	Elem->Value.orec.FPtr = NULL;
      }
      break;

    case DlgTBOX :
      {
	int     count;
	TOGGLE *ToggleList = NULL;

	Elem->LabelPos          = DlgNORTH_WEST;
	Elem->ValuePos          = DlgWEST;
	Elem->Value.trec.RowCol = va_arg (Arguments, int);
	Elem->Value.trec.Number = va_arg (Arguments, int);

	if (Elem->Value.trec.Number >= 0)
	{
	  Elem->Value.trec.Toggle = 
	    (SWITCHREC*)malloc (sizeof(SWITCHREC) * Elem->Value.trec.Number);
	  for (count = 0; count < Elem->Value.trec.Number; ++count)
	  {
	    DLG_strncp (Elem->Value.trec.Toggle[count].Item,
			va_arg(Arguments, char*), DLG_LLEN);
	    Elem->Value.trec.Toggle[count].Value  = va_arg (Arguments, dual);
	    Elem->Value.trec.Toggle[count].Index  = count;
	    Elem->Value.trec.Toggle[count].Parent = Elem;
	  }
	}
	else
	{
	  Elem->Value.trec.Number *= -1;
	  ToggleList = va_arg (Arguments, TOGGLE*);
	  Elem->Value.trec.Toggle = 
	    (SWITCHREC*)malloc (sizeof(SWITCHREC)*Elem->Value.trec.Number);
	  for (count = 0; count < Elem->Value.trec.Number; ++count)
	  {
	    DLG_strncp (Elem->Value.trec.Toggle[count].Item,
			ToggleList[count].Item, DLG_LLEN);
	    Elem->Value.trec.Toggle[count].Value  = ToggleList[count].Value;
	    Elem->Value.trec.Toggle[count].Index  = count;
	    Elem->Value.trec.Toggle[count].Parent = Elem;
	  }
	} 
	Elem->Value.trec.Result = NULL;
	Elem->Value.trec.Default = 
	  (dual *)malloc (sizeof(dual) * abs(Elem->Value.trec.Number));
	for (count = 0; count < Elem->Value.trec.Number; ++count)
	  Elem->Value.trec.Default[count] = 
	    Elem->Value.trec.Toggle[count].Value;
	Elem->Value.trec.FPtr = NULL;
      }
      break;

    case DlgRBOX :
      {
	int    count, Default;
	char **ItemList = NULL;

	Elem->LabelPos          = DlgNORTH_WEST;
	Elem->LabelPos          = DlgWEST;
	Elem->Value.trec.RowCol = va_arg (Arguments, int);
	Elem->Value.trec.Number = va_arg (Arguments, int);

	if (Elem->Value.trec.Number > 0)
	{
	  Elem->Value.trec.Toggle =
	    (SWITCHREC*)malloc (sizeof(SWITCHREC) * Elem->Value.trec.Number);
	  for (count = 0; count < Elem->Value.trec.Number; ++count)
	  {
	    DLG_strncp (Elem->Value.trec.Toggle[count].Item,
			va_arg(Arguments, char*), DLG_LLEN);
	    Elem->Value.trec.Toggle[count].Value  = FALSE;
	    Elem->Value.trec.Toggle[count].Index  = count;
	    Elem->Value.trec.Toggle[count].Parent = Elem;
	  }
	}
	elsif (Elem->Value.trec.Number == 0)
	  Elem->Value.trec.Toggle = NULL;
	else
	{
	  Elem->Value.trec.Number *= -1;
	  ItemList = va_arg (Arguments, char**);
	  Elem->Value.trec.Toggle =
	    (SWITCHREC*)malloc (sizeof(SWITCHREC) * Elem->Value.trec.Number); 
	  for (count = 0; count < Elem->Value.trec.Number; ++count)
	  {
	    DLG_strncp (Elem->Value.trec.Toggle[count].Item, 
			ItemList[count], DLG_LLEN);
	    Elem->Value.trec.Toggle[count].Value  = FALSE;
	    Elem->Value.trec.Toggle[count].Index  = count;
	    Elem->Value.trec.Toggle[count].Parent = Elem;
	  }
	}
	if ((Default = abs(va_arg(Arguments, int))) > Elem->Value.trec.Number)
	{
	  Default = 0;
	  PRO1 ("WARNING: Default is out of range: %d\n", Default)
	}
	Elem->Value.trec.Toggle[Default].Value = TRUE;
	Elem->Value.trec.Result = NULL;
	Elem->Value.trec.Default = 
	  (dual *)malloc (sizeof(dual) * abs(Elem->Value.trec.Number));
	for (count = 0; count < Elem->Value.trec.Number; ++count)
	  Elem->Value.trec.Default[count] = 
	    Elem->Value.trec.Toggle[count].Value;
	Elem->Value.trec.FPtr   = NULL;
      }
      break;

    case DlgPROC :
      Elem->LabelPos        = DlgNOWHERE;
      Elem->Value.crec.FPtr = va_arg (Arguments, DLGFUNCTION);
      break;

    case DlgARROW :
      Elem->LabelPos        = DlgNOWHERE;
      Elem->ValuePos        = DlgCENTER;
      Elem->Value.arec.Dir  = va_arg (Arguments, int);
      Elem->Value.arec.FPtr = va_arg (Arguments, DLGFUNCTION);
      break;

    case DlgPIC :
      Elem->Value.prec.Format     = va_arg (Arguments, int);
      Elem->Value.prec.SizeX      = va_arg (Arguments, int);
      Elem->Value.prec.SizeY      = va_arg (Arguments, int);
      Elem->Value.prec.PicName[0] = '\0';
      Elem->Value.prec.fg         = 0;
      Elem->Value.prec.bg         = 0;
      Elem->Value.prec.iData      = NULL;

      if (Elem->Value.prec.Format < DlgUNKNOWN || 
		Elem->Value.prec.Format > DlgXPM)
	   PRO_RETURN ("Invalid Image-Type", FALSE)

	 if (Elem->Value.prec.SizeX == 0 || Elem->Value.prec.SizeY == 0)
      {
	   DLG_strncp (Elem->Value.prec.PicName, 
				(char*)va_arg(Arguments, void*), PATH_MAX);
	   Elem->Value.prec.PicName[PATH_MAX] = '\0';
      }
      else
        switch (Elem->Value.prec.Format)
	{
	  case DlgXBM :
	    Elem->Value.prec.Data = 
	      (void*)TOOLCopyXBM ((byte*)va_arg(Arguments, void*),
				  Elem->Value.prec.SizeX, 
				  Elem->Value.prec.SizeY);
	    break;

  	  case DlgXPM :
	    Elem->Value.prec.Data = 
	      (void*)TOOLCopyXPM ((char**)va_arg(Arguments, void*));
	    break;

	  default : 
	    Elem->Value.prec.Data = NULL;
	}

      Elem->Value.prec.FPtr = va_arg (Arguments, DLGFUNCTION);
      if (Elem->Value.prec.FPtr == NULL)
	 {
	   Elem->LabelPos = DlgNOWHERE;
	   Elem->ValuePos = DlgCENTER;
      }
      else
      {
        Elem->LabelPos = DlgCENTER;
	   Elem->ValuePos = DlgNORTH;
      }
      break;

    case DlgIMAGE :
      Elem->Value.xrec.Image = TOOLCopyImage (va_arg(Arguments, XImage*));
      Elem->Value.xrec.FPtr  = va_arg (Arguments, DLGFUNCTION);
      break; 

    case DlgFSEL :
    case DlgDIRSEL :
      {
	char FileSpec[PATH_MAX+1] = "";
	   
	DLG_strncp (Elem->Value.grec.Mask, 
                    va_arg(Arguments, char *), PATH_MAX);
	Elem->Value.grec.Length = va_arg (Arguments, int);
	DLG_strncp (Elem->Value.grec.Default,
		    va_arg(Arguments, char *), PATH_MAX);

        /* When length is zero, the dialog will be directly inserted */
	if (Elem->Value.grec.Length == 0)
	{
	  Elem->LabelPos = DlgNOWHERE;
	  Elem->ValuePos = DlgCENTER;
	}
	else
	  Elem->LabelPos = DlgWEST;

	TOOLParseFileName (Elem->Value.grec.Default, FileSpec, PATH_MAX);
	DLG_strncp        (Elem->Value.grec.Default, FileSpec, PATH_MAX);
	DLG_strncp        (Elem->Value.grec.Result,  FileSpec, PATH_MAX);

#if 1
	TOOLParseFileName (Elem->Value.grec.Mask, FileSpec, PATH_MAX);
#else
	TOOLGetFileDir (Elem->Value.grec.Mask, FileSpec, PATH_MAX);
#endif
	DLG_strncp     (Elem->Value.grec.Mask, FileSpec, PATH_MAX);

	Elem->Value.grec.Button = 0;
        Elem->Value.grec.Files  = (Type == DlgFSEL);
	Elem->Value.grec.Open   = FALSE;
	Elem->Value.grec.FPtr   = NULL;
      }
      break;

    case DlgDIAL :
      Elem->LabelPos            = DlgNOWHERE;
      Elem->ValuePos            = DlgCENTER;
      Elem->Value.erec.Min      = va_arg (Arguments, double);
      Elem->Value.erec.Max      = va_arg (Arguments, double);
      Elem->Value.erec.Val      = va_arg (Arguments, double);
      Elem->Value.erec.Digits   = va_arg (Arguments, int);
      Elem->Value.erec.Decimals = va_arg (Arguments, int);
      Elem->Value.erec.Default  = Elem->Value.erec.Val;
      Elem->Value.erec.FPtr     = NULL;
      break;

    case DlgSEGM :
      Elem->LabelPos            = DlgNOWHERE;
      Elem->ValuePos            = DlgCENTER;
      Elem->Value.hrec.Val      = va_arg (Arguments, double);
      Elem->Value.hrec.Digits   = va_arg (Arguments, int);
      Elem->Value.hrec.Decimals = va_arg (Arguments, int);
      Elem->Value.hrec.Default  = Elem->Value.hrec.Val;
      Elem->Value.hrec.FPtr     = NULL;
      break;

    case DlgUSER :
      Elem->LabelPos         = DlgNOWHERE;
      Elem->ValuePos         = DlgWEST;
      Elem->Value.wrec.FPtr  = va_arg (Arguments, WIDGETFUNCTION);
      Elem->Value.wrec.Param = va_arg (Arguments, void*);
      break;

#if (XmVERSION>=2)
    case DlgNOTE :
      Elem->LabelPos           = DlgNOWHERE;
      Elem->Value.nrec.Style   = va_arg (Arguments, int);
      Elem->Value.nrec.ActPage = 1;
      Elem->Value.nrec.Number  = 0;
      Elem->Value.nrec.Page    = NULL;
      Elem->Value.nrec.Default = Elem->Value.nrec.ActPage;
      Elem->Value.nrec.FPtr    = NULL;
      break;

    case DlgTREE :
      Elem->LabelPos            = DlgNOWHERE;
      Elem->Value.krec.Shadow   = 0;
      Elem->Value.krec.ClosePic = va_arg (Arguments, Pixmap);
      Elem->Value.krec.OpenPic  = va_arg (Arguments, Pixmap);
      Elem->Value.krec.Lines    = va_arg (Arguments, int);
      Elem->Value.krec.Header   = NULL;
      Elem->Value.krec.Data     = NULL;
      Elem->Value.krec.Reason   = 0;
      Elem->Value.krec.Result   = NULL;
      Elem->Value.krec.Default  = NULL;
      Elem->Value.krec.FPtr     = NULL;
      break;
#endif

    default :
      break;
  }

  return Elem;
}
#endif


/* *** LABEL *** */
static DLGELEM *DLG_MakeLabel (int x, int y, char *Title, const char *Label)
{
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type = DlgLABEL;
  Elem->PosX = x;
  Elem->PosY = y;
  Elem->LabelPos = DlgNOWHERE;
  DLG_strncp (Elem->Value.frec.Text, Label, DLG_LLEN);

  return Elem;
}

dual DLGInsertLabel (DIALOG *Dlg, int BoxID, char *Name, 
		                  const char *Label, ...)
{
  char    String[1024];
  va_list Arguments;

  va_start (Arguments, Label);
  vsprintf (String, Label, Arguments);
  va_end (Arguments);

  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			      DLG_MakeLabel(-1, -1, Name, String));
}

dual DLGPlaceLabel (DIALOG *Dlg, int BoxID, int x, int y, char *Name, 
		                    char *Label)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeLabel(x, y, Name, Label));
}

DLGELEM *DLGMakeLabel (char *Name, char *Label)
{
  return DLG_MakeLabel (-1, -1, Name, Label);
}


/* *** BOOL *** */
static DLGELEM *DLG_MakeBool (int x, int y, char *Title, dual Value)
{
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type               = DlgBOOL;
  Elem->PosX               = x;
  Elem->PosY               = y;
  Elem->LabelPos           = DlgNOWHERE;
  Elem->ValuePos           = DlgWEST;
  Elem->Value.drec.Value   = Value;
  Elem->Value.drec.Default = Value;
  Elem->Value.drec.FPtr    = NULL;

  return Elem;
}

dual DLGInsertBool (DIALOG *Dlg, int BoxID, char *Name, dual Value)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID), 
			 DLG_MakeBool(-1, -1, Name, Value));
}

dual DLGPlaceBool (DIALOG *Dlg, int BoxID, int x, int y, char *Name, 
		                   dual Value)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID), DLG_MakeBool(x, y, Name, Value));
}

DLGELEM *DLGMakeBool (char *Name, dual Value)
{
  return DLG_MakeBool (-1, -1, Name, Value);
}



/* *** INT *** */
static DLGELEM *DLG_MakeInt (int x, int y, char *Title, int Value, byte Digits)
{
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type               = DlgINT;
  Elem->PosX               = x;
  Elem->PosY               = y;
  Elem->LabelPos           = DlgWEST;
  Elem->Value.irec.Value   = Value;
  Elem->Value.irec.Digits  = Digits;
  Elem->Value.irec.Default = Value;
  Elem->Value.irec.FPtr    = NULL;

  return Elem;
}

dual DLGInsertInt (DIALOG *Dlg, int BoxID, char *Name, 
		                   int Value, byte Digits)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			      DLG_MakeInt(-1, -1, Name, Value, Digits));
}

dual DLGPlaceInt (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
		                  int Value, byte Digits)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			      DLG_MakeInt(x, y, Name, Value, Digits));
}

DLGELEM *DLGMakeInt (char *Name, int Value, byte Digits)
{
  return DLG_MakeInt (-1, -1, Name, Value, Digits);
}



/* *** DOUBLE *** */
static DLGELEM *DLG_MakeDouble (int x, int y, char *Title, 
				int Value, byte Digits, byte Decimals)
{
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type                = DlgDOUBLE;
  Elem->PosX                = x;
  Elem->PosY                = y;
  Elem->LabelPos            = DlgWEST;
  Elem->Value.rrec.Value    = Value;
  Elem->Value.rrec.Digits   = Digits;
  Elem->Value.rrec.Decimals = Decimals;
  Elem->Value.rrec.Default  = Elem->Value.rrec.Value;
  Elem->Value.rrec.FPtr     = NULL;

  return Elem;
}

dual DLGInsertDouble (DIALOG *Dlg, int BoxID, char *Name,
		                   double Value, byte Digits, byte Decimals)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeDouble(-1, -1, Name, 
				                Value, Digits, Decimals));
}

dual DLGPlaceDouble (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
		                  double Value, byte Digits, byte Decimals)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeDouble(x, y, Name, Value, Digits, Decimals));
}

DLGELEM *DLGMakeDouble (char *Name, double Value, byte Digits, byte Decimals)
{
  return DLG_MakeDouble (-1, -1, Name, Value, Digits, Decimals);
}



/* *** TEXT *** */
static DLGELEM *DLG_MakeText (int x, int y, char *Title, 
			      int Length, const char *Text, 
			      int Cols, int Rows)
{
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type               = DlgTEXT;
  Elem->PosX               = x;
  Elem->PosY               = y;
  Elem->LabelPos           = DlgWEST;
  Elem->Value.srec.Length  = Length;
  Elem->Value.srec.Text    = DLG_strcp (Text);
  Elem->Value.srec.Cols    = Cols;
  Elem->Value.srec.Rows    = Rows;
  Elem->Value.srec.Default = DLG_strcp (Text);
  Elem->Value.srec.FPtr    = NULL;

  return Elem;
}

dual DLGInsertText (DIALOG *Dlg, int BoxID, char *Name,
		                 int Length, const char *Text, 
		                 int Cols, int Rows)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeText(-1, -1, Name, Length, Text, Cols, Rows));
}

dual DLGPlaceText (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
		                int Length, char *Text, int Cols, int Rows)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeText(x, y, Name, Length, Text, Cols, Rows));
}

DLGELEM *DLGMakeText (char *Name, int Length, char *Text, int Cols, int Rows)
{
  return DLG_MakeText (-1, -1, Name, Length, Text, Cols, Rows);
}



/* *** SCALE *** */
static DLGELEM *DLG_MakeScale (int x, int y, char *Title, 
			       double Min, double Max, double Value, 
			       byte Decimals, int Align)
{
  double   swap;
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type                = DlgSCALE;
  Elem->PosX                = x;
  Elem->PosY                = y;
  Elem->LabelPos            = DlgWEST;
  Elem->Value.zrec.Min      = Min;
  Elem->Value.zrec.Max      = Max;
  Elem->Value.zrec.Val      = Value;
  Elem->Value.zrec.Decimals = Decimals;
  Elem->Value.zrec.Side     = Align;
  
  /* prepare correct order (min,max) for motif (but save original) */
  if ((Elem->Value.zrec.reverse = 
       (Elem->Value.zrec.Min > Elem->Value.zrec.Max)))
    {
      swap                 = Elem->Value.zrec.Min;
      Elem->Value.zrec.Min = Elem->Value.zrec.Max;
      Elem->Value.zrec.Max = swap;
    }

  /* range-test*/
  if (Elem->Value.zrec.Val < Elem->Value.zrec.Min)
    Elem->Value.zrec.Val = Elem->Value.zrec.Min;
  elsif (Elem->Value.zrec.Val > Elem->Value.zrec.Max)
    Elem->Value.zrec.Val = Elem->Value.zrec.Max;

  Elem->Value.zrec.Default = Elem->Value.zrec.Val;

  Elem->Value.zrec.FPtr = NULL;

  return Elem;
}

dual DLGInsertScale (DIALOG *Dlg, int BoxID, char *Name,
		                  double Min, double Max, double Value,
		                  int Decimals, int Align)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			      DLG_MakeScale(-1, -1, Name, 
					    Min, Max, Value, Decimals, Align));
}

dual DLGPlaceScale (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
		                 double Min, double Max, double Value,
		                 int Decimals, int Align)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeScale(x, y, Name, 
				       Min, Max, Value, Decimals, Align));
}

DLGELEM *DLGMakeScale (char *Name, double Min, double Max, double Value,
		                   int Decimals, int Align)
{
  return DLG_MakeScale (-1, -1, Name, Min, Max, Value, Decimals, Align);
}



/* *** BAR *** */
static DLGELEM *DLG_MakeBar (int x, int y, char *Title, 
			     double Min, double Max, double Value, 
			     double Step, int Align)
{
  double   swap;
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type            = DlgBAR;
  Elem->PosX            = x;
  Elem->PosY            = y;
  Elem->LabelPos        = DlgWEST;
  Elem->Value.brec.Min  = Min;
  Elem->Value.brec.Max  = Max;
  Elem->Value.brec.Val  = Value;
  Elem->Value.brec.Step = Step;
  Elem->Value.brec.Side = Align;

  if ((Elem->Value.brec.reverse = 
       (Elem->Value.brec.Min > Elem->Value.brec.Max)))
  {
    swap                 = Elem->Value.brec.Min;
    Elem->Value.brec.Min = Elem->Value.brec.Max;
    Elem->Value.brec.Max = swap;
  }

  if (Elem->Value.brec.Val < Elem->Value.brec.Min)
    Elem->Value.brec.Val = Elem->Value.brec.Min;
  elsif (Elem->Value.brec.Val > Elem->Value.brec.Max)
    Elem->Value.brec.Val = Elem->Value.brec.Max;
 
  Elem->Value.brec.Default = Elem->Value.brec.Val;

  Elem->Value.brec.FPtr = NULL;

  return Elem;
}

dual DLGInsertBar (DIALOG *Dlg, int BoxID, char *Name,
		                double Min, double Max, double Value,
		                double Step, dual Align)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			      DLG_MakeBar(-1, -1, Name, 
					  Min, Max, Value, Step, Align));
}

dual DLGPlaceBar (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
		               double Min, double Max, double Value,
		               double Step, dual Align)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			      DLG_MakeBar(x, y, Name, 
					  Min, Max, Value, Step, Align));
}

DLGELEM *DLGMakeBar (char *Name, double Min, double Max, double Value,
		                 double Step, dual Align)
{
  return DLG_MakeBar (-1, -1, Name, Min, Max, Value, Step, Align);
}


/* *** UPDN *** */
static DLGELEM *DLG_MakeUpDn (int x, int y, char *Title, 
			      double Min, double Max, double Value, 
			      double Step)
{
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type            = DlgUPDN;
  Elem->PosX            = x;
  Elem->PosY            = y;
  Elem->LabelPos        = DlgWEST;
  Elem->Value.urec.Min  = Min;
  Elem->Value.urec.Max  = Max;
  Elem->Value.urec.Val  = Value;
  Elem->Value.urec.Step = Step;

  if (Elem->Value.urec.Min < Elem->Value.urec.Max)
    Elem->Value.urec.Val = 
      DLG_UPDN_VAL (Elem->Value.urec.Min,  Elem->Value.urec.Max,
		    Elem->Value.urec.Step, Elem->Value.urec.Val);
  Elem->Value.urec.Decimals = 
    MAX (DLG_GetDecimals(Elem->Value.urec.Min),
	 (MAX(DLG_GetDecimals(Elem->Value.urec.Max),
	      MAX(DLG_GetDecimals(Elem->Value.urec.Val),
		  DLG_GetDecimals(Elem->Value.urec.Step)))));

  Elem->Value.urec.Length = MAX (DLG_GetDigits(Elem->Value.urec.Min),
				 DLG_GetDigits(Elem->Value.urec.Max)) + 
    Elem->Value.urec.Decimals + 2;
  if (Elem->Value.urec.Min >= Elem->Value.urec.Max)
    Elem->Value.urec.Length += 3;
  Elem->Value.urec.Default = Elem->Value.urec.Val;

  Elem->Value.urec.FPtr  = NULL;

  return Elem;
}

dual DLGInsertUpDn (DIALOG *Dlg, int BoxID, char *Name,
		                 double Min, double Max, double Value, 
		                 double Step)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeUpDn(-1, -1, Name, Min, Max, Value, Step));
}

dual DLGPlaceUpDn (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
		                   double Min, double Max, double Value, 
		                   double Step)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeUpDn(x, y, Name, Min, Max, Value, Step));
}

DLGELEM *DLGMakeUpDn (char *Name, double Min, double Max, double Value, 
		                   double Step)
{
  return DLG_MakeUpDn (-1, -1, Name, Min, Max, Value, Step);
}



/* *** LIST *** */
static DLGELEM *DLG_MakeList (int x, int y, char *Title, 
			      int Lines, int Number, va_list Arguments)
{
  int       count;
  char    **ItemList = NULL;
  DLGELEM  *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type              = DlgLIST;
  Elem->PosX              = x;
  Elem->PosY              = y;
  Elem->LabelPos          = DlgNORTH_WEST;
  Elem->ValuePos          = DlgWEST;
  Elem->Value.lrec.Lines  = Lines;
  Elem->Value.lrec.Number = Number;

  if (Elem->Value.lrec.Number > 0)          /* single parameters */
  {
    Elem->Value.lrec.Item = 
      (ITEMTEXT *)malloc (sizeof(ITEMTEXT) * Elem->Value.lrec.Number);
    for (count = 0; count < Elem->Value.lrec.Number; ++count)
      DLG_strncp (Elem->Value.lrec.Item[count], 
		  va_arg(Arguments, char*), DLG_LLEN);
  }
  elsif (Elem->Value.lrec.Number == 0)
    Elem->Value.lrec.Item = NULL;
  else                                      /* list of entries */
  {
    Elem->Value.lrec.Number *= -1;
    ItemList = va_arg (Arguments, char**);
    Elem->Value.lrec.Item =
      (ITEMTEXT *)malloc (sizeof(ITEMTEXT) * Elem->Value.lrec.Number);
    for (count = 0; count < Elem->Value.lrec.Number; ++count)
      DLG_strncp (Elem->Value.lrec.Item[count], ItemList[count], DLG_LLEN);
  }

  Elem->Value.lrec.Result = va_arg (Arguments, int);
  if (Elem->Value.lrec.Result >= Elem->Value.lrec.Number)
  {
    Elem->Value.lrec.Result = 0;
    PRO1 ("WARNING: Default is out of range: %d\n", Elem->Value.lrec.Result)
  }
	
  Elem->Value.lrec.Default = Elem->Value.lrec.Result;
  Elem->Value.lrec.FPtr = NULL;

  return Elem;
}

dual DLGInsertList (DIALOG *Dlg, int BoxID, char *Name, 
		                 int Lines, int Number, ...)
{
  va_list  Arguments;
  DLGELEM *Elem;

  va_start (Arguments, Number);
  Elem = DLG_MakeList (-1, -1, Name, Lines, Number, Arguments);
  va_end (Arguments);

  return DLG_InsertElem (Dlg, DLGBox(BoxID), Elem);
}

dual DLGPlaceList (DIALOG *Dlg, int BoxID, int x, int y, char *Name, 
		                   int Lines, int Number, ...)
{
  va_list  Arguments;
  DLGELEM *Elem;

  va_start (Arguments, Number);
  Elem = DLG_MakeList (x, y, Name, Lines, Number, Arguments);
  va_end (Arguments);

  return DLG_InsertElem (Dlg, DLGBox(BoxID), Elem);
}

DLGELEM *DLGMakeList (char *Name, int Lines, int Number, ...)
{
  va_list  Arguments;
  DLGELEM *Elem;

  va_start (Arguments, Number);
  Elem = DLG_MakeList (-1, -1, Name, Lines, Number, Arguments);
  va_end (Arguments);

  return Elem;
}


/* *** OMENU *** */
static DLGELEM *DLG_MakeOMenu (int x, int y, char *Title, int Number, 
			       va_list Arguments)
{
  int       count;
  char    **ItemList = NULL;
  DLGELEM  *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type              = DlgOMENU;
  Elem->PosX              = x;
  Elem->PosY              = y;
  Elem->LabelPos          = DlgWEST;
  Elem->Value.orec.Number = Number;

  if (Elem->Value.orec.Number > 0)
  {
    Elem->Value.orec.Item = 
      (ITEMREC*)malloc (sizeof(ITEMREC) * Elem->Value.orec.Number);
    for (count = 0; count < Elem->Value.orec.Number; ++count)
    {
      DLG_strncp (Elem->Value.orec.Item[count].Label, 
		  va_arg(Arguments, char*), DLG_LLEN);
      Elem->Value.orec.Item[count].Index    = count;
      Elem->Value.orec.Item[count].Parent   = Elem;
      Elem->Value.orec.Item[count].WidgetID = NULL;
    }
  }
  elsif (Elem->Value.orec.Number == 0)
    Elem->Value.orec.Item = NULL;
  else
  {
    Elem->Value.orec.Number *= -1;
    ItemList = va_arg (Arguments, char**);
    Elem->Value.orec.Item =
      (ITEMREC*)malloc (sizeof(ITEMREC) * Elem->Value.orec.Number);
    for (count = 0; count < Elem->Value.orec.Number; ++count)
    {
      DLG_strncp (Elem->Value.orec.Item[count].Label,
		  ItemList[count], DLG_LLEN);
      Elem->Value.orec.Item[count].Index    = count;
      Elem->Value.orec.Item[count].Parent   = Elem;
      Elem->Value.orec.Item[count].WidgetID = NULL;
    }
  }

  if ((Elem->Value.orec.Result = 
       abs(va_arg(Arguments, int))) > Elem->Value.orec.Number)
  {
    Elem->Value.orec.Result = 0;
    PRO1 ("WARNING: Default is out of range: %d\n", Elem->Value.orec.Result)
  }

  Elem->Value.orec.Default = Elem->Value.orec.Result;
  Elem->Value.orec.FPtr = NULL;

  return Elem;
}

dual DLGInsertOMenu (DIALOG *Dlg, int BoxID, char *Name, int Number, ...)
{
  va_list  Arguments;
  DLGELEM *Elem;

  va_start (Arguments, Number);
  Elem = DLG_MakeOMenu(-1, -1, Name, Number, Arguments);
  va_end (Arguments);

  return DLG_InsertElem (Dlg, DLGBox(BoxID), Elem);
}

dual DLGPlaceOMenu (DIALOG *Dlg, int BoxID, int x, int y, char *Name, 
		                    int Number, ...)
{
  va_list  Arguments;
  DLGELEM *Elem;

  va_start (Arguments, Number);
  Elem = DLG_MakeOMenu(x, y, Name, Number, Arguments);
  va_end (Arguments);
  
  return DLG_InsertElem (Dlg, DLGBox(BoxID), Elem);
}

DLGELEM *DLGMakeOMenu (char *Name, int Number, ...)
{
  va_list  Arguments;
  DLGELEM *Elem;

  va_start (Arguments, Number);
  Elem = DLG_MakeOMenu (-1, -1, Name, Number, Arguments);
  va_end (Arguments);

  return Elem;
}


/* *** TOGGLEBOX *** */
static DLGELEM *DLG_MakeTBox (int x, int y, char *Title, 
			      int RowCols, int Number, va_list Arguments)
{
  int      count;
  TOGGLE  *ToggleList = NULL;
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type              = DlgTBOX;
  Elem->PosX              = x;
  Elem->PosY              = y;
  Elem->LabelPos          = DlgNORTH_WEST;
  Elem->ValuePos          = DlgWEST;
  Elem->Value.trec.RowCol = RowCols;
  Elem->Value.trec.Number = Number;

  if (Elem->Value.trec.Number >= 0)
  {
    Elem->Value.trec.Toggle = 
      (SWITCHREC*)malloc (sizeof(SWITCHREC) * Elem->Value.trec.Number);
    for (count = 0; count < Elem->Value.trec.Number; ++count)
    {
      DLG_strncp (Elem->Value.trec.Toggle[count].Item,
		  va_arg(Arguments, char*), DLG_LLEN);
      Elem->Value.trec.Toggle[count].Value  = va_arg (Arguments, dual);
      Elem->Value.trec.Toggle[count].Index  = count;
      Elem->Value.trec.Toggle[count].Parent = Elem;
    }
  }
  else
  {
    Elem->Value.trec.Number *= -1;
    ToggleList = va_arg (Arguments, TOGGLE*);
    Elem->Value.trec.Toggle = 
      (SWITCHREC*)malloc (sizeof(SWITCHREC)*Elem->Value.trec.Number);
    for (count = 0; count < Elem->Value.trec.Number; ++count)
    {
      DLG_strncp (Elem->Value.trec.Toggle[count].Item,
		  ToggleList[count].Item, DLG_LLEN);
      Elem->Value.trec.Toggle[count].Value  = ToggleList[count].Value;
      Elem->Value.trec.Toggle[count].Index  = count;
      Elem->Value.trec.Toggle[count].Parent = Elem;
    }
  } 
  Elem->Value.trec.Result = NULL;
  Elem->Value.trec.Default = 
    (dual *)malloc (sizeof(dual) * abs(Elem->Value.trec.Number));
  for (count = 0; count < Elem->Value.trec.Number; ++count)
    Elem->Value.trec.Default[count] = Elem->Value.trec.Toggle[count].Value;
  Elem->Value.trec.FPtr = NULL;

  return Elem;
}

dual DLGInsertTBox (DIALOG *Dlg, int BoxID,   char *Name,
		                 int RowCols, int   Number, ...)
{
  va_list  Arguments;
  DLGELEM *Elem;

  va_start (Arguments, Number);
  Elem = DLG_MakeTBox (-1, -1, Name, RowCols, Number, Arguments);
  va_end (Arguments);

  return DLG_InsertElem (Dlg, DLGBox(BoxID), Elem);
}

dual DLGPlaceTBox (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
		                int RowCols, int Number, ...)
{
  va_list  Arguments;
  DLGELEM *Elem;

  va_start (Arguments, Number);
  Elem = DLG_MakeTBox (x, y, Name, RowCols, Number, Arguments);
  va_end (Arguments);

  return DLG_InsertElem (Dlg, DLGBox(BoxID), Elem);
}

DLGELEM *DLGMakeTBox (char *Name, int RowCols, int Number, ...)
{
  va_list  Arguments;
  DLGELEM *Elem;

  va_start (Arguments, Number);
  Elem = DLG_MakeTBox (-1, -1, Name, RowCols, Number, Arguments);
  va_end (Arguments);

  return Elem;
}


/* *** RADIOBOX *** */
static DLGELEM *DLG_MakeRBox (int x, int y, char *Title, 
			      int RowCols, int Number, va_list Arguments)
{
  int       count, Default;
  char    **ItemList = NULL;
  DLGELEM  *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type              = DlgRBOX;
  Elem->PosX              = x;
  Elem->PosY              = y;
  Elem->LabelPos          = DlgNORTH_WEST;
  Elem->LabelPos          = DlgWEST;
  Elem->Value.trec.RowCol = RowCols;
  Elem->Value.trec.Number = Number;

  if (Elem->Value.trec.Number > 0)
  {
    Elem->Value.trec.Toggle =
      (SWITCHREC*)malloc (sizeof(SWITCHREC) * Elem->Value.trec.Number);
    for (count = 0; count < Elem->Value.trec.Number; ++count)
    {
      DLG_strncp (Elem->Value.trec.Toggle[count].Item,
		  va_arg(Arguments, char*), DLG_LLEN);
      Elem->Value.trec.Toggle[count].Value  = FALSE;
      Elem->Value.trec.Toggle[count].Index  = count;
      Elem->Value.trec.Toggle[count].Parent = Elem;
    }
  }
  elsif (Elem->Value.trec.Number == 0)
    Elem->Value.trec.Toggle = NULL;
  else
  {
    Elem->Value.trec.Number *= -1;
    ItemList = va_arg (Arguments, char**);
    Elem->Value.trec.Toggle =
      (SWITCHREC*)malloc (sizeof(SWITCHREC) * Elem->Value.trec.Number); 
    for (count = 0; count < Elem->Value.trec.Number; ++count)
    {
      DLG_strncp (Elem->Value.trec.Toggle[count].Item, 
		  ItemList[count], DLG_LLEN);
      Elem->Value.trec.Toggle[count].Value  = FALSE;
      Elem->Value.trec.Toggle[count].Index  = count;
      Elem->Value.trec.Toggle[count].Parent = Elem;
    }
  }
  if ((Default = abs(va_arg(Arguments, int))) > Elem->Value.trec.Number)
  {
    Default = 0;
    PRO1 ("WARNING: Default is out of range: %d\n", Default)
  }
  Elem->Value.trec.Toggle[Default].Value = TRUE;
  Elem->Value.trec.Result = NULL;
  Elem->Value.trec.Default = 
    (dual *)malloc (sizeof(dual) * abs(Elem->Value.trec.Number));
  for (count = 0; count < Elem->Value.trec.Number; ++count)
    Elem->Value.trec.Default[count] = 
      Elem->Value.trec.Toggle[count].Value;
  Elem->Value.trec.FPtr   = NULL;

  return Elem;
}

dual DLGInsertRBox (DIALOG *Dlg, int BoxID,   char *Name,
		                 int RowCols, int   Number, ...)
{
  va_list  Arguments;
  DLGELEM *Elem;

  va_start (Arguments, Number);
  Elem = DLG_MakeRBox (-1, -1, Name, RowCols, Number, Arguments);
  va_end (Arguments);

  return DLG_InsertElem (Dlg, DLGBox(BoxID), Elem);
}

dual DLGPlaceRBox (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
		                int RowCols, int Number, ...)
{
  va_list  Arguments;
  DLGELEM *Elem;

  va_start (Arguments, Number);
  Elem = DLG_MakeRBox (x, y, Name, RowCols, Number, Arguments);
  va_end (Arguments);

  return DLG_InsertElem (Dlg, DLGBox(BoxID), Elem);
}

DLGELEM *DLGMakeRBox (char *Name, int RowCols, int Number, ...)
{
  va_list  Arguments;
  DLGELEM *Elem;

  va_start (Arguments, Number);
  Elem = DLG_MakeRBox (-1, -1, Name, RowCols, Number, Arguments);
  va_end (Arguments);

  return Elem;
}


/* *** PROC *** */
static DLGELEM *DLG_MakeProc (int x, int y, char *Title, DLGFUNCTION Function)
{
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type            = DlgPROC;
  Elem->PosX            = x;
  Elem->PosY            = y;
  Elem->LabelPos        = DlgNOWHERE;
  Elem->Value.crec.FPtr = Function;

  return Elem;
}

dual DLGInsertProc (DIALOG *Dlg, int BoxID, char *Name, 
		                    DLGFUNCTION Function)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			      DLG_MakeProc(-1, -1, Name, Function));
}

dual DLGPlaceProc (DIALOG *Dlg, int BoxID, int x, int y, char *Name, 
		                    DLGFUNCTION Function)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			      DLG_MakeProc(x, y, Name, Function));
}

DLGELEM *DLGMakeProc (char *Name, DLGFUNCTION Function)
{
  return DLG_MakeProc (-1, -1, Name, Function);
}


/* *** ARROW *** */
static DLGELEM *DLG_MakeArrow (int x, int y, char *Title, 
			       int Dir, DLGFUNCTION Function)
{
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type            = DlgARROW;
  Elem->PosX            = x;
  Elem->PosY            = y;
  Elem->LabelPos        = DlgNOWHERE;
  Elem->ValuePos        = DlgCENTER;
  Elem->Value.arec.Dir  = Dir;
  Elem->Value.arec.FPtr = Function;

  return Elem;
}

dual DLGInsertArrow (DIALOG *Dlg, int BoxID, char *Name,
		                  int Dir, DLGFUNCTION Function)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			      DLG_MakeArrow(-1, -1, Name, Dir, Function));
}

dual DLGPlaceArrow (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
		                 int Dir, DLGFUNCTION Function)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			      DLG_MakeArrow(x, y, Name, Dir, Function));
}

DLGELEM *DLGMakeArrow (char *Name, int Dir, DLGFUNCTION Function)
{
  return DLG_MakeArrow (-1, -1, Name, Dir, Function);
}


/* *** PIC(-BUTTON) *** */
DLGELEM *DLG_MakePic (int x, int y, char *Title, int Format,
		      int SizeX, int SizeY, void *Data, DLGFUNCTION Function)
{
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type                  = DlgPIC;
  Elem->PosX                  = x;
  Elem->PosY                  = y;
  Elem->Value.prec.Format     = Format;
  Elem->Value.prec.SizeX      = SizeX;
  Elem->Value.prec.SizeY      = SizeY;
  Elem->Value.prec.PicName[0] = '\0';
  Elem->Value.prec.fg         = 0;
  Elem->Value.prec.bg         = 0;
  Elem->Value.prec.iData      = NULL;

  if (Elem->Value.prec.Format < DlgUNKNOWN || 
      Elem->Value.prec.Format > DlgXPM)
    PRO_RETURN ("Invalid Image-Type", FALSE)

    if (Elem->Value.prec.SizeX == 0 || Elem->Value.prec.SizeY == 0)
    {
      DLG_strncp (Elem->Value.prec.PicName, (char*)Data, PATH_MAX);
      Elem->Value.prec.PicName[PATH_MAX] = '\0';
    }
    else
      switch (Elem->Value.prec.Format)
      {
        case DlgXBM :
	  Elem->Value.prec.Data = 
	    (void*)TOOLCopyXBM ((byte*)Data, Elem->Value.prec.SizeX, 
				             Elem->Value.prec.SizeY);
	  break;

        case DlgXPM :
	  Elem->Value.prec.Data = (void*)TOOLCopyXPM ((char**)Data);
	  break;

	default : 
	  Elem->Value.prec.Data = NULL;
      }

  Elem->Value.prec.FPtr = Function;
  if (Elem->Value.prec.FPtr == NULL)
  {
    Elem->LabelPos = DlgNOWHERE;
    Elem->ValuePos = DlgCENTER;
  }
  else
  {
    Elem->LabelPos = DlgCENTER;
    Elem->ValuePos = DlgNORTH;
  }

  return Elem;
}

dual DLGInsertPic (DIALOG *Dlg, int BoxID, char *Name, int Format,
	                        int SizeX, int SizeY, void *Data,
	                        DLGFUNCTION Function)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakePic(-1, -1, Name, 
				     Format, SizeX, SizeY, Data, Function));
}

dual DLGPlacePic (DIALOG *Dlg, int BoxID, int x, int y, char *Name, 
	                       int Format, int SizeX, int SizeY, void *Data,
	                       DLGFUNCTION Function)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakePic(x, y, Name, 
				     Format, SizeX, SizeY, Data, Function));
}

DLGELEM *DLGMakePic (char *Name, int Format, int SizeX, int SizeY, void *Data,
	                         DLGFUNCTION Function)
{
  return DLG_MakePic (-1, -1, Name, Format, SizeX, SizeY, Data, Function);
}


/* *** IMAGE(-BUTTON) *** */
DLGELEM *DLG_MakeImage (int x, int y, char *Title, 
		        XImage *Image, DLGFUNCTION Function)
{
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type             = DlgIMAGE;
  Elem->PosX             = x;
  Elem->PosY             = y;
  Elem->Value.xrec.Image = TOOLCopyImage (Image);
  Elem->Value.xrec.FPtr  = Function;

  return Elem;
}

dual DLGInsertImage (DIALOG *Dlg, int BoxID, char *Name, XImage *Image,
		                  DLGFUNCTION Function)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeImage(-1, -1, Name, Image, Function));
}

dual DLGPlaceImage (DIALOG *Dlg, int BoxID, int x, int y, char *Name, 
		                 XImage *Image, DLGFUNCTION Function)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeImage(x, y, Name, Image, Function));
}

DLGELEM *DLGMakeImage (char *Name, XImage *Image, DLGFUNCTION Function)
{
  return DLG_MakeImage (-1, -1, Name, Image, Function);
}


/* *** FILESELECTOR *** */
DLGELEM *DLG_MakeSel (int x, int y, char *Title, int Type,
		      char *FileMask, int Length, char *Default)
{
  char FileSpec[PATH_MAX+1] = "";
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type             = Type;
  Elem->PosX             = x;
  Elem->PosY             = y;
  DLG_strncp (Elem->Value.grec.Mask, FileMask, PATH_MAX);
  Elem->Value.grec.Length = Length;
  DLG_strncp (Elem->Value.grec.Default, Default, PATH_MAX);

  /* When length is zero, the dialog will be directly inserted */
  if (Elem->Value.grec.Length == 0)
  {
    Elem->LabelPos = DlgNOWHERE;
    Elem->ValuePos = DlgCENTER;
  }
  else
    Elem->LabelPos = DlgWEST;

  TOOLParseFileName (Elem->Value.grec.Default, FileSpec, PATH_MAX);
  DLG_strncp        (Elem->Value.grec.Default, FileSpec, PATH_MAX);
  DLG_strncp        (Elem->Value.grec.Result,  FileSpec, PATH_MAX);

  TOOLParseFileName (Elem->Value.grec.Mask, FileSpec, PATH_MAX);
  DLG_strncp        (Elem->Value.grec.Mask, FileSpec, PATH_MAX);

  Elem->Value.grec.Button = 0;
  Elem->Value.grec.Files  = (Type == DlgFSEL);
  Elem->Value.grec.Open   = FALSE;
  Elem->Value.grec.FPtr   = NULL;

  return Elem;
}

dual DLGInsertFileSel (DIALOG *Dlg, int BoxID,  char *Name, 
		                    char *FileMask, int Length, char *Default)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			      DLG_MakeSel(-1, -1, Name, DlgFSEL,
					  FileMask, Length, Default));
}

dual DLGPlaceFileSel (DIALOG *Dlg, int BoxID, int x, int y, char *Name, 
		                   char *FileMask, int Length, char *Default)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeSel(x, y, Name, DlgFSEL,
				     FileMask, Length, Default));
}

DLGELEM *DLGMakeFileSel (char *Name, char *FileMask, int Length, char *Default)
{
  return DLG_MakeSel (-1, -1, Name, DlgFSEL, FileMask, Length, Default);
}


/* *** FILESELECTOR *** */
dual DLGInsertDirSel (DIALOG *Dlg, int BoxID,  char *Name, 
		                   char *DirMask, int Length, char *Default)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeSel(-1, -1, Name, DlgDIRSEL,
				     DirMask, Length, Default));
}

dual DLGPlaceDirSel (DIALOG *Dlg, int BoxID, int x, int y, char *Name, 
		                   char *DirMask, int Length, char *Default)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeSel(x, y, Name, DlgDIRSEL,
				     DirMask, Length, Default));

}

DLGELEM *DLGMakeDirSel (char *Name, char *DirMask, int Length, char *Default)
{
  return DLG_MakeSel (-1, -1, Name, DlgDIRSEL, DirMask, Length, Default);
}



/* *** DIAL *** */
static DLGELEM *DLG_MakeDial (int x, int y, char *Title, 
			      double Min, double Max, double Value,
			      byte Digits, byte Decimals)
{
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type                = DlgDIAL;
  Elem->PosX                = x;
  Elem->PosY                = y;
  Elem->LabelPos            = DlgNOWHERE;
  Elem->ValuePos            = DlgCENTER;
  Elem->Value.erec.Min      = Min;
  Elem->Value.erec.Max      = Max;
  Elem->Value.erec.Val      = Value;
  Elem->Value.erec.Digits   = Digits;
  Elem->Value.erec.Decimals = Decimals;
  Elem->Value.erec.Default  = Value;
  Elem->Value.erec.FPtr     = NULL;

  return Elem;
}

dual DLGInsertDial (DIALOG *Dlg, int BoxID,  char *Name, 
		                 double Min, double Max, double Value, 
		                 byte Digits, byte Decimals)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			      DLG_MakeDial(-1, -1, Name, 
					   Min, Max, Value, Digits, Decimals));
}

dual DLGPlaceDial (DIALOG *Dlg, int BoxID, int x, int y, char *Name, 
		                double Min, double Max, double Value, 
		                byte Digits, byte Decimals)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeDial(x, y, Name, 
				      Min, Max, Value, Digits, Decimals));
}

DLGELEM *DLGMakeDial (char *Name, double Min, double Max, double Value, 
		                   byte Digits, byte Decimals)
{
  return DLG_MakeDial (-1, -1, Name, Min, Max, Value, Digits, Decimals);
}



/* *** 7-SEGMENT-DISPLAY *** */
static DLGELEM *DLG_MakeSegment (int x, int y, char *Title, 
				 double Value, byte Digits, byte Decimals)
{
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type                = DlgSEGM;
  Elem->PosX                = x;
  Elem->PosY                = y;
  Elem->LabelPos            = DlgNOWHERE;
  Elem->ValuePos            = DlgCENTER;
  Elem->Value.hrec.Val      = Value;
  Elem->Value.hrec.Digits   = Digits;
  Elem->Value.hrec.Decimals = Decimals;
  Elem->Value.hrec.Default  = Elem->Value.hrec.Val;
  Elem->Value.hrec.FPtr     = NULL;

  return Elem;
}

dual DLGInsertSegment (DIALOG *Dlg, int BoxID, char *Name, double Value, 
		                       byte Digits, byte Decimals)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			      DLG_MakeSegment(-1, -1, Name, 
					      Value, Digits, Decimals));
}

dual DLGPlaceSegment (DIALOG *Dlg, int BoxID, int x, int y, char *Name, 
		                      double Value, byte Digits, byte Decimals)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeSegment(x, y, Name, Value, Digits, Decimals));
}

DLGELEM *DLGMakeSegment (char *Name, double Value, byte Digits, byte Decimals)
{
  return DLG_MakeSegment (-1, -1, Name, Value, Digits, Decimals);
}



/* *** USER-DEFINED-WIDGET *** */
static DLGELEM *DLG_MakeUser (int x, int y, char *Title, 
			      WIDGETFUNCTION Function, void *Param)
{
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type             = DlgUSER;
  Elem->PosX             = x;
  Elem->PosY             = y;
  Elem->LabelPos         = DlgNOWHERE;
  Elem->ValuePos         = DlgWEST;
  Elem->Value.wrec.FPtr  = Function;
  Elem->Value.wrec.Param = Param;

  return Elem;
}

dual DLGInsertWidget (DIALOG *Dlg, int BoxID, 
		                   WIDGETFUNCTION Function, void *Param)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeUser(-1, -1, NULL, Function, Param));
}

dual DLGPlaceWidget (DIALOG *Dlg, int BoxID, int x, int y,
		                  WIDGETFUNCTION Function, void *Param)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeUser(x, y, NULL, Function, Param));
}



#if (XmVERSION>=2)
/* *** NOTEBOOK *** */
static DLGELEM *DLG_MakeNote (int x, int y, char *Title, int Style)
{
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type               = DlgNOTE;
  Elem->PosX               = x;
  Elem->PosY               = y;
  Elem->LabelPos           = DlgNOWHERE;
  Elem->Value.nrec.Style   = Style;
  Elem->Value.nrec.ActPage = 1;
  Elem->Value.nrec.Number  = 0;
  Elem->Value.nrec.Page    = NULL;
  Elem->Value.nrec.Default = Elem->Value.nrec.ActPage;
  Elem->Value.nrec.FPtr    = NULL;

  return Elem;
}

dual DLGInsertNote (DIALOG *Dlg, int BoxID, char *Name, int Style)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID), 
			      DLG_MakeNote(-1, -1, Name, Style));
}

dual DLGPlaceNote (DIALOG *Dlg, int BoxID, int x, int y, char *Name, int Style)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID), DLG_MakeNote(x, y, Name, Style));
}

dual DLGAddPage (DIALOG *Dlg, int BoxID, int Type, char *Label, int SubBox)
{
  NOTEPAGE *Page, *Ptr;
  DLGBOX   *Box = NULL;
  DLGELEM  *Elem;

  PRO_START ("")

  PRO1 ("\\[%d]...\n", BoxID)

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgNOTE)) == NULL)
    PRO_RETURN ("Invalid BoxID or Type", FALSE)

  Page = (NOTEPAGE *)malloc (sizeof(NOTEPAGE));
  Page->Type = Type;
  strncpy (Page->Label, Label, DLG_LLEN);
  Page->Label[DLG_LLEN] = '\0';
  Page->Box = (DLGBOX*)calloc (1, sizeof(DLGBOX));
  Page->Next = NULL;

  Box = Page->Box;
  Box->BoxID    = SubBox;
  Box->Title[0] = '\0';
  Box->TitlePos = DlgMIDDLE;
  Box->AlignX   = DlgLEFT;
  Box->AlignY   = DlgUP;
  Box->Parent   = NULL;
  Box->Main     = NULL;
  Box->Box      = NULL;
  Box->Border   = DlgNO_LINE[0];
  Box->active   = TRUE;
  Box->Sep[0]   = '\0';
  Box->Dir      = BOXNONE;
  Box->Number   = 0;
  Box->Item     = NULL;
  Box->Cols     = 0;
  Box->Rows     = 0;
  Box->List     = NULL;

  if ((Ptr = Elem->Value.nrec.Page) == NULL)
    Elem->Value.nrec.Page = Page;
  else
  {
    while (Ptr->Next != NULL)
      Ptr = Ptr->Next;
    Ptr->Next = Page;
  }
  (Elem->Value.nrec.Number)++;

  PRO_END
  return TRUE;
}


/* *** TREE *** */
static DLGELEM *DLG_MakeTree (int x, int y, char *Title,
			      Pixmap ClosePic, Pixmap OpenPic, int Lines)
{
  DLGELEM *Elem;

  if ((Elem = DLG_NewElem(Title)) == NULL)
    return NULL;

  Elem->Type                = DlgTREE;
  Elem->PosX                = x;
  Elem->PosY                = y;
  Elem->LabelPos            = DlgNOWHERE;
  Elem->Value.krec.Shadow   = 0;
  Elem->Value.krec.ClosePic = ClosePic;
  Elem->Value.krec.OpenPic  = OpenPic;
  Elem->Value.krec.Lines    = Lines;
  Elem->Value.krec.Header   = NULL;
  Elem->Value.krec.Data     = NULL;
  Elem->Value.krec.Reason   = 0;
  Elem->Value.krec.Result   = NULL;
  Elem->Value.krec.Default  = NULL;
  Elem->Value.krec.FPtr     = NULL;

  return Elem;
}

dual DLGInsertTree (DIALOG *Dlg, int BoxID, char *Name, 
		                 Pixmap ClosePic, Pixmap OpenPic, int Lines)
{
  return DLG_InsertElem (Dlg, DLGBox(BoxID),
			 DLG_MakeTree(-1, -1, Name, ClosePic, OpenPic, Lines));
}

static TREEITEM *DLG_FindTreeItem (TREEITEM *List, int ID)
{
  TREEITEM *Ptr, *Item;

  Ptr = List;
  while (Ptr != NULL && Ptr->ID != ID)
  {
    if (Ptr->Sub != NULL && (Item = DLG_FindTreeItem(Ptr->Sub, ID)) != NULL)
      return Item;

    Ptr = Ptr->Next; 
  }

  return Ptr;
}


void DLG_PrintTree (TREEITEM *Item)
{
  TREEITEM *Ptr;

  Ptr = Item;
  while (Ptr != NULL)
  {
    printf ("%d: %s  [%c] [%c]\n", 
	    Ptr->ID, Ptr->Label,
	    (Ptr->Sub==NULL)?'-':'+', (Ptr->Next==NULL)?'-':'+');
    if (Ptr->Sub != NULL)
    {
      printf ("->\n");
      DLG_PrintTree (Ptr->Sub);
    }
    Ptr = Ptr->Next;
  }
}


dual DLGAddTreeItem (DIALOG *Dlg, int BoxID, const char *Name, 
		                  int ID, int Parent,
		                  dual State, void *UserData)
{
  return DLGAddTreeIcon (Dlg, BoxID, Name, NULL, ID, Parent, State, UserData);
}


dual DLGAddTreeIcon (DIALOG *Dlg, int BoxID, const char *Name, void *Data,
		                  int ID, int Parent, 
		                  dual State, void *UserData)
{
  TREEITEM *Item, *Ptr;
  DLGELEM  *Elem;

  PRO_START ("")

  PRO2 ("\\[%d] #%d...", BoxID, ID)

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgTREE)) == NULL)
    PRO_RETURN1 ("Invalid BoxID or Type: %d", BoxID, FALSE)

  Item = (TREEITEM*)malloc (sizeof(TREEITEM));
  strncpy (Item->Label, Name, DLG_LLEN);
  Item->Label[DLG_LLEN] = '\0';
  Item->ID       = ID;
  Item->Icon     = Data;
  Item->State    = State;
  Item->UserData = UserData;
  Item->gadget   = NULL;
  Item->Sub      = NULL;
  Item->Next     = NULL;

  if ((Ptr = Elem->Value.krec.Data) == NULL)
    Elem->Value.krec.Data = Item;
  elsif ((Ptr = DLG_FindTreeItem(Ptr, Parent)) == NULL)
  {
    Ptr = Elem->Value.krec.Data;
    while (Ptr->Next != NULL)
      Ptr = Ptr->Next;
    Ptr->Next = Item;
  }
  elsif (Ptr->Sub == NULL)
    Ptr->Sub = Item;
  else
  {
    Ptr = Ptr->Sub;
    while (Ptr->Next != NULL)
      Ptr = Ptr->Next;
    Ptr->Next = Item;
  }

  PRO_END
  return TRUE;
}
#endif


#if 0
dual DLGInsertSeparator (DIALOG *Dlg, int BoxID, int Style,
			 int x1, int y1, int length, dual Direction)
{
  DLGELEM *Elem;

  PRO0 ("Insert USER-WIDGET...\n")

  if ((Elem = DLG_CreateElem(Dlg, BoxID, DlgSEP, NULL)) == NULL)
    return FALSE;

  Elem->LabelPos = DLG_NOLABEL;
  Elem->Value.mrec.Style = Style;
  Elem->Value.mrec.x1    = x1;
  Elem->Value.mrec.y1    = y1;
  Elem->Value.mrec.x2    = (Direction==DLG_HORIZONTAL)? (x1+length):x1;
  Elem->Value.mrec.y2    = (Direction==DLG_VERTICAL)? (y1+length):y1;

  return TRUE;
}

dual DLGInsertBox (DIALOG *Dlg, int BoxID, int Style,
		                int x1, int y1, int width, int height)
{
  DLGELEM *Elem;

  PRO0 ("Insert USER-WIDGET...\n")

  if ((Elem = DLG_CreateElem(Dlg, BoxID, DlgBOX, NULL)) == NULL)
    return FALSE;

  Elem->LabelPos = DLG_NOLABEL;
  Elem->Value.mrec.Style = Style;
  Elem->Value.mrec.x1    = x1;
  Elem->Value.mrec.y1    = y1;
  Elem->Value.mrec.x2    = x1+width;
  Elem->Value.mrec.y2    = y1+height;

  return TRUE;
}

#endif



