/* ******************************** */
/* *** (c) Stefan Vietinghoff   *** */
/* *** dialog handling          *** */
/* ******************************** */
#include <math.h>

#include "dlgData.h"
#include "dlgCB.h"
#include "dlgScript.h"

#include <Xm/MwmUtil.h>

#include "dlgicon.xpm"

#if (XmVERSION>=2)
#include <Xm/TransferP.h>
#endif


/* ***** global variables: ***** */
static int n = 0;
static Arg args[NUM_ARGS];
static int XPos = 0;                    /* default dialog-position */
static int YPos = 0;


/* ******************************************************* */
/* ** Creates a new dialog, which contains only one     ** */
/* ** box with given box-ID (e.g. 0 - so to speak top-  ** */
/* ** level-box)                                        ** */
/* ** allocated memory must be freed using the function ** */
/* ** DLGFreeDialog.                                    ** */
/* ** INPUT:  BoxID         identify-number of box      ** */
/* ** OUTPUT: (DIALOG*)  Pointer to allocated mem       ** */
/* ** EXAMPLE:                                          ** */
/* **   DIALOG *Dlg = NULL;                             ** */
/* **   ...                                             ** */
/* **   Dlg = DLGNewDialog (0);                         ** */
/* ******************************************************* */
DIALOG *DLGNewDialog (int BoxID)
{
  DIALOG *Result = NULL;
  DLGBOX *Box = NULL;

  PRO_START ("")

  if (! TOOLIsInitialized())
    PRO_ERROR0 ("Dlg.-Lib has NOT been initialized", NULL)

  Result = (DIALOG*) malloc (sizeof(DLGHEAD));
  Result->Title[0] = '\0';                  /* Empty title               */
  Result->PosX     = 0;                     /* Default-position          */
  Result->PosY     = 0;
  Result->PosSet   = FALSE;                 /* Not set up to now         */
  Result->SizeX    = 0;
  Result->SizeY    = 0;
  Result->Resize   = FALSE;
  Result->active   = TRUE;
  Result->locked   = 0;
  Result->CBFct    = NULL;                  /* No callback function      */
  Result->CheckFct = NULL;
  Result->HelpFct  = NULL;
  Result->Shell    = NULL;                  /* Not created, yet          */
  Result->Form     = NULL;
  Result->ScrWin   = NULL;
  Result->Work     = NULL;
  Result->Ok       = NULL;                  /* Created on openign        */
  Result->Cancel   = NULL;

  Box           = (DLGBOX *)malloc (sizeof(DLGBOX)); 
  Box->BoxID    = BoxID;                   /* Dialog box-identifier      */
  Box->Title[0] = '\0';                    /* No heading                 */
  Box->TitlePos = DlgMIDDLE;
  Box->AlignX   = DlgLEFT;
  Box->AlignY   = DlgUP;
  Box->Parent   = NULL;
  Box->Main     = NULL;
  Box->Box      = NULL;                    /* no shell yet              */
  Box->Border   = DlgNO_LINE[0];           /* default-entries...        */
  Box->active   = TRUE;
  Box->Sep[0]   = '\0';
  Box->Dir      = BOXNONE;                 /* unspecified box-type      */
  Box->Number   = 0;
  Box->Item     = NULL;                    /* and no entries            */
  Box->Cols     = 0;                       /* no subdivisions           */
  Box->Rows     = 0;
  Box->List     = NULL;
  Box->UserData = NULL;
  Result->Box   = Box;

  Result->Buttons.changed = FALSE;
  Result->Buttons.Number  = 2;
  Result->Buttons.Data    = (DLGBUTTON *)malloc (sizeof(DLGBUTTON)*2);
  Result->Buttons.Result  = -1;             /* Default exit-button      */
  memset (Result->Buttons.Types, 0, sizeof(int) * 3);

  Result->Buttons.Data[0].ID        = DlgOK;
  Result->Buttons.Data[0].Type      = OK_ID;
  Result->Buttons.Data[0].Number    = 0;
  strcpy (Result->Buttons.Data[0].Name, "ok");
  Result->Buttons.Data[0].Placement = DlgBUTTON_RIGHT;
  Result->Buttons.Data[0].FPtr      = NULL;
  Result->Buttons.Data[0].Dlg       = Result;
  Result->Buttons.Data[0].Button    = NULL;
  Result->Buttons.Data[0].Width     = 0;
  Result->Buttons.Data[1].ID        = DlgCANCEL;
  Result->Buttons.Data[1].Type      = CANCEL_ID;
  Result->Buttons.Data[1].Number    = 1;
  strcpy (Result->Buttons.Data[1].Name, "cancel");
  Result->Buttons.Data[1].Placement = DlgBUTTON_LEFT;
  Result->Buttons.Data[1].FPtr      = NULL;
  Result->Buttons.Data[1].Dlg       = Result;
  Result->Buttons.Data[1].Button    = NULL;
  Result->Buttons.Data[1].Width     = 0;

  PRO_END
  return Result;
}


/*----------------------------------------------------------------------
  Funktion: lscht alle in einem Dialog enthaltenen Eintrge und 
            gibt den von ihnen belegten Speicher frei.
	    Eine Prfung des Ergebnisses kann mit DLGIsEmpty erreicht 
	    werden.

  INPUT:  Dlg  Zeiger auf den Dialog, dessen Struktur gelscht und
               dessen Speicher freigegeben werden soll
  OUTPUT: ---

  Bsp.: DIALOG *Dlg;
        ...
	DLGFreeDialog (Dlg);
  ----------------------------------------------------------------------*/
void DLGFreeDialog (DIALOG *Dlg)
{
  PRO_START ("\n")

  if (Dlg == NULL)
    PRO_ERROR ("dialog is empty")

  if (! TOOLScriptRunning())
    TOOLDelExitCB ((XtCallbackProc)DLG_ExitCB, Dlg);

  PRO0 ("\\Boxes...")
  DLG_FreeBox (Dlg->Box);

  PRO0 ("\\Buttons...")
  if (Dlg->Buttons.Data != NULL)
    free (Dlg->Buttons.Data);

  PRO0 ("\\Dlg\n")
  free (Dlg);

  PRO_END
}


Widget DLGGetShell (DIALOG *Dlg)
{
  return Dlg->Shell;
}


#define BOX_OFFSET 0
#ifdef _PROTOCOL
  static char AlignStr[] = {'M','L','R','U','D','F','-'};
  static char DirStr[]   = {'H','V','C'};
#endif

/*----------------------------------------------------------------------
  Funktion: Erzeugt aus der initialisierten Dialogr-Struktur eine
            Motif-Widgetbaum-Hierarchie. Dabei werden die zuvor
	    spezifizierten Eintragungen bercksichtigt.
	    (Rahmen, Trennlinien, berschriften, Beschriftungen und
	    ihre Position zu den Werten, Sichtbarkeit,...)
	    Diese Funktion geht rekursiv durch die Dialog-Struktur

  INPUT:  Parent    bergeordnetes Widget, welches bei der nchsthhreren
                    Rekursionstiefe erzeugt wurde. (Bei Beginn von 
		    DLGMakeDialog)
	  Top,Left  oberes und linkes Nachbarwidget
	  Box       Zeiger auf die aktuelle Position in der Dialog-
	            struktur (wird vor dem rek. Aufruf weitergesetz)
	  count     Position innerhalb der Widget-Liste (fortlaufende
	            Zhlvariable)
  OUTPUT: INT       Nummer des letzten benutzten Widgets in der Liste
  ----------------------------------------------------------------------*/
Widget DLG_CreateBox (Widget Parent, Widget Top, Widget Left,
		                     DLGBOX *Box, int Dir)
{
  int      countX, countY;
  int      Part;
  char     Name[DLG_NLEN+1];
  Widget   Master, Form;
  Widget   Last, Cell;
  DLGBOX  *Ptr;
  DLGELEM *Elem = NULL;

  PRO_START ("\n")

  if (Box == NULL)
    PRO_RETURN ("Invalid box", (Dir==DlgVERTICAL)? Top : Left)

  PRO2 ("BOX: %6d Typ: %c  ", Box->BoxID, DirStr[Dir])
  PRO5 ("\\Align: %c %c  Border: '%c' Sep: '%s'  Title: %c\n", 
	AlignStr[Box->AlignX], AlignStr[Box->AlignY],
	Box->Border, Box->Sep,  (Box->Title[0] != '\0')? 'Y':'N');

  n = 0;

  if (Dir != DlgCELL)
  {
    /* Attach to top side (form if first widget in row) */
    XtSetArg (args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (args[n], XmNtopWidget,     Top);             n++;

    /* Attach to left side */
    XtSetArg (args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (args[n], XmNleftWidget,     Left);            n++;

    if ((Dir == DlgHORIZONTAL && Box->AlignY == DlgFILL) || 
	Box->AlignY == DlgDOWN)
    {
      XtSetArg (args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    }
    elsif (Dir == DlgVERTICAL && Box->AlignX == DlgFILL)
    {
      XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    }
  }

  if (Box->Item != NULL)
    strcpy (Name, Box->Item->Name);
  else
    sprintf (Name, "box_%d", Box->BoxID);

  /* Headline/Boder -> frame */
  if (Box->Title[0] != '\0' || Box->Border == DlgNO_LINE[0])
  {
    Box->Main = 
      XtCreateManagedWidget (Name, xmFormWidgetClass, Parent, args, n);
    n = 0;
  }
  else
    Box->Main = Parent;

 /* Header: */
  if (Box->Title[0] != '\0')
  {
    n = 0;
    XtSetArg (args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    if (Box->TitlePos == DlgLEFT || Box->TitlePos == DlgMIDDLE)
    {
      XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    }
    if (Box->TitlePos == DlgRIGHT || Box->TitlePos == DlgMIDDLE)
    {
      XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    }
    Top = XtCreateManagedWidget (Box->Title, xmLabelWidgetClass, 
				             Box->Main, args, n);

    n = 0;
    XtSetArg (args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (args[n], XmNtopWidget,     Top);             n++;
    if (Dir == DlgVERTICAL && Box->AlignX == DlgFILL)
    {
      XtSetArg (args[n], XmNleftAttachment,  XmATTACH_FORM); n++;
      XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    }
    if (Dir == DlgHORIZONTAL && Box->AlignY == DlgFILL)
    {
      XtSetArg (args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    }
  }

  /* Border: */
  if (Box->Border != DlgNO_LINE[0])
  {
    if (Box->Border == DlgSHADOW_IN[0])
      XtSetArg (args[n], XmNshadowType, XmSHADOW_IN);
    elsif (Box->Border == DlgSHADOW_OUT[0])
      XtSetArg (args[n], XmNshadowType, XmSHADOW_OUT);
    elsif (Box->Border == DlgETCHED_OUT[0])
      XtSetArg (args[n], XmNshadowType, XmSHADOW_ETCHED_OUT);
    else
      XtSetArg (args[n], XmNshadowType, XmSHADOW_ETCHED_IN);
    n++;
    Master = XtCreateManagedWidget ("frame", xmFrameWidgetClass, 
				             Box->Main, args, n);
    if (Box->Title[0] == '\0')
      Box->Main = Master;
    n = 0;
  }
  else
    Master = Box->Main;

  /* Create form for value vidget */
  if (Box->Title[0] != '\0' || Box->Border != DlgNO_LINE[0])
    Form = XtCreateManagedWidget (Name, xmFormWidgetClass, Master, args, n);
  else
    Form = Master;

  XtVaSetValues (Form, XmNtopOffset,    3, 
		       XmNleftOffset,   3, 
		       XmNrightOffset,  3,
		       XmNbottomOffset, 3, NULL);
  Box->Box = Box->Main;

  if (Box->Dir == BOXROWS || Box->Dir == BOXCOLS)  /* create ROWS & COLS */
  {
    Part = 1;
    Last = Form;
    for (countY = 0; countY < Box->Rows; ++countY)
      for (countX = 0; countX < Box->Cols; ++countX)
	if ((Ptr = Box->List[countX][countY]) != NULL)
	{
	  Ptr->Parent = Master;
	  if ((countX > 0 || countY > 0) && Box->Sep[0] != '\0')
	    Last = DLG_CreateSep (Form, Last, Box, Part++);
	  if (Box->Dir == BOXROWS)
	    Last = DLG_CreateBox (Form, Last, Form, Ptr, DlgVERTICAL);
	  else
	    Last = DLG_CreateBox (Form, Form, Last, Ptr, DlgHORIZONTAL);
	}
    /* if there's still a separator line behind the last element desired */
    if (Box->Sep[Part-1] != '\0')
      Last = DLG_CreateSep (Form, Last, Box, Part);

    if (! Box->active)
      XtSetSensitive (Box->Main, False);
  }
  elsif (Box->Dir == BOXGRID)                      /* create GRID */
  {
    Top = Form;
    for (countY = 0; countY < Box->Rows; ++countY)
    {
      Left = Form;
      for (countX = 0; countX < Box->Cols; ++countX)
        if ((Ptr = Box->List[countX][countY]) != NULL &&
	    (Ptr->Item != NULL || Ptr->List != NULL))
	{
	  Ptr->Parent = Master;
	  sprintf (Name, "cell_%d", Box->BoxID);
	  Cell = XtVaCreateManagedWidget (Name, xmFormWidgetClass, Form,
					  XmNtopAttachment, XmATTACH_WIDGET,
					  XmNtopWidget,     Top,
					  XmNleftWidget,    XmATTACH_WIDGET,
					  XmNleftWidget,    Left,
					  NULL);
	  Left = DLG_CreateBox (Cell, Cell, Cell, Ptr, DlgCELL);
	}
      Top = Left;
    }
  }
  elsif (Box->Dir == BOXARRAY)             /* create array */
  {
    printf ("Box-ARRAY: %d\n", Box->BoxID);
    Box->Box = XtVaCreateManagedWidget ("array", arrayWidgetClass, Box->Main,
					XmNtopAttachment,    XmATTACH_FORM,
					XmNleftAttachment,   XmATTACH_FORM,
					XmNrightAttachment,  XmATTACH_FORM,
					XmNbottomAttachment, XmATTACH_FORM,
					MNautoFill,                   True,
					MNfixedWidth,                 True,
					MNcellDistance,                  2,
					NULL);
    for (countY = 0; countY < Box->Rows; ++countY)
      for (countX = 0; countX < Box->Cols; ++countX)
        if ((Ptr = Box->List[countX][countY]) != NULL &&
	    (Ptr->Item != NULL || Ptr->List != NULL))
	{
	  Ptr->Parent = Master;
	  sprintf (Name, "array_%d", Box->BoxID);
	  Cell = XtVaCreateManagedWidget (Name, xmFormWidgetClass, Box->Box,
					  MNcolumn,      countX,
					  MNrow,         countY,
                                          MNallowResize, True,
					  /* TODO: Je nach Align-Einstellung */
#if 0
	 				  MNxAlignment,  M_ALIGN_LEFT,
	 				  MNyAlignment,  M_ALIGN_TOP,
#else
					  MNxAlignment,  M_ALIGN_FILL,
					  MNyAlignment,  M_ALIGN_CENTER,
#endif
					  MNpadX,        0,
					  MNpadY,        0,
					  NULL);
	  if ((Elem = Ptr->Item) != NULL)
	    DLG_CreateWidget (Cell, Elem, TRUE);
	  else
	    DLG_CreateBox (Cell, Cell, Cell, Ptr, DlgCELL);
	}
  }
  elsif ((Elem = Box->Item) != NULL)    /* Create BOX-ELEMENT */
    while (Elem != NULL)
    {
      DLG_CreateWidget (Form, Elem, TRUE);

      if (Elem->hidden)
	XtSetMappedWhenManaged (Form, False);
      elsif (! Elem->active)
	XtSetSensitive (Form, False);

      Elem = Elem->Next;
    }
  else
    PRO_RETURN ("No Item", Box->Main)

  PRO_END
  return Box->Main;
}


/*----------------------------------------------------------------------
  Funktion: erzeugt die Widget-Struktur des Gesamtdialog, in dem 
            die Besttigungsbutton erstellt werden und die Routine
	    zur Erstellung des Widget-Baums aufgerufen wird.

  INPUT:  Parent  Widget in der Dialog erzeugt werden soll (i.a.
                  die Shell
	  Dlg     Dialogstruktur die in den Widgetbaum umgesetzt 
	          werden soll
  OUTPUT: ---
  ----------------------------------------------------------------------*/
int DLG_CreateDlg (Widget Parent, DIALOG *Dlg)
{
  int        count, n = 0;
  int        amount = 0;
  char      *Text = NULL;
  char       Upper[DLG_LLEN+1];
  DLGBUTTON *Button;
  Widget     Sep, Box;
  Widget     ButtonID;
  Widget     Left, Right;
  XmString   Label;

  PRO_START ("\n")

  if (Dlg == NULL)
    PRO_RETURN ("Invalid Dlg-Ptr", 0)

  XtSetArg (args[0], XmNtopAttachment,    XmATTACH_FORM);
  XtSetArg (args[1], XmNleftAttachment,   XmATTACH_FORM);
  XtSetArg (args[2], XmNrightAttachment,  XmATTACH_FORM);
  XtSetArg (args[3], XmNbottomAttachment, XmATTACH_FORM);
  Dlg->Form = XtCreateManagedWidget ("complete", xmFormWidgetClass,
				                 Parent, args, 4);
  XtAddCallback (Dlg->Form, XmNhelpCallback, (XtCallbackProc)DLG_HelpCB, Dlg);

  if (Dlg->active == FALSE)
    XtSetSensitive (Dlg->Form, False);

  /* **************************** */
  /* ***** Buttons erzeugen ***** */
  /* **************************** */
  if (Dlg->Buttons.Number > 0)
  {
    XtSetArg (args[0], XmNbottomAttachment, XmATTACH_FORM);
    XtSetArg (args[1], XmNbottomOffset,     0);
    XtSetArg (args[2], XmNleftAttachment,   XmATTACH_FORM);
    XtSetArg (args[3], XmNleftOffset,       0);
    XtSetArg (args[4], XmNrightAttachment,  XmATTACH_FORM);
    XtSetArg (args[5], XmNrightOffset,      0);
    Box = XtCreateManagedWidget ("confirmation", xmFormWidgetClass, 
					         Dlg->Form, args, 6);

    XtSetArg (args[0], XmNtopAttachment,    XmATTACH_FORM);
    XtSetArg (args[1], XmNtopOffset,        5);
    XtSetArg (args[2], XmNbottomAttachment, XmATTACH_FORM);
    XtSetArg (args[3], XmNbottomOffset,     5);

    Left = Right = Box;
    /* Create the left-handed buttons */
    for (count = 0; count < Dlg->Buttons.Number; ++count)
    {
      Button = &(Dlg->Buttons.Data[count]);

      if (Button->Type == NO_ID)
	XtSetArg (args[4], XmNleftOffset, 20);
      elsif (Button->Placement & (DlgBUTTON_LEFT | DlgBUTTON_EVEN))
      {
	TOOLstrtoupper(Upper, Button->Name);
	Text = TOOLGetResString (Box, "xmPushButtonWidgetClass", 
				      Button->Name, "labelString", Upper);
	Label = XmStringLtoRCreate (Text, NORMAL_FONT);
	XtSetArg (args[4], XmNlabelString,    Label);
	XtSetArg (args[5], XmNleftAttachment, XmATTACH_WIDGET);
	XtSetArg (args[6], XmNleftOffset,     5);
	XtSetArg (args[7], XmNleftWidget,     Left); 
	ButtonID = XtCreateManagedWidget (Button->Name, 
					  xmPushButtonWidgetClass,
					  Box, args, 8);
	XmStringFree (Label);
	XtAddCallback (ButtonID, XmNactivateCallback, 
		       (XtCallbackProc)DLG_ButtonCB, Button);

        /* Update reference widgets */
	Left = Dlg->Buttons.Data[count].Button = ButtonID;

	if (Button->Type == OK_ID)
	  Dlg->Ok = ButtonID;
	elsif (Button->Type == CANCEL_ID)
	  Dlg->Cancel = ButtonID;

	amount++;
      }
    }

    /* Process right-sided buttons in reverse order, so that they appear in 
       the order in which they are defined */
    for (count = Dlg->Buttons.Number-1; count >= 0; count--)
    {
      Button = &(Dlg->Buttons.Data[count]);

      if (Button->Type == NO_ID)
	XtSetArg (args[4], XmNleftOffset, 20);
      elsif (Button->Placement & DlgBUTTON_RIGHT)
      {
	TOOLstrtoupper(Upper, Button->Name);
	Text = TOOLGetResString (Box, "xmPushButtonWidgetClass", 
				      Button->Name, "labelString", Upper);
	Label = XmStringLtoRCreate (Text, NORMAL_FONT);
	XtSetArg (args[4], XmNlabelString,     Label);
	XtSetArg (args[5], XmNrightAttachment, XmATTACH_WIDGET);
	XtSetArg (args[6], XmNrightOffset,     5);
	XtSetArg (args[7], XmNrightWidget,     Right); 
	ButtonID = XtCreateManagedWidget (Button->Name, 
					  xmPushButtonWidgetClass,
					  Box, args, 8);
	XmStringFree (Label);
	XtAddCallback (ButtonID, XmNactivateCallback, 
		       (XtCallbackProc)DLG_ButtonCB, Button);

        /* Update reference widgets */
	Right = Dlg->Buttons.Data[count].Button = ButtonID;

	if (Button->Type == OK_ID)
	  Dlg->Ok = ButtonID;
	elsif (Button->Type == CANCEL_ID)
	  Dlg->Cancel = ButtonID;

	amount++;
      }
    }

    if (Dlg->Buttons.Types[1] > 0)                 /* Centered buttons */
      for (count = 0; count < Dlg->Buttons.Number; ++count)
      {
	Button = &(Dlg->Buttons.Data[count]);
	if (Button->Type != NO_ID && (Button->Placement & DlgBUTTON_MIDDLE))
	{
	  TOOLstrtoupper(Upper, Button->Name);
	  Text = TOOLGetResString (Box, "xmPushButtonWidgetClass", 
				        Button->Name, "labelString", Upper);
	  Label = XmStringLtoRCreate (Text, NORMAL_FONT);
	  XtSetArg (args[4], XmNlabelString,    Label);
	  XtSetArg (args[5], XmNleftAttachment, XmATTACH_WIDGET);
	  XtSetArg (args[6], XmNleftWidget,     Left);
	  ButtonID = XtCreateManagedWidget (Button->Name, 
					    xmPushButtonWidgetClass,
					    Box, args, 9);
	  XmStringFree (Label);
	  XtAddCallback (ButtonID, XmNactivateCallback, 
			           (XtCallbackProc)DLG_ButtonCB, Button);
	  Left = Dlg->Buttons.Data[count].Button = ButtonID;

	  if (Button->Type == OK_ID)
	    Dlg->Ok = ButtonID;
	  elsif (Button->Type == CANCEL_ID)
	    Dlg->Cancel = ButtonID;

	  amount++;
	}
    }

    /* Create an additional fill widget to prevent overlapping buttons */
    XtVaCreateManagedWidget ("fill", xmFormWidgetClass, Box,
			     XmNleftAttachment,  XmATTACH_WIDGET,
			     XmNleftWidget,      Left,
			     XmNrightAttachment, XmATTACH_WIDGET,
			     XmNrightWidget,     Right,
			     NULL);

    PRO1 ("NUMBER OF BUTTONS: %d\n", amount)

    /* separator-line: */
    XtSetArg (args[0], XmNbottomAttachment, XmATTACH_WIDGET);
    XtSetArg (args[1], XmNbottomWidget,     Box);
    XtSetArg (args[2], XmNbottomOffset,     5);
    XtSetArg (args[3], XmNleftAttachment,   XmATTACH_FORM);
    XtSetArg (args[4], XmNleftOffset,       0);
    XtSetArg (args[5], XmNrightAttachment,  XmATTACH_FORM);
    XtSetArg (args[6], XmNrightOffset,      0);
    XtSetArg (args[7], XmNorientation,      XmHORIZONTAL);
    XtSetArg (args[8], XmNseparatorType,    XmSHADOW_ETCHED_IN);
    XtSetArg (args[9], XmNmargin,           5);
    Sep = XtCreateManagedWidget ("sepLine", xmSeparatorWidgetClass,
				            Dlg->Form, args, 10);
  }
  else
    Sep = Dlg->Form;

  if (Dlg->Ok != NULL)
  {
    XtInstallAccelerators (Dlg->Ok, Dlg->Ok);
    if (Dlg->Cancel != NULL)
    {
      XtInstallAccelerators (Dlg->Ok,     Dlg->Cancel);
      XtInstallAccelerators (Dlg->Cancel, Dlg->Ok);
    }
  }
  if (Dlg->Cancel != NULL)
    XtInstallAccelerators (Dlg->Cancel, Dlg->Cancel);

  if (Dlg->Resize)
  {
    Dlg->ScrWin = 
      XtVaCreateManagedWidget ("scrWin",
			       xmScrolledWindowWidgetClass, Dlg->Form,
			       XmNscrollingPolicy,          XmAUTOMATIC,
			       XmNscrollBarDisplayPolicy,   XmAS_NEEDED,
			       XmNvisualPolicy,             XmCONSTANT,
			       XmNtopAttachment,            XmATTACH_FORM,
			       XmNleftAttachment,           XmATTACH_FORM,
			       XmNrightAttachment,          XmATTACH_FORM,
			       XmNbottomAttachment,         XmATTACH_WIDGET,
			       XmNbottomWidget,             Sep,
			       NULL);
    XtSetArg (args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
  }
  else
  {
    Dlg->ScrWin = Dlg->Form;
    XtSetArg (args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (args[n], XmNbottomWidget,     Sep);             n++;
  }

  XtSetArg (args[n], XmNbottomOffset,     5);             n++;
  XtSetArg (args[n], XmNtopAttachment,    XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNtopOffset,        5);             n++;
  XtSetArg (args[n], XmNleftAttachment,   XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNleftOffset,       5);             n++;
  XtSetArg (args[n], XmNrightAttachment,  XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNrightOffset,      5);             n++;

  /* create widget-tree */
  Dlg->Box->Parent = Dlg->Work = 
    XtCreateManagedWidget ("work", xmFormWidgetClass, Dlg->ScrWin, args, n);

  DLG_CreateBox (Dlg->Work, Dlg->Work, Dlg->Work, Dlg->Box, DlgVERTICAL);

#if 0
{
  XtWidgetGeometry  Geo;

  Geo.request_mode = CWWidth | CWHeight;
  XtQueryGeometry (Dlg->Work, NULL, &Geo);
  if (Geo.request_mode & (CWWidth | CWHeight))
    printf ("WORK1: %d x %d\n", Geo.width, Geo.height);
}
#endif

  PRO_END
  return amount;
} 


/*----------------------------------------------------------------------
  Funktion: Aufruf zur Erzeugung des gesamten Dialogs

  INPUT:  Dlg       Dialogstruktur des zu erzeugenden Dialogs
          Title     Beschriftung des Titelbalkens (soweit im Stil vorhanden)
	  Style     gewnschter Dialogstil
	  Function  Prozedur die bei Beendingung des Dialogs aufgerufen
	            wird. NULL, falls das Programm auf die Rckkehr warten soll
  OUTPUT: ---
  ----------------------------------------------------------------------*/
void DLG_StartDlg (DIALOG *Dlg, char *Title, int Style, 
		                   CALLFUNCTION  Function)
{
  dual          DefPos = FALSE;
  unsigned int  WinW, WinH, bh;
#if 0
  unsigned int  FormW, FormH;
#endif
  unsigned int  border, bits;
  int           NewX = 0, NewY = 0;
  int           MaxW = 0;
  int           WinX, WinY, bx, by;
  int           ScrX, ScrY;
  int           xMouse = 0, yMouse = 0;
  int           ButtonNo, count;
  int           Sum = 0, Space;
  double        PosX, PosY;
  char          geo_string[12];
  char          DefStyle[10];
  char         *StyleStr = NULL;
  Widget        Shell, Button;
  Window        Dummy;
  XSizeHints    Hints;
#if 0
  Pixmap        Image;
#endif

  PRO_START ("\n")

  /* init global dialog attributes: */
  Dlg->Buttons.Result = -1;
  Dlg->CBFct          = Function;

  /* get dialog position from resourcefile: */
  PosX = TOOLGetResDouble (AppShell, "shellWidgetClass", Title, 
			             "xPosition", (double)Dlg->PosX);
  PosY = TOOLGetResDouble (AppShell, "shellWidgetClass", Title, 
			             "yPosition", (double)Dlg->PosY);

  ScrX = DisplayWidth  (XDisplay, XScreen);
  ScrY = DisplayHeight (XDisplay, XScreen);

  if (PosX >= 0.0 && PosY >= 0.0)
  {
    TOOLGetScrValue (XDisplay, XScreen, PosX, PosY, &NewX, &NewY);
    if (NewX != Dlg->PosX || NewY != Dlg->PosY)
    {
      Dlg->PosX = NewX;
      Dlg->PosY = NewY;
      Dlg->PosSet = TRUE;
    }
  }

  switch (Style)
  {
    case   DlgUSERPOS : strcpy (DefStyle, "USERPOS");
                        break;
    case DlgTRANSIENT : strcpy (DefStyle, "TRANSIENT");
                        break;
    case  DlgMOUSEPOS : strcpy (DefStyle, "MOUSEPOS");
                        break;
    case    DlgNORMAL : strcpy (DefStyle, "NORMAL");
                        break;
              default : PRO1 ("Unknown dialog style %d !\n", Style)
                        strcpy (DefStyle, "NORMAL");
  }
  StyleStr = TOOLGetResString(AppShell, "transientShellWidgetClass",
			                Title, "style", DefStyle);

  n = 0;
  XtSetArg (args[n], XmNsaveUnder,        False);        n++;
  XtSetArg (args[n], XmNdeleteResponse,   XmDESTROY);    n++;
  XtSetArg (args[n], XmNmwmFunctions,     MWM_FUNC_ALL); n++;
  XtSetArg (args[n], XmNinitialState,     NormalState);  n++;
  XtSetArg (args[n], XmNallowShellResize, True);         n++;

  /* the following setting depend on the used wm (check DecorateTransients 
     and WM...Hints in your wm-recouse file e.g. fvwmrc): */
  if (strcasecmp(StyleStr, "TRANSIENT") == 0)
    XtSetArg (args[n], XmNmwmDecorations, MWM_DECOR_BORDER);
  else
  {
    XtSetArg (args[n], XmNtransient, False); n++;
    XtSetArg (args[n], XmNmwmDecorations, MWM_DECOR_ALL);
  }
  n++;

  Shell = XtCreatePopupShell (Title, transientShellWidgetClass, 
			             AppShell, args, n);
  Dlg->Shell = Shell;

  ButtonNo = DLG_CreateDlg (Shell, Dlg);

  DLG_AdjustArrays (Dlg->Box);

  /* Dialog darstellen: */
  PRO0 ("Manage dialog window\n")
  XtSetMappedWhenManaged (Shell, False);
  XtManageChild (Shell);

  PRO0 ("Adjust dialog elements\n")
  DLG_AdjustDlg (Dlg->Box);

  XGetGeometry (XDisplay, XtWindow(Shell), &Dummy,
		&WinX, &WinY, &WinW, &WinH, &border, &bits);

  /* Set specified size of resizeable dialogs */
  if (Dlg->Resize && Dlg->SizeX > 0 && Dlg->SizeY > 0)
  {
    XtVaSetValues (Dlg->ScrWin, XmNwidth,  Dlg->SizeX,
		                XmNheight, Dlg->SizeY,
	                        NULL);
    WinW = Dlg->SizeX;
    WinH = Dlg->SizeY;
  }

#if 0
XGetGeometry (XDisplay, XtWindow(Dlg->Work), &Dummy,
	      &WinX, &WinY, &FormW, &FormH, &border, &bits);
WinW = FormW;
XtUnmanageChild (Shell);
XtUnmanageChild (Dlg->Form);
if (Dlg->ScrWin != Dlg->Form)
{
  XtUnmanageChild (Dlg->ScrWin);
  XtVaSetValues (Dlg->ScrWin, XmNwidth,  FormW + 4,
		              XmNheight, FormH + 4,
	                      NULL);
  XtManageChild (Dlg->ScrWin);
}
 XtVaSetValues (Dlg->Form, XmNwidth,  FormW+4,
		           XmNheight, FormH+4,
	                  NULL);
XtManageChild (Dlg->Form);
XtManageChild (Shell);
#endif

  if (ButtonNo == 1)
  {
    Button = Dlg->Buttons.Data[0].Button;
    /* Center single button vertically */
    if (Dlg->Buttons.Data[0].Placement & DlgBUTTON_FILL)
      XtVaSetValues (Button, XmNwidth, WinW-2, NULL);
    elsif (Dlg->Buttons.Data[0].Placement & DlgBUTTON_MIDDLE)
    {
      XGetGeometry (XDisplay, XtWindow(Button), &Dummy, &bx, &by, 
		    &Dlg->Buttons.Data[0].Width, &bh, &border, &bits);
      XtVaSetValues (Button, 
		     XmNleftOffset, (WinW-Dlg->Buttons.Data[0].Width)/2, 
		     NULL);
    }
  }
  elsif (ButtonNo > 0)
  {
    /* Adjust button size for equal sizes */
    for (count = 0; count < ButtonNo; ++count)     /* find max button size */
    {
      Button = Dlg->Buttons.Data[count].Button;
      XGetGeometry (XDisplay, XtWindow(Button), &Dummy, &bx, &by, 
		    &Dlg->Buttons.Data[count].Width, &bh, &border, &bits);
      MaxW = MAX (MaxW, (int)(Dlg->Buttons.Data[count].Width));
      Sum += Dlg->Buttons.Data[count].Width;
      PRO4 ("Pos: %d/%d  Size: %dx%d\n", 
	    bx, by, Dlg->Buttons.Data[count].Width, bh)
    }

    if ((int)WinW - Sum - 10 * (ButtonNo+1) > 0)
    {
      for (count = 0; count < ButtonNo; ++count)
      {
	Button = Dlg->Buttons.Data[count].Button;
	if (Dlg->Buttons.Data[count].Placement & DlgBUTTON_ADJUST)
        {
	  XtVaSetValues (Button, XmNwidth, MaxW, NULL);
	  Dlg->Buttons.Data[count].Width = MaxW;
	}
      }

      /* Sumarize button sizes */
      Sum = 0;
      for (count = 0; count < ButtonNo; ++count)
	Sum += Dlg->Buttons.Data[count].Width;
    }

    /* Is there still space left for spreading ? */
    if ((int)WinW - 10 - Sum > 0)
    {
      if (Dlg->Buttons.Types[0] > 0)  /* process fill-buttons */
      {
	for (count = 0; count < ButtonNo; ++count)
	  if (Dlg->Buttons.Data[count].Placement & DlgBUTTON_FILL)
	  {
	    Button = Dlg->Buttons.Data[count].Button;
	    XtVaSetValues (Button, XmNwidth,
			   Dlg->Buttons.Data[count].Width + 
			    (int)((WinW-10-Sum)/(double)Dlg->Buttons.Types[0]),
			   NULL);
	  }
      }
      elsif (Dlg->Buttons.Types[1] > 0)  /* process middle buttons */
      {
	/* If there are middle buttons: split space for both sides of them */
	for (count = 0; count < ButtonNo; ++count)
	  if (Dlg->Buttons.Data[count].Placement & DlgBUTTON_MIDDLE)
          {
	    XtVaSetValues (Dlg->Buttons.Data[count].Button, XmNleftOffset,
			   (int)(0.5 * (WinW - 10 - Sum)), NULL);
	    break;
	  }
      }
      elsif (Dlg->Buttons.Types[2] > 0)  /* process equal spaced buttons */
      {
	Space = ((int)WinW - 10 - Sum) / (double)(Dlg->Buttons.Types[2] - 1);

	for (count = 1; count < ButtonNo; ++count)
	  if (Dlg->Buttons.Data[count].Placement & DlgBUTTON_EVEN)
	  {
	    Button = Dlg->Buttons.Data[count].Button;
	    XtVaSetValues (Button, XmNleftOffset, Space, NULL);
	  }
      }   
    }
  }

  border = 10;   /* 2 x border width */
  PRO0 ("setting Window-Position...")
  if (strcasecmp(StyleStr, "USERPOS") == 0 || Dlg->PosX < 0 || Dlg->PosY < 0)
    Hints.flags = USPosition;
  else
  {
    if (Dlg->PosSet)                                  /* positioned Dlg */
    {
      WinX = Dlg->PosX;
      WinY = Dlg->PosY;
    }
    elsif (strcasecmp(StyleStr, "MOUSEPOS") == 0)
    {
      TOOLGetMousePos (&xMouse, &yMouse);
      WinX = MIN (MAX(0, xMouse-25),          ScrX-(int)(WinW+border));
      WinY = MIN (MAX(0, yMouse-(int)WinH-5), ScrY-(int)(WinH+border));
    }
    else                           /* automatic window positioning */
    {
      WinX = XPos += POSSTEP;
      WinY = YPos += POSSTEP;
      DefPos = TRUE;
    }
    sprintf (geo_string, "+%d+%d", WinX, WinY);
    XtVaSetValues (Shell, XmNgeometry, geo_string,
		          XmNx,        WinX,
                          XmNy,        WinY, 
			  NULL);
  }

  Dlg->PosX = WinX;
  Dlg->PosY = WinY;

  XtMapWidget (Shell);
  XtRealizeWidget (Shell);
  XtPopup (Shell, XtGrabNone);

  PRO0 ("\\& Pos\n")
  if (strcasecmp(StyleStr, "USERPOS") == 0)
    Hints.flags = USPosition;
  else
  {
    Hints.x = WinX;
    Hints.y = WinY;
    Hints.flags = PPosition;  /* Program position may be overwriten by WM */
  }

  PRO0 ("\\& Size\n")
  if (! Dlg->Resize)
  {
    /* If dialog is not resizeable, fix the current size */
    Hints.width  = Hints.min_width  = Hints.max_width  = WinW;
    Hints.height = Hints.min_height = Hints.max_height = WinH;
    Hints.flags |= PSize | PMaxSize | PMinSize;
  }
  else
  {
    Hints.width  = Dlg->SizeX;
    Hints.height = Dlg->SizeY;
    Hints.flags |= PSize;
  }
  XSetWMProperties (XDisplay, XtWindow(Shell), NULL, NULL, NULL, 0, 
		              &Hints, NULL, NULL);

  /* create WM-icon (shell must be managed) */
  TOOLSetIcon (Shell, Title, NULL, dlgicon_xpm);

  /* if no callback function is given, use mainloop for modal dialog  */
  if (Dlg->CBFct == NULL)
  {
    PRO0 ("Main-Loop\n")
    TOOLMainLoop (AppCon, Shell);
    PRO0 ("finished\n")
  }

  if (DefPos)                             /* restore default-pos. */
  {
    XPos -= POSSTEP;
    YPos -= POSSTEP;
  }

  PRO_END
}


#if (XmVERSION>=2)
static void DLG_DragCB (Widget w, XtPointer UserData, XtPointer Data)
{
  int Test = (int)UserData;

  printf ("DLG-DRAG CB called: %d\n", Test);
}


static void DLG_DragAction (Widget w, XEvent *Event, 
			              String *params, Cardinal *NumParams)
{
         int            Value;
         Arg            args[4];
	 /*          DragConvertPtr Conv;   */
  static XtCallbackRec  DragCB[] =
  {
    {DLG_DragCB, NULL},
    {NULL, NULL}
  };

  printf ("DRAG-ACTION called\n");

  XtVaGetValues (w, XmNvalue, &Value, NULL);

XtSetArg (args[0], XmNdragOperations,         XmDROP_COPY);
DragCB[0].closure = (XtPointer)Value;
XtSetArg (args[1], XmNdragDropFinishCallback, DragCB);
/*
XtSetArg (args[0], XmNdragMotionCallback,     DragCB);
XtSetArg (args[1], XmNdropSiteLeaveCallback,  DragCB);
*/
/* DragContext = */ XmeDragSource (w, NULL, Event, args, 2);
}
#endif


/*----------------------------------------------------------------------
  Funktion: Aufruf zur Erzeugung des gesamten Dialogs
            Anders als bei DLGMakeDialog wird bei dieser Version 
	    der Inhalt des Dialogs beim Drcken des CANCEL-
	    Buttons nicht gelscht.

  INPUT:  DlgPtr    Dialogstrukturzeiger des zu erzeugenden Dialogs
          Title     Beschriftung des Titelbalkens (soweit im Stil vorhanden)
	  Style     gewnschter Dialogstil
	  Function  Prozedur die bei Beendingung des Dialogs aufgerufen
	            wird. NULL, falls das Programm auf die Rckkehr warten soll
  OUTPUT: ---

  Bsp.: DIALOG *Dlg;
        ...
	DLGMakeDialog (Dlg, "TestDialog", NORMAL, NULL);
  ----------------------------------------------------------------------*/
char *DLGOpenDialog (DIALOG *Dlg, char *Title, int Style, 
		                  CALLFUNCTION Function)
{
  char *Msg = NULL;
#if (XmVERSION>=2)
  static XtActionsRec NewAction[] =
  {
    {"DragAction", DLG_DragAction}
  };
#endif

  PRO_START ("\n")

  if (! TOOLIsInitialized())
    PRO_ERROR0 ("Dlg.-Lib has NOT been initialized", NULL)

  if (Dlg == NULL)
    PRO_ERROR0 ("Try to open an empty dialog", NULL)

  DLG_strncp (Dlg->Title, Title, DLG_TLEN);

#if (XmVERSION>=2)
XtAppAddActions (AppCon, NewAction, sizeof(NewAction)/sizeof(XtActionsRec));
#endif

  if (TOOLScriptRunning())
  {
    if ((Msg = DLG_RunReqScript(Dlg)) != NULL)
    {
      TOOLAddExitCB (DLG_ExitCB, Dlg);
      DLG_StartDlg (Dlg, Title, Style, Function);
      PRO_RETURN (Msg, Msg)
    }

    if (Function != NULL)
      Function (Dlg, Dlg->Buttons.Result);
  }
  else
  {
    TOOLAddExitCB (DLG_ExitCB, Dlg);
    DLG_StartDlg (Dlg, Title, Style, Function);
  }

  PRO_END
  return NULL;
}


/*----------------------------------------------------------------------
  Funktion: schliet den angegebenen Dialog, insofern er existiert

  INPUT:  Dlg     Dialog der geschlossen werden soll
  OUTPUT: ---

  Bsp.: DLGUSTER *Dlg;
        ...
	DLGCloseDialog (Dlg);
  ----------------------------------------------------------------------*/
void DLGCloseDialog (DIALOG *Dlg)
{
  PRO_START ("")

  if (Dlg == NULL)
    PRO_ERROR ("Try to close an empty dialog")

  if (Dlg->Shell != NULL)                          /* close dialog */ 
    XtDestroyWidget (Dlg->Shell);

  Dlg->Buttons.Result = -2;

  PRO_END
}


/*----------------------------------------------------------------------
  Funktion: Initialisiert die Dialoglibrary, falls in der aufrufenden
            Applikation selbst kein Environment zur Verfgung steht.

  INPUT:  argc  Anzahl der Argumente des Programmaufrufes
          argv  Zeiger auf die Liste von Programmargumenten
  OUTPUT: ---

  Bsp.: main (UINT argc, PPCHAR argv)
        {
	  ...
	  DLGInitLib (argc, argv);
	}
  ----------------------------------------------------------------------*/
void DLGInitLib (char *Title, int *argc, char **argv)
{
  PRO_START ("\n")

  if (! TOOLIsInitialized())
  {
    PRO0 ("INITIALIZING TOOLLIB...")
    TOOLInitLib (Title, argc, argv, NULL);
  }
  else
    PRO0 ("TOOLLIB ALREADY INITIALIZED...")

  PRO_END
}


/*----------------------------------------------------------------------
  Funktion: Initialisiert die Dialog-Library, falls in der aufrufenden
            Applikation selbst kein Environment zur Verfgung steht.

  INPUT:  argc  Anzahl der Argumente des Programmaufrufes
          argv  Zeiger auf die Liste von Programmargumenten
  OUTPUT: ---

  Bsp.: ... (Environment z.B. durch libprg erzeugen)
  ----------------------------------------------------------------------*/
void DLGSetLib (Display *XDispl,    XtAppContext AppContext, 
		Widget   ApplShell, int          Screen)
{
  PRO_START ("")

  if (! TOOLIsInitialized())
    TOOLSetLib (XDispl, AppContext, ApplShell, Screen);

  PRO_END
}
