#include <stdio.h>
#include <unistd.h>
#include <math.h>

#include <X11/IntrinsicP.h>

#include <Xm/ManagerP.h>
#include <Xm/DrawingA.h>
#include <Xm/DrawingAP.h>

#include "dlgGlobal.h"
#include "dlgSegment.h"
#include "dlgSegmentP.h"

/* TODO:
   - widget kann noch Fokus erhalten
   - Segment-Destroy erstellen ?
*/


#define SIZE_X 100
#define SIZE_Y 42

 /*    a      */
 /*  +---+    */
 /* f|   |b   */
 /*  +-g-+    */
 /* e|   |c   */
 /*  +---+ .  */
 /*    d   dp */
#define DLG_SEGMENTS 8


typedef struct
{
  int    num;
  XPoint p[6];
} DLG_SEGMENT_ELEM;

typedef struct 
{
  int set[DLG_SEGMENTS];
  int w;
} DLG_SEGMENT_DIGIT;


/* Prototypes: */
static void     DLG_SegmInit    ();
static Boolean  DLG_SegmSetVal  ();
static void     DLG_SegmRedraw  ();
static void     DLG_SegmResize  ();
static void     DLG_SegmExpose  ();


/* resource-list-entries */
static XtResource Resources[] =
{
  /* Name,              Class, Type, Size, Location, DefType, Default */

  /* Form-Widget */
  {XtNwidth,            XtCWidth,
                        XtRDimension,        sizeof(Dimension),
                        XtOffset(CoreWidget, core.width), 
                        XtRImmediate,        (XtPointer) SIZE_X},
  {XtNheight,           XtCHeight,
                        XtRDimension,        sizeof(Dimension),
                        XtOffset(CoreWidget, core.height),  
                        XtRImmediate,        (XtPointer) SIZE_Y},

  /* default is not XmRESIZE_ANY ! ... */
  {XmNresizePolicy,     XmCResizePolicy,
                        XmRChar,              sizeof(unsigned char),
                        XtOffset(XmDrawingAreaWidget, 
				 drawing_area.resize_policy),
                        XmRImmediate,         (XtPointer)XmRESIZE_ANY},

  /* values */
  {DlgNvalue,           DlgCValue,
                        XtRInt,                 sizeof(int),
                        XtOffset(SegmentWidget, segment.Val),
                        XtRImmediate,           (XtPointer)0},
  {DlgNdigits,          DlgCDigits,
                        XtRInt,                 sizeof(int),
                        XtOffset(SegmentWidget, segment.Digits),
                        XtRImmediate,           (XtPointer)5},
  {DlgNdecimals,        DlgCDecimals,
                        XtRInt,                 sizeof(int),
                        XtOffset(SegmentWidget, segment.Decimals),
                        XtRImmediate,           (XtPointer)2},
#if 0
  {XmNbackground,       XmCBackground,
                        XmRPixel,               sizeof(Pixel),
                        XtOffset(SegmentWidget, segment.bg),
                        XtRImmediate,           (XtPointer)CWBackPixel},
  {XmNforeground,       XmCForeground,
                        XmRPixel,               sizeof(Pixel),
                        XtOffset(SegmentWidget, segment.fg),
                        XtRImmediate,           (XtPointer)CWBackPixel},
#endif
};


SegmentClassRec segmentClassRec = 
{
  /* core class: */
  { 
    (WidgetClass)&xmDrawingAreaClassRec,     /* superclass            */
    SEGMENT_CLASS,                           /* class_name            */
    sizeof(SegmentRec),                      /* widget_size           */
    NULL,                                    /* class_iniialize       */
    NULL,                                    /* class_part_initialize */
    FALSE,                                   /* class_inited          */
    (XtInitProc)DLG_SegmInit,                /* initialize            */
    NULL,                                    /* initialize_hook       */
    XtInheritRealize,                        /* realize               */
    NULL,                                    /* actions               */
    0,                                       /* num_actions           */
    Resources,                               /* resources             */
    XtNumber(Resources),                     /* num_resources         */
    NULLQUARK,                               /* xrm_class             */
    TRUE,                                    /* compress_motion       */
    TRUE,                                    /* compress_exposure     */
    TRUE,                                    /* compress_enterleave   */
    FALSE,                                   /* visible_interest      */
    NULL,                                    /* destroy               */
    (XtWidgetProc)DLG_SegmResize,            /* resize                */
    (XtExposeProc)DLG_SegmExpose,            /* expose                */  
    (XtSetValuesFunc)DLG_SegmSetVal,         /* set_values            */
    NULL,                                    /* set_values_hook       */
    XtInheritSetValuesAlmost,                /* set_values_almost     */
    NULL,                                    /* get_values_hook       */
    NULL,                                    /* accept_focus          */
    XtVersion,                               /* version               */
    NULL,                                    /* callback private      */
    XtInheritTranslations,                   /* tm_table              */
    XtInheritQueryGeometry,                  /* query_geometry        */
    NULL,                                    /* display_accelerator   */
    NULL                                     /* extension             */
  },

  /* composite class */
  {
    XtInheritGeometryManager,                /* geometry_manager      */
    XtInheritChangeManaged,                  /* change_managed        */
    XtInheritInsertChild,                    /* insert_child          */
    XtInheritDeleteChild,                    /* delete_child          */
    NULL                                     /* extension             */
  },

  /* constraint class */
  {
    NULL,                                    /* resources             */
    0,                                       /* num_resources         */
    0,                                       /* constraint_size       */
    NULL,                                    /* initialize            */
    NULL,                                    /* destroy               */
    NULL,                                    /* set_values            */
    NULL                                     /* extension             */
  },

  /* XmManager class */
  {
    XtInheritTranslations,                   /* translations          */
    NULL,                                    /* synthetic resources   */
    0,                                       /* num_syn_resources     */
    NULL,                                    /* syn_constraint_res    */
    0,                                       /* num_syn_constr_res    */
    XmInheritParentProcess,                  /* parent_process        */
    NULL                                     /* externsion            */
  },

  /* XmDrawingArea class*/
  {
    NULL                                     /* extension             */
  },

  /* segment class */
  {
    NULL                                     /* extension             */
  }
};


WidgetClass segmentWidgetClass = (WidgetClass)&segmentClassRec;


/* geometry of segment parts */
static DLG_SEGMENT_ELEM Data[DLG_SEGMENTS] =
{
  {6, {{ 3, 2}, { 5, 0}, {15, 0}, {17, 2}, {15, 4}, { 5, 4}}},
  {6, {{18, 3}, {20, 5}, {20,15}, {18,17}, {16,15}, {16, 5}}},
  {6, {{18,19}, {20,21}, {20,31}, {18,33}, {16,31}, {16,21}}},
  {6, {{ 3,34}, { 5,32}, {15,32}, {17,34}, {15,36}, { 5,36}}},
  {6, {{ 2,19}, { 4,21}, { 4,31}, { 2,33}, { 0,31}, { 0,21}}},
  {6, {{ 2, 3}, { 4, 5}, { 4,15}, { 2,17}, { 0,15}, { 0, 5}}},
  {6, {{ 3,18}, { 5,16}, {15,16}, {17,18}, {15,20}, { 5,20}}},
  {4, {{ 0,34}, { 2,32}, { 4,34}, { 2,36}, {-1,-1}, {-1,-1}}}
};

static DLG_SEGMENT_DIGIT Digit[13] = 
{ 
  /* a  b  c  d  e  f  g dp  Width */ 
  { {1, 1, 1, 1, 1, 1, 0, 0}, 24},   /* 0 */
  { {0, 1, 1, 0, 0, 0, 0, 0}, 24},   /* 1 */
  { {1, 1, 0, 1, 1, 0, 1, 0}, 24},   /* 2 */
  { {1, 1, 1, 1, 0, 0, 1, 0}, 24},   /* 3 */
  { {0, 1, 1, 0, 0, 1, 1, 0}, 24},   /* 4 */
  { {1, 0, 1, 1, 0, 1, 1, 0}, 24},   /* 5 */
  { {1, 0, 1, 1, 1, 1, 1, 0}, 24},   /* 6 */
  { {1, 1, 1, 0, 0, 0, 0, 0}, 24},   /* 7 */
  { {1, 1, 1, 1, 1, 1, 1, 0}, 24},   /* 8 */
  { {1, 1, 1, 1, 0, 1, 1, 0}, 24},   /* 9 */
  { {0, 0, 0, 0, 0, 0, 1, 0}, 24},   /* - */
  { {0, 0, 0, 0, 0, 0, 0, 0}, 24},   /*   */
  { {0, 0, 0, 0, 0, 0, 0, 1}, 10}    /* . */
  /* TODO: ':', '+' */
};


/* ************************************* */
/* ** Create a formatted outputstring ** */
/* ** INPUT:  sw  Widget-structure    ** */
/* ** OUTPUT: ---                     ** */
/* ************************************* */ 
static void DLG_CreateString (SegmentWidget sw)
{
  char Format[6];

  if (sw->segment.Decimals == 0)
  {
    sprintf (Format, "%%%dd", sw->segment.Digits);
    sprintf (sw->segment.ValStr, Format, sw->segment.Val);
  }
  else
  {
    sprintf (Format, "%%%d.%df", sw->segment.Digits, sw->segment.Decimals);
    sprintf (sw->segment.ValStr, Format, 
		   sw->segment.Val/pow(10.0, sw->segment.Decimals));
  }
}


static void DLG_SegmInit (SegmentWidget request, SegmentWidget new)
{
  unsigned int count;
  int          SizeX = 0;

  PRO_START ("")

  DLG_CreateString (new);

  /* calculate prefered width */
  for (count = 0; count < strlen(new->segment.ValStr); ++count)
    SizeX += (new->segment.ValStr[count] == '.')? 6 : 15;
  new->core.width = SizeX;

  DLG_SegmResize (new);

  PRO_END
}


static Boolean DLG_SegmSetVal (SegmentWidget old, SegmentWidget request, 
			       SegmentWidget new)
{
  PRO_START ("")

  PRO3 ("\\%d:%d:%d...", request->segment.Val, 
	                 request->segment.Digits, request->segment.Decimals)

  new->segment.Val      = request->segment.Val;
  new->segment.Digits   = request->segment.Digits;
  new->segment.Decimals = request->segment.Decimals;

  DLG_CreateString (new);

  PRO_END
  return True;
}


/* ******************************** */
/* ** redraws a segment-wiget    ** */
/* ** INPUT:  sw  segment-widget ** */
/* ** OUTPUT: ---                ** */
/* ******************************** */
static void DLG_SegmRedraw (SegmentWidget sw)
{
  unsigned int  count;
  int           Pos, num;
  int           Index, segment;
  XPoint        Points[7];
  Display      *XDispl;
  GC            SegmGC;
  Window        Win;
  Colormap      SegmCMap;
  Pixel         color;

  PRO_START ("")

  /* important: must be realized before drawing (drawable must exist) */
  if (! XtIsRealized((Widget)sw))
    PRO_ABORT ("is not realized")

  XDispl   = XtDisplay (sw);
  Win      = XtWindow  (sw);
  SegmGC   = DefaultGC (XDispl, DefaultScreen(XDispl));
  SegmCMap = DefaultColormap (XDispl, DefaultScreen(XDispl));

  PRO3 ("\\%d [%d:%d]...",
	sw->segment.Val, sw->segment.Digits, sw->segment.Decimals)

  /* clear display */
  XFillRectangle (XDispl, Win, sw->manager.background_GC,
		          1, 1, sw->core.width, sw->core.height);

  XtVaGetValues (XtParent(sw), XmNforeground, &color, NULL);
  XSetForeground (XDispl, SegmGC, color);

  Pos = 5 * sw->segment.XFactor;
  for (count = 0; count < strlen(sw->segment.ValStr); ++count)
  {
    switch ((char)sw->segment.ValStr[count])
    {
      case '-' : Index = 10;
	         break;
      case ' ' : Index = 11;
	         break;
      case '.' : Index = 12;
	         break;
       default : Index = sw->segment.ValStr[count] - '0';
    }

    for (segment = 0; segment < DLG_SEGMENTS; ++segment)
      if (Digit[Index].set[segment])
      {
	for (num = 0; num < Data[segment].num; ++num)
	{
	  Points[num].x = (int)(Pos + Data[segment].p[num].x * 
				      sw->segment.XFactor);
	  Points[num].y = (int)( (3 + Data[segment].p[num].y) * 
				      sw->segment.YFactor);
	}
	Points[num] = Points[0];
	XFillPolygon (XDispl, Win, SegmGC, Points, num+1, 
		              Convex, CoordModeOrigin);
      }
    Pos += (int)(Digit[Index].w * sw->segment.XFactor);
  }

  PRO_END
}


static void DLG_SegmResize (SegmentWidget sw)
{
  unsigned int  count;
  double        len = 6;

  PRO_START ("\n")

  PRO2 ("%dx%d\n", sw->core.width, sw->core.height)

  for (count = 0; count < strlen(sw->segment.ValStr); ++count)
    if (sw->segment.ValStr[count] == '.')
      len += 10.0;
    else
      len += 24.0;

  sw->segment.XFactor = sw->core.width / len;
  sw->segment.YFactor = sw->core.height / 41.0;

  DLG_SegmRedraw (sw);

  PRO_END
}


static void DLG_SegmExpose (SegmentWidget sw, XEvent event, Region region)
{
  DLG_SegmRedraw (sw);
}


/* Convenience-funtions: */
Widget DLGCreateSegment (Widget parent, char *Name, 
			 ArgList args, Cardinal num_args)
{
  Widget Result;

  PRO_START ("")

  Result = XtCreateManagedWidget (Name, segmentWidgetClass, 
				        parent, args, num_args);

  PRO_END
  return Result;
}


void DLGSetSegmentValue (Widget widget, double Value, int Digits, int Decimals)
{
  PRO_START ("")

  XtVaGetValues (widget, DlgNvalue,    (int)(Value * pow(10.0, Decimals) + 
					     GREATER(Value)),
		         DlgNdigits,   Digits,
		         DlgNdecimals, Decimals,
		         NULL);

  PRO_END
}
