#include <math.h>

#include "dlgData.h"
#include "dlgCB.h"
#include "dlgDial.h"
#include "dlgSegment.h"


static Arg args[NUM_ARGS];


void DLGSetLabel (DIALOG *Dlg, int BoxID, char *Value)
{
  char     *Text = NULL;
  XmString  Label;
  DLGELEM  *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgLABEL)) == NULL)
    PRO_ABORT ("DLGSetLabel: Invalid Box")

  DLG_strncp (Elem->Value.frec.Text, Value, DLG_LLEN);

  if (Elem->WidgetID != NULL)
  {
    Text = TOOLGetResString (Elem->Master, "xmlabelWidgetClass", 
			     "label", "labelString", Elem->Value.frec.Text);
    Label = XmStringLtoRCreate (Text, NORMAL_FONT);
    XtSetArg (args[0], XmNlabelString, Label);
    XtSetValues (Elem->WidgetID, args, 1);
    if (Elem->Value.frec.Text[0] != '\0')
      XmStringFree (Label);
  }
  PRO_END
}


void DLGSetBool (DIALOG *Dlg, int BoxID, dual Value)
{
  DLGELEM *Elem;

  PRO_START ("\n")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgBOOL)) == NULL)
    PRO_ABORT ("DLGSetBool: Invalid Box")

  Elem->Value.drec.Value = Value;
  if (Elem->WidgetID != NULL)
  {
    XtSetArg (args[0], XmNset, (Boolean)(Elem->Value.drec.Value));
    XtSetValues (Elem->WidgetID, args, 1);
  }
  PRO_END
}


void DLGSetInt (DIALOG *Dlg, int BoxID, int Value)
{
  char     New[PATH_MAX+1] = "";
  DLGELEM *Elem;

  PRO_START ("\n")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgINT)) == NULL)
    PRO_ABORT ("DLGSetInt: Invalid Box")

  Elem->Value.irec.Value = Value;
  if (Elem->WidgetID != NULL)
  {
    DLG_Val2Str (New, Value, Elem->Value.irec.Digits, 0);
    XmTextSetString (Elem->WidgetID, New);
    XFlush (XDisplay); 
  }
  PRO_END
}


void DLGSetDouble (DIALOG *Dlg, int BoxID, double Value)
{
  char     New[PATH_MAX+1] = "";
  DLGELEM *Elem;

  PRO_START ("\n")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgDOUBLE)) == NULL)
    PRO_ABORT ("DLGSetDouble: Invalid Box")

  Elem->Value.rrec.Value = Value;
  if (Elem->WidgetID != NULL)
  {
    DLG_Val2Str (New, Elem->Value.rrec.Value, Elem->Value.rrec.Digits,
		 Elem->Value.rrec.Decimals);
    XmTextSetString (Elem->WidgetID, New);
    XFlush (XDisplay); 
  }
  PRO_END
}


void DLGSetText (DIALOG *Dlg, int BoxID, char *Text)
{
  DLGELEM *Elem;

  PRO_START ("\n")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgTEXT)) == NULL)
    PRO_ABORT ("DLGSetText: Invalid Box")

  if (Elem->Value.srec.Text != NULL)
    free (Elem->Value.srec.Text);
  Elem->Value.srec.Text = DLG_strcp (Text);
  if (Elem->WidgetID != NULL)
  {
    XmTextSetString (Elem->WidgetID, Elem->Value.srec.Text);
    XFlush (XDisplay); 
    /* TODO:
       if (abs(Elem->Value.srec.Rows) > 1)
       XtSetArg (args[2], XmNeditMode, XmMULTI_LINE_EDIT);
       else
       XtSetArg (args[2], XmNeditMode, XmSINGLE_LINE_EDIT);
    */
  }
  PRO_END
}


void DLGSetScale (DIALOG *Dlg, int BoxID, double Value)
{
  double   Factor;
  DLGELEM *Elem;

  PRO_START ("\n")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgSCALE)) == NULL)
    PRO_ABORT ("DLGSetScale: Invalid Box")

  Elem->Value.zrec.Val = LIMIT (Elem->Value.zrec.Min, Elem->Value.zrec.Max, 
				Value);
				/* MIN(Elem->Value.zrec.Max, 
				   MAX(Elem->Value.zrec.Min, Value)); */
  if (Elem->WidgetID != NULL)
  {
    Factor = pow (10.0, Elem->Value.zrec.Decimals);
    XtSetArg (args[0], XmNvalue, (int)(Elem->Value.zrec.Val * Factor +
				       GREATER(Elem->Value.zrec.Val)));
    XtSetValues (Elem->WidgetID, args, 1);
  }
  PRO_END
}


void DLGSetBar (DIALOG *Dlg, int BoxID, double Value)
{
  DLGELEM *Elem;

  PRO_START ("\n")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgBAR)) == NULL)
    PRO_ABORT ("DLGSetBar: Invalid Box")

  Elem->Value.brec.Val = MIN (Elem->Value.brec.Max,
			      MIN (Elem->Value.brec.Max, Value));

  if (Elem->WidgetID != NULL)
  {
    XtSetArg (args[0], XmNvalue, (int)(Elem->Value.brec.Val * 100)); 
    XtSetValues (Elem->WidgetID, args, 1);
  }
  PRO_END
}


void DLGSetUpDn (DIALOG *Dlg, int BoxID, double Value)
{
  char     New[PATH_MAX+1] = "";
  DLGELEM *Elem;

  PRO_START ("\n")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgUPDN)) == NULL)
    PRO_ABORT ("DLGSetUpDn: Invalid Box")

  Elem->Value.urec.Val = 
    DLG_UPDN_VAL (Elem->Value.urec.Min,  Elem->Value.urec.Max,
		  Elem->Value.urec.Step, Value);

  if (Elem->WidgetID != NULL)
  {
    DLG_Val2Str (New, Elem->Value.urec.Val, Elem->Value.urec.Length,
		                            Elem->Value.urec.Decimals);
    XmTextSetString (Elem->WidgetID, New);
    XFlush (XDisplay); 
  }
  PRO_END
}


void DLGSetList (DIALOG *Dlg, int BoxID, int Selection)
{
  DLGELEM *Elem;

  PRO_START ("")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgLIST)) == NULL)
    PRO_ABORT ("DLGSetList: Invalid Box")

  Elem->Value.lrec.Result = MAX (-1, MIN(Elem->Value.lrec.Number-1,Selection));
  PRO2 ("\\%d -> %d...", Selection, Elem->Value.lrec.Result)

  if (Elem->WidgetID != NULL)
  {
    if (Elem->Value.lrec.Result >= 0)
    {
      XmListSelectPos (Elem->WidgetID, Elem->Value.lrec.Result+1, False);
      XtSetArg (args[0], XmNtopItemPosition, 
		MAX(0, MIN(Elem->Value.lrec.Result+1, 
			   Elem->Value.lrec.Number-Elem->Value.lrec.Lines+1)));
    }
    else
      XtSetArg (args[0], XmNselectedItemCount, 0);

    XtSetValues (Elem->WidgetID, args, 1);
  }

  PRO_END
}


void DLGSetOMenu (DIALOG *Dlg, int BoxID, int Entry)
{
  DLGELEM *Elem;

  PRO_START ("\n")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgOMENU)) == NULL)
    PRO_ABORT ("DLGSetOMenu: Invalid Box")

  Elem->Value.orec.Result = MIN (Elem->Value.orec.Number-1,
				 MAX (0, Entry));
  if (Elem->WidgetID != NULL)
  {
    XtSetArg (args[0], XmNmenuHistory, 
	      Elem->Value.orec.Item[Elem->Value.orec.Result].WidgetID);
    XtSetValues (Elem->WidgetID, args, 1);
  }
  PRO_END
}


void DLGSetTBox (DIALOG *Dlg, int BoxID, dual *Toggle)
{
  int      count;
  DLGELEM *Elem;

  PRO_START ("\n")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgTBOX)) == NULL)
    PRO_ABORT ("DLGSetTBox: Invalid Box")

  for (count = 0; count < Elem->Value.trec.Number; ++count)
    Elem->Value.trec.Toggle[count].Value = Toggle[count];

  if (Elem->WidgetID != NULL)
    for (count = 0; count < Elem->Value.trec.Number; ++count)
    {
      XtSetArg (args[0], XmNset, 
		         (Boolean)Elem->Value.trec.Toggle[count].Value);
      XtSetValues (Elem->Value.trec.Toggle[count].WidgetID, args, 1);
    }
  PRO_END
}


void DLGSetRBox (DIALOG *Dlg, int BoxID, int Button)
{
  int      count, Number;
  DLGELEM *Elem;

  PRO_START ("\n")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgRBOX)) == NULL)
    PRO_ABORT ("DLGSetRBox: Invalid Box")

  Number = MIN (Elem->Value.trec.Number, MAX(0, Button));

  for (count = 0; count < Elem->Value.trec.Number; ++count)
    Elem->Value.trec.Toggle[count].Value = FALSE;
  Elem->Value.trec.Toggle[Number].Value = TRUE;

  if (Elem->WidgetID != NULL)
    for (count = 0; count < Elem->Value.trec.Number; ++count)
    {
      XtSetArg (args[0], XmNset,
		         (Boolean)Elem->Value.trec.Toggle[count].Value);
      XtSetValues (Elem->Value.trec.Toggle[count].WidgetID, args, 1);
    }
  PRO_END
}


void DLGSetProc (DIALOG *Dlg, int BoxID, DLGFUNCTION Function)
{
  DLGELEM *Elem;

  PRO_START ("\n")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgPROC)) == NULL)
    PRO_ABORT ("DLGSetProc: Invalid Box")

  Elem->Value.crec.FPtr = Function;

  PRO_END
}


void DLGSetArrow (DIALOG *Dlg, int BoxID, DLGFUNCTION Function)
{
  DLGELEM *Elem;

  PRO_START ("\n")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgPROC)) == NULL)
    PRO_ABORT ("DLGSetArrow: Invalid Box")

  Elem->Value.arec.FPtr = Function;

  PRO_END
}


void DLGSetPic (DIALOG *Dlg, int BoxID, DLGFUNCTION Function)
{
  DLGELEM *Elem;

  PRO_START ("\n")

  /* Can't change from PIC to PICB or vice versa */
  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgPIC)) == NULL ||
      Elem->Value.prec.FPtr == NULL)
    PRO_ABORT ("DLGSetPic: Invalid Box")

  Elem->Value.prec.FPtr = Function;

  PRO_END
}


void DLGSetImage (DIALOG *Dlg, int BoxID, DLGFUNCTION Function)
{
  DLGELEM *Elem;

  PRO_START ("\n")

  /* Can't change from PIC to PICB or vice versa */
  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgIMAGE)) == NULL ||
      Elem->Value.xrec.FPtr == NULL)
    PRO_ABORT ("DLGSetImage: Invalid Box")

  Elem->Value.xrec.FPtr = Function;

  PRO_END
}


#if (XmVERSION>=2)
  void DLGSetNote (DIALOG *Dlg, int BoxID, int PageNo)
  {
    DLGELEM *Elem;

    PRO_START ("\n")

    if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgNOTE)) == NULL)
      PRO_ABORT ("DLGSetNote: Invalid Box")

    Elem->Value.nrec.ActPage = LIMIT (1, Elem->Value.nrec.Number, PageNo);

    if (Elem->WidgetID != NULL)
    {
      XtSetArg (args[0], XmNcurrentPageNumber, Elem->Value.nrec.ActPage);
      XtSetValues (Elem->WidgetID, args, 1);
    }

    PRO_END
  }
#endif


void DLGSetFileSel (DIALOG *Dlg, int BoxID, char *FileName)
{
  char     New[PATH_MAX+1] = "";
  XmString Label;
  DLGELEM *Elem;

  PRO_START ("\n")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgFSEL)) == NULL)
    PRO_ABORT ("DLGSetFSel: Invalid Box")

  DLG_strncp (Elem->Value.grec.Result, FileName, PATH_MAX);
  if (Elem->WidgetID != NULL)
  {
    DLG_MakeRightText (New, Elem->Value.grec.Result, Elem->Value.grec.Length);
    Label = XmStringLtoRCreate (New, NORMAL_FONT);
    XtSetArg (args[0], XmNlabelString, Label);
    XtSetValues (Elem->WidgetID, args, 1);
    XmStringFree (Label);
  }
  PRO_END
}


void DLGSetDirSel (DIALOG *Dlg, int BoxID, char *DirName)
{
  char     New[PATH_MAX+1] = "";
  XmString Label;
  DLGELEM *Elem;

  PRO_START ("\n")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgDIRSEL)) == NULL)
    PRO_ABORT ("DLGSetFSel: Invalid Box")

  DLG_strncp (Elem->Value.grec.Result, DirName, PATH_MAX);
  if (Elem->WidgetID != NULL)
  {
    DLG_MakeRightText (New, Elem->Value.grec.Result, Elem->Value.grec.Length);
    Label = XmStringLtoRCreate (New, NORMAL_FONT);
    XtSetArg (args[0], XmNlabelString, Label);
    XtSetValues (Elem->WidgetID, args, 1);
    XmStringFree (Label);
  }
  PRO_END
}


void DLGSetDial (DIALOG *Dlg, int BoxID, double Value)
{
  double   Factor;
  DLGELEM *Elem;

  PRO_START ("\n")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgDIAL)) == NULL)
    PRO_ABORT ("DLGSetDial: Invalid Box")

  Elem->Value.erec.Val = MIN(Elem->Value.erec.Max,
			     MAX(Elem->Value.erec.Min, Value));
  if (Elem->WidgetID != NULL)
  {
    Factor = pow (10.0, Elem->Value.erec.Decimals);
    DLGSetDialValue (Elem->WidgetID, (int)(Elem->Value.erec.Val * Factor +
				           GREATER(Elem->Value.erec.Val)));
  }

  PRO_END
}


void DLGSetSegment (DIALOG *Dlg, int BoxID, double Value)
{
  DLGELEM *Elem;

  PRO_START ("\n")

  if ((Elem = DLG_CheckBoxType(Dlg, DLGBox(BoxID), DlgSEGM)) == NULL)
    PRO_ABORT ("DLGSetSegment: Invalid Box")

  Elem->Value.hrec.Val = Value;
  if (Elem->WidgetID != NULL)
  {
    XtSetArg (args[0], DlgNvalue, 
	               (int)(Elem->Value.hrec.Val * 
			     pow(10.0, Elem->Value.hrec.Decimals) + 
			     GREATER(Elem->Value.hrec.Val)));
    XtSetValues (Elem->WidgetID, args, 1);
  }
  PRO_END
}
