/* ********************************* */
/* *** (c) Stefan Vietinghoff    *** */
/* *** general utility functions *** */
/* ********************************* */

#ifdef sunos
#include <sys/types.h>
#include <sys/dir.h>
#define dirent direct
#else
#include <dirent.h>
#endif

#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>

#include <Xm/Protocols.h>

#include "dlgUtils.h"

       int  FSXPos  = 70;
       int  FSYPos  = 50;
static char CompleteName[PATH_MAX+1] = "";

char *ElemName[] =
{
  "EMPTY",
  "LABEL",
  "BOOL",
  "INT",
  "DOUBLE",
  "TEXT",
  "SCALE",
  "BAR",
  "UPDN",
  "LIST",
  "OMENU",
  "TBOX",
  "RBOX",
  "PROC",
  "ARROW",
  "PIC",
  "FSEL",
  "DIAL",
  "SEGM",
  "USER",
#if (XmVERSION>=2)
  "NOTE",
  "TREE",
#endif
  "SEPARATOR",
  "BOX"
};


/* ***************************** */
/* *** common help functions *** */
/* ***************************** */

/*----------------------------------------------------------------------
  Funktion: DLG_strcp kopiert einen String in einen anderen und allokiert 
            dabei den benoetigten Speicher.
	    Der Anwender ist fuer die Freigabe des allokierten Speichers
	    selbst verantwortlich.
  
  INPUT:  String   Quell-Zeichenkette
  OUTPUT: (char *)  Zeiger auf die String-Kopie (Speicher wird allokiert)
  
  Bsp.: char *Ziel = NULL;
        ...
        Ziel = DLG_strcp ("Dies ist ein Test");
   ----------------------------------------------------------------------*/
char *DLG_strcp (const char *String, ...)
{
  va_list  args;
  char    *Result, *Buffer;

  if (String == NULL)
    return NULL;
  
  va_start (args, String);
  
  /* Speicher fr den String allokieren (erst mal 4 kB): */
  Buffer = (char*)calloc(4096, sizeof(char));
  vsprintf (Buffer, String, args);
  
  va_end (args);
  
  /* reduce allocated memory to really used amount */
  Result = (char*)calloc (strlen(Buffer)+1,sizeof(char));
  strcpy (Result, Buffer);
  free (Buffer);

  return Result;
}


/*----------------------------------------------------------------------
  Funktion: DLG_strncp kopiert einen String mit einer maximalen Lnge
            in einen Ergebnisstring. Im Unterschied zu strncp wird bei
            einem NULL-Zeiger das erste Zeichen des Ergebnisstrings mit
            einem '\0' markiert.
  
  INPUT:  Source   Quell-Zeichenkette
          INT      Maximale Lnge von Result
  OUTPUT: Result   Zeiger auf die String-Kopie (mu allokiert sein)
  
  Bsp.: CHAR Ziel[100];
        ...
        DLG_strncp (Ziel, "Dies ist ein Test", 100);
  ----------------------------------------------------------------------*/
void DLG_strncp (char *Result, const char *Source, int Length)
{
  if (Source == NULL)
    Result[0] = '\0';
  else
  {
    strncpy (Result, Source, Length);
    Result[Length-1] = '\0';
  }
}


/*----------------------------------------------------------------------
  Funktion: Ermittlung der Anzahl der Nachommastellen einer double-
            Zahl.
  
  INPUT:  Value  Wert dessen Nachkommastellenanzahl gesucht wird
  OUTPUT: (int)  Anzahl der Nachkommastellen
  
  Bsp.: DLG_GetDecimals (-15.134543);
        liefert den Wert 6
  ----------------------------------------------------------------------*/
int DLG_GetDecimals (double Value)
{
  int count = 0;

  while (fmod(Value, 1.0) != 0.0)   /* mult. mit 10 bis Wert ganzzahlig */
  {
    count++;
    Value *= 10.0;
  }
  return count;
}


/*----------------------------------------------------------------------
  Funktion: liefert die Anzahl der Vorkommastellen einer (double)-Zahl
  
  INPUT:  Value  Wert zu dem die Anzahl der Vorkommastellen bestimmt
                 werden soll
  OUTPUT: (int)  Anzahl der Vorkommastellen
  
  Bsp.: DLG_GetDigits (123.2133);   liefert den Wert 3
  ----------------------------------------------------------------------*/
int DLG_GetDigits (double Value)
{
  int count = 0;
  
  while (floor(fabs(Value)) != 0.0)  /* teile durch 10, bis 1. Stelle = 0 */
  {
    count++;
    Value /= 10.0;
  }
  return count;
}


/*----------------------------------------------------------------------
  Funktion: GetDigit liefert die Ziffer an der durch den Wert Pos ange-
            gebenen Stelle (von hinten gezaehlt). D.h. die 1er Stelle
            erhaelt die Position 1, die 10er Stelle den Wert 2, ...
            oder mit anderen Worten: Pos entspricht dem log(10)-1 der
                                     Wertigkeit der gesuchten Stelle
            Nachkommastellen werden bei der Suche ignoriert (d.h. es
            findet keine Rundung statt)

    INPUT:  Value  Zahl in der die Stelle gesucht werden soll
            Pos    Position der gesuchten Ziffer
    OUTPUT: (int)  Wert an der gesuchten Stelle

  Bsp.: GetDigit (12345.233, 4) liefert den Wert 2
        GetDigit (25390.742, 2) liefert den Wert 9
        usw.
  ----------------------------------------------------------------------*/
int DLG_GetDigit (double Value, int Pos)
{
  int count;
  int Result = 0;

  if (Pos < 0)
    return Result;
  
  for (count = 1; count < Pos; ++count)
    Value /= 10.0;
    
  Result = (int)(floor(Value)) % 10;

  return Result;
}


/*----------------------------------------------------------------------
  Funktion: Ermittelt aus einer List von Trennlinien-Konstanten
            (Zeichen in libdlg.h defininiert) die Trennlinien
	    zwischen zwei geg. Boxen (Number). Ist der String
	    zu kurz wird ' ' (NOLINE) zurckgegeben.

    INPUT:  SepString  String mit den Trennlinienkonstanten
            Number     Nummer der Trennstelle
    OUTPUT: (char)     Konstante der Trennlinie an dieser Stelle

  Bsp.: DLG_GetSeparator (" SDIS", 2) = 'S'
        DLG_GetSeparator (" SS", 5) = ' '
  ----------------------------------------------------------------------*/
char DLG_GetSeparator (char *SepString, int Number)
{
  char Result = 'X';

  if (SepString == NULL || Number > (int)strlen(SepString))
    return ' ';

  Result = SepString[Number-1];
  if (Result != DlgNO_LINE[0]     && Result != DlgDASH_LINE[0] && 
      Result != DlgSINGLE_LINE[0] && Result != DlgDOUBLE_LINE[0] &&
      Result != DlgETCHED_IN[0]   && Result != DlgETCHED_OUT[0])
    Result = 'X';
      
  return Result;
}


/*----------------------------------------------------------------------
  Funktion: Umwandlung einer (double)-Zahl in einen String mit einer
            bestimmten Formatierung. Der Speicher fr den String 
	    wird selbst allokiert und mu somit vom Benutzer wieder
	    freigegeben werden.
	    Vorteil gegenber zu fscanf: 
	    Die Formatierung kann als Parameter angegeben werden und 
	    somit auch whrend des Programmlaufes variieren.

  INPUT:  Value    Zahl die in einen String umgewandelt werden soll
          Length   Gesamtlnge des Strings (d.h. einschlielich Punkt/VZ)
	  Digits   Anzahl der Nachkommastellen
                   (fr Digits = 0 wird auch das Komma weggelassen)
  OUTPUT: (char*)  Zeiger auf die erstellte Zeichenkette
  
  Bsp.: DLGValStr (-134.2346, 10, 2);        liefert "   -134.23"
  ----------------------------------------------------------------------*/
void DLG_Val2Str (char *Result, double Value, int Length, int Digits)
{
  char Pattern[16];

  if (Digits == 0)
  {
    sprintf (Pattern, "%%%dd", Length);        /* Formatstring erzeugen */
    sprintf (Result, Pattern, (int)Value);
  }
  else
  {
    sprintf (Pattern, "%%%d.%df", Length, Digits);
    sprintf (Result, Pattern, Value);
  }
}


/*----------------------------------------------------------------------
  Funktion: Erzeugt einen rechtsbndigen Text der Lnge "Length". Ist 
            der angegeben Text lnger, so werden die VORDEREN Zeichen
	    abgeschnitten.
  
  INPUT:  Text     zu formatierender Text
          Length   gewnschte Lnge des Ergebnisses
  OUTPUT: (char*)  Zeiger auf den erzeugten String
  
  Bsp.: char Result[100];
        ...
	DLG_MakeRightText (Result, "bla bla bla", 20) = "         bla bla bla"
	DLG_MakeRightText (Result, "~/daten/program.c", 12) = "en/program.c"
  ----------------------------------------------------------------------*/
void DLG_MakeRightText (char *Result, char *Text, int Len)
{
  int  length;
  char Mask[16];

  length = strlen (Text);
  if (Len >= length)
  {
    sprintf (Mask, "%%-%ds", Len);
    sprintf (Result, Mask, Text);
  }
  else
  {
    sprintf (Mask, "%%%ds", Len);
    sprintf (Result, Mask, Text+(length-Len+1));
  }
}  


/*----------------------------------------------------------------------
  Funktion: Komplettiert einen gegebenen Dateinamen soweit dieses eindeutig
            mglich ist. (funktioniert wie Tab in bash)
  
  INPUT:  FileName  Dateispezifikation, die komplettiert werden soll
  OUTPUT: (int)     Rckgabestatus:
                     0  das im FileName enthaltene Verzeichnis 
		        kann nicht gelesen werden
		     1  es existiert genau eine Datei, auf die 
		        der Dateiname pat
		     2  es existieren mehrere Mglichkeiten, den
		        Dateinamen zu ergnzen.
		    Der ergnzte Name wird in FileName zurckgegeben.
  
  Bsp.: DLG_CompleteFileName ("/users/stefan/lin");
        Existierende Dateien im Verzeichnis "/users/stefan/":
          linux1.0      Rckgabewert 1, Filename = /users/stefan/linux1.0

          linux.MO, linux.HDD und linux.CD
	    Rckgabewert 2, Filename = /users/stefan/linux.
  ----------------------------------------------------------------------*/
static int DLG_Complete (const struct dirent *Entry)
{
  return ((strncmp(Entry->d_name, CompleteName, strlen(CompleteName)) == 0));
}

int DLG_CompleteFileName (char *FileName, int Length)
{
         int      hits = 0;
	 int      count, Pos = 0;
         char     Path[PATH_MAX+1] = "";
  struct dirent **Entry = NULL;

  PRO_START ("")

  TOOLGetFileDir  (FileName, Path,         PATH_MAX);
  TOOLGetFileName (FileName, CompleteName, PATH_MAX);
  hits = scandir ((Path[0]=='\0')?"/":Path, &Entry, DLG_Complete, alphasort);

  if (hits <= 0)
    PRO_RETURN ("No hits", 0)
  elsif (hits == 1)
    Pos = strlen (Entry[0]->d_name);
  else
  {
    Pos = strlen (CompleteName);
    while (Pos < (int)strlen(Entry[0]->d_name))
    {
      count = 1;
      while (count < hits &&
	     Entry[0]->d_name[Pos] == Entry[count]->d_name[Pos])
	count++;
      if (count < hits)
	break;
      Pos++;
    }
  }

  sprintf (FileName, "%s/", Path);
  strncat (FileName, Entry[0]->d_name, Pos);
  if (hits == 1 && opendir(FileName) != NULL)
    strcat (FileName, "/");

  for (count = 0; count < hits; ++count)
    free (Entry[count]);
  free (Entry);

  PRO_END
  return (hits==1)? 1:2;
}


/* ********************************************************* */
/* ** Universal timer. Checks, if the left mouse-button   ** */
/* ** is pressed. While it is not released an user-       ** */
/* ** update-function is called.                          ** */
/* ** INPUT:  Data   Button-data with following inform.:  ** */
/* **                - primary or secondary delay         ** */
/* **                - element to update by user-function ** */
/* **                - widget of button                   ** */
/* **                - sizeof button                      ** */
/* **                - user-function                      ** */
/* **         Timer  Identifier of timer                  ** */
/* ** OUTPUT: ---                                         ** */
/* ********************************************************* */
void DLG_Timer (TIMER_DATA *Data, XtIntervalId Timer)
{
  int          PosX, PosY, x, y;
  unsigned int Mask;
  Window       RootWin, ChildWin;

  XQueryPointer(XDisplay, XtWindow(Data->widget),
		&RootWin, &ChildWin, &x, &y, &PosX, &PosY, &Mask);
  if (Mask & Button1Mask)
  {
    Data->Function (Data->Data, Data->Param);
    XtAppAddTimeOut (AppCon, Data->start? DLG_START_DELAY : DLG_NEXT_DELAY, 
			     (XtTimerCallbackProc)DLG_Timer, Data);
  }
  Data->start = FALSE;
}


/* ********************************************************* */
/* ** Universal timer. Checks, if the left mouse-button   ** */
/* ** is pressed. While it is not released an user-       ** */
/* ** update-function is called.                          ** */
/* ** INPUT:  Data   Button-data with following inform.:  ** */
/* **                - primary or secondary delay         ** */
/* **                - element to update by user-function ** */
/* **                - widget of button                   ** */
/* **                - sizeof button                      ** */
/* **                - user-function                      ** */
/* **         Timer  Identifier of timer                  ** */
/* ** OUTPUT: ---                                         ** */
/* ********************************************************* */
void DLG_UpdateSizeAndPos (DIALOG *Dlg)
{
  Dimension PosX,  PosY;
  Dimension SizeX, SizeY;

  if (Dlg != NULL && Dlg->Shell != NULL && XtIsManaged(Dlg->Shell))
  {
    XtVaGetValues (Dlg->Shell, XmNx,      &PosX,
		               XmNy,      &PosY,
		               XmNwidth,  &SizeX,
		               XmNheight, &SizeY, 
		               NULL); 
    Dlg->PosX  = (int)PosX;
    Dlg->PosY  = (int)PosY;
    Dlg->SizeX = (int)SizeX;
    Dlg->SizeY = (int)SizeY;
  }
} 


