#ifndef _DLG_LIB_H
#define _DLG_LIB_H

/* ********************************************** */
/* **             Dialog Lib 0.9               ** */
/* **    (c) 1995 by Stefan Vietinghoff        ** */
/* ** helpful library for generating intactive ** */
/* ** user interfaces                          ** */
/* ********************************************** */

/* Before you report any bugs you should remember:

      *** IT'S NOT A BUG - IT'S A FEATURE ! ***  

   This library needs the libbase-package !
   If not already done directly, you must initialize it with either the 
   DLGInitLib- or DLGSetLib-function before you can use it. 
*/

#ifdef __cplusplus
extern "C" {
#endif

#include <limits.h>

#include <Xm/Xm.h>

#include <libtools.h>

/*
  just implemented elements:
   x seperator (horizontal/veritcal)
   x title (headlines)
   - - - - - - - - - - - - - - - - - - - - - - 
   x dialogbox
   x symbolbox
   - - - - - - - - - - - - - - - - - - - - - - 
   x fileSelector (selbstndig)
   - - - - - - - - - - - - - - - - - - - - - -
   x arrow-button
   x bool-button
   x push-button
   x scroll-bar
   x text-box
   x scale
   x radio-box
   x (scrolled) list
   x int-/double-value
   x option-menu
   x toggle-group
   x UpDn-Button
   x file-selector
   x arrow-button
   x label
   x picture(-button)
   x dial
   + 7-segement-display
   x notebook
     combo-box
     dir-tree
   x user-widget
*/

/* *** Definitions: *** */
#define DLG_NLEN    64
#define DLG_TLEN    64
#define DLG_LLEN   512
#define DLG_VLEN   128
#define MAX_SEP     32

#define DLG_START_DELAY 1000
#define DLG_NEXT_DELAY    50

#define DlgCANCEL  0
#define DlgOK      1
#define DlgHELP    2
#define DlgCLOSE   3
#define DlgSAVE    4
#define DlgSTOP    5
#define DlgAPPLY   6
#define DlgACCEPT  7
#define DlgDONE    8
#define DlgDEFAULT 9


/* *** Enumerations: *** */
enum DLG_MSG_TYPES
{
  DlgERROR = 0,
  DlgINFO,
  DlgMESSAGE,
  DlgQUESTION,
  DlgWARNING,
  DlgWORKING
};

enum DLG_ELEM_TYPES
{
  DlgEMPTY = -1,       /* -1: no element                  */
  DlgLABEL,            /*  0: label                       */
  DlgBOOL,             /*  1: toggle-button (radio-box)   */ 
  DlgINT,              /*  2: integer                     */  
  DlgDOUBLE,           /*  3: double                      */
  DlgTEXT,             /*  4: text-(box)                  */
  DlgSCALE,            /*  5: scale-bar                   */
  DlgBAR,              /*  6: scroll-bar                  */
  DlgUPDN,             /*  7: up/donw-arrow-selection     */
  DlgLIST,             /*  8: (scrolled) list             */
  DlgOMENU,            /*  9: option-menu                 */
  DlgTBOX,             /* 10: toggle-box                  */
  DlgRBOX,             /* 11: radio-box                   */
  DlgPROC,             /* 12: function-call (push-button) */
  DlgARROW,            /* 13: arrow-button                */
  DlgPIC,              /* 14: XPM-label/button            */
  DlgIMAGE,            /* 15: Image label/button          */
  DlgFSEL,             /* 16: File-selektor               */
  DlgDIRSEL,           /* 17: Directory-selector          */
  DlgDIAL,             /* 18: Dial-arrow                  */
  DlgSEGM,             /* 19: 7-segment-display           */
  DlgUSER,             /* 20: User-defined widget         */

#if (XmVERSION>=2)
    DlgNOTE,           /* 21: Notebook                    */
    DlgTREE,           /* 22: Tree                        */
#endif

  DlgSEP,              /* 23: Separator-Line              */
  DlgBOX               /* 24: Frame-Box                   */
};


/* Seperator-types: */
#define DlgNO_LINE     " "
#define DlgDASH_LINE   "A"
#define DlgSINGLE_LINE "L"
#define DlgDOUBLE_LINE "D"
/* Border-types: */
#define DlgSHADOW_IN   "I"
#define DlgSHADOW_OUT  "S"
/* both types: */
#define DlgETCHED_IN   "E"
#define DlgETCHED_OUT  "O"

/* label/value-orientation: */
#define DlgNOWHERE     (1L<<0)
#define DlgCENTER      (1L<<1)
#define DlgWEST        (1L<<2)
#define DlgNORTH_WEST  (1L<<3)
#define DlgNORTH       (1L<<4)
#define DlgNORTH_EAST  (1L<<5)
#define DlgEAST        (1L<<6)
#define DlgSOUTH_EAST  (1L<<7)
#define DlgSOUTH       (1L<<8)
#define DlgSOUTH_WEST  (1L<<9)

enum BUTTON_PLACEMENT
{
  DlgBUTTON_LEFT   = (1L<<0),
  DlgBUTTON_MIDDLE = (1L<<1),
  DlgBUTTON_RIGHT  = (1L<<2),
  DlgBUTTON_EVEN   = (1L<<3)
};

#define DlgBUTTON_ADJUST (1L<<4)        /* Resize button-size */
#define DlgBUTTON_FILL   (1L<<5)        /* enlarge button     */


enum DIRECTIONS
{
  DlgMIDDLE = 0,
  DlgLEFT,
  DlgRIGHT,
  DlgUP,
  DlgDOWN,
  DlgFILL,
  DlgNO_CHANGE
};

enum PICTYPE
{
  DlgUNKNOWN = -1,
  DlgANY,
  DlgXBM,
  DlgXPM
  /* TODO: ? 
     GIF, TIFF, PCX, PNG */
};

enum DLG_OPEN_STYLES
{
  DlgNORMAL = 0,
  DlgUSERPOS,
  DlgTRANSIENT,
  DlgMOUSEPOS
};

enum DLG_NOTE_STYLES
{
  DlgNOBINDING = 0,
  DlgSOLID,
  DlgSPIRAL
};

enum DLG_PAGE_STYLES
{
  DlgPAGE = 0,
  DlgMAJOR,
  DlgMINOR
};

enum DLG_ORIENTATIONS
{
  DlgHORIZONTAL = 0,
  DlgVERTICAL,
  DlgCELL
};

enum DLG_CBSTATUS
{
  DlgCONFIRM = -1,
  DlgDENY
};

enum DLG_TREE_REASON
{
  DlgOTHER = -1,
  DlgCURSOR,
  DlgRETURN,
  DlgMOUSE1,
  DlgMOUSE2,
  DlgMOUSE3
};


/* ********************************** */
/* *** structure of element-types *** */
/* ********************************** */

struct  DLGHEAD;
typedef struct DLGHEAD DIALOG;

typedef int (*DLGFUNCTION)(DIALOG *Dlg, int BoxID);
typedef void (*CALLFUNCTION)(DIALOG *Dlg, int ID);
typedef dual (*CHECKFUNCTION)(DIALOG *Dlg, int ButtonID);
typedef Widget (*WIDGETFUNCTION)(DIALOG *Dlg, Widget Parent, void *Param);
typedef char ITEMTEXT[DLG_LLEN];


typedef struct ITEMREC
{
         char     Label[DLG_LLEN];
	 int      Index;
	 Widget   WidgetID;
  struct DLGELEM *Parent;
} ITEMREC;


typedef struct TOGGLE
{
  char  Item[DLG_LLEN];
  dual  Value;
} TOGGLE;


typedef struct SWITCHREC
{
         char    Item[DLG_LLEN];
	 dual    Value;
	 int     Index;
	 Widget  WidgetID;
  struct DLGELEM *Parent;
} SWITCHREC;


typedef struct LABELREC
{
  char Text[DLG_LLEN];
} LABELREC;


typedef struct BOOLREC
{
  dual        Value;
  dual        Default;
  DLGFUNCTION FPtr;
} BOOLREC;


typedef struct INTREC
{
  int         Value;
  int         Default;
  byte        Digits;
  DLGFUNCTION FPtr;
} INTREC;


typedef struct REALREC
{
  byte        Digits;
  byte        Decimals;
  double      Value;
  double      Default;
  DLGFUNCTION FPtr;
} REALREC;


typedef struct STRINGREC
{
  int          Length;
  char        *Text;
  char        *Default;
  int          Cols;
  int          Rows;
  DLGFUNCTION  FPtr;
} STRINGREC;


typedef struct PUSHREC
{
  DLGFUNCTION FPtr;
} PUSHREC;


typedef struct ARROWREC
{
  int         Dir;
  DLGFUNCTION FPtr;
} ARROWREC;


typedef struct BARREC
{
  double      Min;
  double      Max;
  double      Val;
  double      Default;
  double      Step;
  int         Side;
  dual        reverse;
  DLGFUNCTION FPtr;
} BARREC;

 
typedef struct UPDNREC
{
  double      Min;
  double      Max;
  double      Val;
  double      Default;
  double      Step;
  int         Length;
  int         Decimals;
  DLGFUNCTION FPtr;
} UPDNREC;


typedef struct SCALEREC
{
  double      Min;
  double      Max;
  double      Val;
  double      Default;
  int         Decimals;
  int         Side;
  dual        reverse;
  DLGFUNCTION FPtr;
} SCALEREC;


typedef struct OMENUREC
{
  int          Number;
  ITEMREC     *Item;
  int          Result;
  int          Default;
  DLGFUNCTION  FPtr;
} OMENUREC;  

typedef struct LISTREC
{
  int         Number;
  ITEMTEXT   *Item;
  int         Lines;
  int         Result;
  int         Default;
  DLGFUNCTION FPtr;
} LISTREC;


typedef struct TOGGLEREC
{
  int          RowCol;
  int          Number;
  SWITCHREC   *Toggle;
  dual        *Result;
  dual        *Default;
  DLGFUNCTION  FPtr;
} TOGGLEREC;


typedef struct PICREC
{
  char         Format;
  int          SizeX, SizeY;
  char         PicName[PATH_MAX+1];
  Pixel        fg, bg;
  void        *Data;
  void        *iData;
  XImage      *Image;
  DLGFUNCTION  FPtr;
} PICREC;


typedef struct IMAGEREC
{
  XImage      *Image;
  DLGFUNCTION  FPtr;
} IMAGEREC;


typedef struct NOTEPAGE
{
         int       Type;
         char      Label[DLG_LLEN+1];
  struct DLGBOX   *Box;
  struct NOTEPAGE *Next;
} NOTEPAGE;


typedef struct NOTEREC
{
  int          Style;
  int          ActPage;
  int          Default;
  int          Number;
  NOTEPAGE    *Page;
  DLGFUNCTION  FPtr;
} NOTEREC;


typedef struct TREEITEM
{
         int       ID;
         char      Label[DLG_LLEN+1];
         void     *Icon;
         dual      State;
         void     *UserData;
         Widget    gadget;
  struct TREEITEM *Sub;
  struct TREEITEM *Next;
} TREEITEM;


typedef struct TREEREC
{
  int           Shadow;
  int           Lines;
  char        **Header;
  Pixmap        ClosePic;
  Pixmap        OpenPic;
  int           Reason;
  TREEITEM     *Data;
  TREEITEM     *Result;
  TREEITEM     *Default;
  DLGFUNCTION   FPtr;
} TREEREC;


typedef struct FILESELREC
{
  char         Mask[PATH_MAX+1];
  char         Default[PATH_MAX+1];
  int          Length;
  int          Button;
  char         Result[PATH_MAX+1];
  char         Default2[PATH_MAX+1];
  dual         Open;
  dual         Files;
  DLGFUNCTION  FPtr;
} FILESELREC;


typedef struct DIALREC
{
  double      Min;
  double      Max;
  double      Val;
  double      Default;
  double      Step;
  byte        Digits;
  byte        Decimals;
  DLGFUNCTION FPtr;
} DIALREC;


typedef struct SEGMENTREC
{
  double      Val;
  double      Default;
  byte        Digits;
  byte        Decimals;
  DLGFUNCTION FPtr;
} SEGMENTREC;


typedef struct WIDGETREC
{
  WIDGETFUNCTION  FPtr;
  void           *Param;
} WIDGETREC;


typedef struct MARKREC
{
  byte Style;
  int  x1, y1;
  int  x2, y2;
} MARKREC;


/* default-buttons: */
enum BUTTON_ID
{
  NO_ID = -1,
  CANCEL_ID,
  OK_ID,
  CLOSE_ID,
  HELP_ID,
  SAVE_ID,
  STOP_ID,
  APPLY_ID,
  ACCEPT_ID,
  DONE_ID,
  DEFAULT_ID,
  USER_ID
};


/* Convenience definitions for dialog buttons: */
#define NO_BUTTON      NO_ID,      "", \
                       DlgBUTTON_RIGHT | DlgBUTTON_ADJUST, NULL, -1
#define CANCEL_BUTTON  CANCEL_ID,  "cancel", \
                       DlgBUTTON_RIGHT | DlgBUTTON_ADJUST, NULL, DlgCANCEL
#define OK_BUTTON      OK_ID,      "ok", \
                       DlgBUTTON_RIGHT | DlgBUTTON_ADJUST, NULL, DlgOK
#define CLOSE_BUTTON   CLOSE_ID,   "close", \
                       DlgBUTTON_RIGHT | DlgBUTTON_ADJUST, NULL, DlgCLOSE
#define HELP_BUTTON    HELP_ID,    "help",  \
                       DlgBUTTON_RIGHT | DlgBUTTON_ADJUST, NULL, DlgHELP
#define SAVE_BUTTON    SAVE_ID,    "save", \
                       DlgBUTTON_RIGHT | DlgBUTTON_ADJUST, NULL, DlgSAVE
#define STOP_BUTTON    STOP_ID,    "stop", \
                       DlgBUTTON_RIGHT | DlgBUTTON_ADJUST, NULL, DlgSTOP
#define APPLY_BUTTON   APPLY_ID,   "apply", \
                       DlgBUTTON_RIGHT | DlgBUTTON_ADJUST, NULL, DlgAPPLY
#define ACCEPT_BUTTON  ACCEPT_ID,  "accept", \
                       DlgBUTTON_RIGHT | DlgBUTTON_ADJUST, NULL, DlgACCEPT
#define DONE_BUTTON    DONE_ID,    "done", \
                       DlgBUTTON_RIGHT | DlgBUTTON_ADJUST, NULL, DlgDONE
#define DEFAULT_BUTTON DEFAULT_ID, "default", \
                       DlgBUTTON_RIGHT | DlgBUTTON_ADJUST, NULL, DlgDEFAULT
#define USER_BUTTON    USER_ID

typedef struct DLGBUTTON
{
  int           ID;
  int           Type;
  int           Number;
  char          Name[DLG_LLEN+1];
  int           Placement;
  CALLFUNCTION  FPtr;
  DIALOG       *Dlg;
  Widget        Button;
  unsigned int  Width;
} DLGBUTTON;

typedef struct BUTTONINFO
{
  int        Number;                    /* amount of buttons    */
  DLGBUTTON *Data;                      /* vector of buttons    */
  int        Result;                    /* id of result-button  */
  dual       changed;
  int        Types[3];                  /* # of fill/center/space buttons */
} BUTTONINFO;

#define DlgGRID(BoxID,Col,Row) ((BoxID * 100 + Col) * 100 + Row)


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* simlified dialog-structure: 

DIALOG**    DIALOG*        DLGHEAD
   +---+  :  +---+        +-------+
   | o-+--:->| o-+------->|       |
   +---+  :  +---+        |       |      DLGBOX
..........:   ^ ^         +---+---+     +-------+
              | +---------+-o | o-+---->|       |<----------------+
   +----------+           +---+---+     |       |                 |
   |                                    |       |                 |
   |    DLGELEM                  +----->|       |     DLGBOX      |
   |   +-------------------------+-+    +---+---+    +-------+    |
   +-->|                         | |<---+-o V o-+--->|       |<-+ |
   |   |                         | |    +-------+    |       |  | |
   |   +-------------------------+-+                 +-------+  | |
   +---+--------------o          | |                 |   o---+--+ |
       +---+---+---+---+---+---+-+-+                 +-------+    |
       |   |   |   |   |   |   | o |                 |   o---+----+
       +---+---+---+---+---+---+---+                 +-------+
*/

typedef struct DLGELEM
{
           char      Name[DLG_NLEN+1];  /* box-title                      */
           int       BoxID;             /* box-identifier                 */

           int       PosX, PosY;        /* elem-offset in form-widget  v  */

  unsigned long   LabelPos;             /* label-adjustment               */
  unsigned long   ValuePos;

         dual     hidden;               /* is element visible ?           */
         dual     active;               /* is element sensitive           */

         int      Type;                 /* type-identifier                */
         union    {                     /* still free letters: jqvy       */
	            LABELREC     frec;
	            BOOLREC      drec;
	            INTREC       irec;
	            REALREC      rrec;
	            STRINGREC    srec;
	            PUSHREC      crec;
	            BARREC       brec;
	            UPDNREC      urec;
	            SCALEREC     zrec;
	            TOGGLEREC    trec;
	            OMENUREC     orec;
	            LISTREC      lrec;
	            ARROWREC     arec;
	            PICREC       prec;
     	            IMAGEREC     xrec;
	            FILESELREC   grec;
	            DIALREC      erec;
	            SEGMENTREC   hrec;
                    NOTEREC      nrec;
	            TREEREC      krec; 
                    WIDGETREC    wrec;
	            MARKREC      mrec;
	          } Value;              /* type-specific value            */

         int      NumArgs;              /* number of user-arguments       */
         ArgList  UserArgs;             /* list of user-arguments         */

         Pixel    fg, bg;               /* color of element               */

         Widget   Master;               /* main (form-)widget of element  */
         Widget   Label;                /* Label-widget                   */
         Widget   WidgetID;             /* value-widget of element        */
         DIALOG  *Dlg;                  /* pointer to dialog              */
  struct DLGELEM *Next;                 /* ptr to next elem (if pos-list) */
} DLGELEM;


typedef struct DLGBOX
{
         int        BoxID;              /* Box-ID (auch Name denkbar)       */
	 char       Title[DLG_TLEN+1];  /* Box heading                      */
         char       TitlePos;           /* Position of Title                */
         int        AlignX, AlignY;     /* box-alignment                    */
	 Widget     Parent, Main, Box;  /* Box-widgets                      */
	 char       Border;             /* border around box                */
	 dual       active;             /* Box sensitiv ?                   */
	 char       Sep[MAX_SEP+1];     /* separators between rows/cols     */
         byte       Dir;                /* 0=elem, 1=col, 2=row, 3=pos-list */
         int        Number;             /* Number of items                  */
  struct DLGELEM   *Item;               /* box-item (possibly. itemlist)    */
         int        Rows, Cols;         /* # subboxes                       */
         void      *UserData;           /* user-informationen               */
  struct DLGBOX  ***List;               /* pointer to subboxes              */
} DLGBOX;


/* This structure contains info of complete dialog */
typedef struct DLGHEAD
{
  char           Title[DLG_TLEN+1];     /* title of whole dialog            */

  dual           PosSet;                /* flag, if position has been set   */
  int            PosX, PosY;            /* position of dialog               */
  int            SizeX, SizeY;          /* size of dialog                   */
  dual           Resize;                /* is dialog resizeable             */

  dual           active;                /* dialog sensitive ?               */
  byte           locked;                /* dialog buttons locked ?          */

  DLGBOX        *Box;                   /* Ptr onto dialog-struct itself    */
  BUTTONINFO     Buttons;               /* button-data                      */

  Widget         Shell;                 /* dialog shell-widget              */
  Widget         Form;                  /* main-form-widget                 */
  Widget         ScrWin;                /* Scrolled-window                  */
  Widget         Work;                  /* work-form-widget                 */
  Widget         Ok;                    /* widget for OK-button             */
  Widget         Cancel;                /* widget for CANCEL-button         */

  CALLFUNCTION   CBFct;                 /* Callback-function                */
  CHECKFUNCTION  CheckFct;              /* Check-function                   */
  CALLFUNCTION   HelpFct;               /* Help-funktion                    */
} DLGHEAD;


/* --- Resource-Function --------------------------------------------- */

char *DLGGetResourceEntry   (char *Name, char *Default);

/* --- Dialogwindow -------------------------------------------------- */

int   DLGOpenTextDialog (char *DialogID, int Type, char *Msg, int Number, ...);
int   DLGOpenPicDialog  (char *DialogID, int Format, int XSize, int YSize,
		         void *Data, char *Msg, int Number, ...);
void  DLGSetDialogPos   (double xPos, double yPos);

/* --- File-Selector ------------------------------------------------- */

char *DLGOpenFileSel    (const char *Title, 
		         char *Result, int Length, char *Mask, ...);
char *DLGOpenFileSel2   (const char *Title, 
			 char *Result, int Length, char *Mask, ...);
char *DLGOpenDirSel     (const char *Title, 
			 char *Result, int Length, char *Mask, ...);
void  DLGSetFSPos       (double xPos, double yPos);

/* --- dialog-Functions ---------------------------------------------- */

Widget DLGCreateWidget   (Widget Parent, DLGELEM *Elem, ArgList args, int Num);
Widget DLGVaCreateWidget (Widget Parent, DLGELEM *Elem, ...);
void   DLGFreeElem       (DLGELEM *Elem);
Widget DLGGetShell       (DIALOG *Dlg);

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
dual  DLGInsertLabel     (DIALOG *Dlg, int BoxID, char *Name,
			               const char *Label, ...);
dual  DLGInsertBool      (DIALOG *Dlg, int BoxID, char *Name, dual Value);
dual  DLGInsertInt       (DIALOG *Dlg, int BoxID, char *Name,
			               int Value, byte Digits);
dual  DLGInsertDouble    (DIALOG *Dlg, int BoxID, char *Name,
			               double Value, byte Digits, 
			               byte Decimals);
dual  DLGInsertText      (DIALOG *Dlg, int BoxID, char *Name,
			               int Length, const char *Text, 
			               int Cols, int Rows);
dual  DLGInsertScale     (DIALOG *Dlg, int BoxID, char *Name,
			               double Min, double Max, double Value,
			               int Decimals, int Align);
dual  DLGInsertBar       (DIALOG *Dlg, int BoxID, char *Name,
			               double Min, double Max, double Value,
			               double Step, dual Align);
dual  DLGInsertUpDn      (DIALOG *Dlg, int BoxID, char *Name,
		                       double Min, double Max, double Value,
		                       double Step);
dual  DLGInsertList      (DIALOG *Dlg, int BoxID, char *Name, 
			               int Lines, int Number, ...);
dual  DLGInsertOMenu     (DIALOG *Dlg, int BoxID, char *Name,
		 	               int Number, ...);
dual  DLGInsertTBox      (DIALOG *Dlg, int BoxID, char *Name,
			               int RowCols, int Number, ...);
dual  DLGInsertRBox      (DIALOG *Dlg, int BoxID, char *Name,
		                       int RowCols, int Number, ...);
dual  DLGInsertProc      (DIALOG *Dlg, int BoxID, char *Name,
			               DLGFUNCTION Function);
dual  DLGInsertArrow     (DIALOG *Dlg, int BoxID, char *Name,
			               int Dir, DLGFUNCTION Function);
dual  DLGInsertPic       (DIALOG *Dlg, int BoxID, char *Name, int Format, 
			               int SizeX, int SizeY, void *Data,
			               DLGFUNCTION Function);
dual  DLGInsertImage     (DIALOG *Dlg, int BoxID, char *Name, XImage *Image, 
			               DLGFUNCTION Function);
dual  DLGInsertFileSel   (DIALOG *Dlg, int BoxID, char *Name, 
			               char *FileMask, int Length, 
			               char *Default);
dual  DLGInsertDirSel    (DIALOG *Dlg, int BoxID, char *Name, 
			               char *DirMask, int Length, 
			               char *Default);
dual  DLGInsertDial      (DIALOG *Dlg, int BoxID, char *Name,
			               double Min, double Max, double Value,
			               byte Digits, byte Decimals);
dual  DLGInsertSegment   (DIALOG *Dlg, int BoxID, char *Name, double Value,
		                       byte Digits, byte Decimals);
dual  DLGInsertSeparator (DIALOG *Dlg, int BoxID, int Style,
			               int x1, int y1, int length,
			               dual Direction);
dual  DLGInsertBox       (DIALOG *Dlg, int BoxID, int Style,
		                       int x1, int y1, 
			               int width, int height);
dual  DLGInsertWidget    (DIALOG *Dlg, int BoxID, WIDGETFUNCTION Function, 
		                       void *Param);

dual  DLGPlaceLabel      (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
		                       char *Label);
dual  DLGPlaceBool       (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
			               dual Value);
dual  DLGPlaceInt        (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
		                       int Value, byte Digits);
dual  DLGPlaceDouble     (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
		                       double Value, byte Digits, 
		                       byte Decimals);
dual  DLGPlaceText       (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
		                       int Length, char *Text, 
		                       int Cols, int Rows);
dual  DLGPlaceScale      (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
			               double Min, double Max, double Value,
			               int Decimals, int Align);
dual  DLGPlaceBar        (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
			               double Min, double Max, double Value,
		                       double Step, dual Align);
dual  DLGPlaceUpDn       (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
			               double Min, double Max, double Value,
			               double Step);
dual  DLGPlaceList       (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
			               int Lines, int Number, ...);
dual  DLGPlaceOMenu      (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
			               int Number, ...);
dual  DLGPlaceTBox       (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
			               int RowCols, int Number, ...);
dual  DLGPlaceRBox       (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
			               int RowCols, int Number, ...);
dual  DLGPlaceProc       (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
			               DLGFUNCTION Function);
dual  DLGPlaceArrow      (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
		                       int Dir, DLGFUNCTION Function);
dual  DLGPlacePic        (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
			               int Format, 
			               int SizeX, int SizeY, void *Data,
			               DLGFUNCTION Function);
dual  DLGPlaceImage      (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
			               XImage *Image, DLGFUNCTION Function);
dual  DLGPlaceFileSel    (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
			               char *FileMask, int Length, 
			               char *Default);
dual  DLGPlaceDirSel     (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
			               char *DirMask, int Length, 
			               char *Default);
dual  DLGPlaceDial       (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
			               double Min, double Max, double Value,
			               byte Digits, byte Decimals);
dual  DLGPlaceSegment    (DIALOG *Dlg, int BoxID, int x, int y, char *Name,
			               double Value,
			               byte Digits, byte Decimals);
dual  DLGPlaceSeparator  (DIALOG *Dlg, int BoxID, int x, int y, int Style,
			               int x1, int y1, int length,
			               dual Direction);

DLGELEM  *DLGMakeLabel    (char *Name, char *Label);
DLGELEM  *DLGMakeBool     (char *Name, dual Value);
DLGELEM  *DLGMakeInt      (char *Name, int Value, byte Digits);
DLGELEM  *DLGMakeDouble   (char *Name, double Value, 
			               byte Digits, byte Decimals);
DLGELEM  *DLGMakeText     (char *Name, int Length, char *Text, 
			               int Cols, int Rows);
DLGELEM  *DLGMakeScale    (char *Name, double Min, double Max, double Value,
			               int Decimals, int Align);
DLGELEM  *DLGMakeBar      (char *Name, double Min, double Max, double Value,
			               double Step, dual Align);
DLGELEM  *DLGMakeUpDn     (char *Name, double Min, double Max, double Value,
			               double Step);
DLGELEM  *DLGMakeList     (char *Name, int Lines, int Number, ...);
DLGELEM  *DLGMakeOMenu    (char *Name, int Number, ...);
DLGELEM  *DLGMakeTBox     (char *Name, int RowCols, int Number, ...);
DLGELEM  *DLGMakeRBox     (char *Name, int RowCols, int Number, ...);
DLGELEM  *DLGMakeProc     (char *Name, DLGFUNCTION Function);
DLGELEM  *DLGMakeArrow    (char *Name, int Dir, DLGFUNCTION Function);
DLGELEM  *DLGMakePic      (char *Name, int Format, 
			               int SizeX, int SizeY, void *Data,
			               DLGFUNCTION Function);
DLGELEM  *DLGMakeImage    (char *Name, XImage *Image, DLGFUNCTION Function);
DLGELEM  *DLGMakeFileSel  (char *Name, char *FileMask, int Length, 
			               char *Default);
DLGELEM  *DLGMakeDirSel   (char *Name, char *DirMask, int Length, 
			               char *Default);
DLGELEM  *DLGMakeDial     (char *Name, double Min, double Max, double Value,
			               byte Digits, byte Decimals);
DLGELEM  *DLGMakeSegment  (char *Name, double Value,
			               byte Digits, byte Decimals);

void      DLGCloseDialog      (DIALOG *Dlg);
void      DLGFreeDialog       (DIALOG *Dlg);
void     *DLGGetUserData      (DIALOG *Dlg, int BoxID);
int       DLGGetExitButton    (DIALOG *Dlg);
dual      DLGIsEmpty          (DIALOG *Dlg);
void      DLGMakeCols         (DIALOG *Dlg, int BoxID, int Number,
			                    char *Separator, ...);
void      DLGMakeRows         (DIALOG *Dlg, int BoxID, int Number, 
		                            char *Separator, ...);
void      DLGMakeGrid         (DIALOG *Dlg, int BoxID, int Rows, int Cols);
void      DLGMakeArray        (DIALOG *Dlg, int BoxID, int Rows, int Cols);
void      DLGMakeButton       (DIALOG *Dlg, int Type, char *Label,
			                    int Placement,
			                    CALLFUNCTION Function, 
			                    int ButtonID);
DIALOG   *DLGNewDialog        (             int BoxID);
char     *DLGOpenDialog       (DIALOG *Dlg, char *Title, int Style, 
			                    CALLFUNCTION Function);
void      DLGSetBorder        (DIALOG *Dlg, int BoxID, char *Style);
void      DLGSetColor         (DIALOG *Dlg, int BoxID, 
			                    int r, int g, int b, dual front);
void      DLGSetBoxActive     (DIALOG *Dlg, int BoxID, dual Active);
void      DLGSetBoxPosition   (DIALOG *Dlg, int BoxID, int PosX, int PosY);  
void      DLGSetFunction      (DIALOG *Dlg, int BoxID, DLGFUNCTION Function);
void      DLGSetTitle         (DIALOG *Dlg, int BoxID, const char *Title, 
			                    char Pos);
void      DLGSetButtonActive  (DIALOG *Dlg, int Button, dual active);
void      DLGSetCheckFunction (DIALOG *Dlg, CHECKFUNCTION CheckFct);
void      DLGSetHelpFunction  (DIALOG *Dlg, CALLFUNCTION  HelpFct);
void      DLGSetLabelPos      (DIALOG *Dlg, int BoxID, 
			                    unsigned long LabelPos,
			                    unsigned long ValuePos);
void      DLGSetBoxAlignment  (DIALOG *Dlg, int BoxID,
				            int AdjustX, int AdjustY);
void      DLGSetMouseCursor   (DIALOG *Dlg, int CursorID);
void      DLGSetMousePos      (DIALOG *Dlg, int BoxID);
void      DLGSetPos           (DIALOG *Dlg, double PosX, double PosY);
void      DLGSetSize          (DIALOG *Dlg, double SizeX, double SizeY,
	 		                    dual Resize);
void      DLGSetDlgActive     (DIALOG *Dlg, dual Active);
void      DLGSetUserData      (DIALOG *Dlg, int BoxID, void *Data);
void      DLGShowBox          (DIALOG *Dlg, int BoxID, dual show);
void      DLGShowDialog       (DIALOG *Dlg, dual State);

void      DLGSetLabel         (DIALOG *Dlg, int BoxID, char *Value);
void      DLGSetBool          (DIALOG *Dlg, int BoxID, dual Value);
void      DLGSetInt           (DIALOG *Dlg, int BoxID, int Value);
void      DLGSetDouble        (DIALOG *Dlg, int BoxID, double Value);
void      DLGSetText          (DIALOG *Dlg, int BoxID, char *Text);
void      DLGSetScale         (DIALOG *Dlg, int BoxID, double Value);
void      DLGSetBar           (DIALOG *Dlg, int BoxID, double Value);
void      DLGSetUpDn          (DIALOG *Dlg, int BoxID, double Value);
void      DLGSetList          (DIALOG *Dlg, int BoxID, int Selection);
void      DLGSetOMenu         (DIALOG *Dlg, int BoxID, int Entry);
void      DLGSetTBox          (DIALOG *Dlg, int BoxID, dual *Toggle);
void      DLGSetRBox          (DIALOG *Dlg, int BoxID, int Button);
void      DLGSetProc          (DIALOG *Dlg, int BoxID, DLGFUNCTION Function);
void      DLGSetArrow         (DIALOG *Dlg, int BoxID, DLGFUNCTION Function);
void      DLGSetPic           (DIALOG *Dlg, int BoxID, DLGFUNCTION Function);
void      DLGSetImage         (DIALOG *Dlg, int BoxID, DLGFUNCTION Function);
void      DLGSetSegment       (DIALOG *Dlg, int BoxID, double Value);
void      DLGSetDial          (DIALOG *Dlg, int BoxID, double Value);
void      DLGSetFileSel       (DIALOG *Dlg, int BoxID, char *FileName);
void      DLGSetDirSel        (DIALOG *Dlg, int BoxID, char *DirName);
void      DLGSetArguments     (DIALOG *Dlg, int BoxID, 
			                 ArgList Arguments, int Number);
void      DLGSetTreeState  (DIALOG *Dlg, int BoxID, dual State);

void      DLGSetDefault    (DIALOG *Dlg, int BoxID);

void  DLGChangeInt     (DIALOG *Dlg, int BoxID, int Digits);
void  DLGChangeDouble  (DIALOG *Dlg, int BoxID, int Digits, int Decimals);
void  DLGChangeText    (DIALOG *Dlg, int BoxID, int Length, int Rows);
void  DLGChangeScale   (DIALOG *Dlg, int BoxID, double Min, double Max, 
		                                int Decimals);
void  DLGChangeBar     (DIALOG *Dlg, int BoxID, double Min, double Max,
		                                double Step);
void  DLGChangeUpDn    (DIALOG *Dlg, int BoxID, double Min, double Max,
		                                double Step);
void  DLGChangeList    (DIALOG *Dlg, int BoxID, int Lines, int Number, ...);
void  DLGChangeOMenu   (DIALOG *Dlg, int BoxID, int Number, ...);
void  DLGChangeProc    (DIALOG *Dlg, int BoxID, char *Name, 
			                        DLGFUNCTION Function);
void  DLGChangeArrow   (DIALOG *Dlg, int BoxID, int Direction);
void  DLGChangePic     (DIALOG *Dlg, int BoxID, int SizeX, int SizeY, 
			                        void *Data);
void  DLGChangeImage   (DIALOG *Dlg, int BoxID, XImage *Image);
void  DLGChangeSegment (DIALOG *Dlg, int BoxID, int Digits, int Decimals);

void     DLGGetPos         (DIALOG *Dlg, int *PosX,  int *PosY);
void     DLGGetSize        (DIALOG *Dlg, int *SizeX, int *SizeY);

void    *DLGGetValue       (DIALOG *Dlg, int BoxID);
dual     DLGGetBool        (DIALOG *Dlg, int BoxID);
int      DLGGetInt         (DIALOG *Dlg, int BoxID);
double   DLGGetDouble      (DIALOG *Dlg, int BoxID);
char    *DLGGetText        (DIALOG *Dlg, int BoxID);
double   DLGGetScale       (DIALOG *Dlg, int BoxID);
double   DLGGetBar         (DIALOG *Dlg, int BoxID);
double   DLGGetUpDn        (DIALOG *Dlg, int BoxID);
int      DLGGetList        (DIALOG *Dlg, int BoxID);
int      DLGGetOMenu       (DIALOG *Dlg, int BoxID);
dual     DLGGetTBox        (DIALOG *Dlg, int BoxID, int Number);
int      DLGGetRBox        (DIALOG *Dlg, int BoxID);
char    *DLGGetFileSel     (DIALOG *Dlg, int BoxID);
char    *DLGGetDirSel      (DIALOG *Dlg, int BoxID);
double   DLGGetDial        (DIALOG *Dlg, int BoxID);

Widget   DLGGetBoxWidget   (DIALOG *Dlg, int BoxID);

#if (XmVERSION>=2)
  dual  DLGInsertNote     (DIALOG *Dlg, int BoxID, char *Name, int Style);
  dual  DLGAddPage        (DIALOG *Dlg, int BoxID, int Type, char *Label, 
			                int SubBox);
  void  DLGSetNote        (DIALOG *Dlg, int BoxID, int PageNo);
  int   DLGGetNote        (DIALOG *Dlg, int BoxID);

  dual  DLGInsertTree     (DIALOG *Dlg, int BoxID, char *Name, 
			                Pixmap ClosePic, Pixmap OpenPic,
      			                int Lines);
  dual  DLGAddTreeItem    (DIALOG *Dlg, int BoxID, const char *Name, 
			                int ID, int Parent, 
			                dual State, void *UserData);
  dual  DLGAddTreeIcon    (DIALOG *Dlg, int BoxID, const char *Name, 
			                void *Data, int ID, int Parent, 
			                dual State, void *UserData);
  TREEITEM *DLGGetTree    (DIALOG *Dlg, int BoxID);
  int       DLGGetReason  (DIALOG *Dlg, int BoxID);
  dual      DLGFreeTree   (DIALOG *Dlg, int BOXID);
  dual      DLGBuildTree  (DIALOG *Dlg, int BoxID);
#endif

void    DLGWriteAppDefs   (FILE      *fp,  DIALOG *Dlg);
void    DLGWriteDialog    (DIALOG *Dlg, char *FileName);

/* --- Convenience-Functions -------------------------------------- */

Widget  DLGCreateSep      (Widget Parent, DLGELEM *Elem);
Widget  DLGCreateDial     (Widget Parent, String Name, 
			   ArgList user_args, Cardinal num_args,
			   XtCallbackProc CBProc, XtPointer UserData);

/* --- Initialization-Functions ----------------------------------- */

void    DLGInitLib       (char *Title, int *argc, char **argv);
void    DLGSetLib        (Display *XDispl,    XtAppContext AppContext, 
			  Widget   ApplShell, int          Screen);

#ifdef __cplusplus
} /* for C++ V2.0 */
#endif

#endif
