/* ************************************************************************* */
/* POLYEDER.C                                                                */
/* besteht aus Funktionen zum Erstellen und Verndern von Polyedern          */
/* ------------------------------------------------------------------------- */
/* Von            : Stefan Vietinghoff und Heike Ochs-Vietinghoff            */
/* letzte nderung: 23.06.1997 von Heike                                     */
/* ************************************************************************* */


#include <math.h>

#include "libgeo.h"


/* ************************************************************************* */
/* GEONewPolyeder                                                            */
/* erzeugt leeres Polyeder.                                                  */
/* ************************************************************************* */
POLYEDER *GEONewPolyeder (void)
{
  POLYEDER *Result;

  Result = (POLYEDER*)malloc (sizeof(POLYEDER));

  Result->Number = 0;
  Result->List   = NULL;
  Result->Last   = NULL;

  return Result;
}


/* ************************************************************************* */
/* GEOAddPolyeder                                                            */
/* fgt neues Polygon in Polyeder ein.                                       */
/* ************************************************************************* */
POLYEDER *GEOAddPolygon (POLYEDER *Polyeder, POLYGON *Polygon)
{
  int       count;
  POLYGON  *Ptr, *Next;
  POLYEDER *Result;

  if (Polygon == NULL)                    /* Polygon leer -> Ende            */
    return Polyeder;

  if ((Result = Polyeder) == NULL)        /* Polyeder leer -> neues erzeugen */
    Result = GEONewPolyeder ();

#if 1
  while (Polygon != NULL)
  {
    Next = Polygon->Next;
    Polygon->Next = NULL;

    if (GEOTestPolygon(Polygon) == 1)
    {
printf ("TESSELATING.............................................\n");

      GEOTriangulatePolygon (Polygon);
      for (count = 0; count < Polygon->Number-2; ++count)
      {
	Ptr = GEONewPolygon ();
	Ptr->Points   = Polygon->Points;
	Ptr->Number   = 3;
	Ptr->Vertex   = (int *)malloc (sizeof(int) * 3);
	Ptr->Triangle = (TRIANGLE *)malloc (sizeof(TRIANGLE));
	Ptr->Vertex[0] = Ptr->Triangle[0][0] = Polygon->Triangle[count][0];
	Ptr->Vertex[1] = Ptr->Triangle[0][1] = Polygon->Triangle[count][1];
	Ptr->Vertex[2] = Ptr->Triangle[0][2] = Polygon->Triangle[count][2];
	Ptr->Info.Center = GEOCalcPolygonCenter (Ptr);
	GEOCalcHesseForm (Ptr);

        if (Result->List == NULL)
	  Result->List = Ptr;
	else
	  Result->Last->Next = Ptr;
	Result->Last = Ptr;
	(Result->Number)++;
      }
      GEOFreePolygon (Polygon, FALSE);
    }
    else
    {
      if (Result->List == NULL)
	Result->List = Polygon;
      else
	Result->Last->Next = Polygon;
      Result->Last = Polygon;
      (Result->Number)++;
    }
    Polygon = Next;
  }
#else
  if (Result->List == NULL)               /* Polygonliste von Polyeder leer  */
    Result->List = Polygon;               /* -> Listenanfang  setzen         */
  else                                    /* sonst:                          */
    Result->Last->Next = Polygon;         /* -> Polygon hinten anhngen      */

  while (Polygon != NULL)
  {
    Result->Last = Polygon;               /* Last aktualisieren              */
    (Result->Number)++;                   /* Polygonanzahl erhhen           */
    Polygon = Polygon->Next;
  }
#endif

  return Result;
}


/* ************************************************************************* */
/* GEODeletePolygon                                                          */
/* Removes a polyeder's polygon.                                             */
/* ************************************************************************* */
void GEODeletePolygon (POLYEDER *Polyeder, POLYGON *Polygon)
{
  POLYGON *Ptr, *Last = NULL;

  if (Polyeder == NULL || Polygon == NULL)
    return;

  Ptr = Polyeder->List;
  while (Ptr != NULL)
  {
    if (Ptr == Polygon)
    {
      if (Last == NULL)                 /* First polygon in the list */
	Polyeder->List = Ptr->Next;
      else
	Last->Next = Ptr->Next;

      if (Ptr->Next == NULL)            /* Update last-pointer       */
	Polyeder->Last = Last;

      (Polyeder->Number)--;             /* reduce polygon amount     */
      GEOCalcPolyederInfo (Polyeder);
      return;
    }
    Last = Ptr;
    Ptr = Ptr->Next;
  }
}



/* ************************************************************************* */
/* GEOCopyPolyeder                                                           */
/* Create a copy of a polyeder.                                              */
/* ************************************************************************* */
POLYEDER *GEOCopyPolyeder (POLYEDER *Polyeder, VECTOR3D *VertexList)
{
  POLYGON  *Ptr, *Polygon;
  POLYEDER *Result;

  if (Polyeder == NULL)
    return NULL;

  Result = GEONewPolyeder ();

  Result->Number = Polyeder->Number;
  Result->List   = NULL;

  Ptr = Polyeder->List;
  while (Ptr != NULL)     /* Copy all polygons */
  {
    Polygon = GEOCopyPolygon (Ptr);
    if (VertexList != NULL)
      Polygon->Points = VertexList;

    if (Result->List == NULL)
      Result->List = Polygon;
    else
      Result->Last->Next = Polygon;
    Result->Last = Polygon;

    Ptr = Ptr->Next;
  }

  return Result;
}


/* ********************************************************************** */
/* ** Locates the polygon in the given polyeder, with the specified    ** */
/* ** Id-number.                                                       ** */
/* ** INPUT:  Polyeder   contains all it's polygons.                   ** */
/* **         ID         Number of the searched polygon.               ** */
/* ** OUTPUT: (POLYGON*) Pointer to the found polygon or NULL if not   ** */
/* **                    found.                                        ** */
/* ********************************************************************** */
POLYGON *GEOGetPolygon (POLYEDER *Polyeder, int ID)
{
  POLYGON *Polygon;

  Polygon = Polyeder->List;
  while (Polygon != NULL)
  {
    if (Polygon->Info.ID == ID)
      return Polygon;
    Polygon = Polygon->Next;
  }
  return NULL;
}


/* ********************************************************************** */
/* ** Checks, if the polyeder contains a polygon, which includes the   ** */
/* ** point (represented by the x-, y-, and z-component).              ** */
/* ** INPUT:  Polyeder    Polyederdata containing all the polygons     ** */
/* **         x, y, z     Coordinates of the point.                    ** */
/* ** OUTPUT: (POLYGON*)  If a polygon was found this is the pointer   ** */
/* **                     to it, otherwise a NULL-pointer will be      ** */
/* **                     returned.                                    ** */
/* ********************************************************************** */
POLYGON *GEOFindPolygon (POLYEDER *Polyeder, double x, double y, double z)
{
  double    Dist;
  VECTOR3D  p;
  POLYGON  *Polygon;

  p = GEOMakeVector (x, y, z);

  Polygon = Polyeder->List;
  while (Polygon != NULL)
  {
    Dist = GEOAreaPointDist (Polygon, p);

    if (fabs(Dist) < D_DIST && GEOPointInPolygon(Polygon, p))
    {
      PRO1 ("FindPolygon: Dist = %f\n", Dist)
      return Polygon;
    }

    Polygon = Polygon->Next;
  }

  return NULL;
}


/* ************************************************************************* */
/* GEOFreePolyeder                                                           */
/* gibt von Polyeder belegten Speicher frei.                                 */
/* ************************************************************************* */
void GEOFreePolyeder (POLYEDER *Polyeder)
{
  POLYGON *Ptr, *Next;

  if (Polyeder == NULL)
    return;

  Ptr = Polyeder->List;
  while (Ptr != NULL)                       /* Polygonliste freigeben        */
  {
    Next = Ptr->Next;
    GEOFreePolygon (Ptr, FALSE);
    Ptr = Next;
  }

  free (Polyeder);
}


/* ************************************************************************* */
/* GEOPolyederInfo                                                           */
/* berechnet Volumen, Oberflche und Schwerpunkt des bergebene Polyeder.    */
/* ************************************************************************* */
INFO_3D GEOCalcPolyederInfo (POLYEDER *Polyeder)
{
  int       count, No;
  double    Volume, Area;
  VECTOR3D  Center;
  VECTOR3D *Points;
  POLYGON  *Polygon = NULL;
  INFO_3D   Result;

  PRO_START (" ")

  Result.Center = Vector_0;
  Result.Min    = Vector_0;
  Result.Max    = Vector_0;
  Result.Area   = 0.0;
  Result.Volume = 0.0;
  Result.Bubble = 0.0;

  Polygon = Polyeder->List;

  /* Initialize min. and max. polyeder size */
  if (Polygon != NULL && Polygon->Number > 0)
    Result.Min = Result.Max = Polygon->Points[Polygon->Vertex[0]];

  while (Polygon != NULL)
  {
    Points = Polygon->Points;

    if (Polygon->Triangle != NULL)
    {
      Polygon->Info.Area = 0.0;
      for (count = 0; count < Polygon->Number-2; ++count)
      {
	GEOTetraederVolume (Points[Polygon->Triangle[count][0]], 
			    Points[Polygon->Triangle[count][1]], 
			    Points[Polygon->Triangle[count][2]], 
			    &Area, &Volume);
	Polygon->Info.Area += Area;              /* area of polygon    */
	Result.Volume += Volume;                 /* volume of polyeder */
	Result.Area   += Area;                   /* area of polyeder   */

	Center = GEOTetraederCenter (Points[Polygon->Triangle[count][0]], 
				     Points[Polygon->Triangle[count][1]], 
				     Points[Polygon->Triangle[count][2]]);
	Result.Center.x += Center.x * Volume;
	Result.Center.y += Center.y * Volume;
	Result.Center.z += Center.z * Volume;
      }
    }
    else
      PRO0 ("Polygon is not triangulated !\n")
	
    /* Compute min and max-coordinates of polyeder */
    for (count = 0; count < Polygon->Number; ++count)
    {
      No = Polygon->Vertex[count];
      if (Points[No].x < Result.Min.x)
	Result.Min.x = Points[No].x;
      if (Points[No].x > Result.Max.x)
	Result.Max.x = Points[No].x;
      if (Points[No].y < Result.Min.y)
	Result.Min.y = Points[No].y;
      if (Points[No].y > Result.Max.y)
	Result.Max.y = Points[No].y;
      if (Points[No].z < Result.Min.z)
	Result.Min.z = Points[No].z;
      if (Points[No].z > Result.Max.z)
	Result.Max.z = Points[No].z;
    }

    GEOCalcHesseForm (Polygon);

    Polygon = Polygon->Next;
  }

  if (Result.Volume != 0.0)
  {
    Result.Center.x /= Result.Volume;
    Result.Center.y /= Result.Volume;
    Result.Center.z /= Result.Volume;
  }

  PRO_END
  return Result;
}


#if 0
void GEOPolyederNormals (POLYEDER *Polyeder, int *Amount)
{
  int      count, No;
  VERTEX  *Points;
  POLYGON *Polygon = NULL;

  PRO_START (" ")

  Polygon = Polyeder->List;
  while (Polygon != NULL)
  {
    Points = Polygon->Points;

    for (count = 0; count < Polygon->Number; ++count)
    {
      No = Polygon->Vertex[count];
      if ((Points[No].Normal.x == 0.0 && Points[No].Normal.y == 0.0 &&
	  Points[No].Normal.z == 0.0) || 
	  GEOVectorAngle(Points[No].Normal, Polygon->Info.n) < 15.0)
	Points[No].Normal = GEOVectorAdd (Points[No].Normal, Polygon->Info.n);
      else
	Points[No].Normal = Vector_0;

      Amount[No]++;
    }
    Polygon = Polygon->Next;
  }

  PRO_END
}
#endif


/* ********************************************************************** */
/* ** This function returns both polygons, which posseses the edge     ** */
/* ** Point1->Point2 or vice versa.                                    ** */
/* ** INPUT:  Polyeder     Polyeder-data                               ** */
/* **         Point1/2    Index into point-list of the vertexes        ** */
/* ** INOUT:  Polygon1/2  Pointer onto the resulting polygons          ** */
/* ** OUTPUT: (BOOL)      TRUE, if two polygons using this edge exists ** */
/* **                     FALSE, otherwise.                            ** */
/* ** NOTE:   This function does only find the polygons, if their edge ** */
/* **         has the same indizes. That means also, that the polygons ** */
/* **         will not be found if their common edge has the same      ** */
/* **         coordinates but diffeneret point indizes !               ** */
/* ********************************************************************** */   
dual GEOGetEdgePolygons (POLYEDER *Polyeder, int Point1, int Point2, 
			 POLYGON **Polygon1, POLYGON **Polygon2)
{
  dual     found;
  int      count, Number = 0;
  POLYGON *Polygon;

  Polygon = Polyeder->List;
  while (Polygon != NULL && Number < 2)
  {
    found = FALSE;
    count = 0;
    while (! found && count < Polygon->Number-1)
    {
      found = ((Polygon->Vertex[count] == Point1 && 
		Polygon->Vertex[count+1] == Point2) ||
	       (Polygon->Vertex[count] == Point2 &&
		Polygon->Vertex[count+1] == Point1));
      count++;
    }
    if (found)
    {
      if (Number == 0)
	*Polygon1 = Polygon;
      else
	*Polygon2 = Polygon;
      Number++;
    }
    Polygon = Polygon->Next;
  }

  return (Number == 2);
}


POLYGON *GEOFindNeighbour (POLYEDER *Polyeder, POLYGON *Polygon,
			   int Index1, int Index2)
{
  int      count;
  POINT3D  Point1A, Point1B;
  POINT3D  Point2A, Point2B;
  POINT3D *Points;
  POLYGON *Poly;

  if (Index1 < 0 || Index1 >= Polygon->Number || 
      Index2 < 0 || Index2 >= Polygon->Number)
    return NULL;

  Point1A = Polygon->Points[Polygon->Vertex[Index1]];
  Point1B = Polygon->Points[Polygon->Vertex[Index2]];

  Poly = Polyeder->List;
  while (Poly != NULL)
  {
    if (Poly != Polygon)
    {
      Points = Poly->Points;
      for (count = 0; count < Poly->Number; ++count)
      {
	Point2A = Points[Poly->Vertex[count]];
	Point2B = Points[Poly->Vertex[(count+1) % Poly->Number]];
	if ((VEC_EQUAL(Point2A, Point1A) && VEC_EQUAL(Point2B, Point1B)) ||
	    (VEC_EQUAL(Point2A, Point1B) && VEC_EQUAL(Point2B, Point1A)))
	  return Poly;
      }
    }
    Poly = Poly->Next;
  }

  return NULL;
}


/* ********************************************************************** */
/* ** Free the normal list of each polygon in the polyeder             ** */
/* ** INPUT:   Polyeder   Polyeder data                                ** */
/* ********************************************************************** */   
void GEOFreePolyederNormals (POLYEDER *Polyeder)
{
  POLYGON *Polygon;

  Polygon = Polyeder->List;
  while (Polygon != NULL)
  {
    GEOFreePolygonNormals (Polygon);
    Polygon = Polygon->Next;
  }
}


/* ********************************************************************** */
/* ** For each point in the point-list create a list with all polygons ** */
/* ** of this polyeder which contains this point (index)               ** */
/* ** INPUT:   Polyeder    Polyeder data                               ** */
/* ** IN/OUT:  Pointlist   List of object points                       ** */
/* ********************************************************************** */   
void GEOBuildPolygonList  (POLYEDER *Polyeder, POINTELEM *PointList)
{
  int          count, No;
  VECTOR3D    *Points;
  POLYGON     *Polygon = NULL;
  POLYGONELEM *Elem;

  Polygon = Polyeder->List;
  while (Polygon != NULL)
  {
    Points = Polygon->Points;
    for (count = 0; count < Polygon->Number; ++count)
    {
      No = Polygon->Vertex[count];

      Elem = (POLYGONELEM *)malloc (sizeof(POLYGONELEM));
      Elem->Polygon = Polygon;
      Elem->done    = FALSE;
      Elem->Next    = NULL;

      if (PointList[No].PolygonList == NULL)
	PointList[No].PolygonList = Elem;
      else
	PointList[No].Last->Next = Elem;
      PointList[No].Last = Elem;

      (PointList[No].Number)++;
    }
    Polygon = Polygon->Next;    
  }
}


int GEOGetTriangles (POLYEDER *Polyeder)
{
  int      Result = 0;
  POLYGON *Polygon;

  PRO_START ("")

  if (Polyeder == NULL)
    PRO_RETURN ("empty polyeder", 0)

  Polygon = Polyeder->List;
  while (Polygon != NULL)
  {
    Result += Polygon->Number - 2;
    Polygon = Polygon->Next;
  }

  PRO_END
  return Result;
}

