#include <math.h>

#include "libgeo.h"


H_MATRIX GEOTranslate (H_MATRIX M, VECTOR3D Vec)
{
  H_MATRIX Result;

  Result.m[P][X] += Vec.x;
  Result.m[P][Y] += Vec.y;
  Result.m[P][Z] += Vec.z;

  return Result;
}


/* ********************************************************************** */
/* ** Create a tranformation-matrix with offsets in x-, y, and         ** */
/* ** z-direction.                                                     ** */
/* ********************************************************************** */
H_MATRIX GEOMakeTrans (double dx, double dy, double dz)
{
  H_MATRIX Result = Matrix_1;

  Result.m[P][X] = dx;
  Result.m[P][Y] = dy;
  Result.m[P][Z] = dz;

  return Result;
}


/* ********************************************************************** */
/* ** Create a transformation matrix which is rotated around the given ** */
/* ** direction-vector. This vector must be normalized.                ** */
/* ********************************************************************** */
H_MATRIX GEOMakeRot (VECTOR3D Dir, double Angle)
{
  double s, c;
  double komp;
  H_MATRIX Result = Matrix_1;

  s = sin (RAD(Angle));
  c = cos (RAD(Angle));
  komp = 1.0 - c;

  if (Dir.y == 0.0 && Dir.z == 0.0)     /* rotation around x-axis         */
  {
    Result.m[O][Y] = c;
    Result.m[O][Z] = s;
    Result.m[A][Y] = -s;
    Result.m[A][Z] = c;
  }
  elsif (Dir.x == 0.0 && Dir.z == 0.0)  /* rotation around y-axis         */
  {
    Result.m[N][X] = c;
    Result.m[N][Z] = -s;
    Result.m[A][X] = s;
    Result.m[A][Z] = c;
  }
  elsif (Dir.x == 0.0 && Dir.y == 0.0)  /* rotation around z-axis         */
  {
    Result.m[N][X] = c;
    Result.m[N][Y] = s;
    Result.m[O][X] = -s;
    Result.m[O][Y] = c;
  }
  else                                  /* complex rotation               */
  {
    Result.m[N][X] = Dir.x * Dir.x * komp + c;
    Result.m[N][Y] = Dir.x * Dir.y * komp + Dir.z * s;
    Result.m[N][Z] = Dir.x * Dir.z * komp - Dir.y * s;
    Result.m[O][X] = Dir.y * Dir.x * komp - Dir.z * s;
    Result.m[O][Y] = Dir.y * Dir.y * komp + c;
    Result.m[O][Z] = Dir.y * Dir.z * komp + Dir.x * s;
    Result.m[A][X] = Dir.z * Dir.x * komp + Dir.y * s;
    Result.m[A][Y] = Dir.z * Dir.y * komp - Dir.x * s;
    Result.m[A][Z] = Dir.z * Dir.z * komp + c;
  }

  return Result;
}


H_MATRIX GEOMakeTransform (H_MATRIX M1, H_MATRIX M2)
{
  H_MATRIX M1Inv, Result;

  M1Inv = GEOMatrixInv (M1);
 
  Result = GEOMatrixProd (M1Inv, M2);

  return Result;
}
