#include "libgeo.h"


POLYGON *ReadPolygon (FILE *fp, int ID)
{
  int      count;
  char     letter;
  char     Word[10];
  POLYGON *Polygon;

  Polygon = GEONewPolygon ();

  /* first lines may be comments */
  do
  {
    fscanf (fp, "%s", Word);
    do 
      letter = fgetc (fp);
    while (letter != '\n');
  } while (Word[0] == '#');
  Polygon->Number = atoi (Word);
    
  Polygon->Points = (POINT3D *)malloc(sizeof(POINT3D) * Polygon->Number);
  Polygon->Vertex =     (int *)malloc(sizeof(int)     * Polygon->Number);
  for (count = 0; count < Polygon->Number; ++count)
  {
    fscanf (fp, "%lf %lf %lf\n", &Polygon->Points[count].x,
	                         &Polygon->Points[count].y,
	                         &Polygon->Points[count].z);
    Polygon->Vertex[count] = count;
  }

  GEOCalcHesseForm (Polygon);
  Polygon->Info.ID = ID;
  Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);

  return Polygon;
}

void Frame2Euler ()
{
  H_MATRIX Matr;
  EULER    Euler;
 
  Matr = GEORPY2Frame (343.0, 47.0, 100.0, -180.0, 0.0, -90.0);
  GEOFrame2Euler (Matr, &Euler.px, &Euler.py, &Euler.pz, 
	                &Euler.z1, &Euler.y, &Euler.z2);
  WRITE_HMATRIX (stderr, Matr, "Matrix")
  printf ("Euler-Angle: %.4f  %.4f  %.4f\n", Euler.z1, Euler.y, Euler.z2);
}


void LineBubbleCut ()
{
  POINT3D  M, Pos;
  VECTOR3D Dir;

  /* Cut between a line and a bubble */
  M   = GEOMakeVector (1, 1, 1);
  Pos = GEOMakeVector (0, 0, 0);
  Dir = GEOMakeVector (2, 1, 1);
  printf ("%d Results\n", GEOComputeBubbleCut(M, 1, Pos, Dir, NULL, NULL));
}


void SolveLinearProgram ()
{
  LGS MA, Mx;

  /* Compute the solution of a linear program */
  MA.Size = 3;
  MA.Field[0][0] =  3;
  MA.Field[0][1] =  8;
  MA.Field[0][2] = -1;
  MA.Field[0][3] =  1;
  MA.Field[1][0] =  5;
  MA.Field[1][1] =  6;
  MA.Field[1][2] =  0;
  MA.Field[1][3] = 23;
  MA.Field[2][0] =  1;
  MA.Field[2][1] =  7;
  MA.Field[2][2] =  1;
  MA.Field[2][3] = -3;
  GEOShowLGS ("One solution", MA);
  Mx = GEOSolveLGS (MA);
  GEOPrintLGS (Mx);

  MA.Size = 2;
  MA.Field[0][0] =  3;
  MA.Field[0][1] =  4;
  MA.Field[0][2] =  5;
  MA.Field[1][0] =  6;
  MA.Field[1][1] =  8;
  MA.Field[1][2] =  7;
  GEOShowLGS ("No solution", MA);
  Mx = GEOSolveLGS (MA);
  GEOPrintLGS (Mx);

  MA.Size = 2;
  MA.Field[0][0] =  1;
  MA.Field[0][1] =  2;
  MA.Field[0][2] =  3;
  MA.Field[1][0] =  2;
  MA.Field[1][1] =  4;
  MA.Field[1][2] =  6;
  GEOShowLGS ("No solution", MA);
  Mx = GEOSolveLGS (MA);
  GEOPrintLGS (Mx);

  MA.Size = 3;
  MA.Field[0][0] =  2;
  MA.Field[0][1] = -2;
  MA.Field[0][2] =  0;
  MA.Field[0][3] = -4;
  MA.Field[1][0] =  1;
  MA.Field[1][1] = -1;
  MA.Field[1][2] =  0;
  MA.Field[1][3] = -2;
  MA.Field[2][0] =  0;
  MA.Field[2][1] = -1;
  MA.Field[2][2] =  0;
  MA.Field[2][3] =  0;
  GEOShowLGS ("One solution", MA);
  Mx = GEOSolveLGS (MA);
  GEOPrintLGS (Mx);

}

void LineLineDistance ()
{
  POINT3D  Pos1, Pos2;
  POINT3D  Point1, Point2;
  VECTOR3D Dir1, Dir2;

  /* Compute the minimum distance and the nearest point between two lines */
  Pos1 = GEOMakeVector (0.0, 0.0, 0.0);
  Dir1 = GEOMakeVector (0.0, 0.0, 1.0);
  Pos2 = GEOMakeVector (1.0, 0.0, 0.0);
  Dir2 = GEOMakeVector (2.0, 1.0, 0.0);  
  printf ("Line-Dist:      %f\n", GEOLineLineDist(Pos1, Dir1, Pos2, Dir2));
  GEOLineLineLot (Pos1, Dir1, Pos2, Dir2, &Point1, &Point2);
  printf ("Nearest Points:\n");
  printf ("(%.5f / %.5f / %.5f)\n", Point1.x, Point1.y, Point1.z);
  printf ("(%.5f / %.5f / %.5f)\n", Point2.x, Point2.y, Point2.z);
}


int main (int argc, char **argv)
{
  FILE      *fp = NULL;
  POLYGON   *Polygon1 = NULL;
  POLYGON   *Polygon2 = NULL;
  INTERSECT *Contact = NULL;

#if 0
  Frame2Euler ();
#endif

#if 0
  LineBubbleCut ();
#endif

#if 0
  LineLineDistance ();
#endif

#if 0
  SolveLinearProgram ();
#endif

#if 1
  /* Compute the contact between two polygons */
  if (argc < 2)
  {
    printf ("Syntax: tstgeo <data-file>\n");
    exit (1);
  }

  if ((fp = fopen(argv[1], "r")) == NULL)
  {
    printf ("Error opening data-file\n");
    exit (1);
  }

  /* 
     Dateien: Nur schneller Hack zum Test der Kontakt-Berechnung
     - daher keine Fehlerprfung !
     Datei-Format:
         # Comment
         Number of vertexes in Polygon1
	 x11 y11 z11
	 x12 y12 z12
	 ...
	 Number of vertexes in Polygon2
	 x21 y21 z21
	 x22 y22 z22
	 ...
  */

  Polygon1 = ReadPolygon (fp, 1);
  Polygon2 = ReadPolygon (fp, 2);

  Contact = GEOPolygonIntersection (Polygon1, Polygon2);
  /* 
     GEOPrintPolygon (Contact);
  */

#if 0
  GEOTriangulatePolygon (Polygon1);
  GEOTriangulatePolygon (Polygon2);

  if ((Contact = GEOPolygonContact (Polygon1, Polygon2)) != NULL)
  {
    printf ("Contact-Polygon computed: %d\n", Contact->Number);
    for (count = 0; count < Contact->Number; ++count)
      printf ("%f %f %f\n", Contact->Points[count].x,
	      Contact->Points[count].y, Contact->Points[count].z);

    free (Contact->Points);
    GEOFreePolygon (Contact, TRUE);
  }
  else
    printf ("No Contact-Polygon found");
#endif

  GEOFreePolygon (Polygon1, TRUE);
  GEOFreePolygon (Polygon2, TRUE);
#endif
  return 0;
}
