#ifndef _HELP_GLOBAL_H
#define _HELP_GLOBAL_H

#include <Xm/PushB.h>
#include <Xm/ToggleB.h>
#include <Xm/Text.h>

#include "libhelp.h"


#define HTML_BEHAVE   FALSE             /* Newline after headings ?       */
#define DOUBLE_CLICK  FALSE             /* single/double click on ref's   */

#define SCALE         0                 /* Default font-scale             */
#define MARK_REFS     1                 /* Mark refs with small-pixmap    */
#define SHADOW_REF    0                 /* Draw HREF's with shadow        */
#define BUTTON_REF    0                 /* Draw Text-HRefs as button      */
#define MOUSE_DRAG    +                 /* direction of mouse-dragging    */
                                        /* + means pulling the document   */

#if BUTTON_REF
#if ! SHADOW_REF
#undef SHADOW_REF
#define SHADOW_REF 1
#endif
#endif


/* Drawing the text into a pixmap wastes a lot of memory. So the information
   will be stored in an interal data-structure which allows a fast redrawing
   on demand, but also causes an larger overhead if used over net.
   Decide, which version is best for you. The allowed combinations are:
   USE_PIXMAP USE_LABEL   description
   ------------------------------------------------------------------------
        0        0        draws data into a drawin-area by expose-ev.
        1        0        use pixmap for backup-data from which the
                          needed data will be copied
        1        1        use a label-widget, that means: let X/motif
                          manage the refresh
        0        1        not allowed (changing label makes no sense)
*/

#define USE_PIXMAP  0
#define USE_LABEL   0

#if USE_LABEL
#if ! USE_PIXMAP
#undef USE_PIXMAP
#define USE_PIXMAP 1
#endif
#endif


#define MAX_URI       5                 /* max. action-URIs in a document */
#define MAX_SECT      5                 /* depth of subsection hierarchie */
#define MAX_INDENT   10                 /* max. indent depth storage      */ 

#define HELP_CASE     1
#define HELP_SUBSTR   2
#define HELP_GLOBAL   4

#define DEF_METHOD  (HELP_SUBSTR)       /* Default search method          */


/* Elem types: */
#define ELEM_TEXT   (1L<<0)             /* text               */
#define ELEM_VSPACE (1L<<1)             /* <BR>/<P>           */
#define ELEM_LINE   (1L<<2)             /* <HR>               */
#define ELEM_NAME   (1L<<3)             /* label              */
#define ELEM_PRE    (1L<<4)             /* preformatted text  */
#define ELEM_ITEM   (1L<<5)             /* item               */
#define ELEM_ENUM   (1L<<6)             /* enumeration        */
#define ELEM_IMG    (1L<<7)             /* images             */
#define ELEM_SECT   (1L<<8)             /* section            */
#define ELEM_IND    (1L<<9)             /* indent             */
#define ELEM_NIND   (1L<<10)            /* no indent          */
#define ELEM_REF    (1L<<11)            /* reference          */
#define ELEM_CLEAR  (1L<<12)            /* jump to free line  */
#define ELEM_HSPACE (1L<<13)            /* HSPACE             */
#define ELEM_TABLE  (1L<<14)            /* Table              */
#define ELEM_INPUT  (1L<<15)            /* input-field        */


/* Definitions used for proving help-text-correctness */
enum TEST_BITS
{
  EM_BIT = 0,
  BD_BIT,
  REF_BIT,
  INDENT_BIT,
  COLOR_BIT,
  ULINE_BIT,
  FORM_BIT
};


/* helpSearch.c: */
#if SEARCHING
  dual        HELP_LocateHit       (HELPDATA *Data, int HitNo);
  void        HELP_DoSearchCB      (Widget widget, HELPDATA *Data, 
				                   XmAnyCallbackStruct *Ptr);
  int         HELP_DataSearch      (HELPDATA *Data, char *Pattern, int Method);
#endif

/* helpMain.c: */
void        HELP_OpenDlg         (Widget Parent, char *Name, 
				  char *Msg, char **Image);
void        HELP_MarkHRef        (HELPDATA *Data, int ID);
void        HELP_Jump            (HELPDATA *Data, const char *FileName, 
				                  char *Label);
void        HELP_RefreshText     (HELPDATA *Data, int start, int end);
void        HELP_UpdateWindow    (HELPDATA *Data, char *Label, int PosY, 
				                  dual Refresh);


#endif
