/* ********************************************************************** */
/* ** HelpWindow function                                              ** */
/* ** (c) 1998/1999 Stefan Vietinghoff                                 ** */
/* ********************************************************************** */

/* 
   Warning: 
   Do not try to understand this program - it might be too confusing !
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>

#include <X11/keysym.h>

#include <Xm/DialogS.h>
#include <Xm/MessageB.h>

#include <Xm/MainW.h>
#include <Xm/MwmUtil.h>
#include <Xm/Form.h>
#include <Xm/ScrolledW.h>
#include <Xm/DrawingA.h>
#include <Xm/Separator.h>
#include <Xm/ScrollBar.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>

#include <Xm/ToggleB.h>

#include "helpGlobal.h"

#include "helpicon.xpm"
#include "badimg.xpm"
#if MARK_REFS
#include "refmark.xpm"
#endif


#if USE_PIXMAP
#define GRAPHIC Data->Pic
#else
#define GRAPHIC Data->Win
#endif

#define MIN_WIDTH   400                 /* use scroll-bar if smaller      */

#define MARGIN        5                 /* window- and image-margin       */
#define KEY_STEP      6                 /* scroll-step with arrow-keys    */
#define BIG_SKIP     10                 /* <P> = <BR> + BIG_SKIP          */

#define HELP_NO_REF   -1                /* ID for non-reference-types     */


enum HELP_ALIGN
{
  HELP_LEFT = 0,
  HELP_CENTER,
  HELP_RIGHT,
  HELP_FILL
};

enum CMD_CODE                                      /* supported commands: */
{
  NO_CMD = -1,                                       /* No/bad command    */
  E_CMD,                                             /* end-of-command    */

  TEXT_CMD,                                          /* normal text       */

  TITLE_S_CMD,                                       /* title (Start/End) */
  TITLE_E_CMD,

  BR_CMD,                                            /* newline           */
  P_S_CMD,                                           /* paragraph start   */
  P_E_CMD,                                           /* paragraph_end     */
  HR_CMD,                                            /* hor. line         */

  EM_S_CMD,                                          /* emphasized        */
  EM_E_CMD,
  BD_S_CMD,                                          /* bold              */
  BD_E_CMD,
  BG_CMD,                                            /* background image  */
  BGCOLOR_CMD,                                       /* background color  */
  HREF_CMD,                                          /* reference         */
  NAME_CMD,                                          /* label             */
  A_E_CMD,                                           /* special cmd (End) */
  REM_CMD,                                           /* remark/comment    */
  ITEM_S_CMD,                                        /* itemize           */
  ITEM_E_CMD,
  ENUM_S_CMD,                                        /* enumeration       */
  ENUM_E_CMD,
  ITEM_CMD,                                          /* item              */
  QUOTE_S_CMD,                                       /* quotation         */
  QUOTE_E_CMD,
  IMG_CMD,                                           /* image             */
  CLEAR_CMD,                                         /* skip to free line */
  PRE_S_CMD,                                         /* preformated text  */
  PRE_E_CMD,
  COLOR_S_CMD,                                       /* color             */
  COLOR_E_CMD,
  ALIGN_CMD,                                         /* alignment         */
  ULINE_S_CMD,                                       /* underline         */
  ULINE_E_CMD,
  DL_S_CMD,                                          /* Descr list        */
  DL_E_CMD,
  DT_CMD,                                            /* descr title       */
  VSPACE_CMD,                                        /* vertical space    */
  HSPACE_CMD,                                        /* horizontal space  */
  TABLE_S_CMD,                                       /* table             */
  TABLE_E_CMD,
  FORM_S_CMD,                                        /* formular          */
  FORM_E_CMD,
  INPUT_CMD,                                         /* input             */
  SECTN_CMD,                                         /* (Sub)Section      */
  HN_S_CMD,                                          /* font size         */
  HN_E_CMD
};


enum INPUT_TYPES
{
  TEXTVALUE = 0,
  CHECKBOX,
  SUBMIT
  /* SELECT, TEXTAREA, RADIO */
};


/* various types and definitions for parsing the text...                  */

typedef struct
{
  int     SizeX, SizeY;
  XImage *Image;
  dual    loaded;
  char    alt[LABELLEN+1];
} IMG_DATA;


typedef struct
{
  int    Type;
  char   Name[LABELLEN+1];
  int    Size;
  char   Value[LABELLEN+1];
  dual   checked;
  Widget widget;
} INPUT_DATA;


typedef struct
{
  byte   Attr[5];                                 /* Font attributes      */
  byte   ColorID;                                 /* Color index          */
  byte   Sect[MAX_SECT];                          /* sections 1..5        */
  dual   Head;                                    /* Chapter Headline     */
  dual   skip;                                    /* skip on DT           */ 
  short  ID;                                      /* HREF-ID              */
  int    Number;                                  /*                      */
  byte   align;                                   /* alignment            */
  void  *Param;                                   /* misc params          */
} HELP_ATTR;

typedef struct
{
  char  Key[20];
  short Code;
} HELP_KEYWORD;


/* ********************************************************************** */
/* ** corresponding keywords (in alphabetic order) If two commands     ** */
/* ** begins with the same letters, we need to define the longer one   ** */
/* ** first ).                                                         ** */
/* ** If you want to change the keywords for some formattings, change  ** */
/* ** them here in the left column (e.g. if you want to use <U> for    ** */
/* ** underlined text, then you need to modify or copy+change the two  ** */
/* ** UNDERLINE-lines)                                                 ** */
/* ********************************************************************** */
static HELP_KEYWORD Keys[] =
{
  {"A HREF",      HREF_CMD},
  {"A NAME",      NAME_CMD},
  {"ALIGN",       ALIGN_CMD},
  {"/A",          A_E_CMD},
  {"BACKGROUND",  BG_CMD},
  {"BD",          BD_S_CMD},
  {"/BD",         BD_E_CMD},
  {"BF",          BD_S_CMD},
  {"/BF",         BD_E_CMD},
  {"BGCOLOR",     BGCOLOR_CMD},
  {"BGIMAGE",     BG_CMD},
  {"BLOCKQUOTE",  QUOTE_S_CMD},
  {"/BLOCKQUOTE", QUOTE_E_CMD},
  {"BQ",          QUOTE_S_CMD},
  {"/BQ",         QUOTE_E_CMD}, 
  {"BR",          BR_CMD},
  {"B",           BD_S_CMD},
  {"/B",          BD_E_CMD},
  {"CLEAR",       CLEAR_CMD},
  {"CODE",        PRE_S_CMD},
  {"/CODE",       PRE_E_CMD},
  {"COLOR",       COLOR_S_CMD},
  {"/COLOR",      COLOR_E_CMD},
  {"DD",          QUOTE_S_CMD},
  {"DL",          DL_S_CMD},
  {"/DL",         DL_E_CMD},
  {"DT",          DT_CMD},
  {"EM",          EM_S_CMD},
  {"/EM",         EM_E_CMD},
  {"ENUM",        ENUM_S_CMD},
  {"/ENUM",       ENUM_E_CMD},
  {"FORM",        FORM_S_CMD},
  {"/FORM",       FORM_E_CMD},
  {"HR",          HR_CMD},
  {"HSPACE",      HSPACE_CMD},
  {"IMG",         IMG_CMD},
  {"INPUT",       INPUT_CMD},
  {"ITEMIZE",     ITEM_S_CMD},
  {"/ITEMIZE",    ITEM_E_CMD},
  {"ITEM",        ITEM_CMD},
  {"IT",          EM_S_CMD},
  {"/IT",         EM_E_CMD},
  {"I",           EM_S_CMD},
  {"/I",          EM_E_CMD},
  /* LH.../HL LIST HEADER */
  {"LI",          ITEM_CMD},
  {"OL",          ENUM_S_CMD},
  {"/OL",         ENUM_E_CMD},
  {"PRE",         PRE_S_CMD},
  {"/PRE",        PRE_E_CMD},
  {"P",           P_S_CMD},
  {"/P",          P_E_CMD},
  {"QUOTE",       QUOTE_S_CMD},
  {"/QUOTE",      QUOTE_E_CMD},
  {"STRONG",      BD_S_CMD},
  {"/STRONG",     BD_E_CMD},
  {"TABLE",       TABLE_S_CMD},
  {"/TABLE",      TABLE_E_CMD},
  {"TEXT",        COLOR_S_CMD},
  {"/TEXT",       COLOR_E_CMD},
  {"TITLE",       TITLE_S_CMD},
  {"/TITLE",      TITLE_E_CMD},
  {"UL",          ITEM_S_CMD},
  {"/UL",         ITEM_E_CMD},
  {"UNDERLINE",   ULINE_S_CMD},
  {"/UNDERLINE",  ULINE_E_CMD},
  {"U",           ULINE_S_CMD},
  {"/U",          ULINE_E_CMD},
  {"VERBATIM",    PRE_S_CMD},
  {"/VERBATIM",   PRE_E_CMD},
  {"VSPACE",      VSPACE_CMD},
  {"!--",         REM_CMD}
};

#define KEYWORDS (int)(sizeof(Keys)/sizeof(HELP_KEYWORD))


/* ------------------- */
/* -- S Y M B O L S -- */
/* ------------------- */
typedef struct
{
  char Key[20];
  char Symbol;
} HELP_SYMBOL; 

/* supported symbols - add new symbols, if needed                         */
static HELP_SYMBOL Symbols[] = 
{
  {"Auml",       ''},
  {"Ouml",       ''},
  {"Uuml",       ''},
  {"amp",        '&'},
  {"auml",       ''},
  {"brvbar",     ''},
  {"copy",       ''},
  {"deg",        ''},
  {"frac14",     ''},
  {"frac12",     ''},
  {"frac34",     ''},
  {"gt",         '>'},
  {"micro",      ''},
  {"nbsp",       ' '},
  {"not",        ''},
  {"ouml",       ''},
  {"laquo",      ''},
  {"lt",         '<'},
  {"plusmn",     ''},
  {"quot",       '"'},
  {"raquo",      ''},
  {"reg",        ''},
  {"sect",       ''},
  {"sup2",       ''},
  {"sup3",       ''},
  {"szlig",      ''},
  {"times",      ''},
  {"uuml",       ''},
  {"#09",       '\t'}
};

#define SYMBOLS (int)(sizeof(Symbols)/sizeof(HELP_SYMBOL))


/* --------------- */
/* -- F O N T S -- */
/* --------------- */
#define FONT_MASK "-adobe-helvetica-%s-%s-*-*-%s-*-*-*-*-*-iso8859-1"


/* Font attributes */
enum FONT_ATTR
{
  FONT_WEIGHT = 0,
  FONT_SLANT,
  FONT_SIZE,
  FONT_PRE,
  FONT_UL
};

#define WEIGHT_DEF 0
#define SLANT_DEF  0
#define SIZE_DEF   2
#define PRE_DEF    0
#define UL_DEF     0

#define FONT_ATTRIBS 3

static char  FontMask[128] = FONT_MASK;
static char *Weight[]  = {"medium", "bold"};
static char *Slant[]   = {"r",      "o"};
static char *SizeTab[] = {"50", "40", "34", "25", "20", 
			  "18", "15", "12", "10"};

#define WEIGHT_NO (sizeof(Weight)/sizeof(char*))
#define SLANT_NO  (sizeof(Slant)/sizeof(char*))
#define SIZE_NO   (sizeof(SizeTab)/sizeof(char*))

static int AttrNo[FONT_ATTRIBS] = {WEIGHT_NO, SLANT_NO, SIZE_NO};

#define FONT_AMOUNT (int)(WEIGHT_NO * SLANT_NO * SIZE_NO)

/* Look-Up-Table into size-table */
static byte Size[][4] = {{5, 6, 7, 8},      /* Font-scale 0 (small)       */
			 {4, 5, 6, 7},
                         {2, 3, 5, 6},      /*            2 (medium)      */
                         {1, 2, 4, 5},
                         {0, 1, 3, 4}};     /*            4 (large)       */

#define SCALES (int)(sizeof(Size)/(sizeof(byte)*4))

typedef struct
{
  dual         loaded;
  byte         Attr[5];
  XFontStruct *DrawFont;
  byte         Space;                   /* width of space-character       */
  byte         Height;
} FONT_DATA;

static FONT_DATA Fonts[FONT_AMOUNT+1];

/* ----------------- */
/* -- C O L O R S -- */
/* ----------------- */
enum HELP_COLOR_ID                       /* insert your own colors here    */
{
  HELP_BG = 0,
  HELP_BLACK,
  HELP_WHITE,
  HELP_GREY25,
  HELP_GREY50,
  HELP_GREY75,
  HELP_BLUE,
  HELP_RED,
  HELP_GREEN,
  HELP_YELLOW,
  HELP_ORANGE,
  HELP_FGREEN,
  HELP_BROWN,
  HELP_GOLD
};

static char *ColorName[] = {"BACKGROUND",
                            "BLACK", "WHITE", "GREY25", "GREY50", "GREY75",
			    "BLUE",  "RED", "GREEN", "YELLOW", "ORANGE",
                            "FORESTGREEN", "BROWN", "GOLD"};

#define HELP_COLORS (int)(sizeof(ColorName)/sizeof(char*))

static int    ColorNo = HELP_COLORS;
static Pixel *Colors;


/* - - - - - - - - - - - - - -  */

static byte   Initialized = 0;
static int    MouseX, MouseY;
static GC     XOR_GC;
#if USE_PIXMAP
  static dual    done = FALSE;
#endif
#if MARK_REFS
  static dual    RefOK = FALSE;
  static XImage *RefPic = NULL;
  static int     RefX = 0, RefY = 0;
#endif
static HELPDATA *HelpData = NULL;


/* Prototypes: */
static void  HELP_GetSize       (HELPDATA *Data);
static void  HELP_HandleEvents  (Widget widget, HELPDATA *Data, 
			        XEvent *Event);
static void  HELP_FollowCB      (Widget widget, HELPDATA *Data, 
			        XmAnyCallbackStruct *Ptr);
static void  HELP_BackCB        (Widget widget, HELPDATA *Data, 
			        XmAnyCallbackStruct *Ptr);
static void  HELP_SubmitCB      (Widget widget, HELPDATA *Data,
			        XmAnyCallbackStruct *Ptr);
static void  HELP_ExitCB        (Widget widget, HELPDATA *Data, 
			        XmAnyCallbackStruct *Ptr);
static int   HELP_GetArg        (char *Text, 
			        char *Argument, char *Result, int Pos);
static void  HELP_FreeTable     (TABLE_DATA *Table);
static HELP_ELEM *HELP_MakeElem (int Type, void *Info, byte Scale,
				short space, HELP_ATTR *Attr);
#if 1
  static void  HELP_PrintText  ();
#endif


/* ********************************************************************** */
/* ** Why in the hell tells isspace that some characters are spaces.   ** */
/* ** This may be ok for english people, but what's about people       ** */
/* ** speaking German or other foreign languages ?                     ** */
/* ** Nevertheless, these function change this behavior for some       ** */
/* **  characters.                                                     ** */
/* ** INPUT:  letter  character to check                               ** */
/* ** OUTPUT: (dual)  TRUE if letter is space-char                     ** */
/* ********************************************************************** */
static dual HELP_IsSpace (char letter)
{
  return (isspace(letter) &&
	  letter != '' && letter != '' && letter != '' &&
	  letter != '' && letter != '');
}


#if 0
/* ********************************************************************** */
/* ** This function prints some size data. It is used for debug        ** */
/* ** purpose only.                                                    ** */
/* ** INPUT:  Data    Pointer to help-text data-structure              ** */
/* ** OUTPUT:                                                          ** */
/* ********************************************************************** */
static void HELP_PrintSizes (HELPDATA *Data)
{
  int       PosX, PosY;
  Dimension SizeX, SizeY;

  printf ("Size:    %d x %d\n", Data->SizeX, Data->SizeY);
  if (Data->Shell != NULL)
  {
    XtVaGetValues (Data->Shell, XmNwidth,  &SizeX, 
		                XmNheight, &SizeY,
		                NULL);
    printf ("Shell:    %d x %d\n", SizeX, SizeY);
  }
  printf ("Win:     %d / %d\n", Data->WinX, Data->WinY);
  printf ("Lines:   %d\n",      Data->Lines);
  if (Data->ScrWin != NULL)
  {
    XtVaGetValues (Data->ScrWin, XmNwidth,  &SizeX, 
		                 XmNheight, &SizeY,
		                 NULL);
    printf ("ScrWin:  %d x %d\n", SizeX, SizeY);
  }

  if (Data->ClipWin != NULL)
  {
    XtVaGetValues (Data->ClipWin, XmNwidth,  &SizeX, 
		                  XmNheight, &SizeY,
		                  NULL);
    printf ("ClipWin: %d x %d\n", SizeX, SizeY);
  }

  if (Data->Work != NULL)
  {
    XtVaGetValues (Data->Work, XmNwidth,  &SizeX, 
		               XmNheight, &SizeY,
		               NULL);
    printf ("Work:    %d x %d\n", SizeX, SizeY);
  }

  if (Data->ScrBarX != NULL)
  {
    XtVaGetValues (Data->ScrBarX, XmNvalue, &PosX, NULL);
    printf ("PosX:     %d\n", PosX);
  }

  if (Data->ScrBarY != NULL)
  {
    XtVaGetValues (Data->ScrBarX, XmNvalue, &PosY, NULL);
    printf ("PosY:     %d\n", PosY);
  }
}
#endif


/* ********************************************************************** */
/* ** Creates an dialog-box which displays a messages (e.g. file not   ** */
/* ** found). HELP_DlgCB is the callback-function which destroys the   ** */
/* ** dialog-shell.                                                    ** */
/* ** INPUT:  Parent      parent widget                                ** */
/* **         Name        name of the dialog-box                       ** */
/* **         Msg         text to display                              ** */
/* **         Image       optional image data or NULL for no image     ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
static void HELP_DlgCB (Widget widget, void *Data, XmAnyCallbackStruct *Ptr)
{
  XtDestroyWidget ((Widget)Data);
}

void HELP_OpenDlg (Widget Parent, char *Name, char *Msg, char **Image)
{
  int      n = 0;
  Arg      args[10];
  Pixmap   PixMap = XmUNSPECIFIED_PIXMAP;
  Widget   Shell, Dialog;
  XmString Message, Button;

  PRO_START (" ")

  Shell = XtCreatePopupShell (Name, xmDialogShellWidgetClass, Parent, NULL, 0);

  Message = XmStringLtoRCreate (Msg,  NORMAL_FONT);
  Button  = XmStringLtoRCreate ("OK", NORMAL_FONT);
  XtSetArg (args[n], XmNmessageString,     Message); n++;
  XtSetArg (args[n], XmNmessageAlignment,  XmALIGNMENT_BEGINNING); n++;
  XtSetArg (args[n], XmNcancelLabelString, Button); n++;
  XtSetArg (args[n], XmNdialogType,        XmDIALOG_INFORMATION); n++;

  if (Image != NULL && 
      (PixMap = TOOLMakeXPM(Image, NULL, NULL, 
			    TOOLGetColor("Grey"))) != XmUNSPECIFIED_PIXMAP)
  {
    XtSetArg (args[n], XmNsymbolPixmap, PixMap); n++;
  }

  Dialog = XmCreateMessageBox (Shell, "helpDialog", args, n);

  XmStringFree (Message);
  XmStringFree (Button);

  XtAddCallback (Dialog, XmNcancelCallback, (XtCallbackProc)HELP_DlgCB,
		 (void*)Shell);

  XtUnmanageChild (XmMessageBoxGetChild(Dialog, XmDIALOG_OK_BUTTON));
  XtUnmanageChild (XmMessageBoxGetChild(Dialog, XmDIALOG_HELP_BUTTON));

  XtSetMappedWhenManaged (Shell, False);
  XtManageChild (Dialog);

  XtRealizeWidget (Shell);
  XtPopup (Shell, XtGrabNonexclusive);
  XFlush  (XDisplay);

  TOOLMainLoop (AppCon, Shell);

  if (PixMap != XmUNSPECIFIED_PIXMAP)
    XmDestroyPixmap (XtScreen(Shell), PixMap);

  PRO_END
}


/* ********************************************************************** */
/* ** Prove if a command has been already started (nested commands) or ** */
/* ** if a command wasn't finished (syntax-error)                      ** */
/* ** INPUT:  Data    help-structure                                   ** */
/* **         ID      command-ID                                       ** */
/* **         state   -1: close command if it's open                   ** */
/* **                  0: open command if not yet done                 ** */
/* **                  1: open (nested) command                        ** */
/* ** OUTPUT: (dual)  TRUE, if the state is correct                    ** */
/* ********************************************************************** */
static dual HELP_TestFlag (HELPDATA *Data, byte ID, short state)
{
  dual Result = TRUE;

  switch (state)
  {
    case -1 : 
      if ((Result = (Data->Flags[ID] > 0)))
	(Data->Flags[ID])--;
      else
	PRO1 ("WARNING: %s-Cmd has been allready finished\n", TestText[ID])
      break;

    case 0 :
      if ((Result = (Data->Flags[ID] == 0)))
	(Data->Flags[ID])++;
      else
	PRO1 ("WARNING: Nested Commands are not allowed: %s\n", TestText[ID])
      break;

    default :
      (Data->Flags[ID])++;
      break;
  }
  return Result;
}


/* ********************************************************************** */
/* ** calculates a font-index to some font-attributes and a scale-     ** */
/* ** factor. Then loads the font, if not yet done                     ** */
/* ** INPUT:  Scale    scale-factor [0..2]                             ** */
/* **         Attr     font attributes                                 ** */
/* ** OUTPUT: (short)  index into font-table                           ** */
/* ********************************************************************** */
static short HELP_FontIndex (byte Scale, byte *Attr)
{
  int   count, Faktor = 1;
  short Result = 0;
  char  FontName[128];

  if (! Attr[FONT_PRE])
  {
    for (count = 0; count < FONT_ATTRIBS-1; ++count)
    {
      Result += Attr[count] * Faktor;
      Faktor *= AttrNo[count];
    }
    Result += Size[Scale][Attr[FONT_SIZE]] * Faktor;
    Result++;
  }

  /* check, if font was loaded */
  while (! Fonts[Result].loaded && Result >= 0)
  {
    if (Result == 0)
      sprintf (FontName, "fixed");    /* should be exist on all machines */
    else
    {
      PRO5 ("FONT %d [%s][%s][%s] -> ID: %d\n", Scale,
	    Weight[Attr[FONT_WEIGHT]], Slant[Attr[FONT_SLANT]], 
	    SizeTab[Size[Scale][Attr[FONT_SIZE]]], Result)
      sprintf (FontName, FontMask, Weight[Attr[FONT_WEIGHT]], 
	                           Slant[Attr[FONT_SLANT]], 
	                           SizeTab[Size[Scale][Attr[FONT_SIZE]]]);  
    }
    memcpy (Fonts[Result].Attr, Attr, sizeof(byte)*5);

    Fonts[Result].DrawFont = XLoadQueryFont (XDisplay, FontName);
    if ((Fonts[Result].loaded = (Fonts[Result].DrawFont != NULL)))
    {
      Fonts[Result].Space = XTextWidth (Fonts[Result].DrawFont, " ", 1);
      Fonts[Result].Height = Fonts[Result].DrawFont->ascent + 
	                     Fonts[Result].DrawFont->descent;
    }
    elsif (Result > 0)
      Result = 0;
    else
      Result = -1;
    
  }

  return Result;
}


/* ********************************************************************** */
/* ** Changes a text's scale-factor and updates it's data. It also     ** */
/* ** loads new fonts if needed.                                       ** */
/* ** INPUT:  Data     Text-data                                       ** */
/* **         Scale    new scale factor                                ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
static dual HELP_ChangeScale (HELPDATA *Data, int Scale)
{
  byte         Attr[5];
  int          len;
  XFontStruct *Font;
  HELP_ELEM   *Elem;

  PRO_START ("\n")

  if (Data->Param.Scale == Scale)
    PRO_RETURN ("Scale not changed", FALSE)

  PRO1 ("Scale: %d\n", Scale)

  Elem = Data->first;
  while (Elem != NULL)
  {
    memcpy (Attr, Fonts[Elem->FontID].Attr, sizeof(byte)*5);
    Attr[FONT_SIZE] = Elem->Index;
    Elem->FontID = HELP_FontIndex (Scale, Attr);

    if (Elem->Type == ELEM_TEXT || Elem->Type == ELEM_SECT)
    {
      Font = Fonts[Elem->FontID].DrawFont;
      len = strlen ((char*)Elem->Data);
      Elem->SizeX = XTextWidth (Font, (char*)Elem->Data, len);
      Elem->SizeY = Fonts[Elem->FontID].Height;
    }
    Elem = Elem->Next;
  }
  Data->Param.Scale = Scale;

  PRO_END
  return TRUE;
}


/* ********************************************************************** */
/* ** HELP_EvalFileSpec parses a file-specification with the following  ** */
/* ** features:                                                        ** */
/* ** - skips 'file:'-keyword at the beginning (if exists)             ** */
/* ** - reads the filename and evaluates sys-vars                      ** */
/* ** - looks for labels                                               ** */
/* ** INPUT:  CurrDir   actual directory                               ** */
/* **         FileSpec  file with path relative to CurrDir             ** */
/* ** OUTPUT: (REF)     Structure containing the data                  ** */
/* ********************************************************************** */
static REF HELP_EvalFileSpec (char *CurrDir, const char *FileSpec, int *Pos)
{
  int  Index;
  char Type[6];
  char FileName[PATH_MAX+1];
  char Complete[PATH_MAX+1];
  REF  Result = {0, "", "", NULL};

  if (FileSpec[*Pos] == '\0')
    return Result;

  /* skip a preceeding 'FILE:' */
  strncpy (Type, FileSpec+(*Pos), 5);
  Type[5] = '\0';
  if (strncasecmp(Type, "FILE:", 5) == 0)
    (*Pos) += 5;

  /* skip '//' or './' */
  if (FileSpec[(*Pos)+1] == '/')
  {
    if (FileSpec[*Pos] == '/')
      (*Pos)++;
    elsif (FileSpec[*Pos] == '.')
      (*Pos) += 2;
  }

  /* get filename... */
  Index = 0;
  while (FileSpec[*Pos] != '#' && FileSpec[*Pos] != '"' && 
	 FileSpec[*Pos] != '\0' && Index < PATH_MAX)
    FileName[Index++] = FileSpec[(*Pos)++];
  FileName[Index] = '\0';

  if (FileName[0] != '\0')
  {
    TOOLParseFileName (FileName, Complete, PATH_MAX);

    if (CurrDir[0] == '\0' || Complete[0] == '/')
      strcpy (Result.FileName, Complete);
    elsif (CurrDir[strlen(CurrDir)-1] == '/')
      sprintf (Result.FileName, "%s%s", CurrDir, Complete);
    else
      sprintf (Result.FileName, "%s/%s", CurrDir, Complete);
  }
  else
    Result.FileName[0] = '\0';

  /* ...and now the label (if exists) */
  Index = 0;
  if (FileSpec[*Pos] == '#')
  {
    (*Pos)++;
    while (FileSpec[*Pos] != '"' && FileSpec[*Pos] != '\0' && Index < LABELLEN)
      Result.Label[Index++] = FileSpec[(*Pos)++];
  }
  (*Pos)++;
  Result.Label[Index] = '\0';
  Result.geo = NULL;

  return Result;
}


/* ********************************************************************** */
/* ** Converts the section-No, which is stored in the Attr-structure   ** */
/* ** as a short-vector into a string. Section-No "0" will not be      ** */
/* ** convererted (exept the one)                                      ** */
/* ** INPUT:  Result   Pointer to allocated string-memory              ** */
/* **         Attr     Attributes containing the vector                ** */
/* ** OUTPUT: (char*)  (duplication of Result-pointer                  ** */
/* ********************************************************************** */
static char *HELP_MakeSectNo (char *Result, HELP_ATTR *Attr)
{
  int count;

  sprintf (Result, "%d", Attr->Sect[0]); 
  count = 1;
  while (count < MAX_SECT && Attr->Sect[count] > 0)
  {
    sprintf (Result, "%s.%d", Result, Attr->Sect[count]);
    count++;
  }
  strcat (Result, " ");

  return Result;
}


/* ********************************************************************** */
/* ** This function allocates and initializes a structure for keeping  ** */
/* ** the complete help-information. At it's first run it also         ** */
/* ** initializes global infos such as fonts and colors                ** */
/* ** INPUT:  Title        Name of help-window                         ** */
/* ** OUTPUT: (HELPDATA*)  allocated and initial. structure            ** */
/* ********************************************************************** */
static HELPDATA *HELP_InitHelpData (const char *Title)
{
  int         count;
  char       *Mask = NULL;
  XGCValues   GCVal;
  HELPDATA   *Result = NULL;
  XtResource  ResData[1];

  PRO_START ("\n")

  /* allocates momory for the general helptext-structure */
  PRO0 ("malloc...")
  if ((Result = (HELPDATA*)calloc(sizeof(HELPDATA), 1)) == NULL)
    PRO_RETURN ("MEMORY EXHAUSTED !", NULL)

  /*
     Read an alternate font-mask from resource-database. 
     Use %s for weight, slant and size ! e.g.:
     *helpWin*fontMask: -adobe-courier-%s-%s-*-*-%s-*-*-*-*-*-iso8859-1
  */
  ResData[0].resource_name   = "fontMask";
  ResData[0].resource_class  =  "helpWin";
  ResData[0].resource_type   =  XmRString;
  ResData[0].resource_size   =        128;
  ResData[0].resource_offset =          0;
  ResData[0].default_type    =  XmRString;
  ResData[0].default_addr    =   FontMask;
  
  /* Read entry from resouce-database: */
  XtGetSubresources (AppShell, &Mask, "helpWin", "shellWidgetClass", 
	             ResData, 1, (ArgList)NULL, 0);
  strncpy (FontMask, Mask, 127);
  FontMask[127] = '\0';

  /* if never run before, then load the fonts and build color table       */
  if (Initialized == 0)
  {
    PRO0 ("\\Fonts...\n")

    for (count = 0; count < FONT_AMOUNT+1; ++count)
      Fonts[count].loaded = FALSE;

    Colors = (Pixel *)malloc (sizeof(Pixel) * HELP_COLORS);
    for (count = 1; count < HELP_COLORS; ++count)
      Colors[count] = TOOLGetColor (ColorName[count]);
  }

  strncpy (Result->Title, Title, PATH_MAX);
  Result->Title[PATH_MAX] = '\0';
  Result->FileName[0] = '\0';
  Result->CurrDir[0]  = '\0';
  Result->MemID       = 0;
  Result->Shell       = NULL;

#if SEARCHING
  Result->Search.Method  = DEF_METHOD;
  Result->Search.Pattern = NULL;
  Result->Search.Hits    = 0;
  Result->Search.Files   = 0;
  Result->Search.Index   = 0;
  Result->Search.List    = NULL;
  Result->Search.Shell   = NULL;
  Result->Search.updated = FALSE;
#endif

  Result->Param.HTML  = HTML_BEHAVE;
  Result->Param.Click = DOUBLE_CLICK;
  Result->Param.Scale = SCALE;

  Result->Prev = NULL;
  Result->Next = NULL;

  Result->bg   = 0;

  GCVal.foreground = TOOLGetColor ("Black");
  GCVal.background = TOOLGetColor ("grey");
  GCVal.font       = XLoadFont (XDisplay, "fixed");

  XOR_GC = XCreateGC (XDisplay, RootWindow(XDisplay, XScreen),
		      GCForeground | GCBackground | GCFont, &GCVal);
  XSetFunction (XDisplay, XOR_GC, GXxor);

  Initialized++;
  PRO1 ("init: %d\n", Initialized)

  PRO_END
  return Result;
}


/* ********************************************************************** */
/* ** Creates a new help-data-structure and adds it into the global    ** */
/* ** list                                                             ** */
/* ** INPUT:  Title        Name of the new help-data-window            ** */
/* ** OUTPUT: (HELPDATA*)  Pointer to created help-data-structure      ** */
/* ********************************************************************** */
static HELPDATA *HELP_AddData (const char *Title)
{
  HELPDATA *Data, *Ptr;

  Data = HELP_InitHelpData (Title);

  if (HelpData == NULL)
    HelpData = Data;
  else
  {
    Ptr = HelpData;
    while (Ptr->Next != NULL)
      Ptr = Ptr->Next;

    Data->Prev = Ptr;
    Ptr->Next = Data;
  }

  return Data;
}


/* ********************************************************************** */
/* ** Checks, if a given help-data-pointer is stored in the list       ** */
/* ** INPUT:  Data   Searched data                                     ** */
/* ** OUTPUT: (dual) TRUE, if data was found in the list               ** */
/* **                FALSE, otherwise                                  ** */
/* ********************************************************************** */ 
static dual HELP_TestData (HELPDATA *Data)
{
  HELPDATA *Ptr;

  Ptr = HelpData;
  while (Ptr != NULL && Ptr != Data)
    Ptr = Ptr->Next;

  return (Ptr == Data);
}


/* ********************************************************************** */
/* ** Removes the data from the global list.                           ** */
/* ** INPUT:  Data  Specific data-pointer                              ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
static void HELP_DelData (HELPDATA *Data)
{
  if (HelpData == NULL || Data == NULL)
    return;

  if (HelpData == Data)
  {
    HelpData = Data->Next;
    Data->Prev = NULL;
  }
  else
  {
    Data->Prev->Next = Data->Next;
    if (Data->Next != NULL)
      Data->Next->Prev = Data->Prev;
  }
}


/* ********************************************************************** */
/* ** cleans the memory used by href's geometry-data                   ** */
/* ** INPUT:  Data  help-text data structure                           ** */
/* ** OUTPUT: none                                                     ** */
/* ********************************************************************** */
static void HELP_ResetData (HELPDATA *Data)
{
  int       count;
  HELP_GEO *Ptr, *Next;

  /* Free references' geometry-list */
  for (count = 0; count < Data->RNo; ++count)
  {
    Ptr = Data->HRef[count].geo;
    while (Ptr != NULL)
    {
      Next = Ptr->Next;
      free (Ptr);
      Ptr = Next;
    }
    Data->HRef[count].geo = NULL;
  }
}


/* *********************************************************************** */
/* ** Deallocates the memory used by an help-text-element               ** */
/* ** INPUT:  Elem  Pointer to help-text element                        ** */
/* ** OUTPUT: ---                                                       ** */
/* *********************************************************************** */
static void HELP_FreeElem (HELP_ELEM *Elem)
{
  IMG_DATA   *Image = NULL;
  INPUT_DATA *Input = NULL;

  if (Elem->Type == ELEM_IMG)
  {
    Image = (IMG_DATA *)Elem->Data;
    if (Image != NULL && Image->loaded)
    {
      free (Image->Image->data);
      /* XDestroyImage free's the image-data, which will crash 
	 in DEBUG-mode because of memory-control. Works, while 
	 XDestroyImage proves if data->ptr is NULL */
      Image->Image->data = NULL;
      XDestroyImage (Image->Image);
      free (Image);
    }
  }
  elsif (Elem->Type == ELEM_INPUT)
  {
    Input = (INPUT_DATA*)Elem->Data;
    if (Input->widget != NULL)
      XtDestroyWidget (Input->widget);
  }
  elsif (Elem->Type == ELEM_TABLE)
    HELP_FreeTable ((TABLE_DATA*)Elem->Data);
  elsif (Elem->Type != ELEM_CLEAR && Elem->Data != NULL)
    free (Elem->Data);

  free (Elem);
}

/* ********************************************************************** */
/* ** Frees all the data allocated by a help-text. If the last help-   ** */
/* ** text will be freed, ist frees also global fonts                  ** */
/* ** INPUT:  Data       help-text to be freed                         ** */
/* **         ResetOnly  dealloc only mem for loading new text         ** */
/* ** OUTPUT: none                                                     ** */
/* ********************************************************************** */
static void HELP_FreeHelpData (HELPDATA *Data, dual ResetOnly)
{
  short       count;
  HELP_ELEM  *Elem, *Next;
#if SEARCHING
  HIT_DATA   *Hit, *NextHit;
#endif

  PRO_START ("\n")

  /* remove all elements from elem-list                                   */
  PRO0 ("ELEMs...\n")
  Elem = Data->first;
  while (Elem != NULL)
  {
    Next = Elem->Next;
    HELP_FreeElem (Elem);
    Elem = Next;
  }
  Data->first = NULL;
  Data->last  = NULL;

  PRO0 ("REFs...\n")
  HELP_ResetData (Data);

  PRO0 ("BACKGROUND...\n")
  if (Data->bgImage != NULL)
  {
    XDestroyImage (Data->bgImage);
    Data->bgImage = NULL;
  }

  if (ResetOnly)
    PRO_ABORT ("finished")

#if SEARCHING
  PRO0 ("SEARCH-PATTERN...\n")
  if (Data->Search.Pattern != NULL)
    free (Data->Search.Pattern);
  Data->Search.Pattern = NULL;

  Hit = Data->Search.List;
  while (Hit != NULL)
  {
    NextHit = Hit->Next;
    free (Hit);
    Hit = NextHit;
  }
  Data->Search.List = NULL;
#endif

#if USE_PIXMAP
  PRO0 ("PIXMAP...\n")
  if (Data->Pic != None)
    XFreePixmap (XDisplay, Data->Pic);
#endif

  PRO0 ("GC...\n")
    XFreeGC (XDisplay, Data->DrawGC);

  free (Data);

  /* If this is the last opened help-win, we have to free the fonts */
  if (Initialized == 1)
  {
    PRO0 ("FONTs:")
    for (count = 0; count < FONT_AMOUNT+1; ++count)
      if (Fonts[count].loaded)
      {
	PRO1 ("\\.%d", count)
	XFreeFont (XDisplay, Fonts[count].DrawFont);
      }

    /* TODO: dynamisch allokierte Farben (> HELP_COLORS) freigeben ? */
    free (Colors);
    PRO0 ("\\\n")
  }
  Initialized--;

  PRO_END
} 


/* ********************************************************************** */
/* ** Evaluates a label such as <A NAME="...">                         ** */
/* ** INPUT:  Text     Complete label                                  ** */
/* ** OUTPUT: (LABEL)  Structure containing the Name                   ** */
/* **                  (Position will be added later)                  ** */
/* ********************************************************************** */
static LABEL HELP_EvalLabel (char *Text, int *Pos)
{
  int   Index;
  LABEL Result = {0, 0, ""};

  while (Text[*Pos] != '"' && Text[*Pos] != '\0')
    (*Pos)++;
  (*Pos)++;

  Index = 0;
  while (Text[*Pos] != '"' && Text[*Pos] != '\0' && Index < LABELLEN)
    Result.Name[Index++] = Text[(*Pos)++];
  Result.Name[Index] = '\0';
  (*Pos)++;

  return Result;
}


/* ********************************************************************** */
/* ** Interprets a complete href-cmd                                   ** */
/* ** INPUT:  Word   Reference text                                    ** */
/* ** OUTPUT: (REF)  Reference structure with filename+label           ** */
/* ********************************************************************** */
static REF HELP_EvalHRef (HELPDATA *Data, char *Text, int *Pos)
{
  while (Text[*Pos] != '"' && Text[*Pos] != '\0')
    (*Pos)++;
  (*Pos)++;

  return (HELP_EvalFileSpec(Data->CurrDir, Text, Pos));
}


/* ********************************************************************** */
/* ** Checks, if the specified image file has a known type and loads   ** */
/* ** it if possible.                                                  ** */
/* ** INPUT:  Data      Pointer to help data                           ** */
/* **         FileName  Image filename                                 ** */
/* ** INOUT:  SizeX                                                    ** */
/* **         SizeY     Returns the size of a loaded image             ** */
/* ** OUTPUT: (XImage*) Image Pointer or NULL when loading failed.     ** */
/* ********************************************************************** */
static XImage *HELP_LoadImage (HELPDATA *Data,
			      char *FileName, int *SizeX, int *SizeY)
{
  char    Complete[PATH_MAX+1];
  XImage *Result = NULL;

  if (Data->CurrDir[0] == '\0' || FileName[0] == '/')
    strcpy (Complete, FileName);
  elsif (Data->CurrDir[strlen(Data->CurrDir)-1] == '/')
    sprintf (Complete, "%s%s", Data->CurrDir, FileName);
  else
    sprintf (Complete, "%s/%s", Data->CurrDir, FileName);

  switch (TOOLGetImageType(Complete))
  {
    case XPM_FORMAT :
      Result = TOOLLoadXPM2 (Complete, SizeX, SizeY, Colors[HELP_BG]);
      break;
    case PNG_FORMAT :
      Result = TOOLLoadPNG  (Complete, SizeX, SizeY, FALSE);
      break;
    case JPG_FORMAT :
      Result = TOOLLoadJPG  (Complete, SizeX, SizeY);
      break;
    default :
      PRO0 ("unsupported image format\n")
  }

  return Result;
}


/* ********************************************************************** */
/* ** Extracts the filename of a image-cmd, parse it for system-vars   ** */
/* ** and loads the image data                                         ** */
/* ** INPUT:  Word         image command                               ** */
/* ** OUTPUT: (IMG_DATA*)  data stored in help-structure               ** */
/* **         NULL         error while loading the image               ** */
/* ********************************************************************** */
static IMG_DATA *HELP_EvalImg (HELPDATA *Data, char *Text, int *Pos)
{
  int       Index;
  char      FileName[PATH_MAX+1];
  char      FileSpec[PATH_MAX+1];
  IMG_DATA *Result = NULL;

  while (Text[*Pos] != '"' && Text[*Pos] != '\0')
    (*Pos)++;
  if (Text[*Pos] != '\0')
    (*Pos)++;

  /* get the image-filename */
  Index = 0;
  while (Text[*Pos] != '"' && Text[*Pos] != '>' && 
	 Text[*Pos] != '\0' && Index < PATH_MAX)
    FileSpec[Index++] = Text[(*Pos)++];
  FileSpec[Index] = '\0';
  TOOLParseFileName (FileSpec, FileName, PATH_MAX);
  (*Pos)++;

  Result = (IMG_DATA*)malloc (sizeof(IMG_DATA));
  Result->Image  = NULL;
  Result->loaded = FALSE;

  Result->alt[0] = '\0';                /* Read alternate description     */
  HELP_GetArg(Text, "ALT", Result->alt, *Pos);
  *Pos = HELP_GetArg (Text, NULL, NULL, *Pos) - 1;

  /* load image... */
  Result->Image = HELP_LoadImage (Data, FileName, 
				 &Result->SizeX, &Result->SizeY);

  if (Result->Image == NULL && Result->alt[0] == '\0')
    Result->Image = TOOLMakeXPM2 (badimg_xpm, &Result->SizeX, &Result->SizeY,
		                              Colors[HELP_BG]);

  Result->loaded = (Result->Image != NULL);

  if (! Result->loaded && Result->alt[0] == '0')
  {
    free (Result);
    Result = NULL;
  }

  return Result;
}


/* ********************************************************************** */
/* ** Searches for the next HTML-tag, which is specified by 'Tag'. If  ** */
/* ** no tag-name is given (NULL), the next tag will be located.       ** */
/* ** INPUT:  Text   Complete HELP-text                                ** */
/* **         Tag    Tag to look for                                   ** */
/* ** INOUT:  Pos    text-position                                     ** */
/* ** OUTPUT: (dual) TRUE, if tag was found, FALSE otherwise           ** */
/* ********************************************************************** */
static dual HELP_FindTag (char *Text, char *Tag, int *Pos)
{
  do
  {
    while (Text[*Pos] != '<' && Text[*Pos] != '\0')
      (*Pos)++;

    if (Text[*Pos] == '\0')
      return FALSE;

    (*Pos)++;
    if (Tag == NULL || Tag[0] == '\0' ||
	strncasecmp(Text+(*Pos), Tag, strlen(Tag)) == 0)
      return TRUE;
  } while (Text[*Pos] != '\0');

  return FALSE;
}


/* ********************************************************************** */
/* ** This function looks for a given argument in the actual tag. On   ** */
/* ** success 0 will be returned and the value will be delieverd in    ** */
/* ** 'Result'. The attribut-value might be enclose in brackets        ** */
/* ** INPUT:  Text      Complete text                                  ** */
/* **         Argument  name of the tag-attribute                      ** */
/* **         Pos       positon, where to start searching              ** */
/* ** OUTPUT: Result    argument-value                                 ** */
/* **         (int)     0, on success. If no argument is specified or  ** */
/* **                   it was NOT found, then the position behind the ** */
/* **                   closing tag-bracket will be returned.          ** */
/* ********************************************************************** */
static int HELP_GetArg (char *Text, char *Argument, char *Result, int Pos)
{
  dual quoted;
  int  Index, len = 0;

  if (Argument != NULL)
    len = strlen (Argument);
  else
    len = -1;

  while (Text[Pos] != '>' && Text[Pos] != '\0')
  {
    if ((len > 0 && strncasecmp(Text+(Pos), Argument, len) == 0) ||
        len == 0)
    {
      Pos += len;
      while (Text[Pos] != '=' && Text[Pos] != '\0')
	Pos++;
      if (Text[Pos] != '\0')
	Pos++;
      while (HELP_IsSpace(Text[Pos]) && Text[Pos] != '\0')
	Pos++;
      if ((quoted = (Text[Pos] == '"')))
	Pos++;

      Index = 0;
      while ((! HELP_IsSpace(Text[Pos]) || quoted) &&
	     Text[Pos] != '"' && Text[Pos] != '>' && Index < 40)
      {
	Result[Index++] = Text[Pos++];
      }
      Result[Index] = '\0';

      if (Text[Pos] == '"')
	Pos++;

      return 0;
    }
    Pos++;
  }
  if (Text[Pos] == '>')
    Pos++;

  return Pos;
}


/* ********************************************************************** */
/* ** Deallocates the memory, which was used by a table                ** */
/* ** INPUT:  Table  Data of the table                                 ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
static void HELP_FreeTable (TABLE_DATA *Table)
{
  int        row, col;
  HELP_ELEM *Elem, *Next; 

  if (Table == NULL)
    return;

  if (Table->Data != NULL)
  {
    for (row = 0; row < Table->Rows; ++row)
    {
      for (col = 0; col < Table->Cols; ++col)
      {
	Elem = Table->Data[row][col].List;
	while (Elem != NULL)
	{
	  Next = Elem->Next;
	  HELP_FreeElem (Elem);
	  Elem = Next;
	}
      }
      free (Table->Data[row]);
    }
    free (Table->Data);
  }

  free (Table->w);
  free (Table->h);

  free (Table);
}


/* ********************************************************************** */
/* ** Parses table-data and stores it in a grid-array-structure.       ** */
/* ** INPUT:  Data  Pointer to help-data                               ** */
/* **         Text           Complete help-text                        ** */
/* ** INOUT:  Pos            Start- and end-position of the table      ** */
/* ** OUTPUT: (TABLE_DATA *) Pointer to created table-data             ** */
/* **                        NULL, if an error occured                 ** */
/* ********************************************************************** */
static TABLE_DATA *HELP_EvalTable (HELPDATA *Data, 
				  char *Text, int *Pos, HELP_ATTR *Attr)
{
  dual        space;
  int         row, col, Diff;
  int         countX, countY;
  int         SizeX, SizeY, MaxSize = 0;
  int         count, Index;
  int         MaxCols = 0;
  int         Weight = 0;
  char        Argument[41];
  HELP_ELEM  *Elem, *Last = NULL;
  TABLE_DATA *Table;

  Table = (TABLE_DATA*)malloc (sizeof(TABLE_DATA));
  Table->SizeX  = 0;
  Table->SizeY  = 10;
  Table->Border = 0;
  Table->Rows   = 0;
  Table->Cols   = 0;
  Table->Data   = NULL;

  if (HELP_GetArg(Text, "BORDER", Argument, *Pos) == 0)
    Table->Border = atoi (Argument);

  if (HELP_GetArg(Text, "WIDTH", Argument, *Pos) == 0)
    Table->SizeX = atoi (Argument);

  if (HELP_GetArg(Text, "HEIGHT", Argument, *Pos) == 0)
    Table->SizeY = atoi (Argument);

  *Pos = HELP_GetArg (Text, NULL, NULL, *Pos);

  Index = *Pos;
  while (HELP_FindTag(Text, NULL, &Index) && 
	 strncasecmp(Text+Index, "/TABLE", 6) != 0)
  {
    if (strncasecmp(Text+Index, "TR", 2) == 0)
    {
      Index += 3;

      (Table->Rows)++;
      Table->Cols = 0;
    }
    elsif (strncasecmp(Text+Index, "TH", 2) == 0 ||
	   strncasecmp(Text+Index, "TD", 2) == 0)
    {
      Index += 2;

      if (HELP_GetArg(Text, "COLSPAN", Argument, Index) == 0)
	Table->Cols += atoi(Argument);
      else
	(Table->Cols)++;

      Index = HELP_GetArg (Text, NULL, NULL, Index); /* search End-Of-Tag */
    }
    else
      Index++;

    if (Table->Cols > MaxCols)
      MaxCols = Table->Cols;
  }
  Table->Cols = MaxCols;

  PRO1 ("TABLE-COLS: %d\n", Table->Cols)
  PRO1 ("TABLE-ROWS: %d\n", Table->Rows)

  Table->Data = (TABLE_CELL**)malloc (sizeof(TABLE_CELL*) * Table->Rows);
  for (row = 0; row < Table->Rows; ++row)
  {
    Table->Data[row] = (TABLE_CELL*)malloc(sizeof(TABLE_CELL) * Table->Cols);
    for (col = 0; col < Table->Cols; ++col)
    {
      Table->Data[row][col].Type    = 1;
      Table->Data[row][col].ColSpan = 1;
      Table->Data[row][col].RowSpan = 1;
      Table->Data[row][col].Color   = HELP_BG;
      Table->Data[row][col].align   = HELP_LEFT;
      Table->Data[row][col].List    = NULL;
    }
  }
  Table->w = (int *)malloc (sizeof(int) * Table->Cols);
  for (col = 0; col < Table->Cols; ++col)
    Table->w[col] = 10;
  Table->h = (int *)malloc (sizeof(int) * Table->Rows);
  for (row = 0; row < Table->Rows; ++row)
    Table->h[row] = 10;

  Index = *Pos;
  row = -1;
  col =  0;
  while (HELP_FindTag(Text, NULL, &Index) &&
	 strncasecmp(Text+Index, "/TABLE", 6) != 0)
  {
    if (strncasecmp(Text+Index, "TR", 2) == 0)
    {
      Index += 3;
      row++;
      col = 0;
    }
    elsif (strncasecmp(Text+Index, "TH", 2) == 0 ||
	   strncasecmp(Text+Index, "TD", 2) == 0)
    {
      while (col < Table->Cols && Table->Data[row][col].Type == -1)
	col++;	   
      if (col == Table->Cols)
      {
	HELP_FreeTable (Table);
	return NULL;
      }
      Table->Data[row][col].Type = (Text[Index+1] == 'H');
      Index += 2;

      if (HELP_GetArg(Text, "COLSPAN", Argument, Index) == 0)
	Table->Data[row][col].ColSpan = atoi(Argument);

      if (HELP_GetArg(Text, "ROWSPAN", Argument, Index) == 0)
	Table->Data[row][col].RowSpan = atoi(Argument);

      if (HELP_GetArg(Text, "BGCOLOR", Argument, Index) == 0)
      {
	count = 0;
	while (count < HELP_COLORS && Table->Data[row][col].Color == HELP_BG)
	{
	  if (strcasecmp(Argument, ColorName[count]) == 0)
	    Table->Data[row][col].Color = count;
	  count++;
	}
      }  

      if (HELP_GetArg(Text, "ALIGN", Argument, Index) == 0)
      {
	if (strcasecmp(Argument, "CENTER") == 0)
	  Table->Data[row][col].align = HELP_CENTER;
	elsif (strcasecmp(Argument, "RIGHT") == 0)
	  Table->Data[row][col].align = HELP_RIGHT;
      }
  
      Index = HELP_GetArg (Text, NULL, NULL, Index);   /* Search End-Of-Tag */

      SizeX = 0;
      SizeY = 0;
      while (Text[Index] != '<' && Text[Index] != '\0')
      {
	while (HELP_IsSpace(Text[Index]))
	  Index++;

	count = 0;
	while (! HELP_IsSpace(Text[Index]) && Text[Index] != '<' &&
	       count < 40 && Text[Index] != '\0')
	  Argument[count++] = Text[Index++];
        Argument[count++] = '\0';

	space = HELP_IsSpace(Text[Index]);
	if (Table->Data[row][col].Type == 1)
	{
	  Weight = Attr->Attr[FONT_WEIGHT];
	  Attr->Attr[FONT_WEIGHT] = 1;
	}
	Elem = HELP_MakeElem (ELEM_TEXT, (void *)Argument, Data->Param.Scale,
			     (space)? (-1):0, Attr);
	if (Table->Data[row][col].Type == 1)
	  Attr->Attr[FONT_WEIGHT] = Weight;

	if (Table->Data[row][col].List == NULL)
	  Table->Data[row][col].List = Elem;
	else
	  Last->Next = Elem;
	Last = Elem;

       	if (SizeY == 0)
	  SizeY = Elem->SizeY;

	if (SizeX > 0 && 
	    ((Table->SizeX == 0 && 
	      SizeX+Elem->SizeX >= 600 / Table->Cols) ||
	     (Table->SizeX > 0 &&
	      SizeX+Elem->SizeX >= Table->SizeX / Table->Cols)))
	{
	  SizeY += Elem->SizeY;
	  MaxSize = MAX (MaxSize, SizeX);
	  SizeX = Elem->SizeX;
	}
	else
	  SizeX += Elem->SizeX;

	if (Elem->space > 0)
	  SizeX += Elem->space;
	elsif (Elem->space < 0)
	  SizeX += Fonts[Elem->FontID].Space;
      }

      if (Table->Data[row][col].ColSpan == 1)
	Table->w[col] = MAX (Table->w[col], MAX(MaxSize, SizeX)+6);

      if (Table->Data[row][col].RowSpan == 1)
	Table->h[row] = MAX (Table->h[row], SizeY+4);

      for (countX=col+1; countX<col+Table->Data[row][col].ColSpan; ++countX)
	Table->Data[row][countX].Type = -1;
      for (countY=row+1; countY<row+Table->Data[row][col].RowSpan; ++countY)
        for (countX=col; countX<col+Table->Data[row][col].ColSpan; ++countX)
	  Table->Data[countY][countX].Type = -1;

      col += Table->Data[row][col].ColSpan;
    }
    else
      while (Text[Index] != '>' && Text[Index] != '\0')
	Index++;
  }

  SizeX = (Table->Cols+1) * (Table->Border-1);
  for (col = 0; col < Table->Cols; ++col)
    SizeX += Table->w[col];
  if (SizeX < Table->SizeX)
  {
    Diff = (Table->SizeX-SizeX) / Table->Cols;
    for (col = 0; col < Table->Cols; ++col)
      Table->w[col] += Diff;
    Table->SizeX = SizeX + Diff * Table->Cols;
  }
  else
    Table->SizeX = SizeX;

  SizeY = (Table->Rows+1) * (Table->Border-1);
  for (row = 0; row < Table->Rows; ++row)
    SizeY += Table->h[row];
  if (SizeY < Table->SizeY)
  {
    Diff = (Table->SizeY-SizeY) / Table->Rows;
    for (row = 0; row < Table->Rows; ++row)
      Table->h[row] += Diff;
    Table->SizeY = SizeY + Diff * Table->Rows;
  }
  else
    Table->SizeY = SizeY;

#ifdef PROTOCOL
  PRO2 ("TABLE: %d x %d\n", Table->SizeX, Table->SizeY);

  for (col = 0; col < Table->Cols; ++col)
    PRO1 ("\\%d  ", Table->w[col])
  PRO0 ("\n")
  for (row = 0; row < Table->Rows; ++row)
    PRO1 ("%d\n", Table->h[row])

  for (row = 0; row < Table->Rows; ++row)
    for (col = 0; col < Table->Cols; ++col)
    {
      PRO2 ("[%d][%d]: ", col, row)
      PRO3 ("\\  %c %dx%d\n",
	    (Table->Data[row][col].Type == -1)? '-':'+',
	    Table->Data[row][col].ColSpan, Table->Data[row][col].RowSpan)
    }
#endif

  *Pos = Index;

  return Table;
}


/* ********************************************************************** */
/* ** Reads the parameter of a form-command. We assume that the first  ** */
/* ** attribute will be action="...". Other attributes don't matter.   ** */
/* ** The Uniform Resource Identifier (URI) should be a simplified     ** */
/* ** http-location: name of an executeable shell-script               ** */
/* ********************************************************************** */
static dual HELP_EvalForm (HELPDATA *Data, char *Text, int *Pos)
{
  int Index;

  while (Text[*Pos] != '"' && Text[*Pos] != '\0')
    (*Pos)++;
  if (Text[*Pos] != '\0')
    (*Pos)++;

  Index = 0;
  while (Text[*Pos] != '"' && Text[*Pos] != '>' && 
	 Text[*Pos] != '\0' && Index < PATH_MAX)
    Data->URI[Data->UNo][Index++] = Text[(*Pos)++];
  Data->URI[Data->UNo][Index] = '\0';

  while (Text[*Pos] != '>' && Text[*Pos] != '\0')
    (*Pos)++;

  return (Text[*Pos] != '\0');
}


/* ********************************************************************** */
/* ** Reads the parameters of an input-statement                       ** */
/* ********************************************************************** */
static INPUT_DATA *HELP_EvalInput (char *Text, int *Pos)
{
  dual        quoted;
  int         Index;
  char        Keyword[LABELLEN+1];
  char        Value[LABELLEN+1];
  INPUT_DATA *Result = NULL;

  Result = (INPUT_DATA*)malloc (sizeof(INPUT_DATA));
  Result->Type     = TEXTVALUE;
  Result->Name[0]  = '\0';
  Result->Size     = 20;
  Result->Value[0] = '\0';
  Result->checked  = FALSE;
  Result->widget   = NULL;

  while (Text[*Pos] != '>' && Text[*Pos] != '\0')
  {
    while (HELP_IsSpace(Text[*Pos]) && Text[*Pos] != '\0')
      (*Pos)++;

    Index = 0;
    while (! HELP_IsSpace(Text[*Pos]) && 
	   Text[*Pos] != '=' && Text[*Pos] != '>')
      Keyword[Index++] = Text[(*Pos)++];
    Keyword[Index] = '\0';

    while (HELP_IsSpace(Text[*Pos]) && Text[*Pos] != '\0')
      (*Pos)++;

    if (Text[*Pos] == '=')
      (*Pos)++;

    while (HELP_IsSpace(Text[*Pos]) && Text[*Pos] != '\0')
      (*Pos)++;

    if ((quoted = (Text[*Pos] == '"')))
      (*Pos)++;

    Index = 0;
    while ((! HELP_IsSpace(Text[*Pos]) || quoted) && 
	   Text[*Pos] != '"' && Text[*Pos] != '>')
      Value[Index++] = Text[(*Pos)++];
    Value[Index] = '\0';

    if (Text[*Pos] == '"')
      (*Pos)++;

    if (strncasecmp(Keyword, "TYPE", 4) == 0)
      if (strncasecmp(Value, "CHECKBOX", 8) == 0)
	Result->Type = CHECKBOX;
      elsif (strncasecmp(Value, "SUBMIT", 6) == 0)
	Result->Type = SUBMIT;
      elsif (strncasecmp(Value, "TEXT", 4) == 0)
	Result->Type = TEXTVALUE;
      else
	PRO1 ("Input-Type: '%s' not supported\n", Value)
    elsif (strncasecmp(Keyword, "NAME", 4) == 0)
    {
      strncpy (Result->Name, Value, LABELLEN);
      Result->Name[LABELLEN] = '\0';
    }
    elsif (strncasecmp(Keyword, "SIZE", 4) == 0)
      Result->Size = atoi(Value);
    elsif (strncasecmp(Keyword, "VALUE", 5) == 0)
    {
      strncpy (Result->Value, Value, LABELLEN);
      Result->Value[LABELLEN] = '\0';
    }
    elsif (strncasecmp(Keyword, "CHECKED", 7) == 0)
      Result->checked = TRUE;
    else
      PRO2 ("Invalid Argument: %s = '%s'", Keyword, Value)
  }

  if (Result->Name[0] == '\0' && Result->Type != SUBMIT)
  {
    PRO0 ("Missing NAME-attribute\n")
    free (Result);
    return NULL;
  }

  return Result;
}


/* ********************************************************************** */
/* ** Calculates the Index into the colortable for the given color (if ** */
/* ** found)                                                           ** */
/* ** INPUT:  Word   color-command                                     ** */
/* ** OUTPUT: (int)  Index into colortable                             ** */
/* **         0      ("black"-Index) if bad colorspec                  ** */
/* ********************************************************************** */
static byte HELP_EvalColor (char *Text, int *Pos)
{
  byte Result = 0;
  int  count, Index;
  char Color[12];

  while (Text[*Pos] != '"' && Text[*Pos] != '\0')
    (*Pos)++;
  (*Pos)++;

  /* read color-name */
  Index = 0;
  while (Text[*Pos] != '"' && Text[*Pos] != '\0' && Index < 11)
    Color[Index++] = Text[(*Pos)++];
  Color[Index] = '\0';
  (*Pos)++;

  /* search index */
  count = 0;
  if (Color[0] == '#')
  {
    Colors = realloc (Colors, sizeof(Pixel) * (ColorNo+1));
    Colors[ColorNo] = TOOLGetColor (Color);
    Result = ColorNo++;
  }
  else
  {
    while (count < HELP_COLORS && Result == 0)
    {
      if (strcasecmp(Color, ColorName[count]) == 0)
	Result = count;
      count++;
    }
  }

  return Result;
}


/* ********************************************************************** */
/* ** Parses lines like <CMD=nnn> for their int-value and returns it.  ** */
/* ** INPUT:  Word   command-string                                    ** */
/* ** OUTPUT: (int)  found number nnn                                  ** */
/* ********************************************************************** */
static int HELP_EvalValue (char *Text, int *Pos)
{
  int Result = 0;

  while (Text[*Pos] != '=' && Text[*Pos] != '\0')
    (*Pos)++;
  (*Pos)++;

  if (sscanf(Text+(*Pos), "%d", &Result) == 1)
  {
    while (Text[*Pos] == ' ')           /* update text-position           */
      (*Pos)++;
    while (isdigit(Text[*Pos]))
      (*Pos)++;
  }

  PRO1 ("EVALUATION: %d\n", Result)

  return Result;
}


/* ********************************************************************** */
/* ** Loading a Text given by a filename or direct text and            ** */
/* ** initializes some help-structure informations                     ** */
/* ** INPUT:  Data      help-text structure                            ** */
/* **         TextSpec  filename or direct help-text                   ** */
/* ** OUTPUT: (char*)   file contents or duplicate of text             ** */
/* ********************************************************************** */
static char *HELP_LoadText (HELPDATA *Data, const char *TextSpec)
{
         int   fd = -1, len = 0;
         char  Type[6];
	 char *Result = NULL;
  struct stat  info;

  PRO_START ("\n")

  if (Data == NULL)
    PRO_RETURN ("Invalid data structure", NULL)

  TOOLSetMouseCursor (Data->Shell, ToolWATCH, NULL, NULL); 

  strncpy (Type, TextSpec, 5);
  Type[5] = '\0';
  if (strncasecmp(Type, "TEXT:", 5) == 0)        /* text given directly   */
  {
    TextSpec += 5;
    len = strlen (TextSpec);
    Result = (char*)malloc (sizeof(char) * (len+1));
    strncpy (Result, TextSpec, len);
    Data->FileName[0] = '\0';
  }    
  else                                            /* load text from file  */
  {
    TOOLGetFileName (TextSpec, Data->Title, PATH_MAX);

    if (TextSpec[0] != '\0')
    {
      PRO1 ("Loading: %s\n", TextSpec)

      strncpy (Data->FileName, TextSpec, PATH_MAX);
      Data->FileName[PATH_MAX] = '\0';

      TOOLGetFileDir (TextSpec, Data->CurrDir, PATH_MAX);

      /* try to open file... */
      if (stat(TextSpec, &info) != -1 && (fd = open(TextSpec, O_RDONLY)) != -1)
      {
	len = (int)info.st_size;
	Result = (char*)malloc (sizeof(char) * (len+1));

	if (read(fd, Result, len) == -1)
        {
	  free (Result);
	  Result = NULL;
	}
	Result[len] = '\0';
	close (fd);
      }
    }
  }

  if (Result == NULL)
  {
    TOOLSetMouseCursor (Data->Shell, ToolNORMAL, NULL, NULL);
    PRO_RETURN ("Read failed", NULL)
  }
  else
    Result[len] = '\0';

#ifdef SEARCHING
  Data->Search.updated = FALSE;
#endif

  PRO_END
  return Result;
}


/* ********************************************************************** */
/* ** Creates a new help-text element using the given parameters and   ** */
/* ** attributes.                                                      ** */
/* ** INPUT:  Type           Type of element                           ** */
/* **         Info           additional element-data                   ** */
/* **         Scale          font-size, if text is specified           ** */
/* **         space          Space following the current element       ** */
/* **         Attr           Pointer to actual elem-attributes         ** */
/* ** OUTPUT: (HELP_ELEM *)  Pointer to allocated element-structure    ** */
/* ********************************************************************** */
static HELP_ELEM *HELP_MakeElem (int Type, void *Info, byte Scale,
				short space, HELP_ATTR *Attr)
{
  short        SizeX = 0;
  short        SizeY = 0;
  int          len;
  XFontStruct *Font;
  HELP_ELEM   *Elem = NULL;

  Elem = (HELP_ELEM*)malloc (sizeof(HELP_ELEM));
  Elem->Type   = Type;
  Elem->Index  = Attr->Attr[FONT_SIZE];
  Elem->FontID = HELP_FontIndex (Scale, Attr->Attr);

  /* Calc size of element (depends of it's type) */
  switch (Elem->Type)
  {
    case ELEM_TEXT :
    case ELEM_SECT :
      {
	char *Word = NULL;
        char  Header[255];

	if (Elem->Type == ELEM_TEXT)
	  Word = (char*)Info;
	else
	{
	  Word = HELP_MakeSectNo (Header, Attr);
	  Elem->Type = ELEM_TEXT;
	}
	    
	len = strlen (Word);
	Elem->Data = (void *)malloc (sizeof(char) * (len+1));
	strcpy ((char *)Elem->Data, Word);

	Font = Fonts[Elem->FontID].DrawFont;
	SizeX = XTextWidth (Font, Word, len);
	SizeY = Fonts[Elem->FontID].Height;
      }
      break;

    case ELEM_LINE :
      SizeY = 8;         /* also possible: Fonts[Elem->FontID].Height */
      Elem->Data = NULL;
      break;

    case ELEM_ITEM :
      SizeY = 12;
      Elem->Data = (void *)malloc (sizeof(int));
      *(int*)(Elem->Data) = Attr->Number;
      break;
     
    case ELEM_IMG :
      {
	IMG_DATA *Image = (IMG_DATA*)Info;

	SizeX = Image->SizeX;
	SizeY = Image->SizeY;
	Elem->Data = Image;
      }
      break;

    case ELEM_CLEAR :
      Elem->Data = Attr->Param;
      break;

    case ELEM_TABLE :
      {
	TABLE_DATA *Table = (TABLE_DATA*)Info;

	SizeX = Table->SizeX;
	SizeY = Table->SizeY;
	Elem->Data = Table;
      }
      break;

    case ELEM_INPUT :
      {
	Dimension   w, h;
	INPUT_DATA *Input = (INPUT_DATA*)Info;

        XtVaGetValues (Input->widget, XmNwidth, &w, 
		                      XmNheight, &h,
		                      NULL);
	SizeX = w;
	SizeY = h+2;
	Elem->Data = Input;
      }
      break;

#if MARK_REFS
    case ELEM_REF :
      SizeX = RefX + 4;
      SizeY = RefY;
      Elem->Data = NULL;
      break;
#endif

    default:
      Elem->Data = NULL;
  }

  /* set various informations */
  Elem->space   = space;
  Elem->line    = Attr->Attr[FONT_UL];
  Elem->align   = Attr->align;
  Elem->ColorID = Attr->ColorID;
  Elem->ID      = Attr->ID;
  Elem->SizeX   = SizeX;
  Elem->SizeY   = SizeY;
  Elem->fits    = FALSE;
  Elem->Next    = NULL;

  return Elem;
}


/* ********************************************************************** */
/* ** Adds a help text-element to the internal elem-list using the     ** */
/* ** given font and attributes.                                       ** */
/* ** INPUT:   Data   structure containing the elem-list               ** */
/* **          Type   type of element                                  ** */
/* **          Info   elem-data (text or other data)                   ** */
/* **          space  meaning depends on element-type and value:       ** */
/* **                 ! ELEM_VSPACE:                                   ** */
/* **                   <0 : Font-space [neg. Fontindex-1]             ** */
/* **                   >0 : absolute space behind elem [Pixel]        ** */
/* **                 ELEM_VSPACE: size of vertical space              ** */
/* **          Attr   additional attributes                            ** */
/* ** OUTPUT:  ---                                                     ** */
/* ********************************************************************** */
static void HELP_AppendElem (HELPDATA *Data, int Type, void *Info,
			                    short space, HELP_ATTR *Attr)
{
  HELP_ELEM   *Elem = NULL;

  if (Attr->Attr[FONT_PRE] && !(Type & (ELEM_TEXT | ELEM_VSPACE)))
    return;

  Elem = HELP_MakeElem (Type, Info, Data->Param.Scale, space, Attr);

  if (Data->first == NULL)              /* append elem at end of list     */
    Data->first = Elem;
  else
    Data->last->Next = Elem;

  Data->last  = Elem;
}


/* ********************************************************************** */
/* ** looks in the table for a special symbol and returns it the       ** */
/* ** keyword could be terminated by a semicolon                       ** */
/* ** INPUT:  Word    alias (keyword) for the symbol                   ** */
/* ** OUTPUT: (char)  found symbol                                     ** */
/* **         '\0'    symbol-key is unknown                            ** */
/* **         len     pointer to length of found symbol-key            ** */
/* ********************************************************************** */
static char HELP_GetSymbol (char *Word, int *len)
{
  int  count = 0;
  char Value[4];
  char Result = '\0';

  /* compare with all known symbol-keywords */
  *len = 0;

  if (isdigit(Word[0]) || Word[0] == '#')
  {
    if (Word[0] == '#')
      (*len)++;

    while (isdigit(Word[(*len)+count]) && count < 3)
    {
      Value[count] = Word[(*len)+count];
      count++;
    }
    Value[count] = '\0';
    *len += count;
    Result = (char)atoi(Value);
  }
  else
    while (count < SYMBOLS && Result == '\0')
      if (strncmp(Word, Symbols[count].Key, strlen(Symbols[count].Key)) == 0)
      {
	Result = Symbols[count].Symbol;
	*len = strlen (Symbols[count].Key);
      }
      else
	count++;

  /* if not found in list, then skip this expression */
  if (Result == '\0')
    while (! HELP_IsSpace(Word[*len]) && Word[*len] != ';')
      (*len)++;

  /* skip terminating semicolon */
  if (Word[*len] == ';')
    (*len)++;

  return Result;
}


/* ********************************************************************** */
/* ** Reads a word or symbol at the given text position and inserts it ** */
/* ** into elem-list                                                   ** */
/* ** INPUT:   Data   help-structure                                   ** */
/* **          Text   help-text                                        ** */
/* **          Pos    acktual text position                            ** */
/* **          Attr   Attributs                                        ** */
/* ** OUTPUT:  (int)  New text position                                ** */
/* ********************************************************************** */
static void HELP_ReadWord (HELPDATA *Data, 
			  char *Text, int *Pos, HELP_ATTR *Attr)
{
  dual  space;
  short count = 0;
  int   len;
  char  Symbol;
  char  Word[255];

  /* read word till next space or max. length of word */
  while (! HELP_IsSpace(Text[*Pos]) && count < 254 &&
	 (Text[*Pos] != '<' || (Attr->Attr[FONT_PRE] &&
		  	        strncasecmp(Text+(*Pos), "</PRE", 5) != 0)))
    if (Text[*Pos] == '&' && ! Attr->Attr[FONT_PRE])
    {
      if ((Symbol = HELP_GetSymbol(Text+(*Pos)+1, &len)) != '\0')
        Word[count++] = Symbol;
      (*Pos) += len+1;
    }
    else
      Word[count++] = Text[(*Pos)++];
  Word[count] = '\0';

  /* store the word as plain text */
  space = ((HELP_IsSpace(Text[*Pos]) && ! Attr->Attr[FONT_PRE]) ||
	   (Attr->Attr[FONT_PRE] && HELP_IsSpace(Text[*Pos]) && 
	    Text[*Pos] != '\n'));
  HELP_AppendElem (Data, ELEM_TEXT, (void *)Word, (space)? (-1):0, Attr);
  if (Attr->Attr[FONT_PRE] && Text[*Pos] == '\n')
    HELP_AppendElem (Data, ELEM_VSPACE, NULL, 0, Attr);

  if (space || (Attr->Attr[FONT_PRE] && Text[*Pos] == '\n'))
    (*Pos)++;
}

/* ********************************************************************** */
/* ** Generates a command-identifier from a given command-sequence.    ** */
/* ** In some cases the attributes will be changed                     ** */
/* ** INPUT:  Word   Text of command                                   ** */
/* ** OUTPUT: (int)  Command-ID (see above)                            ** */
/* **         Attr   changed attributes                                ** */
/* **                (if Textsize- or align-cmd)                       ** */
/* ********************************************************************** */
static int HELP_GetCmd (char *Text, HELP_ATTR *Attr, int *Pos)
{
  int  count = 0, len;
  int  Size, Sect;
  int  Result = NO_CMD;
#ifdef _PROTOCOL
  char Test[20];
#endif

  if (Text[*Pos] == '>' || Text[*Pos] == '\0')
  {
    (*Pos)++;
    return E_CMD;
  }

  while (HELP_IsSpace(Text[*Pos]))
    (*Pos)++;

#ifdef _PROTOCOL
  strncpy (Test, Text+(*Pos), 10);
  Test[10] = '\0';
  PRO1 ("Identifying '%s' ", Test)
#endif

  /* comparison with keyword-list */
  while (count < KEYWORDS && Result == NO_CMD)
  {
    len = strlen (Keys[count].Key);
    if (strncasecmp(Text+(*Pos), Keys[count].Key, len) == 0 &&
	(isspace(Text[(*Pos)+len]) || 
	 Text[(*Pos)+len] == '>' || Text[(*Pos)+len] == '='))
    {
      Result = Keys[count].Code;
      (*Pos) += len;
    }
    count++;
  }

  /* not found - maybe a font-size or section command... */
  if (Result == NO_CMD)
  {
    if (toupper(Text[*Pos]) == 'H' &&
	Text[(*Pos)+1] > '0' && Text[(*Pos)+1] < '6')
    {
      Result = HN_S_CMD;
      Size = Text[(*Pos)+1] - '1';
      Attr->Attr[FONT_SIZE] = (Size > 1)? (Size-1) : Size;
      (*Pos) += 2;
    }
    elsif (strncasecmp(Text+(*Pos), "/H", 2) == 0 && 
	   Text[(*Pos)+2] > '0' && Text[(*Pos)+2] < '6')
    {
      Result = HN_E_CMD;
      Attr->Attr[FONT_SIZE] = 2;
      Attr->align = HELP_LEFT;
      (*Pos) += 3;
    }
    elsif (strncasecmp(Text+(*Pos), "SECT", 4) == 0 && 
	   ((Text[(*Pos)+4] > '0' && Text[(*Pos)+4] < '5') ||
	    Text[(*Pos)+4] == '>'))
    {
      Result = SECTN_CMD;
      Sect = (Text[(*Pos)+4]=='>')? 0 : (Text[(*Pos)+4]-'0');
      (Attr->Sect[Sect])++;
      for (count = Sect+1; count < MAX_SECT; ++count)
	Attr->Sect[count] = 0;
      Attr->Attr[FONT_SIZE] = (Sect+1) / 2;
      Attr->Head = TRUE;
      (*Pos) += 4;
      if (Sect > 0)
	(*Pos)++;
    }
    else
      Result = NO_CMD;
  }

  if (Result != NO_CMD)
    PRO1 ("\\=> %d\n", Result)
  else
  {
    PRO0 ("\\ BAD COMMAND !\n")
    (*Pos)++;
  }

  return Result;
}


/* ********************************************************************** */
/* ** Read the following command adds it to the elem-list and proves   ** */
/* ** the correctness of the help text                                 ** */
/* ** INPUT:  Data   help-structure                                    ** */
/* **         Text   help text                                         ** */
/* **         Pos    actual text position                              ** */
/* **         Attr   actual attributes                                 ** */
/* ** OUTPUT: (int)  new text position                                 ** */
/* **         Attr   (changed) attributes                              ** */
/* ********************************************************************** */
static void HELP_ReadCmd (HELPDATA *Data, char *Text, int *Pos, 
			  HELP_ATTR *Attr)
{
  short count = 0, space = 0;
  int   Cmd = NO_CMD;

  while ((Cmd = HELP_GetCmd(Text, Attr, Pos)) != E_CMD)
    switch (Cmd)
    {
      case BR_CMD :
      case P_S_CMD :
      case P_E_CMD :
	HELP_AppendElem (Data, ELEM_VSPACE, NULL, 
			 (Cmd==P_S_CMD || Cmd==P_E_CMD)? BIG_SKIP:0, Attr);
	if (Attr->Head)
	{
	  Attr->Attr[FONT_SIZE] = 2;
	  Attr->Head = FALSE;
        }
	break;

      case VSPACE_CMD :
	if ((space = HELP_EvalValue(Text, Pos)) > 0)
	  HELP_AppendElem (Data, ELEM_VSPACE, NULL, space, Attr);
	break;

      case HSPACE_CMD :
        space = HELP_EvalValue (Text, Pos);
	HELP_AppendElem (Data, ELEM_HSPACE, NULL, space, Attr);
	break;

      case EM_S_CMD :
	Attr->Attr[FONT_SLANT] = 1;
	HELP_TestFlag (Data, EM_BIT, 0);
	break;

      case EM_E_CMD :
	Attr->Attr[FONT_SLANT] = 0;
	HELP_TestFlag (Data, EM_BIT, -1);
	break;

      case BD_S_CMD :
	Attr->Attr[FONT_WEIGHT] = 1;
	HELP_TestFlag (Data, BD_BIT, 0);
	break;

      case BD_E_CMD :
	Attr->Attr[FONT_WEIGHT] = 0;
	HELP_TestFlag (Data, BD_BIT, -1);
	break;

      case HN_S_CMD :
	break;

      case HN_E_CMD :
        if (Data->Param.HTML)
	  HELP_AppendElem (Data, ELEM_VSPACE, NULL, 0, Attr);
	break;

      case COLOR_S_CMD :
	Attr->ColorID = HELP_EvalColor (Text, Pos);
	HELP_TestFlag (Data, COLOR_BIT, 0);
	break;

      case COLOR_E_CMD :
	Attr->ColorID = HELP_BLACK;
	HELP_TestFlag (Data, COLOR_BIT, -1);
	break;

      case ULINE_S_CMD :
	Attr->Attr[FONT_UL] = 1;
	HELP_TestFlag (Data, ULINE_BIT, 0);
        break;

      case ULINE_E_CMD :
	Attr->Attr[FONT_UL] = 0;
	HELP_TestFlag (Data, ULINE_BIT, -1);
        break;

      case HREF_CMD :
#if MARK_REFS
	if (! RefOK)
	{
	  RefPic = TOOLMakeXPM2 (refmark_xpm, &RefX, &RefY, Colors[HELP_BG]);
	  RefOK = TRUE;
	}
#endif
	if (Data->RNo < MAX_REFS)
	{
#if MARK_REFS
	  if (RefPic != NULL)
	    HELP_AppendElem (Data, ELEM_REF, NULL, 0, Attr);
#endif
	  Data->HRef[Data->RNo]   = HELP_EvalHRef (Data, Text, Pos);
	  Attr->ID                = Data->RNo;
	  Attr->ColorID           = HELP_BLUE;
	  Attr->Attr[FONT_WEIGHT] = 1;
	  (Data->RNo)++;
	}
	HELP_TestFlag (Data, REF_BIT, 0);
	break;

      case NAME_CMD :
	if (Data->LNo < MAX_REFS)
	{
	  Data->Label[Data->LNo] = HELP_EvalLabel (Text, Pos);
	  Attr->ID = Data->LNo;
	  HELP_AppendElem (Data, ELEM_NAME, NULL, 0, Attr);
	  Attr->ID = HELP_NO_REF;
	  (Data->LNo)++;
	}
	HELP_TestFlag (Data, REF_BIT, 0);
	break;

      case A_E_CMD :
	Attr->ID                = HELP_NO_REF;
	Attr->ColorID           = HELP_BLACK;
	Attr->align             = HELP_LEFT;
	Attr->Attr[FONT_WEIGHT] = 0;
	HELP_TestFlag (Data, REF_BIT, -1);
	break;

      case ITEM_S_CMD :
      case QUOTE_S_CMD :
	HELP_AppendElem (Data, ELEM_IND, NULL, 0, Attr);
	HELP_TestFlag (Data, INDENT_BIT, 1);
	break;

      case ENUM_S_CMD :
	Attr->Number = 1;
	HELP_AppendElem (Data, ELEM_ENUM, NULL, 0, Attr);
	HELP_TestFlag (Data, INDENT_BIT, 1);
	break;

      case ITEM_CMD :
	HELP_AppendElem (Data, ELEM_ITEM, NULL, 0, Attr);
        if (Attr->Number > 0)
	  (Attr->Number)++;
	break;

      case DL_S_CMD :
	HELP_AppendElem (Data, ELEM_VSPACE, NULL, 
			      (Data->Param.HTML)? BIG_SKIP:0, Attr);
	Attr->skip = FALSE;
	HELP_TestFlag (Data, INDENT_BIT, 1);
	break;

      case DT_CMD :
	if (Attr->skip)
	  HELP_AppendElem (Data, ELEM_NIND,  NULL, 0, Attr);
	Attr->skip = TRUE;
	break;

      case DL_E_CMD :
	Attr->skip = TRUE;

      case ITEM_E_CMD :
      case ENUM_E_CMD :
      case QUOTE_E_CMD :
	Attr->Number = 0;
	HELP_AppendElem (Data, ELEM_NIND,  NULL, 0, Attr);
	HELP_TestFlag (Data, INDENT_BIT, -1);
	break;

      case IMG_CMD :
        {
	  IMG_DATA *Image = NULL;

	  if ((Image = HELP_EvalImg(Data, Text, Pos)) != NULL)
	  {
            if (Image->loaded)
	      HELP_AppendElem (Data, ELEM_IMG, (void *)Image, 0, Attr);
	    elsif (Image->alt[0] != '\0')
	    {
	      Attr->ColorID = HELP_GREEN;
	      HELP_AppendElem (Data, ELEM_TEXT, Image->alt, 0, Attr);
	      free (Image);
	      Attr->ColorID = HELP_BLACK;
	    }
	  }
	}
        break;

      case TABLE_S_CMD :
	{
	  TABLE_DATA *Table;

	  if ((Table = HELP_EvalTable(Data, Text, Pos, Attr)) != NULL)
	    HELP_AppendElem (Data, ELEM_TABLE, (void *)Table, 0, Attr);
	}
        break;

      case TABLE_E_CMD :
	break;

      case FORM_S_CMD :
	if (Data->UNo < MAX_URI)
	{
          HELP_EvalForm (Data, Text, Pos);
	  (Data->UNo)++;
	}
	HELP_TestFlag (Data, FORM_BIT, 0);
	break;

      case FORM_E_CMD :
	HELP_TestFlag (Data, FORM_BIT, -1);
	break;

      case INPUT_CMD :
	{
	  Arg         args[10];
	  XmString    Label;
	  INPUT_DATA *Input = NULL;
	  
	  if ((Input = HELP_EvalInput(Text, Pos)) != NULL)
	  {
	    PRO4 ("INPUT: %d %s='%s' Size:%d\n", 
		  Input->Type, Input->Name, Input->Value, Input->Size)

	    Label = XmStringLtoRCreate ("", XmSTRING_DEFAULT_CHARSET);
	    XtSetArg (args[0], XmNbackground,         Colors[HELP_BG]);
	    XtSetArg (args[1], XmNforeground,         Colors[Attr->ColorID]);
	    XtSetArg (args[2], XmNlabelString,        Label);
	    XtSetArg (args[3], XmNhighlightThickness, 0);
	    XtSetArg (args[4], XmNmarginHeight,       2);

	    switch (Input->Type)
	    {
	      case TEXTVALUE :
		XtSetArg (args[5], XmNvalue,   Input->Value);
		XtSetArg (args[6], XmNcolumns, Input->Size);
		Input->widget = XtCreateWidget (Input->Name,
						xmTextWidgetClass,
						Data->Work, args, 7);
		break;

	      case CHECKBOX :
		XtSetArg (args[5], XmNset,           Input->checked);
		XtSetArg (args[6], XmNindicatorSize, 14);
		XtSetArg (args[7], XmNspacing,       0);
		Input->widget = XtCreateWidget (Input->Name,
						xmToggleButtonWidgetClass,
						Data->Work, args, 8);
		break;

	      case SUBMIT :
		{
		  XmString Title;

		  Title = XmStringLtoRCreate (Input->Value,
					      XmSTRING_DEFAULT_CHARSET);
		  XtSetArg (args[5], XmNlabelString, Title);
		  Input->widget = XtCreateWidget (Input->Name, 
						  xmPushButtonWidgetClass,
						  Data->Work , args, 6);
		  XtAddCallback (Input->widget, XmNactivateCallback,
				 (XtCallbackProc)HELP_SubmitCB, Data);
		  XmStringFree (Title);
		}
	        break;

	      default :
		PRO0 ("Invalid Input-Type !\n")
	    }
	    XmStringFree (Label);
            Attr->ID = Data->UNo-1;
	    if (Input->widget != NULL)
	      HELP_AppendElem (Data, ELEM_INPUT, (void *)Input, 0, Attr);
	    else
	      free (Input);
	  }
	}
        break;

      case CLEAR_CMD :
	if (Text[*Pos] == '=')
	{
	  (*Pos)++;
	  Attr->Param = (void *)(strncasecmp(Text+(*Pos), "ALL", 3) == 0);
	}
	HELP_AppendElem (Data, ELEM_CLEAR, NULL, 0, Attr);
	Attr->Param = NULL;
	break;

      case HR_CMD :
	HELP_AppendElem (Data, ELEM_LINE, NULL, 0, Attr);
	break;

      case ALIGN_CMD :
	if (Text[*Pos] == '=')
        {
	  (*Pos)++;
	  if (strncasecmp(Text+(*Pos), "CENTER", 6) == 0)
	  {
	    Attr->align = HELP_CENTER;
	    (*Pos) += 6;
	  }
	  elsif (strncasecmp(Text+*(Pos), "RIGHT", 5) == 0)
	  {
	    Attr->align = HELP_RIGHT;
	    (*Pos) += 5;
	  }
	  elsif (strncasecmp(Text+(*Pos), "LEFT", 4) == 0)
	  {
	    Attr->align = HELP_LEFT;
	    (*Pos) += 4;
	  }
	  elsif (strncasecmp(Text+(*Pos), "FILL", 4) == 0)
	  {
	    Attr->align = HELP_FILL;
	    (*Pos) += 4;
	  }
	}
	HELP_TestFlag (Data, REF_BIT, 0);	
	break;

      case SECTN_CMD :
	HELP_AppendElem (Data, ELEM_VSPACE, NULL, BIG_SKIP, Attr);
        HELP_AppendElem (Data, ELEM_SECT,   NULL,        0, Attr);
	break;

      case PRE_S_CMD :
	if (Data->Param.HTML)
	  HELP_AppendElem (Data, ELEM_VSPACE, NULL, 0, Attr);
	Attr->Attr[FONT_PRE] = 1;
	break;

      case PRE_E_CMD :
	Attr->Attr[FONT_PRE] = 0;
	if (Data->Param.HTML)
	  HELP_AppendElem (Data, ELEM_VSPACE, NULL, 0, Attr);
	break;

      case REM_CMD :
	if (strncmp(Text+(*Pos), "-->", 3) != 0)
	{
	  while (Text[*Pos] != '\0' &&
		 (Text[*Pos] != '-' || strncmp(Text+(*Pos), "-->", 3) != 0))
	    (*Pos)++;
	  (*Pos) += 2;
	}
	break;

      case TITLE_S_CMD :
	count = 0;
	(*Pos)++;
	while (Text[*Pos] != '<' && count < PATH_MAX && Text[*Pos] != '\0')
	  Data->Title[count++] = Text[(*Pos)++];
	Data->Title[count] = '\0';
	(*Pos)++;
	break;

      case TITLE_E_CMD :
	break;

      case BG_CMD :
	{
	  char FileName[41];

	  if (HELP_GetArg(Text, "", FileName, *Pos) == 0)
	    Data->bgImage = HELP_LoadImage (Data, FileName, 
					          &Data->bgX, &Data->bgY);
	  *Pos = HELP_GetArg (Text, NULL, NULL, *Pos);
	}
	break;

      case BGCOLOR_CMD :
	  Data->bg = HELP_EvalColor (Text, Pos);
	  break;

      default :
#if 0
	if (Text[0] != '\0')
	  PRO2 ("UNKNOWN COMMAND(%d): %s\n", Cmd, Text)

#endif
	break;
    }

  if (HELP_IsSpace(Text[*Pos]) &&
      Data->last != NULL && Data->last->Type == ELEM_TEXT)
  {
    if (Cmd != ULINE_E_CMD && Cmd != HSPACE_CMD)
      Data->last->space = -1;
    else
      HELP_AppendElem (Data, ELEM_TEXT, " ", 0, Attr);
  }
}


/* ********************************************************************** */
/* ** Parses complete help text an store it contents into an elem-list ** */
/* ** INPUT:   Data    help-struct where to store the result           ** */
/* **          Text    help text                                       ** */
/* ** OUTPUT:  (dual)  returns the success of this function            ** */
/* ********************************************************************** */
static dual HELP_ParseText (HELPDATA *Data, char *Text)
{
  int       len, count;
  int       Pos = 0;
  char      Space[256] = "";
  HELP_ATTR Attr = {{WEIGHT_DEF, SLANT_DEF, SIZE_DEF, PRE_DEF, UL_DEF}, 
	            HELP_BLACK,  {0, 0, 0, 0, 0}, FALSE, TRUE,
	            HELP_NO_REF,  0,              HELP_LEFT, NULL};

  PRO_START ("\n")

  if (Data == NULL)
  {
    free (Text);
    PRO_RETURN ("Invalid Data", FALSE)
  }

  /* Reset the help-structure */
  HELP_FreeHelpData (Data, TRUE);

  Data->RNo = 0;
  Data->LNo = 0;
  Data->UNo = 0;
  for (count = 0; count < TESTBITS; ++count)
    Data->Flags[count] = 0;
#if USE_PIXMAP
  Data->Pic = None;
#endif
  Data->Selection = -1;
  Data->overHRef  = FALSE;

  /* evaluate help text */
  len = strlen (Text);
  while (Pos < len)
  {
    while (Pos < len && HELP_IsSpace(Text[Pos]))
    {
      if (Attr.Attr[FONT_PRE])
      {
	if (Text[Pos] == '\n')
	  HELP_AppendElem (Data, ELEM_VSPACE, NULL, Fonts[0].Height, &Attr);
        elsif (Text[Pos] == ' ')
	{
	  count = 0;
	  while (Text[Pos] == ' ' && count < 255)
	  {
	    Space[count++] = ' ';
	    Pos++;
	  }
	  Pos--;
	  Space[count] = '\0';
	  HELP_AppendElem (Data, ELEM_TEXT, Space, 0, &Attr);
	}
      }
      Pos++;
    }

    if (Text[Pos] == '<' && 
	(! Attr.Attr[FONT_PRE] || strncasecmp(Text+Pos, "</PRE", 5) == 0))
    {
      Pos++;
      HELP_ReadCmd (Data, Text, &Pos, &Attr);
    }
    else
      HELP_ReadWord (Data, Text, &Pos, &Attr);
  }

  /* check for commands which has not been closed */
  for (count = 0; count < 4; ++count)
    if (Data->Flags[count] > 0)
      PRO1 ("Command not finished: %s\n", TestText[count])

  if (Data->Work != NULL)
  {
    if (Data->bg > 0 && Data->bg < ColorNo)
      XtVaSetValues (Data->Work, XmNbackground, Colors[Data->bg], NULL);
    else
    {
      Colors[HELP_BG] = TOOLGetColor("grey");
      XtVaSetValues (Data->Work, XmNbackground, Colors[HELP_BG], NULL);
    }
    Data->hl = Colors[HELP_BG] ^ Colors[HELP_RED];
  }

  PRO_END
  return TRUE;
}


/* ********************************************************************** */
/* ** Adds a reference(-word) (including it's size and position) to    ** */
/* ** the internal reference-list                                      ** */
/* ** INPUT:   Data  help-structure                                    ** */
/* **          Elem  element which belongs to the reference            ** */
/* **          PosX                                                    ** */
/* **          PosY  position in the formated text                     ** */
/* ** OUTPUT:  ---                                                     ** */
/* ********************************************************************** */
static void HELP_AppendRef (HELPDATA *Data, HELP_ELEM *Elem, int PosX, int PosY)
{
  HELP_GEO *geo = NULL;
  HELP_GEO *Ptr = NULL;

  geo = (HELP_GEO*)malloc (sizeof(HELP_GEO));
  geo->PosX = PosX;
  geo->PosY = PosY;
  if (Elem->Type == ELEM_TEXT)
    geo->PosY += Fonts[Elem->FontID].DrawFont->descent;
  geo->w    = Elem->SizeX;
  geo->h    = Elem->SizeY;  
  geo->Next = NULL;

  if ((Ptr = Data->HRef[Elem->ID].geo) == NULL)
    Data->HRef[Elem->ID].geo = geo;
  else
  {
    while (Ptr->Next != NULL)
      Ptr = Ptr->Next;
    Ptr->Next = geo;
  }
}


/* ********************************************************************** */
/* ** Produce a list which contains                                    ** */
/* **  i) the vertical position of each line                           ** */
/* ** ii) a pointer into the elem-list for the first elem of each line ** */
/* ** INPUT:   Data   help-structure                                   ** */
/* **          SizeX  width of the output-window/pixmap                ** */
/* **          Seek   help-element, which we want to search            ** */
/* ** OUTPUT:  (int)  Y-position of searched element (if found)        ** */
/* ********************************************************************** */
static int HELP_CalcLines (HELPDATA *Data, int SizeX, HELP_ELEM *Seek)
{
  dual       LF = TRUE;                 /* LineFeed                       */
  dual       EOL = FALSE;               /* EndOfLine                      */
  short      align = HELP_LEFT;
  int        PosX, PosY = 10;
  int        count = 0, EOINo = 0;      /* line-counter    */
  int        Result = 0;
  int        Index = 0;                 /* Indent-counter                 */
  int        Words = 0;                 /* Words in a line                */
  int        StartX[MAX_INDENT];        /* x-Indent-Positions             */
  int        EOI[MAX_INDENT];           /* End Of Image (Y-Pos)           */
  HELP_ELEM *Elem;

  PRO_START (" ")
  PRO1 ("\\%d...", SizeX)

  PosX = StartX[0]    = MARGIN;
  EOI[0]              = MARGIN;
  Data->Lines         = 0;
  Data->Line[0].SizeY = 0;
  Data->Line[0].Max   = 0;

  HELP_ResetData (Data);

  /* run through elem-list */
  Elem = Data->first;
  while (Elem != NULL && count < MAX_LINES-1)
  {
    /* check for new line... */
    EOL = (PosX+Elem->SizeX >= SizeX-MARGIN);
    if (EOL || LF || ! (Elem->Type & (ELEM_TEXT | ELEM_REF | ELEM_NAME | 
				      ELEM_LINE | ELEM_IMG | ELEM_TABLE | 
				      ELEM_INPUT | ELEM_HSPACE))) 
    {
      if (Elem->Type == ELEM_VSPACE)
      {
	if (Data->Line[count].SizeY == 0 && Elem->space <= 0)
	  Data->Line[count].SizeY = Fonts[0].Height;
	else
	  PosY += Elem->space;
      }     

      /* setting alignment of complete line */
      if (align != HELP_LEFT)
      {
	if (align == HELP_CENTER)
	  Data->Line[count].PosX = (SizeX - PosX) / 2;
	elsif (align == HELP_RIGHT)
	  Data->Line[count].PosX = SizeX - MARGIN - PosX;
	elsif (align == HELP_FILL && Words > 1 && EOL)
	  Data->Line[count].Space = (SizeX-MARGIN-PosX) / (double)(Words-1);
	align = HELP_LEFT;
      }

      /* flow around pictures */
      if (Elem->Type == ELEM_CLEAR)
      {
	if ((int)Elem->Data == 1)
	{
	  Index = 0;
	  if (EOINo > 0)
	    EOINo--;
	  while (EOINo > 0 && EOI[EOINo] < EOI[EOINo-1])
	    EOINo--;
	  PosY = EOI[EOINo];
	}
        elsif (EOINo > 0) /* jump to EndOfImage  */
	{
	  EOINo--;
	  PosY = EOI[EOINo];
	  Index = MAX (0, Index-1);
	} 
      }
      PosY += Data->Line[count].SizeY;

      /* normal EOI reached ? */
      if (EOI[MAX(EOINo-1,0)] > 0 && PosY > EOI[MAX(EOINo-1,0)])
      {
	if (EOINo > 0)
	{
	  EOINo--;
	  Index = MAX (0, Index-1);
	}
      }

      count++;
      Data->Line[count].PosX  = StartX[MIN(Index, MAX_INDENT-1)];
      Data->Line[count].PosY  = PosY;
      Data->Line[count].SizeY = 
	(Elem->Type != ELEM_IMG && Elem->Type != ELEM_TABLE)? Elem->SizeY : 0;
      Data->Line[count].Max   = 0;
      Data->Line[count].Nr    = (Elem->Type != ELEM_VSPACE)? 0 : 1;
      Data->Line[count].List  = Elem;
      Data->Line[count].Space = 0.0;
      PosX = Data->Line[count].PosX;
      LF = FALSE;
      Words = 0;
    }

    if (Elem->Type == ELEM_HSPACE)
      PosX += Elem->space;

    /* use last alignment-value in a line: not very nice but should suffice */
    if (Elem->align != HELP_LEFT)
      align = Elem->align;

    /* reference found ? */
    if (Elem->ID != HELP_NO_REF && Elem->Type != ELEM_NAME)
      HELP_AppendRef (Data, Elem, PosX, PosY);

    if (Seek != NULL && Elem == Seek)
      Result = PosY;

    switch (Elem->Type)
    {
      case ELEM_TEXT :
	Data->Line[count].SizeY = MAX (Data->Line[count].SizeY, Elem->SizeY);
	PosX += Elem->SizeX;
	if (Elem->space > 0)
	  PosX += Elem->space;
	elsif (Elem->space < 0)
	  PosX += Fonts[Elem->FontID].Space;
	Words++;
	break;

      case ELEM_VSPACE :
      case ELEM_CLEAR :
	break;

      case ELEM_LINE :
	Data->Line[count].SizeY = MAX (Data->Line[count].SizeY, Elem->SizeY);
        PosX = SizeX;
        break;

      case ELEM_ITEM :
	break;

      case ELEM_NAME :
        if (Elem->ID != HELP_NO_REF)
	{
	  Data->Label[Elem->ID].PosX = PosX;
	  Data->Label[Elem->ID].PosY = PosY;
	}
	break;

#if MARK_REFS
      case ELEM_REF :
	if (RefPic != NULL)
	{
	  Data->Line[count].SizeY = 
	    MAX (Data->Line[count].SizeY, Elem->SizeY);
	  PosX += Elem->SizeX;
	}
	break;
#endif

      case ELEM_IMG :
	{
	  int Step;

	  Index++;
	  if (! (LF = (Index >= MAX_INDENT)))
	  {
	    PosX += MARGIN;
	    StartX[Index] = PosX + Elem->SizeX;

	    EOI[EOINo] = PosY + Elem->SizeY /* + MARGIN */;

	    for (Step = 0; Step < EOINo; ++Step)
	      if (EOI[Step] < EOI[EOINo])
		EOI[Step] = EOI[EOINo];

	    Data->Line[count].Max = MAX(Data->Line[count].Max, Elem->SizeY);

	    if (EOINo < MAX_INDENT-1)
	      EOINo++;
	  }
	  else
	    Data->Line[count].SizeY = 
	      MAX (Data->Line[count].SizeY, Elem->SizeY+10);
	  PosX += Elem->SizeX;
	}
	break;

      case ELEM_TABLE :
	{
	  int Step;

	  Index++;
	  if (! (LF = (Index >= MAX_INDENT)))
	  {
	    PosX += MARGIN;
	    StartX[Index] = PosX + Elem->SizeX;

	    EOI[EOINo] = PosY + Elem->SizeY /* + MARGIN */;

	    for (Step = 0; Step < EOINo; ++Step)
	      if (EOI[Step] < EOI[EOINo])
		EOI[Step] = EOI[EOINo];

	    Data->Line[count].Max = MAX(Data->Line[count].Max, Elem->SizeY);

	    if (EOINo < MAX_INDENT-1)
	      EOINo++;
	  }
	  else
	    Data->Line[count].SizeY = 
	      MAX (Data->Line[count].SizeY, Elem->SizeY+10);
	  PosX += Elem->SizeX;
	}
	break;

      case ELEM_INPUT :
	{
	  INPUT_DATA *Input = (INPUT_DATA*)Elem->Data;

	  XtVaSetValues (Input->widget, XmNx, PosX,
			                XmNy, PosY+7,
			                NULL);
	  XtManageChild (Input->widget);

	  PosX += Elem->SizeX+4;
	  Data->Line[count].SizeY = MAX (Data->Line[count].SizeY, 
					 Elem->SizeY);
	}
	break;

      case ELEM_IND :
      case ELEM_ENUM :
	Index++;
        if (Index < MAX_INDENT)
	  StartX[Index] = StartX[Index-1] + ((Elem->Type==ELEM_IND)? 15 : 25);
	LF = TRUE;
	break;

      case ELEM_NIND :
	Index = MAX (0, Index-1);
	LF = TRUE;
	break;

      default :
	PRO1 ("ELEM-TYPE unknown: %d\n", Elem->Type)
    }
    Data->Line[count].Max = 
      MAX(Data->Line[count].Max, MAX(Data->Line[count].SizeY, Elem->SizeY));

    if (Elem->Type != ELEM_VSPACE)
      (Data->Line[count].Nr)++;

    Elem = Elem->Next;
  }
  Data->Lines = count+1;

  Data->SizeX = SizeX;
  Data->SizeY = MAX (Data->WinY, MAX(PosY+Data->Line[count].Max+10,
				     EOI[MAX(EOINo-1,0)]+10)); 
  Data->MaxY = Data->SizeY - Data->WinY;

  if (Data->Work != NULL)
    XtVaSetValues (Data->Work, XmNwidth, Data->SizeX-2,
		               XmNheight, Data->SizeY,
		               NULL);

  PRO_END
  return Result;
}


/* ********************************************************************** */
/* ** Draws a border using a the specified mode.                       ** */
/* ** INPUT:  Data          Pointer to help-data                       ** */
/* **         PosX, PosY    Upper left corner of the box               ** */
/* **         SizeX, SizeY  width and height of the border             ** */
/* **         Mode          0: normal black box                        ** */
/* **                       1: shadow on bottom-right                  ** */
/* **                       2: shadow on upper-left corner             ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
static void HELP_DrawBorder (HELPDATA *Data, 
			    int PosX, int PosY, int SizeX, int SizeY,
			    int Mode)
{
  XSetForeground (XDisplay, Data->DrawGC, 
		  (Mode & 1)? Colors[HELP_WHITE]:Colors[HELP_BLACK]);
  if (Mode == 0)
    XDrawRectangle (XDisplay, GRAPHIC, Data->DrawGC, PosX, PosY, SizeX, SizeY);
  else
  {
    XDrawLine (XDisplay, GRAPHIC, Data->DrawGC,
	                 PosX, PosY, PosX+SizeX, PosY);
    XDrawLine (XDisplay, GRAPHIC, Data->DrawGC,
	                 PosX, PosY, PosX, PosY+SizeY);

    XSetForeground (XDisplay, Data->DrawGC, 
		    (Mode & 1)? Colors[HELP_BLACK]:Colors[HELP_WHITE]);
    XDrawLine (XDisplay, GRAPHIC, Data->DrawGC,
	                 PosX+SizeX, PosY, PosX+SizeX, PosY+SizeY);
    XDrawLine (XDisplay, GRAPHIC, Data->DrawGC,
	                 PosX, PosY+SizeY, PosX+SizeX, PosY+SizeY);
  }
}


/* ********************************************************************** */
/* ** DrawBackground fills the background of a specified range with a  ** */
/* ** specified pixmap.                                                ** */
/* ** INPUT:  Data          Pointer to help-data                       ** */
/* **         x1, y1        Upper left corner of the box               ** */
/* **         x2, y2        Lower right corner                         ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
static void HELP_DrawBackground (HELPDATA *Data, int x1, int y1, int x2, int y2)
{
  int PosX, PosY;
  int SizeX, SizeY;
  int OffsX, OffsY;

  if (Data->bgImage == NULL || Data->bgX == 0 || Data->bgY == 0)
    XClearArea (XDisplay, GRAPHIC, x1, y1, x2-x1, y2-y1, False);
  else
  {
    PosY = y1;
    OffsY = PosY % Data->bgY;
    while (PosY < y2)
    {
      PosX = x1;
      OffsX = PosX % Data->bgX;
      SizeY = MIN (Data->bgY-OffsY, y2-PosY);
      while (PosX < x2)
      {
	SizeX = MIN (Data->bgX-OffsX, x2-PosX);
	XPutImage (XDisplay, GRAPHIC, Data->DrawGC, Data->bgImage,
		             OffsX, OffsY, PosX, PosY, SizeX, SizeY);
	PosX += SizeX;
	OffsX = 0;
      }
      PosY += SizeY;
      OffsY = 0;
    }
  }
}


/* ********************************************************************** */
/* ** redraws the text between the vertical position 'start' and 'end' ** */
/* ** using the line-list                                              ** */
/* ** INPUT:  Data   help-structure                                    ** */
/* **         start  start of redrawing [Pixel]                        ** */
/* **         end    end of output [Pixel]                             ** */
/* ** OUTPUT: none                                                     ** */
/* ********************************************************************** */
void HELP_RefreshText (HELPDATA *Data, int start, int end)
{
  short      ColorID = 0; 
  short      FontID = 0;
  int        line = 1, count;
  int        PosX, PosY;
  int        SizeX, SizeY, Num = 0;
  int        Sum = 0;
  double     Space = 0.0;
  XTextItem  Word;
  HELP_ELEM *Elem;

  PRO_START (" ")

  PRO2 ("\\Range: [%d..%d]...", start, end)

  Word.chars  = NULL;
  Word.nchars = 0;
  Word.delta  = 0;

#if USE_PIXMAP
  if (Data->Pic != None)
    XFreePixmap (XDisplay, Data->Pic);
  Data->Pic = XCreatePixmap (XDisplay, Data->Win,
			     Data->SizeX-6, Data->SizeY, 8);
  XSetForeground (XDisplay, Data->DrawGC, Colors[HELP_BG]);
  XFillRectangle (XDisplay, Data->Pic, Data->DrawGC,
		            0, 0, Data->SizeX, Data->SizeY);
#endif

  XSetForeground (XDisplay, Data->DrawGC, Colors[ColorID]);

  /* compute starting line */
  while (line+1 < Data->Lines && 
	 Data->Line[line+1].PosY + 
	 MAX(Data->Line[line+1].SizeY, Data->Line[line+1].Max) < start)
    line++;
  line--;

  PosY = end;
  /* draw the elements until end is reached */
  while (line < Data->Lines && Data->Line[MAX(0,line-1)].PosY <= end)
  {
    if (Data->Line[line].Space != 0.0)       /* initialized fill-line     */
    {
      Space = 0.0;
      Sum   = 0;
    }

    PosX = Data->Line[line].PosX;
    PosY = Data->Line[line].PosY + Data->Line[line].SizeY;
    Elem = Data->Line[line].List;
    for (count = 0; count < Data->Line[line].Nr; ++count)
    {
      if (Elem->ColorID != ColorID)
      {
	ColorID = Elem->ColorID;
	XSetForeground (XDisplay, Data->DrawGC, Colors[ColorID]);
      }

      if (Elem->FontID != FontID)
	FontID = Elem->FontID;

      switch (Elem->Type)
      {
        case ELEM_TEXT :
	  Word.chars  = (char *)Elem->Data;
	  Word.nchars = strlen ((char *)Elem->Data);
	  Word.font   = Fonts[FontID].DrawFont->fid;
#if SHADOW_REF
	  if (Elem->ID != HELP_NO_REF)
	  {
#if BUTTON_REF
	    XSetForeground (XDisplay, Data->DrawGC, Colors[HELP_WHITE]);
	    XDrawText (XDisplay, GRAPHIC, Data->DrawGC, 
		                 PosX-1, PosY-1, &Word, 1);
#endif
	    XSetForeground (XDisplay, Data->DrawGC, Colors[HELP_BLACK]);
	    XDrawText (XDisplay, GRAPHIC, Data->DrawGC, 
		                 PosX+1, PosY+1, &Word, 1);
	    XSetForeground (XDisplay, Data->DrawGC, Colors[ColorID]);
	  }
#endif
	  XDrawText (XDisplay, GRAPHIC, Data->DrawGC, PosX, PosY, &Word, 1);

#if SEARCHING
	  if (Elem->fits)             /* mark hits of search            */
	  {
	    XSetForeground (XDisplay, Data->DrawGC, Colors[HELP_RED]);
	    XDrawRectangle (XDisplay, GRAPHIC, Data->DrawGC, 
			              PosX-2,   PosY-Elem->SizeY+
			              Fonts[FontID].DrawFont->descent, 
			              Elem->SizeX+3, Elem->SizeY); 
	    XSetForeground (XDisplay, Data->DrawGC, Colors[ColorID]);
	  }
#endif

	  SizeX = Elem->SizeX;
	  if (Elem->space > 0)
	    SizeX += Elem->space;
	  elsif (Elem->space < 0)
	    SizeX += Fonts[Elem->FontID].Space;

	  if (Elem->line)
	    XDrawLine (XDisplay, GRAPHIC, Data->DrawGC, 
		                 PosX, PosY+1, PosX+SizeX, PosY+1);
	  PosX += SizeX;

	  if (Data->Line[line].Space != 0.0)
	  {
	    Space += Data->Line[line].Space;
	    if (Space - Sum > 1.0)
	    {
	      PosX += (int)(Space-Sum);
	      Sum += (int)(Space-Sum);
	    }
	  }
	  break;

	case ELEM_VSPACE :
#if 0
	  PosY += Elem->space;
#endif
        case ELEM_CLEAR :
	case ELEM_NAME :
	case ELEM_IND :
	case ELEM_PRE :
	  break;

        case ELEM_HSPACE :
	  PosX += Elem->space;
	  break;

	case ELEM_NIND :
	  Num = 0;
	  break;

	case ELEM_LINE :
	  {
	    int OffsY;

	    OffsY = (int)(Data->Line[line].SizeY * 0.30);
	    XSetForeground (XDisplay, Data->DrawGC, Colors[HELP_BLACK]);
	    XDrawLine (XDisplay, GRAPHIC,  Data->DrawGC, 
	       PosX+1, PosY-OffsY+1, Data->SizeX-MARGIN-3, PosY-OffsY+1);
	    XDrawPoint (XDisplay, GRAPHIC, Data->DrawGC,
	       Data->SizeX-MARGIN-3, PosY-OffsY);
	    XSetForeground (XDisplay, Data->DrawGC, Colors[HELP_WHITE]);
	    XDrawLine (XDisplay, GRAPHIC,  Data->DrawGC,
	       PosX, PosY-OffsY-1, Data->SizeX-MARGIN-4, PosY-OffsY-1);
	    XDrawPoint (XDisplay, GRAPHIC, Data->DrawGC, PosX, PosY-OffsY);
	    XSetForeground (XDisplay, Data->DrawGC, Colors[ColorID]);
	    if (ColorID != HELP_BLACK)
	      XDrawLine (XDisplay, GRAPHIC,  Data->DrawGC, 
		PosX+1, PosY-OffsY, Data->SizeX-MARGIN-4, PosY-OffsY);
	  }
	  break;

	case ELEM_ITEM :
	  {
	    char NumStr[4];
	    int  width, Number;

	    Number = *(int*)Elem->Data;
	    if (Number == 0)
	      XFillArc (XDisplay, GRAPHIC, Data->DrawGC,
			          PosX-12, PosY-Data->Line[line].SizeY/2-2, 
			          8, 8, 0, 360*64);
	    else
	    {
	      sprintf (NumStr, "%2d.", Number % 100);  NumStr[3] = '\0';
	      Word.chars  = NumStr;
	      Word.nchars = strlen (NumStr);
	      Word.font   = Fonts[FontID].DrawFont->fid;
	      width = XTextWidth (Fonts[FontID].DrawFont, NumStr, Word.nchars);
	      XDrawText (XDisplay, GRAPHIC, Data->DrawGC, 
			 PosX-width-6, PosY, &Word, 1);
	    }
	  }
	  break;

	case ELEM_ENUM :
	  Num = 1;
	  break;

	case ELEM_IMG :
	  {
	    IMG_DATA *Image = (IMG_DATA*)Elem->Data;

	    XPutImage (XDisplay, GRAPHIC, Data->DrawGC, Image->Image,
		       0, 0, PosX, PosY-10, Elem->SizeX, Elem->SizeY);
	    if (Elem->ID != HELP_NO_REF)
	    {
	      Data->HRef[Elem->ID].geo->PosX = PosX;
	      Data->HRef[Elem->ID].geo->PosY = PosY-10;
	    }
	    PosX += Elem->SizeX + MARGIN;
	  }
	  break;

        case ELEM_TABLE :
	  {
	    int         row, col, Border;
	    int         x, y, Index, px;
	    int         TextX, TextY;
	    int         sum;
	    TABLE_DATA *Table = (TABLE_DATA*)Elem->Data;
	    HELP_ELEM  *Ptr, *Ptr2;

	    y = PosY - Data->Line[line].SizeY + 5;
	    if ((Border = Table->Border-1) > 0)
	      HELP_DrawBorder (Data, PosX, y, Table->SizeX, Table->SizeY, 1);

	    y += Border;
	    for (row = 0; row < Table->Rows; ++row)
	    {
	      x = PosX + Border;
	      for (col = 0; col < Table->Cols; ++col)
	      {
		if (Table->Data[row][col].Type != -1)
		{
		  SizeX = 0;
		  for (Index = col;
		       Index < col+Table->Data[row][col].ColSpan; ++Index)
		    SizeX += Table->w[Index];
		  SizeX += (Table->Data[row][col].ColSpan-1) * Border;

		  SizeY = 0;
		  for (Index = row;
		       Index < row+Table->Data[row][col].RowSpan; ++Index)
		    SizeY += Table->h[Index];
		  SizeY += (Table->Data[row][col].RowSpan-1) * Border;

		  if (Table->Data[row][col].Color != HELP_BG)
		  {
		    XSetForeground (XDisplay, Data->DrawGC, 
				    Colors[Table->Data[row][col].Color]);
		    XFillRectangle (XDisplay, GRAPHIC, Data->DrawGC,
				    x+1, y+1, SizeX-2, SizeY-2);
		    XSetForeground (XDisplay, Data->DrawGC, Colors[ColorID]);
		  }

		  if (Border >= 0)
   		    HELP_DrawBorder (Data, x, y, SizeX, SizeY,
				    (Border==0)? 0:2);
		  XSetForeground (XDisplay, Data->DrawGC, Colors[ColorID]);

                  /* Draw cell-data (text) */
                  px = x + 3;
		  TextY = 0;
		  Ptr = Table->Data[row][col].List;
		  while (Ptr != NULL)
		  {
		    Word.chars  = (char *)Ptr->Data;
		    Word.nchars = strlen ((char *)Ptr->Data);
		    Word.font   = Fonts[Ptr->FontID].DrawFont->fid;

		    if (px + Ptr->SizeX >= x + SizeX - 3)
		    {
		      TextY += Ptr->SizeY;
		      px = x + 3;
		    }
		    if (px == x+3 && 
			Table->Data[row][col].align != HELP_LEFT)
		    {
		      Ptr2 = Ptr;
		      sum = 0;
		      while (Ptr2 != NULL && sum +Ptr->SizeX < SizeX - 6)
		      {
			sum += Ptr2->SizeX;
			if (Ptr2->space > 0)
			  sum += Ptr2->space;
			elsif (Ptr2->space < 0)
			  sum += Fonts[Ptr2->FontID].Space;
			Ptr2 = Ptr2->Next;
		      }
		      if (Table->Data[row][col].align == HELP_CENTER)
			px += (SizeX - sum) / 2;
		      elsif (Table->Data[row][col].align == HELP_RIGHT)
			px += SizeX - sum;
		    }

		    XDrawText (XDisplay, GRAPHIC, Data->DrawGC, 
			       px, y+TextY+14, &Word, 1);

		    TextX = Ptr->SizeX;
		    if (Ptr->space > 0)
		      TextX += Ptr->space;
		    elsif (Ptr->space < 0)
		      TextX += Fonts[Ptr->FontID].Space;

		    if (Ptr->line)
		      XDrawLine (XDisplay, GRAPHIC, Data->DrawGC, 
				 px, y+TextY+14, px+TextX, y+TextY+14);
		    px += TextX;

		    Ptr = Ptr->Next;
		  }
		}
                x += Table->w[col] + Border;
	      }
	      y += Table->h[row] + Border;
	    }
	    XSetForeground (XDisplay, Data->DrawGC, Colors[ColorID]);
	    PosX += Elem->SizeX + MARGIN;
	  }
	  break;

        case ELEM_INPUT :
	  PosX += Elem->SizeX+4;
	  break;

#if MARK_REFS
        case ELEM_REF :
	  if (RefPic != NULL)
	  {
	    XPutImage (XDisplay, GRAPHIC, Data->DrawGC, RefPic,
		       0, 0, PosX, PosY-RefY, Elem->SizeX, Elem->SizeY);
	    PosX += Elem->SizeX;
	  }
	  break;
#endif

	default :
	  PRO1 ("INVALID COMMAND-TYPE: %d\n", Elem->Type)
      }
      Elem = Elem->Next;
    }
    line++;
  }

  if (line == Data->Lines && PosY < end)
  {
    XSetForeground (XDisplay, Data->DrawGC, Colors[HELP_BG]);
    XFillRectangle (XDisplay, Data->Win, Data->DrawGC, 
		              0, end, Data->SizeX, Data->WinY);
  }

#if USE_LABEL
  XtVaSetValues (Data->Work, XmNlabelPixmap, Data->Pic, NULL);
#endif

  PRO_END
}


/* ********************************************************************** */
/* ** Locates the index of a given label in the label-list             ** */
/* ** INPUT:   Data   help-structure (contains label-list)             ** */
/* **          Label  label string to look for                         ** */ 
/* ** OUTPUT:  (int)  index of label in the list                       ** */
/* **          -1     if label not in list                             ** */
/* ********************************************************************** */
static int HELP_SearchLabel (HELPDATA *Data, char *Label)
{
  int count = 0;

  while (count < Data->LNo)
    if (strcmp(Data->Label[count].Name, Label) == 0)
      return count;
    else
      count++;

  return -1;
}


/* ********************************************************************** */
/* ** search a href, which matches the current mouse pos. and delivers ** */
/* ** it's index into the data's reference-list or -1 if no href found ** */
/* ** at this position.                                                ** */
/* ** INPUT:  Data   help-structure (with ref-list)                    ** */
/* ** OUTPUT: (int)  index of reference                                ** */
/* **         -1     No reference at mouse position                    ** */
/* ********************************************************************** */
static int HELP_LocateHREF (HELPDATA *Data)
{
  int       count = 0;
  HELP_GEO *Ptr = NULL;

  while (count < Data->RNo)
  {
    Ptr = Data->HRef[count].geo;
    while (Ptr != NULL)
    {
      if (MouseX >= Ptr->PosX && MouseX <= Ptr->PosX + Ptr->w &&
	  MouseY >= Ptr->PosY && MouseY <= Ptr->PosY + Ptr->h)
      {
	PRO2 ("HREF: %s # %s\n",
	      Data->HRef[count].FileName, Data->HRef[count].Label)
	return count;
      }
      else
	Ptr = Ptr->Next;
    }
    count++;
  }
  return -1;
}


/* ********************************************************************** */
/* ** Sets the value of a scrollbar. Uses XmScrollBarSetValues because ** */
/* ** XtSetArg do not actualize the scrolled window                    ** */
/* ** INPUT:  ScrBar  the scrollbar-widget                             ** */
/* **         Value   preferred value (may be changed to fit into      ** */
/* **                                  scrollbar-range)                ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
static void HELP_SetScrBar (Widget ScrBar, int Value)
{
  int Dummy;
  int Size, Inc, Page;

  XmScrollBarGetValues (ScrBar, &Dummy, &Size, &Inc, &Page);
  XmScrollBarSetValues (ScrBar, MAX(0, Value), Size, Inc, Page, True);    
}


/* ********************************************************************** */
/* ** Uses the clipping window to fetch the actual size of the visible ** */
/* ** area of the scrolled win. The result will be directly written    ** */
/* ** into the structure.                                              ** */
/* ** INPUT:  Data  helptext structure                                 ** */
/* ** OUTPUT: none                                                     ** */
/* ********************************************************************** */
static void HELP_GetSize (HELPDATA *Data)
{
  Dimension width, height;

  XtVaGetValues (Data->ClipWin, XmNwidth,  &width,
		                XmNheight, &height,
		                NULL);

  Data->WinX = (int)width;
  Data->WinY = (int)height;
}


/* ********************************************************************** */
/* ** Refreshes the complete visible window/picture area               ** */
/* ** INPUT:  Data     helpdata-structure                              ** */
/* **         Label    if given: used to calc the start pos            ** */
/* **         PosY     otherwise this value is used                    ** */
/* **         Refresh  a complete refresh is needed                    ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
void HELP_UpdateWindow (HELPDATA *Data, char *Label, int PosY, dual Refresh)
{
  int Index = -1;

  HELP_GetSize (Data);

  if (Label != NULL && Label[0] != '\0')
    if ((Index = HELP_SearchLabel(Data, Label)) > -1)
      PosY = Data->Label[Index].PosY;

  PosY = MAX (0, MIN(PosY, Data->MaxY));

  HELP_SetScrBar (Data->ScrBarY, PosY);

  if (Refresh)
  {
#if USE_PIXMAP
    HELP_RefreshText (Data, 0, Data->SizeY);
    XCopyArea (XDisplay, Data->Pic, Data->Win, Data->DrawGC, 
	                 0, PosY, Data->SizeX, Data->WinY, 0, PosY);
#else

#if 0
    XClearWindow (XDisplay, Data->Win);
#else
    HELP_DrawBackground (Data, 0, PosY, Data->SizeX, PosY+Data->WinY);
#endif

  HELP_RefreshText (Data, PosY, PosY+Data->WinY);

#endif
  }
}


/* ********************************************************************** */
/* ** Marks/Unmarks a reference (afer selecting it) by inverting it's  ** */
/* ** area. Multiple reference parts are possible (e.g. words.         ** */
/* ** INPUT:  Data  helptext-data                                      ** */
/* **         ID    Number of the reference                            ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
void HELP_MarkHRef (HELPDATA *Data, int ID)
{
  HELP_GEO *Ptr = NULL;

  /* 
     TODO:
     Markierung auch beim Scrollen erhalten
  */
  PRO1 ("Marking %d\n", ID);

  if (Data->Param.Click && (Ptr = Data->HRef[ID].geo) != NULL)
  {
    XSetForeground (XDisplay, XOR_GC, Colors[HELP_BG]);
    while (Ptr != NULL)
    {
      XFillRectangle (XDisplay, Data->Win, XOR_GC, 
		                Ptr->PosX, Ptr->PosY, Ptr->w, Ptr->h);
      Ptr = Ptr->Next;
    }
  }
}


/* ********************************************************************** */
/* ** Locates the first element which is displayed at the given        ** */
/* ** window-line. Needed to reposition the text after changing the    ** */
/* ** window-size.                                                     ** */
/* ** INPUT:  Data          helptext-data                              ** */
/* **         PosY          Position of (first) window-line            ** */
/* ** OUTPUT: (HELP_ELEM*)  first element in the line                  ** */
/* ** ******************************************************************* */
static HELP_ELEM *HELP_GetFirstElem (HELPDATA *Data, int PosY)
{
  int count = 0;

  while (count < Data->Lines && Data->Line[count].PosY+5 < PosY)
    count++;

  return Data->Line[count].List;
}


/* ********************************************************************** */
/* ** Event handler reacting on the following events which belongs to  ** */
/* ** the output window:                                               ** */
/* ** - events evoked by moving the mouse (in this window)             ** */
/* ** - expose events for redrawing the text                           ** */
/* ** - resizing events (configurations)                               ** */
/* ** - pressing the left mouse button                                 ** */
/* ** - moving arround using the cursor-keys                           ** */
/* ** INPUT:   widget  widget which caused the event                   ** */
/* **          Data    the helptext-structure                          ** */
/* **          event   event data (reason, ...)                        ** */
/* ** OUTPUT:  ---                                                     ** */
/* ********************************************************************** */
static void HELP_HandleEvents (Widget widget, HELPDATA *Data, XEvent *Event)
{
  static int py;

  switch (Event->type)
  {
    case MotionNotify :
      {
	int           PosY;
        XMotionEvent *Motion = (XMotionEvent *)Event;

	MouseX = Motion->x;
	MouseY = Motion->y;

	if (Motion->state & Button1Mask && py >= 0)    /* Drag with mouse */
	{
	  /* devide new scrollbar-difference by two, because the 
             mouse-position is relative to the scrolled windows and would
             be changed twice the time. Formula is equivalent to
	       Scroll_Y_Start +/- 0.5*(MouseY-Mouse_Y_Start),
	     but needs only one start-value and. */
	  PosY = MIN (py-(MOUSE_DRAG 0.75*MouseY), Data->MaxY);
	  HELP_SetScrBar (Data->ScrBarY, PosY);
	}
	elsif (HELP_LocateHREF(Data) >= 0)
	{
	  if (! Data->overHRef)
	    TOOLSetMouseCursor (Data->Shell, ToolHAND, "blue", "white");
	  Data->overHRef = TRUE;
	}
	elsif (Data->overHRef)
	{
	  TOOLSetMouseCursor (Data->Shell, ToolNORMAL, NULL, NULL);
	  Data->overHRef = FALSE;
	}
      }
      break;

    case Expose :
      {
        XExposeEvent *expose = (XExposeEvent*)Event;

	PRO4 ("EXPOSE [%d][%d]: %dx%d\n",
	      expose->x, expose->y, expose->width, expose->height)

	HELP_DrawBackground (Data, expose->x, expose->y, 
			          expose->x + expose->width, 
			          expose->y + expose->height);

#if USE_PIXMAP
	if (! done)
	  XCopyArea (XDisplay, Data->Pic, expose->window, Data->DrawGC,
		     expose->x, expose->y, expose->width, expose->height,
		     expose->x, expose->y);
	else
	  done = FALSE;
#else
	HELP_RefreshText (Data, expose->y, expose->y + expose->height);
#endif
      }
      break;

    case ConfigureNotify :
      {
        int        SizeX;
	int        PosY;
	HELP_ELEM *Elem;

        HELP_GetSize (Data);
        SizeX = MAX (MIN_WIDTH, Data->WinX);

        if (Data->SizeX != SizeX)
	{
	  PRO4 ("RESIZE: %dx%d  CLIP: %dx%d\n",
		Data->SizeX, Data->SizeY, Data->WinX, Data->WinY)

	  XtVaGetValues (Data->ScrBarY, XmNvalue, &PosY, NULL);
          Elem = HELP_GetFirstElem (Data, PosY);

	  SizeX += 2;
	  PosY = HELP_CalcLines (Data, SizeX, Elem);
	  HELP_UpdateWindow (Data, NULL, PosY, TRUE);

          if (Data->Param.Click && Data->Selection > -1)
	    HELP_MarkHRef (Data, Data->Selection);
	}
	/* Data->MaxY = Data->SizeY - Data->WinY; */
      }
      break;

    case ButtonPress :
      {
	int Number;

	if (Event->xbutton.button == Button1)
	{
	  py = -1;
	  PRO2 ("CLICK: %d / %d\n", MouseX, MouseY)
	  if ((Number = HELP_LocateHREF(Data)) >= 0)
	  {
	    if (Data->Param.Click &&
		(Data->Selection == -1 || Data->Selection != Number))
	    {
	      if (Data->Selection >= 0)
		HELP_MarkHRef (Data, Data->Selection);
	      HELP_MarkHRef (Data, Number);
	      XtSetSensitive (Data->Follow, TRUE);
	      Data->Selection = Number;
	    }
	    else
	    {
	      Data->Selection = Number;
	      HELP_FollowCB (widget, Data, NULL);
	    }
	  }
	  elsif (Data->Param.Click && Data->Selection > -1)
	  {
	    HELP_MarkHRef (Data, Data->Selection);
	    Data->Selection = -1;
	  }
	  else
	  {
	    XtVaGetValues (Data->ScrBarY, XmNvalue, &py, NULL);
	    py += MOUSE_DRAG 0.5 * MouseY;
	    TOOLSetMouseCursor (Data->Shell, 42, NULL, NULL);
	  }
	}
      }
      break;

    case ButtonRelease :
      TOOLSetMouseCursor (Data->Shell, ToolNORMAL, NULL, NULL);
      break;  

    case KeyPress :
    {
      XKeyPressedEvent *KeyEvent = (XKeyPressedEvent*)(&(Event->xkey));
      int               Modifier;
      KeySym            Symbol;

      Symbol   = XLookupKeysym (KeyEvent, 0);
      Modifier = KeyEvent->state;

      switch (Symbol)
      {
        case XK_Left : 
	  HELP_SetScrBar (Data->ScrBarX, 0);
	  break;

	case XK_Right :
	  HELP_SetScrBar (Data->ScrBarX, Data->SizeX-Data->WinX);
	  break;

	case XK_Up :
	case XK_Down :
        case XK_Page_Up :
        case XK_Page_Down :
        case XK_Home :
        case XK_End :
	{
	  int PosY;

	  XtVaGetValues (Data->ScrBarY, XmNvalue, &PosY, NULL);

	  switch (Symbol)
	  {
	    case XK_Down :
	      PosY = MIN (Data->MaxY, PosY + KEY_STEP);
	      break;
  	    case XK_Up : 
	      PosY = MAX (0, PosY - KEY_STEP);
	      break;
  	    case XK_Page_Down :
	      PosY = MIN (Data->MaxY, PosY + Data->WinY);
	      break;
	    case XK_Page_Up :
	      PosY = MAX (0, PosY - Data->WinY);
	      break;
	    case XK_Home :
	      PosY = 0;
	      break;
  	    case XK_End :
	      PosY = Data->MaxY;
	      break;
  	    default :
	      PRO0 ("Invalid Key")
	  }
	  HELP_SetScrBar (Data->ScrBarY, PosY);
	}
	break;

        case XK_KP_Subtract :
        case XK_KP_Add :
	  {
	    dual       done = FALSE;
	    int        PosY;
	    HELP_ELEM *Elem;

	    if (Symbol == XK_KP_Add)
	      done = HELP_ChangeScale (Data, MIN(Data->Param.Scale+1,SCALES-1));
	    else
	      done = HELP_ChangeScale (Data, MAX(0,Data->Param.Scale-1));
	    if (done)
	    {
	      HELP_GetSize (Data);
	      XtVaGetValues (Data->ScrBarY, XmNvalue, &PosY, NULL);
	      Elem = HELP_GetFirstElem (Data, PosY);
	      PosY = HELP_CalcLines (Data, Data->SizeX, Elem);
	      HELP_UpdateWindow (Data, NULL, PosY, TRUE);
	    }
	  }
          break;

        case XK_Escape :
	     HELP_ExitCB (widget, Data, NULL);
	     break;

        case XK_s :
        case XK_S :
          if (Modifier == ControlMask)            /* Alt-Key: Mod1Mask    */
		  HELP_DoSearchCB (widget, Data, NULL);
		break;

#if 1
        case XK_Print :
	  HELP_PrintText (Data, TRUE);
	  break;

        case XK_Insert :
	  {
	    int   PosY;
	    char *Text = NULL;

	    if ((Text = HELP_LoadText(Data, Data->FileName)) == NULL)
	      return;

	    if (HELP_ParseText(Data, Text))
	    {
	      XtVaGetValues (Data->ScrBarY, XmNvalue, &PosY, NULL);
	      HELP_CalcLines (Data, Data->SizeX, NULL);
	      HELP_UpdateWindow (Data, NULL, PosY, TRUE);
	    }
	    TOOLSetMouseCursor (Data->Shell, ToolNORMAL, NULL, NULL);
	    free (Text);
	  }
	  break;
#endif

	default : ;
      }
    }
    break;

    default : ;
  }
}


/* ********************************************************************** */
/* ** This is a help function to print out the internal data           ** */
/* ** INPUT:  Data  help-structue                                      ** */
/* **         Text  if set to FALSE, the elem-list is skiped           ** */
/* ** OUTPUT: none                                                     ** */
/* ********************************************************************** */
static void HELP_PrintText (HELPDATA *Data, dual Text)
{
  int        count;
  REF        HRef;
  HELP_ELEM *Elem;

  PRO_START ("\n")

  if (Text)
  {
    Elem = Data->first;
    while (Elem != NULL)
    {
      PRO3 ("ENTRY[%c]: %3dx%3d ", 
	    (Elem->Type==ELEM_TEXT)?'T':
	    ((Elem->Type==ELEM_ITEM)?'I':
	     ((Elem->Type==ELEM_LINE)?'L':
	      ((Elem->Type==ELEM_VSPACE)?'V':
	       ((Elem->Type==ELEM_HSPACE)?'H':
		((Elem->Type==ELEM_NAME)?'N':
		 ((Elem->Type==ELEM_PRE)?'P':'Q')))))),
	          Elem->SizeX, Elem->SizeY)
      PRO3 ("\\ %d %s |%s|\n", 
	    Elem->FontID, Elem->Data, (Elem->space>0)?" ":"")
      Elem = Elem->Next;
    }
  }

  PRO0 ("\\\n")
  PRO1 ("HREFs: %d\n", Data->RNo)
  for (count = 0; count < Data->RNo; ++count)
  {
    HRef = Data->HRef[count];
    PRO1 ("%2d: ", count)
    if (HRef.geo != NULL)
      PRO2 ("\\[%d][%d] ", HRef.geo->PosX, HRef.geo->PosY)
    PRO2 ("\\%s @ %s\n", HRef.Label, HRef.FileName)
  }

  PRO0 ("\\\n")
  PRO1 ("LABELs: %d\n", Data->LNo)
  for (count = 0; count < Data->LNo; ++count)
    PRO4 ("%2d: [%d][%d] %s\n", count, 
	  Data->Label[count].PosX, Data->Label[count].PosY, 
	  Data->Label[count].Name)

  PRO0 ("\\\n")
  PRO1 ("LINEs: %d\n", Data->Lines)
  for (count = 0; count < Data->Lines; ++count)
    if (Data->Line[count].List != NULL)
      PRO5 ("%3d: #%d [%3d][%3d] %s\n", 
	    count, Data->Line[count].Nr, Data->Line[count].PosX,
	    Data->Line[count].PosY, Data->Line[count].List->Data)

  PRO_END
}


/* ********************************************************************** */
/* ** internal function which creates the help-window                  ** */
/* ** INPUT:  same as HELPOpenHelpWin                                   ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
static HELPDATA *HELP_CreateHelpWin (      Widget  Parent, 
				     const char   *Title, const char *TextSpec,
					   char   *Pattern,     byte  Method,
				           HELPCB  HelpCB)
{
  int        n = 0, Pos = 0;
  char      *Text = NULL;
  Dimension  width;
  XmString   Label;
  Widget     Form, Line, Buttons;
  XGCValues  GCVal;
  Arg        args[10];
  REF        Target = {0, "", "", NULL};
  HELPDATA  *Data = NULL;

  PRO_START (" ")

  /* create new data-structure */
  Data = HELP_AddData (Title);
  Data->Shell  = Parent;
  Data->HelpCB = HelpCB;

  XtAddCallback (Data->Shell, XmNdestroyCallback, 
			      (XtCallbackProc)HELP_ExitCB, Data);

  Form = XtCreateManagedWidget ("form", xmFormWidgetClass, 
				        Data->Shell, NULL, 0);

  /* now, read and parse the text... */
  if (strncasecmp(TextSpec, "TEXT:", 5) != 0)
  {
    Target = HELP_EvalFileSpec ("", TextSpec, &Pos);
    Text = HELP_LoadText (Data, Target.FileName);
  }
  else 
    Text = HELP_LoadText (Data, TextSpec);

  /* Create a form-widget for the buttons usded for vertical centered elems */
  PRO0 ("\\Buttons...")
  Buttons = XtVaCreateManagedWidget ("buttons", xmFormWidgetClass, Form,
				     XmNleftAttachment,   XmATTACH_FORM,
				     XmNleftOffset,       5,
				     XmNrightAttachment,  XmATTACH_FORM,
				     XmNrightOffset,      5,
				     XmNbottomAttachment, XmATTACH_FORM,
				     XmNbottomOffset,     5,
				     NULL);

  /* BACK-button: */
  Label = XmStringLtoRCreate ("BACK", XmSTRING_DEFAULT_CHARSET);
  Data->Back = XtVaCreateManagedWidget ("back", xmPushButtonWidgetClass, 
					Buttons, 
					XmNleftAttachment, XmATTACH_WIDGET,
					XmNleftWidget,     Buttons,
					XmNlabelString,    Label,
					NULL);
  XtAddCallback (Data->Back, XmNactivateCallback, 
		             (XtCallbackProc)HELP_BackCB, Data);
  XmStringFree (Label);
  XtSetSensitive (Data->Back, FALSE);

  /* FOLLOW-button: */
  n = 0;
  XtSetArg (args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg (args[n], XmNleftWidget,     Data->Back);      n++;

  if (Data->Param.Click)
  {
    Label = XmStringLtoRCreate ("FOLLOW", XmSTRING_DEFAULT_CHARSET);
    XtSetArg (args[n], XmNlabelString, Label); n++;
    Data->Follow = XtCreateManagedWidget ("follow", xmPushButtonWidgetClass,
				                    Buttons, args, n);
    XtAddCallback (Data->Follow, XmNactivateCallback, 
			                  (XtCallbackProc)HELP_FollowCB, Data);
    XmStringFree (Label);
    XtSetSensitive (Data->Follow, FALSE);

    XtSetArg (args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (args[n], XmNleftWidget,     Data->Follow);    n++;
  }

#if SEARCHING
  /* SEARCH-button: */ 
  PRO0 ("\\Searching...")
  Label =  XmStringLtoRCreate ("SEARCH", XmSTRING_DEFAULT_CHARSET);
  XtSetArg (args[n], XmNleftOffset,  10);    n++;
  XtSetArg (args[n], XmNlabelString, Label); n++;
  Data->Find = XtCreateManagedWidget ("search", xmPushButtonWidgetClass,
				                Buttons, args, n);
  XtAddCallback (Data->Find, XmNactivateCallback, 
		             (XtCallbackProc)HELP_DoSearchCB, Data);
  XmStringFree (Label);
#endif

  /* CLOSE-button: */
  Label = XmStringLtoRCreate ("CLOSE", XmSTRING_DEFAULT_CHARSET);
  Data->Close = XtVaCreateManagedWidget ("close", xmPushButtonWidgetClass,
					 Buttons,
					 XmNlabelString,     Label,
					 XmNrightAttachment, XmATTACH_FORM,
					 XmNrightOffset,     5,
					 NULL);
  XtAddCallback (Data->Close, XmNactivateCallback, 
		              (XtCallbackProc)HELP_ExitCB, Data);
  XmStringFree (Label);

  Line = XtVaCreateManagedWidget ("separator", xmSeparatorWidgetClass, Form,
				  XmNorientation,      XmHORIZONTAL,
				  XmNseparatorType,    XmSHADOW_ETCHED_IN,
				  XmNleftAttachment,   XmATTACH_FORM,
				  XmNrightAttachment,  XmATTACH_FORM,
				  XmNbottomAttachment, XmATTACH_WIDGET,
				  XmNbottomWidget,     Buttons,
				  XmNbottomOffset,     5,
				  NULL);

  /* SCROLLED-window: */
  PRO0 ("\\ScrWin...\n")
  Data->ScrWin = 
    XtVaCreateManagedWidget ("scrWin", xmScrolledWindowWidgetClass, Form, 
			     XmNscrollingPolicy,        XmAUTOMATIC,
			     XmNscrollBarDisplayPolicy, XmAS_NEEDED,
			     XmNvisualPolicy,           XmCONSTANT,
			     XmNtopAttachment,          XmATTACH_FORM,
			     XmNleftAttachment,         XmATTACH_FORM,
			     XmNrightAttachment,        XmATTACH_FORM,
			     XmNbottomAttachment,       XmATTACH_WIDGET,
			     XmNbottomWidget,           Line,
			     XmNbottomOffset,           3,
			     NULL);
  XtVaGetValues (Data->ScrWin, XmNverticalScrollBar,   &Data->ScrBarY,
		               XmNhorizontalScrollBar, &Data->ScrBarX,
 		               XmNwidth,               &width,
		               XmNclipWindow,          &Data->ClipWin,
		               NULL);

#if USE_LABEL
  Data->Work = XtVaCreateManagedWidget ("textArea", xmLabelWidgetClass,
					Data->ScrWin,
					XmNmarginLeft,  0,
					XmNmarginWidth, 0,
					XmNalignment,   XmALIGNMENT_BEGINNING,
					NULL);

  XtAddEventHandler (Data->Work, KeyPressMask    | PointerMotionMask |
		                 EnterWindowMask | LeaveWindowMask   |
		                 ButtonPressMask | ButtonReleaseMask
		                 /* | StructureNotifyMask */, FALSE,
		                 (XtEventHandler)HELP_HandleEvents, Data);
#else
  Data->Work = XtCreateManagedWidget ("textArea", xmDrawingAreaWidgetClass,
					          Data->ScrWin, NULL, 0);
  XtAddEventHandler (Data->Work, KeyPressMask    | PointerMotionMask |
		                 EnterWindowMask | LeaveWindowMask   |
		                 ButtonPressMask | ButtonReleaseMask |
		                 ExposureMask    | StructureNotifyMask, FALSE,
		                (XtEventHandler)HELP_HandleEvents, Data);
#endif

  XtVaGetValues (Data->Work, XmNbackground, &Colors[HELP_BG], NULL);
  /*   Data->hl      = Colors[HELP_BG] ^ Colors[HELP_RED]; */
  Data->bg      = 0;
  Data->bgImage = NULL;
  Data->bgX     = 0;
  Data->bgY     = 0;

#if 0
#endif

  if (Text != NULL && ! HELP_ParseText(Data, Text))
  {
    free (Text);
    TOOLSetMouseCursor (Data->Shell, ToolNORMAL, NULL, NULL);
    PRO_RETURN ("Parse ERROR", FALSE)
  }

  if (Data->bg > 0 && Data->bg < ColorNo)
    Colors[HELP_BG] = Colors[Data->bg];
  Data->hl      = Colors[HELP_BG] ^ Colors[HELP_RED];

#if SEARCHING
  HELP_DataSearch (Data, Pattern, DEF_METHOD);
#endif

  HELP_CalcLines (Data, width, NULL);

  /* Get the window of working area, needed for drawing */
  Data->Win = XtWindow (Data->Work);

  XtAddEventHandler (Data->ClipWin, StructureNotifyMask, False,
		                    (XtEventHandler)HELP_HandleEvents, Data);

  XtVaSetValues (Data->ScrWin, XmNworkWindow, Data->Work, NULL);

  GCVal.graphics_exposures = False;
  Data->DrawGC = XCreateGC (XDisplay, Data->Win, GCGraphicsExposures, &GCVal);

#if USE_PIXMAP
  HELP_RefreshText (Data, 0, Data->SizeY);
#if USE_LABEL
  XtVaSetValues (Data->Work, XmNlabelType, XmPIXMAP,
		             XmNlabelPixmap, Data->Pic,
		             NULL);
#endif
#endif

  /* Set window-name and icon */
  XStoreName (XDisplay, XtWindow(Data->Shell), Data->Title);
  TOOLSetIcon (Data->Shell, Data->Title, NULL, helpicon_xpm); 

  XtRealizeWidget (Data->Shell);
  XtPopup (Data->Shell, 
	   (Data->HelpCB != NULL)? XtGrabNonexclusive : XtGrabExclusive);

  if (Target.Label[0] != '\0')
  {
    PRO1 ("JUMP to label: %s\n", Target.Label)
    HELP_GetSize (Data);
    HELP_UpdateWindow (Data, Target.Label, 0, FALSE);
  }
#if SEARCHING
  elsif (Data->Search.Hits > 0)
    HELP_LocateHit (Data, 1);
#endif

  TOOLSetMouseCursor (Data->Shell, ToolNORMAL, NULL, NULL);

  if (Text == NULL)
  {
    HELP_OpenDlg (Data->Shell, "fileError", "Error opening file", NULL);
    HELP_ExitCB (NULL, Data, NULL);
  }
  else
    free (Text);

  PRO_END
  return Data;
}


/* ********************************************************************** */
/* ** function for creating a help-window. The textspec could contain  ** */
/* ** a label and/or a leading 'file:'.                                ** */
/* ** INPUT:  Title     Name of the window (also used in the ad-file)  ** */
/* **         TextSpec  filename or help-text                          ** */
/* **         SizeX                                                    ** */
/* **         SizeY     size of the window at the beginning            ** */
/* **         Pattern   Search-pattern                                 ** */
/* **         Method    Search method (see above)                      ** */
/* ** OUTPUT: ---                                                      ** */
/* ** EXAMPLE:                                                         ** */
/* **   HELPOpenHelpWindow ("helpWin",                                 ** */
/* **                      "/usr/data/help/doc.hlp#toc",               ** */
/* **                      400, 400,                                   ** */
/* **                      "search me", HELP_CASE | HELP_SUBSTR);      ** */
/* ********************************************************************** */
HELPWIN *HELPOpenHelpWin (const char    *Title,   const char   *TextSpec, 
			        double   PosX,          double  PosY,
			        double   SizeX,         double  SizeY,
			        char    *Pattern,       byte    Method,
			        HELPCB   HelpCB)
{
  int      x, y, w, h;
  Widget   Shell;
  Arg      args[10];
  HELPWIN *Result;

  PRO_START ("\n")

  /* Set window's position and size */
  TOOLGetScrValue (XDisplay, XScreen, PosX,  PosY,  &x, &y);
  TOOLGetScrValue (XDisplay, XScreen, SizeX, SizeY, &w, &h);

  XtSetArg (args[0], XmNx,      x);
  XtSetArg (args[1], XmNy,      y);
  XtSetArg (args[2], XmNwidth,  w);
  XtSetArg (args[3], XmNheight, h);

  if (TOOLIsRunning())             /* is there a main-program running ? */
  {
    PRO0 ("Window...\n")
    XtSetArg (args[4], XmNmwmFunctions,        MWM_FUNC_ALL);
    XtSetArg (args[5], XmNmwmDecorations,      MWM_DECOR_ALL);
    XtSetArg (args[6], XmNdeleteResponse,      XmDESTROY);
    XtSetArg (args[7], XmNkeyboardFocusPolicy, XmPOINTER);
    Shell = XtCreatePopupShell (Title, transientShellWidgetClass,
				       AppShell, args, 8);
    /* 
       TODO:
       durch das frhe Managen erhlt das Fenster die Gre 0. Nur durch die
       Grenangabe wird das Fenster grer dargestellt -> spter managen !
    */
    XtManageChild (Shell);
  }
  else
  {
    PRO0 ("Shell...\n")
    Shell = AppShell;

    XtSetValues (Shell, args, 4);
    XtPopup (Shell, XtGrabNonexclusive);
  }

  Result = HELP_CreateHelpWin (Shell, Title, TextSpec, 
			              Pattern, Method, HelpCB);

  PRO_END
  return Result;
}


dual HELPGetSize (HELPWIN *HelpWin, int *SizeX, int *SizeY)
{
  Dimension w, h;

  if (HelpWin == NULL)
    return FALSE;

  XtVaGetValues (HelpWin->Shell, XmNwidth,  &w,
		                 XmNheight, &h, 
		                 NULL);

  if (SizeX != NULL)
    *SizeX = w;
  if (SizeY != NULL)
    *SizeY = h;

  return TRUE;
}


dual HELPGetPos (HELPWIN *HelpWin, int *PosX, int *PosY)
{
  Dimension x, y;

  if (HelpWin == NULL)
    return FALSE;

  XtVaGetValues (HelpWin->Shell, XmNx, &x, 
		                 XmNy, &y, 
		                 NULL);

  if (PosX != NULL)
    *PosX = x;
  if (PosY != NULL)
    *PosY = y;

  return TRUE;
}


void HELPLoad (HELPWIN *HelpWin, 
	       const char *TextSpec, char *Pattern, byte Method)
{
  char *HashPtr;
  char  FileName[PATH_MAX+1];
  char  Label[LABELLEN+1] = "";

  strncpy (FileName, TextSpec, PATH_MAX);
  if ((HashPtr = strstr(FileName, "#")) != NULL)
  {
    strncpy (Label, HashPtr+1, LABELLEN);
    Label[LABELLEN] = '\0';
    *HashPtr = '\0';
  }

  PRO2 ("Loading %s # %s\n", FileName, Label)

  HELP_Jump (HelpWin, FileName, Label);
}


/* ********************************************************************** */
/* ** Function saves the actual helptext-position, loads a new help    ** */
/* ** text and/or jumps to the actual label                            ** */
/* ** INPUT:  Data      Pointer to complete help-structure             ** */
/* **         FileName  New filename                                   ** */
/* **         Label     Name of the selected label                     ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
void HELP_Jump (HELPDATA *Data, const char *FileName, char *Label)
{
  dual  NewFile = FALSE;
  int   OldPosY;
  char  OldFile[PATH_MAX+1];
  char *Text = NULL;

  PRO0 (" ")

  XtVaGetValues (Data->ScrBarY, XmNvalue, &OldPosY, NULL);

  strcpy (OldFile, Data->FileName);

  NewFile = (FileName[0] != '\0' && strcmp(FileName, Data->FileName) != 0);
  if (NewFile)
  {
    PRO1 ("NEW FILE: %s\n", FileName)
    
    if ((Text = HELP_LoadText(Data, FileName)) == NULL)
      return;

    Data->bg = 0;
    if (! HELP_ParseText(Data, Text))
    {
      free (Text);
      TOOLSetMouseCursor (Data->Shell, ToolNORMAL, NULL, NULL);
      return;
    }
    free (Text);

    HELP_CalcLines (Data, Data->SizeX, NULL);
    XStoreName (XDisplay, XtWindow(Data->Shell), Data->Title);

#if SEARCHING
    HELP_DataSearch (Data, Data->Search.Pattern, Data->Search.Method);
#endif
  }
  else
    HELP_MarkHRef (Data, Data->Selection);

  PRO2 ("save: %s at Pos %d\n", OldFile, OldPosY)
  strcpy (Data->Memory[Data->MemID].FileName, OldFile);
  Data->Memory[Data->MemID].Scroll = OldPosY;

  HELP_UpdateWindow (Data, Label, 0, NewFile);

  Data->Selection = -1;

  if (Data->MemID == 0)
  {
    if (Data->Param.Click)
      XtSetSensitive (Data->Follow, FALSE);
    XtSetSensitive (Data->Back, TRUE);
  }
  (Data->MemID)++;

  if (NewFile)
    TOOLSetMouseCursor (Data->Shell, ToolNORMAL, NULL, NULL);
}


/* ********************************************************************** */
/* ** Callback-routine which is called to follow a refrence            ** */
/* ** INPUT:  widget  calling widget (here follow-button or ScrWin)    ** */
/* **         Data    help-structure                                   ** */
/* **         Ptr     Callback-structure (not used here)               ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
static void HELP_FollowCB (Widget widget, HELPDATA *Data, 
				         XmAnyCallbackStruct *Ptr)
{
  char Label[LABELLEN+1];

  PRO0 ("FOLLOW\n")

  if (widget == NULL)
    return;

  if (Data->Selection >= 0 && Data->MemID < MEM_SIZE)
  {
    strcpy (Label, Data->HRef[Data->Selection].Label);
    HELP_Jump (Data, Data->HRef[Data->Selection].FileName, Label);
  }
}


/* ********************************************************************** */
/* ** Callback for moving back to the last position                    ** */
/* ** INPUT:  widget  calling widget                                   ** */
/* **         Data    help-structure                                   ** */
/* **         Ptr     callback-data (not used)                         ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
static void HELP_BackCB (Widget widget, HELPDATA *Data, 
			               XmAnyCallbackStruct *Ptr)
{
  dual  NewFile = FALSE;
  byte  MemID;
  char *Text = NULL;

  PRO0 ("BACK\n")

  if (widget == NULL)
    return;

  if (Data->MemID > 0)
  {
    MemID = --(Data->MemID);

    PRO2 ("jump back: %s at Pos %d\n", 
	  Data->Memory[MemID].FileName, Data->Memory[MemID].Scroll)

    NewFile = (Data->Memory[MemID].FileName[0] != '\0' &&
	       strcmp(Data->Memory[MemID].FileName, Data->FileName) != 0);
    if (NewFile)         
    {
      PRO1 ("NEW FILE: %s\n", Data->Memory[MemID].FileName)
      if ((Text = HELP_LoadText(Data, Data->Memory[MemID].FileName)) == NULL)
	return;
      if (! HELP_ParseText(Data, Text))
      {
	free (Text);
	TOOLSetMouseCursor (Data->Shell, ToolNORMAL, NULL, NULL);
	return;
      }
      free (Text);

#if SEARCHING
      HELP_DataSearch (Data, Data->Search.Pattern, Data->Search.Method);
#endif
      HELP_CalcLines (Data, Data->SizeX, NULL);
      XStoreName (XDisplay, XtWindow(Data->Shell), Data->Title);
    }

    HELP_UpdateWindow (Data, NULL, Data->Memory[MemID].Scroll, NewFile);

    if (Data->Param.Click)
      XtSetSensitive (Data->Follow, FALSE);
    if (MemID == 0)
      XtSetSensitive (Data->Back, FALSE);

    if (NewFile)
      TOOLSetMouseCursor (Data->Shell, ToolNORMAL, NULL, NULL);
  }
}


/* ********************************************************************** */
/* ** Callback, which handles a submit-command called by an input-     ** */
/* ** form using the get-method.                                       ** */
/* ** INPUT:  widget  calling widget                                   ** */
/* **         Data    help-text-data                                   ** */
/* **         Ptr     Pointer to callback-data                         ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
static void HELP_SubmitCB (Widget widget, HELPDATA *Data,
			                 XmAnyCallbackStruct *Ptr)
{
  dual        first = TRUE;
  Boolean     Button;
  int         ID = -1;
  char       *Text;
  char        Arguments[1024] = "";
  INPUT_DATA *Input;
  HELP_ELEM  *Elem;

  Elem = Data->first;
  while (Elem != NULL && ID == -1)
  {
    if (Elem->Type == ELEM_INPUT)
    {
      Input = (INPUT_DATA*)Elem->Data;
      if (Input->Type == SUBMIT && Input->widget == widget)
	ID = Elem->ID;
    }
    Elem = Elem->Next;
  }

  if (ID == -1)
    return;

  Elem = Data->first;
  while (Elem != NULL)
  {
    if (Elem->Type == ELEM_INPUT && Elem->ID == ID)
    {
      Input = (INPUT_DATA*)Elem->Data;
      switch (Input->Type)
      {
        case TEXTVALUE :
	  XtVaGetValues (Input->widget, XmNvalue, &Text, NULL);
          strcat (Arguments, first?"?":"&");
	  strcat (Arguments, Input->Name);
	  strcat (Arguments, "=");
	  while (*Text != '\0')
          {
	    if (*Text == ' ')
	      strcat (Arguments, "+");
	    else
	      strncat (Arguments, Text, 1);
	    Text++;
	  }
	  first = FALSE;
	  break;

        case CHECKBOX :
	  XtVaGetValues (Input->widget, XmNset, &Button, NULL);
	  if (Button)
	  {
	    strcat (Arguments, first?"?":"&");
	    strcat (Arguments, Input->Name);
	    strcat (Arguments, "=");
	    strcat (Arguments, Input->Value);
	    first = FALSE;
	  }
	  break;

        case SUBMIT :
	  break;

        default :
	  PRO1 ("Invalid INPUT-Type: %d\n", Input->Type)
      }
    }
    Elem = Elem->Next;
  }
  printf ("Arguments: '%s'\n", Arguments);
  printf ("URI: '%s'\n", Data->URI[ID]);

  /* system ("Data->URI[ID] Arguments"); */
}


/* ************************************************************ */
/* ** Callback for closing the help-window manages the       ** */
/* ** following jobs:                                        ** */
/* ** - frees the data allocated in helptext-structure       ** */
/* ** - pop down and destroy the shell widget                ** */
/* ** - removing the popdown-callback from the application   ** */
/* **   shell (used to free data if the application will be  ** */
/* **   closed while a help-window is still open)            ** */
/* ** If this routine is called from the close-button, it    ** */
/* ** will only do a pop down and a deletion first.          ** */
/* ** Thereby a destroy callback will be generated which     ** */
/* ** calls this function a second time to free the memory.  ** */
/* ** INPUT:  widget  close- or WM-destroy-button            ** */
/* **         Data    help-data                              ** */
/* **         Ptr     callback data (reason)                 ** */
/* ** OUTPUT: none                                           ** */
/* ************************************************************ */
static void HELP_ExitCB (Widget widget, HELPDATA *Data, 
			                XmAnyCallbackStruct *Ptr)
{
  PRO_START ("\n")

  if (Data == NULL || ! HELP_TestData(Data))
    PRO_ABORT ("No data")

  if (Data->HelpCB != NULL)
  {
    Data->HelpCB (Data);
    /* 
       Do not access data any more, because the data might be deleted in 
       the callback routine 
    */
  }
  else
    HELPClose (Data);

  PRO_END
}


void HELPClose (HELPWIN *HelpWin)
{
  Widget Shell = NULL;

  PRO_START ("\n")

  if (HelpWin == NULL)
    return;

  Shell = (Widget)(HelpWin->Shell);
  HELP_DelData (HelpWin);
  HELP_FreeHelpData (HelpWin, FALSE);
  XtDestroyWidget (Shell);

  PRO_END
}
