#ifdef sunos
#include <sys/types.h>
#include <sys/dir.h>
#define dirent direct
#else
#include <dirent.h>
#endif

#include <sys/stat.h>
#include <ctype.h>
#include <unistd.h>

#include "helpGlobal.h"

#include <Xm/Form.h>
#include <Xm/Separator.h>
#include <Xm/Label.h>
#include <Xm/List.h>


#if SEARCHING
  static void  HELP_SearchCB (Widget widget, HELPDATA *Data,
	  	                             XmAnyCallbackStruct *Ptr);
  static void  HELP_ClearCB  (Widget widget, HELPDATA *Data,
		                             XmAnyCallbackStruct *Ptr);
  static void  HELP_CloseCB  (Widget widget, HELPDATA *Data,
			                     XmAnyCallbackStruct *Ptr);
  static void  HELP_HitCB    (Widget widget, HELPDATA *Data,
			                     XmListCallbackStruct *Ptr);


static int HELP_Select (const struct dirent *Entry)
{
  return (strcmp(Entry->d_name, ".") != 0 && 
	  strcmp(Entry->d_name, "..") != 0);
}


static HIT_DATA *HELP_FileLocate (char *FileName, char *Pattern, int Method)
{
  int       count;
  char      letter;
  char      Title[64] = "";
  char      Word[64];
  FILE     *fp = NULL;
  HIT_DATA *Result = NULL;

  if ((fp = fopen(FileName, "r")) == NULL)
    return NULL;

  while (! feof(fp) && Result == NULL)
  {
    letter = fgetc (fp);
    if (letter == '<')
    {
      fscanf (fp, "%5s", Word);
      if (strncasecmp(Word, "TITLE", 5) == 0)
      {
	while (! feof(fp) && fgetc(fp) != '>')
	  ;
	count = 0;
	while (count < 63 && ! feof(fp) && (letter = fgetc(fp)) != '<')
	  Title[count++] = letter;
	Title[count] = '\0';
      }

      while (! feof(fp) && fgetc(fp) != '>')
	;
    }
    elsif (letter == Pattern[0] ||
	   ((Method & HELP_CASE) && toupper(letter) == toupper(Pattern[0])))
    {
      fscanf (fp, "%s", Word);
      if (strncmp(Word, Pattern+1, strlen(Pattern)-1) == 0 ||
	  (Method & HELP_CASE &&
	   strncasecmp(Word, Pattern+1, strlen(Pattern)-1) == 0))
	Result = (HIT_DATA*)malloc (sizeof(HIT_DATA));
    }
  }
  fclose (fp);

  if (Result != NULL)
  {
    if (Title[0] == '\0')               /* hit, but no title found        */
      TOOLGetFileName (FileName, Result->Title, 64);
    else
      strcpy (Result->Title, Title);

    strncpy (Result->FileName, FileName, PATH_MAX);
    Result->FileName[PATH_MAX] = '\0';

    Result->Next = NULL;
  }

  return Result;
}


static dual HELP_MatchMask (char *FileName, char *Mask)
{
  int len1, len2;

  len1 = strlen (FileName);
  len2 = strlen (Mask);

  return (strcmp(FileName+len1-len2+1, Mask+1) == 0);
}
    

static HIT_DATA *HELP_FileSearch (char *Directory, char *Mask, char *Pattern, 
				 HIT_DATA **Result, int *Amount)
{
         int        count, Number;
         char       FileName[PATH_MAX+1];
	 HIT_DATA  *Hit = NULL;
  struct stat       Buffer;
  struct dirent   **Entry = NULL;

  if ((Number = scandir(Directory, &Entry, HELP_Select, alphasort)) > 0)
  {
    for (count = 0; count < Number; ++count)
    {
      if (strcmp(Entry[count]->d_name, ".") != 0 && 
	  strcmp(Entry[count]->d_name, "..") != 0)
      {
	sprintf (FileName, "%s/%s", Directory, Entry[count]->d_name); 
	PRO1 ("Searching: %s\n", FileName)

	/* call recursive with subdirs... */
	if (stat(FileName, &Buffer) == 0 && 
	    (Buffer.st_mode & S_IFMT) == S_IFDIR)
	  HELP_FileSearch (FileName, Mask, Pattern, Result, Amount);
	elsif (HELP_MatchMask(FileName, Mask) &&
	       (Hit = HELP_FileLocate(FileName, Pattern, 0)) != NULL)
	{
	  if (*Result == NULL)
	    *Result = Hit;
	  else
	  {
	    Hit->Next = *Result;
	    *Result = Hit;
	  }
	  (*Amount)++;
	}
      }
      free (Entry[count]);
    }
    free (Entry);
  }
  return *Result;
}


/* ************************************************************ */
/* ** searches a given pattern in the data's text-elems and  ** */
/* ** returns the amount of hits.                            ** */
/* ** INPUT:   Data     help-structure                       ** */
/* **          Pattern  search-pattern                       ** */
/* ** OUTPUT:  (int)    number of found hits                 ** */
/* ************************************************************ */
int HELP_DataSearch (HELPDATA *Data, char *Pattern, int Method)
{
  char      *comp1, *comp2;
  HELP_ELEM *Elem;

  PRO_START ("")

  if (Data->Search.updated)
    PRO_RETURN ("search is up to date", 0)

  if (Data->Search.Pattern != NULL &&
      strcmp(Data->Search.Pattern, Pattern) != 0)
    free (Data->Search.Pattern);
  Data->Search.Method = Method;
  Data->Search.Hits   = 0;
  Data->Search.Index  = 0;

  Elem = Data->first;

  PRO2 ("\\method: %d: '%s'...", Method, Pattern)

  if (Pattern == NULL || Pattern[0] == '\0')
  {
    Data->Search.Pattern = NULL;
    while (Elem != NULL)
    {
      if (Elem->Type == ELEM_TEXT)
	Elem->fits = FALSE;
      Elem = Elem->Next;
    }
  }
  else
  {
    if (Data->Search.Pattern == NULL ||
	strcmp(Data->Search.Pattern, Pattern) != 0)
    {
      Data->Search.Pattern = (char*)malloc (sizeof(char)*(strlen(Pattern)+1));
      strcpy (Data->Search.Pattern, Pattern);
    }

    if (! (Method & HELP_CASE))
    {
      comp2 = (char*)malloc (sizeof(char) * (strlen(Pattern)+1));
      TOOLstrtoupper (comp2, Pattern);
    }
    else
      comp2 = Pattern;

    while (Elem != NULL)
    {
      if (Elem->Type == ELEM_TEXT)
      {
	if (! (Method & HELP_CASE))
	{
	  comp1 = (char*)malloc (sizeof(char) *
				 (strlen((char*)Elem->Data)+1));
	  TOOLstrtoupper (comp1, (char*)Elem->Data);
	}
	else
	  comp1 = (char*)Elem->Data;

	if (Method & HELP_SUBSTR)
	  Elem->fits = (strstr(comp1, comp2) != NULL);
	else
	  Elem->fits = (strcmp(comp1, comp2) == 0);

	if (! (Method & HELP_CASE))
	  free (comp1);

	if (Elem->fits)
	  (Data->Search.Hits)++;
      }
      Elem = Elem->Next;  
    }
    if (! (Method & HELP_CASE))
      free (comp2);
  }
  PRO1 ("\\%d hits found...", Data->Search.Hits)
  Data->Search.updated = TRUE;

  PRO_END
  return Data->Search.Hits;
}


/* ************************************************************ */
/* ** look for the n-th hit from last search and display the ** */
/* ** appropriate line at top of window                      ** */
/* ** INPUT:  Data    helptext-data                          ** */
/* **         HitNo   No of hit, which should be displayed   ** */
/* ** OUTPUT: (dual)  TRUE, if selected hit exists           ** */
/* ************************************************************ */
dual HELP_LocateHit (HELPDATA *Data, int HitNo)
{
  int        count = 0;
  int        line = 0, Index;
  HELP_ELEM *Elem;

  PRO_START ("")

  if (HitNo < 1 || HitNo > Data->Search.Hits)
    PRO_RETURN ("HitNo out of range", FALSE)

  while (line < Data->Lines && count < HitNo)
  {
    Elem = Data->Line[line].List;
    Index = 0;
    while (Index < Data->Line[line].Nr && count < HitNo)
    {
      if (Elem->fits)
	count++;
      Elem = Elem->Next;
      Index++;
    }
    line++;
  }

  if (line < Data->Lines)
  {
    HELP_UpdateWindow (Data, NULL, Data->Line[line-1].PosY, FALSE);
    Data->Search.Index = HitNo;
  }

  PRO_END
  return TRUE;
}


/* ************************************************************** */
/* Search-window contains elements for starting a text-search. ** */
/* INPUT:  widget  associated widget                           ** */
/*         Data    Pointer to help-data                        ** */
/*         Ptr     callback-data (not used yet)                ** */
/* OUTPUT: none                                                ** */
/* ************************************************************** */
void HELP_DoSearchCB (Widget widget, HELPDATA *Data, XmAnyCallbackStruct *Ptr)
{
  int            count = 0, n;
  Arg            args[15];
  XmString       Label;
  XmStringTable  List;
  Widget         Form, Line1, Line2;
  Widget         Descr, Box, Button;
  HIT_DATA      *Hit;

  if (widget == NULL || Data->Search.Shell != NULL)
    return;

  XtSetSensitive (Data->Find, FALSE); 
  XtSetSensitive (Data->Close, FALSE); 

  Data->Search.Shell = 
    XtCreatePopupShell ("helpSearch", topLevelShellWidgetClass, 
			              Data->Shell, args, 0);

  Form = XtVaCreateManagedWidget ("form", xmFormWidgetClass, 
				          Data->Search.Shell,
				          XmNtopAttachment,    XmATTACH_FORM,
	 			          XmNleftAttachment,   XmATTACH_FORM,
				          XmNrightAttachment,  XmATTACH_FORM,
				          XmNbottomAttachment, XmATTACH_FORM,
				          NULL);

  /* Buttons: */
  Label = XmStringLtoRCreate ("SEARCH", XmSTRING_DEFAULT_CHARSET);
  Button = XtVaCreateManagedWidget ("start", xmPushButtonWidgetClass, Form, 
				    XmNbottomAttachment, XmATTACH_FORM,
				    XmNbottomOffset,     5,
				    XmNleftAttachment,   XmATTACH_FORM,
				    XmNleftOffset,       5,
				    XmNlabelString,      Label,
				    NULL);
  XtAddCallback (Button, XmNactivateCallback, 
		         (XtCallbackProc)HELP_SearchCB, Data);
  XmStringFree (Label);

  Label = XmStringLtoRCreate ("CLEAR", XmSTRING_DEFAULT_CHARSET);
  Button = XtVaCreateManagedWidget ("clear", xmPushButtonWidgetClass, Form, 
				    XmNbottomAttachment, XmATTACH_FORM,
				    XmNbottomOffset,     5,
				    XmNlabelString,      Label,
				    XmNleftAttachment,   XmATTACH_WIDGET,
				    XmNleftWidget,       Button,
				    XmNleftOffset,       10,
				    NULL);
  XtAddCallback (Button, XmNactivateCallback, 
	                 (XtCallbackProc)HELP_ClearCB, Data);
  XmStringFree (Label);

  Label = XmStringLtoRCreate ("CLOSE", XmSTRING_DEFAULT_CHARSET);
  Button = XtVaCreateManagedWidget ("close", xmPushButtonWidgetClass, Form,
				    XmNbottomAttachment, XmATTACH_FORM,
				    XmNbottomOffset,     5,
				    XmNlabelString,      Label,
				    XmNrightAttachment,  XmATTACH_FORM,
				    XmNrightOffset,      5,
				    NULL);
  XtAddCallback (Button, XmNactivateCallback, 
	   	         (XtCallbackProc)HELP_CloseCB, Data);
  XmStringFree (Label);

  /* Separator */
  Line1 = XtVaCreateManagedWidget ("separator", xmSeparatorWidgetClass, Form,
				   XmNorientation,      XmHORIZONTAL,
				   XmNseparatorType,    XmSHADOW_ETCHED_IN,
				   XmNleftAttachment,   XmATTACH_FORM,
				   XmNrightAttachment,  XmATTACH_FORM,
				   XmNbottomAttachment, XmATTACH_WIDGET,
				   XmNbottomWidget,     Button,
				   XmNbottomOffset,     3,
				   NULL);

  /* Elements */
  Form = XtVaCreateManagedWidget ("elems", xmFormWidgetClass, Form,
				  XmNtopAttachment   , XmATTACH_FORM,
				  XmNtopOffset,        5,
				  XmNleftAttachment,   XmATTACH_FORM,
				  XmNleftOffset,       5,
				  XmNrightAttachment,  XmATTACH_FORM,
				  XmNrightOffset,      5,
				  XmNbottomAttachment, XmATTACH_WIDGET,
				  XmNbottomWidget,     Line1,
				  XmNbottomOffset,     5,
				  NULL);

  Box = XtVaCreateManagedWidget ("elems", xmFormWidgetClass, Form,
				 XmNtopAttachment   , XmATTACH_FORM,
				 XmNleftAttachment,   XmATTACH_FORM,
				 XmNrightAttachment,  XmATTACH_FORM,
				 NULL);

  Descr = XtVaCreateManagedWidget ("pattern", xmLabelWidgetClass, Box,
				   XmNtopAttachment,    XmATTACH_FORM,
				   XmNbottomAttachment, XmATTACH_FORM,
				   NULL);

  XtSetArg (args[0], XmNleftAttachment,  XmATTACH_WIDGET);
  XtSetArg (args[1], XmNleftWidget,      Descr);
  XtSetArg (args[2], XmNrightAttachment, XmATTACH_FORM);
  if (Data->Search.Pattern != NULL)
    XtSetArg (args[3], XmNvalue, Data->Search.Pattern);
  else
    XtSetArg (args[3], XmNvalue, "");
  Data->Search.Input = XtCreateManagedWidget ("patternVal", xmTextWidgetClass,
					                    Box, args, 4);

  Label =  XmStringLtoRCreate ("CASE", XmSTRING_DEFAULT_CHARSET);
  Data->Search.Option[0] = 
    XtVaCreateManagedWidget ("case", xmToggleButtonWidgetClass, Form,
			     XmNlabelString,    Label,
			     XmNtopAttachment, XmATTACH_WIDGET,
			     XmNtopWidget,     Data->Search.Input,
			     XmNset, 
  			      ((Data->Search.Method & HELP_CASE) == HELP_CASE),
			     NULL);

  Label =  XmStringLtoRCreate ("SUBSTR", XmSTRING_DEFAULT_CHARSET);
  Data->Search.Option[1] = 
    XtVaCreateManagedWidget ("substr", xmToggleButtonWidgetClass, Form,
			     XmNlabelString,    Label,
			     XmNtopAttachment,  XmATTACH_WIDGET,
			     XmNtopWidget,      Data->Search.Input,
			     XmNleftAttachment, XmATTACH_WIDGET,
			     XmNleftWidget,     Data->Search.Option[0],
			     XmNset,
			     ((Data->Search.Method & HELP_SUBSTR)==HELP_SUBSTR),
			     NULL);

  Label = XmStringLtoRCreate ("GLOBAL", XmSTRING_DEFAULT_CHARSET);
  Data->Search.Option[2] =
    XtVaCreateManagedWidget ("global", xmToggleButtonWidgetClass, Form,
			     XmNlabelString,    Label,
			     XmNtopAttachment,  XmATTACH_WIDGET,
			     XmNtopWidget,      Data->Search.Input,
			     XmNleftAttachment, XmATTACH_WIDGET,
			     XmNleftWidget,     Data->Search.Option[1],
			     XmNset, 
			     ((Data->Search.Method & HELP_GLOBAL)==HELP_GLOBAL),
			     NULL);

  Line2 = XtVaCreateManagedWidget ("separator", xmSeparatorWidgetClass, Form, 
				   XmNorientation,     XmHORIZONTAL,
				   XmNseparatorType,   XmSHADOW_ETCHED_IN,
				   XmNtopAttachment,   XmATTACH_WIDGET,
				   XmNtopWidget,       Data->Search.Option[2],
				   XmNtopOffset,       5,
				   XmNleftAttachment,  XmATTACH_FORM,
				   XmNrightAttachment, XmATTACH_FORM,
				   NULL);

  Descr = XtVaCreateManagedWidget ("hits", xmLabelWidgetClass, Form,
				   XmNtopAttachment,  XmATTACH_WIDGET,
				   XmNtopWidget,      Line2,
				   XmNtopOffset,      5,
				   XmNleftAttachment, XmATTACH_FORM,
				   XmNleftOffset,     10,
				   NULL);

  List = (XmStringTable)malloc (sizeof(XmString) * Data->Search.Files);
  Hit = Data->Search.List;
  while (Hit != NULL)
  {
    List[count++] = XmStringCreate (Hit->Title, XmSTRING_DEFAULT_CHARSET);
    Hit = Hit->Next;
  }
  n = 0;
  XtSetArg (args[n], XmNtopAttachment,          XmATTACH_WIDGET); n++;
  XtSetArg (args[n], XmNtopWidget,              Descr); n++;
  XtSetArg (args[n], XmNtopOffset,              5); n++;
  XtSetArg (args[n], XmNleftAttachment,         XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNrightAttachment,        XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNbottomAttachment,       XmATTACH_WIDGET); n++;
  XtSetArg (args[n], XmNbottomWidget,           Line1); n++;
  XtSetArg (args[n], XmNlistSizePolicy,         XmVARIABLE); n++;
  XtSetArg (args[n], XmNscrollBarDisplayPolicy, XmAS_NEEDED); n++;
  XtSetArg (args[n], XmNselectionPolicy,        XmSINGLE_SELECT); n++;
  XtSetArg (args[n], XmNvisibleItemCount,       5); n++;
  if (Data->Search.Files > 0)
  {
    XtSetArg (args[n], XmNitemCount,              Data->Search.Files); n++;
    XtSetArg (args[n], XmNitems,                  List); n++;
  }
  Data->Search.ListW = XmCreateScrolledList (Form, "hitList", args, n);
  XtManageChild (Data->Search.ListW);
  XtAddCallback (Data->Search.ListW, XmNsingleSelectionCallback, 
		                           (XtCallbackProc)HELP_HitCB, Data);

  for (count = 0; count < Data->Search.Files; ++count)
    XmStringFree (List[count]);
  free (List);

  XtManageChild (Data->Search.Shell);
  XtRealizeWidget (Data->Search.Shell);
  XtPopup (Data->Search.Shell, XtGrabNonexclusive);

  TOOLMainLoop (AppCon, Data->Search.Shell);
}


/* ********************************************************************** */
/* ** Callback for searching string in Data. If the environment        ** */
/* ** variable 'HELPDIR' is defined, the search will be started there. ** */
/* ** Otherwise the directory of the actual help-text will be used     ** */
/* ** INPUT:  widget  calling widget                                   ** */
/* **         Data    help-structure                                   ** */
/* **         Ptr     callback-data (not used)                         ** */
/* ** OUTPUT: none                                                     ** */
/* ********************************************************************** */
static void HELP_SearchCB (Widget widget, HELPDATA *Data,
		                         XmAnyCallbackStruct *Ptr)
{
  int            hits = 0, count = 0;
  int            Method = 0;
  char           Ext[10], Mask[12];
  char          *Pattern = NULL;
  char          *Directory = NULL;
  Boolean        Option;
  Arg            args[2];
  XmStringTable  List;
  HIT_DATA      *Hits, *NextHit;
#if USE_PIXMAP
  int      PosY;
#endif

  if (widget == NULL)
    TOOLSetMouseCursor (Data->Search.Shell, ToolWATCH, NULL, NULL);

  XtVaGetValues (Data->Search.Input, XmNvalue, &Pattern, NULL);

  if (Data->Search.Pattern == NULL ||
      strcmp(Pattern, Data->Search.Pattern) != 0)
    Data->Search.updated = FALSE;

  XtSetArg (args[0], XmNset, &Option);
  XtGetValues (Data->Search.Option[0], args, 1);
  if (Option)
    Method += HELP_CASE;

  XtSetArg (args[0], XmNset, &Option);
  XtGetValues (Data->Search.Option[1], args, 1);
  if (Option)
    Method += HELP_SUBSTR;

  XtSetArg (args[0], XmNset, &Option);
  XtGetValues (Data->Search.Option[2], args, 1);
  if (Option)
    Method += HELP_GLOBAL;

  if (Method != Data->Search.Method)
    Data->Search.updated = FALSE;

  PRO1 ("Searching '%s'...\n", Pattern)
  PRO3 ("Options:   %c %c %c\n",
	(Method & HELP_CASE)? 'C':'-', (Method & HELP_SUBSTR)? 'S':'-',
	Option? 'R':'-')

  if (! Data->Search.updated)
  {
    if (Option && Pattern[0] != '\0')
    {
      TOOLGetFileExt (Data->FileName, Ext, 10);
      sprintf (Mask, "*.%s", Ext);

      if ((Directory = getenv("HELPDIR")) == NULL)
	Directory = Data->CurrDir;

      PRO1 ("Dir:       %s\n", Directory)
      PRO1 ("Mask:      %s\n", Mask)

      Hits = Data->Search.List;
      while (Hits != NULL)
      {
	NextHit = Hits->Next;
	free (Hits);
	Hits = NextHit;
      }
      Data->Search.List  = NULL;
      Data->Search.Files = 0;

      Data->Search.List = HELP_FileSearch (Data->CurrDir, Mask, Pattern, 
					  &Data->Search.List, 
					  &Data->Search.Files);
      List = (XmStringTable)malloc (sizeof(XmString) * Data->Search.Files);
      Hits = Data->Search.List;
      while (Hits != NULL)
      {
	List[count++] = XmStringCreate (Hits->Title, XmSTRING_DEFAULT_CHARSET);
	PRO1 ("===> %s\n", Hits->Title)
	Hits = Hits->Next;
      }
      XtSetArg (args[0], XmNitemCount, count);
      XtSetArg (args[1], XmNitems,     List);
      XtSetValues (Data->Search.ListW, args, 2);

      for (count = 0; count < Data->Search.Files; ++count)
	XmStringFree (List[count]);
      free (List);
    }
    elsif (Pattern[0] == '\0')
    {
      XtSetArg (args[0], XmNitemCount, 0);
      XtSetValues (Data->Search.ListW, args, 1);
    }
  }

  if ((hits = HELP_DataSearch(Data, Pattern, Method)) > 0 || 
      Pattern[0] == '\0')
  {
    XClearWindow (XDisplay, Data->Win);
    HELP_RefreshText (Data, 0, Data->SizeY);

#if USE_PIXMAP
      XtSetArg (args[0], XmNvalue, &PosY);
      XtGetValues (Data->ScrBarY, args, 1);

      XCopyArea (XDisplay, Data->Pic, Data->Win, Data->DrawGC, 
	  	           0, PosY, Data->SizeX, Data->WinY, 0, PosY);
#endif
    if (Data->Param.Click && Data->Selection >= 0)
      HELP_MarkHRef (Data, Data->Selection);
  }

  if (Data->Search.Hits <= 0 || 
      ! HELP_LocateHit(Data, Data->Search.Index+1))
    if (Pattern[0] != '\0')
      HELP_OpenDlg (Data->Search.Shell, "searchMsg", 
		                       "search string not found", NULL);

  TOOLSetMouseCursor (Data->Search.Shell, ToolNORMAL, NULL, NULL);
}


/* *********************************************************** */
/* ** Clears the actual marked search-hits in the text.     ** */
/* ** INPUT:  widget  Calling widget                        ** */
/* **         Data    Pointer to complete help-data         ** */
/* **         Ptr     Callback-data                         ** */
/* ** OUTPUT: none                                          ** */
/* *********************************************************** */
static void HELP_ClearCB (Widget widget, HELPDATA *Data,
			                XmAnyCallbackStruct *Ptr)
{
  Arg args[1];

  XtSetArg (args[0], XmNvalue, "");
  XtSetValues (Data->Search.Input, args, 1);

  HELP_SearchCB (widget, Data, Ptr);
}


/* *********************************************************** */
/* ** Closes the search-window and activates the locked     ** */
/* ** buttons.                                              ** */
/* ** INPUT:  widget  Calling widget                        ** */
/* **         Data    Pointer to complete data-structure    ** */
/* **         Ptr     Callback-data                         ** */
/* ** OUTPUT: none                                          ** */
/* *********************************************************** */
static void HELP_CloseCB (Widget widget, HELPDATA *Data,
			                XmAnyCallbackStruct *Ptr)
{
  HELP_ClearCB (widget, Data, Ptr);

  XtPopdown (Data->Search.Shell);
  XtDestroyWidget (Data->Search.Shell);

  Data->Search.Shell = NULL;

  XtSetSensitive (Data->Find, TRUE); 
  XtSetSensitive (Data->Close, TRUE); 
}

/* *********************************************************** */
/* ** Jumps to a hit-position, which was selected in the    ** */
/* ** list of results                                       ** */
/* ** INPUT:  widget  Calling widget                        ** */
/* **         Data    Pointer to complete data-structure    ** */
/* **         Ptr     Callback-data                         ** */
/* ** OUTPUT: none                                          ** */
/* *********************************************************** */
static void HELP_HitCB (Widget widget, HELPDATA *Data,
		                      XmListCallbackStruct *Ptr)
{
  int       Selection, count = 0;
  HIT_DATA *Hit;

  if (widget == NULL)
    return;

  Selection = Ptr->item_position-1;
  Hit = Data->Search.List;
  while (count < Selection && Hit != NULL)
  {
    count++;
    Hit = Hit->Next;
  }

  if (Hit == NULL || strcmp(Data->FileName, Hit->FileName) == 0)
    return;

  Data->Search.updated = FALSE;

  HELP_Jump (Data, Hit->FileName, "");
}

#endif
