#ifndef _HELP_LIB_H
#define _HELP_LIB_H

#include <limits.h>

#include <libtools.h>

#ifdef __cplusplus
extern "C" {
#endif


#define SEARCHING     1                 /* enable/disable Search-fucntion */
#define LABELLEN     40                 /* max. length of labelnames      */
#define MAX_REFS    200                 /* max. amount of references      */
#define MAX_LINES  4096                 /* max. amount of textlines       */
#define MEM_SIZE    100                 /* number of stored positions     */

static char *TestText[] = 
{
  "<EM>", "<BD>", "<REF>", "<ITEMIZE>", "<COLOR>", "<UNDERLINE>", "<FORM>"
};

#define TESTBITS (int)(sizeof(TestText)/sizeof(char*))


struct HELPDATA;
typedef struct HELPDATA HELPWIN;
typedef void (*HELPCB)(HELPWIN *Help);

typedef struct HIT_DATA
{
         char      Title[64];
         char      FileName[PATH_MAX+1];
  struct HIT_DATA *Next;
} HIT_DATA;


typedef struct
{
  char     *Pattern;              /* actual search pattern        */
  byte      Method;               /* Coded search-attributes      */
  int       Hits, Files;          /* Amount of Hits/Files         */
  HIT_DATA *List;                 /* list of found files          */
  int       Index;                /* actual index of hit          */
  Widget    Shell, Input;         /* used widgets...              */
  Widget    CaseW, SubStrW;
  Widget    ListW;
  Widget    Option[3];
  dual      updated;
} HELP_SEARCH;


typedef struct HELP_GEO
{
  unsigned short     PosX, PosY;
  unsigned short     w, h;
  struct   HELP_GEO *Next;
} HELP_GEO; 


typedef struct REF
{
  int       Scroll;                     /* Scrollbar-Pos fr Memory */
  char      FileName[PATH_MAX+1];
  char      Label[LABELLEN+1];
  HELP_GEO *geo;
} REF;


typedef struct
{
  unsigned short PosX, PosY;
           char  Name[LABELLEN+1];
} LABEL;


typedef struct HELP_ELEM
{
  unsigned long       Type;             /* Type of Element (ID)           */
           void      *Data;             /* additional data                */
           short      space;            /* horizontal/vertical space      */
           dual       line;             /* underlined text                */
           byte       align;            /* text alignment (HELP_ALIGN)    */
           byte       Index;            /* font size-index                */
           byte       FontID;           /* font ID                        */
           byte       ColorID;          /* color ID (HELP_COLOR_ID)       */
           short      ID;               /* HRef-number                    */
           short      SizeX, SizeY;     /* Size of element                */
           dual       fits;             /* fits search Argument           */
  struct   HELP_ELEM *Next;             /* Pointer to next element        */
} HELP_ELEM;


typedef struct
{
  unsigned short      PosX, PosY;
           short      SizeY, Max;
  unsigned short      Nr;
           double     Space;
           HELP_ELEM *List;
} HELP_LINE;


typedef struct TABLE_COL
{
  int        Type;                      /* TD/TH/empty */
  int        ColSpan, RowSpan;
  int        Color;
  int        align;
  HELP_ELEM *List;
} TABLE_CELL;

typedef struct
{
  int          SizeX, SizeY;
  int          Border;
  int          Rows, Cols;
  char        *Caption;
  int         *w, *h;
  TABLE_CELL **Data;  
} TABLE_DATA;


typedef struct
{
  dual HTML;
  dual Click;
  byte Scale;
} HELP_PARAM;


typedef struct HELPDATA
{
  char         Title[PATH_MAX+1];       /* Titlestring                    */
  char         FileName[PATH_MAX+1];    /* Filename of displayed text     */
  char         CurrDir[PATH_MAX+1];     /* Current directory              */

  HELP_ELEM   *first, *last;            /* List of elements               */
  byte         RNo;                     /* Amount of references in text   */
  REF          HRef[MAX_REFS];          /* References                     */
  byte         LNo;                     /* Amount of labels               */ 
  LABEL        Label[MAX_REFS];         /* Labels                         */
  byte         UNo;                     /* Amount of URIs                 */
  char         URI[5][PATH_MAX+1];      /* URIs for forms in the text     */

  short        Lines;                   /* Amount of (displayed) lines    */
  HELP_LINE    Line[MAX_LINES+1];       /* Line-list                      */

  HELP_PARAM   Param;                   /* Help parameters                */

  Widget       Shell;                   /* Referenced widgets             */
  Widget       ScrWin, ClipWin;
  Widget       ScrBarX, ScrBarY;
  Widget       Work;
  Widget       Follow, Back;
  Widget       Find, Close;

  REF          Memory[MEM_SIZE];        /* buffer for stored text         */
  byte         MemID;                   /* index into array               */

#if USE_PIXMAP
  Pixmap       Pic;
#endif
#if SEARCHING
  HELP_SEARCH  Search;                  /* Search data                    */
#endif

  Window       Win;                     /* context for drawins            */
  GC           DrawGC;

  byte         Flags[TESTBITS];         /* flags for correctness test     */ 

  int          SizeX, SizeY;            /* size of formated text          */
  int          PosX, PosY;              /* top-/left-position in scr-win  */
  int          WinX, WinY;              /* size of displayed screen       */
  int          MaxY;
  short        Selection;
  dual         overHRef;                /* Mouse is over HREF             */
  Pixel        hl;                      /* highlight color                */
  byte         bg;                      /* background color               */
  XImage      *bgImage;                 /* background image               */
  int          bgX, bgY;                /* size of background image       */

  HELPCB       HelpCB;                  /* User help callback function    */

  struct HELPDATA *Prev;                /* Pointer to last Help-Window    */
  struct HELPDATA *Next;                /* Pointer to next Help-Window    */
} HELPDATA;


HELPWIN *HELPOpenHelpWin   (const char   *Title,   const char   *TextSpec,
			          double  PosX,          double  PosY,   
			          double  SizeX,         double  SizeY, 
 			          char   *Pattern,       byte    Method,
			          HELPCB  HelpCB);
void     HELPLoad          (HELPWIN *HelpWin, 
			    const char *TextSpec, char *Pattern, byte Method);

dual     HELPGetSize       (HELPWIN *HelpWin, int *SizeX, int *SizeY);
dual     HELPGetPos        (HELPWIN *HelpWin, int *PosX,  int *PosY);
void     HELPClose         (HELPWIN *HelpWin);


#ifdef __cplusplus
} /* for C++ V2.0 */
#endif

#endif
