#ifndef LIBPRG_H
#define LIBPRG_H

#include <limits.h>

#include <libtools.h>

#ifdef __cplusplus
extern "C" {
#endif

#define MAX_ARGS 15
#define MAX_DESCRLEN 255
#define MAX_TITLE 80


typedef void (*MENU_FUNCTION)();
typedef void (*FUNCTION)(void);

#define TITLE(Name) {Name, PrgFUNC, "", "", ' ', TRUE, NULL, NULL, NULL, NULL}
#define ENTRY(Name,Acc,AccText,Mnemonic,Active,Function,Data) \
              {Name, PrgFUNC, Acc, AccText, Mnemonic, Active, NULL, \
	       (MENU_FUNCTION)Function, NULL, (void*)Data}
#define SWITCH(Name,Acc,AccText,Mnemonic,Active,Function,Value) \
               {Name, PrgTOGL, Acc, AccText, Mnemonic, Active, NULL, \
	       (MENU_FUNCTION)Function, NULL, (void*)Value}
#define ICON(Name,Acc,AccText,Mnemonic,Active,Function,Pic) \
             {Name, PrgCHECK, Acc, AccText, Mnemonic, Active, NULL, \
              (MENU_FUNCTION)Function, NULL, (void*)Pic}
#define SUB_MENU(Name,Acc,AccText,Mnemonic,Active,SubMenu) \
                 {Name, PrgSUB, Acc, AccText, Mnemonic, Active, NULL, NULL, \
                  (MENU_ENTRY*)SubMenu, \
                   (void*)(sizeof(SubMenu)/sizeof(MENU_ENTRY))}
#define MENU_LINE(Type) \
                  {NULL, PrgLINE, "", "", ' ', FALSE, NULL, NULL, NULL, \
                   (void*)Type}

#define FUNC_MENU(Name,Mnemonic,Active,Menu) \
                   {Name, PrgFUNC_MENU, Mnemonic, Active, NULL, \
	            sizeof(Menu)/sizeof(MENU_ENTRY), (MENU_ENTRY*)Menu}
#define HELP_MENU(Name,Mnemonic,Active,Menu) \
                   {Name, PrgHELP_MENU, Mnemonic, Active, NULL, \
	            sizeof(Menu)/sizeof(MENU_ENTRY), (MENU_ENTRY*)Menu}

#define MAIN_MENU(Name,MenuCols) \
                  {Name, NULL, sizeof(MenuCols)/sizeof(MENU_COLUMN), MenuCols}


#define TOOL_ICON(Data,Function,Param,Description) \
                  PRGAddToolIcon (-1, -1, (void *)Data, \
				  (MENU_FUNCTION)Function, (void *)Param, \
				  Description)
#define TOOL_SPACE(Space) PRGAddToolSpace (Space);


enum MENU_TYPES
{
  PrgFUNC_MENU = 0,
  PrgSUB_MENU,
  PrgHELP_MENU
};

enum ENTRY_TYPES
{
  PrgFUNC = 0,
  PrgTOGL,
  PrgCHECK,
  PrgSUB,
  PrgLINE
};

enum TOOLBAR_POS
{
  PrgNO_TOOL = 0,
  PrgWIN_TOOL,
  PrgTOP_TOOL,
  PrgLEFT_TOOL,
  PrgRIGHT_TOOL
};

enum SEPARATOR_TYPES
{
  PrgSINGLE_LINE = 0,
  PrgDOUBLE_LINE,
  PrgSINGLE_DASH,
  PrgDOUBLE_DASH,
  PrgETCHED_IN,
  PrgETCHED_OUT,
  PrgDASH_IN,
  PrgDASH_OUT
};

typedef struct MENU_ENTRY
{
         char          *Name;
         int            Type;
         char           Acc[20];
         char           AccText[15];
         char           Mnemonic;
         dual           active;
         Widget         widget;
         MENU_FUNCTION  FPtr;
  struct MENU_ENTRY    *Sub;
         void          *Data;
} MENU_ENTRY;

typedef struct
{
  char        *Name;
  int          Type;
  char         Mnemonic;
  dual         active;
  Widget       widget;
  int          Rows;
  MENU_ENTRY  *Entry;
} MENU_COLUMN;

typedef struct
{
  char         *Title;
  Widget        widget;
  int           Cols;
  MENU_COLUMN  *Column;
} MENU_BAR;


typedef struct ICONDATA
{
         int           SizeX, SizeY;
	 int           Dist;
         void         *Data;
         char          PicName[PATH_MAX+1];
         Pixmap        PixMap;
         MENU_FUNCTION FPtr;
         void         *Param;
         dual          active;
         char          Descr[MAX_DESCRLEN+1];
	 Widget        Display;
	 XtIntervalId  Timer;
  struct ICONDATA *Next;
} ICONDATA;

typedef struct
{
  int        Pos;
  int        Number;
  int        Delay;
  ICONDATA  *Entry;
  ICONDATA  *Last;
  dual       active;
  Widget     widget;
  Widget     Shell;
} TOOLDATA;

typedef struct
{
  Widget    Output;
} INFODATA;


typedef struct
{
  char     Title[MAX_TITLE+1];
  int      PosX, PosY;
  int      SizeX, SizeY;

  Widget   Shell;
  Widget   List;

  int      LineNo;
  char   **Lines;
} LOGWIN;


typedef struct
{
  char      Title[MAX_TITLE+1];
  int       PosX, PosY;
  double    SizeX, SizeY;

  MENU_BAR *MenuBar;
  TOOLDATA *ToolBar;
  INFODATA *Info;

  Widget    Main;

  LOGWIN   *Log;

  char     *ScriptName;
} PROG;


extern void    PRGInitLib           (char *Title, int *argc, char **argv,
				     char **FallbackResources);
extern void    PRGSetPos            (double PosX,  double PosY);
extern void    PRGSetSize           (double SizeX, double SizeY);
extern Widget  PRGCreateMenu        (Widget Parent, char *Name, int x, int y, 
				     MENU_ENTRY *Entry, int Number);
extern Widget  PRGCreateMenuBar     (Widget Parent, MENU_BAR *MenuBar);
extern Widget  PRGCreateProgram     (char *Title, MENU_BAR *MenuBar);
extern void    PRGStartProgram      (MENU_FUNCTION Function, void *Param);
extern void    PRGExitProgram       (void);

extern dual    PRGGetPos            (int *PosX,  int *PosY);
extern dual    PRGGetSize           (int *SizeX, int *SizeY);

extern dual    PRGSetMenuActive     (char *Name, dual active);
extern dual    PRGSetToolBarActive  (dual active);
extern void    PRGShowToolBar       (dual visible);

extern void    PRGMakeToolBar       (int Position);
extern void    PRGAddToolIcon       (int SizeX, int SizeY,   void *Data,
				     MENU_FUNCTION Function, void *Param, 
				     char *Description);
extern void    PRGAddToolSpace      (int Space);
extern dual    PRGSetToolBarActive  (dual active);
extern void    PRGSetDelayTime      (int Delay);

extern char   *PRGRunScript         (char *FileName);

extern int     PRGPrintf            (const char *Message, ...);

extern dual    PRGOpenLogWindow     (const char *Title, int SizeX, int SizeY);
extern void    PRGSetLogPos         (double PosX, double PosY);
extern int     PRGLogPrintf          (const char *Message, ...);
extern dual    PRGClearLog          ();

extern int     PRGStartAnim         (MENU_FUNCTION Function, void *Param,
				     int Delay);
extern dual    PRGStopAnim          (int ID);


#ifdef __cplusplus
} /* for C++ V2.0 */
#endif

#endif

