#include <stdarg.h>

#include <Xm/DialogS.h>
#include <Xm/MwmUtil.h>
#include <Xm/Form.h>
#include <Xm/Text.h>

#include "prgGlobal.h"


void ScrollCB (Widget w, caddr_t *Data, XmScrollBarCallbackStruct *Ptr)
{
  printf ("ScrollCB\n");
}


dual PRGOpenLogWindow (const char *Title, int SizeX, int SizeY)
{
  Arg     args[16];
  LOGWIN *Log;

  if (Program == NULL)
    return FALSE;
 
  if (Program->Log != NULL)
  {
    PRGClearLog ();
    return TRUE;
  }

  Program->Log = Log = (LOGWIN *)malloc (sizeof(LOGWIN));
  if (Title != NULL && Title[0] != '\0')
  {
    strncpy (Log->Title, Title, MAX_TITLE);
    Log->Title[MAX_TITLE] = '\0';
  }
  else
    strcpy (Log->Title, "Log-Window");

  Log->PosX  = 0;
  Log->PosY  = 0;
  Log->SizeX = SizeX;
  Log->SizeY = SizeY;

  Log->Shell = XtVaCreatePopupShell (Log->Title, transientShellWidgetClass, 
				     AppShell, 
				     XmNwidth,               Log->SizeX,
				     XmNheight,              Log->SizeY,
				     XmNx,                   Log->PosX,
				     XmNy,                   Log->PosY,
				     XmNallowShellResize,    True,
				     XmNmwmFunctions,        MWM_FUNC_ALL,
				     XmNmwmDecorations,      MWM_DECOR_ALL,
				     XmNkeyboardFocusPolicy, XmPOINTER,
				     NULL);

  XtSetArg (args[0], XmNresizeWidth,            True);
  XtSetArg (args[1], XmNresizeHeight,           True);
  XtSetArg (args[2], XmNscrollingPolicy,        XmAUTOMATIC);
  XtSetArg (args[3], XmNscrollBarDisplayPolicy, XmAS_NEEDED);
  XtSetArg (args[4], XmNscrollHorizontal,       True);
  XtSetArg (args[5], XmNscrollVertical,         True);
  XtSetArg (args[6], XmNwordWrap,               False);
  XtSetArg (args[7], XmNeditMode,               XmMULTI_LINE_EDIT);
  XtSetArg (args[8], XmNeditable,               False);  
  Log->List = XmCreateScrolledText (Log->Shell, "list", args, 9);
  XtManageChild (Log->List);

#if 0
  XtVaGetValues (Log->List, XmNhorizontalScrollBar, &ScrollBar, NULL);
  if (ScrollBar != NULL)
    XtAddCallback (ScrollBar, XmNvalueChangedCallback, 
		              (XtCallbackProc)ScrollCB, NULL);
#endif

  XmTextEnableRedisplay (Log->List);

  XtPopup (Log->Shell, XtGrabNone);

  return TRUE;
}


extern void PRGSetLogPos (double PosX, double PosY)
{
    int ScrX, ScrY;

  PRO_START ("")

  if (Program == NULL || Program->Log == NULL)
    PRO_ABORT ("No program defined !\n")

  ScrX = DisplayWidth  (XDisplay, XScreen);
  ScrY = DisplayHeight (XDisplay, XScreen);

  if (PosX > 0.0 && PosX < 1.0)
    Program->Log->PosX = (int)(PosX * ScrX);
  elsif (PosX >= 0.0)
    Program->Log->PosX = (int)MIN (PosX, ScrX);
  else
    Program->Log->PosX = -1;

  if (PosY > 0.0 && PosY < 1.0)
    Program->Log->PosY = (int)(PosY * ScrY);
  elsif (PosY >= 0.0)
    Program->Log->PosY = (int)MIN (PosY, ScrY);
  else
    Program->Log->PosY = -1;

  XtVaSetValues (Program->Log->Shell, XmNx, Program->Log->PosX, 
		                      XmNy, Program->Log->PosY,
		                      NULL);

  PRO_END
}


int PRGLogPrintf (const char *Msg, ...)
{
  int             Result = 0;
  char            Message[1024];
  va_list         Arguments;
  LOGWIN         *Log;
  XmTextPosition  Pos;

  if (Program == NULL || Message == NULL)
    return -1;

  va_start (Arguments, Msg);
  Result = vsprintf (Message, Msg, Arguments);
  va_end (Arguments);

  if (Message[0] == '\\')
    PRO1 ("\\%s\n", Message+1)
  else
    PRO1 ("%s\n", Message)

  if ((Log = Program->Log) == NULL)
    return 0;

  Pos = XmTextGetLastPosition (Log->List);

  XmTextInsert       (Log->List, Pos+1, Message);
  XmTextShowPosition (Log->List, Pos + (XmTextPosition)strlen(Message));

  XFlush (XDisplay);

  return Result;  
}


dual PRGClearLog ()
{
  if (Program != NULL && Program->Log != NULL)
    XmTextSetString (Program->Log->List, "");

  return TRUE;
}
