/* (Program- and) Dialog Construction Tools */
/* (c) 1995 by Stefan Vietinghoff               */

#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <ctype.h>
#include <string.h>

#include "prgGlobal.h"

/* Shell-Typen: */
#include <X11/Shell.h>

/* Geometrie-Management: */
#include <Xm/Form.h>

/* Font-Cursor */
#include <X11/cursorfont.h>

/* Widget-Klassen: */
#include <Xm/MainW.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
#include <Xm/Text.h>

#include "tux.xpm"


#define START_TIMER_DELAY 500


static void  PRG_ExitCB (Widget widget, void *Data, XmAnyCallbackStruct *Ptr);


       PROG           *Program = NULL;
static Arg             args[MAX_ARGS];
static MENU_FUNCTION   StartFunction = NULL;


/* ******************************************* */
/* ** creates and initializes a program-    ** */
/* ** structure                             ** */
/* ** INPUT:  ---                           ** */
/* ** OUTPUT: (PPROG)  pointer to program-  ** */
/* **                  data                 ** */
/* ******************************************* */
PROG *PRG_NewProgram (void)
{
  PROG *Result = NULL;

  PRO_START ("")

  Result = (PROG*)calloc (1, sizeof(PROG));
  Result->Title[0] = '\0';
  Result->PosX     = -1;
  Result->PosY     = -1;
  Result->SizeX    =  0;
  Result->SizeY    =  0;

  Result->MenuBar = NULL;
  Result->ToolBar = NULL;
  Result->Info    = NULL;

  Result->Log     = NULL;

  Result->Main = (Widget)NULL;

  Result->ScriptName = NULL;

  PRO_END
  return Result;
}


/* ********************************************************************** */
/* ** Initalizes the program-library. Because, this library usess the  ** */
/* ** TOOL-lib, it will be initialized, too (if not yet done).         ** */
/* ** Call this function before parsing the program-parameters,        ** */
/* ** because XtAppInitialize (calls itself XtDisplayInitialize)       ** */
/* ** extracts the X-Windows relevant parameters and returns the       ** */
/* ** reduced list.                                                    ** */
/* ** INPUT:  Title  Title of the created programm (also used for      ** */
/* **                recource-entries.                                 ** */
/* **         argc   amount of programm-arguments                      ** */
/* **         argv   pointer to this arguements                        ** */
/* ** OUPUT:  ---                                                      ** */
/* ********************************************************************** */ 
void PRGInitLib (char *Title, int *argc, char **argv, char **FallbackResources)
{
  PRO_START ("\n")

  if (! TOOLIsInitialized())
    TOOLInitLib (Title, argc, argv, FallbackResources);

  TOOLAddExitCB ((XtCallbackProc)PRG_ExitCB, NULL);

  Program = PRG_NewProgram();

  PRO_END
}


static Widget PRG_MakeInfoArea (Widget Parent, PROG *Program)
{
  XmString Label;

  if (Program == NULL || Program->Info != NULL)
    return NULL;

  Program->Info = (INFODATA*)malloc (sizeof(INFODATA));

  Label = XmStringLtoRCreate (" ", XmSTRING_DEFAULT_CHARSET);
  XtSetArg (args[ 0], XmNlabelString,            Label);
  XtSetArg (args[ 1], XmNleftAttachment,         XmATTACH_FORM);
  XtSetArg (args[ 2], XmNleftOffset,             5);
  XtSetArg (args[ 3], XmNrightAttachment,        XmATTACH_FORM);
  XtSetArg (args[ 4], XmNrightOffset,            5);
  XtSetArg (args[ 5], XmNbottomAttachment,       XmATTACH_FORM);
  XtSetArg (args[ 6], XmNbottomOffset,           5);
  XtSetArg (args[ 7], XmNmarginHeight,           5);
  XtSetArg (args[ 8], XmNeditable,               FALSE);
  Program->Info->Output = XmCreateText (Parent, "info", args, 9);
  XtManageChild (Program->Info->Output);

  XmStringFree (Label);

  return Program->Info->Output;
}


static Widget PRG_MakeWorkArea (PROG *Program)
{
  int    n = 0;
  Widget Frame;
  Widget Result;

  PRO_START ("\n")

  if (Program == NULL)
    PRO_RETURN ("Invalid Program", NULL)

  /* left */
  if (Program->ToolBar != NULL && Program->ToolBar->Pos == PrgLEFT_TOOL)
  {
    XtSetArg (args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (args[n], XmNleftWidget,     Program->ToolBar->widget);
  }
  else
    XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); 
  n++;

  /* top */
  if (Program->ToolBar != NULL && Program->ToolBar->Pos == PrgTOP_TOOL)
  {
    XtSetArg (args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (args[n], XmNtopWidget,     Program->ToolBar->widget);
  }
  elsif (Program->MenuBar != NULL)
  {
    XtSetArg (args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (args[n], XmNtopWidget,     Program->MenuBar->widget);
  }
  else
    XtSetArg (args[n], XmNtopAttachment, XmATTACH_FORM);
  n++;

  /* bottom */
  if (Program->Info)
  {
    XtSetArg (args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (args[n], XmNbottomWidget,     Program->Info->Output); n++;
    XtSetArg (args[n], XmNbottomOffset,     5);
  }
  else
    XtSetArg (args[n], XmNbottomAttachment, XmATTACH_FORM);
  n++;

  /* right */
  XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;

  Frame = XtCreateManagedWidget ("frame", xmFrameWidgetClass, 
				          Program->Main, args, n);

  Result = XtCreateManagedWidget ("work", xmFormWidgetClass, Frame, NULL, 0);

  PRO_END
  return Result;
}


/* ********************************************************************** */
/* ** (De-)activates a Menu-Entry or -Column. The Name specified the   ** */
/* ** appropriate element using the menu-hierarchy.                    ** */
/* ** INPUT:   Name    specification of the menu-range.                ** */
/* **          active  TRUE for activating the entry, FALSE to         ** */
/* **                  deactivate it.                                  ** */
/* ** OUPUT:   (bool)  success of this operation.                      ** */
/* ** EXAMPLE: PRGSetMenuActive ("/system/save", FALSE);               ** */
/* **            deactivates the save-function in the submenu "system" ** */
/* ********************************************************************** */
dual PRGSetMenuActive (char *Name, dual active)
{
  return (PRG_SetMenuActive(Program, Name, active));
}


/* ********************************************************************** */
/* ** Create a new program                                             ** */
/* ** INPUT:  Title     Name of Program                                ** */
/* **         MenuBar   menu-structure                                 ** */
/* ** OUTPUT: (Widget)  Work-Area (Form). Might be filled by user      ** */
/* ** EXAMPLE:                                                         ** */
/* **   Widget Work;                                                   ** */
/* **   ...                                                            ** */
/* **   static MENU_BAR Menu = ...                                     ** */
/* **   Work = PRGCreateProgram ("prog", Menu);                        ** */
/* ********************************************************************** */
Widget PRGCreateProgram (char *Title, MENU_BAR *MenuBar)
{
  Widget Work, Menu;
  Widget Info;
  Widget MainWin;

  PRO_START ("\n")

  if (Program == NULL)
    PRO_RETURN ("Bad Program", NULL)

  if (! TOOLIsInitialized())
  {
    XtErrorMsg ("InitErr", "StartPrg", "Error",
		"Library is not initialized", NULL, NULL);
    PRO_RETURN ("BaseLib is NOT initialized", NULL)
  }

  XtSetMappedWhenManaged (AppShell, FALSE);

  strncpy (Program->Title, Title, MAX_TITLE);
  Program->Title[MAX_TITLE] = '\0';

  XtSetArg (args[0], XmNtitle, Title);
  MainWin = XmCreateMainWindow (AppShell, "mainWindow", args, 1);
  XtManageChild (MainWin);

  XtSetArg (args[0], XmNtopAttachment,    XmATTACH_FORM);
  XtSetArg (args[1], XmNleftAttachment,   XmATTACH_FORM);
  XtSetArg (args[2], XmNrightAttachment,  XmATTACH_FORM);
  XtSetArg (args[3], XmNbottomAttachment, XmATTACH_FORM);
  Program->Main = XmCreateForm (MainWin, "form", args, 4);
  XtManageChild (Program->Main);

  PRO0 ("Menu-Bar...\n")
  PRG_SetMenuBar (Program, MenuBar);
  Menu = PRGCreateMenuBar (Program->Main, Program->MenuBar);

  PRO0 ("Display...\n")
  Info = PRG_MakeInfoArea (Program->Main, Program);

  PRO0 ("ToolBar...\n")
  PRG_MakeToolBar (Program);

  TOOLInitGCs ();

  PRO0 ("Working Area...\n")
  Work = PRG_MakeWorkArea (Program);

  PRO_END
  return Work;
}


/* ********************************************************************** */
/* ** defines the default position of the program-window               ** */
/* ** INPUT:  PosX                                                     ** */
/* **         PosY  0<Val<1: pos in percent of screen-size (/100)      ** */ 
/* **               1<=Val:  absolute posotion [Pixel]                 ** */
/* ** OUTPUT: ---                                                      ** */
/* ** EXAMPLE:                                                         ** */
/* **   PRGSetPos (0.2, 300);                                          ** */
/* **   positions at 20% of width, at height 300                       ** */
/* ********************************************************************** */
void PRGSetPos (double PosX, double PosY)
{
  int ScrX, ScrY;

  PRO_START ("")

  if (Program == NULL)
    PRO_ABORT ("No program defined !\n")

  ScrX = DisplayWidth  (XDisplay, XScreen);
  ScrY = DisplayHeight (XDisplay, XScreen);

  if (PosX > 0.0 && PosX < 1.0)
    Program->PosX = (int)(PosX * ScrX);
  elsif (PosX >= 0.0)
    Program->PosX = (int)MIN (PosX, ScrX);
  else
    Program->PosX = -1;

  if (PosY > 0.0 && PosY < 1.0)
    Program->PosY = (int)(PosY * ScrY);
  elsif (PosY >= 0.0)
    Program->PosY = (int)MIN (PosY, ScrY);
  else
    Program->PosY = -1;

  PRO_END
}


/* ********************************************************************** */
/* ** defines the default size of the program-window                   ** */
/* ** INPUT:  SizeX                                                    ** */
/* **         SizeY   Val<0:   percent of other direction's size       ** */
/* **                 0<Val<1: size in percent of screen-size (/100)   ** */
/* **                 1<=Val:  absolute size [Pixel]                   ** */
/* **                 Val=0:   use prefered size                       ** */
/* ** OUTPUT: ---                                                      ** */
/* ** EXAMPLE:                                                         ** */
/* **   PRGSetSize (0.5, 0);            50 % width and prefered height ** */
/* **   PRGSetSize (0, -2.0);        twice the hight of perfered width ** */
/* ********************************************************************** */
void PRGSetSize (double SizeX, double SizeY)
{
  int ScrX, ScrY;

  PRO_START ("")

  if (Program == NULL)
    PRO_ABORT ("No program defined !\n")

  ScrX = DisplayWidth  (XDisplay, XScreen);
  ScrY = DisplayHeight (XDisplay, XScreen);

  if (SizeX > 0.0 && SizeX < 1.0)
    Program->SizeX = floor(SizeX * ScrX);
  elsif (SizeX >= 1.0)
    Program->SizeX = MIN (floor(SizeX), ScrX);
  elsif (SizeX < 0.0)
    Program->SizeX = SizeX;
  else
    Program->SizeX = 0.0;

  if (SizeY > 0.0 && SizeY < 1.0)
    Program->SizeY = floor(SizeY * ScrY);
  elsif (SizeY >= 1.0)
    Program->SizeY = MIN (floor(SizeY), ScrY);
  elsif (SizeY < 0.0)
    Program->SizeY = SizeY;
  else
    Program->SizeY = 0.0;

  PRO_END
}


dual PRGGetPos (int *PosX, int *PosY)
{
  Dimension x, y;

  if (Program == NULL)
    PRO_RETURN ("No program defined !", FALSE)

  XtVaGetValues (AppShell, XmNx, &x,
		           XmNy, &y,
		           NULL);

  Program->PosX = x;
  Program->PosY = y;

  if (PosX != NULL)
    *PosX = Program->PosX;
  if (PosY != NULL)
    *PosY = Program->PosY;

  return TRUE;
}


dual PRGGetSize (int *SizeX, int *SizeY)
{
  Dimension w, h;

  if (Program == NULL)
    PRO_RETURN ("No program defined !", FALSE)

  XtVaGetValues (AppShell, XmNwidth,  &w,
		           XmNheight, &h,
		           NULL);

  Program->SizeX = w;
  Program->SizeY = h;

  if (SizeX != NULL)
    *SizeX = (int)Program->SizeX;
  if (SizeY != NULL)
    *SizeY = (int)Program->SizeY;

  return TRUE;
}


static void PRG_StartTimer (void *Param, XtIntervalId Timer)
{
  if (StartFunction != NULL)
#ifndef __CPP__
    StartFunction (Param);
#else
    StartFunction ();
#endif

  if (Program->ScriptName != NULL)
    PRGRunScript (Program->ScriptName);
}


static void PRG_EvHandler (Widget widget, void *Param, XEvent *event)
{
  if (widget != AppShell)
    return;

  switch (event->type)
  {
    case ReparentNotify :
      XFlush (XDisplay);
      if (Program->ScriptName != NULL || StartFunction != NULL)
	XtAppAddTimeOut (AppCon, START_TIMER_DELAY, 
			 (XtTimerCallbackProc)PRG_StartTimer, Param);
      XtRemoveEventHandler (AppShell,
			    VisibilityChangeMask | StructureNotifyMask, 
			    FALSE, (XtEventHandler)PRG_EvHandler, Param);
      break;

    default :
      PRO1 ("EVENT: %d\n", event->type)
  }
}


/* ********************************************************************** */
/* ** starts the program                                               ** */
/* ** INPUT:   Function  to be called after starting                   ** */
/* **          Param     parameters for function                       ** */
/* ** OUTPUT:  ---                                                     ** */
/* ** EXAMPLE: PRGStartProgram (NULL, NULL);                           ** */
/* ********************************************************************** */
void PRGStartProgram (MENU_FUNCTION Function, void *Param)
{
  int       n = 0;
  char      geo_string[12];
  Dimension w, h;

  PRO_START ("\n")

  if (Program == NULL)
    PRO_ABORT ("No program defined !")

  XtRealizeWidget (AppShell);
  StartFunction = Function;

  XtAddEventHandler (AppShell, VisibilityChangeMask | StructureNotifyMask, 
		     FALSE, (XtEventHandler)PRG_EvHandler, Param);

  if (Program->PosX != -1 && Program->PosY != -1)
  {
    sprintf (geo_string, "+%d+%d", Program->PosX, Program->PosY);
    XtSetArg (args[n], XmNgeometry, geo_string); n++;
    XtSetArg (args[n], XmNx, Program->PosX); n++;
    XtSetArg (args[n], XmNy, Program->PosY); n++;
  }

  if (Program->SizeX > 0.0)
  {
    XtSetArg (args[n], XmNwidth, (int)Program->SizeX); n++;
  }
  if (Program->SizeY > 0.0)
  {
    XtSetArg (args[n], XmNheight, (int)Program->SizeY); n++;
  }
  XtSetValues (AppShell, args, n);

  if (Program->SizeX < 0.0)
  {
    XtVaGetValues (AppShell, XmNheight, &h, NULL);
    XtVaSetValues (AppShell, XmNwidth, h * (-1.0 * Program->SizeX), NULL);
  }
  elsif (Program->SizeY < 0.0)
  {
    XtVaGetValues (AppShell, XmNwidth, &w, NULL);
    XtVaSetValues (AppShell, XmNheight, w * (-1.0 * Program->SizeY), NULL);
  }

  TOOLSetIcon (AppShell, Program->Title, NULL, tux_xpm);

  XFlush (XDisplay);

  PRO0 ("\\Shell...")
  XtPopup (AppShell, XtGrabNone);

  if (Program->ToolBar != NULL && Program->ToolBar->Shell != NULL)
  {
    PRO0 ("\\ToolBar...")
    XtRealizeWidget (Program->ToolBar->Shell);
    XtPopup (Program->ToolBar->Shell, XtGrabNone);
  }
  PRO0 ("\\\n")

  PRO0 ("\\MainLoop...\n")
  XtAppMainLoop (AppCon);

  /* never reached */
  PRO_END
}


void PRGExitProgram (void)
{
  PRO_START ("\n")

  TOOLExitLib ();

  PRO_END
}


static void PRG_ExitCB (Widget widget, void *Data, XmAnyCallbackStruct *Ptr)
{
  PRO_START ("\n")

  if (Program == NULL)
    PRO_ABORT ("Nothing to do")

  PRO0 ("Menu...\n")
  if (Program->MenuBar != NULL)
    free (Program->MenuBar);

  if (Program->ToolBar != NULL)
    PRG_FreeToolBar (Program->ToolBar);

  if (Program->Info != NULL)
    free (Program->Info);

  PRO0 ("Program\n")
  free (Program);
  Program = NULL;

  PRO_END
}
