#include <Xm/RowColumn.h>
#include <Xm/Separator.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/ToggleB.h>
#include <Xm/CascadeB.h>

#if (XmVERSION>=2 && XmREVISION>=1)
#include <Xm/IconG.h>
#endif

#include "prgGlobal.h"

static void PRG_CallFktCB (Widget widget, MENU_ENTRY *Entry,
			                  XmAnyCallbackStruct *Data);

static unsigned char SeparatorType[] =
{
  XmSINGLE_LINE,           XmDOUBLE_LINE, 
  XmSINGLE_DASHED_LINE,    XmDOUBLE_DASHED_LINE,
  XmSHADOW_ETCHED_IN,      XmSHADOW_ETCHED_OUT, 
  XmSHADOW_ETCHED_IN_DASH, XmSHADOW_ETCHED_OUT_DASH
};

static Arg args[MAX_ARGS];



static MENU_ENTRY *PRG_SearchFunction (MENU_ENTRY *Entry, 
				       int Number, char *Name)
{
  int         row, len;
  int         NameLen;
  MENU_ENTRY *Result;

  if (Entry == NULL)
    return NULL;

  NameLen = strlen(Name);
  len = 0;
  while (len < NameLen && Name[len] != '/')
    len++;

  for (row = 0; row < Number; ++row)
  {
    if (Entry->Name != NULL && strncmp(Name, Entry->Name, len) == 0)
      switch (Entry->Type)
      {
        case PrgSUB :
	  Result = PRG_SearchFunction (Entry->Sub, 
				       (int)Entry->Data, Name+len+1);
	  if (Result != NULL)
	    return Result;
	  break;

	default :
	  if (len == NameLen)
	    return Entry;
      }
    Entry++;
  }
  return NULL;
}


MENU_ENTRY *PRG_SearchMenuEntry (MENU_BAR *MenuBar, char *Name)
{
  int         count = 0;
  int         len = 0, NameLen;
  MENU_COLUMN Column;

  if (Name == NULL)
    return NULL;

  NameLen = strlen(Name);
  while (len < NameLen && Name[len] != '/')
    len++;

  while (count < MenuBar->Cols)
  {
    Column = MenuBar->Column[count];
    if (strncmp(Name, Column.Name, len) == 0)
      return PRG_SearchFunction (Column.Entry, Column.Rows, Name+len+1);
    else
      count++;
  }
  return NULL;
} 


void PRG_SetShortKey (Widget Parent, Widget Button, MENU_ENTRY *Entry)
{
  int       n = 0;
  char     *Acc;
  char     *AccText = NULL;
  char      Mnemonic;
  XmString  String = NULL;

  if (Entry->Acc[0] != '\0')
  {
    Acc = TOOLGetResString (Parent, "xmPushButtonWidgetClass", 
			    Entry->Name, XmNaccelerator, Entry->Acc);
    XtSetArg (args[n], XmNaccelerator, Acc); n++;
  }

  if (Entry->AccText[0] != '\0')
  {
    AccText = TOOLGetResString (Parent, "xmPushButtonWidgetClass", 
				Entry->Name, XmNacceleratorText, 
				Entry->AccText);
    String = XmStringLtoRCreate (AccText, XmSTRING_DEFAULT_CHARSET);
    XtSetArg (args[n], XmNacceleratorText, String); n++;
  }

  if (Entry->Mnemonic != ' ' && Entry->Mnemonic != '\0')
  {
    Mnemonic = TOOLGetResChar (Parent, "xmPushButtonWidgetClass", 
			       Entry->Name, XmNmnemonic, Entry->Mnemonic);
    XtSetArg (args[n], XmNmnemonic, Mnemonic); n++;
  }

  XtSetValues (Button, args, n);

  if (AccText != NULL)
    XmStringFree (String);
}


Widget PRG_CreateMenu (Widget Parent, char *Title, char Mnemonic,
		       MENU_ENTRY *Entry, int Number, dual Pulldown)
{
  int      start = 0, n = 0, row;
  char     Mnemonic2;
  XmString Text = NULL;
  Widget   PullDown, Cascade;
  Widget   Result;

  if (Parent == NULL || Entry == NULL)
    return NULL;

  if (Entry->Name == NULL || Entry->Name[0] == '\0')
  {
    XtSetArg (args[n], XmNtearOffModel, XmTEAR_OFF_ENABLED); n++;
#if (XmVERSION>=2)
      Text = XmStringLtoRCreate (Title, XmSTRING_DEFAULT_CHARSET);
      XtSetArg (args[n], XmNtearOffTitle, Text); n++;
#endif
    start = 1;
    Entry++;
  }

  if (Pulldown)
    PullDown = XmCreatePulldownMenu (Parent, Title, args, n);
  else
  {
    XtSetArg (args[0], XmNwhichButton, Button1);
    PullDown = XmCreatePopupMenu (Parent, Title, args, 1);
  }

  if (Text != NULL)
    XmStringFree (Text);

  for (row = start; row < Number; ++row)
  {  
    if (Entry->Name == NULL || Entry->Name[0] == '\0')
    {
      if ((int)Entry->Data >= 0 && (int)Entry->Data <= PrgDASH_OUT)
	  Entry->widget = 
	    XtVaCreateManagedWidget ("seperator", xmSeparatorWidgetClass, 
				     PullDown, XmNseparatorType, 
				     SeparatorType[(int)Entry->Data], NULL);
    }
    else
    {
      PRO2 ("MENUENTRY: %d %s\n", Entry->Type, Entry->Name)
      switch (Entry->Type)
      {
        case PrgFUNC :
	  if (Entry->FPtr != NULL)
	  {
	    Entry->widget =
	      XtCreateManagedWidget (Entry->Name, xmPushButtonWidgetClass,
				                  PullDown, NULL, 0);
	    PRG_SetShortKey (PullDown, Entry->widget, Entry);
	    XtAddCallback (Entry->widget, XmNactivateCallback, 
			   (XtCallbackProc)PRG_CallFktCB, Entry);
	  }
	  else
	    Entry->widget = 
	      XtCreateManagedWidget (Entry->Name, xmLabelWidgetClass,
				                  PullDown, NULL, 0);
	  break;

	case PrgTOGL :
	  XtSetArg (args[0], XmNvisibleWhenOff, True);
          XtSetArg (args[1], XmNfillOnSelect,   True);
	  XtSetArg (args[2], XmNset, (Entry->Data==NULL)? FALSE:Entry->Data);
	  XtSetArg (args[3], XmNselectColor,    TOOLGetColor("Red"));
  	  Entry->widget = 
	    XtCreateManagedWidget (Entry->Name, xmToggleButtonWidgetClass,
					        PullDown, args, 3);
          PRG_SetShortKey (PullDown, Entry->widget, Entry);

	  XtAddCallback (Entry->widget, XmNvalueChangedCallback,
			                (XtCallbackProc)PRG_CallFktCB, Entry);
	  break;

        case PrgCHECK :
        {
#if 0
	  XmString Text, Check, Complete;

	  Text = XmStringSegmentCreate ("Bla Bla ", NORMAL_FONT,
					XmSTRING_DIRECTION_L_TO_R, False);
	  Check = XmStringSegmentCreate ("\165 ", "symbol",
					 XmSTRING_DIRECTION_L_TO_R, False);
	  Complete = XmStringConcat (Text, Check);
	  XtSetArg (args[0], XmNlabelString, Complete);
	  Entry->widget = 
	    XtCreateManagedWidget ("check", xmPushButtonWidgetClass,
				            PullDown, args, 1);
#endif
#if (XmVERSION>=2 && XmREVISION>=1)
	  Pixmap   Pix;
          Pixel    bg, fg;
	  XmString Name;
	  
	  XtVaGetValues (Parent, XmNbackground, &bg,
			         XmNforeground, &fg,
			         NULL);
	  Pix = TOOLMakeXPM ((char **)Entry->Data, NULL, NULL, bg);
	  if (Pix != XmUNSPECIFIED_PIXMAP)
	  {
	    Entry->widget = XtCreateManagedWidget ("ig", xmIconGadgetClass,
						         PullDown, NULL, 0);
	    Name = XmStringCreateLocalized (Entry->Name);
	    XtVaSetValues (Entry->widget, XmNlabelString,      Name, 
			                  XmNviewType,         XmSMALL_ICON,
			                  XmNsmallIconPixmap,  Pix,
			                  XmNspacing,          10,
			                  XmNvisualEmphasis,   XmNOT_SELECTED,
			                  XmNshadowThickness,  0,
			                  NULL, NULL);
	    XtAddCallback (Entry->widget, XmNactivateCallback, 
			                  (XtCallbackProc)PRG_CallFktCB, 
			                  Entry);
	    PRG_SetShortKey (PullDown, Entry->widget, Entry);
	  }
#endif
	}
        break;

	case PrgSUB :
	  PRO1 ("SUBMENU: %d\n", (int)Entry->Data)
	  Entry->widget =
	    PRG_CreateMenu (PullDown, Entry->Name, Entry->Mnemonic,
			              Entry->Sub, (int)Entry->Data, TRUE);
	  break;

	default :
	  break;
      }
      if (! Entry->active && Entry->widget != NULL)
	XtSetSensitive (Entry->widget, (Boolean)FALSE);
    }
    Entry++;
  }

  if (Pulldown)
  {
    XtSetArg (args[0], XmNsubMenuId, PullDown);
    n = 1;
    if (Mnemonic != ' ' && Mnemonic != '\0')
    {
      Mnemonic2 = TOOLGetResChar (Parent, "xmPushButtonWidgetClass", 
				          Title, XmNmnemonic, Mnemonic);
      XtSetArg (args[n], XmNmnemonic, Mnemonic2); n++;
    }
    Cascade = XtCreateManagedWidget (Title, xmCascadeButtonWidgetClass, 
				            Parent, args, n);
    Result = Cascade;
  }
  else
    Result = PullDown;

  return Result;
}


static void PRG_FktTimer (MENU_ENTRY *Entry, XtIntervalId ID)
{
  if (Entry->Data != NULL || Entry->Type == PrgTOGL)
    ((void(*)(void*))Entry->FPtr) (Entry->Data);
  else
    ((void(*)(void))Entry->FPtr) (); 
}


static void PRG_CallFktCB (Widget widget, MENU_ENTRY *Entry,
			                  XmAnyCallbackStruct *Data)
{
  Boolean Value;

  if (Entry == NULL || Entry->FPtr == NULL)
    return;

  if (Entry->Type == PrgTOGL)
  {
    XtSetArg (args[0], XmNset, &Value);
    XtGetValues (widget, args, 1);

    Entry->Data = (void*)(dual)Value;
  }

  /* Clean message in info-area */
  PRGPrintf (" ");

  XtAppAddTimeOut (AppCon, 50, (XtTimerCallbackProc)PRG_FktTimer, Entry);
}


void PRG_SetMenuBar (PROG *Program, MENU_BAR *MenuBar)
{
  if (MenuBar != NULL)
  {
    if (Program->MenuBar != NULL)
      PRO0 ("MenuBar already exists")
    else
    {
      Program->MenuBar = (MENU_BAR*)malloc(sizeof(MENU_BAR));
      *(Program->MenuBar) = *MenuBar;
    }
  }
}


Widget PRGCreateMenuBar (Widget Parent, MENU_BAR *MenuBar)
{
  int         count;
  MENU_COLUMN Column;

  if (MenuBar == NULL)
    return NULL;

  XtSetArg (args[0], XmNtopAttachment,   XmATTACH_FORM);
  XtSetArg (args[1], XmNleftAttachment,  XmATTACH_FORM);
  XtSetArg (args[2], XmNrightAttachment, XmATTACH_FORM);
  MenuBar->widget = XmCreateMenuBar (Parent, "menuBar", args, 3);

  for (count = 0; count < MenuBar->Cols; ++count)
  {
    Column = MenuBar->Column[count];

    PRO1 ("MENU-COL: %s\n", Column.Name)
    MenuBar->Column[count].widget = 
      PRG_CreateMenu (MenuBar->widget, Column.Name, Column.Mnemonic,
		                       Column.Entry, Column.Rows, TRUE);
    /*
       TODO:
       Cascade-Widget speichern, um spter darauf zugreifen zu knnen
       (unmanagen ganzer Menspalten)
    */

    if (Column.Type == PrgHELP_MENU)
    {
      XtSetArg (args[0], XmNmenuHelpWidget, MenuBar->Column[count].widget);
      XtSetValues (MenuBar->widget, args, 1);
    }
  }
  XtManageChild (MenuBar->widget);

  return MenuBar->widget;
}


dual PRG_SetMenuActive (PROG *Program, char *Name, dual active)
{
  int          count = 0, Pos;
  int          count1 = 0, count2 = 0;
  int          len1, len2;
  Widget       widget = NULL;
  MENU_ENTRY  *Entry;
  MENU_COLUMN  Column;

  PRO_START (" ")

  if (Program->MenuBar == NULL)
    PRO_RETURN ("No Menubar defined", FALSE)

  if (Name == NULL)
    PRO_RETURN ("Invalid Name of menu(-entry)", FALSE)

  len1 = strlen (Program->MenuBar->Title);
  len2 = strlen (Name);

  if (Name[0] == '/')
    count2++;

  /* check menu-title */
  while (count1 < len1 && count2 < len2 &&
	 Program->MenuBar->Title[count1++] == Name[count2++])
    ;

  if (strcmp(Name, "/") == 0 || 
      (count1 == len1 && Program->MenuBar->Title[count1] == Name[count2]))
    widget = Program->MenuBar->widget;
  elsif (Name[count2++] == '/')
  {
    Pos = count2;
    while (count < Program->MenuBar->Cols && widget == NULL)
    {
      count2 = Pos;
      Column = Program->MenuBar->Column[count];
      len1 = strlen (Column.Name);
      count1 = 0;
      while (count1 < len1 && count2 < len2 &&
	     Column.Name[count1++] == Name[count2++])
	;

      if (count1 == len1 && Column.Name[count1] == Name[count2])
	widget = Column.widget;
      elsif (Name[count2++] == '/')
	if ((Entry = PRG_SearchFunction(Column.Entry, 
					 Column.Rows, Name+count2)) != NULL)
	  widget = Entry->widget;
	    
      count++;
    }
  }

  if (widget == NULL)
    PRO_RETURN ("Menu-entry not found", FALSE)

  XtSetSensitive (widget, (Boolean)active);
  XFlush (XDisplay);

  PRO_END
  return TRUE;
}


static void UnmapCB (Widget W, XtPointer ClientData, XtPointer CallData)
{
  XtUnmapWidget (W);
} 


Widget PRGCreateMenu (Widget Parent, char *Name, int x, int y, 
		      MENU_ENTRY *Entry, int Number)
{ 
  Dimension px, py;
  Widget    Shell, Menu;

  if (Parent == NULL || Entry == NULL)
    return NULL;

  XtVaGetValues (Parent, XmNx, &px,
		         XmNy, &py,
		         NULL);

  Shell = XtCreatePopupShell (Name, overrideShellWidgetClass,
			      Parent, NULL, 0);

  Menu = PRG_CreateMenu (Shell, "menu", ' ', Entry, Number, FALSE);
  XtVaSetValues (Menu, XmNx, px + x,
		       XmNy, py + y,
		       NULL);

  XtAddCallback (Menu, XmNunmapCallback, UnmapCB, NULL);

  XtManageChild (Menu);

#if 0
  XtSetMappedWhenManaged (Shell, False);
  XtManageChild (Shell);
  XtMapWidget (Shell);
  XtRealizeWidget (Shell);
#endif

  return Menu;
}
