#include <Xm/Form.h>
#include <Xm/PushB.h>
#include <Xm/MwmUtil.h>
#include <X11/xpm.h>

#include "prgGlobal.h"


static int Dist = 0;
static Arg args[MAX_ARGS];


static void PRG_ButtonCB  (Widget widget, ICONDATA *Entry,
			                           XmAnyCallbackStruct *Data);


static Widget PRG_CreateButton (Widget Parent, ICONDATA *Entry)
{
  char    FileName[PATH_MAX+1];
  char   *Text = NULL;
  Pixel   Color;
  Pixmap  PixMap;
  Widget  Result = NULL;

  PRO_START ("\n")

  Result = XtCreateWidget ("button", xmPushButtonWidgetClass, Parent, NULL, 0);
  XtAddCallback (Result, XmNarmCallback, (XtCallbackProc)PRG_ButtonCB, Entry);

  XtVaGetValues (Result, XmNbackground, &Color, NULL);

  if (Entry->PicName[0] != '\0')
  {
    Text = TOOLGetResString (Parent, "xmPushButtonWidgetClass", "pic",
			             "fileName", Entry->PicName);
    TOOLParseFileName (Text, FileName, PATH_MAX);
    if ((PixMap = 
	   TOOLLoadXPM(FileName, NULL, NULL, Color)) == XmUNSPECIFIED_PIXMAP)
      PRO_RETURN ("Invalid pixmap-file", NULL)
  }
  elsif ((PixMap = TOOLMakeXPM((char**)Entry->Data, 
			       NULL, NULL, Color))==XmUNSPECIFIED_PIXMAP)
    PRO_RETURN ("Invalid pixmap-data", NULL)

  XtSetArg (args[0], XmNlabelType,              XmPIXMAP);
  XtSetArg (args[1], XmNlabelPixmap,            PixMap);
  XtSetArg (args[2], XmNlabelInsensitivePixmap, PixMap);
  XtSetArg (args[3], XmNpushButtonEnabled,      TRUE);
  XtSetArg (args[4], XmNshadowType,             XmSHADOW_ETCHED_IN);
  XtSetValues (Result, args, 5);

  XtManageChild (Result);

  PRO_END
  return Result;
}


static Widget PRG_CreateToolBar (Widget Parent, Widget Descr, 
				                TOOLDATA *ToolBar)
{
  int       count = 0;
  ICONDATA *Entry = NULL;
  Widget    Button, last;
  Widget    Result;
  Arg       args[10];

  PRO_START ("\n")

  if (ToolBar == NULL || ToolBar->Pos == PrgNO_TOOL)
    PRO_RETURN ("Invalid ToolBar-Data", NULL)

  if (ToolBar->Pos == PrgWIN_TOOL)
  {
    XtSetArg (args[0], XmNdeleteResponse, XmDESTROY);
    XtSetArg (args[1], XmNtransient,      False);
    XtSetArg (args[2], XmNmwmDecorations,   
	               MWM_DECOR_TITLE | MWM_DECOR_BORDER | MWM_DECOR_RESIZEH);
    XtSetArg (args[3], XmNmwmFunctions,     
	               MWM_FUNC_RESIZE | MWM_FUNC_MOVE | MWM_FUNC_CLOSE);
    Parent = XtCreatePopupShell ("toolBar", transientShellWidgetClass,
				            AppShell, args, 4);
    /* TODO:
       Vergrerung verbieten ?
    */
    ToolBar->Shell = Parent;
  }
  ToolBar->widget = XtCreateManagedWidget ("toolBar", xmFormWidgetClass, 
					              Parent, NULL, 0);
  Result = last = ToolBar->widget;

  if (ToolBar->Pos == PrgTOP_TOOL)
  {
    XtSetArg (args[0], XmNtopAttachment,   XmATTACH_FORM);
    XtSetArg (args[1], XmNbottomAttachment,XmATTACH_FORM);
    XtSetArg (args[2], XmNleftAttachment,  XmATTACH_WIDGET);
  }
  else
  {
    XtSetArg (args[0], XmNleftAttachment,  XmATTACH_FORM);
    XtSetArg (args[1], XmNrightAttachment, XmATTACH_FORM);
    XtSetArg (args[2], XmNtopAttachment,   XmATTACH_WIDGET);
  }

  Entry = ToolBar->Entry;
  while (Entry != NULL)
  {
    Button = PRG_CreateButton (Result, Entry);

    if (ToolBar->Pos == PrgTOP_TOOL)
    {
      XtSetArg (args[3], XmNleftWidget, last);
      XtSetArg (args[4], XmNleftOffset, Entry->Dist);
    }
    else
    {
      XtSetArg (args[3], XmNtopWidget, last);
      XtSetArg (args[4], XmNtopOffset, Entry->Dist);
    }
    XtSetValues (Button, args, 5);
    last = Button;

    if (Entry->Descr[0] != '\0')
    {
#if 0
      Entry->Display = Descr;
      PRG_InstallDescrText (Button, Entry, ToolBar->Delay);
#else
      PRG_InstallDescrWin (Button, Entry, ToolBar->Delay);
#endif
    }
    Entry = Entry->Next;
    count++;
  }

  PRO_END
  return Result;
}


Widget PRG_MakeToolBar (PROG *Program)
{
  int    n;
  Widget Tools, Descr = NULL;

  if (Program->Info != NULL)
    Descr = Program->Info->Output;

  Tools = PRG_CreateToolBar (Program->Main, Descr, Program->ToolBar);

  if (Program->ToolBar)
  {
    n = 0;
    if (Program->MenuBar != NULL && Program->ToolBar->Pos != PrgWIN_TOOL)
    {
      XtSetArg (args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
      XtSetArg (args[n], XmNtopWidget,     Program->MenuBar->widget); n++;
    }

    if (Program->Info != NULL && Program->ToolBar->Pos == PrgLEFT_TOOL)
    {
      XtSetArg (args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
      XtSetArg (args[n], XmNbottomWidget,     Program->Info->Output); n++;
    }
    XtSetValues (Tools, args, n);
  }

  return Tools;
}


static void PRG_ButtonCB (Widget widget, ICONDATA *Entry,
			                 XmAnyCallbackStruct *Data)
{
  if (Entry == NULL || Entry->FPtr == NULL)
    return;

  PRG_RemoveDescrWin ();

  /* Clean message in info-area */
  PRGPrintf (" ");

  if (Entry->Param == NULL)
    ((void(*)(void))Entry->FPtr) ();
  else
    ((void(*)(void*))Entry->FPtr) (Entry->Param);
}


void PRGMakeToolBar (int Position)
{
  PRO_START ("")

  if (Program == NULL)
    PRO_ABORT ("No Program given")

  if (Program->ToolBar != NULL)
    PRO_ABORT ("Toolbar already installed")

  Program->ToolBar = (TOOLDATA*)calloc (1, sizeof(TOOLDATA));
  Program->ToolBar->Pos    = Position;
  Program->ToolBar->Number = 0;
  Program->ToolBar->Delay  = DELAY_TIME;
  Program->ToolBar->Entry  = NULL;
  Program->ToolBar->Last   = NULL;
  Program->ToolBar->active = TRUE;
  Program->ToolBar->widget = NULL;
  Program->ToolBar->Shell  = NULL;

  PRO_END
}


dual PRGSetToolBarActive (dual active)
{
  ICONDATA *Entry;

  if (Program == NULL || Program->ToolBar == NULL)
    return FALSE;

  Program->ToolBar->active = active;
  Entry = Program->ToolBar->Entry;
  while (Entry != NULL)
  {
    Entry->active = active;
    Entry = Entry->Next;
  }

  if (Program->ToolBar->widget != NULL)
    XtSetSensitive (Program->ToolBar->widget, (Boolean)active);
  
  return TRUE;
}


void PRGShowToolBar (dual visible)
{
  if (Program != NULL && 
      Program->ToolBar != NULL && Program->ToolBar->widget != NULL)
  {
    if (visible)
    {
      XtManageChild (Program->ToolBar->widget);
      XtMapWidget   (Program->ToolBar->widget);
    }
    else
    {
      XtUnmanageChild (Program->ToolBar->widget);
      XtUnmapWidget   (Program->ToolBar->widget);
    }

    XtRealizeWidget (Program->Main);
  }
}


/* ************************************************* */
/* ** Inserts an icon into the toolbar. Places it ** */
/* ** on the right-hand/bottom-side of the last   ** */
/* ** one.                                        ** */
/* ** INPUT:  SizeX                               ** */
/* **         SizeY        Image size xpm-file    ** */
/* **                      0: filename, -1: data  ** */
/* **         Data         image data (see above) ** */
/* **         Function     user-function          ** */
/* **         Param        function-parameter     ** */
/* **         Description  popup-helptext         ** */
/* ** OUTPUT: ---                                 ** */
/* ** EXAMPLE: ...                                ** */
/* **  PRGAddToolIcon (0, 0, "$HOME/picture.xpm", ** */
/* **                  MyFunction, NULL, "foo");  ** */
/* ************************************************* */
void PRGAddToolIcon (int SizeX, int SizeY, void *Data,
		     MENU_FUNCTION Function, void *Param, char *Description)
{
  ICONDATA *Entry = NULL;

  PRO_START ("")

  if (Program == NULL || Program->ToolBar == NULL)
    PRO_ABORT ("Bad Program or NO Toolbar found")

  Entry = (ICONDATA*)calloc (sizeof(ICONDATA), 1);

  Entry->SizeX  = SizeX;
  Entry->SizeY  = SizeY;
  Entry->Dist   = Dist;
  Entry->FPtr   = Function;
  Entry->Param  = Param;
  Entry->active = Program->ToolBar->active;
  if (Description != NULL)
  {
    strncpy (Entry->Descr, Description, MAX_DESCRLEN);
    Entry->Descr[MAX_DESCRLEN] = '\0';
  }
  else
    Entry->Descr[0] = '\0';

  if (SizeX == 0 || SizeY == 0)
  {
    strncpy (Entry->PicName, (char *)Data, PATH_MAX);
    Entry->PicName[PATH_MAX] = '\0';
    Entry->Data  = NULL;
  }
  else
  {
    Entry->Data = (void *)TOOLCopyXPM ((char **)Data);
    Entry->PicName[0] = '\0';
  }

  Entry->Display = NULL;
  Entry->Timer   = 0;
  Entry->Next    = NULL;

  if (Program->ToolBar->Entry == NULL)
  {
    Program->ToolBar->Entry = Entry;
    Program->ToolBar->Last = Entry;
  }
  else
  {
    Program->ToolBar->Last->Next = Entry;
    Program->ToolBar->Last = Entry;
  }
  Dist = 0;

  (Program->ToolBar->Number)++;

  PRO_END
}


/* ************************************************ */
/* ** Set the time between entering a toolbar-   ** */
/* ** icon and displaying the description-text   ** */
/* ** INPUT:  Delay  delay-time in msec          ** */
/* **                default: 500 (DELAY_TIME)   ** */
/* ** OUTPUT: ---                                ** */
/* ************************************************ */
void PRGSetDelayTime (int Delay)
{
  PRO_START ("")

  if (Program == NULL || Program->ToolBar == NULL)
    PRO_ABORT ("Bad program or undefined toolbar")

  Program->ToolBar->Delay = Delay;

  PRO_END
}


/* ******************************************* */
/* ** Insert space before next toolbar-icon ** */
/* ** INPUT:  Space  size of space [pixel]  ** */
/* ** OUTPUT: ---                           ** */
/* ** EXAMPLE: PRGAddToolSpace (15);        ** */
/* ******************************************* */
void PRGAddToolSpace (int Space)
{
  PRO_START ("")

  if (Program == NULL || Program->ToolBar == NULL)
    PRO_ABORT ("Bad program or undefined toolbar")

  Dist = Space;

  PRO_END
}


void PRG_FreeToolBar (TOOLDATA *ToolBar)
{
  ICONDATA *Entry = NULL;
  ICONDATA *Next;

  if (ToolBar == NULL)
    PRO_ABORT ("Nothing to do")

  Entry = ToolBar->Entry;
  while (Entry != NULL)
  {
    Next = Entry->Next;
    if (Entry->Data != NULL)
      free (Entry->Data);
    free (Entry);
    Entry = Next;
  }

  free (ToolBar);
}
