#include <Xm/Label.h>

#include "prgGlobal.h"


typedef struct
{
  Widget        Button;
  char         *Text;
  int           Delay;
  Widget        Shell;
  XtIntervalId  Timer;
  Window        Win;
} DESCR_DATA;


typedef struct
{
  MENU_FUNCTION  Function;
  void          *Param;
  XtIntervalId   Timer;
  int            Delay;
} ANIM_DATA;

static Arg        args[MAX_ARGS];
static DESCR_DATA Descr;
static ANIM_DATA  Anim[5] = {{NULL, NULL, (XtIntervalId)-1, 1000},
					    {NULL, NULL, (XtIntervalId)-1, 1000},
					    {NULL, NULL, (XtIntervalId)-1, 1000},
					    {NULL, NULL, (XtIntervalId)-1, 1000},
					    {NULL, NULL, (XtIntervalId)-1, 1000}};


static void PRG_ShowDescr (ICONDATA *Data, XtIntervalId Timer)
{
  XmString Label;

  PRO1 ("%s\n", Data->Descr)

  Data->Timer = 0;

  Label = XmStringLtoRCreate (Data->Descr, XmSTRING_DEFAULT_CHARSET);
  XtSetArg (args[0], XmNlabelString,  Label);
  XtSetValues (Data->Display, args, 1);
  if (Data->Descr[0] != '\0')
    XmStringFree (Label);
}

static void PRG_DescrTextHandler (Widget widget, ICONDATA *Data, XEvent *event)
{
  XmString Label;

  if (widget == NULL)
    return;

  PRO_START ("")

  switch (event->type)
  {
    case EnterNotify :
      if (Data->Timer == 0 && Data->active)
	Data->Timer = XtAppAddTimeOut (AppCon, Descr.Delay, 
			          (XtTimerCallbackProc)PRG_ShowDescr, Data);
      break;

    case LeaveNotify :
      if (Data->Timer != 0)   /* icon was left in before timeout reached */
      {
	   XtRemoveTimeOut (Data->Timer);
	   Data->Timer = 0;
      }
      Label = XmStringLtoRCreate (" ", XmSTRING_DEFAULT_CHARSET);
      XtSetArg (args[0], XmNlabelString,  Label);
      XtSetValues (Data->Display, args, 1);
      XmStringFree (Label);
      break;

    default : 
      break;
  }

  PRO_END
}

void PRG_InstallDescrText (Widget widget, ICONDATA *Entry, int Delay)
{
  Descr.Delay = Delay;

  XtAddEventHandler (widget, EnterWindowMask | LeaveWindowMask, False,
		             (XtEventHandler)PRG_DescrTextHandler, Entry);
}


/* ----------------------------------------------------------- */

static void PRG_OpenDescr (char *Text, XtIntervalId Timer)
{
  int      MouseX, MouseY;
  XmString Label;

  Descr.Timer = 0;

  if (TOOLGetMousePos(&MouseX, &MouseY) != Descr.Win)
    return;

  PRO1 ("DESCRIPTION: %s\n", Text)

  Descr.Shell = XtVaCreatePopupShell ("descr", overrideShellWidgetClass,
				      Descr.Button,
				      XmNx, MouseX+10,
				      XmNy, MouseY+10,
				      NULL);

  Label = XmStringLtoRCreate (Text, XmSTRING_DEFAULT_CHARSET);
  XtVaCreateManagedWidget ("descrLabel", xmLabelWidgetClass, Descr.Shell,
			                 XmNlabelString, Label,
			                 XmNalignment,   XmALIGNMENT_BEGINNING,
			                 NULL);
			   
  if (Text[0] != '\0')
    XmStringFree (Label);

  XtRealizeWidget (Descr.Shell);
  XtPopup (Descr.Shell, XtGrabNone);
}


void PRG_RemoveDescrWin (void)
{
  if (Descr.Timer != 0)
  {
    XtRemoveTimeOut (Descr.Timer);
    Descr.Timer = 0;
  }

  if (Descr.Shell != NULL)
  {
    XtUnmapWidget (Descr.Shell);
    XtDestroyWidget (Descr.Shell);
    Descr.Shell = NULL;
  }
}


static void PRG_DescrWinHandler (Widget widget, ICONDATA *Entry, XEvent *event)
{
  PRO_START ("")

  switch (event->type)
  {
    case EnterNotify :
      PRO0 ("\\ENTERING...")
      if (Descr.Timer == 0)
      {
	Descr.Win = TOOLGetMousePos (NULL, NULL);
	Descr.Button = widget;
	Descr.Shell  = NULL;
	if (Entry->active)
	  Descr.Timer = 
	    XtAppAddTimeOut (AppCon, Descr.Delay, 
			     (XtTimerCallbackProc)PRG_OpenDescr, Entry->Descr);
      }
      break;

    case LeaveNotify :
      PRO0 ("\\LEAVING...")
      PRG_RemoveDescrWin ();
      break;

    case FocusIn :
      PRO0 ("\\FOCUS IN...")
      break;

    case FocusOut :
      PRO0 ("\\FOCUS OUT...")
      PRG_RemoveDescrWin ();
      break;

    default : 
      PRO1 ("\\UNKNOWN EVENT: %d...", event->type)
      break;
  }

  PRO_END
}


void PRG_InstallDescrWin (Widget widget, ICONDATA *Entry, int Delay)
{
  Descr.Delay = Delay;

  if (Entry != NULL && Entry->Descr[0] != '\0')
    XtAddRawEventHandler (widget, EnterWindowMask | LeaveWindowMask |
			          FocusChangeMask, False, 
			          (XtEventHandler)PRG_DescrWinHandler, Entry);
}


static void PRG_AnimTimer (ANIM_DATA *Data, XtIntervalId Timer)
{
  if (Data->Timer == (XtIntervalId)-1)
    return;

#ifndef __CPP__
    Data->Function (Data->Param);
#else
    Data->Function ();
#endif

  if (Data->Timer == (XtIntervalId)-1)
    return;

  Data->Timer = XtAppAddTimeOut (AppCon, Data->Delay, 
				 (XtTimerCallbackProc)PRG_AnimTimer, Data);
}

int PRGStartAnim (MENU_FUNCTION Function, void *Param, int Delay)
{
  int count = 0;

  while (count < 5 && Anim[count].Function != NULL)
    count++;

  if (count == 5)
    return -1;

  Anim[count].Function = Function;
  Anim[count].Param    = Param;
  Anim[count].Timer    = 0;
  Anim[count].Delay    = Delay;

  PRG_AnimTimer (&Anim[count], (XtIntervalId)-1);

  return count;
}


dual PRGStopAnim (int ID)
{
  if (ID < 0 || ID >= 5 || Anim[ID].Timer == (XtIntervalId)-1)
    return FALSE;

  XtRemoveTimeOut (Anim[ID].Timer);
  Anim[ID].Function = NULL;
  Anim[ID].Param    = 0;
  Anim[ID].Timer    = (XtIntervalId)-1;
  Anim[ID].Delay    = 1000;

  return TRUE;
}
