/*-----------------------------------------------------------------------
  Author    : Stefan Vietinghoff
  FileName  : tstDlg.c
  Date      : November 1996
  Function  : demonstration of dialog-library
  -----------------------------------------------------------------------*/
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>

#include <Xm/Form.h>
#include <Xm/Label.h>          /* Needed only for user-widget */

#include <libdlg.h>
#include <libprg.h>
#include <libhelp.h>
#include <Array.h>

/* Icons */
#include <icons/penguin.xbm>
#include <icons/look.xpm>
#include <icons/new.xpm>
#include <icons/stefan.xpm>
#include <icons/ball_red.xpm>
#include <icons/ball_green.xpm>
#include <icons/ball_blue.xpm>
#include <icons/ball_yellow.xpm>
#include <icons/plus.xpm>
#include <icons/minus.xpm>

#define elsif else if
#define MIN(x, y) (((x) < (y))? (x) : (y))

int Power[] = {1, 2, 4, 8, 16, 32, 64, 128};

#define DLGCOLORS 5

typedef struct
{
  int BoxID;
  int r[DLGCOLORS];
  int g[DLGCOLORS];
  int b[DLGCOLORS];
} DLGDATA;

typedef struct
{
  char *PrgName;
} PRG_PARAM;

/* Statische Funktionen: */
static void ProgramExit ();

/* Prototypen: */
int SpaceFkt   (DIALOG *Dlg, int BoxID);
int StartProc  (DIALOG *Dlg, int BoxID);
int DisplayPos (DIALOG *Dlg, int BoxID);


 
/*-----------------------------------------------------------------------
                    D E F I N E S   und   M A K R O S
  -----------------------------------------------------------------------*/
/* Es scheint uebersichtlicher, anstelle der Indizes 0, 1, ..., die      */
/* folgenden Konsanten zu verwenden                                      */
#define WIN_OUT 0				  /* Textfenster fuer    */
						  /* allgemeine  Ausgaben*/
#define PIC_OUT -1				  /* Graphicfenster mit  */
						  /* geladenem Bild 	 */
#define COL_OUT 2				  /* Fenster fuer 	 */
						  /* Malfunktionen  	 */


static PRG_PARAM Param = {NULL};
double ColorTable[][3] = 
{
  {1.00, 1.00, 1.00},      /* Value          */
  {0.75, 0.75, 0.75},      /* NULL           */
  {1.00, 1.00, 1.00},      /* White          */
  {0.70, 0.70, 0.70},      /* Grey           */
  {0.00, 0.00, 0.00},      /* Black          */
  {1.00, 0.00, 0.00},      /* Red            */
  {0.00, 0.80, 0.00},      /* Green          */
  {0.00, 0.00, 1.00},      /* Blue           */
  {1.00, 1.00, 0.00},      /* Yellow         */
  {0.00, 1.00, 1.00},      /* Cyan           */
  {1.00, 0.00, 1.00},      /* Purple         */
  {1.00, 0.60, 0.30},      /* Orange         */
  {1.00, 0.00, 1.00},      /* Magenta        */
  {1.00, 0.70, 1.00},      /* Cherry         */
  {0.00, 0.75, 1.00},      /* DeepSkyBlue    */
  {0.55, 0.27, 0.16},      /* (Saddle)Brown  */
  {0.20, 0.80, 0.20},      /* Lime (Green)   */
  {0.80, 0.70, 0.60},      /* Tan            */ 
  {1.00, 0.84, 0.00},      /* Gold           */
  {0.27, 0.51, 0.70},      /* Steel Blue     */ 
  {0.00, 0.52, 0.54},      /* Turquoise4     */
  {0.63, 0.32, 0.18},      /* Sienna (Brown) */
  {1.00, 0.85, 0.72},      /* PeachPuff      */
  {0.80, 0.52, 0.25},      /* Peru           */
  {1.00, 0.98, 0.80},      /* LemonChiffon   */
  {0.73, 0.33, 0.82},      /* MediumOrchid   */
  {0.54, 0.53, 0.47},      /* CornSilk4      */
  {0.11, 0.52, 0.93},      /* DodgerBlue2    */
  {1.00, 0.08, 0.57},      /* DeepPink       */
  {0.26, 0.43, 0.93},      /* RoyalBlue2     */
  {0.26, 0.80, 0.50},      /* SeaGreen3      */
  {0.93, 0.46, 0.00},      /* DarkOrange2    */
  {0.00, 0.80, 0.40},      /* SpringGreen3   */
  {1.00, 0.19, 0.19}       /* FireBrick1     */
};

#define MAX_COLORS (sizeof(ColorTable) / (sizeof(double)*3))

static char *ColorNames[] = 
{
  "VALUES",
  NULL,
  "WHITE",
  "GREY",
  "BLACK",
  "RED",
  "GREEN",
  "BLUE",
  "YELLOW",
  "CYAN",
  "PURPLE",
  "ORANGE",
  "MAGENTA",
  "CHERRY",
  "LT BLUE",
  "BROWN",
  "LIME",
  "TAN",
  "GOLD",
  "STEEL BLUE",
  "TURQUOISE",
  "SIENNA",
  "PEACHPUFF",
  "PERU",
  "LEMONCHIFFON",
  "MEDORCHID",
  "CORNSILK",
  "DODGERBLUE",
  "DEEPPINK",
  "ROYALBLUE",
  "SEAGREEN",
  "DARKORANGE",
  "SPRINGGREEN",
  "FIREBRICK"
};



/* ********************************* */
/* *** allgemeine Hilfsprogramme *** */
/* ********************************* */

static void SetBit (char ** Data, int SizeX, int PosX, int PosY)
{
  int Offset, HBytes;

  HBytes = (int)(SizeX/8);
  Offset = PosY*HBytes + (PosX/8);
  /* TODO:
     Power ersetzen durch 1 << (PosX % 8);
  */
  (*Data)[Offset] |= Power[PosX % 8];
}


static char *strcp (char *String, ...)
{
  char    *Result, *Work;
  va_list  args;

  if (String == NULL)
    return NULL;
  
  va_start (args, String);
  Work = malloc (sizeof(char)*(strlen(String)+255+1));
  vsprintf (Work, String, args);
  va_end (args);
  
  Result = calloc (strlen(Work)+1, sizeof(char));
  strcpy (Result, Work);
  free (Work);
  
  return Result;
}


static void InitPrg (void)
{
  char *Msg;

  if (Param.PrgName != NULL)
  {
    PRGPrintf ("Running script...");
    if ((Msg = PRGRunScript(Param.PrgName)) != NULL)
      DLGOpenTextDialog ("scriptError", DlgERROR, Msg, 1, "OK");
    PRGPrintf ("\\ready !");
  }
}


/*----------------------------------------------------------------------
  Funktion: Funktion wird von ExitError aufgerufen !!

  allg. Syntax : VOID MyExit (VOID)

  Datum der letzten Aenderung: 08.05.92
  ----------------------------------------------------------------------*/
static void ProgramExit ()
{
  if (DLGOpenTextDialog("exitProgram", DlgQUESTION, 
			               "Really want to exit the Programm ?", 
                                       2, "YES", "NO") == 1)
  {
    PRGExitProgram ();
    exit (0);
  }
}


/*----------------------------------------------------------------------
  Funktion: Infotext in eigenem Fenster ausgeben

  allg. Syntax : VOID ProgramInfo (VOID)

  Datum der letzten Aenderung: 16.05.92
  ----------------------------------------------------------------------*/
static void ProgramInfo (void)
{
  DLGSetDialogPos (0.4, 0.45);
  DLGOpenPicDialog ("INFO", DlgXPM, -1, -1, stefan_xpm, 
			    "           TSTDLG V1.1\n"
			    "\n"
			    "      Test-program for the\n"
			    "         dialog-library"
			    "\n"
			    "\n"
			    "   Copyright: Stefan Vietinghoff\n"
			    "   Irisweg 34, D-90427 Nrnberg\n"
		            "eMail: Stefan.Vietinghoff@epost.de", 0);
}


void Dialog0 (void)
{
  PRGPrintf ("This dialog can be closed by clicking on it with the left\n"
	     "mouse button\n");
  DLGSetDialogPos (0.5, 0.5);
  DLGOpenTextDialog ("Dialog0", DlgINFO, "Welcome to DlgLib", 0);
  PRGPrintf ("Dialog finished\n");
  DLGSetDialogPos (-1, -1);
}

void Dialog1 (void)
{
  PRGPrintf ("Button %d pressed\n", 
	     DLGOpenTextDialog ("Dialog1", DlgERROR, "Hey Bill", 1, "OK"));
}

void Dialog2 (void)
{
  PRGPrintf ("Button %d pressed\n", 
	     DLGOpenTextDialog ("Dialog2", DlgINFO, 
				"Allgemeine Schutzverletzungen", 
				2, "ARD", "Winzigweich"));
}

void Dialog3 (void)
{
  PRGPrintf ("Button %d pressed\n",
	     DLGOpenTextDialog ("Dialog3", DlgMESSAGE, 
				"Why do you want to crash\n"
				"your system today ?", 
				1, "Take Linux"));
}

void Dialog4 (void)
{
  PRGPrintf ("Button %d pressed\n",
	     DLGOpenTextDialog("Dialog4", DlgQUESTION, 
			       "Test: Wie spt ist es ?", 
			       3, "5 Knochen nach\nSteinzeit",
			       "2 Lffel nach Nesquick -\nZeit zum umrhren",
			       "Kurz vor gleich"));
}

void Dialog5 (void)
{
  PRGPrintf ("Button %d pressed\n", 
	     DLGOpenTextDialog("Dialog5", DlgWARNING, 
			                  "Attention: I'm formatting the HDD", 
			                  2, "No problem", "I'll help you"));
}

void Dialog6 (void)
{
  PRGPrintf ("Button %d pressed\n", 
	     DLGOpenTextDialog("Dialog6", DlgWORKING, 
			       "Test: keine Lust mehr\nich hr fr heute auf",
			       1, "Tsch"));
}

void Dialog7 (void)
{
  DLGSetDialogPos (0.5, 0.5);
  PRGPrintf ("To Close this dialog: click on it !\n");
  DLGOpenPicDialog ("Dialog7", DlgXPM, -1, -1, new_xpm, NULL, 0);
  PRGPrintf ("You got it !\n");
}


void Dialog8 (void)
{
  PRGPrintf ("Button %d pressed\n",
	     DLGOpenPicDialog("Dialog7", DlgXBM, 0, 0, "/pic/icons/msdos.xbm",
			                 NULL, 1, "Ganz meiner Meinung"));
}

void Dialog9 (void)
{
  PRGPrintf ("Button %d pressed\n",
	     DLGOpenPicDialog("Dialog8", DlgXBM,
			      (int)penguin_width, (int)penguin_height, 
			      penguin_bits, "take linux", 
			      2, "yeaahhh", "???"));
}

void Dialog10 (void)
{
  PRGPrintf ("Button %d pressed\n", 
	     DLGOpenPicDialog("Dialog9", DlgANY,
			      20, 20, NULL, NULL, 2, "na klar", "???"));
}


/* *************** */
/* *** Dialog: *** */
/* *************** */

int SpaceFkt (DIALOG *Dlg, int BoxID)
{
  /* dummy-function: do nothing... */
  return DlgCONFIRM;
}


void PtrSplit (void)
{
  int        count, Number;
  int       *ColList = NULL;
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGInsertUpDn (Dlg, 0, "count", 1.0, 10.0, 2.0, 1.0);

  DLGOpenDialog (Dlg, "SPLITASK", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
  {
    Number = (int) *(double*)DLGGetValue (Dlg, 0);
    DLGFreeDialog (Dlg);

    Dlg = DLGNewDialog (0);

    ColList = (int*)calloc (Number, sizeof(int));
    for (count = 0; count < Number; ++count)
      ColList[count] = 10+count;

    DLGMakeCols (Dlg, 0, -Number, DlgNO_LINE, ColList);

    for (count = 0; count < Number; ++count)
      DLGInsertLabel (Dlg, 10+count, "colum", "Column %d", count);
    
    DLGOpenDialog (Dlg, "PTRSPLIT", DlgNORMAL, NULL);

    free (ColList);
  }
  DLGFreeDialog (Dlg);
}


/* ********************************************************************* */
/* *** zeigt wie Boxen und Menues mit einer Zeigerliste erzeugt      *** */
/* *** werden                                                        *** */
/* ********************************************************************* */
void PtrDlg (void)
{
  int         count;
  int         Result;
  dual       *TResult = NULL;
  char      **Menu = NULL;
  char      **RBox = NULL;
  TOGGLE     *TBox = NULL;
  DIALOG  *Dlg = NULL;

  PRGPrintf ("The Menu and the Boxes in this dialog are created as\n"
	     "a allocated pointer-structure\n");

  Dlg = DLGNewDialog (0);

  DLGMakeCols (Dlg,  0, 2, DlgNO_LINE, 10, 20);
  DLGMakeRows (Dlg, 20, 2, DlgNO_LINE, 21, 22);

  TBox = (TOGGLE*) calloc (6, sizeof(TOGGLE));
  strncpy (TBox[0].Item, "toggle1", DLG_LLEN);   TBox[0].Value = TRUE;
  strncpy (TBox[1].Item, "toggle2", DLG_LLEN);   TBox[1].Value = FALSE;
  strncpy (TBox[2].Item, "toggle3", DLG_LLEN);   TBox[2].Value = FALSE;
  strncpy (TBox[3].Item, "toggle4", DLG_LLEN);   TBox[3].Value = TRUE;
  strncpy (TBox[4].Item, "toggle5", DLG_LLEN);   TBox[4].Value = FALSE;
  strncpy (TBox[5].Item, "toggle6", DLG_LLEN);   TBox[5].Value = TRUE;
  DLGInsertTBox (Dlg, 10, "prttbox", 2, -6, TBox);

  Menu = (char**)calloc (7, sizeof(char*));
  Menu[0] = strcp ("option1");
  Menu[1] = strcp ("option2");
  Menu[2] = strcp ("option3");
  Menu[3] = strcp ("option4");
  Menu[4] = NULL;
  Menu[5] = strcp ("option5");
  Menu[6] = strcp ("option6");
  DLGInsertOMenu (Dlg, 21, "prtMenu", -7, Menu, 1);

  RBox = (char**)calloc (3, sizeof (char*));
  RBox[0] = strcp ("Radio1");
  RBox[1] = strcp ("Radio2");
  RBox[2] = strcp ("Radio3");
  DLGInsertRBox (Dlg, 22, "prtBox", 2, -3, RBox, 1);

  DLGOpenDialog (Dlg, "PTRDLG", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
  {
    free (TBox);
    TResult = *(dual**) DLGGetValue (Dlg, 10);

    for (count = 0; count < 6; ++count)
      PRGPrintf ("Toggle %d is %s\n", count, TResult[count]? "TRUE" : "FALSE");

    Result = *(int*)DLGGetValue (Dlg, 21);
    PRGPrintf ("Menu selection is %s\n", Menu[Result]);
    for (count = 0; count < 7; ++count)
      if (Menu[count] != NULL)
        free (Menu[count]);
    free (Menu);

    Result = *(int*)DLGGetValue (Dlg, 22);
    PRGPrintf ("Radio selection is %s\n", RBox[Result]);
  }    
  for (count = 0; count < 3; ++count)
    free (RBox[count]);
  free (RBox);

  DLGFreeDialog (Dlg);
}



/* ********************************************************************* */
/* *** Aufruf einer Funktion mit bergabe von Parametern mittels     *** */
/* *** DLGSetFunction                                                *** */
/* ********************************************************************* */
int ParamFkt (DIALOG *Dlg, int BoxID)
{
  PRGPrintf ("Got actual value: %f\n", *(double*)DLGGetValue(Dlg, BoxID));

  return DlgCONFIRM;
}

void ParamCall (void)
{
  DIALOG *Dlg = NULL;

  PRGPrintf ("Calling a function with the actual dialog-parameters\n"
	     "using the DLGSetFunction function\n");

  Dlg = DLGNewDialog (0);

  DLGInsertUpDn (Dlg, 0, "paramID", 0.0, 12.0, 3.5, 0.25);

  DLGSetFunction (Dlg, 0, ParamFkt);

  DLGOpenDialog (Dlg, "PARAMCALL", DlgNORMAL, NULL);

  DLGFreeDialog (Dlg);
}



/* ********************************************************************* */
/* *** Darstellung der (z.Zt.) untersttzten Label-Positionen        *** */
/* ********************************************************************* */
void LabelPos (void)
{
  DIALOG *Dlg = NULL;


  PRGPrintf ("This is a example of the implemented label-positions\n");

  Dlg = DLGNewDialog (10);

  DLGMakeGrid (Dlg, 10, 3, 3);
  
  DLGInsertText (Dlg, DlgGRID(10,0,0), "northWest", 40, "NORTHWEST", 10, 1);
  DLGInsertText (Dlg, DlgGRID(10,1,0), "north",     30, "NORTH",     10, 1);
  DLGInsertText (Dlg, DlgGRID(10,2,0), "northEast", 30, "NORTHEAST", 10, 1);
  DLGInsertText (Dlg, DlgGRID(10,0,1), "west",      20, "WEST",      10, 1);
  DLGInsertText (Dlg, DlgGRID(10,1,1), "nowhere",   10, "NOWHERE",   10, 1);
  DLGInsertText (Dlg, DlgGRID(10,2,1), "east",      10, "EAST",      10, 1);
  DLGInsertText (Dlg, DlgGRID(10,0,2), "southWest", 10, "SOUTHWEST", 10, 1);
  DLGInsertText (Dlg, DlgGRID(10,1,2), "south",     10, "SOUTH",     10, 1);
  DLGInsertText (Dlg, DlgGRID(10,2,2), "southEast", 10, "SOUTHEAST", 10, 1);

  DLGSetLabelPos (Dlg, DlgGRID(10,0,0), DlgNORTH_WEST,   DlgWEST);
  DLGSetLabelPos (Dlg, DlgGRID(10,1,0),      DlgNORTH, DlgCENTER);
  DLGSetLabelPos (Dlg, DlgGRID(10,2,0), DlgNORTH_EAST, DlgCENTER);
  DLGSetLabelPos (Dlg, DlgGRID(10,0,1),       DlgWEST,   DlgEAST);
  DLGSetLabelPos (Dlg, DlgGRID(10,1,1),    DlgNOWHERE, DlgCENTER);
  DLGSetLabelPos (Dlg, DlgGRID(10,2,1),       DlgEAST, DlgCENTER);
  DLGSetLabelPos (Dlg, DlgGRID(10,0,2), DlgSOUTH_WEST, DlgCENTER);
  DLGSetLabelPos (Dlg, DlgGRID(10,1,2),      DlgSOUTH, DlgCENTER);
  DLGSetLabelPos (Dlg, DlgGRID(10,2,2), DlgSOUTH_EAST, DlgCENTER);  

  DLGSetBorder (Dlg, DlgGRID(10,0,0), DlgETCHED_IN);
  DLGSetBorder (Dlg, DlgGRID(10,1,0), DlgETCHED_IN);
  DLGSetBorder (Dlg, DlgGRID(10,2,0), DlgETCHED_IN);
  DLGSetBorder (Dlg, DlgGRID(10,0,1), DlgETCHED_IN);
  DLGSetBorder (Dlg, DlgGRID(10,1,1), DlgETCHED_IN);
  DLGSetBorder (Dlg, DlgGRID(10,2,1), DlgETCHED_IN);
  DLGSetBorder (Dlg, DlgGRID(10,0,2), DlgETCHED_IN);
  DLGSetBorder (Dlg, DlgGRID(10,1,2), DlgETCHED_IN);
  DLGSetBorder (Dlg, DlgGRID(10,2,2), DlgETCHED_IN);

  /* 
   Also possible:
   DLGSetLabelPos (Dlg, 22,       DlgEAST,       DlgWEST);
   DLGSetLabelPos (Dlg, 22,     DlgCENTER,      DlgNORTH);
   DLGSetLabelPos (Dlg, 22,      DlgSOUTH,      DlgNORTH);
   DLGSetLabelPos (Dlg, 22, DlgNORTH_EAST, DlgSOUTH_WEST);
   ...
  */

  DLGOpenDialog (Dlg, "LABELPOS", DlgNORMAL, NULL);

  DLGFreeDialog (Dlg);
}


/* ********************************************************************* */
/* *** Demonstation der Wertnderung von Items                       *** */
/* ********************************************************************* */
int BarControl (DIALOG *Dlg, int BoxID)
{
  double Value;

  Value = *(double*)DLGGetValue (Dlg, BoxID);
  /* Range: 0.3 - 9.4 */
  if (Value < 2.5)
    DLGSetLabel (Dlg, 10, "This is low grade");
  elsif (Value < 3.6)
    DLGSetLabel (Dlg, 10, "This is lower medium grade");
  elsif (Value < 6.2)
    DLGSetLabel (Dlg, 10, "This is medium grade");
  elsif (Value < 8.1)
    DLGSetLabel (Dlg, 10, "This is upper medium grade");
  else
    DLGSetLabel (Dlg, 10, "This is high grade");

  return DlgCONFIRM;
}

void ChangeValue (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGMakeRows (Dlg, 0, 2, DlgNO_LINE, 10, 20);

  DLGInsertLabel (Dlg, 10, "chgLabel", "Start grade");
  DLGInsertBar   (Dlg, 20,  "chgBar", 0.3, 9.4, 5.2, 0.6, DlgHORIZONTAL);

  DLGSetFunction (Dlg, 20, BarControl);

  DLGOpenDialog (Dlg, "CHANGEVALUE", DlgNORMAL, NULL);

  DLGFreeDialog (Dlg);
}


/* ********************************************************************* */
/* *** In diesem Dialog wird der Einsatz von Space-Items zur         *** */
/* *** Festlegung der Anordnung der Items                            *** */
/* ********************************************************************* */
void ControlDlg (void)
{
  DIALOG *Dlg = NULL;

  PRGPrintf ("this dialog demonstates the using of dummy-items to\n"
	     "position the active items to specific places\n");

  Dlg = DLGNewDialog(0); 

  DLGMakeCols (Dlg,  0, 3, DlgNO_LINE, 10, 20, 30);
  DLGMakeRows (Dlg, 10, 3, DlgNO_LINE, 11, 12, 13);
  DLGMakeRows (Dlg, 20, 3, DlgNO_LINE, 21, 22, 23);  
  DLGMakeRows (Dlg, 30, 3, DlgNO_LINE, 31, 32, 33);  

  DLGInsertProc  (Dlg, 11, "X",                NULL);
  DLGShowBox     (Dlg, 11, FALSE);
  DLGInsertArrow (Dlg, 12, "left",    DlgLEFT, SpaceFkt);  
  DLGInsertArrow (Dlg, 21, "up",        DlgUP, SpaceFkt);  
  DLGInsertArrow (Dlg, 22, "set",   DlgMIDDLE, SpaceFkt);  
  DLGInsertArrow (Dlg, 23, "down",    DlgDOWN, SpaceFkt);  
  DLGInsertProc  (Dlg, 31, "X",                NULL);
  DLGShowBox     (Dlg, 31, FALSE);
  DLGInsertArrow (Dlg, 32, "right",  DlgRIGHT, SpaceFkt);  

  DLGOpenDialog (Dlg, "VIEWCONTROL", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgCANCEL)
    PRGPrintf ("CANCEL selected\n"); 

  /* keine Auswertung ntig! */
  DLGFreeDialog (Dlg);
}


/* ********************************************************************* */
/* *** Dieser Dialog demonstriert die unterschiedlichen              *** */
/* *** Gestaltungsmglichkeiten der Dialogaufteilung                 *** */
/* ********************************************************************* */
void AdornDlg (void)
{
  DIALOG *Dlg = NULL;

  PRGPrintf ("here's a demonstration of the various separators which\n"
	     "are supported by dialoglib. Also the functions DLGSetBoxTitle\n"
	     "and DLGSetBorder are used\n");

  Dlg = DLGNewDialog (0);

  DLGMakeCols (Dlg,  0, 2,                    
	            DlgSINGLE_LINE, 10, 20);
  DLGMakeRows (Dlg, 10, 4, 
	            DlgNO_LINE DlgDASH_LINE DlgSINGLE_LINE, 11, 12, 13, 14);
  DLGMakeRows (Dlg, 20, 4, 
	            DlgDOUBLE_LINE DlgETCHED_IN DlgETCHED_OUT, 21, 22, 23, 24);

  DLGInsertLabel (Dlg, 11, "label1", "Noline:");
  DLGInsertLabel (Dlg, 12, "label2", "Dashline:");
  DLGInsertLabel (Dlg, 13, "label3", "SingleLine:");
  DLGInsertLabel (Dlg, 14, "label4", "");
  DLGInsertLabel (Dlg, 21, "label5", "DoubleLine:");
  DLGInsertLabel (Dlg, 22, "label6", "Etched IN:");
  DLGInsertLabel (Dlg, 23, "label7", "Etched OUT:");
  DLGInsertLabel (Dlg, 24, "label8", "");

  DLGSetBorder (Dlg, 0, DlgETCHED_IN);

  DLGSetTitle (Dlg, 0, "Border", DlgMIDDLE);

  DLGOpenDialog (Dlg, "ADORNDLG", DlgNORMAL, NULL);
 
  DLGFreeDialog (Dlg);
}


/* ********************************************************************* */
/* *** Dieser Dialog demonstriert die Abhngigkeit eines Items       *** */
/* *** von dem aktuellen Wert eines anderen Eintrags                 *** */
/* ********************************************************************* */
int SwitchSens (DIALOG *Dlg, int BoxID)
{
  dual status;

  status = DLGGetBool (Dlg, BoxID);

  /*  DLGSetBoxActive (Dlg, 20, status); */

  DLGShowBox (Dlg, 20, status);

  return DlgCONFIRM;
}


void SensDlg (void)
{
  DIALOG *Dlg = NULL;

  PRGPrintf ("The <UPDN>-item depends directly from the <DUAL>-button\n");

  Dlg = DLGNewDialog (0);
  DLGMakeRows (Dlg, 0, 2, DlgSINGLE_LINE, 10, 20);

  DLGInsertBool (Dlg, 10, "sensitive", FALSE);
  DLGInsertUpDn (Dlg, 20, "depend", 0.0, 10.0, 4.5, 0.5);

  DLGSetFunction  (Dlg, 10, SwitchSens);
  DLGSetBoxActive (Dlg, 20, FALSE);

  DLGOpenDialog (Dlg, "SENSDLG", DlgNORMAL, NULL);

  DLGFreeDialog (Dlg);
}



static void ActivateB (DIALOG *Dlg, int ButtonID)
{
  DLGSetBoxActive (Dlg, 0, FALSE);
  DLGSetButtonActive (Dlg, 2, FALSE);
}

void ActiveB (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGInsertLabel  (Dlg, 0, "label", "Insensitiver Dialog");

  DLGMakeButton (Dlg, OK_BUTTON);
  DLGMakeButton (Dlg, USER_BUTTON, "deactivate", DlgBUTTON_RIGHT, ActivateB, 2);
  DLGMakeButton (Dlg, CANCEL_BUTTON);
  DLGOpenDialog (Dlg, "DEPENDDLG", DlgNORMAL, NULL);

  DLGFreeDialog (Dlg);
}


/* ********************************************************************* */
/* *** Dieser Dialog wird vollstndig insensitiviert und ist nach    *** */
/* *** ca. 10 Sek. wieder nderbar.                                  *** */
/* ********************************************************************* */
void DependExit (DIALOG *Dlg, int Button)
{
  DLGFreeDialog (Dlg);
}


void ActiveDlg (void)
{
  DIALOG *Dlg = NULL;
  time_t     Sec1, Sec2;

  PRGPrintf ("At opening this dialog is completely insensitivated.\n"
	     "After about 10 sec. this state will be changed !\n");

  Dlg = DLGNewDialog (0);
  DLGMakeRows (Dlg, 0, 2, DlgNO_LINE, 10, 20); 

  DLGInsertLabel  (Dlg, 10, "label", "Insensitiver Dialog");
  DLGInsertDouble (Dlg, 20, "value", 91.7, 6, 1);

  DLGSetDlgActive (Dlg, FALSE);

  DLGOpenDialog (Dlg, "DEPENDDLG", DlgNORMAL, DependExit);

  Sec1 = time (&Sec1);
  do 
    Sec2 = time (&Sec2);
  while (Sec2 < Sec1 + 5);

  DLGSetDlgActive (Dlg, TRUE);
}


/* ********************************************************************* */
/* *** Dieser Dialog wird bei bettigung des Buttons fr             *** */
/* *** 10 Sekunden unsichtbar                                        *** */
/* ********************************************************************* */
int HideRoutine (DIALOG *Dlg, int BoxID)
{
  time_t Sec1, Sec2;

  DLGShowDialog (Dlg, FALSE);

  Sec1 = time (&Sec1);
  do 
    Sec2 = time (&Sec2);
  while (Sec2 < Sec1 + 5);

  DLGShowDialog (Dlg, TRUE);

  return DlgCONFIRM;
}

int DeActivate (DIALOG *Dlg, int BoxID)
{
  time_t Sec1, Sec2;

  DLGSetDlgActive (Dlg, FALSE);

  Sec1 = time (&Sec1);
  do 
    Sec2 = time (&Sec2);
  while (Sec2 < Sec1 + 5);

  DLGSetDlgActive (Dlg, TRUE);

  return DlgCONFIRM;
}

void HideDlg (void)
{
  DIALOG *Dlg = NULL;

  PRGPrintf ("here's a demo of making a dialog (in)visible. For\n"
	     "example, if the button is pressed the dialog will\n"
	     "be hidden for about 10 seconds. Then it will be\n"
	     "displayed again\n\n");

  PRGPrintf ("Click to hide dialog\n");

  Dlg = DLGNewDialog (0);
  DLGMakeCols (Dlg, 0, 2, DlgNO_LINE, 10, 20);

  DLGInsertProc (Dlg, 10, "hide",       HideRoutine);
  DLGInsertProc (Dlg, 20, "deactivate", DeActivate);

  DLGOpenDialog (Dlg, "hideDialog", DlgNORMAL, NULL);

  DLGFreeDialog (Dlg);
}


/* ********************************************************************* */
/* *** Dieser Dialog wird auf ein Ereignis (hier Mouse-Button)       *** */
/* *** hin geschlossen                                               *** */
/* ********************************************************************* */
void MsgDlg (DIALOG *Dlg, int Button)
{
  DLGOpenTextDialog ("closeMsg", DlgINFO, "Click in Graphic-Window",
		                 1, "mach ich");

  DLGOpenDialog (Dlg, "autoClose", DlgNORMAL, MsgDlg);
}  


void AutoClose (void)
{
  DIALOG *Dlg = NULL;

  PRGPrintf ("Clicking in a graphic-window calls the DLGCloseDialog\n"
	     "function, which closes the dialog, but don't free the\n"
	     "dialog structure\n");

  Dlg = DLGNewDialog (0);

  DLGInsertLabel (Dlg, 0, "autoMsg", 
    "Press Mouse-Button in a\ngraphic-window to close\n    this dialog");

  DLGOpenDialog (Dlg, "autoClose", DlgNORMAL, MsgDlg);

  /* TODO: 
     Auf Maustaste warten 
  */
  if (DLGGetExitButton(Dlg) == DlgCANCEL)
    DLGCloseDialog (Dlg);

  DLGFreeDialog (Dlg);
}


/* ********************************************************************* */
/* *** Der folgende Dialog ruft beim Beenden eine Callback-          *** */
/* *** Routine auf                                                   *** */
/* ********************************************************************* */
void CallBack (DIALOG *Dlg, int Button)
{
  /* Diese Funktion darf nicht als lokale Funktion in CallBackDlg hinein
     gezogen werden, da durch den parallelen Ablauf diese Prozedur be-
     endet wird und somit auch CallBack nicht mehr existieren wrde 
  */
  if (Button == 1)
    DLGOpenTextDialog ("CallBack", DlgINFO, "OK pressed...", 1, "all right");
  else
    DLGOpenTextDialog ("CallBack", DlgINFO, "CANCEL selected", 1, "corect");

  DLGFreeDialog (Dlg);
}


void CallBackDlg (void)
{
  DIALOG *Dlg = NULL;

  PRGPrintf ("When exiting this dialog a callback (CB) function will\n"
	     "be startet, fetching the full dialog-structure as\n"
	     "parameter\n");

  Dlg = DLGNewDialog (0);
  DLGInsertLabel (Dlg, 0, "cbDlg", "Close to start CallBack");

  DLGOpenDialog (Dlg, "CLOSE", DlgNORMAL, CallBack);

  PRGPrintf ("This text was written after the dialog has been opened !\n");
}


/* ********************************************************************* */
/* *** Das folgende Beispiel demonstriert die Funktion zum ndern    *** */
/* *** der Farbe eine Elements                                       *** */
/* *** Die Fr jede Grundfarbe (rot, grn, blau) kann ein Wert von   *** */
/* *** 0 bis 15 und somit 16^3 = 4096 Farben erzeugt werden.         *** */
/* *** Ich hoffe das reicht frs erste.                              *** */
/* ********************************************************************* */
int ChangeBox (DIALOG *Dlg, int BoxID)
{
  DLGDATA *DlgData;

  DlgData = (DLGDATA *)DLGGetUserData (Dlg, 0);
  DlgData->BoxID = BoxID;

  DLGSetScale (Dlg, 21, (*DlgData).r[BoxID-11]);
  DLGSetScale (Dlg, 22, (*DlgData).g[BoxID-11]);
  DLGSetScale (Dlg, 23, (*DlgData).b[BoxID-11]);

  return DlgCONFIRM;
}

int ColorControl (DIALOG *Dlg, int BoxID)
{
  int      r, g, b;
  int      Color;
  DLGDATA *DlgData;

  if (BoxID < 24)
  {
    DlgData = (DLGDATA *)DLGGetUserData (Dlg, 0);
    r = (int)DLGGetScale (Dlg, 21);
    g = (int)DLGGetScale (Dlg, 22);
    b = (int)DLGGetScale (Dlg, 23);
    DlgData->r[DlgData->BoxID-11] = r;
    DlgData->g[DlgData->BoxID-11] = g; 
    DlgData->b[DlgData->BoxID-11] = b;
    
    DLGSetColor (Dlg, DlgData->BoxID, r, g, b, FALSE);
  }
  else
  {
    Color = DLGGetOMenu (Dlg, 24);
    if (Color == 0)
      DLGSetColor (Dlg, 24, 255, 255, 255, FALSE);
    else
    {
      DLGSetColor (Dlg, 24, 
		   (int)(ColorTable[Color][0]*255), 
		   (int)(ColorTable[Color][1]*255), 
		   (int)(ColorTable[Color][2]*255), FALSE);
      /* Set foreground for dark colors */
      if (Color == 4)
	DLGSetColor (Dlg, 24, 255, 255, 255, TRUE);
      else
	DLGSetColor (Dlg, 24, 0, 0, 0, TRUE);
    }
  }

  return DlgCONFIRM;
}

void ChangeColor (void)
{
  int      count, Depth;
  int      Boxes[DLGCOLORS];
  char     Title[7];
  DLGDATA  DlgData = {11, {240, 240,   0,   0, 176},
	                  {  0, 240, 144,   0, 176},
	                  {  0,   0,   0, 240, 176}};
  DIALOG  *Dlg = NULL;

  Depth = TOOLGetDepth ();

  for (count = 0; count < DLGCOLORS; ++count)
    Boxes[count] = 11+count;

  Dlg = DLGNewDialog (0);
  DLGSetUserData (Dlg, 0, &DlgData); 

  DLGMakeRows (Dlg,  0,          2, DlgNO_LINE, 10, 20);
  DLGMakeCols (Dlg, 10, -DLGCOLORS, DlgNO_LINE, Boxes);
  DLGMakeCols (Dlg, 20,          4, DlgNO_LINE, 21, 22, 23, 24);

  for (count = 0; count < DLGCOLORS; ++count)
  {
    sprintf (Title, "COLOR%d", count+1);
    DLGInsertPic (Dlg, 11+count, Title, DlgXBM, 25, 25, NULL, ChangeBox);
    DLGSetColor  (Dlg, 11+count, 
		       DlgData.r[count], DlgData.g[count], DlgData.b[count], 
                       FALSE);
  }
  DLGInsertScale (Dlg, 21, "red",   0.0, 255, DlgData.r[0], 0, DlgVERTICAL);
  DLGInsertScale (Dlg, 22, "green", 0.0, 255, DlgData.g[0], 0, DlgVERTICAL);
  DLGInsertScale (Dlg, 23, "blue",  0.0, 255, DlgData.b[0], 0, DlgVERTICAL);
  DLGInsertOMenu (Dlg, 24, "color", -34, ColorNames, 6);

  DLGSetLabelPos (Dlg, 10, DlgNOWHERE, DlgCENTER);
  DLGSetLabelPos (Dlg, 20,   DlgNORTH, DlgSOUTH);

  DLGSetFunction (Dlg, 20, ColorControl);

  ColorControl (Dlg, 24);

  DLGOpenDialog (Dlg, "colorControl", DlgNORMAL, NULL);

  DLGFreeDialog (Dlg);
}


void CallButton (DIALOG *Dlg, int ButtonID)
{
  char Msg[255];

  sprintf (Msg, "Button %d pressed\n", ButtonID);
  DLGOpenTextDialog ("BUTTON", DlgINFO, Msg, 1, "OK");
  DLGCloseDialog (Dlg);
}

void ButtonTest (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGInsertBool   (Dlg, 0, "label",  FALSE);

  DLGMakeButton (Dlg, OK_BUTTON);
  DLGMakeButton (Dlg, USER_BUTTON, "close",    DlgBUTTON_RIGHT, NULL,       5);
  DLGMakeButton (Dlg, USER_BUTTON, "function", DlgBUTTON_RIGHT, CallButton, 4);
  DLGMakeButton (Dlg, CANCEL_BUTTON);

  DLGOpenDialog (Dlg, "buttonTest", DlgNORMAL, NULL);
  PRGPrintf ("ExitButton = %d\n", DLGGetExitButton(Dlg));

  DLGFreeDialog (Dlg);
}


static int DigitBlock (DIALOG *Dlg, int BoxID)
{
  PRGPrintf ("Button: %d\n", BoxID);

  return DlgCONFIRM;
}


void GridTest (void)
{
  DIALOG *Dlg;

  Dlg = DLGNewDialog (0);
  DLGMakeCols (Dlg, 0, 3, DlgETCHED_IN DlgETCHED_IN, 10, 20, 30);
  
  DLGMakeGrid (Dlg, 10, 3, 3);
  DLGInsertArrow (Dlg, DlgGRID(10,1,0), "up",     DlgUP,       NULL);
  DLGInsertArrow (Dlg, DlgGRID(10,0,1), "left",   DlgLEFT,     NULL);
  DLGInsertArrow (Dlg, DlgGRID(10,1,1), "center", DlgMIDDLE,   NULL);
  DLGInsertArrow (Dlg, DlgGRID(10,2,1), "right",  DlgRIGHT,    NULL);
  DLGInsertArrow (Dlg, DlgGRID(10,1,2), "down",   DlgDOWN,     NULL);

  DLGMakeGrid (Dlg, 20, 3, 4);
  DLGInsertProc (Dlg, DlgGRID(20,0,0), "7", DigitBlock);
  DLGInsertProc (Dlg, DlgGRID(20,1,0), "8", DigitBlock);
  DLGInsertProc (Dlg, DlgGRID(20,2,0), "9", DigitBlock);
  DLGInsertProc (Dlg, DlgGRID(20,0,1), "4", DigitBlock);
  DLGInsertProc (Dlg, DlgGRID(20,1,1), "5", DigitBlock);
  DLGInsertProc (Dlg, DlgGRID(20,2,1), "6", DigitBlock);
  DLGInsertProc (Dlg, DlgGRID(20,0,2), "1", DigitBlock);
  DLGInsertProc (Dlg, DlgGRID(20,1,2), "2", DigitBlock);
  DLGInsertProc (Dlg, DlgGRID(20,2,2), "3", DigitBlock);
  DLGInsertProc (Dlg, DlgGRID(20,0,3), "0", DigitBlock);
  DLGInsertProc (Dlg, DlgGRID(20,1,3), " .", DigitBlock);
  DLGInsertProc (Dlg, DlgGRID(20,2,3), "<", DigitBlock);

  DLGMakeGrid (Dlg, 30, 3, 4);
  DLGInsertLabel (Dlg, DlgGRID(30,0,0), "label11", "label11");
  DLGInsertLabel (Dlg, DlgGRID(30,1,0), "label21", "label21");
  DLGInsertLabel (Dlg, DlgGRID(30,2,0), "label31", "label31");
  DLGInsertPic   (Dlg, DlgGRID(30,1,1), "cell22", 
		                             DlgXPM, -1, -1, look_xpm, NULL);
  DLGInsertLabel (Dlg, DlgGRID(30,0,2), "label13", "label13");
  DLGInsertLabel (Dlg, DlgGRID(30,1,2), "label23", "label23");
  DLGInsertLabel (Dlg, DlgGRID(30,0,3), "label14", "long-label-14");
  DLGInsertLabel (Dlg, DlgGRID(30,1,3), "label24", "label24");
  DLGInsertLabel (Dlg, DlgGRID(30,2,3), "label34", "label34");

  DLGOpenDialog (Dlg, "arrayTest", DlgNORMAL, NULL);
  DLGFreeDialog (Dlg);
}


void ArrayTest ()
{
  DIALOG *Dlg;

  Dlg = DLGNewDialog (10);

  DLGMakeArray (Dlg, 10, 5, 5);
#if 0
  DLGInsertArrow (Dlg, DlgGRID(10,1,0), "up",     DlgUP,     NULL);
  DLGInsertArrow (Dlg, DlgGRID(10,0,1), "left",   DlgLEFT,   NULL);
  DLGInsertArrow (Dlg, DlgGRID(10,1,1), "center", DlgMIDDLE, NULL);
  DLGInsertArrow (Dlg, DlgGRID(10,2,1), "right",  DlgRIGHT,  NULL);
  DLGInsertArrow (Dlg, DlgGRID(10,1,2), "down",   DlgDOWN,   NULL);
#else
  DLGInsertProc (Dlg, DlgGRID(10,0,0), "Button1",        NULL);
  DLGInsertProc (Dlg, DlgGRID(10,1,1), "2nd Button",     NULL);
  DLGInsertProc (Dlg, DlgGRID(10,2,2), "The 3rd button", NULL);
  DLGInsertProc (Dlg, DlgGRID(10,3,3), "Knpke",         NULL);
  DLGInsertProc (Dlg, DlgGRID(10,4,4), "Hello World",    NULL);
  DLGSetBoxAlignment (Dlg, 10, DlgFILL, DlgFILL);
#endif

  DLGOpenDialog (Dlg, "arrayTest", DlgNORMAL, NULL);
  DLGFreeDialog (Dlg);
}


void KeepPos ()
{
  static int PosX = -1, PosY = -1;
  DIALOG *Dlg;

  Dlg = DLGNewDialog (0);
  DLGInsertLabel (Dlg, 0, "keepPos", "Restore position on next opening");
  DLGSetPos (Dlg, PosX, PosY);
  DLGOpenDialog (Dlg, "keepPosDlg", DlgNORMAL, NULL);
  DLGGetPos (Dlg, &PosX, &PosY);
  DLGFreeDialog (Dlg);
}



void CheckDlg ()
{
}



/* ********************************************************************** */
/* *** Hier wird gezeigt, wie ein Bitmap manuelle verndert wird      *** */
/* *** wichtig ist, da die Reihenfolge der Bits verdreht (LSB links) *** */
/* *** ist, d.h. will man 1101 (1=Punkt) darstellen so ist der Wert   *** */
/* *** 1+2+8 = 0x0b und nicht 8+4+1 = 0x0d einzutragen.               *** */
/* *** Grund: XBM-Format                                              *** */
/* ********************************************************************** */
static int UpdateXBM (DIALOG *Dlg, int BoxID)
{
  int     count1, count2;
  double  Value;
  char   *Data = NULL; 

  Value = *(double*)DLGGetValue (Dlg, BoxID);

  Data = calloc (32, sizeof(char));
  for (count1 = 3; count1 < 5; ++count1)
    for (count2 = 0; count2 < 8; ++count2)
      SetBit (&Data, 32, (int)(Value)+count2, count1);

  for (count1 = 0; count1 < 32; ++count1)
  {
    SetBit (&Data, 32, count1, 0);
    SetBit (&Data, 32, count1, 7);
  }
  for (count2 = 0; count2 < 8; ++count2)
  {
    SetBit (&Data, 32,  0, count2);
    SetBit (&Data, 32, 31, count2);
  }

  DLGChangePic (Dlg, 20, 32, 8, Data);
  free (Data);

  return DlgCONFIRM;
}
  
void XBMChange (void)
{
  unsigned char Data[] = 
    {0xff,0xff,0xff,0xff,   /* xxxxxxxxxxx 32 x xxxxxxxxxxxx */
     0x01,0x00,0x00,0x80,   /* x                           x */
     0x01,0x00,0x00,0x80,   /* x                           x */
     0x01,0xf0,0x0f,0x80,   /* x         xxxxxxxx          x */
     0x01,0xf0,0x0f,0x80,   /* x         xxxxxxxx          x */  
     0x01,0x00,0x00,0x80,   /* x                           x */
     0x01,0x00,0x00,0x80,   /* x                           x */
     0xff,0xff,0xff,0xff};  /* xxxxxxxxxxxxxxxxxxxxxxxxxxxxx */
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);
  DLGMakeCols (Dlg, 0, 2, DlgSINGLE_LINE, 10, 20);

  DLGInsertScale (Dlg, 10, "position", 0.0, 23.0, 12.0, 0, DlgHORIZONTAL);
  DLGInsertPic   (Dlg, 20, "output", DlgXBM, 32, 8, Data, NULL);

  DLGSetFunction (Dlg, 10, UpdateXBM);

  DLGOpenDialog (Dlg, "XBMTest", DlgNORMAL, NULL);

  DLGFreeDialog (Dlg);
}


static void FormTest (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGPlaceInt   (Dlg, 0, 60, 60, "posInt", 776, 6);
  DLGPlaceOMenu (Dlg, 0, 50, 25, "posMenu", 5, 
		 "Option1", "Option2", "Option3", "Option4", "Option5", 1);


  DLGOpenDialog (Dlg, "placeTest", DlgNORMAL, NULL);

  DLGFreeDialog (Dlg);
}


void WidgetTest ()
{
  Widget   Shell, Form;
  Widget   Menu;
  DLGELEM *Elem;
  
  Shell = XtVaCreatePopupShell ("widget",
				topLevelShellWidgetClass, AppShell,
				XmNtitle,                 "menu",
				NULL);

  Form = XtCreateManagedWidget ("form", xmFormWidgetClass, 
				        Shell, NULL, 0);

  Elem = DLGMakeOMenu ("test", 4, "Entry1", "Entry2", NULL, "Entry4", 1);
#if 0
  Menu = DLGCreateWidget (Form, Elem, NULL, 0);
#else
  Menu = DLGVaCreateWidget (Form, Elem, XmNbackground,  TOOLGetColor("Yellow"),
			                XmNforeground,  TOOLGetColor("Red"),
			                XmNborderWidth, 10,
                                        
			                NULL);
#endif

  XtRealizeWidget (Shell);
  XtPopup         (Shell, XtGrabExclusive);

  TOOLMainLoop (AppCon, Shell);

  PRGPrintf ("RESULT: %d\n", Elem->Value.orec.Result);
  DLGFreeElem (Elem);
}


/* ********************************************************************* */
/* *** Die folgenden Dialog des Demo-Menu's zeichen jeweils einen    *** */
/* *** Item-Typ                                                      *** */
/* ********************************************************************* */
void ChangeLabel (DIALOG *Dlg, int ButtonID)
{
  DLGSetLabel (Dlg, 0, "labelChanged");
}

void LabelItem (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);
  DLGMakeRows (Dlg, 0, 2, DlgNO_LINE, 10, 20);

  DLGInsertLabel (Dlg, 10, "label1", "Insert Item");
  DLGInsertLabel (Dlg, 20, "label2:", "Label Value");
  DLGSetLabelPos (Dlg, 20, DlgWEST, DlgEAST);

  DLGMakeButton (Dlg, OK_BUTTON);
  DLGMakeButton (Dlg, USER_BUTTON, "change", DlgBUTTON_RIGHT, ChangeLabel, 2);
  DLGMakeButton (Dlg, CANCEL_BUTTON);
  DLGOpenDialog (Dlg, "labelDlg", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
    PRGPrintf ("no return value\n");

  DLGFreeDialog (Dlg);
}


void ChangeBool (DIALOG *Dlg, int ButtonID)
{
  dual Value;

  Value = DLGGetBool (Dlg, 0);
  PRGPrintf ("Actual Value is %s", Value? "TRUE":"FALSE");
  DLGSetBool (Dlg, 0, !Value);
}

void BoolItem (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGInsertBool (Dlg, 0, "dual", FALSE);

  DLGMakeButton (Dlg, OK_BUTTON);
  DLGMakeButton (Dlg, USER_BUTTON, "change", DlgBUTTON_RIGHT, ChangeBool, 2);
  DLGMakeButton (Dlg, CANCEL_BUTTON);
  DLGOpenDialog (Dlg, "dualDlg", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
    PRGPrintf ("result is %s\n", DLGGetBool(Dlg, 0)? "TRUE":"FALSE");

  DLGFreeDialog (Dlg);
}


void ChangeInt (DIALOG *Dlg, int ButtonID)
{
  PRGPrintf ("Actual INT is %d", DLGGetInt(Dlg, 0));

  DLGChangeInt (Dlg, 0, 8);
  DLGSetInt (Dlg, 0, 228);
}

void IntItem (void)
{
  Arg args[4];
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGInsertInt (Dlg, 0, "int", 117, 6);

  XtSetArg (args[0], XmNmarginWidth, 30);
  XtSetArg (args[1], XmNmarginHeight, 30);
  DLGSetArguments (Dlg, 0, args, 2);

  DLGMakeButton (Dlg, OK_BUTTON);
  DLGMakeButton (Dlg, USER_BUTTON, "change", DlgBUTTON_RIGHT, ChangeInt, 2);
  DLGMakeButton (Dlg, CANCEL_BUTTON);
  DLGOpenDialog (Dlg, "intDlg", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
    PRGPrintf ("result is %d\n", DLGGetInt(Dlg, 0));

  DLGFreeDialog (Dlg);
}


void ChangeDouble (DIALOG *Dlg, int ButtonID)
{
  PRGPrintf ("Actual DOUBLE is %d", DLGGetDouble(Dlg, 0));
  DLGChangeDouble (Dlg, 0, 5, 2);
  DLGSetDouble (Dlg, 0, 12.71);
}

void DoubleItem (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGInsertDouble (Dlg, 0, "double", 1015.433, 9, 3);

  DLGMakeButton (Dlg, OK_BUTTON);
  DLGMakeButton (Dlg, USER_BUTTON, "change", DlgBUTTON_RIGHT, ChangeDouble, 2);
  DLGMakeButton (Dlg, CANCEL_BUTTON);
  DLGOpenDialog (Dlg, "doubleDlg", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
    PRGPrintf ("result is %.3f\n", DLGGetDouble(Dlg, 20));

  DLGFreeDialog (Dlg);
}


void ChangeText (DIALOG *Dlg, int ButtonID)
{
  PRGPrintf ("Actual Text 1 is %s", DLGGetText(Dlg, 10));
  DLGShowBox (Dlg, 10, FALSE);
  DLGChangeText (Dlg, 10, 40, 2);
  DLGShowBox (Dlg, 10, FALSE);
  DLGSetText (Dlg, 10, "Changed Text");
}

void TextItem (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGMakeRows (Dlg, 0, 3, DlgNO_LINE, 10, 20, 30);
  DLGInsertText (Dlg, 10, "text1", 80, "1.st TextItem", 20, 1);
  DLGInsertText (Dlg, 20, "text2", 80, "2.nd TextItem (1.st line)\n"
		                         "(2.nd line)", 20, -2);
  DLGInsertText (Dlg, 30, "text3", 80, "Only vertical scrollbar\n", -20, 1);

  DLGMakeButton (Dlg, OK_BUTTON);
  DLGMakeButton (Dlg, USER_BUTTON, "change", DlgBUTTON_RIGHT, ChangeText, 2);
  DLGMakeButton (Dlg, CANCEL_BUTTON);
  DLGOpenDialog (Dlg, "textDlg", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
  {
    printf ("result 1 is %s\n", (char*)DLGGetValue (Dlg, 10));
    printf ("result 2 is %s\n", DLGGetText(Dlg, 20));
    printf ("result 3 is %s\n", DLGGetText(Dlg, 30));
  }
  DLGFreeDialog (Dlg);
}


void ChangeScale (DIALOG *Dlg, int ButtonID)
{
  PRGPrintf ("Actual Scale 1 is %f", DLGGetScale(Dlg, 11));
  DLGChangeScale (Dlg, 11, -100.0, 0.0, 1);
  DLGSetScale (Dlg, 11, -78.0);
}

void ScaleItem (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGMakeCols (Dlg,  0, 3, DlgNO_LINE, 10, 20, 30);
  DLGMakeRows (Dlg, 10, 2, DlgNO_LINE, 11, 12);

  DLGInsertScale (Dlg, 11, "scale1",   0.0, 100.0, 40.0, 0, DlgHORIZONTAL);
  DLGInsertScale (Dlg, 12, "scale2", 128.0,   1.0, 14.0, 1, DlgHORIZONTAL);
  DLGInsertScale (Dlg, 20, "scale3",   0.0, 255.0, 64.0, 0, DlgVERTICAL);
  DLGInsertScale (Dlg, 30, "scale4",  60.7,   4.3, 20.8, 1, DlgVERTICAL);

  DLGMakeButton (Dlg, OK_BUTTON);
  DLGMakeButton (Dlg, USER_BUTTON, "change", DlgBUTTON_RIGHT, ChangeScale, 2);
  DLGMakeButton (Dlg, CANCEL_BUTTON);
  DLGOpenDialog (Dlg, "scaleDlg", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
  {
    printf ("1.st scale is %f\n", *(double*)DLGGetValue(Dlg, 11));
    printf ("2.nd scale is %f\n", *(double*)DLGGetValue(Dlg, 12));
    printf ("3.rd scale is %f\n", DLGGetScale(Dlg, 20));
    printf ("4.th scale is %f\n", DLGGetScale(Dlg, 30));
  }
  DLGFreeDialog (Dlg);
}


void ChangeBar (DIALOG *Dlg, int ButtonID)
{
  PRGPrintf ("Actual Bar 1 is %f", DLGGetBar(Dlg, 10));
  DLGChangeBar (Dlg, 10, 0.0, 250.0, 128.0);
  DLGSetBar (Dlg, 10, -10.0);
}

void BarItem (void)
{
  double     Result;
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGMakeCols (Dlg,  0, 2, DlgNO_LINE, 10, 20);

  DLGInsertBar (Dlg, 10, "bar1",   0.0, 100.0, 40.0, 1.0, DlgHORIZONTAL);
  DLGInsertBar (Dlg, 20, "bar2", 128.0,   2.0, 14.0, 2.0, DlgVERTICAL);

  DLGMakeButton (Dlg, OK_BUTTON);
  DLGMakeButton (Dlg, USER_BUTTON, "change", DlgBUTTON_RIGHT, ChangeBar, 2);
  DLGMakeButton (Dlg, CANCEL_BUTTON);
  DLGOpenDialog (Dlg, "barDlg", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
  {
    Result = *(double*)DLGGetValue (Dlg, 10);
    printf ("1.st bar is %f\n", Result);
    printf ("2.nd bar is %f\n", DLGGetBar(Dlg, 20));
  }
  DLGFreeDialog (Dlg);
}


void ChangeUpDn (DIALOG *Dlg, int ButtonID)
{
  PRGPrintf ("Actual UpDn 1 is %d", DLGGetUpDn(Dlg, 10));
  DLGChangeUpDn (Dlg, 10, 0.0, 1.0, 0.15);
  DLGSetUpDn (Dlg, 10, 0.40);
}

void UpDnItem (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGMakeRows (Dlg, 0, 2, DlgSINGLE_LINE, 10, 20);
  DLGInsertUpDn (Dlg, 10, "upDn1", 0.0, 100.0, 40.0, 1.5);
  DLGInsertUpDn (Dlg, 20, "upDn2", 0.0, 0.0, 40.0, 1.5);

  DLGMakeButton (Dlg, OK_BUTTON);
  DLGMakeButton (Dlg, USER_BUTTON, "change", DlgBUTTON_RIGHT, ChangeUpDn, 2);
  DLGMakeButton (Dlg, CANCEL_BUTTON);
  DLGOpenDialog (Dlg, "upDnDlg", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
  {
    printf ("result1 is %f\n", *(double*)DLGGetValue(Dlg, 10));
    printf ("result2 is %f\n", DLGGetUpDn(Dlg, 20));
  }
  DLGFreeDialog (Dlg);
}


void ChangeList (DIALOG *Dlg, int ButtonID)
{
  PRGPrintf ("Actual List 1 is %d", DLGGetList(Dlg, 10));
  DLGChangeList (Dlg, 10, 4, 8, "a", "b", "c", "d", "e", "f", "g", "h");
  DLGSetList (Dlg, 10, 2);
}

void ListItem (void)
{
  int        Result;
  DIALOG *Dlg = NULL;
  char      *Entries[] = {"one", "two", "three", "four", "five", "six"};

  Dlg = DLGNewDialog (0);

  DLGMakeCols (Dlg, 0, 2, DlgETCHED_IN, 10, 20);

  DLGInsertList(Dlg, 10, "list1", 4, 4,
		         "Entry1", "Entry2", "Entry3", "Entry4", 0);
  DLGInsertList(Dlg, 20, "list2", 3, -6, Entries, 1);

  DLGMakeButton (Dlg, OK_BUTTON);
  DLGMakeButton (Dlg, USER_BUTTON, "change", DlgBUTTON_RIGHT, ChangeList, 2);
  DLGMakeButton (Dlg, CANCEL_BUTTON);
  DLGOpenDialog (Dlg, "listDlg", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
  {
    Result = *(int*)DLGGetValue (Dlg, 10);
    printf ("1.st selection is Entry%d\n", Result);
    printf ("2.nd selection is ScrEntry%d\n", DLGGetList(Dlg, 20));
    printf ("3.rd selection is %d\n", DLGGetList(Dlg, 30));
  }
  DLGFreeDialog (Dlg);
}


void ChangeOMenu (DIALOG *Dlg, int ButtonID)
{
  PRGPrintf ("Actual OMenu 1 %d", DLGGetOMenu(Dlg, 10));
  DLGChangeOMenu (Dlg, 10, 7, "a", NULL, "c", "d", NULL, "f", "g");
  DLGSetOMenu (Dlg, 10, 2);
}

void MenuItem (void)
{
  int        Result;
  char      *Options[] = {"option_0", "option_1"};
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGMakeRows (Dlg, 0, 2, DlgNO_LINE, 10, 20);
  DLGInsertOMenu (Dlg, 10, "menu1", 5,
                           "Entry1", "Entry2", "Entry3", NULL, "Entry4", 1);
  DLGInsertOMenu (Dlg, 20, "menu2", -2, Options, 0);
  DLGSetLabelPos (Dlg, 20, DlgNOWHERE, DlgCENTER);

  DLGMakeButton (Dlg, OK_BUTTON);
  DLGMakeButton (Dlg, USER_BUTTON, "change", DlgBUTTON_RIGHT, ChangeOMenu, 2);
  DLGMakeButton (Dlg, CANCEL_BUTTON);
  DLGOpenDialog (Dlg, "menuDlg", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
  {
    Result = *(int*)DLGGetValue (Dlg, 10);
    printf ("menu-selection 1 is Entry %d\n", Result);
    printf ("menu-selection 2 is Entry %d\n", DLGGetOMenu(Dlg, 20));
  }
  DLGFreeDialog (Dlg);
}


void ChangeToggle (DIALOG *Dlg, int ButtonID)
{
  dual Buttons[8] = {FALSE, FALSE, TRUE, FALSE, FALSE, TRUE, TRUE, FALSE};

  DLGSetTBox (Dlg, 10, Buttons);
}

void TBoxItem (void)
{
  dual      *Result = NULL;
  int        count;
  TOGGLE     Buttons[8] = {{"toggle_1",  TRUE},
			   {"toggle_2",  TRUE},
			   {"toggle_3", FALSE},
			   {"toggle_4",  TRUE},
			   {"toggle_5", FALSE}, 
			   {"toggle_6", FALSE},
			   {"toggle_7",  TRUE}, 
			   {"toggle_8", FALSE}};
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGMakeCols (Dlg, 0, 2, DlgNO_LINE, 10, 20);

  DLGInsertTBox (Dlg, 10, "tBox1", 2, 8,
                          "Toggle1",  TRUE, "Toggle2", FALSE,
                          "Toggle3", FALSE, "Toggle4",  TRUE,
                          "Toggle5",  TRUE, "Toggle6",  TRUE,
                          "Toggle7",  TRUE, "Toggle8", FALSE);
  DLGInsertTBox (Dlg, 20, "tBox2", -3, -8, Buttons);

  DLGMakeButton (Dlg, OK_BUTTON);
  DLGMakeButton (Dlg, USER_BUTTON, "change", DlgBUTTON_RIGHT, ChangeToggle, 2);
  DLGMakeButton (Dlg, CANCEL_BUTTON);
  DLGOpenDialog (Dlg, "tBoxDlg", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
  {
    Result = *(dual**) DLGGetValue (Dlg, 10);
    printf ("result 1 is:\n");
    for (count = 0; count < 8; ++count)
      printf ("Toggle%d=%c ", count, Result[count]? 'T': 'F');
    printf ("\n");

    printf ("result 2 is:\n");
    for (count = 0; count < 8; ++count)
      printf ("Toggle%d=%c ", count, DLGGetTBox(Dlg, 20, count)? 'T': 'F');
    printf ("\n");
  }
  DLGFreeDialog (Dlg);
}


int RadioCB (DIALOG *Dlg, int BoxID)
{
  printf ("Radio: %d = %d\n", BoxID, DLGGetRBox(Dlg, BoxID));
  return DlgCONFIRM;
}

void ChangeRadio (DIALOG *Dlg, int ButtonID)
{
  DLGSetRBox (Dlg, 10, 4);
}

void RBoxItem (void)
{
  int        Result;
  char      *Buttons[] = {"button_1", "button_2", "button_3", "button_4",
			  "button_5", "button_6", "button_7", "button_8"};
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGMakeCols (Dlg, 0, 2, DlgNO_LINE, 10, 20);

  DLGInsertRBox (Dlg, 10, "rBox1", 2, 7,
                          "Button1", "Button2", "Button3", "Button4", 
                          "Button5", "Button6", "Button7", 2);
  DLGInsertRBox (Dlg, 20, "rBox2", -3, -8, Buttons, 4);

  DLGSetFunction (Dlg, 10, RadioCB);
  DLGSetFunction (Dlg, 20, RadioCB);

  DLGMakeButton (Dlg, OK_BUTTON);
  DLGMakeButton (Dlg, USER_BUTTON, "change", DlgBUTTON_RIGHT, ChangeRadio, 2);
  DLGMakeButton (Dlg, CANCEL_BUTTON);
  DLGOpenDialog (Dlg, "rBoxDlg", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
  {
    Result = *(int*)DLGGetValue (Dlg, 10);
    printf ("result of radio-box 1 is Button %d\n", Result);
    printf ("result of radio-box 2 is Button %d\n", DLGGetRBox(Dlg, 20));
  }
  DLGFreeDialog (Dlg);
}


int ProcCall (DIALOG *Dlg, int BoxID)
{
  PRGPrintf ("PROC-function started\n");

  return DlgCONFIRM;
}

void ProcItem (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);
  DLGInsertProc (Dlg, 0, "proc", ProcCall);

  DLGOpenDialog (Dlg, "procDlg", DlgNORMAL, NULL);

  DLGFreeDialog (Dlg);
}


int ArrowCall (DIALOG *Dlg, int BoxID)
{
  PRGPrintf ("ARROW-function started\n");

  return DlgCONFIRM;
}

void ArrowItem (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);
  DLGMakeCols (Dlg, 0, 2, DlgNO_LINE, 10, 20);

  DLGInsertArrow (Dlg, 10, "arrow1",  DlgLEFT, ArrowCall);
  DLGInsertArrow (Dlg, 20, "arrow2", DlgRIGHT, ArrowCall);

  DLGOpenDialog (Dlg, "arrowDlg", DlgNORMAL, NULL);

  DLGFreeDialog (Dlg);
}


void PicItem (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGMakeCols (Dlg, 0, 2, DlgNO_LINE, 10, 20);
  DLGInsertPic (Dlg, 10, "pic1", DlgXBM,
		         penguin_width, penguin_height, penguin_bits, NULL);
  DLGInsertPic (Dlg, 20, "pic2", DlgXPM, -1, -1, new_xpm, NULL);

  DLGOpenDialog (Dlg, "picDlg", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
    PRGPrintf ("no return value\n");

  DLGFreeDialog (Dlg);
}


int PicBCall (DIALOG *Dlg, int BoxID)
{
  PRGPrintf ("PictureButton-function called\n");

  return DlgCONFIRM;
}

void PicBItem (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);
  DLGMakeCols (Dlg, 0, 2, DlgNO_LINE, 10, 20);

  DLGInsertPic (Dlg, 10, "picB1", DlgXBM,
		     penguin_width, penguin_height, penguin_bits, PicBCall);
  DLGInsertPic (Dlg, 20, "picB2", DlgXPM, 0, 0, "./tux.xpm", PicBCall);

  DLGOpenDialog (Dlg, "picBDlg", DlgNORMAL, NULL);

  DLGFreeDialog (Dlg);
}


int ImageBCall (DIALOG *Dlg, int BoxID)
{
  XImage *Image;

  PRGPrintf ("ImageButton-function called");

  Image = TOOLLoadJPG ("./penguin2.jpg", NULL, NULL);
  DLGChangeImage (Dlg, 10, Image);

  return DlgCONFIRM;
}

void ImageItem (void)
{
  XImage *Image;
  DIALOG *Dlg = NULL;

  if ((Image = TOOLLoadJPG("./penguin.jpg", NULL, NULL)) == NULL)
    PRGPrintf ("Image not './penguin.jpg' not found");

  Dlg = DLGNewDialog (0);
  DLGMakeCols (Dlg, 0, 2, DlgNO_LINE, 10, 20);
  
  DLGInsertImage (Dlg, 10, "image1", Image, NULL);
  DLGInsertImage (Dlg, 20, "imageB2", Image, ImageBCall);

  DLGOpenDialog (Dlg, "imageBDlg", DlgNORMAL, NULL);

  DLGFreeDialog (Dlg);
}


#if (XmVERSION>=2)
void ChangeNote (DIALOG *Dlg, int ButtonID)
{
  DLGSetNote (Dlg, 0, 2);
}

void NoteItem (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGInsertNote  (Dlg,  0, "note", DlgSPIRAL);

  DLGAddPage     (Dlg,  0, DlgMAJOR, "page1", 10);
  /* now continue with box 10 like in 'normal' dialogs... */
  DLGMakeRows    (Dlg, 10, 4, 
		       DlgETCHED_IN DlgNO_LINE DlgETCHED_IN, 11, 12, 13, 14);
  DLGInsertLabel (Dlg, 11, "page1", "This is page 1");
  DLGInsertScale (Dlg, 12, "scale", 0.0, 60.0, 23.1, 1, DlgHORIZONTAL);
  DLGInsertPic   (Dlg, 13, "pic", DlgXPM, -1, -1, new_xpm, NULL);
  DLGInsertOMenu (Dlg, 14, "menu", 2, "YES", "NO", 0);

  DLGAddPage     (Dlg,  0, DlgPAGE, "page2", 20);
  DLGInsertLabel (Dlg, 20, "page2", "This is page 2");

  DLGAddPage     (Dlg,  0, DlgMINOR,  "page2-1", 30);
  DLGInsertLabel (Dlg, 30, "page2-1", "This is page 2-1");

  DLGAddPage     (Dlg,  0, DlgMINOR, "sub", 40);
  DLGInsertLabel (Dlg, 40, "sub1", "This is sub-page 1");

  DLGAddPage     (Dlg,  0, DlgMAJOR, "page3", 50);
  DLGInsertLabel (Dlg, 50, "page3", "This is page 3");

  DLGAddPage     (Dlg,  0, DlgMINOR, "page4", 60);
  DLGInsertLabel (Dlg, 60, "page4", "This is page 4");

  DLGMakeButton (Dlg, OK_BUTTON);
  DLGMakeButton (Dlg, USER_BUTTON, "change", DlgBUTTON_RIGHT, ChangeNote, 2);
  DLGMakeButton (Dlg, CANCEL_BUTTON);
  DLGOpenDialog (Dlg, "noteDlg", DlgNORMAL, NULL);

  PRO1 ("Actual Page was: %d\n", DLGGetNote(Dlg, 0))

  DLGFreeDialog (Dlg);
}


void TreeItem (void)
{
  TREEITEM *Item;
  DIALOG   *Dlg = NULL;
  Pixmap   ClosePic, OpenPic;

  Dlg = DLGNewDialog (0);

  ClosePic = TOOLMakeXPM (plus_xpm,  NULL, NULL, TOOLGetColor("grey90"));
  OpenPic  = TOOLMakeXPM (minus_xpm, NULL, NULL, TOOLGetColor("grey90"));

  DLGInsertTree  (Dlg, 0, "tree",  ClosePic, OpenPic, 0);
  DLGAddTreeIcon (Dlg, 0, "test1",  ball_yellow,  1,  0, TRUE, NULL);
  DLGAddTreeIcon (Dlg, 0, "test2",  ball_red,     2,  1, TRUE, NULL);
  DLGAddTreeIcon (Dlg, 0, "test3",  ball_red,     3,  1, TRUE, NULL);
  DLGAddTreeIcon (Dlg, 0, "test4",  ball_blue,    4,  3, TRUE, NULL);
  DLGAddTreeIcon (Dlg, 0, "test5",  ball_green,   5,  4, TRUE, NULL);
  DLGAddTreeIcon (Dlg, 0, "test6",  ball_green,   6,  4, TRUE, NULL);
  DLGAddTreeIcon (Dlg, 0, "test7",  ball_blue,    7,  3, TRUE, NULL);
  DLGAddTreeIcon (Dlg, 0, "test7",  ball_green,   8,  7, TRUE, NULL);
  DLGAddTreeIcon (Dlg, 0, "test8",  ball_red,     9,  1, TRUE, NULL);

  DLGAddTreeIcon (Dlg, 0, "test9",  ball_yellow, 11, 10, TRUE, NULL);
  DLGAddTreeIcon (Dlg, 0, "test10", ball_red,    12, 11, TRUE, NULL);
  DLGAddTreeIcon (Dlg, 0, "test11", ball_red,    13, 11, TRUE, NULL);

  DLGSetSize (Dlg, 0, 0, TRUE);

  DLGOpenDialog (Dlg, "treeDlg", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
    if ((Item = DLGGetTree(Dlg, 0)) != NULL)
      printf ("Selection is: %d\n", Item->ID);

  XFreePixmap (XDisplay, ClosePic);
  XFreePixmap (XDisplay, OpenPic);

  DLGFreeDialog (Dlg);
}
#endif


void FSelItem (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGMakeCols (Dlg,  0, 2, DlgNO_LINE, 10, 20);
  DLGMakeRows (Dlg, 10, 2, DlgNO_LINE, 11, 12);
  DLGMakeRows (Dlg, 20, 2, DlgNO_LINE, 21, 22);

  DLGInsertFileSel (Dlg, 11, "fSel1",   "$HOME/*", 30, "$HOME/test");    /* */
  DLGInsertFileSel (Dlg, 12, "fSel2",   "$HOME/*",  0, "$HOME/test");    /* */
  DLGInsertDirSel  (Dlg, 21, "dirSel1", "$HOME/tux*", 30, "$HOME/project"); /* */
  DLGInsertDirSel  (Dlg, 22, "dirSel2", "$HOME/tux*",  0, "$HOME/project"); /* */

  DLGSetSize (Dlg, 0, 0, FALSE);

  DLGOpenDialog (Dlg, "fSelDlg", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
  {
    printf ("1st filename is %s\n",  (char*)DLGGetValue(Dlg, 11));
    printf ("2nd filename is %s\n",         DLGGetFileSel(Dlg, 12));
    printf ("1st directory is %s\n",        DLGGetDirSel(Dlg, 21));
    printf ("1st directory is %s\n", (char*)DLGGetValue(Dlg, 22));
  }
  DLGFreeDialog (Dlg);
}


int DialTest (DIALOG *Dlg, int BoxID)
{
  if (DLGGetDial(Dlg, BoxID) == 50.0)
    DLGSetDial (Dlg, BoxID, 30.7);

  return DlgCONFIRM;
}

void DialItem (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGMakeCols (Dlg, 0, 2, DlgNO_LINE, 10, 20);
  DLGInsertDial (Dlg, 10, "dial1", 10.0, 100.0, 20.0, 5, 1);
  DLGInsertDial (Dlg, 20, "dial2",  0.0, 100.0, 40.0, 5, 0);

  DLGSetFunction (Dlg, 10, DialTest);

  DLGOpenDialog (Dlg, "dialDlg", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgOK)
  {
    printf ("1st dial-value is %f\n", DLGGetDial(Dlg, 10));
    printf ("2nd dial-value is %f\n", DLGGetDial(Dlg, 20));
  }
  DLGFreeDialog (Dlg);
}


static int ControlSegm (DIALOG *Dlg, int BoxID)
{
  DLGSetSegment (Dlg, 20, DLGGetUpDn(Dlg, 10));

  return DlgCONFIRM;
}

void SegmItem (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGMakeCols (Dlg, 0, 2, DlgNO_LINE, 10, 20);
  DLGInsertUpDn    (Dlg, 10, "control", 0.0, 100.0, 40.0, 1.0);
  DLGInsertSegment (Dlg, 20, "segment", 127.3, 8, 2);

  DLGSetFunction (Dlg, 10, ControlSegm);

  DLGOpenDialog (Dlg, "segmentDlg", DlgNORMAL, NULL);

  DLGFreeDialog (Dlg);
}


static Widget CreateWidget (DIALOG *Dlg, Widget Parent, void *Param)
{
  Widget Result;
  Arg    args[1];

  PRO1 ("Creating User-Widget. User-Param %d\n", (int)Param)

  /* Insert attributes... */
  Result = XtCreateManagedWidget ("user", xmLabelWidgetClass, Parent, args, 0);
  /* Add callback... */

  return Result;
}

void UserItem (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGInsertWidget (Dlg, 0, CreateWidget, (void*)10);

  DLGMakeButton (Dlg, NO_BUTTON);
  DLGMakeButton (Dlg, OK_BUTTON);
  DLGMakeButton (Dlg, NO_BUTTON);
  DLGMakeButton (Dlg, CANCEL_BUTTON);  
  DLGOpenDialog (Dlg, "userDlg", DlgNORMAL, NULL);

  DLGFreeDialog (Dlg);
}


/* ********************************************************************* */
/* ***    Dieser Dialog demonstriert die meisten unterstzten        *** */
/* ***                          Item-Typen                           *** */
/* ********************************************************************* */
int ChangeProc (DIALOG *Dlg, int BoxID)
{
  dual Value;

  Value = DLGGetBool (Dlg, BoxID);

  if (! Value)
  {
    DLGSetMousePos  (Dlg, 12);

    DLGSetInt       (Dlg, 10, 9876);
    DLGSetDouble    (Dlg, 11, 0.815);
    DLGSetText      (Dlg, 30, "Changed Text");
    DLGSetUpDn      (Dlg, 41, 18.0);
    DLGSetList      (Dlg, 13, 0);
    DLGSetRBox      (Dlg, 55, 5);
    DLGSetOMenu     (Dlg, 40, 1);
    DLGSetScale     (Dlg, 66, 7.2);
  }
  else
  {
    DLGSetMousePos  (Dlg, 30);

    DLGSetBoxActive (Dlg, 16, TRUE);
    DLGSetBoxActive (Dlg, 58, TRUE);
    DLGSetBoxActive (Dlg, 50, TRUE);
    DLGSetBoxActive (Dlg, 77, TRUE);
    DLGSetScale     (Dlg, 66, 3.9);
  }
  DLGSetBoxActive (Dlg, 50, Value);

  return DlgCONFIRM;
}

static int ChangeSegment (DIALOG *Dlg, int BoxID)
{
  DLGSetSegment (Dlg, 63, DLGGetBar(Dlg, 67));

  return DlgCONFIRM;
}

static int WriteAppDef (DIALOG *Dlg, int BoxID)
{
  char  FileName[PATH_MAX+1] = "$HOME/.appdefaults/Hello.ad";
  char  FileMask[PATH_MAX+1] = "$HOME/.appdefaults/*.ad"; /* */
  FILE *fp = NULL;

  if (DLGOpenFileSel("selectAppFile", FileName, PATH_MAX, FileMask) != NULL)
    if ((fp = fopen(FileName, "a+")) != NULL)
    {
      DLGWriteAppDefs (fp, Dlg);
      fclose (fp);
      DLGOpenTextDialog ("readyMsg", DlgINFO, 
			             "AppDef-file written !", 1, "OK");
    }

  return DlgCONFIRM;
}

static void CompleteHelp (DIALOG *Dlg, int BoxID)
{
  char Msg[255];
  char Help[50] = "$HOME/project/help/libprg/complete.hlp";

  switch (BoxID)
  {
    case -1 :
      sprintf (Msg, "%s\n\nYou called the help-function from BoxID %d\n",
	            "This is a dialog with most\n    element-types\n", BoxID);
      DLGOpenTextDialog ("HELPMSG", DlgINFO, Msg, 1, "OK");
      return;
      break;

    case 10 : strcat (Help, "#int");
              break;
    case 11 : strcat (Help, "#double");
              break;
    case 12 :
    case 13 : strcat (Help, "#list");
              break;
    case 16 : strcat (Help, "#dual");
              break;
    case 17 : strcat (Help, "#picb");
              break;
    case 30 : strcat (Help, "#text");
              break;
    case 31 : strcat (Help, "#filesel");
              break;
    case 40 : strcat (Help, "#omenu");
              break;
    case 41 : strcat (Help, "#updn");
              break;
    case 52 : strcat (Help, "#toggle");
              break;
    case 62 : strcat (Help, "#proc");
              break;
    case 66 : strcat (Help, "#scale");
              break;
    case 67 : strcat (Help, "#bar");
              break;
    case 69 : strcat (Help, "#dial");
              break;
    case 72 :
    case 76 :
    case 77 :
    case 78 :
    case 82 : strcat (Help, "#arrow");
              break;

    default : PRGPrintf ("Description for element %d is missing !", BoxID);
  }
  HELPOpenHelpWin ("complete", Help, 0, 0, 500, 300, NULL, 0, NULL);
}

int CBTest (DIALOG *Dlg, int BoxID)
{
  printf ("CB-Test: %d\n", BoxID);

#if 0
  return DlgCONFIRM;
#else
  return DlgDENY;
#endif
}

void Complete ()
{
  int        count;
  dual      *Result = NULL;
  DIALOG *Dlg = NULL;

  /* create new dialog. with boxID 0: */
  Dlg = DLGNewDialog (0);

  /* split dialog into subboxes: */
  DLGMakeCols  (Dlg,  0, 3, DlgSINGLE_LINE DlgETCHED_IN,  5,  6,  7    );
  DLGMakeRows  (Dlg,  5, 3,                  DlgNO_LINE,  8,  9, 15    );
  DLGMakeRows  (Dlg,  8, 2,                  DlgNO_LINE, 10, 11        );
  DLGMakeRows  (Dlg,  9, 3,                  DlgNO_LINE, 12, 13, 14    );
  DLGMakeRows  (Dlg,  6, 3,                  DlgNO_LINE, 16, 17, 18    );
  DLGMakeRows  (Dlg,  7, 3,                  DlgNO_LINE, 20, 21, 22    );

  DLGMakeCols  (Dlg, 20, 2,              DlgDOUBLE_LINE, 30, 31        );
  DLGMakeCols  (Dlg, 21, 3,                  DlgNO_LINE, 50, 52, 55    );
  DLGMakeRows  (Dlg, 50, 2,                  DlgNO_LINE, 40, 41        );
  DLGMakeCols  (Dlg, 22, 2,                  DlgNO_LINE, 60, 69        );
  DLGMakeRows  (Dlg, 60, 2,                  DlgNO_LINE, 61, 65        );
  DLGMakeCols  (Dlg, 61, 2,                  DlgNO_LINE, 62, 63        );
  DLGMakeCols  (Dlg, 65, 2,                DlgDASH_LINE, 66, 67        );
  DLGMakeGrid  (Dlg, 18, 3, 3);

  DLGSetBoxAlignment (Dlg, 15, DlgFILL,   DlgDOWN);
  DLGSetBoxAlignment (Dlg, 18, DlgMIDDLE, DlgUP);
  DLGSetBoxAlignment (Dlg, 55, DlgRIGHT,  DlgUP);
  DLGSetBoxAlignment (Dlg, 63, DlgRIGHT,  DlgUP);
  DLGSetBoxAlignment (Dlg, 65, DlgLEFT,   DlgDOWN);
  DLGSetBoxAlignment (Dlg, 69, DlgRIGHT,  DlgUP);

  /* define dialog items: */
  DLGInsertBool     (Dlg, 16, "dual", TRUE);
  DLGInsertPic      (Dlg, 17, "bitMap", DlgXBM, 
		              penguin_width, penguin_height, penguin_bits, 
		              StartProc);
  DLGInsertInt      (Dlg, 10, "int", 1234, 8);
  DLGInsertDouble   (Dlg, 11, "double", 107.85, 8, 3);
  DLGInsertList     (Dlg, 12, "list", 3, 3, "list1", "list2", "list3", 0);
  DLGInsertList     (Dlg, 13, "scrList",  3, 5,
                              "scrList1", "scrList2", "scrList3", 
		              "scrList4", "scrList5", 2);
  /* Box 14 left empty */
  DLGInsertPic      (Dlg, 15, "pic", DlgXPM, -1, -1, new_xpm, NULL);
  DLGInsertText     (Dlg, 30, "text", 65, 
		     "This is a Text-Widget-Test\nNow the 2nd Line", 25, 2);
  DLGInsertFileSel  (Dlg, 31, "fileSel", "$HOME/*", 30, "$HOME/test"); /* */
  DLGInsertTBox     (Dlg, 52, "box", 2, 8,
                              "toggle1",  TRUE, "toggle2", FALSE,
                              "toggle3", FALSE, "toggle4",  TRUE,
		              "toggle5",  TRUE, "toggle6",  TRUE,
		              "toggle7",  TRUE, "toggle8", FALSE);
  DLGInsertRBox     (Dlg, 55, "radio", 2, 7,
		              "radio1", "radio2", "radio3", "radio4", 
		              "radio5", "radio6", "radio7", 2);
  DLGInsertProc     (Dlg, 62, "writeAppDef", WriteAppDef);
  DLGInsertSegment  (Dlg, 63, "segment",     134.15, 6, 2);
  DLGInsertScale    (Dlg, 66, "scale",       3.5, 9.2, 5.12, 2, DlgHORIZONTAL);
  DLGInsertBar      (Dlg, 67, "bar",         3.15, 344.15, 134.15, 1.0, 
		                                                DlgHORIZONTAL);
  DLGInsertDial     (Dlg, 69, "dial",        0.0, 100.0, 43.2, 5, 1);
  DLGInsertArrow    (Dlg, DlgGRID(18,1,0), "up",    DlgUP,     SpaceFkt);
  DLGInsertArrow    (Dlg, DlgGRID(18,0,1), "left",  DlgLEFT,   SpaceFkt);
  DLGInsertArrow    (Dlg, DlgGRID(18,1,1), "set",   DlgMIDDLE, SpaceFkt);
  DLGInsertArrow    (Dlg, DlgGRID(18,2,1), "right", DlgRIGHT,  SpaceFkt);
  DLGInsertArrow    (Dlg, DlgGRID(18,1,2), "down",  DlgDOWN,   SpaceFkt);
  DLGInsertOMenu    (Dlg, 40, "menu", 5, "test1", "test2", "test3",
		                         NULL, "test4", 2);
  DLGInsertUpDn     (Dlg, 41, "upDown", 0.0, 100.0, 36.0, 1.25);

  DLGSetFunction (Dlg, 16, ChangeProc);
  DLGSetFunction (Dlg, 67, ChangeSegment);

  /* define box-titles: */
  DLGSetTitle (Dlg,  0, "headline", DlgMIDDLE);
  DLGSetTitle (Dlg, 41, "boxTitle", DlgLEFT);

  /* set some borders: */
  DLGSetBorder (Dlg,  8, DlgETCHED_IN);
  DLGSetBorder (Dlg,  9, DlgETCHED_IN);
  DLGSetBorder (Dlg, 18, DlgSHADOW_IN);
  DLGSetBorder (Dlg, 21, DlgETCHED_IN);
  DLGSetBorder (Dlg, 41, DlgETCHED_OUT);
  DLGSetBorder (Dlg, 63, DlgSHADOW_IN);
  DLGSetBorder (Dlg, 65, DlgETCHED_IN);
  DLGSetBorder (Dlg, 69, DlgSHADOW_OUT);

  DLGSetLabelPos (Dlg,  9, DlgNORTH_WEST, DlgNORTH_EAST);
  DLGSetLabelPos (Dlg, 15, DlgNOWHERE,    DlgCENTER);
  DLGSetLabelPos (Dlg, 16, DlgNORTH,      DlgCENTER);
  DLGSetLabelPos (Dlg, 21, DlgNOWHERE,    DlgCENTER);
  DLGSetLabelPos (Dlg, 61, DlgWEST,       DlgEAST);
  DLGSetLabelPos (Dlg, 69, DlgNOWHERE,    DlgCENTER);

  DLGSetSize (Dlg, 0, 0, FALSE);

  /* specify help-function (activated by pressing F1): */
  DLGSetHelpFunction (Dlg, CompleteHelp);

#if 1
  DLGMakeButton (Dlg, DEFAULT_ID, "default", 
		      DlgBUTTON_LEFT | DlgBUTTON_ADJUST,   NULL, DlgDEFAULT);
  DLGMakeButton (Dlg, HELP_ID,    "help",    
		      DlgBUTTON_MIDDLE | DlgBUTTON_ADJUST,   NULL, DlgHELP);
  DLGMakeButton (Dlg, CANCEL_ID,  "cancel",  
		      DlgBUTTON_MIDDLE | DlgBUTTON_ADJUST, NULL, DlgCANCEL);
  DLGMakeButton (Dlg, OK_ID,      "ok",  
		      DlgBUTTON_RIGHT | DlgBUTTON_ADJUST,  NULL, DlgOK);
#else
  /* TODO: Eigenstndige Demo-Dlgs erstellen: */
  DLGMakeButton (Dlg, DEFAULT_ID, "default", 
		      DlgBUTTON_LEFT | DlgBUTTON_FILL,   NULL, DlgDEFAULT);
  DLGMakeButton (Dlg, HELP_ID,    "help",    
		      DlgBUTTON_MIDDLE | DlgBUTTON_FILL,   NULL, DlgHELP);
  DLGMakeButton (Dlg, CANCEL_ID,  "cancel",  
		      DlgBUTTON_MIDDLE | DlgBUTTON_FILL, NULL, DlgCANCEL);
  DLGMakeButton (Dlg, OK_ID,      "ok",  
		      DlgBUTTON_RIGHT | DlgBUTTON_FILL,  NULL, DlgOK);


  DLGMakeButton (Dlg, DEFAULT_ID, "default", 
		      DlgBUTTON_EVEN | DlgBUTTON_ADJUST,   NULL, DlgDEFAULT);
  DLGMakeButton (Dlg, HELP_ID,    "help",    
		      DlgBUTTON_EVEN | DlgBUTTON_ADJUST,   NULL, DlgHELP);
  DLGMakeButton (Dlg, CANCEL_ID,  "cancel",  
		      DlgBUTTON_EVEN | DlgBUTTON_ADJUST, NULL, DlgCANCEL);
  DLGMakeButton (Dlg, OK_ID,      "ok",  
		      DlgBUTTON_EVEN | DlgBUTTON_ADJUST,  NULL, DlgOK);
#endif

  /* open dialog-window: */
  DLGOpenDialog (Dlg, "complete", DlgNORMAL, NULL); 

  /* check result: */
  switch (DLGGetExitButton(Dlg))
  {
    case DlgOK :
      /* evaluate dialog values: */
      printf ("OK -> Auswertung:\n");
      printf ("Bool: %c\n",    DLGGetBool(Dlg, 16)? 'T':'F');
      printf ("Int:  %d\n",    DLGGetInt(Dlg, 10));
      printf ("Real: %f\n",    DLGGetDouble(Dlg, 11));
      printf ("List: %d\n",    DLGGetList(Dlg, 12));
      printf ("Scroll: %d\n",  DLGGetList(Dlg, 13));
      printf ("Text: %s\n",    DLGGetText(Dlg, 30));
      printf ("FileSel: %s\n", DLGGetFileSel(Dlg, 31));
      printf ("Scale: %f\n",   DLGGetScale(Dlg, 66));
      printf ("Bar: %f\n",     DLGGetBar(Dlg, 67)); 
      printf ("Menu: %d\n",    DLGGetOMenu(Dlg, 40));
      printf ("UpDn: %f\n",    DLGGetUpDn(Dlg, 41));
      Result = *(dual **)DLGGetValue (Dlg, 52);
      printf ("Toggle: ");
      for (count = 0; count < 8; ++count)
	printf ("%c  ", Result[count]? 'T': 'F');
      printf ("\n");
      printf ("Radio: %d\n", DLGGetRBox(Dlg, 55));
      PRO0 ("try to read a non-existing box...")
      if (DLGGetValue(Dlg, 120) == NULL)
	PRO0 ("\\not found\n")
      else
	PRO0 ("\\valid\n")
      break;

    case DlgCANCEL :
      printf ("CANCEL selected\n");
      break;

    default :
      printf ("CLOSED with Command or Window-Close\n");
  }

  /* Speicher des Dialogs wieder freigeben: */
  DLGFreeDialog (Dlg);
}


int StartProc (DIALOG *Dlg, int BoxID)
{
  DLGOpenTextDialog ("callBack", DlgINFO, "Function started", 0);

  return DlgCONFIRM;
}


/* ********************************************************************* */
/* *** Dieser Dialog wurde aus fplot2 geklaut und demonstriert       *** */
/* *** die teilweise komplexen Zusammenhnge der einzelnen Items und *** */
/* *** ihre Realisierung mit der Dlg.Lib                             *** */
/* ********************************************************************* */
int CheckDevice (DIALOG *Dlg, int Button)
{
  if (Button == 1)
  {
    DLGSetBoxActive (Dlg, 71, TRUE);
    if (*(int*) DLGGetValue (Dlg, 71) == 6)
      DLGSetBoxActive (Dlg, 72, TRUE);
  }
  else
  {
    DLGSetBoxActive (Dlg, 71, FALSE);
    DLGSetBoxActive (Dlg, 72, FALSE);
  }

  return DlgCONFIRM;
}


int CheckOther (DIALOG *Dlg, int BoxID)
{
  if (*(int*)DLGGetValue (Dlg, 71) == 5)
    DLGSetBoxActive (Dlg, 72, TRUE);
  else
    DLGSetBoxActive (Dlg, 72, FALSE);

  return DlgCONFIRM;
}

int CheckDistortX (DIALOG *Dlg, int BoxID)
{
  double YFill, Value;

  Value = *(double*)DLGGetValue (Dlg, BoxID);

  YFill = *(double*) DLGGetValue (Dlg, 11);
  if (! DLGGetBool(Dlg, 22))  /* allow distortion */
    DLGSetUpDn (Dlg, 12, YFill * DLGGetBool(Dlg, 42)? 0.7 : 1.0);

  DisplayPos (Dlg, 0);

  return DlgCONFIRM;
}

int CheckDistortY (DIALOG *Dlg, int BoxID)
{
  double XFill, Value;

  Value = *(double*)DLGGetValue (Dlg, BoxID);

  XFill = DLGGetUpDn(Dlg, 12);
  if (! DLGGetBool(Dlg, 22))
    DLGSetUpDn (Dlg, 11, XFill * DLGGetBool(Dlg,42)? (1000/700.0) : 1.0);

  DisplayPos (Dlg, 0);

  return DlgCONFIRM;
}

int CheckCenter (DIALOG *Dlg, int BoxID)
{
  dual Center;

  Center = DLGGetBool (Dlg, BoxID);

  DLGSetBoxActive (Dlg, 34, !Center);
  DLGSetBoxActive (Dlg, 36, !Center);

  return DlgCONFIRM;
}


int CheckAdjust (DIALOG *Dlg, int BoxID)
{
  dual Adjust;

  Adjust = DLGGetBool (Dlg, BoxID);

  if (Adjust)
    DLGSetUpDn (Dlg, 12, DLGGetUpDn(Dlg,11)*1.2);
  else
    DLGSetUpDn (Dlg, 11, DLGGetUpDn(Dlg,12)*1.2*0.7);

  return DlgCONFIRM;
}


int DisplayPos (DIALOG *Dlg, int BoxID)
{
  int   count1, count2;
  int   row, col;
  int   PosX, PosY;
  char *Data = NULL;
 
  Data = calloc (72*100, sizeof(char));

  col = (int)(DLGGetUpDn(Dlg, 11) * 0.72);
  row = (int)DLGGetUpDn (Dlg, 12);

  if (*(dual*)DLGGetValue (Dlg, 41))
  {
    PosX = ( 72-col)/2;
    PosY = (100-row)/2;
  }
  else
  {
    PosX = (int)(*(double*)DLGGetValue (Dlg, 36) * 0.72);
    PosY = (int)*(double*)DLGGetValue (Dlg, 34);
  }

  for (count2 = 0; count2 < 72; ++count2)
  {
    SetBit (&Data, 72, count2, 0);
    SetBit (&Data, 72, count2, 99);
  }
  for (count1 = 0; count1 < 99; ++count1)
  {
    SetBit (&Data, 72,  0, count1);
    SetBit (&Data, 72, 71, count1);
  }

  for (count1 = 0; count1 < row; ++count1)
    for (count2 = 0; count2 < MIN(col, 72-PosX); ++count2)
      SetBit (&Data, 72, PosX+count2, PosY+count1);

  DLGChangePic (Dlg, 37, 72, 100, Data);
  free (Data);

  return DlgCONFIRM;
}


void Parameter (void)
{
  int        count1, count2;
  char      *Data = NULL;
  DIALOG *Dlg = NULL;

  Data = calloc (72*100, sizeof(char));
  for (count2 = 0; count2 < 72; ++count2)
  {
    SetBit (&Data, 72, count2, 0);
    SetBit (&Data, 72, count2, 99);
  }
  for (count1 = 0; count1 < 99; ++count1)
  {
    SetBit (&Data, 72,  0, count1);
    SetBit (&Data, 72, 71, count1);
  }

  for (count1 = 0; count1 < 70 ; ++count1)
    for (count2 = 0; count2 < 57; ++count2)
      SetBit (&Data, 72, 10+count2, 15+count1);

  Dlg = DLGNewDialog (0);
  DLGMakeRows (Dlg,  0, 2, DlgSINGLE_LINE,  5, 30);
  DLGMakeRows (Dlg,  5, 2,     DlgNO_LINE, 10, 20);
  DLGMakeCols (Dlg, 10, 2,     DlgNO_LINE, 11, 12);
  DLGMakeCols (Dlg, 30, 2,     DlgNO_LINE, 31, 40);
#if 0
  DLGMakeCols (Dlg, 31, 2,     DlgNO_LINE, 32, 35);
  DLGMakeRows (Dlg, 32, 2,     DlgNO_LINE, 33, 34);
  DLGMakeRows (Dlg, 35, 2,     DlgNO_LINE, 36, 37);
#else
  DLGMakeGrid (Dlg, 31, 2, 2);
#endif
  DLGMakeRows (Dlg, 40, 3,     DlgNO_LINE, 41, 42, 43);

  DLGInsertUpDn  (Dlg, 11, "xFill", 1.0, 200.0, 80.0, 0.5);
  DLGInsertUpDn  (Dlg, 12, "yFill", 1.0, 200.0, 70.0, 0.5);
  DLGInsertBool  (Dlg, 20, "allow distortion", TRUE);
#if 0
  DLGInsertArrow (Dlg, 33, "x", DlgUP, NULL);
  DLGShowBox     (Dlg, 33, FALSE);
  DLGInsertScale (Dlg, 34, "yOffset", 100.0, 0.0, 0.0, 0, DlgVERTICAL);
  DLGInsertScale (Dlg, 36, "xOffset", 0.0, 100.0, 0.0, 0, DlgHORIZONTAL);
  DLGInsertPic   (Dlg, 37, "page", DlgXBM, 72, 100, Data, NULL);
#else
  DLGInsertScale (Dlg, DlgGRID(31,1,0), 
		       "xOffset", 0.0, 100.0, 0.0, 0, DlgHORIZONTAL);
  DLGInsertScale (Dlg, DlgGRID(31,0,1), 
		       "yOffset", 100.0, 0.0, 0.0, 0, DlgVERTICAL);
  DLGInsertPic   (Dlg, DlgGRID(31,1,1), "page", DlgXBM, 72, 100, Data, NULL);
#endif
  DLGInsertBool  (Dlg, 41, "center", TRUE);
  DLGInsertBool  (Dlg, 42, "adjust", FALSE);
  DLGInsertRBox  (Dlg, 43, "Allignment", 1, 2, "Portrait", "Landscape", 0);


  DLGSetLabelPos (Dlg, 31, DlgNOWHERE, DlgCENTER);
  DLGSetLabelPos (Dlg, 20, DlgNOWHERE, DlgCENTER);
  DLGSetBoxAlignment (Dlg, 20, DlgMIDDLE, DlgUP);
#if 0
  DLGSetLabelPos (Dlg, 37, DlgNOWHERE, DlgCENTER);
  DLGSetLabelPos (Dlg, 33,   DlgNORTH, DlgCENTER);
#endif

  DLGSetTitle (Dlg, 31, "Offset", DlgMIDDLE);

  DLGSetBorder   (Dlg, 31, DlgETCHED_IN);

#if 1
  DLGSetBoxActive (Dlg, DlgGRID(31,0,1), FALSE);
  DLGSetBoxActive (Dlg, DlgGRID(31,1,0), FALSE);
#else
  DLGSetBoxActive (Dlg, 34, FALSE);
  DLGSetBoxActive (Dlg, 36, FALSE);
#endif

  DLGSetFunction (Dlg, 11, CheckDistortX);
  DLGSetFunction (Dlg, 12, CheckDistortY);
  DLGSetFunction (Dlg, 41, CheckCenter);
  DLGSetFunction (Dlg, 42, CheckAdjust);
#if 0
  DLGSetFunction (Dlg, 34, DisplayPos);
  DLGSetFunction (Dlg, 36, DisplayPos);
#endif

  DLGOpenDialog (Dlg, "psPrefs", DlgNORMAL, NULL);

  if (DLGGetExitButton(Dlg) == DlgCANCEL)
    PRGPrintf ("CANCEL selected");

  DLGFreeDialog (Dlg);
  free (Data);
}


int CallParameter (DIALOG *Dlg, int BoxID)
{ 
  DLGSetButtonActive (Dlg, 0, FALSE);
  Parameter ();
  DLGSetButtonActive (Dlg, 0, TRUE);

  return DlgCONFIRM;
}

void PrintDlg (void)
{
  char      **PrtList = NULL;
  DIALOG  *Dlg = NULL;

  Dlg = DLGNewDialog (0);

  DLGMakeRows (Dlg,  0, 2,   DlgETCHED_IN, 10, 50);
  DLGMakeCols (Dlg, 10, 2, DlgSINGLE_LINE, 20, 30);
  DLGMakeRows (Dlg, 20, 2,     DlgNO_LINE, 21, 22);
  DLGMakeRows (Dlg, 30, 3,     DlgNO_LINE, 31, 32, 33);
  DLGMakeCols (Dlg, 50, 2,     DlgNO_LINE, 60, 70);
  DLGMakeRows (Dlg, 70, 2,     DlgNO_LINE, 71, 72);
 
  DLGInsertPic  (Dlg, 21, "icon", DlgXPM, 0, 0, "icons/printer.xpm", NULL);
  DLGInsertBool (Dlg, 22, "preview", TRUE);
  DLGInsertProc (Dlg, 31, "psParam",  StartProc);
  DLGInsertProc (Dlg, 32, "psLayout", CallParameter);
  DLGInsertProc (Dlg, 33, "psAxis",   StartProc);
  DLGInsertRBox (Dlg, 60, "device", 1, 2, "Printer", "File", 0);
  PrtList = (char**)malloc (sizeof(char*) * 6);
  PrtList[0] = strcp ("HP_Deskjet");
  PrtList[1] = strcp ("LaserJet4P");
  PrtList[2] = strcp ("NEC (C)P6");
  PrtList[3] = strcp ("Epson FX80");
  PrtList[4] = NULL;
  PrtList[5] = strcp ("other");
  DLGInsertOMenu (Dlg, 71, "printer", -6, PrtList, 1);
  DLGInsertText (Dlg, 72, "other", 30, "lp", 10, 1);

  DLGSetLabelPos (Dlg, 30, DlgWEST, DlgEAST);

  DLGSetFunction (Dlg, 60, CheckDevice);
  DLGSetFunction (Dlg, 71, CheckOther);

  DLGSetBoxActive (Dlg, 72, FALSE);

  DLGSetPos (Dlg, 150.0, 200.0);
  DLGOpenDialog (Dlg, "PRINT", DlgNORMAL, NULL);

  free (PrtList[0]);
  free (PrtList[1]);
  free (PrtList[2]);
  free (PrtList[3]);
  free (PrtList[5]);
  free (PrtList);

  if (DLGGetExitButton(Dlg) == DlgCANCEL)
    PRGPrintf ("CANCEL selected\n");

  DLGFreeDialog (Dlg);
}


/* ************************************* */
/* *** Aufruf des DLG-Fileselektors: *** */
/* ************************************* */
void FileSel (void)
{
  char FileName[PATH_MAX+1] = "$HOME/project/src/test.c";
  char FileMask[PATH_MAX+1] = "$HOME/project/src/*.c,*.h";

  DLGSetFSPos (450, 200);
  if (DLGOpenFileSel ("selectFile", FileName, PATH_MAX, FileMask) != NULL)
  {
    printf ("File: %s\n", FileName);
    printf ("Mask: %s\n", FileMask);
  }
  else
    PRGPrintf ("FS aborted\n");
}


void FileSel2 (void)
{
  char FileName[PATH_MAX+1] = "$HOME/project/src/test.c";
  char FileMask[PATH_MAX+1] = "$HOME/project/src/*.c,*.h";

  DLGSetFSPos (450, 200);
  if (DLGOpenFileSel2 ("selectFile", FileName, PATH_MAX, FileMask) != NULL)
  {
    printf ("File: %s\n", FileName);
    printf ("Mask: %s\n", FileMask);
  }
  else
    PRGPrintf ("FS aborted\n");
}


void DirSel (void)
{
  char DirName[PATH_MAX+1] = "/tmp";
  char DirMask[PATH_MAX+1] = "/*";

  if (DLGOpenDirSel("SelectDirectory", DirName, PATH_MAX, DirMask) != NULL)
  {
    printf ("Dir : %s\n", DirName);
    printf ("Mask: %s\n", DirMask);
  }
  else
    PRGPrintf ("FS aborted\n");
}


/* **************************************** */
/* *** Unterschiedliche Dialog-Typen *** */
/* **************************************** */

/* *** works only, if NoPPostion of WM is diabled !!! *** */
void UserPosDlg (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);
  DLGInsertLabel (Dlg, 0, "styleID", "UserPos");
  DLGOpenDialog (Dlg, "USERPOS", DlgUSERPOS, NULL);

  DLGFreeDialog (Dlg);
}


void TransientDlg (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);
  DLGInsertLabel (Dlg, 0, "styleID", "Transient");
  DLGOpenDialog (Dlg, "TRANSIENT", DlgTRANSIENT, NULL);

  DLGFreeDialog (Dlg);
}


void NormalDlg (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);
  DLGInsertLabel (Dlg, 0, "styleID", "Normal");
  DLGOpenDialog (Dlg, "DLG_NORMAL", DlgNORMAL, NULL);

  DLGFreeDialog (Dlg);
}


void MousePosDlg (void)
{
  DIALOG *Dlg = NULL;

  Dlg = DLGNewDialog (0);
  DLGInsertLabel (Dlg, 0, "styleID", "MousePos");
  DLGOpenDialog (Dlg, "MOUSEPOS", DlgMOUSEPOS, NULL);

  DLGFreeDialog (Dlg);
}


/* ****************************************** */
/* *** bersetzung mittels Resource-Datei *** */
/* ****************************************** */
void Translate (void)
{
  PRGPrintf ("Translation of \"Welt\" is: %s\n", 
	     DLGGetResourceEntry ("WorldTest", "Welt"));
}


/* ************* */
/* *** Hilfe *** */
/* ************* */
void Help (void)
{
  HELPOpenHelpWin ("HELPWINDOW", "$HOME/project/help/dialog/libdlg.hlp", 
		                 0, 0, 450, 300, NULL, 0, NULL);
}


  /* Column 1: System */
  static MENU_ENTRY MenuCol1[] = 
  {
    /* name,       mnemon., sh-key-txt, short-key, function-ptr */
    ENTRY ("Exit", "", "", ' ', TRUE, ProgramExit, NULL)
  };

  /* Column 2: Dialog */
  static MENU_ENTRY MenuCol2[] = 
  {
    ENTRY ("PtrSplit",    "", "", ' ', TRUE, PtrSplit,    NULL),
    ENTRY ("PtrDlg",      "", "", ' ', TRUE, PtrDlg,      NULL),
    ENTRY ("SetFkt",      "", "", ' ', TRUE, ParamCall,   NULL),
    ENTRY ("LabelPos",    "", "", ' ', TRUE, LabelPos,    NULL),
    ENTRY ("ChangeValue", "", "", ' ', TRUE, ChangeValue, NULL),
    ENTRY ("Space",       "", "", ' ', TRUE, ControlDlg,  NULL),
    ENTRY ("Adornment",   "", "", ' ', TRUE, AdornDlg,    NULL),
    ENTRY ("Sensitive",   "", "", ' ', TRUE, SensDlg,     NULL),
    ENTRY ("ActiveDlg",   "", "", ' ', TRUE, ActiveDlg,   NULL),
    ENTRY ("ActiveB",     "", "", ' ', TRUE, ActiveB,     NULL),
    ENTRY ("HideDlg",     "", "", ' ', TRUE, HideDlg,     NULL),
    ENTRY ("AutoClose",   "", "", ' ', TRUE, AutoClose,   NULL),
    ENTRY ("CB-Dlg",      "", "", ' ', TRUE, CallBackDlg, NULL),
    ENTRY ("ColorTest",   "", "", ' ', TRUE, ChangeColor, NULL),
    ENTRY ("ButtonTest",  "", "", ' ', TRUE, ButtonTest,  NULL),
    ENTRY ("GridTest",    "", "", ' ', TRUE, GridTest,    NULL),
    ENTRY ("ArrayTest",   "", "", ' ', TRUE, ArrayTest,   NULL),
    ENTRY ("KeepPos",     "", "", ' ', TRUE, KeepPos,     NULL),
    ENTRY ("ValidTest",   "", "", ' ', TRUE, CheckDlg,    NULL)
  };

  /* Column 3: Boxtypes */
  static MENU_ENTRY MenuCol3[] = 
  {
    ENTRY ("LABEL",           "", "", ' ', TRUE, LabelItem,  NULL),
    ENTRY ("BOOL",            "", "", ' ', TRUE, BoolItem,   NULL),
    ENTRY ("INT",             "", "", ' ', TRUE, IntItem,    NULL),
    ENTRY ("DOUBLE",          "", "", ' ', TRUE, DoubleItem, NULL),
    ENTRY ("TEXT",            "", "", ' ', TRUE, TextItem,   NULL),
    ENTRY ("SCALE",           "", "", ' ', TRUE, ScaleItem,  NULL),
    ENTRY ("BAR",             "", "", ' ', TRUE, BarItem,    NULL),
    ENTRY ("UPDN",            "", "", ' ', TRUE, UpDnItem,   NULL),
    ENTRY ("LIST",            "", "", ' ', TRUE, ListItem,   NULL),
    ENTRY ("OMENU",           "", "", ' ', TRUE, MenuItem,   NULL),
    ENTRY ("TBOX",            "", "", ' ', TRUE, TBoxItem,   NULL),
    ENTRY ("RBOX",            "", "", ' ', TRUE, RBoxItem,   NULL),
    ENTRY ("PROC",            "", "", ' ', TRUE, ProcItem,   NULL),
    ENTRY ("ARROW",           "", "", ' ', TRUE, ArrowItem,  NULL),
    ENTRY ("PIC",             "", "", ' ', TRUE, PicItem,    NULL),
    ENTRY ("PICB",            "", "", ' ', TRUE, PicBItem,   NULL),
    ENTRY ("IMAGE",           "", "", ' ', TRUE, ImageItem,  NULL),
    ENTRY ("FSEL",            "", "", ' ', TRUE, FSelItem,   NULL),
#if (XmVERSION>=2)
    ENTRY ("NOTE",            "", "", ' ', TRUE, NoteItem,   NULL),
    ENTRY ("TREE",            "", "", ' ', TRUE, TreeItem,   NULL),
#endif
    ENTRY ("DIAL",            "", "", ' ', TRUE, DialItem,   NULL),
    ENTRY ("SEGMENT",         "", "", ' ', TRUE, SegmItem,   NULL),
    ENTRY ("USER",            "", "", ' ', TRUE, UserItem,   NULL)
  };

  /* Column 4: Demos */
  static MENU_ENTRY MenuCol4[] = 
  {
    ENTRY ("Complete",        "", "", ' ', TRUE, Complete,   NULL),
    ENTRY ("Printer",         "", "", ' ', TRUE, PrintDlg,   NULL),
    ENTRY ("Parameter",       "", "", ' ', TRUE, Parameter,  NULL),
    ENTRY ("BMPTest",         "", "", ' ', TRUE, XBMChange,  NULL),
    ENTRY ("FormTest",        "", "", ' ', TRUE, FormTest,   NULL),
    ENTRY ("WidgetTest",      "", "", ' ', TRUE, WidgetTest, NULL)
    
  };

  /* Submenus of 5.th column */
  static MENU_ENTRY TextDialogMenu[] = 
  {
    ENTRY ("Mouse-Dialog",    "", "",  ' ', TRUE, Dialog0, NULL),
    ENTRY ("Error-Dialog",    "", "",  ' ', TRUE, Dialog1, NULL),
    ENTRY ("Info-Dialog",     "", "",  ' ', TRUE, Dialog2, NULL),
    ENTRY ("Msg-Dialog",      "", "",  ' ', TRUE, Dialog3, NULL),
    ENTRY ("Question-Dialog", "", "",  ' ', TRUE, Dialog4, NULL),
    ENTRY ("Warn-Dialog",     "", "",  ' ', TRUE, Dialog5, NULL),
    ENTRY ("Work-Dialog",     "", "",  ' ', TRUE, Dialog6, NULL)
  };

  static MENU_ENTRY PicDialogMenu[] = 
  {
    ENTRY ("Symbol-Box(0)",   "", "",  ' ', TRUE, Dialog7,  NULL),
    ENTRY ("Symbol-Box(1)",   "", "",  ' ', TRUE, Dialog8,  NULL),
    ENTRY ("Symbol-Box(2)",   "", "",  ' ', TRUE, Dialog9,  NULL),
    ENTRY ("Symbol-Box(3)",   "", "",  ' ', TRUE, Dialog10, NULL)
  };

  /* Column 5: Message-Boxen */
  static MENU_ENTRY MenuCol5[] = 
  {
    SUB_MENU ("TextDialogs",  "", "", ' ', TRUE, TextDialogMenu),
    SUB_MENU ("PicDialogs",   "", "", ' ', TRUE, PicDialogMenu)
  };

  /* Column 6: FileSelektor */
  static MENU_ENTRY MenuCol6[] = 
  {
    ENTRY ("Select File",     "", "",  ' ', TRUE, FileSel,  NULL),
    ENTRY ("Select File2",    "", "",  ' ', TRUE, FileSel2, NULL),
    ENTRY ("Select Dir",      "", "",  ' ', TRUE, DirSel,   NULL)
  };

  /* Column 7: Dialog-Styles */
  static MENU_ENTRY MenuCol7[] = 
  {
    ENTRY ("UserPos",         "", "",  ' ', TRUE, UserPosDlg,  NULL),
    ENTRY ("Transient",       "", "",  ' ', TRUE, TransientDlg, NULL),
    ENTRY ("Normal",          "", "",  ' ', TRUE, NormalDlg,   NULL),
    ENTRY ("MousePos",        "", "",  ' ', TRUE, MousePosDlg, NULL)
  };

  /* Column 8: Translation */
  static MENU_ENTRY MenuCol8[] =
  {
    ENTRY ("Translate",       "", "", ' ', TRUE, Translate, NULL)
  };

  static MENU_ENTRY MenuCol9[] =
  {
    ENTRY     ("helpWindow",      "", "",  ' ', TRUE, Help,        NULL),
    MENU_LINE (PrgETCHED_IN),
    ENTRY     ("info",            "", "", ' ',  TRUE, ProgramInfo, NULL)
  };

  /* Menu-columns */
  static MENU_COLUMN MenuCols[] = 
  {
    FUNC_MENU ("System",       ' ', TRUE, MenuCol1),
    FUNC_MENU ("Dialog",       ' ', TRUE, MenuCol2), 
    FUNC_MENU ("Boxen",        ' ', TRUE, MenuCol3), 
    FUNC_MENU ("Demo",         ' ', TRUE, MenuCol4), 
    FUNC_MENU ("Dialog",       ' ', TRUE, MenuCol5),
    FUNC_MENU ("FileSelector", ' ', TRUE, MenuCol6),
    FUNC_MENU ("OpenStyles",   ' ', TRUE, MenuCol7),
    FUNC_MENU ("Translation",  ' ', TRUE, MenuCol8),
    HELP_MENU ("help",         ' ', TRUE, MenuCol9)
  };

  static MENU_BAR Menu[] =
  {
    MAIN_MENU ("tstDlg", MenuCols)
  };

/*----------------------------------------------------------------------
			 H A U P T P R O G R A M M
  ----------------------------------------------------------------------*/
int main (int argc, char *argv[])
{
  int count = 1;

  PRGInitLib (argv[0], &argc, argv, NULL);

  while (count < argc)
  {
    if (strncmp(argv[count], "-cmd", 4) == 0)
    {
	 count++;
	 if (count == argc)
	   exit (0);
	 Param.PrgName = argv[count];
    }
    count++;
  }

  PRGCreateProgram (argv[0], Menu);

  PRGStartProgram (InitPrg, NULL);

  exit (0);
  return 0;
} 
