#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "memlist.h"

#define MEM_FKT_LEN 32
#define MEM_MSG_LEN 128
#define MEM_VAL_LEN 35

#define MEM_ID_LEN 5
#define MEM_ID "DEBUG"

#ifndef MIN
#define MIN(a,b) (((a)<(b)) ? (a):(b))
#endif

#define MEM_LIB_MSG(Function, FileName, LineNo, CallFkt, Msg)\
        { \
	  fprintf (stderr, "ERROR: %s\n", Function); \
	  fprintf (stderr, "       called from %s[%d]: %s\n", \
		           FileName, LineNo, CallFkt); \
	  fprintf (stderr, "%s\n", Msg); \
	}

#define MEM_LIB_ERROR(Function, FileName, LineNo, CallFkt, Msg)\
        { \
	  MEM_LIB_MSG (Function, FileName, LineNo, CallFkt, Msg); \
	  strncpy (ErrorMsg, Msg, MEM_MSG_LEN); \
	  ErrorMsg[MEM_MSG_LEN] = '\0'; \
	  strncpy (ErrorFkt, CallFkt, MEM_FKT_LEN); \
	  ErrorFkt[MEM_FKT_LEN] = '\0'; \
	  StatusFlag = 2; \
	  MEMExitLib (); \
	}


typedef struct
{
  char   *Memory;
  size_t  MemSize;
  char    FileName[MEM_FKT_LEN+1];
  int     LineNo;
} MEM_ENTRY;


/* Prototypes: */
static  void    MEM_HandleSignal   (int);
static  void    MEM_HandleSegv     (int);
static  int     MEM_CmpMemPtr      (void *Ptr1, void *Ptr22);
static  int     MEM_TestPtr        (char *Ptr,  size_t size);

        void    MEMExitLib         (void);
        void    MEMFree            (char   *FileName, int   LineNo, 
				    char   *CallFkt,  void *Memory);
        size_t  MEMFreeAllMemory   (size_t *MemSize);


static int      StatusFlag = 0;
static char     ErrorMsg[MEM_MSG_LEN+1];
static char     ErrorFkt[MEM_FKT_LEN+1];
static size_t   MemCount = 0;
static MEMLIST *MemList = NULL;
static MEMCELL *ActCell = NULL;


static void MEM_HandleSegv (int Signal)
{
  fprintf (stderr, "*** FATAL SEGMENTATION - ERROR\n");
  exit (Signal);
}


static void MEM_HandleSignal (int Signal)
{
  char *reason = "UNKNOWN_SIGNAL";

  switch (Signal)
  {
    case   SIGHUP: reason = "HANGUP"; 
                   signal (SIGHUP, MEM_HandleSignal);
                   break;
    case   SIGINT: MEMExitLib ();
                   break;
    case  SIGQUIT: reason = "quit (ASCII FS)";   
                   signal (SIGQUIT, MEM_HandleSignal);
                   break;
    case   SIGILL: reason = "ILLEGAL INSTRUCTION (NOT RESET WHEN CAUGHT)"; 
                   break;
    case  SIGTRAP: reason = "TRACE TRAP (NOT RESET WHEN CAUGHT)"; 
                   break;
    case  SIGABRT: reason = "USED BY ABORT, REPLACE SIGIOT IN THE FUTURE"; 
                   signal (SIGABRT, MEM_HandleSignal);
                   break;
    case   SIGFPE: reason = "FLOATING POINT EXCEPTION"; 
                   signal (SIGFPE, MEM_HandleSignal);
                   break;
    case  SIGKILL: reason = "KILL (CANNOT BE CAUGHT OR IGNORED)"; 
                    break;
    case   SIGBUS: reason = "BUS ERROR"; 
                   signal (SIGBUS, MEM_HandleSignal);
                   break;
    case  SIGSEGV: reason = "SEGMENTATION VIOLATION"; 
                   signal (SIGSEGV, MEM_HandleSegv);   
                   break;
    case  SIGPIPE: reason = "WRITE ON A PIPE WITH NO ONE TO READ IT"; 
                   break;
    case  SIGALRM: reason = "ALARM CLOCK"; 
                   break;
    case  SIGTERM: reason = "SOFTWARE TERMINATION SIGNAL FROM KILL";
                   signal (SIGTERM, MEM_HandleSignal);
                   break;
    case  SIGUSR1: reason = "USER DEFINED SIGNAL 1"; 
                   break;
    case  SIGUSR2: reason = "USER DEFINED SIGNAL 2"; 
                   break;
    case   SIGPWR: reason = "POWER-FAIL RESTART"; 
                   break;
    case SIGWINCH: reason = "WINDOW CHANGE"; 
                   break;
    case   SIGCLD: reason = "DEATH OF A CHILD"; 
                   break; 
         default : fprintf (stderr, "UNKNOWN SIGNAL: %d\n", Signal);
  }
  
  StatusFlag = 2;
  fprintf (stderr, "\n");
  fprintf (stderr, "INTERRUPT: Signal: %s\n", reason);

  MEMExitLib ();
}


static int MEM_TestPtr (char *Memory, size_t MemSize)
{
  if (strncmp(Memory, MEM_ID, MEM_ID_LEN) != 0)
    return -1;

  Memory += MemSize - MEM_ID_LEN;
  if (strncmp(Memory, MEM_ID, MEM_ID_LEN))
    return 1;

  return 0;
}


static int MEM_CmpMemPtr (void *Ptr1, void *Ptr2)
{
  return (((MEM_ENTRY *)Ptr1)->Memory == ((MEM_ENTRY *)Ptr2)->Memory);
}


static char MEM_Printable (char letter)
{
  char Result = ' ';

  if ((letter > 0x1f && letter < 0x7f) ||
      letter == '' || letter == '' || letter == '' || 
      letter == '' || letter == '')
    Result = letter;

  return Result;
}


void MEM_CheckMemoryList (void)
{
  unsigned int  count;
  char          letter;
  char         *Memory = NULL;
  size_t        MemSize, Size;
  MEMCELL      *CurCell = NULL;

  if (MemList == NULL)
  {
    fprintf (stderr, "Empty Memory-list");
    return;
  }

  fprintf (stderr, "\n");
  fprintf (stderr, "M e m o r y - C h e c k:\n");
  fprintf (stderr, "========================\n\n");

  if ((CurCell = MemList->Start) != NULL)
  {
    fprintf (stderr, "File            #line:    size   reason: data\n");
    fprintf (stderr, "-----------------------------------------------"
	             "-----------------------------\n");

    while (CurCell != NULL)
    {
      Memory  = (((MEM_ENTRY *)CurCell->ValPtr)->Memory) - MEM_ID_LEN;
      Size = ((MEM_ENTRY *)CurCell->ValPtr)->MemSize;
      MemSize = Size + 2 * MEM_ID_LEN;

      fprintf (stderr, "%-15s #%4d: [%6d] ",
      	               ((MEM_ENTRY *)CurCell->ValPtr)->FileName,
	               ((MEM_ENTRY *)CurCell->ValPtr)->LineNo, Size);
      switch (MEM_TestPtr(Memory, MemSize))
      {
        case -1 :
	  fprintf (stderr, "  (BOM): ");
	  break;
        case 1 :
	  fprintf (stderr, "  (EOM): ");
	  break;
        default :
	  fprintf (stderr, "(ALLOC): ");
      }

      /* Print contents of allocated memory */
      for (count = 0; count < MIN(Size,MEM_VAL_LEN); ++count)
      {
	letter = ((MEM_ENTRY *)CurCell->ValPtr)->Memory[count];
	fprintf (stderr, "%c", MEM_Printable(letter));
      }
      fprintf (stderr, "\n");

      CurCell = CurCell->Next;
    }

    fprintf (stderr, "-----------------------------------------------"
	             "-----------------------------\n");
    fprintf (stderr, "END-OF-MEMORY-CHECK\n");
    StatusFlag = 2;
  }
  else
    fprintf (stderr, "NO MEMORY ALLOCATED\n");
}


/* -------------- MEMORY-Management ------------------- */
void MEMInitLib (void)
{
  if (StatusFlag != 0)
  {
    fprintf (stderr, "MEM-Library has been already initialized");
    return;
  }

  signal (SIGHUP,  MEM_HandleSignal);
  signal (SIGINT,  MEM_HandleSignal);
  signal (SIGQUIT, MEM_HandleSignal);
  signal (SIGBUS,  MEM_HandleSignal);
  signal (SIGFPE,  MEM_HandleSignal);
  signal (SIGABRT, MEM_HandleSignal);
  signal (SIGTERM, MEM_HandleSignal);
  signal (SIGSEGV, MEM_HandleSignal);

  StatusFlag = 1;
}


void MEMExitLib (void)
{
  size_t remain, last = 0;

  if (StatusFlag == 0)
  {
    fprintf (stderr, "MEM-Library has not been initialized\n");
    return;
  }

  signal (SIGHUP,  SIG_DFL);
  signal (SIGINT,  SIG_DFL);
  signal (SIGQUIT, SIG_DFL);
  signal (SIGBUS,  SIG_DFL);
  signal (SIGFPE,  SIG_DFL);
  signal (SIGABRT, SIG_DFL);
  signal (SIGTERM, SIG_DFL);
  signal (SIGSEGV, SIG_DFL);

  if (StatusFlag == 2)
    fprintf (stderr, "#\n"
	             "#----------------------------------\n"
	             "#           MEM-Library\n"
	             "#----------------------------------\n"
	             "#\n");

  MEM_CheckMemoryList ();
  remain = MEMFreeAllMemory (&last);

  if (StatusFlag == 2)
  { 
#if 0
    fprintf (stderr, "ERROR in Function %s:\n", Error.Function);
    if (Error.Msg[0] != '\0')
      fprintf (stderr, "  %s\n\n", Error.Msg);
#endif
    fprintf (stderr, "Memory-Check: Before cleaning = %d\n", last);
    fprintf (stderr, "              After cleaning =  %d\n", remain);
    fprintf (stderr, "READY\n\n");
  }

  exit (StatusFlag - 1);
}


void *MEMMalloc (char *FileName, int LineNo, char *CallFkt, size_t MemSize)
{
  MEM_ENTRY MemEntry;

  if (StatusFlag == 0)
    MEMInitLib ();

  if ((MemEntry.MemSize = MemSize) == 0)
  {
    fprintf (stderr, "libmem: allocating 0 bytes");
    return NULL;
  }

  if ((MemEntry.Memory = (char *)malloc(MemEntry.MemSize + 
			                sizeof(char)*(2*MEM_ID_LEN))) == NULL)
    MEM_LIB_ERROR (__FUNCTION__, FileName, LineNo, CallFkt, 
		   "       -------------- MEMORY EXHAUSTED -----------\n"
		   "       -------------------------------------------\n"
		   "       ----- NO MORE MEMORY CAN BE ALLOCATED -----\n"
		   "       --------------- PROGRAM ABORTED -----------\n")

  strncpy (MemEntry.Memory, MEM_ID, MEM_ID_LEN);
  MemEntry.Memory += MEM_ID_LEN;
  strncpy (MemEntry.Memory + MemEntry.MemSize, MEM_ID, MEM_ID_LEN);

  strncpy (MemEntry.FileName, FileName, MEM_FKT_LEN);
  MemEntry.FileName[MEM_FKT_LEN] = '\0';
  MemEntry.LineNo = LineNo;

  if (MemList == NULL)
  {
    MemList = MEM_InitList (sizeof(MEM_ENTRY));
    MEM_InsertList (MemList, NULL, &MemEntry);
    ActCell = MemList->Start;
  }
  else
    MEM_InsertList (MemList, MemList->End, &MemEntry);

  MemCount += MemEntry.MemSize;

  return (MemEntry.Memory);
}


void *MEMCalloc (char *FileName, int LineNo, char *CallFkt,
		 size_t Count, size_t Size)
{
  char *Memory = NULL;

  if (StatusFlag == 0)
    MEMInitLib ();

  if (Size == 0 || Count == 0)
  {
    fprintf (stderr, "nothing to allocate");
    return NULL;
  }

  Memory = (char *)MEMMalloc (FileName, LineNo, CallFkt, Size * Count);
  memset (Memory, '\0', Count * Size);

  return (Memory);
}


void *MEMRealloc(char *FileName, int LineNo, char *CallFkt,
		 void *OldMemPtr, size_t NewMemSize)
{
  MEM_ENTRY  NewMemEntry;
  MEMCELL   *CurCell = NULL;

  if (StatusFlag == 0)
    MEMInitLib ();

  if (NewMemSize == 0)
  {
    MEMFree (FileName, LineNo, CallFkt, OldMemPtr);
    fprintf (stderr, "New Size is 0 -> freeing");
    return NULL;
  }

  if (OldMemPtr == NULL)
    return MEMMalloc (FileName, LineNo, CallFkt, NewMemSize);

  NewMemEntry.Memory  = (char *)OldMemPtr;
  NewMemEntry.MemSize = NewMemSize;
  strncpy (NewMemEntry.FileName, FileName, MEM_FKT_LEN);
  NewMemEntry.FileName[MEM_FKT_LEN] = '\0';
  NewMemEntry.LineNo = LineNo;

  CurCell = MEM_GetListItem (MemList, &NewMemEntry, ActCell, MEM_CmpMemPtr);

  if (CurCell == NULL)
    MEM_LIB_ERROR (__FUNCTION__, FileName, LineNo, CallFkt,
		   "       No Memory allocated at this adress")

  if (!(NewMemEntry.Memory = 
	   (char *)realloc(NewMemEntry.Memory - MEM_ID_LEN, 
				    NewMemEntry.MemSize + 2*MEM_ID_LEN)))
    MEM_LIB_ERROR (__FUNCTION__, FileName, LineNo, CallFkt, 
		   "       -------------- MEMORY EXHAUSTED -----------\n"
		   "       -------------------------------------------\n"
		   "       ----- NO MORE MEMORY CAN BE ALLOCATED -----\n"
		   "       --------------- PROGRAM ABORTED -----------\n")
    
  /* Es findet eine Prfung der Randbereiche statt.                     */
  strncpy (NewMemEntry.Memory, MEM_ID, MEM_ID_LEN);
  NewMemEntry.Memory += MEM_ID_LEN;
  strncpy (NewMemEntry.Memory + NewMemEntry.MemSize, MEM_ID, MEM_ID_LEN);

  MemCount -= ((MEM_ENTRY *)CurCell->ValPtr)->MemSize;
  ActCell = MEM_DeleteList (MemList, CurCell);
  MEM_InsertList (MemList, MemList->End, &NewMemEntry);
  MemCount += NewMemEntry.MemSize;

  return (NewMemEntry.Memory);
}


void MEMFree (char *FileName, int LineNo, char *CallFkt, void *Memory)
{
  int        status;
  MEM_ENTRY  MemEntry;
  MEM_ENTRY *Ptr = NULL;
  MEMCELL   *CurCell = NULL;

  if (StatusFlag == 0)
    MEMInitLib ();

  if (MemList == NULL)
  {
    MEM_LIB_MSG (__FUNCTION__, FileName, LineNo, CallFkt,
			  "       No memory allocated yet")
    return;
  }

  if (Memory == NULL)
  {
    MEM_LIB_MSG (__FUNCTION__, FileName, LineNo, CallFkt,
			  "       Pointer is NULL")
    return;
  }

  MemEntry.Memory  = (char *)Memory;
  MemEntry.MemSize = 0;

  CurCell = MEM_GetListItem (MemList, &MemEntry, ActCell, MEM_CmpMemPtr);
   
  /* 
     Problem: The original C-memory-functions will be overwritten
              by the functions in this library. This method works fine
	      while the memory will be completely managed by these functions.
              But some Standarg-C-functions, e.g. scandir will allocate
	      memory using the original C-library-function 'malloc'. 
	      When trying to free this range using the modified 'free' 
              (MEMFree) this area will not be recognised !
  */
  if (CurCell == NULL)
    if (strncmp((char*)Memory, MEM_ID, MEM_ID_LEN) == 0)
      MEM_LIB_ERROR (__FUNCTION__, FileName, LineNo, CallFkt,
		     "       No memory allocated at this adress")
    else
      free ((char*)Memory);
  else
  {
    Ptr = (MEM_ENTRY *)CurCell->ValPtr;
    if ((status = MEM_TestPtr(Ptr->Memory - MEM_ID_LEN, 
			      Ptr->MemSize + 2*MEM_ID_LEN)) != 0) 
    {
      fprintf (stderr, "FREE-ERROR %d: "__FUNCTION__"\n", status);
      fprintf (stderr, "               Range Error\n");
      fprintf (stderr, "               allocation at Pos: %s[%d]\n",
	               Ptr->FileName, Ptr->LineNo);
      fprintf (stderr, "            free       at Pos: %s[%d]\n", 
	               FileName, LineNo);
      MEMExitLib ();
    }
    else
    {
      free ((char*)Memory - MEM_ID_LEN);
      MemCount -= Ptr->MemSize;
      ActCell = MEM_DeleteList (MemList, CurCell);
    }
  }
}


void MEMExit (char *FileName, int LineNo, char *CallFkt, int status)
{
  MEM_LIB_MSG (__FUNCTION__, FileName, LineNo, CallFkt, "Exiting")

  if (StatusFlag != 0)
    MEMExitLib ();
  else
    exit (status);
}
 

size_t MEMCountMemory (void)
{
  return MemCount;
}


size_t MEMFreeAllMemory (size_t *MemSize)
{
  if (MemList == NULL)
    return 0;

  *MemSize = MemCount;

  ActCell = MemList->Start;

  while (ActCell != NULL)
  {
    free ((((MEM_ENTRY *)ActCell->ValPtr)->Memory) - MEM_ID_LEN);
    MemCount -= ((MEM_ENTRY *)ActCell->ValPtr)->MemSize;
    ActCell = MEM_DeleteList (MemList, ActCell);
  }

  return MemCount;
}



