#include <stdlib.h>
#include <string.h>

#include "memlist.h"

#ifndef elsif
#define elsif else if
#endif


MEMLIST *MEM_InitList (int ElemSize) 
{
  MEMLIST *Result = NULL;

  Result = (MEMLIST *)calloc (1, sizeof(MEMLIST));
  Result->Start   = NULL;
  Result->End     = NULL;
  Result->ValSize = ElemSize;

  return Result;
}


MEMCELL *MEM_InsertList (MEMLIST *List, MEMCELL *ElemPtr, void *Value)
{
  MEMCELL *Result;

  Result = (MEMCELL *)calloc (1, sizeof(MEMCELL));
  Result->ValPtr = calloc (List->ValSize, 1);
  memcpy (Result->ValPtr, Value, List->ValSize);
      
  if (ElemPtr != NULL) 
  {
    if (ElemPtr->Next != NULL)
      ElemPtr->Next->Prev= Result;
    else
      List->End = Result;

    Result->Next  = ElemPtr->Next;
    ElemPtr->Next = Result;
    Result->Prev   = ElemPtr;
  } 
  else 
  {
    Result->Next = List->Start;
    Result->Prev  = NULL;
    List->Start   = Result;

    if (Result->Next != NULL)
      Result->Next->Prev = Result;
    else 
      List->End = Result;
  }
  
  return Result;
}


MEMCELL *MEM_DeleteList (MEMLIST *List, MEMCELL *ElemPtr)
{
  MEMCELL *Next1 = NULL;
  MEMCELL *Next2 = NULL;

  if (ElemPtr != NULL)  
  {
    if ((Next1 = ElemPtr->Next) != NULL)
      ElemPtr->Next->Prev = ElemPtr->Prev;
    else
      List->End = ElemPtr->Prev;
    if ((Next2 = ElemPtr->Prev) != NULL)
      ElemPtr->Prev->Next = ElemPtr->Next;
    else
      List->Start = ElemPtr->Next;
    
    free (ElemPtr->ValPtr);
    free (ElemPtr);
    return (Next1 == NULL)? Next2 : Next1;
  }
  return NULL;
}


MEMCELL *MEM_GetListItem (MEMLIST *List, void *Data, MEMCELL *StartElem,
			  int (*CmpFunc)(void *, void *))
{
  MEMCELL *CurCell = NULL;
  MEMCELL *LeftCell = NULL; 
  MEMCELL *RightCell = NULL;
  int      DirFlag = 0;

  CurCell = (StartElem == NULL)? List->Start : StartElem;
  
  if (CurCell != NULL) 
  {
    LeftCell  = CurCell->Prev;
    RightCell = CurCell->Next;
    
    while (CurCell != NULL && !(*CmpFunc)(Data, CurCell->ValPtr))
    {
      DirFlag ^= 1;

      if (LeftCell != NULL && (DirFlag || RightCell == NULL)) 
      {
	CurCell = LeftCell;
	LeftCell = LeftCell->Prev;
      } 
      elsif (RightCell != NULL) 
      {
	CurCell = RightCell;
	RightCell = RightCell->Next;
      } 
      else 
	CurCell = NULL;
    }
    
  }
  return CurCell;
}
