/* ********************************************************************** */
/* ***                           Net-Library                          *** */
/* *** Submitting Client-/Server-applications with socket-connections *** */
/* ********************************************************************** */

#ifndef LIBNET_H
#define LIBNET_H

#if 1
/* controls using of a thread based socket-server. */
#define NET_THREADS
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef NET_THREADS
#include <pthread.h>
#endif

#include <stdarg.h>

#include <debug.h>


#define NET_IP_ADDR1(addr) ((addr & 0xff000000) >> 24)
#define NET_IP_ADDR2(addr) ((addr & 0x00ff0000) >> 16)
#define NET_IP_ADDR3(addr) ((addr & 0x0000ff00) >>  8)
#define NET_IP_ADDR4(addr)  (addr & 0x000000ff)

#define NET_PROTOCOL         AF_INET
#define NET_TYPE             SOCK_STREAM
#define SOCKET_BUF_SIZE      4096
/* Amount of Protocol bytes (message-length) preceeding each message */
#define NET_LEN_BYTES        0
#define NET_MAX_CONNECTIONS  5
#define NET_MAX_SERVER       10

typedef char *(*SERVER_FUNCTION)(int, char*);


/* ********************************************************************** */
/* ** The used SocketIDs are different from the system's socket-call:  ** */
/* ** They representate a index into a internal-array. The usage       ** */
/* ** of this socket-id's provides the following advantages:           ** */
/* **   + Controlls, if a socket is open/close (will be set to -1, if  ** */
/* **     closed)                                                      ** */
/* **   + All open sockets could be closed by a single function        ** */
/* **     (NETCloseAllServers)                                         ** */
/* **   + Controlls, if a socket-id is valid (range 0..max)            ** */
/* ********************************************************************** */

/* Client-Functions: */
int         NETInitClient       (const char *HostName, int PortNo);
dual        NETSendCommand      (int SocketID, char *Msg);
char       *NETSend             (int SocketID, char *Msg, ...);
dual        NETCloseClient      (int SocketID);

/* Server-Functions: */
int         NETInitServer       (const char *HostName, int PortNo);
int         NETRunServer        (int SocketID, SERVER_FUNCTION Function, 
				 int MaxConnections);
dual        NETCloseServer      (int SocketID);
void        NETCloseAllServers  (void);

#ifdef NET_THREADS
pthread_t  *NETStartServer      (int SocketID, SERVER_FUNCTION Function, 
				 int MaxConnections);
dual        NETStopServer       (pthread_t *NetThread);
#endif

#ifdef __cplusplus
}
#endif

#endif
