#include <math.h>

#include "libobj.h"


#define NO_ISECT           0
#define COPLANAR_ISECT     1
#define TRANSVERSAL_ISECT  2

#define INSIDE            -1
#define COMMON             0
#define OUTSIDE            1


/* Help structure containing the intersection structure and additional 
   particulary informations */
typedef struct
{
  INTERSECT *Intersect;
  int        Side;
  int        Index;
} CONTACT;


typedef struct SUBPOLY
{
         int      Depth;
         POLYGON *Polygon;
  struct SUBPOLY *Next;
} SUBPOLY;


#if 0
typedef struct
{ 
  OBJECT    *Obj;
  OBJECT    *Obj1, *Obj2;
  INTERSECT *List;
} CSG_DATA;
#endif


#if 0
/* Returns the index (number) of the given intersection in the polygon's
   edge by counting the intersections which have a lower lambda value. */
static int OBJ_GetIndex (INTERSECT *List, POLYGON *Poly, 
			 int Index, double Lambda)
{
  int        count, Side;
  int        Result = 1;
  INTERSECT *Ptr = List;

  while (Ptr != NULL)
  {
    for (Side = 0; Side < 2; ++Side)
    {
      if (Ptr->Poly[Side] == Poly)
	for (count = 0; count < Ptr->Number[Side]; ++count)
	  if (Ptr->Index[Side][count] == Index && 
	      Ptr->Lambda[Side][count] < Lambda)
	    ++Result;
    }
    Ptr = Ptr->Next;
  }

  return Result;
}
#endif


/* Search for unhandled polygons in the intersection list */
static INTERSECT *OBJ_UnhandledIntersection (INTERSECT *List)
{
  INTERSECT *Ptr = List;
  
  while (Ptr != NULL && Ptr->done)
    Ptr = Ptr->Next;

  return Ptr;
}


static POLYGON *OBJ_AddPoint (OBJECT *Obj, POLYGON *Poly, POINT3D Point)
{
  int Index;

  Index = OBJAddVertex (Obj, Point);

  if (Poly == NULL)
  {
printf ("New Polygon\n");
    Poly = GEONewPolygon ();
 }

  /* Update pointer to vertex list due to changes when calling realloc */
  Poly->Points = Obj->VertexList; 

  Poly->Vertex = realloc (Poly->Vertex, sizeof(int) * (Poly->Number + 1));
  Poly->Vertex[(Poly->Number)++] = Index;

  /* TODO: PRO4 () */
printf ("    Add #%d: %.3f / %.3f / %.3f\n", 
	Poly->Number, Point.x, Point.y, Point.z);

  return Poly; 
}


static void OBJ_PrintList (INTERSECT *List, const char *Name)
{
  int        count, Side;
  POINT3D    Min, Max;  
  INTERSECT *Ptr = List;

  printf ("%s:\n", Name);

  while (Ptr != NULL)
  {
    printf ("  %d: Processed: %s\n", Ptr->No, Ptr->done? "YES":"NO");
    for (Side = 0; Side < 2; ++Side)
    {
      GEOGetPolygonBounds (Ptr->Poly[Side], &Min, &Max);
      printf ("  %c Polygon-Bounds(ID%d):"
	      " (%.2f %.2f %.2f) - (%.2f %.2f %.2f)\n",
	      'A'+Side, Ptr->Poly[Side]->Info.ID, 
	      Min.x, Min.y, Min.z, Max.x, Max.y, Max.z);    
    }

    for (Side = 0; Side < 2; ++Side)
      for (count = 0; count < Ptr->Number[Side]; ++count)
	printf ("    %c%d: Edge=%d: %.3f / %.3f / %.3f Lambda=%.4f\n", 
		'A'+Side, count+1, Ptr->Index[Side][count],
		Ptr->Points[Side][count].x,
		Ptr->Points[Side][count].y,
		Ptr->Points[Side][count].z,
		Ptr->Lambda[Side][count]);
    printf ("\n");

    Ptr = Ptr->Next;
  }

  /* ... and now sorted for sides */
  for (Side = 0; Side < 2; ++Side)
  {
    printf ("Object %d\n", Side+1);
    Ptr = List;
    while (Ptr != NULL)
    {
      if (Ptr->Number[Side] > 0)
      {
	printf ("  Polygon %d\n", Ptr->Poly[Side]->Info.ID);
	for (count = 0; count < Ptr->Number[Side]; ++count)
	  printf ("    %d: #%d = %.3f / %.3f / %.3f\n", 
		  count+1, Ptr->Index[Side][count],
		  Ptr->Points[Side][count].x,
		  Ptr->Points[Side][count].y,
		  Ptr->Points[Side][count].z);
      }
      Ptr = Ptr->Next;
    }
  }
}


/* This routine decides if a point lies eighter inside (-1), outside (+1) the
   other polyeder or if it is part of the intersection */
static int OBJ_GetSide (INTERSECT *List, POLYGON *Polygon, POINT3D Point)
{
  int         SignCount = 0;
  int        Side, count, Result;
  double     Dist;
  POLYGON   *Poly;
  INTERSECT *Ptr = List;

  PRO_START (" ")

  while (Ptr != NULL)
  {
    /* Find out the corresponding polygon in the other polyeder */
    if (Ptr->Poly[0] == Polygon)
      Poly = Ptr->Poly[1];
    elsif (Ptr->Poly[1] == Polygon)
      Poly = Ptr->Poly[0];
    else
      Poly = NULL;

    if (Poly != NULL)
    {
      /* First, check relation to polygon in other polyeder */
      if ((Dist = GEOAreaPointDist(Poly, Point)) > 0.0)
	SignCount++;
      else	                        /* Do we have a polygon point */
	for (Side = 0; Side < 2; ++Side)
	  for (count = 0; count < Ptr->Number[Side]; ++count)
	    if (VEC_EQUAL(Ptr->Points[Side][count], Point))
	      PRO_RETURN ("Intersection point", COMMON)
    }

    Ptr = Ptr->Next;
  }

  Result = (SignCount > 0)? OUTSIDE:INSIDE;

  PRO_END
  return Result;
}


static INTERSECT *OBJ_GetIntersectionList (OBJECT *Object1, OBJECT *Object2)
{
  int         count = 0;
  POINT3D    Min1, Max1;
  POINT3D    Min2, Max2;
  POLYGON   *Poly1, *Poly2;
  INTERSECT *List = NULL, *Intersect;

  PRO_START ("\n")

  /* The calculation needs only to be done, if the bouding boxes of both
     objects do intersect. */
  if (! GEOIntersectBounds(Object1->Info.Min, Object1->Info.Max,
			   Object2->Info.Min, Object2->Info.Max))
    PRO_RETURN ("No intersection", NULL)

#if 1
/* TODO: PRO6 () */
printf ("Obj-Bounds1: (%.2f %.2f %.2f) - (%.2f %.2f %.2f)\n",
	Object1->Info.Min.x, Object1->Info.Min.y, Object1->Info.Min.z,
	Object1->Info.Max.x, Object1->Info.Max.y, Object1->Info.Max.z);
Poly1 = Object1->Data.Polyeder->List;
while (Poly1 != NULL) 
{
  GEOGetPolygonBounds (Poly1, &Min1, &Max1);
  printf (" Poly %d: (%.2f %.2f %.2f) - (%.2f %.2f %.2f)\n",
	  Poly1->Info.ID, Min1.x, Min1.y, Min1.z, Max1.x, Max1.y, Max1.z);
  Poly1 = Poly1->Next;
}
printf ("Obj-Bounds2: (%.2f %.2f %.2f) - (%.2f %.2f %.2f)\n",
	Object2->Info.Min.x, Object2->Info.Min.y, Object2->Info.Min.z,
	Object2->Info.Max.x, Object2->Info.Max.y, Object2->Info.Max.z);
Poly2 = Object2->Data.Polyeder->List;
while (Poly2 != NULL) 
{
  GEOGetPolygonBounds (Poly2, &Min2, &Max2);
  printf (" Poly %d: (%.2f %.2f %.2f) - (%.2f %.2f %.2f)\n",
	  Poly2->Info.ID, Min2.x, Min2.y, Min2.z, Max2.x, Max2.y, Max2.z);
  Poly2 = Poly2->Next;
}
  count = 0;
#endif

  /* Initialize polygon flags */ 
  Poly1 = Object1->Data.Polyeder->List;
  while (Poly1 != NULL)
  {
    Poly1->Flags = 0;
    Poly1 = Poly1->Next;
  }

  Poly2 = Object2->Data.Polyeder->List;
  while (Poly2 != NULL)
  {
    Poly2->Flags = 0;
    Poly2 = Poly2->Next;
  }

  /* Compute polygon intersections for all combinations */
  Poly1 = Object1->Data.Polyeder->List;
  while (Poly1 != NULL)
  {
    GEOGetPolygonBounds (Poly1, &Min1, &Max1);

    Poly2 = Object2->Data.Polyeder->List;
    while (Poly2 != NULL)
    {
      GEOGetPolygonBounds (Poly2, &Min2, &Max2);

      /* Test for intersection of polygons bounding boxes */
      if (GEOIntersectBounds(Min1, Max1, Min2, Max2))
      {
        if ((Intersect = GEOPolygonIntersection(Poly1, Poly2)) != NULL)
	{
	  Intersect->No = count++;
          Intersect->Next = List;
	  List = Intersect;
	}
      }

      Poly2 = Poly2->Next;
    }
    Poly1 = Poly1->Next;
  }

  PRO_END
  return List;
}


/* Search for an intersection in the edge from index Index to Index+1.
   Lambda is a value between 0.0 (Index) and 1.0 (Index+1) and describes
   the position of the intersection. If there are more than one intersection, 
   the Lambda value will be interpreted as follows:
           -1: Look for an intersection which lies inside the opposite polygon
          0.0: Search for an intersection with a MINIMUM lambda value
          1.0: Search for an intersection with a MAXIMUM lambda value
   0<Lambda<1: Search the intersection with the next larger lambda value
*/
static CONTACT OBJ_FindIntersection (INTERSECT *List, POLYGON *Poly,
				     int Index, double Lambda)
{
  int        count, Side;
  dual       first = TRUE;
  INTERSECT *Ptr = List;
  CONTACT    Result = {NULL, -1, -1};

  /* Search for an interior intersection */
  if (Index == -1)
  {
    while (Ptr != NULL)
    {
      if (! Ptr->done)
	for (Side = 0; Side < 2; ++Side)
	{
	  if (Ptr->Poly[Side] == Poly && Ptr->Number[Side] == 0)
	  {
	    Result.Intersect = Ptr;
	    Result.Side      = 1-Side;      /* Contact is in other polygon */
	    Result.Index     = 0;
	    return Result;
	  }
	}
      Ptr = Ptr->Next;
    }
    return Result;
  }

  /* Search for an edge-intersection */
  while (Ptr != NULL)
  {
    for (Side = 0; Side < 2; ++Side)
    {
      if (Ptr->Poly[Side] == Poly)
      {
	for (count = 0; count < Ptr->Number[Side]; ++count)
	{
	  if (Ptr->Index[Side][count] == Index)
	  {
	    if (first && (Lambda == 0.0 || Lambda == 1.0 || 
			  Ptr->Lambda[Side][count] > Lambda))
            {                              /* Initialize first occurence */
	      Result.Intersect = Ptr;
	      Result.Side      = Side;
	      Result.Index     = count;
	      first = FALSE;
	    }
	    elsif ((Lambda == 0.0 &&             /* find smallest lambda */
		    Ptr->Lambda[Side][count] <
		    Result.Intersect->Lambda[Result.Side][Result.Index]) ||
		   (Lambda == 1.0 &&                 /* find last lambda */
		    Ptr->Lambda[Side][count] >
		    Result.Intersect->Lambda[Result.Side][Result.Index]) ||
		   (Lambda > 0.0 && Lambda < 1.0 &&  /* find next lambda */
		    Ptr->Lambda[Side][count] > Lambda &&
		    Ptr->Lambda[Side][count] < 
		    Result.Intersect->Lambda[Result.Side][Result.Index]))
	    {
	      Result.Intersect = Ptr;
	      Result.Side      = Side;
	      Result.Index     = count;
	    }
	  }
	}
      }
    }
    Ptr = Ptr->Next;
  }

  return Result;
}			    


/* The function follows a given intersection point. First we search the 
   corresponding entry in the intersection list, The other values of this
   entry will be added to a new polygon. At the end of each entry we
   search for the next list entry. This action will be continued until
   we reach the starting polygon again. While propagating the intersections
   we add the points into the new polygon.
*/
static CONTACT OBJ_TraceIntersection (OBJECT *Obj, POLYGON *Polygon,
				      INTERSECT *List, POLYGON *Poly,
				      POINT3D Point)
{
  dual       found = FALSE; 
  int        count;
  int        count1 = 0, Index = 0;
  int        Side, Side1 = -1;
  INTERSECT *Ptr = List;
  CONTACT    Contact = {NULL, -1, -1};

  PRO_START ("\n")

  if (List == NULL)
    PRO_RETURN ("Empty List", Contact)

  /* Find an entry in the intersection list with the given polygon (Poly) and
     the defined intersection point (Point) */
  while (Ptr != NULL && ! found)
  {
    if ((! Ptr->done) && (Ptr->Poly[0] == Poly || Ptr->Poly[1] == Poly))
    {
      Side1 = 0;
      while (Side1 < 2 && ! found)
      {
 	for (count1 = 0; count1 < Ptr->Number[Side1]; ++count1)
	  if ((found = (GEOVectorDist(Ptr->Points[Side1][count1],Point)<1e-5)))
	    break;
	if (! found)
	  ++Side1;
      }
    }
    if (! found)
      Ptr = Ptr->Next;
  }

  if (! found || Ptr == NULL)
  {
printf ("ERROR: Polygon not found / Trace finished\n");
    PRO_RETURN ("Polygon not found", Contact)
  }
printf ("  FOUND: %d  Lambda=%.4f\n", Ptr->No, Ptr->Lambda[Side1][count1]);

  /* 
     First, process the remaining points of the first entry (without the 
     current point), then switch to the other side and then process all 
     points of this entry.
  */
  Side = Side1;
  for (count = 1; count < Ptr->Number[Side]; ++count)
  {
    Index = (count1 + count) % Ptr->Number[Side];
    OBJ_AddPoint (Obj, Polygon, Ptr->Points[Side][Index]);
  }
  Side = 1 - Side;                      /* switch to other side ...         */
  if (Ptr->Number[Side] > 0)
  {                                     /* and now process the second entry */
    for (Index = 0; Index < Ptr->Number[Side]; ++Index)
      OBJ_AddPoint (Obj, Polygon, Ptr->Points[Side][Index]);
    Index = Ptr->Number[Side]-1;        /* same as Index--                  */
  }
  else
    Side = Side1;

  Ptr->done = TRUE;                     /* Mark contact as processed        */

  if (Ptr->Poly[Side] != Poly)
    Contact = OBJ_TraceIntersection (Obj, Polygon, List, Poly, 
				     Ptr->Points[Side][Index]);
  else
  {
    Contact.Intersect = Ptr;
    Contact.Side      = Side;
    Contact.Index     = Index;
  }

  PRO_END
  return Contact;
}


static int OBJ_GetDepth (POLYGON *List, POLYGON *Poly)
{
  int      count = 0;
  POINT3D  Point;
  POLYGON *Ptr = List;

  if (Poly->Number == 0)
    return 0;

  Point = Poly->Points[Poly->Vertex[0]];
  while (Ptr != NULL)
  {
    if (Ptr != Poly && GEOPointInPolygon(Ptr, Point))
      count++;
    Ptr = Ptr->Next;
  }

#if 1
  printf ("Depth: %.3f %.3f %.3f = %d\n", Point.x, Point.y, Point.z, count);
#endif
 
  return count;
}


/* Search for a polygon with depth 1 respective to the given polygon */
/* If the given Polygon is NULL, this function returns the first polygon */
/* with depth zero. */
static POLYGON *OBJ_FindPolygon (POLYGON **List, POLYGON *Poly)
{
#if 0
  int      Depth;
#endif
  POLYGON *Ptr = *List, *Last = NULL;

  /* TODO:
     Setzt voraus, dass in der Liste die bereits bearbeiteten Polygon gelscht
     wurden.
  */

#if 0
  if (Poly == NULL)
#endif
    while (Ptr != NULL)
    {
      if (OBJ_GetDepth(*List, Ptr) == 0)
      {
	/* Remove this polygon from list */
	if (Last == NULL)
	  *List = Ptr->Next;
	else
	  Last->Next = Ptr->Next;

	return Ptr;
      }
      else
	Last = Ptr;
      Ptr = Ptr->Next;
    }
#if 0
  else
  {
    Depth = OBJ_GetDepth(*List, Ptr);
    while (Ptr != NULL)
    {
      if (Ptr != Poly && OBJ_GetDepth(*List, Ptr) == Depth+1 &&
	  OBJ_GetDepth(Poly, Ptr) > 0)
	return Ptr;
      Ptr = Ptr->Next;
    }
  }
#endif

  return NULL;
}


/* Polygon enthlt bereits ein das uere (Kopie oder Zeiger auf 
   aus Intersection-List entferntes Polygon. Suche nun rekursiv alle direkten
   Unterpolygone und fge sie als Lcher in das Polygon ein. Entferne
   dabei die bearbeiten Polygon aus der Intersection Liste. Werden keine
   direkten Subpolygone gefunden, wird das Polygon selbst zurckgegeben. */
/* Ausserhalb wird diese Routine mit allen ueren Polygonen aufgerufen, bis
   es keine Polygone mehr gibt */
static POLYGON *OBJ_DoInclude (OBJECT *Obj, POLYGON *Polygon,
			       POLYGON **List, int Side, int Op)
{
  int      count, count1, count2;
  int      Index1, Index2;
  int      Dist, Min;
  int      Depth;
  POINT3D  Point;
  POLYGON *Ptr = *List;
  POLYGON *Last = NULL, *Next;
  POLYGON *Result = Polygon, *New;

printf ("DO INCLUDE: %c\n", (Polygon!=NULL)? 'Y':'N');

  while (Ptr != NULL)
  {
/* FEHLER: BERCKSICHTIGT NICHT EIN EVTL. BERGEBENES POLYGON, D.H. NUR 
           INNERHALB DIESES POLYGONS NACH NCHSTEN SUBPOLYGON(EN) SUCHEN. 
           ERST RELEVANT, WENN MEHRERE SUBPOLYGON MIT GLEICHER TIEFE EX. */

    Depth = OBJ_GetDepth (*List, Ptr);
    if (Depth < 1)                     /* Do we have a direct subpolygon */
    {
      Index1 = 0;
      Index2 = 0;

      /* Search the vertices in the polygon and the hole with the minimum 
         distance to each other. This is neccessary, because if we have more
         than one subpolygons the result is *not* convex. */
      if (Result != NULL)
      {
	Min = GEOVectorDist (Result->Points[Result->Vertex[0]],
			     Ptr->Points[Ptr->Vertex[0]]);
	for (count1 = 0; count1 < Result->Number; ++count1)
	{
	  Point = Result->Points[Result->Vertex[count1]];
	  for (count2 = 0; count2 < Ptr->Number; ++count2)
	  {
	    Dist = GEOVectorDist (Point, Ptr->Points[Ptr->Vertex[count2]]);
	    if (Dist < Min)
	    {
	      Min = Dist;
	      Index1 = count1;
	      Index2 = count2;
	    }
	  }
	}
      }

      /* Add the hole's vertices + two vertices for the connection */
      New = GEONewPolygon ();

      if (Result != NULL)
	for (count = 0; count <= Index1; ++count)
	{
	  Result->Points = Obj->VertexList;
	  OBJ_AddPoint (Obj, New, Result->Points[Result->Vertex[count]]);
	}

      for (count = 0; count <= Ptr->Number; ++count)
      {
	Ptr->Points = Obj->VertexList;
	OBJ_AddPoint (Obj, New, 
	              Ptr->Points[Ptr->Vertex[(Index2+count) % Ptr->Number]]);
      }

      if (Result != NULL)
      {
	for (count = Index1; count < Result->Number; ++count)
	{
	  Result->Points = Obj->VertexList;
	  OBJ_AddPoint (Obj, New, Result->Points[Result->Vertex[count]]);
	}

        /* Remove the old result and store the new one. */
	GEOFreePolygon (Result, FALSE);
      }
      Result = New;
    }
    Ptr = Ptr->Next;
  }

  /* Now remove all processed subpolygons from the list. */
  Ptr = *List;
  while (Ptr != NULL)
  {
    Next = Ptr->Next;
    
/* FEHLER: BEIM LSCHEN DER BEARBEITETEN EINTRGE NDERT SICH DIE TIEFE DER
           RESTLICHEN SUBPOLYGONE -> (BAUM-)DATENSTRUKTUR ? */

    Depth = OBJ_GetDepth (*List, Ptr);
/* NDERN WIE OBEN (NUR INNERHALB DES POLYGONS) */
    if (Depth < 1)
    {
      if (Last == NULL)
	*List = Next;
      else
	Last->Next = Next;

      GEOFreePolygon (Ptr, FALSE);
      return Result;
    }
    else
      Last = Ptr;
 
    Ptr = Next;
  }

  return Result;
}


/* Process contacts within the polygon
   We have to distinguish several different cases depending of the objects
   no and the operation:
   - Intersection:
      1st object: polygon has same orientation as base polygon
      2nd object: same as intesection of 1st object
   - Difference:
      1st object: create a hole, that means a complementary polygon with 
                  opposite orientation
      2nd object: create a polygon with same orientation as outer polygon
   - Union:
      1st object: create a complementary polygon with same orientation
      2nd object: like 1st object of union
*/
static POLYGON *OBJ_MakeIncludes (OBJECT *Obj, POLYGON *Polygon,
				  INTERSECT *List, POLYGON *Poly, int Side,
				  int Op)
{
  dual       first = TRUE;
  dual       Direction;
  int        count, Index = 0;
  int        Depth;
  double     Dist, Min = 0.0;
  double     Angle;
  POINT3D    Point;
  CONTACT    Contact;
  INTERSECT *Intersect;
  POLYGON   *Sub, *Next;
  POLYGON   *Sub1 = NULL, *Sub2;
  POLYGON   *Ptr = NULL;
  POLYGON   *Result = NULL;

  PRO_START ("\n")
  /* TODO:
     Unterscheidung, bei mehreren verschachtelten Intersections
  */

  printf ("INCLUDE: Polygon exists: %s\n", (Polygon==NULL)? "NO":"YES");

  /* Search all interior intersections of this polygon */
  do
  {
    Sub = NULL;
    Contact = OBJ_FindIntersection (List, Poly, -1, 0.0);
    if (Contact.Intersect != NULL)
    {
printf ("NEW SUB-POLYGON...\n");
      Sub = GEONewPolygon ();
      while ((Intersect = Contact.Intersect) != NULL)
      {
	Intersect->done = TRUE;
	Sub = OBJ_AddPoint (Obj, Sub,
			    Intersect->Points[Contact.Side][Contact.Index]);
	Contact = OBJ_TraceIntersection (Obj, Sub, List, Poly,
	            Intersect->Points[Contact.Side][Contact.Index]);
      }
      if (Sub->Number < 3)
	GEOFreePolygon (Sub, FALSE);
      else
      {
     	Sub->Points = Obj->VertexList;
	GEOTriangulatePolygon (Sub);
	GEOCalcHesseForm      (Sub);

	/* Add this sub-polygon to the list */
	Sub->Next = Ptr;
	Ptr = Sub;
      }
printf ("...SUB-POLYGON FINISHED\n");
    }
  } while (Sub != NULL);

  if ((Sub = Ptr) == NULL)
    PRO_RETURN ("No inclusion found", NULL)

  /* Update all sub-polygons' vertex-pointer */
  while (Sub != NULL)
  {
    Sub->Points = Obj->VertexList;
    Sub = Sub->Next;
  };
  Sub = Ptr;

  Direction = (Op == CSG_UNION || (Op == CSG_DIFFERENCE && Side == 0));

#if 1
  /* Swap orientation of all polygons with odd/even depth (depends on 
     direction) */
  while (Sub != NULL)
  {
    Depth = OBJ_GetDepth (Ptr, Sub);
    Angle = GEOVectorAngle (Poly->Info.n, Sub->Info.n);
printf ("DEPTH: %d  Angle: %.3f  Dir: %c\n", Depth, Angle, Direction? 'Y':'N');
    if ((Depth & 1) != Direction && Angle > 90.0)
      GEOSwapOrientation (Sub);

    Sub = Sub->Next;
  }
printf ("OK\n");
  Sub = Ptr;

  if (Direction)
  {
    /* In this case the first included polygon is counter clockwise and so
       we need the (surrounding) original polyon. If it does not exist, we
       have to copy it. */
    if ((Sub = Polygon) == NULL)
    {
printf ("Copy Polygon\n");
      Sub = GEONewPolygon ();
      for (count = 0; count < Poly->Number; ++count)
	OBJ_AddPoint (Obj, Sub, Poly->Points[Poly->Vertex[count]]);
    }
    else
      Sub->Points = Obj->VertexList;

    GEOSwapOrientation (Sub);
  }
  else
    Sub = NULL;

  /* Process all internal intersections of this polygon */
  do
  {
    if (Sub != NULL)
    {
      Sub = OBJ_DoInclude (Obj, Sub, &Ptr, Side, Op); 

      Sub->Next = Result;
      Result = Sub;
    }

printf ("SEARCH NEXT SUBPOLYGON\n");
    Sub = OBJ_FindPolygon (&Ptr, NULL);
  } while (Sub != NULL);
      
#if 0
    if ((Sub2 = OBJ_FindPolygon(Ptr, Sub1)) != NULL)
    {
      if (Direction)
	GEOSwapOrientation (Sub2);
    }
    Direction = ! Direction;

    if (Sub1 != NULL && Sub2 != NULL)
    {
      /* Search hole-vertex with minimum distance to last polygon vertex */
      Point = Sub1->Points[Sub1->Vertex[Sub1->Number-1]];
      for (count = 0; count < Sub2->Number; ++count)
      {
	Dist = GEOVectorDist (Point, Sub2->Points[Sub2->Vertex[count]]);
	if (first || Dist < Min)
	{
	  Index = count;
	  Min   = Dist;
	  first = FALSE;
	}
      }

      /* Add the hole's vertices + two vertices for the connection */
      for (count = 0; count <= Sub2->Number; ++count)
      {
	Sub1 = OBJ_AddPoint (Obj, Sub1,
		    Sub2->Points[Sub2->Vertex[(Index+count) % Sub2->Number]]);
      	Sub2->Points = Obj->VertexList;
      }
      Result = OBJ_AddPoint (Obj, Sub1, Point);

      GEOFreePolygon (Sub2, FALSE);

      /* TODO:
	 Sub2 aus Liste lschen und Sub1 an Ergebnisliste anhngen */
    }
    else
    {
      Sub1 = Sub2;
    }
  } while (Sub2 != NULL);
#endif
#endif


#if 0
/* Funktionsfhig fr eine Verschachtelungstiefe */
  while (Sub != NULL)
  {
    Next = Sub->Next;

    OBJ_GetDepth (Ptr, Sub);
    /* TODO:
       Odd-Even-Regel anwenden -> Hole oder Subpoly, ggf. zwei benachbarte
       verschachtelte Polygone zu einem Ring erzeugen:
       Untere Formel gilt nur fr Depth = 0 (Spezialfall von even)
       Beachten: Mehrere eigenstndige Verschachtelungen mglich
                 -> zuerst alle auf Ebene 0 bearbeiten, ab dann aber 
		    nur die jeweils in einem Subpolygon weiter untersuchen
       Erste Idee:
       Bestimme das nchste (Sub-)Polygon (evtl. auch das Original-Polygon)
       gem der auszufhrenden Op und Side. Suche nach allen Sub-Polygonen, 
       die in diesem Polygn liegen und fge sie (invers) hinzu, indem zuvor
       ein Punkt mit krzestem Abstand gesucht wird. Rekursiv fortsetzen.
    */

    Angle = GEOVectorAngle (Poly->Info.n, Sub->Info.n);
    if ((Op == CSG_INTERSECTION && Angle > 90.0) ||
	(Op == CSG_DIFFERENCE   && Angle < 90.0) ||
	(Op == CSG_UNION        && Angle < 90.0))
      GEOSwapOrientation (Sub);

    if (Op == CSG_INTERSECTION || (Op == CSG_DIFFERENCE && Side == 1))
    {
printf ("SUB-POLY\n");
      /* Add the sub polygon */
      Sub->Info.Center = GEOCalcPolygonCenter (Sub);
      GEOTriangulatePolygon (Sub);
      Result = Sub;
    }
    elsif (Op == CSG_UNION || (Op == CSG_DIFFERENCE && Side == 0))
    {
printf ("HOLE-POLY\n");
      /*
	Create a hole in the current polygon using the sub polygon. If there's
	no polygon computed yet, make a copy of the object's polygon and 
	insert the hole. 
      */
      /*
	TODO:
	Wenn die Behandlung mehrerer Intersections eingefgt wird, darf hier
	keine zweite Kopie angefertigt werden. 
      */
      if ((Result = Polygon) == NULL)
      {
printf ("COPY POLYGON FOR INTERIOR INTERSECTION\n");
         Result = GEONewPolygon ();
	 for (count = 0; count < Poly->Number; ++count)
	   OBJ_AddPoint (Obj, Result, Poly->Points[Poly->Vertex[count]]);
      }
      else
	Result->Points = Obj->VertexList;

      /* Search hole-vertex with minimum distance to last polygon vertex */
      Point = Result->Points[Result->Vertex[Result->Number-1]];
      for (count = 0; count < Sub->Number; ++count)
      {
	Dist = GEOVectorDist (Point, Sub->Points[Sub->Vertex[count]]);
	if (first || Dist < Min)
	{
	  Index = count;
	  Min   = Dist;
	  first = FALSE;
	}
      }

      /* Add the hole's vertices + two vertices for the connection */
      for (count = 0; count <= Sub->Number; ++count)
      {
	Result = OBJ_AddPoint (Obj, Result,
			Sub->Points[Sub->Vertex[(Index+count) % Sub->Number]]);
      	Sub->Points = Obj->VertexList;
	/* TODO: 
	   Muss fr alle Sub-Polygone gemacht werden, da diese nicht in der
	   Polygonliste des objektes eingehngt sind und sonst automatisch
	   aktualisiert werden wrden. */
      }
      Result = OBJ_AddPoint (Obj, Result, Point);

      GEOFreePolygon (Sub, FALSE);

      if (Polygon != NULL)
	Result = NULL;
    }
    else
      printf ("INVALID PARAMETER\n");

    Sub = NULL;
  }
#endif

  PRO_END
  return Result;
}


/* This routine computes the intersection polygon which is defined by the
   given Intersection (Polygon and Contact). Flags contains the information 
   about processed vertices in the polygon */
static POLYGON *OBJ_SubPoly (OBJECT *Obj, INTERSECT *List, 
			     INTERSECT *Intersect, dual *Flags,
			     int Mode, int Side, int Index)
{
  dual     Stop = FALSE;
  int      count, Start;
  double   Lambda;
  POINT3D  Point;
  POLYGON *Poly;
  POLYGON *Polygon = NULL;
  CONTACT  Contact;

  PRO_START ("\n")

  if (Intersect == NULL)
    PRO_RETURN ("Empty intersection", NULL)

  if (Intersect->Number[Side] == 0)
  {
    /*    Intersect->done = TRUE; */
    PRO_RETURN ("Internal intersection", NULL)
  }

  /* Save some start values */
  Poly   = Intersect->Poly[Side];
  count  = Intersect->Index[Side][Index];
  Start  = count;
  Lambda = Intersect->Lambda[Side][Index];

  /* Initalize the first contact */
  Contact.Intersect = Intersect;
  Contact.Side      = Side;
  Contact.Index     = Index;

  do
  {
    if ((Intersect = Contact.Intersect) != NULL)
    {                                   /* Let's process the intersection */
      Point = Intersect->Points[Contact.Side][Contact.Index];
      Polygon = OBJ_AddPoint (Obj, Polygon, Point);    /* Add first point */
      /* Follow the intersection until we reach another contact           */
      printf ("-->\n");
      Contact = OBJ_TraceIntersection (Obj, Polygon, List, Poly, Point);
      printf ("<--\n");
      if ((Intersect = Contact.Intersect) == NULL)
      {
	if (Polygon->Number < 3)
	{
	  GEOFreePolygon (Polygon, FALSE);
	  Polygon = NULL;
	}
	PRO_RETURN ("ERROR: Intersection not complete", Polygon)
      }

      /* Check, if we have reached the start intersection */
      count = Intersect->Index[Contact.Side][Contact.Index];
      Stop = (count == Start &&
	      Intersect->Lambda[Contact.Side][Contact.Index] <= Lambda);
      if (! Stop)
	Contact = OBJ_FindIntersection (List, Poly, count,
			      Intersect->Lambda[Contact.Side][Contact.Index]);
    }
    else
    {
      count = (count + 1) % Poly->Number;  /* increase vertex-no            */
      if (! Flags[count])                  /* Vertex was not processed, yet */
      {
	Point = Poly->Points[Poly->Vertex[count]];
	/* Check, if the vertex has desired state (inside/outside) */
	if (OBJ_GetSide(List, Poly, Point) == Mode)
	{
	  /* Add vertex to the polygon, mark it as processed and search for
             the *next* intersection in the following edge (Parameter 0.0) */
	  Polygon = OBJ_AddPoint (Obj, Polygon, Point);
	  Flags[count] = TRUE;
	  Contact = OBJ_FindIntersection (List, Poly, count, 0.0);
	}
      }
      Stop =
	(count == Start && 
	 (Contact.Intersect == NULL ||
	  Contact.Intersect->Lambda[Contact.Side][Contact.Index] <= Lambda));
    }
  } while (! Stop);

  PRO_END
  return Polygon;
}


static CONTACT OBJ_GetContact (INTERSECT *List, POLYGON *Poly, int Index, 
			       int No)
{
  int        count, Side;
  int        count1, count2;
  int        Number = 0;
  INTERSECT *Ptr = List;
  CONTACT   *Contacts = NULL;
  CONTACT    Contact1, Contact2, Swap;
  CONTACT    Contact = {NULL, -1, -1};

  while (Ptr != NULL)
  {
    /* Collect all contacts of this polygon's edge */
    if (! Ptr->done)
      for (Side = 0; Side < 2; ++Side)
	if (Ptr->Poly[Side] == Poly)
	  for (count = 0; count < Ptr->Number[Side]; ++count)
	    if (Ptr->Index[Side][count] == Index)
	    {
	      Contacts = (CONTACT *)realloc (Contacts, 
					     sizeof(CONTACT) * (Number+1));
	      Contacts[Number].Intersect = Ptr;
	      Contacts[Number].Side      = Side;
	      Contacts[Number].Index     = count;
	      ++Number;
	    }
    Ptr = Ptr->Next;
  }

  if (Number >= No)
  {
    /* Sort the collected contacts. Use bubble-sort because normaly we have
       one or two intersections */
    for (count1 = 0; count1 < Number-1; ++count1)
    {
      /* Sort contacts by ascending lambda-values */
      Contact1 = Contacts[count1];
      for (count2 = Number-1; count2 > count1; count2--)
      {
	Contact2 = Contacts[count2];
	if (Contact1.Intersect->Lambda[Contact1.Side][Contact1.Index] >
	    Contact2.Intersect->Lambda[Contact2.Side][Contact2.Index])
	{
	  Swap             = Contacts[count1];
	  Contacts[count1] = Contacts[count2];
	  Contacts[count2] = Swap;
	}
      }
    }

    /* Select the desired contact */
    Contact = Contacts[No-1];
  }
  
  free (Contacts);

  return Contact;
}
    

/*
  Process one polygon. 
*/
static POLYGON *OBJ_MakePoly (OBJECT *Obj, INTERSECT *List, POLYGON *Poly,
 			      int Op, int Mode, int Side)
{
  int        count = 0, No;
  int        Type;
  dual      *Flags;
  POINT3D   *Points;
  POLYGON   *Polygon, *Ptr;
  POLYGON   *Result = NULL;
  INTERSECT *Intersect;
  CONTACT    Contact;

  PRO_START ("\n")

  if (Poly == NULL || Poly->Number == 0)
    PRO_RETURN ("Empty polygon", NULL)

  /* Mark all vertices as not processed */
  Flags = (dual *)malloc (sizeof(dual) * Poly->Number);
  memset (Flags, FALSE, sizeof(dual) * Poly->Number);

  Points = Poly->Points;

printf ("POLYGON %d\n", Poly->Info.ID);
  /* Process all polygon's edges */
  for (count = 0; count < Poly->Number; ++count)
  {
    Type = OBJ_GetSide (List, Poly, Points[Poly->Vertex[count]]);
    No = (Type == Mode)? 1:2;

    do
    {
      Contact = OBJ_GetContact (List, Poly, count, No);
      if ((Intersect = Contact.Intersect) != NULL)
      {
        if ((Polygon = OBJ_SubPoly(Obj, List, Intersect, Flags, Mode,
				   Contact.Side, Contact.Index)) != NULL)
	{
	  Polygon->Next = Result;
	  Result = Polygon;
	}
	No = 2;
      }
    } while (Intersect != NULL);      
  }

  free (Flags);

  Polygon = OBJ_MakeIncludes (Obj, Result, List, Poly, Side, Op);
  if ((Ptr = Polygon) != NULL)
  {
    while (Ptr->Next != NULL)
      Ptr = Ptr->Next;
    Ptr->Next = Result;
    Result = Polygon;
  }

  PRO_END
  return Result;
}


/* Find adjacent polygons and add them to the polyeder */
static void OBJ_Propagate (OBJECT *Object, OBJECT *Obj, POLYGON *Poly, 
			   dual Orientation)
{
  int      count;
  int      Start, End, Step;
  POLYGON *Polygon = NULL, *Neighbour;

  if (Poly->Flags == 1)
    return;

  Poly->Flags = 1;
  Polygon = GEONewPolygon ();

  if (Orientation)
  {
    Start = 0;
    End   = Poly->Number;
    Step  = 1;
  }
  else
  {
    Start = Poly->Number-1;
    End   = -1;
    Step  = -1;
  }

  count = Start;
  while (count != End)
  {
    OBJ_AddPoint (Object, Polygon, Poly->Points[Poly->Vertex[count]]);
    Neighbour = GEOFindNeighbour (Obj->Data.Polyeder, Poly, 
				  count, (count+1) % Poly->Number);
    if (Neighbour != NULL && Neighbour->Flags == 0)
      OBJ_Propagate (Object, Obj, Neighbour, Orientation);

    count += Step;
  }

  Polygon->Points      = Object->VertexList;
  Polygon->Info.ID     = Object->Data.Polyeder->Number;
  Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);
  GEOAddPolygon (Object->Data.Polyeder, Polygon);
}


/* Usually, not all polygons of both objects do intersect each other. So, some 
   of them lie completely inside the other object. The intersection list only 
   contains polygons which directly intersects with one or more polygons
   of the other object. This procedure finds the adjacent polygons of an
   object. Depending of the operation (Intersection/Difference/Union) we
   must search for polygons inside or outside the intersection and possibly
   swap the order of the polygons vertexes (which depends on the side, from
   which can be seen.
*/
static void OBJ_AddPolygons (OBJECT *Object, INTERSECT *List, OBJECT *Obj,
			     int Mode, int Side, dual Orientation)
{
  int        count;
  int        State[2];
  POLYGON   *Poly, *Neighbour;
  INTERSECT *Ptr = List;

  while (Ptr != NULL)
  {
    Ptr->Poly[Side]->Flags = 1;
    Ptr = Ptr->Next;
  }
  
  Ptr = List;
  while (Ptr != NULL)
  {
    Poly = Ptr->Poly[Side];

    State[1] = OBJ_GetSide (List, Poly, Poly->Points[Poly->Vertex[0]]);
    for (count = 0; count < Poly->Number; ++count)
    {
      State[0] = State[1];
      State[1] = 
        OBJ_GetSide (List, Poly, 
		     Poly->Points[Poly->Vertex[(count+1) % Poly->Number]]);
      if (State[0] == Mode && State[1] == Mode)
      {
	Neighbour = GEOFindNeighbour (Obj->Data.Polyeder, Poly, 
				      count, (count+1) % Poly->Number);
	if (Neighbour != NULL && Neighbour->Flags == 0)
	  OBJ_Propagate (Object, Obj, Neighbour, Orientation);
      }
    }

    Ptr = Ptr->Next;
  }
}


static void OBJ_Finalization (OBJECT *Obj, INTERSECT *List)
{
  POLYGON   *Poly;
  INTERSECT *Ptr = List, *Next;

  /* Due to reallocation of object-vertexes, the polygons' vertex-pointer
     may be wrong, now. So we have to correct this. */
  Poly = Obj->Data.Polyeder->List;
  while (Poly != NULL)
  {
    Poly->Points = Obj->VertexList;
    Poly = Poly->Next;
  }

  /* Compute the object's informations */
  OBJCalcInfo (Obj);

  /* Clean up unused intersection list */
  while (Ptr != NULL)
  {
    Next = Ptr->Next;
    GEOFreeIntersection (Ptr);
    Ptr = Next;
  };
}


/* OBJIntersection
   Up to now, this routine computes the contact-points between two objects. 
   Also, for each intersected polygon, there will be figured out, which points
   are inside or outside the other polyeder.
*/
OBJECT *OBJIntersection (OBJECT *Object1, H_MATRIX WTO1, 
			 OBJECT *Object2, H_MATRIX WTO2, 
			 int Color)
{
  int        count = 0, Side;
  double     Angle;
  POLYGON   *Poly, *Next;
  INTERSECT *List = NULL, *Intersect;
  INTERSECT *Ptr = NULL;
  OBJECT    *Obj = NULL;
  OBJECT    *Obj1, *Obj2;

  PRO_START ("\n")

WRITE_HMATRIX (stdout, WTO1, "WTO1");
WRITE_HMATRIX (stdout, WTO2, "WTO2");

  /* Create a copy of both objects and transform their coordinates */
  Obj1 = OBJCopyObject (Object1);
  OBJPlace (Obj1, WTO1);

  Obj2 = OBJCopyObject (Object2);
  OBJPlace (Obj2, WTO2);

  /* Compute list of intersections */
  if ((List = OBJ_GetIntersectionList(Obj1, Obj2)) == NULL)
    PRO_RETURN ("Empty intersection list", NULL)

OBJ_PrintList (List, "List");

  /* Create a new object */
  Obj = OBJNewObject ();
  Obj->ColorIndex    = -Color;
  Obj->Type          = PART_PLANE;
  Obj->Data.Polyeder = GEONewPolyeder ();

  /* Use intersection list to create polygons of the intersection object */
  for (Side = 0; Side < 2; ++Side)
  {
printf ("============= Object %d ===============\n", Side+1);
    /* Reset processing flags */ 
    Ptr = List;
    while (Ptr != NULL)
    {
      Ptr->done = FALSE;
      Ptr = Ptr->Next;
    }

    while ((Intersect = OBJ_UnhandledIntersection(List)) != NULL)
    {
      if ((Poly = OBJ_MakePoly(Obj, List, Intersect->Poly[Side],
			       CSG_INTERSECTION, INSIDE, Side)) != NULL)
      {
	while (Poly != NULL)
	{
	  Poly->Points      = Obj->VertexList;
	  Poly->Info.ID     = count++;
	  Poly->Info.Center = GEOCalcPolygonCenter (Poly);
	  GEOTriangulatePolygon (Poly);
	  GEOCalcHesseForm      (Poly);

	  /* Intersections within a polygon might have wrong orientation */
	  Angle = GEOVectorAngle (Intersect->Poly[Side]->Info.n, Poly->Info.n);
	  if (Angle > 90.0)
	    GEOSwapOrientation (Poly);

	  Next = Poly->Next;
	  Poly->Next = NULL;
	  GEOAddPolygon (Obj->Data.Polyeder, Poly);

	  Poly = Next;
	}
      }
      else                /* An error occured while creating the polygon */
      {
printf ("Error creating polygon\n");
        Intersect->done = TRUE;
      }
    }
  }

  OBJ_AddPolygons (Obj, List, Obj1, INSIDE, 0, TRUE);
  OBJ_AddPolygons (Obj, List, Obj2, INSIDE, 1, TRUE);

printf ("INTERSECTION READY !!!\n");

  OBJ_Finalization (Obj, List); 

  OBJFreeObject (Obj1);
  OBJFreeObject (Obj2);

  PRO_END
  return Obj;
}


/* OBJDifference
   subtracts one object from the other.
*/
OBJECT *OBJDifference (OBJECT *Object1, H_MATRIX WTO1, 
		       OBJECT *Object2, H_MATRIX WTO2,
		       int Color)
{
  int        count = 0, Side;
  double     Angle;
  POLYGON   *Poly, *Next;
  POLYGON   *SplitList;
  INTERSECT *List = NULL, *Intersect;
  INTERSECT *Ptr = NULL;
  OBJECT    *Obj = NULL;
  OBJECT    *Obj1, *Obj2;

  PRO_START ("\n")

WRITE_HMATRIX (stdout, WTO1, "WTO1");
WRITE_HMATRIX (stdout, WTO2, "WTO2");

  /* Create a copy of both objects and transform their coordinates */
  Obj1 = OBJCopyObject (Object1);
  OBJPlace (Obj1, WTO1);

  Obj2 = OBJCopyObject (Object2);
  OBJPlace (Obj2, WTO2);

  /* Compute list of intersections */
  if ((List = OBJ_GetIntersectionList(Obj1, Obj2)) == NULL)
    PRO_RETURN ("Empty intersection list", NULL)

OBJ_PrintList (List, "List");

  /* Create a new object */
  Obj = OBJNewObject ();
  Obj->ColorIndex    = -Color;
  Obj->Type          = PART_PLANE;
  Obj->Data.Polyeder = GEONewPolyeder ();

  /* Use intersection list to create polygons of the intersection object */
  for (Side = 0; Side < 2; ++Side)
  {
printf ("============= Object %d ===============\n", Side+1);
    /* Reset processing flags */ 
    Ptr = List;
    while (Ptr != NULL)
    {
      Ptr->done = FALSE;
      Ptr = Ptr->Next;
    }
    while ((Intersect = OBJ_UnhandledIntersection(List)) != NULL)
    {
      if ((Poly = OBJ_MakePoly(Obj, List, Intersect->Poly[Side], 
			       CSG_DIFFERENCE,
			       (Side==0)?OUTSIDE:INSIDE, Side)) != NULL)
      {
	while (Poly != NULL)
	{
	  Poly->Points      = Obj->VertexList;
	  Poly->Info.ID     = count++;
	  Poly->Info.Center = GEOCalcPolygonCenter (Poly);

	  GEOTriangulatePolygon (Poly);
	  GEOCalcHesseForm      (Poly);

	  /* Intersections within a polygon might have wrong orientation */
	  Angle = GEOVectorAngle (Intersect->Poly[Side]->Info.n, Poly->Info.n);
	  if ((Side == 0 && Angle > 90.0) || (Side == 1 && Angle < 90.0))
	    GEOSwapOrientation (Poly);

	  Next = Poly->Next;
          Poly->Next = NULL;

	  /* On subtraction the resulting polygon may be concave. So we must
	     check this and split the polygon in such a case. */
	  if (GEOTestPolygon(Poly) == 1)
          {                               /* Split a concave polygon */
	    printf ("Tessellating...\n");
	    SplitList = GEOTessellatePolygon (Poly);
	    GEOAddPolygon (Obj->Data.Polyeder, SplitList);
	    GEOFreePolygon (Poly, FALSE);
	    printf ("tessellated\n");
	  }
	  else
	    GEOAddPolygon (Obj->Data.Polyeder, Poly);

	  Poly = Next;
	}
      }
      else                /* An error occured while creating the polygon */
      {
printf ("Error creating polygon\n");
	Intersect->done = TRUE;
      }
    }
  }

  OBJ_AddPolygons (Obj, List, Obj1, OUTSIDE, 0, TRUE);
  OBJ_AddPolygons (Obj, List, Obj2, INSIDE,  1, FALSE);

printf ("INTERSECTION READY !!!\n");

  OBJ_Finalization (Obj, List); 

  OBJFreeObject (Obj1);
  OBJFreeObject (Obj2);

  PRO_END
  return Obj;
}


/* OBJUnion
   Compute the union of both objects.
*/
OBJECT *OBJUnion (OBJECT *Object1, H_MATRIX WTO1, 
		  OBJECT *Object2, H_MATRIX WTO2,
		  int Color)
{
  int        count = 0, Side;
  double     Angle;
  POLYGON   *Poly, *Next;
  POLYGON   *SplitList;
  INTERSECT *List = NULL, *Intersect;
  INTERSECT *Ptr = NULL;
  OBJECT    *Obj = NULL;
  OBJECT    *Obj1, *Obj2;

  PRO_START ("\n")

WRITE_HMATRIX (stdout, WTO1, "WTO1");
WRITE_HMATRIX (stdout, WTO2, "WTO2");

  /* Create a copy of both objects and transform their coordinates */
  Obj1 = OBJCopyObject (Object1);
  OBJPlace (Obj1, WTO1);

  Obj2 = OBJCopyObject (Object2);
  OBJPlace (Obj2, WTO2);

  /* Compute list of intersections */
  if ((List = OBJ_GetIntersectionList(Obj1, Obj2)) == NULL)
    PRO_RETURN ("Empty intersection list", NULL)

OBJ_PrintList (List, "List");

  /* Create a new object */
  Obj = OBJNewObject ();
  Obj->ColorIndex    = -Color;
  Obj->Type          = PART_PLANE;
  Obj->Data.Polyeder = GEONewPolyeder ();
                            
  /* Use intersection list to create polygons of the intersection object */
  for (Side = 0; Side < 2; ++Side)
  {
printf ("============= Object %d ===============\n", Side+1);
    /* Reset processing flags */ 
    Ptr = List;
    while (Ptr != NULL)
    {
      Ptr->done = FALSE;
      Ptr = Ptr->Next;
    }
    while ((Intersect = OBJ_UnhandledIntersection(List)) != NULL)
    {
      if ((Poly = OBJ_MakePoly(Obj, List, Intersect->Poly[Side], 
			            CSG_UNION, OUTSIDE, Side)) != NULL)
      {
	while (Poly != NULL)
	{
	  Poly->Points      = Obj->VertexList;
	  Poly->Info.ID     = count++;
	  Poly->Info.Center = GEOCalcPolygonCenter (Poly);

	  GEOTriangulatePolygon (Poly);
	  GEOCalcHesseForm      (Poly);

	  /* Intersections within a polygon might have wrong orientation */
	  Angle = GEOVectorAngle (Intersect->Poly[Side]->Info.n, Poly->Info.n);
	  if (Angle > 90.0)
	    GEOSwapOrientation (Poly);

	  Next = Poly->Next;
	  Poly->Next = NULL;

	  /* On subtraction the resulting polygon can be convex. So we must
	     check this and split the polygon in such a case. */
	  if (GEOTestPolygon(Poly) == 1)
          {                               /* Split a concave polygon */
	    printf ("Tessellating...\n");
	    SplitList = GEOTessellatePolygon (Poly); 
	    GEOAddPolygon (Obj->Data.Polyeder, SplitList);
	    GEOFreePolygon (Poly, FALSE);
	    printf ("finished\n");
	  }
	  else
	    GEOAddPolygon (Obj->Data.Polyeder, Poly);

	  Poly = Next;
	}
      }
      else                /* An error occured while creating the polygon */
      {
printf ("Error creating polygon\n");
	Intersect->done = TRUE;
      }
    }
  }

  OBJ_AddPolygons (Obj, List, Obj1, OUTSIDE, 0, TRUE);
  OBJ_AddPolygons (Obj, List, Obj2, OUTSIDE, 1, TRUE);

printf ("INTERSECTION READY !!!\n");

  OBJ_Finalization (Obj, List); 

  OBJFreeObject (Obj1);
  OBJFreeObject (Obj2);

  PRO_END
  return Obj;
}


/* 
   OBJGetIntersections returns a list of intersection points 
*/
INTERSECT *OBJGetIntersections (OBJECT *Object1, H_MATRIX WTO1, 
				OBJECT *Object2, H_MATRIX WTO2)
{
  INTERSECT *List = NULL;
  OBJECT    *Obj1, *Obj2;

  PRO_START ("\n")

  /* Create a copy of both objects and transform their coordinates */
  Obj1 = OBJCopyObject (Object1);
  OBJPlace (Obj1, WTO1);

  Obj2 = OBJCopyObject (Object2);
  OBJPlace (Obj2, WTO2);

  /* Compute list of intersections */
  List = OBJ_GetIntersectionList (Obj1, Obj2);

OBJ_PrintList (List, "List");

  PRO_END
  return List;
}  
