#include "libobj.h"


dual OBJ_ReadLine (FILE *fp)
{
  char letter = ' ';

  while (! feof(fp) && letter != '\n')
    letter = fgetc (fp);

  return (! feof(fp));
}


OBJECT *OBJLoadV11 (FILE *fp, int Version, int *PolygonID, dual check)
{
  int      count, index;
  int      TextNo, TextureNo;
  int      EdgeNo, PlaneNo;
  int      Color, Value;
  GEO_EDGE Edge;
  POLYGON *Polygon;
  OBJECT  *Result = NULL;

  PRO_START ("\n")

  /* ***** Color-Index ***** */
  if (fscanf(fp, "%d\n", &Color) < 1)
    PRO_RETURN ("No color-index found !", NULL)

  Result = OBJNewObject ();
  Result->ColorIndex = OBJGetIGRIPColor (Color);

  /* ***** Vertices ***** */
  fscanf (fp, "%d\n", &Value);
  Result->VertexNo = Value;
  PRO1 ("# NODES: %d\n", Result->VertexNo)
  if (Result->VertexNo > 0)
  {
    Result->VertexList = 
      (POINT3D *)malloc (sizeof(POINT3D) * Result->VertexNo);
    for (count = 0; count < (int)Result->VertexNo; ++count)
    {
      if (fscanf(fp, "%lf %lf %lf\n", 
	             &Result->VertexList[count].x,
	             &Result->VertexList[count].y,
	             &Result->VertexList[count].z) < 3)
      {
	OBJFreeObject (Result);
	PRO_RETURN1 ("Node %d has invalid coordinates !", count+1, NULL)
      }

      PRO4 ("%4d: [%6.2f][%6.2f][%6.2f]\n", count,
                  Result->VertexList[count].x,
                  Result->VertexList[count].y,
                  Result->VertexList[count].z)
    }
  }

  /* ***** Edges/Lines ***** */
  fscanf (fp, "%d\n", &EdgeNo);
  PRO1 ("# EDGES: %d\n", EdgeNo)
  if (EdgeNo > 0)
  {
    EdgeNo++;
    Result->Type = PART_CURVE;
    PRO1 ("# LINES: %d\n", EdgeNo)

    Result->Data.Curve = (LINES *)malloc (sizeof(LINES));
    Result->Data.Curve->Number = EdgeNo;
    Result->Data.Curve->List   = (int *)malloc (sizeof(int) * EdgeNo);
    count = 0;
    while (count < EdgeNo-1 && ! feof(fp))
    {
      fscanf (fp, "%d %d\n", &Edge.start, &Edge.end);
      Result->Data.Curve->List[count] = Edge.start;
      PRO1 ("\\ %d", Result->Data.Curve->List[count]);
      count++;
    }
    PRO0 ("\\\n")
    Result->Data.Curve->List[EdgeNo-1] = Edge.end;
  }

  /* ***** Planes ***** */
  fscanf (fp, "%d\n", &PlaneNo);
  PRO1 ("# PLANES: %d\n", PlaneNo)  
  if (PlaneNo != 0)
  {
    Result->Type = PART_PLANE;

    Result->Data.Polyeder = GEONewPolyeder ();
    Result->Data.Polyeder->Number = abs(PlaneNo);

    count = 0;
    while (count < abs(PlaneNo) && ! feof(fp))
    {
      Polygon = GEONewPolygon ();
      fscanf (fp, "%d ", &Polygon->Number);
      PRO2 ("[%d: #%d]:", count, Polygon->Number)
      Polygon->Vertex = (int *)malloc (sizeof(int)*Polygon->Number);
      for (index = 0; index < Polygon->Number; ++index)
      {	
	fscanf (fp, "%d ", &Polygon->Vertex[index]);
	PRO1 ("\\ %d", Polygon->Vertex[index])
      }
      PRO0 ("\\\n")
      Polygon->Points = Result->VertexList;

      /* check polygon and calculate triangulation */
      if (! GEOCalcHesseForm(Polygon) || 
	  (check && GEOTestPolygon(Polygon) > 0))
      {
	GEOFreePolygon (Polygon, FALSE);
	OBJFreeObject (Result);
	PRO_RETURN1 ("Polygon %d is concave or not planar !", count+1, NULL)
      }
      Polygon->Info.ID     = *PolygonID;
      Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);
      (*PolygonID)++;

      /* triangulate polygon */
      GEOTriangulatePolygon (Polygon);

      if (Result->Data.Polyeder->List == NULL)
	Result->Data.Polyeder->List = Polygon;
      else
	Result->Data.Polyeder->Last->Next = Polygon;

      Result->Data.Polyeder->Last = Polygon;

      count++;
    }
    /* 
       TODO:
       Polyeder prfen: Euler -> Warnung ?
    */
  }

  Result->Info = OBJCalcInfo (Result);

  /* ***** Text ***** */
  if (fscanf(fp, "%d\n", &TextNo) < 1)
  {
    OBJFreeObject (Result);
    PRO_RETURN ("No text-amount found !", NULL)
  }
  PRO1 ("TEXTS:    %d\n", TextNo)
  for (count = 0; count < TextNo; ++count)
    for (index = 0; index < 5; ++index)
      OBJ_ReadLine (fp);

  /* ***** Texture ***** */
  if (Version == 11)
  {
    if (fscanf(fp, "%d\n", &TextureNo) < 1)
      TextureNo = -1;
    PRO1 ("TEXTURES: %d\n", TextureNo)
    if (TextureNo != 0)
      for (count = 0; count < TextureNo; ++count)
	OBJ_ReadLine (fp);
  }

  Result->Next = NULL;

  PRO_END
  return Result;
}


OBJECT *OBJLoadV12 (FILE *fp, int *PolygonID, dual check)
{
  int       count, index;
  int       TextNo, TextureNo;
  int       PointNo, Value;
  int       EdgeNo, PlaneNo;
  int       Color, Edge;
  GEO_EDGE *Edges = NULL;
  POLYGON  *Polygon;
  OBJECT   *Result = NULL;

  PRO_START ("\n")

  /* ***** Color-Index ***** */
  if (fscanf(fp, "%d\n", &Color) < 1)
    PRO_RETURN ("No color-index found !", NULL)

  Result = OBJNewObject ();
  Result->ColorIndex = OBJGetIGRIPColor (Color);

  fscanf (fp, "*\n");

  /* ***** Vertices ***** */
  fscanf (fp, "%d\n", &Value);
  Result->VertexNo = Value;
  PRO1 ("# NODES: %d\n", Result->VertexNo)
  if (Result->VertexNo > 0)
  {
    Result->VertexList = 
      (POINT3D *)malloc (sizeof(POINT3D) * Result->VertexNo);
    count = 0;
    while (count < (int)Result->VertexNo && !feof(fp))
    {
      if (fscanf(fp, "%lf %lf %lf\n", &Result->VertexList[count].x,
		                      &Result->VertexList[count].y,
		                      &Result->VertexList[count].z) < 3)
      {
	OBJFreeObject (Result);
	PRO_RETURN1 ("Node %d has invalid coordinates !", count+1, NULL)
      }

      PRO4 ("%d: [%6.2f][%6.2f][%6.2f]\n", count,
                 Result->VertexList[count].x,
	         Result->VertexList[count].y,
                 Result->VertexList[count].z)
      count++;
    }
  }

  /* ***** Points ***** */
  fscanf (fp, "%d\n", &PointNo);
  PRO1 ("# POINTS: %d\n", PointNo);
  for (count = 0; count < PointNo; ++count)
    OBJ_ReadLine (fp);

  /* ***** Edges **** */
  fscanf (fp, "%d\n", &EdgeNo);
  PRO1 ("# EDGES: %d\n", EdgeNo)  
  if (EdgeNo > 0)
  {
    Edges = (GEO_EDGE *)malloc (sizeof(GEO_EDGE) * EdgeNo);
    for (count = 0; count < EdgeNo; ++count)
    {
      fscanf (fp, "%d %d\n", &Edges[count].start, &Edges[count].end);
      PRO2 ("  %d -> %d\n", Edges[count].start, Edges[count].end)
    }
  }

  /* ***** Planes ***** */
  fscanf (fp, "%d\n", &PlaneNo);
  PRO1 ("# PLANES: %d\n", PlaneNo)
  if (PlaneNo != 0)
  {
    Result->Type = PART_PLANE;

    Result->Data.Polyeder = GEONewPolyeder ();
    Result->Data.Polyeder->Number = abs(PlaneNo);

    count = 0;
    while (count < abs(PlaneNo) && ! feof(fp))
    {
      Polygon = GEONewPolygon ();
      fscanf (fp, "%d ", &Polygon->Number);
      PRO2 ("[%d: #%d]:", count+1, Polygon->Number)
      Polygon->Vertex = (int *)malloc (sizeof(int) * Polygon->Number);
      for (index = 0; index < Polygon->Number; ++index)
      {	
	fscanf (fp, "%d ", &Edge);
	if (Edge > 0)
	  Polygon->Vertex[index] = Edges[Edge-1].start;
	else
	  Polygon->Vertex[index] = Edges[-Edge-1].end;
	PRO2 ("\\ %d[%d]", Edge, Polygon->Vertex[index])
      }
      PRO0 ("\\\n")
      Polygon->Points = Result->VertexList;

      /* check polygon and calculate triangulation */
      if (! GEOCalcHesseForm(Polygon) || 
	  (check && GEOTestPolygon(Polygon) > 0))
      {
	GEOFreePolygon (Polygon, FALSE);
	OBJFreeObject (Result);
	free (Edges);
	PRO_RETURN1 ("Polygon %d is concave or not planar !", count+1, NULL)
      }
      Polygon->Info.ID     = *PolygonID;
      Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);
      (*PolygonID)++;

      /* triangulate polygon */
      GEOTriangulatePolygon (Polygon);

      if (Result->Data.Polyeder->List == NULL)
	Result->Data.Polyeder->List = Polygon;
      else
	Result->Data.Polyeder->Last->Next = Polygon;

      Result->Data.Polyeder->Last = Polygon;

      count++;
    }
  }
  elsif (EdgeNo > 0)
  {
    Result->Type = PART_CURVE;
    Result->Data.Curve = (LINES *)malloc (sizeof(LINES));
    Result->Data.Curve->Number = EdgeNo+1;
    Result->Data.Curve->List = (int *)malloc (sizeof(int) * (EdgeNo+1));
    for (count = 0; count < EdgeNo; ++count)
      Result->Data.Curve->List[count] = Edges[count].start;
    Result->Data.Curve->List[count] = Edges[count-1].end;

    OBJ_ReadLine (fp);
  }
  else
    Result->Type = NONE;

  if (Edges != NULL)
    free (Edges);

  Result->Info = OBJCalcInfo (Result);

  /* ***** Text ***** */
  if (fscanf(fp, "%d\n", &TextNo) < 1)
  {
    OBJFreeObject (Result);
    PRO_RETURN ("No text-amount found !", NULL)
  }
  PRO1 ("TEXTS:    %d\n", TextNo)
  for (count = 0; count < TextNo; ++count)
    for (index = 0; index < 5; ++index)
      OBJ_ReadLine (fp);

  /* ***** Texture ***** */
  if (fscanf(fp, "%d\n", &TextureNo) < 1)
    TextureNo = -1;
  PRO1 ("TEXTURES: %d\n", TextureNo)
  if (TextureNo != 0)
    for (count = 0; count < TextureNo; ++count)
      OBJ_ReadLine (fp);

  Result->Next = NULL;

  PRO_END
  return Result;
}


dual OBJWriteV11 (FILE *fp, OBJECT *Obj, dual triangle)
{
  int  count, Sum = 0;
  POLYGON      *Ptr;

  PRO_START (" ")

  fprintf (fp, "%d\n", OBJSetIGRIPColor(Obj->ColorIndex)); /* color-index */

  fprintf (fp, "%d\n", Obj->VertexNo);
  for (count = 0; count < (int)Obj->VertexNo; ++count)
    fprintf (fp, "%.6f %.6f %.6f\n", Obj->VertexList[count].x,
	                             Obj->VertexList[count].y,
	                             Obj->VertexList[count].z);

  switch (Obj->Type)
  {
    case PART_CURVE :
      fprintf (fp, "%d\n", Obj->Data.Curve->Number);
      for (count = 0; count < Obj->Data.Curve->Number-1; ++count)
	fprintf (fp, "%d %d\n", Obj->Data.Curve->List[count], 
		                Obj->Data.Curve->List[count+1]);
      fprintf (fp, "0\n");
      break;

    case PART_PLANE :
      fprintf (fp, "0\n");

      if (triangle)
      {
	Ptr = Obj->Data.Polyeder->List;
	while (Ptr != NULL)
	{
	  Sum += Ptr->Number-2;
	  Ptr = Ptr->Next;
	}
      }
      else
	Sum = Obj->Data.Polyeder->Number;

      fprintf (fp, "%d\n", Sum);

      Ptr = Obj->Data.Polyeder->List;
      while (Ptr != NULL)
      {
	if (triangle)
	  for (count = 0; count < Ptr->Number-2; ++count)
	    fprintf (fp, "3 %d %d %d\n",          Ptr->Triangle[count][0], 
	                 Ptr->Triangle[count][1], Ptr->Triangle[count][2]);
	else
	{
	  fprintf (fp, "%d", Ptr->Number);
	  for (count = 0; count < Ptr->Number; ++count)
	    fprintf (fp, " %d", Ptr->Vertex[count]);
	  fprintf (fp, "\n");
	}
	Ptr = Ptr->Next;
      }
      break;

    default :
      PRO0 ("Invalid Object-Type\n")
  }

  fprintf (fp, "0\n");

  fprintf (fp, "0\n");

  PRO_END
  return TRUE;
}


dual OBJWriteV12 (FILE *fp, OBJECT *Obj)
{
  int      count, index;
  POLYGON *Polygon;

  PRO_START (" ")

  /* ***** Color-Index ***** */
  fprintf (fp, "-%d\n", Obj->ColorIndex);

  fprintf (fp, "*\n");

  /* ***** Vertices ***** */
  fprintf (fp, "%d\n", Obj->VertexNo);
  for (count = 0; count < (int)Obj->VertexNo; ++count)
    fprintf (fp, "%.6f %.6f %.6f\n", 
	         Obj->VertexList[count].x,
	         Obj->VertexList[count].y,
	         Obj->VertexList[count].z);
    
  /* ***** Edges/Lines *****/
  if (Obj->Type == PART_CURVE)
  {
    fprintf (fp, "%d\n", Obj->Data.Curve->Number-1);

    /* TODO:
       Linie kann auch unterbrochen sein !
    */
    for (count = 0; count < Obj->Data.Curve->Number-1; ++count)
      fprintf (fp, "%d %d\n", 
	           Obj->Data.Curve->List[count],
	           Obj->Data.Curve->List[count+1]);
    fprintf (fp, "0\n");
  }
  if (Obj->Type == PART_PLANE)
  {
    fprintf (fp, "0\n");
    fprintf (fp, "0\n");

    fprintf (fp, "%d\n", Obj->Data.Polyeder->Number);
    if (Obj->Data.Polyeder->Number > 0)
    {
      Polygon = Obj->Data.Polyeder->List;
      for (count = 0; count < Obj->Data.Polyeder->Number; ++count)
      {
	fprintf (fp, "%d", Polygon->Number);
	for (index = 0; index < Polygon->Number; ++index)
	  fprintf (fp, " %d", Polygon->Vertex[index]);
	fprintf (fp, "\n");
	Polygon = Polygon->Next;
      }
    }
    fprintf (fp, "0\n");
  }

  /* ***** Text ***** */
  fprintf (fp, "0\n");

  /* ***** Texture ***** */
  fprintf (fp, "0\n");

  PRO_END
  return TRUE;
}


