#include "libpart.h"


static PART *PART_CSGOperation (int Operation,
				const char *Name, H_MATRIX Frame,
				const PART *Part1, H_MATRIX WTP1, 
				const PART *Part2, H_MATRIX WTP2,
				int Color)
{
  OBJECT *Obj, *ObjList = NULL;
  OBJECT *Obj1, *Obj2;
  PART   *Part = NULL;

  PRO_START ("\n")

  if (Part1 == NULL || Part2 == NULL)
    PRO_RETURN ("empty part", NULL)

  Obj1 = Part1->ObjectList;
  while (Obj1 != NULL)
  {
    Obj2 = Part2->ObjectList;
    while (Obj2 != NULL)
    {
      switch (Operation)
      {
        case CSG_UNION :
	  Obj = OBJUnion (Obj1, WTP1, Obj2, WTP2, Color);	  
	  break;

        case CSG_INTERSECTION :
	  Obj = OBJIntersection (Obj1, WTP1, Obj2, WTP2, Color);	  
	  break;

        case CSG_DIFFERENCE :
	  Obj = OBJDifference (Obj1, WTP1, Obj2, WTP2, Color);
	  break;

        default :
	  PRO_RETURN1 ("Invalid Operation", Operation, NULL)
      }
	  
      if (Obj == NULL)
      {
	/* If there is no intersection between both objects, we may copy 
	   one or both objects respecting the calculated function */
	switch (Operation)
	{
  	  case CSG_UNION :              /* Copy objects of both parts */
	    Obj = OBJCopyObject (Obj1);
       	    Obj->Next = OBJCopyObject (Obj2);
	    break;

  	  case CSG_DIFFERENCE :         /* Copy object of first part */
	    Obj = OBJCopyObject (Obj1);
	    break;
	
  	  default :                     /* No copying for intersection */
	    break;
	}
      }

      if (Obj != NULL)                  /* Add object to the object list */
      {
	if (ObjList != NULL)
	  Obj->Next = ObjList;
	ObjList = Obj;
      }
	
      Obj2 = Obj2->Next;
    }
    Obj1 = Obj1->Next;
  }

  Part = PARTCreate (Name, ObjList);

  PRO_END
  return Part;
}


PART *PARTIntersection (const char *Name, H_MATRIX Frame,
			const PART *Part1, H_MATRIX WTP1, 
			const PART *Part2, H_MATRIX WTP2,
			int Color)
{
  return PART_CSGOperation (CSG_INTERSECTION,
			    Name, Frame, Part1, WTP1, Part2, WTP2, Color);
}


PART *PARTDifference (const char *Name, H_MATRIX Frame,
		      const PART *Part1, H_MATRIX WTP1, 
		      const PART *Part2, H_MATRIX WTP2,
		      int Color)
{
  return PART_CSGOperation (CSG_DIFFERENCE,
			    Name, Frame, Part1, WTP1, Part2, WTP2, Color);
}


PART *PARTUnion (const char *Name, H_MATRIX Frame,
		 const PART *Part1, H_MATRIX WTP1, 
		 const PART *Part2, H_MATRIX WTP2,
		 int Color)
{
  return PART_CSGOperation (CSG_UNION,
			    Name, Frame, Part1, WTP1, Part2, WTP2, Color);
}


INTERSECT *PARTGetIntersections (const PART *Part1, H_MATRIX WTP1, 
				 const PART *Part2, H_MATRIX WTP2)
{
  OBJECT    *Obj1, *Obj2;
  INTERSECT *Result = NULL;
  INTERSECT *List, *Ptr;

  PRO_START ("\n")

  if (Part1 == NULL || Part2 == NULL)
    PRO_RETURN ("empty part", NULL)

  Obj1 = Part1->ObjectList;
  while (Obj1 != NULL)
  {
    Obj2 = Part2->ObjectList;
    while (Obj2 != NULL)
    {
      if ((List = OBJGetIntersections(Obj1, WTP1, Obj2, WTP2)) != NULL)
      {
	if ((Ptr = Result) == NULL)
	  Result = List;
	else
        {
	  while (Ptr->Next != NULL)
	    Ptr = Ptr->Next;
	  Ptr->Next = List;
	}
      }

      Obj2 = Obj2->Next;
    }
    Obj1 = Obj1->Next;
  }

  PRO_END
  return Result;
}
