#include <math.h>

#include "partGlobal.h"


typedef struct EDGELIST
{
         GEO_EDGE  Edge;
  struct EDGELIST *Next;
} EDGELIST;


static EDGELIST *EdgeList = NULL;


static void PART_Prove (int start, int end)
{
  EDGELIST *Elem, *Last;
  EDGELIST *Ptr;

  Ptr = Last = EdgeList;
  while (Ptr != NULL)
  {
    if (Ptr->Edge.start == end && Ptr->Edge.end == start)
    {
      if (Ptr == EdgeList)
        EdgeList = EdgeList->Next;
      else
	Last->Next = Ptr->Next;
      PRO2 ("Clearing %d -> %d\n", Ptr->Edge.start, Ptr->Edge.end)
      free (Ptr);
      return;
    }
    Last = Ptr;
    Ptr = Ptr->Next;
  }

  Elem = (EDGELIST*)malloc (sizeof(EDGELIST));
  Elem->Edge.start = start;
  Elem->Edge.end   = end;
  Elem->Next       = NULL;

  if (EdgeList == NULL)
    EdgeList = Elem;
  else
    Last->Next = Elem;
}


static dual PART_IsClosedPolyeder (POLYEDER *Polyeder)
{
  int       count;
  POLYGON  *Polygon;
  EDGELIST *Ptr, *Next;

  Polygon = Polyeder->List;
  while (Polygon != NULL)
  {
    for (count = 0; count < Polygon->Number; ++count)
      PART_Prove (Polygon->Vertex[count], 
		 Polygon->Vertex[(count+1) % Polygon->Number]);
    Polygon = Polygon->Next;
  }

  if ((Ptr = EdgeList) != NULL)
  {
    while (Ptr != NULL)
    {
      Next = Ptr->Next;
      PRO2 ("Remaining edge: %d -> %d\n", Ptr->Edge.start, Ptr->Edge.end)
      free (Ptr);
      Ptr = Next;
    }
    EdgeList = NULL;
    return FALSE;
  }
    
  return TRUE;
}


dual PARTIsClosed (PART *Part)
{
  int     count = 1;
  OBJECT *Object;

  PRO_START ("\n")

  if (Part == NULL)
    PRO_RETURN ("Empty part", TRUE)

  Object = Part->ObjectList;
  while (Object != NULL)
  {
    PRO1 ("Object %d:\n", count)
    if (Object->Type == PART_PLANE)
      if (! PART_IsClosedPolyeder(Object->Data.Polyeder))
	PRO_RETURN ("  NOT closed", FALSE)
      else
	PRO0 ("  closed\n")
    else
      PRO0 ("  Lines\n");
    Object = Object->Next;
    count++;
  }

  PRO_END
  return TRUE;
}


dual PARTIsConvex (PART *Part)
{
  int      plane;
  POLYGON *Ptr;
  OBJECT  *Object;

  PRO_START ("\n")

  if (Part == NULL)
    PRO_RETURN ("Empty part", TRUE)

  Object = Part->ObjectList;
  while (Object != NULL)
  {
    if (Object->Type == PART_PLANE)
    {
      plane = 1;
      Ptr = Object->Data.Polyeder->List;
      PRO0 ("Polyeder:\n")
      while (Ptr != NULL)
      {
	PRO1 ("  %d:\n", plane)

	if (! GEOCalcHesseForm(Ptr))
	  PRO_RETURN ("error computing hesse-form", FALSE)

	switch (GEOTestPolygon(Ptr))
	{
   	  case 0 : 
	    PRO0 ("convex and planar\n")
	    break;
  	  case 1 : 
	    PRO_RETURN ("concav", FALSE)
          case 2 : 
	    PRO_RETURN ("not planar", FALSE)
     	  case 3 :
	    PRO_RETURN ("all edges collinear", FALSE)
	  default : 
	    PRO_RETURN ("unkown error", FALSE)
        }

        Ptr = Ptr->Next;
	plane++;
      }
      PRO0 ("\\\n")
    }
    Object = Object->Next;
  }

  PRO_END
  return TRUE;
}


dual PARTCheckEuler (PART *Part)
{
  OBJECT *Object;

  PRO_START ("\n")

  if (Part == NULL)
    PRO_RETURN ("Empty part", TRUE)

  Object = Part->ObjectList;
  while (Object != NULL)
  {
    if (Object->Type == PART_PLANE)
    {
      if (! OBJIsEuler(Object))
	PRO_RETURN ("failed", FALSE)
      else
	PRO0 ("ok\n")
      Object = Object->Next;
    }
  }

  PRO_END
  return TRUE;
}


#if 0
dual PARTCheckPart (PART *Part)
{
  int      count = 0, index;
  OBJECT  *Object;
  POLYGON *Polygon;

  PRO_START ("\n")

  if (Part == NULL)
    PRO_RETURN ("Empty part", TRUE)

  Object = Part->ObjectList;
  while (Object != NULL)
  {
    PRO1 ("Object %d: \n", count);

    switch (Object->Type)
    {
      case LINES :
	PRO0 ("  curve\n")
	break;

      case PLANES :
	PRO0 ("  Euler-function: ...")
	if (! PART_IsEulerObject(Object))
	  PRO0 ("\\E - K + F != 2\n")
        else
	  PRO0 ("\\OK\n")

	if (! PART_IsClosedPolyeder(&Object->Data.Polyeder))
	  PRO_RETURN ("not closed", FALSE)
	else
	  PRO0 ("  closed\n")

	PRO0 ("  Hesse:")
	if (! GEOCalcHesseForm(Polygon))
	  PRO_RETURN ("error calculating hesse-form", FALSE)

	PRO0 ("  convex/planar:")
	index = 0;
	Polygon = Object->Data.Polyeder.List;
	while (Polygon != NULL)
	{
	  PRO1 ("\\ %d", index)
	  if (GEOTestPolygon(Polygon) > 0)
	    PRO_RETURN ("not convex", FALSE)
	  Polygon = Polygon->Next;
	  index++;
	}
	PRO0 ("\\\n")
	break;

      default :
	PRO_RETURN ("invalid object-type", FALSE)
    }
    Object = Object->Next;
    count++;
  }

  PRO_END
  return TRUE;
}
#endif



