#include "libpart.h"


PART *PARTCreateArrow (const char *Name, H_MATRIX Frame, 
		       double Radius1,   double   Len1, 
		       double Radius2,   double   Len2, 
		       double Angle,     int      Facettes, 
		       double Segment,   int      Color)
{
  OBJECT  *Obj;
  PART    *Part;

  PRO_START ("\n")

  Obj = OBJCreateArrow (Frame,
			Radius1, Len1, Radius2, Len2, Angle, Facettes,
			Segment, Color);
  Part = PARTCreate (Name, Obj);

  PRO_END
  return Part;
}


PART *PARTCreateBolid (const char *Name, H_MATRIX Frame,
		       double  Radius1,  double   Radius2,   double Radius3,
		       double  SizeZ,    int      Facettes1, int    Facettes2, 
		       double  Segment,  int      Color)
{
  OBJECT  *Obj;
  PART    *Part;

  PRO_START ("\n")

  Obj = OBJCreateBolid (Frame,
			Radius1, Radius2, Radius3, SizeZ, Facettes1, Facettes2,
			Segment, Color);
  Part = PARTCreate (Name, Obj);

  PRO_END
  return Part;
}


PART *PARTCreateCone (const char *Name, H_MATRIX Frame,
		      double  Radius1,  double   Radius2, double Radius3, 
		      double  SizeZ,    double   Segment, int    Facettes, 
		      int     Color)
{
  OBJECT *Obj;
  PART   *Part;

  PRO_START ("\n")

  Obj = OBJCreateCone (Frame, Radius1, Radius2, Radius3, SizeZ, Segment,
		       Facettes, Color);
  Part = PARTCreate (Name, Obj);

  PRO_END
  return Part;
}


PART *PARTCreateCylinder (const char *Name, H_MATRIX Frame,
			  double  Radius,   double   Depth,
			  double  Segment,  int      Facettes, int Color)
{
  OBJECT *Obj;
  PART   *Part;

  PRO_START ("\n")

  Obj = OBJCreateCylinder (Frame, Radius, Depth, Segment, Facettes, Color);
  Part = PARTCreate (Name, Obj);
  
  PRO_END
  return Part;
}


PART *PARTCreateGear (const char *Name, H_MATRIX Frame,
		      double  Radius1,  double Radius2, 
		      double  Radius3,  double Depth1,  double Depth2,
		      int     Teeth,    double Skew,    int    Color)
{
  OBJECT *Obj;
  PART   *Part;

  PRO_START ("\n")

  Obj = OBJCreateGear (Frame, Radius1, Radius2, Radius3, Depth1, Depth2,
		       Teeth, Skew, Color);
  Part = PARTCreate (Name, Obj);

  PRO_END
  return Part;
}


PART *PARTCreateIcosahedron (const char *Name, H_MATRIX Frame,
			     double  Radius1,  double   Radius2, 
			     int     Density,  int      Iterations, int Color)
{
  OBJECT *Obj;
  PART   *Part;

  PRO_START ("\n")

  Obj = OBJCreateIcosahedron (Frame, Radius1, Radius2, Density, Iterations,
			      Color);
  Part = PARTCreate (Name, Obj);

  PRO_END
  return Part;
}


PART *PARTCreatePipe (const char *Name,   H_MATRIX Frame,
		      double  Radius1,    double   Radius2, 
		      double  Thickness1, double   Thickness2,
		      double  SizeZ,      double   Segment, 
		      int     Facettes,   int      Color)
{
  OBJECT *Obj;
  PART   *Part;

  PRO_START ("\n")

  Obj = OBJCreatePipe (Frame, Radius1, Radius2, Thickness1, Thickness2,
		       SizeZ, Segment, Facettes, Color);
  Part = PARTCreate (Name, Obj);

  PRO_END
  return Part;
}


PART *PARTCreatePrism (const char *Name, H_MATRIX Frame,
		       double  Width1,   double   Width2,  double OffsX,
		       double  Depth1,   double   Depth2,  double OffsY,
		       double  Height1,  double   Height2, double OffsZ,
		       int     Color)
{
  OBJECT *Obj;
  PART   *Part;

  PRO_START ("\n")

  Obj = OBJCreatePrism (Frame, Width1, Width2, OffsX, Depth1, Depth2, OffsY,
			Height1, Height2, OffsZ, Color);
  Part = PARTCreate (Name, Obj);

  PRO_END
  return Part;
}


PART *PARTCreateQuader (const char *Name, H_MATRIX Frame,
			double  SizeX,    double   SizeY, double SizeZ, 
			int     Color)
{
  OBJECT *Obj;
  PART   *Part;

  PRO_START ("\n")

  Obj = OBJCreateQuader (Frame, SizeX, SizeY, SizeZ, Color);
  Part = PARTCreate (Name, Obj);

  PRO_END
  return Part;
}


PART *PARTCreateRot (const char *Name, H_MATRIX  Frame,
		     int     Number,   POINT2D  *Points, 
		     double  Segment,  int       Facettes,
		     double  Radius,   int       Color)
{
  OBJECT *Obj;
  PART   *Part;

  PRO_START ("\n")

  Obj = OBJCreateRot (Frame, Number, Points, Segment, Facettes, Radius, Color);
  Part = PARTCreate (Name, Obj);

  PRO_END
  return Part;
}


PART *PARTCreateScrew (const char *Name, H_MATRIX Frame,
		       double  Radius1,  double   Radius2, 
		       double  Height,   double   Increase,  double Offset,
		       int     Facettes, dual     Direction, int    Color)
{
  OBJECT *Obj;
  PART   *Part;

  PRO_START ("\n")

  Obj = OBJCreateScrew (Frame, Radius1, Radius2, Height, Increase, Offset,
			Facettes, Direction, Color);
  Part = PARTCreate (Name, Obj);

  PRO_END
  return Part;
}


PART *PARTCreateSlice (const char *Name,  H_MATRIX Frame,
		       double  Radius1,   double   Radius2, 
		       double  Radius3,   double   SizeZ, 
		       int     Facettes1, int      Facettes2, 
		       double  Segment,   int      Color)
{
  OBJECT *Obj;
  PART   *Part = NULL;

  PRO_START ("\n")

  Obj = OBJCreateSlice (Frame, Radius1, Radius2, Radius3, SizeZ, 
			Facettes1, Facettes2, Segment, Color);
  Part = PARTCreate (Name, Obj);

  PRO_END
  return Part;
}


PART *PARTCreateSphere (const char *Name, H_MATRIX Frame,
			double  Radius, 
			double  SegmentX, double   SegmentZ,
			int     Facettes, int      Color)
{
  OBJECT *Obj;
  PART   *Part;

  PRO_START ("\n")

  Obj = OBJCreateSphere (Frame, Radius, SegmentX, SegmentZ, Facettes, Color);
  Part = PARTCreate (Name, Obj);

  PRO_END
  return Part;
}


PART *PARTCreateSpring (const char *Name,  H_MATRIX Frame,
			double  Radius1,   double   Radius2,   double Radius3,
			double  Height,    double   Increase,
		        int     Facettes1, int      Facettes2, 
			dual    Direction, int      Color)
{
  OBJECT *Obj;
  PART   *Part;

  PRO_START ("\n")

  Obj = OBJCreateSpring (Frame, Radius1, Radius2, Radius3, Height, 
			 Increase, Facettes1, Facettes2, Direction, Color);
  Part = PARTCreate (Name, Obj);

  PRO_END
  return Part;
}


PART *PARTCreateSweep (const char *Name, H_MATRIX  Frame,
		       int    Number,    POINT2D *Points, 
		       double Height,    int       Color)
{
  OBJECT *Obj;
  PART   *Part;

  PRO_START ("\n")

  Obj = OBJCreateSweep (Frame, Number, Points, Height, Color);
  Part = PARTCreate (Name, Obj);

  PRO_END
  return Part;
}


PART *PARTCreateTorus (const char *Name,  H_MATRIX Frame,
		       double  Radius1,   double   Radius2, 
		       int     Facettes1, int      Facettes2,
		       double  Segment,   int      Color)
{
  OBJECT *Obj;
  PART   *Part;

  PRO_START ("\n")

  Obj = OBJCreateTorus (Frame, Radius1, Radius2, Facettes1, Facettes2,
			Segment, Color);
  Part = PARTCreate (Name, Obj);

  PRO_END
  return Part;
}


PART *PARTCreateWedge (const char *Name, H_MATRIX Frame,
		       double  Width1,   double   Width2,
		       double  Height1,  double   Height2,
		       double  Depth1,   double   Depth2, int Color)
{
  OBJECT *Obj;
  PART   *Part;

  PRO_START ("\n")

  Obj = OBJCreateWedge (Frame, Width1, Width2, Height1, Height2, 
			Depth1, Depth2, Color);
  Part = PARTCreate (Name, Obj);

  PRO_END
  return Part;
}


