#include "partGlobal.h"


#define DXF_BUFSIZE 2048


enum DFX_GROUP
{
  DXF_NONE = -1,
  DXF_UNKNOWN,
  DXF_HEADER,
  DXF_TABLES,
  DXF_BLOCKS,
  DXF_ENTITY
};


typedef struct
{
  int  GroupCode;
  char Buffer[DXF_BUFSIZE];
} DXF_LINE; 


typedef struct
{
  int      Color;
  char     Name[80];
  int      VecNo;
  POINT3D *VecList;
  /* POINT3D VecList[9]; */
  int      AngleNo;
  double   AngleList[10];
  int      FloatNo;
  double   FloatList[9];
  int      IntNo;
  int      IntList[9];
  int      Flag;
  /* Text, Size, ... */
} DXF_OBJECT;


#if 0
static int DXFColors[] = 
{
   1, /* Black     */  
   8, /* Red       */
  13, /* Yellow    */
  34, /* Green     */
  25, /* Cyan      */
  22, /* Blue      */
  29, /* Magenta   */
   6, /* White     */
   2, /* Dark Grey */
   5, /* Lt. Grey  */
   6, /* White     */
};
#endif


static DXF_OBJECT PART_NewDXFObject (const char *Name)
{
  int        count;
  DXF_OBJECT Obj;

  /* Initialize object values */
  Obj.Color = 2;

  strncpy (Obj.Name, Name, 79);
  Obj.Name[79] = '\0';

  Obj.VecNo = 0;
#if 0
  for (count = 0; count < 9; ++count)
    Obj.VecList[count] = Vector_0;
#else
  Obj.VecList = NULL;
#endif
  Obj.AngleNo = 0;
  for (count = 0; count < 10; ++count)
    Obj.AngleList[count] = 0.0;
  Obj.FloatNo = 0;
  for (count = 0; count < 9; ++count)
    Obj.FloatList[count] = 0.0;
  Obj.IntNo = 0;
  for (count = 0; count < 9; ++count)
    Obj.IntList[count] = 0.0;
  Obj.Flag = 0;

  return Obj;
}


static void PART_SetDXFCoordinate (DXF_OBJECT *Obj, int Index, DXF_LINE Line)
{
  POINT3D Default = Vector_0;

  if (Obj == NULL)
    return;

  if (strcasecmp(Obj->Name, "POLYLINE") == 0)
    Index = Obj->VecNo;

  if (Index >= Obj->VecNo)
  {
    Obj->VecList = realloc (Obj->VecList, sizeof(POINT3D) * (Index+1));
    while (Obj->VecNo <= Index)
    {
      Obj->VecList[Obj->VecNo] = Default;
      (Obj->VecNo)++;
    }
  }

  if (Line.GroupCode >= 10 && Line.GroupCode <= 18)
  {
    Obj->VecList[Index].x = atof (Line.Buffer);
    PRO2 ("X[%d] = %.3f\n", Index, Obj->VecList[Index].x)
  }
  elsif (Line.GroupCode >= 20 && Line.GroupCode < 28)
  {
    Obj->VecList[Index].y = atof (Line.Buffer);
    PRO2 ("Y[%d] = %.3f\n", Index, Obj->VecList[Index].y)
  }
  elsif (Line.GroupCode >= 30 && Line.GroupCode < 37)
  {
    Obj->VecList[Index].z = atof (Line.Buffer);
    PRO2 ("Z[%d] = %.3f\n", Index, Obj->VecList[Index].z)
  }
}



static dual PART_ReadDXFLine (FILE *fp, DXF_LINE *Line)
{
  dual Result = TRUE;
  char Letter;
  int  count = 0;

  if (feof(fp))
    return FALSE;

  if (fscanf(fp, "%d", &(Line->GroupCode)) < 1)
    return FALSE;

  if (feof(fp))
    return FALSE;

  do
    Letter = fgetc (fp);
  while (Letter != '\0' && Letter != '\n');

  Letter = fgetc (fp);
  while (count < DXF_BUFSIZE-1 && Letter != '\0' && Letter != '\n')
  {
    Line->Buffer[count++] = Letter;
    Letter = fgetc (fp);
  }
  Line->Buffer[count] = '\0';

printf ("DXF #%3d: '%s'\n", Line->GroupCode, Line->Buffer);

  return Result;
}


static int PART_GetSection (FILE *fp)
{
  DXF_LINE Line;

  while (PART_ReadDXFLine(fp, &Line))
  {
    if (Line.GroupCode == 0)
    {
      if (strncasecmp(Line.Buffer, "EOF", 3) == 0)
	return DXF_NONE;
      if (strncasecmp(Line.Buffer, "SECTION", 7) == 0)
      {
	if (! PART_ReadDXFLine(fp, &Line))
	  return DXF_NONE;
      
	if (Line.GroupCode == 2)
	{
	  if (strncasecmp(Line.Buffer, "HEADER", 6) == 0)
	    return DXF_HEADER;
	  if (strncasecmp(Line.Buffer, "TABLES", 6) == 0)
	    return DXF_TABLES;
	  if (strncasecmp(Line.Buffer, "BLOCKS", 6) == 0)
	    return DXF_BLOCKS;
	  if (strncasecmp(Line.Buffer, "ENTITIES", 8) == 0)
	    return DXF_ENTITY;

	  return DXF_UNKNOWN;
	}
      }
    }
  }
  return DXF_NONE;
}


static dual PART_ReadSection (FILE *fp)
{
  DXF_LINE Line;

  while (PART_ReadDXFLine(fp, &Line))
  {
    if (Line.GroupCode == 0 && strncasecmp(Line.Buffer, "ENDSEC", 6) == 0)
      return TRUE;
  }
  return FALSE;
}
    

#if 0
static OBJECT *PART_GetObject (PART *Part, int No)
{
  OBJECT *Ptr, *Obj;

  Ptr = Part->ObjectList;
  while (Ptr != NULL && Ptr->Next != NULL && Ptr->ObjNo != No)
    Ptr = Ptr->Next;

  if (Ptr == NULL || Ptr->ObjNo != No)
  {
    Obj = OBJNewObject ();
    Obj->ObjNo    = No;
    Obj->Type     = PART_PLANE;
    Obj->VertexNo = 0;
    Obj->Data.Polyeder = GEONewPolyeder ();

    if (Part->ObjectList == NULL)
      Part->ObjectList = Obj;
    else
      Ptr->Next = Obj;

    (Part->ObjectNo)++;
  }
  else
    Obj = Ptr;

  return Obj;
}


static int PART_InsertVertex (OBJECT *Obj, POINT3D Vertex)
{
  int count = 0;

  while (count < Obj->VertexNo && 
	 GEOVectorLength (GEOVectorSub(Obj->VertexList[count], Vertex)) > 0.1)
    count++;

  if (count == Obj->VertexNo)
  {
    Obj->VertexList = (POINT3D *)realloc (Obj->VertexList,
					  sizeof(POINT3D) * (Obj->VertexNo+1));
    Obj->VertexList[Obj->VertexNo] = Vertex;
    (Obj->VertexNo)++;
  }

  return count;
}
#endif


static dual PART_AddDXFObject (OBJECT *Obj, DXF_OBJECT Object)
{
         int       count;
  static int       PolyNo = 0;
         POLYGON  *Polygon;

  if (strcasecmp(Object.Name, "3DFACE") == 0 ||
      strcasecmp(Object.Name, "POLYLINE") == 0)
  {
    if (Object.VecNo > 2)
    {
      Polygon = GEONewPolygon ();
      Polygon->Vertex = (int *)malloc (sizeof(int) * Object.VecNo);
      for (count = 0; count < Object.VecNo; ++count)
      {
	Polygon->Vertex[count] = OBJAddVertex (Obj, Object.VecList[count]);
	PRO3 ("  VERTEX: %.2f %.2f %.2f\n", Object.VecList[count].x, 
	      Object.VecList[count].y, Object.VecList[count].z)
      }
      Polygon->Points = Obj->VertexList;
      Polygon->Number = Object.VecNo;

      GEOCalcHesseForm (Polygon);
      Polygon->Info.ID = PolyNo;
      Polygon->Info.Center = GEOCalcPolygonCenter (Polygon);
      GEOTriangulatePolygon (Polygon);

      OBJAddPolygon (Obj, Polygon);
    }
  }
  elsif (strcasecmp(Object.Name, "POLYLINE") == 0)
  {
    
  }
  else
    printf ("Unsuported Object: '%s'\n", Object.Name);
  /*
    skipped Keywords:
    LINE, 3DLINE, POINT, CIRCLE, ARC, TRACE, SOLID, TEXT, 
    SHAPE, INSERT, ATTDEF, ATTRIB, POLYLINE, VERTEX, SEQEND, 
    DIMENSION,
  */

  PRO0 ("DONE\n")
  return TRUE;
}


static dual PART_ReadEntity (FILE *fp, PART *Part)
{
  DXF_LINE    Line;
  DXF_OBJECT  Obj = PART_NewDXFObject ("");
  OBJECT     *Object;
 
  Object = OBJNewObject ();
  Object->Data.Polyeder = GEONewPolyeder ();
  Object->Type = PART_PLANE;

  while (PART_ReadDXFLine(fp, &Line))
  {
    switch (Line.GroupCode)
    {
      case 0 :
	if (strncasecmp(Line.Buffer, "ENDSEC", 6) == 0)
	{
	  OBJCalcInfo (Object);
	  PARTAddObject (Part, Object);
	  return TRUE;
	}
        elsif (strncasecmp(Line.Buffer, "ENDSEQ", 6) == 0)
	{
	  PRO0 ("ENDSEQ\n")
	  OBJCalcInfo (Object);
	  PARTAddObject (Part, Object);
	}
	else
	  PART_AddDXFObject (Object, Obj);

	Obj = PART_NewDXFObject (Line.Buffer);
	break;

      case 1 :
      case 2 :
      case 3 :
      case 4 :
      case 5 :
      case 6 :
      case 7 :
      case 8 :
	/* Obj = PART_GetObject (Part, atoi(Line.Buffer)); */
      case 9 :
	break;

      default :
	if (Line.GroupCode >= 10 && Line.GroupCode <= 18)
	  PART_SetDXFCoordinate (&Obj, Line.GroupCode-10, Line);
	elsif (Line.GroupCode >= 20 && Line.GroupCode <= 28)
	  PART_SetDXFCoordinate (&Obj, Line.GroupCode-20, Line);
	elsif (Line.GroupCode >= 30 && Line.GroupCode <= 37)
	  PART_SetDXFCoordinate (&Obj, Line.GroupCode-30, Line);
	elsif (Line.GroupCode >= 40 && Line.GroupCode <= 48)  /* Float value */
        {
	  Obj.FloatList[Line.GroupCode-40] = atof (Line.Buffer);
	  Obj.FloatNo = MAX (Obj.FloatNo, Line.GroupCode-40);
	  PRO2 ("FLOAT[%d] = %.3f\n",
		Line.GroupCode-40, Obj.FloatList[Line.GroupCode-40])
	}
        elsif (Line.GroupCode >= 50 && Line.GroupCode <= 58)  /* Angular val */
	{
	  Obj.AngleList[Line.GroupCode-50] = atof (Line.Buffer);
	  Obj.AngleNo = MAX (Obj.AngleNo, Line.GroupCode-50);
	  PRO2 ("ANGLE[%d] = %.3f\n",
		Line.GroupCode-50, Obj.AngleList[Line.GroupCode-50])
	}
	elsif (Line.GroupCode == 62)                         /* Elem color */
	{
	  Obj.Color = atoi (Line.Buffer);
	  PRO1 ("COLOR=%d\n", Obj.Color)
	}
        elsif (Line.GroupCode == 66)
	{
          PRO0 ("ENTITY FOLLOWS\n")
	}
	elsif (Line.GroupCode >= 70 && Line.GroupCode <= 78) /* Int value */
	{
	  Obj.IntList[Line.GroupCode-70] = atoi (Line.Buffer);
	  Obj.IntNo = MAX (Obj.IntNo, Line.GroupCode-70);
	  PRO2 ("INT[%d] = %.3f\n", 
		Line.GroupCode-70, Obj.IntList[Line.GroupCode-70])
	}
	elsif (Line.GroupCode >= 70 && Line.GroupCode <= 78)
	{
	  /* x-, y-, z-components of extrusion direction (>= AutoCAD V10.0) */
	}
	elsif (Line.GroupCode == 999)
	{
	  /* Comment since AutoCAD Version 9.0 */
	}
	else
	  printf ("unknown GroupCode: #%d\n", Line.GroupCode);
    }
  }

  return FALSE;
}


PART *PARTLoadDXF (FILE *fp, char *PartName, dual check)
{
  int   Section = DXF_NONE;
  PART *Result = NULL;

  PRO_START ("\n")

  if (fp == NULL)
    PRO_RETURN ("invalid file-pointer", NULL)

  Result = PARTNewPart ();
  Result->PartName = (char *)malloc (sizeof(char) * strlen(PartName)+1);
  strcpy (Result->PartName, PartName);

  while (! feof(fp))
  {
    Section = PART_GetSection (fp);
    switch (Section)
    {
      /* global/default cad-parameter */
      case DXF_HEADER :
	/* 
	   skipped Keywords: 
           $ACADVER, AC..., $INSBASE, $EXTMIN, $EXTMAX, $LIMMIN, $LIMMAX,
	   $ORTHOMODE, $REGENMODE, $FILLMODE, $QTEXTMODE, $MIRRTEXT,
	   $DRAGMODE, $LTSCALE, $OSMODE, $ATTMODE, $TEXTSIZE, $TRACEWID,
	   $TEXTSTYLE, STANDARD, $CLAYER, $CELTYPE, $BYLAYER, $CECOLOR,
	   $DIMSCALE, $DIMASZ, $DIMEXO, $DIMDLI, $DIMRND, $DIMDLE, $DIMEXE
	   $DIMTP, $DIMTM, $DIMTXT, $DIMCEN, $DIMTSZ, $DIMTOL, $DIMLIM, 
	   $DIMTIH, $DIMTOH, $DIMSE1, $DIMSE2, $DIMTAD, $DIMZIN, $DIMBLK,
	   $DIMASO, $DIMSHO, $DIMPOST, $DIMAPOST, $DIMALT, $DIMALTF, $DIMLFAC,
	   $DIMTOFL, $DIMTVP, $DIMTIX, $DIMSOXD, $DIMSAH, $DIMBLK1, $DIMBLK2,
	   $DIMSTYLE, *UNNAMED, $DIMCLRD, $DIMCLRE, $DIMCLRT, $DIMTFAC, 
	   $DIMGAP, $LUNITS, $LUPREC, $SKETCHINC, $FILLETRAD, $AUNITS,
	   $AUPREC, $MENU, $ELEVATION, $PELEVATION, $THICKNESS, $LIMCHECK,
	   $BLIPMODE, $CHAMFERA, $CHAMFERB, $SKPOLY, $TDCREATE, $TDUPDATE,
	   $TDINDWG, $TDUSRTIMER, $USRTIMER, $ANGBASE, $ANGDIR, $DPMODE,
	   $PDSIZE, $PLINEWID, $COORDS, $SPLFRAME, $SPLINETYPE, $SPLINESEGS,
	   $ATTDIA, $ATTREQ, $SHANDLING, $HANDSEED, $SURFTAB1, $SURFTAB2,
	   $SURFTYPE, $SURFU, $SURFV, $UCSNAME, $UCSORG, $UCSXDIR, $UCSYDIR,
	   $PUCSNAME, $PUCSORG, $PUCSXDIR, $PUCSYDIR, 
	   $USERI1, $USERI2, $USERI3, $USERI4, $USERI5, 
	   $USERR1, $USERR2, $USERR3, $USERR4, $USERR5, $WORKDVIEW, $SHADEDGE,
	   $SHADEDIF, $TILEMODE, $MAXACTVP, $PINSBASE, $PLIMCHECK, $PEXTMIN,
	   $PEXTMAN, $PLIMMIN, $PLIMMAX, $UNITMODe, $VISRETAIN, $PLINEGEN,
	   $PSLTSCALE, $TREEDEPTH, $DWGCODEPAGE
	*/
	PART_ReadSection (fp);
	break;

      case DXF_TABLES :
	/* skipped Keywords: ENDTAB, ... */
	PART_ReadSection (fp);
	break;

      case DXF_BLOCKS :
	/* skipped Keywords: LTYPE, LAYER, STYLE, VIEW, UCS, VPORT, ENDBLK */
	PART_ReadSection (fp);
	break;

      case DXF_ENTITY :
	PART_ReadEntity (fp, Result);
	break;

      default :
	if (Section != DXF_NONE)
	{
	  printf ("Unknown Section: %d", Section);
	  PART_ReadSection (fp);
	}
	else
	  break;
    }
  }

  PARTWritePro (stdout, Result);

printf ("calc info\n");
  PARTCalcInfo (Result);
  PART_CalcCoordSys (Result);

  PRO_END
  return Result;
}



#if 0

typedef struct {
  t_LL            PolyVertices;
}               t_Polygon;


typedef struct {
  float           x;
  float           y;
  float           z;
  float           radius;
  float		  startangle;
  float		  endangle;
}               t_Sphere;


#define BUFSIZE	2048
#define FABS(x) ((x<0.0)?-x:x)

int groupcode;
char inname[80], outname[80],gfname[80];
char linbuf[BUFSIZE];
FILE *FpIn, *FpOut;
long primitives;
long degenerates;
char curobj[80];
int curcolor;
float curthick;
float xcoords[10];
float ycoords[10];
float zcoords[10];
float floats[10];
float angles[10];
int ints[10];
float max_x, max_y, max_z, min_x, min_y, min_z;
t_LL Polygons;
t_LL Lines;
t_LL Circles;
t_LL Arcs;
t_Vertex Point;
t_Sphere Circle;
t_Polygon Poly;


int main(argc, argv)
     int argc;
     char **argv;
{
  primitives = 0L;
  degenerates = 0L;
  Polygons=ConsLL();
  Circles=ConsLL();
  Arcs=ConsLL();
  Lines=ConsLL();

  printf("\n\nAutoCad DXF to Xmgf gf Data file Translator\n");
  printf("\n Converted from AutoCad DXF to DKB by Aaron A. Collins\n");
  printf("\n dxf2gf by Paul Hoad 16/7/93\n");

  if (argc < 2) {
    fprintf (stderr, "Usage: dxf2gf infile.[dxf] outfile\n");
    exit (0);
  }

  if ((FpIn = fopen (argv[1], "r")) == NULL) {
    fprintf (stderr, "File %s Does not exist\n", argv[1]);
    exit (-1);
  }

  if ((FpOut = fopen (argv[2], "w")) == NULL) {
    fprintf (stderr, "File %s Does not exist\n", argv[2]);
    exit (-1);
  }

  printf("\nPlease wait; Processing...");

  curobj[0] = '\0';	/* not working on any object currently */
  curcolor = 7;		/* and it also doesn't have a color yet... */
  max_x = max_y = max_z = -10000000.0;	/* init bounding limits */
  min_x = min_y = min_z =  10000000.0;

find:	
  while (!feof(FpIn))	/* run file up to the "ENTITIES" section */
    {
      if (getline())		/* get a group code and a line */
	goto stopit;
      if (groupcode == 0)	/* file section mark */
	{
	  if (strstr(linbuf, "EOF"))
	    goto stopit;
	  if (strstr(linbuf, "SECTION"))
	    {
	      if (getline())
		goto stopit;
	      if (groupcode != 2)
		continue;
	      if (strstr(linbuf, "ENTITIES"))
		break;
	    }
	}
    }
  while (!feof(FpIn))		/* scan ENTITIES section */
    {
      if (getline())		/* get a group code and a line */
	break;
      if (groupcode < 10)	/* cardinal group codes */
	{
	  switch(groupcode)
	    {
	    case 0: /* start of entity, table, file sep */
	      if (strstr(linbuf, "EOF"))
		{
		  writeobj(); /* dump object */
		  goto stopit;
		}
	      if (strstr(linbuf, "ENDSEC"))
		{
		  writeobj(); /* dump object */
		  goto find;
		}
	      writeobj();	/* dump old object */
	      curobj[0] = '\0'; /* reset object */
	      curcolor = 7;
	      strcpy(curobj, linbuf);	/* get new */
	      break;
	    case 1:	/* primary text value for entity (?)*/
	      break;
	    case 2: /* block name, attribute tag, etc */
	    case 3:	/* other names */
	    case 4:
	      break;
	    case 5:	/* entity handle (hex string) */
	      break;
	    case 6: /* line type name */
	      break;
	    case 7: /* text style name */
	      break;
	    case 8: /* layer name */
	      break;
	    case 9: /* variable name ID (only in header)*/
	      break;
	    }
	}
      else if (groupcode >= 10 && groupcode < 19) /* Some X coord */
	{
	  sscanf(linbuf, "%f", &(xcoords[groupcode-10]));
	  if (xcoords[groupcode-10] > max_x)
	    max_x = xcoords[groupcode-10];
	  if (xcoords[groupcode-10] < min_x)
	    min_x = xcoords[groupcode-10];
	}
      else if (groupcode >= 20 && groupcode < 29) /* Some Y coord */
	{
	  sscanf(linbuf, "%f", &(ycoords[groupcode-20]));
	  if (ycoords[groupcode-20] > max_y)
	    max_y = ycoords[groupcode-20];
	  if (ycoords[groupcode-20] < min_y)
	    min_y = ycoords[groupcode-20];
	}
      else if (groupcode >= 30 && groupcode < 38) /* Some Z coord */
	{
	  sscanf(linbuf, "%f", &(zcoords[groupcode-30]));
	  if (zcoords[groupcode-30] > max_z)
	    max_z = zcoords[groupcode-30];
	  if (zcoords[groupcode-30] < min_z)
	    min_z = zcoords[groupcode-30];
	}
      else if (groupcode == 38) /* entity elevation if nonzero */
	{
	}
      else if (groupcode == 39) /* entity thickness if nonzero */
	{
	}
      else if (groupcode >= 40 && groupcode < 49) /* misc floats */
	{
	  sscanf(linbuf, "%f", &(floats[groupcode-40]));
	}
      else if (groupcode == 49) /* repeated value groups */
	{
	}
      else if (groupcode >= 50 && groupcode < 59) /* misc angles */
	{
	  sscanf(linbuf, "%f", &(angles[groupcode-50]));
	}
      else if (groupcode == 62) /* Color number */
	{
	  sscanf(linbuf, "%6d", &curcolor);
	}
      else if (groupcode == 66) /* "entities follow" flag */
	{
	}
      else if (groupcode >= 70 && groupcode < 79) /* misc ints */
	{
	  sscanf(linbuf, "%f", &(ints[groupcode-70]));
	}
      else if (groupcode == 210 || groupcode == 220 || groupcode == 230)
	{	/* X, Y, Z components of extrusion direction */
	}
    }
stopit: 
  fclose(FpIn);
  
  /* My Output Bits */
  OutputGF(FpOut);
  printf("Finished.\n\nTotal objects written to output file: %ld\n\n", primitives);
  printf("Total degenerate triangles removed from scene: %ld\n\n", degenerates);
  printf ("X bounding values range from %f to %f\n", min_x, max_x);
  printf ("Y bounding values range from %f to %f\n", min_y, max_y);
  printf ("Z bounding values range from %f to %f\n", min_z, max_z);
  exit(0);
}


int getline()		/* read a group code and the next line from FpIn */
{
  fgets(linbuf, BUFSIZE, FpIn);	/* get a line from .DXF */
  if (feof(FpIn))
    return(1);
  sscanf(linbuf, "%3d", &groupcode);  /* scan out group code */
  fgets(linbuf, BUFSIZE, FpIn);	/* get a line from .DXF */
  if (feof(FpIn))
    return(1);
  return(0);
}

void writeobj()	/* dump out current object we should have all info on */
{
  if ((!strstr(curobj, "LINE")) && strstr(curobj,"POLYLINE")) /* One Line */
    {
      Poly.PolyVertices = ConsLL ();
      Point.x = xcoords[0]; Point.y = ycoords[0]; Point.z = zcoords[0];
      InsLastLL (Poly.PolyVertices, Point);
      Point.x = xcoords[1]; Point.y = ycoords[1]; Point.z = zcoords[1];
      InsLastLL (Poly.PolyVertices, Point);
      InsLastLL (Lines, Poly);
      finishobj(curcolor);
      return;
    }
  else if (strstr(curobj, "POINT"))	/* an itty, bitty sphere! */
    {
      Poly.PolyVertices = ConsLL ();
      Point.x = xcoords[0]; Point.y = ycoords[0]; Point.z = zcoords[0];
      InsLastLL (Poly.PolyVertices, Point);
      InsLastLL (Polygons, Poly);
      finishobj(curcolor);
      return;
    }
  else if (strstr(curobj, "CIRCLE"))	/* a VERY short cylinder! */
    {
      Circle.x = xcoords[0]; Circle.y = ycoords[0]; Circle.z = zcoords[0];
      Circle.radius = floats[0];
      InsLastLL (Circles, Circle);
      finishobj(curcolor);
      return;
    }
  else if (strstr(curobj, "ARC"))		/* not implemented for now */
    {
      Circle.x = xcoords[0]; Circle.y = ycoords[0]; Circle.z = zcoords[0];
      Circle.radius = floats[0];
      Circle.startangle = angles[0];
      Circle.endangle = angles[1];
      InsLastLL (Arcs, Circle);
      return;
    }
  else if (strstr(curobj, "TRACE"))	/* 2 back-to-back triangles */
    {
      if (checkdegen(0, 1, 2))
	{
	  degenerates++;
	  return;
	}
      Poly.PolyVertices = ConsLL ();
      Point.x = xcoords[0]; Point.y = ycoords[0]; Point.z = zcoords[0];
      InsLastLL (Poly.PolyVertices, Point);
      Point.x = xcoords[1]; Point.y = ycoords[1]; Point.z = zcoords[1];
      InsLastLL (Poly.PolyVertices, Point);
      Point.x = xcoords[2]; Point.y = ycoords[2]; Point.z = zcoords[2];
      InsLastLL (Poly.PolyVertices, Point);
      InsLastLL (Polygons, Poly);
      finishobj(curcolor);
      
      if (checkdegen(0, 3, 2))
	{
	  degenerates++;
	  return;
	}
      Poly.PolyVertices = ConsLL ();
      Point.x = xcoords[0]; Point.y = ycoords[0]; Point.z = zcoords[0];
      InsLastLL (Poly.PolyVertices, Point);
      Point.x = xcoords[1]; Point.y = ycoords[1]; Point.z = zcoords[1];
      InsLastLL (Poly.PolyVertices, Point);
      Point.x = xcoords[2]; Point.y = ycoords[2]; Point.z = zcoords[2];
      InsLastLL (Poly.PolyVertices, Point);
      InsLastLL (Polygons, Poly);
      finishobj(curcolor);
      return;
    }
  else if (strstr(curobj, "SOLID"))	/* 1 or 2 triangles */
    {
      if (checkdegen(0, 1, 2))
	{
	  degenerates++;
	  return;
	}
      Poly.PolyVertices = ConsLL ();
      Point.x = xcoords[0]; Point.y = ycoords[0]; Point.z = zcoords[0];
      InsLastLL (Poly.PolyVertices, Point);
      Point.x = xcoords[1]; Point.y = ycoords[1]; Point.z = zcoords[1];
      InsLastLL (Poly.PolyVertices, Point);
      Point.x = xcoords[2]; Point.y = ycoords[2]; Point.z = zcoords[2];
      InsLastLL (Poly.PolyVertices, Point);
      InsLastLL (Polygons, Poly);
      if (xcoords[2] == xcoords[3] && ycoords[2] == ycoords[3] && zcoords[2] == zcoords[3])
	return;	/* one triangle was enough... */
      
      if (checkdegen(0, 3, 2))
	{
	  degenerates++;
	  return;
	}
      Poly.PolyVertices = ConsLL ();
      Point.x = xcoords[0]; Point.y = ycoords[0]; Point.z = zcoords[0];
      InsLastLL (Poly.PolyVertices, Point);
      Point.x = xcoords[1]; Point.y = ycoords[1]; Point.z = zcoords[1];
      InsLastLL (Poly.PolyVertices, Point);
      Point.x = xcoords[2]; Point.y = ycoords[2]; Point.z = zcoords[2];
      InsLastLL (Poly.PolyVertices, Point);
      InsLastLL (Polygons, Poly);
      finishobj(curcolor);
      return;
    }
  else if (strstr(curobj, "TEXT"))	/* not implemented for now */
    {
      return;
    }
  else if (strstr(curobj, "SHAPE"))	/* these look very hard */
    {
      return;
    }
  else if (strstr(curobj, "BLOCK"))	/* these look very hard */
    {
      return;
    }
  else if (strstr(curobj, "ENDBLK"))	/* these look very hard */
    {
      return;
    }
  else if (strstr(curobj, "INSERT"))	/* these look very hard */
    {
      return;
    }
  else if (strstr(curobj, "ATTDEF"))	/* not implemented for now */
    {
      return;
    }
  else if (strstr(curobj, "ATTRIB"))	/* not implemented for now */
    {
      return;
    }
  else if (strstr(curobj, "POLYLINE"))	/* these look fairly hard */
    {
      Poly.PolyVertices = ConsLL ();
      return;
    }
  else if (strstr(curobj, "VERTEX"))	/* these look fairly hard */
    {
      Point.x = xcoords[0]; Point.y = ycoords[0]; Point.z = zcoords[0];
      InsLastLL (Poly.PolyVertices, Point);
      return;
    }
  else if (strstr(curobj, "SEQEND"))	/* these look fairly hard */
    {
      InsLastLL (Polygons, Poly);
      return;
    }
  else if (strstr(curobj, "3DLINE"))	/* a VERY skinny triangle! */
    {
      if (xcoords[0] == xcoords[1] && ycoords[0] == ycoords[1] && zcoords[0] == zcoords[1])
	{
	  degenerates++;
	  return;
	}
      Poly.PolyVertices = ConsLL ();
      Point.x = xcoords[0]; Point.y = ycoords[0]; Point.z = zcoords[0];
      InsLastLL (Poly.PolyVertices, Point);
      Point.x = xcoords[1]; Point.y = ycoords[1]; Point.z = zcoords[1];
      InsLastLL (Poly.PolyVertices, Point);
      InsLastLL (Lines, Poly);
      finishobj(curcolor);
      
      return;
    }
  else if (strstr(curobj, "3DFACE"))	/* 1 or 2 triangles */
    {
      if (checkdegen(0, 1, 2))
	{
	  degenerates++;
	  return;
	}
      Poly.PolyVertices = ConsLL ();
      Point.x = xcoords[0]; Point.y = ycoords[0]; Point.z = zcoords[0];
      InsLastLL (Poly.PolyVertices, Point);
      Point.x = xcoords[1]; Point.y = ycoords[1]; Point.z = zcoords[1];
      InsLastLL (Poly.PolyVertices, Point);
      Point.x = xcoords[2]; Point.y = ycoords[2]; Point.z = zcoords[2];
      InsLastLL (Poly.PolyVertices, Point);
      InsLastLL (Polygons, Poly);
      finishobj(curcolor);
      
      if (xcoords[2] == xcoords[3] && ycoords[2] == ycoords[3] && zcoords[2] == zcoords[3])
	return;	/* one triangle was enough... */
      
      if (checkdegen(0, 3, 2))
	{
	  degenerates++;
	  return;
	}
      Poly.PolyVertices = ConsLL ();
      Point.x = xcoords[0]; Point.y = ycoords[0]; Point.z = zcoords[0];
      InsLastLL (Poly.PolyVertices, Point);
      Point.x = xcoords[1]; Point.y = ycoords[1]; Point.z = zcoords[1];
      InsLastLL (Poly.PolyVertices, Point);
      Point.x = xcoords[2]; Point.y = ycoords[2]; Point.z = zcoords[2];
      InsLastLL (Poly.PolyVertices, Point);
      InsLastLL (Polygons, Poly);
      finishobj(curcolor);
      return;
    }
  else if (strstr(curobj, "DIMENSION"))	/* not implemented for now */
    {
      return;
    }
  return;	/* no current object defined... */
}


void finishobj(color)	/* conclude a DKB-style object definition */
int color;
{
  float red, green, blue;

  lookupcolor(color, &red, &green, &blue);
  printf(".");		/* activity echo (happy dots) */
  primitives++;		/* count another output file primitive */
}


void lookupcolor(color, red, green, blue) /* basic AutoCAD 9-color pallette */
int color;
float *red, *green, *blue;
{
  switch (color)
    {
    case 0:	/* black */
      *red = *green = *blue = 0.0;
      break;
    case 1: /* red */
      *red = 1.0;
      *blue = *green = 0.0;
      break;
    case 2: /* yellow */
      *red = *green = 1.0;
      *blue = 0.0;
      break;
    case 3:	/* green */
      *green = 1.0;
      *red = *blue = 0.0;
      break;
    case 4: /* cyan */
      *blue = *green = 1.0;
      *red = 0.0;
      break;
    case 5: /* blue */
      *blue = 1.0;
      *red = *green = 0.0;
      break;
    case 6: /* magenta */
      *blue = *red = 1.0;
      *green = 0.0;
      break;
    case 8:	/* dk. grey */
      *red = *green = *blue = 0.5;
      break;
    case 9: /* lt. grey */
      *red = *green = *blue = 0.75;
      break;
    case 7: /* white */
    default: /* make anything else white (?) */
      *red = *green = *blue = 1.0;
    }
  return;
}


int checkdegen(a, b, c)		/* check for degenerate triangle structure */
int a, b, c;
{
  if (
      (xcoords[a] == xcoords[b] &&
       ycoords[a] == ycoords[b] &&
       zcoords[a] == zcoords[b]) || 
      (xcoords[b] == xcoords[c] &&
       ycoords[b] == ycoords[c] &&
       zcoords[b] == zcoords[c]) || 
      (xcoords[a] == xcoords[c] &&
       ycoords[a] == ycoords[c] &&
       zcoords[a] == zcoords[c]))
    return(1);
  return(0);
}


void OutputGF(FpOut)
  FILE *FpOut;
{
  t_Polygon *OutPoly;
  t_Vertex *OutPoint;
  t_Sphere *OutSphere;
  char BaseName[100];

  strcpy(BaseName,"dx2gf");
  fprintf(stderr,"%d, Polygons\n",SizeLL(Polygons));

  fprintf (FpOut, "@Set %s\nFormat Polygon3D\n", BaseName);
  ForeachLL_M (Polygons, OutPoly) {
    fprintf (FpOut, "%d ", SizeLL (OutPoly->PolyVertices));
    ForeachLL_M (OutPoly->PolyVertices, OutPoint) {
    fprintf (FpOut, "%1.03f %1.03f %1.03f ", OutPoint->x, OutPoint->y, OutPoint->z);
    }
    fprintf (FpOut, "\n");
  }
  fprintf (FpOut, "@\n");

  fprintf(stderr,"%d, Circles\n",SizeLL(Circles));
  /* Circles */
  fprintf (FpOut, "@Set %s\nFormat Circle3D\n", BaseName);
  ForeachLL_M (Circles, OutSphere) {
      fprintf (FpOut, "%f %f %f %f\n", OutSphere->x, OutSphere->y, OutSphere->z,OutSphere->radius);
  }
  fprintf (FpOut, "@\n");

  fprintf(stderr,"%d, Lines\n",SizeLL(Lines));
  /* Lines */
  fprintf (FpOut, "@Set %s\nFormat Line3D\n", BaseName);
  ForeachLL_M (Lines, OutPoly) {
    ForeachLL_M (OutPoly->PolyVertices, OutPoint) {
    fprintf (FpOut, "%1.03f %1.03f %1.03f ", OutPoint->x, OutPoint->y, OutPoint->z);
    }
    fprintf (FpOut, "\n");
  }
  fprintf (FpOut, "@\n");

  fprintf(stderr,"%d, Arcs\n",SizeLL(Arcs));
  /* Arcs */
  fprintf (FpOut, "@Set %s\nFormat Arc3D\n", BaseName);
  ForeachLL_M (Arcs, OutSphere) {
      fprintf (FpOut, "%f %f %f %f %f %f\n", OutSphere->x, OutSphere->y, OutSphere->z,OutSphere->radius,OutSphere->startangle, OutSphere->endangle);
  }
  fprintf (FpOut, "@\n");
}
#endif

