#include <limits.h>

#include "partGlobal.h"

  
dual PART_ReadLine (FILE *fp)
{
  char letter = ' ';

  while (! feof(fp) && letter != '\n')
    letter = fgetc (fp);

  return (! feof(fp));
}


dual PART_LoadMatrix (FILE *fp, H_MATRIX *Matrix)
{
  int      row;
  H_MATRIX Matr = Matrix_1;

  for (row = 0; row < 4; ++row)
    if (fscanf(fp, "%lf %lf %lf\n", 
	           &Matr.m[row][X], &Matr.m[row][Y], &Matr.m[row][Z]) < 3)
      return FALSE;

  if (Matrix != NULL)
    *Matrix = Matr;

  return TRUE;
}


static PART *PART_Load (FILE *fp, int Version, dual check)
{
  int     count;
  int     CurveNo, SurfNo;
  int     ObjectNo;
  int     PolygonID;
  char    letter;
  OBJECT *Obj = NULL;
  OBJECT *Last = NULL;
  PART   *Result = NULL;

  PRO_START ("\n")

  if (feof(fp))
    PRO_RETURN ("unexpected end-of-file", NULL)

  Result = PARTNewPart ();

  /* Coordinate systems */
  if (fscanf(fp, "%d\n", &Result->CSNo) < 1)
  {
    PARTFreePart (Result);
    PRO_RETURN ("Missing amount of coordinate-systems", NULL)
  }
  PRO1 ("COORD-SYSTEMS: %d\n", Result->CSNo)
  if (Result->CSNo > 0)
  {
    Result->CS = (H_MATRIX *)malloc (sizeof(H_MATRIX) * Result->CSNo);
    for (count = 0; count < Result->CSNo; ++count)
      if (! PART_LoadMatrix(fp, &(Result->CS[count])))
      {
	PARTFreePart (Result);
	PRO_RETURN ("error while reading matrix", NULL)
      }
  }

  /* Curves */
  if (fscanf(fp, "%d\n", &CurveNo) < 1)
    CurveNo = -1;
  PRO1 ("CURVES:   %d\n", CurveNo)
  if (CurveNo < 0)
  {
    PARTFreePart (Result);
    PRO_RETURN ("Invalid Curve-amount", NULL)
  }
  elsif (CurveNo > 0)
    for (count = 0; count < CurveNo * 9; ++count)
      do
      {
	letter = fgetc (fp);
      } while (letter != '\n' && ! feof(fp));

  /* Surfaces */
  if (fscanf(fp, "%d\n", &SurfNo) < 1)
    SurfNo = -1;
  PRO1 ("SURFACES: %d\n", SurfNo);
  if (SurfNo != 0)
  {
    PARTFreePart (Result);
    PRO_RETURN ("Invalid surface-amount", NULL)
  }

  /* now load all objects... */
  fscanf (fp, "%d\n", &ObjectNo);
  PRO1 ("OBJECTS: %d\n", ObjectNo)
  PolygonID = 0;
  for (count = 0; count < ObjectNo; ++count)
  {
    PRO1 ("Object %d\n", count)
    switch (Version)
    {
      case  9 :
      case 10 : 
      case 11 : Obj = OBJLoadV11 (fp, Version, &PolygonID, check);
	        break;
      case 12 : Obj = OBJLoadV12 (fp, &PolygonID, check);
	        break;
      default : PRO1 ("Invalid file-version %d\n", Version)
 	        Obj = NULL;
    }
    if (Obj == NULL)
    {
      PRO1 ("ERROR loading object %d\n", count+1)
      PARTFreePart (Result);
      PRO_RETURN ("error loading object", NULL)
    }

    if (Result->ObjectList == NULL)
      Result->ObjectList = Obj;
    else
      Last->Next = Obj;

    Last = Obj;
    (Result->ObjectNo)++;
  }

  /* Calc part informations */
  PARTCalcInfo (Result);

  /* Compute vertex normals for smooth-rendering using OpenGL */
  PARTCalcNormals (Result, 30.0);

  /* Calculate part-CS... */
  PART_CalcCoordSys (Result);

  PRO_END
  return Result;
}


/* ********************************************************************* */
/* ** Loads the from file 'FileName' and give it the specified name   ** */
/* ** On success a pointer to the created part will returned.         ** */
/* ** Otherwise a NULL-pointer will be returned.                      ** */
/* ********************************************************************* */  
PART *PARTLoad (const char *FileName, char *PartName, dual check)
{
  int   Version;
  char  FileDescr[PATH_MAX+1];
  FILE *fp = NULL;
  PART *Result = NULL;

  PRO_START ("\n")

  if (FileName == NULL)
    ERROR_RETURN0 ("Invalid FileName", NULL)

  TOOLParseFileName (FileName, FileDescr, PATH_MAX);
  PRO1 ("FILENAME: %s\n", FileDescr)

  if ((fp = fopen(FileDescr, "r")) == NULL)
    ERROR_RETURN1 ("Error opening part-file:\n'%s'", FileDescr, NULL)

  if (fscanf(fp, "%d\n", &Version) < 1)
    TOOLSetErrorMsg ("Error loading part:\n%s\nFailed to read version-number",
		     FileDescr);
  else
  {
    PRO1 ("VERSION: %d\n", Version)

    Result = PART_Load (fp, Version % 100, check);
  }
  fclose (fp);

  if (Result == NULL)
    PRO_RETURN (TOOLGetErrorMsg(), NULL)

  if (PartName != NULL)
  {
    Result->PartName = (char *)malloc (sizeof(char) * (strlen(PartName)+1));
    strcpy (Result->PartName, PartName);
  }
  Result->FileName = (char *)malloc (sizeof(char) * (strlen(FileName)+1));
  strcpy (Result->FileName, FileName);

  PRO_END
  return Result;
}


/* ********************************************************************** */
/* ** Save a part with given filename. If 'triangle' is TRUE all       ** */
/* ** polygons will be saved triangulated.                             ** */
/* ********************************************************************** */ 
dual PARTSave (const char *FileName, PART *Part, dual triangle)
{
  dual  Result;
  FILE *fp = NULL;

  PRO_START (" ")
  PRO1 ("\\%s...\n", FileName)

  if ((fp = fopen(FileName, "w+")) == NULL)
    PRO_RETURN ("Error while creating file", FALSE)

  Result = PART_Write11 (fp, Part, triangle);
  /* Result = PART_Write12 (fp, Part); */

  fclose (fp);

  PRO_END
  return Result;
}


/* ********************************************************************** */
/* ** print out protocol of part-structure for testing-purpose. The    ** */
/* ** text will be added to an opened file                             ** */
/* ********************************************************************** */
void PARTWritePro (FILE *fp, PART *Part)
{
  int      count, index;
  double   px, pz, py, r, p, y;
  OBJECT  *Object;
  FEATURE *Feature;

  PRO_START (" ")

  if (fp == NULL || Part == NULL)
    PRO_ABORT ("Invalid File or Part")

  fprintf (fp, "PartName:  %s\n", Part->PartName);
  fprintf (fp, "FileName:  %s\n", Part->FileName);  
  fprintf (fp, "\n");

  WRITE_HMATRIX (fp, Part->WTP, "[WTP]")
  fprintf (fp, "\n");
  GEOFrame2RPY (Part->WTP, &px, &py, &pz, &y, &p, &r);
  fprintf (fp, "RPY(WTP): %f %f %f %f %f %f\n", px, py, pz, r, p, y);
  fprintf (fp, "\n");
  WRITE_HMATRIX (fp, Part->PTC, "[PTC]")
  fprintf (fp, "     0.0000    0.0000    0.0000    1.0000\n");
  WRITE_VECTOR3D (fp, Part->Info.Center, "Center");
  fprintf (fp, "Bubble: %f\n", Part->Info.Bubble);
  fprintf (fp, "Volume: %f\n", Part->Info.Volume);
  fprintf (fp, "\n");

  fprintf (fp, "[Objects:]\n");
  fprintf (fp, "Number: %d\n", Part->ObjectNo);
  fprintf (fp, "\n");

  count = 1;
  Object = Part->ObjectList;
  while (Object != NULL)
  {
    fprintf (fp, "Object %d:\n", count);
    fprintf (fp, "Color: %d\n", Object->ColorIndex);
    fprintf (fp, "Points: %d\n", Object->VertexNo);
    for (index = 0; index < (int)Object->VertexNo; ++index)
      fprintf (fp, "   %d: %9.5f  %9.5f  %9.5f\n", index,
	           Object->VertexList[index].x, 
	           Object->VertexList[index].y, 
	           Object->VertexList[index].z);
    fprintf (fp, "\n");

    switch (Object->Type)
    {
      case PART_CURVE :
	fprintf (fp, "   Type: CURVE\n");
	fprintf (fp, "   Lines: %d\n", Object->Data.Curve->Number);
	for (index = 0; index < Object->Data.Curve->Number; ++index)
	  fprintf (fp, "    %d", Object->Data.Curve->List[index]);
	fprintf (fp, "\n");
	break;

      case PART_PLANE :
	{
	  int      No = 1;
	  POLYGON *Ptr;

	  fprintf (fp, "  Type: POLYEDER\n");
	  fprintf (fp, "  Polygons: %d\n", Object->Data.Polyeder->Number);
	  Ptr = Object->Data.Polyeder->List;
	  while (Ptr != NULL)
	  {
	    fprintf (fp, "   Polygon %d:\n", No);
	    fprintf (fp, "     #%d:", Ptr->Number);
	    for (index = 0; index < Ptr->Number; ++index)
	      fprintf (fp, "    %d", Ptr->Vertex[index]);
	    fprintf (fp, "\n");

	    fprintf (fp, "     ID:     %d\n", Ptr->Info.ID);
	    fprintf (fp, "     Center: %9.5f  %9.5f  %9.5f\n",
		         Ptr->Info.Center.x, Ptr->Info.Center.y,
		         Ptr->Info.Center.z);
	    fprintf (fp, "     Area:   %9.5f\n", Ptr->Info.Area);
	    fprintf (fp, "     HNF:\n");
	    fprintf (fp, "              %8.4f\n", Ptr->Info.n.x);
	    fprintf (fp, "              %8.4f * x %s%8.4f = 0\n", 
		         Ptr->Info.n.y, (Ptr->Info.a>=0.0)?"+ ":"",
		         Ptr->Info.a);
	    fprintf (fp, "              %8.4f\n", Ptr->Info.n.z);

	    fprintf (fp, "     Triangles:\n");
	    for (index = 0; index < Ptr->Number-2; ++index)
	      fprintf (fp, "       %d: %d %d %d\n", index+1, 
		           Ptr->Triangle[index][0], Ptr->Triangle[index][1],
		           Ptr->Triangle[index][2]);

	    Ptr = Ptr->Next;
	    No++;
	  }
	}
        break;

      default :
	fprintf (fp, "Type: NONE\n");
    }
    fprintf (fp, "\n");
    fprintf (fp, "[Info:]\n");
    fprintf (fp, "Center: %9.5f  %9.5f  %9.5f\n",
	         Object->Info.Center.x, Object->Info.Center.y, 
	         Object->Info.Center.z);
    fprintf (fp, "Area:   %9.5f\n", Object->Info.Area);
    fprintf (fp, "Volume: %9.5f\n", Object->Info.Volume);
    fprintf (fp, "Radius: %9.5f\n", Object->Info.Bubble);

    Object = Object->Next;
    count++;
  }

  fprintf (fp, "Features:\n");
  Feature = Part->FeatureList;
  count = 1;
  while (Feature != NULL)
  {
    fprintf (fp, "%d: ", count);
    WRITE_FEATURE (fp, Feature)
    fprintf (fp, "  ID: %d\n", Feature->ID);
    fprintf (fp, "   #: %d\n", Feature->used);
    WRITE_HMATRIX (fp, Feature->CTF, "CTF")
    Feature = Feature->Next;
    count++;
  }

  PRO_END
}


/* ********************************************************************** */
/* ** Create a protocol-file and write the part's information into it. ** */
/* ********************************************************************** */
dual PARTSavePro (char *FileName, PART *Part)
{
  FILE *fp = NULL;

  PRO_START ("\n")

  if ((fp = fopen(FileName, "w+")) == NULL)
    PRO_RETURN ("Error while opening file", FALSE)

  PARTWritePro (fp, Part);

  fclose (fp);

  PRO_END
  return TRUE;
}


/* ********************************************************************** */
/* ** Write a part's geometric-data into one or more ply-files. If the ** */
/* ** Part consists out of several objects, each object will be        ** */
/* ** written into an own ply-file. The name of those files will be    ** */
/* ** created by adding the extension '.x' to the basename (where x is ** */
/* ** the number of the object).                                       ** */
/* ** INPUT:  Part     Part for which the PLY-files should be created. ** */
/* **         BaseName Basic filename (may be expanded by an           ** */
/* **                  extension)                                      ** */
/* ** OUTPUT: (dual)   TRUE: file(s) created successfully              ** */
/* **                  FALSE: error while writing a PLY-file           ** */
/* ********************************************************************** */
dual PARTWritePLY (PART *Part, char *BaseName)
{
  int      Nr = 1;
  int      count, Index;
  char     FileName[PATH_MAX+1];
  POLYGON *Polygon;
  OBJECT  *Object;
  FILE    *fp = NULL;

  PRO_START ("\n")

  Object = Part->ObjectList;
  while (Object != NULL)
  {
    if (Object->Type == PART_PLANE)
    {
      if (Part->ObjectNo == 1)
	fp = fopen(BaseName, "w+");
      else
      {
	sprintf (FileName, "%s.%d", BaseName, Nr);
	fp = fopen(FileName, "w+");
	Nr++;
      }

      if (fp == NULL)
	PRO_RETURN ("Error while opening file", FALSE)

      fprintf (fp, "%%P\n");
      for (count = X; count <= W; ++count)
	fprintf (fp, "%10.6f  %10.6f  %10.6f  %10.6f\n",
		     Matrix_1.m[N][count], Matrix_1.m[O][count],
	             Matrix_1.m[A][count], Matrix_1.m[P][count]);

      fprintf (fp, "%%S\n");
      fprintf (fp, "1.000000\n");

      fprintf (fp, "%%V\n");
      fprintf (fp, "%d\n", Object->VertexNo);
      for (Index = 0; Index < (int)Object->VertexNo; ++Index)
	fprintf (fp, "%10.6f  %10.6f  %10.6f\n", 
	             Object->VertexList[Index].x,
	             Object->VertexList[Index].y,
	             Object->VertexList[Index].z);

      fprintf (fp, "%%F\n");
      fprintf (fp, "%d\n", Object->Data.Polyeder->Number);
      Polygon = Object->Data.Polyeder->List;
      while (Polygon != NULL)
      {
	fprintf (fp, "%d", Polygon->Number);
	for (Index = 0; Index < Polygon->Number; ++Index)
	  fprintf (fp, " %d", Polygon->Vertex[Index]);
        fprintf (fp, "\n");

        Polygon = Polygon->Next;
      }

      fclose (fp);
    }
    Object = Object->Next;
  }

  PRO_END
  return TRUE;
}

  
