#include <stdarg.h>

#include <stdio.h>
#include <math.h>

#include "partGlobal.h"


#if 0
int PART_GetEdge (OBJECT *Obj, int start, int end)
{
  int count = 0;

  PRO_START (" ")

  if (Obj == NULL)
    PRO_RETURN ("invalid object", 0)

  PRO2 ("\\%d -> %d: ", start, end)

  while (count < Obj->EdgeNo)
  {
    if (Obj->Edges[count].Start == start && Obj->Edges[count].End == end)
      PRO_RETURN ("->found", count+1)
    elsif (Obj->Edges[count].Start == end && Obj->Edges[count].End == start)
      PRO_RETURN ("<-found", -(count+1))
    count++;
  }

  PRO0 ("\\realloc...")
  if (Obj->Edges == NULL)
    Obj->Edges = (EDGE_REC*)malloc (sizeof(EDGE_REC));
  else
    Obj->Edges = (EDGE_REC*)realloc (Obj->Edges,
				     sizeof(EDGE_REC) * (Obj->EdgeNo+1));
  Obj->Edges[Obj->EdgeNo].Start = start;
  Obj->Edges[Obj->EdgeNo].End   = end;
  (Obj->EdgeNo)++;
  PRO1 ("\\inserted: %d...", Obj->EdgeNo);

  PRO_END
  return Obj->EdgeNo;
}
#endif


/* ********************************************************************** */
/* ** Allocate a new part and initialize it with default-values        ** */
/* ** INPUT:  none                                                     ** */
/* ** OUTPUT: (PART*)  Pointer to new created part                     ** */
/* ********************************************************************** */
PART *PARTNewPart (void)
{
  PART *Result = NULL;

  PRO_START (" ")

  Result = (PART*)malloc (sizeof(PART));

  Result->PartName = NULL;
  Result->FileName = NULL;

  Result->CSNo = 0;
  Result->CS   = NULL;
  Result->WTP  = Matrix_1;
  Result->PTC  = Matrix_1;

  Result->Scale[X] = 1.0;
  Result->Scale[Y] = 1.0;
  Result->Scale[Z] = 1.0;

  Result->ObjectNo   = 0;
  Result->ObjectList = NULL;

  Result->Info.Center = Vector_0;
  Result->Info.Min    = Vector_0;
  Result->Info.Max    = Vector_0;
  Result->Info.Volume = 0.0;
  Result->Info.Area   = 0.0;
  Result->Info.Bubble = 0.0;

  Result->FeatureList = NULL;

  Result->NormalAngle = 30.0;
  Result->updated = FALSE;

  Result->Next = NULL;

  PRO_END
  return Result;
}


/* ********************************************************************** */
/* ** Create a new part containing the specified object.               ** */
/* ** INPUT:  Name     Name of the new part                            ** */
/* **         Object   Object data of the part.                        ** */
/* ** OUTPUT: (PART*)  destination-part                                ** */
/* ********************************************************************** */
PART *PARTCreate (const char *Name, OBJECT *ObjectList)
{
  OBJECT *Object;
  PART   *Part;

  PRO_START ("\n")

  if (ObjectList == NULL)
    PRO_RETURN ("No object defined", NULL)

  Part = PARTNewPart ();
  if (Name != NULL)
  {
    Part->PartName = (char *)malloc (sizeof(char) * (strlen(Name)+1));
    strcpy (Part->PartName, Name);
  }

  Part->ObjectList = ObjectList;

  Object = ObjectList;                          /* count number of objects */
  while (Object != NULL)
  {
    (Part->ObjectNo)++;
    Object = Object->Next;
  }

  PARTCalcInfo (Part);
  PARTCalcNormals (Part, Part->NormalAngle);

  PRO_END
  return Part;
}


/* ********************************************************************** */
/* ** Duplicate a complete part. For the features, we need to create   ** */
/* ** new pointers onto the copied data.                               ** */
/* ** INPUT:  Part     source-part                                     ** */
/* ** OUTPUT: (PART*)  destination-part                                ** */
/* ********************************************************************** */
PART *PARTCopyPart (PART *Part)
{
  int      count;
  OBJECT  *Object, *LastObj = NULL;
  OBJECT  *ObjPtr1 /* , *ObjPtr2 */; 
  FEATURE *Feature;
  FEATURE *FeatPtr, *LastFeat = NULL;
  PART    *Result;

  PRO_START ("\n")

  /* create new part-structure */
  Result = PARTNewPart ();

  /* copy data... */
  if (Part->PartName != NULL)
  {
    Result->PartName = malloc (sizeof(char) * (strlen(Part->PartName)+1));
    strcpy (Result->PartName, Part->PartName);
  }

  if (Part->FileName != NULL)
  {
    Result->FileName = malloc (sizeof(char) * (strlen(Part->FileName)+1));
    strcpy (Result->FileName, Part->FileName);
  }

  Result->CSNo = Part->CSNo;
  Result->CS   = (H_MATRIX *)malloc (sizeof(H_MATRIX) * Result->CSNo);
  memcpy (Result->Scale, Part->Scale, sizeof(double)*3);
  Result->WTP = Part->WTP;
  Result->PTC = Part->PTC;

  /* copy objects (set polygon's Vertex-pointer to copied list) */
  ObjPtr1 = Part->ObjectList;
  while (ObjPtr1 != NULL)
  {
    Object = OBJCopyObject (ObjPtr1);
    if (Result->ObjectList == NULL)
      Result->ObjectList = Object;
    else
      LastObj->Next = Object;
    LastObj = Object;

    (Result->ObjectNo)++;
    ObjPtr1 = ObjPtr1->Next;
  }

  Result->NormalAngle = Part->NormalAngle;
  Result->Info = Part->Info;

  PARTCalcNormals (Result, Result->NormalAngle);

  /* and now the features (find new vectors for object and polygons) */
  FeatPtr = Part->FeatureList;
  while (FeatPtr != NULL)
  {
    Feature = PARTNewFeature ();
    Feature->Type = FeatPtr->Type;
    Feature->ID   = FeatPtr->ID;

    for (count = 0; count < 3; ++count)
      if (FeatPtr->Polygon[count] != NULL)
        PARTGetPolygon (Result, FeatPtr->Polygon[count]->Info.ID, 
		                &Feature->Object, &Feature->Polygon[count]);

#if 0
    ObjPtr1 = Part->ObjectList;
    ObjPtr2 = Result->ObjectList;
    while (ObjPtr1 != NULL && ObjPtr1 != FeatPtr->Object)
    {
      ObjPtr1 = ObjPtr1->Next;
      ObjPtr2 = ObjPtr2->Next;
    }
    Feature->Object = ObjPtr2;
#endif    
    Feature->CTF = FeatPtr->CTF;

    if (Result->FeatureList == NULL)
      Result->FeatureList = Feature;
    else
      LastFeat->Next = Feature;
    LastFeat = Feature;

    FeatPtr = FeatPtr->Next;
  }

  Result->updated = Part->updated;

  PRO_END
  return Result;
}


/* ********************************************************************** */
/* ** Merges two parts. create decides, if a new copy should be        ** */
/* ** created.                                                         ** */
/* ** INPUT:  Part1    source-part (reference part)                    ** */
/* **         Part2    merged part                                     ** */
/* **         create   TRUE: a new part will be created                ** */
/* **                  FALSE: compose existing parts                   ** */
/* ** OUTPUT: (PART*)  new part with both subparts                     ** */
/* ********************************************************************** */
PART *PARTMergeParts (PART *Part1, PART *Part2, dual create)
{
  OBJECT   *Object;
  OBJECT   *ObjPtr, *ObjPtr2;
  FEATURE  *Feature, *NextFeature;
  PART     *Part;
  H_MATRIX  P1TP2;

  PRO_START (" ")

  if (Part1 == NULL)
    PRO_RETURN ("Part1 is empty", Part2)

  if (Part2 == NULL)
    PRO_RETURN ("Part2 is empty", Part1)

  if (create)
    Part = PARTCopyPart (Part1);
  else
    Part = Part1;

  ObjPtr = Part->ObjectList;
  while (ObjPtr != NULL && ObjPtr->Next != NULL)
    ObjPtr = ObjPtr->Next;
  ObjPtr2 = Part2->ObjectList;

  P1TP2 = GEOMatrixProd (GEOMatrixInv(Part1->WTP), Part2->WTP);

  if (create)
    while (ObjPtr2 != NULL)
    { 
      Object = OBJCopyObject (ObjPtr2);
      OBJPlace (Object, P1TP2);
		     
      if (ObjPtr == NULL)
	Part->ObjectList = Object;
      else
	ObjPtr->Next = Object;
      ObjPtr = Object;

      (Part->ObjectNo)++;
      ObjPtr2 = ObjPtr2->Next;
    }
  else
  {
    if (ObjPtr == NULL)
      Part->ObjectList = ObjPtr2;
    else
      ObjPtr->Next = ObjPtr2;

    while (ObjPtr2 != NULL)
    {
      OBJPlace (ObjPtr2, P1TP2);
      ObjPtr2 = ObjPtr2->Next;
    }

    Part->ObjectNo += Part2->ObjectNo;

    if (Part2->PartName != NULL)
      free (Part2->PartName);
    if (Part2->FileName != NULL)
      free (Part2->FileName);
    if (Part->CS != NULL)
      free (Part2->CS);

    /*
      TODO: Features kopieren statt lschen
    */
    Feature = Part2->FeatureList;
    while (Feature != NULL)
    {
      NextFeature = Feature->Next;
      free (Feature);
      Feature = NextFeature;
    }
  }

  /*
    TODO: 
    Features kopieren (noch nicht bentigt) !
  */

  PRO_END
  return Part;
}


PART *PARTSmashPart (PART *Part)
{
  int     count = 0;
  OBJECT *Object;
  PART   *NewPart, *LastPart = NULL;
  PART   *PartList = NULL;

  PRO_START (" ")

  if (Part == NULL)
    return NULL;

  Object = Part->ObjectList;
  while (Object != NULL)
  {
    NewPart = PARTNewPart ();
    PARTSetName (NewPart, "%s_%d", Part->PartName, count);
    NewPart->ObjectList = OBJCopyObject (Object);
    NewPart->ObjectNo   = 1;
    PARTCalcInfo (NewPart);

    if (PartList == NULL)
      PartList = NewPart;
    else
      LastPart->Next = NewPart;
    LastPart = NewPart;

    Object = Object->Next;
    count++;
  }

  PRO_END;
  return PartList;
}


/* ********************************************************************** */
/* ** Scales a part absolute in x-, y- and z-direction. If a scale-    ** */
/* ** value is zero, this size in this direction istn't changed.       ** */
/* ** erased.                                                          ** */
/* ** INPUT:  Part         the part-data                               ** */
/* **         sx, sy, sz   ABSOLUTE scale-values in x-, y- and z-dir   ** */
/* ** OUTPUT: none                                                     ** */
/* ********************************************************************** */
void PARTScalePart (PART *Part, double sx, double sy, double sz)
{
  double  Scale[3] = {1.0, 1.0, 1.0};
  OBJECT *Obj;

  if (sx != 0.0)
    Scale[X] = sx / Part->Scale[X];
  if (sy != 0.0)
    Scale[Y] = sy / Part->Scale[Y];
  if (sz != 0.0);
    Scale[Z] = sz / Part->Scale[Z];

  Obj = Part->ObjectList;
  while (Obj != NULL)
  {
    OBJScale (Obj, Scale[X], Scale[Y], Scale[Z]);
    Obj = Obj->Next;
  }

  if (sx != 0.0)
    Part->Scale[X] = sx;
  if (sy != 0.0)
    Part->Scale[Y] = sy;
  if (sz != 0.0)
    Part->Scale[Z] = sz;

  PARTCalcInfo (Part);
}


void PARTSetName (PART *Part, const char *Name, ...)
{
  char    PartName[1024];
  va_list Arguments;

  if (Part == NULL)
    return;

  if (Part->PartName != NULL)
    free (Part->PartName);

  va_start (Arguments, Name);
  vsprintf (PartName, Name, Arguments);
  va_end (Arguments);

  Part->PartName = (char *)malloc (sizeof(char) * (strlen(PartName)+1));
  strcpy (Part->PartName, PartName);
}


void PARTSetOrigin (PART *Part, H_MATRIX Matrix)
{
  OBJECT *Obj;

  if (Part == NULL)
    return;

  Obj = Part->ObjectList;
  while (Obj != NULL)
  {
    OBJPlace (Obj, Matrix);
    Obj = Obj->Next;
  }
}


/* ********************************************************************** */
/* ** Returns the part's scale factor. NULL may be given when          ** */
/* ** single directions aren't needed.                                 ** */
/* ** INPUT:  Part        the part-data                                ** */
/* ** IN/OUT: sx, sy, sz  Pointer for returned data                    ** */
/* ** OUTPUT: none                                                     ** */
/* ********************************************************************** */
void PARTGetScale (PART *Part, double *sx, double *sy, double *sz)
{
  if (sx != NULL)
    *sx = Part->Scale[X];

  if (sy != NULL)
    *sy = Part->Scale[Y];

  if (sz != NULL)
    *sz = Part->Scale[Z];
}


/* ********************************************************************** */
/* ** deallocate a part's memory. All objects and features will be     ** */
/* ** erased.                                                          ** */
/* ** INPUT:  Part  the part-data                                      ** */
/* ** OUTPUT: none                                                     ** */
/* ********************************************************************** */
void PARTFreePart (PART *Part)
{
  OBJECT  *Obj, *NextObj;
  FEATURE *Feature, *NextFeature;

  PRO_START ("\n")

  if (Part == NULL)
    PRO_ABORT ("Empty Part")

  if (Part->PartName != NULL)
    free (Part->PartName);

  if (Part->FileName != NULL)
    free (Part->FileName);

  if (Part->CS != NULL)
    free (Part->CS);

  Obj = Part->ObjectList;
  while (Obj != NULL)
  {
    NextObj = Obj->Next;
    OBJFreeObject (Obj);
    Obj = NextObj;
  }

  Feature = Part->FeatureList;
  while (Feature != NULL)
  {
    NextFeature = Feature->Next;
    free (Feature);
    Feature = NextFeature;
  }

  free (Part);

  PRO_END
}


/* ********************************************************************** */
/* ** Add some (one or more) objects to the part's objectlist          ** */
/* ** INPUT:  Part     Pointer to part-structure                       ** */
/* **         Object   pointer to new object(s)                        ** */
/* ** OUTPUT: none                                                     ** */
/* ********************************************************************** */
void PARTAddObject (PART *Part, OBJECT *ObjectList)
{
  OBJECT *Obj;

  if (Part == NULL || ObjectList == NULL)
    return;

  if ((Obj = Part->ObjectList) == NULL)
    Part->ObjectList = ObjectList;
  else
  {
    while (Obj->Next != NULL)
      Obj = Obj->Next;
    Obj->Next = ObjectList;
  }

  Obj = ObjectList;
  while (Obj != NULL)
  {
    (Part->ObjectNo)++;
    Obj = Obj->Next;
  } 
}


/* ********************************************************************** */
/* ** Check, if there exists a polygon which includes the given x-, y- ** */
/* ** and z-coordinates. On success TRUE will be returned and 'Object  ** */
/* ** points onto the object and Polygon points onto the polygon.      ** */
/* ** The coordinates must be given relative to the part's coordinate- ** */
/* ** system.                                                          ** */
/* ** 'Object' and 'Polygon' are optional parameters. If they are NULL ** */
/* ** this routine performs only a boolean-check.                      ** */
/* ** INPUT:  Part     Pointer to part-structure                       ** */
/* **         x, y, z  coordinates of the given point (interpreted as  ** */
/* **                  local-coordinates)                              ** */
/* ** OUTPUT: (bool)   TRUE, if Polygon was found, FALSE if not        ** */
/* **         Object   Object including the polygon                    ** */
/* **         Polygon  Polygon, which includes the given point         ** */
/* ********************************************************************** */
dual PARTFindPolygon (PART *Part, double x, double y, double z, 
		      OBJECT **Object, POLYGON **Polygon)
{
  POLYGON *Poly = NULL;
  OBJECT  *Obj;

  Obj = Part->ObjectList;
  while (Obj != NULL)
  {
    if (Obj->Type == PART_PLANE &&
	(Poly = GEOFindPolygon(Obj->Data.Polyeder, x, y, z)) != NULL)
    {
      if (Object != NULL)
	*Object = Obj;
      if (Polygon != NULL)
	*Polygon = Poly;

      return TRUE;
    }
    Obj = Obj->Next;
  }

  return FALSE;
}


/* ********************************************************************** */
/* ** Locate a polygon with the given ID-Number, which is unique.      ** */
/* ** Object and Polygon are optional. So if NULL is passed the data   ** */
/* ** won't be returned.                                               ** */
/* ** INPUT:  Part     Pointer to the part                             ** */
/* **         ID       Identifier of searched polygon                  ** */
/* ** OUTPUT: (dual)   TRUE, if polygon was found, FALSE otherwise     ** */
/* **         Object   Object, in which the polygon was found          ** */
/* **         Polygon  Polygon with this ID                            ** */
/* ********************************************************************** */
dual PARTGetPolygon (PART *Part, int ID, OBJECT **Object, POLYGON **Polygon)
{
  dual     Result = FALSE;
  OBJECT  *Obj;
  POLYGON *Poly;

  PRO_START (" ")

  if (ID < 0)
    PRO_RETURN ("Invalid ID", FALSE)

  Obj = Part->ObjectList;
  while (! Result && Obj != NULL)
  {
    if (Obj->Type == PART_PLANE)
      if ((Poly = GEOGetPolygon(Obj->Data.Polyeder, ID)) != NULL)
      {
	if (Object != NULL)
	  *Object = Obj;
	if (Polygon != NULL)
	  *Polygon = Poly;

	Result = TRUE;
      }

    Obj = Obj->Next;
  }
  PRO1 ("\\Polygon%sfound...", Result? " ":" NOT ")

  PRO_END
  return Result;
}


/* ********************************************************************* */
/* ** Place part-coordinate-system in center of biggest (sub-)object  ** */
/* ********************************************************************* */
void PART_CalcCoordSys (PART *Part)
{
  double  MaxVol = 0;
  OBJECT *Object;
  OBJECT *MaxObj = NULL;

  if (Part == NULL)
    return;

  /* place Part-coord-sys in center of object with biggest bubble */
  Object = Part->ObjectList;
  while (Object != NULL)
  {
    if (Object->Info.Volume >= MaxVol)
    {
      MaxObj = Object;
      MaxVol = Object->Info.Volume;
    }
    Object = Object->Next;
  }

  Part->PTC.m[P][X] = MaxObj->Info.Center.x;
  Part->PTC.m[P][Y] = MaxObj->Info.Center.y;
  Part->PTC.m[P][Z] = MaxObj->Info.Center.z;
}


/* ********************************************************************** */
/* ** Calculate a part's area, volume, center of gravity and           ** */
/* ** bubble-radius using the data of it's objects.                    ** */
/* ********************************************************************** */ 
void PARTCalcInfo (PART *Part)
{
  unsigned int  count;
  VECTOR3D      Vector;
  OBJECT       *Object;

  if (Part == NULL)
    return;

  Part->Info.Center = Vector_0;
  Part->Info.Min    = Vector_0;
  Part->Info.Max    = Vector_0;
  Part->Info.Area   = 0.0;
  Part->Info.Volume = 0.0;
  Part->Info.Bubble = 0.0;

  /* first, let's compute the parts volume as the sum of all objects and  */
  /* the center                                                           */
  Object = Part->ObjectList;
  while (Object != NULL)
  {
    Part->Info.Volume += Object->Info.Volume;
    /* The part's area is assumed to be the sum of all object-areas. That */
    /* might be incorrect, but up to now there's no use of this value yet */
    Part->Info.Area   += Object->Info.Area;

    Part->Info.Center.x += Object->Info.Center.x * Object->Info.Volume;
    Part->Info.Center.y += Object->Info.Center.y * Object->Info.Volume;
    Part->Info.Center.z += Object->Info.Center.z * Object->Info.Volume;

    Object = Object->Next;
  }

  if (Part->Info.Volume != 0.0)
  {
    Part->Info.Center.x /= Part->Info.Volume;
    Part->Info.Center.y /= Part->Info.Volume;
    Part->Info.Center.z /= Part->Info.Volume;
  }

  /* Find out the largest distance between center and any object-         */
  /* point. We needn't use the Object->bubble-info, because it uses the   */
  /* object's center, which might not be the part's center.               */
  /* Also compute the part's range                                        */
  if ((Object = Part->ObjectList) != NULL)
  {
    Part->Info.Min = Object->Info.Min;
    Part->Info.Max = Object->Info.Max;
  }
  while (Object != NULL)
  {
    for (count = 0; count < Object->VertexNo; ++count)
    {
      Vector = GEOVectorSub (Object->VertexList[count], Part->Info.Center);
      Part->Info.Bubble = MAX (Part->Info.Bubble, VEC_LEN(Vector));
    }

    Part->Info.Min.x = MIN (Part->Info.Min.x, Object->Info.Min.x);
    Part->Info.Max.x = MAX (Part->Info.Max.x, Object->Info.Max.x);
    Part->Info.Min.y = MIN (Part->Info.Min.y, Object->Info.Min.y);
    Part->Info.Max.y = MAX (Part->Info.Max.y, Object->Info.Max.y);
    Part->Info.Min.z = MIN (Part->Info.Min.z, Object->Info.Min.z);
    Part->Info.Max.z = MAX (Part->Info.Max.z, Object->Info.Max.z);

    Object = Object->Next;
  }
}


/* ********************************************************************** */
/* ** Calculates a part's vertex-normals used for e.g. OpenGL-         ** */
/* ** rendering. zero-vectors will cause to use the polygon-normal     ** */
/* ** INPUT:  Part    Pointer to specific part                         ** */
/* ** OUTPUT: none                                                     ** */
/* ********************************************************************** */
/* TODO: 
   Verbesserungen: Vergleiche Nachbarn, gltte Seite mit geringerem Winkel.
*/

void PARTCalcNormals (PART *Part, double Range)
{
  int          count;
  int          Points;
  double       Angle, Min;
  VECTOR3D     Sum, Avg;
  POLYGONELEM *Ptr1, *Ptr2;
  POLYGONELEM *Poly1, *Poly2;
  POLYGONELEM *Next;
  POINTELEM   *PointList;
  OBJECT      *Object;

  /* Clear old normal vectors */
  Object = Part->ObjectList;
  while (Object != NULL)
  {
    if (Object->Type == PART_PLANE)
      GEOFreePolyederNormals (Object->Data.Polyeder);
    Object = Object->Next;
  }

  if (Range == 0.0)
    return;

  Object = Part->ObjectList;
  while (Object != NULL)
  {
    Points = Object->VertexNo;
    PRO1 ("<= Object-Points: %d\n", Points)

    /* Create a list of all polygons using this vertex for each vertex. */ 
    PointList = (POINTELEM *)malloc (sizeof(POINTELEM) * Points);
    for (count = 0; count < Points; ++count)
    {
      /* Initialize point-list */
      PointList[count].Number      = 0;
      PointList[count].PolygonList = NULL;
    }
    if (Object->Type == PART_PLANE)
      GEOBuildPolygonList (Object->Data.Polyeder, PointList);

    for (count = 0; count < Points; ++count)
    {
      PRO2 ("Point %d:  %d Polygons: ", count+1, PointList[count].Number)

      /* Compute averaging normal vector */
      Avg = Vector_0;
      Ptr1 = PointList[count].PolygonList;
      while (Ptr1 != NULL)
      {
	Avg = GEOVectorAdd (Avg, Ptr1->Polygon->Info.n);
	Ptr1 = Ptr1->Next;
      }

      /* If the distance of each polygon is less than half the specified range,
	 the average vector is suited for all the polygons */
      Ptr1 = PointList[count].PolygonList;
      while (Ptr1 != NULL && 
	     GEOVectorAngle(Avg, Ptr1->Polygon->Info.n) < 0.5 * Range)
	Ptr1 = Ptr1->Next;

      if (Ptr1 == NULL)
      {
	PRO0 ("\\Set average normal")
	Ptr1 = PointList[count].PolygonList;
	while (Ptr1 != NULL) 
	{
	  GEOSetVertexNormal (Ptr1->Polygon, count, Avg);
	  Ptr1 = Ptr1->Next;
	}
      }
      else              /* The average normal does not fits our range value */
      {
	/* Two different strategies: */
#if 1
	/* Strategy 1: (seems to be better, but may vary from case to case)
	   Find the two adjacent polygons which are not yet processed and 
	   have the smallest angle. If they exist and the angle is smaller
           than the defined range sets the normal vectors. Repeat until there
           are no unprocessed polygons left or angle is larger than range.
	   So we will process the polygons with ascending angles.
	*/
	do 
	{
	  Min = 0.0;
	  Poly1 = NULL;
          Poly2 = NULL;
	  Ptr1 = PointList[count].PolygonList;
	  while (Ptr1 != NULL)
	  {
	    if (! Ptr1->done)
	    {
	      Ptr2 = Ptr1->Next;
	      while (Ptr2 != NULL)
	      {
		if (! Ptr2->done &&
		    GEOAreNeighbours(Ptr1->Polygon, Ptr2->Polygon))
		{
		  Angle = GEOVectorAngle (Ptr1->Polygon->Info.n,
					  Ptr2->Polygon->Info.n);
		  if (Poly1 == NULL || Poly2 == NULL || Angle < Min)
		  {
		    Min   = Angle;               /* Store minimum angle and */
		    Poly1 = Ptr1;
		    Poly2 = Ptr2;                /* the appropriate polygon */
		  }
		}
		Ptr2 = Ptr2->Next;
	      }
	    }
	    Ptr1 = Ptr1->Next;
	  }

	  if (Poly1 != NULL && Poly2 != NULL)
	  {
	    PRO1 ("\\%.2f", Min)
	    if (Min > 0.0 && Min <= Range+0.001)
	    {
	      Sum = GEOVectorAdd (Poly1->Polygon->Info.n,
				  Poly2->Polygon->Info.n);
	      Avg = GEOVectorNorm (GEOVectorDiv(Sum, 2.0));
	      GEOSetVertexNormal (Poly1->Polygon, count, Avg);
	      GEOSetVertexNormal (Poly2->Polygon, count, Avg);
	    }
	    Poly1->done = TRUE;
	    Poly2->done = TRUE;

	    /* Abort, if angle is larger than the range */
	    if (Min > Range+0.001)  
	      Poly1 = Poly2 = NULL;
	  }
	  else
	    PRO0 ("\\not found")
	} while (Poly1 != NULL && Poly2 != NULL);
#else
	/* Strategy 2:
	   Step through the polygonlist. For each polygon search for another
	   adjacent polygon with the smalles angle. This means we process the
	   local minimum - not the global, like the method above. The order
	   in which the polygons are processed depends on the order in the
	   polygonlist.
	*/
	Poly1 = PointList[count].PolygonList;
	while (Poly1 != NULL)
	{
       	  if (! Poly1->done)      /* Vertex of this polygon wasn't processed */
	  {
	    /* Find neighbour-polygon with smallest angle between normal-vec */
	    Poly2 = NULL;
	    Min   = 0.0;
	    Ptr2  = PointList[count].PolygonList;
	    while (Ptr2 != NULL)
	    {
	      if (Ptr2 != Poly1 && ! Ptr2->done && 
		  GEOAreNeighbours(Poly1->Polygon, Ptr2->Polygon))
	      {
		Angle = GEOVectorAngle (Poly1->Polygon->Info.n,
					Ptr2->Polygon->Info.n);
		if (Poly2 == NULL || Angle < Min)
	        {
		  Min   = Angle;                 /* Store minimum angle and */
		  Poly2 = Ptr2;                  /* the appropriate polygon */
		}
	      }
	      Ptr2 = Ptr2->Next;
	    }

	    if (Poly2 != NULL)	           /* A neighbour polygon was found */
	    {
	      PRO1 ("\\ %.2f", Min)
	      if (Min > 0.0 && Min <= Range+0.001)
	      {
		PRO0 ("\\*")
		Sum = GEOVectorAdd (Poly1->Polygon->Info.n, 
				    Poly2->Polygon->Info.n);
		Avg = GEOVectorNorm (GEOVectorDiv(Sum, 2.0));
		GEOSetVertexNormal (Poly1->Polygon, count, Avg);
 		GEOSetVertexNormal (Poly2->Polygon, count, Avg);
		Poly1->done = TRUE;
		Poly2->done = TRUE;
	      }
	    }
	    else
	      PRO0 ("\\not found")
	  }
	  Poly1 = Poly1->Next;
	}                                                 
#endif
      }                                 /* else */
      PRO0 ("\\\n")
    }

    /* clean up PointList data */
    for (count = 0; count < Points; ++count)
    {
      Poly1 = PointList[count].PolygonList;
      while (Poly1 != NULL)
      {
	Next = Poly1->Next;
	free (Poly1);
	Poly1 = Next;
      }
    }

    Object = Object->Next;
    free (PointList);
  }
}


/* ********************************************************************** */
/* ** Set a part's World-to-Part-frame                                 ** */
/* ** INPUT:  Part    Pointer to specific part                         ** */
/* **         Matrix  New W2P-frame                                    ** */
/* ** OUTPUT: none                                                     ** */
/* ********************************************************************** */
void PARTSetFrame (PART *Part, H_MATRIX Matrix)
{
  if (Part != NULL)
    Part->WTP = Matrix;
}


/* ********************************************************************** */
/* ** Counts a part's polygons                                         ** */
/* ** INPUT:  Part   Pointer to the part-data                          ** */
/* ** OUTPUT: (int)  Number of polygons in the part                    ** */
/* ********************************************************************** */ 
int PARTPolygonAmount (PART *Part)
{
  int     Result = 0;
  OBJECT *Object;

  PRO_START ("")

  if (Part == NULL)
    PRO_RETURN ("empty part", 0)

  Object = Part->ObjectList;
  while (Object != NULL)
  {
    if (Object->Type == PART_PLANE)
      Result += Object->Data.Polyeder->Number;
    Object = Object->Next;
  }

  PRO_END
  return Result;
}

/* ********************************************************************** */
/* ** Counts a part's triangles                                        ** */
/* ** INPUT:  Part   Pointer to the part-data                          ** */
/* ** OUTPUT: (int)  Number of triangles in the part.                  ** */
/* ********************************************************************** */ 
int PARTTriangleAmount (PART *Part)
{
  int     Result = 0;
  OBJECT *Object;

  PRO_START ("")

  if (Part == NULL)
    PRO_RETURN ("empty part", 0)

  Object = Part->ObjectList;
  while (Object != NULL)
  {
    if (Object->Type == PART_PLANE)
      Result += OBJGetTriangles (Object);
    Object = Object->Next;
  }

  PRO_END
  return Result;
}


/* ********************************************************************** */
/* ** Counts the number of points in a part                            ** */
/* ** INPUT:  Part   Pointer to complete part                          ** */
/* ** OUTPUT: (int)  Number of part-points                             ** */
/* ********************************************************************** */ 
int PARTPointAmount (PART *Part)
{
  int     Result = 0;
  OBJECT *Object;

  if (Part == NULL)
    return 0;

  Object = Part->ObjectList;
  while (Object != NULL)
  {
    Result += Object->VertexNo;
    Object = Object->Next;
  }

  return Result;
}


void PARTSetColor (PART *Part, int Color)
{
  OBJECT *Obj;

  Obj = Part->ObjectList;
  while (Obj != NULL)
  {
    Obj->ColorIndex = Color;
    Obj = Obj->Next;
  }
}
