#include <assert.h>

#include <libsim3d.h>


#define func(x) ((x)*(x)*(x) - (x)) /* x^3 - x */


CURVE::CURVE ()
{
}


double CURVE::GetValue (double Time)
{
  if (Time < 0.0 || Time > 1.0)
    return 0.0;

  return 1.0;
}


SPLINE::SPLINE (int Number, POINT2D *Points) :
  CURVE  (),
  Number (Number)
{
  int    count;
  double Factor;

  // The x-coordinate needs to have ascending order !

  assert (Number > 1 && Points != NULL);

  this->Points = new POINT2D [Number];
  Factor = Points[Number-1].x - Points[0].x;
  for (count = 0; count < Number; count++)
  {
    // Norm x-range to 0..1
    this->Points[count].x = (Points[count].x - Points[0].x) / Factor;
    this->Points[count].y = Points[count].y;
  }

  // Compute distances between two adjacent points 
  u = new double[Number];
  for (count = 0; count < Number-1; ++count)
    u[count] = Points[count+1].x - Points[count].x;

  // Calc weighted values for the direct neighbours
  d = new double[Number];
  for (count = 1; count < Number-1; ++count)
    d[count] = 2.0 * (Points[count+1].x - Points[count-1].x);

  w = new double[Number];
  for (count = 1; count < Number-1; ++count)
    w[count] = 6.0 * (  (Points[count+1].y - Points[count  ].y) / u[count  ]
		      - (Points[count  ].y - Points[count-1].y) / u[count-1]);

  for (count = 1; count < Number-2; ++count)
  { 
    w[count+1] -= w[count] * u[count]/d[count];
    d[count+1] -= u[count] * u[count]/d[count];
  }

  // Combine the upper values:
  k = new double[Number];
  k[0] = 0.0;
  for (count = Number-2; count > 0; count--)
    k[count] = (w[count] - u[count] * k[count+1]) / d[count];
  k[Number-1] = 0.0;
}


SPLINE::~SPLINE ()
{
  if (Points != NULL)
    delete Points;

  if (d != NULL)
    delete d;

  if (u != NULL)
    delete u;

  if (w != NULL)
    delete w;

  if (k != NULL)
    delete k;
}


double SPLINE::GetValue (double Time)
{
  int    count = 0;
  double t, Result;

  // Check, if time is in the range
  if (Time < 0.0 || Time > 1.0) 
    return 0.0;

  while (count < Number-1 && Points[count+1].x < Time)
    count++;

  t = (Time - Points[count].x) / u[count];
  Result =    t  * Points[count+1].y + 
           (1-t) * Points[count].y + 
           u[count] * u[count] * (func(t)*k[count+1] + 
				  func(1-t)*k[count]) / 6.0;

  return Result;
}
