#include <assert.h>

#include "libsim3d.h"


#if 0
#define TEST_MATRIX(Matr) \
        { \
          assert ((Matr)[N][W] == 0.0);  \
          assert ((Matr)[O][W] == 0.0);  \
          assert ((Matr)[A][W] == 0.0);  \
          assert ((Matr)[P][W] == 1.0 || (Matr)[P][W] == 0.0);  \
        }


MATRIX::MATRIX (double Value /* = 1.0 */) :
  Rows     (4),
  Cols     (4)
  normal   (HVECTOR(Value,   0.0,   0.0, 0.0)),
  open	   (HVECTOR(  0.0, Value,   0.0, 0.0)),
  approach (HVECTOR(  0.0,   0.0, Value, 0.0)),
  position (HVECTOR(  0.0,   0.0,   0.0, 1.0))
{
}


MATRIX::MATRIX (const HVECTOR &n, const HVECTOR &o, 
		const HVECTOR &a, const HVECTOR &p) :
  Rows     (4),
  Cols     (4),
  normal   (n),
  open     (o),
  approach (a),
  position (p)
{
  TEST_MATRIX (*this);
}


MATRIX::MATRIX (double n1, double n2, double n3,
		double o1, double o2, double o3,
		double a1, double a2, double a3,
		double p1, double p2, double p3) :
  Rows     (4),
  Cols     (4),
  normal   (n1, n2, n3, 0.0), 
  open     (o1, o2, o3, 0.0),
  approach (a1, a2, a3, 0.0),
  position (p1, p2, p3, 1.0)
{
}


MATRIX::~MATRIX ()
{
}


void MATRIX::Init (double Value)
{
  TEST_MATRIX (*this);

  normal  [X] = normal  [Y] = normal  [Z] = Value;
  open    [X] = open    [Y] = open    [Z] = Value;
  approach[X] = approach[Y] = approach[Z] = Value;
  position[X] = position[Y] = position[Z] = Value;

  Rows = 4;
  Cols = 4;
}


int MATRIX::operator == (const MATRIX &Matr) const
{
  TEST_MATRIX (*this);
  TEST_MATRIX (Matr);

  return ((normal   == Matr.normal)   &&
	  (open	    == Matr.open)	   &&
	  (approach == Matr.approach) &&
	  (position == Matr.position));
}


MATRIX& MATRIX::operator += (const MATRIX &Matr)
{
  TEST_MATRIX (*this);
  TEST_MATRIX (Matr);

  normal   += Matr.normal;
  open     += Matr.open;
  approach += Matr.approach;
  position += Matr.position;
  position[W] = 1.0;            // correct HVECTOR w value

  return *this;
}


MATRIX& MATRIX::operator -= (const MATRIX &Matr)
{
  TEST_MATRIX (*this);
  TEST_MATRIX (Matr);

  normal   -= Matr.normal;
  open	   -= Matr.open;
  approach -= Matr.approach;
  position -= Matr.position;
  position[W] = 1.0;            // correct HVECTOR w value

  return *this;
}


MATRIX MATRIX::operator * (const MATRIX &Matr) const
{
  int spalte, zeile, k;

  TEST_MATRIX (*this);
  TEST_MATRIX (Matr);

  MATRIX Result (0.0);               // result of multiplication

  // rotation
  for (spalte = 0; spalte <= 2; ++spalte)
    for (zeile = 0; zeile <= 2; ++zeile)
      for (k = 0; k <= 3; k++)
        Result[spalte][zeile] += (*this)[k][zeile] * Matr[spalte][k];

  // translation
  for (zeile = 0; zeile <= 2; ++zeile)
    for (k = 0; k <= 3; k++)
      Result[3][zeile] += (*this)[k][zeile] * Matr[3][k];
 
  return Result;
}


int MATRIX::operator != (const MATRIX &Matr) const
{
  return !(*this == Matr);
}


MATRIX MATRIX::operator + (const MATRIX &Matr) const
{
  return MATRIX (*this) += Matr;
}


MATRIX MATRIX::operator - (const MATRIX &Matr) const
{
  return MATRIX (*this) -= Matr;
}


MATRIX& MATRIX::operator *= (const MATRIX &Matr)
{
  return *this = *this * Matr;
}


MATRIX MATRIX::operator + () const
{
  TEST_MATRIX (*this);

  return MATRIX (+normal, +open, +approach, +position);
}


MATRIX MATRIX::operator - () const
{
  TEST_MATRIX (*this);

  return MATRIX (-normal, -open, -approach, -position);
}


HVECTOR MATRIX::operator * (const HVECTOR &Vec) const
{
  int zeile, k;

  TEST_MATRIX (*this);

  HVECTOR Result (0.0, 0.0, 0.0, Vec[W]);

  for (zeile = 0; zeile <= 2; ++zeile)  
    for (k = 0; k <= 3; ++k)
      Result[zeile] += (*this)[k][zeile]*Vec[k];
 
  return Result;
}


HVECTOR &MATRIX::operator [] (int i)
{
  TEST_MATRIX (*this);

  assert (i >= 0 && i <= 3);

  switch(i) 
  {
    case  0 : return normal;
    case  1 : return open;
    case  2 : return approach;
    case  3 :
    default : return position;
  }

  return normal;
}


HVECTOR MATRIX::operator [] (int i) const
{
  TEST_MATRIX (*this);

  switch(i) 
  {
    case  0 : return normal;
    case  1 : return open;
    case  2 : return approach;
    case  3 : return position;
  }

  return normal;
}


void MATRIX::Write (const char *Title /* = NULL */) const
{
  PRO4 ("[%6.4f %6.4f %6.4f %6.4f]\n",
	normal[X], open[X], approach[X], position[X])
  PRO4 ("[%6.4f %6.4f %6.4f %6.4f]\n",
	normal[Y], open[Y], approach[Y], position[Y])
  PRO4 ("[%6.4f %6.4f %6.4f %6.4f]\n",
	normal[Z], open[Z], approach[Z], position[Z])
  PRO4 ("[%6.4f %6.4f %6.4f %6.4f]\n",
	normal[W], open[W], approach[W], position[W])
}


int MATRIX::operator <= (const MATRIX &Matr) const
{
  TEST_MATRIX (*this);
  TEST_MATRIX (Matr);
 
  return ((normal   <= Matr.normal)   &&
	  (open	    <= Matr.open)	   &&
	  (approach <= Matr.approach) &&
	  (position <= Matr.position));
}

MATRIX operator * (double s, const MATRIX &Matr)
{
  TEST_MATRIX (Matr);

  return MATRIX (s * Matr.normal,   s * Matr.open, 
		 s * Matr.approach, s * Matr.position);
}
#endif

