#include <math.h>
#include <assert.h>

#include <debug.h>

#include "libsim3d.h"


double Alpha[5] = {0.0, 90.0, 0.0, 0.0, 90.0};


MITSUBISHI::MITSUBISHI (INTERFACE *Interface) :
  ROBOT (Interface),
  /*
  DHP_D (0.0, 165.0,   0.0,   0.0,  0.0, 0.0),
  DHP_A (0.0,  85.0, 220.0, 160.0, 30.5, 0.0)
  */
  DHP_D (0.0,   0.0,   0.0,  0.0, 0.0, 0.0),
  DHP_A (0.0, 220.0, 160.0, 30.5, 0.0, 0.0)
{
  OldSpeed = GetSpeed ();               // save speed-settings

  OpenHand ();
}


MITSUBISHI::~MITSUBISHI ()
{
  SetSpeed (OldSpeed);                  // Restore old speed
}


JOINTS MITSUBISHI::InvKin (const FRAME &Frame)
{
  double  s0, c0, s1, s2, c2, s4, c4, s123, c123;
  double  s11, c11, t11, s12, t12;
  double  Joint123, d;
  JOINTS  Joint;
  HVECTOR  p2;
  FRAME   TCP;

  PRO_START (" ")

  TCP = (! Base) * Frame * (! Offset);

  // ================ Joint 1 ==================
  Joint[0] = atan2 (TCP[P][Y], TCP[P][X]);
  s0 = sin (Joint[0]);
  c0 = cos (Joint[0]);

  p2 = TCP[P] - (DHP_A[3] * TCP[A]);      // Distance Joint 1 to Joint 3
  d = p2.Len ();

  // ================ Joint 3 ===================
  c2 = (SQR(DHP_A[1]) + SQR(DHP_A[2]) - SQR(d)) / (2.0 * DHP_A[1] * DHP_A[2]);
  if (c2 < -1.0 || c2 > 1.0)
  {
    PRO0 ("Out Of Workspace: Joint 3")
    throw KINEMATIC::OutOfWorkspace (1<<3);
  }

  s2 = sqrt (1.0 - SQR(c2));
  Joint[2] = atan2 (s2, c2) - PI;

  // ================ Joint 2 ==================
  s11 = p2[Z] / d;
  c11 = sqrt (SQR(p2[X]) + SQR(p2[Y])) / d;
  t11 = atan2 (s11, c11);
  s12 = DHP_A[2] * s2 / d;
  t12 = asin (s12);

  Joint[1] = t11 + t12;
  s1 = sin (Joint[1]);

  // Sum of joint 1 -3                                          /* Home[3] */
  Joint123 = atan2 (TCP[A][Z], c0 * TCP[A][X] * s0 * TCP[A][Y]) + 0.5 * PI;
  s123 = sin (Joint123);
  c123 = cos (Joint123);

  // ================ Joint 4 ==================
  Joint[3] = Joint123 - (Joint[1] + Joint[2]);

  // ================ Joint 5 ==================
  s4 = s0 * TCP[O][X] - c0 * TCP[O][Y];
  c4 = c123 * (c0 * TCP[O][X] + s0 * TCP[O][Y]) + s123 * TCP[O][Z];
  Joint[4] = atan2 (s4, c4) - 0.5 * PI;

  Joint.Deg ();

  PRO5 ("Joint: %.2f %.2f %.2f %.2f %.2f\n", 
	Joint[0], Joint[1], Joint[2], Joint[3], Joint[4])

  PRO_END
  return Joint;
}


JOINTS MITSUBISHI::InvKin (const FRAME  &Frame,
			         int     Configuration,
			   const JOINTS &LastJoint,
			         bool    CheckOOW /* = TRUE */,
			         bool    ValidLastJoint /* = TRUE */)
{
  return InvKin (Frame);
}

JOINTS MITSUBISHI::InvKin (const FRAME &Frame,
			         int    Configuration,  
			         bool   CheckOOW /* = TRUE */)
{
  return InvKin (Frame);
}


int MITSUBISHI::InvKin (const FRAME  &Frame,
			      JOINTS &Joint,
			const JOINTS &LastJoint,
			      bool    CheckOOW /* = TRUE */)
{
  Joint = InvKin (Frame);

  return 0;
}


JOINTS MITSUBISHI::InvKin (const FRAME  &Frame,
			   const JOINTS &LastJoint,
			         bool    CheckOOW)
{
  return InvKin (Frame);
}


int MITSUBISHI::InvKin (const FRAME  &Frame, 
                              JOINTS &Joint)
{
  Joint = InvKin (Frame);

  return 0;
}


FRAME MITSUBISHI::FwdKin (const JOINTS &Joint, bool Check /* = TRUE */)
{
  PRO_START (" ")
#if 0
  double  Euler[6];
  char   *Buffer;
  FRAME   TCP;

  if (SocketID < 0 ||
      (Buffer = NETSend(SocketID, "RPY")) == NULL ||
      sscanf(Buffer, "%lf %lf %lf %lf %lf %lf",
	             &Euler[0], &Euler[1], &Euler[2], 
	             &Euler[3], &Euler[4], &Euler[6]) < 6)
    PRO0 ("Error computing Forward kinematics")

  TCP.LoadRotRPY (Euler[5], Euler[4], Euler[3]);
  TCP[P][X] = Euler[0];
  TCP[P][Y] = Euler[1];
  TCP[P][Z] = Euler[2];
#else
  int     count;
  FRAME   TCP;

  for (count = 0; count < 5; count++)
  {
    TCP *= FRAME().SetDHP (Joint[count], DHP_D[count], DHP_A[count], Alpha[count]);
    TCP.Write (NULL, "TCP...");
  }
#endif

  TCP = Base * TCP;

  PRO_END
  return TCP;
}


int MITSUBISHI::CalcConfiguration (const JOINTS &Joint, 
	  	            	         bool    Check /* = TRUE */)
{
  // BitNo = JointNo: 
  // 0: smaller angle,  1: larger angle
  return 0; 
}

int MITSUBISHI::CalcConfiguration (const JOINTS &Joint, 
				   const FRAME  &Frame, 
				         bool    Check /* = TRUE */)
{
  return 0;
}


JOINTS MITSUBISHI::CheckFrame (const FRAME &Frame)
{
  JOINTS Joint;

  return Joint;
}


void MITSUBISHI::CheckJoint (const JOINTS &Joint)
{
  int error (0);                       // binary pattern for bad joints

  for (int i = 0; i < 5; ++i)
    if (Joint[i] < JointMin[i] || Joint[i] > JointMax[i])
    {
      error += (1 << i);
      PRO4 ("OutOfWorkspace Joint %d: %f Range (%f - %f)\n",
	    i+1, Joint[i], JointMin[i], JointMax[i])
    }

  if (error > 0) 
  {
    PRO1 ("OutOfWorkspace: %d\n", error)
    throw KINEMATIC::OutOfWorkspace (error);
  }
}


void MITSUBISHI::CheckOutOfLimit(const JOINTS &Joint, const JOINTS &LastJoint)
{
  int error (0);                      // binary pattern for bad joints

  for (int count = 0; count <= 5; count++)
    if (fabs(LastJoint[count]-Joint[count]) > EpsilonJoint[count])
    {
      error += (1 << count);
      PRO4 ("OutOfLimit Joint #%d: %.5f -> %.5f epsilon:%.3f\n",
	    count+1, LastJoint[count], Joint[count], EpsilonJoint[count])
    }

  if (error > 0) 
  {
    PRO1 ("OutOfLimit: %d\n", error)
    throw KINEMATIC::OutOfLimit (error); // Auerhalb des Gltigkeitsbereiches!
  }
}


#ifdef TODO
bool MITSUBISHI::SetSpeed (double Speed)  // m/sec */
{
  if (Speed <= 0.0)
    return FALSE;

  this->Speed = Speed;
  SetInterpPerMeter ((1000 / GetClock()) / Speed);

  return TRUE;
}
#endif

