#include <libsim3d.h>


MOTION::MOTION (const FRAME   &Start,   const FRAME  &End,
		/* const PROFILE &Profile,  */     double  Duration) :
  Number   (2),
  Duration (Duration)
{
  FRAME Delta ((! Start) * End);

  Frames = new FRAME[Number];

  Frames[0] = Start;
  Frames[1] = End;

  // TODO: Pos = new HVECTOR (Number-1);
  // TODO: Rot = new HVECTOR (Number-1);
  // TOOD: Angle = new double (Number-1);

  Pos = Frames[1][P] - Frames[0][P];
  Delta.CalcRotVector (Rot, Angle);
}


MOTION::MOTION (const MOTION &Motion) :
  Number   (Motion.Number),
  Duration (Motion.Duration),
  Pos      (Motion.Pos),
  Rot      (Motion.Rot),
  Angle    (Motion.Angle)
{
  int count;

  Frames = new FRAME[Number];
  for (count = 0; count < Number; ++count)
    Frames[count] = Motion.Frames[count];
}


MOTION::~MOTION ()
{
  if (Frames != NULL)
    delete Frames;
}


FRAME MOTION::GetStep (double Time) const
{
  double Factor = Time / Duration;
  FRAME  GoalRot, Result;

  GoalRot    = Frames[0];
  GoalRot[P] = HVECTOR ();

  Result.SetRotVector (Rot, Angle * Factor);
  Result = GoalRot * Result;
  Result[P] = Frames[0][P] + Factor * Pos;

  return Result; 
}


double MOTION::GetDuration () const
{
  return Duration;
}
