#include <assert.h>

#include "libsim3d.h"


RANGE::RANGE (double     Val       /* =  0.0      */, 
	      double     MinVal    /* =  1.0      */,
	      double     MaxVal    /* = -1.0      */,
	      RANGE_TYPE JointType /* = ROTATORIC */,
	      double     Precision /* = 0.001     */) :
  Type     (JointType),
  Value    (Val),
  Min      (MinVal),
  Max      (MaxVal),
  Accuracy (Precision)
{
}


double &RANGE::operator [] (int Index)
{
  assert (Index >= MINIMUM && Index <= MAXIMUM);

  switch (Index)
  {
    case MINIMUM : return Min;
    case MAXIMUM : return Max;
    case VALUE   :
    default      : return Value;
  }
}


double RANGE::operator [] (int Index) const
{
  assert (Index >= MINIMUM && Index <= MAXIMUM);

  switch (Index)
  {
    case MINIMUM : return Min;
    case MAXIMUM : return Max;
    case VALUE   :
    default      : return Value;
  }
}  


RANGE &RANGE::operator = (const RANGE &Range)
{
  if (this != &Range)                          // check for self-assignment
  {
    Type     = Range.Type;
    Value    = Range.Value;
    Min      = Range.Min;
    Max      = Range.Max;
    Accuracy = Range.Accuracy;
    Check ();
  }

  return *this;
} 


bool RANGE::Check () const
{
  if (Min <= Max && (Min > Value || Value > Max))
    throw RANGE_EXCEPTION (Min, Max, Value);

  return TRUE;
}


bool RANGE::Check (double Value) const
{
  return ((Min > Max) || (Min <= Value && Max >= Value));
}


RANGE::RANGE_TYPE RANGE::GetType () const
{ 
  return Type;     
}


double RANGE::GetValue () const
{
  return Value;
} 


double &RANGE::GetValue ()
{ 
  return Value;
}


double RANGE::GetMin () const
{ 
  return Min;
}


double RANGE::GetMax () const
{
  return Max;
}


double RANGE::GetAccuracy () const
{
  return Accuracy; 
}


void RANGE::SetType (RANGE_TYPE Type /* = ROTATORIC */) 
{ 
  this->Type = Type;
}


void RANGE::SetValue (double Value /* = 0.0 */)
{ 
  this->Value = Value; 
  //	   Check (); 
}


void RANGE::SetRange (double Min /* = 1.0 */, double Max /* = -1.0 */) 
{
  this->Min = Min; 
  this->Max = Max;
}


void RANGE::SetAccuracy (double Accuracy /* = RANGE_ACCURACY */)
{
  this->Accuracy = Accuracy;
}
