#include <string.h>

#include <libnet.h>

#include "libsim3d.h"


SOCKET::SOCKET (const char *HostName /* = NULL */, int PortNo /* = 2030 */)
{
  PRO2 ("Connecting '%s' at Port %d...", HostName, PortNo)

  if (PortNo > 0)
    SocketID = NETInitClient (HostName, PortNo);

  PRO0 ("\\done\n")

  // Socket.Send ("PROTOCOL ON");       // print debug-Messages on terminal
  // Socket.Send ("DEBUG ON");          // confirm each command with MCP
}


SOCKET::~SOCKET ()
{
  if (SocketID >= 0)
  {
    NETSend (SocketID, "EXIT");
    NETCloseClient (SocketID);
  }
}


char *SOCKET::SendCmd (const char *Cmd, ...)
{
  char    *Buffer = NULL;
  char     Message[255];
  va_list  Arguments;

  va_start (Arguments, Cmd);
  vsprintf (Message, Cmd, Arguments);
  va_end   (Arguments);

  if (SocketID >= 0)
    Buffer = NETSend (SocketID, Message);

  return Buffer;
}


bool SOCKET::SetSpeed (double Speed) const // m/sec */
{
  char *Buffer = NULL;

  if (SocketID < 0 || (Buffer = NETSend(SocketID, 
					"SPEED %.2f", Speed)) == NULL)
  {
    PRO0 ("Error setting speed\n")
    return FALSE;
  }

  return TRUE;
}


double SOCKET::GetSpeed () const
{
  char *Buffer = NULL;

  if (SocketID < 0 || (Buffer = NETSend(SocketID, "GET SPEED(1)")) == NULL)
  {
    PRO0 ("Error getting speed\n")
    return -1.0;
  }

  return atof (Buffer);
}


bool SOCKET::OpenHand () const
{
  if (SocketID >= 0)
    NETSend (SocketID, "OPEN");

  return TRUE;
}


bool SOCKET::CloseHand () const
{
  if (SocketID >= 0)
    NETSend (SocketID, "CLOSE"); 

  return TRUE;
}


int SOCKET::GetHand () const
{ 
  char *Buffer = NULL;

  if (SocketID < 0 || (Buffer = NETSend(SocketID, "GET HAND")) == NULL)
  {
    PRO0 ("Error getting hand-status\n")
    return -1;
  }

  return (Buffer[0] == '1');
}


bool SOCKET::Place (const JOINTS &Joint) const
{
  int   count;
  char *Buffer = NULL;
  char  Value[16];
  char  Command[255];

  sprintf (Command, "JOINT %d ", Joint.GetNumber());
  for (count = 0; count < Joint.GetNumber(); count++)
  {
    sprintf (Value, " %.3f", Joint[count]);
    strcat (Command, Value);
  }

  if (SocketID < 0 || (Buffer = NETSend(SocketID, Command)) == NULL)
  {
    PRO0 ("Error moving robot\n")
    return FALSE;
  }

  return (Buffer[0] == 'O');
}


bool SOCKET::Break () const
{
  if (SocketID < 0)
  {
    PRO0 ("No connection\n")
    return FALSE;
  }

  return (NETSend(SocketID, "WAIT") != FALSE);
}


JOINTS SOCKET::GetRobotPos () const
{
  char   *Buffer = NULL;
  JOINTS  Joint (0.0, 0.0, 0.0, 0.0, 0.0, 0.0);

  PRO_METHOD ("SOCKET", "");

  if (SocketID < 0 || 
      (Buffer = NETSend(SocketID, "HERE")) == NULL ||
      sscanf(Buffer, "%lf %lf %lf %lf %lf %lf",
	             &Joint[0], &Joint[1], &Joint[2], 
	             &Joint[3], &Joint[4], &Joint[5]) < 6)
    PRO1 ("Error reading Robot-position: '%s'\n", Buffer)

  PRO6 ("Joints: %.1f %.1f %.1f %.1f %.1f %.1f...",
	Joint[0], Joint[1], Joint[2], Joint[3], Joint[4], Joint[5])

  PRO_END
  return Joint;
}


FRAME SOCKET::GetPos () const
{
  JOINTS Joint;

  Joint = GetRobotPos ();

#ifdef TODO
  // Does not really compute the fwd-kinematic
  return FwdKin (Joint);
#endif
  return FRAME ();
}


FRAME SOCKET::Teach (const char *Msg /* = NULL */, ...)
{
  char     Message[255];
  char    *Buffer;
  JOINTS   Joint;
  FRAME    Result;
  va_list  Arguments;

  va_start (Arguments, Msg);
  vsprintf (Message, Msg, Arguments);
  va_end   (Arguments);

  if ((Buffer = NETSend(SocketID, "TEACH %s", Message)) == NULL ||
      sscanf(Buffer, "%lf %lf %lf %lf %lf %lf",
	             &Joint[0], &Joint[1], &Joint[2], 
                     &Joint[3], &Joint[4], &Joint[5]) < 6)
    PRO0 ("Error reading Teaching-position\n")

#ifdef TODO
  Result = FwdKin (Joint, FALSE);
#endif
  return Result;
}


