#ifndef _LIBTOOLS_H
#define _LIBTOOLS_H

#ifdef __cplusplus
extern "C" {
#endif

#include <X11/cursorfont.h>

#include <Xm/Xm.h>
#include <X11/xpm.h>

#include <debug.h>

/* Constants: */
#define CFG_DELTA_RADIUS     22         /* accuracy of contact-radius     */

#define MAX_COLOR 256
#define NORMAL_FONT XmSTRING_DEFAULT_CHARSET 

#define ERROR_RETURN0(Msg,Result) \
        { \
	  TOOLSetErrorMsg (Msg); \
	  PRO_RETURN (Msg, Result) \
	}
#define ERROR_RETURN1(Msg,Val,Result) \
        { \
	  TOOLSetErrorMsg (Msg, Val); \
	  PRO_RETURN1 (Msg, Val, Result) \
	}
#define ERROR_RETURN2(Msg,Val1,Val2,Result) \
        { \
	  TOOLSetErrorMsg (Msg, Val1, Val2); \
	  PRO_RETURN2 (Msg, Val1, Val2, Result) \
	}

/* Macros: */
#define NOT(x) ((int)(++x %2))

#define MILLI_SECOND(uSec) ((uSec)*1000)
#define SECOND(uSec) ((uSec)*1000000)

enum CURSOR_SHAPES
{
  ToolMAGNIFIER = -1,
  ToolQUESTION  = XC_question_arrow,
  ToolCIRCLE    = XC_circle,
  ToolLEFT      = XC_left_ptr,
  ToolPLUS      = XC_plus,
  ToolDOT       = XC_dot,
  ToolMOUSE     = XC_mouse,
  ToolCROSS     = XC_crosshair,
  ToolHAND      = XC_hand1,
  ToolPENCIL    = XC_pencil,
  ToolTEXT      = XC_xterm,
  ToolWATCH     = XC_watch,
  ToolNORMAL    = XC_top_left_arrow,
  ToolRESTORE   = XC_num_glyphs+2
};

enum PICTURE_TYPES
{
  UNKNOWN_FORMAT = -1,
  XBM_FORMAT,
  XPM_FORMAT,
#if (LIB_PNG==YES)
  PNG_FORMAT,
#endif
  JPG_FORMAT
};


typedef int (*TOOLPRINTFKT)(const char *format, ...);

typedef struct TOOL_LINE
{
         int        LineNo;
         char       Type;
         char      *Cmd;
         char      *Val;
         dual       optional;
  struct TOOL_LINE *Next;
} TOOL_LINE;


#define CFG_KEYWORD_LEN 40
#define CFG_STRING_LEN 1024

enum CONFIG_TYPES
{
  CFG_BOOL = 0,
  CFG_INT,
  CFG_DOUBLE,
  CFG_STRING
};

typedef struct
{
  int    CfgID;
  char   Keyword[CFG_KEYWORD_LEN+1];
  int    Type;
  double Value;                         /* bool/int/double                */
  char   String[CFG_STRING_LEN+1];      /* string                         */
} CONFIG_ENTRY;

#define CFG_ENTRIES(CfgEntries) (int)(sizeof(CfgEntries)/sizeof(CONFIG_ENTRY))


/* important environment for the libraries: */
extern XtAppContext  AppCon;
extern Widget        AppShell;
extern Display      *XDisplay;
extern int           XScreen;

extern GC            defGC, drawGC, xorGC, copyGC;
extern Colormap      CMap;


/* Usefull functions which can be called WITHOUT initializing this lib */
void       TOOLstrtoupper        (char *Result, char *Str);
char      *TOOLParseFileName     (const char *FileName, 
				  char *Result, int Length);
char      *TOOLGetFileDir        (const char *FileName, 
				  char *Result, int Length);
char      *TOOLGetFileName       (const char *FileName, 
				  char *Result, int Length);
char      *TOOLGetFileExt        (const char *FileName, 
				  char *Result, int Length);
dual       TOOLTestPattern       (const char *Pattern, const char *FileName);

/* ***** Message-Functions: ***** */
void       TOOLSetErrorMsg       (char *Msg, ...);
void       TOOLSetErrorStatus    (int status);
char      *TOOLGetErrorMsg       (void);
int        TOOLGetErrorStatus    (void);
void       TOOLSetPrintFunction  (TOOLPRINTFKT Function);
int        TOOLPrintf            (char *Msg, ...);

/* *** Config-Files: *** */
dual       TOOLInsertEntries     (CONFIG_ENTRY *Entry, int Amount);
dual       TOOLFreeEntries       (void);

dual       TOOLGetConfigBool     (int CfgID);
int        TOOLGetConfigInt      (int CfgID);
double     TOOLGetConfigDouble   (int CfgID);
char      *TOOLGetConfigString   (int CfgID);

dual       TOOLSetConfigBool     (int CfgID, dual    Value);
dual       TOOLSetConfigInt      (int CfgID, int     Value);
dual       TOOLSetConfigDouble   (int CfgID, double  Value);
dual       TOOLSetConfigString   (int CfgID, char   *String);

dual       TOOLLoadConfig        (char *FileName, char *IDName);
dual       TOOLSaveConfig        (char *FileName, char *IDName);


/* Script functions (no initialization needed */
int        TOOLOpenScript        (char *FileName);
void       TOOLCloseScript       (void);
dual       TOOLScriptRunning     (void);
dual       TOOLScriptOK          (void);
TOOL_LINE *TOOLGetScriptCmd      (void);
dual       TOOLNextScriptCmd     (void);
#if 0
dual       TOOLStopScript        (void *Function, void *Data);
dual       TOOLContScript        (void);
#endif

int        TOOLParse             (const char *Data, const char *Mask, ...);


/* X functions */
int        TOOLGetResInt         (Widget  Parent, char  *Class, char *Spec, 
				  char   *Name,   int    Default);
double     TOOLGetResDouble      (Widget  Parent, char  *Class, char *Spec, 
				  char   *Name,   double Default);
char       TOOLGetResChar        (Widget  Parent, char  *Class, char *Spec, 
				  char   *Name,   char   Default);
char      *TOOLGetResString      (Widget  Parent, char  *Class, char *Spec, 
				  char   *Name,   char  *Default);

void       TOOLInitGCs           (void);

int        TOOLGetDepth          (void);
int        TOOLGetBytesPerPixel  (void);
void       TOOLInitLib           (char *Title, int *argc, char **argv,
				  char **FallbackResources);
void       TOOLSetLib            (Display *XDispl,    XtAppContext AppContext, 
				  Widget   ApplShell, int          Screen);
void       TOOLAddExitCB         (XtCallbackProc Function, void *Data);
void       TOOLDelExitCB         (XtCallbackProc Function, void *Data);
void       TOOLExitLib           (void);
dual       TOOLIsInitialized     (void);
dual       TOOLIsRunning         (void);

Pixel      TOOLGetColor          (char *ColorStr);
Window     TOOLGetMousePos       (int *MouseX, int *MouseY);
void       TOOLSetIcon           (Widget Shell, char *IconName, 
			                        char *FileName, void *Data);
void       TOOLMainLoop          (XtAppContext AppCon, Widget Shell);
void       TOOLWait              (long uSec);
void       TOOLWaitForMapping    (Display *Disp, Window Win);
void       TOOLGetScrValue       (Display *Displ, int Screen,
				  double x, double y, int *ScrX, int *ScrY);

int        TOOLGetImageType      (const char *FileName);
Pixmap     TOOLImage2Pixmap      (XImage *Image);
XImage    *TOOLCreateImage       (int SizeX, int SizeY, 
				  double r, double g, double b);
XImage    *TOOLLoadImage         (const char *FileName);
XImage    *TOOLScaleImage        (XImage *Image, int SizeX, int SizeY);
XImage    *TOOLCopyImage         (XImage *Image);
void       TOOLDestroyImage      (XImage *Image);
void       TOOLHSV2RGB           (double  h, double  s, double  v,
				  double *r, double *g, double *b);

/* Images (XBM) */
unsigned char *TOOLLoadXBMData   (char *FileName, int *sizeX, int *SizeY,
			                          int *HotX, int *HotY);
Pixmap     TOOLLoadXBM           (char *FileName, int *width, int *height,
				                  int *HotX, int *HotY);
Pixmap     TOOLMakeXBM           (Display *XDisplay, char *Data, 
				  int SizeX, int SizeY, Pixel fg, Pixel bg);
byte      *TOOLMakeEmptyXBM      (int SizeX, int SizeY);
byte      *TOOLLightenXBM        (byte *PicData, int width, int height);
byte      *TOOLCopyXBM           (byte *Data, int SizeX, int SizeY);

/* Images (XPM) */
char     **TOOLCopyXPM           (char **Data);
Pixmap     TOOLLoadXPM           (char *FileName, int *SizeX, int *SizeY,
				                  Pixel Color);
byte      *TOOLLoadXPM_RGBA      (char *FileName, double Alpha, 
				  int *SizeX, int *SizeY);
Pixmap     TOOLMakeXPM           (char **Data,    int *SizeX, int *SizeY,
				                  Pixel Color);

XImage    *TOOLLoadXPM2          (const char *FileName, int *SizeX, int *SizeY,
				                        Pixel Color);
XImage    *TOOLMakeXPM2          (char **Data,    int *SizeX, int *SizeY,
				                  Pixel Color);

Pixmap     TOOLxpm2xbm           (char *Data, int SizeX, int SizeY, 
				              Drawable d);

void       TOOLSetMouseCursor    (Widget Shell, int CursorID, 
				  char *fgColor, char *bgColor);

/* Images (PNG) */
XImage    *TOOLLoadPNG           (const char *FileName, 
				  int *SizeX, int *SizeY, dual check);

XImage    *TOOLLoadJPG           (const char *FileName, 
				  int *SizeX, int *SizeY);

byte      *TOOLMakeRGBA          (XImage *Image, double Alpha, 
     				                 int *SizeX, int *SizeY);

#ifdef __cplusplus
} /* for C++ V2.0 */
#endif

#endif
