#include <string.h>

#include "libtools.h"


char  *ColorName[] = {"", "", 
		      "RED", "ORANGE", "YELLOW", "GREEN", 
		      "BLUE", "MAGENTA", "GREY", "BROWN"};
static CONFIG_ENTRY *Config = NULL;
static int           ConfigNo = 0;
static char          DebMsg[512] = " ";


dual TOOLInsertEntries (CONFIG_ENTRY *Entry, int Amount)
{
  Config   = Entry;
  ConfigNo = Amount;

  return TRUE;
}

dual TOOLFreeEntries (void)
{
  return TRUE;
}


/* ********************************************************************** */
/* ** Locate a configuration-entry using it's id-number and type       ** */
/* ********************************************************************** */
static dual TOOL_GetConfigEntry (int CfgID, int Type, CONFIG_ENTRY **Entry)
{
  int count = 0;

  while (count < ConfigNo && 
	 (Config[count].CfgID != CfgID || Config[count].Type != Type))
    count++;

  if (count == ConfigNo)
  {
    PRO2 ("Could not find Config-Entry %d with Type %d\n", CfgID, Type)
    return FALSE;
  }

  *Entry = &Config[count];
  return TRUE;
}


/* ********************************************************************** */
/* ** locate a Config-entry using it's keyword                         ** */
/* ********************************************************************** */
static dual TOOL_SearchConfigEntry (char *Keyword, CONFIG_ENTRY **Entry)
{
  int count = 0;

  while (count < ConfigNo && strncmp(Config[count].Keyword, Keyword, 
		                     strlen(Config[count].Keyword)) != 0)
    count++;

  if (count == ConfigNo)
  {
    PRO1 ("Could not find Config-Entry '%s'\n", Keyword)
    return FALSE;
  }

  *Entry = &Config[count];
  return TRUE;
}


/* ********************************************************************** */
/* ** Delievers a boolean config-entry with given identifier           ** */
/* ** INPUT:  CfgID   Configuration-number                             ** */
/* ** OUTPUT: (bool)  Actual value of entry, FALSE if not found        ** */
/* ********************************************************************** */
dual TOOLGetConfigBool (int CfgID)
{
  CONFIG_ENTRY *Entry = NULL;

  if (! TOOL_GetConfigEntry(CfgID, CFG_BOOL, &Entry))
    return FALSE;

  return (dual)Entry->Value;
}

/* ********************************************************************** */
/* ** Returns the config-value for the specified identifier            ** */
/* ** INPUT:  CfgID  Number of entry                                   ** */
/* ** OUPUT:  (int)  actual int-value or 0 if not found                ** */
/* ********************************************************************** */
int TOOLGetConfigInt (int CfgID)
{
  CONFIG_ENTRY *Entry = NULL;

  if (! TOOL_GetConfigEntry(CfgID, CFG_INT, &Entry))
    return 0;

  return (int)Entry->Value;
}

/* ********************************************************************** */
/* ** Gives back the actual double-config-value with ID 'CfgID'        ** */
/* ** INPUT:  CfgID     Config-number of double-entry                  ** */
/* ** OUTPUT: (double)  Actual value of entry or 0.0 if error          ** */
/* ********************************************************************** */   
double TOOLGetConfigDouble (int CfgID)
{
  CONFIG_ENTRY *Entry = NULL;

  if (! TOOL_GetConfigEntry(CfgID, CFG_DOUBLE, &Entry))
    return 0.0;

  return (double)Entry->Value;
}

/* ********************************************************************* */
/* ** Get the actual config-value of a string-entry                   ** */
/* ** INPUT:  CfgID    Identifier of entry                            ** */
/* ** OUTPUT: (char*)  char-pointer to actual value                   ** */
/* ********************************************************************* */  
char *TOOLGetConfigString (int CfgID)
{
  CONFIG_ENTRY *Entry = NULL;

  if (! TOOL_GetConfigEntry(CfgID, CFG_STRING, &Entry))
    return NULL;

  return (char *)Entry->String;
}


/* ********************************************************************** */
/* ** Set the Config-value for a boolean-entry                         ** */
/* ** INPUT:  CfgID   Configuration-id                                 ** */
/* **         Value   New value of entry                               ** */
/* ** OUTPUT: (bool)  TRUE, if updated successfully                    ** */
/* ********************************************************************** */
dual TOOLSetConfigBool (int CfgID, dual Value)
{
  CONFIG_ENTRY *Entry = NULL;

  if (! TOOL_GetConfigEntry(CfgID, CFG_BOOL, &Entry))
    return FALSE;

  Entry->Value = (double)Value;
  return TRUE;
}

/* ********************************************************************** */
/* ** Set value of an config-int-entry                                 ** */
/* ** INPUT:  CfgID   Config-ID of entry                               ** */
/* **         Value   New int-value                                    ** */
/* ** OUTPUT: (dual)  TRUE on success, FALSE otherwise                 ** */
/* ********************************************************************** */
dual TOOLSetConfigInt (int CfgID, int Value)
{
  CONFIG_ENTRY *Entry = NULL;

  if (! TOOL_GetConfigEntry(CfgID, CFG_INT, &Entry))
    return FALSE;

  Entry->Value = (double)Value;
  return TRUE;
}

/* *********************************************************************** */
/* ** Set a double-entry to a new value                                 ** */
/* ** INPUT:  CfgID   Identifier of the config-entry                    ** */
/* **         Value   New value                                         ** */
/* ** OUTPUT: (dual)  TRUE, if changing was successfully                ** */
/* **                 FALSE otherwise                                   ** */
/* *********************************************************************** */
dual TOOLSetConfigDouble (int CfgID, double Value)
{
  CONFIG_ENTRY *Entry = NULL;

  if (! TOOL_GetConfigEntry(CfgID, CFG_DOUBLE, &Entry))
    return FALSE;

  Entry->Value = (double)Value;
  return TRUE;
}

/* ********************************************************************** */
/* ** Set a config-entry to a new string-value                         ** */
/* ** INPUT:  CfgID   Corresponding config-identifier                  ** */
/* **         String  New String to store                              ** */
/* ** OUTPUT: (dual)  TRUE: string was copied, FALSE: error occured    ** */
/* ********************************************************************** */
dual TOOLSetConfigString (int CfgID, char *String)
{
  CONFIG_ENTRY *Entry = NULL;

  if (! TOOL_GetConfigEntry(CfgID, CFG_STRING, &Entry))
    return FALSE;

  strncpy (Entry->String, String, CFG_STRING_LEN);
  return TRUE;
}


/* ********************************************************************** */
/* ** Loads a config file and store the read values in the config-data ** */
/* ** INPUT:  FileName  Config-file-name                               ** */
/* **         IDName    Identifier-string to check for appropriate     ** */
/* **                   config-file                                    ** */
/* ** OUTPUT: (dual)    TRUE: file was read successfully               ** */
/* **                   FALSE: an error occured while reading          ** */
/* ********************************************************************** */
dual TOOLLoadConfig (char *FileName, char *IDName)
{
  dual          Error = FALSE;
  int           count;
  char          cVal, letter;
  char          String[80];
  int           iVal;
  double        dVal;
  CONFIG_ENTRY *Entry = NULL;
  FILE         *fp = NULL;

  PRO_START (" ")

  if ((fp = fopen(FileName, "r+")) == NULL)
  {
    TOOLSetErrorMsg ("Error reading config-file");
    PRO_RETURN ("Error reading config-file", FALSE)
  }

  if (fscanf(fp, "%s", String) < 1 ||
      strncmp(String, IDName, strlen(IDName)) != 0)
  {
    fclose (fp);
    TOOLSetErrorMsg ("Invalid config-file");
    PRO_RETURN ("Invalid config-file", FALSE)
  }

  DebMsg[0] = '\0';

  while (! feof(fp) && fscanf(fp, "%s", String) == 1)
  {
    do 
      letter = fgetc(fp);
    while (! feof(fp) && letter != '\n');

    if (String[0] != '#')
    {
      if ((Error = (! TOOL_SearchConfigEntry(String, &Entry))))
	TOOLSetErrorMsg ("Unkonwn keyword: %s", String);
      else
      {
	cVal = ' ';
	switch (Entry->Type)
	{
	  case CFG_BOOL :
	    if ((Error = (fscanf(fp, "%s", String) < 1)))
	      TOOLSetErrorMsg ("Invalid BOOL-value for Keyword '%s'",
			       Entry->Keyword);
	    else
	      Entry->Value = (double)(String[0] == 'T' || String[0] == 'Y');
	    break;

          case CFG_INT :
	    if ((Error = (fscanf(fp, "%d", &iVal) < 1)))
	      TOOLSetErrorMsg ("Invalid INT-value for Keyword '%s'",
			       Entry->Keyword);
	    else
	      Entry->Value = (double)(iVal);
	    break;

          case CFG_DOUBLE :
	    if ((Error = (fscanf(fp, "%lf", &dVal) < 1)))
	      TOOLSetErrorMsg ("Invalid DOUBLE-value for Keyword '%s'",
			       Entry->Keyword);
	    else
	      Entry->Value = dVal;
	    break;

          case CFG_STRING :
	    count = 0;
	    cVal = fgetc (fp);
	    while (! feof(fp) && (cVal == ' ' || cVal == '\t'))
	      cVal = fgetc (fp);
	    while (! feof(fp) && cVal != '\n' && count < CFG_STRING_LEN)
	    {
	      Entry->String[count++] = cVal;
	      cVal = fgetc (fp);
	    }
	    Entry->String[count] = '\0';
	    break;

         default :
	   Error = TRUE;
	   TOOLSetErrorMsg ("Invalid data-type for keyword '%s'", 
			    Entry->Keyword);
	}

	while (! feof(fp) && cVal != '\n')     /* Read until end-of-line    */
	  cVal = fgetc(fp);
      }

      if (Error)
      {
	fclose (fp);
	PRO_RETURN (TOOLGetErrorMsg(), FALSE)
      }
    }
  }
  fclose (fp);

  PRO_END
  return TRUE;
}


/* ************************************************************************ */
/* ** Save all config-values in a config-file with specified name        ** */
/* ** INPUT:  FileName  Name of target-config-file                       ** */
/* **         IDName    Identifier-name - used to check if a config-file ** */
/* **                   is accordingly an special application-program.   ** */
/* ** OUTPUT: (dual)    TRUE, when file was saved successfully           ** */
/* **                   FALSE, when writing failed.                      ** */
/* ************************************************************************ */
dual TOOLSaveConfig (char *FileName, char *IDName)
{
  int           count;
  CONFIG_ENTRY  Entry;
  FILE         *fp = NULL;

  PRO_START (" ")

  if ((fp = fopen(FileName, "w+")) == NULL)
    PRO_RETURN ("Error creating config-file", FALSE)

  fprintf (fp, "%s\n", IDName);
  fprintf (fp, "\n");
  fprintf (fp, "\n");

  for (count = 0; count < ConfigNo; ++count)
  {
    Entry = Config[count];
    fprintf (fp, "%s:\n", Entry.Keyword);
    switch (Config[count].Type)
    {
      case CFG_BOOL :
	fprintf (fp, "%s\n", ((dual)Entry.Value==TRUE)? "YES":"NO");
	break;
      case CFG_INT :
	fprintf (fp, "%d\n", (int)Entry.Value);
	break;
      case CFG_DOUBLE :
       	fprintf (fp, "%f\n", (double)Entry.Value);
        break;
      case CFG_STRING :
       	fprintf (fp, "%s\n", (char *)Entry.String);
        break;
      default :
	PRO1 ("Invalid Config-Type: %d\n", Entry.Type)
    }
    fprintf (fp, "\n");
  }
  fclose (fp);

  PRO_END
  return TRUE;
}
