
#include "libtools.h"

#define HAVE_BOOLEAN

#include <jpeglib.h>


XImage *TOOLLoadJPG (const char *FileName, int *SizeX, int *SizeY)
{
         int                     rs, count;
	 int                     w, h, Index;
	 int                     BytesPerPixel;
         byte                   *Data;
	 byte                   *ImageData, *ImgPtr;
         FILE                   *fp;
	 XColor                  Color;
	 Colormap                CMap;
	 XImage                 *Result;
         JSAMPROW               *RowPtr = NULL;
  struct jpeg_decompress_struct  cinfo;
  struct jpeg_error_mgr          jerr;
         Pixel                  *Buffer;

  PRO_START ("\n")

  /* Open jpeg file */
  if ((fp = fopen(FileName, "r")) == NULL)
    PRO_RETURN ("Open error", NULL)

  cinfo.err = jpeg_std_error (&jerr);
  jpeg_create_decompress (&cinfo);
  jpeg_stdio_src         (&cinfo, fp);

  /* Read jpeg header */
  if (jpeg_read_header(&cinfo, TRUE) == 0)
  {
    fclose (fp);
    PRO_RETURN ("Error reading header", NULL)
  }

  /* specify output values */
  cinfo.out_color_space = JCS_RGB;
  cinfo.colormap        = NULL;
  cinfo.quantize_colors = TRUE;
  jpeg_start_decompress (&cinfo);

  w = cinfo.output_width;
  h = cinfo.output_height;
  Data = (byte *)malloc (sizeof(byte) * w * h * 3);

  /* Read all image lines */
  RowPtr = (JSAMPROW *)malloc (sizeof(JSAMPROW) * h);
  rs = w * 3;
  for (count = 0; count < h; ++count)
  {
    RowPtr[count] = (byte *)&Data[count*rs];
    jpeg_read_scanlines (&cinfo, &RowPtr[count], 1);
  }

  jpeg_finish_decompress  (&cinfo);
  jpeg_destroy_decompress (&cinfo);

  fclose (fp);
  
  /* Create colormap */
  CMap = DefaultColormap (XDisplay, XScreen);

  Color.pad   = 0;
  Color.flags = DoRed | DoGreen | DoBlue;
  Buffer = (Pixel *)malloc (sizeof(Pixel) * cinfo.actual_number_of_colors);
  for (count = 0; count < cinfo.actual_number_of_colors; ++count)
  {
    Color.red   = cinfo.colormap[0][count] * 255;  /* Scale to 0..65535 */
    Color.green = cinfo.colormap[1][count] * 255;
    Color.blue  = cinfo.colormap[2][count] * 255;
    if (! XAllocColor(XDisplay, CMap, &Color))
    {
      printf ("Error allocating more than %d colors\n", count);
      break;
    }
    Buffer[count] = Color.pixel;
  }   

  BytesPerPixel = TOOLGetBytesPerPixel ();

  PRO1 ("Bytes/Pixel: %d\n", BytesPerPixel)

  ImageData = (byte *)malloc (sizeof(byte) * w * h * BytesPerPixel);
  ImgPtr = ImageData;
  switch (BytesPerPixel)
  {
    case 1 :
      for (count = 0; count < h; ++count)
	for (Index = 0; Index < w; ++Index)
	{
	  /* TODO: Datentyp abhaengig von Tiefe */
	  *(byte *)ImgPtr = (byte)Buffer[RowPtr[count][Index]];
	  ImgPtr += BytesPerPixel;
	}
      break;

    case 2 :
      for (count = 0; count < h; ++count)
	for (Index = 0; Index < w; ++Index)
	{
	  /* TODO: Datentyp abhaengig von Tiefe */
	  *(short *)ImgPtr = (short)Buffer[RowPtr[count][Index]];
	  ImgPtr += BytesPerPixel;
	}
      break;

    case 4 :
      for (count = 0; count < h; ++count)
	for (Index = 0; Index < w; ++Index)
	{
	  /* TODO: Datentyp abhaengig von Tiefe */
	  *(int *)ImgPtr = (int)Buffer[RowPtr[count][Index]];
	  ImgPtr += BytesPerPixel;
	}
      break;

    default :
      printf ("TOOLLoadJPEG: Unsupported Bytes/Pixel: %d\n", BytesPerPixel);
  }

  free (Buffer);
  free (RowPtr);
  free (Data);

  /* Create the image */
  Result = XCreateImage (XDisplay, DefaultVisual(XDisplay, XScreen), 
			 TOOLGetDepth(), ZPixmap, 0,
			 (char *)ImageData, w, h, 8, 0);
  /* Do not free ImageData here. Will be used by XImage ! */

  if (SizeX != NULL)
    *SizeX = w;
  if (SizeY != NULL)
    *SizeY = h;

  PRO_END
  return Result;
}
