#include <stdarg.h>

#include "libtools.h"


/* ********************************************************************** */
/* ** Complex applications should be devided into libraries. To get a  ** */
/* ** better code-reuse and a better portability to other systems,     ** */
/* ** should better do no output. All the output should be done by the ** */
/* ** main application !                                               ** */
/* ** This means to reach all messages up to the highest level, what   ** */
/* ** might be very complicting e.g. if you want to document the       ** */
/* ** progress of a calculation.                                       ** */
/* ** So we use a small trick:                                         ** */
/* ** The application registers output-function in one of the          ** */
/* ** low-level-libraries (like the libtools). Also we export an own   ** */
/* ** output-function, which calls the registered function (if it was  ** */
/* ** defined) or use the standard-output (if no function was          ** */
/* ** registered.                                                      ** */
/* ********************************************************************** */

static char          ErrorMessage[2048] = "";
static int           ErrorStatus        = 0;
static TOOLPRINTFKT  ToolPrintFkt       = printf;


void TOOLSetErrorMsg (char *Msg, ...)
{
  va_list Arguments;

  if (Msg != NULL)
  {
    va_start (Arguments, Msg);
    vsprintf (ErrorMessage, Msg, Arguments);
    va_end   (Arguments);
  }
  else
    ErrorMessage[0] = '\0';
}


void TOOLSetErrorStatus (int status)
{
  ErrorStatus = status;
}


char *TOOLGetErrorMsg (void)
{
  return ErrorMessage;
}


int TOOLGetErrorStatus (void)
{
  return ErrorStatus;
}


/* ************************************************************************ */
/* ** Registers a output-function, which should be used to print text    ** */
/* ** from libraries.                                                    ** */
/* ** INPUT:  Function  User output-function (syntax like printf)        ** */
/* ** OUTPUT: none                                                       ** */
/* ************************************************************************ */
void TOOLSetPrintFunction (TOOLPRINTFKT Function)
{
  ToolPrintFkt = Function;
}

/* *********************************************************************** */
/* ** Print out a message using either the registered output-function   ** */
/* ** or a standard-printf.                                             ** */
/* ** INPUT:  Msg    output-string                                      ** */
/* **         ...    variable parameter-list                            ** */
/* ** OUTPUT: (int)  return-value of used output-function               ** */
/* *********************************************************************** */
int TOOLPrintf (char *Msg, ...)
{
  int     Result = 1;
  char    Message[2048];
  va_list Arguments;

  va_start (Arguments, Msg);
  vsprintf (Message, Msg, Arguments);
  va_end   (Arguments);

  if (ToolPrintFkt != NULL)
    Result = ToolPrintFkt (Message);
  else
  {
    Result = fprintf (stderr, Message);
    fflush (stderr);
  }

  return Result;
}


