#if (LIB_PNG==YES)

#include <png.h>
#endif

#include "libtools.h"

#define NO_PNG_CHECK


#if (LIB_PNG==YES)
XImage *TOOLLoadPNG (const char *FileName, int *SizeX, int *SizeY, dual check)
{
  unsigned int   count;
  unsigned int   countX, countY;
  unsigned int   ColorNo = 0;
  int            bit, BPP;
  byte          *Data = NULL, *Ptr;
  FILE          *fp = NULL;
  png_byte     **PNGData;
#if 1
  png_color_16   bg, *Imgbg;
#endif
  png_struct    *PNG;
  png_info      *PNGInfo, *PNGEnd;
  XColor        *Colors = NULL;
  Colormap       CMap;
  XImage        *Image = NULL;
  unsigned char  Header[16];

  PRO_START ("\n")

  if ((fp = fopen(FileName, "rb")) == NULL)
    PRO_RETURN ("Error reading PNG-file", Image)

  if (check)
  {
    fread (Header, 1, sizeof(Header), fp);
    if (! png_check_sig(Header, sizeof(Header)))
      PRO_RETURN ("Not a PNG-file", Image)
  }

  if ((PNG = png_create_read_struct(PNG_LIBPNG_VER_STRING, 
				    NULL, NULL, NULL)) == NULL)
  {
    fclose (fp);
    PRO_RETURN ("Error creating PNG-Struct", Image)
  }

  if ((PNGInfo = png_create_info_struct(PNG)) == NULL)
  {
    fclose (fp);
    png_destroy_read_struct (&PNG, (png_info**)NULL, (png_info**)NULL);
    PRO_RETURN ("Error creating PNG-Info", Image)
  }

  if ((PNGEnd = png_create_info_struct(PNG)) == NULL)
  {
    fclose (fp);
    png_destroy_read_struct (&PNG, (png_info**)NULL, (png_info**)NULL);
    PRO_RETURN ("Error creating PNG-End", Image)
  }

  png_init_io (PNG, fp);
  if (check)
    png_set_sig_bytes (PNG, sizeof(Header));

  png_read_info (PNG, PNGInfo);

  PRO2 ("PNG-Size: %d x %d\n", PNGInfo->width, PNGInfo->height)

  if (png_get_bKGD(PNG, PNGInfo, &Imgbg))
    png_set_background (PNG, Imgbg, PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
  else
    png_set_background (PNG, &bg, PNG_BACKGROUND_GAMMA_SCREEN, 0, 1.0);

  /* allocate memory and read image data */
  PNGData = (png_byte **)malloc (sizeof(png_byte *) * PNGInfo->height);
  for (count = 0; count < PNGInfo->height; ++count)
    PNGData[count] = (png_byte *)malloc (png_get_rowbytes(PNG, PNGInfo));

  png_read_image (PNG, PNGData);

  /* read comments, ... */
  png_read_end (PNG, PNGEnd);
  fclose (fp);

  PRO0 ("PNG-Info\n")
  PRO1 ("PNG-Depth:      %d\n", PNGInfo->bit_depth)
  PRO1 ("PNG-Color:      %d\n", PNGInfo->color_type)
  PRO1 ("PNG-Filter:     %d\n", PNGInfo->filter_type)
  PRO1 ("PNG-Interlace:  %d\n", PNGInfo->interlace_type)
  PRO1 ("PNG-Channels:   %d\n", PNGInfo->channels)
  PRO1 ("PNG-Bits/Pixel: %d\n", PNGInfo->pixel_depth)

  CMap = DefaultColormap (XDisplay, XScreen);
  
  switch (PNGInfo->color_type)
  {
    case PNG_COLOR_TYPE_GRAY :
      PRO0 ("Greyscale\n")
 
      if (PNGInfo->bit_depth < 8)
	png_set_expand (PNG);

      ColorNo = 1 << PNGInfo->bit_depth;
      Colors = (XColor *)malloc (sizeof(XColor) * ColorNo);
      for (count = 0; count < ColorNo; ++count)
      {
	Colors[count].pad   = 0;
	Colors[count].flags = DoRed | DoGreen | DoBlue;
	Colors[count].red = Colors[count].green = Colors[count].blue = 
	  (int)(255.0 / (double)(ColorNo-1) * count) << 8;
	if (! XAllocColor(XDisplay, CMap, &Colors[count]))
	{
	  PRO1 ("Error allocating more than %d colors\n", count)
	  break;
	}
      }
      break;

    case PNG_COLOR_TYPE_PALETTE :
      PRO0 ("Color palette\n")

      ColorNo = PNGInfo->num_palette;      
      PRO1 ("Palette-No: %d\n", ColorNo)

      if (PNGInfo->bit_depth < 8)
	png_set_expand (PNG);

      Colors = (XColor *)malloc (sizeof(XColor) * ColorNo);
      for (count = 0; count < ColorNo; ++count)
      {
	/* Scale to 0..65535 */
	Colors[count].pad   = 0;
	Colors[count].flags = DoRed | DoGreen | DoBlue;
	Colors[count].red   = PNGInfo->palette[count].red   << 8; 
	Colors[count].green = PNGInfo->palette[count].green << 8;
	Colors[count].blue  = PNGInfo->palette[count].blue  << 8;
	if (! XAllocColor(XDisplay, CMap, &Colors[count]))
	{
	  PRO1 ("Error allocating more than %d colors\n", count)
	  break;
	}
      }
      break;

    case 6 :
      PRO0 ("Unsupported color type 6\n")
      break;

    default :
      PRO1 ("COLOR-Type %d not Supported - yet\n", PNGInfo->color_type)
  }

  BPP = TOOLGetBytesPerPixel ();

  if (Colors != NULL)
  {
    Ptr = Data = (byte *)malloc (sizeof(byte) * 
				 PNGInfo->width * PNGInfo->height * BPP);
    count = 0;

    if (PNGInfo->pixel_depth == 1)                    /* binary image ? */
    {                            /* inner loop: -> (int)((width/8)+0.5) */
      switch (BPP)
      {
        case 1 :
	  for (countY = 0; countY < PNGInfo->height; ++countY)
	    for (countX = 0; countX < (PNGInfo->width+0.5)/8; ++countX)
	      for (bit = 7; bit >= 0; bit--)
		if (8*countX+bit < PNGInfo->width)
		{
		  *(byte *)Ptr = 
		    (byte)Colors[(PNGData[countY][countX] & 
				  (1<<bit)) != 0].pixel;
		  Ptr += BPP;
		}
	  break;

        case 2 :
	  for (countY = 0; countY < PNGInfo->height; ++countY)
	    for (countX = 0; countX < (PNGInfo->width+0.5)/8; ++countX)
	      for (bit = 7; bit >= 0; bit--)
		if (8*countX+bit < PNGInfo->width)
		{
		  *(short *)Ptr = 
		    (short)Colors[(PNGData[countY][countX] & 
				  (1<<bit)) != 0].pixel;
		  Ptr += BPP;
		}
	  break;

        case 4 :
	  for (countY = 0; countY < PNGInfo->height; ++countY)
	    for (countX = 0; countX < (PNGInfo->width+0.5)/8; ++countX)
	      for (bit = 7; bit >= 0; bit--)
		if (8*countX+bit < PNGInfo->width)
		{
		  *(int *)Ptr = 
		    (int)Colors[(PNGData[countY][countX] & 
				 (1<<bit)) != 0].pixel;
		  Ptr += BPP;
		}
	  break;

        default :
	  PRO1 ("Unsupported Pixel depth: %d\n", BPP)
      }
    }
    else 
      switch (BPP)
      {
        case 1 :
	  for (countY = 0; countY < PNGInfo->height; ++countY)
	    for (countX = 0; countX < PNGInfo->width; ++countX)
	    {
	      *(byte *)Ptr = (byte)Colors[PNGData[countY][countX]].pixel;
	      Ptr += BPP;
	    }
	  break;

        case 2 :
	  for (countY = 0; countY < PNGInfo->height; ++countY)
	    for (countX = 0; countX < PNGInfo->width; ++countX)
	    {
	      *(short *)Ptr = (short)Colors[PNGData[countY][countX]].pixel;
	      Ptr += BPP;
	    }
	  break;

        case 4 :
	  for (countY = 0; countY < PNGInfo->height; ++countY)
	    for (countX = 0; countX < PNGInfo->width; ++countX)
	    {
	      *(int *)Ptr = (int)Colors[PNGData[countY][countX]].pixel;
	      Ptr += BPP;
	    }
	  break;

        default :
	  PRO1 ("Unsupported Pixel depth: %d\n", BPP)
      }

    free (Colors);
  }

  for (count = 0; count < PNGInfo->height; ++count)
    free (PNGData[count]);
  free (PNGData);

  if (Data != NULL)
  {
    PRO0 ("Image...\n")
    Image = XCreateImage (XDisplay, DefaultVisual(XDisplay, XScreen),
			            TOOLGetDepth(), ZPixmap, 0, (char *)Data, 
			            PNGInfo->width, PNGInfo->height,
			            8, 0);
    /* Don't free Image data, because XCreateImage does not copy this data ! */
  }

  if (SizeX != NULL)
    *SizeX = PNGInfo->width;
  if (SizeY != NULL)
    *SizeY = PNGInfo->height;

  png_destroy_read_struct (&PNG, (png_info**)NULL, (png_info**)NULL);

  PRO_END
  return Image;
}
#endif
