#include <stdarg.h>
#include <ctype.h>

#include "libtools.h"


#define MAX_NAME_LEN 128


#define PARSE_RETURN(Msg) \
        { \
          free (Start); \
          PRO_RETURN (Msg, 0) \
        } 

/* Parse-Strukturen */
enum TC_PARSE_CODE
{
  NOCODE = 0, 
  KEYCODE, TEXTCODE, REMCCODE, REMSCODE,
  INTCODE, FLOATCODE, CHARCODE, STRINGCODE, LINECODE, DUALCODE, IDCODE
};

enum TC_BACKET_CODE
{
  NOBRACE = 0, 
  LESSBRACE, RNDBRACE, NDXBRACE, SETBRACE, SETSETBRACE
};

typedef struct 
{
  int   IBracket;       /* innere Klammer     */
  int   OBracket;       /* uere Klammer     */
  int   Code;           /* Wert-Code          */
  int   Length;         /* max. Stringlnge   */
  int   Number;         /* Anzahl IDs         */
  int   Amount;         /* {{...}}-hufigkeit */
  char  IDWord[30][25]; /* max 30 ID-Wrter mit je 25 Zeichen */
} PARSE_INFO;


static dual Error = FALSE;


static char *TOOL_GetData (const char *Text)
{
  int   count1 = 0, count2 = 0;
  int   len;
  char *Result = NULL;

  PRO_START (" ")

  if (Text == NULL)
    PRO_RETURN ("No text found", Result)

  len = strlen (Text);
  Result = (char *)calloc (len+2, sizeof(char));

  do
  {
    if ((Result[count2] = Text[count1++]) == '#')
    {
      Result[count2] = '\n';
      do
        count1++;
      while (Text[count1] != '\n' && Text[count1] != '\0');      
    }
  } while (count1 <= len && Result[count2++] != '\0');

  Result[count2]   = '\0';
  
  PRO_END
  return Result;
}


static dual TOOL_IsSpace (char letter)
{
  return (isspace(letter) && letter != '' && letter != '');
}


static char TOOL_GetASCII (char **Data)
{
  char letter;
  
  if ((letter = **Data) != '\0')
    (*Data)++;
  
  return letter;
}


static char TOOL_SkipSpace (char **Data, dual EOL)
{
  char letter;

  do
    if ((letter = TOOL_GetASCII(Data)) == '\\')
      if (**Data == '\\')
      {
        do
          letter = TOOL_GetASCII (Data);
        while (letter != '\n');
        letter = TOOL_GetASCII (Data);
      }
  while (TOOL_IsSpace(letter) && (letter != '\n' || !EOL) && letter != '\0');

  return letter;
}


static dual TOOL_CheckData (char *Data)
{
  int count = 0;
  int Result;

  if (Data == NULL)
    return FALSE;

  while(TOOL_IsSpace(*Data) && *Data != '\0')
  {
    Data++;
    count++;
  }
  
  Result = (*Data != '\0');
  /*  (*Data) -= count; */

  return Result;
}


static char TOOL_GetChar (char **Data, dual SkipSpace)
{
  char letter = '\0';

  PRO_START (" ")

  if (Data == NULL || *Data == NULL)
    PRO_RETURN ("No data found", '\0')

  if (SkipSpace)
    letter = TOOL_SkipSpace (Data, FALSE);
  else
    letter = TOOL_GetASCII (Data);

  PRO_END
  return letter;
}


static int TOOL_GetInt (char **Data, dual SkipSpace)
{
  int  count = 0, Result = 0;
  char letter;
  char Value[20];

  PRO_START (" ")

  if (Data == NULL || *Data == NULL)
    PRO_RETURN ("No data found", 0)

  Error = FALSE;

  if (SkipSpace)
    do
      letter = TOOL_GetASCII (Data);
    while (!isdigit(letter) && letter!='-' && letter!='+');
  else   
    letter = TOOL_SkipSpace (Data, FALSE);

  if (letter == '\0' ||
      (!isdigit(letter) && letter!='-' && letter!='+'))
  {
    Error = TRUE;
    PRO_RETURN ("Invalid charater found", 0)
  }

  while ((isdigit(letter) || letter == '-' || letter == '+'
          || letter == 'E' || letter == 'e') && count < 19)
  {
    Value[count++] = letter;
    letter = TOOL_GetASCII (Data);
  }

  (*Data)--;
  Value[count] = '\0';
  Result = atoi (Value);

  PRO_END
  return Result;
}


static double TOOL_GetDouble (char **Data, dual SkipSpace)
{
  int    count = 0;
  char   letter;
  char   Value[20];
  double Result = 0.0;

  PRO_START (" ")

  Error = FALSE;

  if (Data == NULL || *Data == NULL)
    PRO_RETURN ("No data found", 0.0)

  if (SkipSpace)
    do
      letter = TOOL_GetASCII (Data);
    while (!isdigit(letter) && letter != '\0' && 
           letter!='-' && letter!='+' && letter!='.');
  else   
    letter = TOOL_SkipSpace (Data, FALSE);

  if (letter == '\0' ||
      (!isdigit(letter) && letter != '-' && letter != '+' && letter != '.'))
  {
    Error = TRUE;
    PRO_RETURN ("Invalid character found", 0.0)
  }

  while ((isdigit(letter) || letter == '.' || letter == '-'  ||
	  letter == 'E' || letter == 'e' || letter == '+') &&
	 letter != '\0' && count < 19)
  {
    Value[count++] = letter;
    letter = TOOL_GetASCII (Data);
  }
  (*Data)--;
  Value[count] = '\0';
  Result = atof (Value);

  PRO_END
  return Result;
}


static char *TOOL_GetString (char **Data, char *Result, int Length, 
			     dual EndOfLine)
{
  int  count = 0;
  char letter;  

  PRO_START (" ")

  if (Data == NULL || *Data == NULL)
    PRO_RETURN ("No data found", NULL)
  if (Result == NULL || Length == 0)
    PRO_RETURN ("Invalid buffer", NULL)

  Error = FALSE;

  Result[0] = '\0';
  
  if (((letter=TOOL_SkipSpace(Data, EndOfLine)) == '\n') && EndOfLine)
  {
    Error = TRUE;
    PRO_RETURN ("Nothing found", NULL)
  }

  while (count < Length && letter != '\0' &&
         letter != ' ' && letter != '\t' && (letter != '\n' || ! EndOfLine))
  {
    Result[count++] = letter;
    letter = TOOL_GetASCII (Data);
  }
  (*Data)--;
  Result[count] = '\0';
  
  PRO_END
  return Result;
}


static char *TOOL_GetLine (char **Data, char *Result, int Length)
{
  int  count = 0;
  char letter;

  PRO_START (" ")

  if (Data == NULL || *Data == NULL)
    PRO_RETURN ("No data found", NULL)
  if (Result == NULL || Length == 0)
    PRO_RETURN ("invalid buffer", NULL)

  Error = FALSE;

  Result[0] = '\0';

  if ((letter = TOOL_SkipSpace(Data, TRUE)) == '\n')
  {
    Error = TRUE;
    PRO_RETURN ("Empty Line found", NULL)
  }

  Result[count++] = letter;
  letter = TOOL_GetASCII (Data);
  while (count < Length-1 && letter != '\n' && letter != '\0')
  {
    if (letter == '\\')
    {
      if (**Data == '\\')
        do
          letter = TOOL_GetASCII (Data);
        while (letter != '\n' && letter != '\0');
      Result[count++] = '\\';
    }
    Result[count++] = letter;
    
    letter = TOOL_GetASCII (Data);
  }
  (*Data)--;
  Result[count] = '\0';
  
  PRO_END
  return Result;
}


int TOOLParse (const char *Text, const char *Mask, ...)
{
  int          Result = 0;
  int          IDVal;
  int          IntVal;
  double       DoubleVal;
  char         CharVal;
  char        *TextVal = NULL;

  dual        *RetDual = NULL;
  int         *RetInt = NULL;
  int         *RetID = NULL;
  double      *RetDouble = NULL;
  char        *RetChar = NULL;
  char        *RetText = NULL;
  char        *RetLine = NULL;
  void       **RetMulti = NULL;
  char        *RetData = NULL;

  va_list      Arguments;
  dual         ready = FALSE;
  dual         EOL = TRUE;
  dual         found = TRUE, ok = FALSE;
  int          count, mc = 0;
  int          repeat = 0, Size;
  char         KeyWord[80] = "";
  char         TextWord[80] = "";
  char         String[1024];
  PARSE_INFO   Parse;
  char        *Data = NULL, *Start;
  char        *save = NULL;

  PRO_START ("\n")

  if (Text == NULL)
    PRO_RETURN ("No text found", 0)

  Start = Data = TOOL_GetData (Text);

  va_start (Arguments, Mask);

  Parse.OBracket = NOBRACE;
  Parse.Amount   = 0;
  while (Mask[mc] != '\0')
  {
    /* ***** nchsten Maskeneintrag suchen & auswerten: ***** */
    ready = FALSE;
    Parse.IBracket = NOBRACE;
    Parse.Code     = NOCODE;
    Parse.Number   = 0;
    while (! ready && Mask[mc] != '\0')
    {
      switch (Mask[mc])
      {
        case '<' : 
          if (Parse.IBracket != NOBRACE)
            Parse.OBracket = Parse.IBracket;
          Parse.IBracket = LESSBRACE;
          break;

        case '(' : 
          if (Parse.IBracket != NOBRACE)
            Parse.OBracket = Parse.IBracket;
          Parse.IBracket = RNDBRACE;
          break;

        case '[' : 
          Parse.IBracket = NDXBRACE;
          break;

        case '{' : 
          if (Mask[mc+1] == '{')
          {
            Parse.IBracket = SETSETBRACE;
            mc++;
            repeat = mc;
            Size = 0;
            while (Mask[mc] != '}' || Mask[mc+1] != '}')
              if (Mask[mc++] == '%')
              {
                if (isdigit(Mask[mc]))
                {
                  count = 0;
                  do
                    String[count++] = Mask[mc++];
                  while (isdigit(Mask[mc]));
                  String[count] = '\0';
                  Parse.Length = atoi (String);
                }
                else
                  Parse.Length = MAX_NAME_LEN;

                switch (Mask[mc])
                {
                  case 'd' : Size += sizeof(int);
                             break;
                  case 'f' : Size += sizeof(double);
                             break;
                  case 'c' : Size += sizeof(char);
                             break;
                  case 's' : Size += sizeof(char) * Parse.Length;
                             break;
                  case 'l' : Size += sizeof(char) * Parse.Length;
                             break;
                   default : PARSE_RETURN ("Unknown Data-type")
                }
                mc++;
              }
            mc = repeat;
            RetData = (char*)malloc (Size * Parse.Amount);
            RetMulti = (void**)va_arg (Arguments, void**);
            *RetMulti = (void*)RetData;
          }
          else
            Parse.IBracket = SETBRACE;
          break; 

        case '%' : 
          mc++;
          if (Parse.IBracket != NOBRACE)
          {
            if (isdigit(Mask[mc]))
            {
              count = 0;
              do
                String[count++] = Mask[mc++];
              while (isdigit(Mask[mc]));
              String[count] = '\0';
              Parse.Length = atoi (String);
            }
            else
              Parse.Length = MAX_NAME_LEN;

            switch (Mask[mc])
            {
              case 'd' : Parse.Code = INTCODE;
                         break;
              case 'f' : Parse.Code = FLOATCODE;
                         break;
              case 'c' : Parse.Code = CHARCODE;
                         break;
              case 's' : Parse.Code = STRINGCODE;
                         break;
              case 'l' : Parse.Code = LINECODE;
                         break;
               default : PARSE_RETURN ("Unknown data-type")
            }
            mc++;
            while (Mask[mc] != ')' && Mask[mc] != ']' && Mask[mc] != '}')
              mc++;
          }
         else
          {
            Parse.Code = TEXTCODE;
            TextWord[0] = '%';
            count = 1;
            while (Mask[mc] != ' ' && Mask[mc] != '\t' && Mask[mc] != '\n')
              TextWord[count++] = Mask[mc++];
            TextWord[count] = '\0';
          }
          ready = TRUE;
          break;

        case ' ' : 
       case '\n' : 
       case '\t' : 
         break;

        case '#' :
          if (Mask[mc+1] == '#')
          { 
            Parse.Code = REMSCODE; mc++;
          }
         else
            Parse.Code = REMCCODE;
          ready = TRUE;
          break;

        case ')' : 
          if (Parse.OBracket != RNDBRACE)
            PARSE_RETURN ("  Too much ')'-Brackets")
          Parse.OBracket = NOBRACE;
          break;

        case ']' : 
          if (Parse.OBracket != NDXBRACE)
	    PARSE_RETURN ("Too much ']'-Brackets")
          Parse.OBracket = NOBRACE;
          break;

        case '}' :
          if (Parse.OBracket != SETBRACE && Parse.OBracket != SETSETBRACE)
	    PARSE_RETURN ("Too much '}'-Brackets")

          if (Parse.OBracket == SETSETBRACE)
            if (Mask[mc+1] != '}')
	      PARSE_RETURN ("Bad close-brackets '}}'")
            else if (Parse.Amount > 1)
            {
              Parse.Amount--;
	      mc = repeat;
            }
            else
            {
              Parse.OBracket = NOBRACE; 
	      mc++;
            }
          else
            Parse.OBracket = NOBRACE;                        
          break;

        default : 
          if (Parse.IBracket == NOBRACE)
          {
            count = 0;
            while (Mask[mc] != ' ' && Mask[mc] != '\t' && Mask[mc] != '\n')
              TextWord[count++] = Mask[mc++];
            TextWord[count] = '\0';
            Parse.Code = TEXTCODE;
            ready = TRUE;
          }
          else if (Parse.IBracket == LESSBRACE)
          {
            count = 0;
            while (Mask[mc] != '>')
              KeyWord[count++] = Mask[mc++];
            KeyWord[count] = '\0';
            Parse.Code = KEYCODE;
            ready = TRUE;
          }
          else if (Parse.IBracket > LESSBRACE)
          {
            Parse.Number = 1; count = 0;
            Parse.Length = MAX_NAME_LEN;
            Parse.Code = IDCODE;
            IDVal = 0;
            while (Mask[mc] != ')' && Mask[mc] != ']' && Mask[mc] != '}')
            {
              if (Mask[mc] == '|')
              {
                Parse.IDWord[Parse.Number-1][count] = '\0';
                Parse.Number++; count = 0;
              }
              else
              {
                Parse.IDWord[Parse.Number-1][count] = Mask[mc];
                count++;
              }
              mc++;
            }
            Parse.IDWord[Parse.Number-1][count] = '\0';
            if (Mask[mc] == '}' && Mask[mc+1] == '}')
              mc++;
            ready = TRUE;
          }
      }
      mc++;
    }

    if (!ready && Mask[mc] == '\0' && 
        (Parse.IBracket != NOBRACE || Parse.Code != NOCODE))
      PARSE_RETURN ("End of Mask reached !")

    /* geparsten Code in der Datei ermitteln und in Ergebnis schreiben */
    if (Parse.Code == KEYCODE && 
        (Parse.OBracket != SETSETBRACE || Parse.Amount > 0))
    {
      PRO1 ("Parse.KeyWord '%s'\n", KeyWord)
      save = Data;
      TOOL_GetString (&Data, String, 1024, TRUE);
      while (!(found = (strcmp(String, KeyWord) == 0)) && TOOL_CheckData(Data))
        TOOL_GetString (&Data, String, 1024, TRUE);
      if (! TOOL_CheckData(Data))
      {
        if (Parse.OBracket == NDXBRACE)
        {
          PRO0 ("unnessesary")
	  Data = save;
        }
        else
	  PARSE_RETURN ("NOT FOUND")
      }
      PRO1 ("found '%s'\n", String)
    }
    else if (Parse.Code == TEXTCODE &&
             (Parse.OBracket != SETSETBRACE || Parse.Amount > 0))
    {
      PRO1 ("Parse.TextWord %s\n", TextWord)

      TOOL_GetString (&Data, String, 1024, FALSE);
      if (strcmp (String, TextWord) != 0)
	PARSE_RETURN ("NOT FOUND")
      PRO1 ("found '%s'", String)
    }
    else if (Parse.Code != NOCODE && 
             (Parse.OBracket != SETSETBRACE || Parse.Amount > 0))
    {
      PRO0 ("Parse...")
      if (Parse.IBracket == NDXBRACE)
        save = Data;
      switch (Parse.Code)
      {
        case REMCCODE :
          CharVal = TOOL_GetChar (&Data, TRUE);
          PRO1 ("Char-Remark %d", CharVal)
          break;

        case REMSCODE : 
          TOOL_GetString (&Data, String, 1024, TRUE);
          PRO1 ("String-Remark %s", String) 
          break;

        case INTCODE :
          PRO0 ("\\INT\n")
          if (Parse.IBracket == RNDBRACE || Parse.IBracket == NDXBRACE)
          {
            IntVal = TOOL_GetInt (&Data, FALSE);
            if (Parse.OBracket != SETSETBRACE)
            {
              RetInt = (int*)va_arg(Arguments, int*);
              if (Parse.OBracket != NDXBRACE || found)
                if (! Error)
                  Parse.Amount = *RetInt = IntVal;
                else if (Parse.IBracket == NDXBRACE)
                {
		  Data = save;
                  Parse.Amount = 0;
                }
                else
		  PARSE_RETURN ("Invalid brace")
              else
		Data = save;
            }
            else if (Parse.Amount > 0)
            {
              memcpy ((void*)RetData, (void*)&IntVal, sizeof(int));
              RetData += sizeof (int);
            }
            PRO1 ("INT: %d\n", IntVal)
            Result++;
          }
          break;

        case FLOATCODE :
          PRO0 ("\\DOUBLE\n")
          if (Parse.IBracket == RNDBRACE || Parse.IBracket == NDXBRACE)
          {
            DoubleVal = TOOL_GetDouble (&Data, FALSE);
            if (Parse.OBracket != SETSETBRACE)
            {
              RetDouble = (double*)va_arg(Arguments, double*);
              if (Parse.OBracket != NDXBRACE || found)
                if (! Error)
                  *RetDouble = DoubleVal;
                else if (Parse.IBracket == NDXBRACE)
		  Data = save;
                else
		  PARSE_RETURN ("Invalid brace")
              else
		Data = save;
            }
            else if (Parse.Amount > 0)
            {
              memcpy ((void*)RetData, (void*)&DoubleVal, sizeof(double));
              RetData += sizeof (double);
            }
            PRO1 ("DOUBLE: %f\n", DoubleVal)
	    Result++;
          }
          break;

        case CHARCODE :
          PRO0 ("\\CHAR\n")
          if (Parse.IBracket == RNDBRACE || Parse.IBracket==NDXBRACE)
          {
            CharVal = TOOL_GetChar (&Data, FALSE);
            if (Parse.OBracket != SETSETBRACE)
            {
              RetChar = (char*)va_arg(Arguments, char*);
              if (Parse.OBracket != NDXBRACE || found)
                if (ok)
                  *RetChar = CharVal;
                else if (Parse.IBracket == NDXBRACE)
		  Data = save;
                else
		  PARSE_RETURN ("Invalid brace")
              else
		Data = save;
            }
            else if (Parse.Amount > 0)
            {
              memcpy ((void*)RetData, (void*)&CharVal, sizeof(char));
              RetData += sizeof (char);
            }
	    PRO1 ("CHAR: <%c>\n", CharVal)
	    Result++;
          }
          break;

        case STRINGCODE :
          PRO0 ("\\STRING\n")
          if (Parse.IBracket == RNDBRACE || Parse.IBracket == NDXBRACE)
          {
            TextVal = (char*)malloc (sizeof(char) * Parse.Length);
            TOOL_GetString (&Data, TextVal, Parse.Length, TRUE);
            if (Parse.OBracket != SETSETBRACE)
            {
              RetText = (char*)va_arg(Arguments, char*);
              if (Parse.OBracket != NDXBRACE || found)
	      {
                if (ok || Parse.IBracket == NDXBRACE)
                  strcpy (RetText, TextVal);
                else
                {
                  free (TextVal);
		  PARSE_RETURN ("Invalid index-brace")
                }
	      }
            }
            else if (Parse.Amount > 0)
            {
              memcpy ((void*)RetData, (void*)TextVal, 
                      sizeof(char)*Parse.Length);
              RetData += sizeof (char) * Parse.Length;
            }
            PRO1 ("TEXT: <%s>\n", TextVal)
            free (TextVal);
            Result++;
          }
          break;

        case LINECODE :
          PRO0 ("\\LINE\n")
          if (Parse.IBracket == RNDBRACE || Parse.IBracket == NDXBRACE)
          {
            TextVal = (char*)malloc (sizeof(char) * Parse.Length);
            TOOL_GetLine (&Data, TextVal, Parse.Length);
            if (Parse.OBracket != SETSETBRACE)
            {
              RetLine = (char*)va_arg(Arguments, char*);
              if (Parse.OBracket != NDXBRACE || found)
	      {
                if (ok || Parse.IBracket == NDXBRACE)
                  strcpy (RetLine, TextVal);
                else
                {
                  free (TextVal);
		  PARSE_RETURN ("Invalid index-braces")
		}
	      }
            }
            else if (Parse.Amount > 0)
            {
              memcpy ((void*)RetData, (void*)TextVal,
		      sizeof(char) * Parse.Length);
              RetData += sizeof(char) * Parse.Length;
            }
          }
          free (TextVal);
          PRO1 ("LINE: <%s>\n", TextVal)
          Result++;
          break;

        case IDCODE :
          PRO0 ("\\ID\n")
          IDVal = 0;
          count = 0;
          do
          {
            if (!(EOL= ! TOOL_GetString(&Data, String, Parse.Length, TRUE)))
            {
              ok = FALSE;
              count = 0;
              while (count < Parse.Number && ! ok)
                if ((ok = (strcmp(String, Parse.IDWord[count]) == 0)))
                  IDVal += 1 << count;
                else
                  count++;
              if (! ok)
		PARSE_RETURN ("Invalid Keyword")
            }
          }
          while (! EOL && Parse.IBracket == SETBRACE && Parse.Number > 1);

          if (Parse.Number == 2 && Parse.IBracket != SETBRACE)
          {
            RetDual = (dual *)va_arg (Arguments, dual *);
            if (found)
              *RetDual = (dual)(IDVal-1);
            PRO1 ("ID: %d\n", *RetDual)
            Result++;
          }
          else
          {
            RetID = (int *)va_arg (Arguments, int *);
            if (found)
	    {
              if (Parse.IBracket == RNDBRACE)
                *RetID = count;
              else
                *RetID = IDVal;
	      PRO1 ("ID: %d\n", IDVal)
	      Result++;
	    }
          }
          break;

        default : 
	  break;
      }
    }
  }
  free (Start);

  PRO_END
  return Result;
}

