#include <math.h>

#if (LIB_PNG==YES)
#include <png.h>
#endif

#include "libtools.h"


int TOOLGetImageType (const char *FileName)
{
  int   Type = UNKNOWN_FORMAT;
  char  Buffer[128];
  FILE *fp;

  PRO_START (" ")

  if ((fp = fopen(FileName, "r")) == NULL)
    PRO_RETURN ("Read error", UNKNOWN_FORMAT)

  fread (Buffer, sizeof(char), sizeof(Buffer), fp);

  fclose (fp);

  if (strstr(Buffer, "#define") != NULL)
    Type = XBM_FORMAT;
   elsif (strncmp(Buffer+6, "JFIF", 4) == 0)
     Type = JPG_FORMAT;
#if (LIB_PNG==YES)
   elsif (png_check_sig((unsigned char *)Buffer, sizeof(Buffer)))
    Type = PNG_FORMAT;
#endif
  elsif (strstr(Buffer, "char") != NULL)
    Type = XPM_FORMAT;


  PRO_END
  return Type;
}


static int TOOL_GetMSBPos (unsigned long Value)
{ 
  int count = 0;

  while (Value != 0)
  {
    Value = Value >> 1;
    count++;
  }

  return count;
}


byte *TOOLMakeRGBA (XImage *Image, double Alpha, int *SizeX, int *SizeY)
{
  unsigned int   r, g, b, a = 1;
  unsigned long  Pixel;
  int            RBit, GBit, BBit;
  int            count = 0;
  int            countX, countY;
  int            Components = 3;
  byte          *Result = NULL;

  PRO_START ("\n")
  
  if (Image == NULL || Image->format != ZPixmap)
    PRO_RETURN ("No/Invalid Image", NULL)

  if (Alpha >= 0.0 && Alpha <= 1.0)
  {
    a          = (unsigned int)(Alpha * 255);
    Components = 4;
  }

  PRO2 ("Size:        %d x %d\n", Image->width, Image->height)
  PRO1 ("Bitmap-Unit: %d\n", Image->bitmap_unit)
  PRO1 ("BitOrder:    %s\n", (Image->bitmap_bit_order==LSBFirst)? "LSB":"MSB")
  PRO1 ("Bytes/Line:  %d\n", Image->bytes_per_line)
  PRO1 ("Bits/Pixel:  %d\n", Image->bits_per_pixel)
  PRO3 ("Mask:        R:%06lx G:%06lx B:%06lx\n", 
	Image->red_mask, Image->green_mask, Image->blue_mask)

  /* Compte the position of the most significant bit in the mask */
  RBit = TOOL_GetMSBPos (Image->red_mask);
  GBit = TOOL_GetMSBPos (Image->green_mask);
  BBit = TOOL_GetMSBPos (Image->blue_mask);

  PRO3 ("Bits: R=%d G=%d  B=%d\n", RBit, GBit, BBit)

  Result = (byte *)malloc (sizeof(byte) * 
			   Image->width * Image->height * Components);
  for (countY = 0; countY < Image->height; ++countY)
    for (countX = 0; countX < Image->width; ++countX)
    {
      /* Read the pixel from image data */
      Pixel = XGetPixel (Image, countX, countY);

      /* normalize to 8-bit values (use most-significant-bits) */
      r = (Pixel & Image->red_mask);
      if (RBit > 8)
	r = r >> (RBit - 8);
      elsif (RBit < 8)
	r = r << (8 - RBit);

      g = (Pixel & Image->green_mask);
      if (GBit > 8)
	g = g >> (GBit - 8);
      elsif (GBit < 8)
	g = g << (8 - GBit);

      b = (Pixel & Image->blue_mask);
      if (BBit > 8)
	b = b >> (BBit - 8);
      elsif (BBit < 8)
	b = b << (8 - BBit);

      Result[count*Components + 0] = (byte)r;
      Result[count*Components + 1] = (byte)g;
      Result[count*Components + 2] = (byte)b;
      if (Components == 4)
        Result[count*Components + 3] = (byte)a;

      count++;
    }

  if (SizeX != NULL)
    *SizeX = Image->width;
  if (SizeY != NULL)
    *SizeY = Image->height;

  PRO_END
  return Result;
}


Pixmap TOOLImage2Pixmap (XImage *Image)
{
  Pixmap Result = (Pixmap)NULL;

  PRO_START ("")

  if (Image == NULL)
    PRO_RETURN ("No Image", Result)

  Result = XCreatePixmap (XDisplay, DefaultRootWindow(XDisplay),
			  Image->width, Image->height, TOOLGetDepth());
  XPutImage (XDisplay, Result, DefaultGC (XDisplay, XScreen), Image, 
	               0, 0, 0, 0, Image->width, Image->height);

  PRO_END
  return Result;
}


XImage *TOOLCreateImage (int SizeX, int SizeY, double r, double g, double b)
{
  int       count;
  int       Depth, BPP;
  byte     *Data, *Ptr;
  XColor    Color;
  Colormap  CMap;
  XImage   *Image;

  PRO_START ("")

  Depth = TOOLGetDepth         ();
  BPP   = TOOLGetBytesPerPixel ();

  CMap = DefaultColormap (XDisplay, XScreen);

  Color.pad = 0;
  Color.flags = DoRed | DoGreen | DoBlue;
  Color.red   = r * 65535;
  Color.green = g * 65535;
  Color.blue  = b * 65535;
  if (! XAllocColor(XDisplay, CMap, &Color))
  {
    PRO0 ("Alloc Error\n")
    Color.pixel = TOOLGetColor ("red");
  }

  /* Allocate memory for image data */
  Ptr = Data = (byte *)malloc (sizeof(byte) * SizeX * SizeY * BPP);
  switch (BPP)
  {
    case 1 :
      for (count = 0; count < SizeX * SizeY; ++count)
      {
	*(byte *)Ptr = (byte)Color.pixel;
	Ptr += BPP;
      }
      break;

    case 2 :
      for (count = 0; count < SizeX * SizeY; ++count)
      {
	*(short *)Ptr = (short)Color.pixel;
	Ptr += BPP;
      }
      break;

    case 4 :
      for (count = 0; count < SizeX * SizeY; ++count)
      {
	*(int *)Ptr = (int)Color.pixel;
	Ptr += BPP;
      }
      break;

    default :
      PRO1 ("Unsupported Bytes per Pixel: %d\n", BPP)
  }

  Image = XCreateImage (XDisplay, DefaultVisual(XDisplay, XScreen),
		        Depth, ZPixmap, 0, (char *)Data,
			SizeX, SizeY, 8, 0);

  /* Do not free memory here. Will be managed by XImage (XDestroyImage) */

  PRO_END
  return Image;
}


XImage *TOOLScaleImage (XImage *Image, int SizeX, int SizeY)
{
  int     countX, countY;
  int     PosX, PosY;
  double  StepX, StepY;
  Pixel   Point;
  XImage *Result;

  PRO_START ("\n")

  if (Image == NULL)
    PRO_RETURN ("Empty Image", NULL)

  PRO4 ("ScaleImage: %d x %d -> %d x %d\n", 
        Image->width, Image->height, SizeX, SizeY)

  if (Image->width == SizeX && Image->height == SizeY)
    Result = TOOLCopyImage (Image);
  else
  {
    /* Create an initial white image */
    Result = TOOLCreateImage (SizeX, SizeY, 1.0, 1.0, 1.0);

    /* Compute step-increment in both directions */
    StepX = (double)Image->width  / (double)SizeX;
    StepY = (double)Image->height / (double)SizeY;

    for (countY = 0; countY < SizeY; ++countY)
      for (countX = 0; countX < SizeX; ++countX)
      {
	PosX = countX * StepX;
	PosY = countY * StepY;
	Point = XGetPixel (Image, PosX, PosY);
	XPutPixel (Result, countX, countY, Point);
      }
  }

  PRO_END
  return Result;
}


XImage *TOOLLoadImage (const char *FileName)
{
  XImage *Image = NULL;

  PRO_START ("\n")

  switch (TOOLGetImageType(FileName))
  {
    case XPM_FORMAT :
      Image = TOOLLoadXPM2 (FileName, NULL, NULL, TOOLGetColor("grey"));
      break;

    case PNG_FORMAT :
      Image = TOOLLoadPNG (FileName, NULL, NULL, FALSE);
      break;

    case JPG_FORMAT :
      Image = TOOLLoadJPG (FileName, NULL, NULL);
      break;

    default :
      PRO0 ("Image-type not supported\n")
  }

  PRO_END
  return Image;
}


XImage *TOOLCopyImage (XImage *Image)
{
  char   *Data;
  XImage *Result;

  PRO_START ("")

  if (Image == NULL)
    PRO_RETURN ("Copy image", NULL)

  /* XSubImage seems to be buggy */
  Data = (char *)malloc (sizeof(char) * Image->height * Image->bytes_per_line);
  Result = XCreateImage (XDisplay, DefaultVisual(XDisplay, XScreen),
			 Image->depth, Image->format, 0, Data, 
			 Image->width, Image->height,
			 Image->bitmap_pad, Image->bytes_per_line);
  memcpy (Result->data, Image->data, 
	  sizeof(char) * Image->height * Image->bytes_per_line);

  PRO_END
  return Result;
}


void TOOLDestroyImage (XImage *Image)
{
  char *Data;

  PRO_START ("")

  /* 
     When createing an image using our tool-functions, the memory might be
     managed by our debug-function. Because XCreateImage does not copy the
     image-data, but frees the data on XDestroyImage, but we have to manage
     the memory by ourself. So, let's save the data-ptr and set the pointer 
     in the image-structure to NULL. Then call XDestroyImage and free the data.
  */

  if (Image == NULL)
    PRO_ABORT ("No Image")

  Data = Image->data;
  Image->data = NULL;

  XDestroyImage (Image);

  if (Data != NULL)
    free (Data);

  PRO_END
}


void TOOLHSV2RGB (double  h, double  s, double  v,
		  double *r, double *g, double *b)
{
  double xh = fmod (h*360.0, 360) / 60.0;
  double i  = floor(xh);
  double f  = xh - i;
  double p1 = v * (1 - s);
  double p2 = v * (1 - (s * f));
  double p3 = v * (1 - (s * (1-f)));

  switch((int)i)
  {
    case 0: *r =  v; 
            *g = p3; 
	    *b = p1; 
            break;
    case 1: *r = p2; 
            *g =  v; 
	    *b = p1; 
            break;
    case 2: *r = p1; 
            *g =  v; 
	    *b = p3; 
	    break;
    case 3: *r = p1; 
            *g = p2;
	    *b =  v;
	    break;
    case 4: *r = p3; 
            *g = p1;
	    *b =  v;
	    break;
    case 5: *r =  v;
            *g = p1;
	    *b = p2;
  }
}



