#include <sys/stat.h>
#include <ctype.h>

#include "libtools.h"


typedef struct
{
  TOOL_LINE *Data;
  TOOL_LINE *Ptr;
} TOOL_SCRIPT;


static TOOL_SCRIPT  Script = {NULL, NULL};
static int          running = 0;
#if 0
  static void        *ScriptFkt = NULL;
  static void        *ScriptData = NULL;
#endif


static TOOL_LINE *TOOL_NewLine (void)
{
  TOOL_LINE *Result;

  Result = (TOOL_LINE*)malloc (sizeof(TOOL_LINE));
  Result->LineNo   = 0;
  Result->Type     = ' ';
  Result->Cmd      = NULL;
  Result->Val      = NULL;
  Result->optional = FALSE;
  Result->Next     = NULL;

  return Result;
}


dual TOOL_ReadComment (FILE *fp)
{
  char letter;

  do
    letter = getc (fp);
  while (letter != '\n' && letter != EOF);

  return (letter != EOF);
}


char TOOL_GetChar (FILE *fp, int *line)
{
  char letter;

  do
  {
    letter = getc (fp);
    if (letter == '#')
    {
      if (! TOOL_ReadComment(fp))
	return EOF;

      letter = '\n';
    }

    if (letter == '\n')
      (*line)++;
  } while (isspace(letter) && letter != EOF);

  return letter;
}


int TOOLOpenScript (char *FileName)
{
  dual       optional;
  int        count1, count2, line = 2;
  char       letter = ' ';
  char       Type = ' ';
  char       Cmd[256], Val[256];
  FILE      *fp = NULL;
  TOOL_LINE *Line = NULL, *Last = NULL;

  PRO_START ("\n")

  if (running > 0)
    PRO_RETURN ("another script is always running", -1)

  if ((fp = fopen(FileName, "r+")) == FALSE)
    PRO_RETURN ("Error while opening file", -1)

  if (getc(fp)!='S' || getc(fp)!='C' || getc(fp)!='R' || 
      getc(fp)!='I' || getc(fp)!='P' || getc(fp)!='T')
    PRO_RETURN ("File is not a script-file !", 1)

  while (letter != '\n')
    letter = getc(fp);

  while (letter != EOF)
  {
    Type     = ' ';
    Cmd[0]   = '\0';
    Val[0]   = '\0';
    optional = FALSE;

    letter = TOOL_GetChar (fp, &line);

    if (letter == EOF)
      break;

    if (letter == '[')
    {
	 optional = TRUE;
	 letter = TOOL_GetChar (fp, &line);
    }

    if (! isalpha(letter))
    {
      Type = letter;
      letter = TOOL_GetChar (fp, &line);
      count1 = 0;
    }
    else
    {
      Cmd[0] = Type;
      Type = ' ';
      count1 = 1;
    }

    /* Read Keyword... */
    if (letter != EOF)
    {
      while (! isspace(letter) && letter != ':' && 
		   count1 < 255 && letter != EOF)
      {
	   Cmd[count1++] = letter;
	   letter = getc (fp);
      }
      Cmd[count1] = '\0';
    }

    /* skip spaces and ':' */
    while (letter == ' ' || letter == '\t')
	 letter = getc (fp);

    if (letter == ':')
      letter = getc (fp);

    while (letter == ' ' || letter == '\t')
      letter = getc(fp);

    if (letter == EOF)
    {
	 TOOLCloseScript ();
      PRO_RETURN ("unexpected EOF", line)
    }

    /* Read value */
    count2 = 0;
    while (letter != '\n' && (letter != ']' || ! optional) && 
		 letter != EOF && count2 < 255)
    {
      Val[count2++] = letter;
      letter = getc (fp);
	 if (letter == '\\')
      {
        letter = getc (fp);
	   if (letter == '\\')
	   {
		do 
		  letter = getc (fp);
		while (letter != '\n' && letter != EOF);
		line++;
	   }
	   else
		switch (letter)
		{
		  case 'n' : Val[count2++] = '\n';
		             break;
 	   	  case 't' : Val[count2++] = '\t';
		             break;
  	    	  case ' ' : Val[count2++] = ' ';
		             break;
 		   default : Val[count2++] = '\\';
		             Val[count2++] = letter;
		}
	   letter = getc (fp);
	 }	   
    }
    if (optional && count2 > 0 && Val[count2-1] == ']')
	 count2--;

    Val[count2] = '\0';

    Line = TOOL_NewLine ();
    Line->LineNo = line;
    Line->Type   = Type;
    if (count1 > 0)
    {
      Line->Cmd = (char*)malloc (sizeof(char) * (count1+1));
      strcpy (Line->Cmd, Cmd);
    }
    if (count2 > 0)
    {
      Line->Val = (char*)malloc (sizeof(char) * (count2+1));
      strcpy (Line->Val, Val);
    }
    Line->optional = optional;

    if (letter == '\n')
      line++;

    if (Script.Data == NULL)
      Script.Data = Line;
    else
      Last->Next = Line;

    Last = Line;
  }

  Script.Ptr = Script.Data;

  fclose (fp);

#if 1
  PRO0 ("SCRIPT:\n")
  PRO0 ("===================================\n")
  Line = Script.Data;
  while (Line != NULL)
  {
    PRO1 ("\\%c", Line->optional?'[':' ')
    PRO4 ("\\#%2d: <%c> %s:  %s", 
		Line->LineNo, Line->Type, Line->Cmd, (Line->Val==NULL)?" ":Line->Val)
    PRO1 ("\\%c", Line->optional?']':' ')
    PRO0 ("\\\n")
    Line = Line->Next;
  }
  PRO0 ("===================================\n")
#endif
  running++;

  PRO_END
  return 0;
}


#if 0
dual TOOLStopScript (void *Function, void *Data)
{
  if (ScriptFkt == NULL)
    return FALSE;

  ScriptFkt  = Function;
  ScriptData = Data;  

  return TRUE;
}


dual TOOLContScript (void)
{
  void *Function (void*);
  void *Data;

  if (ScriptFkt == NULL)
    return FALSE;

  Function = (void*)(*)ScriptFkt;
  Data     = ScriptData;

  ScriptFkt  = NULL;
  ScriptData = NULL;

  Function (Data);
  return TRUE;
}
#endif


void TOOLCloseScript ()
{
  TOOL_LINE *Line, *NextLine;

  PRO_START ("")

  if (running <= 0 || Script.Data == NULL)
    PRO_ABORT ("No script open !")

  Line = Script.Data;
  while (Line != NULL)
  {
    NextLine = Line->Next;

    if (Line->Cmd != NULL)
      free (Line->Cmd);
    if (Line->Val != NULL)
      free (Line->Val);

    free (Line);

    Line = NextLine;
  }
  Script.Data = NULL;
  Script.Ptr  = NULL;

  running--;

  PRO_END
}


dual TOOLScriptRunning (void)
{
  return (running > 0);
}


dual TOOLScriptOK (void)
{
  if (running == 0 || Script.Ptr == NULL)
  {
    TOOLCloseScript ();
    return FALSE;
  }

  return (Script.Ptr != NULL);
}


/* ********************************************************************** */
/* ** Returns a pointer to the actual script-line's data               ** */
/* ** INPUT:  ---                                                      ** */
/* ** OUTPUT: (TOOL_LINE*)  Pointer to sommand-data                    ** */
/* **                       NULL, if no script or end is reached       ** */
/* ********************************************************************** */
TOOL_LINE *TOOLGetScriptCmd (void)
{
  PRO_START (" ")

  if (running == 0)
    PRO_RETURN ("No script running", NULL)

  if (Script.Ptr == NULL)
    PRO_RETURN ("End of script reached", NULL)

  PRO_END
  return Script.Ptr;
} 


/* ********************************************************************** */
/* ** Jumps to the next script-command                                 ** */
/* ** INPUT:  ---                                                      ** */
/* ** OUTPUT: (dual)  TRUE on success, FALSE otherwise                 ** */
/* ********************************************************************** */
dual TOOLNextScriptCmd (void)
{
  PRO_START (" ")

  if (running == 0)
    PRO_RETURN ("No script running", FALSE)

  if (Script.Ptr == NULL)
    PRO_RETURN ("End of script reached", FALSE)

  Script.Ptr = Script.Ptr->Next;

  PRO_END
  return TRUE;
}
