/* *********************************** */
/* *** (c) Stefan Vietinghoff      *** */
/* *** basic X-Window functions    *** */
/* *********************************** */

#include <sys/time.h>

#include <stdio.h>
#include <unistd.h>
#include <ctype.h>

#include "libtools.h"

#include "magnify.xbm"
#include "magnify_mask.xbm"


#define ToolClass "ToolLib"


#define FABS(x) (((x) < 0)? (-x):(x))


typedef struct
{
  int   Value;
  float Real;
  char  Char;
  char *String;
} RESDATATYPE;

typedef char CSYMB[4];
typedef struct
{
  int    Index;
  CSYMB  Symbol;
  XColor Color;
  char   CName[20];
  int    amount;
} TOOL_PIXEL;


Display      *XDisplay = NULL;
XtAppContext  AppCon = NULL;
Widget        AppShell = NULL;
int           XScreen = 0;

GC            defGC, drawGC, xorGC, copyGC;
Colormap      CMap;
#if 0
  Colormap      CLUT;
  static Pixel  CTable[216];
  static int    CIndex = 0;
#endif


static void TOOL_StartCB (Widget widget, void *Data, XmAnyCallbackStruct *Ptr);
static void TOOL_ExitCB  (Widget widget, void *Data, XmAnyCallbackStruct *Ptr);

static dual         initialized = FALSE;
static dual         running = FALSE;
static Arg          args[10];
static RESDATATYPE  ResResult;
static Cursor       MouseCursor = None;


/*----------------------------------------------------------------------
  Funktion: Ermittelt einen Zeichenketteneintrag  aus der zur Anwendung
            gehrenden Resourcedatei
  
  INPUT:  (Parent, Class, Spec, Name : siehe TOOLGetResString)
          Default  Defaultwert, der bei erfolgloser Suche zurckgegeben
                   wird.
  OUTPUT: (int)    gelesener Eintrag bzw. Defaultwert
  
  Bsp.: Widget widget;
        int    Value;
	...
	Value = TOOLGetResInt (widget, "xmTextWidgetClass", "Wert", 
	                               "Eintrag", 129);
  ----------------------------------------------------------------------*/
int TOOLGetResInt (Widget  Parent, char *Class, char *Spec, 
		   char   *Name,   int   Default)
{
  XtResource ResData[1];

  ResData[0].resource_name   = Name;
  ResData[0].resource_class  = ToolClass;
  ResData[0].resource_type   = XmRInt;
  ResData[0].resource_size   = sizeof(int);
  ResData[0].resource_offset = XtOffset (RESDATATYPE*, Value);
  ResData[0].default_type    = XmRInt;
  ResData[0].default_addr    = &Default;

  ResResult.Value = 0;
  XtGetSubresources (Parent, &ResResult, Spec, Class, 
		             ResData, 1, (ArgList)NULL, 0);

  return ResResult.Value;
}


/*----------------------------------------------------------------------
  Funktion: quivalent zu TOOLGetResInt fr Fliekommawerte
            (eigentlich nicht ganz korrekt, da keine double, sondern
	    nur float-Werte gelesen werden.
  
  INPUT:  (Parent, Class, Spec, Name : siehe TOOLGetResString)
          Default  Defaultwert, der bei erfolgloser Suche zurckgegeben wird.
  OUTPUT: (double) gelesener Eintrag bzw. Defaultwert

  Bsp.: Widget widget;
        double Value;
	...
	Value = TOOLGetResDouble (widget, "xmTextWidgetClass", "wert", 
	                                   "eintrag", 129.0);
  ----------------------------------------------------------------------*/
double TOOLGetResDouble (Widget  Parent, char *Class, char *Spec, 
			 char   *Name,   double Default)
{
  float       DefValue;
  XtResource  ResData[1];

  DefValue = (float) Default;
  ResData[0].resource_name   = Name;
  ResData[0].resource_class  = ToolClass;
  ResData[0].resource_type   = XmRFloat;
  ResData[0].resource_size   = sizeof(float);
  ResData[0].resource_offset = XtOffset (RESDATATYPE*, Real);
  ResData[0].default_type    = XmRFloat;
  ResData[0].default_addr    = &DefValue;

  /* Ergebnis aus der Resourcedatei lesen: */
  ResResult.Real = 0.0;
  XtGetSubresources (Parent, &ResResult, Spec, Class, 
		     ResData, 1, (ArgList)NULL, 0);

  return (double)ResResult.Real;
}


/*----------------------------------------------------------------------
  Function: Reads a character out of the application-resource-file
  
  INPUT:  (Parent, Class, Spec, Name : see TOOLGetResString)
          Default  Default-value, which will delivered if the specified
	           entry is missing.
  OUTPUT: (char)   resource- or default-character

  Bsp.: Widget widget;
        char   Value;
	...
	Value = TOOLGetResChar (widget, "xmTextWidgetClass", "value", 
	                                "entry", 'a');
  ----------------------------------------------------------------------*/
char TOOLGetResChar (Widget Parent, char *Class, char *Spec, 
                                    char *Name,  char Default)
{
  XtResource  ResData[1];

  ResData[0].resource_name   = Name;
  ResData[0].resource_class  = ToolClass;
  ResData[0].resource_type   = XmRChar;
  ResData[0].resource_size   = sizeof(char);
  ResData[0].resource_offset = XtOffset (RESDATATYPE*, Char);
  ResData[0].default_type    = XmRChar;
  ResData[0].default_addr    = &Default;

  ResResult.Char = '\0';
  XtGetSubresources (Parent, &ResResult, Spec, Class,
                     ResData, 1, (ArgList)NULL, 0);

  return ResResult.Char;
}


/*----------------------------------------------------------------------
  Funktion: Durchsucht die Resourcedatei nach einem zu den Parametern
            passenden Eintrag und liefert diesen bei Erfolg zurck.
	    Andernfalls wird der Defaultwert zurckgeliefert.
	    Von den Werten wird keine Kopie bergeben !
  
  INPUT:  Parent   bergeordnetes Widget, auf das sich der Resourceeintrag 
                   beziehen soll
          Class    Klasse, die das Widget angehrt
	  Spec     Name des eigenen Widgets
	  Name     Name des Resourceeintrags (z.B. foreground,...)
	  Default  Defaultstring, der zurckgeliefert wird, wenn kein
	           passender Eintrag in der Resource-Datei existiert
  OUTPUT: (char *)  Zeiger auf die gelesene Zeichenkette bzw. den Default-
                   eintrag.
  
  Bsp.: Widget widget;
        ...
	TOOLGetResString (widget, "xmLabelWidgetClass", "Titel", 
	                          "foreground", "Red");
  ----------------------------------------------------------------------*/
char *TOOLGetResString (Widget Parent, char *Class, char *Spec, 
		                       char *Name,  char *Default)
{
  XtResource ResData[1];

  ResData[0].resource_name   = Name;         /* Resource-Struktur init. */
  ResData[0].resource_class  = ToolClass;
  ResData[0].resource_type   = XmRString;
  ResData[0].resource_size   = sizeof(char) * 80;
  ResData[0].resource_offset = XtOffset (RESDATATYPE*, String);
  ResData[0].default_type    = XmRString;
  ResData[0].default_addr    = Default;

  /* und Resourcedatei auslesen: */
  ResResult.String = NULL;
  XtGetSubresources (Parent, &ResResult, Spec, Class,
		             ResData, 1, (ArgList)NULL, 0);

  return ResResult.String;
}


/* ****************************************** */
/* ** return the depth of current screen   ** */
/* ** INTPUT:  ---                         ** */
/* ** OUTPUT: (int)  depth of screen [bit] ** */
/* ****************************************** */
int TOOLGetDepth (void)
{
  return DisplayPlanes (XDisplay, XScreen); /* DefaultScreen(XDisplay)); */
}


/*----------------------------------------------------------------------
  Funktion: Determine the number of bytes needed for each pixel
  ----------------------------------------------------------------------*/
int TOOLGetBytesPerPixel (void)
{
  int Depth;

  Depth = TOOLGetDepth();
  if (Depth <= 8)
    return 1;
  elsif (Depth <= 16)
    return 2;
  else
    return 4;
}


/*----------------------------------------------------------------------
  Funktion: Umwandlung einer Farb-Zeichenkette in einen von Motif ver-
            wendeten Farbtyp
  
  INPUT:  ColorStr  Farbname
  OUTPUT: (Pixel)   Motif Farb-Datentyp
  
  Bsp.: Pixel color;
        ...
        TOOLGetColor ("Red");
  ----------------------------------------------------------------------*/
Pixel TOOLGetColor (char *ColorStr)
{
  XrmValue from, to;
  XColor   Color;
  Pixel    Result = 0;

  PRO_START ("")
  PRO1 ("\\%s...", ColorStr)

  if (! initialized || ColorStr == NULL || ColorStr[0] == '\0')
    PRO_RETURN ("Base-Lib is not initialized", (Pixel)0)

  if (ColorStr[0] == '#')
  {
    if (XParseColor(XDisplay, CMap, ColorStr, &Color) &&
	XAllocColor(XDisplay, CMap, &Color))
      Result = Color.pixel;
  }
  else
  {
    from.addr = ColorStr;
    from.size = strlen(ColorStr)+1;
    if (from.size < sizeof(String))
      from.size = sizeof(String);
    to.addr = NULL;
    XtConvert (AppShell, XmRString, &from, XmRPixel, &to);
  
    if (to.addr != NULL)
      Result = *((Pixel*)to.addr);
  }

  PRO_END
  return Result;
}


/*----------------------------------------------------------------------
  Funktion: Ermittelt die aktuelle Mausposition
  
  INPUT:  MouseX, MouseY  Rckgabewerte des Mauszeigerkoordinaten
  OUTPUT: none
  
  Bsp.: int MouseX, MouseY;
        ...
	TOOLGetMousePos (&MouseX, &MouseY);
  ----------------------------------------------------------------------*/
Window TOOLGetMousePos (int *MouseX, int *MouseY)
{
           int NewX = 0, NewY = 0;
	   int PosX = 0, PosY = 0;
  unsigned int Mask;
  Window       RootWin, Win1, Win2;

  RootWin = RootWindow (XDisplay, XScreen);
  XQueryPointer (XDisplay, RootWin, &Win1, &Win2,
		 &PosX, &PosY, &NewX, &NewY, &Mask);

  if (MouseX)
    *MouseX = PosX;

  if (MouseY)
    *MouseY = PosY;

  return Win2;
}


/*----------------------------------------------------------------------
  Funktion: Erzeugt ein Pixmap als Icon fr einen verkleinertes 
            (iconisiertes) Dialogfenster
            Bildparameter: max 6 Farben (der Standard-Palette ?)
  
  INPUT:  Shell     Shell-Widget des Dialogfensters
          FileName  Name der XPM-Datei
  OUTPUT: ---
  
  Bsp.: TOOLSetIcon (Shell, "blabla", "./dlgicon.xpm", NULL);
  ----------------------------------------------------------------------*/
void TOOLSetIcon (Widget Shell, char *IconName, char *FileName, void *Data)
{
  Pixel  bg;
  Pixmap PixMap;

  PRO_START ("\n")

  if (! initialized)
    PRO_ABORT ("Base-Lib not initialized")

  if (FileName == NULL && Data == NULL)
    PRO_ABORT ("Invalid Data")

  XtVaGetValues (Shell, XmNbackground, &bg, NULL);

  if (FileName != NULL && FileName[0] != '\0')
    PixMap = TOOLLoadXPM (FileName, NULL, NULL, bg);
  else
    PixMap = TOOLMakeXPM ((char**)Data, NULL, NULL, bg);

  XtSetArg (args[0], XmNiconPixmap, PixMap);
  XtSetValues (Shell, args, 1);

  PRO1 ("Name: %s\n", IconName)
  XSetIconName (XDisplay, XtWindow(Shell), IconName);

  PRO_END
}


/*----------------------------------------------------------------------
  Function: Sets the shape of the mouse-cursor to a specified value
  
  INPUT:  Shell     shell-widget, which shows the given cursor
          CursorID  cursor-shape-identifier 
	            (see "/usr/include/X11/cursorfont.h" for detail)
  OUTPUT: ---
  
  Example: Widget Shell;
           ...
           TOOLSetIcon (Shell, XC_clock);
  ----------------------------------------------------------------------*/
void TOOLSetMouseCursor (Widget Shell, int CursorID, 
			 char *fgColor, char *bgColor)
{
  int    HotX = -1, HotY = -1;
  int    SizeX, SizeY;
  XColor fg, bg;
  Pixmap Source, Mask;

  PRO_START ("")

  if (Shell == NULL || ! XtIsManaged(Shell))
    PRO_ABORT ("Shell not managed")

  if (CursorID < ToolMAGNIFIER || 
      (CursorID > 0 && (CursorID % 2) == 1) || CursorID > 154)
    PRO_ABORT ("Invalid Cursor-ID")

  if (fgColor != NULL && fgColor[0] != '\0')
    XAllocNamedColor (XDisplay, CMap, fgColor, &fg, &fg);
  else
    XAllocNamedColor (XDisplay, CMap, "black", &fg, &fg);

  if (bgColor != NULL && bgColor[0] != '\0')
    XAllocNamedColor (XDisplay, CMap, bgColor, &bg, &bg);
  else
    XAllocNamedColor (XDisplay, CMap, "white", &bg, &bg);

  if (CursorID < 0)
  {
    switch (CursorID)
    {
      case ToolMAGNIFIER :
	SizeX = magnify_width;
	SizeY = magnify_height;
	HotX  = 5;
	HotY  = 5;
	Source = XCreatePixmapFromBitmapData (XDisplay, 
					      XDefaultRootWindow(XDisplay),
					      (char*)magnify_bits,
					      magnify_width, magnify_height,
					      0xffffff, 0x000000, 1);
	Mask = XCreatePixmapFromBitmapData (XDisplay, 
					    XDefaultRootWindow(XDisplay),
					    (char*)magnify_mask_bits,
					    magnify_mask_width, 
					    magnify_mask_height,
					    0xffffff, 0x000000, 1);
	break;
      default :
        PRO_ABORT ("Invalid CursorID")
    }

    if (HotX == -1)
      HotX = 0.5 * SizeX;
    if (HotY == -1)
      HotY = 0.5 * SizeY;

    if (MouseCursor != None)
      XFreeCursor (XDisplay, MouseCursor);

    if (Source != None)
    {
      MouseCursor = XCreatePixmapCursor (XDisplay, Source, Mask, 
					           &fg, &bg, HotX, HotY);
      XFreePixmap (XDisplay, Source);
    }
    if (Mask != None)
      XFreePixmap (XDisplay, Mask);
  }
  elsif (CursorID < ToolRESTORE)
  {
    if (MouseCursor != None)
      XFreeCursor (XDisplay, MouseCursor);

    MouseCursor = XCreateFontCursor (XDisplay, (unsigned int)CursorID);

    if ((fgColor != NULL && fgColor[0] != '\0') ||
	(bgColor != NULL && bgColor[0] != '\0'))
      XRecolorCursor (XDisplay, MouseCursor, &fg, &bg);
  }
  else
  {
    XUndefineCursor (XDisplay, XtWindow(Shell));
    return;
  }

  XDefineCursor (XDisplay, XtWindow(Shell), MouseCursor);

  XFlush (XDisplay);

  PRO_END
}


/*----------------------------------------------------------------------
  Funktion: Startet eine eigene Hauptschleife, welche beim Schlieen des 
            in Shell angegebenen Shell-Widgets wieder verlassen wird.

  INPUT:  AppCon  Motif-Application-Context in dem sich die Shell be-
                  findet
          Shell   Shell-Widget, welches die Schleife startet und auf
                  dessen Beendigung geachtet werden soll.
  OUTPUT: ---
  
  Bsp.: Widget Shell;
        ...
        TOOLMainLoop (AppCon, Shell);
  ----------------------------------------------------------------------*/
void TOOLMainLoop (XtAppContext AppCon, Widget Shell)
{
  XEvent Event;

  PRO_START ("\n")

  for (;;)
  {
    if (XtAppPeekEvent(AppCon, &Event))          /* is there an event ? */
    {
      XtAppNextEvent (AppCon, &Event);           /* yes: get it         */

      /* if event closes the application-window, then exit mainloop:    */
      if (Event.type == DestroyNotify &&
          Event.xdestroywindow.window == XtWindow(Shell))
        PRO_ABORT ("finished")
      else
        XtDispatchEvent (&Event);                /* evaluate event      */
    }
    else
      XtAppProcessEvent (AppCon, XtIMTimer);  
  }
  PRO_END
}


/* ********************************************************************** */
/* ** This function performs an active waiting-loop. A sleep-call      ** */
/* ** would stop the complete program. This wait routine, processes    ** */
/* ** all events (e.g. refreshing the screen). So, it does not blocks  ** */
/* ** the program.                                                     ** */
/* ** The Time will be checked on every event. If there are no events  ** */
/* ** pending, a timeout will finish the loop.                         ** */
/* ** WARNING: THIS FUNCTION DOES NOT WORK WHEN SWITCHING DAYS ON      ** */
/* ** MIDNIGHT !                                                       ** */
/* ** INPUT:  mSec   duration of the loop.                             ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
static void TOOLWaitTimer (void *Ptr, XtIntervalId Timer)
{
  /* Reset Ptr-Flag ? */
}

void TOOLWait (long uSec)
{
	 long int     Diff;
  struct timeval      Start, Time;
         XtIntervalId Timer = -1;
         XEvent       Event;
#if 0
	 long int     Sec;
  struct timeval      End;
#endif

  if (uSec == 0)
    return;

  if (uSec >= 1000)
    Timer = XtAppAddTimeOut (AppCon, uSec * 0.001, 
			     (XtTimerCallbackProc)TOOLWaitTimer, NULL);

  gettimeofday (&Start, NULL);

#if 0
  /* Compute End-Time */
  /* Diese Version, bei der die Rechenzeit in der Schleife minimiert wird, 
     funktioniert aus noch nicht geklrter Ursache nicht:
     Die Schleife bleibt gelegentliche stehen !!! */
  End.tv_usec = Start.tv_usec + uSec;
  Sec = End.tv_usec * 0.000001;
  End.tv_usec -= Sec * 1000000;
  End.tv_sec   = Start.tv_sec + Sec;

  printf ("Start: %ld : %ld\n", Start.tv_sec, Start.tv_usec);
  printf ("End:   %ld : %ld\n",   End.tv_sec,   End.tv_usec);
#endif

  do
  {
    if (XtAppPeekEvent (AppCon, &Event))         /* is there an event ? */
    {
      XtAppNextEvent (AppCon, &Event);           /* yes: get it         */
      XtDispatchEvent (&Event);                  /* evaluate event      */
    }
    else
      XtAppProcessEvent (AppCon, XtIMTimer);

    gettimeofday (&Time, NULL);
    Diff = (Time.tv_sec - Start.tv_sec) * 1000000 + 
           (Time.tv_usec - Start.tv_usec);
  } /*  while (Time.tv_sec < End.tv_sec || Time.tv_usec < End.tv_usec); */
    while (Diff < uSec);

  if (Timer >= 0)
    XtRemoveTimeOut (Timer);
}



void TOOLWaitForMapping (Display *Disp, Window Win)
{
  XEvent Event;

  while (1)
  {
    XNextEvent (Disp, &Event);
    if (Event.type == MapNotify && Event.xmap.window == Win)
      break;
  }
}



void TOOLGetScrValue (Display *Displ, int Screen, 
		      double x, double y, int *ScrX, int *ScrY)
{
  int SizeX, SizeY;

  if (ScrX != NULL)
  {
    if (x > 0.0)
    {
      /* Set size depending of screen size: */
      SizeX = DisplayWidth  (Displ, Screen);
      
      if (x > 0.0 && x <= 1.0)
	*ScrX = (int)(x * SizeX);
      else
	*ScrX = MIN ((int)FABS(x), SizeX);
    }
    else
      *ScrX = 0;
  }

  if (ScrY != NULL)
  {
    if (y > 0.0)
    {
      SizeY = DisplayHeight (Displ, Screen);

      if (y > 0.0 && y <= 1.0)
	*ScrY = (int)(y * SizeY);
      else
	*ScrY = MIN ((int)FABS(y), SizeY); 
    }
    else
      *ScrY = 0;
  }
}


void TOOLInitGCs (void)
{
  XGCValues GCVal;
#if 0
  dual   ok = TRUE;
  int    r, g, b;
  XColor color;

  XParseColor (XDispl, CMap, "Green", &color);
  XAllocColor (XDispl, CMap, &color);
#endif

  CMap = DefaultColormap (XDisplay, XScreen);
#if 0
    /* 
       allocate 216 standarized colors (CLUT), that means all combinations 
       of values 0, 51, 102, 153, 204, 255 (see c't-magazine 3/97 page 152)
    */
    CLUT = XAllocStandardColormap();    /* color look-up-table */
    color.pad = 0;
    color.flags = DoRed | DoGreen | DoBlue;

    r = 0;
    while (r < 5 && ok)
    {
      g = 0;
      while (g < 5 && ok)
      {
	b = 0;
	while (b < 5 && ok)
	{
	  color.red   = r*51 * 256;
	  color.green = g*51 * 256;
	  color.blue  = b*51 * 256;
	  if ((ok = XAllocColor(XDisplay, CMap, &color)))
	    CTable[CIndex++] = color.pixel;
	  b++;
        }
	g++;
      }
      r++;
    }
    if (! ok)
      fprintf (stderr, "Only %d colors can be allocated\n", CIndex);
#endif

  XScreen = DefaultScreen (XDisplay);

  defGC = DefaultGC (XDisplay, XScreen);
  GCVal.graphics_exposures = False;
  XChangeGC (XDisplay, defGC, GCGraphicsExposures, &GCVal);

  GCVal.foreground = TOOLGetColor ("Black");
  GCVal.background = TOOLGetColor ("grey");
  GCVal.font       = XLoadFont (XDisplay, "fixed");

  drawGC = XCreateGC (XDisplay, RootWindow(XDisplay, XScreen),
		      GCForeground | GCBackground | GCFont, &GCVal);
  XSetFunction (XDisplay, drawGC, GXset);

  xorGC  = XCreateGC (XDisplay, RootWindow(XDisplay, XScreen),
		      GCForeground | GCBackground | GCFont, &GCVal);
  XSetFunction (XDisplay, xorGC, GXxor);

  copyGC = XCreateGC (XDisplay, RootWindow(XDisplay, XScreen),
		      GCForeground | GCBackground | GCFont, &GCVal);
  XSetFunction (XDisplay, copyGC, GXcopy);
}


/* ********************************************************************** */
/* ** Initializes the TOOL-lib by creating the following (internal)    ** */
/* ** application-environment: App-shell, App-context, screen,         ** */
/* ** display, graphic-context and color-table.                        ** */
/* ** Call this function before parsing the program-parameters,        ** */
/* ** because XtAppInitialize (calls itself XtDisplayInitialize)       ** */
/* ** extracts the X-Windows relevant parameters and returns the       ** */
/* ** reduced list.                                                    ** */
/* ** INPUT:  Title  title of application (mostly argv[0])             ** */
/* **         argc   program-param-count                               ** */
/* **         argv   program parameters                                ** */
/* ** OUTPUT: ---                                                      ** */
/* ********************************************************************** */
void TOOLInitLib (char *Title, int *argc, char **argv, 
		  char **FallbackResources)
{
  char ClassName[128] = "Application";

  PRO_START ("")

  if (initialized)
    PRO_ABORT ("\\already initalized...")

  if (Title != NULL && Title[0] != '\0')
  {
    strncpy (ClassName, Title, 127);
    ClassName[0] = toupper (Title[0]);
  }
  elsif (*argc > 0 && argv[0] != '\0')
  {
    strncpy (ClassName, argv[0], 127);
    ClassName[0] = toupper (argv[0][0]);
  }

  ClassName[127] = '\0';
  PRO1 ("\\Title: %s...", ClassName)

  XtSetArg (args[0], XmNdeleteResponse, XmUNMAP);
  if ((AppShell = XtAppInitialize (&AppCon, ClassName, NULL, 0,
				   argc, argv, FallbackResources, 
				   args, 1)) == NULL)
    PRO_ABORT ("\\AppInitialize failed...")

  if ((XDisplay = XtDisplay(AppShell)) == NULL)
    XtErrorMsg ("OpenErr", "TOOLInitLib", "Error", 
		           "Opening display failed", NULL, NULL);

  XtAddCallback (AppShell, XmNpopupCallback,   
			            (XtCallbackProc)TOOL_StartCB, NULL);
  XtAddCallback (AppShell, XmNpopdownCallback, 
			            (XtCallbackProc)TOOL_ExitCB, NULL);

  XScreen = DefaultScreen (XDisplay);

  TOOLInitGCs ();
 
  initialized = TRUE;

  PRO_END
}


/* ****************************************************** */
/* ** Initializes the TOOL-lib by accepting the user's ** */
/* ** initialized application-environment.             ** */
/* ** GC and color-table will be initialized local     ** */
/* ** INPUT:  XDispl      Display initialized by user  ** */
/* **         AppContext  user's application context   ** */
/* **         AppShell           applictaion shell     ** */
/* **         Screen      valid screen                 ** */
/* ** OUTPUT: ---                                      ** */
/* ****************************************************** */
void TOOLSetLib (Display *XDispl,    XtAppContext AppContext, 
                 Widget   ApplShell, int          Screen)
{
  PRO_START ("")

  if (initialized)
    PRO_ABORT ("\\already initialized...")

  XDisplay = XDispl;
  AppCon   = AppContext;
  AppShell = ApplShell;
  XScreen  = Screen;

  TOOLInitGCs ();

  initialized = TRUE;

  PRO_END
}


/* ************************************************* */
/* ** Adds a function into the popdown-CB-list.   ** */
/* ** This function will be called before popping ** */
/* ** down the app-shell to give the user-program ** */
/* ** the possibility to free it's allocated      ** */
/* ** memory.                                     ** */
/* ** INPUT:  Function  user-(Exit-)function      ** */
/* **         Data      data passed to function   ** */
/* ** OUTPUT: ---                                 ** */
/* ************************************************* */
void TOOLAddExitCB (XtCallbackProc Function, void *Data)
{
  PRO_START ("")

  /* 
     add the CB-function and sort the CB-list so, that the final Callback 
     (which kills the program) will be called at last.
  */
  XtRemoveCallback (AppShell, XmNpopdownCallback, 
				          (XtCallbackProc)TOOL_ExitCB, NULL);
  XtAddCallback    (AppShell, XmNpopdownCallback, 
				          (XtCallbackProc)Function,    Data);
  XtAddCallback    (AppShell, XmNpopdownCallback, 
				          (XtCallbackProc)TOOL_ExitCB, NULL);

  PRO_END
}  


/* *************************************************** */
/* ** Removes an user's exit-function (added with   ** */
/* ** TOOLAddExitCB) from CB-list, if both          ** */
/* ** function-ptr and data-ptr found in the list   ** */
/* ** INPUT:  Function  user's exit-function        ** */
/* **         Data      specified data              ** */
/* ** OUTPUT: ---                                   ** */
/* *************************************************** */
void TOOLDelExitCB (XtCallbackProc Function, void *Data)
{
  PRO_START ("")

  XtRemoveCallback (AppShell, XmNpopdownCallback, Function, Data);

  PRO_END
}


static void TOOL_StartCB (Widget widget, void *Data, XmAnyCallbackStruct *Ptr)
{
  if (widget != NULL)
    running = TRUE;
}


static void TOOL_ExitCB (Widget widget, void *Data, XmAnyCallbackStruct *Ptr)
{
  if (widget != NULL)
  {
    XtDestroyWidget (AppShell);
    XFreeColormap (XDisplay, CMap);
    XCloseDisplay (XDisplay);
  }
  exit (0);
}


/* **************************************** */
/* ** Stops the application by poping it ** */
/* ** down (activate inserted callback-  ** */
/* ** procs including TOOL_ExitCB)       ** */
/* ** INPUT:  ---                        ** */
/* ** OUTPUT: ---                        ** */
/* **************************************** */
void TOOLExitLib (void)
{
  XtPopdown (AppShell);
}


/* ************************************** */
/* ** check, if library is initialized ** */
/* ** INPUT:  ---                      ** */
/* ** OUTPUT: ---                      ** */
/* ************************************** */    
dual TOOLIsInitialized (void)
{
  return initialized;
}


dual TOOLIsRunning (void)
{
  return running;
}

