/* ************************************ */
/* *** (c) Stefan Vietinghoff       *** */
/* *** helpful routines for icons   *** */
/* ************************************ */
#include <unistd.h>

#include "libtools.h"


/*----------------------------------------------------------------------
  Funktion: Extra Routine zum kopieren von Bitmap-Daten, da die DLG_strcp
            Routine durch die mglichen 0-Werte nicht funktioniert.
	    Da in der Regel die Gre des Bildes und nicht die Anzahl
	    der Datenwerte bekannt ist, wird aus dieser die Anzahl der
	    Bytes berechnet.
  
  INPUT:  Data          zu kopierende Bitmap-Daten
          SizeX, SizeY  Gre des Bildes 
  OUTPUT: (byte*)       Zeiger Kopie der Bilddaten
  
  Bsp.: char *xbm;
        ...
        xbm = TOOLCopyXBM (Daten, 20, 20);
  ----------------------------------------------------------------------*/
byte *TOOLCopyXBM (byte *Data, int SizeX, int SizeY)
{
  int   Number, count;
  byte *Result = NULL;
  
  if (Data == NULL) 
    return Result;

  /* Anzahl der Bytes bestimmen: */
  if ((SizeX % 8) == 0)
    Number = (int)((SizeX / 8) * SizeY);
  else
    Number = (int)(((SizeX / 8) + 1) * SizeY);

  Result = (byte*)calloc (Number, sizeof(byte));
  for (count = 0; count < Number; ++count)
    Result[count] = Data[count];

  return Result;
}


#if 0
INT TOOL_GetPicFormat (PCHAR Data)
{
  INT   count = 0;
  INT   Result = DLG_UNKNOWN;
  PCHAR Ptr = Data;

  PRO_START ("["__FUNCTION__"]...")

  if (Ptr == NULL)
    PRO_RETURN ("Invalid data-pointer", DLG_UNKNOWN)

  while (Result == TOOL_UNKNOWN && Ptr[count] != '\0')
  {
    PRO1 ("\\%c", Ptr[count])
    if (strncmp(Ptr+count, "#define", 7) == 0)
      Result = XBM;
    elsif (strncmp(Ptr+count, "static", 6) == 0)
      Result = XPM;
    else
      count++;
  }

  PRO_END
  return Result;
}
#endif


Pixmap TOOLMakeXBM (Display *XDisplay, char *Data, int SizeX, int SizeY,
   		                                   Pixel fg, Pixel bg)
{
  Pixmap Result;

  Result = XCreatePixmapFromBitmapData (XDisplay, DefaultRootWindow(XDisplay),
					Data, SizeX, SizeY, fg, bg,
					TOOLGetDepth());
  if (Result == None)
    Result = XmUNSPECIFIED_PIXMAP;

  return Result;
}


/*----------------------------------------------------------------------
  Funktion: Diese Funktion erzeugt ein leeres Bild mit den gegebenen 
            Ausmaen.
  
  INPUT:  Data     SizeX, SizeY  Gre des zu erzeugenden Bildes 
  OUTPUT: (byte*)  Zeiger Kopie der Bilddaten
  
  Bsp.: byte *xbm;
        ...
        xbm = TOOLMakeEmptyXBM (20, 20);
  ----------------------------------------------------------------------*/
byte *TOOLMakeEmptyXBM (int SizeX, int SizeY)
{
  int   Number;
  byte *Result = NULL;
  
  if ((SizeX % 8) == 0)
    Number = (int)((SizeX / 8) * SizeY);
  else
    Number = (int)(((SizeX / 8) + 1) * SizeY);

  Result = (byte*)calloc (Number, sizeof(byte));

  return Result;
}


/*----------------------------------------------------------------------
  Funktion: liest den Inhalt einer X-Bitmap (.xbm) Datei und gibt das
            ergebnis als Zeichenkette, sowie die Gre zurck.
	    Kann die Datei nicht gelesen werden/tritt ein Fehler auf,
	    so wird der NULL-Zeiger zurckgeliefert und die 
	    Grenangaben sind ungltig
  
  INPUT:  FileName  Name der zu lesenden Bitmap-Datei
  OUTPUT: (byte*)   Zeiger auf die Liste der gelesenen Bilddaten
          width     Zeiger auf die Breite des Bildes
	  height    Zeiger auf die Bildhhe

  Bsp.: int            XSize, YSize;
        unsigned char *PicData = NULL;
	...
	PicData = TOOLLoadXBMData ("/pic/stefan/rolf.bmp", &XSize, &YSize,
	                           NULL, NULL);
  ----------------------------------------------------------------------*/
unsigned char *TOOLLoadXBMData (char *FileName, int *SizeX, int *SizeY,
	      			                int *HotX, int *HotY)
{
  unsigned int   width, height;
  int            SpotX, SpotY;
  unsigned char *Result = NULL;

  PRO_START ("")

  XReadBitmapFileData (FileName, &width, &height, &Result, &SpotX, &SpotY);

  if (SizeX != NULL)
    *SizeX = width;
  if (SizeY != NULL)
    *SizeY = height;
  if (HotX != NULL)
    *HotX = SpotX;
  if (HotY != NULL)
    *HotY = SpotY;

  PRO_END
  return Result;
}


Pixmap TOOLLoadXBM (char *FileName, int *SizeX, int *SizeY,
		                    int *HotX, int *HotY)
{
  unsigned int width, height;
  int          SpotX, SpotY;
  Pixmap       Result = XmUNSPECIFIED_PIXMAP;

  XReadBitmapFile (XDisplay, XDefaultRootWindow(XDisplay), FileName, 
		   &width, &height, &Result, &SpotX, &SpotY);

  if (SizeX != NULL)
    *SizeX = width;
  if (SizeY != NULL)
    *SizeY = height;
  if (HotX != NULL)
    *HotX = SpotX;
  if (HotY != NULL)
    *HotY = SpotY;

  return Result;
}

/*----------------------------------------------------------------------
  Funktion: hellt ein Bitmap durch &-Verknpfung mit einem schachbrett-
            artigen Muster auf. Dabei wird nur jedes 2. Pixel sichtbar.
	    Diese Funktion dient dazu, Bilddaten fr die insensitive 
	    Darstellung in Bitmap-Buttons zu erzeugen.  
	    Der Benutzer ist fr die Freigabe des durch die erzeugten
	    Bilddaten belegten Speichers verantwortlich.

  INPUT:  PicData  Zeiger auf die Bilddaten (z.B. mit TOOLLoadXBMData geladen)
          width
	  height   Gre des gegebenen Bildes
  OUTPUT: (byte*)  aufgehellte Bilddaten
                   Zeigt PicData auf NULL oder tritt ein Fehler ein,
		   wird NULL zurckgeliefert
  
  Bsp.: byte *Result = NULL;
        byte *PicData;
	...
	Result = TOOLLightenXBM (PicData, 30, 40);
  ----------------------------------------------------------------------*/
byte *TOOLLightenXBM (byte *PicData, int width, int height)
{
  int   Row, Col;
  int   Number = 0, Elems;
  byte *Result = NULL;

  PRO_START ("")

  if (PicData == NULL)
    PRO_RETURN ("Invalid Pic-Data", NULL)
  
  /* Speicher fr Ergebnis allokieren: */
  Elems = ((width % 8) == 0)? (width/8):((width/8)+1);
  Result = (byte*) calloc ((Elems*height)+1, sizeof(char));

  /* abwechselnd die Bytes mit 170 und 85 UND-verknpfen: */
  for (Row = 0; Row < height; ++Row)
    for (Col = 0; Col < Elems; ++Col)
    {
      if (((Row+Col) % 2) == 0)
	Result[Number] = PicData[Number] & 170;  /* 10101010 */
      else
	Result[Number] = PicData[Number] & 85;   /* 01010101 */
      Number++;
    }

  PRO_END
  return Result;
}
