#include <unistd.h>
#include <ctype.h>

#include "toolPic.h"

#include "error.xpm"


char **TOOLCopyXPM (char **Data)
{
  int    count, Size = 0;
  int    h, w, c;
  char **Result;

  if (Data == NULL)
    return NULL;

  sscanf (Data[0], "%d %d %d", &w, &h, &c);

  PRO3 ("\\Data: %dx%d [%d]\n", w, h, c)

  for (count = 0; count < h+c+1; ++count)
    Size += strlen (Data[count]) + 1;

  Result = (char**)malloc (sizeof(char)*Size);
  memcpy (Result, Data, Size);

  return Result;
}


Pixmap TOOLxpm2xbm (char *Data, int SizeX, int SizeY, Drawable d)
{
  XImage *Image;
  Pixmap  Result = (Pixmap)0;
  GC      copyGC;

  PRO_START (" ")

  if (! TOOLIsInitialized())
    PRO_RETURN ("Base-Lib is not initialized", Result)

  copyGC = DefaultGC (XDisplay, XScreen);

  Image = XCreateImage (XDisplay, DefaultVisual(XDisplay, XScreen),
                        TOOLGetDepth(), ZPixmap, 0, 
			Data, SizeX, SizeY, TOOLGetDepth(), 0);

  Result = XCreatePixmap (XDisplay, d, SizeX, SizeY, TOOLGetDepth());
  XPutImage (XDisplay, Result, copyGC, Image, 0, 0, 0, 0, SizeX, SizeY);

  PRO_END
  return Result;
}


Pixmap TOOLLoadXPM (char *FileName, int *SizeX, int *SizeY, Pixel Color)
{
  int            width = 0, height = 0;
  Pixmap         Result = XmUNSPECIFIED_PIXMAP;
  Pixmap         Shape;
  XpmAttributes  Attr;
  XpmColorSymbol Symbol;

  PRO_START (" ")

  Symbol.name  = NULL;
  Symbol.value = "none";
  Symbol.pixel = Color;

  Attr.colorsymbols = &Symbol;
  Attr.numsymbols   = 1;
  Attr.valuemask    = XpmColorSymbols;

  if ((FileName != NULL && FileName[0] != '\0' &&
       XpmReadFileToPixmap(XDisplay, DefaultRootWindow(XDisplay),
			   FileName, &Result, &Shape, &Attr) == XpmSuccess) ||
      XpmCreatePixmapFromData(XDisplay, DefaultRootWindow(XDisplay),
			      error_xpm, &Result, &Shape, &Attr) != XpmSuccess)
  {
    width  = Attr.width;
    height = Attr.height;
    PRO2 ("\\[%d]x[%d]...", width, height)
  }
  else
  {
    PRO0 ("\\failed...")
    Result = XmUNSPECIFIED_PIXMAP;
  }

  /* XpmFreeAttributes ? */

  if (SizeX != NULL)
    *SizeX = width;
  if (SizeY != NULL)
    *SizeY = height;

  PRO_END
  return Result;
}


byte *TOOLLoadXPM_RGBA (char *FileName, double Alpha, int *SizeX, int *SizeY)
{
  unsigned int  count, Components = 3;
  unsigned int  r, g, b, a = 1;
  int           status;
  byte         *Result = NULL;
  XpmImage      Image;
  XpmInfo       Info;

  PRO_START (" ")

  Info.valuemask = 0;
  if ((status = XpmReadFileToXpmImage(FileName, &Image, &Info)) != XpmSuccess)
    PRO_RETURN1 ("XPM-Error %d", status, NULL)

  /* TODO:
     Prfen, ob width und Height == 2^n + 2*Border:
       fmod(log(Image.width)/log(2), 1) != 0.0
  */
  
  if (Alpha >= 0.0 && Alpha <= 1.0)
  {
    a          = (unsigned int)(Alpha * 255);
    Components = 4;
  }

  Result = (byte *)malloc (sizeof(byte) * 
			   Image.width * Image.height * Components);
  for (count = 0; count < Image.width * Image.height; ++count)
  {
    sscanf (Image.colorTable[Image.data[count]].c_color, "#%02x%02x%02x",
	    &r, &g, &b);
    Result[count*Components + 0] = (byte)r;
    Result[count*Components + 1] = (byte)g;
    Result[count*Components + 2] = (byte)b;
    if (Components == 4)
      Result[count*Components + 3] = (byte)a;
  }

  if (SizeX != NULL)
    *SizeX = Image.width;
  if (SizeY != NULL)
    *SizeY = Image.height;

  PRO_END
  return Result;
}


Pixmap TOOLMakeXPM (char **Data, int *SizeX, int *SizeY, Pixel Color)
{
  int           width = 0, height = 0;
  Pixmap        Result = XmUNSPECIFIED_PIXMAP;
  XpmAttributes Attr;
  XpmColorSymbol Symbol;

  PRO_START (" ")

  Symbol.name  = NULL;
  Symbol.value = "none";
  Symbol.pixel = Color;

  Attr.colorsymbols = &Symbol;
  Attr.numsymbols   = 1;
  Attr.valuemask    = XpmColorSymbols;

  #if 0
    Attr.valuemask = XpmVisual | XpmColormap | XpmDepth;
    Attr.visual   = DefaultVisual   (XDisplay, XScreen);
    Attr.colormap = DefaultColormap (XDisplay, XScreen);
    Attr.depth    = TOOLGetDepth ();
  #endif

  if ((Data != NULL && 
       XpmCreatePixmapFromData(XDisplay, DefaultRootWindow(XDisplay),
                               Data, &Result, NULL, &Attr)==XpmSuccess) ||
      XpmCreatePixmapFromData(XDisplay, DefaultRootWindow(XDisplay),
			      error_xpm, &Result, NULL, &Attr)==XpmSuccess)
  {
    width  = Attr.width;
    height = Attr.height;
    PRO2 ("\\[%d]x[%d]...", width, height)
  }
  else
  {
    PRO0 ("\\failed...")
    Result = XmUNSPECIFIED_PIXMAP;
  }

  if (SizeX != NULL)
    *SizeX = width;
  if (SizeY != NULL)
    *SizeY = height;

  PRO_END
  return Result;
}


XImage *TOOLLoadXPM2 (const char *FileName, 
		      int *SizeX, int *SizeY, Pixel Color)
{
  int             width = 0, height = 0;
  int             ID = 0;
  XImage         *Result = NULL;
  XImage         *Shape;
  XpmAttributes   Attr;
  XpmColorSymbol  Symbol;

  PRO_START (" ")
  PRO1 ("\\%s...", FileName)

  Symbol.name  = NULL;
  Symbol.value = "None";
  Symbol.pixel = Color;

  Attr.colorsymbols = &Symbol;
  Attr.numsymbols   = 1;
  Attr.valuemask    = XpmColorSymbols;

  if ((ID = XpmReadFileToImage(XDisplay, (char *)FileName, 
			       &Result, &Shape, &Attr)) == XpmSuccess)
  {
    width  = Attr.width;
    height = Attr.height;
  }
  else
  {
    PRO1 ("\\FAILED: %d...", ID)
    Result = NULL;
  }

  if (SizeX != NULL)
    *SizeX = width;
  if (SizeY != NULL)
    *SizeY = height;

  PRO_END
  return Result;
}


XImage *TOOLMakeXPM2 (char **Data, int *SizeX, int *SizeY, Pixel Color)
{
  int             width = 0, height = 0;
  int             ID = 0;
  XImage         *Result = NULL;
  XImage         *Shape;
  XpmAttributes   Attr;
  XpmColorSymbol  Symbol;

  PRO_START (" ")

  Symbol.name  = NULL;
  Symbol.value = "none";
  Symbol.pixel = Color;

  Attr.colorsymbols = &Symbol;
  Attr.numsymbols   = 1;
  Attr.valuemask    = XpmColorSymbols;

  if ((ID = XpmCreateImageFromData(XDisplay, Data, 
				   &Result, &Shape, &Attr)) == XpmSuccess)
  {
    width  = Attr.width;
    height = Attr.height;
  }
  else
  {
    PRO1 ("FAILED: %d...", ID)
    Result = NULL;
  }

  if (SizeX != NULL)
    *SizeX = width;
  if (SizeY != NULL)
    *SizeY = height;

  PRO_END
  return Result;
}


