#define USE_THREAD

#ifdef USE_THREAD
#include <pthread.h>
#endif

#include <ctype.h>

#include <libview.h>

#include "tuxCAD.h"

#define CLI_CHECK_WORLD \
        if ((World = GetCurrent()) == NULL && \
	    (World = SelectWorld(TRUE)) == NULL) \
        { \
          strcpy (Command, "No active workcell"); \
          return FAIL; \
        }


typedef int (*CLIFUNCTION)(char*, int);

typedef struct
{
  char         Keyword[50];
  CLIFUNCTION  Fkt;
  char         Msg[40];
} CLI_MACRO;



static int    CLI_Attach            (char *Command, int Offset);
static int    CLI_ClearAll          (char *Command, int Offset);
static int    CLI_Create            (char *Command, int Offset);
static int    CLI_Delete            (char *Command, int Offset);
static int    CLI_ExeGSLMacro       (char *Command, int Offset);
static int    CLI_Exit              (char *Command, int Offset);
static int    CLI_Fill              (char *Command, int Offset);
static int    CLI_GetDevName        (char *Command, int Offset);
static int    CLI_InquireParts      (char *Command, int Offset);
static int    CLI_InvokeButton      (char *Command, int Offset);
static int    CLI_LoadPart          (char *Command, int Offset);
static int    CLI_LoadDevice        (char *Command, int Offset);
static int    CLI_LoadWorkCell      (char *Command, int Offset);
static int    CLI_Open              (char *Command, int Offset);
static int    CLI_Place             (char *Command, int Offset);
static int    CLI_Rename            (char *Command, int Offset);
static int    CLI_Rotate            (char *Command, int Offset);
static int    CLI_SaveDevice        (char *Command, int Offset);
static int    CLI_SelectDev         (char *Command, int Offset);
static int    CLI_Set               (char *Command, int Offset);
static int    CLI_TransferToMenu    (char *Command, int Offset);
static int    CLI_Translate         (char *Command, int Offset);
static int    CLI_UpdateGraphics    (char *Command, int Offset);

static int    CLI_Robot             (char *Command, int Offset);

static int    CLI_SelectFace        ();


/* List of simulated CLI-Macros */
static CLI_MACRO Macro[] =
{
  {"ATTACH DEVICE",            CLI_Attach,           "Attach Device"},
  {"ATTACH",                   CLI_Attach,           "Attach"},
  {"CLEAR ALL",                CLI_ClearAll,         "Clear all"},
  {"CLOSE POLYGON",            NULL,                 "Close Polygon"}, //
  {"CREATE PATH",              NULL,                 "Create Path"},
  {"CREATE SPHERE",            NULL,                 "Create Sphere"}, // 
  {"CREATE TAG",               NULL,                 "Create Tag"},
  {"CREATE",                   CLI_Create,           "Create ..."},
  {"DELETE PATH",              NULL,                 "Delete Path"},
  {"DELETE TAG",               NULL,                 "Delete Tag"},
  {"DELETE",                   CLI_Delete,           "Delete Part/Device ..."},
  {"EXECUTE",                  CLI_ExeGSLMacro,      "Exe GSL-Macro"},
  {"EXIT",                     CLI_Exit,             "Exit tuxCAD"},
  {"FILL",                     CLI_Fill,             "Fill"},
  {"INQUIRE ALL DEVICE NAMES", CLI_GetDevName,       "Get Device-Name"},
  {"INQUIRE_FILENAME",         NULL,                 "Get FileName"},
  {"INQUIRE",                  CLI_InquireParts,     "Get Parts"},
  {"INVOKE MENU BUTTON",       CLI_InvokeButton,     "Invoke Button"},
  {"MOVE ALONG",               NULL,                 "Move along Path"},
  {"MULTI VIEWS",              NULL,                 "Multi Views"},
  {"OPEN PART",                CLI_Open,             "Open Part"},
  {"OPEN POLYGON",             NULL,                 "Open Polygon"}, //
  {"PLACE TAG",                NULL,                 "Place Tag"},
  {"PLACE",                    CLI_Place,            "Place Part/Device"},
  {"POP",                      NULL,                 "Pop"},
  {"PUSH",                     NULL,                 "Push"},
  {"RENAME",                   CLI_Rename,           "Rename Device"},
  {"RESTORE",                  NULL,                 "Restore"},
  {"RETRIEVE PART",            CLI_LoadPart,         "Load Part"},
  {"RETRIEVE DEVICE",          CLI_LoadDevice,       "Load Device"},
  {"RETRIEVE WORKCELL",        CLI_LoadWorkCell,     "Retrieve Workcell"},
  {"ROTATE",                   CLI_Rotate,           "Rotate ..."}, //
  {"SAVE",                     CLI_SaveDevice,       "Save Device"}, 
  {"SET",                      CLI_Set,              "Set ..."},
  {"TRANSFER TO MENU",         CLI_TransferToMenu,   "Show tuxCAD-Menu"},
  {"TRANSLATE",                CLI_Translate,        "Translate ..."},
  {"UPDATE GRAPHICS",          CLI_UpdateGraphics,   "Update Graphics"},

  /* Additional commands used from amp: */
  {"ACTIVATE VIEW",            NULL,                 "Activate View"},
  {"MESSAGE MODE",             NULL,                 "Set Message mode"},
  {"RESET ALL",                NULL,                 "Reset all"},
  {"SCALE WORLD",              NULL,                 "Scale World"},
  // {"SET UNITS",             NULL,                 "SET UNITS"},
  {"STARTMSG",                 NULL,                 "Start Message"},

  /* Commands added for igrip-simulation */
  {"ROBOT",                    CLI_Robot,            "Simulate Robot"},
  {"SELECT",                   CLI_SelectDev,        "Select ..."}
};

#define CLI_MACROS (int)(sizeof(Macro)/sizeof(CLI_MACRO))


typedef struct
{
  char     DevName[DEV_NAME_LEN+1];
  char     PartName[PART_NAME_LEN+1];
  char     FileName[PATH_MAX+1];
  int      Mode;
  int      Face;
  POINT3D  Pos;
} FEATURE_DATA;

typedef struct
{
  PART       *Part;
  char        PartName[PART_NAME_LEN+1];
  char        DevName[DEV_NAME_LEN+1];
} CREATE_DATA;


static char          Msg[128] = "0: OK";
static int           Number = 0;
static char          DirMask[PATH_MAX+1] = CAD_PATH"PARTS/*.part"; /* */
static CREATE_DATA   Create = {NULL, "",  ""};
static FEATURE_DATA  Feature = {"Node", "Pendel.part", 
			        CAD_PATH"PARTS/pendel/pendel.part",
			        0, 0, 
			        {0.0, 0.0, 0.0}};
static bool          AutoRefresh = TRUE;
       SOCKET_DATA   Server = {-1, 0, NULL};
#ifdef USE_THREAD
static pthread_t *NetThread = NULL;
#endif


static void CLI_SkipSpace (char *Message, int &Pos)
{
  if (Message == NULL)
    return;

  while (isspace(Message[Pos]))
    Pos++;
}


static void CLI_SkipKeyWord (char *Message, int &Pos, char *KeyWord)
{
  int len;

  if (Message == NULL)
    return;

  len = strlen (KeyWord);
  while (Message[Pos] != '\0' && strncmp(Message+Pos, KeyWord, len) != 0)
    Pos++;

  Pos += len;
}


static void GetFileName (char *Command, char *FileName)
{
  int count1 = 0, count2 = 0;
  int len;

  len = strlen (Command);
  while (count1 < len && Command[count1] != '\'')
    count1++;
  count1++;

  while (count1 < len && Command[count1] != '\'')
    FileName[count2++] = Command[count1++];
  FileName[count2] = '\0';
}


static int ParseName (char *Command, char *DevName, char *PartName)
{
  int count = 0, len;
  int count1 = 0, count2 = 0;

  len = strlen (Command);
  CLI_SkipSpace(Command, count);

  if (strncasecmp(Command+count, "DEVICE", 6) == 0)
    count += 7;
  elsif (strncasecmp(Command+count, "PART", 4) == 0)
    count += 5;

  if (Command[count] == '\'')
    count++;

  while (count < len && Command[count] != ':' && Command[count] != '\'' &&
	 Command[count] != ' ' && Command[count] != '\0')
    DevName[count1++] = Command[count++];
  DevName[count1] = '\0';

  if (Command[count] == ':')
  {
    count++;
    while (count < len && Command[count] != '\'' && 
	   Command[count] != ' ' && Command[count] != '\0')
      PartName[count2++] = Command[count++];
    PartName[count2] = '\0';
  }
  else
    PartName[0] = '\0';

  if (Command[count] == '\'')
    count++;

  return count;
}


static int WritePos (DEVICE3D *Dev, PART3D *Part)
{
  double  rpy[6] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
  FILE   *fp = NULL;

  if ((fp = fopen("/tmp/gsl.dat", "w+")) != NULL)
  {
    if (Dev == NULL)
      fprintf (fp, "Dummy");
    else
      fprintf (fp, "%s", Dev->GetName());

    fprintf (fp, ":");

    if (Part == NULL)
      fprintf (fp, "Dummy");
    else
    {
      fprintf (fp, "%s", Part->GetPart()->PartName);
      GEOFrame2RPY (Part->GetWTP(),
		    &rpy[0], &rpy[1], &rpy[2], &rpy[3], &rpy[4], &rpy[5]);
      PRO3 ("yaw: %f  pitch: %f  roll: %f\n", rpy[3], rpy[4], rpy[5])
    }
    fprintf (fp, "\n");

    fprintf (fp, "%.4f %.4f %.4f  %.4f %.4f %.4f\n", 
   	         rpy[0], rpy[1], rpy[2], rpy[5], rpy[4], rpy[3]);
    fclose (fp);
  }
  return OK;
}


static char *ExeCLI (int SocketID, char *Command)
{
  int count = 0;
  int Result = 0;

  assert (SocketID > -1);

  if (Command == NULL || Command[0] == '\0')
  {
    PRO0 ("Empty Command received")
    return "OK";
  }

  PRO1 ("Cmd: %s\n", Command)
  PRGPrintf ("Cmd: %s", Command);

  // TODO:
  // Schreibe den Befehl in einer Pipe und warte auf eine Rckmeldung.
  // (Shared-Memory-Verfahren scheint Probleme zu verursachen)
  // pthread_cond_init und pthread_cond_wait bis Rueckgabewert

  while (count < CLI_MACROS && strncasecmp(Command, Macro[count].Keyword, 
					   strlen(Macro[count].Keyword)) != 0)
    count++;

  if (count == CLI_MACROS)
  {
    PRO0 ("??????????? UNKNOWN COMMAND ???????????? \n")
    sprintf (Msg, "1: ERROR");
  }
  elsif (Macro[count].Fkt != NULL)
  {
    Result = Macro[count].Fkt (Command, strlen(Macro[count].Keyword));
    switch (Result)
    {
      case -1 :
	Msg[0] = '\0';
	PRO0 ("EXIT")
        return NULL;
	break;

      case 0 :
	sprintf (Msg, "%d: OK", Result);
	break;

      default :
	sprintf (Msg, "%d: %s", Result, Command);
    }  
  }
  else
  {
    PRO1 ("!!!!!!!!!!!! MISSING FUNCTION: %s\n", Macro[count].Msg)
    strcpy (Msg, "0: OK");
  }
 
  PRO1 ("CLI-RETURN: %s\n", Msg)

  return Msg;
}


bool SplitName (char *Name, char *DevName, char *PartName)
{
  int count = 0, index = 0;
  int len;

  if (Name[0] == '\'')
    count++;

  PRO2 ("DEVICE: '%s'  PART: '%s'\n", DevName, PartName)

  len = strlen (Name);
  while (count < len && Name[count] != ':')
    DevName[index++] = Name[count++];
  DevName[index] = '\0';

  index = 0;
  while (count < len && Name[count] != '\'')
    PartName[index++] = Name[count++];
  PartName[index] = '\0';

  return TRUE;
}


bool RunCLIServer (char *HostName, int PortNo)
{
  bool Result = FALSE;

  assert (HostName != NULL || PortNo > 0);

  if (PortNo > 0 && (Server.ID = NETInitServer(HostName, PortNo)) < 0)
    return FALSE;

  PRGPrintf ("Running CLI-server...");

  if (Server.ID >= 0)
#ifdef USE_THREAD
    Result = ((NetThread = NETStartServer(Server.ID, ExeCLI, 1)) != NULL);
#else
    Result = (NETRunServer(Server.ID, ExeCLI, 1) >= 0);
#endif

  PRGPrintf ("\\%s", Result ? "running !" : "failed !");

  return Result;
}


void StartCLIServer ()
{
  RunCLIServer("", 2030);
}


void CloseCLIServer ()
{
  if (Server.PortNo > 0 && Server.ID >= 0)
    NETCloseServer (Server.ID);

  Server.ID = -1;
}


dual GetGSLName (char *Command, char *Result)
{
  int count1, count2=0;
  int len;

  Result[0] = '\0';

  if (Command[0] == '\0')
    return FALSE;

  len = strlen (Command);
  count1 = len - 2;
  while (count1 > 0 && Command[count1] != '/' && Command[count1] != '\'')
    count1--;
  count1++;

  while (count1 < len && Command[count1] != '\'')
    Result[count2++] = Command[count1++];
  Result[count2] = '\0';

  return TRUE;
}


static void GetNameOfFile (char *FileSpec, char *FileName)
{
  int count1, count2 = 0;
  int len;

  len = strlen (FileSpec);
  count1 = len-1;
  while (count1 > 0 && FileSpec[count1] != '/')
    count1--;

  if (FileSpec[count1] == '/')
    count1++;

  while (count1 < len)
    FileName[count2++] = FileSpec[count1++];
  FileName[count2] = '\0';
}


static int CheckMode (DIALOG *Dlg, int BoxID)
{
  int Mode;

  Mode = DLGGetOMenu (Dlg, BoxID);
  DLGSetBoxActive (Dlg, 22, (Mode == 0));
  DLGSetBoxActive (Dlg, 30, (Mode == 1));

  return DlgCONFIRM;
}


static int SetName (DIALOG *Dlg, int BoxID)
{
  assert (BoxID > -1);

  char FileSpec[PATH_MAX+1];
  char FileName[PATH_MAX+1];

  strcpy (FileSpec, DLGGetFileSel(Dlg, 14));
  GetNameOfFile (FileSpec, FileName);
  DLGSetText (Dlg, 13, FileName);

  return DlgCONFIRM;
}

// **************************************************************************

static int CLI_ExeGSLMacro (char *Command, int Offset)
{ 
  int       Button = 1;
  double    x, y, z;
  double    rpy[6];
  char      GSLName[SOCKET_BUF_SIZE+1];
  char      Name[SOCKET_BUF_SIZE+1];
  char      DevName[DEV_NAME_LEN+1];
  char      PartName[PART_NAME_LEN+1];
  FILE     *fp;
  PART3D   *Part = NULL;
  DEVICE3D *Dev = NULL;
  WORLD    *World = GetCurrent ();

  if (! GetGSLName(Command+Offset, GSLName))
    PRO_RETURN ("Invalid GSL-Macro-Name", 0)
    
  PRO1 ("GSL-Name: %s\n", GSLName)

  if (strncmp(GSLName, "face.gsl", 8) == 0)
    if (World != NULL)
    {
      Part = World->PickPart (&x, &y, &z);
      World->Select (Part);
      if ((fp = fopen("/tmp/gsl.dat", "w+")) != NULL)
      {
	fprintf (fp, "%d\n", Button);
	if (Part != NULL)
	{
          GEOFrame2RPY (Part->GetWTP(),
			&rpy[0], &rpy[1], &rpy[2], &rpy[3], &rpy[4], &rpy[5]);
	  PRO3 ("yaw: %f  pitch: %f  roll: %f\n", rpy[3], rpy[4], rpy[5])
	  fprintf (fp, "%s:%s\n", 
		       Part->GetParent()->GetName(), Part->GetName());
	  fprintf (fp, "%s\n",    Part->GetPart()->FileName);
	  fprintf (fp, "%f %f %f\n", x, y, z);
	  fprintf (fp, "%.4f %.4f %.4f  %.4f %.4f %.4f\n", 
		       rpy[0], rpy[1], rpy[2], rpy[5], rpy[4], rpy[3]);
	}
	fclose (fp);
      }
    }
    else
      return CLI_SelectFace ();
  elsif (strncasecmp(GSLName, "WORLDPOS", 8) == 0)
  {
    if (World != NULL && (fp = fopen("/tmp/gsl.dat", "r+")) != NULL)
    {
      fscanf (fp, "%s", Name);
      ParseName (Name, DevName, PartName);
      PRO2 ("DEV: '%s'  PART: '%s' \n", DevName, PartName)
      if (DevName[0] != '\0')
	Dev = World->GetDev (DevName);
      if (PartName[0] != '\0')
	Part = World->GetPart (DevName, PartName);
    }
    return WritePos (Dev, Part);
  }
  elsif (strncasecmp(GSLName, "GETPOS", 6) == 0)
  {
    PRO0 ("TODO: GETPOS\n")
    return WritePos (NULL, NULL);
  }

  return OK;
}


static int CLI_Attach (char *Command, int Offset)
{
  char      DevName[DEV_NAME_LEN+1], PartName[PART_NAME_LEN+1];
  PART3D   *Part = NULL;
  DEVICE3D *Dev = NULL;
  WORLD    *World;

  CLI_CHECK_WORLD

  Offset += ParseName (Command+Offset, DevName, PartName);

  if ((Dev = World->GetDev(DevName)) == NULL)
  {
    PRO1 ("Device '%s' not found\n", DevName)
    sprintf (Command, "1: Device not found '%s'", DevName);
    return FAIL;
  }

  while (Command[Offset] != '\0' && strncmp(Command+Offset, "AT", 2) != 0)
    Offset++;
  Offset += 2;

  ParseName (Command+Offset, DevName, PartName);

  PRO3 ("ATTACH '%s' -> '%s':'%s'\n", Dev->GetName(), DevName, PartName)

  if ((Part = World->GetPart(DevName, PartName)) != NULL)
  {
    World->Remove (Dev);
    Part->AttachDev   (Dev);
  }

  return OK;
}


static int CLI_ClearAll (char *Command, int Offset)
{
  DeleteWorld ();

  return OK;
}


static int CLI_Create (char *Command, int Offset)
{
  double    Param[6];
  DEVICE3D *Dev;
  WORLD    *World;

  CLI_CHECK_WORLD

  CLI_SkipSpace (Command, Offset);

  if (strncasecmp(Command+Offset, "CONE", 4) == 0)
  {
    Offset += 5;
    sscanf (Command+Offset, "%lf, %lf, %lf, %lf, %lf",
	    &Param[0], &Param[1], &Param[2], &Param[3], &Param[4]);
    PRO5 ("CREATE CONE <%.1f %.1f %.1f %.1f %.1f>\n",
	  Param[0], Param[1], Param[2], Param[3], Param[4])
    if ((Create.Part = PARTCreateCone(Create.PartName, Matrix_1,
				      0.0, Param[0], Param[1],
				      Param[2], Param[3], 
				      (int)Param[4], -1)) != NULL)
    {
      if ((Dev = World->GetDev(Create.DevName)) == NULL)
      {
	PRO0 ("NEW\n")
	Dev = new DEVICE3D (Create.DevName, new PART3D(Create.Part));
	World->AddDevice (Dev);
      }
      else
      {
	PRO0 ("APPEND\n")
	Dev->AddPart (NULL, new PART3D(Create.Part));
      }

      Create.Part = NULL;
    }
  }

  return OK; 
}


static int CLI_Delete (char *Command, int Offset)
{
  char      DevName[DEV_NAME_LEN+1], PartName[PART_NAME_LEN+1];
  PART3D   *Part;
  DEVICE3D *Dev;
  WORLD    *World;

  CLI_CHECK_WORLD

  Offset += ParseName (Command+Offset, DevName, PartName);

  if (PartName[0] == '\0')
  {
    if ((Dev = World->GetDev(DevName)) != NULL)
      World->Delete (Dev);
  }
  elsif ((Part = World->GetPart(DevName, PartName)) != NULL)
    World->Delete (Part);

  PRO2 ("DELETE '%s':'%s'\n", DevName, PartName)

  if (AutoRefresh)
    World->Redraw ();

  return OK;
}


static int CLI_Fill (char *Command, int Offset)
{
  WORLD *World;

  PRO_START ("")

  CLI_CHECK_WORLD

  if (World != NULL)
    printf ("CURRWORLD: %s\n", World->GetName());
  else
    printf ("NO World active\n");

  World->Fill ();

  PRO_END
  return OK;
}


static int CLI_Exit (char *Command, int Offset)
{
  PRGExitProgram ();

  return -1;
}


static int CLI_GetDevName (char *Command, int Offset)
{
  sprintf (Command, "0: 1, Dummy;");

  return OK;
}


static int CLI_InquireParts (char *Command, int Offset)
{
  char Name[40];

  sscanf (Command+Offset, "'%s'", Name);  

  sprintf (Command, "0: 1, %s;", Name);

  return OK;
}


static int CLI_InvokeButton (char *Command, int Offset)
{
  char Msg[80];

  sprintf (Msg, "Button %s invoked", Command+Offset);

  DLGOpenTextDialog ("invokeButton", DlgMESSAGE, Msg, 1, "OK");

  return OK;
}


static int CLI_LoadPart (char *Command, int Offset)
{
  char      FileName[PATH_MAX+1];
  DEVICE3D *Dev;
  WORLD    *World;

  CLI_CHECK_WORLD

  GetFileName (Command+Offset, FileName);
  PRO1 ("LOAD PART: '%s'\n", FileName)

  if ((Dev = World->LoadPart(FileName)) == NULL)
    return FAIL;

  World->Select (Dev);

  return OK;
}


static int CLI_LoadDevice (char *Command, int Offset)
{
  char   FileName[PATH_MAX+1];
  WORLD *World;

  CLI_CHECK_WORLD

  GetFileName (Command+Offset, FileName);
  PRO1 ("LOAD DEVICE: '%s'\n", FileName)

  if (! ImportDevice(World, FileName))
    return FAIL;

  if (AutoRefresh)
    World->Redraw ();

  return OK;
}

static int CLI_Open (char *Command, int Offset)
{
  Offset += ParseName (Command+Offset, Create.DevName, Create.PartName);

  PRO2 ("OPEN PART '%s':'%s'\n", Create.DevName, Create.PartName)
  Create.Part = NULL;

  return OK;
}


static int CLI_Place (char *Command, int Offset)
{
  double    RPY[6];
  char      DevName[DEV_NAME_LEN+1], PartName[PART_NAME_LEN+1];
  H_MATRIX  M;
  PART3D   *Part;
  DEVICE3D *Dev;
  WORLD    *World;

  CLI_CHECK_WORLD

  Offset += ParseName (Command+Offset, DevName, PartName);

  Offset += 10;
  sscanf (Command+Offset, "%lf, %lf, %lf, %lf, %lf, %lf",
	  &RPY[0], &RPY[1], &RPY[2], &RPY[3], &RPY[4], &RPY[5]);
  M = GEORPY2Frame (RPY[0], RPY[1], RPY[2], RPY[3], RPY[4], RPY[5]);

  if (PartName[0] == '\0')
  {
    if ((Dev = World->GetDev(DevName)) != NULL)
      Dev->UpdatePos (M); // TODO: UpdatePos ist falsch -> Place 
    else
      PRO1 ("Device '%s' not found\n", DevName)
  }
  else
  {
    if ((Part = World->GetPart(DevName, PartName)) != NULL)
      Part->SetRel (M);
    else
      PRO2 ("Part '%s':'%s' not found\n", DevName, PartName)
  }

  PRO2 ("PLACE '%s':'%s' AT ", DevName, PartName)
  PRO6 ("\\%.1f %.1f %.1f %.1f %.1f %.1f\n", 
	RPY[0], RPY[1], RPY[2], RPY[3], RPY[4], RPY[5]) 

  if (AutoRefresh)
    World->Redraw ();

  return OK;
}


static int CLI_Rename (char *Command, int Offset)
{
  char              DevName[DEV_NAME_LEN+1], PartName[PART_NAME_LEN+1];
  DEVICE3D         *Dev;
  WORLD            *World = GetCurrent ();
  LIST<DEVICE3D *> *List = GetDevBuffer (World);

  CLI_CHECK_WORLD

  Offset += ParseName (Command+Offset, DevName, PartName);
  if (strncmp(DevName, "CD", 2) == 0 && List != NULL && List->Size() == 1)
    Dev = List->Get (0);
  else 
    Dev = World->GetDev (DevName);

  if (Dev == NULL)
  {
    PRO1 ("Device '%s' not found\n", DevName)
    sprintf (Command, "1: Device not found '%s'", DevName);
    return FAIL;
  }

#if 0
  while (Command[Offset] != '\0' && strncmp(Command+Offset, "TO", 2) != 0)
    Offset++;
  Offset += 2;
#else
  CLI_SkipKeyWord (Command, Offset, "TO");
#endif

  ParseName (Command+Offset, DevName, PartName);

  PRO2 ("RENAME '%s' -> '%s'\n", Dev->GetName(), DevName)
  Dev->SetName (DevName);

  return OK;
}


static int CLI_Rotate (char *Command, int Offset)
{
  int       count, No = 1;
  char      DevName[DEV_NAME_LEN+1], PartName[PART_NAME_LEN+1];
  char      Axis[20];
  double    Val, Rot[3] = {0.0, 0.0, 0.0};
  PART3D   *Part;
  DEVICE3D *Dev;
  WORLD    *World;

  CLI_CHECK_WORLD

  Offset += ParseName (Command+Offset, DevName, PartName);
  CLI_SkipSpace (Command, Offset);

  if (strncasecmp(Command+Offset, "ABOUT", 5) == 0)
    sscanf (Command+Offset+6, "%s BY %lf ", Axis, &Val);
  elsif (strncasecmp(Command+Offset, "TO", 2) == 0)
    sscanf (Command+Offset+3, "%lf, %lf, %lf", &Rot[0], &Rot[1], &Rot[2]);
#if 0
  CLI_SkipKeyWord (Command, Offset, "IN ");
#endif

  if (PartName[0] == '\0')
  {
    Dev = World->GetDev (DevName);
    Part = Dev->GetPartList();
  }
  else
    Part = World->GetPart (DevName, PartName);

  if (Part != NULL)
  {
    PRO5 ("ROTATE '%s':'%s': %.3f %.3f %.3f\n",
	  DevName, PartName, Rot[0], Rot[1], Rot[2]);

    for (count = 1; count <= No; ++count)
    {
      Part->SetVal (ROT_X, Part->GetVal(ROT_X) + (Rot[0] / No) * count);
      Part->SetVal (ROT_Y, Part->GetVal(ROT_Y) + (Rot[1] / No) * count);
      Part->SetVal (ROT_Z, Part->GetVal(ROT_Z) + (Rot[2] / No) * count);

      if (AutoRefresh)
	World->Redraw ();
    }
  }
  else
    PRO0 ("TODO: ROTATE DEV\n")

  return OK;
}


static int CLI_LoadWorkCell (char *Command, int Offset)
{
  WORLD *World = GetCurrent ();

  char FileName[PATH_MAX+1];

  GetFileName (Command+Offset, FileName);
  PRO1 ("LOAD WORLD: '%s'\n", FileName)

  if (World == NULL)
    NewWorld (FileName);

  World->LoadWorld (FileName);

  return OK;
}


static int CLI_SaveDevice (char *Command, int Offset)
{
  char FileName[PATH_MAX+1];
  char DevName[DEV_NAME_LEN+1];

  sscanf (Command+Offset, "%s AS '%s'", DevName, FileName);
  PRO1 ("DEVNAME: %s\n", DevName)
  PRO1 ("FILENAME: %s\n", FileName);

  return 0;
}


static int CLI_SelectFace ()
{
  dual      Error;
  int       Button;
  FILE     *fp = NULL;
  POINT3D   Pos = Vector_0;
  POLYGON  *Polygon;
  PART     *Part;
  DIALOG   *Dlg = NULL;

  do
  {
    Error = FALSE;

    Dlg = DLGNewDialog (0);
    DLGMakeCols (Dlg,  0, 2,              DlgETCHED_IN, 10, 20);
    DLGMakeRows (Dlg, 10, 4,                DlgNO_LINE, 11, 12, 13, 14);
    DLGMakeRows (Dlg, 20, 3, DlgETCHED_IN DlgETCHED_IN, 21, 22, 30);
    DLGMakeRows (Dlg, 30, 3,                DlgNO_LINE, 31, 32, 33);

    DLGInsertText    (Dlg, 12, "devName",  40, Feature.DevName, 16, 1);
    DLGInsertText    (Dlg, 13, "partName", 40, Feature.PartName, 16, 1);
    DLGInsertFileSel (Dlg, 14, "fileSel", DirMask, 20, Feature.FileName);
    DLGSetFunction  (Dlg, 14, SetName);
    if (Number > 0)
    {
      DLGSetBoxActive (Dlg, 13, FALSE);
      DLGSetBoxActive (Dlg, 14, FALSE);
    }

    DLGSetTitle     (Dlg, 20, "Position", DlgMIDDLE);
    DLGInsertOMenu  (Dlg, 21, "mode", 2, "FACE", "POS", Feature.Mode);
    DLGInsertInt    (Dlg, 22, "face", Feature.Face, 4);
    DLGInsertDouble (Dlg, 31, "xPosition", Feature.Pos.x, 8, 5);
    DLGInsertDouble (Dlg, 32, "yPosition", Feature.Pos.y, 8, 5);  
    DLGInsertDouble (Dlg, 33, "zPosition", Feature.Pos.z, 8, 5);
    DLGSetBoxActive (Dlg, 22, (Feature.Mode == 0));
    DLGSetBoxActive (Dlg, 30, (Feature.Mode == 1));
    DLGSetFunction  (Dlg, 21, CheckMode);

    DLGMakeButton (Dlg, OK_BUTTON);
    DLGMakeButton (Dlg, USER_BUTTON, "abort", DlgBUTTON_RIGHT, NULL, 2);
    DLGMakeButton (Dlg, CANCEL_BUTTON);

    DLGOpenDialog (Dlg, "defineFeature", DlgNORMAL, NULL);

    Button = DLGGetExitButton (Dlg);
    if (Button == 1)
    {
      strcpy (Feature.DevName,  DLGGetText(Dlg, 12));
      strcpy (Feature.PartName, DLGGetText(Dlg, 13));
      strcpy (Feature.FileName, DLGGetFileSel(Dlg, 14));
      TOOLGetFileDir (Feature.FileName, DirMask, PATH_MAX);
      strcat (DirMask, "/*.part"); /* */

      Feature.Mode  = DLGGetOMenu  (Dlg, 21);
      Feature.Face  = DLGGetInt    (Dlg, 22);
      Feature.Pos.x = DLGGetDouble (Dlg, 31);
      Feature.Pos.y = DLGGetDouble (Dlg, 32);
      Feature.Pos.z = DLGGetDouble (Dlg, 33);
      if (Feature.Mode == 0)
	if ((Part = PARTLoad(Feature.FileName, 
			     Feature.PartName, FALSE)) != NULL)
	{
	  PARTGetPolygon (Part, Feature.Face, NULL, &Polygon);
	  Pos = Polygon->Info.Center;
	  PARTFreePart (Part);
	}
	else
	{
	  Button = 2;
	  Error = (DLGOpenTextDialog("badPart", DlgWARNING, TOOLGetErrorMsg(),
				                2, "RETRY", "ABORT") == 1);
	}
      else
	Pos = Feature.Pos;
    }

    if ((fp = fopen("/tmp/gsl.dat", "w+")) != NULL)
    {
      fprintf (fp, "%d\n", Button);
      if (Button == 1)
      {
	fprintf (fp, "%s:%s\n",     Feature.DevName, Feature.PartName);
	fprintf (fp, "%s\n",        Feature.FileName);
	fprintf (fp, "%f  %f %f\n", Pos.x, Pos.y, Pos.z);
	fprintf (fp, "0.0  0.0  0.0  0.0  0.0  0.0\n");
      }
      fclose (fp);
    }

    DLGFreeDialog (Dlg);
  } while (Error);

  return OK;
}


static int CLI_Set (char *Command, int Offset)
{
  bool      Mode;
  int       count;
  char      DevName[DEV_NAME_LEN+1], PartName[PART_NAME_LEN+1];
  PART3D   *Part = NULL;
  DEVICE3D *Dev = NULL;
  WORLD    *World;

  CLI_CHECK_WORLD

  CLI_SkipSpace (Command, Offset);
  if (strncasecmp(Command+Offset, "GRAPHICS MODE", 13) == 0)
  {
    Offset += 14;
    AutoRefresh = (strncasecmp(Command+Offset, "ON", 2) == 0);
    PRO1 ("SET GRAPHICS <%s>\n", AutoRefresh? "ON":"OFF")
  }
  elsif (strncasecmp(Command+Offset, "GRID", 4) == 0)
    PRO0 ("TODO: GRID\n")
  else
  {
    Offset += ParseName (Command+Offset, DevName, PartName);

    if (PartName[0] == '\0')
      Dev = World->GetDev (DevName);
    else
      Part = World->GetPart (DevName, PartName);

    if (Dev == NULL && Part == NULL)
    {
      printf ("PART '%s':'%s' not found\n", DevName, PartName);
      return FAIL;
    }

    CLI_SkipSpace (Command, Offset);
    if (strncasecmp(Command+Offset, "HILIGHT", 7) == 0)
    {
      Offset += 8;
      Mode = (strncasecmp(Command+Offset, "ON", 2) == 0);
      PRO1 ("SET HILIGHT <%s>\n", Mode? "ON":"OFF")
      if (Dev != NULL)
	if (Mode)
	  Dev->SetState (Dev->GetState() | STATE_SELECTED);
	else
	  Dev->SetState (Dev->GetState() & (~STATE_SELECTED));
      elsif (Part != NULL)
	if (Mode)
	  Part->SetState (STATE_SELECTED);
	else
	  Part->SetState (STATE_NORMAL);
    }
    elsif (strncasecmp(Command+Offset, "COLOR", 5) == 0)
    {
      Offset += 9;
      count = 0;
      while (count < COLOR::COLORS && 
	     strncasecmp(Command+Offset+1, ColorNames[count], 
			 strlen(ColorNames[count])) != 0)
	count++;
      if (count < COLOR::COLORS)
      {
	PRO1 ("COLOR <%s>\n", ColorNames[count])
	if (Dev != NULL)
	  Dev->SetColorID ((COLOR::ID)count);
	else
	  Part->SetColorID ((COLOR::ID)count);
      }
      else
	printf ("UNKNOWN COLOR\n");
    }
    elsif (strncasecmp(Command+Offset, "RENDER", 6) == 0)
      PRO0 ("TODO: RENDER\n")
    else
      PRO1 ("UNKNOWN SET-COMMAND: '%s'\n", Command+Offset);
  }

  if (AutoRefresh)
    World->Redraw ();

  return OK;
}


static int CLI_TransferToMenu (char *Command, int Offset)
{
  DLGOpenTextDialog ("tucCADMenu", DlgMESSAGE, 
		     "tuxCAD-MENU activated", 1, "OK");

  return OK;
}


static int CLI_Translate (char *Command, int Offset)
{
  char      DevName[DEV_NAME_LEN+1], PartName[PART_NAME_LEN+1];
  double    Trans[3];
  PART3D   *Part;
  DEVICE3D *Dev;
  WORLD    *World;

  CLI_CHECK_WORLD

  Offset += ParseName (Command+Offset, DevName, PartName);
  CLI_SkipKeyWord (Command, Offset, "BY");
  sscanf (Command+Offset, "%lf, %lf, %lf", &Trans[0], &Trans[1], &Trans[2]);
#if 0
  CLI_SkipKeyWord (Command, Offset, "IN");
#endif

  if (PartName[0] == '\0')
  {
    Dev = World->GetDev (DevName);
    Part = Dev->GetPartList();
  }
  else
    Part = World->GetPart (DevName, PartName);

  if (Part != NULL)
  {
    PRO5 ("TRANSLATE '%s':'%s' BY %.3f %.3f %.3f\n",
	  DevName, PartName, Trans[0], Trans[1], Trans[2]);

    Part->SetVal (POS_X, Part->GetVal(POS_X) + Trans[0]);
    Part->SetVal (POS_Y, Part->GetVal(POS_Y) + Trans[1]);
    Part->SetVal (POS_Z, Part->GetVal(POS_Z) + Trans[2]);

    if (AutoRefresh)
      World->Redraw ();
  }
  else
    PRO0 ("TODO: TRANS DEV\n")

  return OK;
}


static int CLI_UpdateGraphics (char *Command, int Offset)
{
  WORLD *World;

  CLI_CHECK_WORLD

  World->Redraw ();

  return OK;
}


static int CLI_SelectDev (char *Command, int Offset)
{
  char      DevName[DEV_NAME_LEN+1], PartName[PART_NAME_LEN+1];
  DEVICE3D *Dev;
  WORLD    *World;

  CLI_CHECK_WORLD

  CLI_SkipSpace (Command, Offset);

  if (strncasecmp(Command+Offset, "PART BY", 7) == 0)
  {
    PRO0 ("TODO: SELECT BY MOUSE\n")
  }
  else
  {
    Offset += ParseName (Command+Offset, DevName, PartName);

    if ((Dev = World->GetDev(DevName)) == NULL)
    {
      PRGPrintf ("Device '%s' not found", DevName);
      sprintf (Command, "1: Device not found '%s'", DevName);
      return FAIL;
    }
    else
      World->Select (Dev);
  }

  return OK;
}


static int CLI_Robot (char *Command, int Offset)
{
  return CLI_SelectDev (Command, Offset);
}    
