#include "tuxCAD.h"


#define NET_TIMEOUT 10


typedef void (*SOCKETFUNCTION)(int SocketID, int pid, void *Data);

typedef struct
{
  int             Timer;
  SOCKETFUNCTION  Function;
  int             Port;
  int             PID;
  void           *Data;
  dual            New;
} HL_SOCKET_DATA;


static HL_SOCKET_DATA Connect = {1, NULL, -1, 0, NULL, FALSE};
static char Command[SOCKET_BUF_SIZE] = "B#0.0#0.0#0.0#0.0#0.0#0.0";
static int  PortNo = 2020;
static char HostName[256] = "sirius.galaxy.de";

SOCKET_DATA Client   = {-1, 0, NULL};


/* ======================================================================= */

/* ********************************************************************** */
/* ** 2nd callback-function for igrip-connectivity. Cycles until       ** */
/* ** connection is established or the timeout is reached. The         ** */
/* ** calls itself by starting a adding a timout to the x event-queue. ** */
/* ********************************************************************** */
static void StartIGRIP2 (void *Data, XtIntervalId Timer)
{
  int SocketID = -1;

  assert (Data == NULL);
  assert (int(Timer) >= 0);

  if ((SocketID = NETInitClient(HostName, Connect.Port)) < 0)
  {
    /* Does the timer reached the timeout-limit?                          */
    if ((Connect.Timer % NET_TIMEOUT) != 0)
    {
      PRGPrintf ("\\.");
      XtAppAddTimeOut (AppCon, 1000, (XtTimerCallbackProc)StartIGRIP2, NULL);
    }                                   /* Ask for reconnection           */
    elsif (DLGOpenTextDialog("connectError", DlgQUESTION, 
			     "Connection failed\n"
			     "Try to reconnect ?", 2, "retry", "abort") == 1) 
    {
      PRGPrintf ("Reconnecting...");
      XtAppAddTimeOut (AppCon, 1000, 
		       (XtTimerCallbackProc)StartIGRIP2, NULL);
    }
    else                                /* User aborted                   */
    {
      PRGPrintf ("Connection error");
      if (Connect.Function != NULL)
	Connect.Function (-1, Connect.PID, Connect.Data);
    }
  }
  else                                  /* Connection established         */
  {
    PRGPrintf ("ready !");

    if (Connect.Function != NULL)
    {
#if 0
      if (Connect.New)
      {
        Connect.PID = HL_GetLastPID ();
	PRO1 ("PID: %d\n", Connect.PID)
      }
#endif
      Connect.Function (SocketID, Connect.PID, Connect.Data);
    }
  }

  (Connect.Timer)++;
}


/* ********************************************************************** */
/* ** 1st connection-callback-function. Writes a short message and     ** */
/* ** starts the 2nd callback-function.                                ** */
/* ********************************************************************** */
static void StartIGRIP1 (void *Data, XtIntervalId Timer)
{
  PRGPrintf ("Connecting IGRIP");

  StartIGRIP2 (Data, Timer);
}

/* ********************************************************************** */
/* ** Start a new IGRIP-session using the specified port-address.      ** */
/* ** INPUT:  Port      Port-number for the connection to IGRIP        ** */
/* ** OUTPUT: Function  Callback-function, which will be called when   ** */
/* **                   connection was established or aborted. This    ** */
/* **                   function is called with the following params.  ** */
/* **                     SocketID:  Identifier for the connection     ** */
/* **                                >= 0, if connected successfully   ** */
/* **                                -1, if connecting failed          ** */
/* **                     Data:      Pointer to the specified          ** */
/* **                                user-data                         ** */
/* **         Data      void-casted pointer to users-specific data     ** */
/* ** OUTPUT: (dual)    FALSE if starting IGRIP fails, otherwise TRUE  ** */
/* ********************************************************************** */
dual StartIGRIP (int Port, char *Param, SOCKETFUNCTION Function, void *Data)
{
  int PID = 0;

  assert (Param == NULL);

  /* Get the name of the IGRIP host-computer */
  /* If we use the localhost, we will start IGRIP directly                */
  PRO2 ("Using Host: <%s>: %d\n", HostName, Port)

  /* Initialize the data-structure, because the X's timeout-function      */
  /* allows only one user-value !                                         */
  Connect.Timer    = 1;
  Connect.Function = Function;
  Connect.Port     = Port;
  Connect.Data     = Data;
  Connect.PID      = PID;
  XtAppAddTimeOut (AppCon, 10, (XtTimerCallbackProc)StartIGRIP1, NULL);
 
  return TRUE;
}

/* ======================================================================= */

void UserFunction (int SocketID, int pid, void *Data)
{
  assert (pid != 0);
  assert (Data == NULL);

  if (SocketID < 0)
    PRGPrintf ("Connecting failed !");
  else
    Client.ID = SocketID;
}


void SendData ()
{
  char   *Response = NULL;
  DIALOG *Dlg;

  if (Client.ID >= 0)
  {
    Dlg = DLGNewDialog (0);
    DLGInsertText (Dlg, 0, "cmd", SOCKET_BUF_SIZE, Command, 40, 1);

    DLGOpenDialog (Dlg, "insertCommand", DlgNORMAL, NULL);

    if (DLGGetExitButton(Dlg) == DlgOK)
    {
      strcpy (Command, DLGGetText(Dlg, 0));
      Response = NETSend (Client.ID, Command);
    }
    DLGFreeDialog (Dlg);
  }
  else
    PRGPrintf ("No Connection open...");
}


void StartClient ()
{
  StartIGRIP (PortNo, "", UserFunction, NULL);
}


void CloseClient ()
{
  if (Client.ID >= 0)
    NETCloseClient (Client.ID);

  Client.ID = -1;
}
