#include "tuxCAD.h"


static void ColorMenuCB (DIALOG *Dlg, int BoxID, int Mask)
{
  int ColorID;
  int Color[3];

  ColorID = DLGGetOMenu (Dlg, BoxID);
  if (! (Mask & OBJ_MASK))
    ColorID++;

  if (ColorID == 1)                     // Values
  {
    Color[RED]   = (int)(DLGGetScale(Dlg, BoxID+2) * 255);
    Color[GREEN] = (int)(DLGGetScale(Dlg, BoxID+3) * 255);
    Color[BLUE]  = (int)(DLGGetScale(Dlg, BoxID+4) * 255); 
  }
  elsif (ColorID == 38)                 // Texture
  {
    Color[RED] = Color[GREEN] = Color[BLUE] = (int)(0.6 * 255);
  }
  else                                  // Indexed colors 
  {
    Color[RED]   = (int)(ColorTable[ColorID][0] * 255);
    Color[GREEN] = (int)(ColorTable[ColorID][1] * 255);
    Color[BLUE]  = (int)(ColorTable[ColorID][2] * 255);
  }

  DLGSetColor (Dlg, BoxID, Color[RED], Color[GREEN], Color[BLUE], FALSE);

  // set foreground depending on brightness of actual color
  // Take into account, that green is brighter than red and blue
  if (Color[RED] + 2*Color[GREEN] + Color[BLUE] < 384)
    DLGSetColor (Dlg, BoxID, 255, 255, 255, TRUE);
  else
    DLGSetColor (Dlg, BoxID, 0, 0, 0, TRUE);

  // Set RGB-Scale sensitivity
  DLGSetBoxActive (Dlg, BoxID+2, (ColorID==1));
  DLGSetBoxActive (Dlg, BoxID+3, (ColorID==1));
  DLGSetBoxActive (Dlg, BoxID+4, (ColorID==1));
}


COLOR ColorCB (DIALOG *Dlg, int BaseID, COLOR Color, int BoxID, int Mask)
{
  int ColorID;

  ColorMenuCB (Dlg, BaseID+1, Mask);

  ColorID = DLGGetOMenu (Dlg, BaseID+1);
  if (Mask & OBJ_MASK)
    ColorID--;

  if (ColorID == 0)
    if (Mask & ALPHA_MASK)
      Color.SetValues (DLGGetScale(Dlg, BaseID+3), DLGGetScale(Dlg, BaseID+4),
		       DLGGetScale(Dlg, BaseID+5), DLGGetScale(Dlg, BaseID+6));
    else
      Color.SetValues (DLGGetScale(Dlg, BaseID+3), DLGGetScale(Dlg, BaseID+4),
		       DLGGetScale(Dlg, BaseID+5));
  else
  {
    Color.SetID ((COLOR::ID)ColorID);
    if (Mask & ALPHA_MASK)
      Color.SetAlpha (DLGGetScale(Dlg, BaseID+6));
  }

  return Color;
}


void SetColor (DIALOG *Dlg, int BaseID, COLOR Color, int Mask)
{
  if (Mask & OBJ_MASK)
    DLGSetOMenu (Dlg, BaseID+1, Color.GetID()+1);
  else
    DLGSetOMenu (Dlg, BaseID+1, Color.GetID());

  DLGSetScale (Dlg, BaseID+3, Color.GetRed());
  DLGSetScale (Dlg, BaseID+4, Color.GetGreen());
  DLGSetScale (Dlg, BaseID+5, Color.GetBlue());
  if (Mask & ALPHA_MASK)
    DLGSetScale (Dlg, BaseID+6, Color.GetAlpha());

  ColorMenuCB (Dlg, BaseID+1, Mask);
}


void SelectColor (DIALOG *Dlg, int BaseID, int ColorID, int Mask)
{
  DLGSetOMenu (Dlg, BaseID+1, ColorID);
  ColorMenuCB (Dlg, BaseID+1, Mask);
}


/* Range: BaseID...BaseID+6 */
void ColorDialog (DIALOG *Dlg, int BaseID, COLOR Color, int Mask)
{
  int Colors;

  PRO_START ("")

  DLGMakeRows (Dlg, BaseID,    2, DlgNO_LINE, BaseID+1, BaseID+2);
  if (Mask & ALPHA_MASK)
    DLGMakeRows (Dlg, BaseID+2, 4, DlgNO_LINE, 
		                   BaseID+3, BaseID+4, BaseID+5, BaseID+6);
  else
    DLGMakeRows (Dlg, BaseID+2, 3, DlgNO_LINE, BaseID+3, BaseID+4, BaseID+5);
  
  if (Mask & OBJ_MASK)
  {
    Colors = (Mask & TEX_MASK)? COLOR::COLORS :  37;
    DLGInsertOMenu (Dlg, BaseID+1, "color2", -Colors, ColorNames, 
		                             Color.GetID()+1);
  }
  else
  {
    Colors = (Mask & TEX_MASK)? 38 :  36;
    DLGInsertOMenu (Dlg, BaseID+1, "color", -Colors, ColorNames+1, 
		                            Color.GetID());
  }

  DLGInsertScale (Dlg, BaseID+3, "red",   
		                 0, 1.0, Color.GetRed(),   2, DlgHORIZONTAL);
  DLGInsertScale (Dlg, BaseID+4, "green", 
		                 0, 1.0, Color.GetGreen(), 2, DlgHORIZONTAL);
  DLGInsertScale (Dlg, BaseID+5, "blue",  
		                 0, 1.0, Color.GetBlue(),  2, DlgHORIZONTAL);
  if (Mask & ALPHA_MASK) 
    DLGInsertScale (Dlg, BaseID+6, "alpha", 
		         0, 1.0, Color.GetAlpha(), 2, DlgHORIZONTAL);

  ColorMenuCB (Dlg, BaseID+1, Mask);

  PRO_END
}
