#include "tuxCAD.h"


#define IS_CURRENT(World) (Current != NULL && Current->World == World)


LIST<DATA *>  Data;
DATA         *Current = NULL;


DATA::DATA (WORLD *World) :
  World      (World),
  Action     (NORMAL_MODE),
  Robot      (NULL),
  MouseX     (0),
  MouseY     (0),
  SelParts   (),
  SelDevs    ()
{
  assert (World != NULL);

  Scheduler    = new SCHEDULER     (World);

  PartDialog   = new PART_DIALOG   (World); 
  DevDialog    = new DEV_DIALOG    (World);
  WorldDialog  = new WORLD_DIALOG  (World);
  TreeDialog   = new TREE_DIALOG   (World);
  CreateDialog = new CREATE_DIALOG (World);
}


DATA::~DATA ()
{
  PRO_METHOD ("~DATA", "\n")

  ClearSelection ();

  if (PartDialog != NULL)
    delete PartDialog;

  if (DevDialog != NULL)
    delete DevDialog;

  if (WorldDialog != NULL)
    delete WorldDialog;

  if (TreeDialog != NULL)
    delete TreeDialog;

  if (CreateDialog != NULL)
    delete CreateDialog;

  if (Scheduler != NULL)
    delete Scheduler;

  if (World != NULL)
    delete World;
  
  PRO_END
}


void DATA::Select (PART3D *Part, SELECTION Mode)
{
  if (Mode == UNSELECT || (Mode == SWAP && SelParts.Exist(Part)))
  {
    SelParts.Delete (Part);
    World->Select (Part, FALSE);
  }
  else
  {
    SelParts.Add (Part);
    World->Select (Part, TRUE);
  }
}

 
void DATA::Select (DEVICE3D *Dev, SELECTION Mode)
{
  if (Mode == UNSELECT || (Mode == SWAP && SelDevs.Exist(Dev)))
  {
    SelDevs.Delete (Dev);
    World->Select (Dev, FALSE);
  }
  else
  {
    SelDevs.Add (Dev);
    World->Select (Dev, TRUE);
  }

  printf ("Select %d: DEV-Buffer = #%d\n", Action, SelDevs.Size());
}


bool DATA::Selected (PART3D *Part) const
{
  return SelParts.Exist (Part);
}


bool DATA::Selected (DEVICE3D *Dev) const
{
  return SelDevs.Exist (Dev);
}


LIST<PART3D *> *DATA::GetPartBuffer ()
{
  return &SelParts;
}


LIST<DEVICE3D *> *DATA::GetDevBuffer ()
{
  return &SelDevs;
}


void DATA::ClearSelection ()
{
  while (SelParts.Size() > 0)
    Select (SelParts.Get(0), UNSELECT);

  while (SelDevs.Size() > 0)
    Select (SelDevs.Get(0), UNSELECT);
}


int DATA::SelectedParts () const
{
  return SelParts.Size ();
}


int DATA::SelectedDevs () const
{
  return SelDevs.Size ();
}


void DATA::SetMouse (int MouseX, int MouseY)
{
  this->MouseX = MouseX;
  this->MouseY = MouseY;
}


int DATA::GetMouseX () const
{
  return MouseX;
}


int DATA::GetMouseY () const
{
  return MouseY;
}



DATA *GetData (WORLD *World)
{
  int   count;
  DATA *Ptr;

  for (count = 0; count < Data.Size(); ++count)
  {
    Ptr = Data.Get (count);
    if (Ptr->World == World)
      return Ptr;
  }

  return NULL;
}


DATA *GetData (int No)
{
  assert (No >= 0 && No < Data.Size());

  return Data.Get (No);
}
  

WORLD *GetWorld (int No)
{
  assert (No >= 0 && No < Data.Size());

  return Data.Get (No)->World;
}


void InsWorld (WORLD *World)
{
  Current = new DATA (World);

  Data.Add (Current);
}


void DelWorld (WORLD *World)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
  {
    PRO0 ("Deleting buffer\n")
    Data.Delete (Current);
    delete (Current);
    Current = NULL;
  }
}


int GetWorldNo ()
{
  return Data.Size ();
}


WORLD *GetCurrent ()
{
  if (Current == NULL)
    return NULL;
  else
    return Current->World;
}


void SetCurrent (WORLD *World)
{
  if (! IS_CURRENT(World))
    Current = GetData (World);
}


void SwapSelection (WORLD *World, PART3D *Part)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
    Current->Select (Part, DATA::SWAP);
}


void SwapSelection (WORLD *World, DEVICE3D *Dev)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
    Current->Select (Dev, DATA::SWAP);
}


bool IsSelected (WORLD *World, PART3D *Part)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
    return Current->Selected (Part);

  return FALSE;
}


bool IsSelected (WORLD *World, DEVICE3D *Dev)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
    return Current->Selected (Dev);

  return FALSE;
}


LIST<PART3D *> *GetPartBuffer (WORLD *World)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
    return Current->GetPartBuffer ();

  return NULL;
}


LIST<DEVICE3D *> *GetDevBuffer (WORLD *World)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
    return Current->GetDevBuffer ();

  return NULL;
}


void ClearSelection (WORLD *World)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
    Current->ClearSelection ();
}


void SetAction (WORLD *World, DATA::ACTION Action /* = DATA::NORMAL_MODE */)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
  {
    Current->Action = Action;
    WorldMouseCursor (World);
  }
}


DATA::ACTION GetAction (WORLD *World)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
    return Current->Action;
  else
    return DATA::NORMAL_MODE;
}


void SetMouse (WORLD *World, int MouseX, int MouseY)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
    Current->SetMouse (MouseX, MouseY);
}


int GetMouseX (WORLD *World)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
    return Current->GetMouseX ();
  return 0;
}


SCHEDULER *GetScheduler (WORLD *World)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
    return Current->Scheduler;
  return NULL;
}


int GetMouseY (WORLD *World)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
    return Current->GetMouseY ();
  return 0;
}


void OpenTreeDialog (WORLD *World)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
    Current->TreeDialog->Open ();
}


void OpenPartDialog (WORLD *World, PART3D *Part)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
    Current->PartDialog->Open (Part);
}


void OpenDevDialog (WORLD *World, DEVICE3D *Dev)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
    Current->DevDialog->Open (Dev);
}


void OpenWorldDialog (WORLD *World)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
    Current->WorldDialog->Open ();
}


void OpenCreateDialog (WORLD *World)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
    Current->CreateDialog->Open ();
}


void UpdateDialogs (WORLD *World, unsigned long Mask /* = DLG_ALL */)
{
  if (IS_CURRENT(World) || (Current = GetData(World)) != NULL)
  {
    if (Mask & DLG_PART)
      Current->PartDialog->Update ();
    if (Mask & DLG_DEV)
      Current->DevDialog->Update ();
    if (Mask & DLG_WORLD)
      Current->WorldDialog->Update ();
    if (Mask & DLG_TREE)
      Current->TreeDialog->Update ();
    if (Mask & DLG_CREATE)
      Current->CreateDialog->Update ();
  }
}
