#include <dlfcn.h>
#include <dirent.h>
#include <math.h>

#ifdef _DYNAMIC_
  #include <libfilter.h>
#endif

#include "tuxCAD.h"


#ifndef _DYNAMIC_

#include "filter/filter_igrip/igrip.cc"


DEVICE3D *ImportDevice (WORLD *World, const char *FileSpec)
{
  return ImportDevice (World, FileSpec, TOOLGetConfigString(CFG_DATA_PATH));
}


WORLD *ImportWorld (WORLD *World, const char *FileSpec)
{
  return ImportWorld (World, FileSpec, TOOLGetConfigString(CFG_DATA_PATH));
}


void ListFormats ()
{
  printf ("IGRIP\n");
}

#else

#define FUNC_GETNAME      "_Z7GetNamev"  
// gcc V2.95:             "GetName__Fv"
#define FUNC_IMPORTDEV    "_Z12ImportDeviceP5WORLDPKcS2_"  
// gcc V2.95:             "ImportDevice__FP5WORLDPCcT1"
#define FUNC_IMPORTWORLD  "_Z11ImportWorldP5WORLDPKcS2_"
// gcc V2.95:             "ImportWorld__FP5WORLDPCcT1"


static int ImportSelect (const struct dirent *Entry)
{
  return (strncmp(Entry->d_name, "lib", 3) == 0 &&
	  strstr(Entry->d_name, ".so") != NULL);
}


FILTER::FILTER (const char *LibDir)
{
  strncpy (this->LibDir, LibDir, PATH_MAX);
  this->LibDir[PATH_MAX] = '\0';
}


void *FILTER::Execute (WORLD *World, const char *FileSpec, FILTERFUNC Function)
{
         void       *Result = NULL;
         int         count, Number;
	 char       *Error;
         char        FileName[PATH_MAX+1];
         void       *LibPtr;
         NAME_FUNC   NameFunc;
  struct dirent    **Entry = NULL;

  if ((Number = scandir(LibDir, &Entry, ImportSelect, alphasort)) > 0)
  {
    count = 0;
    while (count < Number && Result == NULL)
    {
      sprintf (FileName, "%s/%s", LibDir, Entry[count]->d_name); 
      printf ("reading '%s' ...\n", FileName);

      if ((LibPtr = dlopen(FileName, RTLD_LAZY)) != NULL)
      {
	NameFunc = (NAME_FUNC)dlsym (LibPtr, FUNC_GETNAME);
	if ((Error = dlerror()) == NULL)
	  printf ("Format %d: %s\n", count+1, NameFunc());
	else
	  printf ("Symbol 'GetName' not found\n");

        Result = Function (LibPtr, World, FileSpec);

	dlclose (LibPtr);
	printf ("ok\n");
      }
      else
	printf ("failed\n");

      ++count;
    }
  }

  for (count = 0; count < Number; ++count)
    free (Entry[count]);    
  free (Entry);

  return Result;
}


static void *PrintFormat (void *LibPtr, WORLD *World, const char *FileSpec)
{
  char         *Error;
  IMP_DEV_FUNC  ImportDevFunc;

  Error = NULL;
  printf ("  Symbol 'ImportDevice...");
  ImportDevFunc = (IMP_DEV_FUNC)dlsym (LibPtr, FUNC_IMPORTDEV);
  if (ImportDevFunc == NULL || (Error = dlerror()) != NULL)
  {
    printf ("not found\n");
    if (Error != NULL)
      printf ("    '%s'\n", Error);
  }
  else
    printf ("ok\n");

  Error = NULL;
  printf ("  Symbol 'ImportWorld...");
  ImportDevFunc = (IMP_DEV_FUNC)dlsym (LibPtr, FUNC_IMPORTWORLD);
  if (ImportDevFunc == NULL || (Error = dlerror()) != NULL)
  {
    printf ("not found\n");
    if (Error != NULL)
      printf ("    '%s'\n", Error);
  }
  else
    printf ("ok\n");

  return NULL;
}


void ListFormats ()
{
  FILTER Filter (TOOLGetConfigString(CFG_LIBS_PATH));

  Filter.Execute (NULL, NULL, PrintFormat);
}


static void *LoadDevice (void *LibPtr, WORLD *World, const char *FileSpec)
{
  char         *Error;
  IMP_DEV_FUNC  ImportFunc;

  ImportFunc = (IMP_DEV_FUNC)dlsym (LibPtr, FUNC_IMPORTDEV);
  if ((Error = dlerror()) != NULL)
    PRGPrintf ("Error starting: %s", Error);
    
  return ImportFunc(World, FileSpec, TOOLGetConfigString(CFG_DATA_PATH));
}

DEVICE3D *ImportDevice (WORLD *World, const char *FileSpec)
{
  FILTER Filter (TOOLGetConfigString(CFG_LIBS_PATH));

  return (DEVICE3D *)Filter.Execute (World, FileSpec, LoadDevice);
}


static void *LoadWorld (void *LibPtr, WORLD *World, const char *FileSpec)
{
  char           *Error;
  IMP_WORLD_FUNC  ImportFunc;

  ImportFunc = (IMP_WORLD_FUNC)dlsym (LibPtr, FUNC_IMPORTWORLD);
  if ((Error = dlerror()) != NULL)
    printf ("Error accessing library symbol 'ImportWorld': %s", Error);

  return ImportFunc (World, FileSpec, TOOLGetConfigString(CFG_DATA_PATH));
}


WORLD *ImportWorld (WORLD *World, const char *FileSpec)
{
  FILTER Filter (TOOLGetConfigString(CFG_LIBS_PATH));

  return (WORLD *)Filter.Execute (World, FileSpec, LoadWorld);
}

#endif
