#include <math.h>

#include "tuxCAD.h"


#define FONT_TEXT  "TuxCAD !"
#define FONT_DIR   "FONTS"
#define FONT_FILE  "times.ttf"
#define FONT_SIZE  220


typedef struct
{
  char      FontMask[PATH_MAX+1];
  char      FontFile[PATH_MAX+1];
  int       FontSize;
  double    FontDepth;
  double    Precision;
  char      Text[256];
  int       PosX, PosY;

  MATERIAL  Material;

  bool      Relation;
  int       Function;
} FONT_DATA;


static FONT_DATA Fonts = 
{
  CAD_PATH FONT_DIR"/*.ttf", /* */   // FontMask 
  CAD_PATH FONT_DIR"/"FONT_FILE,     // FontFile
  64,                                // Size
  10.0,                              // Depth
  0.2,                               // Precision
  FONT_TEXT,                         // Text
  -1, -1,                            // Dialog-position

  MATERIAL (COLOR::MEDRED),          // Material

  FALSE,                              // Relation
  -1                                  // Function
};


void Transform (POINT3D *Point, VECTOR3D *Normal)
{
  switch (Fonts.Function)
  {
    case 0 :
      Point->y += sin (Point->x / 50.0) * 100.0;
      break;

    case 1 :
      Point->y *= cos (Point->x / 50.0);
      Point->z  = Point->y * sin (Point->x / 50.0);
      break;

    default :
      ;
  }

#if 0
  Pos[X] = Point->x /* + OffsX */;
  Pos[Y] = Point->y + DeltaY;
  Angle = 180.0 * Point->x / Vec->getAdvance ();
  Pos[Z] = sin (RAD(Angle)) * 25.0;
#endif
}


static int TextCB (DIALOG *Dlg, int BoxID)
{
  int Page;

  Page = DLGGetNote (Dlg, 0);
  switch (Page)
  {
    case 1 :
      {
	strncpy (Fonts.FontFile, DLGGetFileSel(Dlg, 110), PATH_MAX);
	Fonts.FontFile[PATH_MAX] = '\0';
	TOOLGetFileDir (Fonts.FontFile, Fonts.FontMask, PATH_MAX);
	strcat (Fonts.FontMask, "/*.ttf"); /* */
	Fonts.FontSize    = (int)DLGGetScale (Dlg, 120);
	Fonts.FontDepth   =      DLGGetUpDn  (Dlg, 130);
	Fonts.Precision   =      DLGGetScale (Dlg, 140);
	strncpy (Fonts.Text, DLGGetText(Dlg, 150), 255);
	Fonts.Text[255] = '\0';
      }
      break;

    case 2 :
      MaterialCB (Dlg, 200, &(Fonts.Material), BoxID, ALPHA_MASK | TEX_MASK); 
      break;

    default :
      break;
  }

  return DlgCONFIRM;
}


void MakeText ()
{
  DEVICE3D *Dev;
  DIALOG   *Dlg;
  WORLD    *World;
  
  PRGPrintf ("Create 3D-text...");

  if ((World = SelectWorld(TRUE)) == NULL)
  {
    PRGPrintf ("\\Function aborted");
    return;
  }

  Dlg = DLGNewDialog (0);
  DLGInsertNote (Dlg, 0, "note", DlgNOBINDING);
    DLGAddPage (Dlg, 0, DlgMAJOR, "text", 100);
      DLGMakeRows (Dlg, 100, 5, DlgNO_LINE, 110, 120, 130, 140, 150);

      DLGInsertFileSel (Dlg, 110, "font",  Fonts.FontMask, 50, Fonts.FontFile);
      DLGInsertScale (Dlg, 120, "size",    1.0, 1000.0, Fonts.FontSize, 0, 
		                           DlgHORIZONTAL);
      DLGInsertUpDn  (Dlg, 130, "depth",   0.0, 1000.0, Fonts.FontDepth, 1.0);
      DLGInsertScale (Dlg, 140, "precision", 0.0, 1.0, Fonts.Precision, 2, 
		                             DlgHORIZONTAL);
      DLGInsertText  (Dlg, 150, "text",      255, Fonts.Text, 40, 1);

    DLGAddPage (Dlg, 0, DlgMAJOR, "material", 200);
      MaterialDialog (Dlg, 200, &Fonts.Material, ALPHA_MASK | TEX_MASK);

    DLGAddPage (Dlg, 0, DlgMAJOR, "transform", 300);
      DLGMakeRows (Dlg, 300, 5, DlgNO_LINE, 310, 320, 330, 340, 350);

      DLGInsertOMenu (Dlg, 310, "relation", 2, "absolute", "relative", 
		                               Fonts.Relation);
      
      // Fonthhe, -dicke, drehung, Verzerrung, Kurve, 

  DLGSetFunction (Dlg, 0,   TextCB);
  DLGSetPos      (Dlg, Fonts.PosX, Fonts.PosY);

  DLGMakeButton (Dlg, DEFAULT_BUTTON);
  DLGMakeButton (Dlg, HELP_BUTTON);
  DLGMakeButton (Dlg, OK_BUTTON);

  DLGOpenDialog (Dlg, "fontSelection", DlgNORMAL, NULL);  
  if (DLGGetExitButton(Dlg) == DlgOK)
  {
    DLGGetPos (Dlg, &Fonts.PosX, &Fonts.PosY);

    Dev = UPDATE(World)->CreateText (Fonts.FontFile, 
				     Fonts.FontSize, Fonts.FontDepth, 
				     Fonts.Precision, Fonts.Material, 
				     Transform, FALSE,
				     Fonts.Text);
    World->AddDevice (Dev);
    ClearSelection   (World);
    World->Fill      ();
  }
  DLGFreeDialog (Dlg);
}


static int WriteCB (DIALOG *Dlg, int BoxID)
{
  static int       count = 1;
  static double    width = 0.0;
  static DEVICE3D *Last = NULL;
         DEVICE3D *Dev;
         WORLD    *World;

  World = (WORLD *)DLGGetUserData (Dlg, 0);
  if (BoxID == 10)
  {
    Dev = World->CreateText ("/users/stefan/IGRIP/FONTS/times.ttf", 64, 10.0,
			     0.5, MATERIAL(COLOR::MEDRED), NULL, TRUE,
			     "%d", count);
    if (Last == NULL)
      World->AddDevice (Dev);
    else
    {
      // Works only with single characters. Otherwise, the text must be
      // attached to last part !
      Dev->GetPartList()->SetVal (POS_X, width);
      Last->GetPartList()->AttachDev (Dev);
    }
    Last = Dev;
    width = World->GetTextWidth ("/users/stefan/IGRIP/FONTS/times.ttf", 64, 
				 NULL, TRUE, "%d", count);
    count = (count+1) % 10 ;
  }
  else
  {
    // letztes Zeichen lschen
  }

  World->Fill ();

  return DlgCONFIRM;
}


void MakeText2 ()
{
  DIALOG *Dlg;
  WORLD  *World;

  PRGPrintf ("Create 3d-Text");

  if ((World = SelectWorld(TRUE)) == NULL)
  {
    PRGPrintf ("\\function aborted");
    return;
  }

  Dlg = DLGNewDialog (0);
  DLGMakeRows (Dlg, 0, 2, DlgNO_LINE, 10, 20);
  DLGInsertProc (Dlg, 10, "add", WriteCB);
  DLGInsertProc (Dlg, 20, "del", WriteCB);
  DLGSetUserData (Dlg, 0, World);

  DLGOpenDialog (Dlg, "font", DlgNORMAL, NULL);
  if (DLGGetExitButton(Dlg) == DlgOK)  
  {
  }
  DLGFreeDialog (Dlg);
}
