#include <X11/keysym.h>

#include "tuxCAD.h"


#define GRANULARITY 0.1                 /* Granularity of mouse actions */


enum KEY_BITS
{
  KEY_SHIFT = (1<<0),                   // 1 = 0000000000000001
  KEY_CTRL  = (1<<1),                   // 2 = 0000000000000010
  KEY_ALT   = (1<<2)                    // 4 = 0000000000000100
};


static unsigned long KeyMask = 0x00;
static int           M1x = 0, M1y = 0;
static int           OldRectX = -1;
static int           OldRectY = -1;


static int GetKeyMask ()
{
  char Keys[32];

   XQueryKeymap (XDisplay, Keys);

  if (Keys[4] & 0x20)
    KeyMask |= KEY_CTRL;
  else
    KeyMask &= ~KEY_CTRL;
  if (Keys[6] & 0x04 || Keys[7] & 0x40)
    KeyMask |= KEY_SHIFT;
  else
    KeyMask &= ~KEY_SHIFT;

  return KeyMask;
}


// TODO: TranslateWorldAbs / TranslateWorldRel
static void TranslateWorld (WORLD *World, int Button, int Dist)
{
  int     SizeX, SizeY;
  RPY     Camera = World->GetCamera ();
  VIEW   *View   = World->GetView ();
  double *Range  = View->GetRange ();

  World->GetWinSize (&SizeX, &SizeY);
  if (Button & Button1Mask)
    Camera.px += 0.25 * Dist / (double)SizeX * (Range[MAX_X] - Range[MIN_X]);
  if (Button & Button2Mask)
    Camera.py -= 0.25 * Dist / (double)SizeX * (Range[MAX_Y] - Range[MIN_Y]);
  if (Button & Button3Mask)
    Camera.pz -= 0.25 * Dist / (double)SizeX * (Range[MAX_Z] - Range[MIN_Z]);

  View->SetCamera (Camera);
  PRGPrintf ("Camera-Pos: %.2f %.2f %.2f [mm]", 
	     Camera.px, Camera.py, Camera.pz);
}


// TODO RotateWorldAbs / RotateWorldRel
static void RotateWorld (WORLD *World, int Button, int Dist)
{
  RPY   Camera = RPY_0;
  VIEW *View = World->GetView ();

  if (Button & Button1Mask)
    Camera.y = Dist;
  if (Button & Button2Mask)
    Camera.p = Dist;
  if (Button & Button3Mask)
    Camera.r = Dist;
  View->Rotate (Camera.y, Camera.p, Camera.r);

  Camera = World->GetCamera ();
  PRGPrintf ("Camera-Rot: %.2f %.2f %.2f []", Camera.y, Camera.p, Camera.r);
}


static void PanWorld (WORLD *World, int Button, int DistX, int DistY)
{
  double Dist;
  double OffsX = 0.0, OffsY = 0.0;

  if (Button & Button3Mask)
  {
    OffsX = - DistX;
    OffsY =   DistY;
  }
  else
  {
    Dist = (abs(DistX) > abs(DistY))? DistX : DistY;
    if (Button & Button1Mask)
      OffsX = - Dist;
    if (Button & Button2Mask)
      OffsY =   Dist;
  }

  World->GetView()->Pan (OffsX, OffsY);
}


static void FlyWorld (WORLD *World, int Button, int DistX, int DistY)
{
  int       count;
  double    COI[3];
  double   *Range;
  RPY       Camera = RPY_0;
  H_MATRIX  Cam, Trans = Matrix_1;
  H_MATRIX  NewCOI = Matrix_1;
  VIEW     *View   = World->GetView ();

  Camera = World->GetCamera ();

  if (Button & Button1Mask)
  {
#if 0
    Range = View->GetRange ();
    Range[MIN_Z] += DistX;
    Range[MAX_Z] += DistX;
    View->SetRange (Range);
#endif
#if 1
    Cam = GEORPY2Frame (Camera.px, Camera.py, Camera.pz, 
			Camera.y,  Camera.p,  Camera.r);
    Trans.m[P][Z] = DistX;
    Cam = GEOMatrixProd (Cam, Trans);
    Camera.px = Cam.m[P][X];
    Camera.py = Cam.m[P][Y];
    Camera.pz = Cam.m[P][Z];
    View->SetCamera (Camera);
#endif
  }
  elsif (Button & Button2Mask)
  {
#if 1
  Range = View->GetRange ();
  Cam = GEORPY2Frame (Camera.px, Camera.py, Camera.pz, 
		      Camera.y,  Camera.p,  Camera.r);
  Trans = GEORPY2Frame (0.0, 0.0, 0.0, DistX, DistY, 0.0);
  View->GetCOI (&COI[X], &COI[Y], &COI[Z]);
  NewCOI.m[P][X] = COI[X];
  NewCOI.m[P][Y] = COI[Y];
  NewCOI.m[P][Z] = COI[Z];
  NewCOI = GEOMatrixProd (Trans, NewCOI);
  for (count = X; count <= Z; ++count)
  {
    Range[2*count]   += NewCOI.m[P][count] - COI[count];
    Range[2*count+1] += NewCOI.m[P][count] - COI[count];
  }
  printf ("COI: %.2f %.2f %.2f\n", 
	  NewCOI.m[P][X] - COI[X], NewCOI.m[P][Y] - COI[Y],
	  NewCOI.m[P][Z] - COI[Z]);

  View->SetRange (Range);

  Camera.r = DistX;
  Camera.y = DistY;
  View->Rotate (Camera.y, Camera.p, Camera.r);    
#else
    Camera.r = DistX;
    Camera.y = DistY;
    View->Rotate (Camera.y, Camera.p, Camera.r);    
#endif
  }
  elsif (Button & Button3Mask)
    SetAction (World, DATA::NORMAL_MODE);
}


static bool ZoomWorld (WORLD *World, int Button, int MouseX, int MouseY, 
		       double Dist)
{
  int       SizeX, SizeY;
  double    Factor;
  WINDOW3D *Win = World->GetWin ();

  if (Dist == 0.0)
    return TRUE;

  if (Button & Button1Mask)
  {
    double ARatio1, ARatio2;

    World->GetWinSize (&SizeX, &SizeY);

    XSetForeground (Win->GetDisplay(), xorGC, 0xffffffff);
    XDrawRectangle (Win->GetDisplay(), Win->GetWindow(), 
		    xorGC, MIN(M1x,OldRectX), MIN(M1y,OldRectY),
		    abs(OldRectX-M1x), abs(OldRectY-M1y));

    OldRectX = MouseX;
    OldRectY = MouseY;

    // On Shift: keep aspect ratio
    if ((KeyMask & KEY_SHIFT) && (MouseY != M1y))
    {
      ARatio1 = double(SizeX) / double(SizeY);
      ARatio2 = (MouseX-M1x) / (double)(MouseY-M1y);
      if (ARatio1 > ARatio2)
	OldRectY = M1y + (int)((MouseX-M1x) / ARatio1);
      elsif (ARatio1 < ARatio2)
	OldRectX = M1x + (int)((MouseY-M1y) / ARatio1);
    }

    XDrawRectangle (Win->GetDisplay(), Win->GetWindow(),
		    xorGC, MIN(M1x,OldRectX), MIN(M1y,OldRectY),
		    abs(OldRectX-M1x), abs(OldRectY-M1y));
    return FALSE;
  }
  elsif (Button & Button2Mask)
    PRGPrintf ("Zoom in Center-Pos");
  elsif (Button & Button3Mask)
  {
#if 1
    Factor = 1.0 - Dist * 0.005;
#else
    // This formular changes the zoom factor proportianal to the quadratic 
    // mouse motion. So, this allows a fine adjustment on small motion, but 
    // also large changes on faster motion.
    Dist *= 0.01;
    Factor = 1.0 / (1.0 + SQR(Dist));
    if (Dist < 0.0)
      Factor = 2.0 - Factor;
#endif
    World->SetScale (Factor);
    PRGPrintf ("Scale: %.2f %%", World->GetView()->GetScale()*100.0);
  }

  return TRUE;
}


static bool PartMouseHandle (WORLD *World, const LIST<PART3D *> &List,
			     int Button, int MouseX, int MouseY)
{
  int          count, Dist;
  int          OldX, OldY;
  PART3D      *Part = NULL;
  DATA::ACTION Action = GetAction (World);

  OldX = GetMouseX (World);
  OldY = GetMouseY (World);
  if (abs(MouseX - OldX) > abs(MouseY - OldY))
    Dist = MouseX - OldX;
  else
    Dist = MouseY - OldY;

  if (Dist == 0)
    return FALSE;

  SetMouse (World, MouseX, MouseY);

  if (List.Size() == 0)
    return FALSE;

  if (KeyMask & KEY_SHIFT)              // Shift increases motion
    Dist *= 10;

  switch (Action)
  {
    // ************* TRANSLATE RELATIVE/ABSOLUTE ****************
    case DATA::TRANS_ABSOLUTE :
    case DATA::TRANS_RELATIVE :
    {
      double   Change;
      H_MATRIX Rel = Matrix_1;
      H_MATRIX WTP; 

      Change = GRANULARITY * (double)Dist;

      if (Button & Button1Mask)
	Rel.m[P][X] = Change;
      if (Button & Button2Mask)
	Rel.m[P][Y] = Change;
      if (Button & Button3Mask)
	Rel.m[P][Z] = Change;

      for (count = 0; count < List.Size(); ++count)
      {
	Part = List.Get (count);
	if (Action == DATA::TRANS_ABSOLUTE)
	  WTP = GEOMatrixProd (Rel, Part->GetMatrix());
	else
	  WTP = GEOMatrixProd (Part->GetMatrix(), Rel);
	Part->SetMatrix (WTP);
      }
      if (List.Size() == 1)
	PRGPrintf ("Pos: %.2f %.2f %.2f [mm]", 
		   WTP.m[P][X], WTP.m[P][Y], WTP.m[P][Z]);
    }
    break;

    // *************** ROTATE ABSOLUTE/RELATIVE *****************
    case DATA::ROT_ABSOLUTE :
    case DATA::ROT_RELATIVE :
    {
      double   Change;
      RPY      Rel = RPY_0;
      H_MATRIX T, WTP, Old;

      Change = GRANULARITY * (double)Dist;

      if (Button & Button1Mask)
	Rel.y = Change;
      if (Button & Button2Mask)
	Rel.p = Change;
      if (Button & Button3Mask)
	Rel.r = Change;
      T = GEORPY2Frame (0.0, 0.0, 0.0, Rel.y, Rel.p, Rel.r);

      for (count = 0; count < List.Size(); ++count)
      {
	Part = List.Get (count);

	Old = Part->GetMatrix ();
	if (Action == DATA::ROT_ABSOLUTE)
	  WTP = GEOMatrixProd (T, Old);
	else
	  WTP = GEOMatrixProd (Old, T);

	WTP.m[P][X] = Old.m[P][X];        // Keep position (don't rotate
	WTP.m[P][Y] = Old.m[P][Y];        // arround origin of world-CS)
	WTP.m[P][Z] = Old.m[P][Z];
	Part->SetMatrix (WTP);
	Rel = GEOGetRPY (WTP);
      }
      if (List.Size() == 1)
	PRGPrintf ("Rot: %.2f %.2f %.2f []", Rel.y, Rel.p, Rel.r);
    }
    break;

    // ***************** JOINT MOVE TRANSLATE ***********************
    case DATA::MOVE_TRANSLATE :
    {
      int Axes;

      for (count = 0; count < List.Size(); ++count)
      {
	Part = List.Get (count);

	Axes = Part->GetJoint().GetAxes ();
	if ((Button & Button1Mask) && (Axes & (1<<ROT_X)))
	  Part->SetVal (POS_X, Part->GetVal(POS_X)+Dist);
	if ((Button & Button2Mask) && (Axes & (1<<ROT_Y)))
	  Part->SetVal (POS_Y, Part->GetVal(POS_Y)+Dist);
	if ((Button & Button3Mask) && (Axes & (1<<ROT_Z)))
	  Part->SetVal (POS_Z, Part->GetVal(POS_Z)+Dist);
      }
      if (List.Size() == 1)
	PRGPrintf ("Joint: %.2f %.2f %.2f [mm]", Part->GetVal(POS_X), 
		   Part->GetVal(POS_Y), Part->GetVal(POS_Z));
    }
    break;

    // ***************** JOINT MOVE ROTATE **************************
    case DATA::MOVE_ROTATE :
    {
      int Axes;

      for (count = 0; count < List.Size(); ++count)
      {
	Part = List.Get (count);

	Axes = Part->GetJoint().GetAxes ();

	if ((Button & Button1Mask) && (Axes & (1<<ROT_X)))
	  Part->SetVal (ROT_X, MOD(Part->GetVal(ROT_X)+Dist,
				   -180.0, 180.0, 360.0));
	if ((Button & Button2Mask) && (Axes & (1<<ROT_Y)))
	  Part->SetVal (ROT_Y, MOD(Part->GetVal(ROT_Y)+Dist,
				   -180.0, 180.0, 360.0));
	if ((Button & Button3Mask) && (Axes & (1<<ROT_Z)))
	  Part->SetVal (ROT_Z, MOD(Part->GetVal(ROT_Z)+Dist,
				   -180.0, 180.0, 360.0));
      }
      if (List.Size() == 1)
	PRGPrintf ("Joint: %.2f %.2f %.2f []", Part->GetVal(ROT_X), 
		   Part->GetVal(ROT_Y), Part->GetVal(ROT_Z));
    }
    break;

    default :
      if (Action != -1)
	printf ("Invalid Action: %d\n", Action);
      return FALSE;
  }
  return TRUE;
}


static bool DevMouseHandle (WORLD *World, const LIST<DEVICE3D *> &List, 
			    int Button, int MouseX, int MouseY)
{
  int           count, Dist;
  int           OldX, OldY;
  DEVICE3D     *Dev = NULL;
  DATA::ACTION  Action = GetAction (World);

  OldX = GetMouseX (World);
  OldY = GetMouseY (World);
  if (abs(MouseX - OldX) > abs(MouseY - OldY))
    Dist = MouseX - OldX;
  else
    Dist = MouseY - OldY;

  if (Dist == 0)
    return FALSE;

  SetMouse (World, MouseX, MouseY);

  if (List.Size() == 0)
    return FALSE;

  switch (Action)
  {
    // ************* TRANSLATE RELATIVE/ABSOLUTE ****************
    case DATA::TRANS_ABSOLUTE :
    case DATA::TRANS_RELATIVE :
    {
      H_MATRIX Rel = Matrix_1;
      H_MATRIX WTD;
 
      if (Button & Button1Mask)
	Rel.m[P][X] = Dist;
      if (Button & Button2Mask)
	Rel.m[P][Y] = Dist;
      if (Button & Button3Mask)
	Rel.m[P][Z] = Dist;

      for (count = 0; count < List.Size(); ++count)
      {
	Dev = List.Get (count);

	if (Action == DATA::TRANS_ABSOLUTE)
	  WTD = GEOMatrixProd (Rel, Dev->GetWTD());
	else
	  WTD = GEOMatrixProd (Dev->GetWTD(), Rel);
	Dev->SetWTD (WTD);
      }
      if (List.Size() == 1)
	PRGPrintf ("Pos: %.2f %.2f %.2f [mm]", 
		   WTD.m[P][X], WTD.m[P][Y], WTD.m[P][Z]);
    }
    break;

    // *************** ROTATE ABSOLUTE/RELATIVE *****************
    case DATA::ROT_ABSOLUTE : 
    case DATA::ROT_RELATIVE :
    {
      RPY      Rel = RPY_0;
      H_MATRIX T, WTD, Old;

      if (Button & Button1Mask)
	Rel.y = Dist;
      if (Button & Button2Mask)
	Rel.p = Dist;
      if (Button & Button3Mask)
	Rel.r = Dist;
      T = GEORPY2Frame (0.0, 0.0, 0.0, Rel.y, Rel.p, Rel.r);

      for (count = 0; count < List.Size(); ++count)
      {
	Dev = List.Get (count);

	Old = Dev->GetWTD ();
	if (Action == DATA::ROT_ABSOLUTE)
	  WTD = GEOMatrixProd (T, Old);
	else
	  WTD = GEOMatrixProd (Old, T);

	WTD.m[P][X] = Old.m[P][X];        // Keep position (don't rotate
	WTD.m[P][Y] = Old.m[P][Y];        // arround origin of world-CS)
	WTD.m[P][Z] = Old.m[P][Z];
	Dev->SetWTD (WTD);
	Rel = GEOGetRPY (WTD);
      }
      if (List.Size() == 1)
	PRGPrintf ("Rot: %.2f %.2f %.2f []", Rel.y, Rel.p, Rel.r);
    }
    break;

    // *************** MOVE TCP ABSOLUTE/RELATIVE *****************
    case DATA::TCP_TRANS_REL :
    case DATA::TCP_TRANS_ABS :
    case DATA::TCP_ROT_REL   :
    case DATA::TCP_ROT_ABS   :
    {
             int        count;
	     double    *J = NULL;
             VECTOR3D   Diff = {0.0, 0.0, 0.0};
             H_MATRIX   Next, Rel = Matrix_1;
             JOINTS     Joint;
             DEVICE3D  *Dev;
      static H_MATRIX   WTT, WTR, RTT, WTM;
      static PART3D    *Mount = NULL;
      static ROBOT     *Robot; 

      if (Button & Button1Mask)
	Diff.x = Dist;
      if (Button & Button2Mask)
	Diff.y = Dist;
      if (Button & Button3Mask)
	Diff.z = Dist;

      if (Action == DATA::TCP_TRANS_REL || Action == DATA::TCP_TRANS_ABS)
	Rel = GEOSetMatrixVector (Rel, Diff, P);
      else
	Rel = GEORPY2Frame (0.0, 0.0, 0.0, Diff.x, Diff.y, Diff.z);

      for (count = 0; count < List.Size(); ++count)
      {
	Dev = List.Get (count);

	if ((Robot = Dev->GetRobot()) == NULL)
	{
	  PRGPrintf (MSG_NO_KINEMATIC);
	  return FALSE;
	}

	if ((Mount = Dev->GetMount()) == NULL)          // Get mount part
	{
	  PRGPrintf (MSG_NO_TCP);
	  return FALSE;
	}

	WTR = Dev->GetWTD ();                           // World_T_Robot
	WTT = Dev->GetTCP ();                           // World_T_TCP
	RTT = GEOMatrixProd (GEOMatrixInv(WTR), WTT);   // Robot_T_TCP
#if 0
	// Test, ob TuxCAD-TCP mit TCP aus Kinematik bereinstimmt
	// Nicht bei HEXA mglich, da (noch) keine Vorwrtskinematik existiert.
	double *JointVec;
	int     JointNo = Dev->GetJoints (&JointVec);
	FRAME TestF = Robot->FwdKin (JOINTS(JointNo, JointVec));
	RTT = TO_MATRIX (TestF);
	free (JointVec);
	WRITE_MATRIX (stdout, RTT, "ROBOT-RTT")
#endif

	// Compute next frame
	if (Action == DATA::TCP_TRANS_REL || Action == DATA::TCP_ROT_REL)
	  Next = GEOMatrixProd (RTT, Rel);      // Relative transformation
	else
	  Next = GEOMatrixProd (Rel, RTT);      // Absolute transformation

	if (Action == DATA::TCP_ROT_ABS)       // Restore pos, if abs rotate
	{
	  Next.m[P][X] = RTT.m[P][X];
	  Next.m[P][Y] = RTT.m[P][Y];
	  Next.m[P][Z] = RTT.m[P][Z];
	}

	try
	{
	  Joint = Robot->InvKin (TO_FRAME(Next));
	  J = new double[Joint.GetNumber()];
	  for (count = 0; count < Joint.GetNumber(); ++count)
	    J[count] = Joint[count];
	  Dev->SetJoint (J, Joint.GetNumber());

	  RTT = Next;
	  RPY rpy = GEOGetRPY (RTT);
	  if (List.Size() == 1)
	    PRGPrintf ("TCP: %.2f %.2f %.2f [mm]  %.2f %.2f %.2f",  
		       rpy.px, rpy.py, rpy.pz, rpy.y, rpy.p, rpy.r);

	  // Update mount plate of hexa-robots
	  if (strcmp(Dev->GetName(), "hexa") == 0 && Mount != NULL)
	  {
	    WTM = Mount->GetMatrix ();
	    if (Action == DATA::TCP_TRANS_REL || Action == DATA::TCP_ROT_REL)
	      WTT = GEOMatrixProd (WTM, Rel);
	    else
	    {
	      WTT = GEOMatrixProd(Rel, WTM);
	      if (Action == DATA::TCP_ROT_ABS)
	      {
		WTT.m[P][X] = WTM.m[P][X];
		WTT.m[P][Y] = WTM.m[P][Y];
		WTT.m[P][Z] = WTM.m[P][Z];
	      }
	    }
	    Mount->SetMatrix (WTT);
	  }
	}
	catch (KINEMATIC::OutOfWorkspace OOW)
	{
	  PRGPrintf ("OOW: [%x] Pos: %.2f %.2f %.2f [mm]", OOW.ErrorJoint, 
		     Next.m[P][X], Next.m[P][Y], Next.m[P][Z]);
	}
	if (J != NULL)
	  delete (J);
      }
    }
    break;
    
    default :
      printf ("Invalid action: %d\n", Action);
      return FALSE;
  }
  return TRUE;
}


static void SelectRegion (WORLD *World, int Button, int MouseX, int MouseY)
{
  WINDOW3D *Win = World->GetWin ();

  XSetForeground (Win->GetDisplay(), xorGC, 0xffffffff);
  XDrawRectangle (Win->GetDisplay(), Win->GetWindow(), 
		  xorGC, MIN(M1x,OldRectX), MIN(M1y,OldRectY),
		  abs(OldRectX-M1x), abs(OldRectY-M1y));

  OldRectX = MouseX;
  OldRectY = MouseY;

  XDrawRectangle (Win->GetDisplay(), Win->GetWindow(),
		  xorGC, MIN(M1x,OldRectX), MIN(M1y,OldRectY),
		  abs(OldRectX-M1x), abs(OldRectY-M1y));
}


bool WorldMouseHandle (WORLD *World, int Button, int MouseX, int MouseY)
{
  int Dist;
  int OldX, OldY;

  if ((Button & (Button1Mask | Button2Mask | Button3Mask)) ||
      GetAction(World) == DATA::FLY_MODE)
  {
    if (GetAction(World) == DATA::NORMAL_MODE)
    {
      SelectRegion (World, Button, MouseX, MouseY);
      return FALSE;
    }

    OldX = GetMouseX (World);
    OldY = GetMouseY (World);
    if (abs(MouseX - OldX) > abs(MouseY - OldY))
      Dist = MouseX - OldX;
    else
      Dist = MouseY - OldY;

    if (Dist == 0)
      return FALSE;

    SetMouse (World, MouseX, MouseY);

    switch (GetAction(World))
    {
      // ************************ TRANSLATE *****************************
      case DATA::TRANS_ABSOLUTE :
	TranslateWorld (World, Button, Dist);
	break;
      case DATA::TRANS_RELATIVE :
	TranslateWorld (World, Button, Dist);
	break;

      // ************************* ROTATE *******************************
      case DATA::ROT_ABSOLUTE :
	RotateWorld (World, Button, Dist);
	break;
      case DATA::ROT_RELATIVE :
	RotateWorld (World, Button, Dist);
	break;

      // ************************* PANNING ******************************
      case DATA::PAN_MODE :
	PanWorld (World, Button, MouseX-OldX, MouseY-OldY);
	break;
      case DATA::FLY_MODE :
	FlyWorld (World, Button, MouseX-OldX, MouseY-OldY);
	break;

      // ************************** ZOOM ********************************
      case DATA::ZOOM_INTERACTIVE:
	if (! ZoomWorld(World, Button, MouseX, MouseY, Dist))
	  return FALSE;
	break;

      default :
	return FALSE;
    }
  }

  return TRUE;
}


void ButtonHandle (WORLD *World, int Button, int MouseX, int MouseY)
{
  DATA::ACTION Action = GetAction (World);

  if (Button == 0)
    return;

  if (Action == DATA::ZOOM_INTERACTIVE || Action == DATA::NORMAL_MODE)
  {
    switch (Button)
    {
      case 1 :                // start region select
      case 2 :
	PRGPrintf ("Drag Region...");

	M1x = OldRectX = MouseX;
	M1y = OldRectY = MouseY;
	break;

      case -1 :               // mouse button released (selection finished)
      case -2 :
	{
	  int Swap, Hits, count;

	  MouseX = OldRectX;
	  MouseY = OldRectY;

	  if (MouseX < M1x)
	  {
	    Swap = M1x; M1x = MouseX; MouseX = Swap;
	  }
	  if (MouseY < M1y)
	  {
	    Swap = M1y; M1y = MouseY; MouseY = Swap;
	  }

          PRO4 ("Select: %d / %d - %d / %d\n", M1x, M1y, MouseX, MouseY)
          if (GetAction(World) == DATA::ZOOM_INTERACTIVE)
	    World->ZoomView (M1x, M1y, MouseX, MouseY);
	  else                // region selection
	  {
	    WINDOW3D *Win = World->GetWin ();
	    XDrawRectangle (Win->GetDisplay(), Win->GetWindow(),
			    xorGC, MIN(M1x,OldRectX), MIN(M1y,OldRectY),
			    abs(OldRectX-M1x), abs(OldRectY-M1y));

	    PART3D **PartList = NULL;
	    Hits = World->SearchParts (M1x, M1y, MouseX, MouseY, &PartList);
	    if (Hits > 0)
	    {
	      if (Button == -1)
		for (count = 0; count < Hits; ++count)
		  SwapSelection (World, PartList[count]);
	      else
		// TODO: Wenn mehrere Teile eines Devices selektiert wurden
		//       sollte das Device nicht stndig geswapped werden.
		//       => bereits behandelte Devs merken !!!
		for (count = 0; count < Hits; ++count)
		  SwapSelection (World, PartList[count]->GetParent());

	      delete PartList;
	    }
	  }
	}
	break;

      default :
	break;
    }
  }
  else
  {
    switch (GetAction(World))
    {
      case DATA::ZOOM_DISCRETE:
	if (Button == 1 || Button == 3)
        {
	  World->SetScale ((Button == 1)? 0.5 : 2.0);
	  World->Redraw ();               // Update current view only
	  glXWaitGL ();                   // Suppress delayes drawings
	}
	break;

      default :
	if (KeyMask & KEY_CTRL)
	  switch (Button)                 // Multi-Select Part/Device
	  {
	    case 1 : 
	      {
		PART3D *Part;

		if ((Part = ChoosePart("Select part...",
				       World, MouseX, MouseY)) != NULL)
		  SwapSelection (World, Part);
	      }
	      break;

	    case 2 :
	      {
		DEVICE3D *Dev;

		if ((Dev = ChooseDev("Select device...",
				     World, MouseX, MouseY)) != NULL)
		  SwapSelection (World, Dev);
	      }
	      break;

	    default :
	      break;
	  }
    }
  }
}


void KeyHandle (WORLD *World, KeySym Key, int Modifier)
{
  if (Modifier >= 0 && (Modifier & 1) == 0)       // button pressed 
    switch (Key)
    {
      case XK_Escape :
	SetAction        (World, DATA::NORMAL_MODE);
	ClearSelection   (World);
	World->Redraw    ();
        break;

      case XK_Shift_L :
      case XK_Shift_R :
	KeyMask |= KEY_SHIFT;
	break;

      case XK_Control_L :
	KeyMask |= KEY_CTRL;
	break;

      case XK_F1 :
	Manual ();
	break;

      case XK_Print :
	World->Print ();
	break;

      default :
	break;
    }
  elsif (Modifier < 0)                            // button released
    switch (Key)
    {
      case XK_Shift_L :
      case XK_Shift_R :
	KeyMask &= ~KEY_SHIFT;
	break;

      case XK_Control_L :
	KeyMask &= ~KEY_CTRL;
	break;

      default :
	break;
    }
}


void DestroyHandle (WORLD *World)
{
  if (DLGOpenTextDialog("destroyWorld", DlgQUESTION,
			MSG_ASK_DELETE_WORLD, 2, MSG_NO, MSG_YES) == 2)
    DeleteWorld (World);
}


void MouseHandle (WORLD *World, int Button, int MouseX, int MouseY)
{
  bool              changed = FALSE;
  double            px, py, pz;
  PART3D           *Ptr;
  LIST<PART3D *>   *Parts;
  LIST<DEVICE3D *> *Devs;

  if ((Button & (Button1Mask | Button2Mask | Button3Mask)) || 
      GetAction(World) == DATA::FLY_MODE)
  {
    if (MouseX != GetMouseX(World) || MouseY != GetMouseY(World))
    {
      if (KeyMask & KEY_CTRL)
	WorldMouseHandle (World, Button, MouseX, MouseY);
      elsif ((Parts = GetPartBuffer(World)) != NULL && Parts->Size() > 0)
	changed = PartMouseHandle (World, *Parts, Button, MouseX, MouseY);
      elsif ((Devs = GetDevBuffer(World)) != NULL && Devs->Size() > 0)
	changed = DevMouseHandle (World, *Devs, Button, MouseX, MouseY);
      else                                        // Nothing selected
	changed = WorldMouseHandle (World, Button, MouseX, MouseY);

      if (changed)
      {
	World->Redraw ();                         // update current view 
	glXWaitGL ();                             // prevent delayed drawings
	XSync (XDisplay, True);
      }
    }
  }
  else
  {
    PRGPrintf (MSG_MOUSE, MouseX, MouseY);
    if (TOOLGetConfigBool(CFG_SHOW_COORDS))
    {
      // Show optional coordinates. May cause performance problems on large
      // workcells
      if (MouseX > 0 && MouseY > 0 &&
	  (Ptr = World->GetPart(MouseX, MouseY, &px, &py, &pz)) != NULL)
	PRGPrintf ("\\ Part: '%s' Pos: %.3f %.3f %.3f", 
		   Ptr->GetName(), px, py, pz);
    }
    SetMouse (World, MouseX, MouseY);
  }
}


void ClickHandle (WORLD *World, int Button, int MouseX, int MouseY)
{
  // We do not recognize, if a key was pressed or released while a menu is 
  // open. So, lets update the keyboad state here...
  GetKeyMask ();

  switch (Button)
  {
    case 1 :
      {
	PART3D *Part;
  
	if (KeyMask & KEY_CTRL)
	  if ((Part = World->GetPart(MouseX, MouseY)) != NULL)
	  {
	    printf ("Selected: %s\n", Part->GetName());
	    SwapSelection (World, Part);
	  }
      }
      break;

    case 2 :
      if (KeyMask & KEY_CTRL)
      {
	DEVICE3D *Dev;
	
	if ((Dev = World->GetDev(MouseX, MouseY)) != NULL)
	{
	  printf ("Selected: %s\n", Dev->GetName());
	  SwapSelection (World, Dev);
	}
      }
      break;

    default :
      {
        LIST<PART3D *> *Parts;
	LIST<DEVICE3D *> *Devs;

	if ((Parts = GetPartBuffer(World)) != NULL && Parts->Size() > 0)
          OpenPartMenu (World->GetShell(), *Parts, MouseX, MouseY);
	elsif ((Devs = GetDevBuffer(World)) != NULL && Devs->Size() > 0)
	  OpenDevMenu (World->GetShell(), *Devs, MouseX, MouseY);
      }
  }
}


void WorldSelectHandle (WORLD *World, int Button, int MouseX, int MouseY)
{
  SetCurrent (World);
  GetKeyMask ();

  switch (Button)
  {
    case 1 :
      {
	PART3D         *Part;
	LIST<PART3D *> *Parts;

        if ((Part = ChoosePart(MSG_SELECT_PART,
			       World, MouseX, MouseY)) == NULL)
	{
	  PRGPrintf (MSG_NO_PART);
	  SetAction (World, DATA::NORMAL_MODE);
	}
	else
	{
	  PRGPrintf (MSG_SELECTED_PART, 
		     Part->GetParent()->GetName(), Part->GetName());
	  if (! (KeyMask & KEY_CTRL))
	    ClearSelection (World);

	  SwapSelection (World, Part);

	  Parts = GetPartBuffer (World);
	  OpenPartMenu (World->GetShell(), *Parts, MouseX, MouseY);
	}
      }
      break;

    case 2 :
      {
	PART3D           *Mount;
	DEVICE3D         *Dev;
	LIST<DEVICE3D *> *Devs;

	PRO0 ("Select Device\n")
        if (/* (Dev = World->GetDev(MouseX, MouseY)) == NULL  && */
	    (Dev = ChooseDev(MSG_SELECT_DEV, 
	    World, MouseX, MouseY)) == NULL)
	  PRGPrintf (MSG_ABORT);
	else
	{
	  Mount = Dev->GetMount ();
	  PRGPrintf (MSG_SELECTED_MOUNT,
		     Dev->GetName(), (Mount==NULL)? "-":(Mount->GetName()));
	  if (! (KeyMask & KEY_CTRL))
	    ClearSelection (World);

	  SwapSelection (World, Dev);

	  Devs = GetDevBuffer (World);
	  OpenDevMenu (World->GetShell(), *Devs, MouseX, MouseY);
	}
      }
      break;

    case 3 :
      {
	PRGPrintf (MSG_SELECTED_WORLD, World->GetName());
	ClearSelection (World);
	OpenWorldMenu (World->GetShell(), World, MouseX, MouseY);
      }
      break;

    default :
      break;
  }
}


void SetWorldHandles (WORLD *World)
{
  World->SetUserMouseHandle   ((MOUSE_HANDLE)  MouseHandle);
  World->SetUserButtonHandle  ((BUTTON_HANDLE) ButtonHandle);
  World->SetUserSelectHandle  ((SELECT_HANDLE) WorldSelectHandle);  
  World->SetUserClickHandle   ((CLICK_HANDLE)  ClickHandle);
  World->SetUserKeyHandle     ((KEY_HANDLE)    KeyHandle);
  World->SetUserDestroyHandle ((DESTROY_HANDLE)DestroyHandle);
}
