#include "tuxCAD.h"

#include <icons/penguin.xpm>


HELP_DIALOG HelpDlg;


void ShowSyntax ()
{
  // Show program syntax
  printf ("##################################"
	  "   TuxCAD V%d.%d 2001-2003\n", VER, REV);
  printf (" TTTTT             CCC    A   DDD  \n");
  printf ("   T              C   C  A A  D  D \n");
  printf ("   T  u  u  x  x C      A   A D   D"
	  "  Simple 3D-CAD and simulation program\n"); 
  printf ("   T  u  u   xx  C      AAAAA D   D\n");
  printf ("   T  u  u   xx   C   C A   A D  D "
	  "  written by Stefan Vietinghoff\n");
  printf ("   T   uu u x  x   CCC  A   A DDD  "
	  "  Nuremberg, Germany\n");
  printf ("##################################\n");
  printf ("\n");
  printf ("Syntax:\n"
	  "  tuxCAD [-cfg Config-File]\n"
	  "         [-s PortNo] [-h hostname]\n"
	  "         [-pro Protocol-File]\n"
	  "         [part-file | dev-file | world-file]\n");
  printf ("    -cfg     Load configuration data file\n"
	  "    -s       Start as CLI-Server\n"
	  "    -pro     creates the named protocol file\n"
	  "    file(s)  load and display the specified data\n");
}


void Manual ()
{
  // Open help window without special label
  HelpDlg.Open ("Manual", "tuxCAD.hlp", "");
}


void About ()
{
  char Message[256];

  // VER and REV will be defined in the Makefile
  sprintf (Message, "              TuxCAD\n"
                    "      Version %d Revision %d\n"
	            "\n"
                    "  OpenGL-based CAD- and Simulation-\n"
	            "              program\n"
	      	    "\n"
	            "   Copyright: Stefan Vietinghoff\n"
	            "          D-90427 Nrnberg\n"
 	            "       EMail: stefan@tuxcad.de", 
	            VER, REV);

  // Display a info dialog 
  DLGSetDialogPos (0.4, 0.45);
  DLGOpenPicDialog ("INFO", DlgXPM, -1, -1, penguin_xpm, Message, 0);
}


HELP_DIALOG::HELP_DIALOG () :
  SizeX   (400),
  SizeY   (300),
  PosX    (0),
  PosY    (0),
  HelpWin (NULL)
{  
}


bool HELP_DIALOG::Open (const char *Title, 
			const char *FileSpec, const char *Label)
{
  char HelpLoc[PATH_MAX+1];

  PRO_METHOD ("HELP_DIALOG", "\n")

  if (Label == NULL || Label[0] == '\0')
#if 0
    sprintf (HelpLoc, "%s/help/%s", 
	              TOOLGetConfigString(CFG_TUXCAD_PATH), FileSpec);
#else
  sprintf (HelpLoc, "/users/stefan/project/help/tuxCAD/%s", FileSpec);
#endif
	   
  else
    sprintf (HelpLoc, "%s/help/%s#%s",
	              TOOLGetConfigString(CFG_TUXCAD_PATH), FileSpec, Label);

  printf ("Opening '%s'\n", HelpLoc);

  if (HelpWin != NULL)
    HELPLoad (HelpWin, HelpLoc, NULL, 0);
  else
    HelpWin = HELPOpenHelpWin (Title, HelpLoc, PosX, PosY, SizeX, SizeY, 
			       NULL, 0, CloseHelp);

  PRO_END
  return TRUE;
}


void HELP_DIALOG::Close ()
{
  if (HelpWin == NULL)
    return;

  // Save help windows position and size
  HELPGetPos  (HelpWin, &PosX, &PosY);
  HELPGetSize (HelpWin, &SizeX, &SizeY);

  HELPClose (HelpWin);
  HelpWin = NULL;
}  


void CloseHelp (HELPWIN *HelpWin)
{
  HelpDlg.Close ();
}


