#include "tuxCAD.h"


void PartInfo (PART3D *Part)
{
  int       count, Axes;
  char      AxisString[6][20] = {"", "", "", "", "", ""};
  RPY       WTP;
  COLOR     Color;
  PART     *Ptr = Part->GetPart();
  DEVICE3D *Dev;
  DIALOG   *Dlg;
 
  Dev   = Part->GetParent ();
  WTP   = GEOGetRPY (Part->GetWTP()); 
  Color = Part->GetMaterial()->GetColor();

  Dlg = DLGNewDialog (10);
  DLGMakeGrid (Dlg,  10, 2, 14);

  // Label:
  DLGInsertLabel (Dlg, DlgGRID(10,0, 0), "name",       "Name:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 1), "device",     "Device:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 2), "wtp",        "WTP:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 3), "material",   "Material:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 4), "color",      "Color:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 5), "polys",      "# Polygons:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 6), "points",     "# Points:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 7), "triangles:", "# Triangles:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 8), "whd:",       "W/H/D:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 9), "volume",     "Volume:");
  DLGInsertLabel (Dlg, DlgGRID(10,0,10), "surface",    "Surface:");
  DLGInsertLabel (Dlg, DlgGRID(10,0,11), "center",     "Center:");
  DLGInsertLabel (Dlg, DlgGRID(10,0,12), "radius",     "Radius:");
  DLGInsertLabel (Dlg, DlgGRID(10,0,13), "dof",        "DOF:");

  // Values:
  DLGInsertLabel (Dlg, DlgGRID(10,1, 0), "nameVal",   Part->GetName());
  DLGInsertLabel (Dlg, DlgGRID(10,1, 1), "deviceVal", 
		                        (Dev != NULL)? Dev->GetName() : "---");
  DLGInsertLabel (Dlg, DlgGRID(10,1, 2), "wtpVal", 
		  "%.3f[mm] %.3f[mm] %.3f[mm]   %7.2f %7.2f %7.2f",
		  WTP.px, WTP.py, WTP.pz, WTP.y, WTP.p, WTP.r);
  DLGInsertLabel (Dlg, DlgGRID(10,1, 3), "materialVal", " ");
  DLGInsertLabel (Dlg, DlgGRID(10,1, 4), "colorVal",
		  "Name:   %-38s\n"
		  "Values: R=%.4f  G=%.4f  B=%.4f  A=%.4f",
		  ((Color.GetName() != NULL)? Color.GetName() : "---"),
		  Color[RED], Color[GREEN], Color[BLUE], Color[ALPHA]);
  DLGInsertLabel (Dlg, DlgGRID(10,1, 5), "polysVal",  
		                         "%d", PARTPolygonAmount(Ptr));
  DLGInsertLabel (Dlg, DlgGRID(10,1, 6), "pointsVal", 
		                         "%d", PARTPointAmount(Ptr));
  DLGInsertLabel (Dlg, DlgGRID(10,1, 7), "trianglesVal",
		                         "%d", PARTTriangleAmount(Ptr));
  DLGInsertLabel (Dlg, DlgGRID(10,1, 8), "whdVal",
		                         "%.2f [mm] %.2f [mm] %.2f [mm]",
 		                         (Ptr->Info.Max.x - Ptr->Info.Min.x),
		                         (Ptr->Info.Max.y - Ptr->Info.Min.y),
 		                         (Ptr->Info.Max.z - Ptr->Info.Min.z));
  DLGInsertLabel (Dlg, DlgGRID(10,1, 9), "volumeVal",  
		                         "%.4f [mm]", Ptr->Info.Volume);
  DLGInsertLabel (Dlg, DlgGRID(10,1,10), "surfaceVal", 
		                         "%.4f [mm]", Ptr->Info.Area);
  DLGInsertLabel (Dlg, DlgGRID(10,1,11), "centerVal", 
		                         "%.3f [mm] %.3f [mm] %.3f [mm]", 
		                         Ptr->Info.Center.x, 
		                         Ptr->Info.Center.y, 
		                         Ptr->Info.Center.z);
  DLGInsertLabel (Dlg, DlgGRID(10,1,12), "radiusVal", 
		                         "%.2f [mm]", Ptr->Info.Bubble);
  Axes = Part->GetJoint().GetAxes();
  if (Axes & (1<<POS_X))
    sprintf (AxisString[0], "  TransX=%.2f", Part->GetVal(POS_X));
  if (Axes & (1<<POS_Y))
    sprintf (AxisString[1], "  TransY=%.2f", Part->GetVal(POS_Y));
  if (Axes & (1<<POS_Z))
    sprintf (AxisString[2], "  TransZ=%.2f", Part->GetVal(POS_Y));
  if (Axes & (1<<ROT_X))
    sprintf (AxisString[3], "  RotX=%.2f",   Part->GetVal(ROT_X));
  if (Axes & (1<<ROT_Y))
    sprintf (AxisString[4], "  RotY=%.2f",   Part->GetVal(ROT_Y));
  if (Axes & (1<<ROT_Z))
    sprintf (AxisString[5], "  RotZ=%.2f",   Part->GetVal(ROT_Z));
  DLGInsertLabel (Dlg, DlgGRID(10,1,13), "dofVal", "%s%s%s%s%s%s", 
		  AxisString[0], AxisString[1], AxisString[2],
		  AxisString[3], AxisString[4], AxisString[5]);

  for (count = 0; count < 14; ++count)
    DLGSetBorder (Dlg, DlgGRID(10,1,count), DlgETCHED_IN);
  DLGSetLabelPos (Dlg, 10, DlgNOWHERE, DlgWEST);
  
  DLGSetTitle (Dlg, 10, "partInfo", DlgMIDDLE);

  DLGMakeButton (Dlg, OK_ID, "ok", 
		      DlgBUTTON_MIDDLE | DlgBUTTON_FILL, NULL, DlgOK);
  DLGOpenDialog (Dlg, "partInfo", DlgNORMAL, NULL);
  DLGFreeDialog (Dlg);
}


void DevInfo (DEVICE3D *Dev)
{
  int       count, Rows = 3;
  int       JointNo;
  double   *JointVec;
  char      JointVal[8];
  char     *JointStr;
  RPY       TCP, WTD, Offset;
  FRAME     Base;
  PART3D   *Mount;
  DEVICE3D *Tool = NULL;
  ROBOT    *Robot;
  DIALOG   *Dlg;

  if ((Robot = Dev->GetRobot()) != NULL)
    Rows += 8;

  WTD = GEOGetRPY (Dev->GetWTD());

  Dlg = DLGNewDialog (10);
  DLGMakeGrid (Dlg, 10, 2, Rows);

  // Labels:
  DLGInsertLabel (Dlg, DlgGRID(10,0, 0), "name",    "Name:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 1), "parts",   "# Parts:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 2), "wtd",     "WTD:");

  // Values:
  DLGInsertLabel (Dlg, DlgGRID(10,1, 0), "nameVal",  Dev->GetName());
  DLGInsertLabel (Dlg, DlgGRID(10,1, 1), "partsVal", "%d", Dev->GetPartNo());
  DLGInsertLabel (Dlg, DlgGRID(10,1, 2), "wtdVal", 
		  "%.3f[mm] %.3f[mm] %.3f[mm]   %7.2f %7.2f %7.2f",
		  WTD.px, WTD.py, WTD.pz, WTD.y, WTD.p, WTD.r);

  if (Robot != NULL)
  {
    TCP    = GEOGetRPY (Dev->GetTCP());
    Offset = GEOGetRPY (Dev->GetOffset());
    if ((Mount = Dev->GetMount()) != NULL)
      Tool = Mount->GetAttach ();
    Base    = Robot->GetBase ();
    JointNo = Dev->GetJoints (&JointVec);

    // Labels:
    DLGInsertLabel (Dlg, DlgGRID(10,0, 3), "base",    "Base:");
    DLGInsertLabel (Dlg, DlgGRID(10,0, 4), "offset",  "Offset:");
    DLGInsertLabel (Dlg, DlgGRID(10,0, 5), "tcp",     "TCP:");
    DLGInsertLabel (Dlg, DlgGRID(10,0, 6), "mount",   "Mount:");
    DLGInsertLabel (Dlg, DlgGRID(10,0, 7), "tool",    "Tool:");
    DLGInsertLabel (Dlg, DlgGRID(10,0, 8), "joints",  "# Joints:");
    DLGInsertLabel (Dlg, DlgGRID(10,0,10), "dhp",     "DHP:");

    // Values:
    DLGInsertLabel (Dlg, DlgGRID(10,1, 3), "baseVal", "%.2f %.2f %.2f",
		                           Base[P][X], Base[P][Y], Base[P][Z]);
    DLGInsertLabel (Dlg, DlgGRID(10,1, 4), "offsetVal",  
		         "%.3f[mm] %.3f[mm] %.3f[mm]  %7.2f %7.2f %7.2f",
		         Offset.px, Offset.py, Offset.pz, 
		         Offset.y,  Offset.p,  Offset.r);
    DLGInsertLabel (Dlg, DlgGRID(10,1, 5), "tcpVal", 
		         "%.3f[mm] %.3f[mm] %.3f[mm]  %7.2f %7.2f %7.2f",
		         TCP.px, TCP.py, TCP.pz,  TCP.y, TCP.p, TCP.r);
    DLGInsertLabel (Dlg, DlgGRID(10,1, 6), "mountVal",
		         (Mount != NULL)? Mount->GetName() : "---");
    DLGInsertLabel (Dlg, DlgGRID(10,1, 7), "toolVal",  
		         (Tool != NULL)? Tool->GetName() : "---");
    DLGInsertLabel (Dlg, DlgGRID(10,1, 8), "jointsVal", "%d", JointNo);

    JointStr = (char *)malloc (sizeof(char) * JointNo * 20 * 3);
    strcpy (JointStr, " No:    ");
    for (count = 0; count < JointNo; ++count)
    {
      sprintf (JointVal, "%7d ", count);
      strcat (JointStr, JointVal);
    }

    strcat (JointStr, "\n Type:  ");
    for (count = 0; count < JointNo; ++count)
    {
      sprintf (JointVal, "%7s ", "Rot");
      strcat (JointStr, JointVal);
    }

    strcat (JointStr, "\n Value:  ");
    for (count = 0; count < JointNo; ++count)
    {
      sprintf (JointVal, "%7.2f ", JointVec[count]);
      strcat (JointStr, JointVal);
    }

    DLGInsertLabel (Dlg, DlgGRID(10,1, 9), "jointNoVal", JointStr);

    free (JointStr);

    DLGInsertLabel (Dlg, DlgGRID(10,1,10), "dhpVal", "DHP-Parameter...");
  }

  for (count = 0; count < 11; ++count)
    DLGSetBorder (Dlg, DlgGRID(10,1,count), DlgETCHED_IN);
  DLGSetLabelPos (Dlg, 10, DlgNOWHERE, DlgWEST);

  DLGSetTitle (Dlg, 10, "deviceInfo", DlgMIDDLE);

  DLGMakeButton (Dlg, OK_ID, "ok", 
		      DlgBUTTON_MIDDLE | DlgBUTTON_FILL, NULL, DlgOK);
  DLGOpenDialog (Dlg, "devInfo", DlgNORMAL, NULL);
  DLGFreeDialog (Dlg);
}


void TriangleCount (DEVICE3D *Dev, PART3D *Part, void *Ptr)
{
  if (Part != NULL)
    *((int *)Ptr) += PARTTriangleAmount (Part->GetPart());
}


void WorldInfo (WORLD *World)
{
  int     count; 
  int     Triangles = 0;
  int     LightNo = 0;
  int     ViewNo = 0;
  int     WinX, WinY;
  double  Range[6];
  LIGHT  *Light;
  VIEW   *View;
  DIALOG *Dlg;

  World->GetSize (&Range[MIN_X], &Range[MAX_X],
		  &Range[MIN_Y], &Range[MAX_Y],
		  &Range[MIN_Z], &Range[MAX_Z]);
  World->GetWinSize (&WinX, &WinY);

  World->UserAction (TriangleCount, &Triangles);

  for (count = 0; count < 4; ++count)
    if ((View = World->GetView(count)) != NULL)
      ViewNo++;

  for (count = 0; count < 8; ++count)
    if ((Light = World->GetLight((LIGHT::NUMBER)count)) != NULL)
      if (Light->GetMode() != LIGHT::OFF)
	LightNo++;

  Dlg = DLGNewDialog (10);
  DLGMakeGrid (Dlg, 10, 2, 11);

  // Labels:
  DLGInsertLabel (Dlg, DlgGRID(10,0, 0), "name",      "Name:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 1), "window",    "Window:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 2), "devices",   "# Devices:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 3), "parts",     "# Parts:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 4), "views",     "# Views:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 5), "range",     "Range:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 8), "scale",     "Scale:");
  DLGInsertLabel (Dlg, DlgGRID(10,0, 9), "lights",    "# Lights:");
  DLGInsertLabel (Dlg, DlgGRID(10,0,10), "triangles", "# Triangles:");

  // Values:
  DLGInsertLabel (Dlg, DlgGRID(10,1, 0), "nameVal",   World->GetName());
  DLGInsertLabel (Dlg, DlgGRID(10,1, 1), "windowVal", "%d x %d", WinX, WinY);
  DLGInsertLabel (Dlg, DlgGRID(10,1, 2), "devicesVal", 
		                         "%d", World->GetDevNum());
  DLGInsertLabel (Dlg, DlgGRID(10,1, 3), "partsVal",  
		                         "%d", World->GetPartNum());
  DLGInsertLabel (Dlg, DlgGRID(10,1, 4), "viewsVal",  "%d", ViewNo);
  DLGInsertLabel (Dlg, DlgGRID(10,1, 5), "rangeXVal", "X: %.3f..%.3f [mm]", 
		                         Range[MIN_X], Range[MAX_X]);
  DLGInsertLabel (Dlg, DlgGRID(10,1, 6), "rangeYVal", "Y: %.3f..%.3f [mm]", 
		                         Range[MIN_Y], Range[MAX_Y]);
  DLGInsertLabel (Dlg, DlgGRID(10,1, 7), "rangeZVal", "Z: %.3f..%.3f [mm]", 
		                         Range[MIN_Z], Range[MAX_Z]);
  DLGInsertLabel (Dlg, DlgGRID(10,1, 8), "scaleVal", "%5.1f [%%]", 
		                         100.0*World->GetScale());
  DLGInsertLabel (Dlg, DlgGRID(10,1, 9), "lightsVal", "%d", LightNo);
  DLGInsertLabel (Dlg, DlgGRID(10,1,10), "trianglesVal", "%d", Triangles);

  for (count = 0; count < 11; ++count)
    DLGSetBorder (Dlg, DlgGRID(10,1,count), DlgETCHED_IN);
  DLGSetLabelPos (Dlg, 10, DlgNOWHERE, DlgWEST);

  DLGSetTitle (Dlg, 10, "worldInfo", DlgMIDDLE);

  DLGMakeButton (Dlg, OK_ID, "ok", 
		      DlgBUTTON_MIDDLE | DlgBUTTON_FILL, NULL, DlgOK);
  DLGOpenDialog (Dlg, "worldInfo", DlgNORMAL, NULL);
  DLGFreeDialog (Dlg);

  // #Lights, Projection, alpha, blending, Camera(s)

  //  World->Print ();
}
