#include <math.h>

#include "tuxCAD.h"


// TODO: Light-Placement, 
//       attenuation factor (OFF / CONSTANT + LINEAR + QUADRATIC) nicht fr
//          directional lights, nur positional + spot
//       Spot-Lights: Auto-Direction to Center of Objects / to origin,


static void ActivateLightDlg (DIALOG *Dlg, int BaseID, int Mode)
{
  DLGSetBoxActive (Dlg, BaseID+12, (Mode == 3 || Mode == 4));
  DLGSetBoxActive (Dlg, BaseID+20, (Mode == 3 || Mode == 4));
  DLGSetBoxActive (Dlg, BaseID+30, (Mode == 4));
  DLGSetBoxActive (Dlg, BaseID+41, (Mode == 4));
  DLGSetBoxActive (Dlg, BaseID+42, (Mode == 4));
  DLGSetBoxActive (Dlg, BaseID+50, (Mode > 1));
  DLGSetBoxActive (Dlg, BaseID+55, (Mode == 3 || Mode == 4));
}


LIGHT *LightCB (DIALOG *Dlg, int BaseID, LIGHT *Light, int BoxID)
{
  if (Light == NULL)
    return NULL;

  switch (BoxID-BaseID)
  {
    case 11 :
      {
	int Mode;

	if ((Mode = DLGGetOMenu(Dlg, BoxID)) == 0)
	  Light->SetMode (LIGHT::OFF);
	else
	  Light->SetMode ((LIGHT::MODE)(Mode-1));

	ActivateLightDlg (Dlg, BaseID, Mode);
      }
      break;

    case 12 :
      Light->SetPlacement ((LIGHT::PLACEMENT)DLGGetOMenu(Dlg, BoxID));
      break;

    case 13 :
      {
	// Sizes: 0, 0.125, 0.25, 0.5, 1.0, 2.0, 4.0, 8.0, 16.0, 32.0, ...
	double Size = 0.125 * pow (2.0, DLGGetOMenu(Dlg, BoxID));

	Light->SetSize (Size);
      }
      break;

    case 21 :
    case 22 :
    case 23 :
      {
	POINT3D Pos;

	Pos.x = DLGGetDouble (Dlg, BaseID+21);
	Pos.y = DLGGetDouble (Dlg, BaseID+22);
	Pos.z = DLGGetDouble (Dlg, BaseID+23);
	Light->SetPosition (Pos);
      }
      break;

    case 31 :
    case 32 :
    case 33 :
      {
	VECTOR3D Dir;

	Dir.x = DLGGetDouble (Dlg, BaseID+31);
	Dir.y = DLGGetDouble (Dlg, BaseID+32);
	Dir.z = DLGGetDouble (Dlg, BaseID+33);
	Light->SetDirection (Dir);
      }
      break;

    case 41 :
      Light->SetCutOff (DLGGetScale(Dlg, BoxID));
      break;

    case 42 :
      Light->SetSpotExponent (DLGGetScale(Dlg, BoxID));
      break;

#if 0
    case 43 :
      Light->GetShadow()->SetMode ((SHADOW::MODE)DLGGetOMenu(Dlg, BoxID));
      break;
#endif

    case 50 :
      Light->SetIntensity (DLGGetScale(Dlg, BoxID));
      break;

    case 56 :
    case 57 :
    case 58 :
      Light->SetAttenuation (DLGGetDouble(Dlg, BaseID+56),
			     DLGGetDouble(Dlg, BaseID+57),
			     DLGGetDouble(Dlg, BaseID+58));
      break;

    default :
      if (BoxID-BaseID < 70)
	Light->SetAmbient (ColorCB(Dlg, BaseID+60, Light->GetAmbient(),  
				                   BoxID, ALPHA_MASK));
      elsif (BoxID-BaseID < 80)
	Light->SetDiffuse (ColorCB(Dlg, BaseID+70, Light->GetDiffuse(),
				                   BoxID, ALPHA_MASK));
      else
	Light->SetSpecular (ColorCB(Dlg, BaseID+80, Light->GetSpecular(), 
				                    BoxID, ALPHA_MASK));
  }

  return Light;
}


void SetLight (DIALOG *Dlg, int BaseID, LIGHT *Light)
{
  int       Mode = 0;
  double   *Attenuation;
  VECTOR3D  Vec;

  if (Light == NULL)
    return;

  if ((Mode = int(Light->GetMode())) > 0)
    Mode++;

  DLGSetOMenu (Dlg, BaseID+11, Mode);
  DLGSetOMenu (Dlg, BaseID+12, (int)Light->GetPlacement());
  DLGSetOMenu (Dlg, BaseID+13, (int)Light->GetSize());

  Vec = Light->GetPosition ();
  DLGSetDouble (Dlg, BaseID+21, Vec.x);
  DLGSetDouble (Dlg, BaseID+22, Vec.y);
  DLGSetDouble (Dlg, BaseID+23, Vec.z);

  Vec = Light->GetDirection ();
  DLGSetDouble (Dlg, BaseID+31, Vec.x);
  DLGSetDouble (Dlg, BaseID+32, Vec.y);
  DLGSetDouble (Dlg, BaseID+33, Vec.z);

  DLGSetScale (Dlg, BaseID+41, Light->GetCutOff());
  DLGSetScale (Dlg, BaseID+42, Light->GetSpotExponent());
#if 0
  DLGSetOMenu (Dlg, BaseID+43, (Light->GetShadow()->GetMode()==SHADOW::ON));
#endif

  DLGSetScale (Dlg, BaseID+50, Light->GetIntensity());

  Attenuation = Light->GetAttenuation ();
  DLGSetDouble (Dlg, BaseID+56, Attenuation[0]);
  DLGSetDouble (Dlg, BaseID+57, Attenuation[1]);
  DLGSetDouble (Dlg, BaseID+58, Attenuation[2]);

  ActivateLightDlg (Dlg, BaseID, Mode);

  SetColor (Dlg, BaseID+60, Light->GetAmbient(),  ALPHA_MASK);
  SetColor (Dlg, BaseID+70, Light->GetDiffuse(),  ALPHA_MASK);
  SetColor (Dlg, BaseID+80, Light->GetSpecular(), ALPHA_MASK);
}


void LightDialog (DIALOG *Dlg, int BaseID, LIGHT *Light)
{
  int       Mode = 0;
  int       Size;
  double   *Attenuation;
  VECTOR3D  Vec;

  PRO_START ("")

  if (Light == NULL)
    PRO_ABORT ("NULL-pointer")

  DLGMakeRows (Dlg, BaseID,    3, DlgETCHED_IN DlgNO_LINE,  
	                          BaseID+1, BaseID+3, BaseID+5);
  DLGMakeCols (Dlg, BaseID+ 1, 4, DlgETCHED_IN DlgETCHED_IN DlgETCHED_IN,
		                  BaseID+10, BaseID+20, BaseID+30, BaseID+40);
  DLGMakeRows (Dlg, BaseID+10, 3, DlgNO_LINE, BaseID+11, BaseID+12, BaseID+13);
  DLGMakeRows (Dlg, BaseID+20, 3, DlgNO_LINE, BaseID+21, BaseID+22, BaseID+23);
  DLGMakeRows (Dlg, BaseID+30, 3, DlgNO_LINE, BaseID+31, BaseID+32, BaseID+33);
#if 1
  DLGMakeRows (Dlg, BaseID+40, 2, DlgNO_LINE, BaseID+41, BaseID+42);
#else
  DLGMakeRows (Dlg, BaseID+40, 3, DlgNO_LINE, BaseID+41, BaseID+42, BaseID+43);
#endif
  DLGMakeCols (Dlg, BaseID+ 3, 2, DlgNO_LINE, BaseID+50, BaseID+55);
  DLGMakeCols (Dlg, BaseID+55, 3, DlgNO_LINE, BaseID+56, BaseID+57, BaseID+58);
  DLGMakeCols (Dlg, BaseID+ 5, 3, DlgNO_LINE, BaseID+60, BaseID+70, BaseID+80);

  DLGSetBorder (Dlg, BaseID+60, DlgETCHED_IN);
  DLGSetBorder (Dlg, BaseID+70, DlgETCHED_IN);
  DLGSetBorder (Dlg, BaseID+80, DlgETCHED_IN);

  DLGSetTitle (Dlg, BaseID+60, "ambient",  DlgLEFT);
  DLGSetTitle (Dlg, BaseID+70, "diffuse",  DlgLEFT);
  DLGSetTitle (Dlg, BaseID+80, "specular", DlgLEFT);

  if ((Mode = int(Light->GetMode())) > 0)
    Mode++;
  DLGInsertOMenu  (Dlg, BaseID+11, "lightMode", 5, 
		                   "off", NULL, "directional", "positional", 
		                   "spot", Mode);
  DLGInsertOMenu  (Dlg, BaseID+12, "place", 2, "stationary", "relative",
		                   (int)Light->GetPlacement());
  if (Light->GetSize() == 0.0)
    Size = 0;
  else
    Size = int (log(8.0*Light->GetSize()) / log(2.0));
  DLGInsertOMenu  (Dlg, BaseID+13, "show", 11, "hide", 
		                   "size1", "size2", "size3", "size4", 
		                   "size5", "size6", "size7", "size8", 
		                   "size9", "size10",
		                   Size);

  Vec = Light->GetPosition ();
  DLGInsertDouble (Dlg, BaseID+21, "posX", Vec.x, 8, 2);
  DLGInsertDouble (Dlg, BaseID+22, "posY", Vec.y, 8, 2);
  DLGInsertDouble (Dlg, BaseID+23, "posZ", Vec.z, 8, 2);

  Vec = Light->GetDirection ();
  DLGInsertDouble (Dlg, BaseID+31, "dirX", Vec.x, 8, 2);
  DLGInsertDouble (Dlg, BaseID+32, "dirY", Vec.y, 8, 2);
  DLGInsertDouble (Dlg, BaseID+33, "dirZ", Vec.z, 8, 2);

  DLGInsertScale (Dlg, BaseID+41, "cutOff", 0.0, 90.0, 
		                  Light->GetCutOff(), 1, DlgHORIZONTAL);
  DLGInsertScale (Dlg, BaseID+42, "spotExp", 0.0, 10.0, 
		                  Light->GetSpotExponent(), 1, DlgHORIZONTAL);
#if 0
  DLGInsertOMenu (Dlg, BaseID+43, "shadow", 2, "off", "on", 
		                  (Light->GetShadow()->GetMode()==SHADOW::ON));
#endif

  DLGInsertScale  (Dlg, BaseID+50, "intensity", 0.0, 1.0, 
		                   Light->GetIntensity(), 2, DlgHORIZONTAL);

  Attenuation = Light->GetAttenuation ();
  DLGInsertDouble (Dlg, BaseID+56, "attenuation", Attenuation[0], 7, 2);
  DLGInsertDouble (Dlg, BaseID+57, "dd+",         Attenuation[1], 7, 2);
  DLGInsertDouble (Dlg, BaseID+58, "d+",          Attenuation[2], 7, 2);

  ActivateLightDlg (Dlg, BaseID, Mode);

  ColorDialog (Dlg, BaseID+60, Light->GetAmbient(),  ALPHA_MASK);
  ColorDialog (Dlg, BaseID+70, Light->GetDiffuse(),  ALPHA_MASK);
  ColorDialog (Dlg, BaseID+80, Light->GetSpecular(), ALPHA_MASK);

  PRO_END
}

