#include "tuxCAD.h"


template<class Type /* = void * */>
  LIST<Type>::LIST () :
    Number (0),
    List   (NULL),
    Last   (NULL)
{
}


template<class Type /* = void * */>
  LIST<Type>::~LIST ()
{
  Clear ();
}


template<class Type /* = void * */>
  void LIST<Type>::Add (const Type &Value)
{
  LISTELEM<Type> *New;

  if (Exist(Value))
    return;

  New = new LISTELEM<Type> (Value);
  New->Prev = Last;

  if (List == NULL)
    List = New;
  else
    Last->Next = New;
  Last = New;

  Number++;
}


template<class Type /* = void * */>
  void LIST<Type>::Delete (const Type &Value)
{
  LISTELEM<Type> *Ptr = List;

  if (Value == NULL)
    return;

  if (Ptr->Data == Value)
  {
    List = Ptr->Next;
    if (List == NULL)
      Last = NULL;
    else
      List->Prev = NULL;
  }
  else
  {
    while (Ptr != NULL && Ptr->Data != Value)
      Ptr = Ptr->Next;
    if (Ptr != NULL)                    // Element found...
    {
      Ptr->Prev->Next = Ptr->Next;
      if (Ptr->Next == NULL)
	Last = Ptr->Prev;
      else
	Ptr->Next->Prev = Ptr->Prev;
    }
  }

  if (Ptr != NULL)
  {
    delete (Ptr);
    Number--;
  }
}


template<class Type /* = void * */>
  bool LIST<Type>::Exist (const Type &Value) const
{
  LISTELEM<Type> *Ptr = List;

  while (Ptr != NULL && Ptr->Data != Value)
    Ptr = Ptr->Next;

  return (Ptr != NULL);
}


template<class Type /* = void * */>
  int LIST<Type>::Size () const
{
  return Number;
}


template<class Type /* = void * */>
  Type LIST<Type>::Get (int Index) const
{
  LISTELEM<Type> *Ptr = List;

  if (Index < 0 || Index >= Number)
    return NULL;

  while (Ptr != NULL && Index > 0)
  {
    Ptr = Ptr->Next;
    Index--;
  }

  return Ptr->Data;
}


template<class Type /* = void * */>
  void LIST<Type>::Clear ()
{
  LISTELEM<Type> *Ptr = List, *Next;

  while (Ptr != NULL)
  {
    Next = Ptr->Next;
    delete (Ptr);
    Ptr = Next;
  }

  List   = NULL;
  Last   = NULL;
  Number = 0;
}


// g++ does not automatically instantiate templates if they are defined in 
// separate files (e.g. header and implementation). We need to fix this with
// the following lines. See file "FAQ_for_g++_and_libg++..." for details...
template class LIST<DATA *>;
template class LIST<BUFFER *>;
template class LIST<PART3D *>;
template class LIST<DEVICE3D *>;


template<class Type /* = void * */>
  SORTLIST<Type>::SORTLIST () :
    LIST<Type> ()
{
}


template<class Type /* = void * */>
  void SORTLIST<Type>::Add (const Type &Value)
{
  LISTELEM<Type> *New;
  LISTELEM<Type> *Ptr = List;

  New = new LISTELEM<Type> (Value);

  if (List == NULL)
    List = Last = New;
  elsif (Comp(Ptr->Data, Value) > 0)
  {
    New->Next = List;
    List = Ptr->Prev = New;
  }
  else
  {
    while (Ptr->Next != NULL && Comp(Ptr->Next->Data, Value) <= 0)
      Ptr = Ptr->Next;
    New->Next = Ptr->Next;
    New->Prev = Ptr;
    Ptr->Next = New;

    if (New->Next != NULL)
      New->Next->Prev = New;
    else
      Last = New;
  }

  Number++;
}


template class SORTLIST<TASK *>;



