/* Material-Editor */
#include <Xm/Form.h>

#include "tuxCAD.h"


// Einstellungen fr Preview-Window
static PART3D *Part   = NULL;
static WORLD  *PWorld = NULL;

static char TextureMask[PATH_MAX+1] = CAD_PATH"TEXTURES/*.xpm,*.png,*jpg"; /**/
static char TextureFile[PATH_MAX+1] = CAD_PATH"TEXTURES/water128.jpg";


MATERIAL MaterialEditor (MATERIAL Material)
{
  Widget Shell, Form, Area;

  Shell = XtVaCreatePopupShell ("preview",
				topLevelShellWidgetClass, AppShell,
				XmNtitle,             "Material",
				XmNtransient,         True,
				XmNmappedWhenManaged, True,
				XmNwidth,             250, 
				XmNheight,            250,
				NULL);
  Form = XtVaCreateManagedWidget ("form", xmFormWidgetClass, Shell, NULL);
  // Manage after world creation because of window mapping
  Area = XtVaCreateWidget ("area", xmFormWidgetClass,   Form,
			           XmNtopAttachment,  XmATTACH_FORM,
			           XmNtopOffset,      50,
			           XmNleftAttachment, XmATTACH_FORM,
			           XmNleftOffset,     50,
			           XmNwidth,          150,
			           XmNheight,         150, 
			           NULL);

  XtPopup (Shell, XtGrabNonexclusive);

  PWorld = new WORKCELL (Area, "preview");
  Part = new SPHERE (PWorld, "sphere", GEOMakeRot(Vector_Y, -90.0), 
		                       10.0, 180.0, 180.0, 32, Material);
  PWorld->AddDevice (new DEVICE3D("preview", Part));
  PWorld->Fill ();
  PWorld->SetScale (0.5);

  XtManageChild (Area);

  TOOLMainLoop (AppCon, Shell);

  return Material;
}


void MaterialEdit ()
{
  MATERIAL Material;
  
  Material = MaterialEditor (MATERIAL(COLOR::LIME));
}


static void ShowMaterial (DIALOG *Dlg, int BaseID, MATERIAL *Material, 
			  bool active)
{
  char    *TexFile;
  TEXTURE *Texture;
  XImage  *Image, *Image1;

  DLGSetBoxActive (Dlg, BaseID+30, active);
  if (active)
  {
    if ((TexFile = DLGGetFileSel(Dlg, BaseID+32)) != NULL)
    {
      Material->LoadTexture (TexFile);
      Texture = Material->GetTexture ();
      if (Texture != NULL && (Image1 = Texture->GetImage()) != NULL)
      {
	if (Image1->width != 128 || Image1->height != 128)
	  Image = TOOLScaleImage (Image1, 128, 128);
	else
	  Image = TOOLCopyImage (Image1);

	DLGChangeImage (Dlg, BaseID+33, Image);
	TOOLDestroyImage (Image);
	DLGSetOMenu (Dlg, BaseID+38, Texture->GetMode());
      }
    }
  }
}


MATERIAL *MaterialCB (DIALOG *Dlg, int BaseID, MATERIAL *Material,
		     int BoxID, int Mask)
{
  TEXTURE *Texture = Material->GetTexture ();

  switch (BoxID-BaseID)
  {
    case 1 :
      Material->SetMaterialID (MATERIAL::ID(DLGGetOMenu(Dlg, BoxID)-1));
      DLGSetScale (Dlg, BaseID+21, Material->GetAmbient());
      DLGSetScale (Dlg, BaseID+22, Material->GetDiffuse());
      DLGSetScale (Dlg, BaseID+23, Material->GetSpecular());
      DLGSetScale (Dlg, BaseID+24, Material->GetEmission());
      DLGSetScale (Dlg, BaseID+25, Material->GetShininess());
      if (Material->HasTexture())
	DLGSetFileSel (Dlg, BaseID+32, Material->GetTexture()->GetFileName());
      DLGSetBoxActive (Dlg, BaseID+25, (Material->GetSpecular() > 0.0));
      ShowMaterial (Dlg, BaseID, Material, Material->HasTexture());
      break;

    case 21 :
      //      Material->SetMaterialID (MATERIAL::VALUES);
      Material->SetAmbient (DLGGetScale(Dlg, BoxID));
      break;

    case 22 :
      //      Material->SetMaterialID (MATERIAL::VALUES);
      Material->SetDiffuse (DLGGetScale(Dlg, BoxID));
      break;

    case 23 :
      {
	double Specular = DLGGetScale (Dlg, BoxID);

	//	Material->SetMaterialID (MATERIAL::VALUES);
	Material->SetSpecular (Specular);
	DLGSetBoxActive (Dlg, BaseID+25, (Specular > 0.0));
      }
      break;

    case 24 :
      //      Material->SetMaterialID (MATERIAL::VALUES);
      Material->SetEmission (DLGGetScale(Dlg, BoxID));
      break;

    case 25 :
      //      Material->SetMaterialID (MATERIAL::VALUES);
      Material->SetShininess (DLGGetScale(Dlg, BoxID));
      break;

    case 36 :
      if (Texture != NULL)
	Texture->SetMapping ((TEXTURE::MAPPING)DLGGetOMenu(Dlg, BoxID));
      break;

    case 37 :
      if (Texture != NULL)
	Texture->SetFilter ((TEXTURE::FILTER)DLGGetOMenu(Dlg, BoxID));
      break;

    case 38 :
      if (Texture != NULL)
	Texture->SetMode ((TEXTURE::MODE)DLGGetOMenu(Dlg, BoxID));
      break;

    case 39 :
      if (Texture != NULL)
	Texture->SetStretch (DLGGetScale(Dlg, BoxID),
			     DLGGetScale(Dlg, BoxID));
      break;

    default :
      {
	COLOR Color = ColorCB (Dlg, BaseID+10, 
				    Material->GetColor(), BoxID, Mask);
        Material->SetColor (Color);
	DLGSetBoxActive (Dlg, BaseID+20, (Color.GetID() != COLOR::TEXTURE));
	ShowMaterial (Dlg, BaseID, Material, (Color.GetID()==COLOR::TEXTURE));
      }
  }

  if (Part != NULL && PWorld != NULL)
  {
    Part->SetMaterial (*Material);
    PWorld->Redraw ();
  }

  return Material;
}


// Create and handle only the sub-dialog
void MaterialDialog (DIALOG *Dlg, int BaseID, MATERIAL *Material, int Mask)
{
  double   Specular;
  XImage  *Image = NULL;
  XImage  *Image1 = NULL;
  TEXTURE *Texture;

  PRO_START ("")

  if (Material == NULL)
    return;

  Texture = Material->GetTexture ();

  if (Mask & TEX_MASK)
  {
    DLGMakeRows (Dlg, BaseID,    3, DlgETCHED_IN DlgETCHED_IN, 
		                      BaseID+1, BaseID+5, BaseID+30);
    DLGMakeCols (Dlg, BaseID+30, 2, DlgETCHED_IN, BaseID+31, BaseID+35);
    DLGMakeRows (Dlg, BaseID+31, 2, DlgNO_LINE,   BaseID+32, BaseID+33);
    DLGMakeRows (Dlg, BaseID+35, 4, DlgNO_LINE, 
		                 BaseID+36, BaseID+37, BaseID+38, BaseID+39);
  }
  else
    DLGMakeRows (Dlg, BaseID,  2, DlgETCHED_IN, BaseID+1, BaseID+5);

  DLGMakeCols (Dlg, BaseID+5,  3, DlgETCHED_IN, BaseID+10, BaseID+20, 
	                                        BaseID+28);
  DLGMakeRows (Dlg, BaseID+20, 5, DlgNO_LINE,   BaseID+21, BaseID+22, 
	                                        BaseID+23, BaseID+24, 
	                                        BaseID+25);

  DLGInsertOMenu (Dlg, BaseID+1, "material", -9, MaterialNames, 0);  
  ColorDialog (Dlg, BaseID+10, Material->GetColor(), Mask);

  Specular = Material->GetSpecular ();

  // TODO: Bild auf 2^n x 2^n skalieren (64 x 64)
  DLGInsertScale (Dlg, BaseID+21, "ambient",  
		                  0.0, 1.0, Material->GetAmbient(),
		                  2, DlgHORIZONTAL);
  DLGInsertScale (Dlg, BaseID+22, "diffuse",  
		                  0.0, 1.0, Material->GetDiffuse(),
		                  2, DlgHORIZONTAL);
  DLGInsertScale (Dlg, BaseID+23, "specular", 0.0, 1.0, Specular,
		                  2, DlgHORIZONTAL);
  DLGInsertScale (Dlg, BaseID+24, "emission", 
		                  0.0, 1.0, Material->GetEmission(),
		                  2, DlgHORIZONTAL);
  DLGInsertScale (Dlg, BaseID+25, "shininess", 
		                  0.0, 128.0, Material->GetShininess(), 
		                  1, DlgHORIZONTAL);

  DLGSetBoxActive (Dlg, BaseID+25, (Specular > 0.0));

  if (Mask & TEX_MASK)
  {
    if (Texture == NULL || Texture->GetFileName() == NULL || 
                           Texture->GetFileName()[0] == '\0')
      strcpy (TextureFile, CAD_PATH"TEXTURES/water128.jpg");
    else
    {
      strncpy (TextureFile, Texture->GetFileName(), PATH_MAX);
      TextureFile[PATH_MAX] = '\0';
    }
    DLGInsertFileSel (Dlg, BaseID+32, "textureFile",
		                      TextureMask, 40, TextureFile);

    if (Texture == NULL || (Image1 = Texture->GetImage()) == NULL)
      Image = TOOLCreateImage (128, 128, 0.9, 0.9, 0.9);
    elsif (Image1->width != 128 || Image1->height != 128)
      Image = TOOLScaleImage (Image1, 128, 128);
    else
      Image = TOOLCopyImage (Image1);

    DLGInsertImage (Dlg, BaseID+33, "textImg", Image, NULL);

    DLGInsertOMenu (Dlg, BaseID+36, "mapping",  2, "clamp", "repeat", 
		                    (Texture==NULL)? 0:Texture->GetMapping());
    DLGInsertOMenu (Dlg, BaseID+37, "filter", 2, "nearest", "linear", 
		     	            (Texture==NULL)? 0:Texture->GetFilter());
    DLGInsertOMenu (Dlg, BaseID+38, "textureMode", 4, "modulate", "replace",
		                                      "decal",    "blend",
		                    (Texture==NULL)? 2:Texture->GetMode());
    // TODO: color for blend-mode
    DLGInsertScale (Dlg, BaseID+39, "scale", 0.1, 100.0, 
		            (Texture==NULL)? 1.0:Texture->GetStretchS(), 1,
		            DlgHORIZONTAL);

    DLGSetLabelPos (Dlg, BaseID+33, DlgNOWHERE, DlgEAST);

    TOOLDestroyImage (Image);
  }

  DLGSetLabelPos  (Dlg, BaseID+1, DlgWEST, DlgCENTER);


  PRO_END
}

