#include "tuxCAD.h"


#define TRANS_MASK ((1<<POS_X) | (1<<POS_Y) | (1<<POS_Z)) 
#define ROT_MASK   ((1<<ROT_X) | (1<<ROT_Y) | (1<<ROT_Z))


static void PartMenuCB   (int ID);
static void DevMenuCB    (int ID);
static void WorldMenuCB  (int ID);

static char PartName[PATH_MAX+1] = "test.part";
static char PartMask[PATH_MAX+1] = CAD_PATH"PARTS/*.part"; /* */

static char DevName[PATH_MAX+1] = "test.dev";
static char DevMask[PATH_MAX+1] = CAD_PATH"DEVICES/*.dev"; /* */



// *************************************************************************
// ** Name:                                                               **
// ** Description:                                                        **
// **                                                                     **
// ** Parameter:                                                          **
// **                                                                     **
// *************************************************************************

// *************************************************************************
// ***                                                                   ***
// ***                           P A R T - M E N U                       ***
// ***                                                                   ***
// *************************************************************************
static MENU_ENTRY PartTransMenu[] =
{
  ENTRY     ("absolute",   "", "", ' ', TRUE, PartMenuCB, 101),
  ENTRY     ("relative",   "", "", ' ', TRUE, PartMenuCB, 102),
};

static MENU_ENTRY PartRotMenu[] =
{
  ENTRY     ("absolute",   "", "", ' ', TRUE, PartMenuCB, 201),
  ENTRY     ("relative",   "", "", ' ', TRUE, PartMenuCB, 202),
};

static MENU_ENTRY PartTransformMenu[] =
{
  ENTRY     ("absolute",   "", "", ' ', TRUE, PartMenuCB, 301),
  ENTRY     ("relative",   "", "", ' ', TRUE, PartMenuCB, 302),
};

static MENU_ENTRY PartJointMenu[] =
{
  ENTRY     ("translate",  "", "", ' ', TRUE, PartMenuCB, 401),
  ENTRY     ("rotate",     "", "", ' ', TRUE, PartMenuCB, 402),
};

static MENU_ENTRY PartEditMenu[] =
{
  ENTRY     ("delete",     "", "", ' ',  TRUE, PartMenuCB, 500),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("cut",        "", "", ' ',  TRUE, PartMenuCB, 501),
  ENTRY     ("copy",       "", "", ' ',  TRUE, PartMenuCB, 502),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("scale",      "", "", ' ', TRUE,  PartMenuCB, 510),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("merge",      "", "", ' ', TRUE,  PartMenuCB, 520),
  ENTRY     ("smash",      "", "", ' ', TRUE,  PartMenuCB, 521),
  ENTRY     ("setOrigin",  "", "", ' ', TRUE,  PartMenuCB, 522),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("chain",      "", "", ' ', TRUE,  PartMenuCB, 530),
  ENTRY     ("attach",     "", "", ' ', TRUE,  PartMenuCB, 531),
  ENTRY     ("detachAll",  "", "", ' ', FALSE, PartMenuCB, 532),
  ENTRY     ("mount",      "", "", ' ', FALSE, PartMenuCB, 533)
};

static MENU_ENTRY PartCSGMenu[] =
{
  ENTRY     ("intersection", "", "", ' ',  TRUE, PartMenuCB, 601),
  ENTRY     ("difference",   "", "", ' ',  TRUE, PartMenuCB, 602),
  ENTRY     ("union",        "", "", ' ',  TRUE, PartMenuCB, 603),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("contacts",     "", "", ' ',  TRUE, PartMenuCB, 610)
};


static MENU_ENTRY PartMenu[] = 
{
  TITLE     ("partMenu"),
  MENU_LINE (PrgDOUBLE_LINE),
  ENTRY     ("properties",    "", "", ' ', TRUE,  PartMenuCB, 1),
  MENU_LINE (PrgETCHED_IN),
  SUB_MENU  ("translate",     "", "", ' ', TRUE,  PartTransMenu),
  SUB_MENU  ("rotate",        "", "", ' ', TRUE,  PartRotMenu),
  SUB_MENU  ("transform",     "", "", ' ', TRUE,  PartTransformMenu),
  SUB_MENU  ("joint",         "", "", ' ', TRUE,  PartJointMenu),
  MENU_LINE (PrgETCHED_IN),
  SUB_MENU  ("edit",          "", "", ' ', TRUE,  PartEditMenu),
  SUB_MENU  ("csg",           "", "", ' ', TRUE,  PartCSGMenu),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("save",          "", "", ' ', TRUE,  PartMenuCB,  80),
  ENTRY     ("write",         "", "", ' ', TRUE,  PartMenuCB,  81),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("info",          "", "", ' ', TRUE,  PartMenuCB,  90)
};


static int ChangeReference (DIALOG *Dlg, int BoxID)
{
  bool            Reference;
  int             count, Index;
  double          Values[6];
  H_MATRIX        Frame;
  PART3D         *Part;
  LIST<PART3D *> *List;

  if ((List = GetPartBuffer(GetCurrent())) == NULL)
    return DlgDENY;

  Reference = DLGGetBool (Dlg, BoxID);

  for (count = 0; count < List->Size(); ++count)
  {
    Part = List->Get (count);

    if (Reference)
    {
      for (Index = PX; Index <= RZ; ++Index)
	Values[count] = Part->GetVal (Index);

      Frame = GEORPY2Frame (Values[PX], Values[PY], Values[PZ],
			    Values[RX], Values[RY], Values[RZ]);
      Frame = GEOMatrixProd (Part->GetRel(), Frame);
      GEOFrame2RPY (Frame, &Values[PX], &Values[PY], &Values[PZ],
		           &Values[RX], &Values[RY], &Values[RZ]);
      DLGSetDouble (Dlg, 111, Values[PX]);
      DLGSetDouble (Dlg, 112, Values[PY]);
      DLGSetDouble (Dlg, 113, Values[PZ]);
      DLGSetDouble (Dlg, 121, Values[RX]);
      DLGSetDouble (Dlg, 122, Values[RY]);
      DLGSetDouble (Dlg, 123, Values[RZ]);
    }
  }
  return DlgCONFIRM;
}


static int AbsChange (DIALOG *Dlg, int BoxID)
{
  int             count, Dir;
  PART3D         *Part;
  WORLD          *World = GetCurrent ();
  LIST<PART3D *> *List;

  if ((List = GetPartBuffer(World)) == NULL)
    return DlgDENY;

  for (count = 0; count < List->Size(); ++count)
  {
    Part = List->Get (count);

    Dir = ((BoxID-100)/20) * 3 + ((BoxID-101) % 10);
    Part->SetVal (Dir, DLGGetDouble(Dlg, BoxID));
    World->Redraw ();
  }

  return DlgCONFIRM;
}


static void UpdatePart (DIALOG *Dlg, int ButtonID)
{
  int             count;
  PART3D         *Part;
  WORLD          *World = GetCurrent ();
  LIST<PART3D *> *List;

  if ((List = GetPartBuffer(World)) == NULL)
    return;

  for (count = 0; count < List->Size(); ++count)
  {
    Part = List->Get (count);

    Part->SetPos (DLGGetDouble(Dlg, 111),
		  DLGGetDouble(Dlg, 112),
		  DLGGetDouble(Dlg, 113));
    Part->SetRot (DLGGetDouble(Dlg, 121),
		  DLGGetDouble(Dlg, 122),
		  DLGGetDouble(Dlg, 123));
  }
}


static void ClickPos (DIALOG *Dlg, int ButtonID)
{
  int             count;
  double          px, py, pz;
  double          rx, ry, rz;
  H_MATRIX        M;
  PART3D         *Part2, *Part1;
  WORLD          *World = GetCurrent ();
  LIST<PART3D *> *List;

  PRGPrintf ("Select Point");

  if ((List = GetPartBuffer(World)) == NULL)
    return;

  World->SetMouseCursor (ToolNORMAL);
  if ((Part2 = World->PickPart(&px, &py, &pz)) == NULL)
    return;

  for (count = 0; count < List->Size(); ++count)
  {
    Part1 = List->Get (count);

    //   CurrWorld->SelectPart (Part);
    DLGSetDouble (Dlg, 11, -px);
    DLGSetDouble (Dlg, 12, -py);  
    DLGSetDouble (Dlg, 13, -pz);

    M = GEOMatrixProd (GEOMatrixInv(Part1->GetWTP()), Part2->GetWTP());
    GEOFrame2RPY (M, &px, &py, &pz, &rx, &ry, &rz);

    printf ("Trans: %.1f %.1f %.1f  Rot:  %.1f %.1f %.1f\n",
	    px, py, pz, rx, ry, rz);
  }
}


static void SetOrigin (DIALOG *Dlg, int ButtonID)
{
  int             count;
  RPY             Pos = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
  H_MATRIX        Rel;
  PART3D         *Part;
  WORLD          *World = GetCurrent ();
  LIST<PART3D *> *List;

  if ((List = GetPartBuffer(World)) != NULL)
    for (count = 0; count < List->Size(); ++count)
    {
      Part = List->Get (count);
      switch (ButtonID)
      {
        case -3 :
	  if (Part != NULL)
	  {
	    Rel = Part->GetWTP ();
	    Part->SetMatrix (Matrix_1);
	  }
	  break;

        default :
	  Pos.px = DLGGetDouble (Dlg, 11);
	  Pos.py = DLGGetDouble (Dlg, 12);
	  Pos.pz = DLGGetDouble (Dlg, 13);
	  Pos.y  = DLGGetDouble (Dlg, 21);
	  Pos.p  = DLGGetDouble (Dlg, 22);
	  Pos.r  = DLGGetDouble (Dlg, 23);
	  Rel = GEORPY2Frame (Pos.px, Pos.py, Pos.pz, Pos.y, Pos.p, Pos.r);
      }

      if (Part != NULL)
      {
	PARTSetOrigin (Part->GetPart(), Rel);
	Part->Create ();
	World->Redraw ();
      }
    }
  else
    PRGPrintf ("No part selected");
}


static void CloseDlg (DIALOG *Dlg, int ButtonID)
{
  DLGCloseDialog (Dlg);

  DLGFreeDialog (Dlg);
}


#define CONTACT_SIZE 0.025

static void ShowIntersection (WORLD *World, PART3D *Part1, PART3D *Part2)
{
  int        count, Side;
  char       PartName[20];
  INTERSECT *List, *Ptr;
  PART3D    *Part;
  DEVICE3D  *Dev = NULL;

  if ((List = PARTGetIntersections(Part1->GetPart(), Part1->GetWTP(),
				   Part2->GetPart(), Part2->GetWTP())) != NULL)
  {
    Dev = new DEVICE3D ("contact");
    World->AddDevice (Dev);
  }

  Ptr = List;
  while (Ptr != NULL)
  {
    for (Side = 0; Side < 2; ++Side)
      for (count = 0; count < Ptr->Number[Side]; ++count)
      {
	sprintf (PartName, "contact_%d_%d", Side+1, count+1);
	Part = new SPHERE (World, PartName, Matrix_1, CONTACT_SIZE, 
			   180.0, 360.0, 12, MATERIAL(COLOR(COLOR::MEDBLUE)));
	Part->SetMatrix (GEOMakeTrans(Ptr->Points[Side][count].x,
				      Ptr->Points[Side][count].y,
				      Ptr->Points[Side][count].z));
	Dev->AddPart (Part);
      }
    Ptr = Ptr->Next;
  }
}


static void PartMenuCB (int ID)
{
  unsigned long   UpdateMode = 0;
  int             count;
  DIALOG         *Dlg;
  PART3D         *Part;
  WORLD          *World = GetCurrent ();
  LIST<PART3D *> *List;

  if ((List = GetPartBuffer(World)) == NULL)
    return;

#if 0
  for (count = 0; count < List->Size(); ++count)
  {
    Part = List->Get (count);
    // Set the part's visible parameter while it is activated
    // e.g. unhighlight it, to see the shading or draw a bounding box (should
    // be set to old Value when deselected...)
    World->Select (Part, FALSE);
    Part->SetMode (Part->GetMode() + PART3D::DRAW_BOX);
  }
#endif

  switch (ID)
  {
    case 1 :                          // Edit part properties
      if (List->Size() == 1)
      {
	Part = List->Get (0);
	PartProperties (World, Part);
      }
      break;

    case 80 :                           // Save part
      if (List->Size() == 1)
      {
	Part = List->Get (0);
	if (DLGOpenFileSel("savePart", PartName, PATH_MAX, PartMask) != NULL)
	{
	  PARTSave (PartName, Part->GetPart(), FALSE);
	  TOOLGetFileDir (PartName, PartMask, PATH_MAX);
	  strcat (PartMask, "/*.part"); /* */
	}
      }
      break;

    case 81 :                           // Write part
      if (List->Size() == 1)
      {
	FILE *fp = NULL;
	Part = List->Get (0);
	if (DLGOpenFileSel("savePart", PartName, PATH_MAX, "/tmp/") != NULL)
	  if ((fp = fopen(PartName, "w+")) != NULL)
	  {
	    Part->Write (fp);
	    fclose (fp);
	  }
      }
      break;

    case 90 :                           // Show part info
      if (List->Size() == 1)
      {
	Part = List->Get (0);
	PartInfo (Part);
      }
      break;

    case 101 :                          // Translate part absolute
      SetAction (World, DATA::TRANS_ABSOLUTE);
      UpdateMode = DLG_PART;
      break;

    case 102 :                          // Translate part relative
      SetAction (World, DATA::TRANS_RELATIVE);
      UpdateMode = DLG_PART;
      break;

    case 201 :                          // Rotate part absolute
      SetAction (World, DATA::ROT_ABSOLUTE);
      UpdateMode = DLG_PART;
      break;

    case 202 :                          // Rotate part relative
      SetAction (World, DATA::ROT_RELATIVE);
      UpdateMode = DLG_PART;
      break;

    case 301 :                          // Transform part absolute
    case 302 :                          //                relative
      {
	int    count, Axes;
	double Values[6];

	if (List->Size() == 1)
	{
	  Part = List->Get (0);

	  Dlg = DLGNewDialog (0);
	  DLGMakeRows (Dlg, 0, 2, DlgETCHED_IN, 10, 100);
	  DLGMakeCols (Dlg, 100, 3, DlgETCHED_IN DlgETCHED_IN, 110, 120, 130);
	  DLGMakeRows (Dlg, 110, 3, DlgNO_LINE, 111, 112, 113);
	  DLGMakeRows (Dlg, 120, 3, DlgNO_LINE, 121, 122, 123);
	  DLGMakeRows (Dlg, 130, 3, DlgNO_LINE, 131, 132, 133);

	  DLGInsertBool (Dlg, 10, "world", FALSE);
	  DLGSetFunction (Dlg, 10, ChangeReference);
	  for (count = PX; count <= RZ; ++count)
	    Values[count] = Part->GetVal (count);
	  Axes = Part->GetJoint().GetAxes ();

	  DLGInsertDouble (Dlg, 111, "px", Values[PX], 10, 4);
	  if (! (Axes & (1<<POS_X)))
	    DLGSetBoxActive (Dlg, 111, FALSE);
	  DLGInsertDouble (Dlg, 112, "py", Values[PY], 10, 4);
	  if (! (Axes & (1<<POS_Y)))
	    DLGSetBoxActive (Dlg, 112, FALSE);
	  DLGInsertDouble (Dlg, 113, "pz", Values[PZ], 10, 4);
	  if (! (Axes & (1<<POS_Z)))
	    DLGSetBoxActive (Dlg, 113, FALSE);
	  DLGSetFunction (Dlg, 110, AbsChange);
	  DLGInsertDouble (Dlg, 121, "rx", Values[RX], 8, 3);
	  if (! (Axes & (1<<ROT_X)))
	    DLGSetBoxActive (Dlg, 121, FALSE);
	  DLGInsertDouble (Dlg, 122, "ry", Values[RY], 8, 3);
	  if (! (Axes & (1<<ROT_Y)))
	    DLGSetBoxActive (Dlg, 122, FALSE);
	  DLGInsertDouble (Dlg, 123, "rz", Values[RZ], 8, 3);
	  if (! (Axes & (1<<ROT_Z)))
	    DLGSetBoxActive (Dlg, 123, FALSE);
	  DLGSetFunction (Dlg, 120, AbsChange);
	  DLGInsertDouble (Dlg, 131, "scaleX", 1.0, 8, 2);
	  DLGInsertDouble (Dlg, 132, "scaleY", 1.0, 8, 2);
	  DLGInsertDouble (Dlg, 133, "scaleZ", 1.0, 8, 2);

	  DLGMakeButton (Dlg, USER_ID, "apply", DlgBUTTON_RIGHT, 
			                        UpdatePart, -1);
	  DLGMakeButton (Dlg, CLOSE_BUTTON);
	  DLGOpenDialog (Dlg, "absChange", DlgNORMAL, CloseDlg);
	  UpdateMode = DLG_PART;
	}
      }
      break;

    case 401 :                          // Translate joint
      SetAction (World, DATA::MOVE_TRANSLATE);
      UpdateMode = DLG_PART | DLG_DEV;
      break;

    case 402 :                          // Rotate joint
      SetAction (World, DATA::MOVE_ROTATE);
      UpdateMode = DLG_PART | DLG_DEV;
      break;

    case 500 :                           // Delete part
      // TODO: Abfrage, ob attached-te Devices hochgezogen werden sollen
      if (DLGOpenTextDialog("deletePart(s)", DlgWARNING, 
			    "Really delete this part(s) ?", 
			    2, "NO", "YES") == 1)
        FUNC_ABORT
        // TODO: Momentan noch vom Benutzer nicht nachvollziehbar, weil 
        // attached- und chain-parts nicht markiert sind !

      for (count = List->Size()-1; count >= 0; count--)
      {
	Part = List->Get (count);
	World->Delete  (Part);
      }
      List->Clear ();
      World->Redraw  ();
      UpdateMode = DLG_TREE;
      break;

    case 501 :                           // Cut part
      for (count = List->Size()-1; count >= 0; count--)
      {
	Part = List->Get (count);
	World->Remove (Part);
	Buffer.Add (Part);
      }
      World->Redraw ();
      UpdateMode = DLG_TREE;
      break;

    case 502 :                           // Copy part
      for (count = 0; count < List->Size(); ++count)
      {
	Part = List->Get (count);
	Buffer.Add (new PART3D(Part, World));
      }
      break;

    case 510 :                          // Scale part
      {
	bool     KeepPos = TRUE;
	double   Scale[3];
	H_MATRIX WTP;

	// TODO: 1 und mehrere Parts unterscheiden...

	if (List->Size() == 1)
	{
	  Part = List->Get (0);
	  PARTGetScale (Part->GetPart(), &Scale[X], &Scale[Y], &Scale[Z]);
	}
	else
	  Scale[X] = Scale[Y] = Scale[Z] = 1.0;

	Dlg = DLGNewDialog (0);
	DLGMakeRows (Dlg,  0, 4, DlgNO_LINE, 10, 20, 30, 40);
	DLGInsertDouble (Dlg, 10, "scaleX", Scale[X], 8, 3);
	DLGInsertDouble (Dlg, 20, "scaleY", Scale[Y], 8, 3);
	DLGInsertDouble (Dlg, 30, "scaleZ", Scale[Z], 8, 3);
	DLGInsertBool   (Dlg, 40, "keepPos", KeepPos);
	DLGOpenDialog   (Dlg, "scalePart", DlgNORMAL, NULL);
	if (DLGGetExitButton(Dlg) == DlgOK)
	{
	  Scale[X] = DLGGetDouble (Dlg, 10);
	  Scale[Y] = DLGGetDouble (Dlg, 20);
	  Scale[Z] = DLGGetDouble (Dlg, 30);
	  KeepPos  = DLGGetBool   (Dlg, 40);
	  for (count = 0; count < List->Size(); ++count)
	  {
	    Part = List->Get (count);

	    if (KeepPos)                // Save position, if it should be kept
	      WTP = Part->GetMatrix ();

	    Part->Scale (Scale[X], Scale[Y], Scale[Z]);
	
	    if (KeepPos)                // Restore old position
	      Part->SetMatrix (WTP);
	  }
	  World->Redraw ();
	  UpdateMode = DLG_PART;
	}
	DLGFreeDialog (Dlg);
      }
      break;

    case 520 :                          // Merge parts
      {
	PART3D   *Part2;
	DEVICE3D *Dev1, *Dev2, *Attach, *Ptr;

	Part = List->Get (0);
	if (List->Size() == 1)          // We need at least two parts to merge
	{
	  do
	    Part2 = ChoosePart (MSG_SELECT_PART, World);
	  while (Part2 == Part);
	  
	  if (Part2 != NULL)
	    SwapSelection (World, Part2);
	}

	Dev1 = Part->GetParent ();
	for (count = 1; count < List->Size(); ++count)
	{
	  Part2 = List->Get (count);

	  // locate device of 2nd part
	  if ((Dev2 = Part2->GetParent()) != NULL)
	    if (Dev2->Remove(Part2))
	      Part2->SetNext (NULL);

	  if ((Attach = Part2->GetAttach()) != NULL)
	    Part2->SetAttach (NULL);      // detach attached devices

	  // If we have attached devices, add them as next devices
	  // We might also call AddDevice, but this would disturb the order.
	  if ((Ptr = Attach) != NULL)
	  {
	    while (Ptr != NULL && Ptr->GetNext() != NULL)
	      Ptr = Ptr->GetNext ();

	    Ptr->SetNext (Dev2->GetNext());
	    Dev2->SetNext (Attach);
	  }

	  // Merge both parts without attached devices (destroys part two)
	  World->MergeParts (Part, Part2, CHAIN_MASK);

	  // if there's no other part in the device, delete it 
	  if (Dev2 != NULL && Dev2->GetPartNo() == 0)
	    World->Delete (Dev2);
	}
	World->Redraw ();
	UpdateMode |= DLG_TREE;
      }
      break;

    case 521 :                          // Smash part
      for (count = 0; count < List->Size(); ++count)
      {
	Part = List->Get (count);
	World->Smash   (Part);
      }
      ClearSelection (World);
      World->Redraw  ();
      UpdateMode = DLG_TREE;
      break;

    case 522 :                          // Set part origin
      {
	RPY Pos = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};

	Dlg = DLGNewDialog (0);
	DLGMakeCols (Dlg,  0, 2, DlgNO_LINE, 10, 20, 30);
	DLGMakeRows (Dlg, 10, 3, DlgNO_LINE, 11, 12, 13);
	DLGMakeRows (Dlg, 20, 3, DlgNO_LINE, 21, 22, 23);
	DLGInsertDouble (Dlg, 11, "px", Pos.px, 10, 3);
	DLGInsertDouble (Dlg, 12, "py", Pos.py, 10, 3);
	DLGInsertDouble (Dlg, 13, "pz", Pos.pz, 10, 3);
	DLGInsertDouble (Dlg, 21, "rx", Pos.y,  10, 3);
	DLGInsertDouble (Dlg, 22, "ry", Pos.p,  10, 3);
	DLGInsertDouble (Dlg, 23, "rz", Pos.r,  10, 3);
	DLGMakeButton (Dlg, USER_ID, "click",  DlgBUTTON_RIGHT, ClickPos,  -1);
	DLGMakeButton (Dlg, USER_ID, "apply",  DlgBUTTON_RIGHT, SetOrigin, -2);
	DLGMakeButton (Dlg, USER_ID, "origin", DlgBUTTON_RIGHT, SetOrigin, -3);
	DLGMakeButton (Dlg, CLOSE_BUTTON);
	DLGOpenDialog (Dlg, "relPlace", DlgNORMAL, NULL);
	/* TODO: Place part ... */
	DLGFreeDialog (Dlg);
	UpdateMode = DLG_PART;
      }
      break;

    case 530 :                          // Chain
      {
	PART3D *Ptr;

	if (List->Size() == 1)
	{
	  Part = List->Get (0);
	  PRGPrintf ("Select a part to add in chain...");
	  do
	  {
	    if ((Ptr = ChoosePart("select part...", World)) == NULL)
	      break;
	    elsif (Ptr == Part)
	      PRGPrintf ("Invalid part selected !");
	  } while (Ptr == NULL);	      

	  if (Ptr != NULL)
	  {
	    World->Remove (Ptr);
	    Part->AddChain (Ptr);
	    UpdateMode = DLG_TREE;
	  }
	}
      }
      break;

    case 531 :                          // Attach device
      {
	PART3D   *Ptr;
	DEVICE3D *Dev;

	if (List->Size() == 1)
	{
	  Part = List->Get (0);

	  PRGPrintf ("Select device to attach...");
	  do 
	  {
	    Dev = NULL;
	    if ((Ptr = ChoosePart("Select device...", World)) == NULL)
	      break;
	    elsif (Ptr == Part)
	      PRGPrintf ("Invalid device selected !");
	    else
	      Dev = Ptr->GetParent ();
	  } while (Dev == NULL);

	  if (Dev != NULL)
	  {
	    H_MATRIX PTD = Matrix_1;

	    World->Remove (Dev);
	    PTD = GEOMatrixProd (GEOMatrixInv(Part->GetWTP()), Dev->GetWTD());
	    Dev->SetWTD     (PTD);
	    Part->AttachDev (Dev);
	    UpdateMode = DLG_TREE;
	  }
	}
	else
	  PRGPrintf ("\\"MSG_ABORT);
      }
      break;

    case 532 :                          // Detach all
      // TODO: Sinnvoll / ntig ???
      UpdateMode = DLG_TREE;
      break;

    case 533 :                          // Mount 
      {
	UpdateMode = DLG_TREE;
      }
      break;

    case 601 :  // CSG-Operations
    case 602 :
    case 603 :
    case 610 :
      {
	char     *DevName;
	PART3D   *CSG = NULL, *Part2;
	DEVICE3D *Dev;

	if (List->Size() == 1)
	{
	  Part = List->Get (0);
	  do
	    Part2 = ChoosePart ("Select a second part...", World);
	  while (Part2 == Part);

	  if (Part2 != NULL)
	  {
	    if (ID == 601)
	      // = new INTERSECTION (World, "intersection", Matrix_1, Part,
	      //                            Part2, Part->GetMaterial());
	      CSG = *Part % *Part2; 
	    elsif (ID == 602)
	      // = new DIFFERENCE (World, "difference", Matrix_1, Part,
	      //                          Part2, Part->GetMaterial());
	      CSG = *Part - *Part2;
	    elsif (ID == 603)
	      // = new UNION (World, "union", Matrix_1, Part,
	      //                     Part2, Part->GetMaterial());
              CSG = *Part + *Part2;
	    else
	      ShowIntersection (World, Part, Part2);

	    if (CSG != NULL)
	    {
	      DevName = new char[strlen(Part->GetName()) + 
				 strlen(Part2->GetName()) + 2];
	      sprintf (DevName, "%s_%s", Part->GetName(), Part2->GetName());
	      Dev = new DEVICE3D (DevName, CSG);
	      delete DevName;
	      World->AddDevice (Dev);
	    }
#if 1
	    if (CSG != NULL || ID == 610)
	    {
	      // Make source parts transparent
	      Part->GetMaterial()->SetAlpha  (0.50);
	      Part->Create ();
	      Part2->GetMaterial()->SetAlpha (0.50);
	      Part2->Create ();
	      World->IncAlpha (2);
	    }
#endif
	    World->Redraw ();
	    UpdateMode = DLG_TREE;
	  }
	}
      }
      break;

    default : 
      PRO1 ("Invalid Function %d\n", ID)
  }

  WorldMouseCursor (World);
  UpdateDialogs    (World, UpdateMode);
}


void OpenPartMenu (Widget Shell, const LIST<PART3D *> &Parts, 
		   int MouseX /* = -1 */, int MouseY /* = -1 */)
{
  int     Axes;
  PART3D *Part;

  if (Parts.Size() == 0)
    return;

  if (MouseX == -1 || MouseY == -1)
    TOOLGetMousePos (&MouseX, &MouseY);

  if (Parts.Size() == 1 && (Part = Parts.Get(0)) != NULL)
  {
    // Trans- and Rotate-menu is only active when part is not a joint
    PartMenu[4].active = PartMenu[5].active = PartMenu[6].active = 
      (Part->GetPrev() == NULL);

    // Insensitivate non existing DOF and open Menu
    Axes = Part->GetJoint().GetAxes ();
    PartJointMenu[0].active = (Axes & TRANS_MASK);
    PartJointMenu[1].active = (Axes & ROT_MASK);
  }
  PRGCreateMenu (Shell, "partMenu", MouseX, MouseY,
		        PartMenu, sizeof(PartMenu)/sizeof(MENU_ENTRY));
}



// *************************************************************************
// ***                                                                   ***
// ***                       D E V I C E - M E N U                       ***
// ***                                                                   ***
// *************************************************************************
static MENU_ENTRY DevTransMenu[] =
{
  ENTRY     ("absolute",   "", "", ' ', TRUE, DevMenuCB, 101),
  ENTRY     ("relative",   "", "", ' ', TRUE, DevMenuCB, 102),
};

static MENU_ENTRY DevRotMenu[] =
{
  ENTRY     ("absolute",   "", "", ' ', TRUE, DevMenuCB, 201),
  ENTRY     ("relative",   "", "", ' ', TRUE, DevMenuCB, 202),
};

static MENU_ENTRY TCPTransMenu[] =
{
  ENTRY     ("absolute",   "", "", ' ', TRUE,  DevMenuCB, 401),
  ENTRY     ("relative",   "", "", ' ', TRUE,  DevMenuCB, 402)
};

static MENU_ENTRY TCPRotMenu[] =
{
  ENTRY     ("absolute",   "", "", ' ', TRUE,  DevMenuCB, 501),
  ENTRY     ("relative",   "", "", ' ', TRUE,  DevMenuCB, 502)
};

static MENU_ENTRY ToolMenu[] = 
{
  ENTRY     ("open",    "", "", ' ', TRUE,  DevMenuCB, 601),
  ENTRY     ("middle",  "", "", ' ', TRUE,  DevMenuCB, 602),
  ENTRY     ("close",   "", "", ' ', TRUE,  DevMenuCB, 603)
};

static MENU_ENTRY DevEditMenu[] = 
{
  ENTRY     ("delete",     "", "", ' ', TRUE,  DevMenuCB, 70),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("copy",       "", "", ' ', TRUE,  DevMenuCB, 80),
  ENTRY     ("cut",        "", "", ' ', TRUE,  DevMenuCB, 81),
  ENTRY     ("paste",      "", "", ' ', FALSE, DevMenuCB, 82),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("scale",      "", "", ' ', TRUE,  DevMenuCB, 90),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("smash",      "", "", ' ', TRUE,  DevMenuCB, 92),
  MENU_LINE (PrgETCHED_IN),
#if 1
  // sollte diese Richtung beibehalten werden, d.h. man whlt das zuerst
  // das zu attachende Device an und dann das Part, an das es attached werden
  // soll ? Anders herum gibt's die Funktion bereits im Part-Men.
  ENTRY     ("attach",     "", "", ' ', TRUE,  DevMenuCB, 95),
#endif
  ENTRY     ("detach",     "", "", ' ', TRUE,  DevMenuCB, 96)
};

static MENU_ENTRY DevMenu[] = 
{
  TITLE     ("deviceMenu"),
  MENU_LINE (PrgDOUBLE_LINE),
  ENTRY     ("properties", "", "", ' ', TRUE, DevMenuCB, 100),
  MENU_LINE (PrgETCHED_IN),
  SUB_MENU  ("translate",  "", "", ' ', TRUE, DevTransMenu),
  SUB_MENU  ("rotate",     "", "", ' ', TRUE, DevRotMenu),
  MENU_LINE (PrgETCHED_IN),
  SUB_MENU  ("edit",       "", "", ' ', TRUE, DevEditMenu),
  MENU_LINE (PrgETCHED_IN),
  SUB_MENU  ("tcpTrans",   "", "", ' ', TRUE, TCPTransMenu),
  SUB_MENU  ("tcpRot",     "", "", ' ', TRUE, TCPRotMenu),
  SUB_MENU  ("tool",       "", "", ' ', TRUE, ToolMenu),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("save",       "", "", ' ', TRUE, DevMenuCB, 150),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("info",       "", "", ' ', TRUE, DevMenuCB, 160)
};


static void DevMenuCB (int ID)
{
  unsigned long     UpdateMode = 0;
  int               count;
  DEVICE3D         *Dev;
  DIALOG           *Dlg;
  WORLD            *World = GetCurrent ();;
  LIST<DEVICE3D *> *List;

  if ((List = GetDevBuffer(World)) == NULL)
    return;

  switch (ID)
  {
    case 101 :                          // Translate device absolute
      SetAction (World, DATA::TRANS_ABSOLUTE);
      UpdateMode = DLG_PART | DLG_DEV;
      break;

    case 102 :                          // Translate device relative
      SetAction (World, DATA::TRANS_RELATIVE);
      UpdateMode = DLG_PART | DLG_DEV;
      break;

    case 201 :                          // Rotate device absolute
      SetAction (World, DATA::ROT_ABSOLUTE);
      UpdateMode = DLG_PART | DLG_DEV;
      break;

    case 202 :                          // Rotate device relative
      SetAction (World, DATA::ROT_RELATIVE);
      UpdateMode = DLG_PART | DLG_DEV;
      break;

    case 401 :                          // Translate TCP absolute
      for (count = 0; count < List->Size(); ++count)
      {
	Dev = List->Get (count);

	World->Select (Dev, FALSE);
	SetAction (World, DATA::TCP_TRANS_ABS);
	UpdateMode = DLG_PART | DLG_DEV;
      }
      break;

    case 402 :                          // Translate TCP relative
      for (count = 0; count < List->Size(); ++count)
      {
	Dev = List->Get (count);

	World->Select (Dev, FALSE);
	SetAction (World, DATA::TCP_TRANS_REL);
	UpdateMode = DLG_PART | DLG_DEV;
      }
      break;

    case 501 :                          // Rotate TCP absolute
      for (count = 0; count < List->Size(); ++count)
      {
	Dev = List->Get (count);

	World->Select (Dev, FALSE);
	SetAction (World, DATA::TCP_ROT_ABS);
	UpdateMode = DLG_PART | DLG_DEV;
      }
      break;

    case 502 :                          // Rotate TCP relative
      for (count = 0; count < List->Size(); ++count)
      {
	Dev = List->Get (count);

	World->Select (Dev, FALSE);
	SetAction (World, DATA::TCP_ROT_REL);
	UpdateMode = DLG_PART | DLG_DEV;
      }
      break;

    // TODO: dynamische Aktionen von Tool def. (nicht nur ffnen/schliessen)
    case 601 :                          // Open tool
    case 602 :                          // Tool middle position
    case 603 :                          // Close tool
      {
	PART3D   *Mount;
	DEVICE3D *Tool;

	for (count = 0; count < List->Size(); ++count)
        {
	  Dev = List->Get (count);

	  if ((Mount = Dev->GetMount()) == NULL ||
	      (Tool = Mount->GetAttach()) == NULL)
	    PRGPrintf ("No tool found");
	  else
	  {
	    PRGPrintf ("Mount: %s", Tool->GetName());
	    if (ID == 601)
	      Tool->SetAllJoints (-1.0);
	    elsif (ID == 603)
	      Tool->SetAllJoints (1.0);
	    else
	      Tool->SetAllJoints (0.0);
	  }
	}
	World->Redraw ();
	UpdateMode = DLG_PART | DLG_DEV;
      }
      break;

    case  70 :                          // Delete device
      if (DLGOpenTextDialog("deleteDev", DlgWARNING, 
			    "Really delete this device(s) ?", 
			    2, "NO", "YES") == 1)
        FUNC_ABORT

      for (count = List->Size()-1; count >= 0; count--)
      {
	Dev = List->Get (count);
	World->Delete  (Dev);
      }
      List->Clear ();
      World->Redraw  ();
      UpdateMode = DLG_TREE;
      break;

    case 80 :                          // Copy device
      for (count = 0; count < List->Size(); ++count)
      {
	Dev = List->Get (count);
        Buffer.Add (new DEVICE3D(Dev, World));
      }
      break;

    case 81 :                          // Cut device
      for (count = 0; count < List->Size(); ++count)
      {
	Dev = List->Get (count);
	World->Remove (Dev);
	Buffer.Add (Dev);
      }
      World->Redraw ();
      UpdateMode = DLG_TREE;
      break;

    case 90 :                          // Scale device
      {
	bool     KeepPos = TRUE;
	double   Scale[3];
	H_MATRIX WTD;

	if (List->Size() == 1)
	{
	  Dev = List->Get (0);
	  PARTGetScale (Dev->GetPartList()->GetPart(), 
			&Scale[X], &Scale[Y], &Scale[Z]);
	}
	else
	  Scale[X] = Scale[Y] = Scale[Z] = 1.0;

	Dlg = DLGNewDialog (0);
	DLGMakeRows (Dlg,  0, 4, DlgNO_LINE, 10, 20, 30, 40);
	DLGInsertDouble (Dlg, 10, "scaleX", Scale[X], 8, 3);
	DLGInsertDouble (Dlg, 20, "scaleY", Scale[Y], 8, 3);
	DLGInsertDouble (Dlg, 30, "scaleZ", Scale[Z], 8, 3);
	DLGInsertBool   (Dlg, 40, "keepPos", KeepPos);
	DLGOpenDialog (Dlg, "scaleDevice", DlgNORMAL, NULL);
	if (DLGGetExitButton(Dlg) == DlgOK)
	{
	  Scale[X] = DLGGetDouble (Dlg, 10);
     	  Scale[Y] = DLGGetDouble (Dlg, 20);
	  Scale[Z] = DLGGetDouble (Dlg, 30);
	  KeepPos  = DLGGetBool   (Dlg, 40);
	  for (count = 0; count < List->Size(); ++count)
	  {
	    Dev = List->Get (count);

	    if (KeepPos)                // Save old position if desired
	      WTD = Dev->GetWTD ();

	    Dev->Scale (Scale[X], Scale[Y], Scale[Z]);

	    if (KeepPos)                // Restore old position
	      Dev->SetWTD (WTD); 
	  }
	  World->Redraw ();
	  UpdateMode = DLG_PART | DLG_DEV;
	}
	DLGFreeDialog (Dlg);
      }      
      break;

    case  92 :                          // Smash device
      for (count = 0; count < List->Size(); ++count)
      {
	Dev = List->Get (count);
	World->Smash   (Dev);
      }
      ClearSelection (World);
      World->Redraw  ();
      UpdateMode = DLG_TREE;
      break;      

#if 1
    case  95 :                          // Attach device
      {
	PART3D *Part;

	if (List->Size() == 1)
	{
	  Dev = List->Get (0);

	  PRGPrintf ("Select part to which the device should be attached...");
	  do
	  {
	    if ((Part = ChoosePart("Select part...", World)) == NULL)
	      break;
	    elsif (Part->GetParent() == Dev)
	    {
	      PRGPrintf ("Invalid part selected !");
	      Part = NULL;
	    }
	  } while (Part == NULL);

	  if (Part != NULL)
	  {
	    H_MATRIX PTD;

	    World->Remove (Dev);
	    PTD = GEOMatrixProd (GEOMatrixInv(Part->GetWTP()), Dev->GetWTD());
	    Dev->SetWTD (PTD);
	    Part->AttachDev (Dev);
	    World->Redraw ();
	    UpdateMode = DLG_TREE;
	  }
	}
	else
	  PRGPrintf ("\\"MSG_ABORT);
      }
      break;
#endif

    case 96 :                           // Detach device
      {
	H_MATRIX  WTD;
	PART3D   *Part;

	for (count = 0; count < List->Size(); ++count)
	{
	  Dev = List->Get (count);
	  if ((Part = Dev->GetFix()) != NULL)
	  {
	    WTD = GEOMatrixProd (Part->GetWTP(), Dev->GetWTD());
	    Part = Part->GetNext ();
	    World->Detach (Dev);
	    Dev->SetWTD (WTD);
	  }
	}
	World->Redraw ();
	UpdateMode = DLG_TREE;
      }
      break;

    case 100 :                          // Device properties
      if (List->Size() == 1)
      {
	Dev = List->Get (0);
	DevProperties (World, Dev);
      }
      break;

    case 150 :                          // Save device
      if (List->Size() == 1)
      {
	Dev = List->Get (0);
        sprintf (DevName, "%s.dev", Dev->GetName());
	DevName[PATH_MAX] = '\0';
	if (DLGOpenFileSel("saveDev", DevName, PATH_MAX, DevMask) != NULL &&
	    OverwriteFile(DevName))
	{
	  Dev->Save (DevName);
	  TOOLGetFileDir (DevName, DevMask, PATH_MAX);
	  strcat (DevMask, "/*.dev"); /* */
	}
      }
      break;

    case 160 :                          // Show device infos
      if (List->Size() == 1)
      {
	Dev = List->Get (0);
	DevInfo (Dev);
      }
      break;

    default :
      PRO1 ("Function %d\n", ID)
  }

  WorldMouseCursor (World);
  UpdateDialogs    (World, UpdateMode);
}


void OpenDevMenu (Widget Shell, const LIST<DEVICE3D *> &Devs,
		  int MouseX /* = -1 */, int MouseY /* = -1 */)
{
  DEVICE3D *Dev;
  ROBOT    *Robot;

  if (Devs.Size() == 0)
    return;

  if (MouseX == -1 || MouseY == -1)
    TOOLGetMousePos (&MouseX, &MouseY);

  if (Devs.Size() == 1 && (Dev = Devs.Get(0)) != NULL)
  {
    Robot = Dev->GetRobot();    /* || World->SelectedDevs() > 1 */
    DevMenu[ 9].active = (Robot != NULL);
    DevMenu[10].active = (Robot != NULL);
    DevMenu[11].active = (Robot != NULL);

    DevEditMenu[8].active = (Dev->GetFix() != NULL);
  }
  PRGCreateMenu (Shell, "deviceMenu", MouseX, MouseY, 
		 DevMenu, sizeof(DevMenu)/sizeof(MENU_ENTRY));
}


// *************************************************************************
// ***                                                                   ***
// ***                   W O R L D - M E N U                             ***
// ***                                                                   ***
// *************************************************************************
static MENU_ENTRY WorldZoomMenu[] = 
{
  ENTRY ("interactive",       "", "", ' ', TRUE, WorldMenuCB, 401),
  ENTRY ("step",              "", "", ' ', TRUE, WorldMenuCB, 402),
  ENTRY ("fill",              "", "", ' ', TRUE, WorldMenuCB, 403)
};

static MENU_ENTRY WorldLoadMenu[] =
{
  ENTRY ("part",              "", "", ' ', TRUE, WorldMenuCB, 801),
  ENTRY ("dev",               "", "", ' ', TRUE, WorldMenuCB, 802),
  ENTRY ("world",             "", "", ' ', TRUE, WorldMenuCB, 803),
};

static MENU_ENTRY WorldFileMenu[] =
{
  SUB_MENU  ("load",          "", "", ' ', TRUE,  WorldLoadMenu),
  ENTRY     ("save",          "", "", ' ', TRUE,  WorldMenuCB, 90),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("import",        "", "", ' ', FALSE, WorldMenuCB, 95),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("print",         "", "", ' ', TRUE, WorldMenuCB, 96)
};

MENU_ENTRY WorldViewMenu[] =
{
  ENTRY     ("tree",          "", "", ' ', TRUE, WorldMenuCB, 1001),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("single",        "", "", ' ', TRUE, WorldMenuCB, 1002),
  ENTRY     ("double",        "", "", ' ', TRUE, WorldMenuCB, 1003),
  ENTRY     ("quad",          "", "", ' ', TRUE, WorldMenuCB, 1004),
  ENTRY     ("1+3",           "", "", ' ', TRUE, WorldMenuCB, 1005),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("coi",           "", "", ' ', TRUE, WorldMenuCB, 1006)
};

static MENU_ENTRY WorldSelectMenu[] = 
{
  ENTRY     ("allParts",      "", "", ' ', FALSE, WorldMenuCB, 41),
  ENTRY     ("allDevices",    "", "", ' ', FALSE, WorldMenuCB, 42),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("vertex",        "", "", ' ', FALSE, WorldMenuCB, 43),
  ENTRY     ("edge",          "", "", ' ', FALSE, WorldMenuCB, 44),
  ENTRY     ("coordinates",   "", "", ' ', TRUE,  WorldMenuCB, 45),
  ENTRY     ("polygonCenter", "", "", ' ', TRUE,  WorldMenuCB, 46)
};

static MENU_ENTRY WorldMeasureMenu[] = 
{
  // TODO
  MENU_LINE (PrgETCHED_IN)
};

MENU_ENTRY WorldEditMenu[] =
{
  ENTRY     ("clear",         "", "", ' ', TRUE,  WorldMenuCB, 50),
  ENTRY     ("delete",        "", "", ' ', TRUE,  WorldMenuCB, 60),
  ENTRY     ("copy",          "", "", ' ', FALSE, WorldMenuCB, 70),
  ENTRY     ("paste",         "", "", ' ', FALSE, WorldMenuCB, 75)
};

MENU_ENTRY WorldSimuMenu[] = 
{
  ENTRY     ("run",           "", "", ' ', FALSE, WorldMenuCB, 2001),
  ENTRY     ("stop",          "", "", ' ', FALSE, WorldMenuCB, 2002),
  ENTRY     ("step",          "", "", ' ', FALSE, WorldMenuCB, 2003),
  ENTRY     ("reset",         "", "", ' ', FALSE, WorldMenuCB, 2004),
  ENTRY     ("set",           "", "", ' ', FALSE, WorldMenuCB, 2005),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("control",       "", "", ' ', FALSE, WorldMenuCB, 2010)
};

MENU_ENTRY WorldMenu[] =
{
  TITLE     ("worldMenu"),
  MENU_LINE (PrgDOUBLE_LINE),
  ENTRY     ("properties",    "", "", ' ', TRUE,  WorldMenuCB, 110),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("translate",     "", "", ' ', TRUE,  WorldMenuCB, 10),
  ENTRY     ("rotate",        "", "", ' ', TRUE,  WorldMenuCB, 20),
  // TODO: rotate around COI (Middle of Ortho)
  ENTRY     ("fly",           "", "", ' ', TRUE,  WorldMenuCB, 25),
  ENTRY     ("pan",           "", "", ' ', TRUE,  WorldMenuCB, 30),
  SUB_MENU  ("zoom",          "", "", ' ', TRUE,  WorldZoomMenu),
  MENU_LINE (PrgETCHED_IN),
  SUB_MENU  ("edit",          "", "", ' ', TRUE,  WorldEditMenu),
  MENU_LINE (PrgETCHED_IN),
  SUB_MENU  ("select",        "", "", ' ', TRUE,  WorldSelectMenu),
  SUB_MENU  ("view",          "", "", ' ', TRUE,  WorldViewMenu),
  SUB_MENU  ("measure",       "", "", ' ', FALSE, WorldMeasureMenu),
  MENU_LINE (PrgETCHED_IN),
  SUB_MENU  ("simulator",     "", "", ' ', FALSE, WorldSimuMenu),
  MENU_LINE (PrgETCHED_IN),
  SUB_MENU  ("file",          "", "", ' ', TRUE,  WorldFileMenu),
  MENU_LINE (PrgETCHED_IN),
  ENTRY     ("info",          "", "", ' ', TRUE,  WorldMenuCB, 120)
};


int WorldMenuEntries = sizeof(WorldMenu) / sizeof(MENU_ENTRY);


static void WorldMenuCB (int ID)
{
  unsigned long  UpdateMode = 0;
  WORLD         *World = GetCurrent ();

  if (World == NULL)
  {
    printf ("Invalid World\n");
    return;
  }

  switch (ID)
  {
    case  10 :                          // Translate world
      SetAction (World, DATA::TRANS_ABSOLUTE);
      UpdateMode = DLG_WORLD;
      break;

    case  20 :                          // Rotate world
      SetAction (World, DATA::ROT_ABSOLUTE);
      UpdateMode = DLG_WORLD;
      break;

    case  25 :                          // Fly
      SetAction (World, DATA::FLY_MODE);
      UpdateMode = DLG_WORLD;
      break;

    case  30 :                          // Pan
      SetAction (World, DATA::PAN_MODE);
      UpdateMode = DLG_WORLD;
      break;

    case 401 :                          // Zoom interactive
      SetAction (World, DATA::ZOOM_INTERACTIVE);
      UpdateMode = DLG_WORLD;
      break;

    case 402 :                          // Zoom by discrete steps
      SetAction (World, DATA::ZOOM_DISCRETE);
      UpdateMode = DLG_WORLD;
      break;

    case 403 :                          // Fill world
      World->Fill ();
      UpdateMode = DLG_WORLD;
      break;

    case  43 :                          // Select vertex
    case  44 :                          //        edge
    case  45 :                          //        coordinate system
    case  46 :                          //        polygon center
      {
	double  px, py, pz;
	PART3D *Part;

	PRGPrintf ("select point");
	World->GetWin()->SetMouseCursor (ToolNORMAL);
	if ((Part = World->PickPart(&px, &py, &pz)) != NULL)
	  if (ID == 43)                 // Show vertex coodinates
	  {
	    // TODO: suche Vertex
	  }
  	  elsif (ID == 44)              // Print edge coodinates
	  {
	    // TODO: suche Kante
	  }
	  elsif (ID == 45)              // Show selection coodinates 
	    PRGPrintf ("Point: %.3f %.3f %.3f", px, py, pz);
	  elsif (ID == 46)
	  {
	    POLYGON *Polygon;
	    if (PARTFindPolygon(Part->GetPart(), px, py, pz, NULL, &Polygon))
	      PRGPrintf ("Polygon: %.3f %.3f %.3f", 
			 Polygon->Info.Center.x, Polygon->Info.Center.y, 
			 Polygon->Info.Center.z);
	    else
	      PRGPrintf ("No polygon found\n");
	  }
	  else
	    PRGPrintf ("Pos: %.3f %.3f %.3f", px, py, pz);
	else
	  PRGPrintf ("\\"MSG_ABORT);
      }
      break;

    case  50 :                          // Clear world
      if (DLGOpenTextDialog("clearWorld", DlgWARNING, 
			    "Really clear this world ?", 2, "NO", "YES") == 1)
        FUNC_ABORT
      World->Clear ();
      UpdateMode = DLG_TREE;
      break;

    case  60 :                          // Delete world
      if (DLGOpenTextDialog("deleteWorld", DlgWARNING, 
			    "Really delete this world ?", 2, "NO", "YES") == 1)
        FUNC_ABORT
      DeleteWorld (World);
      UpdateMode = DLG_TREE;
      break;

    case  70 :                          // Copy buffer
      break;

    case 75 :                           // Paste buffer
      {
	int               count;
	DEVICE3D         *Dev;
	LIST<DEVICE3D *> *List;

	if ((List = Buffer.GetBuffer()) == NULL)
	  return;

	for (count = 0; count < List->Size(); ++count)
	{
	  Dev = List->Get (count);
	  World->AddDevice (new DEVICE3D(Dev));
	}
	World->Redraw ();
	UpdateMode = DLG_TREE;
      }
      break;

    case 801 :                          // Load part
      LoadPart (USE_WORLD);
      UpdateMode = DLG_TREE;
      break;

    case 802 :                          // Load device
      LoadDev (USE_WORLD);
      UpdateMode = DLG_TREE;
      break;

    case 803 :                          // Load world
      LoadWorld ();
      UpdateMode = DLG_TREE;
      break;

    case  90 :                          // Save world
      SaveWorld (0);
      break;

    case  96 :                          // Print world
      World->Print ();
      break;

    case 110 :                          // World properties
      WorldProperties (USE_WORLD);
      break;

    case 120 :                          // World info
      WorldInfo (World);
      break;

    case 1001 :                         // View world tree
      OpenTreeDialog (World);
      break;

    case 1002 :                         // Single-View
      World->SetViews (1);
      break;

    case 1003 :                         // Double-View
      World->SetViews (2 | 4);
      break;

    case 1004 :                         // Quad-View
      World->SetViews (8 | 16 | 32 | 64);
      break;

    case 1005 :                         // 1+3 View
      World->SetViews (128 | 256 | 512 | 1024);
      break;

    case 1006 :                         // COI
      {
        H_MATRIX COI = Matrix_1;
        
        World->GetCOI (&COI.m[P][X], &COI.m[P][Y], &COI.m[P][Z]);
        UPDATE(World)->DrawFrame (COI);
      }
      break;

    default : 
      SetAction (World, DATA::NORMAL_MODE);
      PRO1 ("Invalid Function %d\n", ID)
  }

  WorldMouseCursor (World);
  UpdateDialogs    (World, UpdateMode);
}


void OpenWorldMenu (Widget Shell, WORLD *World, 
		    int MouseX /* = -1 */, int MouseY /* = -1 */)
{
  int ViewNo;

  assert (World != NULL);

  ViewNo = World->GetWin()->GetViewNo ();;

  if (MouseX == -1 || MouseY == -1)
    TOOLGetMousePos (&MouseX, &MouseY);

  WorldEditMenu[3].active = (Buffer.Size() > 0);

  WorldViewMenu[2].active = (ViewNo > 1);
  WorldViewMenu[3].active = (ViewNo != 2);
  WorldViewMenu[4].active = (ViewNo < 4);

  PRGCreateMenu (Shell, "worldMenu", MouseX, MouseY, 
		        WorldMenu, WorldMenuEntries);

  if (World != NULL)
    World->Redraw ();
}


