#include "tuxCAD.h"


/* (+ = changeable)
    PART-Properies:
    Attrib:      +Name, +FileName, Parent, 
                 +Color, +Material (Regler fr Ambient, Diffuse, ...), +Texture
                 +Render (Wire, Point, ...), +Blend (visible, transparent),
                 +Shade(flat/smooth), +Show Edges, ...
    Values:      +Position+Orientation (RPY/Frame): rel/abs, 
                 +Joint-Value (rel/abs), Range, CS
    Shape:       +Scale (abs/rel), *Rotation
    Info:         #Points, #Faces, #Radius, #Volume, ... 
                 +Attachments (del/add), +DOF(dx,dy,dz,rx,ry,rz), +Ranges, ...
    Operations:  +Add/Sub/XOr/
*/


int PART_ChangeCB (DIALOG *Dlg, int BoxID)
{
  PART_DIALOG *PartDlg = (PART_DIALOG *)DLGGetUserData (Dlg, 0);

  if (PartDlg != NULL)
    PartDlg->ChangeCB (BoxID);

  return DlgACCEPT;
}


PART_DIALOG::PART_DIALOG (WORLD *World) :
  CAD_DIALOG (),
  World      (World),
  Part       (NULL),
  Handle     (NULL)
{
  PRO_METHOD ("PART_DIALOG", "")

  assert (World != NULL);

  PRO_END
}


bool PART_DIALOG::Open (PART3D *Part)
{
  int     Axes, Mode;
  double  rpy[6];

  PRO_METHOD ("PART_DIALOG", "\n")

  if ((this->Part = Part) == NULL)
    PRO_RETURN ("No part specified", FALSE)

  if (! CAD_DIALOG::Open())
    PRO_RETURN ("Opening dialog failed", FALSE)

  Axes = Part->GetJoint().GetAxes ();
  GEOFrame2RPY (Part->GetMatrix(), &rpy[0], &rpy[1], &rpy[2],
		                   &rpy[3], &rpy[4], &rpy[5]);

  DLGInsertNote (Dlg, 0, "note", DlgNOBINDING);
    DLGAddPage (Dlg, 0, DlgMAJOR, "global", 100);
      DLGMakeRows (Dlg, 100, 5, DlgNO_LINE, 110, 120, 160, 165, 190);
      DLGMakeRows (Dlg, 120, 2, DlgNO_LINE, 125, 130);
      DLGMakeCols (Dlg, 130, 4, DlgNO_LINE, 140, 145, 150, 155);
      DLGMakeRows (Dlg, 140, 3, DlgNO_LINE, 141, 142, 143);
      DLGMakeRows (Dlg, 145, 3, DlgNO_LINE, 146, 147, 148);
      DLGMakeRows (Dlg, 150, 3, DlgNO_LINE, 151, 152, 153);
      DLGMakeRows (Dlg, 155, 3, DlgNO_LINE, 156, 157, 158);
      DLGMakeCols (Dlg, 165, 2, DlgNO_LINE, 166, 167);

      DLGInsertText  (Dlg, 110, "name", 50,
		      (Part->GetName()==NULL)? " " : Part->GetName(), 30, 1);
      DLGInsertOMenu (Dlg, 125, "ref", 2, "local", "world", 0);

      DLGInsertLabel (Dlg, 141, "frame", "Frame");
      DLGInsertLabel (Dlg, 146, "xdir", "x"); 
      DLGInsertLabel (Dlg, 151, "ydir", "y"); 
      DLGInsertLabel (Dlg, 156, "zdir", "z"); 
      DLGInsertDouble (Dlg, 147, "pos", rpy[0], 7, 2);
      DLGInsertDouble (Dlg, 152, "ry",  rpy[1], 7, 2);
      DLGInsertDouble (Dlg, 157, "rz",  rpy[2], 7, 2);
      DLGInsertDouble (Dlg, 148, "dir", rpy[3], 7, 2);
      DLGInsertDouble (Dlg, 153, "ry",  rpy[4], 7, 2);
      DLGInsertDouble (Dlg, 158, "rz",  rpy[5], 7, 2);
      DLGSetLabelPos (Dlg, 146, DlgEAST, DlgEAST);
      DLGSetLabelPos (Dlg, 151, DlgEAST, DlgEAST);
      DLGSetLabelPos (Dlg, 156, DlgEAST, DlgEAST);
      DLGSetLabelPos (Dlg, 152, DlgNOWHERE, DlgNOWHERE);
      DLGSetLabelPos (Dlg, 157, DlgNOWHERE, DlgNOWHERE);
      DLGSetLabelPos (Dlg, 153, DlgNOWHERE, DlgNOWHERE);
      DLGSetLabelPos (Dlg, 158, DlgNOWHERE, DlgNOWHERE);
      DLGInsertScale (Dlg, 160, "normalAngle", 0.0, 180.0, 
		      Part->GetNormalAngle(), 1, DlgHORIZONTAL);
      Mode = Part->GetMode ();
      DLGInsertTBox (Dlg, 166, "mode", 1, 3, 
		               "part",  ((Mode & PART3D::DRAW_PART) != 0), 
		               "box",   ((Mode & PART3D::DRAW_BOX)  != 0), 
		               "coord", ((Mode & PART3D::DRAW_AXES)   != 0));
      BoxDialog (Dlg, 167, Part->GetBox());
      DLGSetBorder   (Dlg, 120, DlgETCHED_IN);
      DLGSetBorder   (Dlg, 167, DlgETCHED_IN);
      DLGSetBoxActive (Dlg, 167, (Part->GetMode() > 1));
      // TODO: + Linienfarbe
      DLGInsertScale (Dlg, 190, "lines", 0.0, 8.0, 
		                         Part->GetMaterial()->GetLine(), 
		                         0, DlgHORIZONTAL);

    DLGAddPage (Dlg, 0, DlgMAJOR, "material", 200);
      DLGMakeRows (Dlg, 200, 2, DlgETCHED_IN, 220, 280);
      MaterialDialog (Dlg, 220, Part->GetMaterial(), 
		                OBJ_MASK | TEX_MASK | ALPHA_MASK);

    DLGAddPage (Dlg, 0, DlgMAJOR, "kinematic", 300);
      JointDialog (Dlg, 300, Part);

  DLGSetUserData     (Dlg, 0, this);
  DLGSetFunction     (Dlg, 0, PART_ChangeCB);
  DLGSetHelpFunction (Dlg, HelpDlgCB);

  DLGMakeButton (Dlg, DEFAULT_BUTTON);
  DLGMakeButton (Dlg, HELP_BUTTON);
  DLGMakeButton (Dlg, OK_BUTTON);
  DLGOpenDialog (Dlg, "partProperties", DlgNORMAL, NULL);

  Close ();

  PRO_END
  return TRUE;
}


void PART_DIALOG::ChangeCB (int BoxID)
{
#if (XmVERSION >= 2)
  int Page = DLGGetNote (Dlg, 0);

  PRO_METHOD ("PART_DIALOG", "")

  CAD_DIALOG::ChangeCB (BoxID);

  switch (Page)
  {
    case 1 :                             // Global
      {
	char *PartName;

	switch (BoxID)
        {
          case 110 :
	    PartName = DLGGetText (Dlg, BoxID);
	    PARTSetName (Part->GetPart(), PartName);
	    break;

  	  case 160 :
	    Part->SetNormalAngle (DLGGetScale(Dlg, BoxID));
	    World->Redraw ();
	    break;

	  case 166 :
	    {
	      int Mode = 0;

	      if (DLGGetTBox(Dlg, BoxID, 0))
		Mode += PART3D::DRAW_PART;
	      if (DLGGetTBox(Dlg, BoxID, 1))
		Mode += PART3D::DRAW_BOX;
	      if (DLGGetTBox(Dlg, BoxID, 2))
		Mode += PART3D::DRAW_AXES;
	      Part->SetMode (Mode);
	      DLGSetBoxActive (Dlg, 167, (Mode > 1));
	      World->Redraw ();
	    }
	    break;

 	  case 190 :
	    Part->GetMaterial()->SetLine ((int)DLGGetScale(Dlg, BoxID));
	    Part->Create ();	    
	    World->Redraw ();
	    break;

   	  default :
	    BoxCB (Dlg, 167, Part->GetBox(), BoxID);
	    World->Redraw ();
	}
      }
      break;

    case 2 :                            // Material
      {
	double    OldAlpha;
	MATERIAL *Material = Part->GetMaterial ();

	OldAlpha = Material->GetAlpha ();
	MaterialCB (Dlg, 220, Material, BoxID, 
		              OBJ_MASK | TEX_MASK | ALPHA_MASK);
	if ((OldAlpha == 1.0 || OldAlpha == 0.0) && 
	    Material->GetAlpha() > 0.0 && Material->GetAlpha() < 1.0)
	  World->IncAlpha ();
	elsif (OldAlpha > 0.0 && OldAlpha < 1.0 && 
	       (Material->GetAlpha() == 0.0 || Material->GetAlpha() == 1.0))
	  World->DecAlpha ();

	Part->Create ();
	World->Redraw ();
      }
      break;

    case 3 :
      {
	JointCB (Dlg, 300, Part, BoxID);
#if 0
	int count, Axes = 0;
	for (count = 0; count < 6; ++count)
	  if (DLGGetTBox(Dlg, 310, count))
	    Axes += (1<<count);
	Part->GetJoint().SetAxes (Axes);
#endif
      }
      break;

    default :
      PRO0 ("Invalid Page\n");
  }
#endif

  if (Handle != NULL)
    Handle (Dlg, BoxID, Part);

  PRO_END
}


void PART_DIALOG::Update ()
{
  PRO_METHOD ("PART_DIALOG", "\n")

  if (Dlg == NULL)
    PRO_ABORT ("Dialog is not open")

  // TODO: ...

  PRO_END
}


void PART_DIALOG::Set (PART3D *Part)
{
  PRO_METHOD ("PART_DIALOG", "\n")

  if (Dlg == NULL)
    PRO_ABORT ("Dialog is not open")

  if ((this->Part = Part) == NULL)
    PRO_ABORT ("No part defined")

  // TODO:...

  PRO_END
}


void PART_DIALOG::HelpCB (int ID)
{
  int  Page = DLGGetNote (Dlg, 0);
  char Label[20] = "";

  switch (Page)
  {
    case 1 :
      if (ID == 110)
	strcpy (Label, "name");
      elsif (ID == 125 || (ID >= 147 && ID <= 158))
	strcpy (Label, "frame");
      elsif (ID == 160)
	strcpy (Label, "angel");
      elsif (ID == 166)
	strcpy (Label, "mode");
      elsif (ID >= 167)
	strcpy (Label, "box");
      else
	strcpy (Label, "global");
      break;

    case 2 :
      if (ID >= 220 && ID < 280)
	strcpy (Label, "material");
      elsif (ID == 280)
	strcpy (Label, "lines");
      else
	strcpy (Label, "material");
      break;

    case 3 :
      strcpy (Label, "kinematic");
      break;

    default : 
      break;
  }
  HelpDlg.Open ("PartProperty", "dlg_part.hlp", Label);
}


void PART_DIALOG::OnChange (CHANGE_HANDLE Handle)
{
  this->Handle = Handle;
}


void PartProperties (WORLD *World, PART3D *Part)
{
  if (World == NULL)
    return;

  // Unhighlight current part, so color changes are visible
  World->Select (Part, FALSE);

  OpenPartDialog (World, Part);

#if 1
  // Do not keep selection of this part
  ClearSelection (World);
#else
  // Highlight part again
  World->Select (Part, TRUE);
#endif
}



/*
  DEVICE-Properties:
    +Name, DOF, +Kinematic & DH-Param, TCP-Offset, +Pos/Orientation, 
*/

int DEV_ChangeCB (DIALOG *Dlg, int BoxID)
{
  DEV_DIALOG *DevDlg = (DEV_DIALOG *)DLGGetUserData (Dlg, 0);

  if (DevDlg != NULL)
    DevDlg->ChangeCB (BoxID);

  return DlgACCEPT;
}


DEV_DIALOG::DEV_DIALOG (WORLD *World) :
  CAD_DIALOG (),
  World      (World),
  Dev        (NULL),
  Handle     (NULL)
{
  PRO_METHOD ("DEV_DIALOG", "")

  assert (World != NULL);

  PRO_END
}


bool DEV_DIALOG::Open (DEVICE3D *Dev)
{
  PRO_METHOD ("WORLD_DIALOG", "\n")

  if ((this->Dev = Dev) == NULL)
    PRO_RETURN ("No device specified", FALSE)

  if (! CAD_DIALOG::Open())
    PRO_RETURN ("Opening dialog failed", FALSE)

  DLGInsertNote (Dlg, 0, "page", DlgNOBINDING);
    DLGAddPage (Dlg, 0, DlgMAJOR, "global", 100);
      DLGInsertText   (Dlg, 100, "name", 40, Dev->GetName(), 40, 1);
      // Pos & Rot 

    DLGAddPage (Dlg, 0, DlgMAJOR, "kinematic", 200);
      DLGMakeRows (Dlg, 200, 2, DlgETCHED_IN, 210, 25);
      DLGMakeCols (Dlg, 210, 2, DlgNO_LINE, 211, 212);

      DLGInsertLabel (Dlg, 211, "dof", "DOF:");
      DLGInsertLabel (Dlg, 212, "dof", "%d", Dev->GetDOF());
      // DHP,  (Array)
      DLGMakeGrid (Dlg, 25, 5, 7);
      DLGInsertLabel (Dlg, DlgGRID(25,1,0), "a",     "a");
      DLGInsertLabel (Dlg, DlgGRID(25,2,0), "d",     "d");
      DLGInsertLabel (Dlg, DlgGRID(25,3,0), "alpha", "alpha");
      DLGInsertLabel (Dlg, DlgGRID(25,4,0), "delta", "delta");

      DLGInsertLabel (Dlg, DlgGRID(25,0,1), "0",      "0");
      DLGInsertLabel (Dlg, DlgGRID(25,1,1), "a0",     "a0");
      DLGInsertLabel (Dlg, DlgGRID(25,2,1), "d0",     "d0");
      DLGInsertLabel (Dlg, DlgGRID(25,3,1), "alpha0", "alpha0");
      DLGInsertLabel (Dlg, DlgGRID(25,4,1), "delta0", "delta0");

      DLGInsertLabel (Dlg, DlgGRID(25,0,2), "1",      "1");
      DLGInsertLabel (Dlg, DlgGRID(25,1,2), "a1",     "a1");
      DLGInsertLabel (Dlg, DlgGRID(25,2,2), "d1",     "d1");
      DLGInsertLabel (Dlg, DlgGRID(25,3,2), "alpha1", "alpha1");
      DLGInsertLabel (Dlg, DlgGRID(25,4,2), "delta1", "delta1");

      DLGInsertLabel (Dlg, DlgGRID(25,0,3), "2",      "2");
      DLGSetLabelPos (Dlg, 25, DlgNOWHERE, DlgWEST);

    DLGAddPage (Dlg, 0, DlgMAJOR, "tools", 300);

  DLGSetUserData     (Dlg, 0, this);
  DLGSetFunction     (Dlg, 0, DEV_ChangeCB);
  DLGSetHelpFunction (Dlg, HelpDlgCB);

  DLGMakeButton (Dlg, DEFAULT_BUTTON);
  DLGMakeButton (Dlg, HELP_BUTTON);
  DLGMakeButton (Dlg, OK_BUTTON);
  DLGOpenDialog (Dlg, "devProperties", DlgNORMAL, NULL);

  Close ();

  PRO_END
  return TRUE;
}


void DEV_DIALOG::Update ()
{
  PRO_METHOD ("DEV_DIALOG", "\n")

  if (Dlg == NULL)
    PRO_ABORT ("Dialog is not open")

  // TODO: ...

  PRO_END
}


void DEV_DIALOG::Set (DEVICE3D *Dev)
{
  PRO_METHOD ("DEV_DIALOG", "\n")

  if (Dlg == NULL)
    PRO_ABORT ("Dialog is not open")

  if ((this->Dev = Dev) == NULL)
    PRO_ABORT ("No device defined")

  // TODO:

  PRO_END
}


void DEV_DIALOG::ChangeCB (int BoxID)
{  
  PRO_METHOD ("DEV_DIALOG", "")

  CAD_DIALOG::ChangeCB (BoxID);

  // Nothing to do, yet.

  if (Handle != NULL)
    Handle (Dlg, BoxID, Dev);

  PRO_END
}


void DEV_DIALOG::HelpCB (int ID)
{
  HelpDlg.Open ("DevConfig", "dlg_dev.hlp", "");
}


void DEV_DIALOG::OnChange (CHANGE_HANDLE Handle)
{
  this->Handle = Handle;
}


void DevProperties (WORLD *World, DEVICE3D *Dev)
{
  if (World == NULL || Dev == NULL)
    return;

   // Unhighlight the selected device
  World->Select (Dev, FALSE);

  OpenDevDialog (World, Dev);

  // Highlight device again
  World->Select (Dev, TRUE);
}



/*  
   WORLD-Properties:
     +Name, +Views, +Render(front/back), +Shade (flat/smooth), 
     +Shadow(on/off,Face,...), +Units(mm/cm/m/inch), Spiegelung (an/aus)
     +Blend(alpha), +ViewPort(min-max), +Projection(orthographic/perspective)
     +Grid(on/size/mode,no), +Color(bg/grid), 
     +Light(#,pos,brightness,color), Fog +Camera, 
     +Motion Blur (on/off/Size,...)
*/

int WORLD_ChangeCB (DIALOG *Dlg, int BoxID)
{
  WORLD_DIALOG *WorldDlg = (WORLD_DIALOG *)DLGGetUserData (Dlg, 0);

  if (WorldDlg != NULL)
    WorldDlg->ChangeCB (BoxID);

  return DlgACCEPT;
}


WORLD_DIALOG::WORLD_DIALOG (WORLD *World) :
  CAD_DIALOG (),
  World      (World),
  Handle     (NULL)
{
  PRO_METHOD ("WORLD_DIALOG", "\n")

  assert (World != NULL);

  PRO_END
}


bool WORLD_DIALOG::Open ()
{
  double   *Range;
  FLOOR    *Floor;
  WINDOW3D *Win;

  PRO_METHOD ("WORLD_DIALOG", "\n")

  if (! CAD_DIALOG::Open())
    PRO_RETURN ("Opening dialog failed", FALSE)

  Win   = World->GetWin   ();
  Range = World->GetRange ();

#if (XmVERSION >= 2)
  DLGInsertNote (Dlg, 0, "page", DlgNOBINDING);
    // Global:
    DLGAddPage (Dlg, 0, DlgMAJOR, "global", 100);
      DLGMakeRows     (Dlg, 100, 2, DlgNO_LINE, 101, 180);
      DLGMakeCols     (Dlg, 101, 2, DlgETCHED_IN, 105, 150);
      DLGMakeRows     (Dlg, 105, 2, DlgETCHED_IN, 110, 120);
      DLGMakeRows     (Dlg, 120, 6, DlgNO_LINE, 121, 122, 123, 124, 125, 126);
      DLGMakeCols     (Dlg, 150, 2, DlgNO_LINE, 160, 170);
      DLGMakeCols     (Dlg, 180, 2, DlgNO_LINE, 181, 190),
      DLGMakeRows     (Dlg, 181, 4, DlgNO_LINE, 182, 183, 184, 185);
      DLGInsertText   (Dlg, 110, "name", 40, World->GetName(), 20, 1);
      DLGInsertOMenu  (Dlg, 121, "surface", 4, 
		                 "point", "wireframe", "hiddenline", "solid",
	                    (int)Win->GetPolygonMode());
      DLGInsertOMenu  (Dlg, 122, "shading", 3, "none", "flat", "smooth", 
		            (int)Win->GetShading());
      DLGInsertOMenu  (Dlg, 123, "blending", 2, "solid", "alpha", 
		            (int)Win->GetBlending());
      DLGInsertOMenu  (Dlg, 124, "culling", 4, "off", "front", "back", "both",
		            (int)Win->GetCulling());
      DLGInsertOMenu  (Dlg, 125, "surfaceBack", 4, 
		                 "point", "wireframe", "hiddenline", "solid",
	                    (int)Win->GetPolygonMode(WINDOW3D::BACK_SIDE));
      DLGInsertOMenu  (Dlg, 126, "antialiasing", 3, 
		                 "off", "fastest", "nicest", 
		            (int)Win->GetAliasing());
      COLOR Color = World->GetPartHighLight();
      ColorDialog (Dlg, 160, Color, ALPHA_MASK);
      ColorDialog (Dlg, 170, World->GetDevHighLight(), ALPHA_MASK);

      DLGInsertOMenu (Dlg, 182, "coordinates", 2, "hide", "show",
		      TOOLGetConfigBool(CFG_SHOW_COORDS));

      DLGSetBorder (Dlg, 160, DlgETCHED_IN);
      DLGSetBorder (Dlg, 170, DlgETCHED_IN);

      DLGSetTitle (Dlg, 160, "partHighLight", DlgLEFT);
      DLGSetTitle (Dlg, 170, "devHighLight",  DlgLEFT);

    // View:
    DLGAddPage (Dlg, 0, DlgMAJOR, "view", 200);
      DLGMakeRows     (Dlg, 200, 2, DlgNO_LINE,   210, 215);
      DLGMakeCols     (Dlg, 210, 2, DlgNO_LINE,   211, 212);
      DLGInsertOMenu  (Dlg, 211, "views", 6, 
		                 "full", "half", "quarter", "mixed", 
		                 NULL, "userDefined", 0);
      DLGInsertOMenu  (Dlg, 212, "viewNumber", 4, "1", "2", "3", "4", 0);
      ViewDialog (Dlg, 215, World, 0);

    // Light:
    DLGAddPage (Dlg, 0, DlgMAJOR, "light", 300);
      DLGMakeRows     (Dlg, 300, 2, DlgNO_LINE, 305, 310);
      DLGInsertOMenu  (Dlg, 305, "lightNumber", 8, 
		                 "light_1", "light_2", "light_3", "light_4",
		                 "light_5", "light_6", "light_7", "light_8", 
		                 0);
      DLGSetLabelPos  (Dlg, 305, DlgWEST, DlgCENTER);
      DLGSetBorder    (Dlg, 310, DlgETCHED_IN);
      LightDialog     (Dlg, 310, World->GetLight(LIGHT::LIGHT1));

    // Shadow (Subpage):
      DLGAddPage (Dlg, 0, DlgMINOR, "shadow", 400);
        ShadowDialog   (Dlg, 400, World->GetLight(LIGHT::LIGHT1)->GetShadow());

    // Fog:
    DLGAddPage (Dlg, 0, DlgMAJOR, "fog", 500);
      FogDialog (Dlg, 500, Win->GetFog());

    // Background:
    DLGAddPage (Dlg, 0, DlgMAJOR, "background", 600);
      ColorDialog (Dlg, 600, World->GetWin()->GetBackground(), BASE_MASK);

    // Floor:
    DLGAddPage (Dlg, 0, DlgMAJOR, "floor", 700);
      Floor = World->GetFloor (); 
      DLGMakeCols     (Dlg, 700, 2, DlgNO_LINE, 705, 750);
      DLGMakeRows     (Dlg, 705, 2, DlgNO_LINE, 710, 720);
      DLGMakeRows     (Dlg, 720, 2, DlgNO_LINE, 730, 740);
      DLGMakeRows     (Dlg, 730, 5, DlgNO_LINE, 731, 732, 733, 734, 735);
      DLGMakeRows     (Dlg, 740, 2, DlgNO_LINE, 741, 742);
      DLGInsertOMenu  (Dlg, 710, "floorShow", 2, "OFF", "ON", 
		                                 int(Floor->GetMode()));
      DLGInsertUpDn   (Dlg, 731, "sizeX",     0, 250, Floor->GetSizeX(), 1);
      DLGInsertUpDn   (Dlg, 732, "sizeY",     0, 250, Floor->GetSizeY(), 1);
      DLGInsertDouble (Dlg, 733, "spacingX",  Floor->GetSpaceX(), 8, 3);
      DLGInsertDouble (Dlg, 734, "spacingY",  Floor->GetSpaceY(), 8, 3);
      DLGInsertDouble (Dlg, 735, "height",    (*Floor)[PZ], 8, 3);
      DLGInsertScale  (Dlg, 741, "lineWidth", 0, 8, Floor->GetLineWidth(), 
		                              0, DlgHORIZONTAL);
      ColorDialog (Dlg, 742, Floor->GetGridColor(), BASE_MASK);
      MaterialDialog (Dlg, 750, Floor->GetMaterial(), TEX_MASK | ALPHA_MASK);
      DLGSetLabelPos  (Dlg, 710, DlgWEST, DlgCENTER);

      DLGSetBorder    (Dlg, 705, DlgETCHED_IN);
      DLGSetBorder    (Dlg, 750, DlgETCHED_IN);
      DLGSetBorder    (Dlg, 730, DlgETCHED_IN);
      DLGSetBorder    (Dlg, 740, DlgETCHED_IN);
#endif

  DLGSetBoxActive (Dlg, 125, ((int)Win->GetCulling() < 2));

  DLGSetUserData     (Dlg, 0, this);
  DLGSetFunction     (Dlg, 0, WORLD_ChangeCB);
  DLGSetHelpFunction (Dlg, HelpDlgCB);

  DLGMakeButton (Dlg, DEFAULT_BUTTON);
  DLGMakeButton (Dlg, HELP_BUTTON);
  DLGMakeButton (Dlg, OK_BUTTON);
  DLGOpenDialog (Dlg, "worldProperties", DlgNORMAL, NULL);

  Close ();
  
  PRO_END
  return TRUE;
}


void WORLD_DIALOG::ChangeCB (int BoxID)
{
  int       Page = DLGGetNote (Dlg, 0);
  WINDOW3D *Win = World->GetWin ();

  CAD_DIALOG::ChangeCB (BoxID);

  switch (Page)
  {
    case  1 :  // Global
      switch (BoxID)
      {
        case 110 :
	  World->SetName (DLGGetText(Dlg, BoxID));
	  break;
	  
        case 121 : 
	  Win->SetPolygonMode ((WINDOW3D::POLYGON_MODE)DLGGetOMenu(Dlg, 
								   BoxID));
	  break;

        case 122 : 
	  Win->SetShading  ((WINDOW3D::SHADING)DLGGetOMenu(Dlg, BoxID));
	  break;

        case 123 : 
	  Win->SetBlending ((WINDOW3D::BLENDING)DLGGetOMenu(Dlg, BoxID));
	  break;

        case 124 : 
	  Win->SetCulling ((WINDOW3D::CULLING)DLGGetOMenu(Dlg, BoxID));
	  DLGSetBoxActive (Dlg, 125, (DLGGetOMenu(Dlg, BoxID) < 2));
	  break;

        case 125 :
	  Win->SetPolygonMode ((WINDOW3D::POLYGON_MODE)DLGGetOMenu(Dlg, BoxID),
			       WINDOW3D::BACK_SIDE);
	  break;

        case 126 :
	  Win->SetAliasing ((WINDOW3D::ALIASING)DLGGetOMenu(Dlg, BoxID));
	  break;

        case 182 :
	  TOOLSetConfigBool (CFG_SHOW_COORDS, (dual)DLGGetOMenu(Dlg, BoxID));
	  break;

        default :
	  {
	    double OldAlpha;
	      
	    if (BoxID < 170)
	    {
	      OldAlpha = World->GetPartHighLight()[ALPHA];
	      World->SetPartHighLight(ColorCB(Dlg, 160,
					           World->GetPartHighLight(), 
					           BoxID, ALPHA_MASK));
	      if ((OldAlpha == 1.0 || OldAlpha == 0.0) &&
		  World->GetPartHighLight()[ALPHA] > 0.0 &&
		  World->GetPartHighLight()[ALPHA] < 1.0)
		World->IncAlpha ();
	      elsif (OldAlpha > 0.0 && OldAlpha < 1.0 &&
		     (World->GetPartHighLight()[ALPHA] == 0.0 ||
		      World->GetPartHighLight()[ALPHA] == 1.0))
		World->DecAlpha ();
	    }
	    elsif (BoxID < 180)
	    {
	      OldAlpha = World->GetDevHighLight()[ALPHA];
	      World->SetDevHighLight(ColorCB(Dlg, 170,
					     World->GetDevHighLight(), BoxID,
					     ALPHA_MASK));
	      if ((OldAlpha == 1.0 || OldAlpha == 0.0) &&
		  World->GetDevHighLight()[ALPHA] > 0.0 &&
		  World->GetDevHighLight()[ALPHA] < 1.0)
		World->IncAlpha ();
	      elsif (OldAlpha > 0.0 && OldAlpha < 1.0 &&
		     (World->GetDevHighLight()[ALPHA] == 0.0 ||
		      World->GetDevHighLight()[ALPHA] == 1.0))
		World->DecAlpha ();
	    }
	    else
	      PRO1 ("WorldPropertyCB: Not implemented, yet: %d\n", BoxID)
	  }
      }
      break;

    case 2 :  // Views
      {
	int Selection = DLGGetOMenu (Dlg, 211);
	int ViewNo    = DLGGetOMenu (Dlg, 212);

	if (BoxID == 211)
	{
	  switch (Selection)
	  {
	    case 0 : World->SetViews (1);
	             break;
	    case 1 : World->SetViews (2 | 4);
	             break;
	    case 2 : World->SetViews (8 | 16 | 32 | 64);
  	             break;
	    case 3 : World->SetViews (128 | 256 | 512 | 1024);
	             break;
  	    default : PRGPrintf ("Sorry not supported yet");
	  }
	  DLGSetBoxActive (Dlg, 212, (Selection > 0));
	  SetView (Dlg, 215, World, ViewNo);
	}
	elsif ((BoxID == 212 && ! SetView(Dlg, 215, World, ViewNo)) ||
	       (BoxID != 212 && 
		ViewCB(Dlg, 215, World, ViewNo, BoxID) == NULL))
	  DLGSetOMenu (Dlg, 212, 0);
      }
      break;

    case 3 :  // Lights
      {
	int     LightNo = (int)DLGGetOMenu (Dlg, 305);
	LIGHT *Light = World->GetLight ((LIGHT::NUMBER)DLGGetOMenu(Dlg, 305));
	SHADOW *Shadow = World->GetLight((LIGHT::NUMBER)LightNo)->GetShadow ();

	if (BoxID == 305)             // Change the light source / shadow
	{
	  SetLight  (Dlg, 310, Light);
	  SetShadow (Dlg, 400, Shadow);
	}
	else
	  LightCB (Dlg, 310, Light, BoxID);
      }
      break;

    case 4 : // Shadow
      // Set shadow color for current lights
      {
	int     LightNo = (int)DLGGetOMenu (Dlg, 305);
	SHADOW *Shadow = World->GetLight((LIGHT::NUMBER)LightNo)->GetShadow ();

	ShadowCB (Dlg, 400, Shadow, BoxID);
      }
      break;

    case 5 :   // Fog (subpage)
      FogCB (Dlg, 500, Win->GetFog(), BoxID);
      break;

    case 6 :  // Background
      Win->SetBackground (ColorCB(Dlg, 600, Win->GetBackground(),
				       BoxID, BASE_MASK));
      break;

    case 7 :  // Grid
      {
	double    OldAlpha;
	FLOOR    *Floor = World->GetFloor ();
	MATERIAL *Material = Floor->GetMaterial();

	switch (BoxID)
	{
	  case 710 :
	    Floor->SetMode ((FLOOR::MODE)DLGGetOMenu(Dlg, BoxID));
	    break;
	     
	  case 731 :
	  case 732 :
	    Floor->SetSize ((int)DLGGetUpDn(Dlg, 731), 
			    (int)DLGGetUpDn(Dlg, 732));
	    break;

	  case 733 :
	  case 734 :
	    Floor->SetSpacing (DLGGetDouble(Dlg, 733), DLGGetDouble(Dlg, 734));
	    break;

	  case 735 :
	    Floor->SetPos (0.0, 0.0, DLGGetDouble(Dlg, BoxID));
	    break;

	  case 741 :
	    Floor->SetLineWidth ((int)DLGGetScale(Dlg, BoxID));
	    break;

	  default :
	    if (BoxID < 750) 
	      Floor->SetGridColor (ColorCB(Dlg, 742, Floor->GetGridColor(),
					             BoxID, BASE_MASK));
	    else
	    {
	      // All changes will be handled by the material page...
	      OldAlpha = Material->GetAlpha();
	      MaterialCB (Dlg, 750, Material, BoxID, TEX_MASK);
	      if ((OldAlpha == 1.0 || OldAlpha == 0.0) && 
		  Material->GetAlpha() > 0.0 && Material->GetAlpha() < 1.0)
		World->IncAlpha ();
	      elsif (OldAlpha > 0.0 && OldAlpha < 1.0 && 
		     (Material->GetAlpha() == 0.0 || 
		      Material->GetAlpha() == 1.0))
		World->DecAlpha ();
	      Floor->Create ();
	    }
	    break;
	}
      }
      break;

    default : 
      break;
  }

  if (Handle != NULL)
    Handle (Dlg, BoxID, World);

  World->Redraw ();
}


void WORLD_DIALOG::Update ()
{
  PRO_METHOD ("WORLD_DIALOG", "\n")

  if (Dlg == NULL)
    PRO_ABORT ("Dialog is not open")

  // TODO: ...

  PRO_END
}


void WORLD_DIALOG::HelpCB (int ID)
{
  int  Page = DLGGetNote (Dlg, 0);
  char Label[20] = "";

  switch (Page)
  {
    case 1 :
      strcpy (Label, "global");
      break;

    case 2 :
      strcpy (Label, "view");
      break;

    case 3 :
      strcpy (Label, "light");
      break;

    case 4 :
      strcpy (Label, "shadow");
      break;

    case 5 :
      strcpy (Label, "fog");
      break;

    case 6 :
      strcpy (Label, "background");
      break;

    case 7 :
      strcpy (Label, "floor");
      break;

    default :
      break;
  }

  HelpDlg.Open ("WorldConfig", "dlg_world.hlp", Label);
}


void WORLD_DIALOG::OnChange (CHANGE_HANDLE Handle)
{
  this->Handle = Handle;
}


void WorldProperties (int Select)
{
  WORLD *World = GetCurrent();

  PRO_START ("")

  if (World == NULL || Select == SELECT_WORLD)
    if ((World = SelectWorld(FALSE)) == NULL)
    {
      PRGPrintf ("Function aborted");
      PRO_ABORT ("Function aborted")
    }

  OpenWorldDialog (World);

  PRO_END;
}
