#include "tuxCAD.h"


typedef struct
{
  int       JointNo;
  int       Ref;
  DEVICE3D *Robot;
} TEACH_DATA;


#if 0
int ROBOT::SetJoint (int No, double Val)
{
  if (No > JointNo)
    return -1;

  if (Val < Joints[No].Min || Val > Joints[No].Max)
    return -2;

  Joints[No].Val = Val;

  return 1;
}
#endif

static void MakeText (DEVICE3D *Robot, char *Text)
{
  int      count;
  double   Euler[6] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
  double   Min[6], Val[6], Max[6];
  H_MATRIX WTH;
  RPY      rpy;

  WTH = Robot->GetTCP ();

PRO_HMATRIX (WTH, "WTH")

  GEOFrame2RPY   (WTH, &rpy.px, &rpy.py, &rpy.pz, &rpy.y, &rpy.p, &rpy.r);
  GEOFrame2Euler (WTH, &Euler[0], &Euler[1], &Euler[2], 
		       &Euler[3], &Euler[4], &Euler[5]);

  PRO6 ("EULER: %.3f %.3f %.3f %.3f %.3f %.3f\n",
        Euler[0], Euler[1], Euler[2], Euler[3], Euler[4], Euler[5])

  for (count = 0; count < 6; ++count)
    Robot->GetRange (count, Min[count], Val[count], Max[count]);

  sprintf (Text, 
	   "JOINTS: J1:%7.2f J2:%7.2f J3:%7.2f J4:%7.2f J5:%7.2f J6:%7.2f\n"
	   "RPY:    PX:%7.2f PY:%7.2f PZ:%7.2f  Y:%7.2f  P:%7.2f  R:%7.2f\n"
	   "EULER:  PX:%7.2f PY:%7.2f PZ:%7.2f Z1:%7.2f  Y:%7.2f Z2:%7.2f",
	   Val[0], Val[1], Val[2], Val[3], Val[4], Val[5],
	   rpy.px, rpy.py, rpy.pz, rpy.y, rpy.p, rpy.r,
	   Euler[0], Euler[1], Euler[2], Euler[3], Euler[4], Euler[5]);

WTH = GEOEuler2Frame (Euler[0], Euler[1], Euler[2], 
		      Euler[3], Euler[4], Euler[5]);
PRO_HMATRIX (WTH, "FrameEulerFrame")
}


static void SetDisplay (DEVICE3D *Robot, DIALOG *Dlg)
{
  char Text[256];

  MakeText (Robot, Text);

  DLGSetText (Dlg, 10, Text);
}


void SelectChange (DIALOG *Dlg, int BoxID)
{
  double      Min(0.0), Max(0.0), Val(0.0);
  DEVICE3D   *Robot;
  TEACH_DATA *Teach;

  assert (Dlg != NULL);

  Teach = (TEACH_DATA *)DLGGetUserData (Dlg, 0);

  Robot = Teach->Robot;
  switch (BoxID)
  {
    case 30 :
      switch (Teach->Ref)
      {
        case 0 :
        case 1 :
	  // TODO: Inverse Kinematik berechnen
	  PRO0 ("No Inv. Kinematic defined\n")
	  break;

        case 2 :
#if 0
	    Val = DLGGetScale (Dlg, 30);
	    Robot->GetRange (Teach->JointNo, Min, Max, Val);
	    Robot->Joint[Teach->JointNo]->Part->WTP =
	    GEOMatrixProd (Robot->Joint[Teach->JointNo]->Part->WTP,
			   GEOMakeRot (GEOMakeVector(0.0, 0.0, 1.0), 
				       Val-Robot->Joint[Teach->JointNo]->Val));
	  Robot->Joint[Teach->JointNo]->Val = Val;
#endif
	  break;

        default :
	  PRO0 ("Undefined Reference\n")
      }
      GetCurrent()->Redraw ();
      break;

    case 50 :
      Teach->Ref = DLGGetRBox (Dlg, 50);
      break;

    case 101 :
    case 102 : 
    case 103 :
    case 104 :
    case 105 :
    case 106 :
      Teach->JointNo = BoxID - 100;
      Robot->GetRange (Teach->JointNo, Min, Max, Val);
      DLGChangeScale (Dlg, 30, Min, Max, 1);
      DLGSetScale (Dlg, 30, Val);
      break;

    default :
      printf ("unknown component\n");
  }

  SetDisplay (Robot, Dlg);
}

#include "./pic/button.xpm"
#include "./pic/edit.xpm"
#include "./pic/disp.xpm"
#include "./pic/clrerr.xpm"
#include "./pic/ruler.xpm"
#include "./pic/grip.xpm"
#include "./pic/estop.xpm"
#include "./pic/x1.xpm"
#include "./pic/y2.xpm"
#include "./pic/z3.xpm"
#include "./pic/rx4.xpm"
#include "./pic/ry5.xpm"
#include "./pic/rz6.xpm"
#include "./pic/t1.xpm"

int ButtonCB (DIALOG *Dlg, int BoxID)
{
  if (Dlg == NULL)
    return DlgDENY;

  switch (BoxID)
  {
    default :
      PRO1 ("Unsupported Button %d\n", BoxID)
  }
  return DlgCONFIRM;
}



void TeachPanel ()
{
  int       row, col;
  char      Text[400];
  DIALOG   *Dlg;
  DEVICE3D *Robot;
#if 0
  TEACH_DATA  Teach = {1, 2, NULL};
#endif

  PRGPrintf ("Select Robot");
  if (! CheckWorld() || (Robot = GetCurrent()->PickDev()) == NULL)
  {
    PRGPrintf ("No Robot found\n");
    return;
  }

  PRGPrintf ("OK");

  MakeText (Robot, Text);

  Dlg = DLGNewDialog (0);

  DLGMakeRows (Dlg,   0, 4, DlgNO_LINE, 10, 20, 30, 100);
  DLGMakeCols (Dlg,  20, 9, DlgNO_LINE, 21, 22, 23, 24, 25, 26, 27, 28, 29);
  DLGMakeCols (Dlg,  30, 9, DlgNO_LINE, 31, 32, 33, 34, 35, 36, 37, 38, 39);
  DLGMakeCols (Dlg, 100, 3, DlgNO_LINE, 200, 300, 400);
  DLGMakeRows (Dlg, 200, 2, DlgNO_LINE, 210, 220);
  DLGMakeCols (Dlg, 220, 2, DlgNO_LINE, 230, 250);
  DLGMakeRows (Dlg, 250, 2, DlgNO_LINE, 251, 255);
  DLGMakeRows (Dlg, 255, 3, DlgSINGLE_LINE DlgSINGLE_LINE, 256, 257, 258);
  DLGMakeRows (Dlg, 300, 2, DlgNO_LINE, 310, 350);
  DLGMakeRows (Dlg, 310, 2, DlgSINGLE_LINE, 311, 320);
  DLGMakeCols (Dlg, 311, 2, DlgSINGLE_LINE, 312, 313);
  DLGMakeCols (Dlg, 320, 3, DlgNO_LINE,     321, 322, 323);
  DLGMakeRows (Dlg, 400, 8, DlgNO_LINE,
	                    401, 402, 403, 404, 405, 406, 407, 408);
  DLGMakeGrid (Dlg, 350, 3, 5);

  DLGSetBorder (Dlg,  20, DlgSHADOW_IN);
  DLGSetBorder (Dlg,  30, DlgSHADOW_IN);
  DLGSetBorder (Dlg, 251, DlgSHADOW_IN);
  DLGSetBorder (Dlg, 255, DlgSHADOW_IN);
  DLGSetBorder (Dlg, 310, DlgSHADOW_IN);
  DLGSetBorder (Dlg, 350, DlgSHADOW_IN);
  DLGSetBorder (Dlg, 400, DlgSHADOW_IN);

  DLGInsertText  (Dlg,  10, "display", 65, Text, 65, 3);
  DLGSetLabelPos (Dlg,  10, DlgNORTH_WEST, DlgWEST);  

  DLGInsertPic  (Dlg,  21, "edit",   DlgXPM, -1, -1, edit_xpm,   ButtonCB);
  DLGInsertPic  (Dlg,  22, "B2",     DlgXPM, -1, -1, button_xpm, ButtonCB);
  DLGInsertPic  (Dlg,  23, "disp",   DlgXPM, -1, -1, disp_xpm,   ButtonCB);
  DLGInsertPic  (Dlg,  24, "B4",     DlgXPM, -1, -1, button_xpm, ButtonCB);
  DLGInsertPic  (Dlg,  25, "clrErr", DlgXPM, -1, -1, clrerr_xpm, ButtonCB);
  DLGInsertPic  (Dlg,  26, "B6",     DlgXPM, -1, -1, button_xpm, ButtonCB);
  DLGInsertPic  (Dlg,  27, "B7",     DlgXPM, -1, -1, button_xpm, ButtonCB);
  DLGInsertPic  (Dlg,  28, "B8",     DlgXPM, -1, -1, button_xpm, ButtonCB);
  DLGInsertPic  (Dlg,  29, "B9",     DlgXPM, -1, -1, button_xpm, ButtonCB);

  DLGInsertPic  (Dlg, 210, "ruler",  DlgXPM, -1, -1, ruler_xpm,   NULL);
  DLGInsertPic  (Dlg, 230, "grip",   DlgXPM, -1, -1, grip_xpm,    NULL);

  DLGInsertPic  (Dlg, 251, "slow",   DlgXPM, -1, -1, button_xpm,   ButtonCB);
  DLGInsertPic  (Dlg, 256, "S1",     DlgXPM, -1, -1, button_xpm,   ButtonCB);
  DLGInsertPic  (Dlg, 257, "S2",     DlgXPM, -1, -1, button_xpm,   ButtonCB);
  DLGInsertPic  (Dlg, 258, "S3",     DlgXPM, -1, -1, button_xpm,   ButtonCB);

  DLGInsertPic  (Dlg, 312, "estop",    DlgXPM, -1, -1, estop_xpm,   ButtonCB);
  DLGInsertPic  (Dlg, 313, "manHalt",  DlgXPM, -1, -1, button_xpm,  ButtonCB);
  DLGInsertPic  (Dlg, 321, "runHold",  DlgXPM, -1, -1, button_xpm,  ButtonCB);
  DLGInsertPic  (Dlg, 322, "disPwr",   DlgXPM, -1, -1, button_xpm,  ButtonCB);
  DLGInsertPic  (Dlg, 323, "compPwr",  DlgXPM, -1, -1, button_xpm,  ButtonCB);

  for (row = 0; row < 5; ++row)
    for (col = 0; col < 3; ++col)
      DLGInsertPic (Dlg, DlgGRID(350,col,row), " ", 
		    DlgXPM, -1, -1, button_xpm, ButtonCB);

  DLGInsertPic  (Dlg, 401, "x1",    DlgXPM, -1, -1, x1_xpm, ButtonCB);
  DLGInsertPic  (Dlg, 402, "y2",    DlgXPM, -1, -1, y2_xpm, ButtonCB);
  DLGInsertPic  (Dlg, 403, "z3",    DlgXPM, -1, -1, z3_xpm, ButtonCB);
  DLGInsertPic  (Dlg, 404, "rx4",   DlgXPM, -1, -1, rx4_xpm, ButtonCB);
  DLGInsertPic  (Dlg, 405, "ry5",   DlgXPM, -1, -1, ry5_xpm, ButtonCB);
  DLGInsertPic  (Dlg, 406, "rz6",   DlgXPM, -1, -1, rz6_xpm, ButtonCB);
  DLGInsertPic  (Dlg, 407, "t1",    DlgXPM, -1, -1, t1_xpm, ButtonCB);
  DLGInsertPic  (Dlg, 408, "step",  DlgXPM, -1, -1, button_xpm, ButtonCB);
  DLGSetLabelPos (Dlg,  20, DlgNOWHERE, DlgCENTER);
  DLGSetLabelPos (Dlg,  30, DlgNOWHERE, DlgCENTER);
  DLGSetLabelPos (Dlg, 100, DlgNOWHERE, DlgCENTER);

  DLGOpenDialog (Dlg, "teachPanel",  DlgNORMAL, NULL);

  DLGFreeDialog (Dlg);
}


