#include <Xm/MainW.h>
#include <Xm/MwmUtil.h>
#include <Xm/Form.h>
#include <Xm/RowColumn.h>
#include <Xm/ScrolledW.h>
#include <Xm/DrawingA.h>
#include <Xm/PushB.h>
#include <Xm/ScrollBar.h>
#include <Xm/Label.h>
#include <Xm/Separator.h>

#include "tuxCAD.h"

#include "pic/start.xpm"
#include "pic/reverse.xpm"
#include "pic/back.xpm"
#include "pic/stop.xpm"
#include "pic/step.xpm"
#include "pic/run.xpm"
#include "pic/end.xpm"


#define X_SCALE   5
#define RULER_Y   25
#define FONT_MASK "-adobe-helvetica-medium-r-*-*-10-*-*-*-*-*-iso8859-1" 

/* TODO:
   - besseres Refresh (aktuellen Position nicht immer sichtbar )
   - Parameter-Dlg fr Scheduler: Speed, Skalierung (Zoom), Font? 
   - Parameter-Dlg fr Task: Farbe, Start-/End-Frame & -Zeit, Motionparam.
   - Tasks verschieben, stretchen, ...
   - Beim Springen der aktuellen Zeit mssen alle Parts aktualisiert werden
     (auch diejenigen, die nicht in diesem Zeitpunkt einen Task besitzen, d.h.
     Position aus Endframe des letzen Tasks ermitteln oder fr Startzeit 
     entweder Position zu Beginn merken, oder Startframe von erstem Task 
     bestimmen.
*/

static bool UpdateDialog (SCHEDULER *Scheduler, double Time, TASK_DIALOG *Dlg)
{
  Dlg->DrawTime ();
  Dlg->Refresh ((int)Time-1, 0, (int)Time+1, 300);

  return TRUE;
}


TASK_DIALOG::TASK_DIALOG (WORLD *World, SCHEDULER *Scheduler /* = NULL */) :
  World   (World),
  LineNo  (0),
  Lines   (NULL),
  Shell   (NULL),
  ScrWin  (NULL),
  ClipWin (NULL),
  ScrBarX (NULL),
  ScrBarY (NULL),
  Work    (NULL),
  SizeY   (100),
  LineY   (0),
  Win     (None),
  Font    (NULL)
{
  if (Scheduler == NULL)
    this->Scheduler = new SCHEDULER (World);
  else
    this->Scheduler = Scheduler;

  Scheduler->SetEvent ((SCHEDULER::SCHEDULER_EVENT)UpdateDialog, this);
}


TASK_DIALOG::~TASK_DIALOG ()
{
  if (Scheduler != NULL)
    delete Scheduler;
}


static void SimulatorHandle (Widget widget, TASK_DIALOG *Dlg, XEvent *Event)
{
  switch (Event->type)
  {
    case Expose :
      {
	XExposeEvent *expose = (XExposeEvent *)Event;

	Dlg->Refresh (expose->x, expose->y, expose->width, expose->height);
      }
      break;
      
    default :
      break;
  }
}


void TASK_DIALOG::SetAbsTime (double Time)
{
  SCHEDULER *Scheduler = GetScheduler ();

  DrawTime ();
  //  Scheduler->SetTime (Time);
  Scheduler->Execute (Time);
  Refresh ((int)Time-1, 0, (int)Time+2, SizeY);
}


void TASK_DIALOG::SetRelTime (double Time)
{
  double     AbsTime;
  SCHEDULER *Scheduler = GetScheduler ();

  AbsTime = Scheduler->GetTime() + Time;
  SetAbsTime (AbsTime);
}


static void StartCB (Widget widget, TASK_DIALOG *Dlg, XtPointer Data)
{
  Dlg->SetAbsTime (0.0);
}


static void ReverseCB (Widget widget, TASK_DIALOG *Dlg, XtPointer Data)
{
  double     Speed;
  SCHEDULER *Scheduler = Dlg->GetScheduler ();

  Speed = Scheduler->GetSpeed ();
  if (Speed > 0.0) 
    Speed *= -1.0;
  Scheduler->SetSpeed (Speed);
  Scheduler->Run ();
}


static void BackCB (Widget widget, TASK_DIALOG *Dlg, XtPointer Data)
{
  Dlg->SetRelTime (-1.0);
}


static void StopCB (Widget widget, TASK_DIALOG *Dlg, XtPointer Data)
{
  SCHEDULER *Scheduler = Dlg->GetScheduler ();
  Scheduler->Stop ();
}


static void StepCB (Widget widget, TASK_DIALOG *Dlg, XtPointer Data)
{
  Dlg->SetRelTime (1.0);
}


static void RunCB (Widget widget, TASK_DIALOG *Dlg, XtPointer Data)
{
  double     Speed;
  SCHEDULER *Scheduler = Dlg->GetScheduler ();

  Speed = Scheduler->GetSpeed ();
  if (Speed < 0.0)
    Speed *= -1.0;
  Scheduler->SetSpeed (Speed);
  Scheduler->Run ();
}


static void EndCB (Widget widget, TASK_DIALOG *Dlg, XtPointer Data)
{
  Dlg->SetAbsTime (0.0);
}


int TASK_DIALOG::GetLine (TASK *Task) const
{
  int   count;
  char *Name;

  if (Lines != NULL)
    for (count = 0; count < LineNo; ++count)
    {
      Name = Task->GetName ();
      if (Name != NULL && strcmp(Lines[count].Name, Name) == 0)
      {
	delete Name;
	return count;
      }
      delete Name;
    }

  return -1;
}


bool TASK_DIALOG::Open ()
{
  int        count, Line;
  int        PicX, PicY;
  Dimension  h;
  Arg        args[10];
  Pixel      back;
  XGCValues  GCVal;
  Widget     Form, Sep;
  Widget     List, Label;
  Pixmap     Pic;
  TASK      *Task;

  PRO_METHOD ("TASK_DIALOG", "\n")

  Lines = new TASK_LINE[10];

  for (count = 0; count < Scheduler->GetTaskNo(); ++count)
  {
    Task = Scheduler->GetTask (count);

    if ((Line = GetLine(Task)) < 0)     // if the task does not exist, add it.
    {
      Lines[LineNo].Name  = Task->GetName ();
      Lines[LineNo].Color = TOOLGetColor  ("ForestGreen");
      ++LineNo;
    }
  }

  if (Shell != NULL)
    PRO_RETURN ("Opening dialog failed", FALSE)

  XtSetArg (args[0], XmNwidth,               400);
  XtSetArg (args[1], XmNheight,              300);
  XtSetArg (args[2], XmNmwmFunctions,        MWM_FUNC_ALL);
  XtSetArg (args[3], XmNmwmDecorations,      MWM_DECOR_ALL);
  XtSetArg (args[4], XmNdeleteResponse,      XmDESTROY);
  XtSetArg (args[5], XmNkeyboardFocusPolicy, XmPOINTER);
  Shell = XtCreatePopupShell ("taskControl", transientShellWidgetClass,
                                             AppShell, args, 6);
  XtManageChild (Shell);

  Form = XtVaCreateManagedWidget ("form", xmFormWidgetClass, Shell,
				  XmNleftAttachment,   XmATTACH_FORM,
				  XmNleftOffset,       5,
				  XmNrightAttachment,  XmATTACH_FORM,
				  XmNrightOffset,      5,
				  XmNbottomAttachment, XmATTACH_FORM,
				  XmNbottomOffset,     5,
				  NULL);    
  XtVaGetValues (Form, XmNbackground, &back, NULL);

  Label = XtVaCreateManagedWidget ("time", xmLabelWidgetClass, Form,
				   XmNtopAttachment,    XmATTACH_FORM,
				   XmNleftAttachment,   XmATTACH_FORM,
				   XmNalignment,        XmALIGNMENT_END,
				   NULL);

  List = XtVaCreateManagedWidget ("list", xmRowColumnWidgetClass, Form,
				  XmNtopAttachment,  XmATTACH_FORM,
				  XmNtopOffset,      RULER_Y,
				  XmNleftAttachment, XmATTACH_FORM,
				  XmNorientation,    XmVERTICAL,
				  NULL);

  /* List of CAD-objects */
  for (count = 0; count < LineNo; ++count)
    Label = XtVaCreateManagedWidget (Lines[count].Name, 
				     xmPushButtonWidgetClass, List,
				     NULL);

  Actions[0] = 
    XtVaCreateManagedWidget ("add", xmPushButtonWidgetClass, Form,
		                    XmNbottomAttachment, XmATTACH_FORM,
		                    XmNleftAttachment,   XmATTACH_FORM,
		                    NULL);
  //  XtAddCallback (Actions[0], XmNactivateCallback, (XtCallbackProc)Step, this);

  Actions[1] = 
    XtVaCreateManagedWidget ("del", xmPushButtonWidgetClass, Form,
		                    XmNbottomAttachment, XmATTACH_FORM,
		                    XmNleftAttachment,   XmATTACH_WIDGET,
			            XmNleftWidget,       Actions[0],
			            NULL);
  //  XtAddCallback (Actions[1], XmNactivateCallback, (XtCallbackProc)Step, this);

  Pic = TOOLMakeXPM (start_data, &PicX, &PicY, back);
  Actions[3] = 
    XtVaCreateManagedWidget ("start", xmPushButtonWidgetClass, Form,
		                      XmNbottomAttachment, XmATTACH_FORM,
  			              XmNbottomOffset,     3,
		                      XmNleftAttachment,   XmATTACH_WIDGET,
			              XmNleftWidget,       Actions[1],
 			              XmNleftOffset,       30,
			              XmNlabelType,        XmPIXMAP,
			              XmNlabelPixmap,      Pic,
		                      NULL);
  XtAddCallback (Actions[3], XmNactivateCallback, 
		             (XtCallbackProc)StartCB, this);

  Pic = TOOLMakeXPM (reverse_data, &PicX, &PicY, back);
  Actions[4] = 
    XtVaCreateManagedWidget ("reverse", xmPushButtonWidgetClass, Form,
		                        XmNbottomAttachment, XmATTACH_FORM,
  			                XmNbottomOffset,     3,
		                        XmNleftAttachment,   XmATTACH_WIDGET,
			                XmNleftWidget,       Actions[3],
			                XmNlabelType,        XmPIXMAP,
			                XmNlabelPixmap,      Pic,
		                        NULL);
  XtAddCallback (Actions[4], XmNactivateCallback, 
		             (XtCallbackProc)ReverseCB, this);

  Pic = TOOLMakeXPM (back_data, &PicX, &PicY, back);
  Actions[5] = 
    XtVaCreateManagedWidget ("back", xmPushButtonWidgetClass, Form,
		                      XmNbottomAttachment, XmATTACH_FORM,
  			              XmNbottomOffset,     3,
		                      XmNleftAttachment,   XmATTACH_WIDGET,
			              XmNleftWidget,       Actions[4],
			              XmNlabelType,        XmPIXMAP,
			              XmNlabelPixmap,      Pic,
		                      NULL);
  XtAddCallback (Actions[5], XmNactivateCallback,
		             (XtCallbackProc)BackCB, this);

  Pic = TOOLMakeXPM (stop_data, &PicX, &PicY, back);
  Actions[6] = 
    XtVaCreateManagedWidget ("stop", xmPushButtonWidgetClass, Form,
		                      XmNbottomAttachment, XmATTACH_FORM,
  			              XmNbottomOffset,     3,
		                      XmNleftAttachment,   XmATTACH_WIDGET,
			              XmNleftWidget,       Actions[5],
			              XmNlabelType,        XmPIXMAP,
			              XmNlabelPixmap,      Pic,
		                      NULL);
  XtAddCallback (Actions[6], XmNactivateCallback,  
                             (XtCallbackProc)StopCB, this);

  Pic = TOOLMakeXPM (step_data, &PicX, &PicY, back);
  Actions[7] = 
    XtVaCreateManagedWidget ("step", xmPushButtonWidgetClass, Form,
		                     XmNbottomAttachment, XmATTACH_FORM,
  			             XmNbottomOffset,     3,
		                     XmNleftAttachment,   XmATTACH_WIDGET,
			             XmNleftWidget,       Actions[6],
			             XmNlabelType,        XmPIXMAP,
			             XmNlabelPixmap,      Pic,
		                     NULL);
  XtAddCallback (Actions[7], XmNactivateCallback,
		             (XtCallbackProc)StepCB, this);

  Pic = TOOLMakeXPM (run_data, &PicX, &PicY, back);
  Actions[8] = 
    XtVaCreateManagedWidget ("run", xmPushButtonWidgetClass, Form,
		                    XmNbottomAttachment, XmATTACH_FORM,
  			            XmNbottomOffset,     3,
		                    XmNleftAttachment,   XmATTACH_WIDGET,
			            XmNleftWidget,       Actions[7],
			            XmNlabelType,        XmPIXMAP,
			            XmNlabelPixmap,      Pic,
		                    NULL);
  XtAddCallback (Actions[8], XmNactivateCallback, (XtCallbackProc)RunCB, this);

  Pic = TOOLMakeXPM (end_data, &PicX, &PicY, back);
  Actions[9] = 
    XtVaCreateManagedWidget ("end", xmPushButtonWidgetClass, Form,
		                    XmNbottomAttachment, XmATTACH_FORM,
  			            XmNbottomOffset,     3,
		                    XmNleftAttachment,   XmATTACH_WIDGET,
			            XmNleftWidget,       Actions[8],
			            XmNlabelType,        XmPIXMAP,
			            XmNlabelPixmap,      Pic,
		                    NULL);
  XtAddCallback (Actions[9], XmNactivateCallback,
		             (XtCallbackProc)EndCB, this);

  /* Create a separator */
  Sep = XtVaCreateManagedWidget ("separator", xmSeparatorWidgetClass, Form,
				 XmNleftAttachment,   XmATTACH_FORM,
				 XmNrightAttachment,  XmATTACH_FORM,
				 XmNbottomAttachment, XmATTACH_WIDGET,
				 XmNbottomWidget,     Actions[7],
				 XmNbottomOffset,     2,
				 XmNseparatorType,    XmSHADOW_ETCHED_IN,
				 NULL);
  ScrWin =
    XtVaCreateManagedWidget ("scrWin", xmScrolledWindowWidgetClass, Form,
                             XmNscrollingPolicy,        XmAUTOMATIC,
                             XmNscrollBarDisplayPolicy, XmAS_NEEDED,
                             XmNvisualPolicy,           XmCONSTANT,
                             XmNtopAttachment,          XmATTACH_FORM,
                             XmNleftAttachment,         XmATTACH_WIDGET,
			     XmNleftWidget,             List,
                             XmNrightAttachment,        XmATTACH_FORM,
                             XmNbottomAttachment,       XmATTACH_WIDGET,
			     XmNbottomWidget,           Sep,
                             XmNbottomOffset,           2,
                             NULL);
  XtVaGetValues (ScrWin, XmNverticalScrollBar,   &ScrBarY,
                         XmNhorizontalScrollBar, &ScrBarX,
                         XmNclipWindow,          &ClipWin,
                         NULL);           

  XtVaGetValues (List, XmNheight, &h, NULL);
  SizeY = RULER_Y + h+1;
  LineY = h / LineNo;

  Work = 
    XtVaCreateManagedWidget ("textArea", xmDrawingAreaWidgetClass, ScrWin, 
			     XmNwidth, (int)(Scheduler->GetMaxTime()*X_SCALE)+1,
			     XmNheight, SizeY, 
			     NULL);
  XtAddEventHandler (Work, KeyPressMask    | PointerMotionMask |
                           EnterWindowMask | LeaveWindowMask   |
                           ButtonPressMask | ButtonReleaseMask |
                           ExposureMask    | StructureNotifyMask, FALSE,
                           (XtEventHandler)SimulatorHandle, this);
  Win = XtWindow (Work);

  GCVal.graphics_exposures = False;
  DrawGC = XCreateGC (XDisplay, Win, GCGraphicsExposures, &GCVal);
  XSetForeground (XDisplay, DrawGC, TOOLGetColor("Black"));

  XtVaGetValues (Work, XmNbackground, &bg, NULL);
  XSetForeground (XDisplay, xorGC, bg ^ TOOLGetColor("Red"));

  if ((Font = XLoadQueryFont(XDisplay, FONT_MASK)) == NULL)
    printf ("ERROR LOADING FONT: '%s'\n", FONT_MASK);

  XtRealizeWidget (Shell);
  XtPopup (Shell, XtGrabNonexclusive);

  PRO_END
  return TRUE;
}


SCHEDULER *TASK_DIALOG::GetScheduler ()
{
  return Scheduler;
}


void TASK_DIALOG::Refresh (int x, int y, int w, int h)
{
  int     count;
  double  Start, End;
  TASK   *Task;

  DrawRuler (x, x+w+1);

  for (count = 0; count < Scheduler->GetTaskNo(); ++count)
  {
    Task = Scheduler->GetTask (count);
    Start = Task->GetStart() * X_SCALE;
    End   = Task->GetEnd()   * X_SCALE;

    if ((Start >= x && Start <= x+w) || (End >= x || End <= x+w) || 
	(Start < x && End > x+w))
      DrawTask (Task);
  }

  DrawTime ();
}


void TASK_DIALOG::DrawRuler (int x1, int x2) const
{
  int       countX, countY;
  int       SizeX;
  char      ValStr[10];
  XTextItem Word;

  XSetForeground (XDisplay, DrawGC, TOOLGetColor("White"));
  XFillRectangle (XDisplay, Win, DrawGC, x1, 0, x2-x1, RULER_Y);

  // TODO:
  // Es gibt noch probleme, wenn ein Wert zuvor in den neu gezeichneten 
  // Bereich hereinragt, aber deren Mittelposition nicht zum refresh-Bereich
  // gehrt. Das hat zur Folge, dass ein Teil des Textes berschrieben wird.
  // => ggf. auch noch die vorhergehende und nchste Skalierung zeichnen 

  XSetForeground (XDisplay, DrawGC, TOOLGetColor("Black"));
  countX = x1 - (x1 % X_SCALE);         // Calc preceeding raster value
  while (countX <= x2)
  {
    if ((countX % (10*X_SCALE)) == 0)
    {
      XDrawLine (XDisplay, Win, DrawGC, countX, 12, countX, RULER_Y);
      sprintf (ValStr, "%d", (int)(0.2*countX));
      Word.chars  = ValStr;
      Word.nchars = strlen (ValStr);
      Word.font   = Font->fid;
      Word.delta  = 0;
      SizeX = XTextWidth (Font, ValStr, Word.nchars);
      XDrawText (XDisplay, Win, DrawGC, (int)(countX-0.5*SizeX), 10, &Word, 1);
    }
    elsif ((countX % (5*X_SCALE)) == 0)
      XDrawLine (XDisplay, Win, DrawGC, countX, 16, countX, RULER_Y);    
    else
      XDrawLine (XDisplay, Win, DrawGC, countX, 20, countX, RULER_Y);

    countX += X_SCALE;
  }

  for (countY = 0; countY <= LineNo; ++countY)
    XDrawLine (XDisplay, Win, DrawGC, x1, RULER_Y + countY*LineY, 
	                              x2, RULER_Y + countY*LineY);
}


void TASK_DIALOG::DrawTask (TASK *Task) const
{
  int    n = 2, h, Line;
  int    Start, End, PosY;
  int    count;
  XPoint Points[6];

  if (Task == NULL)
    return;

  if ((Line = GetLine(Task)) < 0) 
    return;
  Start = (int)(Task->GetStart() * X_SCALE);
  End   = (int)(Task->GetEnd()   * X_SCALE);
  PosY  = (int)(RULER_Y + Line*LineY + 0.5*LineY+1);

  h = (int)(0.25 * LineY);
  if (End - Start <= 2*h)
    h = (int)(0.5 * (End - Start));

  Points[0].x = Start;      Points[0].y = PosY;
  Points[1].x = Start+h;    Points[1].y = PosY-h;
  if (End - Start > 2*h)
  {
    Points[n].x = End-h;   Points[n].y = PosY-h;  ++n;
  }
  Points[n].x = End;     Points[n].y = PosY;  ++n;
  if (End - Start > 2*h)
  {
    Points[n].x = End-h;   Points[n].y = PosY+h;  ++n;
  }
  Points[n].x = Start+h;    Points[n].y = PosY+h;  ++n;

  XSetForeground (XDisplay, DrawGC, Lines[Line].Color);
  XFillPolygon (XDisplay, Win, DrawGC, Points, n, Convex, CoordModeOrigin);
  XSetForeground (XDisplay, DrawGC, TOOLGetColor("Black"));
  for (count = 0; count < n-1; ++count)
    XDrawLine (XDisplay, Win, DrawGC, Points[count].x, Points[count].y,
 	                              Points[count+1].x, Points[count+1].y);
  XDrawLine (XDisplay, Win, DrawGC, Points[n-1].x, Points[n-1].y,
	                            Points[0].x,   Points[0].y);
}


void TASK_DIALOG::DrawTime () const
{
  /* Draw current position */
  XDrawLine (XDisplay, Win, xorGC, (int)Scheduler->GetTime()*X_SCALE, RULER_Y,
 	                           (int)Scheduler->GetTime()*X_SCALE, 1000);
}


// Idee fr Simulator-Dialog:
//
// Object       Time  0   5  10  15  20  25  30  35  40  45  50 ...
//                    +--+---+---+---+---+---+---+---+---+---+--- - -   ^ 
// Part1(-name)       |         []=========[]    |     []======[]*     | |
// Dev1(-name)        |   []=======[]            |                     | |
// Dev2(-name)        |            []==[]        |                     | |
// TCP1(Robot)        |      []=============[]===|========[]           [=]
// Part2(-name)       |                          |           []=====   | |
// ...                |                        <-|->                   | |
//                    +--------------------------+---------------- -    v
//                    <============[  ]=============================>
// 
// [Add] [Del] [Info]    [<|] [<<] [[]] [>] [>>] [|>]       [Close] [Help]
//
// * Click auf einen Node oder Task-Balken: 
//     Anzeige und nderung des Frames und der Parameter:
// 
//  +-------+-----+------+-------+
//  | Start | End | Time | Param | ...
//  +-------+-----+------+-------+---------------+
//  |                                            |
//  |                                          
//  |                                         
//
// Tasks lassen sich verschieben, Start / Endzeit/-frame verndern, 


void MotionControl ()
{
  PART3D      *Part1, *Part2, *Part3;
  DEVICE3D    *Dev;
  MATERIAL     Material = MATERIAL (COLOR::MEDRED);
  WORLD       *World;
  TASK        *Task[9];
  SCHEDULER   *Scheduler;
  TASK_DIALOG *Dlg;

  World = NewWorld ("Tasks");
  Scheduler = GetScheduler (World);
  Scheduler->SetSpeed (1.0);

  Part1 = new QUADER (World, "quader1", Matrix_1, 10.0, 10.0, 10.0, Material);
  Dev = new DEVICE3D ("device1", Part1);
  World->AddDevice (Dev);
  Part2 = new QUADER (World, "quader2", Matrix_1,  5.0,  5.0,  5.0, Material);
  Dev = new DEVICE3D ("device2", Part2);
  World->AddDevice (Dev);
  Part3 = new QUADER (World, "quader3", Matrix_1,  1.0,  1.0,  1.0, Material);
  Dev->AddPart (Part3);

  Task[0] = new PART_TASK (Part1, MOTION(FRAME(), FRAME(), 50.0), 0.0);
  Scheduler->AddTask (Task[0]);
  Task[1] = new PART_TASK (Part1, MOTION(FRAME(), FRAME(), 25.0), 80.0);
  Scheduler->AddTask (Task[1]);
  Task[2] = new PART_TASK (Part2, MOTION(FRAME(), FRAME(), 40.0), 30.0);
  Scheduler->AddTask (Task[2]);
  Task[3] = new PART_TASK (Part1, MOTION(FRAME(), FRAME(),  2.0), 195.0);
  Scheduler->AddTask (Task[3]);
  Task[4] = new PART_TASK (Part2, MOTION(FRAME(), FRAME(), 10.0), 220.0);
  Scheduler->AddTask (Task[4]);
  Task[5] = new PART_TASK (Part3, MOTION(FRAME(), FRAME(), 72.0), 25.0);
  Scheduler->AddTask (Task[5]);
  Task[6] = new PART_TASK (Part3, MOTION(FRAME(), FRAME(), 70.0), 115.0);
  Scheduler->AddTask (Task[6]);
  Task[7] = new DEV_TASK (Dev,    MOTION(FRAME(), FRAME(), 47.0),  5.0);
  Scheduler->AddTask (Task[7]);
  Task[8] = new TCP_TASK (Dev,    MOTION(FRAME(), FRAME(), 18.0), 61.0);
  Scheduler->AddTask (Task[8]);

  Dlg = new TASK_DIALOG (World, Scheduler);  
  Dlg->Open ();
}


void MotionTest1 ()
{
  PART3D      *Part1, *Part2;
  DEVICE3D    *Dev;
  WORLD       *World;
  H_MATRIX     Matrix;
  FRAME        Start, Goal;
  FRAME        Step;
  TASK        *Task[5];
  SCHEDULER   *Scheduler;
  TASK_DIALOG *Dlg;

  if ((Part1 = ChoosePart("Select part 1...")) == NULL)
    return;

  if ((Part2 = ChoosePart("Select part 2...")) == NULL)
    return;

  if ((Dev = ChooseDev("Select device...")) == NULL)
    return;

  World = GetCurrent ();
  Scheduler = GetScheduler (World);
  Scheduler->SetSpeed (1.0);

  // Add task for part 1
  Matrix = Part1->GetWTP();
  Start  = TO_FRAME (Matrix);
  Matrix = GEOMatrixProd (Matrix, GEORPY2Frame( 0.0, 0.0, 100.0,
					       90.0, 0.0, 180.0));
  Goal   = TO_FRAME (Matrix);
  Task[0] = new PART_TASK (Part1, MOTION(Start, Goal, 50.0), 0.0);
  Scheduler->AddTask (Task[0]);

  // Add task for part 2
  Matrix = Part2->GetWTP();
  Start  = TO_FRAME (Matrix);
  Matrix = GEOMatrixProd (Matrix, GEORPY2Frame(0.0, 0.0, 100.0,
					       0.0, 0.0,   0.0));
  Goal   = TO_FRAME (Matrix);
  Task[1] = new PART_TASK (Part2, MOTION(Start, Goal, 25.0), 25.0);
  Scheduler->AddTask (Task[1]);

  // Add tasks for device
  Matrix = Dev->GetWTD();
  Start  = TO_FRAME (Matrix);
  Matrix = GEOMatrixProd (Matrix, GEORPY2Frame( 0.0, 200.0, 0.0,
					        0.0,  0.0, 0.0));
  Goal   = TO_FRAME (Matrix);
  Task[2] = new DEV_TASK (Dev, MOTION(Start, Goal, 20.0), 10.0);
  Scheduler->AddTask (Task[2]);

  Start = Goal;
  Matrix = GEOMatrixProd (Matrix, GEORPY2Frame(200.0,  0.0, 0.0,
					         0.0,  0.0, 0.0));
  Goal   = TO_FRAME (Matrix);
  Task[3] = new DEV_TASK (Dev, MOTION(Start, Goal, 20.0), 30.0);
  Scheduler->AddTask (Task[3]);

  Dlg = new TASK_DIALOG (World, Scheduler);  
  Dlg->Open ();

#if 0
  // TODO: erst beim Schliessen des Dialogs:
  Scheduler->Clear ();
#endif
}


void MotionTest2 ()
{
  int        count;
  DEVICE3D  *Dev;
  H_MATRIX   WTR, WTT, RTT;
  FRAME      Start, Goal;
  FRAME      Step;
  WORLD     *World;
  TASK      *Task[10] = {NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL};
  SCHEDULER *Scheduler;

  if ((Dev = ChooseDev("Select device...")) == NULL)
    return;

  World = GetCurrent ();
  Scheduler = GetScheduler (World);
  Scheduler->SetSpeed (0.5);

  // Add Task for TCP
  WTR = Dev->GetWTD ();
  WTT = Dev->GetTCP ();
  RTT = GEOMatrixProd (GEOMatrixInv(WTR), WTT);

  Start = TO_FRAME (RTT);

  Goal = Start * FTrans (0.0, 300.0, 0.0);
  Task[0] = new TCP_TASK (Dev, MOTION(Start, Goal, 30.0), 0.0);
  Scheduler->AddTask (Task[0]);

  Start = Goal;
  Goal  = Start * FTrans (300.0, 0.0, 0.0); // * FRotZ (-90.0);
  Task[1] = new TCP_TASK (Dev, MOTION(Start, Goal, 60.0), 30.0);
  Scheduler->AddTask (Task[1]);

  Start = Goal;
  Goal  = TO_FRAME (RTT); // * FRotZ (-90.0);
  Task[2] = new TCP_TASK (Dev, MOTION(Start, Goal, 150.0), 90.0);
  Scheduler->AddTask (Task[2]);

  // Execute all tasks
  Scheduler->Run ();

  for (count = 0; count < 3; ++count)
    if (Task[count] != NULL)
    {
      Scheduler->DelTask (Task[count]);
      delete Task[count];
    }
}
