#include <sys/stat.h>
#include <ctype.h>

#include "tuxCAD.h"

#define NO_DIALOG

static char PartFile[PATH_MAX+1]   = "unnamed.part";
static char PartMask[PATH_MAX+1]   = CAD_PATH"PARTS/*.part,*.?"; /* */
static char DevFile[PATH_MAX+1]    = CAD_PATH"DEVICES/new.dev";
static char DevMask[PATH_MAX+1]    = CAD_PATH"DEVICES/*.dev"; /* */
static char WorldFile[PATH_MAX+1]  = CAD_PATH"WORLDS/new.world";
static char WorldMask[PATH_MAX+1]  = CAD_PATH"WORLDS/*.world"; /* */
static char AutoFile[PATH_MAX+1]   = CAD_PATH;
static char AutoMask[PATH_MAX+1]   = CAD_PATH"/*.world,*.dev,*.part,*.?"; /* */
static char ImportFile[PATH_MAX+1] = CAD_PATH"IMPORTS/new.p5b";
static char ImportMask[PATH_MAX+1] = CAD_PATH"IMPORTS/*.p5b"; /* */
static char CfgFile[PATH_MAX+1]    = CAD_PATH"tuxcad.cfg";
static char CfgMask[PATH_MAX+1]    = CAD_PATH"*.cfg";


bool OverwriteFile (const char *FileName)
{
         char Message[PATH_MAX+30];
  struct stat Info;

  if (stat(FileName, &Info) != 0)   // File does not exist
    return TRUE;                  
  else 
  {
    sprintf (Message, MSG_ASK_FILE_OVERWRITE, FileName);
    if (DLGOpenTextDialog("overwriteFile", DlgWARNING,
			                   Message, 2, MSG_NO, MSG_YES) == 2)
    return TRUE;
  }

  PRGPrintf (MSG_ABORT);

  return FALSE;
}


void NewWorld ()
{
  NewWorld (NULL);
}


void Load (int Select)
{
  char   Extension[32];
  char   ErrorMsg[40+PATH_MAX+1] = "";
  WORLD *World = GetCurrent ();

  PRGPrintf (MSG_LOAD_FILE);

  if (World == NULL || Select == SELECT_WORLD)
    if ((World = SelectWorld(TRUE)) == NULL)
    {
      PRGPrintf ("\\"MSG_ABORT);
      return;
    }

  if (DLGOpenFileSel("load", AutoFile, PATH_MAX, AutoMask) == NULL)
  {
    PRGPrintf ("\\"MSG_ABORT);
    return;
  }

  PRGPrintf (MSG_LOADING, AutoFile);

  TOOLGetFileExt  (AutoFile, Extension, 31);

  if (strcasecmp(Extension, "part") == 0 || 
      (isdigit(Extension[0]) && atoi(Extension) >= 0))
  {
    if (World->LoadPart(AutoFile) == NULL)
      sprintf (ErrorMsg, "Error loading part\n'%s'", AutoFile);
  }
  elsif (strcasecmp(Extension, "dev") == 0)
  {
    if (! ImportDevice(World, AutoFile))
      sprintf (ErrorMsg, "Error loading device\n'%s'", AutoFile);
  }
  elsif (! World->LoadWorld(AutoFile, TOOLGetConfigString(CFG_DATA_PATH)) &&
	 ! ImportWorld(World, AutoFile))
    sprintf (ErrorMsg, "Error loading world\n'%s'", AutoFile);

  World->Fill ();

  if (ErrorMsg[0] != '\0')
    DLGOpenTextDialog ("loadError", DlgWARNING, ErrorMsg, 1, MSG_OK);
  else
    PRGPrintf ("\\"MSG_DONE);
}


void LoadPart (int Select)
{
  WORLD *World = GetCurrent ();

  PRGPrintf (MSG_LOAD_PART);

  if (World == NULL || Select == SELECT_WORLD)
    if ((World = SelectWorld(TRUE)) == NULL)
    {
      PRGPrintf ("\\"MSG_ABORT);
      return;
    }
   
  if (DLGOpenFileSel("loadPart", PartFile, PATH_MAX, PartMask) == NULL)
  {
    PRGPrintf ("\\"MSG_ABORT);
    return;
  }

  if (World->LoadPart(PartFile) == NULL)
    PRGPrintf ("\\"MSG_LOAD_ERROR);
  else
    PRGPrintf ("\\"MSG_DONE);
}


void LoadDev (int Select)
{
  WORLD *World = GetCurrent ();

  PRGPrintf (MSG_LOAD_DEV);

  if (World == NULL || Select == SELECT_WORLD)
    if ((World = SelectWorld(TRUE)) == NULL)
    {
      PRGPrintf ("\\"MSG_ABORT);
      return;
    }

  if (DLGOpenFileSel("loadDev", DevFile, PATH_MAX, DevMask) == NULL)
  {
    PRGPrintf ("\\"MSG_ABORT);
    return;
  }

  if (! ImportDevice(World, DevFile))
    PRGPrintf ("\\"MSG_LOAD_ERROR);
  else
    PRGPrintf ("\\"MSG_DONE);
}


void LoadWorld ()
{
  char   Name[256];
  WORLD *World;

  PRGPrintf (MSG_LOAD_WORLD);

  if ((World = SelectWorld(TRUE)) == NULL)
  {
    PRGPrintf ("\\"MSG_ABORT);
    return;
  }

// TODO:
// Abfrage, ob eine exist. Arbeitswelt gelscht werden soll

  if (DLGOpenFileSel("loadWorld", WorldFile, PATH_MAX, WorldMask) == NULL)
  {
    PRGPrintf ("\\"MSG_ABORT);
    return;
  }
  TOOLGetFileName (WorldFile, Name, 255);

  PRGPrintf ("\\loading '%s'...", Name);
  if (! World->LoadWorld(WorldFile, TOOLGetConfigString(CFG_DATA_PATH)))
    DLGOpenTextDialog ("loadError", DlgWARNING, MSG_LOAD_ERROR, 1, MSG_OK);
  else
    PRGPrintf ("\\"MSG_DONE);
}


void LoadConfig ()
{
  PRO_START ("\n")

  PRGPrintf (MSG_SELECT_CONFIG);

  if (! DLGOpenFileSel("loadCfg", CfgFile, PATH_MAX, CfgMask))
  {
    PRGPrintf (MSG_ABORT);
    PRO_ABORT ("No config file specified")
  }
    
  TOOLGetFileDir (CfgFile, CfgMask, PATH_MAX);
  strcat (CfgMask, "/*.cfg"); /* */

  if (TOOLLoadConfig(CfgFile, "TuxCAD-CONFIG"))
    PRGPrintf (MSG_LOAD_OK);
  else
    PRGPrintf (MSG_ERROR_LOADING, TOOLGetErrorMsg());

  PRO_END 
}


void Import ()
{
  char   Name[256];
  WORLD *World;

  PRGPrintf (MSG_IMPORT);

  if ((World = SelectWorld(TRUE)) == NULL)
  {
    PRGPrintf ("\\"MSG_ABORT);
    return;
  }

  if (DLGOpenFileSel("importData", ImportFile, PATH_MAX, ImportMask) == NULL)
  {
    PRGPrintf ("\\"MSG_ABORT);
    return;
  }
  TOOLGetFileName (ImportFile, Name, 255);

  PRGPrintf ("\\"MSG_LOADING, Name);
  if (! ImportWorld(World, ImportFile))
    DLGOpenTextDialog ("loadError", DlgWARNING, MSG_ERROR_IMPORT, 1, MSG_OK);
  else
    PRGPrintf ("\\"MSG_DONE);
}


void SavePart ()
{
  char    FileName[PATH_MAX+1] = CAD_PATH"PARTS/dummy.part";
  char    DirMask[PATH_MAX+1]  = CAD_PATH"PARTS/*.world"; /* */
  PART3D *Part;

  if (DLGOpenFileSel("saveFile", FileName, PATH_MAX, DirMask) == NULL)
  {
    printf ("function aborted\n");
    return;
  }
  Part = SelectPart ();

  if (OverwriteFile(FileName))
    PARTSave (FileName, Part->GetPart(), FALSE);
}


void SaveDev ()
{
}


void SaveWorld (int Select)
{
  WORLD *World = GetCurrent ();

  if (World == NULL || Select == SELECT_WORLD)
    if ((World = SelectWorld(TRUE)) == NULL)
    {
      PRGPrintf ("\\"MSG_ABORT);
      return;
    }

  if (DLGOpenFileSel("saveWorld", WorldFile, PATH_MAX, WorldMask) == NULL)
  {
    PRGPrintf ("\\"MSG_ABORT);
    return;
  }

#if 0
  if (OverwriteFile(WorldFile))
    World->SaveIgrip (WorldFile);
#endif
}


void SaveConfig ()
{
  PRO_START ("\n")

  PRGPrintf (MSG_SELECT_CONFIG);

  if (! DLGOpenFileSel("saveCfg", CfgFile, PATH_MAX, CfgMask))
  {
    PRGPrintf (MSG_ABORT);
    PRO_ABORT ("No file specified")
  }

  TOOLGetFileDir (CfgFile, CfgMask, PATH_MAX);
  strcat (CfgMask, "/*.cfg"); /* */

  if (TOOLSaveConfig(CfgFile, "TuxCAD-CONFIG"))
    PRGPrintf (MSG_SAVE_OK);
  else
    PRGPrintf (MSG_ERROR_SAVING, TOOLGetErrorMsg());

  PRO_END
}


void ClearWorld ()
{
  WORLD *World;

  if ((World = SelectWorld(FALSE)) != NULL)
    World->Clear ();
}


void DeleteWorld ()
{
  WORLD *World;

  if ((World = SelectWorld(FALSE, XC_pirate)) != NULL)
    DeleteWorld (World);

  WorldMouseCursor (NULL);
}


void PrgConfig ()
{
  char    Mask[PATH_MAX+1];
  COLOR   BG (COLOR::BLACK);
  DIALOG *Dlg;

  Dlg = DLGNewDialog (0);

  DLGInsertNote (Dlg, 0, "params", DlgNOBINDING);
    // Global:  
    DLGAddPage (Dlg, 0, DlgMAJOR, "global", 100);
      DLGMakeRows (Dlg, 100, 4, DlgNO_LINE, 110, 120, 130, 150);
      DLGMakeCols (Dlg, 150, 2, DlgNO_LINE, 151, 160);
      DLGMakeRows (Dlg, 151, 2, DlgNO_LINE, 152, 153);
      sprintf (Mask, "%s/*", TOOLGetConfigString(CFG_TUXCAD_PATH));
      DLGInsertDirSel  (Dlg, 110, "tuxCADPath", Mask, 20,
			     TOOLGetConfigString(CFG_TUXCAD_PATH));
      sprintf (Mask, "%s/*", TOOLGetConfigString(CFG_DATA_PATH));
      DLGInsertDirSel  (Dlg, 120, "dataPath", Mask, 20,
		             TOOLGetConfigString(CFG_DATA_PATH));
      DLGInsertOMenu (Dlg, 152, "floor", 2, "no", "yes",
	                     TOOLGetConfigBool(CFG_CREATE_FLOOR));
      ColorDialog (Dlg, 160, BG, BASE_MASK);
      DLGSetBorder    (Dlg, 160, DlgETCHED_IN);
      DLGSetBoxActive (Dlg, 160, FALSE);

      // TODO: 
      //       Box/Part, Default-Background color
    DLGAddPage (Dlg, 0, DlgMAJOR, "openGL", 200);
      DLGMakeRows (Dlg, 200, 3, DlgNO_LINE, 210, 220, 230);
      DLGInsertOMenu (Dlg, 210, "shading", 3, "none", "flat", "smooth",
		                TOOLGetConfigInt(CFG_GL_SHADING));

  DLGOpenDialog (Dlg, "prgConfig", DlgNORMAL, NULL);

  if (DLGGetExitButton (Dlg) == DlgOK)
  {
    TOOLSetConfigString (CFG_TUXCAD_PATH,        DLGGetDirSel(Dlg, 110));
    TOOLSetConfigString (CFG_DATA_PATH,          DLGGetDirSel(Dlg, 120));
    TOOLSetConfigBool   (CFG_CREATE_FLOOR, (dual)DLGGetOMenu(Dlg, 152));
    TOOLSetConfigInt    (CFG_GL_SHADING,         DLGGetOMenu(Dlg, 210));
  }

  DLGFreeDialog (Dlg);
}


void ExitPrg ()
{
#if 0
  int    PosX, PosY;
  int    SizeX, SizeY;
#endif
  WORLD *World;

  while (GetWorldNo() > 0)
  {
    World = GetData(0)->World;
    DelWorld (World);
  }

  if (Client.PortNo > 0 && Client.ID >= 0)
  {
    PRO0 ("Closing client...\n")
    NETCloseClient (Client.ID);
    Client.ID = -1;
  }

  if (Server.PortNo > 0 && Server.ID >= 0)
  {
    PRO0 ("Closing CLI-Server...\n")
    NETCloseServer (Server.ID);
    Server.ID = -1;
  }

  if (Robot.PortNo > 0 && Robot.ID >= 0)
  {
    PRO0 ("Closing V+ Server...\n")
    NETCloseServer (Server.ID);
    Server.ID = -1;
  }

#if 0
  if (PRGGetPos(&PosX, &PosY))
  {
    TOOLSetConfigInt (CFG_POS_X, PosX);
    TOOLSetConfigInt (CFG_POS_Y, PosY);
  }

  if (PRGGetSize(&SizeX, &SizeY))
  {
    TOOLSetConfigInt (CFG_SIZE_X, SizeX);
    TOOLSetConfigInt (CFG_SIZE_Y, SizeY);
  }
#endif

  if (Param.CfgName != NULL)
  {
    TOOLSaveConfig (Param.CfgName, "TuxCAD-CONFIG");
    free (Param.CfgName);
  }

  PRGExitProgram ();
}
