#include <math.h>

#include "tuxCAD.h"


#ifndef BOW
#define BOW(Angle,Radius) (Angle * PI * Radius / 180.0)
#endif

#ifndef SEGMENT
#define SEGMENT(Bow,Radius) (Bow * 180.0 / (PI * Radius))
#endif

#define CREATE_MINITRIX(Name,ID) \
        { \
          if ((No = RailTable[ID].GL_ID) == -1) \
          { \
	    if (RailTable[ID].Radius <= 0.0)           \
	      RailTable[ID].Part = CreateRail (MiniTrix, Name,        \
				               RailTable[ID].Len);    \
	    elsif (RailTable[ID].Len <= 0.0)           \
	      RailTable[ID].Part = CreateRail (MiniTrix, Name,        \
				               RailTable[ID].Angle,   \
				               RailTable[ID].Radius); \
            else \
	      RailTable[ID].Part = CreateRail (MiniTrix, Name,        \
				               RailTable[ID].Len,     \
				               RailTable[ID].Angle,   \
				               RailTable[ID].Radius); \
            if (RailTable[ID].Part != NULL)       \
              PARTSetOrigin (RailTable[ID].Part, Pos); \
          } \
	}

enum RAIL_TYPE
{
  RAIL_STRAIGHT,
  RAIL_CURVE,
  RAIL_SWITCH
};



/*
      Shape
         <-------Lane------->
       +---+              +---+
       ]   [<---Gauge---->]   [
   +---+   +--------------+   +---+ \
   |                              |  SleeperH
   +------------------------------+ / 
   <--------- SleeperL --------->


                      SleeperW
        <Gap>        /
     +--+   +--+   +--+   +--+
     |  |   |  |   |  |   |  |
     |  +---+  +---+  +---+  |
   =============================
     |  +---+  +---+  +---+  |
     |  |   |  |   |  |   |  |
     |  |   |  |   |  |   |  |
     |  |   |  |   |  |   |  |
     |  +---+  +---+  +---+  |
   =============================
     |  +---+  +---+  +---+  |\ \
     |  |   |  |   |  |   |  |  Offset
     +--+   +--+   +--+   +--+
*/



typedef struct RAIL_ELEM
{
         int        Type;
         PART3D    *Part;
  struct RAIL_ELEM *Prev;
  struct RAIL_ELEM *Next1;
  struct RAIL_ELEM *Next2;
} RAIL_ELEM;


typedef struct
{
  int     No;
  int     Type;
  double  Len;
  double  Radius;
  double  Angle;
  int     GL_ID;
  PART   *Part;
} RAIL_ENTRY;


static RAIL_ENTRY RailTable[] =
{
  {14902, RAIL_STRAIGHT, 312.6,   0.0,  0.0, -1, NULL},
  {14903, RAIL_STRAIGHT,  17.2,   0.0,  0.0, -1, NULL}, 
  {14904, RAIL_STRAIGHT, 104.2,   0.0,  0.0, -1, NULL},
  {14905, RAIL_STRAIGHT,  76.3,   0.0,  0.0, -1, NULL},
  {14906, RAIL_STRAIGHT,  54.2,   0.0,  0.0, -1, NULL},
  {14907, RAIL_STRAIGHT,  50.0,   0.0,  0.0, -1, NULL},
  {14908, RAIL_STRAIGHT,  27.9,   0.0,  0.0, -1, NULL},
  {14909, RAIL_STRAIGHT,  33.6,   0.0,  0.0, -1, NULL},
  {14912, RAIL_CURVE,      0.0, 194.6, 30.0, -1, NULL},
  {14914, RAIL_CURVE,      0.0, 194.6, 24.0, -1, NULL},
  {14916, RAIL_CURVE,      0.0, 194.6,  6.0, -1, NULL},
  {14917, RAIL_CURVE,      0.0, 329.0, 15.0, -1, NULL},
  {14918, RAIL_CURVE,      0.0, 492.6, 15.0, -1, NULL},
  {14922, RAIL_CURVE,      0.0, 228.2, 30.0, -1, NULL},
  {14924, RAIL_CURVE,      0.0, 228.2, 24.0, -1, NULL},    
  {14926, RAIL_CURVE,      0.0, 228.2,  6.0, -1, NULL},
  {14927, RAIL_CURVE,      0.0, 362.6, 15.0, -1, NULL},
  {14928, RAIL_CURVE,      0.0, 526.2, 15.0, -1, NULL},
  {14936, RAIL_SWITCH,     0.0,   0.0,  0.0, -1, NULL},
  {14938, RAIL_SWITCH,     0.0,   0.0,  0.0, -1, NULL},
  {14951, RAIL_SWITCH,   104.2, 194.6, 24.0, -1, NULL},
  {14953, RAIL_SWITCH,     0.0,   0.0,  0.0, -1, NULL},
  {14954, RAIL_SWITCH,     0.0,   0.0,  0.0, -1, NULL},
  {14955, RAIL_SWITCH,     0.0,   0.0,  0.0, -1, NULL},
  {14956, RAIL_SWITCH,     0.0,   0.0,  0.0, -1, NULL},
  {14957, RAIL_SWITCH,     0.0,   0.0,  0.0, -1, NULL},
  {14959, RAIL_SWITCH,     0.0,   0.0,  0.0, -1, NULL},
  {14978, RAIL_SWITCH,     0.0,   0.0,  0.0, -1, NULL},
  {14981, RAIL_SWITCH,     0.0,   0.0,  0.0, -1, NULL},
  {14983, RAIL_SWITCH,     0.0,   0.0,  0.0, -1, NULL}
};

static RAIL_DATA MiniTrix =
{
  {
    {0.7,0.0}, {0.7,0.42}, {0.175,0.56}, {0.175,1.4},
    {0.35,1.54}, {0.35,2.1}, {-0.35,2.1}, {-0.35,1.54},
    {-0.175,1.4}, {-0.175,0.56}, {-0.7,0.42}, {-0.7,0.0}
  },
  9.7 /* 9.0 + 0.7 */, 
  16.8, 1.4, 2.05, 
  2.90
};

static H_MATRIX Pos = Matrix_1;


PART *CreateRail (RAIL_DATA Data, char *Name, double length)
{
  int     count, Number;
  double  Step, Rest;
  double  SleepersLen;
  double  Offset;
  OBJECT *ObjList;
  PART   *Part;

  Step = Data.SleeperW + Data.Gap;

  /* The Rail has Number+1 bars and Number gaps */
  Number = int ((length - Data.SleeperW) / Step);
  SleepersLen = Number * Step + Data.SleeperW;
  Rest = length - SleepersLen;
  Offset = 0.5 * Rest;

  ObjList = OBJCreateQuader (GEOMakeTrans(-0.5*Data.SleeperL, Offset, 0.0),
			     Data.SleeperL, Data.SleeperW, Data.SleeperH,
			     COLOR::BLACK);

  for (count = 1; count < Number+1; ++count)
    OBJAddObject (ObjList,
		  OBJCreateQuader(GEOMakeTrans(-0.5 * Data.SleeperL, 
					       Offset+count*Step, 0.0),
				  Data.SleeperL, Data.SleeperW, Data.SleeperH,
				  COLOR::BLACK));

  /* left side */
  OBJAddObject (ObjList,
		OBJCreateQuader(GEOMakeTrans(-0.5*(Data.Lane+Data.SleeperW),
					     Offset, 0.0),
				Data.SleeperW, SleepersLen, Data.SleeperH,
				COLOR::BLACK));
  OBJAddObject (ObjList,
		OBJCreateSweep(GEORPY2Frame(-0.5*Data.Lane, length, 
					    Data.SleeperH, 90.0, 0.0, 0.0),
			       12, Data.Shape, length, COLOR::SILVER));

  /* right side */
  OBJAddObject (ObjList,
		OBJCreateQuader(GEOMakeTrans(0.5*(Data.Lane-Data.SleeperW),
					     Offset, 0.0),
				Data.SleeperW, SleepersLen, Data.SleeperH,
				COLOR::BLACK));
  OBJAddObject (ObjList,
		OBJCreateSweep(GEORPY2Frame(0.5*Data.Lane, length, 
					    Data.SleeperH, 90.0, 0.0, 0.0),
			       12, Data.Shape, length, COLOR::SILVER));
  Part = PARTCreate (Name, ObjList);

  return Part;
}


static PART *CreateRail (RAIL_DATA Data, char *Name, 
			 double Angle, double Radius)
{
  int       count, Number;
  int       Sign;
  double    Step, Bow, Rest;
  double    SleepersLen;
  double    Offset, Segment;
  H_MATRIX  Place, Trans;
  H_MATRIX  Rel = Matrix_1;
  POINT2D   Points[5];
  OBJECT   *ObjList;
  PART     *Part;

  Sign = SIGN (Angle);
  if (Angle < 0)
    Angle *= -1.0;

  Step = Data.SleeperW + Data.Gap;

  Bow = BOW (Angle, Radius);           // angle -> bow-lenght
  Number = int ((Bow-Data.SleeperW) / Step);  // Number of Sleepers+Gap
  SleepersLen = Number * Step + Data.SleeperW;    

  Rest = Bow - SleepersLen;                    // rest in bow-length

  /* reduce number of sleepers, if offset is less than 1.8 mm on each side */
  if (Rest < 3.6)
    Number--;

  Rest = SEGMENT (3.6, Radius);
  Offset = 0.5 * Rest;
  Segment = Sign * (Angle-Rest-SEGMENT(Data.SleeperW,Radius)) / (double)Number;

  ObjList = OBJCreateQuader (Matrix_1,
			     Data.SleeperL, Data.SleeperW, Data.SleeperH, 
			     COLOR::BLACK);

  Rel.m[P][X] = Radius - 0.5 * Data.SleeperL;

  Trans = GEOMakeTrans (-Rel.m[P][X], 0.0, 0.0);

  for (count = 1; count < Number+1; ++count)
  {
    Place = GEOMatrixProd (GEORotMatrix(Z, count*Segment), Rel);
    OBJAddObject (ObjList, 
		  OBJCreateQuader(GEOMatrixProd (Trans, Place),
				  Data.SleeperL, Data.SleeperW, Data.SleeperH,
				  COLOR::BLACK));
  }

  Angle *= Sign;

  Points[0].x = -0.5 * Data.SleeperW;
  Points[0].y =  0.0;
  Points[1].x =  0.5 * Data.SleeperW;
  Points[1].y =  0.0;
  Points[2].x =  0.5 * Data.SleeperW;
  Points[2].y =        Data.SleeperH;
  Points[3].x = -0.5 * Data.SleeperW;
  Points[3].y =        Data.SleeperH;
  Points[4].x = -0.5 * Data.SleeperW;
  Points[4].y =  0.0;

  /* right side */
  OBJAddObject (ObjList,
		OBJCreateRot(GEOMatrixProd(Trans, 
					   GEORotMatrix(Z, -90.0+Angle-Rest)),
			     5, Points, Angle-Rest, 16,
			     Radius + 0.5 * Data.Lane, COLOR::BLACK));
  OBJAddObject (ObjList,
		OBJCreateRot(GEOMatrixProd(Trans,
			        GEORPY2Frame(0.0, 0.0, Data.SleeperH,
					     0.0, 0.0, -90.0+Angle-Offset)),
			     12, Data.Shape, Angle, 16, Radius + 0.5*Data.Lane,
			     COLOR::SILVER));
		
  /* left side */
  OBJAddObject (ObjList,
		OBJCreateRot(GEOMatrixProd(Trans, 
					   GEORotMatrix(Z, -90.0+Angle-Rest)),
			     5, Points, Angle-Rest, 16,
			     Radius - 0.5 * Data.Lane, COLOR::BLACK));
  OBJAddObject (ObjList,
		OBJCreateRot(GEOMatrixProd(Trans,
			       GEORPY2Frame(0.0, 0.0, Data.SleeperH,
					    0.0, 0.0, -90.0 + Angle-Offset)),
			     12, Data.Shape, Angle, 16, Radius - 0.5*Data.Lane,
			     COLOR::SILVER));

  Part = PARTCreate (Name, ObjList);

  /* Set part's origin */
  Rel.m[P][X] = -0.5 * Data.SleeperL;
  Rel.m[P][Y] = BOW (Offset, Radius);
  PARTSetOrigin (Part, Rel);

  return Part;
}


static PART *CreateRail (RAIL_DATA Data, char *Name, 
			 double Length, double Angle, double Radius)
{
  int       count, Number;
  double    r0, r1, r2;
  double    Alpha1, Offs1;
  double    Alpha2, Offs2;
  double    Step, Bow, Rest;
  double    SleepersLen;
  double    Offset;
  H_MATRIX  Place, Trans;
  H_MATRIX  Rel = Matrix_1;
  POINT2D   Points[6];
  OBJECT   *ObjList;
  PART     *Part;

  Step = Data.SleeperW + Data.Gap;

  r0 = Radius - 0.5 * Data.SleeperL;
  r1 = Radius - 0.5 * Data.Lane;
  r2 = Radius + 0.5 * Data.Lane;

  /* right lane */  
  Alpha1 = acos ((r1 - 0.5 * Data.SleeperW) / r1);
  Offs1 = sin (Alpha1) * r1;

  Alpha2 = acos ((r2 - Data.Lane) / r2);
  Offs2 = sin (Alpha2) * r2;

  /* Sleeper */
  ObjList = OBJCreateQuader (Matrix_1, 
			     Data.SleeperL, Data.SleeperW, Data.SleeperH, 
			     COLOR::BLACK);

  Rel.m[P][X] = Radius - 0.5 * Data.SleeperL;

  Trans = GEOMakeTrans (-Rel.m[P][X], 0.0, 0.0);

  Points[0].x =  0.5 * Data.SleeperL;
  Points[0].y =  Offs1 - 2.0;
  Points[1].x =  0.5 * Data.SleeperL;
  Points[1].y =  Offs2;
  Points[2].x = -0.5 * Data.Lane;
  Points[2].y =  Offs2;
  Points[3].x = cos (Alpha2) * r0 - Radius;
  Points[3].y =  sin (Alpha2) * r0;
  Points[4].x = -0.5 * Data.SleeperL;
  Points[4].y =  Offs1 - 2.0;
  Points[5] = Points[0];

  OBJAddObject (ObjList,
		OBJCreateSweep(GEOMatrixProd(Trans,
					     GEOMakeTrans(Radius, 0.0, 0.0)),
			       6, Points, Data.SleeperH, COLOR::BLACK));
  
  Number = int ((Length-Data.SleeperW - Offs2) / Step);
  Rest = Length - (Number * Step + Data.SleeperW);
  if (Rest < 1.8)
    Number--;

  Bow = BOW (Angle, Radius);           // angle -> bow-lenght
  Number = int ((Bow-Data.SleeperW) / Step);  // Number of Sleepers+Gap
  SleepersLen = Number * Step + Data.SleeperW;    

  Rest = SEGMENT (3.6, Radius);
  Offset = 0.5 * Rest;
  //  Segment = Sign * (Angle-Rest-SEGMENT(Data.SleeperW,Radius)) / (double)Number;

  /* LEFT-SIDE */
  for (count = 1; count <= Number; ++count)
  {
    Place = GEOMakeTrans(Radius - 0.5 * Data.SleeperL,
			 Length-Offset-count*Step, 0.0);
    OBJAddObject (ObjList,
		  OBJCreateQuader(GEOMatrixProd(Trans, Place),
				  Data.SleeperL, Data.SleeperW, Data.SleeperH,
				  COLOR::BLACK));
  }

  Points[0].x = -0.5 * Data.SleeperW;
  Points[0].y =  0.0;
  Points[1].x =  0.5 * Data.SleeperW;
  Points[1].y =  0.0;
  Points[2].x =  0.5 * Data.SleeperW;
  Points[2].y =        Data.SleeperH;
  Points[3].x = -0.5 * Data.SleeperW;
  Points[3].y =        Data.SleeperH;
  Points[4] = Points[0];

  OBJAddObject (ObjList,
		OBJCreateRot(GEOMatrixProd(Trans,
					   GEORotMatrix(Z, -90.0+Angle-Rest)),
			     5, Points, Angle-Rest, 16, // PART A
			     Radius - 0.5 * Data.Lane, COLOR::BLACK));

  OBJAddObject (ObjList,
		OBJCreateRot(GEOMatrixProd(Trans,
			       GEORPY2Frame(0.0, 0.0, Data.SleeperH,
					    0.0, 0.0, -90.0+Angle)),
			     12, Data.Shape, Angle, 16, Radius - 0.5*Data.Lane,
			     COLOR::MEDBLUE));

  /* left side */
  OBJAddObject (ObjList,
		OBJCreateSweep(GEOMatrixProd(Trans,
		   GEORPY2Frame(Radius - 0.5 * (Data.Lane - Data.SleeperW),
				Offs2-3*Data.SleeperW, Data.SleeperH,
				90.0, 0.0, 0.0)),
			       12, Data.Shape, Offs2-Offs1-3*Data.SleeperW,
			       COLOR::LIME));
  OBJAddObject (ObjList,
		OBJCreateQuader(GEOMatrixProd(Trans,
			          GEOMakeTrans(Radius - 0.5 * Data.Lane,
					       Offset + Offs2, 0.0)),
				Data.SleeperW, SleepersLen-Offs2, 
				Data.SleeperH, COLOR::BLACK));
  OBJAddObject (ObjList,
		OBJCreateSweep(GEOMatrixProd(Trans,
		  GEORPY2Frame(Radius - 0.5 * (Data.Lane - Data.SleeperW),
			       Length + Offset - Data.SleeperW, Data.SleeperH,
			       90.0, 0.0, 0.0)),
			       12, Data.Shape, Length-Offs2-Data.SleeperW,
			       COLOR::GOLD));

  /* RIGHT-SIDE */
  /* The Rail has Number+1 bars and Number gaps */
  Number = int ((Length - Data.SleeperW) / Step);
  SleepersLen = Number * Step + Data.SleeperW;
  Rest = Length - SleepersLen;
  Offset = 0.5 * Rest;

  OBJAddObject (ObjList,
		OBJCreateQuader(GEOMatrixProd(Trans, 
	          GEOMakeTrans(Radius + 0.5 * (Data.Lane - Data.SleeperW),
			       Offset, 0.0)),
				Data.SleeperW, SleepersLen, Data.SleeperH, 
				COLOR::BLACK));
  OBJAddObject (ObjList,
		OBJCreateSweep(GEOMatrixProd(Trans,
		   GEORPY2Frame(Radius + 0.5 * Data.Lane,
			        Length, Data.SleeperH, 90.0, 0.0, 0.0)),
			       12, Data.Shape, Length, COLOR::SILVER));

  // PART B
  OBJAddObject (ObjList,
		OBJCreateRot(GEOMatrixProd(Trans,
		  GEORPY2Frame(0.0, 0.0, Data.SleeperH, 
			       0.0, 0.0, -90.0 + Angle - GRAD(Alpha1)-2)),
			     12, Data.Shape, GRAD(Alpha2-Alpha1)-4, 8, 
			     Radius + 0.5*Data.Lane, COLOR::MEDRED));
  OBJAddObject (ObjList,
		OBJCreateRot(GEOMatrixProd(Trans,
		  GEORPY2Frame(0.0, 0.0, Data.SleeperH,
			       0.0, 0.0, -90.0 + Angle)),
			     12, Data.Shape, Angle-GRAD(Alpha2), 8, 
			     Radius + 0.5*Data.Lane, COLOR::SILVER));

  Part = PARTCreate (Name, ObjList);

  return Part;
}


H_MATRIX CalcRel (int ID)
{
  H_MATRIX Trans = Matrix_1;
  H_MATRIX Result = Matrix_1;

  if (RailTable[ID].Radius <= 0.0)
    Result.m[P][Y] += RailTable[ID].Len;
  elsif (RailTable[ID].Len <= 0.0)
  {
    Trans.m[P][X] = RailTable[ID].Radius;
    Result = GEOMatrixProd (
               GEOMatrixProd(GEOMatrixInv(Trans),
			     GEOMakeRot(GEOMakeVector(0.0, 0.0, 1.0), 
					RailTable[ID].Angle)),
	       Trans);
  }
  else
    Result.m[P][Y] += RailTable[ID].Len;

  return Result;
}


void MakeRail (int ID)
{
  int       No;
  char      Name[30];
  PART3D   *Part;
  DEVICE3D *Dev;
  WORLD    *World;

  if ((World = SelectWorld(TRUE)) == NULL)
    return;

  ID--;

  sprintf (Name, "%d", RailTable[ID].No);
  CREATE_MINITRIX (Name, ID);

  if (RailTable[ID].Part != NULL)
  {
    Part = new PART3D (RailTable[ID].Part);
    Dev = new DEVICE3D ("rail", Part);

    World->AddDevice (Dev);
    if (World->GetDevNum() == 1)
    {
      World->Fill ();
      // SetHandles (World);
    }

    Pos = GEOMatrixProd (Pos, CalcRel(ID));
  }
}




