#include <assert.h>

#include <Xm/ScrollBar.h>

#include "tuxCAD.h"

#include <icons/world_s.xpm>
#include <icons/dev.xpm>
#include <icons/gears_blue.xpm>
#include <icons/gears_yellow.xpm>
#include <icons/ball_red.xpm>
#include <icons/ball_green.xpm>
#include <icons/plus.xpm>
#include <icons/minus.xpm>


#define TREEID(x) (x/10*10)


/* TODO: 
   bessere Icons fr Parts (Mutter/Schraube ?), Chains
*/


void TREE_DialogCB (DIALOG *Dlg, int Button)
{
  TREE_DIALOG *TreeDlg = (TREE_DIALOG*)DLGGetUserData (Dlg, 0);

  if (TreeDlg != NULL)
    TreeDlg->ButtonCB (Button);
}


int TREE_SelectCB (DIALOG *Dlg, int BoxID)
{
  TREE_DIALOG *TreeDlg = (TREE_DIALOG *)DLGGetUserData (Dlg, 0);

  if (TreeDlg != NULL)
    TreeDlg->ChangeCB (BoxID);

  return DlgACCEPT;
}


TREE_DIALOG::TREE_DIALOG (WORLD *World) :
  CAD_DIALOG (300, 400),
  World      (World)
{
  PRO_METHOD ("TREE_DIALOG", "\n")

  assert (this->World != NULL);

  // Create pixmaps for tree-buttons
  ClosePic = TOOLMakeXPM (plus_xpm,  NULL, NULL, TOOLGetColor("grey90"));
  OpenPic  = TOOLMakeXPM (minus_xpm, NULL, NULL, TOOLGetColor("grey90"));

  PRO_END
}


TREE_DIALOG::~TREE_DIALOG ()
{
  PRO_METHOD ("~TREE_DIALOG", "")

  XFreePixmap (XDisplay, ClosePic);
  XFreePixmap (XDisplay, OpenPic);

  PRO_END
}


bool TREE_DIALOG::Open ()
{
  PRO_METHOD ("TREE_DIALOG", "\n")

  if (! CAD_DIALOG::Open())                       // Call inherited method
    PRO_RETURN ("Dialog already open", FALSE)

  DLGInsertTree  (Dlg, 0, "WorldName", ClosePic, OpenPic, 0);

  SetTreeWorld (0);

  DLGSetFunction (Dlg, 0, TREE_SelectCB);
  DLGSetSize (Dlg, SizeX, SizeY, TRUE);
  DLGSetPos  (Dlg, PosX,  PosY);

  DLGMakeButton (Dlg, USER_BUTTON, "collapse", 
		      DlgBUTTON_LEFT | DlgBUTTON_ADJUST, TREE_DialogCB, -20);
  DLGMakeButton (Dlg, USER_BUTTON, "expand",   
		      DlgBUTTON_LEFT | DlgBUTTON_ADJUST, TREE_DialogCB, -10);
  DLGMakeButton (Dlg, HELP_BUTTON);
  DLGMakeButton (Dlg, OK_BUTTON);
  // TODO: Buttons fr search/refresh/filter

  DLGSetUserData     (Dlg, 0, this);
  DLGSetHelpFunction (Dlg, HelpDlgCB);

  DLGOpenDialog (Dlg, "worldTree", DlgNORMAL, NULL);

  Close ();

  PRO_END
  return TRUE;
}


void TREE_DIALOG::Update ()
{
  PRO_METHOD ("TREE_DIALOG", "\n")

  if (Dlg == NULL)
    PRO_ABORT ("Dialog not open")

  DLGFreeTree (Dlg, 0);

  SetTreeWorld (0);
  
  DLGBuildTree (Dlg, 0);

  // TODO: 
  // Problem: aktuellen Einstellung beibehalten

  PRO_END
}


void TREE_DIALOG::Expand ()
{
  if (Dlg != NULL)
    DLGSetTreeState (Dlg, 0, TRUE);
}


void TREE_DIALOG::Collapse ()
{
  if (Dlg != NULL)
    DLGSetTreeState (Dlg, 0, FALSE);
}


void TREE_DIALOG::ButtonCB (int Button)
{
  switch (Button)
  {
    case -20 : Collapse ();
               break;
    case -10 : Expand ();
               break;
    default : break;
  }
}


void TREE_DIALOG::ChangeCB (int BoxID)
{
  int       Reason;
  int       MouseX, MouseY;
  int       PosX = 0, PosY = 0;
  Dimension x, y;
  Widget    Shell;
  TREEITEM *Item;
  PART3D   *Part;
  DEVICE3D *Dev;
  WORLD    *World = GetCurrent ();

  if ((Item = DLGGetTree(Dlg, BoxID)) != NULL)
  {
    Reason = DLGGetReason (Dlg, BoxID);
    Shell = DLGGetShell (Dlg);

    if (Reason != DlgCURSOR)
    {
      // Determinate where to place the popup menu. This might also be done
      // using the position of Item->gadget, but needs to corrected by 
      // scrollbar values. I hope this technique will do...
      TOOLGetMousePos (&MouseX, &MouseY);
      XtVaGetValues (Shell, XmNx, &x,
		            XmNy, &y,
		            NULL);
      PosX = MouseX - x + 5;
      PosY = MouseY - y + 5;
    }

    switch (Item->ID % 10)
    {
      case 0 :
	World = (WORLD *)Item->UserData;
	PRGPrintf ("WORLD: %s", World->GetName());
	ClearSelection (World);
	if (Reason != DlgCURSOR)
	  OpenWorldMenu (Shell, World, PosX, PosY);
	break;

      case 1 :
	Dev = (DEVICE3D *)Item->UserData;
	PRGPrintf ("DEVICE: %s", Dev->GetName());
        ClearSelection (World);
	SwapSelection  (World, Dev);
       	if (Reason != DlgCURSOR)
	{
	  LIST<DEVICE3D *> *Devs = GetDevBuffer (World);

	  OpenDevMenu (Shell, *Devs, PosX, PosY);
	}
	break;

      case 2 :
	Part = (PART3D *)Item->UserData;
	PRGPrintf ("PART: %s", Part->GetName());
	ClearSelection (World);
	SwapSelection  (World, Part);
	if (Reason != DlgCURSOR)
	{
	  LIST<PART3D *> *Parts = GetPartBuffer (World);

	  OpenPartMenu (Shell, *Parts, PosX, PosY);
	}
	break;

      default : 
	PRGPrintf ("Invalid Selection");
    }
  }
}


void TREE_DIALOG::HelpCB (int ID)
{
  HelpDlg.Open ("TreeView", "dlg_tree.hlp", "");
}


int TREE_DIALOG::SetTreePart (int BoxID, PART3D *PartList, int ID, int Parent)
{
  int       Parent2;
  PART     *Part;
  PART3D   *Ptr, *Next;
  DEVICE3D *Attach;

  Next = PartList;
  while (Next != NULL)
  {
    if ((Part = Next->GetPart()) != NULL)
      if (Next->GetPrev() != NULL || Next->GetChain() != NULL)  // chain 
	DLGAddTreeIcon (Dlg, BoxID, Part->PartName, ball_red, ID+2, Parent,
			            FALSE, (void*)Next);
      else                                                      // single part
	DLGAddTreeIcon (Dlg, BoxID, Part->PartName, ball_green, ID+2, Parent,
			            FALSE, (void*)Next);

    Parent2 = ID+2;

    if ((Ptr = Next->GetChain()) != NULL)
      ID = SetTreePart (BoxID, Ptr, ID+10, Parent2);
    else
      ID += 10;

    if ((Attach = Next->GetAttach()) != NULL)
      ID = SetTreeDev (BoxID, Attach, ID, Parent2) + 10;

    Next = Next->GetNext ();
  }
  return ID;
}


int TREE_DIALOG::SetTreeDev (int BoxID, DEVICE3D *DevList, int ID, int Parent)
{
  DEVICE3D *Dev;

  Dev = DevList;
  while (Dev != NULL)
  {
    if (Dev->GetFix() != NULL)                    // attached device
      DLGAddTreeIcon (Dlg, BoxID,                
		      (Dev->GetName()!=NULL)?Dev->GetName():"<no name>", 
		      gears_yellow_xpm, ID+1, Parent,
		      FALSE, (void*)Dev);
    elsif (Dev->GetPartNo() == 1)
      DLGAddTreeIcon (Dlg, BoxID, 
		      (Dev->GetName()!=NULL)?Dev->GetName():"<no name>", 
		      dev_xpm, ID+1, Parent,
		      FALSE, (void*)Dev);
    else // autonomeous device
      DLGAddTreeIcon (Dlg, BoxID, 
		      (Dev->GetName()!=NULL)?Dev->GetName():"<no name>", 
		      gears_blue_xpm, ID+1, Parent,
		      FALSE, (void*)Dev);

    ID = SetTreePart (BoxID, Dev->GetPartList(), ID+10, ID+1);
    Dev = Dev->GetNext ();
  }
  return TREEID(ID);
}


void TREE_DIALOG::SetTreeWorld (int BoxID)
{
  DLGAddTreeIcon (Dlg, BoxID, World->GetName(), world_s, 10, 0, TRUE, 
		              (void*)World);

  SetTreeDev (BoxID, World->GetData(), 20, 10);
}


void ShowTree (int Select)
{
  WORLD *World = GetCurrent ();

  PRO_START ("")

  if (World == NULL || Select == SELECT_WORLD)
    if ((World = SelectWorld(FALSE)) == NULL)
    {
      PRGPrintf ("Function aborted\n");
      PRO_ABORT ("Function aborted")
    }

  OpenTreeDialog (World);

  PRO_END
}
