#include <unistd.h>
#include <ctype.h>
#include <string.h>

#include "tuxCAD.h"

#include <icons/bomb.xpm>
#include <icons/load.xpm>
#include <icons/create.xpm>
#include <icons/build.xpm>
#include <icons/tree.xpm>
#include <icons/help2.xpm>


PROGRAM_PARAM Param = 
{
  NULL,                                 // Config filename
  NULL,                                 // Protocol filename
  0,                                    // number of files
  NULL,                                 // Ptr to data filenames
  FALSE                                 // verbose mode
};

static CONFIG_ENTRY Config[] =
{
  {CFG_TUXCAD_PATH,   "TuxCAD-Path",      CFG_STRING,   0, "~/tuxcad"},
  {CFG_DATA_PATH,     "Data-Path",        CFG_STRING,   0, "~/tuxcad/data"},
  {CFG_LIBS_PATH,     "Libs-Path",        CFG_STRING,   0,    
                       "~/tuxcad/data/LIBRARIES"},
  {CFG_POS_X,         "X-Position",       CFG_INT,      0, ""},
  {CFG_POS_Y,         "Y-Position",       CFG_INT,      0, ""},
  {CFG_SIZE_X,        "Width",            CFG_INT,      0, ""},
  {CFG_SIZE_Y,        "Height",           CFG_INT,    150, ""},
  {CFG_CREATE_FLOOR,  "Create Floor",     CFG_BOOL,  TRUE, ""},
  {CFG_SHOW_COORDS,   "Show Coordinates", CFG_BOOL, FALSE, ""}, 
  {CFG_GL_SHADING,    "Shading",          CFG_INT,      0, ""}
  // {CFG_DELTA_DIST,    "Distance-Delta",  CFG_DOUBLE,  5E-2,            ""},
};

static char *CfgNames[] =    /* File specs for possible configuration files */
{ 
  "%s.cfg",
  "~/%s.cfg",
  "~/tuxcad.cfg", 
  "~/.tuxcad"
};


void InitPrg ()
{
  int    count;
  char   FileName[PATH_MAX+1];
  char   ErrorMsg[40+PATH_MAX+1] = "";
  char   Extension[32];
  WORLD *World = NULL;

  OpenGL->Initialize ();
  if (Param.verbose)
    OpenGL->Write ();

  if (Client.PortNo > 0 || Server.PortNo > 0)
    PRGPrintf ("Initalize connection...");

  if (Client.PortNo > 0)
    StartClient ();

  if (Server.PortNo > 0)
  {
    PRO2 ("SERVER: '%s' (%d)\n", Server.HostName, Server.PortNo)
    if (! RunCLIServer(Server.HostName, Server.PortNo))
      PRGPrintf ("Connection failed");
  }
  elsif (Robot.PortNo > 0)
    if (! RunVPLUSServer(Robot.HostName, Robot.PortNo))
      PRGPrintf ("Connection failed");

  for (count = 0; count < Param.FileNo; ++count)
  {
    TOOLGetFileName (Param.InName[count], FileName, PATH_MAX);

    PRGPrintf ("Loading '%s'...", FileName);

    TOOLGetFileExt  (FileName, Extension, 31);

    if (strcasecmp(Extension, "part") == 0 ||
 	(isdigit(Extension[0]) && atoi(Extension) >= 0))
    {
      if (World == NULL)
	World = NewWorld (FileName);
      if (World->LoadPart(Param.InName[count]) == NULL)
        sprintf (ErrorMsg, "Error loading part\n'%s'", Param.InName[count]);
    }
    elsif (strcasecmp(Extension, "dev") == 0)
    {
      if (World == NULL)
	World = NewWorld (FileName);
      if (! ImportDevice(World, Param.InName[count]))
        sprintf (ErrorMsg, "Error loading device\n'%s'", Param.InName[count]);
    }
    else
    {
      World = NewWorld (FileName);
      if (! World->LoadWorld(Param.InName[count]) &&
	  ImportWorld(World, Param.InName[count]) == NULL)
        sprintf (ErrorMsg, "Error loading world\n'%s'", Param.InName[count]);
    }

    if (World != NULL)
      World->Fill ();

    if (ErrorMsg[0] != '\0')
      DLGOpenTextDialog ("loadError", DlgWARNING, ErrorMsg, 1, "ok");
  }
}


void Dummy ()
{
  PRGPrintf ("This is a dummy function");
}


int main (int argc, char *argv[])
{
  int  count;
  char CfgName[PATH_MAX+1];
  char CfgSpec[PATH_MAX+1];

  static MENU_ENTRY LoadMenu[] =
  {
    ENTRY     ("part",   "", "", ' ', TRUE, LoadPart,     SELECT_WORLD),
    ENTRY     ("dev",    "", "", ' ', TRUE, LoadDev,      SELECT_WORLD),
    ENTRY     ("world",  "", "", ' ', TRUE, LoadWorld,    NULL),
    MENU_LINE (PrgETCHED_IN),
    ENTRY     ("config", "", "", ' ', TRUE, LoadConfig,   NULL)
  };

  static MENU_ENTRY SaveMenu[] =
  {
    ENTRY     ("part",  "", "", ' ', TRUE,  SavePart,     NULL),
    ENTRY     ("dev",   "", "", ' ', FALSE, SaveDev,      NULL),
    ENTRY     ("world", "", "", ' ', TRUE,  SaveWorld,    SELECT_WORLD),
    MENU_LINE (PrgETCHED_IN),
    ENTRY     ("config", "", "", ' ', TRUE, SaveConfig,   NULL)
  };

#ifdef VICOS
  static MENU_ENTRY ImportMenu[] = 
  {
    ENTRY     ("vicos", "", "", ' ', TRUE,  Import,       NULL)
  };
#endif

  static MENU_ENTRY SystemMenu[] =
  {
    ENTRY     ("new",        "",         "",     ' ', TRUE, NewWorld,    NULL),
    MENU_LINE (PrgETCHED_IN),
    SUB_MENU  ("load",       "<Key>F3",  "F3",   ' ', TRUE, LoadMenu),
    SUB_MENU  ("save",       "<Key>F2",  "F2",   ' ', TRUE, SaveMenu),
#ifdef VICOS
    SUB_MENU  ("import",     "",         "",     ' ', TRUE, ImportMenu),
#endif
    MENU_LINE (PrgETCHED_IN),
    ENTRY     ("clear",      "",         "",     ' ', TRUE, ClearWorld,  NULL),
    ENTRY     ("delete",     "",         "",     ' ', TRUE, DeleteWorld, NULL),
    MENU_LINE (PrgETCHED_IN),
    ENTRY     ("config",     "",         "",     ' ', TRUE, PrgConfig,   NULL),
    MENU_LINE (PrgETCHED_IN),
    ENTRY     ("exit",     "Alt<Key>F4", "Alt+F4", 'x', TRUE, ExitPrg,   NULL)
  };

  static MENU_ENTRY PropertyMenu[] =
  {
    ENTRY     ("world",     "", "", ' ', TRUE, WorldProperties, SELECT_WORLD)
  };

  static MENU_ENTRY EditMenu[] =
  {
    ENTRY     ("create",    "", "", ' ',  TRUE,  CreatePart,   NULL),
    ENTRY     ("merge",     "", "", ' ',  TRUE,  MergeParts,   NULL),
    ENTRY     ("smash",     "", "", ' ',  TRUE,  SmashPart,    NULL),
    ENTRY     ("mount",     "", "", ' ',  FALSE, Dummy,        NULL)
  };

  static MENU_ENTRY CADMenu[] =
  {
    SUB_MENU  ("edit",         "", "", ' ', TRUE, EditMenu),
    MENU_LINE (PrgETCHED_IN),
    ENTRY     ("viewTree",     "", "", ' ', TRUE, ShowTree,   SELECT_WORLD),
    ENTRY     ("fly",          "", "", ' ', TRUE, Fly,        NULL),
    MENU_LINE (PrgETCHED_IN),
    SUB_MENU  ("properties",   "", "", ' ', TRUE, PropertyMenu),
  };

  static MENU_ENTRY ControlMenu[] = 
  {
    ENTRY     ("run",          "", "", ' ', FALSE, NULL, NULL),
    ENTRY     ("stop",         "", "", ' ', FALSE, NULL, NULL),
    ENTRY     ("step",         "", "", ' ', FALSE, NULL, NULL),
    ENTRY     ("reset",        "", "", ' ', FALSE, NULL, NULL),
    ENTRY     ("set",          "", "", ' ', FALSE, NULL, NULL)
  };

  static MENU_ENTRY SimuMenu[] = 
  {
    ENTRY     ("control",      "", "", ' ', TRUE, MotionControl, NULL),
    ENTRY     ("motionTest1",  "", "", ' ', TRUE, MotionTest1,   NULL),
    ENTRY     ("motionTest2",  "", "", ' ', TRUE, MotionTest2,   NULL),
    SUB_MENU  ("action",       "", "", ' ', TRUE, ControlMenu)
  };

  static MENU_ENTRY RobotMenu[] =
  {
    ENTRY     ("teachPendant", "", "", ' ', TRUE, TeachPanel, NULL)
  };

  static MENU_ENTRY PlotMenu[] = 
  {
    ENTRY     ("plotDemo",     "", "", ' ', TRUE, Plot3D,     NULL)
  };

  static MENU_ENTRY StraightMenu[] =
  {
    ENTRY     ("14902",      "", "", ' ',  TRUE, MakeRail,  MT14902),
    ENTRY     ("14903",      "", "", ' ',  TRUE, MakeRail,  MT14903),
    ENTRY     ("14904",      "", "", ' ',  TRUE, MakeRail,  MT14904),
    ENTRY     ("14905",      "", "", ' ',  TRUE, MakeRail,  MT14905),
    ENTRY     ("14906",      "", "", ' ',  TRUE, MakeRail,  MT14906),
    ENTRY     ("14907",      "", "", ' ',  TRUE, MakeRail,  MT14907),
    ENTRY     ("14908",      "", "", ' ',  TRUE, MakeRail,  MT14908),
    ENTRY     ("14909",      "", "", ' ',  TRUE, MakeRail,  MT14909)
  };

  static MENU_ENTRY CurveMenu[] = 
  {
    ENTRY     ("14912",      "", "", ' ',  TRUE, MakeRail,  MT14912),
    ENTRY     ("14914",      "", "", ' ',  TRUE, MakeRail,  MT14914),
    ENTRY     ("14916",      "", "", ' ',  TRUE, MakeRail,  MT14916),
    ENTRY     ("14922",      "", "", ' ',  TRUE, MakeRail,  MT14922),
    ENTRY     ("14924",      "", "", ' ',  TRUE, MakeRail,  MT14924),
    ENTRY     ("14926",      "", "", ' ',  TRUE, MakeRail,  MT14926),
    ENTRY     ("14917",      "", "", ' ',  TRUE, MakeRail,  MT14917),
    ENTRY     ("14927",      "", "", ' ',  TRUE, MakeRail,  MT14927),
    ENTRY     ("14918",      "", "", ' ',  TRUE, MakeRail,  MT14918),
    ENTRY     ("14928",      "", "", ' ',  TRUE, MakeRail,  MT14928)
  };     

  static MENU_ENTRY SwitchMenu[] = 
  {
    ENTRY     ("14936",      "", "", ' ',  FALSE, MakeRail,  MT14936),
    ENTRY     ("14938",      "", "", ' ',  FALSE, MakeRail,  MT14938),
    ENTRY     ("14951",      "", "", ' ',  TRUE,  MakeRail,  MT14951),
    ENTRY     ("14953",      "", "", ' ',  FALSE, MakeRail,  MT14953),
    ENTRY     ("14954",      "", "", ' ',  FALSE, MakeRail,  MT14954),
    ENTRY     ("14955",      "", "", ' ',  FALSE, MakeRail,  MT14955),
    ENTRY     ("14956",      "", "", ' ',  FALSE, MakeRail,  MT14956),
    ENTRY     ("14957",      "", "", ' ',  FALSE, MakeRail,  MT14957),
    ENTRY     ("14981",      "", "", ' ',  FALSE, MakeRail,  MT14981),
    ENTRY     ("14983",      "", "", ' ',  FALSE, MakeRail,  MT14983),
    MENU_LINE (PrgSINGLE_LINE),
    ENTRY     ("14959",      "", "", ' ',  FALSE, MakeRail,  MT14959),
    ENTRY     ("14978",      "", "", ' ',  FALSE, MakeRail,  MT14978)
  };

  static MENU_ENTRY MiscMenu[] =  
  {
  };

  static MENU_ENTRY TrainMenu[] =    /* Railway */
  {
    SUB_MENU  ("straigt", "", "", ' ', TRUE,  StraightMenu),
    SUB_MENU  ("curve",   "", "", ' ', TRUE,  CurveMenu),
    SUB_MENU  ("switch",  "", "", ' ', TRUE,  SwitchMenu),
    SUB_MENU  ("switch",  "", "", ' ', FALSE, MiscMenu),
  };

  static MENU_ENTRY NetMenu[] =
  {
    ENTRY     ("startClient",  "", "", ' ',  TRUE, StartClient,      NULL),
    ENTRY     ("sendMsg",      "", "", ' ',  TRUE, SendData,         NULL),
    ENTRY     ("closeClient",  "", "", ' ',  TRUE, CloseClient,      NULL),
    MENU_LINE (PrgETCHED_IN),
    ENTRY     ("cliServer",    "", "", ' ',  TRUE, StartCLIServer,   NULL),
    ENTRY     ("exitCLI",      "", "", ' ',  TRUE, CloseCLIServer,   NULL),
    MENU_LINE (PrgETCHED_IN),
    ENTRY     ("vPlusServer",  "", "", ' ',  TRUE, StartVPLUSServer, NULL),
    ENTRY     ("exitVPlus",    "", "", ' ',  TRUE, CloseVPLUSServer, NULL),
  };

  static MENU_ENTRY FontMenu[] =
  {
    ENTRY     ("text",         "", "", ' ',  TRUE, MakeText,         NULL),
    ENTRY     ("type",         "", "", ' ',  TRUE, MakeText2,        NULL)
  };

#ifdef _DYNAMIC_
  static MENU_ENTRY TestMenu[] =
  {
    ENTRY     ("materialPreview", "", "", ' ', TRUE, MaterialEdit,  NULL),
    ENTRY     ("swapToolbar",     "", "", ' ', TRUE, SwapToolbar,   NULL),
    ENTRY     ("ListFormats",     "", "", ' ', TRUE, ListFormats,   NULL),
    ENTRY     ("Preview",         "", "", ' ', TRUE, Preview,       NULL),
    ENTRY     ("LoadTest",        "", "", ' ', TRUE, ReadDataFile,  NULL)
  };
#endif

  static MENU_ENTRY HelpMenu[] = 
  {
    ENTRY     ("manual",      "<Key>F1", "F1", ' ', TRUE, Manual,      NULL),
    MENU_LINE (PrgETCHED_IN),
    ENTRY     ("about",       "",        "",   ' ', TRUE, About,       NULL)
  };

  static MENU_COLUMN MenuCols[] =
  {
    FUNC_MENU ("system",     's', TRUE, SystemMenu),
    FUNC_MENU ("cad",        'c', TRUE, CADMenu),
    FUNC_MENU ("simulation", 'i', TRUE, SimuMenu),
    FUNC_MENU ("robot",      'r', TRUE, RobotMenu),
    FUNC_MENU ("plot",       'p', TRUE, PlotMenu), 
    FUNC_MENU ("train",      't', TRUE, TrainMenu),
    FUNC_MENU ("net",        'n', TRUE, NetMenu),
    FUNC_MENU ("font",       'f', TRUE, FontMenu),
#ifdef _DYNAMIC_
    FUNC_MENU ("test",       ' ', TRUE, TestMenu),
#endif
    HELP_MENU ("help",       'h', TRUE, HelpMenu)
  };

  static MENU_BAR Menu = MAIN_MENU ("mainMenu", MenuCols);

  if (argc == 2 && strcmp(argv[1], "--help") == 0)
  {
    ShowSyntax ();
    exit (1);
  }

  count = 1;
  while (count < argc)
  {
    if (strcmp(argv[count], "-s") == 0)
      Server.PortNo = atoi (argv[++count]);
    elsif (strcmp(argv[count], "-v") == 0)
      Param.verbose = TRUE;
    elsif (strcmp(argv[count], "-cfg") == 0)
    {
      count++;
      if (count == argc)
	ShowSyntax ();
      else
	Param.CfgName = argv[count];
    }
    elsif (strcmp(argv[count], "-robot") == 0)
      Robot.PortNo = atoi (argv[++count]);
    elsif (strcmp(argv[count], "-c") == 0)
      Client.PortNo = atoi (argv[++count]);
    elsif (strcmp(argv[count], "-h") == 0)
    {
      Server.HostName = argv[++count];
      Client.HostName = argv[++count];
      Robot.HostName  = argv[++count];
    }
    elsif (strcmp(argv[count], "-pro") == 0)
    {
      count++;
      if (count == argc)
	ShowSyntax ();
      else
	Param.ProName = argv[count];
    }
    elsif (argv[count][0] != '-')
    {
      if (Param.InName == NULL)
	Param.InName = new char *[argc-1];
      Param.InName[Param.FileNo++] = argv[count];
    }
    else
    {
      ShowSyntax ();
      exit (1);
    }        

    count++;
  }

  PRGInitLib (argv[0], &argc, argv, NULL);

  PRGMakeToolBar (PrgTOP_TOOL);

  PRGAddToolIcon  (-1, -1, bomb_xpm,       ExitPrg,     NULL, "Exit program");
  PRGAddToolSpace (5);
  PRGAddToolIcon  (-1, -1, load_xpm,   (MENU_FUNCTION)Load, 
		                       (void*)SELECT_WORLD, "Load file...");
  // PRGAddToolIcon (-1, -1, save_xpm,    Save,        NULL, "Save data");
  // PRGAddToolIcon (-1, -1, printer_xpm, Print,       NULL, "Print data");
  PRGAddToolIcon  (-1, -1, create_xpm,     CreatePart,  NULL, 
                           "Create a new part");
  PRGAddToolIcon  (-1, -1, build_xpm,      EditPart,    NULL, "Editor");
  PRGAddToolSpace (10);
  PRGAddToolIcon  (-1, -1, tree_xpm,   (MENU_FUNCTION)ShowTree,   
		                       (void*)SELECT_WORLD, 
		           "Open\nTree-Window");
  PRGAddToolSpace (10);
  PRGAddToolIcon  (-1, -1, help2_xpm,      Manual,      NULL, 
		           "Online\nManual");

  PRGCreateProgram (argv[0], &Menu);

  TOOLInsertEntries (Config, sizeof(Config)/sizeof(CONFIG_ENTRY));

  // Load configuration file
  if (Param.CfgName != NULL && access(Param.CfgName, R_OK) == 0)
    TOOLLoadConfig (Param.CfgName, "TuxCAD-CONFIG");
  else
  {
    // Check for one of the default config files
    count = 0;
    while (count < (int)(sizeof(CfgNames)/sizeof(char **)))
    {
      sprintf (CfgSpec, CfgNames[count], argv[0]);
      TOOLParseFileName (CfgSpec, CfgName, PATH_MAX);
      if (access(CfgName, R_OK) == 0)
      {
	Param.CfgName = (char *)malloc (sizeof(char) * (sizeof(CfgName)+1));
	strcpy (Param.CfgName, CfgName);
	TOOLLoadConfig (CfgName, "TuxCAD-CONFIG");
	break;
      }
      count++;
    }
  }

  // Try to open protocol file if specified
  if (Param.ProName != NULL)
    DEBOpenProtocol (Param.ProName, DEB_OVERWRITE);

  PRGSetPos  (TOOLGetConfigInt(CFG_POS_X), TOOLGetConfigInt(CFG_POS_Y));
  PRGSetSize (0, 150);
  PRGStartProgram (InitPrg, NULL);

  if (Param.InName != NULL)
    delete Param.InName;

  return 0;
}
