#ifndef TUXCAD_H
#define TUXCAD_H

// #define USE_STL
// Insert fr Vektoren scheint (zumindest bei meiner g++-Version) nicht 
// korrekt zu funktionieren. Werden die Elemente mit push_back angehngt, 
// geht's. Vorerst eigene Listen

#include <assert.h>
#ifdef USE_STL
  #include <g++/vector.h>
#endif

#include <libtools.h>
#include <libnet.h>

#include <libtools.h>
#include <libdlg.h>
#include <libprg.h>
#include <libhelp.h>

#include <libcad3d.h>

#include "messages.h"


#ifndef OK
  #define OK 0
#endif

#ifndef FAIL
  #define FAIL 1
#endif


#define DOUBLE_CLICK_SPEED 250  // Milliseconds

#define CAD_PATH "$HOME/tuxcad/data/"

#define POS_X 0
#define POS_Y 1
#define POS_Z 2
#define ROT_X 3
#define ROT_Y 4
#define ROT_Z 5

#define  CFG_TUXCAD_PATH      1          // Path of tuxCAD
#define  CFG_DATA_PATH        2          // Path of tuxCAD data
#define  CFG_LIBS_PATH        3          // Path of filter libraries
#define  CFG_POS_X            4          // x-Position of main window
#define  CFG_POS_Y            5          // y-Position of main window
#define  CFG_SIZE_X           6          // width of tuxcad window
#define  CFG_SIZE_Y           7          // height of tuxcad window
#define  CFG_CREATE_FLOOR    10          // Create new worlds with floor ?
#define  CFG_SHOW_COORDS     11          // Show coordinates 
#define  CFG_GL_SHADING     100          // Default shading mode

enum DIALOGS
{
  DLG_PART   = (1L<<0),
  DLG_DEV    = (1L<<1),
  DLG_WORLD  = (1L<<2),
  DLG_TREE   = (1L<<3),
  DLG_CREATE = (1L<<4)
};

#define DLG_ALL 0x1F



namespace TUXCAD
{
  class BUFFER;
  class DATA;
};
  

enum WORLD_SELECTION
{
  USE_WORLD = 1,
  SELECT_WORLD
};


enum RAIL_ID
{
  MT14902 = 1,
  MT14903,
  MT14904,
  MT14905,
  MT14906,
  MT14907,
  MT14908,
  MT14909,
  MT14912,
  MT14914,
  MT14916,
  MT14917,
  MT14918,
  MT14922,
  MT14924,
  MT14926,
  MT14927,
  MT14928,
  MT14936,
  MT14938,
  MT14951,
  MT14953,
  MT14954,
  MT14955,
  MT14956,
  MT14957,
  MT14959,
  MT14978,
  MT14981,
  MT14983
};


typedef struct
{
  POINT2D Shape[12];        // shape of rail
  double  Lane;             // distance between wheels
  double  SleeperL;         // length of sleeper (width of rail)
  double  SleeperH;         // height of sleeper
  double  SleeperW;         // width of sleeper
  double  Gap;              // space between timber
} RAIL_DATA;


#define FUNC_ABORT \
        { \
	  PRGPrintf ("Function aborted"); \
          return; \
        }

typedef struct
{
  int   ID;
  int   PortNo;
  char *HostName;
} SOCKET_DATA;


typedef struct
{
  char   *CfgName;
  char   *ProName;
  int     FileNo;
  char  **InName;
  bool    verbose;
} PROGRAM_PARAM;


// Prototypes:
class SCHEDULER;
class TREE_DIALOG;
class PART_DIALOG;
class DEV_DIALOG;
class WORLD_DIALOG;
class CREATE_DIALOG;
class HELP_DIALOG;


class CAD_EDITOR
{
  public :
    enum EDITOR_COLORS
    {
      Background = 0,
      Raster,
      Line,
      CrossHair
    };

             CAD_EDITOR       (char *Name, int SizeX = 400, int SizeY = 400);

    int      GetWidth         ()   { return SizeX; }
    int      GetHeight        ()   { return SizeX; }
    double   SnapPos          (double Pos, double Raster = -1.0);
    void     Printf           (char *Fmt, ...);
    void     PrintPos         (int x, int y);
    void     Redraw           ();
    void     DrawGrid         ();
    void     DrawLine         (int x1, int y1, int x2, int y2, 
			       EDITOR_COLORS Color);
    void     DrawCrossHair    (int x, int y);
    void     AddPoint         (double x, double y);
    void     DeletePoint      (int No);
    void     Clear            ();
    void     SetColor         (EDITOR_COLORS Color);
    void     ClosePath        ();
    void     SetSnap          (double Value);
    void     TransformPoints  ();
    int      GetNum           ()   { return Num;    }
    POINT2D *GetPoints        ()   { return Points; }
    Widget   GetShell         ()   { return Shell;  }

  private :
    void    SetPos             (int ScrBarX, int ScrBarY);

    int     SizeX, SizeY;
    double  OffsX, OffsY;
    double  Grid, Zoom;
    double  Snap;
    int     Num;
    POINT2D Points[100];

    Widget  Shell, Work;
    Widget  Msg1, Msg2, Msg;
    Widget  ScrBarX, ScrBarY;
    Window  Win;
    Pixmap  Pic;
    Pixel   Colors[5];
    GC      DrawGC, XorGC;
};


template <class Type = void *>class LISTELEM
{
  public :
    LISTELEM (const Type &Value) :
      Data (Value),
      Prev (NULL),
      Next (NULL)
    {
    };

  public :
    Type      Data;
    LISTELEM *Prev;
    LISTELEM *Next;
};

template <class Type = void *>class LIST
{
  public :
    // Constructor / Destructor:
               LIST       ();
    virtual   ~LIST       ();

    // Methods:
    virtual void       Add        (const Type &Value);
    virtual void       Delete     (const Type &Value);
    virtual bool       Exist      (const Type &Value) const;
    virtual int        Size       ()                  const;
    virtual Type       Get        (int Index)         const;
    virtual void       Clear      ();

  protected :

    // Attributes:
    int             Number;
    LISTELEM<Type> *List, *Last;
};


template <class Type = void *>class SORTLIST : public LIST<Type>
{
  public :
    // Constructor / Destructor:
    SORTLIST  ();

    virtual void       Add        (const Type &Value);
};


class DATA
{
  public :
    enum SELECTION
    {
      SELECT,
      UNSELECT,
      SWAP
    };

    enum ACTION
    {
      NORMAL_MODE = -1,                 // No action selected
      TRANS_ABSOLUTE,                   // Translate part/dev absolute
      ROT_ABSOLUTE,                     // Rotate part/dev absolute
      TRANS_RELATIVE,                   // Tranlate part/dev relative
      ROT_RELATIVE,                     // Rotate part/dev relative
      MOVE_TRANSLATE,                   // Move translatoric joint
      MOVE_ROTATE,                      // Move rotatoric joint
      ZOOM_INTERACTIVE,                 // Zoom interactive
      ZOOM_DISCRETE,                    // Zoom discrete
      TCP_TRANS_ABS,                    // Translate TCP absolute
      TCP_ROT_ABS,                      // Rotate TCP absolute
      TCP_TRANS_REL,                    // Translate TCP relative
      TCP_ROT_REL,                      // Rotate TCP relative
      PAN_MODE,                         // Pan world
      FLY_MODE                          // Fly through world
    };

    // Constructor / Destructor:
                      DATA            (WORLD *World);
                     ~DATA            ();

    // Methods:
    void              Select          (PART3D   *Part, SELECTION Mode);
    void              Select          (DEVICE3D *Dev,  SELECTION Mode);
    bool              Selected        (PART3D   *Part) const;
    bool              Selected        (DEVICE3D *Dev)  const;
    void              SetMouse        (int MouseX, int MouseY);
    int               SelectedParts   () const;
    int               SelectedDevs    () const;
    int               GetMouseX       () const;
    int               GetMouseY       () const;
    void              ClearSelection  ();

    LIST<PART3D *>   *GetPartBuffer   ();
    LIST<DEVICE3D *> *GetDevBuffer    ();

    // public Attributes:
    WORLD          *World;
    ACTION          Action;
    ROBOT          *Robot;
    SCHEDULER      *Scheduler;
    TREE_DIALOG    *TreeDialog;
    PART_DIALOG    *PartDialog; 
    DEV_DIALOG     *DevDialog;
    WORLD_DIALOG   *WorldDialog;
    CREATE_DIALOG  *CreateDialog;

  private :
    // Attributes:
    int              MouseX, MouseY;

    LIST<PART3D *>   SelParts;
    LIST<DEVICE3D *> SelDevs;
};


class BUFFER
{
  public :
    // Constructor / Destructor:
                      BUFFER    ();
                     ~BUFFER    ();

    // Methods:
    void              Add       (PART3D *Part);
    void              Add       (DEVICE3D *Dev);
    void              Clear     ();

    int               Size      () const;

    LIST<DEVICE3D *> *GetBuffer ();

  private :
    // Attributes:
    LIST<DEVICE3D *> DevList;
};


class FILTER
{
  public :
    typedef void *(*FILTERFUNC)(void *LibPtr, WORLD *World, 
				const char *FileSpec);

    // Constructor:
          FILTER   (const char *LibDir);

    // Methods:
    void *Execute  (WORLD *World, const char *FileSpec, FILTERFUNC Function);

  private :
    char LibDir[PATH_MAX+1];
};


#define UPDATE(World) ((WORKCELL*)World)

extern PROGRAM_PARAM  Param;
extern LIST<DATA *>   Data;
extern BUFFER         Buffer;
extern SOCKET_DATA    Server;
extern SOCKET_DATA    Client;
extern SOCKET_DATA    Robot;
extern DATA          *Current;
extern HELP_DIALOG    HelpDlg;

extern MENU_ENTRY     WorldEditMenu[];
extern MENU_ENTRY     WorldViewMenu[];
extern MENU_ENTRY     WorldMenu[];
extern int            WorldMenuEntries;


// ************************************************************************
// **                          data.cc                                   **
// ************************************************************************
void              InsWorld          (WORLD *World);
void              DelWorld          (WORLD *World);
void              SetCurrent        (WORLD *World);
WORLD            *GetCurrent        ();
int               GetWorldNo        ();
DATA             *GetData           (int No);
WORLD            *GetWorld          (int No);
void              SwapSelection     (WORLD *World, PART3D *Part);
void              SwapSelection     (WORLD *World, DEVICE3D *Dev);

LIST<PART3D *>   *GetPartBuffer     (WORLD *World);
LIST<DEVICE3D *> *GetDevBuffer      (WORLD *World);

void              ClearSelection    (WORLD *World);
void              SetAction         (WORLD *World, 
				     DATA::ACTION Action = DATA::NORMAL_MODE);
DATA::ACTION      GetAction         (WORLD *World);
void              SetMouse          (WORLD *World, int MouseX, int MouseY);
int               GetMouseX         (WORLD *World);
int               GetMouseY         (WORLD *World);
SCHEDULER        *GetScheduler      (WORLD *World);
void              UpdateDialogs     (WORLD *World, 
				     unsigned long Mask = DLG_ALL);

void              OpenTreeDialog    (WORLD *World);
void              OpenPartDialog    (WORLD *World, PART3D *Part);
void              OpenDevDialog     (WORLD *World, DEVICE3D *Dev);
void              OpenWorldDialog   (WORLD *World);
void              OpenCreateDialog  (WORLD *World);


// ************************************************************************
// **                           handle.cc                                **
// ************************************************************************
bool         WorldMouseHandle  (WORLD *World,
				int Button, int MouseX, int MouseY);
void         ButtonHandle      (WORLD *World, int Button, 
				int MouseX, int MouseY);
void         KeyHandle         (WORLD *World, KeySym Key, int Modifier);
void         DestroyHandle     (WORLD *World);


// ************************************************************************
// **                           world.cc                                 **
// ************************************************************************
WORLD       *NewWorld          (char *Name);
bool         DeleteWorld       (WORLD *World);
void         WorldMouseCursor  (WORLD *World, int Cursor = -1);
WORLD       *SelectWorld       (bool Create, int Cursor = ToolNORMAL,
				int *x = NULL, int *y = NULL,
				int *Button = NULL);

bool         CheckWorld        (bool create = FALSE);


// ************************************************************************
// **                           utils.cc                                 **
// ************************************************************************
void         Fly               ();
PART3D      *ChoosePart        (const char *Msg,
				WORLD *World = NULL, int px = -1, int py = -1);
PART3D      *SelectPart        ();
DEVICE3D    *ChooseDev         (const char *Msg,
				WORLD *World = NULL, int px = -1, int py = -1);


// ************************************************************************
// **                           system.cc                                **
// ************************************************************************
void         NewWorld          ();
void         Load              (int Select);
void         LoadPart          (int Select);
void         LoadDev           (int Select);
void         LoadWorld         ();
void         LoadConfig        ();
void         Import            ();
void         SavePart          ();
void         SaveDev           ();
void         SaveWorld         (int Select);
void         SaveConfig        ();
void         ClearWorld        ();
void         DeleteWorld       ();
void         PrgConfig         ();
void         ExitPrg           ();
bool         OverwriteFile     (const char *FileName);


// ************************************************************************
// **                             handle.cc                              **
// ************************************************************************
void         OpenWorldMenu     (Widget Shell, WORLD *World, 
				int MouseX = -1, int MouseY = -1);
void         OpenPartMenu      (Widget Shell, const LIST<PART3D *> &Parts, 
				int MouseX = -1, int MouseY = -1);
void         OpenDevMenu       (Widget Shell, const LIST<DEVICE3D *> &Devs, 
				int MouseX = -1, int MouseY = -1);
void         SetWorldHandles   (WORLD *World);


// ************************************************************************
// **                              dialog.cc                             **
// ************************************************************************
class CAD_DIALOG
{
  public :
    typedef void (* CHANGE_HANDLE)  (DIALOG *Dlg, int BoxID);

    // Constructor/Destructor:
                   CAD_DIALOG      (int SizeX = -1, int SizeY = -1);
    virtual       ~CAD_DIALOG      ();

    // Methods:
    virtual  bool  Open            ();
    virtual  bool  Close           ();
    virtual  void  Update          () = 0;

    virtual  void  ChangeCB        (int BoxID);
    virtual  void  HelpCB          (int ID);

    virtual  void  OnChange        (CHANGE_HANDLE Handle);

    // Friend Functions:
    friend   void  HelpDlgCB       (DIALOG *Dlg, int ID);

  protected :
    DIALOG        *Dlg;
    int            PosX, PosY;
    int            SizeX, SizeY;
    CHANGE_HANDLE  Handle;
};


// ************************************************************************
// **                             color.cc                              **
// ************************************************************************
#define  BASE_MASK  (0<<0)
#define  OBJ_MASK   (1<<0)
#define  TEX_MASK   (1<<1)
#define  ALPHA_MASK (1<<2)

COLOR        ColorCB           (DIALOG *Dlg, int BaseID, COLOR Color, 
				             int BoxID, int Mask);
void         SetColor          (DIALOG *Dlg, int BaseID, COLOR Color, 
				             int Mask);
void         SelectColor       (DIALOG *Dlg, int BaseID, int ColorID, 
				             int Mask);
void         ColorDialog       (DIALOG *Dlg, int BaseID, COLOR Color, 
		  		             int Mask);

// ************************************************************************
// **                             material.cc                            **
// ************************************************************************
void         MaterialEdit      ();
MATERIAL    *MaterialCB        (DIALOG *Dlg, int BaseID, MATERIAL *Material, 
				             int BoxID, int Mask);
void         MaterialDialog    (DIALOG *Dlg, int BaseID, MATERIAL *Material, 
				             int Mask);

// ************************************************************************
// **                             light.cc                               **
// ************************************************************************
LIGHT       *LightCB           (DIALOG *Dlg, int BaseID, LIGHT *Light, 
			       	             int BoxID);
void         SetLight          (DIALOG *Dlg, int BaseID, LIGHT *Light);
void         LightDialog       (DIALOG *Dlg, int BaseID, LIGHT *Light);


// ************************************************************************
// **                               view.cc                              **
// ************************************************************************
VIEW        *ViewCB            (DIALOG *Dlg, int BaseID, 
				             WORLD *World, int ViewNo, 
			       	             int BoxID);
bool         SetView           (DIALOG *Dlg, int BaseID, 
				             WORLD *World, int ViewNo);
void         ViewDialog        (DIALOG *Dlg, int BaseID, 
				             WORLD *World, int ViewNo);

// ************************************************************************
// **                                fog.cc                              **
// ************************************************************************
FOG         *FogCB             (DIALOG *Dlg, int BaseID, FOG *Fog, int BoxID);
void         FogDialog         (DIALOG *Dlg, int BaseID, FOG *Fog);


// ************************************************************************
// **                             shadow.cc                              **
// ************************************************************************
SHADOW      *ShadowCB          (DIALOG *Dlg, int BaseID, SHADOW *Shadow, 
				             int BoxID);
void         SetShadow         (DIALOG *Dlg, int BaseID, SHADOW *Shadow);
void         ShadowDialog      (DIALOG *Dlg, int BaseID, SHADOW *Shadow);


// ************************************************************************
// **                              freedom.cc                            **
// ************************************************************************
JOINT       *JointCB           (DIALOG *Dlg, int BaseID, PART3D *Part, 
				             int BoxID);
void         JointDialog       (DIALOG *Dlg, int BaseID, PART3D *part);


// ************************************************************************
// **                                box.cc                              **
// ************************************************************************
BOX         *BoxCB             (DIALOG *Dlg, int BaseID, BOX *Box, int BoxID);
void         BoxDialog         (DIALOG *Dlg, int BaseID, BOX *Box);


// ************************************************************************
// **                             craete.cc                              **
// ************************************************************************
class CREATE_DIALOG : public CAD_DIALOG
{
  public :
    typedef void (* CHANGE_HANDLE)  (DIALOG *Dlg, int BoxID, PART3D *Part);

    // Constructor/Destructor:
                  CREATE_DIALOG    (WORLD *World);

    // Methods: 
    virtual  bool  Open            ();
    virtual  void  Update          ();

    virtual  void  ChangeCB        (int BoxID); 
    virtual  void  HelpCB          (int ID);

    virtual  void  OnChange        (CHANGE_HANDLE Handle);
             void  MakePreview     (Widget widget);
             Widget  MakeForm        (Widget Parent);

    // Friend Functions:
    friend   void  CREATE_ChangeCB   ();

  private :
    // Attributes:
    WORLD         *World;
    PART3D        *Part;
    WORLD         *Preview;
    PART3D        *Demo;
    CHANGE_HANDLE  Handle;
    Widget         Form;
};

void         CreatePart         ();


// ************************************************************************
// **                             editor.cc                              **
// ************************************************************************
void         EditPart           ();


// ************************************************************************
// **                               tree.cc                              **
// ************************************************************************
class TREE_DIALOG : public CAD_DIALOG
{
  public :
    // Constructor/Destructor:
                  TREE_DIALOG      (WORLD *World);
    virtual      ~TREE_DIALOG      ();

    // Methods: 
    virtual  bool  Open            ();
    virtual  void  Update          ();
             void  Expand          ();
             void  Collapse        ();

             void  ButtonCB        (int Button);
    virtual  void  ChangeCB        (int BoxID); 
    virtual  void  HelpCB          (int ID);

    // Friend Functions:
    friend   void  TREE_DialogCB   (DIALOG *Dlg, int Button);
    friend   int   TREE_SelectCB   (DIALOG *Dlg, int BoxID);

  private :
    // Methods:
             int   SetTreePart     (int BoxID, 
				    PART3D *PartList, int ID, int Parent);
             int   SetTreeDev      (int BoxID, 
				    DEVICE3D *DevList, int ID, int Parent);
	     void  SetTreeWorld    (int BoxID);

    // Attributes:
    WORLD  *World;
    Pixmap  ClosePic;
    Pixmap  OpenPic;
};

void         ShowTree          (int Select);


// ************************************************************************
// **                          property.cc                               **
// ************************************************************************
class PART_DIALOG : public CAD_DIALOG
{
  public :
    typedef void (* CHANGE_HANDLE) (DIALOG *Dlg, int BoxID, PART3D *Part);

    // Constructor/Destructor:
                  PART_DIALOG   (WORLD *World);

    // Methods:
    virtual bool  Open          (PART3D *Part);
    virtual void  Update        ();
            void  Set           (PART3D *Part);

    virtual void  ChangeCB      (int BoxID);
    virtual void  HelpCB        (int ID);

    virtual void  OnChange      (CHANGE_HANDLE Handle);

    // Friend Methods:
    friend  void  PART_ChangeCB  ();

  private :
    // Attributes:
    WORLD         *World;
    PART3D        *Part;
    CHANGE_HANDLE  Handle;
};


class DEV_DIALOG : public CAD_DIALOG
{
  public :
    typedef void (* CHANGE_HANDLE) (DIALOG *Dlg, int BoxID, DEVICE3D *Dev);

    // Constructor/Destructor:
                   DEV_DIALOG    (WORLD *World);

    // Methods:
    virtual  bool  Open          (DEVICE3D *Dev);
    virtual  void  Update        ();
             void  Set           (DEVICE3D *Dev);

    virtual  void  ChangeCB      (int BoxID);
    virtual  void  HelpCB        (int ID);

    virtual  void  OnChange      (CHANGE_HANDLE Handle);

    // Friend Methods:
    friend   void  DEV_ChangeCB  ();

  private :
    // Attributes:
    WORLD         *World;
    DEVICE3D      *Dev;
    CHANGE_HANDLE  Handle;
};


class WORLD_DIALOG : public CAD_DIALOG
{
  public :
    typedef void (* CHANGE_HANDLE) (DIALOG *Dlg, int BoxID, WORLD *World);

    // Constructor/Destructor:
                   WORLD_DIALOG    (WORLD *World);

    // Methods:
    virtual  bool  Open            ();
    virtual  void  Update          ();

    virtual  void  ChangeCB        (int BoxID);
    virtual  void  HelpCB          (int ID);

    virtual  void  OnChange        (CHANGE_HANDLE Handle);

    // Friend Methods:
    friend   void  WORLD_ChangeCB  ();

  private :
    // Attributes:
    WORLD         *World;
    CHANGE_HANDLE  Handle;
};


void         PartProperties    (WORLD *World, PART3D   *Part);
void         DevProperties     (WORLD *World, DEVICE3D *Dev);
void         WorldProperties   (int Select);

// ************************************************************************
// **                               cad.cc                               **
// ************************************************************************
void         MergeParts        ();
void         SmashPart         ();


// ************************************************************************
// **                             filter.cc                              **
// ************************************************************************
DEVICE3D    *ImportDevice      (WORLD *World, const char *FileSpec);
WORLD       *ImportWorld       (WORLD *World, const char *FileName);
void         ListFormats       ();


// ************************************************************************
// **                            simulation.cc                           **
// ************************************************************************
void         MotionControl     ();
void         MotionTest1       ();
void         MotionTest2       ();


// ************************************************************************
// **                             robot.cc                               **
// ************************************************************************
void         TeachPanel        ();


// ************************************************************************
// **                             plot3d.cc                              **
// ************************************************************************
void         Plot3D            ();


// ************************************************************************
// **                             train.cc                               **
// ************************************************************************
void         MakeRail          (int ID);
PART        *CreateRail        (RAIL_DATA Data, char *Name, double length);


// ************************************************************************
// **                             fonts.cc                               **
// ************************************************************************
void        MakeText           ();
void        MakeText2          ();


// ************************************************************************
// **                             socket.cc                              **
// ************************************************************************
void         StartCLIServer    ();
bool         RunCLIServer      (char *HostName, int PortNo);
void         CloseCLIServer    ();
void         StartVPLUSServer  ();
bool         RunVPLUSServer    (char *HostName, int PortNo);
void         CloseVPLUSServer  ();
void         StartClient       ();
void         SendData          ();
void         CloseClient       ();


// ************************************************************************
// **                               test.cc                              **
// ************************************************************************
void         SwapToolbar       ();
void         ReadDataFile      ();
void         Preview           ();


// ************************************************************************
// **                               info.cc                              **
// ************************************************************************
void         PartInfo          (PART3D   *Part);
void         DevInfo           (DEVICE3D *Dev);
void         WorldInfo         (WORLD    *World);


// ************************************************************************
// **                             help.cc                                **
// ************************************************************************
class HELP_DIALOG
{
  public :
    // Constructor :
          HELP_DIALOG  ();

    // Methods :
            bool  Open       (const char *Title, 
			      const char *FileSpec, const char *Label);
	    void  Close      ();

    friend  void  CloseHelp  (HELPWIN *HelpWin);

  private :
    // Attributes:
    int      SizeX, SizeY;
    int      PosX, PosY;
    HELPWIN *HelpWin;
};

void         Manual            ();
void         About             ();
void         ShowSyntax        ();

// ************************************************************************
// **                             task.cc                                **
// ************************************************************************
class TASK
{
  public :
    typedef int (*TASK_EVENT)(TASK *Task, double Time, void *Data);

    // Constructor:
                    TASK      (const MOTION &Motion, double Start, 
			       TASK_EVENT Event = NULL, void *Data = NULL);
    virtual        ~TASK      ();

    // Methods:
    virtual bool    Execute   (double Time);
            void    SetStart  (double Start);
            void    SetEvent  (TASK_EVENT Event, void *Data = NULL);

	    double  GetStart  () const;
            double  GetEnd    () const;
    virtual char   *GetName   () const;

    friend  int     Comp      (const TASK *Task1, const TASK *Task2);

  public :
    // Attributes:
    // TODO: TaskID (int) ????
    MOTION      Motion;
    double      Start;
    TASK       *Next;
    TASK_EVENT  Event;
    void       *Data;
};

class PART_TASK : public TASK
{
  public :
    // Constructor:
                  PART_TASK (PART3D *Part, const MOTION &Motion, double Start,
			     TASK_EVENT Event = NULL);

    // Methods:
    virtual bool  Execute   (double Time);
    virtual char *GetName   () const;

  private :
    PART3D *Part;
};

class DEV_TASK : public TASK
{
  public :
    // Constructor:
                  DEV_TASK  (DEVICE3D *Dev, const MOTION &Motion, double Start,
			     TASK_EVENT Event = NULL);

    // Methods:
    virtual bool  Execute   (double Time);
    virtual char *GetName   () const;

  private :
    // Attributes :
    DEVICE3D *Dev;
};

class TCP_TASK : public TASK
{
  public :
    // Constructor:
                  TCP_TASK  (DEVICE3D *Dev, const MOTION &Motion, double Start,
			     TASK_EVENT Event = NULL);

    // Methods:
    virtual bool  Execute   (double Time);
    virtual char *GetName   () const;

  private :
    DEVICE3D *Dev;
};


// ************************************************************************
// **                            simulator.cc                            **
// ************************************************************************
class SCHEDULER
{
  public :
    typedef int (*SCHEDULER_EVENT)(SCHEDULER *Scheduler, double Time, 
				   void *Data);

    enum STATUS
    {
      STOPPED = 0,
      RUNNING
    };

            SCHEDULER    (WORLD *World);
           ~SCHEDULER    ();

    void    AddTask      (TASK *Task);
    void    DelTask      (TASK *Task);
    TASK   *GetTask      (int No);
    int     GetTaskNo    () const;
    void    Clear        ();
    void    MoveTask     (TASK *Task, double Dist);
    void    PlaceTask    (TASK *Task, double Time);
    bool    SetTime      (double Time);
    void    SetSpeed     (double Speed);
    void    SetEvent     (SCHEDULER_EVENT Event, void *Data);

    void    Run          ();
    void    Stop         ();
    void    Step         ();
    double  GetTime      () const;
    double  GetSpeed     () const;
    STATUS  GetStatus    () const;
    double  GetMaxTime   () const;

    bool    Execute      (double Time);

    // TODO: Repeat-Funktion ber Zeit / User-Ereignis (z.B. Signal)

  private :
    void    CalcMaxTime  ();

    WORLD            *World;             // Workcell
#ifdef USE_STL
    vector<TASK *>    TaskList; // (1, TASK(MOTION(FZero, FZero, 0.0), 0.0));
#else
    SORTLIST<TASK *>  TaskList;          // Sorted list of tasks
#endif
    double            Time;              // Current Scheduler time
    double            MaxTime;           // finishing time of last task
    double            Speed;             // Stepsize for execution
    STATUS            status;            // Scheduler status
    SCHEDULER_EVENT   Event;
    void             *Data;
};


class TASK_DIALOG
{
  public :
    // Constructor/Destructor:
               TASK_DIALOG   (WORLD *World, SCHEDULER *Scheduler = NULL);
              ~TASK_DIALOG   ();

    // Methods:
    bool       Open          ();

    SCHEDULER *GetScheduler  ();

    void       Refresh       (int x, int y, int w, int h);
    void       SetAbsTime    (double Time);
    void       SetRelTime    (double Time);
    void       DrawTime      () const;

  private :
    // Methods: 
    int        GetLine      (TASK *Task) const;
    void       DrawRuler    (int x1, int x2) const;
    void       DrawTask     (TASK *Task) const;

    // Attributes:
    typedef struct TASK_LINE
    {
      int     PosY, SizeY;
      char   *Name;
      Pixel   Color;
    };

    WORLD       *World;
    SCHEDULER   *Scheduler;
    int          LineNo;
    TASK_LINE   *Lines;

    Widget       Shell;
    Widget       ScrWin, ClipWin;
    Widget       ScrBarX, ScrBarY;
    Widget       Work;
    Widget       Actions[10];

    int          SizeY, LineY;
    Pixel        bg;
    Window       Win;
    GC           DrawGC;
    XFontStruct *Font;
};


// using namespace TUXCAD;


#endif
