#include <math.h>

#include <Xm/Label.h>

#include "tuxCAD.h"

#include <icons/sphere.xpm>


static void UpdateView (int MouseX, int MouseY)
{
  RPY    Camera;
  WORLD *World = GetCurrent ();

  if (World == NULL)
    return;

  Camera.y = - MouseY * 3.0;
  Camera.p = 0.0;
  Camera.r = (30 - MouseX) * 6.0;

  World->SetCamera (Camera);
}


static void SphereHandle (Widget widget, XtPointer Data, XEvent *event)
{
  static int MouseX = 30, MouseY = 30;

  if (widget == NULL)
    return;

  switch (event->type)
  {
    case MotionNotify :
      {
       	XMotionEvent *Motion = (XMotionEvent *)event;

	if (Motion->state & Button1Mask && 
	    SQR(Motion->x-30) + SQR(Motion->y-30) <= 900)
	{

	  XDrawRectangle (XDisplay, XtWindow(widget), xorGC, 
			            MouseX-2, MouseY-2, 4, 4);
	  MouseX = Motion->x;
	  MouseY = Motion->y;
	  XDrawRectangle (XDisplay, XtWindow(widget), xorGC, 
			            MouseX-2, MouseY-2, 4, 4);
	  UpdateView (MouseX, MouseY);
	}
      }
      break;

    case ButtonPress :
      {
	XButtonEvent *Button = (XButtonEvent *)event;

	if (SQR(Button->x-30) + SQR(Button->y-30) <= 900)
	{
	  XDrawRectangle (XDisplay, XtWindow(widget), xorGC, 
			            MouseX-2, MouseY-2, 4, 4);
	  MouseX = Button->x;
	  MouseY = Button->y;
	  XDrawRectangle (XDisplay, XtWindow(widget), xorGC, 
			            MouseX-2, MouseY-2, 4, 4);
	  UpdateView (MouseX, MouseY);
	}
      }
      break;

    case Expose :
      {
	RPY    Camera;
	WORLD *World = GetCurrent ();

	if (World != NULL)
	{
	  Camera = World->GetCamera ();
	  MouseX = (int)(fmod(Camera.r, 360.0) / 6.0) - 3;
	  MouseY = -(int)(fmod(Camera.y, 360.0) / 3.0);

	  XDrawRectangle (XDisplay, XtWindow(widget), xorGC, 
			            MouseX-2, MouseY-2, 4, 4);
	}
      }
      break;

    default :
      break;
  }
}

static Widget CreateSphere (DIALOG *Dlg, Widget Parent, void *Param)
{
  Pixel   Color;
  Pixmap  PixMap;
  Widget Result;
  Arg    args[5];

  Result = XtCreateManagedWidget ("sphere", xmLabelWidgetClass, 
				  Parent, args, 0);

  XtVaGetValues (Result, XmNbackground, &Color, NULL);
  PixMap = TOOLMakeXPM (sphere, NULL, NULL, Color);

  XtSetArg (args[0], XmNlabelType,              XmPIXMAP);
  XtSetArg (args[1], XmNlabelPixmap,            PixMap);
  XtSetArg (args[2], XmNlabelInsensitivePixmap, PixMap);
  XtSetValues (Result, args, 3);

  XtAddEventHandler (Result, PointerMotionMask | ExposureMask | 
		             ButtonPressMask, FALSE,
		             (XtEventHandler)SphereHandle, NULL);

  XSetForeground (XDisplay, xorGC, TOOLGetColor("White"));

  return Result;
}

void Fly ()
{
  DIALOG *Dlg;

  Dlg = DLGNewDialog (0);

  DLGInsertWidget (Dlg, 0, CreateSphere, NULL);
  DLGOpenDialog   (Dlg, "fly", DlgNORMAL, NULL);
  DLGFreeDialog   (Dlg);
}


PART3D *ChoosePart (const char *Msg, WORLD *World /* = NULL */, 
		    int px /* = -1 */, int py /* = -1 */)
{
#if 0
  int      count = 0, PartNo;
  char     Label[20];
  char   **PartNames;
  PART3D **PartList;
  DIALOG  *Dlg;
#endif
  PART3D  *Part;
  WORLD   *SelWorld = World;

  if (px < 0 || py < 0)
    do
    {
      if (Msg != NULL)
	PRGPrintf (Msg);
      if ((SelWorld = SelectWorld(FALSE, ToolNORMAL, &px, &py)) == NULL)
      {
	PRGPrintf (MSG_ABORT);
	return NULL;
      }
    }
    while (World != NULL && SelWorld != World);

  if ((Part = SelWorld->GetPart(px, py)) == NULL)
#if 0
  {
    // No part found -> open part-list dialog
    PartNo = UPDATE(World)->GetPartList(&PartList);
    PartNames = new char *[PartNo];

    for (count = 0; count < PartNo; ++count)
      PartNames[count] = PartList[count]->GetName();

    Dlg = DLGNewDialog (0);
    DLGMakeRows (Dlg, 0, 2, DlgNO_LINE, 10, 20);
    sprintf (Label, "Parts: %d", PartNo);
    DLGInsertLabel (Dlg, 10, "partNo", Label);
    DLGInsertList  (Dlg, 20, "partList", 5, -PartNo, PartNames, 0);
    DLGOpenDialog (Dlg, "selectPart", DlgNORMAL, NULL);
    if (DLGGetExitButton(Dlg) == DlgOK)
      Part = PartList[DLGGetList(Dlg, 20)];

    DLGFreeDialog (Dlg);

    delete PartList;
    delete PartNames;
  }
#else
    PRGPrintf ("No part found");
#endif    

  return Part;
}


PART3D *SelectPart ()
{
  PART3D *Part;
  WORLD  *World = GetCurrent ();

  Part = ChoosePart ("Select part");

  ClearSelection (World);
  SwapSelection (World, Part);

  return Part;
}


DEVICE3D *ChooseDev (const char *Msg, WORLD *World /* = NULL */, 
		     int px /* = -1 */, int py /* = -1 */)
{
#if 0
  int        count = 0, DevNo;
  char       Label[20];
  char     **DevNames;
  DEVICE3D **DevList;
  DIALOG    *Dlg;
#endif
  DEVICE3D  *Dev;

  if (Msg != NULL)
    PRGPrintf (Msg);

  if (World == NULL &&
      (World = SelectWorld(FALSE, ToolNORMAL, &px, &py)) == NULL)
    return NULL;

  if (px < 0 && py < 0 && (Dev = ChooseDev(Msg, World, px, py)) == NULL)
    return NULL;

  if ((Dev = World->GetDev(px, py)) == NULL)
#if 0
  {
    DevNo = World->GetDevList(&DevList);
    DevNames = (char **)malloc (sizeof(char*) * DevNo);

    for (count = 0; count < DevNo; ++count)
      DevNames[count] = DevList[count]->GetName ();

    Dlg = DLGNewDialog (0);
    DLGMakeRows (Dlg, 0, 2, DlgNO_LINE, 10, 20);
    sprintf (Label, "Parts: %d", DevNo);
    DLGInsertLabel (Dlg, 10, "devNo", Label);
    DLGInsertList  (Dlg, 20, "devList", 5, -DevNo, DevNames, 1);
    DLGOpenDialog (Dlg, "selectDevice", DlgNORMAL, NULL);
    if (DLGGetExitButton(Dlg) == DlgOK)
      Dev = DevList[DLGGetList(Dlg, 20)];

    DLGFreeDialog (Dlg);

    free (DevList);
    free (DevNames);
  }
#else
  PRGPrintf ("No device found");
#endif

  return Dev;
}


