#include <math.h>
#include <ctype.h>

#include "tuxCAD.h"


#define VPLUS_CHECK_WORLD \
        if ((World = GetCurrent()) == NULL) \
        { \
          PRO0 ("No current WorkCell") \
          return FALSE; \
        }


typedef struct
{
  DEVICE3D *Robot;
  DEVICE3D *Tool;
  double    Speed;
  bool      Hand; 
  bool      Pro, Debug;
  ROBOT    *Robby;
} VPLUS_ROBOT;


typedef int (*VPLUS_FUNCTION)(char*, int);

typedef struct
{
  char            Keyword[24];
  VPLUS_FUNCTION  Fkt;
  char            Msg[80];
} VPLUS_MACRO;


static int V_Break    (char *Command, int Offset);
static int V_Debug    (char *Command, int Offset);
static int V_Drive    (char *Command, int Offset);
static int V_Euler    (char *Command, int Offset);
static int V_Get      (char *Command, int Offset);
static int V_Hand     (char *Command, int Offset);
static int V_Exit     (char *Command, int Offset);
static int V_GetJoint (char *Command, int Offset);
static int V_RPY      (char *Command, int Offset);
static int V_SetJoint (char *Command, int Offset);
static int V_SetPro   (char *Command, int Offset);
static int V_SetSpeed (char *Command, int Offset);
static int V_Teach    (char *Command, int Offset);
static int V_Hexa     (char *Command, int Offset);
static int V_ShowCS   (char *Command, int Offset);


static VPLUS_MACRO Macro[] =
{
  {"ASK",     V_Get,          "Ask/Get"},
  {"BREAK",   V_Break,        "Break"},
  {"GET",     V_Get,          "Get"},
  {"CLOSE",   V_Hand,         "Close Gripper"},
  {"DEBUG",   V_Debug,        "Set Debug"},
  {"DRIVE",   V_Drive,        "Drive Joint"},
  {"EULER",   V_Euler,        "Euler TCP"},
  {"EXIT",    V_Exit,         "Exit V+"},
  {"HERE",    V_GetJoint,     "Get Joint"},
  {"JOINT",   V_SetJoint,     "Set Joint"},
  {"OPEN",    V_Hand,         "Open Gripper"},
  {"PRO",     V_SetPro,       "Set Protocol"},
  {"SPEED",   V_SetSpeed,     "Set Speed"},
  {"TEACH",   V_Teach,        "Teach Position"},
  {"WAIT",    V_Break,        "Wait"},

  {"HEXA",    V_Hexa,         "Set Hexa"},

  {"CS",      V_ShowCS,       "Show CS"},
  {"RPY",     V_RPY,          "RPY TCP"}
};

#define VPLUS_MACROS (int)(sizeof(Macro)/sizeof(VPLUS_MACRO))


static  char         Msg[512]   = "OK";
static  VPLUS_ROBOT  VPlusRobot = {NULL, NULL, 10.0, TRUE, FALSE, FALSE, NULL};
        SOCKET_DATA  Robot      = {-1, 0, NULL};


static void V_SkipSpace (char *Message, int &Pos)
{
  if (Message == NULL)
    return;

  while (isspace(Message[Pos]))
    Pos++;
}


static dual SetVPLUSRobot (DEVICE3D *Robot, DEVICE3D *Tool, ROBOT *Robby)
{
  if ((VPlusRobot.Robot != NULL && Robot != NULL) || 
      (VPlusRobot.Tool  != NULL && Tool  != NULL))
    return FALSE;

  if (VPlusRobot.Robby != NULL)
    delete VPlusRobot.Robby;

  VPlusRobot.Robot = Robot;
  VPlusRobot.Tool  = Tool;
  VPlusRobot.Robby = Robby;

  return TRUE;
}


static char *ExeVPlus (int SocketID, char *Command)
{
  int count = 0, Result = 0;

  assert (SocketID > -1);

  if (Command == NULL || Command[0] == '\0')
  {
    PRO0 ("Empty Command received")
    return "OK";
  }

  //  PRO1 ("<- Cmd: %s\n", Command)
  if (VPlusRobot.Pro)
    PRGPrintf ("Cmd: %s", Command);
    
  while (count < VPLUS_MACROS && 
	 strncasecmp(Command, Macro[count].Keyword, 
		              strlen(Macro[count].Keyword)) != 0)
    count++;
  if (count == VPLUS_MACROS)
  {
    PRO1 ("Unkonwn Command: %s\n", Command)
    return "ERROR";
  }

  if (VPlusRobot.Pro)
    PRO1 ("  %s\n", Macro[count].Msg)

  if (Macro[count].Fkt != NULL)
  {
    if ((Result = Macro[count].Fkt(Command,
 		  		   strlen(Macro[count].Keyword))) == -1)
    {
      Msg[0] = '\0';
      PRO0 ("EXIT\n")
      return NULL;
    }
	
    if (Result == OK)
      sprintf (Msg, "%s", Command);
    else
      sprintf (Msg, "ERROR: %s", Command);
  }
  else
    strcpy (Msg, "0: OK");
 
  PRO1 ("-> V+ ROBOT: %s\n", Msg)

  return Msg;
}


bool RunVPLUSServer (char *HostName, int PortNo)
{
  bool      Result = FALSE;
  char      RobotName[20] = "???";
  H_MATRIX  M;
  DEVICE3D *RobotDev = NULL;
  DEVICE3D *ToolDev = NULL;
  ROBOT    *Robby = NULL;
  WORLD    *World;

  assert (HostName != NULL || PortNo > 0);

  VPLUS_CHECK_WORLD

  if ((RobotDev = World->GetDev("rx60")) != NULL)
  {
    PRGPrintf ("\\found RX60");
    strcpy (RobotName, "rx60");
    Robby = new RX60 ();
    if ((ToolDev = World->GetDev("hand")) != NULL)
      PRGPrintf ("\\+ hand");
  }
  elsif ((RobotDev = World->GetDev("hexa")) != NULL)
  {
    PRGPrintf ("\\found HEXA");
    strcpy (RobotName, "hexa");
    Robby = new HEXA ();
    if ((ToolDev = World->GetDev("mount")) != NULL)
      PRGPrintf ("\\+ hand");
  }
  elsif ((RobotDev = World->GetDev("rm501")) != NULL)
  {
    PRGPrintf ("\\found RM501");
    strcpy (RobotName, "rm501");
    if ((ToolDev = World->GetDev("greifer")) != NULL)
    {
      PRGPrintf ("\\+ hand");
M = ToolDev->GetWTD ();
WRITE_HMATRIX (stdout, M, "FLANSCH")
    }
  }
  else
  {
    PRGPrintf ("\\Robot not found !");
    return FALSE;
  }

  if (PortNo > 0)
    if ((Server.ID = NETInitServer(HostName, PortNo)) < 0)
    {
      PRGPrintf ("connection failed");
      return FALSE;
    }

  if (Server.ID >= 0)
  {
    SetVPLUSRobot (RobotDev, ToolDev, Robby);
    PRGPrintf ("Running V+ (%s) server at Port %d...", RobotName, PortNo);
    Result = (NETRunServer(Server.ID, ExeVPlus, 1) >= 0);
    PRGPrintf ("Server finished");

    SetVPLUSRobot (NULL, NULL, NULL);
  }

  return Result;
}


void StartVPLUSServer ()
{
  if (! RunVPLUSServer("", 2030))
    PRGPrintf ("failed !");
}


void CloseVPLUSServer ()
{
  if (Server.PortNo > 0 && Server.ID >= 0)
    NETCloseServer (Server.ID);

  Server.ID = -1;
}


// ==================== Command - Evaluation ========================
static int V_Break (char *Command, int Offset)
{
  assert (Command != NULL && Offset > 0);

  return OK;
}


static int V_Get (char *Command, int Offset)
{
  int     Signal;
  int     Result = OK;
  char    Message[40];
  DIALOG *Dlg;

  V_SkipSpace (Command, Offset);

  if (strncasecmp(Command+Offset, "SPEED", 5) == 0)
    sprintf (Command, "%f", VPlusRobot.Speed);
  elsif (strncasecmp(Command+Offset, "SIG", 3) == 0)
  {
    Offset += 3;
    V_SkipSpace (Command, Offset);

    if (sscanf(Command+Offset, "(%d)", &Signal) == 1)
      switch (Signal)
      {
        case 1001 :
	  sprintf (Message, "Signal 1001: Hand Right");
	  break;

        case 1002 :
	  sprintf (Message, "Signal 1002: Hand Left");
	  break;

        case 1003 :
	  sprintf (Message, "Signal 1003: Light-Barrier");
	  break;

        default :
	  sprintf (Message, "Signal %d: Light-Barrier", Signal);
      }
    else
      strcpy (Message, "Unknown Signal !");

    sprintf (Command, "%d", 2 - DLGOpenTextDialog("signal", DlgQUESTION,
						  Message, 2, "yes", "no"));
  }
  elsif (strncasecmp(Command+Offset, "BITS", 3) == 0)
  {
    Dlg = DLGNewDialog (0);
    DLGInsertScale (Dlg, 0, "value", 0.0, 255.0, 0.0, 1, DlgHORIZONTAL);
    DLGOpenDialog (Dlg, Command+Offset, DlgNORMAL, NULL);
    if (DLGGetExitButton(Dlg) == DlgOK)
      sprintf (Command, "%d", (int)DLGGetScale(Dlg, 0));
    else
    {
      strcpy (Command, "canceled");
      Result = FAIL;
    }
    DLGFreeDialog (Dlg);
  }
  else
    Result = FAIL;

  return Result;
}


static int V_Debug (char *Command, int Offset)
{
  V_SkipSpace (Command, Offset);

  VPlusRobot.Debug = (strncasecmp(Command+Offset, "ON", 2) == 0);

  return OK;
}


static int V_Drive (char *Command, int Offset)
{
  int     count, No;
  double  Value = 0.0, Speed = 100.0;
  double  Joint, Step;
  PART3D *Part;

  V_SkipSpace (Command, Offset);

  if (sscanf(Command+Offset, "%d %lf %lf", &No, &Value, &Speed) > 0)
  {
    if (Speed <= 0.0)
      return FAIL;

    Part = VPlusRobot.Robot->GetPartList();
    while (Part != NULL && No > 0)
    {
      Part = Part->GetChain();
      No--;
    }

    if (Part != NULL)
    {
      Joint = Part->GetVal (5);
      for (count = 0; count < 100/Speed; ++count)
      {
	Step = Joint + count * (Value - Joint) / 100.0;
        Part->SetJoint (&Step);
      }
    }
  }

  return OK;
}


static int V_Euler (char *Command, int Offset)
{
  double   Euler[6];
  H_MATRIX TCP;

  if (VPlusRobot.Tool == NULL)
    return FAIL;

  TCP = VPlusRobot.Tool->GetWTD ();  // TODO GetTCP

  GEOFrame2Euler (TCP, &Euler[0], &Euler[1], &Euler[2], 
		       &Euler[3], &Euler[4], &Euler[5]);
  sprintf (Command, "%.3f %.3f %.3f %.3f %.3f %.3f",
     	            Euler[0], Euler[1], Euler[2], 
	            Euler[3], Euler[4], Euler[5]);

  return OK;
}


static int V_Hand (char *Command, int Offset)
{
  int       count;
  double    J[2];
  H_MATRIX  TCP;
  PART3D   *Part;
  //  DEVICE3D *Dev;
  WORLD    *World;

  VPLUS_CHECK_WORLD

  if (VPlusRobot.Tool == NULL ||
      (Part = VPlusRobot.Tool->GetPartList()) == NULL)
  {
    strcpy (Command, "No Tool found");
    return FAIL;
  }

  VPlusRobot.Hand = (strncasecmp(Command, "OPEN", 4) == 0);

  Part = Part->GetChain();
  for (count = 0; count < 2; ++count)
  {
    if (Part == NULL)                      // are there two fingers ?
      return FAIL;

    J[count] = (VPlusRobot.Hand)? 
      Part->GetJoint().GetMin(POS_X):Part->GetJoint().GetMax(POS_X);

    Part = Part->GetNext();
  }
  VPlusRobot.Tool->SetJoint (J, 2);
  World->Redraw ();
 
  // Attach Part near hand
  TCP = VPlusRobot.Tool->GetWTD ();   // TODO: GetTCP
  //  Dev = World->FindNear (TCP);
  // World->Attach (Tool, Dev);

  strcpy (Command, "OK");

  return OK;
}


static int V_Exit (char *Command, int Offset)
{
  CloseVPLUSServer ();

  return -1;
}


static int V_GetJoint (char *Command, int Offset)
{
  int       Result = FAIL;
  char      ValStr[20];
  PART3D   *Part;
  DEVICE3D *Dev;

  if (VPlusRobot.Robot != NULL)
  {
    Command[0] = '\0';

    if ((Dev = VPlusRobot.Robot->GetPartList()->GetAttach()) == NULL)
    {
      /* suppose, that we have a serial robot */
      Part = VPlusRobot.Robot->GetPartList()->GetChain ();

      while (Part != NULL)
      {
	sprintf (ValStr, "%.3f ", Part->GetVal(5));
	strcat (Command, ValStr);
	Part = Part->GetChain ();
      }
    }
    else
      while (Dev != NULL)
      {
	Part = Dev->GetPartList()->GetChain();
	sprintf (ValStr, "%.3f ", Part->GetVal(5));
	strcat (Command, ValStr);
	Dev = Dev->GetNext ();
      }

    Result = OK;
  }

  return Result;
}


static int V_RPY (char *Command, int Offset)
{
  double   RPY[6];
  H_MATRIX TCP;

  if (VPlusRobot.Tool == NULL)
    PRO_RETURN ("No tool defined", FAIL)

  TCP = VPlusRobot.Tool->GetWTD ();   // TODO: GetTCP

  GEOFrame2RPY (TCP, &RPY[0], &RPY[1], &RPY[2], &RPY[3], &RPY[4], &RPY[5]);
  PRO6 ("\\: %.2f %.2f %.2f   %.2f %.2f %.2f\n",
	RPY[0], RPY[1], RPY[2], RPY[3], RPY[4], RPY[5])
  sprintf (Command, "%.3f %.3f %.3f %.3f %.3f %.3f",
     	            RPY[0], RPY[1], RPY[2], RPY[3], RPY[4], RPY[5]);

  return OK;
}


static int V_SetJoint (char *Command, int Offset)
{
  int     No, Number;
  double  J[18];
  FRAME   TCP;
  WORLD  *World;

  VPLUS_CHECK_WORLD

  if ((No = sscanf(Command+Offset, "%d", &Number)) > 0)
  {
    No = 0;
    if (Number == 6)
      No = sscanf (Command+Offset, "%d %lf %lf %lf %lf %lf %lf",
		   &Number, &J[0], &J[1], &J[2], &J[3], &J[4], &J[5]);
    elsif (Number == 18)
      No = sscanf (Command+Offset, "%d %lf %lf %lf %lf %lf %lf"
		                     " %lf %lf %lf %lf %lf %lf"
		                     " %lf %lf %lf %lf %lf %lf",
		   &Number, &J[ 0], &J[ 1], &J[ 2], &J[ 3], &J[ 4], &J[ 5],
		            &J[ 6], &J[ 7], &J[ 8], &J[ 9], &J[10], &J[11],
		            &J[12], &J[13], &J[14], &J[15], &J[16], &J[17]);

    if (No > 0)
    {
      if (VPlusRobot.Robot != NULL)
      {
	VPlusRobot.Robot->SetJoint (J, Number);
        World->Redraw ();
      }

      if (Number == 18)
      {
	JOINTS Joints (18, J);
	TCP = VPlusRobot.Robby->FwdKin (Joints);
TCP.Write (NULL, "Place Tool\n");
// TODO: UpdatePos ist falsch -> Place
	VPlusRobot.Tool->UpdatePos (
	  GEOCreateMatrix(TCP[N][X], TCP[N][Y], TCP[N][Z],
			  TCP[O][X], TCP[O][Y], TCP[O][Z],
			  TCP[A][X], TCP[A][Y], TCP[A][Z],
			  TCP[P][X], TCP[P][Y], TCP[P][Z]));
      }

      strcpy (Command, "OK");
    }
  }

  return OK;
}


static int V_SetPro (char *Command, int Offset)
{
  V_SkipSpace (Command, Offset);

  VPlusRobot.Pro = (strncasecmp(Command+Offset, "ON", 2) == 0);

  return OK;
}


static int V_SetSpeed (char *Command, int Offset)
{
  double Speed;

  V_SkipSpace (Command, Offset);

  if (sscanf(Command+Offset, "%lf", &Speed) == 1)
  {
    VPlusRobot.Speed = Speed;
    strcpy (Command, "OK");
  }

  return OK;
}


static int V_Teach (char *Command, int Offset)
{
  DLGOpenTextDialog ("teach", DlgMESSAGE, Command+Offset, 1, "OK");

  return V_GetJoint (Command, Offset);
}


// =======================================================================
// TODO: Mit Befehl HEXA nur Parameter bergeben, Bewegung mit JOINT
static int V_Hexa (char *Command, int Offset)
{
  int       count;
  double    RPY[6];
  double    J[6], Angle[18];
  JOINTS    Joint, J1, J2, J3;
  H_MATRIX  TCP;
  WORLD    *World;
  static    HEXA   Robot (NULL);

printf ("V_HEXA\n");
  VPLUS_CHECK_WORLD

  if (sscanf(Command+Offset, "%lf %lf %lf %lf %lf %lf " 
                             "%lf %lf %lf %lf %lf %lf ",
	     &RPY[0], &RPY[1], &RPY[2], &RPY[3], &RPY[4], &RPY[5], 
	     &J[0],   &J[1],   &J[2],   &J[3],   &J[4],   &J[5]) < 12)
    return FAIL;

  if (VPlusRobot.Tool != NULL)
  {
    TCP = GEORPY2Frame (RPY[0], RPY[1], RPY[2], RPY[3], RPY[4], RPY[5]);
    VPlusRobot.Tool->UpdatePos (TCP); // UpdatePos ist falsch -> Place
  }

  FRAME  TCPF (TCP.m[N][X], TCP.m[N][Y], TCP.m[N][Z],
	       TCP.m[O][X], TCP.m[O][Y], TCP.m[O][Z],
	       TCP.m[A][X], TCP.m[A][Y], TCP.m[A][Z],
	       TCP.m[P][X], TCP.m[P][Y], TCP.m[P][Z]);

  Joint = Robot.InvKin (TCPF);

printf ("Joints #: %d\n", Joint.GetNumber());

  for (count = 0; count < Joint.GetNumber(); ++count)
    Angle[count] = Joint[count];

  VPlusRobot.Robot->SetJoint (Angle, 18);
  World->Redraw ();

  strcpy (Command, "OK");

  return OK;
}

// =======================================================================

static int V_ShowCS (char *Command, int Offset)
{
  double    RPY[6];
  H_MATRIX  M;
  WORLD    *World;

  VPLUS_CHECK_WORLD

  if (sscanf(Command+Offset, "%lf %lf %lf %lf %lf %lf ",
	     &RPY[0],  &RPY[1],  &RPY[2],  &RPY[3],  &RPY[4],  &RPY[5]) < 6)
    return FAIL;
  
  M = GEORPY2Frame (RPY[0], RPY[1], RPY[2], RPY[3], RPY[4], RPY[5]);
 
  UPDATE(World)->DrawFrame (M);
  strcpy (Command, "OK");

  return OK;
}



/*
  V+ Befehle + einige zustzliche Erweiterungen
  Befehl         Parameter           Beschreibung
- ALIVE          ON $|$ OFF          ON: Server bleibt nach Verbindungsende 
                                     aktiv
o ASK            V+ Befehl       
+ CLOSE                              Schliet den Greifer
+ DEBUG          ON $|$ OFF          Schaltet Debug-Modus an/aus
+ EULER                              Ermittelt die Euler-Koordinaten
	 		             <px py pz z1 y z2>
+ EXIT                               Beendet den V+ Server
- FINISH         PERCENT             Wait until PERCENT percent motion has 
                                     been completed
o GET            V+ Befehl        
+ HERE                               Bestimmt die Gelenkwinkel
		 	             <J1 J2 J3 J4 J5 J6>
- HMOVE          px py pz y p r      Relative Bewegung der Hand
- HROTATE        y p r               Relative Rotation der Hand
- HTRANS         px py pz            Relative Translation der Hand
- INRANGE        px py pz z1 y z2    Prft Euler-Position auf Gltigkeit
- INRANGE_RPY    px py pz y p r      Prft RPY-Postion auf Gltigkeit
- INV_KINEMATIC  px py pz y p r      Berechnet Inverse Kinematik
				     <J1 J2 J3 J4 J5 J6>
+ JOINT          J1 J2 J3 J4 J5 J6   Bewegung zu geg. Gelenkstellung
- KINEMATIC      J1 J2 J3 J4 J5 J6   Berechnet Vorwrtskinematik
				     <px py pz z1 y z2>
- MCS            Parameter           Startet Monitor-Befehl
- MONITOR        Signal [Prioritt]  Startet Monitorroutine
- MOUNT          1..4                Nimmt Tool i auf (nur RX90)
+ OPEN                               ffnet den Greifer
- PLACE[S]       px py pz z1 y z2    Bewegung zu geg. Euler-Position
				     mit 'S': lin. kartesische Interpolation
+ PRO            ON $|$ OFF          Protokolliert eingehende Befehle auf dem 
				     BS
- RPY_MOVE[S]    px py pz y p r      Bewegung zu geg. RPY-Position
				     mit 'S': lin. kartesische Bahn
- TEACH          Msg                 Gibt Meldung auf MCP aus und wartet auf
				     Tastendruck
- ULTRASCHALL    [Avg]               Bestimmt Ultraschallsensor 
				     (Bits 1004-1008)
				     Avg: Bildet Mittelwert 
+ WAIT           [Accuracy]          Wartet bis Zielposition mit geg. 
				     Genauigkeit (z.B. 0.05) erreicht 
				     wurden
				     Ohne Accuracy: wie BREAK
- WTRANS         px py pz            Transformation in Welt-Koordinaten
*/


