#include <math.h>

#include "tuxCAD.h"


static void SetRange (DIALOG *Dlg, int BaseID, double *Range)
{
  DLGSetScale (Dlg, BaseID+46, Range[MIN_X]);
  DLGSetScale (Dlg, BaseID+51, Range[MAX_X]);
  DLGSetScale (Dlg, BaseID+47, Range[MIN_Y]);
  DLGSetScale (Dlg, BaseID+52, Range[MAX_Y]);
  DLGSetScale (Dlg, BaseID+48, Range[MIN_Z]);
  DLGSetScale (Dlg, BaseID+53, Range[MAX_Z]);
}


VIEW *ViewCB (DIALOG *Dlg, int BaseID, WORLD *World, int ViewNo, int BoxID)
{
  double  Clipping[4];
  double  Range[6];
  RPY     Camera;
  VIEW   *View;

  if (World == NULL || (View = World->GetView(ViewNo)) == NULL)
    return NULL;

  switch (BoxID-BaseID)
  {
    case 6 :
    case 7 : 
    case 8 :
    case 9 :   // Clipping-Range
      Clipping[0] = DLGGetScale (Dlg, BaseID+6) * 0.01;
      Clipping[1] = DLGGetScale (Dlg, BaseID+7) * 0.01;
      Clipping[2] = DLGGetScale (Dlg, BaseID+8) * 0.01;
      Clipping[3] = DLGGetScale (Dlg, BaseID+9) * 0.01;
      View->SetClipping (Clipping);
      break;

    case 16 :  // linewidth
      break;

    case 37 :  // projection-mode
      View->SetPrjMode ((VIEW::MODE)DLGGetOMenu(Dlg, BoxID));
      SetRange (Dlg, BaseID, View->GetRange());
      break;

    case 38 :
      // TODO: World->SetScale ();
      if (View->GetPrjMode() == VIEW::PERSPECT)
      {
	double Min, Max;

printf ("Autoadjust Ranges + Camera Position\n");
	World->GetSize (&Range[MIN_X], &Range[MAX_X],
			&Range[MIN_Y], &Range[MAX_Y],
			&Range[MIN_Z], &Range[MAX_Z]);
printf ("Range1:  X: %.2f..%.2f  Y:%.2f..%.2f  Z: %.2f..%.2f\n",
	Range[MIN_X], Range[MAX_X], 
	Range[MIN_Y], Range[MAX_Y],
	Range[MIN_Z], Range[MAX_Z]);

        Min = Range[MIN_X] - 0.75 * fabs (Range[MAX_X] - Range[MIN_X]);
	Max = Min          + 1.50 * fabs (Range[MAX_X] - Range[MIN_X]);
	Range[MIN_X] = Min;
	Range[MAX_X] = Max;

        Min = Range[MIN_Y] - 0.75 * fabs (Range[MAX_Y] - Range[MIN_Y]);
	Max = Min          + 1.50 * fabs (Range[MAX_Y] - Range[MIN_Y]);
	Range[MIN_Y] = Min;
	Range[MAX_Y] = Max;
  
	Min = Range[MIN_Z] -  0.5 * fabs (Range[MAX_Z] - Range[MIN_Z]);
	Max = Min          + 10.0 * fabs (Range[MAX_Z] - Range[MIN_Z]);

        // TODO: aktuelle Kamera-Richtung benutzen um die Kamera-Pos berechnen
        Camera.px = 0.0;
	Camera.py = 0.0;
	Camera.pz = - 1.5 * fabs (Range[MAX_Z] - Range[MIN_Z]);
	Camera.y = Camera.p = Camera.r = 0.0;

	Range[MIN_Z] = Min;
	Range[MAX_Z] = Max;
	View->SetRange (Range);

printf ("Range2:  X: %.2f..%.2f  Y:%.2f..%.2f  Z: %.2f..%.2f\n",
	Range[MIN_X], Range[MAX_X], 
	Range[MIN_Y], Range[MAX_Y],
	Range[MIN_Z], Range[MAX_Z]);
printf ("Camera:  Pos: %.2f/%.2f/%.2f  Rot: %.2f/%.2f%.2f\n",
	Camera.px, Camera.py, Camera.pz, Camera.y, Camera.p, Camera.r);

	View->SetCamera (Camera);
	SetRange (Dlg, BaseID, View->GetRange());

	SetView (Dlg, BaseID, World, ViewNo);
      }
      break;

    case 46 :  // range
    case 47 :
    case 48 :
    case 51 :
    case 52 :
    case 53 :	       
      Range[MIN_X] = DLGGetScale (Dlg, BaseID+46);
      Range[MAX_X] = DLGGetScale (Dlg, BaseID+51);
      Range[MIN_Y] = DLGGetScale (Dlg, BaseID+47);
      Range[MAX_Y] = DLGGetScale (Dlg, BaseID+52);
      Range[MIN_Z] = DLGGetScale (Dlg, BaseID+48);
      Range[MAX_Z] = DLGGetScale (Dlg, BaseID+53);
      View->SetRange (Range);
      break;

    case 57 :  // scale
      View->SetScale (DLGGetScale(Dlg, BaseID+57)*0.01);
      break;

    case 66 :  // camera
    case 67 :
    case 68 :
    case 71 :
    case 72 :
    case 73 :
      Camera.px = DLGGetDouble (Dlg, BaseID+66);
      Camera.py = DLGGetDouble (Dlg, BaseID+67);
      Camera.pz = DLGGetDouble (Dlg, BaseID+68);
      Camera.y  = DLGGetScale  (Dlg, BaseID+71);
      Camera.p  = DLGGetScale  (Dlg, BaseID+72);
      Camera.r  = DLGGetScale  (Dlg, BaseID+73);
      View->SetCamera (Camera);
      break;

    default : 
      View->SetBorder (ColorCB(Dlg, BaseID+17, View->GetBorder(), BoxID, 0));
  }  

  return View;
}


bool SetView (DIALOG *Dlg, int BaseID, WORLD *World, int ViewNo)
{
  double *Clipping;
  RPY     Camera;
  VIEW   *View;

  if (World == NULL || (View = World->GetView(ViewNo)) == NULL)
    return FALSE;

  Clipping = View->GetClipping ();
  DLGSetScale (Dlg, BaseID+6, Clipping[0]*100.0);
  DLGSetScale (Dlg, BaseID+7, Clipping[1]*100.0);
  DLGSetScale (Dlg, BaseID+8, Clipping[2]*100.0);
  DLGSetScale (Dlg, BaseID+9, Clipping[3]*100.0);

  SetColor (Dlg, BaseID+17, View->GetBorder(), BASE_MASK);

  DLGSetOMenu (Dlg, BaseID+37, int(View->GetPrjMode()));
  SetRange (Dlg, BaseID, View->GetRange());

  DLGSetScale (Dlg, BaseID+57, View->GetScale()*100.0);

  Camera = View->GetCamera()->GetLocation ();
  DLGSetDouble (Dlg, BaseID+66, Camera.px);
  DLGSetDouble (Dlg, BaseID+67, Camera.py);
  DLGSetDouble (Dlg, BaseID+68, Camera.pz);
  DLGSetScale  (Dlg, BaseID+71, Camera.y);
  DLGSetScale  (Dlg, BaseID+72, Camera.p);
  DLGSetScale  (Dlg, BaseID+73, Camera.r);

  return TRUE;
}


#define FACTOR 10.0

void ViewDialog (DIALOG *Dlg, int BaseID, WORLD *World, int ViewNo)
{
  double  Size;
  double *Clipping;
  double *Range;
  RPY     Camera;
  VIEW   *View;

  PRO_START ("")

  if (World == NULL)
    PRO_ABORT ("NULL-pointer")

  View = World->GetView (ViewNo);

  DLGMakeCols     (Dlg, BaseID,    2, DlgNO_LINE,   BaseID+1, BaseID+25);
  DLGMakeRows     (Dlg, BaseID+ 1, 2, DlgNO_LINE,   BaseID+5, BaseID+15);
  DLGMakeRows     (Dlg, BaseID+ 5, 4, DlgNO_LINE,   
		                      BaseID+6, BaseID+7, BaseID+8, BaseID+9);
  DLGMakeRows     (Dlg, BaseID+15, 2, DlgNO_LINE,   BaseID+16, BaseID+17);
  DLGMakeRows     (Dlg, BaseID+25, 2, DlgNO_LINE,   BaseID+35, BaseID+55);
  DLGMakeRows     (Dlg, BaseID+35, 2, DlgNO_LINE,   BaseID+36, BaseID+40);
  DLGMakeCols     (Dlg, BaseID+36, 2, DlgNO_LINE,   BaseID+37, BaseID+38);
  DLGMakeCols     (Dlg, BaseID+40, 2, DlgNO_LINE,   BaseID+45, BaseID+50);
  DLGMakeRows     (Dlg, BaseID+45, 3, DlgNO_LINE,  
		                      BaseID+46, BaseID+47, BaseID+48);
  DLGMakeRows     (Dlg, BaseID+50, 3, DlgNO_LINE, 
                                      BaseID+51, BaseID+52, BaseID+53);
  DLGMakeRows     (Dlg, BaseID+55, 2, DlgNO_LINE,   BaseID+56, BaseID+60);
  DLGMakeCols     (Dlg, BaseID+56, 1, DlgNO_LINE,   BaseID+57);
  DLGMakeCols     (Dlg, BaseID+60, 2, DlgETCHED_IN, BaseID+65, BaseID+70);
  DLGMakeRows     (Dlg, BaseID+65, 3, DlgNO_LINE,   
		                      BaseID+66, BaseID+67, BaseID+68);
  DLGMakeRows     (Dlg, BaseID+70, 3, DlgNO_LINE,   
		                      BaseID+71, BaseID+72, BaseID+73);

  DLGSetBorder   (Dlg, BaseID+5,  DlgETCHED_IN);
  DLGSetBorder   (Dlg, BaseID+15, DlgETCHED_IN);
  DLGSetBorder   (Dlg, BaseID+35, DlgETCHED_IN);
  DLGSetBorder   (Dlg, BaseID+55, DlgETCHED_IN);

  DLGSetTitle  (Dlg, BaseID+5,  "viewPort", DlgLEFT);
  DLGSetTitle  (Dlg, BaseID+15, "lines",    DlgLEFT);
  DLGSetTitle  (Dlg, BaseID+35, "range",    DlgLEFT);
  DLGSetTitle  (Dlg, BaseID+55, "camera",   DlgLEFT);

  Clipping = View->GetClipping ();
  DLGInsertScale (Dlg, BaseID+6, "minX", 0.0, 100.0, Clipping[0]*100.0, 
                                                     1, DlgHORIZONTAL);
  DLGInsertScale (Dlg, BaseID+7, "minY", 0.0, 100.0, Clipping[1]*100.0, 
		                                     1, DlgHORIZONTAL);
  DLGInsertScale (Dlg, BaseID+8, "maxX", 0.0, 100.0, Clipping[2]*100.0, 
		                                     1, DlgHORIZONTAL);
  DLGInsertScale (Dlg, BaseID+9, "maxY", 0.0, 100.0, Clipping[3]*100.0, 
		                                     1, DlgHORIZONTAL);

  DLGInsertScale  (Dlg, BaseID+16, "lineWidth", 0.0, 8.0, 1.0, 
		                                0, DlgHORIZONTAL);
  ColorDialog (Dlg, BaseID+17, View->GetBorder(), 0);

  DLGInsertOMenu  (Dlg, BaseID+37, "projection", 2, 
		                   "orthographic", "perspectivic", 
		                   int(View->GetPrjMode()));
  DLGInsertProc   (Dlg, BaseID+38, "auto", NULL);

  Range = View->GetRange ();
  Size = Range[MAX_X] - Range[MIN_X];
  DLGInsertScale  (Dlg, BaseID+46, "minX", Range[MIN_X] - FACTOR * Size,
		                           Range[MAX_X] + FACTOR * Size,
		                           Range[MIN_X], 1, DlgHORIZONTAL);
  DLGInsertScale  (Dlg, BaseID+51, "maxX", Range[MIN_X] - FACTOR * Size,
		                           Range[MAX_X] + FACTOR * Size,
		                           Range[MAX_X], 1, DlgHORIZONTAL);
  Size = Range[MAX_Y] - Range[MIN_Y];
  DLGInsertScale  (Dlg, BaseID+47, "minY", Range[MIN_Y] - FACTOR * Size,
		                           Range[MAX_Y] + FACTOR * Size,
		                           Range[MIN_Y], 1, DlgHORIZONTAL);
  DLGInsertScale  (Dlg, BaseID+52, "maxY", Range[MIN_Y] - FACTOR * Size,
		                           Range[MAX_Y] + FACTOR * Size,
		                           Range[MAX_Y], 1, DlgHORIZONTAL);
  Size = Range[MAX_Z] - Range[MIN_Z];
  DLGInsertScale  (Dlg, BaseID+48, "minZ", Range[MIN_Z] - FACTOR * Size,
		                           Range[MAX_Z] + FACTOR * Size,
		                           Range[MIN_Z], 1, DlgHORIZONTAL);
  DLGInsertScale  (Dlg, BaseID+53, "maxZ", Range[MIN_Z] - FACTOR * Size,
		                           Range[MAX_Z] + FACTOR * Size,
		                           Range[MAX_Z], 1, DlgHORIZONTAL);

  DLGInsertScale  (Dlg, BaseID+57, "scale", 0.0, 1000.0, 
		                            View->GetScale()*100.0, 
		                            2, DlgHORIZONTAL);

  Camera = View->GetCamera()->GetLocation ();
  DLGInsertDouble (Dlg, BaseID+66, "posX", Camera.px, 9, 2);
  DLGInsertDouble (Dlg, BaseID+67, "posY", Camera.py, 9, 2);
  DLGInsertDouble (Dlg, BaseID+68, "posZ", Camera.pz, 9, 2);

  DLGInsertScale  (Dlg, BaseID+71, "rotX", 0.0, 360.0, Camera.y, 
                                           2, DlgHORIZONTAL);
  DLGInsertScale  (Dlg, BaseID+72, "rotY", 0.0, 360.0, Camera.p, 
                                           2, DlgHORIZONTAL);
  DLGInsertScale  (Dlg, BaseID+73, "rotZ", 0.0, 360.0, Camera.r, 
		                           2, DlgHORIZONTAL);

  PRO_END
}
