#ifndef LIBVIEW_H
#define LIBVIEW_H

#ifdef __cplusplus
extern "C" {
#endif

#include <libgeo.h>
#include <libnet.h>


/* define some colors: */
#define VIEW_RED          2
#define VIEW_ORANGE       3
#define VIEW_YELLOW       4
#define VIEW_GREEN        5
#define VIEW_BLUE         6
#define VIEW_MAGENTA      7
#define VIEW_BROWN        8
#define VIEW_WHITE        9
#define VIEW_GREY        10
#define VIEW_BLACK       11
#define VIEW_TRANSPARENT 12
#define VIEW_WIREFRAME   13

#define PART_NAME_LEN  255
#define DEV_NAME_LEN   255
#define FEAT_NAME_LEN   20
#define OBJ_NAME_LEN   DEV_NAME_LEN + PART_NAME_LEN
#define ASM_NAME_LEN   255

#define CFG_HIGHLAP_PATH      8         /* Where to find needed data      */
#define CFG_TMP_FILE         11         /* Temporary data-file            */
#define CFG_CS_FILE          12         /* FileName of used Coord.-System */


/* System-Functions: */
#define VIEWClearAll(SocketID)    VIEWRunCLI (SocketID, "CLEAR ALL")
#define VIEWExit(SocketID)        VIEWRunCLI (SocketID, "EXIT")
#define VIEWFill(SocketID)        VIEWRunCLI (SocketID, "FILL")
#define VIEWMenu(SocketID)        VIEWRunCLI (SocketID, "TRANSFER TO MENU")
#define VIEWUpdate(SocketID)      VIEWRunCLI (SocketID, "UPDATE GRAPHICS")

/* Button-Functions: */
#define VIEWFly(SocketID) \
          VIEWRunCLI (SocketID, "INVOKE MENU BUTTON 'world.fly'")
#define VIEWRotate(SocketID) \
          VIEWRunCLI (SocketID, "INVOKE MENU BUTTON 'world.rotate'")
#define VIEWCruise(SocketID) \
          VIEWRunCLI (SocketID, "INVOKE MENU BUTTON 'world.cruise'")
#define VIEWTurnDev(SocketID) \
          VIEWRunCLI (SocketID, "INVOKE MENU BUTTON 'trndev'")
#define VIEWRotDev(SocketID) \
          VIEWRunCLI (SocketID, "INVOKE MENU BUTTON 'rotdev'")
#define VIEWInvokeButton(SocketID,Button) \
          VIEWRunCLI (SocketID, "INVOKE MENU BUTTON '%s'", Button)


/* Preferences: */
#define VIEWMultiView(SocketID,Nr) \
          VIEWRunCLI (SocketID, "MULTI VIEW %d", Nr)
#define VIEWMsgMode(SocketID,Mode) \
          VIEWRunCLI (SocketID, "MESSAGE MODE %s", Mode)
#define VIEWSetUnit(SocketID,Unit) \
          VIEWRunCLI (SocketID, "SET UNITS TO %s", Unit)
#define VIEWRefreshMode(SocketID,Mode) \
          VIEWRunCLI (SocketID, "SET GRAPHICS MODE %s", Mode?"ON":"OFF")
#define VIEWSetGrid(SocketID,x,y) \
          VIEWRunCLI (SocketID, "SET GRID TO %d, %d", x, y)
#define VIEWSetSpeed(SocketID,Speed) \
          VIEWRunCLI (SocketID, "SET SPEED TO %d", Speed)


/* World-Functions: */
#define VIEWSaveWorld(SocketID,FileName) \
          VIEWRunCLI (SocketID, "SAVE WORKCELL AS '%s'", FileName)
#define VIEWLoadWorld(SocketID,FileName) \
          VIEWRunCLI (SocketID, "RETRIEVE WORKCELL '%s'", FileName)
#define VIEWPushView(SocketID) \
        { \
	  VIEWRunCLI (SocketID, "PUSH WORLD"); \
	  VIEWRunCLI (SocketID, "SAVE ALL POSITIONS"); \
	}
#define VIEWPopView(SocketID) \
        { \
          VIEWRunCLI (SocketID, "RESTORE ALL POSITIONS"); \
	  VIEWRunCLI (SocketID, "POP WORLD"); \
      	}
/* Device-Functions: */
#define VIEWSaveDev(SocketID,DevName,FileName) \
          VIEWRunCLI (SocketID, "SAVE '%s' AS '%s'", DevName, FileName)
#define VIEWLoadDev(SocketID,FileName) \
          VIEWRunCLI (SocketID, "RETRIEVE DEVICE '%s'", FileName)
#define VIEWRenameCD(SocketID,DevName) \
          VIEWRunCLI (SocketID, "RENAME CD TO '%s'", DevName)
#define VIEWDeleteDev(SocketID,DevName) \
          VIEWRunCLI (SocketID, "DELETE '%s'", DevName)
#define VIEWRenameDev(SocketID,DevName1,DevName2) \
          VIEWRunCLI (SocketID, "RENAME '%s' TO '%s'", DevName1, DevName2)
#define VIEWSelectDev(SocketID,Device) \
          VIEWRunCLI (SocketID, "SELECT DEVICE '%s'", Device)
#define VIEWSetDevColor(SocketID,DevName,Color) \
          VIEWRunCLI (SocketID, "SET '%s' COLOR TO $%s", \
		                DevName, IGRIPColors[Color])
#define VIEWHiLightDev(SocketID,DevName,Mode) \
          VIEWRunCLI (SocketID, "SET '%s' HILIGHT %s", \
		                DevName, Mode?"ON":"OFF")
#define VIEWAttachDev(SocketID,Name,DevName) \
          VIEWRunCLI (SocketID, "ATTACH '%s' AT '%s'", Name, DevName)

/* Part-Functions: */
#define VIEWLoadPart(SocketID,FileName) \
          VIEWRunCLI (SocketID, "RETRIEVE PART '%s'", FileName)
#define VIEWOpenPart(SocketID,PartName) \
          VIEWRunCLI (SocketID, "OPEN PART '%s'", PartName)
#define VIEWSelectPart(SocketID,Text) \
          VIEWRunCLI (SocketID, "SELECT PART BY MOUSE " \
		                "WITH NOCONFIRM PROMPT '%s'", Text)
#define VIEWSetPartColor(SocketID,PartName,Color) \
          VIEWRunCLI (SocketID, "SET PART '%s' COLOR TO $%s", \
		                PartName, IGRIPColors[Color])
#define VIEWHiLightPart(SocketID,DevName,PartName,Mode) \
          VIEWRunCLI (SocketID, "SET PART '%s:%s' HILIGHT %s", \
		                DevName, PartName, Mode?"ON":"OFF")
#define VIEWTransPart(SocketID,DevName,PartName,x,y,z) \
          VIEWRunCLI (SocketID, "TRANSLATE PART '%s:%s' TO %.3f, %.3f, %.3f", \
		                DevName, PartName, x, y, z)
#define VIEWAttach(SocketID,Name,DevName,PartName) \
          VIEWRunCLI (SocketID, "ATTACH '%s' AT PART '%s:%s'", \
		                Name, DevName, PartName)


/* Create-Functions: */
#define VIEWOpenPolygon(SocketID)   VIEWRunCLI (SocketID, "OPEN POLYGON")
#define VIEWClosePolygon(SocketID)  VIEWRunCLI (SocketID, "CLOSE POLYGON")
#define VIEWMakeSphere(SocketID,r,n) \
          VIEWRunCLI (SocketID, "CREATE SPHERE %.3f, 360.0, %d", r, n);


/* Path-Functions: */ 
#define VIEWCreatePath(SocketID,PathName,DevName) \
          VIEWRunCLI (SocketID, "CREATE PATH '%s' ON '%s'", PathName, DevName)
#define VIEWCreateTag(SocketID,TagName,PathName) \
            VIEWRunCLI (SocketID, "CREATE TAG '%s' ON PATH '%s'", \
		                  TagName, PathName)
#define VIEWMoveAlong(SocketID,PathName) \
          VIEWRunCLI (SocketID, "MOVE ALONG PATH '%s'", PathName)
#define VIEWDeleteTag(SocketID,TagName) \
          VIEWRunCLI (SocketID, "DELETE TAG '%s'", TagName)
#define VIEWDeletePath(SocketID,PathName) \
          VIEWRunCLI (SocketID, "DELETE PATH '%s'", PathName)

extern char *IGRIPColors[];
extern char *MsgModeText[];
extern char *UnitsText[];


char  *VIEWRunCLI            (int SocketID, char *Command, ...);
dual   VIEWOpenCLIProtocol   (char *FileName);
dual   VIEWCloseCLIProtocol  (void);

int    VIEWMouseSelect       (int SocketID, 
			     char *DevName, char *PartName, char *FileName,
			     double *x, double *y, double *z);
dual   VIEWGetPos            (int SocketID, char *DevName, char *PartName, 
			     char *Reference, H_MATRIX *M);
dual   VIEWInquirePos        (int SocketID, char *DevName, char *PartName, 
			                    H_MATRIX *M);
dual   VIEWGetWorkcell       (int SocketID, char *FileName);
void   VIEWShowFrame         (int SocketID, char *Name, H_MATRIX M);
void   VIEWStartMsg          (int SocketID, char *Msg, ...);
dual   VIEWDrawPolygon       (int SocketID, char *DevName,
			                    POLYGON *Polygon, H_MATRIX WTP, 
			                    int No, int Color);
dual   VIEWDrawSphere        (int SocketID, char *DevName, char *PartName,
			      POINT3D Pos, double r, int n, int Color);

#ifdef __cplusplus
}
#endif

#endif
