#include "viewGlobal.h"


char *IGRIPColors[] = {"", "", 
		       "RED", "ORANGE", "YELLOW", "GREEN", 
		       "BLUE", "MAGENTA", "BROWN", 
		       "WHITE", "GREY", "BLACK", 
		       "", "TRANSPARENT", "WIREFRAME"};




/* ********************************************************************** */
/* ** Draws a Polygon in a given Color.                                ** */
/* ** INPUT:  SocketID  Connection to IGRIP                            ** */
/* **         Name      Name of the device, where to store the polygon ** */
/* **                   in.                                            ** */
/* **         Polgon    Pointer to the first polygon to display        ** */
/* **         WTP       Transformationmatrix for the polygon-points    ** */
/* **                   used for actual position of the parts          ** */
/* **         Color     Index into color-table                         ** */
/* ** OUTPUT: (Bool)    TRUE on success, FALSE otherwise               ** */
/* ********************************************************************** */
dual VIEWDrawPolygon (int SocketID, char *DevName, POLYGON *Polygon, 
		                    H_MATRIX WTP, int No, int Color)
{
  int      count;
  char    *Response = NULL;
  char     PartName[256];
  POINT3D  Point;
  POLYGON *Ptr;

  if ((Ptr = Polygon) == NULL)
    return FALSE;

  Response = VIEWRunCLI(SocketID, "SELECT DEVICE '%s'", DevName);
  if (Response == NULL)
  {
    VIEWRunCLI (SocketID, "RETRIEVE DEVICE '%s/DEVICES/empty.dev'",
	                  TOOLGetConfigString(CFG_HIGHLAP_PATH));
    VIEWRenameCD (SocketID, DevName);
  }

  if (No < 0)
    sprintf (PartName, "%s:polygons", DevName);
  else
    sprintf (PartName, "%s:polygon%d", DevName, No);

  VIEWOpenPart (SocketID, PartName);

  while (Ptr != NULL)
  {
    VIEWOpenPolygon (SocketID);

    for (count = 0; count < Ptr->Number; ++count)
    {
      Point = GEOMatrixMult (WTP, Ptr->Points[Ptr->Vertex[count]]);
      VIEWRunCLI (SocketID, "ADD TO POLY %.3f, %.3f, %.3f",
		            Point.x, Point.y, Point.z);
    }
    VIEWClosePolygon (SocketID);

    Ptr = Ptr->Next;
  }

  VIEWSetPartColor (SocketID, PartName, Color);

  return TRUE;
}


dual VIEWDrawSphere (int SocketID, char *DevName, char *PartName,
		                   POINT3D Pos, double r, int n, int Color)
{
  VIEWRunCLI       (SocketID, "RETRIEVE DEVICE '%s/DEVICES/empty.dev'",
	                      TOOLGetConfigString(CFG_HIGHLAP_PATH));
  VIEWRenameCD     (SocketID, DevName);
  VIEWOpenPart     (SocketID, PartName);
  VIEWTransPart    (SocketID, DevName, PartName, Pos.x, Pos.y, Pos.z);
  VIEWMakeSphere   (SocketID, r, n);
  VIEWSetPartColor (SocketID, PartName, Color);

  return TRUE;
}



