#include <stdarg.h>
#include <string.h>

#include "viewGlobal.h"


static char  Buffer[SOCKET_BUF_SIZE+1] = "";
static FILE *CLIFile = NULL;

char *MsgModeText[] = {"NOMSG", "USERMSG", "ALLMSG"};
char *UnitsText[]   = {"MM", "CM", "METERS", "INCHES", "FEET"};


/* ********************************************************************** */
/* ** Send a CLI-command to IGRIP. This function buffers in- and       ** */
/* ** output itself. So the Command-parameter may be a normal constant ** */
/* ** string. There's also the possibility to specify a variable       ** */
/* ** argument lists                                                   ** */
/* ********************************************************************** */
char *VIEWRunCLI (int SocketID, char *Command, ...)
{
  va_list Arguments;

  PRO_START ("\n");

  va_start (Arguments, Command);
  vsprintf (Buffer, Command, Arguments);
  va_end   (Arguments);

  if (CLIFile != NULL)
    fprintf (CLIFile, "%s\n", Buffer);

#ifndef linux
  PRO1 ("Command: %s\n", Buffer)
#endif

  if (! NETSendCommand(SocketID, Buffer))
    ERROR_RETURN0 ("Connection failed/closed", NULL)

  if (CLIFile != NULL)
    fprintf (CLIFile, " -> %s\n", Buffer);

  if (strncmp(Buffer, "0:", 2) != 0 && strncmp(Buffer, "180:", 4) != 0)
  {
    fprintf (stderr, "VIEW-ERROR: '%s'\n", Buffer);
    fflush (stderr);
    ERROR_RETURN0 (Buffer, NULL)
  }

  PRO_END
  return Buffer;
}



/* ********************************************************************** */
/* ** Opens a protocol-file for storing cli-commands and their         ** */
/* ** response.                                                        ** */
/* ** INPUT:  FileName  Name of the protocol-file                      ** */
/* ** OUTPUT: (dual)    TRUE,  if opening was successfully or a file   ** */
/* **                          was already opened                      ** */
/* **                   FALSE, if FileName is bad or an error occured  ** */
/* ********************************************************************** */
dual VIEWOpenCLIProtocol (char *FileName)
{
  if (FileName == NULL)
    return FALSE;
  elsif (CLIFile != NULL)
    return TRUE;
  else
    return ((CLIFile = fopen(FileName, "w+")) != NULL);
}
  

/* ********************************************************************** */
/* ** Closes an opened cli-protocol-file.                              ** */
/* ** INPUT:   none                                                    ** */
/* ** OUTPUT:  (dual)  TRUE, if file was closed successfully           ** */
/* **                  FALSE otherwise                                 ** */
/* ********************************************************************** */
dual VIEWCloseCLIProtocol (void)
{
  if (CLIFile == NULL)
    return FALSE;

  fclose (CLIFile);
  CLIFile = NULL;

  return TRUE;
}


/* ********************************************************************** */
/* ** extract the devicename out of a complete part-specification in   ** */
/* ** the form 'devname:partname'.                                     ** */
/* ** INPUT:  Name     Complete part-name                              ** */
/* ** INOUT:  DevName  Pointer to a allocated memory-area for the      ** */
/* **                  result                                          ** */
/* ** OUTPUT: (char*)  Pointer to DevName on success or NULL, if error ** */
/* ********************************************************************** */
char *VIEW_ExtractDevName (char *Name, char *DevName)
{
  int count1 = 0, count2 = 0;
  int len;

  if (Name == NULL)
    return NULL;

  len = strlen (Name);
  while (count1 < len && Name[count1] == ' ')
    count1++;

  while (count1 < len && Name[count1] != ':')
    DevName[count2++] = Name[count1++];
  DevName[count2] = '\0';

  return DevName;
}


/* ********************************************************************** */
/* ** Counterpart to VIEW_ExtractDevName. Gets the name of the part    ** */
/* ** from the part-spec. 'devname:partname' -> partname               ** */
/* ********************************************************************** */
char *VIEW_ExtractPartName (char *Name, char *PartName)
{
  int count1 = 0, count2 = 0;
  int len;

  if (Name == NULL)
    return NULL;

  len = strlen (Name);
  count1 = len -1;
  while (count1 > 0 && Name[count1] != ':')
    count1--;

  if (Name[count1] == ':')
    count1++;

  while (count1 < len)
    PartName[count2++] = Name[count1++];
  PartName[count2] = '\0';

  return PartName;
}


/* *********************************************************************** */
/* ** Error-checking routing reading a integer value at a current       ** */
/* ** file-position                                                     ** */
/* *********************************************************************** */
dual VIEW_ReadInt (FILE *fp, int *Value)
{
  dual Result;

  if (feof(fp))
  {
    PRO0 ("Unexpexted end of gsl-data-file\n")
    return FALSE;
  }

  if ((Result = (fscanf(fp, "%d\n", Value) == 1)))
    PRO1 ("Int: %d\n", *Value)
  else
    PRO0 ("Error reading int\n")

  return Result;
}


/* ********************************************************************** */
/* ** Reads a string (until end of line) from a file at the specified  ** */
/* ** position                                                         ** */
/* ********************************************************************** */
dual VIEW_ReadString (FILE *fp, char *String)
{
  dual Result;

  if (feof(fp))
  {
    PRO0 ("Unexpected end of gsl-data-file\n")
    return FALSE;
  }

  if ((Result = (fscanf(fp, "%s\n", String) == 1)))
    PRO1 ("String: %s\n", String)
  else
    PRO0 ("Error reading string\n")

  return Result;
}


/* ********************************************************************** */
/* ** Read the 3d-coordinates from a file. The coordinates must be     ** */
/* ** separated by space-characters (written by GSL-macro)             ** */
/* ********************************************************************** */
dual VIEW_ReadPos (FILE *fp, double *x, double *y, double *z)
{
  dual Result;

  if (feof(fp))
  {
    PRO0 ("Unexpected end of gsl-data-file\n")
    return FALSE;
  }

  if ((Result = (fscanf(fp, "%lf  %lf  %lf\n", x, y, z) == 3)))
    PRO3 ("Pos: [%.3f][%.3f][%.3f]\n", *x, *y, *z)
  else
    PRO0 ("Error reading Position\n")

  return Result;
}


/* ********************************************************************** */
/* ** Get the six coordinates of a frame in rpy-notation. The order    ** */
/* ** is (px py pz r p y).                                             ** */
/* ********************************************************************** */
dual VIEW_ReadRPY (FILE *fp, double *px, double *py, double *pz,
	                     double *r,  double *p,  double *y)
{
  dual Result;

  if (feof(fp))
  {
    PRO0 ("Unexpected end of gsl-data-file\n")
    return FALSE;
  }

  if ((Result = (fscanf(fp, "%lf  %lf  %lf  %lf  %lf  %lf\n", 
			    px, py, pz, r, p, y) == 6)))
  {
    PRO3 ("RPY: %.3f %.3f %.3f  ", *px, *py, *pz)
    PRO3 ("\\%.2f %.2f %.2f\n",    *r,  *p,  *y)
  }
  else
    PRO0 ("Error reading RPY\n")

  return Result;
}


/* ********************************************************************** */
/* ** Shows a coordinate-system at the position, which is given by the ** */
/* ** Matrix M.                                                        ** */
/* ** INPUT:  SocketID  Connection to the IGRIP-output-window          ** */
/* **         Name      Name of the coordinate-system's device         ** */
/* **         M         target-position and orientation of the cs      ** */
/* ** OUTPUT: none                                                     ** */
/* ********************************************************************** */
void VIEWShowFrame (int SocketID, char *Name, H_MATRIX M)
{
  RPY F;

  GEOFrame2RPY (M, &F.px, &F.py, &F.pz, &F.y, &F.p, &F.r);
  
  VIEWRefreshMode (SocketID, FALSE);
  VIEWRunCLI (SocketID, "RETRIEVE DEVICE '%s/DEVICES/%s'",
	                TOOLGetConfigString(CFG_HIGHLAP_PATH),
	                TOOLGetConfigString(CFG_CS_FILE));
  VIEWRenameCD (SocketID, Name);
  VIEWRunCLI (SocketID, 
	      "PLACE '%s' AT COORD %.2f, %.2f, %.2f, %.2f, %.2f, %.2f",
	      Name, F.px, F.py, F.pz, F.y, F.p, F.r);
  VIEWRefreshMode (SocketID, TRUE);
  VIEWUpdate      (SocketID);
}


/* ********************************************************************** */
/* ** This function clears the message-window and write a new string   ** */
/* ** into it.                                                         ** */
/* ** INPUT:  SocketID  Connection to the specific IGRIP-window        ** */
/* **         Msg       Message to display.                            ** */
/* **         ...       variable parameter-list of the message         ** */
/* ** OUTPUT: none                                                     ** */
/* ********************************************************************** */
void VIEWStartMsg (int SocketID, char *Msg, ...)
{
  char    Message[SOCKET_BUF_SIZE+1];
  va_list Arguments;

  va_start (Arguments, Msg);
  vsprintf (Message, Msg, Arguments);
  va_end   (Arguments);

  VIEWRunCLI (SocketID, "STARTMSG '%s'", Message);
}
